//////////////////////////////////////////////////////////////////////////////
//This software module was originally developed by CSELT in the course of   //
//development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.             //
//This software module is an implementation of a part of one or more MPEG-4 //
//Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems        //
//(ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 Systems     //
//(ISO/IEC 14496-1) free license to this software module or modifications   //
//thereof for use in hardware or software products claiming conformance to  //
//the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use this software//
//module in hardware or software products are advised that its use may      //
//infringe existing patents.                                                //
//The original developer of this software module and his/her company, the   //
//subsequent editors and their companies, and ISO/IEC have no liability     //
//for use of this software module or modifications thereof in an            //
//implementation. Copyright is not released for non MPEG-4 Systems          //
//(ISO/IEC 14496-1) conforming products. CSELT S.pA. retains full right     //
//to use the code for his/her own purpose, assign or donate the code to a   //
//third party and to inhibit third parties from using the code for non      //
//MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright      //
//notice must be included in all copies or derivative works.                //
//Copyright  1999.                                                         //
//This work has been partly developed within the framework of the ESPRIT    //
//European Research Project  "MPEG-4 PC".                                   //
//////////////////////////////////////////////////////////////////////////////
//
// File: AR-DS-StreamInfo.cpp
//
// Author: Matthew Leditschke. 
//
// Modified: Gianluca Di Cagno  CSELT gianluca.dicagno@cselt.it, 
//			 Ananda Allys France Telecom
//
// Description:
//   The implementation of a class which encapsulates all of the details
//   of an audio stream that is being played by the DirectSound based
//   audio renderer.
//
// Created: August, 1997.
//
// 
///////////////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "AR-DS-StreamInfo.h"
#include "dx-errors.h"
#include "trace.h"
#include "AudioSourceProxy.h"
#include "Stream.h"

// Define this to have a copy of all data sent to the sound card
// also written to disk.
//#define ALSO_WRITE_TO_DISK

//JLF 11/25/01: Fix for DS versions
#ifndef DSBCAPS_CTRLDEFAULT
#define DSBCAPS_CTRLDEFAULT		DSBCAPS_CTRLPAN | DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY
#endif

// ===============================================================================

//
// StreamInfo::StreamInfo
//
// Constructor, set default values.
//

StreamInfo::StreamInfo(LPDIRECTSOUND pDS, AudioSourceProxy* pSrc,
					   DWORD startTime)
{
	m_pDS = pDS;
	m_pDSBuf = NULL;
	m_startTime = startTime;
	m_pSrc = pSrc;
	m_nextWrite = 0;
	m_bPlaying = FALSE;
	m_bReallyPlaying = FALSE;
	m_dwInitialCTS = 0;

	// This signifies that the audio format is not known
	m_wf.nChannels = 0;

	m_rawOutput = NULL;
#ifdef ALSO_WRITE_TO_DISK
	m_rawOutput = fopen("samplesPlayed.raw","wb");
	assert(m_rawOutput != NULL);
#endif
} /* StreamInfo::StreamInfo */


// ===============================================================================

//
// StreamInfo::~StreamInfo
//
// Destructor, shut down the stream.
//

StreamInfo::~StreamInfo()
{
	if (m_pDSBuf != NULL)
	{
		m_pDSBuf->Stop();
		m_pDSBuf->Release();
	}

	if (m_rawOutput != NULL)
		fclose(m_rawOutput);
} /* StreamInfo::~StreamInfo */


// ===============================================================================


//
// StreamInfo::Create
//
// Create a streaming buffer.  Return DS_OK if created OK.
//

int StreamInfo::Create(DWORD nSamplesPerSec, DWORD nBitsPerSample,
					   WORD nNumChannels)
{
	// Specify the wave format
	memset(&m_wf, 0, sizeof(WAVEFORMATEX));
	m_wf.wFormatTag      = WAVE_FORMAT_PCM; 
	m_wf.nChannels       = nNumChannels;
	m_wf.nSamplesPerSec  = nSamplesPerSec;
	m_wf.wBitsPerSample  = nBitsPerSample;
	m_wf.nBlockAlign     = m_wf.nChannels * m_wf.wBitsPerSample / 8;
	m_wf.nAvgBytesPerSec = nSamplesPerSec * m_wf.nBlockAlign; 
	m_wf.cbSize          = 0;

	// Set up DSBUFFERDESC structure.
	memset(&m_bufDesc, 0, sizeof(DSBUFFERDESC));
	m_bufDesc.dwSize = sizeof(DSBUFFERDESC);
	m_bufDesc.dwFlags = DSBCAPS_CTRLDEFAULT | DSBCAPS_STICKYFOCUS; 
	m_bufDesc.dwBufferBytes = m_wf.nAvgBytesPerSec * BUFFER_LENGTH / 1000;
	m_bufDesc.dwBufferBytes = m_bufDesc.dwBufferBytes / m_wf.nBlockAlign * m_wf.nBlockAlign;
	m_bufDesc.lpwfxFormat = &m_wf;
	
	// Create the DirectSound buffer
	HRESULT res = m_pDS->CreateSoundBuffer(&m_bufDesc, &m_pDSBuf, NULL);

	// store the Threshold for the clock divergence
	m_dwThreshold = BUFFER_UPDATE_TIME;

	return res;
} /* StreamInfo::Create */


// ===============================================================================

//
// StreamInfo::GetMaxWriteSize
//
// Calculate the maximum size of sound buffer data which can be written
// to the given stream.  i.e. how much free space there is in buffer.
//

DWORD StreamInfo::GetMaxWriteSize()
{
    DWORD dwWriteCursor=0, dwPlayCursor=0, dwMaxSize=0;

    // Get current play position
    if (m_pDSBuf->GetCurrentPosition(&dwPlayCursor, &dwWriteCursor) == DS_OK)
    {
		// Handle the special case at the start, when the buffer is
		// completely empty, and the stream is not playing
        if ((m_nextWrite == dwPlayCursor) && (!IsPlaying()))
			dwMaxSize = m_bufDesc.dwBufferBytes;   //was 480*5
 
		else if (m_nextWrite <= dwPlayCursor)
		{
			// Our write position trails play cursor
			dwMaxSize = dwPlayCursor - m_nextWrite;
		}

		else // (m_nextWrite > dwPlayCursor)
		{
			// Play cursor has wrapped
			dwMaxSize = m_bufDesc.dwBufferBytes - m_nextWrite + dwPlayCursor;
		}

		TRACE_PRINT4("nextWrite = %d, playCursor = %d, maxSize = %d NOW = %d\n",
					 m_nextWrite, dwPlayCursor, dwMaxSize, GetTickCount());
		
    }
    else
    {
        // GetCurrentPosition call failed
        assert(0);
        dwMaxSize = 0;
    }

    return (dwMaxSize);
} /* StreamInfo::GetMaxWriteSize */

//
// StreamInfo::ResetWritePointer
//
// Set the write pointer to the first unread byte.
//

void StreamInfo::ResetWritePointer()
{
    DWORD dwWriteCursor=0, dwPlayCursor=0;

    // Get current play position
    if (m_pDSBuf->GetCurrentPosition(&dwPlayCursor, &dwWriteCursor) == DS_OK)
    {
        m_nextWrite = dwPlayCursor;

		TRACE_PRINT3("nextWrite = %d, playCursor = %d NOW = %d\n",
					 m_nextWrite, dwPlayCursor, GetTickCount());
    }
} /* StreamInfo::ResetWritePointer */


// ===============================================================================

//
// StreamInfo::IsReallyPlaying
//
// DS apparently has a difficult startup: reads a few bytes and stops for a while !!!
//

BOOL StreamInfo::IsReallyPlaying()
{
    if (!m_bReallyPlaying)
	{    // Get current play position
		DWORD dwWriteCursor=0, dwPlayCursor=0;
		m_pDSBuf->GetCurrentPosition(&dwPlayCursor, &dwWriteCursor);
 
		// DS apparently has a difficult startup: 
		// reads e few bytes and stops for a while !!!

		m_bReallyPlaying = (dwPlayCursor > GetBytesPerMs()*10);
// Guido	
	    if (m_bReallyPlaying)
		{
			m_dwInitialCTS += dwPlayCursor/GetBytesPerMs();
			m_pSrc -> GetStream() -> AdjustClock (m_dwInitialCTS);
			TRACE_PRINT1("AdjustClock %d \n",m_dwInitialCTS);
	//		m_dwInitialCTS = 0;
		}
	}
	return (m_bReallyPlaying);
} /* StreamInfo::IsReallyPlaying */

// ===============================================================================

//
// StreamInfo::WriteData
//
// Write the given data to the stream buffer.
//


void StreamInfo::WriteData(LPBYTE pData, DWORD nNumBytes, float fIntensity)
//void StreamInfo::WriteData(LPBYTE pData, DWORD nNumBytes)
{
    HRESULT hr;
    LPBYTE lpBuf1 = NULL;
    LPBYTE lpBuf2 = NULL;
    DWORD dwSize1 = 0;
    DWORD dwSize2 = 0;

#ifdef ALSO_WRITE_TO_DISK
	fwrite(pData, nNumBytes, 1, m_rawOutput);
#endif

    
    // fIntensity range 0, 1 ; mapped to -10 000, 0
    hr = m_pDSBuf->SetVolume((long)((fIntensity-1)*10000));

    // Lock the sound buffer
    hr = m_pDSBuf->Lock(m_nextWrite, nNumBytes,(void**) &lpBuf1, &dwSize1,
					(void**)&lpBuf2, &dwSize2, 0);
    assert(hr == DS_OK);

    // Write data to sound buffer. Because the sound buffer is
	// circular, we may have to do two write operations if the locked
	// portion of buffer wraps around to start of buffer.
    assert(lpBuf1 != NULL);
	memcpy(lpBuf1, pData, dwSize1);

	// Second write required?
    if (lpBuf2 != NULL)
    {
		pData += dwSize1;
		memcpy(lpBuf2, pData, dwSize2);
    }

    // Update the buffer offset and unlock sound buffer
    m_nextWrite = (m_nextWrite + nNumBytes) % m_bufDesc.dwBufferBytes;
    m_pDSBuf->Unlock(lpBuf1, dwSize1, lpBuf2, dwSize2);

} /* StreamInfo::WriteData */


// ===============================================================================

//
// StreamInfo::WriteSilence
//
// Write the given number of bytes worth of silence into the buffer.
// Or, simply clean the buffer to avoid audio artifacts, but do not move
// the pointer to give a chance for late CUs to still enter the buffer at
// the correct offset
//

void StreamInfo::WriteSilence(DWORD nNumBytes, BOOL updateOffset)
{
    HRESULT hr;
    LPBYTE lpBuf1 = NULL;
    LPBYTE lpBuf2 = NULL;
    DWORD dwSize1 = 0;
    DWORD dwSize2 = 0;

#ifdef ALSO_WRITE_TO_DISK
	for (DWORD i = 0; i < nNumBytes; i++)
		fputc(0, m_rawOutput);
#endif

    // Lock the sound buffer
    hr = m_pDSBuf->Lock(m_nextWrite, nNumBytes,(void**) &lpBuf1, &dwSize1,
					(void**)	&lpBuf2, &dwSize2, 0);
    assert(hr == DS_OK);

    // Write data to sound buffer. Because the sound buffer is
	// circular, we may have to do two write operations if the locked
	// portion of buffer wraps around to start of buffer.
    assert(lpBuf1 != NULL);
	if (m_wf.wBitsPerSample == 8)
		memset(lpBuf1, 0, dwSize1);
	else
	{
		assert((dwSize1 % sizeof(signed short)) == 0);
		signed short* data = (signed short*) lpBuf1;
		for (unsigned int i = 0; i < dwSize1 / sizeof(signed short); i++)
			data[i] = 0;
	}

	// Second write required?
    if (lpBuf2 != NULL)
	{
		if (m_wf.wBitsPerSample == 8)
			memset(lpBuf2, 0, dwSize2);
		else
		{
			assert((dwSize2 % sizeof(signed short)) == 0);
			signed short* data = (signed short*) lpBuf2;
			for (unsigned int i = 0; i < dwSize2 / sizeof(signed short); i++)
				data[i] = 0;
		}
	}

    // Update the buffer offset and unlock sound buffer
	if (updateOffset)
	    m_nextWrite = (m_nextWrite + nNumBytes) % m_bufDesc.dwBufferBytes;
    m_pDSBuf->Unlock(lpBuf1, dwSize1, lpBuf2, dwSize2);
} /* StreamInfo::WriteSilence */


// ===============================================================================

//
// StreamInfo::Play
//
// Start the stream playing.
//

void StreamInfo::Play()
{
	HRESULT hr = m_pDSBuf->Play(0, 0, DSBPLAY_LOOPING);
	assert(hr == DS_OK);
	if (hr == DS_OK)
	{	m_bPlaying = TRUE;
		SetLastCorrection(m_pSrc->GetStream()->GetClock());
		while (!IsReallyPlaying())
		{
			TRACE_PRINT("AR: PLAY sleeping\n");
			Sleep(10);
		}
	}
} /* StreamInfo::Play */

void StreamInfo::Stop()
{
	HRESULT hr = m_pDSBuf->Stop();
}
void StreamInfo :: SetInitialCTS (DWORD dwCTS)
{
	if (!m_bPlaying && m_dwInitialCTS == 0)
		m_dwInitialCTS = dwCTS;
}
	
// ===============================================================================

double StreamInfo ::GetDivergenceTillGood(int dwClockDivergence,DWORD dwObjectClock)
{
    DWORD dwElapsedTime = dwObjectClock - m_dwLastCorrection;
	if(dwElapsedTime<5000)
		return 0;
	else
	return  (double)dwClockDivergence/(double)dwElapsedTime;
}

void StreamInfo ::UpDateFrequency(double divergenceTillGood)
{
  divergenceTillGood *= m_wf.nSamplesPerSec;
  m_wf.nSamplesPerSec-=divergenceTillGood;
  m_pDSBuf->SetFrequency(m_wf.nSamplesPerSec);
  TRACE_PRINT2("AR: CD NewFrequency %d DivergenceTillGood %f\n",m_wf.nSamplesPerSec,divergenceTillGood);
}