//////////////////////////////////////////////////////////////////////////////
//This software module was originally developed by CSELT in the course of   //
//development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.             //
//This software module is an implementation of a part of one or more MPEG-4 //
//Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems        //
//(ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 Systems     //
//(ISO/IEC 14496-1) free license to this software module or modifications   //
//thereof for use in hardware or software products claiming conformance to  //
//the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use this software//
//module in hardware or software products are advised that its use may      //
//infringe existing patents.                                                //
//The original developer of this software module and his/her company, the   //
//subsequent editors and their companies, and ISO/IEC have no liability     //
//for use of this software module or modifications thereof in an            //
//implementation. Copyright is not released for non MPEG-4 Systems          //
//(ISO/IEC 14496-1) conforming products. CSELT S.pA. retains full right     //
//to use the code for his/her own purpose, assign or donate the code to a   //
//third party and to inhibit third parties from using the code for non      //
//MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright      //
//notice must be included in all copies or derivative works.                //
//Copyright  1999.                                                         //
//This work has been partly developed within the framework of the ESPRIT    //
//European Research Project  "MPEG-4 PC".                                   //
//////////////////////////////////////////////////////////////////////////////
//
// File: AudRend-CSELT-DS.cpp
//
// Author: Matthew Leditschke
//
// Modified: Gianluca Di Cagno, Guido Franceschini,Ananda Allys
//
// Description:
//   The implementation of a class which provides a DirectSound based
//   audio renderer.
//
// Created: August, 1997.
//
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "AudRend-CSELT-DS.h"
#include "AR-DS-StreamInfo.h"
#include "AudioSourceProxy.h"


#include "Sound2DProxy.h"

#include "Caps.h"
#include "Executiv.h"
#include "Presenter.h"
#include "Clock.h"
#include "dx-errors.h"
#include "trace.h"
#ifndef MAKE_TEST_APP
#include "MPEG4Player.h"
#include "FatalError.h"
#include "ThreadList.h"
#include "stream.h"
#endif


 
// =====================================================================

//
// AudioRendererDS::AudioRendererDS
//
// Construct the audio renderer, given a pointer to the current
// Presenter object, and the handle of the window the application
// is running in.
//

AudioRendererDS::AudioRendererDS(Presenter* pPresenter, HWND hWnd)
{
	char errorString[200];
	errorString[0] = 0;

	TRACE_PRINT("Entering AudioRendererDS::AudioRendererDS()\n");

	// Initialise data members
	m_bActive = FALSE;
	m_bPaused = FALSE;
	m_pPresenter = pPresenter;

	// Create the DirectSound object, the interface to the sound card
	HRESULT hResult = DirectSoundCreate(NULL, &m_pDS, NULL);
	if (hResult != DS_OK)
		sprintf (errorString, "DirectSoundCreate() failed.\n  %s",
				 HRESULTtoText(hResult));

	// Set the cooperative level
	if (errorString[0] == 0)
	{
		hResult = m_pDS->SetCooperativeLevel(hWnd, DSSCL_PRIORITY);
		if (hResult != DS_OK)
			sprintf (errorString, "SetCooperativeLevel() failed.\n  %s",
					 HRESULTtoText(hResult));
	}

	// Initialise the audio format output by the sound card
	memset(&m_wFmt, 0, sizeof(WAVEFORMATEX));
	m_wFmt.wFormatTag	  = WAVE_FORMAT_PCM;
	m_wFmt.nChannels      = 0;
	m_wFmt.nSamplesPerSec = 0;
	m_wFmt.wBitsPerSample = 0;
	m_wFmt.cbSize		  = 0;

	// If an error has occurred, display a message and then exit
	if (errorString[0] != 0)
	{
#ifdef STAND_ALONE
		CFatalError fatalErrorDialog;
		fatalErrorDialog.m_message = errorString;
		fatalErrorDialog.DoModal();
#endif
		m_initOK = FALSE;
	}
	else
		m_initOK = TRUE;

	TRACE_PRINT("Leaving AudioRendererDS::AudioRendererDS()\n");
} /* AudioRendererDS::AudioRendererDS */


// =====================================================================

//
// AudioRendererDS::~AudioRendererDS
//
// Class destructor.
//

AudioRendererDS::~AudioRendererDS()
{
	TRACE_PRINT("Entering AudioRendererDS::~AudioRendererDS():\n");

	// Make sure that the renderer is first terminated
	Terminate();

	// Release the main direct sound object
	if (m_pDS != NULL)
		m_pDS->Release();

	TRACE_PRINT("Leaving AudioRendererDS::~AudioRendererDS():\n");
} /* AudioRendererDS::~AudioRendererDS */


// =====================================================================

//
// AudioRendererDS::GetCapabilities
//
// Get the capabilities of this audio renderer.
//

void AudioRendererDS::GetCapabilities (Capabilities* pCaps)
{
#ifdef MAKE_TEST_APP
	pCaps;
#else
	pCaps->AddFormatCapability(MediaFormat (MediaFormat :: PCM), 2);
#endif
} /* AudioRendererDS::GetCapabilities */


// =====================================================================

//
// AudioRendererDS::Init
//
// Initialise the renderer.  The thread is started in Start(), so there
// is nothing to do here.  This method will be called once for every
// file that is opened.
//

void AudioRendererDS::Init()
{
} /* AudioRendererDS::Init */


// =====================================================================

//
// AudioRendererDS::Terminate
//
// Stops rendering and performs cleanups.  This method will be called
// once for every file that is closed.
//

void AudioRendererDS::Terminate()
{
	TRACE_PRINT("Entering AudioRendererDS::Terminate():\n");

	// Flag that the audio renderer is no longer active
	m_bActive = FALSE;

	// Wait for this thread to terminate
	m_mtxThread.P ();
	m_mtxThread.V ();

	TRACE_PRINT("Leaving AudioRendererDS::Terminate():\n");
} /* AudioRendererDS::Terminate */


// =====================================================================

//
// AudioRendererDS::Start
//
// Starts the audio renderer running in a separate thread.  The clock
// will be started by the thread itself.
//

void AudioRendererDS::Start()
{

	// Flag that the audio renderer is now active
	m_bActive = TRUE;

	// Start the main thread of the audio renderer
	ZThread<AudioRendererDS> thread(Main, this, THREAD_PRIORITY_ABOVE_NORMAL);
	int hThread = thread.Begin();
  	if (hThread == -1)
    	throw;
#ifdef STAND_ALONE
		AddToThreadList("AudioRenderer", hThread);
#endif
} /* AudioRendererDS::Start */


// =====================================================================

//
// AudioRendererDS::UpdateOutputFormat
//
// Given a stream which is being added to the list of streams to play,
// update the output format that the sound card is currently using.
//
// The sound card will be set to the maximum sample rate, bits per sample
// and number of channels of all of the input streams.  This means that
// DirectSound will be doing the sample rate conversion if required.
// This has been known to be a problem with some sounds cards, and also
// under WindowsNT, where this is done in software.
//

void AudioRendererDS::UpdateOutputFormat(StreamInfo* pStream)
{
	// Update the sound board settings
	if (m_wFmt.nSamplesPerSec < pStream->m_wf.nSamplesPerSec)
		m_wFmt.nSamplesPerSec = 44100;//pStream->m_wf.nSamplesPerSec;Gianluca to change
	if (m_wFmt.wBitsPerSample < pStream->m_wf.wBitsPerSample)
		m_wFmt.wBitsPerSample = pStream->m_wf.wBitsPerSample;
	if (m_wFmt.nChannels < pStream->m_wf.nChannels)
		m_wFmt.nChannels = pStream->m_wf.nChannels;

	// Set the output format of the sound card
	DSBUFFERDESC primaryDesc;
	memset(&primaryDesc, 0, sizeof(DSBUFFERDESC));
	primaryDesc.dwSize = sizeof(DSBUFFERDESC);
	primaryDesc.dwFlags = DSBCAPS_PRIMARYBUFFER;

	LPDIRECTSOUNDBUFFER primarySB;
	HRESULT hr = m_pDS->CreateSoundBuffer(&primaryDesc, &primarySB, NULL);
	assert(hr == DS_OK);

	m_wFmt.nBlockAlign	   = m_wFmt.nChannels * m_wFmt.wBitsPerSample / 8;
	m_wFmt.nAvgBytesPerSec = m_wFmt.nSamplesPerSec * m_wFmt.nBlockAlign;
	hr = primarySB->SetFormat(&m_wFmt);
	assert(hr == DS_OK);

	// Release the primary buffer as we have finished with it.
	primarySB->Release();

#ifdef _DEBUG
//	printf("Setting the output format to: %d Hz, %d bits/sample, %d channels.\n",
//		   m_wFmt.nSamplesPerSec, m_wFmt.wBitsPerSample, m_wFmt.nChannels);
#endif
} /* AudioRendererDS::UpdateOutputFormat */


// =====================================================================

//
// AudioRendererDS::AddSourceNode
//
// Add the given AudioSourceProxy node to the list of audio source nodes
// which need to be played.  If a node is already in the list, then
// just leave it there.  The time at which the audio is to start
// playing is also given.
//
// It is intended that this method will be called once the audio node
// has been parsed.
//
// The audio renderer will call AudioSource::GetFormat() to ask the
// node of the audio data format, once Init() has been called.
//

void AudioRendererDS::AddSourceNode(AudioSourceProxy* pSourceNode,	
									DWORD startTime,
									DWORD nSamplesPerSec,
									DWORD nBitsPerSample,
									WORD nNumChannels)
{
	assert(pSourceNode != NULL);
	assert(nSamplesPerSec > 0);
	assert(nBitsPerSample > 0);
	assert(nNumChannels > 0);

	// Need to lock the array before accessing to it
	m_mtxStreamArray.P();

	// If this node is already in the array, then return
	for (int i = 0; i < m_streams.GetLength(); i++)
		if (m_streams[i]->m_pSrc == pSourceNode)
		{
			m_mtxStreamArray.V();
			return;
		}

	// Create a stream object for this audio source node
	StreamInfo* pStream = new StreamInfo(m_pDS, pSourceNode, startTime);
	int result = pStream->Create(nSamplesPerSec, nBitsPerSample,
								 nNumChannels);
	assert(result == DS_OK);

	// Update the audio output format of the sound card
	UpdateOutputFormat(pStream);

	// Add this to the array of streams
	m_streams += pStream;


	// And now unlock it so that the main audio renderer thread can
	// access it.
	m_mtxStreamArray.V();
} /* AudioRendererDS::AddSourceNode */


// =====================================================================

//
// AudioRendererDS::RemoveSourceNode
//
// Remove the given AudioSourceProxy node from the list of audio source
// nodes which need to be played.  If the given node is not in the
// list, nothing happens.
//

void AudioRendererDS::RemoveSourceNode(AudioSourceProxy* pSourceNode)
{
	// Need to lock the array before accessing to it
	m_mtxStreamArray.P();

	// Find this node in the array
	for (int i = 0; i < m_streams.GetLength(); i++)
		if (m_streams[i]->m_pSrc == pSourceNode)
		{
			// Remove this node from the array
			StreamInfo* pStream = m_streams[i];
			m_streams -= pStream;

			// Close down this stream
			delete pStream;
			
			//close the stream source 
			if (pSourceNode->GetAudioSource()->GetStream() != NULL)
				pSourceNode->GetAudioSource()->CloseStream();
			
		}
	// And now unlock it so that the main audio renderer thread can
	// access it.
	m_mtxStreamArray.V();

} /* AudioRendererDS::RemoveSourceNode */


// =====================================================================

//
// AudioRendererDS::GetBufferLengthMs
//
// Returns the buffer length, in milliseconds, that the audio renderer has.
// This delay is the average time that it takes samples taken from
// the AudioSourceNode to be played.
//

DWORD AudioRendererDS::GetBufferLengthMs()
{
	return BUFFER_LENGTH;
} /* AudioRendererDS::GetBufferLengthMs */

// =====================================================================
void AudioRendererDS::Main()
{
// WARNING: most of these variables should be per stream.
// No audio mixing is supported by his code.
	// Mark the thread as running
	m_mtxThread.P();
	//SetThreadPriority (GetCurrentThread(), THREAD_PRIORITY_HIGHEST ); 
	unsigned long startingTime = GetTickCount();
	unsigned long updatesDone = 0;
	unsigned int startNextTime = 0;
	DWORD currentTime;
	DWORD objectClock;
	DWORD objectClock1;
	DWORD timeStamp = 0;
	DWORD lastTimeStamp = 0;
	DWORD lastLength = 0;
	DWORD lostBytes = 0;
	long clockDivergence = 0;
	int correctionTime = 0;
//	DWORD skipUntil = 0;
//	double divergenceTillGood = 0;
	bool bufferIsClean = TRUE;
	DWORD bytesInReadBuffer = 0;
	DWORD offsetInReadBuffer = 0;
	DWORD bytesRead = 0;
	DWORD bytesReadToWrite = 0;
	DWORD bytesLostToWrite = 0;
	DWORD bytesSilenceToWrite = 0;
	DWORD totalBytesToWrite = 0;
	DWORD bytesToSkip = 0;
	BOOL skipFlag = false;
	DWORD totalRead = 0;
	DWORD totalWritten = 0;
	DWORD virtuallyLostBytes = 0;
	LPBYTE pData = NULL;

	// This loop runs while the audio renderer is active
	// ie. up to the next Terminate() call.  Every BUFFER_UPDATE_TIME ms
	// the buffers will be updated.
	while (m_bActive)
	{
		// Lock the stream information data
		m_mtxStreamArray.P();
		
		// Reset the clockDivergence, that will be computed only when really playing
		clockDivergence = 0;

		// Read data from each of the active buffers
		TRACE_PRINT1("AR: Using %d streams.\n", m_streams.GetLength());
		for (int i = 0; i < m_streams.GetLength(); i++)
		{
			if(!m_streams[i]->m_pSrc->GetStream()->GetClockReference()->IsActive()&&(!m_bPaused))      
			{
				m_streams[i]->Stop();
				m_bPaused=TRUE;
			}
			if(m_streams[i]->m_pSrc->GetStream()->GetClockReference()->IsActive()&&(m_bPaused))      
			{
				m_streams[i]->Play();
				m_bPaused=FALSE;
			}
			currentTime = GetTickCount() - startingTime;
			// Only get data for streams whose start time has arrived
			if (currentTime >= m_streams[i]->m_startTime)
			{
				// Find out the Object Clock value
				objectClock = m_streams[i]->m_pSrc->GetStream()->GetClock();
				// Find out how much space is available in this buffer
				totalBytesToWrite = m_streams[i]->GetMaxWriteSize();

				// check again the Object Clock value: if it differs from the
				// previous value, then it means that the scheduler has stopped
				// this thread between the objectClock and bytesToWrite
				// computation. We have to recompute them, in order to
				// avoid miscomputation of clockDivergence
				objectClock1 = m_streams[i]->m_pSrc->GetStream()->GetClock();
				while (objectClock1 != objectClock) {
					objectClock = objectClock1;
					totalBytesToWrite = m_streams[i]->GetMaxWriteSize();
					objectClock1 = m_streams[i]->m_pSrc->GetStream()->GetClock();
				}

				// if the DS buffer wraps around, clean it completely
				if (!bufferIsClean && (objectClock > lastTimeStamp + GetBufferLengthMs())) {
					m_streams[i]->WriteSilence(m_streams[i]->m_bufDesc.dwBufferBytes, FALSE);
					TRACE_PRINT1("AR: ClearedA %d bytes.\n", m_streams[i]->m_bufDesc.dwBufferBytes);
					TRACE("AR: ClearedA %d bytes. clock %d \tlastTSplus %d lastTS %d\n", 
						m_streams[i]->m_bufDesc.dwBufferBytes, objectClock, lastTimeStamp + GetBufferLengthMs(), lastTimeStamp);
					bufferIsClean = TRUE;
				}
/*
				// possibly reset the Write Pointer: this is harmful however, since it interacts with
				// other pointers and computations.
				if(bufferIsClean) {
					m_streams[i]->ResetWritePointer();
					totalBytesToWrite = m_streams[i]->m_bufDesc.dwBufferBytes;
					TRACE_PRINT("AR: Reset Write Pointer\n");
				}
*/
				// Repeatedly read data for this buffer, until either the
				// buffer is full, or there is no more data, in which case
				// silence needs to be inserted.
				while (totalBytesToWrite > 0)
				{
					// Read data from the audio node
					bytesReadToWrite = 0;
					bytesLostToWrite = 0;
					bytesSilenceToWrite = 0;
					if (bytesInReadBuffer == 0) {
						m_streams[i]->m_pSrc->GetData(totalBytesToWrite, pData, bytesRead,timeStamp);
						assert(bytesRead <= totalBytesToWrite);
						if (bytesRead == 0)
							break;
						totalRead += bytesRead;
						lastLength = bytesRead;
						bytesInReadBuffer = bytesRead;
						offsetInReadBuffer = 0;
						skipFlag = false;
					}
					bytesReadToWrite = bytesInReadBuffer;

					/*-------------------------------------------------------------
					(Zvi) Ignore frames with zero time stamps
					-------------------------------------------------------------*/
					if (timeStamp == 0)
					{	m_streams[i]->m_pSrc->DataUsed(pData, bytesRead);
						bytesInReadBuffer = 0;
						continue;
					}

					m_streams[i] -> SetInitialCTS (timeStamp);
					if (timeStamp < lastTimeStamp)
					{
						totalRead = 0;
						totalWritten = 0;
					}

					/*-------------------------------------------------------------
					In case the Access Unit arrives too late, skip it, but remember
					that in practice this is similar to having lost it
					-------------------------------------------------------------*/
					if (!skipFlag && m_streams[i]->IsPlaying() && (objectClock > timeStamp))
					{
						bytesToSkip += lastLength;
						virtuallyLostBytes += lastLength;
						skipFlag = true;
					}

					/*-------------------------------------------------------------
					In case the timestamp is too far from the previous one, we have lost at least
					one Access Unit. The check is optimal for AUs of 11 ms.
					-------------------------------------------------------------*/
					if (totalWritten && ((timeStamp - 10) > (lastTimeStamp + (lastLength / m_streams[i]->GetBytesPerMs()))))
					{
						int nLostAUs = (double)(timeStamp - lastTimeStamp) / ((double)(lastLength) / m_streams[i]->GetBytesPerMs()) -1;
						lostBytes += lastLength * nLostAUs;
						lastTimeStamp += (lostBytes / m_streams[i]->GetBytesPerMs());
					}

					/*-------------------------------------------------------------
					If Access Units have been skipped in the past, now not anymore,
					consider them as lost
					-------------------------------------------------------------*/
					if (!skipFlag && virtuallyLostBytes) {
						lostBytes += virtuallyLostBytes;
						virtuallyLostBytes = 0;
					}

					/*-------------------------------------------------------------
					Write into the buffer the number of lost bytes modulo the
					buffer size. This way lip-sync should be preserved
					-------------------------------------------------------------*/
					bytesLostToWrite = lostBytes % m_streams[i]->m_bufDesc.dwBufferBytes;
					TRACE("AR: totToW %5d\tlost %d\tlostToW %d\tsilToW %d\tskipToW %d\treadToW %d\n",
						totalBytesToWrite, lostBytes, bytesLostToWrite, bytesSilenceToWrite, bytesToSkip, bytesReadToWrite);
					TRACE("AR: objectClock %d\ttimeStamp %d\tequivLastTimeStamp %d\ttotalW %d\n",
						objectClock, timeStamp, lastTimeStamp, totalWritten);
					lostBytes = bytesLostToWrite;

					// Compute the clock divergence only if the stream is
					// really playing, to avoid wrong computations
					if (m_streams[i]->IsReallyPlaying())
					{
						// clockDivergence is computed by comparing the CTS of the
						// byte being currently played and the current stream clock. 
						// The CTS of the byte currently played is computed by
						// subtracting the time needed for playing all the bytes
						// in the DS buffer from the CTS of the CU currently fetched.
						
						clockDivergence = timeStamp
//										+ m_streams[i]->GetThreshold()/2 
										+ totalBytesToWrite/m_streams[i]->GetBytesPerMs()
										- GetBufferLengthMs()
										- objectClock;
//TRACE("CD clockDivergence %6d   ", clockDivergence);
						clockDivergence += currentTime - startNextTime;
//TRACE(" %6d\n", clockDivergence);
						TRACE_PRINT1("AR: CD clockDivergenceD %d \n",clockDivergence);
						clockDivergence =0;
					}

					// Align to the Object Clock, in case DS clock is diverging
					// or CUs have been lost
/*
					// If the clock is feed by OCRs, then use that clock as the
					//  master clock, and compensate for drifts in the audioboard.

					if (m_streams[i]->m_pSrc->GetStream()->GetClock()->HasOCR())
					{
					// use clock divergence as before;
					}
					else
					{
						// compute factor
						m_streams[i]->m_pSrc->GetStream()->GetClock()->ModifySpeed(factor);
					}
*/
					// Case 1: DS clock is running too fast (or CUs have been lost)
					// Insert silence
					if (clockDivergence > 80)
					{
						bytesSilenceToWrite = (clockDivergence - 40)*m_streams[i]->GetBytesPerMs();
					}
					// Case 2: DS clock is running too slowly
					// Skip data
					else if (clockDivergence < -80)
					{
						bytesToSkip = (-clockDivergence - 40)*m_streams[i]->GetBytesPerMs();
					}

					if (bytesLostToWrite > totalBytesToWrite)
						bytesLostToWrite = totalBytesToWrite;
					if (bytesLostToWrite > 0)
					{
						m_streams[i]->WriteSilence(bytesLostToWrite, TRUE);
						totalWritten += bytesLostToWrite;
						totalBytesToWrite -= bytesLostToWrite;
						lostBytes -= bytesLostToWrite;
						bufferIsClean = FALSE;
						TRACE_PRINT3("AR: LOST BYTES. lostBytes %d (bytesLostToWrite %d) before timeStamp %d\n",lostBytes,bytesLostToWrite,timeStamp);
					}
					if (bytesSilenceToWrite > totalBytesToWrite)
						bytesSilenceToWrite = totalBytesToWrite;
					if (bytesSilenceToWrite > 0)
					{
						m_streams[i]->WriteSilence(bytesSilenceToWrite, TRUE);
						totalWritten += bytesSilenceToWrite;
						totalBytesToWrite -= bytesSilenceToWrite;
						TRACE_PRINT5("AR: DS is too fast. clockDivergence %d bytesSilenceToWrite %d totalBytesToWrite %d NOW %d SClock %d\n",clockDivergence,bytesSilenceToWrite,totalBytesToWrite,timeGetTime()-m_pPresenter->GetStartTime(),m_streams[i]->m_pSrc->GetStream()->GetClock());
					}
					if (bytesReadToWrite > totalBytesToWrite)
						bytesReadToWrite = totalBytesToWrite;
					if (bytesToSkip)
					{
						int bytesSkipped = min (bytesToSkip, bytesReadToWrite);
						TRACE_PRINT5("AR: DS is too slow. clockDivergence %d bytesToSkip %d bytesSkipped %d NOW %d SClock %d\n",clockDivergence,bytesToSkip,bytesSkipped,timeGetTime()-m_pPresenter->GetStartTime(),m_streams[i]->m_pSrc->GetStream()->GetClock());
						bytesToSkip -= bytesSkipped;
						bytesReadToWrite -= bytesSkipped;
						bytesInReadBuffer -= bytesSkipped;
						offsetInReadBuffer += bytesSkipped;
					}
					if (bytesReadToWrite > 0)
					{
						m_streams[i]->WriteData(pData+offsetInReadBuffer, bytesReadToWrite,
						m_streams[i]->m_pSrc->GetIntensity());
						totalWritten += bytesReadToWrite;
						totalBytesToWrite -= bytesReadToWrite;
						bytesInReadBuffer -= bytesReadToWrite;
						offsetInReadBuffer += bytesReadToWrite;
						bufferIsClean = FALSE;
					}
					// Release this data
					if (bytesInReadBuffer == 0)
					{
						m_streams[i]->m_pSrc->DataUsed(pData, bytesRead);
						offsetInReadBuffer = 0;
						lastTimeStamp = timeStamp;
					}
				}

				// If the buffer was not completely filled, we need to
				// fill the rest of the buffer with silence. This is needed
				// because of the circular buffer used by DirectSound.
				if (!bufferIsClean && (totalBytesToWrite > 0)) {
					m_streams[i]->WriteSilence(totalBytesToWrite, FALSE);
					TRACE_PRINT1("AR: ClearedC %d bytes.\n", totalBytesToWrite);
				}
				// Start the stream playing if it is not already playing
				if (!m_streams[i]->IsPlaying() &&
					totalWritten > 0 &&
					(objectClock - 10 >= lastTimeStamp - totalWritten/m_streams[i]->GetBytesPerMs()))
//				if (!m_streams[i]->IsPlaying() && 
//					(totalWritten > GetBufferLengthMs()*m_streams[i]->GetBytesPerMs()*1/4) &&
//					(objectClock - 100 > lastTimeStamp - totalWritten/m_streams[i]->GetBytesPerMs()))
				{
					m_streams[i]->Play();
					startingTime = GetTickCount() - m_streams[i]->GetThreshold()/2;
					updatesDone = 0;
//					correctionTime = 0;
//					skipUntil = 0;
					TRACE_PRINT("AR: Starting stream playing.\n");
					TRACE("AR: Starting stream playing.\n");
				}
				TRACE_PRINT3("AR: after filling buffer (%d bytes to write) current time %d SClock %d\n",totalBytesToWrite,timeGetTime()-m_pPresenter->GetStartTime(),m_streams[i]-> m_pSrc->GetStream()->GetClock());
//				if (!skipUntil)
//					correctionTime += clockDivergence;
//				TRACE_PRINT1("AR: Correction: time %d\n",correctionTime);
			}
		}
		
		// Unlock the stream information data
		m_mtxStreamArray.V();

		// Wait until it is time for the next buffer update
		updatesDone++;
		unsigned int elapsedTime = GetTickCount() - startingTime;
		startNextTime = updatesDone * BUFFER_UPDATE_TIME + correctionTime;

		// Only sleep if ahead of schedule
		if ((m_bActive) && (startNextTime > elapsedTime)) {
			TRACE_PRINT1("AR: Sleeping for %d ms.\n", startNextTime - elapsedTime);
			Sleep(startNextTime - elapsedTime);
		} else {
			TRACE_PRINT1("AR: Not sleeping (late for %d ms).\n", - (startNextTime - elapsedTime));
		}
	}
	// Clean the DS buffers
	for (int i = 0; i < m_streams.GetLength(); i++)
	{
		m_streams[i]->WriteSilence(m_streams[i]->m_bufDesc.dwBufferBytes, FALSE);
		TRACE_PRINT1("AR: ClearedC %d bytes.\n", m_streams[i]->m_bufDesc.dwBufferBytes);
	}

	// Flag that the thread is terminating
	m_mtxThread.V();
} /* AudioRendererDS::Main */

// =====================================================================
