///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File: AudRend-CSELT-Sgi.cpp
//
// Author: Matthew Leditschke, Gianluca Di Cagno  gianluca.dicagno@cselt.it
//
// Description:
//   The implementation of a class which provides an audio renderer for
//   the SGI platform.
//
// Created: August, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "AudRend-CSELT-Sgi.h"
#include "StreamInfo.h"
#include "Caps.h"
#include "Presenter.h"
#include "Executiv.h"
#include "Clock.h"
#include "AudioSource.h"
#include "trace.h"


// The length of the output queue (in milliseconds)
const int QUEUE_LENGTH = 400;

// How often to update the buffer (in milliseconds)
const int BUFFER_UPDATE_TIME = 200;


// ============================================================================

//
// Define a GetTickCount function as the SGI doesn't have it.
//

static unsigned long GetTickCount()
{
	struct timeval currentTime;
	gettimeofday(&currentTime, NULL);

	unsigned long tickCount = currentTime.tv_sec * 1000 + currentTime.tv_usec / 1000;
	return tickCount;
} /* GetTickCount */


// ============================================================================

//
// AudioRendererSgi::AudioRendererSgi
//
// Construct the audio renderer, given a pointer to the current
// Presenter object.
//

AudioRendererSgi::AudioRendererSgi(Presenter* pPresenter)
{
	TRACE_PRINT("Entering AudioRendererSgi::AudioRendererSgi():\n");

	// Initialise data members
	m_bActive = FALSE;
	m_pPresenter = pPresenter;

	TRACE_PRINT("Leaving AudioRendererSgi::AudioRendererSgi():\n");
} /* AudioRendererSgi::AudioRendererSgi */


// ============================================================================

//
// AudioRendererSgi::~AudioRendererSgi
//
// Class destructor.
//

AudioRendererSgi::~AudioRendererSgi()
{
	TRACE_PRINT("Entering AudioRendererSgi::~AudioRendererSgi():\n");

	// Make sure that the renderer is first terminated
	Terminate();

	TRACE_PRINT("Leaving AudioRendererSgi::~AudioRendererSgi():\n");
} /* AudioRendererSgi::~AudioRendererSgi */


// ============================================================================

//
// AudioRendererSgi::GetCapabilities
//
// Get the capabilities of this audio renderer.
//

void AudioRendererSgi::GetCapabilities (Capabilities* pCaps)
{
#ifdef MAKE_TEST_APP
	pCaps;
#else
	pCaps->AddFormatCapability(FORMAT_PCM_8KHZ_16BIT, 2);
#endif
} /* AudioRendererSgi::GetCapabilities */


// ============================================================================

//
// AudioRendererSgi::Init
//
// Initialise the renderer.  The thread is started by Start(), so this
// method just needs to initialise some data members.
//

void AudioRendererSgi::Init()
{
	TRACE_PRINT("Entering AudioRendererSgi::Init():\n");

	m_audioConfig = NULL;
	m_audioPort = NULL;
	m_mixBuf = NULL;

	TRACE_PRINT("Leaving AudioRendererSgi::Init():\n");
} /* AudioRendererSgi::Init */


// ============================================================================

//
// AudioRendererSgi::Terminate
//
// Stops rendering and performs cleanups.
//

void AudioRendererSgi::Terminate()
{
	TRACE_PRINT("Entering AudioRendererSgi::Terminate():\n");

	// Flag that the audio renderer is no longer active
	m_bActive = FALSE;

	// Wait for this thread to terminate
	m_mtxThread.P ();
	m_mtxThread.V ();

	// Release the audio output related resources
	if (m_audioPort != NULL)
		ALcloseport(m_audioPort);
	if (m_audioConfig != NULL)
		ALfreeconfig(m_audioConfig);

	if (m_mixBuf != NULL)
	{
		delete m_mixBuf;
		m_mixBuf = NULL;
	}

	TRACE_PRINT("Leaving AudioRendererSgi::Terminate():\n");
} /* AudioRendererSgi::Terminate */


// ============================================================================

//
// AudioRendererSgi::StartThread
//
// This is the first function called in the thread, and simply calls the
// Main() member function of the class.
//

void AudioRendererSgi::StartThread(void* params)
{
	AudioRendererSgi* pAudioRendererSgi = (AudioRendererSgi*) params;
	pAudioRendererSgi->Main();
} /* AudioRendererSgi::StartThread */


// ============================================================================

//
// AudioRendererSgi::Start
//
// Starts the audio renderer running in a separate thread.  The clock
// will be started in the thread itself.
//

void AudioRendererSgi::Start()
{
	TRACE_PRINT("Entering AudioRendererSgi::Start():\n");

	// Flag that the audio renderer is now active
	m_bActive = TRUE;

	// Start the main thread of the audio renderer
  	if (ZThread::BeginThread (StartThread, this) == -1)
#ifdef DONT_USE_EXCEPTIONS
    	return;
#else
    	throw;
#endif

	TRACE_PRINT("Leaving AudioRendererSgi::Start():\n");
} /* AudioRendererSgi::Start */


// ============================================================================

//
// AudioRendererSgi::InitAudioOutput
//
// This function is called when the first audio output stream is known.
// The audio output format settings of this first stream are given to
// this function, and the output format of the audio device is set to this.
// If other streams are added which have a different output format, then
// their format will need to be changed in software.
//

void AudioRendererSgi::InitAudioOutput(DWORD samplesPerSec,
									   DWORD bitsPerSample,
									   WORD numChannels)
{
	// Remove the default error handler
	ALseterrorhandler(0);

	// Specify the configuration of the output audio port
	m_audioConfig = ALnewconfig();
	ALsetsampfmt(m_audioConfig, AL_SAMPFMT_TWOSCOMP);
	if (bitsPerSample <= 8)
		ALsetwidth(m_audioConfig, AL_SAMPLE_8);
	else
		ALsetwidth(m_audioConfig, AL_SAMPLE_16);
	ALsetchannels(m_audioConfig, (long)numChannels);

	// Determine and set the length of the output queue (in samples)
	m_outputQueueSize = samplesPerSec * QUEUE_LENGTH / 1000;
	ALsetqueuesize(m_audioConfig, m_outputQueueSize);

	// Create the audio port
	m_audioPort = ALopenport("Mpeg4Player", "w", m_audioConfig);
	if (m_audioPort == (ALport) 0)
	{
		int err = oserror();
		if (err == AL_BAD_NO_PORTS)
		{
			fprintf(stderr, " System is out of audio ports\n");
		}
		else if (err == AL_BAD_DEVICE_ACCESS)
		{
			fprintf(stderr, " Couldn't access audio device\n");
		}
		else if (err == AL_BAD_OUT_OF_MEM)
		{ 
			fprintf(stderr, " Out of memory\n"); 
		} 
		exit(1);  
	} 

	// Change the output sample rate
	long params[2];
	params[0] = AL_OUTPUT_RATE;
	params[1] = samplesPerSec;
	ALsetparams(AL_DEFAULT_DEVICE, params, 2);

	// Remember the output port settings
	m_samplesPerSec  = samplesPerSec;
	m_bitsPerSample  = bitsPerSample;
	m_numChannels    = numChannels;
	m_bytesPerSample = numChannels * bitsPerSample / 8;

	// Allocate a buffer to mix samples from different streams in
	m_mixBuf = new char [m_outputQueueSize * m_bytesPerSample];
} /* AudioRendererSgi::InitAudioOutput */


// ============================================================================

//
// AudioRendererSgi::AddSourceNode
//
// Add the given AudioSource node to the list of audio source nodes
// which need to be played.  If a node is already in the list, then
// just leave it there.  The time at which the audio is to start
// playing is also given.
//
// It is intended that this method will be called once the audio node
// has been parsed.
//
// The audio renderer will call AudioSource::GetFormat() to ask the
// node of the audio data format, once Init() has been called.
//

void AudioRendererSgi::AddSourceNode(AudioSource* pAudioSourceNode,	
									 DWORD startTime,
									 DWORD samplesPerSec,
									 DWORD bitsPerSample,
									 WORD numChannels)
{
	// Need to lock the array before accessing to it
	m_mtxStreamArray.P();

	// If this node is already in the array, then return
	for (int i = 0; i < m_streams.GetLength(); i++)
		if (m_streams[i]->m_pAudioSrc == pAudioSourceNode)
		{
			m_mtxStreamArray.V();
			return;
		}

	// If this is the first node added, then set the output of the audio
	// port to this format
	if (m_streams.GetLength() == 0)
		InitAudioOutput(samplesPerSec, bitsPerSample, numChannels);

	// otherwise abort if the format is different for this new stream
	// (can't do sample rate and format conversion yet)
	else
	{
		if ((samplesPerSec != m_samplesPerSec) ||
			(bitsPerSample != m_bitsPerSample) ||
			(numChannels != m_numChannels))
		{
			fprintf(stderr,"All audio sources must have the same format.\n");
			exit(-1);
		}
	}

	// Create a stream object for this audio source node
	StreamInfo* pStream = new StreamInfo(pAudioSourceNode, startTime,
										 bitsPerSample, numChannels);

	// Add this to the array of streams
	m_streams += pStream;

	// And now unlock it so that the main audio renderer thread can
	// access it.
	m_mtxStreamArray.V();
} /* AudioRendererSgi::AddSourceNode */


// ============================================================================

//
// AudioRendererSgi::RemoveSourceNode
//
// Remove the given AudioSource node from the list of audio source
// nodes which need to be played.  If the given node is not in the
// list, nothing happens. (Debug version will display a warning
// message.)
//

void AudioRendererSgi::RemoveSourceNode(AudioSource* pAudioSourceNode)
{
	// Need to lock the array before accessing to it
	m_mtxStreamArray.P();

	// Find this node in the array
	for (int i = 0; i < m_streams.GetLength(); i++)
		if (m_streams[i]->m_pAudioSrc == pAudioSourceNode)
		{
			// Remove this node from the array
			StreamInfo* pStream = m_streams[i];
			m_streams -= pStream;

			// Close down this stream
			delete pStream;
		}

	// And now unlock it so that the main audio renderer thread can
	// access it.
	m_mtxStreamArray.V();

} /* AudioRendererSgi::RemoveSourceNode */


// ============================================================================

//
// AudioRendererSgi::Main
//
// This is the main loop of the audio renderer thread.
//

void AudioRendererSgi::Main()
{
	TRACE_PRINT("Entering AudioRendererSgi::Main():\n");

	// Have this thread run at a higher than normal priority
	ZThread::SetPriority(THREAD_PRIORITY_ABOVE_NORMAL);

	// Mark the thread as running.
	m_mtxThread.P();

	// Start the main clock running, and record the starting time
	// (which will be zero).
#ifdef MAKE_TEST_APP
	unsigned long startingTime = GetTickCount();
#else
	TRACE_PRINT("AR: Starting the clock running.\n");
	m_pPresenter->GetExecutive()->GetClock()->Start();
	unsigned long startingTime = 0;
#endif
	unsigned long updatesDone = 0;

	// This loop runs while the audio renderer is active
	// ie. up to the next Terminate() call.  Every BUFFER_UPDATE_TIME ms
	// the buffers will be updated.
	while (m_bActive)
	{
		// Lock the stream information data
		m_mtxStreamArray.P();

		// If no streams have been attached, don't do any writing.
		if (m_streams.GetLength())
		{
			// Unlock the stream information data
			m_mtxStreamArray.V();
		}
		else
		{
			// Determine how many samples can be written
			long samplesToWrite = ALgetfillable(m_audioPort);

			// Only read data from the streams if there is room
			if (samplesToWrite > 0)
			{
				// Read data from each of the active buffers, merging the
				// results together into the mixing buffer.
				int streamCount = 0;
				for (int i = 0; i < m_streams.GetLength(); i++)
				{
					DWORD currentTime = GetTickCount() - startingTime;

					// Only get data for streams whose start time has arrived
					if (currentTime >= m_streams[i]->m_startTime)
					{
						streamCount++;

						// Determine number of bytes to read from this source
						long bytesPerSample = m_streams[i]->m_bytesPerSample;
						DWORD bytesToWrite = samplesToWrite * bytesPerSample;
						DWORD bytesWritten = 0;

						// Repeatedly read data for this buffer, until either
						// the buffer is full, or there is no more data, in
						// which case silence needs to be inserted.
						DWORD bytesRead = 1;
						while ((bytesToWrite > 0) && (bytesRead > 0))
						{
							// Read data from the audio node
							LPBYTE pData;
							m_streams[i]->m_pAudioSrc->GetData(bytesToWrite,
															   pData,
															   bytesRead);
				
							if (bytesRead > 0)
							{
								// If this is the first stream, copy it into
								// the mixing buffer.
								if (streamCount == 1)
									memcpy(m_mixBuf+bytesWritten, pData,
										   bytesRead);
	
								// otherwise combine it with the existing data
								else
									CombineData(bytesWritten, pData,
												bytesRead);

								// Release this data
								m_streams[i]->m_pAudioSrc->DataUsed(pData,
																	bytesRead);

								// Update how much data is left to get
								bytesToWrite -= bytesRead;
								bytesWritten += bytesRead;
							}
						}

						// If the buffer was not completely filled and this
						// is the first stream, we need to
						// fill the rest of the buffer with silence.
						if ((streamCount == 1) && (bytesWritten<bytesToWrite))
							WriteSilence(bytesWritten, samplesToWrite);
					}
				}

				// Unlock the stream information data
				m_mtxStreamArray.V();

				// Write these combined samples to the audio port
				ALwritesamps(m_audioPort, m_mixBuf, samplesToWrite);
			}
		}

		// Wait until it is time for the next buffer update
		updatesDone++;
		unsigned int elapsedTime = GetTickCount() - startingTime;
		unsigned int startNextTime = updatesDone * BUFFER_UPDATE_TIME;

		// Only sleep if ahead of schedule
		if (startNextTime > elapsedTime)
			Sleep(startNextTime - elapsedTime);
	}

	// Flag that the thread is terminating
	m_mtxThread.V();

	TRACE_PRINT("Leaving AudioRendererSgi::Main():\n");
} /* AudioRendererSgi::Main */


// ============================================================================

//
// AudioRendererSgi::CombineData
//
// Combine the given audio data with the audio data already in the
// mux buffer.  The offset specifies the offset (in bytes) of where
// to start combining the data.
//

void AudioRendererSgi::CombineData(DWORD offset, LPBYTE pSrc, DWORD lenSrc)
{
	DWORD numSamps  = lenSrc / m_bytesPerSample;

	if (m_bitsPerSample <= 8)
	{
		// Assume 8 bit unsigned format
		unsigned char* oldSamps = (unsigned char*)(m_mixBuf + offset);
		unsigned char* newSamps = (unsigned char*)pSrc;

		for (DWORD i = 0; i < numSamps; i++)
		{
			unsigned short temp = oldSamps[i] + newSamps[i];
			if (temp > 255)
				oldSamps[i] = 255;
			else
				oldSamps[i] = (unsigned char) temp;
		}
	}
	else
	{
		// Assume 16 bit signed format
		signed short* oldSamps = (signed short*)(m_mixBuf + offset);
		signed short* newSamps = (signed short*)pSrc;

		for (DWORD i = 0; i < numSamps; i++)
		{
			signed long temp = oldSamps[i] + newSamps[i];
			if (temp < -32768)
				oldSamps[i] = -32768;
			else if (temp > 32767)
				oldSamps[i] = 32767;
			else
				oldSamps[i] = (signed short) temp;
		}
	}
} /* AudioRendererSgi::CombineData */


// ============================================================================

//
// AudioRendererSgi::WriteSilence
//
// Write silence into the mux buffer, starting at the given offset
// (in bytes), and the fullness specifies the number of samples in
// the buffer.
//

void AudioRendererSgi::WriteSilence(DWORD offset, DWORD fullness)
{
	DWORD startSample = offset / m_bytesPerSample;

	if (m_bitsPerSample <= 8)
	{
		// Assume 8 bit unsigned format
		unsigned char* samples = (unsigned char*)(m_mixBuf + offset);

		for (DWORD i = startSample; i < fullness; i++)
			samples[i] = 0;
	}
	else
	{
		// Assume 16 bit signed format
		signed short* samples = (signed short*)(m_mixBuf + offset);

		for (DWORD i = startSample; i < fullness; i++)
			samples[i] = 0;
	}
} /* AudioRendererSgi::WriteSilence */


// ============================================================================


// Emacs uses these...
// Local Variables:
// mode: c++
// tab-width: 4
// End:
