//////////////////////////////////////////////////////////////////////////////
//This software module was originally developed by CSELT in the course of   //
//development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.             //
//This software module is an implementation of a part of one or more MPEG-4 //
//Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems        //
//(ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 Systems     //
//(ISO/IEC 14496-1) free license to this software module or modifications   //
//thereof for use in hardware or software products claiming conformance to  //
//the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use this software//
//module in hardware or software products are advised that its use may      //
//infringe existing patents.                                                //
//The original developer of this software module and his/her company, the   //
//subsequent editors and their companies, and ISO/IEC have no liability     //
//for use of this software module or modifications thereof in an            //
//implementation. Copyright is not released for non MPEG-4 Systems          //
//(ISO/IEC 14496-1) conforming products. CSELT S.pA. retains full right     //
//to use the code for his/her own purpose, assign or donate the code to a   //
//third party and to inhibit third parties from using the code for non      //
//MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright      //
//notice must be included in all copies or derivative works.                //
//Copyright  1999.                                                         //
//This work has been partly developed within the framework of the ESPRIT    //
//European Research Project  "MPEG-4 PC".                                   //
//////////////////////////////////////////////////////////////////////////////
//
// File:  BRect.cpp
//
// Author: Matthew Leditschke
//
// Modified: Gianluca Di Cagno  gianluca.dicagno@cselt.it
//
// Description:
//   The implementation of a class which stores bounding rectangle
//   information, and provides methods for operating on bounding
//   rectangles.
//
//   Note that this rectangle uses BIFS coordinates, with the origin
//   in the middle of the screen, increasing X to the right, and
//   increasing Y upwards.
//
// Created: September, 1997.
//
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "BRect.h"
#include <assert.h>


// ======================================================================

//
// BRect::MakeEmpty
//
// Make this bounding box empty.
//

void BRect::MakeEmpty()
{
	left = 0;
	top = 0;
	width = 0;
	height = 0;
} /* BRect::MakeEmpty */


// ======================================================================

//
// BRect::SetDimensions
//
// Set the dimensions of the bounding rectangle, given the 4 separate
// values.
//

void BRect::SetDimensions(int l, int t, int w, int h)
{
	left = l;
	top = t;
	width = w;
	height = h;
} /* BRect::SetDimensions */


// ======================================================================

//
// BRect::SetDimensions
//
// Set the dimensions of the bounding rectangle, given another bounding
// rectangle.
//

void BRect::SetDimensions(BRect const& other)
{
	left = other.left;
	top = other.top;
	width = other.width;
	height = other.height;
} /* BRect::SetDimensions */


// ======================================================================

//
// BRect::SetDimensions
//
// Set the given bounding rectangle to be the smallest rectangle
// which completely enclosed the given list of points.
//

void BRect::SetDimensions(PointF const* lpPoints, int numPoints)
{
	if (numPoints == 0)
		MakeEmpty();
	else
	{
		left = lpPoints[0].x;
		top  = lpPoints[0].y;
		int right  = lpPoints[0].x;
		int bottom = lpPoints[0].y;

		for (int i = 1; i < numPoints; i++)
		{
			if (lpPoints[i].x < left)	  left = lpPoints[i].x;
			if (lpPoints[i].x > right)   right = lpPoints[i].x;
			if (lpPoints[i].y > top)	   top = lpPoints[i].y;
			if (lpPoints[i].y < bottom) bottom = lpPoints[i].y;
		}

		width  = right - left + 1;
		height = top - bottom + 1;
	}
} /* BRect::SetDimensions */


// ======================================================================

//
// BRect::Union
//
// The current bounding box becomes the union of itself and
// the given bounding box.
//  - if other box is empty, nothing happens.
//  - if this box is empty, then it is set to the other box.
//

void BRect::Union(BRect const& other)
{
	if (!other.IsEmpty())
	{
		if (IsEmpty())
			SetDimensions(other);
		else
		{
			// In BIFS coordinates, increasing X is from left to right
			if (other.left < left)
			{
				width += left - other.left;
				left = other.left;
			}
			if (other.left + other.width > left + width)
				width = other.left + other.width - left;
			// In BIFS coordinates, increasing Y is from bottom to top
			if (other.top > top)
			{
				height += other.top - top;
				top = other.top;
			}
			if (other.top - other.height < top - height)
				height = top - other.top + other.height;
		}
	}
} /* BRect::Union */


// ======================================================================

//
// BRect::Intersection
//
// The current bounding box becomes the intersection of itself and
// the given bounding box.
//

void BRect::Intersection(BRect const& other)
{
	// If this rectangle is empty, the intersection is also empty
	if (IsEmpty())
		return;

	// If the other rectangle is empty, then the intersection is empty
	if (other.IsEmpty())
	{
		MakeEmpty();
		return;
	}

	// In BIFS coordinates, increasing X is from left to right
	if ((left + width < other.left) || (other.left + other.width < left))
	{
		// Handle non overlapping rectangles
		MakeEmpty();
		return;
	}
	if (other.left > left)
	{
		width -= other.left - left;
		left = other.left;
	}
	if (other.left + other.width < left + width)
		width = other.left + other.width - left;
	assert(width >= 0);

	// In BIFS coordinates, increasing Y is from bottom to top
	if ((top - height > other.top) || (other.top - other.height > top))
	{
		// Handle non overlapping rectangles
		MakeEmpty();
		return;
	}
	if (other.top < top)
	{
		height -= top - other.top;
		top = other.top;
	}
	if (other.top - other.height > top - height)
		height = top - other.top + other.height;
	assert(height >= 0);

} /* BRect::Intersection */


// ======================================================================

//
// BRect::Overlaps
//
// Return TRUE if the current bounding box and the given bounding
// box overlap, otherwise return FALSE.  If either box is empty,
// return FALSE.
//

BOOL BRect::Overlaps(BRect const& other) const
{
	// If either rectangle is empty, return FALSE
	if ((IsEmpty()) || (other.IsEmpty()))
		return FALSE;

	// Is the other box completely to the left of this one?
	if (other.left + other.width < left)
		return FALSE;

	// Is the other box completely to the right of this one?
	if (other.left > left + width)
		return FALSE;

	// Is the other box completely above this one?
	if (other.top - other.height > top)
		return FALSE;

	// Is the other box completely below this one?
	if (other.top < top - height)
		return FALSE;

	// Otherwise they must overlap
	return TRUE;
} /* BRect::Overlaps */


// ======================================================================

//
// operator<<(BRect)
//
// Print out the bounding rectangle to the given stream.
//

ostream& operator<<(ostream& out, BRect const& bRect)
{
	out << "(left = " << bRect.left << ", top = " << bRect.top
		<< ", width = " << bRect.width << ", height = " << bRect.height
		<< ")";

	return out;
} /* operator<<(BRect) */


// ======================================================================

//
// BRect::Print
//
// Print out the current bounding rectangle to the given file.
//

void BRect::Print(FILE* outfile) const
{
	fprintf(outfile, (char*)*this);
} /* BRect::Print */


// ======================================================================

//
// BRect::operator char*
//
// Return a pointer to a string (in static memory) which describes
// the dimensions of the bounding rectangle.
//

BRect::operator char*() const
{
	static char text[80];

	if (IsEmpty())
		sprintf(text, "(empty)");
	else
		sprintf(text, "(left = %d, top = %d, width = %d, height = %d)",
				left, top, width, height);

	return text;
} /* operator char* */


// ======================================================================

//
// BRect::IsPointIn
//
// Return TRUE if the given point is contained in the bounding
// rectangle, otherwise return FALSE.
//

BOOL BRect::IsPointIn(int x, int y) const
{
	if ((x >= left) && (x <= left + width) &&
		(y <= top) && (y >= top - height))
		return TRUE;
	else
		return FALSE;
} /* BRect::IsPointIn */


// ======================================================================

//
// operator==(BRect, BRect)
//
// Return TRUE if the two given bounding rectangles are the same.
//

int operator==(const BRect& rect1,	const BRect& rect2)
{
	if ((rect1.top == rect2.top) &&
		(rect1.left == rect2.left) &&
		(rect1.width == rect2.width) &&
		(rect1.height == rect2.height))
		return TRUE;
	else
		return FALSE;
} /* operator==(BRect, BRect) */

