//////////////////////////////////////////////////////////////////////////////
//This software module was originally developed by CSELT in the course of   //
//development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.             //
//This software module is an implementation of a part of one or more MPEG-4 //
//Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems        //
//(ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 Systems     //
//(ISO/IEC 14496-1) free license to this software module or modifications   //
//thereof for use in hardware or software products claiming conformance to  //
//the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use this software//
//module in hardware or software products are advised that its use may      //
//infringe existing patents.                                                //
//The original developer of this software module and his/her company, the   //
//subsequent editors and their companies, and ISO/IEC have no liability     //
//for use of this software module or modifications thereof in an            //
//implementation. Copyright is not released for non MPEG-4 Systems          //
//(ISO/IEC 14496-1) conforming products. CSELT S.pA. retains full right     //
//to use the code for his/her own purpose, assign or donate the code to a   //
//third party and to inhibit third parties from using the code for non      //
//MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright      //
//notice must be included in all copies or derivative works.                //
//Copyright  1999.                                                         //
//This work has been partly developed within the framework of the ESPRIT    //
//European Research Project  "MPEG-4 PC".                                   //
//////////////////////////////////////////////////////////////////////////////
//
// File: CopyYUVA.cpp
//
// Author: Matthew Leditschke 
//
// Modified: Gianluca Di Cagno  gianluca.dicagno@cselt.it
//
// Description:
//   The implementation of the functions which copy a YUV image
//   to the offscreen RGB surface in the visual renderer.
//
// Created: April, 1997
//
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "VisRend.h"
#include "Copy.h"
#include "trace.h"
#include "ddraw.h"
#include "clipping.h"

// ===============================================================================

//
// SetUpSrcYUVpointers
//
// Set up the source pointers for a YUV image.
//

static void SetUpSrcYUVpointers(SrcYUVAImage const& srcImage,
								CopyParameters const& cpParams,
								unsigned int srcY,
								unsigned char const*& pSrcY,
								unsigned char const*& pSrcU,
								unsigned char const*& pSrcV,
								unsigned char const*& pSrcA,
								unsigned char const*& pEndRowSrcY)
{
	// Ensure that the source image has valid pointers
	assert(srcImage.pY != NULL);
	assert(srcImage.pU != NULL);
	assert(srcImage.pV != NULL);

	// Get pointers to the start of this line in the source frame
	//  - source frame is YUV 4:2:0
	pSrcY = srcImage.pY + srcY * srcImage.pitch + cpParams.srcStartX;
	pSrcU = srcImage.pU
			+ (unsigned int)(srcY / 2) * (unsigned int)(srcImage.pitch / 2)
			+ (unsigned int)(cpParams.srcStartX / 2);
	pSrcV = srcImage.pV
			+ (unsigned int)(srcY / 2) * (unsigned int)(srcImage.pitch / 2)
			+ (unsigned int)(cpParams.srcStartX / 2);

	// Get a pointer to the start of line in the alpha mask, if required
	if (srcImage.pA != NULL)
		pSrcA = srcImage.pA + srcY * srcImage.pitch + cpParams.srcStartX;
	else
		pSrcA = NULL;

	// Get a pointer to the end of the srcY row - terminate for loop
	pEndRowSrcY = srcImage.pY + srcY * srcImage.pitch + cpParams.copyWidth
			+ cpParams.srcStartX;
} /* SetUpSrcYUVpointers */


// =============================================================================

//
// ConvertYUVtripleToRGB:
//
// Convert the given YUV triple to RGB.  The YUV values are assumed to all lie
// in the range 0 .. 255 and the RGB values will be returned in this range as
// well.
//
// The following conversion matrix is used:
//		R = Y + 1.40200 * U
//		G = Y - 0.71414 * U - 0.34414 * V
//		B = Y + 1.77200 * V
//
// For the purpose of speed, the arithmetic is all done using integers, with
// values right shifted by 16 bits.
//
// For some reason beyond my comprehension, this function is not compiled inline,
// even when the \Ob2 compiler option is given.  Thus, a #define style macro
// is used instead. (about 20% faster).
//
/*

#define ConvertYUVtripleToRGB(Y, U, V, R, G, B) \
	/* U and V must be adjusted to lie in the range -128 .. 127   \
	int tempU = U - 128; \
	int tempV = V - 128; \
\
	/* Convert YUV to RGB  \
	int tempY = Y << 16; \
	R = tempY +  91881 * tempV; \
	G = tempY -  46802 * tempV - 22554 * tempU; \
	B = tempY + 116130 * tempU; \
\
	/* Scale down the values and clamp values to the range 0 .. 255  \
	if (R < 0x10000) R = 0; else if (R > 0xff0000) R = 255; else R >>= 16; \
	if (G < 0x10000) G = 0; else if (G > 0xff0000) G = 255; else G >>= 16; \
	if (B < 0x10000) B = 0; else if (B > 0xff0000) B = 255; else B >>= 16
*/
// STEFANO LUT 
#define ConvertYUVtripleToRGB(y, u, v, r, g, b) \
  int tempY =y;\
  tempY <<= 16; \
  r = Clip((tempY              +  91881 * v - 11760828) >> 16); \
  g = Clip((tempY -  22554 * u -  46802 * v +  8877496) >> 16); \
  b = Clip((tempY + 116130 * u              - 14864613) >> 16);

inline void XX_ConvertYUVtripleToRGB (int Y, int U, int V, int& R, int& G, int& B)
{
	// U and V must be adjusted to lie in the range -128 .. 127
	U -= 128;
	V -= 128;

	// Convert YUV to RGB
	Y <<= 16;
	R = Y +  91881 * V;
	G = Y -  46802 * V - 22554 * U;
	B = Y + 116130 * U;

	// Scale down the values and clamp values to the range 0 .. 255
	if (R < 0x10000) R = 0; else if (R > 0xff0000) R = 255; else R >>= 16;
	if (G < 0x10000) G = 0; else if (G > 0xff0000) G = 255; else G >>= 16;
	if (B < 0x10000) B = 0; else if (B > 0xff0000) B = 255; else B >>= 16;
} /* ConvertYUVtripleToRGB */


// ===============================================================================

//
// CopyYUVToRGB555
//
// Copy the given Y, U and V bands to the given surface, in RGB555 format, using
// the alpha mask to determine which pixels are copied.  If the alpha mask is NULL,
// then the entire image is used.
//
// The top-left corner of the given image is copied to the point given by
// (top,left), where (0,0) is the top-left corner of the given surface.
//

void CopyYUVToRGB555(SrcYUVAImage const& srcImage,
					 SurfaceDesc const& surfaceDesc)
{
	CopyParameters cpParams;
	if (SetUpCopyParameters(cpParams, srcImage, surfaceDesc))
		return;

	// For every row of pixels...
	for (unsigned int srcY = cpParams.srcStartY; srcY < cpParams.srcStartY + cpParams.copyHeight; srcY++)
	{
		// Get pointers to the start of this row in each source band,
		// and another for the end of the Y row.
		unsigned char const *pSrcY, *pSrcU, *pSrcV, *pSrcA, *pEndRowSrcY;
		SetUpSrcYUVpointers(srcImage, cpParams, srcY, pSrcY, pSrcU, pSrcV,
							pSrcA, pEndRowSrcY);

		// Get a pointer to the start of this line in the destination frame
		unsigned short* pDest = (unsigned short*)((unsigned char*)((unsigned char*)surfaceDesc.lpSurface + (cpParams.destStartY + srcY - cpParams.srcStartY) * surfaceDesc.lPitch
								+ 2 * cpParams.destStartX));

		// Copy across the line of data, performing alpha masking if required
		unsigned int pixelCount = cpParams.srcStartX;
		while (pSrcY != pEndRowSrcY)
		{
			// Get the alpha mask value
			int alpha = (pSrcA == NULL) ? 255 : *pSrcA++;

			// This means that I can divide by 256 not 255
			if (alpha > 0)
				alpha++;	

			// Convert this YUV triple to RGB
			int red, green, blue, y;
			y = *pSrcY; pSrcY++;
			ConvertYUVtripleToRGB (y, *pSrcU, *pSrcV, red, green, blue);

			if (alpha < 256)
			{
				// Get the RGB values underneath this one
				int oldRed   = (*pDest & 0x7c00) >> 7;
				int oldGreen = (*pDest & 0x03e0) >> 2;
				int oldBlue  = (*pDest & 0x001f) << 3;

				// Apply the alpha mask factor
				ApplyAlphaMaskToRGB(red, green, blue, alpha, oldRed, oldGreen, oldBlue);
			}

			// Combine the values, 5 bits red, 5 bits green, 5 bits blue
			*pDest++ = MAKE_RGB555_PIXEL(red, green, blue);

			// Only advance the U and V pointers every second time
			if ((pixelCount % 2) == 1)
			{
				pSrcU++;
				pSrcV++;
			}
			pixelCount++;
		}
	}
} /* CopyYUVToRGB555 */


// ===============================================================================

//
// CopyYUVToRGB565
//
// Copy the given Y, U and V bands to the given surface, in RGB565 format, using
// the alpha mask to determine which pixels are copied.  If the alpha mask is NULL,
// then the entire image is used.
//
// The top-left corner of the given image is copied to the point given by
// (top,left), where (0,0) is the top-left corner of the given surface.
//

void CopyYUVToRGB565(SrcYUVAImage const& srcImage,
					 SurfaceDesc const& surfaceDesc)
{
	

	CopyParameters cpParams;
	if (SetUpCopyParameters(cpParams, srcImage, surfaceDesc))
		return;

	// For every row of pixels...
	for (unsigned int srcY = cpParams.srcStartY; srcY < cpParams.srcStartY + cpParams.copyHeight; srcY++)
	{
		// Get pointers to the start of this row in each source band,
		// and another for the end of the Y row.
		unsigned char const *pSrcY, *pSrcU, *pSrcV, *pSrcA, *pEndRowSrcY;
		SetUpSrcYUVpointers(srcImage, cpParams, srcY, pSrcY, pSrcU, pSrcV,
							pSrcA, pEndRowSrcY);
        
		// Get a pointer to the start of this line in the destination frame
		unsigned short* pDest = (unsigned short*)((unsigned char*)((unsigned char*)surfaceDesc.lpSurface + (cpParams.destStartY + srcY - cpParams.srcStartY) * surfaceDesc.lPitch
								+ 2 * cpParams.destStartX));

		// Copy across the line of data, performing alpha masking if required
		unsigned int pixelCount = cpParams.srcStartX;
		while (pSrcY != pEndRowSrcY)
		{
			// Get the alpha mask value
			int alpha = (pSrcA == NULL) ? 255 : (*pSrcA++);

			// This means that I can divide by 256 not 255
			if (alpha > 0)
				alpha++;	

			// Convert this YUV triple to RGB
			int red, green, blue;
			
			ConvertYUVtripleToRGB (*pSrcY++, *pSrcU, *pSrcV, red, green, blue);

			if (alpha < 256)
			{
				// Get the RGB values underneath this one
				int oldRed   = (*pDest & 0xf800) >> 8;
				int oldGreen = (*pDest & 0x07e0) >> 3;
				int oldBlue  = (*pDest & 0x001f) << 3;

				// Apply the alpha mask factor
				ApplyAlphaMaskToRGB(red, green, blue, alpha, oldRed, oldGreen, oldBlue);
			}

			// Combine the values, 5 bits red, 6 bits green, 5 bits blue
			*pDest++ = MAKE_RGB565_PIXEL(red, green, blue);
//			*pDest++ = MAKE_RGB565_PIXEL(red, green, blue);
			
			// Only advance the U and V pointers every second time
			if ((pixelCount % 2) == 1)
			{
				pSrcU++;
				pSrcV++;
			}
			pixelCount++;
		}
	}
//((	LPDIRECTDRAWSURFACE )(surfaceDesc.lpSurface))->Blt(NULL,(	LPDIRECTDRAWSURFACE )surfaceDescThird.lpSurface ,NULL,0,NULL);
} /* CopyYUVToRGB565 */


// ===============================================================================

//
// CopyYUVToRGB24bpp
//
// Copy the given Y, U and V bands to the given surface, in RGB24bpp format, using
// the alpha mask to determine which pixels are copied.  If the alpha mask is NULL,
// then the entire image is used.
//
// The top-left corner of the given image is copied to the point given by
// (top,left), where (0,0) is the top-left corner of the given surface.
//

void CopyYUVToRGB24bpp(SrcYUVAImage const& srcImage,
					   SurfaceDesc const& surfaceDesc)
{
	CopyParameters cpParams;
	if (SetUpCopyParameters(cpParams, srcImage, surfaceDesc))
		return;

	// For every row of pixels...
	for (unsigned int srcY = cpParams.srcStartY; srcY < cpParams.srcStartY + cpParams.copyHeight; srcY++)
	{
		// Get pointers to the start of this row in each source band,
		// and another for the end of the Y row.
		unsigned char const *pSrcY, *pSrcU, *pSrcV, *pSrcA, *pEndRowSrcY;
		SetUpSrcYUVpointers(srcImage, cpParams, srcY, pSrcY, pSrcU, pSrcV,
							pSrcA, pEndRowSrcY);

		// Get a pointer to the start of this line in the destination frame
		unsigned char* pDest = (unsigned char*)surfaceDesc.lpSurface
								+ (cpParams.destStartY + srcY - cpParams.srcStartY) * surfaceDesc.lPitch
								+ 3 * cpParams.destStartX;

		// Copy across the line of data, performing alpha masking if required
		unsigned int pixelCount = cpParams.srcStartX;
		while (pSrcY != pEndRowSrcY)
		{
			// Get the alpha mask value
			int alpha = (pSrcA == NULL) ? 255 : *pSrcA++;

			// This means that I can divide by 256 not 255
			if (alpha > 0)
				alpha++;	

			// Convert this YUV triple to RGB
			int red, green, blue, y;
			y = *pSrcY; pSrcY++;
			ConvertYUVtripleToRGB (y, *pSrcU, *pSrcV, red, green, blue);

			if (alpha < 256)
			{
				// Get the RGB values underneath this one
				int oldRed   = pDest[2];
				int oldGreen = pDest[1];
				int oldBlue  = pDest[0];

				// Apply the alpha mask factor
				ApplyAlphaMaskToRGB(red, green, blue, alpha, oldRed, oldGreen, oldBlue);
			}

			// Store the values in the output frame store
			*pDest++ = (unsigned char) blue;
			*pDest++ = (unsigned char) green;
			*pDest++ = (unsigned char) red;

			// Only advance the U and V pointers every second time
			if ((pixelCount % 2) == 1)
			{
				pSrcU++;
				pSrcV++;
			}
			pixelCount++;
		}
	}
} /* CopyYUVToRGB24bpp */


// ===============================================================================

//
// CopyYUVToRGB32bpp
//
// Copy the given Y, U and V bands to the given surface, in RGB32bpp format, using
// the alpha mask to determine which pixels are copied.  If the alpha mask is NULL,
// then the entire image is used.
//
// The top-left corner of the given image is copied to the point given by
// (top,left), where (0,0) is the top-left corner of the given surface.
//

void CopyYUVToRGB32bpp(SrcYUVAImage const& srcImage,
					   SurfaceDesc const& surfaceDesc)
{
	CopyParameters cpParams;
	if (SetUpCopyParameters(cpParams, srcImage, surfaceDesc))
		return;

	// For every row of pixels...
	for (unsigned int srcY = cpParams.srcStartY; srcY < cpParams.srcStartY + cpParams.copyHeight; srcY++)
	{
		// Get pointers to the start of this row in each source band,
		// and another for the end of the Y row.
		unsigned char const *pSrcY, *pSrcU, *pSrcV, *pSrcA, *pEndRowSrcY;
		SetUpSrcYUVpointers(srcImage, cpParams, srcY, pSrcY, pSrcU, pSrcV,
							pSrcA, pEndRowSrcY);

		// Get a pointer to the start of this line in the destination frame
		unsigned long* pDest = (unsigned long*)((unsigned char*)((unsigned char*)surfaceDesc.lpSurface + (cpParams.destStartY + srcY - cpParams.srcStartY) * surfaceDesc.lPitch
								+ 4 * cpParams.destStartX));

		// Copy across the line of data, performing alpha masking if required
		unsigned int pixelCount = cpParams.srcStartX;
		while (pSrcY != pEndRowSrcY)
		{
			// Get the alpha mask value
			int alpha = (pSrcA == NULL) ? 255 : *pSrcA++;

			// This means that I can divide by 256 not 255
			if (alpha > 0)
				alpha++;	

			// Convert this YUV triple to RGB
			int red, green, blue, y;
			y = *pSrcY; pSrcY++;
			ConvertYUVtripleToRGB (y, *pSrcU, *pSrcV, red, green, blue);

			if (alpha < 256)
			{
				// Get the RGB values underneath this one
				int oldRed   = (*pDest & 0xff0000) >> 16;
				int oldGreen = (*pDest & 0xff00) >> 8;
				int oldBlue  = (*pDest & 0xff);

				// Apply the alpha mask factor
				ApplyAlphaMaskToRGB(red, green, blue, alpha, oldRed, oldGreen, oldBlue);
			}

			// Store the values in the output frame store
			*pDest++ = MAKE_RGB32bpp_PIXEL(red, green, blue);

			// Only advance the U and V pointers every second time
			if ((pixelCount % 2) == 1)
			{
				pSrcU++;
				pSrcV++;
			}
			pixelCount++;
		}
	}
} /* CopyYUVToRGB32bpp */


// ========================================================================

//
// CopyYUVToBGR32bpp
//
// Copy the given Y, U and V bands to the given surface, in BGR32bpp format, using
// the alpha mask to determine which pixels are copied.  If the alpha mask is NULL,
// then the entire image is used.
//
// The top-left corner of the given image is copied to the point given by
// (top,left), where (0,0) is the top-left corner of the given surface.
//

void CopyYUVToBGR32bpp(SrcYUVAImage const& srcImage,
					   SurfaceDesc const& surfaceDesc)
{
	CopyParameters cpParams;
	if (SetUpCopyParameters(cpParams, srcImage, surfaceDesc))
		return;

	// For every row of pixels...
	for (unsigned int srcY = cpParams.srcStartY; srcY < cpParams.srcStartY + cpParams.copyHeight; srcY++)
	{
		// Get pointers to the start of this row in each source band,
		// and another for the end of the Y row.
		unsigned char const *pSrcY, *pSrcU, *pSrcV, *pSrcA, *pEndRowSrcY;
		SetUpSrcYUVpointers(srcImage, cpParams, srcY, pSrcY, pSrcU, pSrcV,
							pSrcA, pEndRowSrcY);

		// Get a pointer to the start of this line in the destination frame
		unsigned long* pDest = (unsigned long*)((unsigned char*)((unsigned char*)surfaceDesc.lpSurface + (cpParams.destStartY + srcY - cpParams.srcStartY) * surfaceDesc.lPitch
								+ 4 * cpParams.destStartX));

		// Copy across the line of data, performing alpha masking if required
		unsigned int pixelCount = cpParams.srcStartX;
		while (pSrcY != pEndRowSrcY)
		{
			// Get the alpha mask value
			int alpha = (pSrcA == NULL) ? 255 : *pSrcA++;

			// This means that I can divide by 256 not 255
			if (alpha > 0)
				alpha++;	

			// Convert this YUV triple to RGB
			int red, green, blue, y;
			y = *pSrcY; pSrcY++;
			ConvertYUVtripleToRGB (y, *pSrcU, *pSrcV, red, green, blue);

			if (alpha < 256)
			{
				// Get the RGB values underneath this one
				int oldRed   = (*pDest & 0xff);
				int oldGreen = (*pDest & 0xff00) >> 8;
				int oldBlue  = (*pDest & 0xff0000) >> 16;

				// Apply the alpha mask factor
				ApplyAlphaMaskToRGB(red, green, blue, alpha, oldRed, oldGreen, oldBlue);
			}

			// Store the values in the output frame store
			*pDest++ = MAKE_BGR32bpp_PIXEL(red, green, blue);

			// Only advance the U and V pointers every second time
			if ((pixelCount % 2) == 1)
			{
				pSrcU++;
				pSrcV++;
			}
			pixelCount++;
		}
	}
} /* CopyYUVToBGR32bpp */


// ========================================================================
