//This software module was originally developed by CSELT in the course of   //
//development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.             //
//This software module is an implementation of a part of one or more MPEG-4 //
//Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems        //
//(ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 Systems     //
//(ISO/IEC 14496-1) free license to this software module or modifications   //
//thereof for use in hardware or software products claiming conformance to  //
//the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use this software//
//module in hardware or software products are advised that its use may      //
//infringe existing patents.                                                //
//The original developer of this software module and his/her company, the   //
//subsequent editors and their companies, and ISO/IEC have no liability     //
//for use of this software module or modifications thereof in an            //
//implementation. Copyright is not released for non MPEG-4 Systems          //
//(ISO/IEC 14496-1) conforming products. CSELT S.pA. retains full right     //
//to use the code for his/her own purpose, assign or donate the code to a   //
//third party and to inhibit third parties from using the code for non      //
//MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright      //
//notice must be included in all copies or derivative works.                //
//Copyright  1999.                                                         //
//This work has been partly developed within the framework of the ESPRIT    //
//European Research Project  "MPEG-4 PC".                                   //
//////////////////////////////////////////////////////////////////////////////
// File: FillRGB.cpp
//
// Author: Matthew Leditschke 
//
// Modified: Gianluca Di Cagno  gianluca.dicagno@cselt.it
//
// Description:
//   The implementation of the functions which fill regions of an
//   offscreen RGB surface in the visual renderer.
//
// Created: October, 1997
//
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "VisRend.h"
#include "Copy.h"
#include "BRect.h"


// ===============================================================================

//
// FillRGB555
//
// Copy the given rectangular region of the given surface, in RGB555
// format, with the colour provided, applying the given transparency
// value.
//
// The coordinates of the given rectangle have the origin at the top-left
// corner of the given surface.
//

void FillRGB555(BRect const& drawBRect, SurfaceDesc const& surfaceDesc,
				Color3f const& pixelColor, float transparency)
{
	// Determine the colour to fill
	int newRed   = 255 * pixelColor.r + 0.5;
	int newGreen = 255 * pixelColor.g + 0.5;
	int newBlue  = 255 * pixelColor.b + 0.5;
	unsigned short newPixel = MAKE_RGB555_PIXEL(newRed, newGreen, newBlue);

	// For every row of pixels...
	for (int y = drawBRect.top; y < drawBRect.top + drawBRect.height; y++)
	{
		// Get a pointer to the start of this line in the destination frame
		unsigned short* pDest = (unsigned short*)((unsigned char*)((unsigned char*)surfaceDesc.lpSurface + y * surfaceDesc.lPitch
								+ 2 * drawBRect.left));

		// If there is no transparency, simply set the new values
		if (transparency < ALMOST_ZERO)
		{
			for (int x = 0; x < drawBRect.width; x++)
				pDest[x] = newPixel;
		}

		// Otherwise apply the transparency value
		else
		{
			for (int x = 0; x < drawBRect.width; x++)
			{
				int red   = newRed;
				int green = newGreen;
				int blue  = newBlue;

				// Get the RGB values underneath this one
				int oldRed   = (*pDest & 0x7c00) >> 7;
				int oldGreen = (*pDest & 0x03e0) >> 2;
				int oldBlue  = (*pDest & 0x001f) << 3;

				// Apply the transparency
				ApplyTransparencyToRGB(red, green, blue, transparency,
									   oldRed, oldGreen, oldBlue);

				// Combine the values, 5 bits red, 5 bits green and 5 bits blue
				*pDest++ = MAKE_RGB555_PIXEL(red, green, blue);
			}
		}
	}
} /* FillRGB555 */


// ===============================================================================

//
// FillRGB565
//
// Copy the given rectangular region of the given surface, in RGB565
// format, with the colour provided, applying the given transparency
// value.
//
// The coordinates of the given rectangle have the origin at the top-left
// corner of the given surface.
//

void FillRGB565(BRect const& drawBRect, SurfaceDesc const& surfaceDesc,
				Color3f const& pixelColor, float transparency)
{
	// Determine the colour to fill
	int newRed   = 255 * pixelColor.r + 0.5;
	int newGreen = 255 * pixelColor.g + 0.5;
	int newBlue  = 255 * pixelColor.b + 0.5;
	unsigned short newPixel = MAKE_RGB565_PIXEL(newRed, newGreen, newBlue);

	// For every row of pixels...
	for (int y = drawBRect.top; y < drawBRect.top + drawBRect.height; y++)
	{
		// Get a pointer to the start of this line in the destination frame
		unsigned short* pDest = (unsigned short*)((unsigned char*)((unsigned char*)surfaceDesc.lpSurface + y * surfaceDesc.lPitch
								+ 2 * drawBRect.left));

		// If there is no transparency, simply set the new values
		if (transparency < ALMOST_ZERO)
		{
			for (int x = 0; x < drawBRect.width; x++)
				pDest[x] = newPixel;
		}

		// Otherwise apply the transparency value
		else
		{
			for (int x = 0; x < drawBRect.width; x++)
			{
				int red   = newRed;
				int green = newGreen;
				int blue  = newBlue;

				// Get the RGB values underneath this one
				int oldRed   = (pDest[x] & 0xf800) >> 8;
				int oldGreen = (pDest[x] & 0x07e0) >> 3;
				int oldBlue  = (pDest[x] & 0x001f) << 3;

				// Apply the transparency
				ApplyTransparencyToRGB(red, green, blue, transparency,
									   oldRed, oldGreen, oldBlue);

				// Combine the values, 5 bits red, 6 bits green and 5 bits blue
				pDest[x] = MAKE_RGB565_PIXEL(red, green, blue);
			}
		}
	}
} /* FillRGB565 */


// ===============================================================================

//
// FillRGB24bpp
//
// Copy the given rectangular region of the given surface, in RGB24bpp
// format, with the colour provided, applying the given transparency
// value.
//
// The coordinates of the given rectangle have the origin at the top-left
// corner of the given surface.
//

void FillRGB24bpp(BRect const& drawBRect, SurfaceDesc const& surfaceDesc,
				  Color3f const& pixelColor, float transparency)
{
	// Determine the colour to fill
	int newRed   = 255 * pixelColor.r + 0.5;
	int newGreen = 255 * pixelColor.g + 0.5;
	int newBlue  = 255 * pixelColor.b + 0.5;
	unsigned char newPixelR = (unsigned char) newRed;
	unsigned char newPixelG = (unsigned char) newGreen;
	unsigned char newPixelB = (unsigned char) newBlue;

	// For every row of pixels...
	for (int y = drawBRect.top; y < drawBRect.top + drawBRect.height; y++)
	{
		// Get a pointer to the start of this line in the destination frame
		unsigned char* pDest = (unsigned char*)surfaceDesc.lpSurface
								+ y * surfaceDesc.lPitch
								+ 3 * drawBRect.left;

		// If there is no transparency, simply set the new values
		if (transparency < ALMOST_ZERO)
		{
			for (int x = 0; x < drawBRect.width; x++)
			{
				*pDest++ = newPixelB;
				*pDest++ = newPixelG;
				*pDest++ = newPixelR;
			}
		}

		// Otherwise apply the transparency value
		else
		{
			for (int x = 0; x < drawBRect.width; x++)
			{
				int red   = newRed;
				int green = newGreen;
				int blue  = newBlue;

				// Get the RGB values underneath this one
				int oldRed   = pDest[2];
				int oldGreen = pDest[1];
				int oldBlue  = pDest[0];

				// Apply the transparency
				ApplyTransparencyToRGB(red, green, blue, transparency,
									   oldRed, oldGreen, oldBlue);

				// Store the 8 bit values for blue, green and red
				*pDest++ = (unsigned char) blue;
				*pDest++ = (unsigned char) green;
				*pDest++ = (unsigned char) red;
			}
		}
	}
} /* FillRGB24bpp */


// ===============================================================================

//
// FillRGB32bpp
//
// Copy the given rectangular region of the given surface, in RGB32bpp
// format, with the colour provided, applying the given transparency
// value.
//
// The coordinates of the given rectangle have the origin at the top-left
// corner of the given surface.
//

void FillRGB32bpp(BRect const& drawBRect, SurfaceDesc const& surfaceDesc,
				  Color3f const& pixelColor, float transparency)
{
	// Determine the colour to fill
	int newRed   = 255 * pixelColor.r + 0.5;
	int newGreen = 255 * pixelColor.g + 0.5;
	int newBlue  = 255 * pixelColor.b + 0.5;
	unsigned long newPixel = MAKE_RGB32bpp_PIXEL(newRed, newGreen, newBlue);

	// For every row of pixels...
	for (int y = drawBRect.top; y < drawBRect.top + drawBRect.height; y++)
	{
		// Get a pointer to the start of this line in the destination frame
		unsigned long* pDest = (unsigned long*)((unsigned char*)((unsigned char*)surfaceDesc.lpSurface + y * surfaceDesc.lPitch
								+ 4 * drawBRect.left));

		// If there is no transparency, simply set the new values
		if (transparency < ALMOST_ZERO)
		{
			for (int x = 0; x < drawBRect.width; x++)
				pDest[x] = newPixel;
		}

		// Otherwise apply the transparency value
		else
		{
			for (int x = 0; x < drawBRect.width; x++)
			{
				int red   = newRed;
				int green = newGreen;
				int blue  = newBlue;

				// Get the RGB values underneath this one
				int oldRed   = (*pDest & 0xff0000) >> 16;
				int oldGreen = (*pDest & 0xff00) >> 8;
				int oldBlue  = (*pDest & 0xff);

				// Apply the transparency
				ApplyTransparencyToRGB(red, green, blue, transparency,
									   oldRed, oldGreen, oldBlue);

				// Combine the values, 8 bits red, 8 bits green and 8 bits blue
				*pDest++ = MAKE_RGB32bpp_PIXEL(red, green, blue);
			}
		}
	}
} /* FillRGB32bpp */


// ===============================================================================

//
// FillBGR32bpp
//
// Copy the given rectangular region of the given surface, in BGR32bpp
// format, with the colour provided, applying the given transparency
// value.
//
// The coordinates of the given rectangle have the origin at the top-left
// corner of the given surface.
//

void FillBGR32bpp(BRect const& drawBRect, SurfaceDesc const& surfaceDesc,
				  Color3f const& pixelColor, float transparency)
{
	// Determine the colour to fill
	int newRed   = 255 * pixelColor.r + 0.5;
	int newGreen = 255 * pixelColor.g + 0.5;
	int newBlue  = 255 * pixelColor.b + 0.5;
	unsigned long newPixel = MAKE_BGR32bpp_PIXEL(newRed, newGreen, newBlue);

	// For every row of pixels...
	for (int y = drawBRect.top; y < drawBRect.top + drawBRect.height; y++)
	{
		// Get a pointer to the start of this line in the destination frame
		unsigned long* pDest = (unsigned long*)((unsigned char*)((unsigned char*)surfaceDesc.lpSurface + y * surfaceDesc.lPitch
								+ 4 * drawBRect.left));

		// If there is no transparency, simply set the new values
		if (transparency < ALMOST_ZERO)
		{
			for (int x = 0; x < drawBRect.width; x++)
				pDest[x] = newPixel;
		}

		// Otherwise apply the transparency value
		else
		{
			for (int x = 0; x < drawBRect.width; x++)
			{
				int red   = newRed;
				int green = newGreen;
				int blue  = newBlue;

				// Get the RGB values underneath this one
				int oldRed   = (*pDest & 0xff);
				int oldGreen = (*pDest & 0xff00) >> 8;
				int oldBlue  = (*pDest & 0xff0000) >> 16;

				// Apply the transparency
				ApplyTransparencyToRGB(red, green, blue, transparency,
									   oldRed, oldGreen, oldBlue);

				// Combine the values, 8 bits red, 8 bits green and 8 bits blue
				*pDest++ = MAKE_BGR32bpp_PIXEL(red, green, blue);
			}
		}
	}
} /* FillBGR32bpp */


// ===============================================================================
