//This software module was originally developed by CSELT in the course of   //
//development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.             //
//This software module is an implementation of a part of one or more MPEG-4 //
//Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems        //
//(ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 Systems     //
//(ISO/IEC 14496-1) free license to this software module or modifications   //
//thereof for use in hardware or software products claiming conformance to  //
//the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use this software//
//module in hardware or software products are advised that its use may      //
//infringe existing patents.                                                //
//The original developer of this software module and his/her company, the   //
//subsequent editors and their companies, and ISO/IEC have no liability     //
//for use of this software module or modifications thereof in an            //
//implementation. Copyright is not released for non MPEG-4 Systems          //
//(ISO/IEC 14496-1) conforming products. CSELT S.pA. retains full right     //
//to use the code for his/her own purpose, assign or donate the code to a   //
//third party and to inhibit third parties from using the code for non      //
//MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright      //
//notice must be included in all copies or derivative works.                //
//Copyright  1999.                                                         //
//This work has been partly developed within the framework of the ESPRIT    //
//European Research Project  "MPEG-4 PC".                                   //
//////////////////////////////////////////////////////////////////////////////
//
// File:  PointSet2DProxy.cpp
//
// Author: Antonio Arena, CSELT, arena@kim.cselt.it
//
// Modified:
//
// Description:
//   The implementation of the proxy object used by the presenter
//   to handle the PointSet2DProxy node.
//
// Created: September, 1997.
//
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-CSELT.h"
#include "PointSet2DProxy.h"
#include "Coordinate2DProxy.h"
#include "ColorProxy.h"

// =====================================================================

//
// Presenter::InitPointSet2D
//
// Initialise the proxy data structure and attach it to the render data
// pointer of the given media object node.
//

PointSet2DProxy* Presenter::InitPointSet2D(PointSet2D* pObj)
{
	return new PointSet2DProxy(pObj);   
}


// ======================================================================

//
// PointSet2DProxy::PointSet2DProxy
//
// Class constructor.
//

PointSet2DProxy::PointSet2DProxy(MediaObject* pMO) : GeometryNode(pMO)
{
	m_lpPoints			= NULL;
	m_nAllocatedLength	= 0;
} /* PointSet2DProxy::PointSet2DProxy */


// ======================================================================

//
// PointSet2DProxy::~PointSet2DProxy
//
// Class destructor .
//

PointSet2DProxy::~PointSet2DProxy()
{
	if (m_lpPoints != NULL)		delete [] m_lpPoints;
	
} /* PointSet2DProxy::~PointSet2DProxy */



// =====================================================================

//
// PointSet2DProxy::PreRender
//
// Calculate the bounding rectangle for the set of points, which involves
// calculating where each of the points have been transformed to.
//

void PointSet2DProxy::PreRender(Effects & effects)
{
	DrawableContext *pContext=InitPreRender(effects);
	
	if (!pContext->IsSwitchedOn()) 
		return;

	if(effects.m_ParentGroup)
		effects.m_ParentGroup->AddToDrawableContextList(pContext);

	// If the transparency for this node is 1, then it doesn't need
	// to be drawn, and its bounding rectangle is empty
	if (GetTransparency(pContext) > 1 - ALMOST_ZERO)
	{
		pContext->GetBoundingRect()->MakeEmpty();
		UpdateBoundingRect(pContext);
		return;
	}

	// Allocate memory for the local copy of the points (only if the
	// current array is not large enough).
	m_nLength = GetCoordinate2D()->point.GetLength();
	if (m_nLength > m_nAllocatedLength)
	{
		// Delete the previous arrays - if required
		if (m_lpPoints != NULL)		delete [] m_lpPoints;
	

		// Create the new arrays
		m_lpPoints = new PointF[m_nLength];
	
		m_nAllocatedLength = m_nLength;
	}

	// Determine the position of each of the points once the transform
	// matrix has been applied.
	for (int i = 0; i < m_nLength; i++)
	{
		m_lpPoints[i].x = GetCoordinate2D()->point[i].x; 
		m_lpPoints[i].y = GetCoordinate2D()->point[i].y;
		pContext->m_transform.Apply(m_lpPoints[i].x, m_lpPoints[i].y);
	}

	// Determine the bounding rectangle for this set of points
	pContext->GetBoundingRect()->SetDimensions(m_lpPoints, m_nLength);



	if(effects.m_ParentGroup==NULL)//necessary for postprocessing style nodes 
	UpdateBoundingRect(pContext);

	if(effects.orderHasChanged)
			pContext->m_updateRequired = TRUE;
} /* PointSet2DProxy::PreRender */


// ============================================================================

//
// PointSet2DProxy::Draw
//
// Draw the points which lie within the given clipping rectangle.
// Transparency is supported, along with either the same colour for
// all points, or a different colour for each point.
//

void PointSet2DProxy::Draw(BRect const* pClipBRect,DrawableContext * pContext)
{
	assert(pClipBRect != NULL);
	VisualRenderer* pVR = GetVisualRenderer();

	// The points to draw lie within both the clipping rectangle and the
	// point set's own bounding rectangle.
	BRect visibleRect = *pClipBRect;
	visibleRect.Intersection(pContext->GetClippedBoundingRect());
	if (visibleRect.IsEmpty())
		return;

	// If a colour for each point is given, check that there are enough
	// colours specified
	if (GetColorNode() != NULL)
		assert(GetColorNode()->color.GetLength() >= m_nLength);

	// Get the transparency to draw the points with
	float transparency = GetTransparency(pContext);

	// The diffuse colour is used if the list of colours is not given
	Color3f emissiveColor = GetEmissiveColor(pContext);

	

	// Go through the list of points, drawing those which are in the visible rect
	for (int i = 0; i < m_nLength; i++)
	{	
		int x = m_lpPoints[i].x;
		int y = m_lpPoints[i].y;
		if (visibleRect.IsPointIn(x, y))
		{
			// Determine the colour to draw this point
			Color3f const* pThisColor;
			if (GetColorNode() != NULL)
				pThisColor = &(GetColorNode()->color[i]);
			else
				pThisColor = &emissiveColor;

			pVR->DisplayPoint(x, y, pThisColor, transparency);
		}
	}
} /* PointSet2DProxy::Draw */


// ===========================================================================
