//This software module was originally developed by CSELT in the course of   //
//development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.             //
//This software module is an implementation of a part of one or more MPEG-4 //
//Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems        //
//(ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 Systems     //
//(ISO/IEC 14496-1) free license to this software module or modifications   //
//thereof for use in hardware or software products claiming conformance to  //
//the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use this software//
//module in hardware or software products are advised that its use may      //
//infringe existing patents.                                                //
//The original developer of this software module and his/her company, the   //
//subsequent editors and their companies, and ISO/IEC have no liability     //
//for use of this software module or modifications thereof in an            //
//implementation. Copyright is not released for non MPEG-4 Systems          //
//(ISO/IEC 14496-1) conforming products. CSELT S.pA. retains full right     //
//to use the code for his/her own purpose, assign or donate the code to a   //
//third party and to inhibit third parties from using the code for non      //
//MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright      //
//notice must be included in all copies or derivative works.                //
//Copyright  1999.                                                         //
//This work has been partly developed within the framework of the ESPRIT    //
//European Research Project  "MPEG-4 PC".                                   //
//////////////////////////////////////////////////////////////////////////////
// File: ThreadList-win32.cpp
//
// Author: Matthew Leditschke 
//
// Modified: Gianluca Di Cagno  gianluca.dicagno@cselt.it
//
// Description:
//   The implementation of functions and data which maintain a list of
//   threads that are currently active, only for the win32 platform.
//
// Created: September, 1997.
//
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "ztl.h"
#include <assert.h>
#include "ThreadList.h"


// This structure contains all information about a thread
class ThreadInfo
{
public:
	ThreadInfo(const char* desc, HANDLE hThread)
	{
		threadDesc = _strdup(desc);
		threadHandle = hThread;
	}

	~ThreadInfo()
	{
		free(threadDesc);
	}

	// A string describing the thread
	char* threadDesc;

	// A handle to the thread
	HANDLE threadHandle;
};


// The local storage for the array of thread information.
static ZArray<ThreadInfo*> allThreads;

// The mutex which is used to ensure that only one thread at a time
// is writing to this array
ZMutex mtxArray;


// =====================================================================

//
// AddToThreadList
//
// Add this thread to the list, given a description of the thread
// (which will be used when displaying the list of threads), and the
// thread handle.
//

void AddToThreadList(const char* threadDesc, unsigned long threadHandle)
{
//	afxDump << "Adding " << threadDesc << ", handle " << threadHandle << "\n";

	// If this thread handle is alread in the list, return
	for (int i = 0; i < allThreads.GetLength(); i++)
		if (allThreads[i]->threadHandle == (HANDLE)threadHandle)
			return;

	// Add a new structure for this thread to the array
	mtxArray.P();
	allThreads += new ThreadInfo(threadDesc, (HANDLE)threadHandle);
	mtxArray.V();
} /* AddToThreadList */


// =====================================================================

//
// EmptyThreadList
//
// Delete all of the entries in the thread list.
//

void EmptyThreadList()
{
	mtxArray.P();
	while (allThreads.GetLength() > 0)
	{
		delete allThreads[0];
		allThreads.DeleteFrom(0);
	}
	mtxArray.V();
} /* EmptyThreadList */


// =====================================================================

//
// SubtractFileTimes
//
// Subtract the second time from the first, returning the result in
// the third time.
//

static void SubtractFileTimes(LPFILETIME time1, LPFILETIME time2,
							  LPFILETIME result)
{
/*
	afxDump << "\nTime1:  " << time1->dwHighDateTime
			<< ":" << time1->dwLowDateTime << "\n"
			<< "Time2:  " << time2->dwHighDateTime
			<< ":" << time2->dwLowDateTime << "\n";
*/

	// Perform the subtraction of the low order unsigned 32 bit integers
	DWORD carry;
	if (time1->dwLowDateTime >= time2->dwLowDateTime)
	{
		result->dwLowDateTime = time1->dwLowDateTime - time2->dwLowDateTime;
		carry = 0;
	}
	else
	{
		result->dwLowDateTime = time2->dwLowDateTime - time1->dwLowDateTime;
		result->dwLowDateTime = 0xffffffff - result->dwLowDateTime;
		result->dwLowDateTime++;
		carry = 1;
	}

	// Perform the subtraction of the high order unsigned 32 bit integers
	if (time1->dwHighDateTime >= time2->dwHighDateTime + carry)
		result->dwHighDateTime = time1->dwHighDateTime - time2->dwHighDateTime - carry;
	else
	{
		result->dwHighDateTime = time2->dwHighDateTime + carry - time1->dwHighDateTime;
		result->dwHighDateTime = 0xffffffff - result->dwHighDateTime;
		result->dwHighDateTime++;
	}

//	afxDump << "Result: " << result->dwHighDateTime
//			<< ":" << result->dwLowDateTime << "\n";
} /* SubtractFileTimes */


// =====================================================================

//
// GetThreadInfo
//
// Return a pointer to a multi-line string which contains information
// about all of the threads that are running.  The string is stored
// in static memory.
//

const char** GetThreadInfo()
{
	FILETIME creationFileTime;
	FILETIME exitFileTime;
	FILETIME kernelFileTime;
	FILETIME userFileTime;
	FILETIME currentFileTime;
	FILETIME elapsedFileTime;
	static char text[1000];
	static char* linePtrs[50];
	char* nextLine = text;
	float cpuTime;
	float elapsedTime;
	float cpuLoad;


	// Ensure that the array is big enough
	assert(allThreads.GetLength() < 49);

	// Print out the headers
	sprintf(nextLine, "Thread\tTime\tLoad\tPriority");
	linePtrs[0] = nextLine;
	nextLine += strlen(nextLine) + 1;

	// Go through all of the threads that have been registered
	for (int i = 0; i < allThreads.GetLength();)
	{
//		afxDump << allThreads[i]->threadDesc << ": ";

		BOOL gotInfo = GetThreadTimes(allThreads[i]->threadHandle,
									  &creationFileTime, &exitFileTime,
									  &kernelFileTime, &userFileTime);
		if (!gotInfo)
		{
//			afxDump << "Couldn't get thread times.\n";
			// If there was an error getting the information on this
			// thread, assume that the thread has finished, and so
			// remove it from the list.
			delete allThreads[i];
			allThreads.DeleteFrom(i);
			continue;
		}

		// Determine the amount of CPU time (sum of kernel and user time)
		SYSTEMTIME sysTime;
		FileTimeToSystemTime(&kernelFileTime, &sysTime);
		cpuTime = (float)sysTime.wMinute * 60
					+ (float)sysTime.wSecond
					+ (float)sysTime.wMilliseconds / 1000;
		FileTimeToSystemTime(&userFileTime, &sysTime);
		cpuTime += (float)sysTime.wMinute * 60
					 + (float)sysTime.wSecond
					 + (float)sysTime.wMilliseconds / 1000;

		// Determine how long this thread has been running
		GetSystemTimeAsFileTime(&currentFileTime);
		SubtractFileTimes(&currentFileTime, &creationFileTime, &elapsedFileTime);
		FileTimeToSystemTime(&elapsedFileTime, &sysTime);
		elapsedTime = (float)sysTime.wMinute * 60
						+ (float)sysTime.wSecond
						+ (float)sysTime.wMilliseconds / 1000;

		// And now determine how much of the cpu this thread has used
		cpuLoad = 100 * cpuTime / elapsedTime;

//		afxDump << "cpu time: " << allThreads[i]->cpuTime
//				<< ", elapsed: " << allThreads[i]->elapsedTime
//				<< ", load " << allThreads[i]->cpuLoad << "%\n";

		// Format up the string for this thread
		sprintf(nextLine, "%s\t%.3f\t%.0f\t%d",
				allThreads[i]->threadDesc,
				cpuTime,
				cpuLoad,
				GetThreadPriority(allThreads[i]->threadHandle));
		linePtrs[i + 1] = nextLine;
		nextLine += strlen(nextLine) + 1;

		i++;
	}
	linePtrs[i + 1] = NULL;

	return (const char**)linePtrs;
} /* GetThreadInfo */


// =====================================================================
