///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File: VisRend-CSELT-X11.cpp
//
// Author: Matthew Leditschke, CSELT, Matthew.Leditschke@cselt.it
//
// Description:
//   The implementation of the class representing a VisualRendererX11, which
//   use the X11 library to do the rendering.
//
// Created: August, 1997
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "VisRend-CSELT-X11.h"
#include "Copy.h"
#include "Caps.h"
#include "BRect.h"

#include <stdio.h>
#include "trace.h"


// ========================================================================

//
// VisualRendererX11::FindFullColorVisual
//
// Return a pointer to a full color visual on the given display specified
// by m_display, and also update the member variables of m_depth and
// m_RGBpixelFormat.
//

Visual* VisualRendererX11::FindFullColorVisual()
{
  // Look for all TrueColor visuals
  XVisualInfo vinfo;
  vinfo.c_class = TrueColor;
  int numItems;
  XVisualInfo* vinfo_ret = XGetVisualInfo(m_display, VisualClassMask, &vinfo,
					  &numItems);

  // Return NULL if none are found
  if (numItems == 0)
    return NULL;

  // Find the TrueColor visual with the greatest depth
  int maxDepth = 0;
  while (numItems > 0)
    {
      if (vinfo_ret[numItems - 1].depth > maxDepth)
	maxDepth = vinfo_ret[numItems - 1].depth;
      numItems--;
    }
  XFree(vinfo_ret);

  // If the greatest depth is less than 24, then return NULL
  if (maxDepth < 24)
    return NULL;

  // Find a TrueColor visual with this maximum depth
  if (XMatchVisualInfo(m_display, DefaultScreen(m_display), maxDepth,
		       TrueColor, &vinfo))
    {
      // Check that the bit masks match one which is supported
      if ((vinfo.red_mask == 0x00ff0000) &&
	  (vinfo.green_mask == 0x0000ff00) &&
	  (vinfo.blue_mask == 0x000000ff))
	m_RGBpixelFormat = RGB32bpp;
      else if ((vinfo.red_mask == 0x000000ff) &&
	  (vinfo.green_mask == 0x0000ff00) &&
	  (vinfo.blue_mask == 0x00ff0000))
	m_RGBpixelFormat = BGR32bpp;
      else
	{
	  fprintf(stderr, "im1-2d: Unsupported color bit masks:\n");
	  fprintf(stderr, "          red   = %08x\n",
		  (unsigned int)vinfo.red_mask);
	  fprintf(stderr, "          green = %08x\n",
		  (unsigned int)vinfo.green_mask);
	  fprintf(stderr, "          blue  = %08x\n",
		  (unsigned int)vinfo.blue_mask);
	  return NULL;
	}
      m_depth = maxDepth;
      return vinfo.visual;
    }

  return NULL;
} /* VisualRendererX11::FindFullColorVisual */


// ========================================================================

//
// VisualRendererX11::CreateFullColorWindow
//
// Create a full color window at the given x,y position with the given
// width.  Return the created window, or NULL if it can't be created.
//
// Record the depth of the visual being used in the appropriate data
// member.
//

Window VisualRendererX11::CreateFullColorWindow (int x, int y, int w, int h)
{
  XSetWindowAttributes xswa;
  unsigned int mask;
  unsigned int c_class;
  int screen;


  // Get a TureColor visual to use
  screen = XDefaultScreen(m_display);
  c_class = InputOutput;        /* Could be InputOnly */
  m_visual = FindFullColorVisual();
  if (m_visual == NULL)
    return 0;
 
  // and now create the window
  mask = CWBackPixel | CWColormap | CWBorderPixel;
  xswa.colormap = XCreateColormap(m_display, XRootWindow(m_display, screen),
				  m_visual, AllocNone);
  xswa.background_pixel = BlackPixel(m_display, DefaultScreen(m_display));
  xswa.border_pixel = WhitePixel(m_display, DefaultScreen(m_display));

  return XCreateWindow(m_display, RootWindow(m_display, screen), x, y, w, h,
		       1, m_depth, c_class, m_visual, mask, &xswa);

} /* VisualRendererX11::CreateFullColorWindow */


// ========================================================================

//
// VisualRendererX11::VisualRendererX11
//
// Class constructor - given a pointer to the presenter and the name
// of the display to open.
//
// Initialise the X11 routines.
//

VisualRendererX11::VisualRendererX11(Presenter* pPresenter,
				     const char* displayName)
{
  m_initOK = TRUE;

  // Remember the name of the display being used
  m_displayName = (char const*) displayName;

  // Remember a pointer to the presenter that this renderer is part of
  m_pPresenter = pPresenter;

  // Set the initial size of the window
  m_outputWidth  = 640;
  m_outputHeight = 480;

  // Open a connection with the server
  m_display = XOpenDisplay(m_displayName);
  if (m_display == NULL)
    {
      fprintf (stderr, "im1-2d: Error opening the display '%s'.\n",
	       m_displayName);
      m_initOK = FALSE;
      return;
    }

  // Fill in the hint structure
  XSizeHints hint;
  hint.x = 200;
  hint.y = 300;
  hint.width = 352;
  hint.height = 288;
  hint.flags = PPosition | PSize;

  // Make the window
  m_window = CreateFullColorWindow(hint.x, hint.y, hint.width, hint.height);
  if (m_window == 0)
    {
      fprintf (stderr, "im1-2d: Couldn't create a TrueColor window.\n");
      m_initOK = FALSE;
      return;
    }

  XSelectInput(m_display, m_window, StructureNotifyMask);

  // Tell other applications about this window
  char* titleStr = "IM1-2D";
  XSetStandardProperties(m_display, m_window, titleStr, titleStr, None, NULL,
			 0, &hint);

  // Map window
  XMapWindow(m_display, m_window);

  // Wait for map
  while (1)
    {
      XEvent xev;

      XNextEvent(m_display, &xev);
      if ((xev.type == MapNotify) && (xev.xmap.event == m_window))
	break;
    }

  // Specify the type of input to be received in this window
  XSelectInput(m_display, m_window, NoEventMask);

  // Make up a graphics context to draw into
  m_gc = XCreateGC(m_display, m_window, 0, 0);

  // Create the image structure to draw into
  char dummy;
  m_ximage = XCreateImage (m_display, m_visual, m_depth, ZPixmap,
			   0, &dummy, m_outputWidth, m_outputHeight, 32, 0); 

  // Allocate some memory for this image
  m_imageData = (unsigned int*) malloc(m_outputWidth * m_outputHeight * sizeof(unsigned int));
  m_ximage->data = (char*)m_imageData;

/*
  for (int y = 0; y < 288; y++)
    for (int x = 0; x < 352; x++)
      m_imageData[y * 352 + x] = ((x + y) % 256) << 8;

  // Display the image
  XPutImage(m_display, m_window, m_gc, m_ximage, 0, 0, 0, 0,
            m_outputWidth, m_outputHeight);
*/

  // Erase the output window
  if (m_initOK)
    Clear();
} /* VisualRendererX11::VisualRendererX11 */


// ========================================================================

//
// VisualRendererX11::~VisualRendererX11
//
// Class destructor - shuts down the X11 routines.
//

VisualRendererX11::~VisualRendererX11()
{
  TRACE_PRINT("In VisualRendererX11::~VisualRendererX11():\n");

  XDestroyImage(m_ximage);
  free(m_imageData);
  XFreeGC(m_display, m_gc);
  XCloseDisplay(m_display);
} /* VisualRendererX11::~VisualRendererX11 */


// ========================================================================

//
// VisualRendererX11::Init
//
// Initialise the renderer.
//

void VisualRendererX11::Init()
{
} /* VisualRendererX11::Init */


// ========================================================================

//
// VisualRendererX11::Terminate
//
// Stops rendering and performs cleanups.
// Erase the contents of the window.
//

void VisualRendererX11::Terminate()
{
  Clear();
  Paint();
} /* VisualRendererX11::Terminate */


// ========================================================================

//
// VisualRendererX11::Paint
//
// Update the display in the given rectangle of the window.  The
// bounding rectangle is given in BIFS coordinates.  If the given rectangle
// pointer is NULL, the entire window is redrawn.
//

//static int counter = 0;

void VisualRendererX11::Paint(void* pInfo)
{
  BRect* pBRect = (BRect*)pInfo;

  // Determine the rectangle to copy
  int left, top;
  unsigned int width, height;
  if (pBRect == NULL)
    {
      left = 0;
      top = 0;
      width = m_outputWidth;
      height = m_outputHeight;
    }
  else
    {
      // Convert from BIFS coords to window coords
      left   = pBRect->left + m_outputWidth / 2;
      top    = m_outputHeight / 2 - pBRect->top;
      width  = pBRect->width;
      height = pBRect->height;
    }

  // Display the image
  TRACE_PRINT4("Paint(): left = %d, top = %d, width = %d, height = %d\n",
	       left, top, width, height);

/*
  for (int y = 0; y < 288; y++)
    for (int x = 0; x < 352; x++)
      m_imageData[y * 352 + x] = ((x + y + counter) % 256) << 8;
  counter++;
*/

  // Display the image
  XPutImage(m_display, m_window, m_gc, m_ximage, 0, 0, 0, 0,
	    m_outputWidth, m_outputHeight);

  // For some reason the following doesn't work???
  //  XPutImage(m_display, m_window, m_gc, m_ximage,
  //	    left, top, left, top, width, height);

  TRACE_PRINT("In VisualRendererX11::Paint():\n");
} /* VisualRendererX11::Paint */


// ========================================================================

//
// VisualRendererX11::GetCapabilities
//
// Get the capabilities of this video renderer.
//

void VisualRendererX11::GetCapabilities (Capabilities* pCaps)
{
  pCaps->AddFormatCapability(FORMAT_RGB, 0);
  pCaps->AddFormatCapability(FORMAT_YUV, 1);
} /* VisualRendererX11::GetCapabilities */


// ========================================================================

//
// VisualRendererX11::Clear
//
// Completely erase the specifed region of the offscreen drawing
// area, but do not display the surface.  If the given rectangle
// is NULL, erase the entire offscreen area.
//

void VisualRendererX11::Clear(BRect const* pBRect)
{
  if (pBRect == NULL)
    memset(m_imageData, 0, m_outputWidth * m_outputHeight * 4);
  else
    {
      // Convert from BIFS coords to window coords
      int left   = pBRect->left + m_outputWidth / 2;
      int top    = m_outputHeight / 2 - pBRect->top;
      int bottom = top + pBRect->height;

      for (int y = top; y < bottom; y++)
	memset(&(m_imageData[y * m_outputWidth + left]), 0, pBRect->width * 4);
    }
} /* VisualRendererX11::Clear */


// ========================================================================

//
// VisualRendererX11::SetScreenSize
//
// Change the dimensions of the output window.
//

void VisualRendererX11::SetScreenSize(unsigned int newWidth,
				      unsigned int newHeight)
{
  // Only proceed if the new size is actually different
  if ((m_outputWidth != newWidth) || (m_outputHeight != newHeight))
    {
//      m_outputWidth = newWidth;
//      m_outputHeight = newHeight;

      fprintf (stderr, "VisualRendererX11::Can't do this yet!\n");
    }
} /* VisualRendererX11::SetScreenSize */


// ========================================================================

//
// VisualRendererX11::DisplayRGBAImage
//
// Display the given RGBA image, which includes where to display
// the top-left pixel in the image.  Note that the origin of the
// output display is at the bottom-left of the window.
//
// The image data itself consists of an array of pixels, where each
// pixel stored in a 32 bit unsigned integer, with bits 31-24
// representing alpha, 23-16 for red, 15-8 for green and 7-0 for blue.
// If no alpha mask information exists, the 8 alpha bits should be set
// to 255.
//

void VisualRendererX11::DisplayRGBAImage(SrcRGBAImage const& srcImage)
{
  // Change the (left,top) point from having the origin in the
  // middle of the screen to having the origin at the top-left
  // of the output window.
  SrcRGBAImage flippedSrcImage = srcImage;
  flippedSrcImage.top = m_outputHeight / 2 - flippedSrcImage.top;
  flippedSrcImage.left += m_outputWidth / 2;

  // Specify the properties of the offscreen surface
  SurfaceDesc surfaceDesc;
  surfaceDesc.dwWidth   = m_outputWidth;
  surfaceDesc.dwHeight  = m_outputHeight;
  surfaceDesc.lPitch    = m_outputWidth * 4;
  surfaceDesc.lpSurface = m_imageData;

  // Copy the image (or a portion of it)
  switch (m_RGBpixelFormat)
    {
    case RGB32bpp:
      CopyRGBAToRGB32bpp(flippedSrcImage, surfaceDesc);
      break;
    case BGR32bpp:
      CopyRGBAToBGR32bpp(flippedSrcImage, surfaceDesc);
      break;
    default:
      assert(0);	// Shouldn't get here.
      break;
    }
} /* VisualRendererX11::DisplayRGBAImage */


// ========================================================================

//
// VisualRendererX11::DisplayYUV420AImage
//
// Display the given YUV4:2:0 image, which includes where to display
// the top-left pixel in the image.  Note that the origin of the
// output display is at the bottom-left of the window.
//
// The image data itself consists of an array of pixels for each band
// in the image (ie. one U, U and V array), and another array
// for the alpha mask information.  Each pixel is stored in an
// unsigned char.  The alpha pointer can be NULL to indicate that no
// alpha mask information exists.
//

void VisualRendererX11::DisplayYUV420AImage(SrcYUVAImage const& srcImage)
{
  // The offsets must be even
  assert ((srcImage.left % 2) == 0);
  assert ((srcImage.top % 2) == 0);

  // The source width and height must be even - for now!
  assert ((srcImage.width % 2) == 0);
  assert ((srcImage.height % 2) == 0);

  // Change the (left,top) point from having the origin in the
  // middle of the screen to having the origin at the top-left
  // of the output window.
  SrcYUVAImage flippedSrcImage = srcImage;
  flippedSrcImage.top = m_outputHeight / 2 - flippedSrcImage.top;
  flippedSrcImage.left += m_outputWidth / 2;

  // Specify the properties of the offscreen surface
  SurfaceDesc surfaceDesc;
  surfaceDesc.dwWidth   = m_outputWidth;
  surfaceDesc.dwHeight  = m_outputHeight;
  surfaceDesc.lPitch    = m_outputWidth * 4;
  surfaceDesc.lpSurface = m_imageData;

  // Copy the image (or a portion of it)
  switch (m_RGBpixelFormat)
    {
    case RGB32bpp:
      CopyYUVToRGB32bpp(flippedSrcImage, surfaceDesc);
      break;
    case BGR32bpp:
      CopyYUVToBGR32bpp(flippedSrcImage, surfaceDesc);
      break;
    default:
      assert(0);	// Shouldn't get here.
      break;
    }
} /* VisualRendererX11::DisplayYUV420AImage */


// ========================================================================
