//This software module was originally developed by CSELT in the course of   //
//development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.             //
//This software module is an implementation of a part of one or more MPEG-4 //
//Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems        //
//(ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 Systems     //
//(ISO/IEC 14496-1) free license to this software module or modifications   //
//thereof for use in hardware or software products claiming conformance to  //
//the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use this software//
//module in hardware or software products are advised that its use may      //
//infringe existing patents.                                                //
//The original developer of this software module and his/her company, the   //
//subsequent editors and their companies, and ISO/IEC have no liability     //
//for use of this software module or modifications thereof in an            //
//implementation. Copyright is not released for non MPEG-4 Systems          //
//(ISO/IEC 14496-1) conforming products. CSELT S.pA. retains full right     //
//to use the code for his/her own purpose, assign or donate the code to a   //
//third party and to inhibit third parties from using the code for non      //
//MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright      //
//notice must be included in all copies or derivative works.                //
//Copyright  1999.                                                         //
//This work has been partly developed within the framework of the ESPRIT    //
//European Research Project  "MPEG-4 PC".                                   //
//////////////////////////////////////////////////////////////////////////////
//
// File: VisRend-CSELT.cpp
//
// Authors: Matthew Leditschke 
//
// Modified: Gianluca Di Cagno gianluca.dicagno@cselt.it
//
// Description:
//   The implementation of an abstract base class which performs visual
//   rendering.  Actual implementations of the visual renderer will
//   derive from this class.
//
// Created: August, 1997.
//
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "VisRend.h"

#include "GeometryNode.h"
#include "ShapeProxy.h"
#include "BRect.h"
#include "trace.h"
#include "root.h"
#include "bindable.h"
// Include the details of the Windows 95/NT specific audio renderers
#ifdef WIN32
#include "VisRend-CSELT-DD.h"
#endif

// Include the details of the SGI specific audio renderers
#if sgi
#include "VisRend-CSELT-X11.h"
#endif

// Include the details of the sun specific audio renderers
#if sun
#endif


// ======================================================================

//
// VisualRenderer::Create
//
// Return a pointer to a visual renderer.  Use this to construct
// a visual render appropriate for the current platform.  A pointer
// to the presenter, and a platform specific parameter are given.
//

VisualRenderer* VisualRenderer::Create(Presenter* pPresenter,
									   void* pInfo)
{
	VisualRenderer* pVisRend = NULL;

// Construct a Windows 95/NT specific visual renderer
#ifdef WIN32
	pVisRend = new VisualRendererDD(pPresenter, (HWND)pInfo);
#endif

// Construct an SGI specific visual renderer
#if sgi
	pVisRend = new VisualRendererX11(pPresenter, (const char*)pInfo);
#endif

// Construct a sun specific visual renderer
#if sun
	pVisRend = new VisualRendererX11(pPresenter, (const char*)pInfo);
#endif

	pVisRend->m_pPresenter = pPresenter;

	return pVisRend;
} /* VisualRenderer::Create */


// ======================================================================

//
// VisualRenderer::VisualRenderer
//
// Constructor - empty the list of drawable nodes.
//

VisualRenderer::VisualRenderer()
{
	TRACE_PRINT("Entering VisualRenderer::VisualRenderer()\n");

	m_numberOfDrawables = 0;
	m_contextArray=NULL;
	m_drawableContextSize=0;
	m_clearOnEnd = TRUE;
	BackgroundStack = new BindableStack;
	TRACE_PRINT("Leaving VisualRenderer::VisualRenderer()\n");
} /* VisualRenderer::VisualRenderer */


// ======================================================================

//
// VisualRenderer::~VisualRenderer
//
// Destructor - delete any memory associated with the list of drawable
// nodes.
//

VisualRenderer::~VisualRenderer()
{
	TRACE_PRINT("Entering VisualRenderer::~VisualRenderer()\n");

	
	
	for (unsigned int i=0;i<m_drawableContextSize;i++)
		 delete	m_contextArray[i];	
	
	delete [] m_contextArray;

	

	delete BackgroundStack; 

	TRACE_PRINT("Leaving VisualRenderer::~VisualRenderer()\n");
} /* VisualRenderer::~VisualRenderer */


// ======================================================================

//
// VisualRenderer::Init
//
// Empty the list of drawable nodes (without deleting memory).
//

void VisualRenderer::Init()
{
	TRACE_PRINT("Entering VisualRenderer::Init()\n");

	m_numberOfDrawables = 0;
	
	//context stuff
	if(m_contextArray!=NULL)
		for (unsigned int i=0;i<m_drawableContextSize;i++)
			 delete	m_contextArray[i];	
	m_drawableContextSize=0;	
	//
	
	TRACE_PRINT("Leaving VisualRenderer::Init()\n");
} /* VisualRenderer::Init */


// ======================================================================

//
// VisualRenderer::Terminate
//
// Empty the list of drawable nodes (without deleting memory).
//

void VisualRenderer::Terminate()
{
	TRACE_PRINT("Entering VisualRenderer::Terminate()\n");

	m_numberOfDrawables = 0;
	

	TRACE_PRINT("Leaving VisualRenderer::Terminate()\n");
} /* VisualRenderer::Terminate */


// ======================================================================

//
// VisualRenderer::AddToDrawList
//
// This function is called by a renderable node which needs to
// be rendered.
//
/*
void VisualRenderer::AddToDrawList(DrawableNodePtr pDrawableNode)
{
	// If the array of drawable nodes is not large enough, make it larger
	if (m_numberOfDrawables >= m_drawableArraySize)
	{
		m_drawableArraySize += 10;
		m_drawableArray = (DrawableNodePtr*)realloc(m_drawableArray,
			m_drawableArraySize * sizeof(DrawableNodePtr));
	}

	// Add the details of this presentable media object to the list
	m_drawableArray[m_numberOfDrawables] = pDrawableNode;
	m_numberOfDrawables++;
} /* VisualRenderer:: AddToDrawList */

DrawableContextPtr  VisualRenderer::GetDrawableContext()
{
	// If the array of drawable nodes is not large enough, make it larger
	if (m_numberOfDrawables >= m_drawableContextSize)
	{
		m_drawableContextSize += 50;
		m_contextArray = (DrawableContextPtr*)realloc(m_contextArray,
			m_drawableContextSize * sizeof(DrawableContextPtr));
		for (unsigned int i=m_numberOfDrawables;i<m_drawableContextSize;i++)
			m_contextArray[i]=new DrawableContext;	
	}
	m_contextArray[m_numberOfDrawables]->m_updateRequired=FALSE;
	m_contextArray[m_numberOfDrawables]->m_needsDrawing=FALSE;
	m_contextArray[m_numberOfDrawables]->m_isTransparent=FALSE;
	m_contextArray[m_numberOfDrawables]->m_hasStreamFinished=FALSE;
	m_contextArray[m_numberOfDrawables]->SetAppearance(NULL);
	m_contextArray[m_numberOfDrawables]->SetIsTextNode(FALSE);
	return m_contextArray[m_numberOfDrawables++];
	
	
	
} /* VisualRenderer:: GetDrawableContext */





// ======================================================================

//
// VisualRenderer::DrawScene
//
// Draw all of the nodes which are present in the drawable list.
// Once the nodes have been drawn, this list is emptied.
//

void VisualRenderer::DrawScene()
{
	// Only proceed if there are drawables in the list
	if (m_numberOfDrawables == 0&&m_toBeClean.IsEmpty())
		return;

	
	TRACE_PRINT1("VR: Entering DrawScene(): %d drawables.\n",
				 m_numberOfDrawables);


	// Start with two empty bounding rectangles (in BIFS coordinates)

	BRect needsDrawingRect;
	BRect transparentRect;

	// Go through the list of nodes, determine what area has changed,
	// and what area is transparent.

	unsigned int i;

	for (i = 0; i < m_numberOfDrawables; i++)
	{
		// this is useful when deleting a drawable in the ProcessEvents
		m_contextArray[i]->ResetVisitsList();

	
		// look in the previousBrect list of the drawable associated 
		// to this context if there is a previous object in the position
		// we are going to draw, and add this information to the context,
		// so that we can use it in the next NeedsDrawing test.
		// build also the clippedboundingRect which is the intersection between
		// the clipping Rect and the bounding Rect, to get more accurate 
		// dimensions of the needsDrawing Rect
		m_contextArray[i]->UpdateBoundingRect();
		// If changed, add area of this node to the changed area
		if (m_contextArray[i]->NeedsDrawing())
		{
#ifdef WRITE_TRACE
			TRACE_PRINT1("Node %d needs drawing, rect = ", i);
			m_contextArray[i]->GetClippedBoundingRect()->Print(traceFile);
			TRACE_PRINT("\n");
#endif
			
			needsDrawingRect.Union(m_contextArray[i]->GetClippedBoundingRect());
		}

		// If transparent, add area of this node to the transparent area
		if (m_contextArray[i]->IsTransparent())
		{
#ifdef WRITE_TRACE
			TRACE_PRINT1("Node %d is transparent, rect = ", i);
			m_contextArray[i]->GetClippedBoundingRect()->Print(traceFile);
#endif
			transparentRect.Union(m_contextArray[i]->GetClippedBoundingRect());
		}
	}
// This is used as a garbage collector, it adds to the needsdrawing Rect and 
// transparentRect all the  bounding boxes of the drawables that have not been 
// referenced in the previous "for".
// When a node is switched off, its current bounding box is set to NULL, and the 
// following code will add to the transparent rect (i.e to be cleaned) its 
// previous BBox.
    BRect refreshRect;
	for (i = 0; i < m_numberOfDrawables; i++)
	{
		//compares the current and previous bounding box list of
		//a drawable node
		m_contextArray[i]->BuildRefreshRect(&refreshRect);
		TRACE_PRINT("VR: Refresh Rect:");
		#ifdef WRITE_TRACE
		refreshRect.Print(traceFile);
		#endif
		TRACE_PRINT("\n");
		transparentRect.Union(&refreshRect);
		needsDrawingRect.Union(&refreshRect);
	}
//if a node is deleted, its denstructor is called and m_toBeClean is set to its BBox
//even though the drawing process it's similar to the swtich off, the object is
//not referenced any more in the scene graph so you have to use its denstructor and
//a Visual renderer clean area
	if(!m_toBeClean.IsEmpty())
 	      {
	          transparentRect.Union(m_toBeClean);
			  needsDrawingRect.Union(m_toBeClean);
			  m_toBeClean.MakeEmpty();
          }

	// Ensure that these two rectangles are not larger than the screen
	// itself
#ifdef ORIGIN_CENTER
	BRect screenBRect((int)GetScreenWidth() / -2, (int)GetScreenHeight() / 2,
					  (int)GetScreenWidth(), (int)GetScreenHeight());
#else
    
	// new screen for new valid bifs coordinates
    
	 BRect screenBRect(0,(int)GetScreenHeight(),
    					  (int)GetScreenWidth(),(int)GetScreenHeight());
#endif    

	needsDrawingRect.Intersection(screenBRect);
	transparentRect.Intersection(screenBRect);

	// Reduce the transparent region to be only that area which is
	// both transparent and needs drawing.
	transparentRect.Intersection(needsDrawingRect); //toBEchecked
    
/**/
#ifdef WRITE_TRACE
	TRACE_PRINT("VR: Needs drawing rect: ");
	needsDrawingRect.Print(traceFile);
	TRACE_PRINT("\n");
	TRACE_PRINT("VR: Transparent rect: ");
	transparentRect.Print(traceFile);
	TRACE_PRINT("\n");
#endif
/**/
	//draw background first. Get the Top from the background stack
	BOOL bDrawBackground=FALSE;

		if(!GetBackgroundStack()->IsEmpty())
		{
			
			const BRect* pBRect;

			if(GetBackgroundStack()->GetTop()->GetContext()->GetClippingRect()->IsEmpty())
				pBRect= GetBackgroundStack()->GetTop()->GetContext()->GetBoundingRect();
			else
			{
			//not implemented 
				assert(0);
				pBRect= GetBackgroundStack()->GetTop()->GetContext()->GetClippedBoundingRect();
			}

			if (( GetBackgroundStack()->GetTop()->GetContext()->NeedsDrawing()))
			{				
				// draw  the background,update all the screen
				needsDrawingRect.Union(GetBackgroundStack()->GetTop()->GetScreenBRect());
				//draw what is't on top too..
				transparentRect.Union(GetBackgroundStack()->GetTop()->GetScreenBRect());
				bDrawBackground=TRUE; 
			}
			else	
			if	(transparentRect.Overlaps(pBRect)) 
				bDrawBackground=TRUE;	
						
			
		}

	
	// Keep track of the region which is being drawn (used to
	// determine which nodes which have not changed still need to
	// be drawn).
	BRect drawnRect;
	
	// Only draw things if the changed rectangle is not empty
	if (!needsDrawingRect.IsEmpty())
	{
		
		// Erase the part of the screen which will have transparent
		// objects drawn in it.
		if (!transparentRect.IsEmpty())
		{
			TRACE_PRINT1("VR: Erasing rectangle: %s\n",
						 (char*)transparentRect);
			Clear(&transparentRect);
		
		}

		// And now draw the nodes, from the back to the front, only
		// drawing the part of the objects which lie within the changed
		// area.  Nodes which have not changed are only drawn if they
		// intersect with the transparent area, or something behind them
		// has been redrawn.
		
		drawnRect.MakeEmpty();

			
		TRACE_PRINT1("VR: Original Rect: %s\n",(char*)needsDrawingRect);
		
		if(bDrawBackground)
		{
			GetBackgroundStack()->GetTop()->Draw(&needsDrawingRect,m_contextArray[0]);
			bDrawBackground=FALSE;	
		}

			
		for (i = 0; i < m_numberOfDrawables; i++)
		{
			
			const BRect* pBRect = m_contextArray[i]->GetClippedBoundingRect();
					
			if ((m_contextArray[i]->NeedsDrawing()) ||
				(transparentRect.Overlaps(pBRect))  ||
				(drawnRect.Overlaps(pBRect)))
			{
				
			 
			
			 if(m_contextArray[i]->m_isSwitchedOn)
			 {
				 TRACE_PRINT1("VR: Drawing node %d.\n", i);
				// Here there's a problem :
				// The objects first are tested for NeedsDrawing
				// and so the needsDrawingRect is constructed.
				// THEN you test for Overlaps but the 
				// needsDrawingRectIsAlready constructed!!!!
				
				
                m_nDraw++;
				((DrawableNode *)(m_contextArray[i]->m_pDrawable))->Draw(&needsDrawingRect,m_contextArray[i]); // was :&needsDrawingRect
				BRect tmp(needsDrawingRect);
				tmp.Intersection(pBRect);
				

				//drawnRect.Union(pBRect);
				drawnRect.Union(tmp);
			 }
			}
		/*	if (transparentRect.Overlaps(pBRect)) 
					  
			{
				m_contextArray[i]->m_needsUpdate=0; 
			
			 if(m_contextArray[i]->m_isSwitchedOn)
			 {
				 TRACE_PRINT1("VR: Drawing node %d.\n", i);
			     m_nDraw++;
				 BRect inter;
				 inter.SetDimensions(*pBRect);
				 inter.Intersection(transparentRect);	
				((DrawableNode *)(m_contextArray[i]->m_pDrawable))->Draw(&inter,m_contextArray[i]); // was :&needsDrawingRect
				 drawnRect.Union(inter);

			 }
			}*/
		}

		// And now copy the offscreen drawing area to the on screen,
		// only copying the area that has actually changed.
		TRACE_PRINT1("VR: Updating rectangle: %s\n",
					 (char*)needsDrawingRect);
		TRACE_PRINT1("VR: Number of DrawnObjects: %d\n",
					 m_nDraw);
		m_nDraw=0;

		Paint(&needsDrawingRect);
	}
    
	TRACE_PRINT("VR: Leaving DrawScene()\n");
} /* VisualRenderer::DrawScene */


// ======================================================================

//
// VisualRenderer::RefreshScreen

void VisualRenderer::RefreshScreen()
{
	BRect screenBRect((int)GetScreenWidth() / -2, (int)GetScreenHeight() / 2,
					  (int)GetScreenWidth(), (int)GetScreenHeight());
	
	if((screenBRect.IsEmpty()))
		return; 

	//draw background first
	if(!GetBackgroundStack()->IsEmpty()&&(GetPresenter()->GetScenesDrawn()!=0))
		{
			
			DrawableContext context;
			const BRect* pBRect = GetBackgroundStack()->GetTop()->GetContext()->GetBoundingRect();
			BRect objectVisible = screenBRect;
			objectVisible.Intersection(pBRect);
			GetBackgroundStack()->GetTop()->Draw(&objectVisible,&context);

		}

	Clear(&screenBRect);

	for (unsigned int i = 0; i < m_numberOfDrawables; i++)
		{
			const BRect* pBRect = m_contextArray[i]->GetClippedBoundingRect();
			BRect objectVisible = screenBRect;
			objectVisible.Intersection(pBRect);
			if(m_contextArray[i]->m_isSwitchedOn)
			 {
				 TRACE_PRINT1("VR: Drawing node in Refresh Mode%d.\n", i);
				  DrawableNode * pDrawable = ((DrawableNode *)(m_contextArray[i]->m_pDrawable));           
				  if(pDrawable)
					pDrawable ->Draw(&objectVisible,m_contextArray[i]);
			 
			}
	}
}			
// ======================================================================

//
// VisualRenderer::FindNode
//
// Go through the list of drawable nodes, in front to back depth
// order (opposite to the drawing order), looking for the first which
// is under the given point (which is in BIFS coordinates).
//
// Currently the bounding rectangle is used to determine if the
// point is above the node.  This needs to be improved.
//

DrawableContext * VisualRenderer::FindNode(int x, int y)
{
	// Look through the list, starting with the front most
	for (int i = m_numberOfDrawables - 1; i >= 0; i--)
	{
	 if(m_contextArray[i]!=NULL )
	 {	 
	  if((m_contextArray[i]->IsSwitchedOn())&&(m_contextArray[i]->IsPointOver(x, y)))
	  {
          
		  return m_contextArray[i];
	  }
	 }
    }
	return NULL;
} /* VisualRenderer::FindNode */

// ======================================================================

// Emacs uses these...
// Local Variables:
// mode: c++
// tab-width: 4
// End:
