//This software module was originally developed by CSELT in the course of   //
//development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.             //
//This software module is an implementation of a part of one or more MPEG-4 //
//Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems        //
//(ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 Systems     //
//(ISO/IEC 14496-1) free license to this software module or modifications   //
//thereof for use in hardware or software products claiming conformance to  //
//the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use this software//
//module in hardware or software products are advised that its use may      //
//infringe existing patents.                                                //
//The original developer of this software module and his/her company, the   //
//subsequent editors and their companies, and ISO/IEC have no liability     //
//for use of this software module or modifications thereof in an            //
//implementation. Copyright is not released for non MPEG-4 Systems          //
//(ISO/IEC 14496-1) conforming products. CSELT S.pA. retains full right     //
//to use the code for his/her own purpose, assign or donate the code to a   //
//third party and to inhibit third parties from using the code for non      //
//MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright      //
//notice must be included in all copies or derivative works.                //
//Copyright  1999.                                                         //
//This work has been partly developed within the framework of the ESPRIT    //
//European Research Project  "MPEG-4 PC".                                   //
//////////////////////////////////////////////////////////////////////////////
//
// File: VisRend.h
//
// Authors: Matthew Leditschke 
//
// Modified: Gianluca Di Cagno  gianluca.dicagno@cselt.it
//
// Description:
//   The definition of an abstract base class which performs visual
//   rendering.  Actual implementations of the visual renderer will
//   derive from this class.
//
// Created: August, 1997
//
//
///////////////////////////////////////////////////////////////////////////


#ifndef VIS_REND_H
#define VIS_REND_H


#include "ztypes.h"
#include "types-CSELT.h"
#include "BRect.h"
#include "DrawableContext.h"
#include "ztl.h"
class Presenter;
class Capabilities;
class DrawableNode;
class BRect;
class BindableStack;

// This is the depth that an object must have to be considered the
// back-most object in the drawing order.
const double BACKGROUND_DEPTH = -999999999.0;


// =====================================================================

//
// This class contains the parameters which specify a source image which
// is to be displayed.  Other classes build on this to make actual images
// which can be displayed.
//
class SrcImage
{
public:
	SrcImage() 
	{ left = top = width = height = pitch = 0; scaleX=scaleY=1; }

	// Where to display the top-left corner of this image.  The origin
	// of the scene is in the middle of the screen, with increasing X
	// to the right, and increasing Y upwards.
	int left;
	int top;

	// The width and height of the image (in pixels)
	unsigned int width;
	unsigned int height;

	// The line pitch for this image
	// (number of bytes between start of adjacent rows)
	unsigned int pitch;

	float scaleX;
	float scaleY;
};


// =====================================================================

//
// This class specifies the parameters for an RGBA image which is
// to be displayed.  Pointers to the four bands which make up the
// image are to be specified (although the alpha band is optional).
//
class SrcRGBAImage : public SrcImage
{
public:
	SrcRGBAImage()
		{ pR = pG = pB = pA = NULL; ptrInc = 0; }

	// Pointers to the start of each band in the image
	unsigned char* pR;
	unsigned char* pG;
	unsigned char* pB;
	unsigned char* pA;	// If this is NULL, there is no alpha

	// How much to increment one of the above pointers to get
	// to the next pixel in the same band
	int ptrInc;
};


// =====================================================================

//
// This class specifies the parameters for a YUVA image which is to be
// displayed.  The Y, U, V and A bands are stored separately.
//
// The width and height parameters from SrcImage refer to the dimensions
// of the Y band.  The U and V bands are half this (rounding away from
// zero).
//
class SrcYUVAImage : public SrcImage
{
public:
	SrcYUVAImage()
	{ pY = pU = pV = pA = NULL; }

	// Pointers to the bands of the image,
	unsigned char* pY;
	unsigned char* pU;
	unsigned char* pV;
	unsigned char* pA;	// If this is NULL, there is no alpha
};


// =====================================================================

//
// The definition of a class which describes the parameters of the
// destination surface.
//

class SurfaceDesc
{
public:
	DWORD	dwHeight;
    DWORD	dwWidth;
    LONG	lPitch;
    LPVOID	lpSurface;
};


// =====================================================================

//Drawable list stuff
typedef DrawableNode* DrawableNodePtr;
typedef DrawableContext*  DrawableContextPtr;

class VisualRenderer
{
	//
	// Public functions
	//
public:
	int m_nDraw;
    BRect m_toBeClean;
	// Return a pointer to a visual renderer.  Use this to construct
	// a visual render appropriate for the current platform.  A pointer
	// to the presenter, and a platform specific parameter are given.
	static VisualRenderer* Create(Presenter* pPresenter,
								  void* pInfo = NULL);

	//
	// Functions to be overloaded by actual implementations of the
	// visual renderer.
	//

	// Initialise the renderer.  This will be called once for every file
	// that is opened.  Deriving classes must ensure that this method is
	// called.
	virtual void Init();

	// Stops rendering and performs cleanups, including terminating the
	// audio renderer thread.  This will be called once for every file
	// that is closed.  Deriving classes must ensure that this method is
	// called.
	virtual void Terminate();

	// Called by the Exective when some of the display needs to be
	// redrawn (becuase it has been uncovered).  The given argument
	// specifed the region to redraw.  The region is given in BIFS
	// coordinates.  If the given rectangle pointer is NULL, the
	// entire window is redrawn.
	virtual void Paint(void* pInfo = NULL) = 0;

	// Completely erase the specifed region of the offscreen drawing
	// area, but do not display the surface.  If the given rectangle
	// is NULL, erase the entire offscreen area.
	virtual void Clear(BRect const* pUpdateRect = NULL) = 0;

	// Get the capabilities of this visual renderer
	virtual void GetCapabilities (Capabilities* pPresentationCaps) = 0;

	// Set the output screen size to the given dimensions.
	virtual void SetScreenSize(unsigned int newX, unsigned int newY) = 0;

	// Return the current screen size.
	virtual void GetScreenSize(unsigned int& currX,
							   unsigned int& currY) const = 0;
	virtual unsigned int GetScreenWidth() const = 0;
	virtual unsigned int GetScreenHeight() const = 0;


	// Display the given RGBA image, which includes where to display
	// the top-left pixel in the image.  This uses the BIFS coordinate
	// system, with the origin of the output display is in the middle
	// of the window, with increasing X and Y to the right and up.
	virtual void DisplayRGBAImage(SrcRGBAImage const& srcImage) = 0;

	// Display the given YUV4:2:0 image, which includes where to display
	// the top-left pixel in the image.  This uses the BIFS coordinate
	// system, with the origin of the output display is in the middle
	// of the window, with increasing X and Y to the right and up.
	virtual void DisplayYUV420AImage(SrcYUVAImage const& srcImage) = 0;

	// Display the given shape, which includes where to display
	// the top-left pixel in the image.  This uses the BIFS coordinate
	// system, with the origin of the output display is in the middle
	// of the window, with increasing X and Y to the right and up.
	virtual void DisplayCircle(BRect const& pClipBRect, 
							   PointF const* lpPoints,
							   Color3f color, 
							   int width, 
							   BOOL filled) = 0;
    virtual void DisplayCurve(BRect const& pClipBRect, 
							  PointF const* lpPoints,
							  int const* lpTypes,
							  Color3f color, 
							  int width, 
							  int length) = 0;
    virtual void DisplayLines(BRect const& pClipBRect, 
							  int const& startPos1, int const& startPos2,
							  int const& endPos1, int const& endPos2,
							  Color3f const& pixelColor, 
							  int width) = 0;

	// Draw the portion of the given polygon which lies within the given
	// clipping rectangle.  The polygon is given by an array of points
	// (locations given in BFS coordinates).  The colour, line width
	// and filled flag are also given.
	virtual void DisplayPolygon(BRect const& clipBRect,
								PointF const* lpPoints, int numPoints,
								Color3f const& pixelColor,
								int width, BOOL filled) = 0;

	// Draw a rectangle, whose size and location is given by drawBRect,
	// only showing that portion which lies within the given clipping
	// rectangle.  The colour, line width, filled flag and transparency
	// are also given.  All coordinates are given in BIFS coordinates.
	virtual void DisplayRectangle(BRect const& clipBRect,
								  BRect const& drawBRect,
								  Color3f const& pixelColor,
								  int width, BOOL filled,
								  float transparency) = 0;

	// Set the specified pixel to the given colour, using the provided level
	// of transparency.  The location of the pixel to set is given in BIFS
	// coordinates, with the origin of the output display is in the middle
	// of the window, with increasing X and Y to the right and up.
	virtual void DisplayPoint(int xPos, int yPos, Color3f const* pPixelColor,
							  float transparency) = 0;

	// Return a renderer specific handle to the requested font.
	virtual void* SetFont(LPCTSTR fontName, int fontSize, BOOL bold,
						  BOOL italics, BOOL underline) = 0;

	// Stop using the font which was set using SetFont().  Every call
	// to SetFont() must be accompanied by a call to UnsetFont().
	virtual void UnsetFont(void* pFont) = 0;

	// Return the width and the height of the given text string, using
	// a font return by SetFont().
	virtual void GetTextSize(void* pFont, LPCTSTR string, int length,
							 unsigned int& width, unsigned int& height,
							 unsigned int &ascent,unsigned int &descent) = 0;

	//Set the space between the characters
	virtual void SetCharacterInterspace(void *  pFont,int space) = 0;

	virtual void PixelToLogical(void *  pFont,int& size ) = 0;


	// Display a string with the given length, at the specifed location
	// and using the given colour, clipping the output to the given
	// rectangle.  The x,y point and the clip rectangle are given in
	// BIFS coordinates.
	virtual void DisplayText(void* pFont, LPCTSTR pString, int length,
							 int x, int y, Color3f const& color,
							 BRect const* pClipBRect) = 0;

	//
	// Methods specific to the base class implementation (ie. common
	// to all implementations.
	//

	// Constructor - empty the list of drawable nodes.
	VisualRenderer();

	// Destructor - delete any memory associated with the list of
	// drawable nodes.
	virtual ~VisualRenderer();

	// Call this function before calling PreRender() of the root of
	// the scene, to ensure that the list of drawable nodes is empty.
	void ResetDrawList() { m_numberOfDrawables = 0; }

	// This function is called by a renderable node which needs to
	// be rendered.
	
	void AddToDrawList(DrawableNodePtr pDrawableNode);

	// Drawable Context list management
	DrawableContextPtr  GetDrawableContext();

	
	void ResetDrawableContextAfterRenderList() { m_nContextRendered = 0; }

	//returns the background stack
	BindableStack *	GetBackgroundStack()	{return BackgroundStack;}

		
	// Draw all of the nodes which are present in the drawable list.
	// Once the nodes have been drawn, this list is emptied.
	void DrawScene();

	void RefreshScreen();
	// Go through the list of drawable nodes, in front to back depth
	// order, looking for the first which is under the given point
	// (which is in BIFS coordinates).
	DrawableContext * FindNode(int x, int y);

	// Return TRUE if the video renderer was initialised successfully.
	BOOL InitialisedOK() { return m_initOK; }

	// Return a pointer to the Presenter
	Presenter* GetPresenter() const { return m_pPresenter; }

	// Get and set the flag which indicates if the screen is to be
	// cleared when the end of the file is reached (which happens
	// when the Terminate() method is called.
	BOOL GetClearOnEndFlag() const { return m_clearOnEnd; }
	void SetClearOnEndFlag(BOOL clearOnEnd) { m_clearOnEnd = clearOnEnd; }

	virtual void PostError(HRESULT) = 0;

	virtual void DisplayLogo() =0;
	//
	// Protected implementation details
	//
protected:

	// Remember a pointer to the presenter that this renderer is part of
	Presenter* m_pPresenter;

	// Stores if the audio renderer was initialised successfully.
	BOOL m_initOK;


	//list of contexts
	DrawableContextPtr * m_contextArray;
	
	// The allocated size of this array
	unsigned int m_drawableContextSize;

	
	// The number of items in this array
	unsigned int m_numberOfDrawables;

	unsigned int m_nContextRendered;
	// This is TRUE if the screen should be cleared when the end
	// of a file is reached.
	BOOL m_clearOnEnd;
    
	//contains the list of backgrounds that can be binded to the scene. 
	BindableStack * BackgroundStack;
};


#endif // VIS_REND_H


// Emacs uses these...
// Local Variables:
// mode: c++
// tab-width: 4
// End:
