// -*- c++ -*-

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

  SUBSYSTEM:  Decoder
  FILE:       AnimationStreamDec.h
  AUTHOR:     Mikael Bourges-Sevenier (France Telecom CNET/DIH/ATI)
  OVERVIEW:   Definition of AnimationStreamDecoder
  ---------------------------------------------------------------------------
  
    The purpose of the animation stream decoder is first, to decode the 
    animation mask sent at the beginning of an animation. It then allocates a 
    list of animated nodes. Second, for each animation frame, we call each 
    node of the list to decode itself.
    In addition, the decoder takes care of skipping frames for nodes willing 
    to do so.
    
-------------------------------------------------------------------------*/

#ifndef _ANIMATIONSTREAM_DEC_H_
#define _ANIMATIONSTREAM_DEC_H_


#include "DecoderImp.h"
#include "BifsDec.h"
#include "ODImp.h"




#define CPLUSPLUS

class MediaObject;


#ifdef ANIMATION

#include "ArithDec.h"

class AnimTime {
  int frame_rate; // the reference frame rate of the sequence
  int fr_seconds;   // the fractional reference frame rate
  bool fr_offset;    // true: the frame rate uses the NTSC frequency offset of 1000/1001
  long time_code;
  int tc_hours;  // 0-23
  int tc_minutes;// 0-59
  bool tc_mbit;   // marker bit always 1
  int tc_seconds;// 0-59
  int num_frames_skip;

  float fps; // frame per seconds

public:
  AnimTime() :
    fr_offset(FALSE),tc_mbit(TRUE)
  {
    setFrameRate(0);
    setTimeCode(0);
    setNumSkipFrames(0);
  }

  AnimTime(int frameRate, int timeCode, int framesSkip) :
    fr_offset(FALSE),tc_mbit(TRUE)
  {
    setFrameRate(frameRate);
    setTimeCode(timeCode);
    setNumSkipFrames(framesSkip);
  }

  bool hasFrameRate() { return (frame_rate != 0) ? true : false; }
  bool hasSkipFrames() { return (num_frames_skip != 0) ? true : false; }
  bool hasTimeCode() { return (time_code != 0) ? true : false; }

  void setFrameRate(int frameRate) { frame_rate = frameRate; }
  int getFrameRate() { return frame_rate; }
  int getFrameRateSeconds() { return fr_seconds; }
  void setFrameRateSeconds(int sec) { fr_seconds=sec; }
  bool getFrameRateOffset() { return fr_offset; }
  void setFrameRateOffset(bool b) { fr_offset=b; }

  void setNumSkipFrames(int framesSkip) { num_frames_skip = framesSkip; }
  int getNumSkipFrames() { return num_frames_skip; }
  void setTimeCode(long tc_millis) 
  { 
    int l= (int) (tc_millis / 1000);
    time_code=tc_millis/1000; 
    tc_hours = l / 3600;
    tc_minutes = l / 60;
    tc_seconds = l % 60;
  }
  
  long getTimeCode() { return time_code; }
  int getTimeCodeHours() { return tc_hours; }
  void setTimeCodeHours(int h) { tc_hours=h; }
  int getTimeCodeMinutes() { return tc_minutes; }
  void setTimeCodeMinutes(int m) { tc_minutes=m; }
  int getTimeCodeSeconds() { return tc_seconds; }
  void setTimeCodeSeconds(int s) { tc_seconds=s; }
  bool getTimeCodeMarkerBit() { return tc_mbit; }
  void setTimeCodeMarkerBit(bool mb) { tc_mbit=mb; }

  float getFPS() 
  {
    fps=frame_rate+fr_seconds/16.0;
    if (fr_offset) 
      fps *= (float) (1000.0/1001.0);
    return fps;
  }  
};

#endif	//ANIMATION


class AnimationStreamDecoder : public DecoderImp
{
private:

#ifdef ANIMATION
	int readAFrame();
	int readFrameHeader();
	void readFrameData();
	bool checkAnimStartCode();
	float readFrameRate();
	void  readTimeCode();
	bool readSkipFrames();
#endif //ANIMATION
	
protected:

	int nodeIDbits, routeIDbits;
	
	BOOL isCommandStream;
	BIFSDecoder *m_pBIFSComDec;
	DecoderConfigDescriptor *m_pDecoderConfig;

	// bit manipulation variables
	unsigned char *bufptr,*startbuf; 
	int bufSize;
	unsigned int incnt;     // bit used counter in current read byte
	unsigned int inused;    // number of bits read
	unsigned int inbytecnt; // number of bytes read
	unsigned char inbfr;    // input buffer


#ifdef ANIMATION

	/* The arithmetic decoder is handled by the decoder itself.
	Each animated field has is own model passed to the AAD.   */
	ArithmeticDecoder aad;
	bool randomAccess;
	AnimTime m_aTime; // keep time stuff
	int numNodes; // nodes in mask
	ZArray<bool> nodeMask; // active nodes switch
	ZArray<MediaObject *> pNodeMask; // ptrs to active nodes
	bool isIntra; // intra or predictive frame ?
	
	//FTRD - begin - 7/10/2001 
	FILE*		m_logfile;
	int		m_nbLogBits;
	bool		m_newAqp;
	int frameID;
	//FTRD - end- 7/10/2001 
	
#endif //ANIMATION


public:
	AnimationStreamDecoder ();
	~AnimationStreamDecoder ();

	BOOL m_bStreamFinished;

	void SetConfig(DecoderConfigDescriptor *pDecoderConfig)
	{ m_pDecoderConfig = pDecoderConfig; }

	virtual void Run();
	virtual int GetOptimalOutputSize () {return 0; }
	virtual BOOL SetFormat (Capabilities &formatOptions, DecoderParams &decoderParams);
	virtual BOOL Setup (const BYTE * pSpecificInfo, int nSpecificInfoLength, int nESNumber );
    
	//decode isn't used, AnimationStream overrides the Run() method
	virtual BOOL Decode (LPBYTE /*pInput*/, int /*nInputLength*/, 
						LPBYTE /*pOutput*/, int & /*rnOutputLength*/) 
	{ return TRUE; }


	// bit manipulation methods used by both modes
	void setInputBuffer(unsigned char *AUptr);
	void setBufSize (int inSize) { bufSize = inSize; }
	void initBits();
	int readBits(int n);
	bool readBit();

	//all animation specific
#ifdef ANIMATION

	bool getRandomAccess() {return randomAccess;}
	void markNewAQP(bool state=true) {m_newAqp = state;}
	bool isNewAQP();

	ArithmeticDecoder& getAAD() { return aad; }

	unsigned char *getInputBuffer() { return bufptr; }
	int readBits(int n, char* szDesc, char* szComment=NULL,int logLevel=0);
	bool readBit(char* szDesc, char* szComment=NULL,int logLevel=0);
	float readFloat (char* szDesc, char* szComment=NULL,int logLevel=0);
	float readGenericFloat (bool useEfficientCoding, char* szDesc, char* szComment=NULL,int logLevel=0);
	void alignBits();
	void alignInputBits();
	float readFloat ();
	float readEfficientFloat (int* bitcount = NULL);
	float readGenericFloat (bool useEfficientCoding,int* bitcount = NULL);
	void adjustBits();
	int getBitsIn() { return incnt; }
	void resetBitsIn() { incnt=0; inbfr=0; }
	int getCurrentBufSize() { return (bufSize - (int)inbytecnt); }
	void logEncoded(int val,char* szDesc, char* szComment=NULL,int logLevel=0);
	void log(int nValue, int nBits, char* szDesc, char* szComment,int logLevel=0);
	void log(char* szComment,int logLevel=0);
#endif	//ANIMATION

};

#endif //_ANIMATIONSTREAM_DEC_H_
