/*************************************************************************

Module Name:	AnimationQuantizer.cpp

Abstract:	source file for animation streams decoding, quantization part

Author:		Mikael Bourges-Sevenier (France Telecom CNET/DIH/ATI)
Bug fixes by Tony Walsh (Teltec)

Remarks:        This code uses a single adaptative arithmetic decoder for each
                type of animated fields. We have 9 types: position 3D, 
		position 2d, color, angles, normals, scale, rotation angle,
		rotation normals and size.

*************************************************************************/

#include "defs.h"

#ifdef ANIMATION

#include <stdio.h>
#include <stdlib.h>
#include "BifsDec.h"
#include "AnimationQuantizer.h"
#include "AnimStreamDec.h"
#include <math.h>

// log function only for Zvi's Presenter
#include "Executiv.h"
#include "Presenter.h"

int AnimationQuantizer::nbBounds[]={-1,3,2,-1,1,-1,1,1,1,1,1,1,1,1,-1,-1};

AnimationQuantizer::AnimationQuantizer(int category) :
  INbBits(-1),PNbBits(-1),ncomp(-1),
  isUnitVector(FALSE),offset(0),
  type(category),Quantizer(NULL,NULL),models(NULL)
{
  //FTRD - begin - 20/11/2001 
  for(int i=0;i<3;i++) 
  {
    if (m_pMin)
		IMin[i] = m_pMin[i];
	else
		IMin[i] = -1;
	if (m_pMax)
		IMax[i] = m_pMax[i];
	else
		IMax[i] = -1;
	
	PMin[i]=PMax[i]=-1;
  }
  //FTRD - end 20/11/2001 

  if(category<AQ_LAST) {
    nbounds=nbBounds[type];
    ncomp=3; // Optimize necessary!!!
    if (category != AQ_INT) 
      isUnitVector = (category==AQ_ROTATION ||  category== AQ_NORMAL);
  }
  else 
    fprintf(stderr,"AnimationQuantizer : wrong animType = %d\n",category);

}

AnimationQuantizer::~AnimationQuantizer() {
  if(models) 
    delete[] models;
}

/*-----------------------------------------------------------------
   Initializes arithmetic encoding models. 
-----------------------------------------------------------------*/
bool AnimationQuantizer::initModels() 
{
	//FTRD - begin - 7/10/2001 
  if(!models)
  {
    models=new ArithModel[ncomp];
	return true;
  }
  return false;
  //FTRD - end - 7/10/2001 
}

/*-----------------------------------------------------------------
Resets the statistic of the models
-----------------------------------------------------------------*/
void AnimationQuantizer::resetModels() {
	
	initModels();
	for (int k=0;k<ncomp;k++)
		models[k].init(1<<PNbBits);
	
	//by Marco    
  	bitModel.init(2);

		
}

/*----------------------------------------------------------------------
Reads initial quantization parameters (during parsing of animation mask)
-----------------------------------------------------------------------*/
void AnimationQuantizer::readInitialParameters(AnimationStreamDecoder *pAS,int category) 
{
  int i;

  //FTRD - begin - 7/10/2001 
  //type=category;
  //nbounds=nbBounds[category];
  type = pAS->readBits(4,"type");
  if (type != category)
  {
	  fprintf(stderr,"ERROR : Type in .bifa is incorrect,Exiting\n");
	  exit(1);
  }
  //FTRD - end - 7/10/2001 
  nbounds=nbBounds[type];
 
  
  if(category>=AQ_LAST || category==AQ_NONE) return;

  useDefault=FALSE;
  
  //FTRD - begin - 20/11/2001 
  for(i=0;i<3;i++) 
  {
    if (m_pMin)
		IMin[i] = m_pMin[i];
	else
		IMin[i] = -1;
	if (m_pMax)
		IMax[i] = m_pMax[i];
	else
		IMax[i] = -1;
  }
  //FTRD - end 20/11/2001 

  initModels();

  switch(type) {
  case AQ_COLOR:
  case AQ_BOUNDFLOAT:
		//FTRD - begin - 7/10/2001 
		useDefault=pAS->readBit("useDefault");
		//FTRD - end - 7/10/2001 
  case AQ_POSITION3D:
  case AQ_POSITION2D:
  case AQ_SIZE3D:
  case AQ_SIZE2D:
  case AQ_FLOAT: {
	  if(!useDefault) {
		  //FTRD - begin - 7/10/2001 
		  char buffer[32];
		  int val =0;
		  for(i=0;i<nbounds;i++) 
		  {
			    val = pAS->readBit("useEfficientCoding");
			  int bitcount=0;
			  IMin[i]=pAS->readGenericFloat(val,&bitcount);
			  sprintf(buffer,"%0.3f",IMin[i]);
			  if (val)
			  {
				  unsigned int f;
				  memcpy(&f,&IMin[i],sizeof(unsigned int));
				  pAS->log(f,bitcount,"Imin",buffer);
			  }
			  else
			  {
				  unsigned int f;
				  memcpy(&f,&IMin[i],sizeof(unsigned int));
				  pAS->log(f,32,"Imin",buffer);
			  }
		  }
		  
		  for(i=0;i<nbounds;i++) 
		  {
			  val = pAS->readBit("useEfficientCoding");
			  int bitcount=0;
			  IMax[i]=pAS->readGenericFloat(val,&bitcount);
			  sprintf(buffer,"%0.3f",IMax[i]);
			  //a voir le 32 en cas d'efficient coding??
			  if (val)
			  {
				  unsigned int f;
				  memcpy(&f,&IMax[i],sizeof(unsigned int));
				  pAS->log(f,bitcount,"Imax",buffer);
			  }
			  else
			  {
				  unsigned int f;
				  memcpy(&f,&IMax[i],sizeof(unsigned int));
				  pAS->log(f,32,"Imax",buffer);
			  }
			  
		  }
		  
	  }
	  else
	  {

		  switch (type)
		  {
		  case 4: // Color
		  case 8: //BoundFloat
			  if (IMin[0] == IMax[0])
			  {
				  fprintf(stderr,"UseDefault = true for this field not Implemented\n");
				  exit(1);
			  }
			  break;
		  default: //error
			  fprintf(stderr,"Use of useDefault = true incorrect for this type %d\n",type);
			  exit(1);
		  }
	  }
	  //FTRD - begin - 7/10/2001 

  }
  break;
  case AQ_ANGLE:
    IMin[0]=0; IMax[0]=2*(float) M_PI;
    break;
  case AQ_NORMAL:    
  case AQ_ROTATION:
    for(i=0;i<nbounds;i++) {
      IMin[i]=0;
      IMax[i]=1;
    }
    break;
  case AQ_INT:
	   //FTRD - begin - 7/10/2001 
		IMin[0]=pAS->readBits(32);
		char buffer[16];
		sprintf(buffer,"%0.0f",IMin[0]);
		pAS->log(IMin[0],32,"IminInt",buffer);
		//FTRD - end - 7/10/2001 
		break;
  }
  //FTRD - begin - 7/10/2001 
  char buffer[32];
  INbBits=pAS->readBits(5);
  sprintf(buffer,"%d",INbBits);
  pAS->log(INbBits,5,"INbBits",buffer);
 //FTRD - end - 7/10/2001 
 
  setINbBits(INbBits);
  for(i=0;i<nbounds;i++) 
  {
	  //FTRD - begin - 7/10/2001 
	  int vq=pAS->readBits(INbBits+1);
	  PMin[i]=vq-(1<<INbBits);
	  sprintf(buffer,"%d %d",vq,(int)PMin[i]);
	  pAS->log(vq,INbBits+1,"Pmin",buffer);
	  //FTRD - end - 7/10/2001 
  }
  //FTRD - begin - 7/10/2001 
  PNbBits=pAS->readBits(4,"PNbBits");
  //FTRD - end - 7/10/2001 
  
  //FTRD - begin - 7/10/2001 

  for (int ind=0;ind<nbounds;ind++)
  {
	initial_IMin[ind] = IMin[ind];
	initial_IMax[ind] = IMax[ind];
	initial_PMin[ind] = PMin[ind];
	initial_PMax[ind] = PMax[ind];
  }
  initial_INbBits = INbBits;
  initial_PNbBits = PNbBits;
  //FTRD - end - 7/10/2001 
}

/*-----------------------------------------------------------------
Re-reads quantization parameters during an Intra Frame
-----------------------------------------------------------------*/
void AnimationQuantizer::readSame(AnimationStreamDecoder *pAS) {
	if(type>=AQ_LAST || type==AQ_NONE) return;
	
	if(pAS->readBit("IMinMax")) { // hasMinMax
		int i;
		useDefault=FALSE;
	
		//FTRD - begin - 20/11/2001 
		for(i=0;i<3;i++) 
		{
			if (m_pMin)
				IMin[i] = m_pMin[i];
			else
				IMin[i] = -1;
			if (m_pMax)
				IMax[i] = m_pMax[i];
			else
				IMax[i] = -1;
		}
		//FTRD - end 20/11/2001 
		
		switch(type) {
		case AQ_COLOR:
		case AQ_BOUNDFLOAT:
			useDefault=pAS->readBit("useDefault");
		case AQ_POSITION3D:
		case AQ_POSITION2D:
		case AQ_SIZE3D:
		case AQ_SIZE2D:
		case AQ_FLOAT: 
			if(!useDefault) 
			{
				//FTRD - begin - 7/10/2001 
				char buffer[32];
				int val =0;
				for(i=0;i<nbounds;i++) 
				{
					val = pAS->readBit("useEfficientCoding");
					int bitcount=0;
					IMin[i]=pAS->readGenericFloat(val,&bitcount);
					sprintf(buffer,"%0.3f",IMin[i]);
					//a voir le 32 en cas d'efficient coding??
					if (val)
					{
						unsigned int f;
						memcpy(&f,&IMin[i],sizeof(unsigned int));
						pAS->log(f,bitcount,"Imin",buffer);
					}
					else
					{
						unsigned int f;
						memcpy(&f,&IMin[i],sizeof(unsigned int));
						pAS->log(f,32,"Imin",buffer);
					}	
				}
				for(i=0;i<nbounds;i++) 
				{
					val = pAS->readBit("useEfficientCoding");
					int bitcount=0;
					IMax[i]=pAS->readGenericFloat(val);
					sprintf(buffer,"%0.3f",IMax[i]);
					//a voir le 32 en cas d'efficient coding??
					if (val)
					{
						unsigned int f;
						memcpy(&f,&IMax[i],sizeof(unsigned int));
						pAS->log(f,bitcount,"Imax",buffer);
					}
					else
					{
						unsigned int f;
						memcpy(&f,&IMax[i],sizeof(unsigned int));
						pAS->log(f,32,"Imax",buffer);
					}
					
				}
			}
			else
			{
				
				switch (type)
				{
				case 4: // Color
					IMin[0] =0.0f;
					IMax[0] =1.0f;
					break;
				case 8: //BoundFloat
					fprintf(stderr,"UseDefault = true for BoundFloat not Implemented\n");
					exit(1);
					break;
				default: //error
					fprintf(stderr,"Use of useDefault = true incorrect for this type %d\n",type);
					exit(1);
				}
			}
			//FTRD - end - 7/10/2001 
			break;
		case AQ_ANGLE:
			IMin[0]=0; IMax[0]=2*(float) M_PI;
			break;
		case AQ_NORMAL:
		case AQ_ROTATION:
			for(i=0;i<nbounds;i++) {
				IMin[i]=0;
				IMax[i]=1;
			}
			break;
		case AQ_INT:
			//FTRD - begin - 7/10/2001 
			IMin[0]=pAS->readBits(32);
			char buffer[16];
			sprintf(buffer,"%0.0f",IMin[0]);
			pAS->log(IMin[0],32,"IminInt",buffer);
			//FTRD - end - 7/10/2001 
			break;
		}
	}
	char buffer[32];
	
    if(pAS->readBit("hasINbBits")) {// hasINbBits
		INbBits=pAS->readBits(5);
		sprintf(buffer,"%d",INbBits);
		pAS->log(INbBits,5,"INbBits",buffer);
		
		setINbBits(INbBits);
    }
	
    if(pAS->readBit("PMinMax")) // hasPMinMax
	{
		for(int i=0;i<nbounds;i++) 
		{
			//FTRD - begin - 7/10/2001 
			int vq=pAS->readBits(INbBits+1);
			PMin[i]=vq-(1<<INbBits);
			sprintf(buffer,"%d %d",vq,(int)PMin[i]);
			pAS->log(vq,INbBits+1,"Pmin",buffer);
			//FTRD - end - 7/10/2001 
		}
	}
	
	if(pAS->readBit("hasPNbBits")) // hasPNbBits
		PNbBits=pAS->readBits(4,"PNbBits");
	
	
	//FTRD - begin - 7/10/2001 
	/* Mark for resetting models*/
	pAS->markNewAQP();
	//FTRD - end - 7/10/2001 
}

/*-----------------------------------------------------------------
Sets intra nbBits and prepares scaling factor for unquantization.
-----------------------------------------------------------------*/
void AnimationQuantizer::setINbBits(int iNbBits) {
  INbBits=iNbBits;
  
  int  nbBits = max(0,(isUnitVector) ? INbBits-1 : INbBits);
  for (int k=0;k<ncomp;k++) {
    float span = IMax[k]-IMin[k];
    float smallFloat = 1e-8f;
    if(type==AQ_INT)
      unQuantScale[k]=1;
    else if (span<=smallFloat) 
      unQuantScale[k]=0;
    else if(type==AQ_ROTATION || type==AQ_NORMAL) 
      unQuantScale[k] = span/((1<<(nbBits-1)) - 1);
    else 
      unQuantScale[k] = (nbBits==0) ? span*.5f : span/((1<<nbBits) - 1);
  }    
  offset = (isUnitVector) ? (1<<(INbBits-1)) : 0 ;
}  

/*-----------------------------------------------------------------
Dequantifies a value.
  vq     : quantified value
  compIdx: axis of the component (x,y,z)
-----------------------------------------------------------------*/
float AnimationQuantizer::unQuantize(int vq,int compIdx) {
  int aacIdx = (ncomp==1)?0:compIdx;
  return IMin[aacIdx] + unQuantScale[aacIdx] * (float)vq;
 // return IMin[0] + unQuantScale[0] * (float)vq;
}

/*-----------------------------------------------------------------
Decodes Intra component
-----------------------------------------------------------------*/
int AnimationQuantizer::decodeIComp(AnimationStreamDecoder *pAS) 
{
	//FTRD - begin - 7/10/2001 
  return pAS->readBits(INbBits,"quantValue")- offset;
  //FTRD - end - 7/10/2001 
}

/*-----------------------------------------------------------------
Decodes Predictive component
-----------------------------------------------------------------*/
int AnimationQuantizer::decodePComp(AnimationStreamDecoder *pAS,ArithModel *models,int k) 
{
	//FTRD - begin - 7/10/2001 
	int ret = pAS->getAAD().decodeAndUpdate(models[k]);
	TRACEP(stderr,"%d ",ret);
	//FTRD - end - 7/10/2001 

	return ret + PMin[k];
}

/*-----------------------------------------------------------------
Parses an Intra value
-----------------------------------------------------------------*/
void AnimationQuantizer::ParseAnimValue(AnimationStreamDecoder *pAS,BOOL isIntra)
{
	// decoded value is always stored on vDelta
	// in Intra, vr=vDelta
	// in Predictive, vr=vq1+vDelta except for Normal & Rotation
	
	
	
	if(isIntra) {
		//FTRD - begin - 7/10/2001 
		if (pAS->getRandomAccess() || pAS->isNewAQP())
		{
			if (pAS->getRandomAccess() && !(pAS->isNewAQP()) )
			{
				for (int ind=0;ind<nbounds;ind++)
				{
					IMin[ind] = initial_IMin[ind];
					IMax[ind] = initial_IMax[ind];
					PMin[ind] = initial_PMin[ind];
					PMax[ind] = initial_PMax[ind];
				}
				INbBits = initial_INbBits;
				PNbBits = initial_PNbBits;
				setINbBits(INbBits);
			}
			resetModels();
		}
		//FTRD - end - 7/10/2001 
		
		// read Header
		switch(type) {
		case AQ_NORMAL:
			vq1.direction=pAS->readBits(1,"direction")?-1:1;
		case AQ_ROTATION:
			vq1.orientation=pAS->readBits(2,"orientation");
			break;
		}
		
		//FTRD - begin - 7/10/2001 
		TRACEP(stderr,"\tQuant Value \t");
		//FTRD - end - 7/10/2001 
		
		
		
		
		// read components
		for (int i=0;i<ncomp;i++)
		{
			vq1.vt[i] =  vq1.vq[i] = decodeIComp(pAS);
			
			//FTRD - begin - 7/10/2001 
			TRACEP(stderr,"%0.0f ",vq1.vt[i]);
			//FTRD - end - 7/10/2001 
			
			if ((type == AQ_NORMAL) || (type == AQ_ROTATION))
				vq1.vt[i] = vq1.vt[i] - (1<<(INbBits-1));
			
			
			
			
		}
	}
	else {
		//FTRD - begin - 7/10/2001 
		pAS->markNewAQP(false);
		//FTRD - end - 7/10/2001 

		// read inverse (for Normals)
		if (type==AQ_NORMAL)
			vDelta.inverse=pAS->getAAD().decode(bitModel);
			 //vDelta.inverse=pAS->readBits(1,"inverse");

			
		//FTRD - begin - 7/10/2001 
		TRACEP(stderr,"\tShifted Value ");
		//FTRD - end - 7/10/2001 
		
		// read Components
		for (int i=0;i<ncomp;i++) {
			vDelta.vt[i] = decodePComp(pAS,models,(nbounds==1) ? 0 : i);
			//FTRD - begin - 12/10/2001 
			/*
			//this operation was made before ( in intra frame)
			if ((type == AQ_NORMAL) || (type == AQ_ROTATION))
				vq1.vt[i] = vq1.vt[i] - (1<<(INbBits-1));
			*/
			//FTRD - end - 12/10/2001 
		}
		switch(type) {
		case AQ_POSITION3D:
		case AQ_POSITION2D:
		case AQ_COLOR:
		case AQ_ANGLE:
		case AQ_FLOAT:
			
		case AQ_BOUNDFLOAT:
		case AQ_SIZE3D:
		case AQ_SIZE2D:
		case AQ_INT:
			//FTRD - begin - 7/10/2001 
			TRACEP(stderr,"\tQuant Value \t");
			//FTRD - end - 7/10/2001 
			
			vq1.add(vDelta,ncomp); // now vq1 contains the whole information
			break;
		case AQ_NORMAL: // addDelta stores result in vq1
			addDelta(vq1,vDelta,INbBits,2);
			break;
		case AQ_ROTATION:
			vq1.direction=vDelta.direction=1;
			addDelta(vq1,vDelta,INbBits,3);
			break;
		}
	}
	
	//FTRD - begin - 7/10/2001 
	TRACEP(stderr,"\tValue \t");
	//FTRD - end - 7/10/2001 
	
	if (nbounds == 1)
	{
		for(int i=0;i<ncomp;i++)
			vq1.vq[i] = IMin[0] + unQuantScale[0] * (float)vq1.vt[i];
	}
	else for(int i=0;i<ncomp;i++)
		vq1.vq[i]=unQuantize(vq1.vt[i],i);
}
/*-----------------------------------------------------------------
Parses a Vec3f value
-----------------------------------------------------------------*/
Vec3f AnimationQuantizer::ParseVec3f(AnimationStreamDecoder *pAS,BOOL isIntra)
{
  if(type==AQ_NORMAL) 
    return ParseNormal(pAS,isIntra);
  
  ncomp=3;

  ParseAnimValue(pAS,isIntra);

  Vec3f vr(vq1.vq[0],vq1.vq[1],vq1.vq[2]);

  //FTRD - begin - 7/10/2001 
  TRACEP(stderr,"%f %f %f\n",vq1.vq[0],vq1.vq[1],vq1.vq[2]);
  //FTRD - end - 7/10/2001 
  
  char buffer[64];
  for (int i=0; i<3; i++)
  {
	  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%.3f",vq1.vq[i]);
	  pAS->log(buffer,1);
  }

  return vr;
}

/*-----------------------------------------------------------------
Parses a Vec2f value
-----------------------------------------------------------------*/
Vec2f AnimationQuantizer::ParseVec2f(AnimationStreamDecoder *pAS,BOOL isIntra)
{
  ncomp=2;

  ParseAnimValue(pAS,isIntra);
    
  Vec2f vr(vq1.vq[0],vq1.vq[1]);

  //FTRD - begin - 7/10/2001 
  TRACEP(stderr,"%f %f\n",vq1.vq[0],vq1.vq[1]);
  char buffer[64];
  for (int i=0; i<2; i++)
  {
	  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%.3f",vq1.vq[i]);
	  pAS->log(buffer,1);
  }
  //FTRD - end - 7/10/2001 

  return vr;
}

/*-----------------------------------------------------------------
Parses a Color3f value
-----------------------------------------------------------------*/
Color3f AnimationQuantizer::ParseColor(AnimationStreamDecoder *pAS,BOOL isIntra)
{
  ncomp=3;

  ParseAnimValue(pAS,isIntra);
    
  Color3f vr(vq1.vq[0],vq1.vq[1],vq1.vq[2]);
  
 //FTRD - begin - 7/10/2001 
  TRACEP(stderr,"%f %f %f\n",vq1.vq[0],vq1.vq[1],vq1.vq[2]);
 char buffer[64];
  for (int i=0; i<3; i++)
  {
	  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%.3f",vq1.vq[i]);
	  pAS->log(buffer,1);
  }
  //FTRD - end - 7/10/2001 

  return vr;
}

/*-----------------------------------------------------------------
Parses an Int value
-----------------------------------------------------------------*/
int AnimationQuantizer::ParseInt(AnimationStreamDecoder *pAS,BOOL isIntra)
{
  ncomp=1;

  ParseAnimValue(pAS,isIntra);

  int vr=(int)vq1.vq[0];

 //FTRD - begin - 7/10/2001 
  TRACEP(stderr,"%d\n",vr);
  char buffer[64];
  for (int i=0; i<1; i++)
  {
	  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%d",(int)vq1.vq[i]);
	  pAS->log(buffer,1);
  }
  //FTRD - end - 7/10/2001 

  return vr;
}

/*-----------------------------------------------------------------
Parses a float value
-----------------------------------------------------------------*/
float AnimationQuantizer::ParseFloat(AnimationStreamDecoder *pAS,BOOL isIntra)
{
  ncomp=1;

  ParseAnimValue(pAS,isIntra);

  float vr=vq1.vq[0];

  //FTRD - begin - 7/10/2001 
  TRACEP(stderr,"%f\n",vq1.vq[0]);
  char buffer[64];
  for (int i=0; i<1; i++)
  {
	  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%.3f",vq1.vq[i]);
	  pAS->log(buffer,1);
  }
  //FTRD - end - 7/10/2001 

  return vr;
}

/*-----------------------------------------------------------------
Parses a Rotation value
-----------------------------------------------------------------*/
Rotation AnimationQuantizer::ParseRotation(AnimationStreamDecoder *pAS,BOOL isIntra)
{
  ncomp=3;

  ParseAnimValue(pAS,isIntra);
  vq1.direction=1;

  // convert to native component
  float natComp[4];
  quantToNat (vq1,natComp,ncomp);

  // convert to rotation
  float alpha2=acos(natComp[0]);
  if(fabs(alpha2)<1e-5) {
	  char buffer[64];
	  
	  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%.3f",0.0f);
	  pAS->log(buffer,1);
	  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%.3f",1.0f);
	  pAS->log(buffer,1);
	  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%.3f",0.0f);
	  pAS->log(buffer,1);
	  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%.3f",0.0f);
	  pAS->log(buffer,1);
	  return Rotation(0,1,0,0);
  }


  Rotation vr(natComp[1]/sin(alpha2),natComp[2]/sin(alpha2),natComp[3]/sin(alpha2),2*alpha2);
 
  //FTRD - begin - 7/10/2001 
  TRACEP(stderr,"%f %f %f %f\n",natComp[1]/sin(alpha2),natComp[2]/sin(alpha2),natComp[3]/sin(alpha2),2*alpha2);
  char buffer[64];
  
  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%.3f",natComp[1]/sin(alpha2));
  pAS->log(buffer,1);
  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%.3f",natComp[2]/sin(alpha2));
  pAS->log(buffer,1);
  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%.3f",natComp[3]/sin(alpha2));
  pAS->log(buffer,1);
  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%.3f",2*alpha2);
  pAS->log(buffer,1);
  //FTRD - end - 7/10/2001 
  
  return vr;
}

/*-----------------------------------------------------------------
Parses a normal
-----------------------------------------------------------------*/
Vec3f AnimationQuantizer::ParseNormal(AnimationStreamDecoder *pAS,BOOL isIntra) {
  ncomp=2;
  ParseAnimValue(pAS,isIntra);

  // convert to native component
  float natComp[3];
  quantToNat (vq1,natComp,ncomp);


  Vec3f vr(natComp[0],natComp[1],natComp[2]);

  //FTRD - begin - 7/10/2001 
  TRACEP(stderr,"%f %f %f\n",natComp[0],natComp[1],natComp[2]);
 
  char buffer[64];
  
  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%.3f",natComp[0]);
  pAS->log(buffer,1);
  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%.3f",natComp[1]);
  pAS->log(buffer,1);
  sprintf(buffer,"\t\t\t\t\t\t\t\t\t\t%.3f",natComp[2]);
  pAS->log(buffer,1);
  //FTRD - end - 7/10/2001 
  return vr;
}

BOOL AnimationQuantizer::operator == (const AnimationQuantizer &qp) const {
  if(type==qp.type) return TRUE;
  return FALSE;
}


/*-----------------------------------------------------------------
Translates from quantized components to native components 
for Normal & Rotation only
-----------------------------------------------------------------*/
// static
void AnimationQuantizer :: quantToNat (QVAL &qv, float *natComp, int ncomp)
{
  int j,len=ncomp+1;
  
  int direction   = qv.direction; // only for Normal else 1
  int orientation = qv.orientation;
			  

  float delta=1;
  float vt[4];
  for (j=0;j<ncomp;j++) {
    float v = qv.vq[j];
    vt[j]= (float) tan (v*M_PI/4);
    delta+= vt[j]*vt[j];
  }
  delta = (float)(direction/sqrt(delta));
  
  natComp[(orientation)%len] = delta;
  for (j=0;j<ncomp;j++) {
    natComp[(orientation+j+1)%len] = vt[j]*delta;
  }


}

/*-----------------------------------------------------------------
Computes the sum of a quantized compressed unit vector and a delta
for Normal & Rotation only
-----------------------------------------------------------------*/
// static
void AnimationQuantizer :: addDelta (QVAL &qVal1,QVAL &qDelta,int NbBits,int NBCOMP) {
  //  NbBits = number of bits the quantization is made on
  //  NBCOMP = number of reduced components : 2 for Normals, 3 for Rotations
  //  qVal1  = NBCOMP quantized coordinates of the 1st compressed object
  //  qVal2  = NBCOMP quantized coordinates of the 2nd compressed object
  //  qDelta = NBCOMP quantized coordinates of the delta
  
  QVAL qVal2;
  int scale = (1<<max(NbBits-1,0))-1;   // highest possible quantized value

  // computes the difference in orientation and direction


  int len=NBCOMP+1; // length of decoded value
  int deltaInverse = (qDelta.inverse==FALSE) ? 1 : -1;
  
  int inv=1;
  int diffDirection   = qVal1.direction * deltaInverse;
  int dOri = 0;
  
  qVal2.orientation=qVal1.orientation;
  qVal2.direction=diffDirection;

  // Computes result's compressed components
  for (int j=0;j<NBCOMP;j++) {
    qVal2.vq[j] = qVal1.vt[j]+qDelta.vt[j];
    if (fabs(qVal2.vq[j])>scale) {  // going out of the cube's face
      dOri = j+1;        // relative orientation we are escaping towards
      inv = (qVal2.vq[j]>=0) ? 1 : -1;   // direction are we escaping towards
      break;
    }
  }

  
  // If need to change face --> reindexation
  if (dOri!= 0) {
    int j,k;
    for (j=0;j<NBCOMP-dOri;j++) {
      k = (j+dOri)%NBCOMP;
      qVal2.vq[j] = inv * (qVal1.vq[k]+qDelta.vq[k]);
    }
    k = dOri-1;
    qVal2.vq[NBCOMP-dOri] = inv * 2 * scale - (qVal1.vq[k]+qDelta.vq[k]);
    for (j=NBCOMP-dOri+1;j<NBCOMP;j++) {
      k = (j+dOri-1)%NBCOMP;
      qVal2.vq[j] = inv * (qVal1.vq[k]+qDelta.vq[k]);
    }

    qVal2.orientation=(qVal1.orientation+dOri) % len;
    qVal2.direction=diffDirection*inv;
  }



  qVal1.set(qVal2);
}

#endif ANIMATION
