/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

  SUBSYSTEM:  Decoder
  FILE:       AnimationStreamDec.cpp
  AUTHOR:     Mikael Bourges-Sevenier (France Telecom CNET/DIH/ATI)
  OVERVIEW:   Implementation of ArithmeticDecoder
-------------------------------------------------------------------------*/

/************************************************************************
 *
 *  sac.c, part of tmndecode (H.263 decoder)
 *  Copyright (C) 1996  Telenor R&D, Norway
 *        Karl Olav Lillevold <Karl.Lillevold@nta.no>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Karl Olav Lillevold               <Karl.Lillevold@nta.no>
 *  Telenor Research and Development
 *  P.O.Box 83                        tel.:   +47 63 84 84 00
 *  N-2007 Kjeller, Norway            fax.:   +47 63 81 00 76
 *
 *  Robert Danielsen                  e-mail: Robert.Danielsen@nta.no
 *  Telenor Research and Development  www:    http://www.nta.no/brukere/DVC/
 *  P.O.Box 83                        tel.:   +47 63 84 84 00
 *  N-2007 Kjeller, Norway            fax.:   +47 63 81 00 76
 *  
 ************************************************************************/

/*********************************************************************
 *        SAC Decoder Module
 *        Algorithm as Specified in H26P Annex -E
 *              (c) 1995 BT Labs
 *
 *	Author:	Wayne Ellis <ellis_w_wayne@bt-web.bt.co.uk>
 *
 *********************************************************************/

#include "defs.h"

#ifdef ANIMATION

#include "ArithDec.h"
#include "AnimStreamDec.h"

/*********************************************************************
 *        SAC Decoder Algorithm as Specified in H26P Annex -E
 *
 *        Name:        decode_a_symbol
 *
 *	Description:	Decodes an Aritmetically Encoded Symbol
 *
 *	Input:        none. Uses built-in array of cumulative frequencies.
 *
 *	Returns:	Index to relevant symbol model
 *
 *	Side Effects:	Modifies low, high, length, cum and code_value
 *
 *	Author:        Wayne Ellis <ellis_w_wayne@bt-web.bt.co.uk>
 *
 *********************************************************************/
 
ArithmeticDecoder::ArithmeticDecoder(AnimationStreamDecoder *pAS) :
  nb_of_symbols(256),freq(NULL),cumul_freq(NULL),pin(pAS),
  bitcount(0),opposite_bits(0),zerocount(0)
{
	  
}

ArithmeticDecoder::~ArithmeticDecoder() {
  if(cumul_freq) delete[] cumul_freq;
  if(freq) delete[] freq;
}

/*-----------------------------------------------------------------
Gets a bit out of PSC layer
-----------------------------------------------------------------*/
BOOL ArithmeticDecoder::bit_out_psc_layer()
{
  if(zerocount == ZEROMAX) {
    bitcount++;
    #ifndef LOG_AACVAL
	if(!pin->readBit("AAC","",1)) {
	#else
	if(!pin->readBit()) {
	#endif //LOG_AACVAL
	  
	  //FTRD - begin - 7/10/2001 
	  m_lastvalue<<=1;
	  m_lastvalue|=0;
	  //FTRD - end - 7/10/2001 

      fprintf(stderr, "MARKER BIT error in arithmetic decoding\n");
      return FALSE;
    }
	//FTRD - begin - 7/10/2001 
	m_lastvalue<<=1;
	m_lastvalue|=1;
	//FTRD - end- 7/10/2001 
    
	zerocount = 0;
  }
  
  #ifndef LOG_AACVAL
  Bit = pin->readBit("AAC","",1);
  #else
  Bit = pin->readBit();
  #endif //LOG_AACVAL
  bitcount++;
  //FTRD - begin - 7/10/2001 
  m_lastvalue<<=1;
  if(!Bit)
  {
	  zerocount++;
	  
	  m_lastvalue|=0;
  }
  else
  {
	  zerocount = 0;
	  
	  m_lastvalue|=1;
  }
  //FTRD - end - 7/10/2001 

  
  return TRUE;
}

/*-----------------------------------------------------------------
Fills Decoder FIFO after a fixed word length string has been detected.
-----------------------------------------------------------------*/
BOOL ArithmeticDecoder :: reset()
{
  int i;
  
  code_value = 0;
  low = 0;
  high = TOP;
  zerocount = 0;
  bitcount=0;
  opposite_bits=0; // OO : opposite_bits gets reset here, because it is not in flush()
  
  for (i = 1;   i <= CODEVALNBBITS;   i++) {
    if(!bit_out_psc_layer())
      return FALSE;
    code_value = 2 * code_value + (Bit ? 1 : 0);
  }
  
  return TRUE;
}

/*-----------------------------------------------------------------
Decodes a symbol from stream and updates statistics
-----------------------------------------------------------------*/
int ArithmeticDecoder :: decodeAndUpdate(ArithModel &model)
{
	int val=decode(model);
	update(val,model);
	return val;
}

/*-----------------------------------------------------------------
Decodes a symbol from stream 
-----------------------------------------------------------------*/
int ArithmeticDecoder :: decode(ArithModel &model)
{
	//FTRD - begin - 12/10/2001 
	m_lastvalue =0;
	//FTRD - end- 12/10/2001 

  length = high - low + 1;
  cum = (-1 + (code_value - low + 1) *  model.getCumulFreq(0)) / length;
  
  for (sacindex = 1; model.getCumulFreq(sacindex) > (int) cum; sacindex++);
  
  high  = low - 1 + (length * model.getCumulFreq(sacindex-1)) / model.getCumulFreq(0);
  low  +=           (length * model.getCumulFreq(sacindex  )) / model.getCumulFreq(0);
  
  bitcount=0;
  
  for ( ; ; ) {
    if (high < Q2) 
      opposite_bits=0;
    else if (low >= Q2) {
      code_value -= Q2;
      low -= Q2;
      high -= Q2;
      opposite_bits=0;
    }
    
    else if (low >= Q1 && high < Q3) {
      code_value -= Q1;
      low -= Q1;
      high -= Q1;
      opposite_bits++;
    }
    
    else {
      break;
    }
    
    low *= 2;
    high = 2*high + 1;
    if(!bit_out_psc_layer())
      return DEC_ERROR;
    code_value = 2*code_value + (Bit ? 1 : 0);
  }
  
  //FTRD - begin - 12/10/2001 
  #ifdef LOG_AACVAL
  pin->log(m_lastvalue,bitcount,"encodedValue","",1);
  #endif //LOG_AACVAL
  //FTRD - end - 12/10/2001 
  return (sacindex-1);
}

void ArithmeticDecoder :: adjustBits()
{
  opposite_bits++;
  pin->adjustBits();
}

#endif ANIMATION
