/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS/OD Encoder
FILE:       ScanOD.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implement functions for parsing textual description of
			OD classes.
-------------------------------------------------------------------------*/
#include "defs.h"
#include "ODP.h"
#include "Token.h"
#include "Im1defs.h"

/*-------------------------------------------------------------
Instantiate a descriptor matching the given name
-------------------------------------------------------------*/
BaseDescriptor *SDLField :: CreateClass (LPCSTR szClassName)
{
	for (int nCode = 0; nCode < MPEG4_DESC_TABLE_LENGTH; nCode ++)
		if (m_odFactory [nCode] != NULL)
		{	BaseDescriptor *p = m_odFactory [nCode] -> CreateDescriptor ();
			if (strcmp (p -> GetName (), szClassName) == 0)
				return p;
			delete p;
		}
	return NULL_DESC;
}

/*-------------------------------------------------------------
Parse an integer value from the textual OD description
-------------------------------------------------------------*/
void SDLInteger :: Scan ()
{
	m_nValue = nextToken.GetInt ();
}

/*-------------------------------------------------------------
Parse a float value from the textual OD description
-------------------------------------------------------------*/
void SDLFloat :: Scan ()
{
	m_fValue = nextToken.GetReal ();
}

/*-------------------------------------------------------------
Parse a doub le value from the textual OD description
-------------------------------------------------------------*/
void SDLDouble :: Scan ()
{
	m_fValue = nextToken.GetReal ();
}

/*-------------------------------------------------------------
Parse an array (of any kind) from the textual OD description
-------------------------------------------------------------*/
void SDLArrayBase :: Scan ()
{
	if (nextToken == TID_LEFT_BRACKET)
	{	nextToken.Scan ();
		while (nextToken != TID_RIGHT_BRACKET)
		{
			ScanElement ();
			if (nextToken == TID_COMMA)
				nextToken.Scan ();
		};
		nextToken.Scan ();	
	}
	else
		ScanElement ();
}

/*-------------------------------------------------------------
Parse an integer array from the textual OD description. It can be
a character string, or a real array.
-------------------------------------------------------------*/
void SDLIntegerArray :: Scan ()
{
	if (m_nBits == 8 && nextToken == TID_STRING)
	{	m_array = nextToken.GetString ();
		nextToken.Scan ();
	}
	else
		SDLArrayBase :: Scan ();	
}

/*-------------------------------------------------------------
Parse one element of an integer array from the textual OD description
-------------------------------------------------------------*/
void SDLIntegerArray :: ScanElement ()
{
	InsertElement (nextToken.GetInt ());
}

/*-------------------------------------------------------------
Parse a character string from the textual OD description
-------------------------------------------------------------*/
void SDLString :: Scan ()
{
	m_array = nextToken.GetString ();
	nextToken.Scan ();
}

/*-------------------------------------------------------------
Parse one element of an array of character strings from the textual OD description
-------------------------------------------------------------*/
void SDLStringArray :: ScanElement ()
{
	m_array += nextToken.GetString ();
	nextToken.Scan ();
}

/*-------------------------------------------------------------
Parse a descriptor field from the textual OD description
-------------------------------------------------------------*/
void SDLBaseClass :: Scan ()
{
	int nType;
	m_pClass = CreateNamedDescriptor (nextToken.GetName (), nType);
	if (!m_pClass)
	{	if (nextToken != TID_NAME)
			nextToken.Error ("Unexpected symbol");
		else
			nextToken.Error ("Unrecognized class name");
		nextToken.Scan ();
		return;
	}

	if (nType == -1)
		nextToken.Error ("Unexpected class");
	if (nType == -2)
	{	nType = 1;
		if (nextToken != TID_LEFT_BRACE)
		{	nextToken.Error ("Unexpected symbol");
			nextToken.Scan ();
		}
	}
	else
		nextToken.Scan ();

	m_pClass -> Scan ();
	if (nType == -1)
	{	delete m_pClass;
		m_pClass = NULL;
	}
	else
		m_pClass -> PreEncode (this);
}

/*-------------------------------------------------------------
Create a descriptor for this field, according to the given name.
If the given name is not of a valid descripotor (for this field), it
is either an error (rnResult returns -1), or it is not a name at all.
In the later case, if the field has just one option, we don't need a
name, so if it is not a name of any descriptor it means a descriptor
name is missing. This is legal if the descriptor type can be implied
by this field, in which case rnResult returns -2.
-------------------------------------------------------------*/
BaseDescriptor *SDLBaseClass :: CreateNamedDescriptor (
	LPCSTR szClassName, 
	int &rnResult
)
{
	rnResult = 1;
	BaseDescriptor *pDesc = CreateDescriptor (szClassName);
	if (pDesc == NULL_DESC ||
		stricmp (pDesc -> GetName (), szClassName) != 0
	)
	{
		BaseDescriptor *p = SDLField :: CreateClass (szClassName);
		if (p != NULL)
		{	rnResult = -1;
			delete pDesc;
			return p;
		}
		else
			rnResult = -2;
	}
	return pDesc;
}

/*-------------------------------------------------------------
Parse a descriptor from the textual OD description
-------------------------------------------------------------*/
void BaseDescriptor :: Scan ()
{
	if (nextToken == TID_LEFT_BRACE)
	{	nextToken.Scan ();
		while (nextToken != TID_RIGHT_BRACE)
		{	if (nextToken != TID_NAME)
			{	nextToken.Error ("Unexpected symbol");
				nextToken.Scan ();
			}
			else
			{	const ODFieldTable *pEntry =
					GetODEntry (nextToken.GetName ());
				if (!pEntry)
					nextToken.Error ("Unrecognized field name");
				nextToken.Scan ();
				if (pEntry != NULL)
				{	SDLField *pField =
						(SDLField *) (((char *) this) + pEntry -> m_offset);
					pField -> Scan ();
				}
			}
		}
		nextToken.Scan ();
	}
}

/*-----------------------------------------------------------------
Returns a pointer to the OD table entry for the field whose name is
szName.
-----------------------------------------------------------------*/
const ODFieldTable * BaseDescriptor :: GetODEntry (LPCSTR szName)
{
	const ODFieldTable *pTable = GetFieldTable ();
	for (int i = GetFieldCount (); --i >= 0;)
		if (pTable [i].m_szName && !stricmp (szName, pTable [i].m_szName))
			return pTable + i;
	return NULL;
}

/*-------------------------------------------------------------
Append one element to an OD class array during OD parsing
-------------------------------------------------------------*/
void SDLBaseClassArray :: AddElement (BaseDescriptor *pDesc)
{
	if (m_array.GetLength () > m_nMax)
		nextToken.Error ("Too many array elements");
	m_array += pDesc;
}
