/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Optibase Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Optibase Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1997.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Encoder
FILE:       VrmlPars.cpp
AUTHOR:     Zvi Lifshitz (Optibase Ltd.)
OVERVIEW:   Implementation of VRMLParser, the class that parses tokens from input files
			in VRML format
-------------------------------------------------------------------------*/
#include "defs.h"
#include "VrmlPars.h"

/*-------------------------------------------------------------
Initialize the VRMLParser object by opening the input file
-------------------------------------------------------------*/
bool VRMLParser :: Init (LPCSTR szFileName)
{
	if ((m_fInput = fopen (szFileName, "r")) == NULL)
		return false;
	m_szRootFile = szFileName;
	m_nInputOffset = m_nInputLength = m_nLineNumber = 0;
	m_cNext = GetChar ();
	return true;
}

/*-------------------------------------------------------------
Shutdown the VRMLParser object - close the input file
-------------------------------------------------------------*/
void VRMLParser :: Shutdown ()
{
	if (m_fInput)
		fclose (m_fInput);
}

/*-------------------------------------------------------------
Scan the next token from the input file
-------------------------------------------------------------*/
void VRMLParser :: GetToken (TokenInfo *pInfo)
{
	Scan ();
	pInfo ->m_code = m_code;
	pInfo ->m_nLine = m_nLineNumber;
	pInfo ->m_string = m_string;
}

/*-----------------------------------------------------------------
Read one character from the input file.
-----------------------------------------------------------------*/
int VRMLParser :: GetChar ()
{
	if (m_nInputOffset >= m_nInputLength)
	{	if (feof (m_fInput))
			return EOF;
		if (!fgets (m_szLine, sizeof (m_szLine), m_fInput))
			return EOF;
		m_nInputLength = strlen (m_szLine);
		m_nInputOffset = 0;
		++m_nLineNumber;
	}
	return m_szLine [m_nInputOffset++];
}

/*-----------------------------------------------------------------
Scan a symbol and find its code
-----------------------------------------------------------------*/
void VRMLParser :: Scan ()
{
	/*-----------------------------------------------------------------
	Skip all white spaces (and comments)
	-----------------------------------------------------------------*/
	while (
		m_cNext == ' ' || m_cNext == '\t' || m_cNext == '\n' ||
		m_cNext == '\r' || m_cNext == '#' ||
		m_cNext == '/' && m_szLine [m_nInputOffset] == '/'
	)
	{	// Handle a comment, # or // to the end of the line
		if (m_cNext == '#' || m_cNext == '/')
			while (m_cNext != '\n' && m_cNext != EOF)
				m_cNext = GetChar ();
		m_cNext = GetChar ();
	}
	if (m_cNext == EOF)
	{	m_code = TID_EOF;
		return;
	}

	m_string = "";
	m_code = TID_UNKNOWN;

	/*-----------------------------------------------------------------
	Parse a name
	-----------------------------------------------------------------*/
	if (isalpha (m_cNext) || m_cNext == '_' || m_cNext == '$')
	{	while (
			isalnum (m_cNext) || m_cNext == '_' || m_cNext == '$' ||
			m_cNext == '-'
		)
		{	m_string += m_cNext;
			m_cNext = GetChar ();
		}
		m_code = TID_NAME;
	}

	/*-----------------------------------------------------------------
	Parse a number
	-----------------------------------------------------------------*/
	else if (isdigit (m_cNext) || m_cNext == '-' || m_cNext == '+')
	{	m_code = TID_INTEGER;
		bool bHex = false;
		while (
			isalnum (m_cNext) || m_cNext == '-' || m_cNext == '+' ||
			m_cNext == '.'
		)
		{	m_string += m_cNext;
			bHex |= (tolower (m_cNext) == 'x');
			if (!bHex && (m_cNext == '.' || tolower (m_cNext) == 'e'))
				m_code = TID_NUMBER;
			m_cNext = GetChar ();
		}
	}

	/*-----------------------------------------------------------------
	Parse a string
	-----------------------------------------------------------------*/
	else if (m_cNext == '"')
	{	while ((m_cNext = GetChar ()) != '"')
		{	if (m_cNext == '\\')
				m_cNext = GetChar ();
			if (m_cNext == EOF)
			{	m_string = "Unexpected end of file";
				return;
			}
			/*-------------------------------------------------------------
			Do UTF-8 translation
			-------------------------------------------------------------*/
			if (m_cNext & 0x80)
			{	m_string += 0xc0 | (m_cNext >> 6) & 0x3;
				m_cNext &= 0xbf;
			}
			m_string += m_cNext;
		}
		m_cNext = GetChar ();
		m_code = TID_STRING;
		return;
	}

	/*-------------------------------------------------------------
	Read a complete file into a string
	-------------------------------------------------------------*/
	else if ((m_cNext == '<') || (m_cNext == '('))
	{	while (((m_cNext = GetChar ()) != '>') && (m_cNext != ')'))
		{	if (m_cNext == EOF)
			{	m_string = "Unexpected end of file";
				return;
			}
			m_string += m_cNext;
		}
		m_cNext = GetChar ();
		m_code = TID_STRING;
		char szFileName [300];
		FILE *fExtFile;
		if ((LPCSTR) m_szRootFile != NULL &&
			ZExpandURL (m_szRootFile, m_string, szFileName, 300) == 0
		)
			fExtFile = fopen (szFileName, "rb");
		else
			fExtFile = fopen (m_string, "rb");
		if (fExtFile == NULL)
		{	m_string += " - file not found";
			m_code = TID_UNKNOWN;
		}
		else
		{	int c;
			m_string = "";
			while ((c = fgetc (fExtFile)) != EOF)
				m_string += c;
			fclose (fExtFile);
		}
		return;
	}

	/*-----------------------------------------------------------------
	Parse one character
	-----------------------------------------------------------------*/
	else
	{	m_string += m_cNext;
		m_cNext = GetChar ();
	}

	/*-----------------------------------------------------------------
	Check if a reserved symbol
	-----------------------------------------------------------------*/
	TOKEN_CODE code = GetSymbol (m_string);
	if (code != TID_UNKNOWN)
		m_code = code;
	else if (m_code == TID_UNKNOWN)
		m_string += " - invalid symbol";
}

/*-----------------------------------------------------------------
If quotes are not used, read all characters till a white space 
-----------------------------------------------------------------*/
void VRMLParser :: GetString (ZString &str)
{
	if (m_code != TID_STRING)
		while (
			m_cNext != ' ' && m_cNext != '\t' && m_cNext != '\n' &&
			m_cNext != EOF
		)
		{	m_string += m_cNext;
			m_cNext = GetChar ();
		}
	str = m_string;
}
