/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11
7
SUBSYSTEM:  Buffer management
FILE:       Clock.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of ClockReference.
-------------------------------------------------------------------------*/
#include "defs.h"
#include "Clock.h"
#include "ztl.h"
#include "MediaControl.h"
#include "Inline.h"
#include "ODImp.h"

#ifndef WIN32

// SOLARIS PORTING - put the functions timeGetTime and Sleep in time_def.h
// #include <sys/time.h>
#include "../Include/time_def.h"
#include <unistd.h>

/*
DWORD timeGetTime ()
{
	struct timeval time;
	::gettimeofday(&time);
	return (DWORD) (time.tv_sec * 1000 + time.tv_usec / 1000);
}

void Sleep (DWORD dw)
{
	usleep ((long long) dw * 1000);
}
*/

#endif

/*-----------------------------------------------------------------
Static members intialization
-----------------------------------------------------------------*/
int ClockReference :: m_nGranularity = 20;

/*-----------------------------------------------------------------
Reset the clock to zero
-----------------------------------------------------------------*/
void ClockReference :: Reset ()
{
	m_bReset = true;
	m_dFactor = m_fTransform = m_fSpeed = m_fCumulSpeed = 1;
	m_dwInitOCR = 0;
	for (int i = 0; i < NUM_TAPS; i++)
	{
		m_dDriftEstimate [i] = 1;
	}
}

/*-----------------------------------------------------------------
Return current value of the object
-----------------------------------------------------------------*/
DWORD ClockReference :: GetTime ()
{
	// Guido 6/12/00
	// if the clock was not started yet, return 0 instead of random
	if (m_bReset)
		return 0;
	DWORD dwNow = m_nFrozen > 0 ? m_dwFreezeTime : timeGetTime ();
	DWORD dwTime = m_fCumulSpeed * m_dFactor * (dwNow - m_dwStartTime) + m_dwInitOCR;
	return dwTime - min (int (dwTime), m_nAdjust);
}


/*	max sleep time in wait - 2 reasons for this:
	-it seems that the zemaphore unlocking is not preoperly working on windows (typically if sleeping for 100s
the sempahore is not wacken before a long time), thus the restart of the presentation may lock
	-the clock may not be initialized yet and we haven't droped the data on the input stream (to ensure we keep 
RAP even if not inited). 
*/
#define MAX_CLOCK_WAIT_MS	500

/*-----------------------------------------------------------------
Wait till the object reaches the given time. Return false if didn't have to wait
-----------------------------------------------------------------*/
bool ClockReference :: WaitOn (ZSemaphore &rsmpControl, DWORD dwTime)
{
	DWORD dw;

	if (dwTime >= (dw = GetTime ()) + m_nGranularity)
	{
		DWORD dwWaitTime = (dwTime - dw - m_nGranularity) / m_dFactor / m_fCumulSpeed;
		/*JLF 03/03: weird bug on XP (maybe other platforms), the release of the semaphore timer doesn't 
		reliably work, thus may lock the MediaStream::close() until the end of sleep. 
		the hack is to sleep only for small periods (MAX_CLOCK_WAIT_MS) */
		if (dwWaitTime > MAX_CLOCK_WAIT_MS) dwWaitTime = MAX_CLOCK_WAIT_MS;
		rsmpControl.P (dwWaitTime);
		return true;
	}
	return false;
}

/*-----------------------------------------------------------------
Compare two time stamps
-----------------------------------------------------------------*/
int ClockReference :: CompareTimes (DWORD dwTime, DWORD dwCompareTo)
{
	int nResult;

	/*-------------------------------------------------------------
	Timestamp of zero means no OCR was available at the time the AU
	was dispatched, therefore it should be immediately available.
	-------------------------------------------------------------*/
	if (dwTime == 0)
		return 0;

	if (dwCompareTo == 0)
		dwCompareTo = GetTime ();
	nResult = (long) dwTime - (long) dwCompareTo;
	if (abs (nResult) <= m_nGranularity * m_fCumulSpeed)
		nResult = 0;
	return nResult;
}

/*-----------------------------------------------------------------
Set the clock to a given time.
This function is called to initialize the clock, or to update it when 
a new OCR is received. In the first case, we initialize the clock by
storing the current time (minus the initialization value). In the later
case, we calculate the speed of the clock relatively to our system clock
and store a value that will enable retrieving the controlled speed of
the clock.
-----------------------------------------------------------------*/
void ClockReference :: SetTime (DWORD dwOCR, int nOCRLength)
{
	m_dwLastOCR = dwOCR;
	if (m_bReset)
	{	m_dwRealStartTime = m_dwStartTime = m_dwFreezeTime = timeGetTime ();
		m_dwInitOCR = dwOCR;
		m_nOCRLength = nOCRLength;
		m_bReset = false;
		m_nAdjust = 0;
	}
	else
	{	
		/*-------------------------------------------------------------
		OCR adjustment made by Guido Franceschini (CSELT)

		We estimate the drift between the timebase conveyed by the OCR
		and the STB. We assume that the drift cannot exceed MAX_DRIFT in
		either direction and constrain estimations inside those limits.
		We compute the adjustment factor and use it in subsequent
		clock references only if there is a time slip of more than
		THREESOLD ms between the STB and the time conveyed by the OCR
		We also recalculate the start time in order to avoid discontinuities
		in the STB. However, when estimating the drift, the real start
		time is considered.
		-------------------------------------------------------------*/
		DWORD dwOCRTime = dwOCR - m_dwInitOCR;
		DWORD dwCurrentTime = timeGetTime ();
		DWORD dwElapsedTime = dwCurrentTime - m_dwRealStartTime;
		double dSumDrift = 0;
		for (int i=0; i<NUM_TAPS-1; i++)
		{
			m_dDriftEstimate [i] = m_dDriftEstimate [i+1];
			dSumDrift += m_dDriftEstimate [i];
		}
		m_dDriftEstimate [NUM_TAPS-1] = (((double)dwOCRTime/(double)dwElapsedTime) + dSumDrift)/NUM_TAPS;
		if (m_dDriftEstimate [NUM_TAPS-1] > (1 + MAX_DRIFT))
			m_dDriftEstimate [NUM_TAPS-1] = (1 + MAX_DRIFT);
		if (m_dDriftEstimate [NUM_TAPS-1] < (1 - MAX_DRIFT))
			m_dDriftEstimate [NUM_TAPS-1] = (1 - MAX_DRIFT);
		DWORD dwClockTime = m_dFactor * dwElapsedTime;
		if (abs ((int) (dwClockTime - dwOCRTime)) >= THREESOLD) {
			double dOldFactor = m_dFactor;
			DWORD dwOldStartTime = m_dwStartTime;
			m_dFactor = m_dDriftEstimate [NUM_TAPS-1];
			m_dwStartTime = dwCurrentTime - ((dOldFactor / m_dFactor) * (dwCurrentTime - dwOldStartTime));
		}
	}
}

/*-------------------------------------------------------------
InitFromCTS is called if a stream encountered a CTS before clock was initialized by an OCR.
If no OCR is expected, the function initialize the clock with the CTS value and returns
true
-------------------------------------------------------------*/
bool ClockReference :: InitFromCTS (DWORD dwTime)
{
// Guido 6/12/00
// The original code was to use CTS in case of no OCR.
// Problem : non-sense to call SetTime more than once
// New code makes sure that SetTime is called at most once.
//	if (m_bHasOCR)
	if (m_bHasOCR || !m_bReset)
		return false;
	SetTime (dwTime, 0);
	return true;
}

/*-------------------------------------------------------------
Start is called when a stream starts using a clock. Reset the clock if this is the first
stream using it
-------------------------------------------------------------*/
void ClockReference :: Start ()
{
	if (m_nOpenCounter++ == 0)
		Reset ();
}

/*-------------------------------------------------------------
Stop is called when a stream stops using a clock. Decrement the use counter. If the stream
that called this function is an OCR stream, mark that the clock has been reset.
-------------------------------------------------------------*/
void ClockReference :: Stop (bool bOCRStream)
{
	--m_nOpenCounter;
	if (bOCRStream) 
		m_bReset = true;
}

/*-----------------------------------------------------------------
Adjustment value is the time elapsed since the clock was set to 
a certain value till a unit whose timestamp equals that value is rendered.
Normally this equals the time it takes to decode the unit.
-----------------------------------------------------------------*/
void ClockReference :: Adjust (DWORD dwTime)
{
	m_nAdjust =
		m_dFactor * m_fCumulSpeed * (timeGetTime () - m_dwStartTime) -
		(dwTime - m_dwInitOCR);
}

/*-------------------------------------------------------------
Given residual value of a timestamp (dwTS), its size in bits (nLength)
and resolution (units per second), compare the value to the last OCR used
to set the clock and return the offset which should be added to the timestamp
in order to reconstruct the real timestamp value
-------------------------------------------------------------*/
__int64 ClockReference :: AdjustTimeOffset (
	DWORD dwTS,
	int nLength,
	float fResolutionFactor
)
{
	__int64 dwBase = (__int64)1 << nLength;
	/*-------------------------------------------------------------
	GUIDO 14/6/00: 64 bits shift is wrong
	-------------------------------------------------------------*/
	if (nLength == 64)
		dwBase = 0;
	__int64 dwOCR = (double) m_dwLastOCR / fResolutionFactor;
	/*-------------------------------------------------------------
	GUIDO 02/02/01: previous computation was incorrect for nLength !=
	m_nOCRLength, in particular with TSlength < 32 and OCRlength = 0
	GUIDO 02/02/01: dwOCR contains the clock value in TS units;
	dwOCRMSB contains the MSB of dwOCR (the bits exceeding the TSlength)
	dwOCRLSB contains the LSB of dwOCR (the lower TSlength bits)
	the returned value should represent the most significant
	bits only, otherwise it would add an offset to the transmitted bits
	of the TSs of media that are supposed to be in sync, which is incorrect.
	We take care as well of the current timeline and of the actual
	value of the timestamp, in order to avoid miscomputation in case
	either the timeline or the timestamp are close to their maximum value
	in the permitted range
	-------------------------------------------------------------*/
	__int64 dwOCRMSB = dwOCR & ~(dwBase - 1);
	__int64 dwOCRLSB = dwOCR & (dwBase - 1);
	__int64 diff = (__int64)dwTS - dwOCRLSB;
	if (diff > dwBase/2)
		return (dwOCRMSB - dwBase);
	else if (diff >= 0)
		return (dwOCRMSB);
	else if ((-diff) > dwBase/2)
		return (dwOCRMSB + dwBase);
	else
		return (dwOCRMSB);
}

/*-------------------------------------------------------------
Freeze the clock. bRebuffering is true if this pause was called internally because
the buffer was starved.
-------------------------------------------------------------*/
void ClockReference :: Pause ()
{	
	if (m_nFrozen++ == 0)
		m_dwFreezeTime = timeGetTime ();
}

/*-------------------------------------------------------------
Resume the clock after it was frozen by Pause().
bRebuffering is true if this resume was called internally after finish rebuffering. In this
case resume will not be executed if last pause was triggered manually
-------------------------------------------------------------*/
void ClockReference :: Resume ()
{
	if (--m_nFrozen == 0)
		m_dwStartTime += timeGetTime () - m_dwFreezeTime;
}

/*-------------------------------------------------------------
Modify the clock speed when affected by a TemporalTransform.
If fFactor is negative the clock needs to be resumed after a previous pause
-------------------------------------------------------------*/
void ClockReference :: Scale (float fFactor)
{
	if (fFactor < 0)
		Resume ();
	else if (fFactor == 0)
		Pause ();
	else {
		DWORD dwCurrentTime = timeGetTime();
		m_dwStartTime = dwCurrentTime - ((m_fTransform / fFactor) * (dwCurrentTime - m_dwStartTime));
		m_fTransform = fFactor;
		m_fCumulSpeed = m_fTransform * m_fSpeed;
	}
}

/*-------------------------------------------------------------
set the clock speed
-------------------------------------------------------------*/
void ClockReference :: SetSpeed (float fSpeed)
{
	assert(fSpeed>0);
	if (fSpeed != m_fSpeed) {
		DWORD dwCurrentTime = timeGetTime();
		m_dwStartTime = dwCurrentTime - ((m_fSpeed / fSpeed) * (dwCurrentTime - m_dwStartTime));
		m_fSpeed = fSpeed;
		m_fCumulSpeed = m_fTransform * m_fSpeed;
	}
}

/*-------------------------------------------------------------
Check if a given media time value is within the bounds set by an a associated MediaControl
node
-------------------------------------------------------------*/
bool ClockReference :: IsUnitPlayable (double dTS)
{
	return m_pControl == NULL || m_pControl ->IsUnitPlayable (dTS);
}


void ClockReference :: CheckForInlineOD()
{
	m_pInlineOD = NULL;
	for (int i=0; i<m_AttachedODs.GetLength(); i++) {
		if (m_AttachedODs[i]->GetInline()->GetOD() == m_AttachedODs[i]) {
			m_pInlineOD = m_AttachedODs[i];
		}
	}
}

void ClockReference :: RegisterWithOD(ObjectDescriptorImp *od)
{
	if (m_AttachedODs.Find(od)<0) {
		m_AttachedODs += od;
		CheckForInlineOD();
	}
}

void ClockReference :: UnregisterWithOD(ObjectDescriptorImp *od)
{
	if (m_AttachedODs.Find(od)>=0) {
		m_AttachedODs -= od;
		if (od==m_pInlineOD) m_pInlineOD=NULL;
	}
}


/*-------------------------------------------------------------
The destructor of a clock namespace deletes all the clocks in that space
-------------------------------------------------------------*/
ClockNamespace :: ~ClockNamespace ()
{
	for (int i = m_clocks.GetLength (); --i >= 0;)
		delete m_clocks [i];
}

/*-------------------------------------------------------------
Find a ClockReference object in a namespace
-------------------------------------------------------------*/
ClockReference *ClockNamespace :: GetClock (int nClockID)
{
	for (int i = m_clocks.GetLength (); --i >= 0;)
		if (m_clocks [i] ->GetID () == nClockID)
			return m_clocks [i];
	return NULL;
}

/*-------------------------------------------------------------
Given a stream ID, find a ClockReference object (or create one) in the given
namespace. If nHasOCR >= 0 set the hasOCR field of that object
-------------------------------------------------------------*/
ClockReference *ClockNamespace :: AttachClock (int nClockID, int nHasOCR)
{
	ClockReference *pClock = GetClock (nClockID);
	if (pClock == NULL)
		m_clocks += pClock = new ClockReference (nClockID);

	if (nHasOCR >= 0)
		pClock -> SetOCRFlag (nHasOCR);
	return pClock;
}

/*-------------------------------------------------------------
Pause all the clocks
-------------------------------------------------------------*/
void ClockNamespace::Pause()
{
	for (int i = m_clocks.GetLength (); --i >= 0;)
		m_clocks [i] -> Pause ();
}

/*-------------------------------------------------------------
Resume all the clocks
-------------------------------------------------------------*/
void ClockNamespace::Resume()
{
	for (int i = m_clocks.GetLength (); --i >= 0;)
		m_clocks [i] -> Resume ();

}
