/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Decoder/ BifsEncoder
FILE:       Common.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Functions which are common to the decoder and the encoder
-------------------------------------------------------------------------*/
#include "defs.h"
#include <assert.h>
#include "MediaObj.h"
#include "NodeFld.h"
#include "bifs.h"
#include "Proto.h"
#include "SysCoder.h"
#include "ScriptImp.h" // BF Script
#include "NodeCode.h"  // BF Script

/*-----------------------------------------------------------------
Find address of a field in the node given its BifsTable entry
-----------------------------------------------------------------*/
NodeField *MediaObject :: GetField (const BifsFieldTable *pEntry)
{
	/*-------------------------------------------------------------
	In a regular node the m_offset field is the actual offset of the field in the
	MediaObject object. In a Proto, this is the negative of the field index.
	-------------------------------------------------------------*/
	if (pEntry ->m_offset > 0)
		return (NodeField *) (((char *) this) + pEntry -> m_offset);
	else
	{
		if (GetNodeCode() == CODE_Script) // BF Case script for extra fields.
			return ((ScriptImp*) this) -> GetField ( - pEntry->m_offset - 3);
		else
			return ((Proto *) this) -> GetField (-pEntry ->m_offset);
	}
}

/*-----------------------------------------------------------------
Find address of a field in the node, given its index
-----------------------------------------------------------------*/
NodeField *MediaObject :: GetField (int nIndex)
{
	return GetField (GetFieldTable () + nIndex);
}

#ifdef MPGJ		// SCENE API Integration
/*-----------------------------------------------------------------
Return the number of fields
-----------------------------------------------------------------*/
int MediaObject :: GetNumFields () //--awm 991209 added for MPEGJ node creation to iterate over fields
{
	return GetFieldCount ();
}
#endif MPGJ		// SCENE API Integration

/*-----------------------------------------------------------------
Find index of a field in the node, given its name
-----------------------------------------------------------------*/
int MediaObject::GetFieldID (const char * fieldName)
{
    for (int i = 0; i < GetFieldCount(); i++) {
		if (!GetFieldTable ()) continue;
		const char *n = GetFieldTable ()[i].m_szName;
		/*JLF: ignore case for script tolerance...*/
        if (n && !stricmp(n, fieldName)) {
            return i;
        }
    }
    return -1;
}

/*-------------------------------------------------------------
Find the offset of the 'children' field in this node, if any
-------------------------------------------------------------*/
int MediaObject :: FindChildrenOffset (
	const BifsFieldTable *pFieldTable,
	int nFieldCount
)
{
	for (int i = 0; i < nFieldCount; i++)
		if (pFieldTable [i].m_bChildren)
			return (pFieldTable + i) -> m_offset;
	return -1;
}

/*-----------------------------------------------------------------
SFGenericNode include a static member which is an array of factory
objects which used to instantiate nodes
-----------------------------------------------------------------*/
BifsFactory *SFGenericNode :: m_bifsFactory [MPEG4_NODE_TABLE_LENGTH];

/*-----------------------------------------------------------------
Every SF...Node class has a table of nodes of that type. SFGenericNode
includes all the nodes, therefore GetNodeTableSize() returns the
total amount of nodes defined in MPEG4.
-----------------------------------------------------------------*/
int SFGenericNode :: GetNodeTableSize () const
{
	return MPEG4_TOTAL_NUMBER_OF_NODES;
}

/*-----------------------------------------------------------------
Stores a pointer to a factory object in the factory array. This
function is called from factory object (BifsFactory) constructor.
A factory object for each node is instantiated through the use of the
BIFS_IMPLEMENT_NODE_START macro.
-----------------------------------------------------------------*/
void SFGenericNode :: StoreFactory (int nCode, BifsFactory *pFactory)
{
	assert (nCode >= 0);
	m_bifsFactory [nCode] = pFactory;
}

/*-----------------------------------------------------------------
Use the factory arrays to instantiate a proper node. This is a new
implementation of SFGenericNode :: CreateNode() replacing the old
implementation which used a primitive long switch command.
-----------------------------------------------------------------*/
MediaObject *SFGenericNode :: CreateNode (int nCode)
{
	if (nCode < 0)
		return NULL;
	if (nCode >= MPEG4_NODE_TABLE_LENGTH)
	{	printf ("Invalid node code encountered\n");
		return NULL;
	}
	if (m_bifsFactory [nCode] != NULL)
		return m_bifsFactory [nCode] -> CreateNode ();
	return NULL;
}

/*-------------------------------------------------------------
Proto destructor clean up
-------------------------------------------------------------*/
Proto :: ~Proto ()
{
	RemoveFields ();
}

/*-------------------------------------------------------------
Proto clean up
-------------------------------------------------------------*/
void Proto :: RemoveFields ()
{
    // PORTING to CC 5.3
    int i = 0;
//	for (int i = m_fields.GetLength (); --i >= 0;)
 	for ( i = m_fields.GetLength (); --i >= 0;)
		delete m_fields [i];
	m_fields.Clear ();
#ifdef BIFSDEC
	for (i = m_protoNodes.GetLength (); --i >= 0;)
		m_protoNodes.GetObject (i) ->Remove ();
	m_protoNodes.Remove ();
#endif
}

/*-------------------------------------------------------------
Create a proto field. This involves instantiation of the proper NodeField object,
and adding proper values to the field table for this proto node
-------------------------------------------------------------*/
NodeField *Proto :: CreateInterfaceField (
	SysCoder *pCoder,
	int eventType,
	int fieldType,
	bool isSF,
	LPCSTR szName
)
{
	/*-------------------------------------------------------------
	Add an entry to the BifsFieldTable
	-------------------------------------------------------------*/
	int nOffset = m_fieldTable.GetLength ();
	BifsFieldTable &rEntry = ++m_fieldTable;
	rEntry.m_AnimationType = -1;
	rEntry.m_bChildren = false;
	rEntry.m_dynID = -1;
	rEntry.m_defID = rEntry.m_inID = rEntry.m_outID = -1;
	rEntry.m_fMax [0] = FLT_MAX;
	rEntry.m_fMin [0] = -FLT_MAX;
	rEntry.m_function = NULL;
	rEntry.m_nDimension = 0;
	rEntry.m_nQuantizationType = -1;
	rEntry.m_offset = -nOffset;
	rEntry.m_quantParams = -1;
	if (szName != NULL)
		rEntry.m_szName = m_fieldNames += szName;

	/*-------------------------------------------------------------
	Increment counter in the BifsBounds structure
	-------------------------------------------------------------*/
	pCoder;
	m_bifsBounds.m_nTotalAll++;
	switch (eventType) {
	case ET_FIELD:
		rEntry.m_defID = m_bifsBounds.m_nTotalDefs++;
		TRACEF ("field\t\t\t", 0);
		break;
	case ET_EXPOSED_FIELD:
		rEntry.m_defID = m_bifsBounds.m_nTotalDefs++;
		rEntry.m_inID = m_bifsBounds.m_nTotalIns++;
		rEntry.m_outID = m_bifsBounds.m_nTotalOuts++;
		TRACEF ("exposedField\t", 0);
		break;
	case ET_EVENT_IN:
		rEntry.m_inID = m_bifsBounds.m_nTotalIns++;
		TRACEF ("eventIn\t\t\t", 0);
		break;
	case ET_EVENT_OUT:
		rEntry.m_outID = m_bifsBounds.m_nTotalOuts++;
		TRACEF ("eventOut\t\t", 0);
		break;
	}

	/*-------------------------------------------------------------
	Instantiate the proper field object and add it to the list
	-------------------------------------------------------------*/
	NodeField *pField = NULL;
	switch (isSF) {
	case true:
		switch (fieldType) {
		case FT_BOOL:
			pField = new SFBool;
			TRACEF ("SFBool\t", 0);
			break;
		case FT_FLOAT:
			pField = new SF<float>;
			TRACEF ("SFFloat\t", 0);
			break;
		case FT_TIME:
			pField = new SF<Time>;
			TRACEF ("SFTime\t", 0);
			break;
		case FT_INT32:
			pField = new SF<int>;
			TRACEF ("SFInt32\t", 0);
			break;
		case FT_STRING:
			pField = new SFString;
			TRACEF ("SFString", 0);
			break;
		case FT_VEC3F:
			pField = new SF<Vec3f>;
			TRACEF ("SFVec3f\t", 0);
			break;
		case FT_VEC2F:
			pField = new SF<Vec2f>;
			TRACEF ("SFVec2f\t", 0);
			break;
		case FT_COLOR:
			pField = new SF<Color3f>;
			TRACEF ("SFColor\t", 0);
			break;
		case FT_ROTATION:
			pField = new SF<Rotation>;
			TRACEF ("SFRotation", 0);
			break;
		case FT_IMAGE:
			pField = new SF<Image>;
			TRACEF ("SFImage\t", 0);
			break;
		case FT_NODE:
			pField = new SFWorldNode;
			TRACEF ("SFNode\t", 0);
			break;
		default:
			return NULL;
		}
		break;
	case false:
		switch (fieldType) {
		case FT_BOOL:
			pField = new MF<Bool>;
			TRACEF ("MFBool\t", 0);
			break;
		case FT_FLOAT:
			pField = new MF<float>;
			TRACEF ("MFFloat\t", 0);
			break;
		case FT_TIME:
			pField = new MF<Time>;
			TRACEF ("MFTime\t", 0);
			break;
		case FT_INT32:
			pField = new MF<int>;
			TRACEF ("MFInt32\t", 0);
			break;
		case FT_STRING:
			pField = new MF<ZString>;
			TRACEF ("MFString", 0);
			break;
		case FT_VEC3F:
			pField = new MF<Vec3f>;
			TRACEF ("MFVec3f\t", 0);
			break;
		case FT_VEC2F:
			pField = new MF<Vec2f>;
			TRACEF ("MFVec2f\t", 0);
			break;
		case FT_COLOR:
			pField = new MF<Color3f>;
			break;
		case FT_ROTATION:
			pField = new MF<Rotation>;
			TRACEF ("MFRotation", 0);
			break;
		case FT_IMAGE:
			pField = new MF<Image>;
			TRACEF ("MFImage\t", 0);
			break;
		case FT_NODE:
			pField = new MFWorldNode;
			TRACEF ("MFNode\t", 0);	//AC20011102 was TRACE
			break;
		default:
			return NULL;
		}
	}
	m_fields += pField;
	return pField;
}

/*-------------------------------------------------------------
Static variables that are used to instantiate the type of fields
-------------------------------------------------------------*/
FIELD_TYPE SF<Bool> ::  s_fieldType = SFBOOL;
FIELD_TYPE MF<Bool> ::  s_fieldType = MFBOOL;
FIELD_TYPE SF<Time> ::  s_fieldType = SFTIME;
FIELD_TYPE MF<Time> ::  s_fieldType = MFTIME;
FIELD_TYPE SF<Color3f> ::  s_fieldType = SFCOLOR;
FIELD_TYPE MF<Color3f> ::  s_fieldType = MFCOLOR;
FIELD_TYPE SF<float> ::  s_fieldType = SFFLOAT;
FIELD_TYPE MF<float> ::  s_fieldType = MFFLOAT;
FIELD_TYPE SF<int> ::  s_fieldType = SFINT32;
FIELD_TYPE MF<int> ::  s_fieldType = MFINT32;
FIELD_TYPE SF<Rotation> ::  s_fieldType = SFROTATION;
FIELD_TYPE MF<Rotation> ::  s_fieldType = MFROTATION;
FIELD_TYPE SF<ZString> ::  s_fieldType = SFSTRING;
FIELD_TYPE MF<ZString> ::  s_fieldType = MFSTRING;
FIELD_TYPE SF<Vec2f> ::  s_fieldType = SFVEC2F;
FIELD_TYPE MF<Vec2f> ::  s_fieldType = MFVEC2F;
FIELD_TYPE SF<Vec3f> ::  s_fieldType = SFVEC3F;
FIELD_TYPE MF<Vec3f> ::  s_fieldType = MFVEC3F;
FIELD_TYPE SF<Url> :: s_fieldType = SFURL;
FIELD_TYPE MF<Url> :: s_fieldType = MFURL;
//FIELD_TYPE SF<MediaObject*> :: s_fieldType = SFNODE;
//FIELD_TYPE MF<MediaObject*> :: s_fieldType = MFNODE;
FIELD_TYPE SF<Image> :: s_fieldType = SFIMAGE;
FIELD_TYPE MF<Image> :: s_fieldType = MFIMAGE;
FIELD_TYPE SF<Vec4f> :: s_fieldType = SFVEC4F;
FIELD_TYPE MF<Vec4f> :: s_fieldType = MFVEC4F;
