/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) and edited by Mauro Rossi (CSELT), Guido Franceschini
(CSELT) and Michael Stepping (FTK)
in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Adaptation Layer
FILE:       DataChan.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of DataChannel object
-------------------------------------------------------------------------*/
#include "defs.h"
#include "Channel.h"
#include "Stream.h"
#include "Except.h"
#include "ODP.h"
#include "AL.h"
#include "Executiv.h"
#include "MediaControl.h"
//AC:20021009 - SL extension : odtags  -- begin
#include "odtags.h"
//AC:20021009 - SL extension : odtags  -- end

/*-------------------------------------------------------------
DataChannel constructor
-------------------------------------------------------------*/
DataChannel :: DataChannel (ALManager *pAL, int nESid, ES_Descriptor *pESD, bool bBackChannel) :
	m_pAL (pAL),
	m_nESid (nESid),
	m_nChannelID (0xffffffff),
	m_bStreamOpen (false),
	m_state (CS_IDLE),
	m_pStream (NULL),
	m_pConfig (NULL),
	m_nEncodedAULength (-1),
	m_uuLength (0),
	m_pFile(NULL),
	m_bBackChannel (bBackChannel),
	m_ODDecoder (
		pAL -> GetExecutive (),
		NULL,
		pAL -> GetExecutive () -> GetTraceFile (),
		DetermineTitle (nESid),
		DetermineFlags (pAL -> GetExecutive () -> GetTraceFlags ()),
		0
	)
{
	SetESDParams (pESD);
	m_Duration = -1;
	m_bPrerolled = false;
}

/*-----------------------------------------------------------------
Unregister the channel from its service.
Close the stream before deleting the object.
-----------------------------------------------------------------*/
DataChannel :: ~DataChannel ()
{
	Lock ();
	bool bOpen = m_state >= CS_READY && m_state <= CS_OPEN;
	Unlock ();		// calling ChannelStop while object is locked might cause a deadlock
	if (bOpen)
		ChannelStop ();
	Lock ();
	Close ();
	Unlock ();
	m_pAL -> DeleteChannel (this);
}

/*-----------------------------------------------------------------
Config() binds a MediaStream, copy the config params,
and store values of predefined parameters
-----------------------------------------------------------------*/
//AC:20021009 - SL extension : AbstractSLConfig  -- begin
void DataChannel :: Config (
	MediaStreamImp *pStream,
	/*SLConfigDescriptor*/AbstractSLConfig *pConfigParams,
	int nExpectedAULength,
	ClockReference *pClock
)
//AC:20021009 - SL extension : AbstractSLConfig  -- end
{
	m_pStream = pStream;
	m_pClock = pClock;
	m_nExpectedAULength = nExpectedAULength;

	m_pConfig = pConfigParams;
	m_config.predefined = pConfigParams -> predefined;
	m_config.useAccessUnitStartFlag = (bool) pConfigParams -> useAccessUnitStartFlag;
	m_config.useAccessUnitEndFlag = (bool) pConfigParams -> useAccessUnitEndFlag;
	m_config.useRandomAccessPointFlag = (bool) pConfigParams -> useRandomAccessPointFlag;
	m_config.hasRandomAccessUnitsOnlyFlag = (bool) pConfigParams -> hasRandomAccessUnitsOnlyFlag;
	m_config.usePaddingFlag = (bool) pConfigParams -> usePaddingFlag;
	m_config.useTimeStampsFlag = (bool) pConfigParams -> useTimeStampsFlag;
	m_config.useIdleFlag = (bool) pConfigParams -> useIdleFlag;
	m_config.durationFlag = (bool) pConfigParams -> durationFlag;
	m_config.timeStampResolution = pConfigParams -> timeStampResolution;
	m_config.OCRResolution = pConfigParams -> OCRResolution;
	m_config.timeStampLength = pConfigParams -> timeStampLength;
	m_config.OCRLength = pConfigParams -> OCRLength;
	m_config.AU_Length = pConfigParams -> AU_Length;
	m_config.instantBitrateLength = pConfigParams -> instantBitrateLength;
	m_config.degradationPriorityLength = pConfigParams -> degradationPriorityLength;
	m_config.AU_seqNumLength = pConfigParams -> AU_seqNumLength;
	m_config.packetSeqNumLength = pConfigParams -> packetSeqNumLength;
	m_config.timeScale = pConfigParams -> timeScale;
	m_config.accessUnitDuration = pConfigParams -> accessUnitDuration;
	m_config.compositionUnitDuration = pConfigParams -> compositionUnitDuration;
	m_config.startDecodingTimeStamp = pConfigParams -> startDecodingTimeStamp;
	m_config.startCompositionTimeStamp = pConfigParams -> startCompositionTimeStamp;

	m_fOCRResolutionFactor = 1000. / m_config.OCRResolution;
	m_fTSResolutionFactor = 1000. / m_config.timeStampResolution;
	m_dwPDUSeqNumMask =
		((DWORD) -1) >> (sizeof (DWORD) * 8 - m_config.packetSeqNumLength);
	m_dwAUSeqNumMask =
		((DWORD) -1) >> (sizeof (DWORD) * 8 - m_config.AU_seqNumLength);
	m_bIsOCRStream = m_pClock && m_nESid == m_pClock ->GetID () && m_config.OCRLength > 0;
	m_state = CS_CONFIGURED;
}

/*-----------------------------------------------------------------
Channel is ready for input. A request to the service to add this channels
will be sent after all channels required at this moment will setup.
-----------------------------------------------------------------*/
void DataChannel :: ChannelReady ()
{
	Lock ();
	m_pBuffer = NULL;
	m_dwPreCTS = m_dwCTS = m_dwPreDTS = m_dwDTS = m_dwTimeOffset = m_dwPreOCR = 0;
	m_nAULength = m_nPreSN = m_nAUnumber = 0;
	m_bWaitRandomAccessPoint = m_bNewAU = true;
	m_nExpectedPDUSequenceNumber = m_nExpectedAUSequenceNumber = 
		m_nEncodedAULength = -1;
	m_bOCRInit = false;

	/*-------------------------------------------------------------
	Since timestamps and OCRs are received as residual values, we don't know
	what the real value of the first OCR is. This is not important, since
	sync is achieved by comparing the timestamps to the OCRs. We assume that
	the first OCR offset is OCRLength^2. This way, the following case
	should work well:
	- OCRLength and TimeStampLength are 16 bits.
	- The player logs at a point when the real OCR value is 0x40002.
	- The a stream contains a timestamp whose real value is 0x3fff7.
	- The value of the first OCR parsed by the player is 0x0002.
	- The value of the first timestamp parsed by the player is 0xfff7.
	- Since we assume initial OCR offset of 0x10000, we assume the first
	  OCR is 0x10002.
	- When we see the first timestamp, we can assume its real value is
	  0xfff7, and sync will work well.
	- If the first parsed timestamp is, for instance, 0x0009, we will
	  conclude that its real value is 0x10009, ans sync will still work
	  well.
	-------------------------------------------------------------*/
	m_dwOCROffset = 1 << m_config.OCRLength;
	/*-------------------------------------------------------------
	GUIDO 14/6/00: 32 bits shift is wrong
	-------------------------------------------------------------*/
	if (m_config.OCRLength == 32)
		m_dwOCROffset = 0;

	if (m_pStream)
	{	m_bStreamOpen = true;
		m_pStream -> Open ();
	}
	m_state = CS_READY;
	Unlock ();
}

/*-----------------------------------------------------------------
Called from Service when channel is open. Add it to the list of open channels
and send a PLAY command
-----------------------------------------------------------------*/
void DataChannel :: OnChannelOpen (UINT nChannelID, RESPONSE eResponse)
{
	if (eResponse == RESPONSE_OK)
	{	if (m_state == CS_REQUESTED)
		{	m_nChannelID = nChannelID;
			m_state = CS_OPEN;
			if (m_pClock) {
				m_pClock -> Start ();
				/*if stream owns the clock, pause until we get an AU*/
				if (m_bIsOCRStream || (m_nESid == m_pClock ->GetID()) ) {
					m_pClock -> Pause();
				}
			}
			m_pAL ->GetExecutive () -> AddChannel (this);
			if (!m_bBackChannel)
				Play (m_bPrerolled);

			m_bPrerolled = FALSE;
		}
	}
	else
	{	m_pAL ->GetExecutive () -> Error (CHANNEL_ADD_FAILED);
		Close ();
		m_state = CS_UNAVAILABLE;
	}
}

/*-----------------------------------------------------------------
Get an SL-PDU (i.e. SL header + payload), parse the SL header,
interpret SL information and post the payload to the stream
-----------------------------------------------------------------*/
void DataChannel :: OnPacketReceived (
	const BYTE *pStreamData,
	int nStreamDataLength,
	const BYTE *pAppData,
	int nAppDataLength,
	int nErr
)
{

	nAppDataLength;		// avoid warning in release build
	/*-------------------------------------------------------------
	Erroneous packet is skipped. If the stream has no in-band integrity
	mechanism (i.e. PDU_seqNum), we need to wait for RAP.
	-------------------------------------------------------------*/
	if (nErr)
	{	/*-------------------------------------------------------------
		Temporary: error code with no data is considered end of stream
		-------------------------------------------------------------*/
		if (pStreamData == NULL) 
		{
			/*close stream but don't stop clock since other active streams may still use it*/
			CloseStream();
			return;
		}
		if (m_config.packetSeqNumLength == 0)
		{	SkipTillRAP ();
			m_pAL -> GetExecutive () -> Error (DATA_TRANSMISSION_ERROR);
		}
		return;
	}

	/*-------------------------------------------------------------
	Parse the packet header, or copy it from 'pAppData'
	-------------------------------------------------------------*/
	SL_PacketHeader header (m_bNewAU);
	if (pAppData != NULL)
	//{	assert (nAppDataLength == sizeof SLHeader);
 	{	assert (nAppDataLength == sizeof(SLHeader) ); // SOLARIS PORTING - use sizeof()
		SLHeader *pHeader = ((SLHeader *) pAppData);
		header.accessUnitStartFlag = pHeader -> accessUnitStartFlag;
		header.accessUnitEndFlag = pHeader -> accessUnitEndFlag;
		header.OCRflag = pHeader -> OCRflag;
		header.idleFlag = pHeader -> idleFlag;
		header.paddingFlag = pHeader -> paddingFlag;
		header.paddingBits = pHeader -> paddingBits;
		header.packetSequenceNumber = pHeader -> packetSequenceNumber;
		header.objectClockReference = pHeader -> objectClockReference;
		header.randomAccessPointFlag = pHeader -> randomAccessPointFlag;
		header.AU_sequenceNumber = pHeader -> AU_sequenceNumber;
		header.decodingTimeStampFlag = pHeader -> decodingTimeStampFlag;
		header.compositionTimeStampFlag = pHeader -> compositionTimeStampFlag;
		header.instantBitrateFlag = pHeader -> instantBitrateFlag;
		header.decodingTimeStamp = pHeader -> decodingTimeStamp;
		header.compositionTimeStamp = pHeader -> compositionTimeStamp;
		header.accessUnitLength = pHeader -> accessUnitLength;
		header.instantBitrate = pHeader -> instantBitrate;
		header.degradationPriority = pHeader -> degradationPriority;
	}
	else
	{	m_ODDecoder.Reset (pStreamData, nStreamDataLength);
		header.SetLength (nStreamDataLength * 8);
		header.EncDec (&m_ODDecoder, DECODE, m_pConfig);
		int nHeaderLength = nStreamDataLength - header.GetLength () / 8;
		pStreamData += nHeaderLength;
		nStreamDataLength -= nHeaderLength;
	}

	if (header.idleFlag ||
		(header.paddingFlag && header.paddingBits == 0)
	)
		return;

	/*-------------------------------------------------------------
	Get the OCR and update the clock
	-------------------------------------------------------------*/
	if (header.OCRflag && m_bIsOCRStream)
	{	DWORD dwOCR = header.objectClockReference;
		/*-------------------------------------------------------------
		Guido 14/06/00: take into account initial values of OCR that
		are close to its maximum permitted range
		-------------------------------------------------------------*/
		if (!m_bOCRInit && (dwOCR > m_dwOCROffset/2))
			m_dwOCROffset -= 1 << m_config.OCRLength;

		if (dwOCR + m_dwOCROffset < m_dwPreOCR)
			m_dwOCROffset += 1 << m_config.OCRLength;
		m_dwPreOCR = dwOCR + m_dwOCROffset;
		dwOCR = m_dwPreOCR * m_fOCRResolutionFactor;
		if (m_pClock)
			m_pClock -> SetTime (dwOCR, m_config.OCRLength);
	}

	/*-------------------------------------------------------------
	Now process the AU
	-------------------------------------------------------------*/
	if (m_config.hasRandomAccessUnitsOnlyFlag)
		header.randomAccessPointFlag = header.accessUnitStartFlag;
	if (header.accessUnitStartFlag)
	{
		m_nAUsequenceNumber = header.AU_sequenceNumber;
		/*-------------------------------------------------------------
		Calculate the actual composition time.
		-------------------------------------------------------------*/
		if (header.compositionTimeStampFlag)
		{	DWORD dwCTS = header.compositionTimeStamp;
			DWORD dwDTS = dwCTS;
			if (header.decodingTimeStampFlag)
				dwDTS = header.decodingTimeStamp;
			if (dwDTS + m_dwTimeOffset < m_dwPreDTS )
				if (m_config.timeStampLength < 32)
					m_dwTimeOffset += 1 << m_config.timeStampLength;

			/*-------------------------------------------------------------
			If hasn't been done before, compare the TS to the OCR in
			order to find the TS initial offset.
			-------------------------------------------------------------*/
			if (!m_bOCRInit && m_pClock && m_pClock -> IsOCRInit () ) {
					m_dwTimeOffset = m_pClock -> AdjustTimeOffset (
											dwDTS,
											m_config.timeStampLength,
											m_fTSResolutionFactor
										 );
					m_bOCRInit = true;
			}

			m_dwPreCTS = dwCTS + m_dwTimeOffset;
			m_dwPreDTS = dwDTS + m_dwTimeOffset;
			if (m_dwPreCTS < m_dwPreDTS)
				m_dwPreCTS += 1 << m_config.timeStampLength;
			m_dwCTS = m_dwPreCTS * m_fTSResolutionFactor;
			m_dwDTS = m_dwPreDTS * m_fTSResolutionFactor;
		}
		else
		{	
			/*-------------------------------------------------------------
			All times are calculated relative to the first receive
			OCR.
			-------------------------------------------------------------*/
			if (!m_bOCRInit && m_pClock && m_pClock -> IsOCRInit ())
			{	m_dwPreCTS = m_pClock -> AdjustTimeOffset (0, 0, 1);
				m_bOCRInit = true;
			}
			/*-------------------------------------------------------------
			If timestamp is not included, calculate it out of
			compositionUnitRate
			-------------------------------------------------------------*/
			if (m_config.AU_seqNumLength == 0)
			{	m_dwCTS = m_dwPreCTS;
				m_dwPreCTS += m_config.compositionUnitDuration;
			}
			else
			{	/*-------------------------------------------------------------
				If AUs are counted, calculate the CTS according to the SN
				-------------------------------------------------------------*/
				int n = m_nAUsequenceNumber - m_nPreSN;
				if (n < 0)
					n += m_dwAUSeqNumMask + 1;
				m_nPreSN = m_nAUsequenceNumber;
				m_dwCTS = m_dwPreCTS +=
					n * m_config.compositionUnitDuration;
			}
			m_dwCTS *= m_fTSResolutionFactor;
		}

		/*-------------------------------------------------------------
		If the OCR hasn't been initialized yet, try to initialize it from the CTS. This
		will work if no OCR stream is expected for the clock
		ONLY if stream owns the clock
		-------------------------------------------------------------*/		
		if (!m_bOCRInit && m_pClock && (m_nESid == m_pClock ->GetID()) ) {
			m_bOCRInit = m_pClock -> InitFromCTS (m_dwCTS);
			if (m_bOCRInit) m_pClock ->Resume();

		}

		/*-------------------------------------------------------------
		If clock reference has not been initialized yet, the real CTS of this frame
		cannot be reconstructed and frame should not be displayed. This is signaled
		by setting CTS to zero. If however zero is the real value of the CTS, we change it
		to 1 to avoid skipping this frame
		-------------------------------------------------------------*/
		if (!m_bOCRInit)
			m_dwCTS += m_dwOCROffset;
		/*-------------------------------------------------------------
		GUIDO 14/6/00: the line below is strictly true, however we can guess the CTS
		and be almost sure that we are correct
		-------------------------------------------------------------*/
//			m_dwCTS = 0;
		else if (m_dwCTS == 0)
			m_dwCTS = 1;
		if (m_config.AU_Length > 0)
			m_nEncodedAULength = header.accessUnitLength;
	}
	
	/*-------------------------------------------------------------
	Skip AU if a repeated AU sequence number
	-------------------------------------------------------------*/
	if (m_nExpectedAUSequenceNumber == 
		(int) ((m_nAUsequenceNumber + 1) & m_dwAUSeqNumMask)
	)
		return;

	/*-------------------------------------------------------------
	Skip AU if looking for a RAP and this is not a RAP
	-------------------------------------------------------------*/
	if (m_bWaitRandomAccessPoint &&
		(!header.accessUnitStartFlag || !header.randomAccessPointFlag)
		) {
		return;
	}

	/*-----------------------------------------------------------------
	Calculate the AU length
	-----------------------------------------------------------------*/
	int nOldLength = m_nAULength;
	m_nAULength += nStreamDataLength;
	bool bAUEnd = header.accessUnitEndFlag != 0;
	if (!m_config.useAccessUnitEndFlag &&
		m_nEncodedAULength == m_nAULength
	)
		bAUEnd = true;

	/*-----------------------------------------------------------------
	Integrity checks
	-----------------------------------------------------------------*/
	if (m_nExpectedPDUSequenceNumber >= 0 &&
		m_nExpectedPDUSequenceNumber != header.packetSequenceNumber ||
		m_nExpectedAUSequenceNumber >= 0 &&
		m_nExpectedAUSequenceNumber != m_nAUsequenceNumber ||
		m_bNewAU && !header.accessUnitStartFlag ||
		!m_bNewAU && header.accessUnitStartFlag ||
		bAUEnd && m_nEncodedAULength >= 0 &&
		m_nEncodedAULength != m_nAULength
	)
	{
		/*-----------------------------------------------------------------
		If this PDU is itself a RAP then use it, else skip to next RAP
		-----------------------------------------------------------------*/
//		m_pAL -> GetExecutive () -> Error (DATA_TRANSMISSION_ERROR);
		if (header.accessUnitStartFlag && header.randomAccessPointFlag)
		{
			nOldLength = 0;
			m_nAULength = nStreamDataLength;
		}
		else
		{
			SkipTillRAP ();
			if (m_nExpectedAUSequenceNumber >= 0 &&
				m_nExpectedAUSequenceNumber != m_nAUsequenceNumber
			)
			{	m_nExpectedAUSequenceNumber = -1;
				m_bWaitRandomAccessPoint = true;
			}
			return;
		}
	}
	//AC:20021009 - SL extension -- begin
	/*-----------------------------------------------------------------
	SL Extension.
	-----------------------------------------------------------------*/
	int ESID = GetESid();
	if (header.decodingTimeStampFlag)
	{
		m_pAL->AddDTS(ESID,(int)header.decodingTimeStamp);
	}
	else if (header.compositionTimeStampFlag)
	{
		m_pAL->AddDTS(ESID,(int)header.compositionTimeStamp);
	}


	if (m_pConfig->GetTag()==TAG_ExtSLConfigDescr)
	{		
		ExtSLConfigDescr *  extSL	= (ExtSLConfigDescr *) m_pConfig;
		int					nCount	= extSL->slextDescr.GetLength();
		bool	bCheckDependency	= false;
		int		bOK					= 0;
		for (int i=0; i<nCount; i++)
		{
			switch (extSL->slextDescr[i]->GetTag())
			{
					/*-----------------------------------------------------------------
					TAG_MarkerDescriptor : store the marker.
					-----------------------------------------------------------------*/
					case TAG_MarkerDescriptor:
					{
						//MarkerDescriptor * markerDescr = (MarkerDescriptor *)extSL->slextDescr[i];
						m_pAL->AddMarker(ESID,(int)header.slExtensions[i]);
						break;
					}
					/*-----------------------------------------------------------------
					TAG_DescriptorPointer : check dependency
					-----------------------------------------------------------------*/
					case TAG_DependencyPointer:	
					{
						if (bOK>=0)
						{
							DependencyPointer * depPointer	= (DependencyPointer *)extSL->slextDescr[i];
							int	dESID = (int)depPointer->ESID;
							bOK	= m_pAL->CheckDependency(ESID,dESID,(int)depPointer->mode,(int)header.slExtensions[i]);
							bCheckDependency = true;
							if (bOK<0) SkipTillRAP ();
							break;
						}
					}
			}//switch
		}//for
		if (bOK<0) return;
	}//TAG_ExtSLConfigDescr

	//AC:20021009 - SL extension -- end
	if (m_config.packetSeqNumLength > 0)
		m_nExpectedPDUSequenceNumber =
			header.packetSequenceNumber + 1 & m_dwPDUSeqNumMask;
	if (bAUEnd && m_config.AU_seqNumLength > 0)
		m_nExpectedAUSequenceNumber = ++m_nAUsequenceNumber & m_dwAUSeqNumMask;

	m_bWaitRandomAccessPoint = m_bNewAU = false;

	/*-----------------------------------------------------------------
	If this is a new AU, allocate buffer on the stream, otherwise,
	reallocate if necessary.
	If this is AU end, we know exactly how much to allocate. Otherwise,
	we use the size recommended by the decoder.
	-----------------------------------------------------------------*/
	if (m_bStreamOpen)
	{	if (m_pBuffer == NULL) 
			m_pBuffer = (LPBYTE) m_pStream -> Allocate (
				(bAUEnd || (m_nAULength > m_nExpectedAULength)) ?
				m_nAULength : m_nExpectedAULength
			);
		else if (m_nAULength > m_nExpectedAULength)
			m_pBuffer = (LPBYTE) m_pStream -> Reallocate (m_nAULength);

		if (m_pBuffer == END_OF_STREAM)
		{	m_pBuffer = NULL;
			CloseStream ();
			m_nAULength = 0;
		}
		else if (m_pBuffer == NULL)
		{	SkipTillRAP ();
			m_pAL -> GetExecutive () -> Error (BUFFER_TOO_SMALL);
		}
	}
		
	/*-----------------------------------------------------------------
	Copy the AU payload to the buffer
	-----------------------------------------------------------------*/
	if (m_pBuffer)
	{	memcpy (m_pBuffer + nOldLength, pStreamData, nStreamDataLength);

		/*-----------------------------------------------------------------
		If this is end of AU, dispatch the buffer into the MediaStream.
		-----------------------------------------------------------------*/
		if (bAUEnd && m_pStream && m_bStreamOpen)
		{	
			m_pStream -> Dispatch (m_nAULength, m_dwCTS);
		}
	}
	if (bAUEnd)
	{	m_nAULength = 0;
		m_bNewAU = true;
		m_pBuffer = NULL;
	}
}

/*-----------------------------------------------------------------
SkipTillRAP set flags and variables so decoding stops until next
random access point.
-----------------------------------------------------------------*/
void DataChannel :: SkipTillRAP ()
{
	m_nExpectedPDUSequenceNumber = -1;
	m_bNewAU = true;
	if (m_config.AU_seqNumLength == 0)
		m_bWaitRandomAccessPoint = true;
	m_nAULength = 0;
}

/*-----------------------------------------------------------------
Stop a channel from playing
-----------------------------------------------------------------*/
void DataChannel :: ChannelStop ()
{

	//DO NOT lock the channel since the DMIF sender may be holding the mutex
	//if waiting for space in MediaStream

	//also make sure we close the channel before asking for DMIF to delete it
	Close ();
	m_state = CS_CONFIGURED;

	m_pAL -> RequestChannelDelete (this);
}

/*-------------------------------------------------------------
Close the stream attached to the channel
-------------------------------------------------------------*/
void DataChannel :: CloseStream ()
{
	if (m_bStreamOpen)
	{	m_bStreamOpen = false;
		if (m_pStream)
			m_pStream -> Close ();
	}
}

/*-----------------------------------------------------------------
Close the channel without deleting it. This involves closing the
stream and notifying the attached clock.
-----------------------------------------------------------------*/
void DataChannel :: Close ()
{
	CloseStream ();
	if (m_state == CS_OPEN && m_pClock) 
		m_pClock -> Stop (m_bIsOCRStream);

	m_bOCRInit = false;
}

/*-------------------------------------------------------------
When channel is deleted by DMIF, close the attached stream, if exists.
Delete the channel if deleting was requested before, or if opened
preemptively by DMIF
-------------------------------------------------------------*/
void DataChannel :: OnChannelDelete (REASON reason)
{
	Close ();
	m_state = CS_CONFIGURED;
	if (reason != REASON_OK)
	{	m_state = CS_UNAVAILABLE;
		m_pAL ->GetExecutive () ->Error (UNEXPECTED_CHANNEL_DELETE);
	}
	m_pAL ->GetExecutive () ->DeleteChannel (this);
}

/*-------------------------------------------------------------
Channel shall be requested as soon as the service is connected
Return true if channel is indeed ready.
-------------------------------------------------------------*/
bool DataChannel :: Request ()
{
	if (m_state == CS_READY)
	{	m_state = CS_REQUIRED;
		return true;
	}
	return false;
}

/*-------------------------------------------------------------
If channel is required, mark it as requested and return true
-------------------------------------------------------------*/
BOOL DataChannel :: IsRequired ()
{
	if (m_state == CS_REQUIRED)
	{	m_state = CS_REQUESTED;
		return true;
	}
	else
		return false;
}

/*-------------------------------------------------------------
Display a message error in case user command rejected
-------------------------------------------------------------*/
void DataChannel :: OnUserCommandReject (int)
{
	m_pAL ->GetExecutive () ->Error (SEND_COMMAND_FAILED);
}

/*-------------------------------------------------------------
Sort out the flags that control the parse tracing
-------------------------------------------------------------*/
int DataChannel :: DetermineFlags (int flags)
{
	if (!(flags & TRACE_HEADERS))
		flags = TRACE_NONE;
	else if (!(flags & TRACE_VERBOSE))
		flags = TRACE_PARSE;
	return flags;
}

/*-------------------------------------------------------------
Sort out the title for the parse tracing.
-------------------------------------------------------------*/
LPCSTR DataChannel :: DetermineTitle (int nESid)
{
	char szTitle [140];
	sprintf (szTitle, "SL_PacketHeader (%.110s:%d)", m_pAL -> GetURL (), nESid);
#ifdef TRACE_ENCODING
	m_title.Copy (szTitle);
#endif
	return m_title;
}

/*-------------------------------------------------------------
Initialize values that come from the ES_Descriptor, if available.
If a QoSDescriptor exists, it is re-encoded.
[There is a chance here to manipulate the descriptor, if necessary]
-------------------------------------------------------------*/
void DataChannel :: SetESDParams (ES_Descriptor *pESD)
{
	if (pESD != NULL)
	{	if (pESD ->URLstring.GetLength () > 0)
		{	m_uuData = (LPBYTE) (LPCSTR) pESD ->URLstring;
			m_uuLength = pESD->URLstring.GetLength () + 1;
		}
		else
		{	m_esidAsuuData [0] = HIBYTE (m_nESid);
			m_esidAsuuData [1] = LOBYTE (m_nESid);
			m_uuData = m_esidAsuuData;
			m_uuLength = 2;
		}
		if (pESD -> qosDescr != NULL_DESC)
		{	ODCoder coder (NULL, NULL, NULL, 0, 0);
			pESD -> qosDescr -> EncDec (&coder, CALC_LENGTH, NULL);
			pESD -> qosDescr -> EncDec (&coder, ENCODE, NULL);
			m_qos.Assign (coder.GetData (), coder.GetDataLength ());
		}
	}
}

/*-------------------------------------------------------------
Send data on a back channel
-------------------------------------------------------------*/
void DataChannel :: DataSend (LPBYTE pData, int nLength)
{
	if (m_bBackChannel)
		m_pAL -> DataSend (pData, nLength, m_nChannelID, m_nAUnumber++);
}

/*-------------------------------------------------------------
Send a PLAY command to the server. If there is a MediaControl node attached to the clock
refernce associated with this channel, send a command (or commands) according to the URL and
parameters defined by that node
-------------------------------------------------------------*/
void DataChannel :: Play (BOOL bIsPreroll)
{
	if (m_state == CS_OPEN) {
		MediaControlImp *pControl = m_pClock ->GetMediaControl ();
		if (pControl == NULL) {
			double startTime = m_pClock->GetTime();
			startTime /= 1000;
			m_pAL -> StartChannel(m_nChannelID, startTime, -1, 1);
		} else {
			/*set clock speed*/
			if (!bIsPreroll && pControl->mediaSpeed) m_pClock->SetSpeed(pControl->mediaSpeed>0 ? pControl->mediaSpeed : -pControl->mediaSpeed);
			pControl -> Play (m_nChannelID, bIsPreroll);
		}
	}
}

float DataChannel :: GetBufferOccupancy()
{
	if (m_pStream) return m_pStream->GetBufferOccupancy();
	return 0.0;
}
