/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1997.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Buffer management
FILE:       Stream.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of MediaStreamImp.
			(This version simulation only. Just makes sure Allocate,
			Reallocate, and Dispatch are called properly.
-------------------------------------------------------------------------*/
#include "defs.h"
#include <assert.h>
#include "Stream.h"
#include "Im1defs.h"
#include "MediaControl.h"

int MediaStreamImp :: m_nStreamCount;
FILE * MediaStreamImp :: m_fTraceFile;
ZMutex MediaStreamImp :: m_mtxLockTraceFile;

/*-----------------------------------------------------------------
MediaStreamImp constructor. The simulation allocate a buffer large enough
for all possible AUs
-----------------------------------------------------------------*/
MediaStreamImp :: MediaStreamImp (
	ClockReference *pClock,
	int nBufferSize,
	int nThreshold
) :
	m_smpWaitForMemory (1, 0),
	m_smpWaitForData (1, 0),
	m_smpTimer (1, 0)
{
	/*-----------------------------------------------------------------
	Allocate the stream buffer. If nBufferSize < 0, this is dynamic
	buffer, i.e., memory when Allocate or Reallocate are called.
	-----------------------------------------------------------------*/
	m_nTakenStart = m_nBufferSize = nBufferSize;
	m_nThreshold= nThreshold;
	m_nNextFetch = -1;
	m_nUsedBy = m_nFetched = 0;
	m_pBuffer = NULL;
	m_bDynamic = nBufferSize < 0;
	m_bClockActive = true;
	m_bMultipleCopies = false;
	if (!m_bDynamic)
		m_pBuffer = (LPBYTE) malloc (nBufferSize + 3);
	m_nSN = m_nStreamCount++;		// For tracing
	m_pClock = pClock;
	m_bEndOfStream = false;
	Reset ();
	ResetCounters ();

	m_bInitialBuffering = FALSE;

#ifdef MPGJ
	//*** YT START (108)
	// Initialize variables relating to pause and halfclose
	m_bHalfClose = FALSE;
	m_bPause = FALSE;
	m_bLastBlockFetched = FALSE;
//*** YT END (108)
#endif MPGJ

}

/*-----------------------------------------------------------------
Reset the stream
-----------------------------------------------------------------*/
void MediaStreamImp :: Reset ()
{
	/*-------------------------------------------------------------
	Reset member variables
	-------------------------------------------------------------*/
	m_nTakenStart = m_nBufferSize;
	m_pLastBuffer = NULL;
	m_nNextFetch = -1;
	m_nFreeStart = m_nFetched = 0;
	m_nAllocatedLength = -1;
	m_bWaitingForMemory = m_bWaitingForData = false;
	m_bReallocInProgress = m_bEndOfStream = false;
	m_bFirstRetrieved = false;
	m_bTrace = FALSE;
	m_nOpenToReceive = m_nWaiting = 0;
	m_bOpenToSend = m_bMultipleCopies = false;
	m_bClockActive = true;
	m_nBlockCount = 0;

}

/*-----------------------------------------------------------------
Open for receiving (may be opened by more than one peer)
-----------------------------------------------------------------*/
void MediaStreamImp :: Start ()
{
	if (m_nOpenToReceive > 0)
	{	++m_nOpenToReceive;
		return;
	}
	m_smpReceiving.P ();
	m_smpInUse.P ();
	if (m_nUsedBy++ == 0)	// stream is in used if after opened by both sides
	{	Reset ();			// the first who opens resets
		m_smpInUse.V ();
	}
	++m_nOpenToReceive;
}

/*-----------------------------------------------------------------
Open for sending
-----------------------------------------------------------------*/
void MediaStreamImp :: Open ()
{
	m_smpSending.P ();
	m_smpInUse.P ();
	if (m_nUsedBy++ == 0)	// stream is in used if after opened by both sides
	{	Reset ();			// the first who opens resets
		m_smpInUse.V ();
	}
	m_bOpenToSend = true;
}

/*-----------------------------------------------------------------
MediaStreamImp destructor
-----------------------------------------------------------------*/
MediaStreamImp :: ~MediaStreamImp ()
{
	/*-----------------------------------------------------------------
	Wait till both sides close the stream
	-----------------------------------------------------------------*/
	m_smpReceiving.P ();
	m_smpSending.P ();
	m_mtxLockStream.P ();
	free (m_pBuffer);
	m_mtxLockStream.V ();
	m_smpSending.V ();
	m_smpReceiving.V ();

	/*-----------------------------------------------------------------
	Following lines for tracing
	-----------------------------------------------------------------*/
	if (--m_nStreamCount == 0 && m_fTraceFile != NULL)
	{	fclose (m_fTraceFile);
		m_fTraceFile = NULL;
	}
}

/*-------------------------------------------------------------
Calculate how much buffer memory is occupied
-------------------------------------------------------------*/
inline MediaStreamImp :: GetBufferFullness () const
{
	int nAllocated = m_nFreeStart - m_nTakenStart;
	if (nAllocated < 0)
		nAllocated += m_nBufferSize;
	return nAllocated;
}

float MediaStreamImp :: GetBufferOccupancy()
{
	/*dynamic buffers are never completely filled*/
	if (m_bDynamic) return 0.0;
	float size = GetBufferFullness ();
	size /= m_nBufferSize;
	return size;
}

/*-----------------------------------------------------------------
Allocate() allocates a block of nSize byte and returns its address
-----------------------------------------------------------------*/
LPVOID MediaStreamImp :: Allocate (int nSize, BOOL bWait, BOOL bShove)
{
	assert (m_bOpenToSend);

	nSize += sizeof (int) + sizeof (DWORD);	// for block length and time stamp

	/*-----------------------------------------------------------------
	Return NULL if last allocated block has not been dispatched
	-------------------------------------------------------------*/
	if (m_nAllocatedLength >= 0 && !m_bReallocInProgress) {
		return NULL;
	}

	/*-------------------------------------------------------------
	If this is an empty dynamic buffer reallocate buffer if necessary
	-------------------------------------------------------------*/
	if (nSize > m_nBufferSize)
		if (m_bDynamic &&
			m_nFreeStart == 0 && m_nTakenStart == m_nBufferSize
		)
		{	if (!m_pBuffer)
				m_pBuffer = (LPBYTE) malloc (nSize);
			else
				m_pBuffer = (LPBYTE) realloc (m_pBuffer, nSize);
			m_nTakenStart = m_nBufferSize = nSize;
		}
		else
			return NULL;

	/*-----------------------------------------------------------------
	Check available (contiguous) space.
	-----------------------------------------------------------------*/
	m_mtxLockStream.P ();

	#ifdef MPGJ
	//*** YT START (109)
	// if Pause() is called and lastblock is fetched
	// then shutdown upstream(decoder) by returning END_OF_STREAM
	if(m_bPause && m_bLastBlockFetched) {
		// shutdown sender(decoder)
		m_mtxLockStream.V ();
		return END_OF_STREAM;
	}
	else
		m_bLastBlockFetched = true;
	//*** YT END (109)
	#endif MPGJ

	if (m_bEndOfStream)
	{	m_mtxLockStream.V ();
		return END_OF_STREAM;
	}

	while (
		nSize > m_nTakenStart - m_nFreeStart &&
		(	m_nTakenStart > m_nFreeStart ||
			m_nTakenStart == m_nFreeStart && !m_bReallocInProgress ||
			nSize > m_nBufferSize - m_nFreeStart &&
			nSize > m_nTakenStart
		)
	)
	{
		/*-----------------------------------------------------------------
		Not enough space available. One reason might be that the buffer
		is empty, and this is a reallocation of a block which is
		currently at the middle of the buffer
		-----------------------------------------------------------------*/
		if (m_nTakenStart == m_nFreeStart && m_bReallocInProgress)
		{	m_nFreeStart = 0;
			m_nTakenStart = m_nBufferSize;
			m_nNextFetch = -1;
		}
		/*-----------------------------------------------------------------
		If bShove it TRUE, shove last block
		(or wait for a currently fetched block to be discarded).
		-----------------------------------------------------------------*/
#ifdef MPGJ
	else if (bShove || m_bHalfClose)
#else
	else if (bShove)
#endif

		{	if (m_nFetched > 0)
			{	m_mtxLockStream.V ();
				m_smpLockOutBlock.P ();
				m_smpLockOutBlock.V ();
				m_mtxLockStream.P ();
			}
			else
				Discard ();
		}
		/*-----------------------------------------------------------------
		If the wait flag is true, wait till a space is available
		-----------------------------------------------------------------*/
		else if (bWait) {
			/*unlock rebuffering*/
			if (!m_bClockActive && m_bInitialBuffering) {
				m_bInitialBuffering = FALSE;
				m_bClockActive = true;
				m_pClock ->Resume ();
			}
			m_bWaitingForMemory = true;
			m_mtxLockStream.V ();
			m_smpWaitForMemory.P ();
			m_mtxLockStream.P ();

			if (m_bEndOfStream)
			{	m_mtxLockStream.V ();
				return END_OF_STREAM;
			}
		}
		else
		{	m_mtxLockStream.V ();
			return NULL;
		}
	}

	/*-----------------------------------------------------------------
	Allocate a block of the desired length. Starting either at the next
	available byte, or cycle around the buffer
	-----------------------------------------------------------------*/
	if (m_nBufferSize - m_nFreeStart < nSize)
	{	STOREINT (-1, m_pBuffer + m_nFreeStart);	// Mark as not used
		if (m_nTakenStart == m_nFreeStart)
			m_nTakenStart = m_nBufferSize;
		if (m_nNextFetch == m_nFreeStart)
			m_nNextFetch = -1;
		m_nFreeStart = 0;
	}
	m_nAllocatedLength = nSize - (sizeof (int) + sizeof (DWORD));
	m_bWait = bWait;			// save for realloc
	m_bShove = bShove;
	m_mtxLockStream.V ();
	return m_pBuffer + m_nFreeStart + sizeof (int) + sizeof (DWORD);
}

/*-----------------------------------------------------------------
Reallocate() expands the reserved block, if necessary
-----------------------------------------------------------------*/
LPVOID MediaStreamImp :: Reallocate (int nSize)
{
	if (m_bEndOfStream)
		return END_OF_STREAM;
	assert (nSize >= 0);
	assert (m_bOpenToSend);

	/*-----------------------------------------------------------------
	Reallocation is done by zeroing m_nAllocatedLength and calling
	Allocate with the new size.
	-----------------------------------------------------------------*/

	m_mtxLockStream.P ();


	#ifdef MPGJ
	//*** YT START (109)
	if(m_bPause && m_bLastBlockFetched) {
		// shutdown sender(decoder)
		m_mtxLockStream.V ();
		return END_OF_STREAM;
	}
	else
		m_bLastBlockFetched = true;
	//*** YT END (109)
	#endif MPGJ

	int nPreviousStart = m_nFreeStart;
	int nPreviousLength = m_nAllocatedLength;
	LPVOID pResult;

	m_bReallocInProgress = TRUE;
	m_mtxLockStream.V ();
	pResult = Allocate (nSize, m_bWait, m_bShove);
	m_bReallocInProgress = FALSE;

	/*-----------------------------------------------------------------
	If the stream was closed by the receiving side, tell the receiver
	not to wait till block is discarded, and return END_OF_STREAM
	-----------------------------------------------------------------*/
	if (pResult == END_OF_STREAM)
	{	m_nAllocatedLength = -1;
		if (m_bWaitingForData)
		{	m_bWaitingForData = false;
			m_smpWaitForData.V ();
		}
		return END_OF_STREAM;
	}

	/*-----------------------------------------------------------------
	If realloc successful, but block was shifted in memory, copy the
	contents of the old block.
	-----------------------------------------------------------------*/
	if (pResult != NULL && m_nFreeStart != nPreviousStart)
		memcpy (
			pResult,
			m_pBuffer + nPreviousStart + sizeof (int) + sizeof (DWORD),
			nPreviousLength
		);
	return pResult;
}

/*-----------------------------------------------------------------
Dispatch() dispatches the allocated block into the stream
-----------------------------------------------------------------*/
BOOL MediaStreamImp :: Dispatch (int nSize, DWORD dwTime, BOOL /*bEOU*/)
{
	if (m_bEndOfStream) {
		m_nAllocatedLength = -1;
		return FALSE;
	}
	assert (nSize >= 0);
	assert (m_bOpenToSend);

	if (m_nAllocatedLength < 0)
		return FALSE;

	if (nSize==0) {
		m_nAllocatedLength = -1;
		return TRUE;
	}

	/*-----------------------------------------------------------------
	store the size and time information. Signal to waiting threads.
	-----------------------------------------------------------------*/
//	if (!m_bClockSet)
//		SetClock (dwTime);
	m_mtxLockStream.P ();

	nSize += sizeof (int) + sizeof (DWORD);
	STOREINT (nSize, (m_pBuffer + m_nFreeStart));
	STOREDWORD (dwTime, (m_pBuffer + m_nFreeStart + sizeof (int) ) );

	/*-------------------------------------------------------------
	If this is the only unit in a dynamic buffer, trim the buffer to the
	required size
	-------------------------------------------------------------*/
	if (m_bDynamic &&
		m_nFreeStart == 0 && m_nTakenStart == m_nBufferSize &&
		m_nBufferSize > nSize
	)
	{	m_pBuffer = (LPBYTE) realloc (m_pBuffer, nSize);
		m_nTakenStart = m_nBufferSize = nSize;
	}

	if (m_nNextFetch < 0)
		m_nNextFetch = m_nFreeStart;
	m_nFreeStart += nSize;
	if (m_nFreeStart == m_nBufferSize)
		m_nFreeStart = 0;
	if (m_nTakenStart == m_nBufferSize)
		m_nTakenStart = 0;
	m_nAllocatedLength = -1;

	/*-------------------------------------------------------------
	Indicate finish rebuffering, if necessary
	-------------------------------------------------------------*/
	if (!m_bClockActive && !m_bInitialBuffering && GetBufferFullness () >= m_nThreshold)
	{	m_bClockActive = true;
		m_pClock ->Resume ();
	}

	if (m_bWaitingForData && m_bClockActive)
	{	m_bWaitingForData = FALSE;
		m_smpWaitForData.V ();
	}

	#ifdef MPGJ
	//*** YT START (110)
	// if HalfClose() was called
	// as soon as new block is dispached from upstream
	// fetch the last block queued and discard it
	if(m_bHalfClose) {
		m_nNextFetch = GetNext (m_nNextFetch);
		Discard();
	}
	//*** YT END (110)
	#endif MPGJ
	m_mtxLockStream.V ();
	return TRUE;
}

/*-------------------------------------------------------------
If buffer fullness is monitored and buffer is empty pause the clock
-------------------------------------------------------------*/
inline void MediaStreamImp :: CheckThreshold ()
{
	if (m_bClockActive && !m_bEndOfStream && !m_pClock ->HasOCR () &&
		m_nThreshold > 0 && GetBufferFullness () == 0
	)
	{	m_bClockActive = false;
		m_pClock ->Pause ();
	}
	else if (m_bClockActive && m_bInitialBuffering) {
		if (!m_bWaitingForMemory ) {	
			m_bClockActive = false;
			m_pClock ->Pause ();
		} else {
			m_bInitialBuffering = FALSE;
		}
	}
}

/*-----------------------------------------------------------------
FetchNext fetches the next available block, ignoring its time stamp (which
is only returned in rdwTime.  If bWait is true, the function waits till
a block becomes available.
If successful, the function locks the fetched block.
If end of stream reached, the function returns END_OF_STREAM.
-----------------------------------------------------------------*/
LPVOID MediaStreamImp :: FetchNext (
	int &rnSize,
	DWORD &rdwTime,
	BOOL &bEOU,
	BOOL bWait
)
{
	/*-----------------------------------------------------------------
	Check if a loaded block is available
	-----------------------------------------------------------------*/
	m_mtxLockStream.P ();
	if (m_nOpenToReceive == 0 && m_bEndOfStream)
	{	m_mtxLockStream.V ();
		return END_OF_STREAM;
	}
	assert (m_nOpenToReceive > 0);

	bEOU = FALSE;		// Only in this version. Needs correction.
	CheckThreshold ();

	/*clock paused return NULL*/
	if ( (m_bInitialBuffering && !m_bClockActive) || (!bWait && !m_pClock->IsActive()) ){
		m_mtxLockStream.V ();
		return NULL;
	}

#ifdef MPGJ
	//*** YT START (111)
	// when Pause state ( upstream detached ) and last block was fetched
	// return the block in the head of the queue
	if(m_bPause && m_bLastBlockFetched) {
		// block should be available...
		++m_nFetched;
		m_smpLockOutBlock.P ();
		rnSize = GetBlockLength (m_nTakenStart) - sizeof (int) - sizeof (DWORD);
		rdwTime = (long) m_pClock -> GetTime ();
		m_bFirstRetrieved = TRUE;
		LPVOID pVoid = m_pBuffer + sizeof (int) + sizeof (DWORD) + m_nTakenStart;
		m_mtxLockStream.V ();
		return pVoid;
	}
	//*** YT END (111)
#endif MPGJ

	while (
		m_nNextFetch < 0 ||
		m_nNextFetch == m_nFreeStart && m_nAllocatedLength >= 0
	)
	{
		/*-----------------------------------------------------------------
		If end of stream, return END_OF_STREAM
		-----------------------------------------------------------------*/
		if (m_bEndOfStream)
		{	m_mtxLockStream.V ();
			return END_OF_STREAM;
		}
		/*-----------------------------------------------------------------
		If necessary, wait for a block, otherwise return NULL
		-----------------------------------------------------------------*/
		if (bWait)
		{	m_bWaitingForData = TRUE;
			m_mtxLockStream.V ();
			m_smpWaitForData.P ();
			m_mtxLockStream.P ();
		}
		else
		{	m_mtxLockStream.V ();
			return NULL;
		}
	}

	/*-----------------------------------------------------------------
	Lock a block and return its address, size and time stamp
	-----------------------------------------------------------------*/
	if (m_nFetched++ == 0)
		m_smpLockOutBlock.P ();
	m_nLastSize = rnSize =
		GetBlockLength (m_nNextFetch) - sizeof (int) - sizeof (DWORD);
	m_dwLastTime = rdwTime = GetTimeStamp (m_nNextFetch);
	m_pLastBuffer =
		m_pBuffer + sizeof (int) + sizeof (DWORD) + m_nNextFetch;
	m_nNextFetch = GetNext (m_nNextFetch);
	m_mtxLockStream.V ();
	return m_pLastBuffer;
}

/*-----------------------------------------------------------------
Fetch() fetches a block whose time stamp matured.
It will return a non zero value in rnStatus if it's the same block returned on
previous call. -1 is returned in this field when reaching end of stream.
If bSkip is TRUE it will skip all blocks till the newest one whose time
matches.
If bWait is TRUE it will wait till one block matures.
If nFlag is zero the funtions returns END_OF_STREAM, when all availble units have
been processed and the stream has been closed by the sender, otherwise the last unit
is returned again.
A value of -1 in nFlag indicates that this is only a peeping call for enquiring stream
state, and that the returned unit will not be consumed, therefor buffer state shall not
be affected by the call.
-----------------------------------------------------------------*/
LPVOID MediaStreamImp :: Fetch (
	int &rnSize,
	DWORD &rdwTime,
	BOOL &bEOU,
	int &rnStatus,
	BOOL bSkip,
	BOOL bWait,
	long lDelta,
	BOOL nFlag // SOLARIS PORTING - made BOOl again
)
{

	m_mtxLockStream.P ();

	if (m_nOpenToReceive == 0 && m_bEndOfStream)
	{	m_mtxLockStream.V ();
		return END_OF_STREAM;
	}

	assert (m_nOpenToReceive > 0 || nFlag < 0);
	/*-----------------------------------------------------------------
	Returns same block if a block is already fetched
	-----------------------------------------------------------------*/
	if (m_nFetched > 0)
	{	if (nFlag >= 0)
		{	++m_nFetched;
			m_bMultipleCopies = true;
		}
		rnSize = m_nLastSize;
		rdwTime = m_dwLastTime;
		rnStatus = m_nLastStatus == 0 ? 1 : m_nLastStatus;
		m_mtxLockStream.V ();
		return m_pLastBuffer;
	}


	bEOU = FALSE;		// Only in this version. Needs correction.
	/*-----------------------------------------------------------------
	If first block has been already retrieved and bWait is true, discard
	that block.
	-----------------------------------------------------------------*/
	if (nFlag >= 0)
		CheckThreshold ();

	if (m_bInitialBuffering && !m_bClockActive) {
		m_mtxLockStream.V ();
		return NULL;
	}

#ifdef MPGJ
	//*** YT START (111)
	if(m_bPause && m_bLastBlockFetched) {
		// block should be available...
		++m_nFetched;
		m_smpLockOutBlock.P ();
		rnSize = GetBlockLength (m_nTakenStart) - sizeof (int) - sizeof (DWORD);
		rdwTime = (long) m_pClock -> GetTime () + lDelta;
		rnStatus = TRUE;
		m_bFirstRetrieved = TRUE;
		m_mtxLockStream.V ();
		return m_pBuffer + sizeof (int) + sizeof (DWORD) + m_nTakenStart;
	}
	//*** YT END (111)
#endif MPGJ

	if (m_bFirstRetrieved && bWait)
		Discard ();

	/*-----------------------------------------------------------------
	Check if a loaded block is available
	-----------------------------------------------------------------*/
	while (
		m_nTakenStart >= m_nBufferSize ||
		m_nTakenStart == m_nFreeStart && m_nAllocatedLength >= 0
	)
	{
		/*-----------------------------------------------------------------
		If end of stream, return END_OF_STREAM
		-----------------------------------------------------------------*/
		if (m_bEndOfStream)
		{	m_mtxLockStream.V ();
			return END_OF_STREAM;
		}
		/*-----------------------------------------------------------------
		If necessary, wait for a block, otherwise return NULL
		-----------------------------------------------------------------*/
		if (bWait)
		{	m_bWaitingForData = TRUE;
			m_mtxLockStream.V ();
			m_smpWaitForData.P ();
			if (!m_nOpenToReceive) return END_OF_STREAM;
			m_mtxLockStream.P ();
		}
		else
		{	m_mtxLockStream.V ();
			return NULL;
		}
	}

	/*-----------------------------------------------------------------
	Find the next block which matured. If bSkip is TRUE, find the newest
	matured block.
	-----------------------------------------------------------------*/
	int nNext = m_nTakenStart;
	int nFollowingAU = nNext;
	DWORD dwTime;
	DWORD dwCurrentTime = (long) m_pClock -> GetTime () + lDelta;


	/*if not skipping and first is fetched jump to next frame, otherwise start time checking with first frame*/
	if (m_bFirstRetrieved && !bSkip)  nFollowingAU = GetNext (nNext);


	int nStepped = 0;
	int rt;
	int prevDelta = 0;

	while (bSkip && nFollowingAU >= 0) {
		dwTime = GetTimeStamp (nFollowingAU);
		rt = m_pClock -> CompareTimes( dwTime, dwCurrentTime);

		/*current unit too late*/
		if (rt<0) {
			nNext = nFollowingAU;
			int next = GetNext(nFollowingAU);
			if ((next<0) || (next == nFollowingAU)) break;
			if (!GetTimeStamp(next)) break;
			nFollowingAU = next;
			nNext = next;
			++nStepped;
			prevDelta = -1 * rt;
		} 
		/*current unit is exact match*/
		else if (rt==0) {
			nNext = nFollowingAU;
			break;
		}
		/*current unit is ahead of time - if no previous unit get current one*/
		else if (!prevDelta) {
			nNext = nFollowingAU;
			break;
		}
		/*otherwise compare absolute deltas and fetch min*/
		else if ( prevDelta < rt) {
			nStepped--;
			break;
		} else {
			nNext = nFollowingAU;
			break;
		}
	}

	if (nStepped > 1)
		++m_nSkipped;

	/*-----------------------------------------------------------------
	Now nNext points to the most suitable block.
	If it is a matured block, lock it and return its address.
	Otherwise wait or return the previous frame
	-----------------------------------------------------------------*/
	dwTime = GetTimeStamp (nNext);
	int delta = m_pClock -> CompareTimes (dwTime, dwCurrentTime);

	if (delta > 0)
	{	
		if (bWait) {
			++m_nWaiting;
			m_mtxLockStream.V ();
			while (m_nOpenToReceive > 0 && m_pClock -> WaitOn (m_smpTimer, dwTime));
			m_mtxLockStream.P ();
			--m_nWaiting;
		}
		else
		{	if (m_pLastBuffer != NULL)
			{	rnSize = m_nLastSize;
				rdwTime = m_dwLastTime;
				rnStatus = 1;
				++m_nRepeated;
			}
			m_mtxLockStream.V ();
			return m_pLastBuffer;
		}
	}
	else if (delta < 0)
		++m_nLate;

	/*-----------------------------------------------------------------
	Now we can lock the block and return all the required information
	-----------------------------------------------------------------*/
	if (nFlag >= 0)
	{	++m_nFetched;
		m_smpLockOutBlock.P ();
	}

	m_nLastSize = rnSize = GetBlockLength (nNext) - sizeof (int) - sizeof (DWORD);
	m_dwLastTime = rdwTime = GetTimeStamp (nNext);
	m_nLastStatus = rnStatus = nNext == m_nTakenStart && m_bFirstRetrieved;
	if (nFlag >= 0)
	{	m_nNextFetch = GetNext (m_nTakenStart);
		m_bFirstRetrieved = TRUE;
		/*-----------------------------------------------------------------
		Discard old blocks
		-----------------------------------------------------------------*/
		if (nNext != m_nTakenStart)
		{	m_nTakenStart = nNext;
			if (m_bWaitingForMemory)
			{	m_bWaitingForMemory = FALSE;
				m_smpWaitForMemory.V ();
			}
		}
	}
	m_pLastBuffer = m_pBuffer + sizeof (int) + sizeof (DWORD) + m_nTakenStart;
	/*check for end of stream*/
	if (rnStatus > 0 && m_bEndOfStream) {
		if (GetNext(nNext) <0) {
			m_nLastStatus = rnStatus = -1;
			if (nFlag == 0)
				m_pLastBuffer = END_OF_STREAM;
		}
	}

	/*-------------------------------------------------------------
	Count number of repeated frames and new frames
	-------------------------------------------------------------*/
	if (m_nLastStatus >= 0)
	{	m_nRepeated += m_nLastStatus;
		m_nFrames += m_nLastStatus == 0;
	}

	m_mtxLockStream.V ();
	return m_pLastBuffer;
}

/*-----------------------------------------------------------------
Releases a previously fetched output block. If bDiscard is TRUE, or its
time stamp is zero, discard the block from the stream
-----------------------------------------------------------------*/
void MediaStreamImp :: Release (BOOL bDiscard)
{
	m_mtxLockStream.P ();
	/*-----------------------------------------------------------------
	Do nothing if no block is fetched
	-----------------------------------------------------------------*/
	if (m_nFetched > 0)
	{	if (--m_nFetched == 0)
		{	m_smpLockOutBlock.V ();
			m_bMultipleCopies = false;
		}
#ifdef MPGJ
	//*** YT START (112)
	// If puase state, don't discard
	if(!m_bPause)
	//*** YT END (112)
#endif MPGJ
		if (bDiscard && !m_bMultipleCopies)
			Discard ();
	}
	m_mtxLockStream.V ();
}


/*-----------------------------------------------------------------
Close the stream by the sender
-----------------------------------------------------------------*/
void MediaStreamImp :: Close ()
{
	m_mtxLockStream.P ();
	if (m_bOpenToSend == false) {
		m_mtxLockStream.V ();
		return;
	}

	m_bEndOfStream = true;
	m_bInitialBuffering = FALSE;
	if (!m_bClockActive)
	{	m_bClockActive = true;
		m_pClock ->Resume ();
	}
	if (m_bWaitingForData)
	{	m_bWaitingForData = false;
		m_smpWaitForData.V ();
	}
	if (m_bWaitingForMemory)
	{	m_bWaitingForMemory = false;
		m_smpWaitForMemory.V ();
	}
	if (--m_nUsedBy == 0)
		m_smpInUse.V ();	// stream is in use till released by both sides
	m_mtxLockStream.V ();
	assert (m_nSN >= 0);
	m_bOpenToSend = false;
	m_nAllocatedLength = -1;
	m_smpSending.V ();
}

/*-----------------------------------------------------------------
Discard next block
-----------------------------------------------------------------*/
void MediaStreamImp :: Discard ()
{
	int bl = GetBlockLength (m_nTakenStart);
	m_nTakenStart += bl;

	if (m_nTakenStart >= m_nBufferSize ||
		m_nTakenStart != m_nFreeStart &&
		GetBlockLength (m_nTakenStart) == -1
	)
		m_nTakenStart = 0;

	/*-----------------------------------------------------------------
	If the buffer is now empty and no block is currently
	allocated, we can reset the TakenStart and FreeStart values
	-----------------------------------------------------------------*/
	if (m_nTakenStart == m_nFreeStart && m_nAllocatedLength < 0)
	{	m_nFreeStart = 0;
		m_nTakenStart = m_nBufferSize;
	}
	m_bFirstRetrieved = FALSE;
	if (m_bWaitingForMemory)
	{	m_bWaitingForMemory = FALSE;
		m_smpWaitForMemory.V ();
	}
}

/*-----------------------------------------------------------------
Find the offset of the block following the one whose offset is given
Returns -1 if there are no more available blocks.
-----------------------------------------------------------------*/
int MediaStreamImp :: GetNext (int nOffset)
{
	int i = GetBlockLength (nOffset);

	if (i == -1)
		nOffset = 0;
	else
		nOffset += i;
	if (nOffset >= m_nBufferSize)
		nOffset = 0;
	if (nOffset == m_nFreeStart)
		return -1;
	if (GetBlockLength (nOffset) == -1)
		nOffset = 0;
	if (nOffset == m_nFreeStart)
		return -1;
	return nOffset;
}

/*-----------------------------------------------------------------
Trace a stream action into a file
-----------------------------------------------------------------*/
void MediaStreamImp :: Trace (int nOffset, LPCSTR szAction)
{
	if (!m_bTrace)
		return;

	m_mtxLockTraceFile.P ();
	if (m_fTraceFile == NULL)
		m_fTraceFile = fopen ("Trace.txt", "w");
	fprintf (
		m_fTraceFile,
		"Time: %d, stream #: %d, action: %s, size: %d, time-stamp: %d\n",
		m_pClock -> GetTime (),
		m_nSN,
		szAction,
		GetBlockLength (nOffset),
		GetTimeStamp (nOffset)
	);
	m_mtxLockTraceFile.V ();
}


/*-----------------------------------------------------------------
Close the stream by the receiving peer
-----------------------------------------------------------------*/
void MediaStreamImp :: Terminate ()
{

	m_mtxLockStream.P ();
	if (m_nOpenToReceive == 0) {
		m_mtxLockStream.V ();
		return;
	}
	if (m_nOpenToReceive > 1) {
		--m_nOpenToReceive;
		m_mtxLockStream.V ();
		return;
	}
	Release ();
	m_bEndOfStream = true;
	Abort ();

	/*-----------------------------------------------------------------
	If the sender waits for memory tell it to go
	-----------------------------------------------------------------*/
	if (m_bWaitingForMemory)
	{	m_bWaitingForMemory = false;
		m_smpWaitForMemory.V ();
	}
	/*-------------------------------------------------------------
	If another thread is the receiver, tell it to go
	-------------------------------------------------------------*/
	if (m_bWaitingForData)
	{	m_bWaitingForData = false;
		m_smpWaitForData.V ();
	}

	if (--m_nUsedBy == 0)
		m_smpInUse.V ();	// stream is in use till released by both sides
	if (m_nWaiting > 0)
		m_smpTimer.V ();
	m_mtxLockStream.V ();

	assert (m_nSN >= 0);
	m_nOpenToReceive = 0;
	m_smpReceiving.V ();
}

/*-------------------------------------------------------------
Modify the clock speed of this stream when the stream is affected by a TemporalTransform
-------------------------------------------------------------*/
void MediaStreamImp :: ModifyClock (float fFactor)
{
	m_mtxLockStream.P ();
	if (m_nWaiting > 0)
		m_smpTimer.V ();
	m_pClock ->Scale (fFactor);
	m_mtxLockStream.V ();
}

int MediaStreamImp :: CompareTime(DWORD time1, DWORD time2)
{
	return m_pClock -> CompareTimes (time1, time2);
}

#ifdef MPGJ
//*** YT START (113)
/*-----------------------------------------------------------------
Pause the stream by receiving peer
- no more dispatch - clear out allocated blocks...
- so replay same screen...(?)
-----------------------------------------------------------------*/

void
MediaStreamImp::Pause (BOOL bPause)
{
	m_mtxLockStream.P();
	m_bLastBlockFetched = FALSE;
	m_bPause = bPause;
	m_mtxLockStream.V();
}

/*-----------------------------------------------------------------
Pause the stream by the sending peer
- no more fetch call - so circulate buffer until reopened
-----------------------------------------------------------------*/
void
MediaStreamImp::HalfClose(BOOL bClose)
{
	m_mtxLockStream.P();
	m_bHalfClose = bClose;
	m_mtxLockStream.V();
}
//*** YT END (113)
#endif MPGJ

