/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright ?1997.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS/OD Decoder / Encoder
FILE:       BifsDec.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of SysCoder - the base class for encoding/decoding
			of Systems (OD, BIFS etc.) streams.
-------------------------------------------------------------------------*/
#include "defs.h"
#include <assert.h>
#include "SysCoder.h"

/*-----------------------------------------------------------------
SysCoder constructor for decoding
-----------------------------------------------------------------*/
SysCoder :: SysCoder (FILE *fList, LPCSTR szListTitle, int flags, int version) :
	m_fList (fList),
	m_szType (szListTitle),
	m_flags (flags),
	m_nVersion (version)
{
	m_bDecode = true;
	m_nLastBit = 0;
	m_nTotalBit = 0;
	m_bNewLine = TRUE;

#ifdef BIFSENC
	/*by default the first AU is RAP on all systems streams*/
	m_bSignalRAP = FALSE;
	m_isRAP = TRUE;
#endif
}

/*-------------------------------------------------------------
SysCoder constructor for encoding
-------------------------------------------------------------*/
SysCoder :: SysCoder (
	FILE *fOutput,
	FILE *fList,
	LPCSTR szListTitle,
	int flags,
	int version,
    FILE * fLogs
) :
	m_fOutput (fOutput),
	m_fList (fList),
    m_fLogs (fLogs),
	m_szType (szListTitle),
	m_flags (flags),
	m_nLastByte (0),
	m_nLastBit (0),
	m_bDecode (false),
	m_nVersion (version)
{
    	m_nTotalBit = 0;
}

/*-------------------------------------------------------------
Reset the coder for decoding. The coded unit address and length is
stored in member variables
-------------------------------------------------------------*/
void SysCoder :: Reset (LPCBYTE pData, int nLength)
{
	m_pData = pData;
	m_nSize = nLength;
	m_nLastBit = 8;
}

/*-------------------------------------------------------------
For Decode:
Make sure next bit would be fetched from byte boundary. Return the
number of skipped bits
For Encode:
Pad the current byte and make sure subsequent encoding starts on
a byte boundary.
-------------------------------------------------------------*/
int SysCoder :: Align ()
{
	if (m_bDecode)
	{	int nResult = 8 - m_nLastBit;
		if (nResult > 0)
			ParseInt (nResult, "Align");
		return nResult;
	}
	else if (m_nLastBit > 0)
		EncodeInt (0, 8 - m_nLastBit, "Align");
	return 0;
}

/*-----------------------------------------------------------------
Fetch one byte from the stream
-----------------------------------------------------------------*/
int SysCoder :: GetByte ()
{
	if (m_nSize == 0)
		EndOfUnit ();
	return (int) *m_pData++;
}

/*-----------------------------------------------------------------
Retrieves nBits bits from the BifStream and returns their integer value
Methodology based on GetByte() and GetBits()
-----------------------------------------------------------------*/
int SysCoder :: ParseInt (int nBits, LPCSTR szDesc)
{
	int nReturn = 0, nSaveBits = nBits;
	while (nBits-- > 0)
	{
		nReturn <<= 1;
		nReturn |= GetBit ();
	}
	if(m_fList && nSaveBits > 0)
	{
		if ((m_flags & TRACE_CONFORMANT) > 0) //BF Conformance
		{
			m_nTotalBit += nSaveBits;
			fprintf (
				m_fList,
				"%-4s <>  (%3d,%10d)   %-40s%9X %d\n",
				m_szType,
				nSaveBits,
				m_nTotalBit,
				szDesc,
				nReturn,
				nReturn
				);
		}
		if ((m_flags & TRACE_VERBOSE) > 0)  
		{
			fprintf (
			m_fList,
			"%s: bit (%d) %s = %d (0x%x)\n",
			m_szType,
			nSaveBits,
			szDesc,
			nReturn,
			nReturn
		);
		}
	}
	return nReturn;
}

/*-----------------------------------------------------------------
Retrieves nBits bits from the BifStream and returns their long integer (more than 32 bits)
value Methodology based on GetByte() and GetBits()
-----------------------------------------------------------------*/
ZLargeInt<128> SysCoder :: ParseLongInt (int nBits, LPCSTR szDesc)
{
	ZLargeInt<128> nReturn;
	int nSaveBits = nBits;
	while (nBits-- > 0)
	{
		nReturn <<= 1;
		nReturn += GetBit ();
	}
	if(m_fList && nSaveBits > 0)
	{
		if ((m_flags & TRACE_CONFORMANT) > 0) //BF Conformance
		{
			m_nTotalBit += nSaveBits;
			if (nSaveBits <= 32)
				fprintf (
					m_fList,
					"%-4s <>  (%3d,%10d)   %-40s 0X%X %d\n",
					m_szType,
					nSaveBits,
					m_nTotalBit,
					szDesc,
					(int) nReturn,
					(int) nReturn
				);
			else
				fprintf (
					m_fList,
					"%-4s <>  (%3d,%10d)   %-40s 0X%08X%08X%08X%08X %d\n",
					m_szType,
					nSaveBits,
					m_nTotalBit,
					szDesc,
					nReturn.GetWord (0),
					nReturn.GetWord (1),
					nReturn.GetWord (2),
					nReturn.GetWord (3),
					(int) nReturn
				);
		}
		if ((m_flags & TRACE_VERBOSE) > 0)  
		{
			if (nSaveBits <= 32)
				fprintf (
					m_fList,
					"%s: bit (%d) %s = %d (0x%X)\n",
					m_szType,
					nSaveBits,
					szDesc,
					(int) nReturn,
					(int) nReturn
				);
			else
				fprintf (
					m_fList,
					"%s: bit (%d) %s = %d (0x%08X%08X%08X%08X)\n",
					m_szType,
					nSaveBits,
					szDesc,
					(int) nReturn,
					nReturn.GetWord (0),
					nReturn.GetWord (1),
					nReturn.GetWord (2),
					nReturn.GetWord (3)
				);
		}
	}
	return nReturn;
}

/*-----------------------------------------------------------------
Retrieves one bit from the stream
-----------------------------------------------------------------*/
int SysCoder :: GetBit ()
{
	/*-----------------------------------------------------------------
	Get next byte, if necessary
	-----------------------------------------------------------------*/
	if (m_nLastBit >= 8)
	{
	 	m_nLastByte = GetByte ();
		m_nLastBit = 0;
		m_nSize--;
	}

	/*-----------------------------------------------------------------
	Retrieve the first bit of the byte, shift the byte so next time
	second bit will be the first
	-----------------------------------------------------------------*/
	m_nLastByte <<= 1;
	++m_nLastBit;
	return (m_nLastByte & 0x100) >> 8;
}

/*-------------------------------------------------------------
Skip the given number of input bytes (excluding the current)
-------------------------------------------------------------*/
void SysCoder :: SkipBytes (int nLength)
{
	Align ();
	if (nLength == 0)
		return;
	m_nSize -= nLength;
	if (m_nSize < 0)
		EndOfUnit ();
	m_pData += nLength;

	if(m_fList)
	{
		if (m_flags & TRACE_CONFORMANT)  //BF Conformance
		{
			m_nTotalBit += 8*nLength;
			fprintf (
				m_fList,
				"%-4s <>  (%3d,%10d)   skip unrecognized data\n",
				m_szType,
				8*nLength,
				m_nTotalBit
				);
		}
		if (m_flags & TRACE_VERBOSE)
		{
			fprintf (
			m_fList,
			"%s: bit (%d) skip unrecognized data\n",
			m_szType,
			8 * nLength
		);
		}
	}
}

/*-----------------------------------------------------------------
The following is the float parse function
-----------------------------------------------------------------*/
float SysCoder :: ParseFloat (LPCSTR szDesc)
{
	float fResult;
#ifdef WIN32
	char buf [4] = "\0\0\0";
	int i;
	for (i = 0; i < 32; i++)
		buf[3-i/8] |= GetBit () << (7 - i%8);

	fResult = * (float *) buf;

#else
// SOLARIS PORTING - implemented for Solaris-SPARC
//	fResult = 0.0;
	char buf [4] = "\0\0\0";
	int i;
	for (i = 0; i < 32; i++)
		buf[i/8] |= GetBit () << (7 - i%8);

	fResult = * (float *) buf;
#endif

	if(m_fList)
	{
		if (m_flags & TRACE_CONFORMANT)
		{
			m_nTotalBit += 32;
            // mp begin conformance
            unsigned long iResult = * (unsigned long *) &fResult;
			fprintf (
				m_fList,
				"%-4s <>  ( 32,%10d)   %-40s%9X %f\n",
				m_szType,
				m_nTotalBit,
				szDesc,
				iResult,
				fResult
				);
            // mp end conformance
		}
		if(m_flags & TRACE_VERBOSE)
		{
			fprintf (m_fList, "%s: float %s = %f\n", m_szType, szDesc, fResult);
		}
	}
	return fResult;
}

/*-----------------------------------------------------------------
The following is the double parse function
-----------------------------------------------------------------*/
double SysCoder :: ParseDouble (LPCSTR szDesc)
{
	double dResult;

#ifdef WIN32
	char buf [8] = "\0\0\0\0\0\0\0";
	int i;
	for (i = 0; i < 64; i++)
		buf[7-i/8] |= GetBit () << (7 - i%8);

	dResult = * (double *) buf;

#else
// SOLARIS PORTING - implemented for Solaris-SPARC
//	dResult = 0.0;
	char buf [8] = "\0\0\0\0\0\0\0";
	int i;
	for (i = 0; i < 64; i++)
		buf[i/8] |= GetBit () << (7 - i%8);
	
	dResult = * (double *) buf;
#endif

	if(m_fList)
	{
		if (m_flags & TRACE_CONFORMANT)
		{
			m_nTotalBit += 64;
            // mp begin conformance fix
            unsigned long iResult1 = * (((unsigned long *) &dResult) + 1);
            unsigned long iResult2 = * (unsigned long *) &dResult;
            if (iResult1 > 0) {
                fprintf (
				    m_fList,
				    "%-4s <>  ( 64,%10d)   %-32s %8lX%08lX %lf\n",
				    m_szType,
				    m_nTotalBit,
				    szDesc,
				    iResult1,
				    iResult2,
                    dResult
				    );
            } else {
			    fprintf (
				    m_fList,
				    "%-4s <>  ( 64,%10d)   %-40s%9X %lf\n",
				    m_szType,
				    m_nTotalBit,
				    szDesc,
				    iResult2,
				    dResult
				    );
            }
            // mp end conformance fix
		}
		if (m_flags & TRACE_VERBOSE)
		{
			fprintf (m_fList, "%s: double %s = %lf\n", m_szType, szDesc, dResult);
		}
	}
	return dResult;
}

int SysCoder :: m_nIndent = 0;

/*-----------------------------------------------------------------
Trace functions for debug mode
-----------------------------------------------------------------*/
#ifdef TRACE_ENCODING
void SysCoder :: Tracef (LPCSTR szFormat, ...)
{
	if (!(m_flags & TRACE_PARSE) || m_fList == NULL)
		return;
	if (m_bNewLine)
	{	for (int i = m_nIndent; --i >= 0;)
			putc ('\t', m_fList);
		m_bNewLine = FALSE;
	}
	va_list argptr;
	va_start (argptr, szFormat);
	vfprintf (m_fList, szFormat, argptr);
	va_end (argptr);
//AC:20021009 - SL extension : debug -- begin
	fflush(m_fList);
//AC:20021009 - SL extension : debug -- end
}

void SysCoder :: TraceIndent (int nDirection)
{
	if (m_flags & TRACE_PARSE)
		m_nIndent += nDirection;
}

void SysCoder :: TraceNewLine ()
{
	if (!(m_flags & TRACE_PARSE) || m_fList == NULL)
		return;
	if (!m_bNewLine)
	{	m_bNewLine = TRUE;
		putc ('\n', m_fList);
	}
}
#endif

/*-----------------------------------------------------------------
Encode nValue into nBits bits, and report this action in the list file
-----------------------------------------------------------------*/
void SysCoder :: EncodeInt (int nValue, int nBits, LPCSTR szDesc, LPCSTR szComment)
{
	if ((m_flags & TRACE_CONFORMANT) && m_fLogs) //mp Conformance
	{        
		m_nTotalBit += nBits;
		fprintf (
			m_fLogs,
			"%-4s <>  (%3d,%10d)   %-40s%9X %s\n",
			m_szType,
			nBits,
			m_nTotalBit,
			szDesc,
			nValue,
            szComment ? szComment : ""
			);
	}
    if((m_flags & TRACE_VERBOSE) && m_fList)
	{	if (nBits < sizeof (int) * 8)
			nValue &= (1 << nBits) - 1;
		fprintf (
			m_fList,
			"%s: bit (%d) %s = %d (0x%x) %s\n",
			m_szType,
			nBits,
			szDesc,
			nValue,
			nValue,
            szComment
		);
	}

	nValue <<= sizeof (int) * 8 - nBits;
	while (--nBits >= 0)
	{	WriteBit (nValue < 0);
		nValue <<= 1;
	}
}

/*-----------------------------------------------------------------
Encode a float value
-----------------------------------------------------------------*/
void SysCoder :: EncodeFloat (float fValue, LPCSTR szDesc)
{
	// writes GenericFloat
#ifdef WIN32
	union
	{	float f;
		char sz [4];
	} value;
	value.f = fValue;

	for (int i = 0; i < 32; i++)
		WriteBit ((value.sz [3 - i / 8] & 1 << (7 - i % 8)) != 0);
#else
// SOLARIS PORTING
	union
	{	float f;
		char sz [4];
	} value;
	value.f = fValue;

	for (int i = 0; i < 32; i++)
		WriteBit ((value.sz [ i / 8] & 1 << (7 - i % 8)) != 0);
#endif		// WIN32

	if (m_fLogs && (m_flags & TRACE_CONFORMANT)) // mp conformance
	{
		m_nTotalBit += 32;
        unsigned long iResult = * (unsigned long *) &fValue;
		fprintf (
			m_fLogs,
			"%-4s <>  ( 32,%10d)   %-40s%9X %f\n",
			m_szType,
			m_nTotalBit,
			szDesc,
			iResult,
			fValue
			);
	}
    if((m_flags & TRACE_VERBOSE) && m_fList)
		fprintf (m_fList, "%s: float %s = %f\n", m_szType, szDesc, fValue);
}

/*-----------------------------------------------------------------
Encode a double value
-----------------------------------------------------------------*/
void SysCoder :: EncodeDouble (double dValue, LPCSTR szDesc)
{
#ifdef WIN32
	union
	{	double d;
		char sz [8];
	} value;
	value.d = dValue;

	for (int i = 0; i < 64; i++)
		WriteBit ((value.sz [7 - i / 8] & 1 << (7 - i % 8)) != 0);
#else
// SOLARIS PORTING
	union
	{	double d;
		char sz [8];
	} value;
	value.d = dValue;

	for (int i = 0; i < 64; i++)
		WriteBit ((value.sz [ i / 8] & 1 << (7 - i % 8)) != 0);
#endif		// WIN32

	if ((m_flags & TRACE_CONFORMANT) && m_fList) // mp conformance
	{
		m_nTotalBit += 64;
        unsigned long iResult1 = * (((unsigned long *) &dValue) + 1);
        unsigned long iResult2 = * (unsigned long *) &dValue;
        if (iResult1 > 0) {
            fprintf (
				m_fLogs,
				"%-4s <>  ( 64,%10d)   %-32s %8lX%08lX %lf\n",
				m_szType,
				m_nTotalBit,
				szDesc,
				iResult1,
				iResult2,
                dValue
				);
        } else {
	        fprintf (
		        m_fList,
		        "%-4s <>  ( 64,%10d)   %-40s%9X %lf\n",
		        m_szType,
		        m_nTotalBit,
		        szDesc,
		        iResult2,
		        dValue
		        );
        }
    } else if((m_flags & TRACE_VERBOSE) && m_fList) {
		fprintf (m_fList, "%s: double %s = %lf\n", m_szType, szDesc, dValue);
    }
}

/*-----------------------------------------------------------------
Add one bit to the output stream
-----------------------------------------------------------------*/
void SysCoder :: WriteBit (BOOL bit)
{
	m_nLastByte <<= 1;
	m_nLastByte |= bit;
	if (++m_nLastBit == 8)
	{	m_nLastBit = 0;
		WriteByte ();
		m_nLastByte = 0;
	}
}

/*-----------------------------------------------------------------
Write the last byte to the output stream
-----------------------------------------------------------------*/
void SysCoder :: WriteByte ()
{
	m_buffer += m_nLastByte;
}

/*-------------------------------------------------------------
Reset the buffer and counters
-------------------------------------------------------------*/
void SysCoder :: Reset ()
{
	m_buffer.Cut (0);
	m_nLastByte = 0;
	m_nLastBit = 0;
}

/*-----------------------------------------------------------------
Flush the buffer when finished
-----------------------------------------------------------------*/
void SysCoder :: Flush (DWORD dwTS)
{
	Align ();
	if (m_fOutput && m_buffer.GetLength () > 0) {
#ifdef BIFSENC
		/*JLF 03/03. added RAP indicator*/
		if (m_bSignalRAP) {
			char c = (m_isRAP ? 1 : 0)<<7 | (HIBYTE (HIWORD (dwTS)) & 0x7F);
			fputc (c, m_fOutput);
		} else {
			fputc (HIBYTE (HIWORD (dwTS)), m_fOutput);
		}
		m_isRAP = FALSE;
#else
		fputc (HIBYTE (HIWORD (dwTS)), m_fOutput);
#endif
		fputc (LOBYTE (HIWORD (dwTS)), m_fOutput);
		fputc (HIBYTE (LOWORD (dwTS)), m_fOutput);
		fputc (LOBYTE (LOWORD (dwTS)), m_fOutput);
		Write ();
	}
}

/*-----------------------------------------------------------------
Write the encoded data, preceded by the length, to the output file.
If bShort is true the length is encoded as 16-bit big-endian number,
otherwise it is a 32-bit number.
-----------------------------------------------------------------*/
void SysCoder :: Write (bool bShort)
{
	int nLength = m_buffer.GetLength ();
	if (!bShort)
	{	fputc (HIBYTE (HIWORD (nLength)), m_fOutput);
		fputc (LOBYTE (HIWORD (nLength)), m_fOutput);
	}
	fputc (HIBYTE (LOWORD (nLength)), m_fOutput);
	fputc (LOBYTE (LOWORD (nLength)), m_fOutput);
//	fwrite (m_buffer, m_buffer.GetLength (), 1, m_fOutput);
	fwrite ((BYTE *)m_buffer, m_buffer.GetLength (), 1, m_fOutput); // SOLARIS PORTING - added to BYTE *
	m_buffer.Cut (0);
}

/*-------------------------------------------------------------
Replace the target encoded file
-------------------------------------------------------------*/
void SysCoder :: ReplaceOutputFile (FILE *fNewOutputFile)
{
	if (m_fOutput)
		fclose (m_fOutput);
	m_fOutput = fNewOutputFile;
}

/*-------------------------------------------------------------
Concatenate the buffer of another coder with this one's
-------------------------------------------------------------*/
void SysCoder :: Concatenate (SysCoder *pSource)
{
	int nTotalBits = pSource ->GetOffset ();
	pSource ->Align ();
	pSource ->Reset (pSource->GetData (), pSource ->GetDataLength ());
	while (--nTotalBits >= 0)
		WriteBit (pSource ->GetBit ());
}

/*-------------------------------------------------------------
Parse an SFString value
-------------------------------------------------------------*/
void SysCoder :: ParseSFString (ZString &string)
{
	string.Cut (0);
	int nBits = ParseInt (5, "lengthBits"); // mp : conformance
	int nSize = ParseInt (nBits, "length"); // mp : conformance

	while (--nSize >= 0)
		string += ParseInt (8, "value");
}

/*-------------------------------------------------------------
Encode an SFString value
-------------------------------------------------------------*/
void SysCoder :: EncodeSFString (const ZString &string)
{
	// SOWOO 20030109
	if(m_flags & STRING_FROM_FILE)
	{
		// this part is added by sowoo
		// in this part, 'string' variable contains the file name,
		// so data from this file will be encoded instead of 'string' itself.
		FILE* file = NULL;
		fpos_t nSize;
		file = ::fopen(string, "rb");

		// get the size of string
		::fseek(file, 0L, SEEK_END);
		::fgetpos(file, &nSize);
		
		int nBits = 0;
		int nTemp = nSize;

		while (nSize > 0)
		{	nSize >>= 1;
			nBits++;
		}

		EncodeInt (nBits, 5, "lengthBits");					//AC20011102 : conformance
		EncodeInt (nTemp, nBits, "length");	//AC20011102 : conformance

		::fseek(file, 0L, SEEK_SET);

		int i = 1;
		while(i<=nTemp)
		{
			char szIndex [20];
			if (m_flags & TRACE_VERBOSE)
				sprintf (szIndex, "value [%d]", i-1);
			unsigned char c = ::fgetc(file);
			EncodeInt (c, 8, "value", szIndex);				//AC20011102 : conformance
			i++;
		}
	}
	else
	{
		// this part is the original source code
		int nSize = string.GetLength ();
		int nBits = 0;

		while (nSize > 0)
		{	nSize >>= 1;
			nBits++;
		}

		EncodeInt (nBits, 5, "lengthBits");					//AC20011102 : conformance
		EncodeInt (string.GetLength (), nBits, "length");	//AC20011102 : conformance

		for (int i = 0; i < string.GetLength (); i++)
		{
			char szIndex [20];
			if (m_flags & TRACE_VERBOSE)
				sprintf (szIndex, "value [%d]", i);
			unsigned char c = string [i];
			EncodeInt (c, 8, "value",szIndex);				//AC20011102 : conformance
		}
	}
	// SOWOO end
}
