/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

#if !defined(_sdl_h)	// Sentry, use file only if it's not already included.
#define _sdl_h

/*-------------------------------------------------------------------------
Project Im1Player/ Mux
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  OD Decoder/ Encoder
FILE:       sdl.h
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Declarations for SDL encoding/parsing
-------------------------------------------------------------------------*/

#include "ztl.h"

typedef enum {DECODE, ENCODE, CALC_LENGTH} OPTYPE;

class ODCoder;
class SDLField;

/*-------------------------------------------------------------
A structure describing one field of a descriptor
-------------------------------------------------------------*/
struct ODFieldTable
{
	WORD m_offset;
	LPCSTR m_szName;
};

/*-------------------------------------------------------------
A structure describing one extension entry in an extension table of
an abstract descriptor
-------------------------------------------------------------*/
struct ODExtensionTable
{
	WORD m_tag, m_code;
};

#define NULL_DESC ((BaseDescriptor *)0)

class SDLBaseClass;

/*-------------------------------------------------------------
The base class for all descriptors
-------------------------------------------------------------*/
class BaseDescriptor
{
public:
	BaseDescriptor (int nTagLength = 8, bool bIsExpandable = true) :
		m_nTag (-1),
		m_bProtected (false),
		m_nLength (0),
		m_nTagLength (nTagLength),
		m_bIsExpandable (bIsExpandable)
	{}
	virtual ~BaseDescriptor () {}
	void EncDec (
		ODCoder *pODCoder,
		OPTYPE op,
		BaseDescriptor *pParent = NULL
	);
	virtual void DescEncDec (ODCoder *pODCoder, OPTYPE op);
	virtual LPCSTR GetName () {return NULL;}
	virtual int GetFieldCount () const {return 0;}
	SDLField *GetField (int index);
	virtual const ODFieldTable *GetFieldTable () const {return NULL;}
	virtual int GetTag () {return -1;}
	int GetLength () {return m_nLength;}
	void AddLength (int n) {m_nLength += n;}
	void AlignLength () {m_nLength = (m_nLength + 7) / 8 * 8;}
	int CalcLengthFieldWidth ();
	LPCSTR GetFieldName (SDLField *pField) const;
	int GetNextTag () {return m_nTag;}
	void SetNextTag (int nTag) {m_nTag = nTag;}
	void SetLength (int nLength) {m_nLength = nLength;}
	bool IsProtected () const {return m_bProtected;}
	void Protect (bool bFlag) {m_bProtected = bFlag;}

#if defined(BIFSENC)
	void Scan ();
	virtual void PreEncode (SDLBaseClass *) {}
	const ODFieldTable *GetODEntry (LPCSTR szName);
#endif

protected:
	static BaseDescriptor *CreateExtension (
		int nTag,
		const ODExtensionTable *pTable,
		int nLength
	);
	static BaseDescriptor *CreateExtension (
		LPCSTR szClassName,
		const ODExtensionTable *pTable,
		int nLength
	);

private:
	int m_nLength, m_nTag, m_nTagLength;
	bool m_bProtected, m_bIsExpandable;
};

/*-------------------------------------------------------------
SDLDescriptor implements a non-expandable SDL class
-------------------------------------------------------------*/
class SDLDescriptor : public BaseDescriptor
{
public:
	SDLDescriptor (bool bIsExtended = false) : BaseDescriptor (0, bIsExtended) {}
};


class SDLField
{
public:
	virtual ~SDLField () {}
	virtual void EncDec (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op
	) = 0;
	static void StoreFactory (int nCode, class ODFactoryBase *pFactory);
	static BaseDescriptor *CreateClass (int nCode);

#if defined(BIFSENC)
	virtual void Scan () = 0;
	BaseDescriptor *CreateClass (LPCSTR szClassName);
#endif

private:
	static ODFactoryBase *m_odFactory [];
};

class SDLInteger : public SDLField
{
public:
	SDLInteger (int nBits = 0, bool bAligned = false, int nValue = 0) :
		m_nBits (nBits), m_nValue (nValue), m_bAligned (bAligned)
	{}
	operator int () const {return m_nValue;}
	operator = (const nValue) {return m_nValue = nValue;}
	virtual void EncDec (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op
	);
	void EncDecVarSize (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op,
		int nBits
	);

#if defined(BIFSENC)
	virtual void Scan ();
#endif

private:
	int m_nBits, m_nValue;
	bool m_bAligned;
};

template <int bits> class SDLInt : public SDLInteger
{
public:
	SDLInt (int nValue = 0) : SDLInteger (bits, false, nValue) {}
};

template <int bits> class SDLAlignedInt : public SDLInteger
{
public:
	SDLAlignedInt (int nValue = 0) : SDLInteger (bits, true, nValue) {}
};

class SDLFloat : public SDLField
{
public:
	SDLFloat (float fValue = 0) : m_fValue (fValue) {}
	operator float () const {return m_fValue;}
	virtual void EncDec (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op
	);

#if defined(BIFSENC)
	virtual void Scan ();
#endif

private:
	float m_fValue;
};

class SDLDouble : public SDLField
{
public:
	SDLDouble (float fValue = 0) : m_fValue (fValue) {}
	operator float () const {return m_fValue;}
	virtual void EncDec (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op
	);

#if defined(BIFSENC)
	virtual void Scan ();
#endif

private:
	double m_fValue;
};

class SDLArrayBase : public SDLField
{
public:

#if defined (BIFSENC)
	virtual void Scan ();
	virtual void ScanElement () {}
#endif
};

class SDLIntegerArray : public SDLArrayBase
{
public:
	SDLIntegerArray (int nBits);
	operator LPCSTR () const {return m_array;}
	operator LPCBYTE () const {return m_array;}
	int operator [] (int i) const {return GetElement (i);}
	ZLargeInt<128> GetElement (int i) const;
	int GetLength () const {return m_array.GetLength () / m_nBytes;}
	void InsertElement (ZLargeInt<128> n, int index = -1);
	void StoreElementSize (int nBits);
	virtual void EncDec (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op
	);
	void EncDecArrayVarSize (
	ODCoder *pCoder,
	BaseDescriptor *pDesc,
	OPTYPE op,
	int nLength,
	int nBits
	);
	void EncDecArray (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op,
		int nLength
	);
	void EncDecArrayElement (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op,
		int nIndex
	);

#if defined (BIFSENC)
	virtual void Scan ();
	virtual void ScanElement ();
#endif

protected:
	int m_nBits, m_nBytes;
	ZString m_array;
};

template <int bits> class SDLArray : public SDLIntegerArray
{
public:
	SDLArray () : SDLIntegerArray (bits) {}
};

class SDLLongInteger : public SDLIntegerArray
{
public:
	SDLLongInteger (int bits) : SDLIntegerArray (bits) {}
	ZLargeInt<128> GetValue () const {return SDLIntegerArray :: GetElement (0);}
	virtual void EncDec (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op
	);
};

template <int bits> class SDLLongInt : public SDLLongInteger
{
public:
	SDLLongInt () : SDLLongInteger (bits) {}
};


class SDLString : public SDLArray<8>
{
public:
	operator LPCSTR () const;
	virtual void EncDec (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op
	);
	void EncDecString (
		ODCoder *pCoder,
		BaseDescriptor *pDesc,
		OPTYPE op,
		int bIsUTF16
	);
	void EncDecLongString (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op,
		int bIsUTF16 = false
	);

#if defined(BIFSENC)
	virtual void Scan ();
#endif

private:
	void EncDecStringElements (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op,
		int nLength
	);
};

class SDLStringArray : public SDLArrayBase
{
public:
	const ZString & operator [] (int i) const {return m_array [i];}
	virtual void EncDec (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op
	);
	virtual void EncDecStringArray (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op,
		int length,
		int bIsUTF16 = false
	);
	void EncDecStringArrayElement (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op,
		int nIndex,
		int bIsUTF16 = false
	);
	int GetLength () const {return m_array.GetLength ();}

#if defined(BIFSENC)
	virtual void ScanElement ();
#endif

private:
	ZArray<ZString> m_array;
};

class SDLBaseClass : public SDLField
{
public:
	SDLBaseClass (int min) :
		m_pClass (NULL), m_nMin (min)
	{}
	~SDLBaseClass ();
	BaseDescriptor *DecodeDescriptor (
		ODCoder *pODCoder,
		BaseDescriptor *pParent,
		LPCSTR szFieldName,
		int index
	);
	virtual void EncDec (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op
	);
	BaseDescriptor* operator = (BaseDescriptor *p) {m_pClass = p; return p;}

#if defined(BIFSENC)
	virtual void Scan ();
private:
	virtual BaseDescriptor *CreateDescriptor (LPCSTR szClassName) = 0;
	BaseDescriptor *CreateNamedDescriptor (LPCSTR szClassName,  int &rnResult);
#endif

protected:
	virtual BaseDescriptor *CreateDescriptor (int nTag) = 0;
	virtual bool IsMyTag (int nTag) const = 0;

protected:
	BaseDescriptor *m_pClass;
	int m_nMin;
};

// SOLARIS PORTING
#ifdef WIN32
template <class T, int min = 1> class SDLClass : public SDLBaseClass
#else
template <class T, int min > class SDLClass : public SDLBaseClass
#endif
{
public:
	SDLClass () : SDLBaseClass (min) {}
	T* operator -> () const {return (T*) m_pClass;}
	operator T* () const {return (T*) m_pClass;}
	T* operator = (T *p) {m_pClass = p; return p;}
	BaseDescriptor* operator = (BaseDescriptor *p) {m_pClass = p; return p;}

private:
	virtual bool IsMyTag (int nTag) const;
	virtual BaseDescriptor *CreateDescriptor (int nTag);
	virtual BaseDescriptor *CreateDescriptor (LPCSTR szClassName);
};

class SDLBaseClassArray : public SDLArrayBase
{
public:
	SDLBaseClassArray (int min, int max) : m_nMin (min), m_nMax (max) {}
	~SDLBaseClassArray ();
	virtual void EncDec (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op
	);
	virtual BaseDescriptor *DecodeDescriptor (
		ODCoder *pODCoder,
		BaseDescriptor *pParent,
		LPCSTR szFieldName,
		int index
	) = 0;
	int GetLength () {return m_array.GetLength ();}
	void DeleteFrom (int index) {m_array.DeleteFrom (index);}
	void EncDecArray (
		ODCoder *pODCoder,
		BaseDescriptor *pDesc,
		OPTYPE op,
		int nLength
	);

#if defined(BIFSENC)
protected:
	void AddElement (BaseDescriptor *pDesc);
#endif

protected:
	ZArray<BaseDescriptor*> m_array;
	int m_nMin, m_nMax;
};

// SOLARIS PORTING
#ifdef WIN32
template <class T, int min = 0, int max = 0xffff> class SDLClassArray : public SDLBaseClassArray
#else
template <class T, int min, int max> class SDLClassArray : public SDLBaseClassArray
#endif
{
public:
	SDLClassArray () : SDLBaseClassArray (min, max) {}
	T* operator [] (int i) const {return (T*) m_array [i];}
	operator T * const * () const {return (T * const *) (BaseDescriptor * const *) m_array;}
	T* operator += (T *p)
	{	
		m_array += (BaseDescriptor*) p;
		return p;
	}
	void SetAt (int i, T *p) {m_array [i] = p;}

private:
	virtual bool IsMyTag (int nTag) const;
	virtual BaseDescriptor *CreateDescriptor (int nTag);
	virtual BaseDescriptor *CreateDescriptor (LPCSTR szClassName);
	virtual BaseDescriptor *DecodeDescriptor (
		ODCoder *pODCoder,
		BaseDescriptor *pParent,
		LPCSTR szFieldName,
		int index
	);

#if defined(BIFSENC)
	virtual void ScanElement ();
#endif
};

/*-------------------------------------------------------------
Declare OD factory objects
-------------------------------------------------------------*/
class ODFactoryBase
{
public:
	virtual BaseDescriptor *CreateDescriptor () = 0;
};

template <class T> class ODFactory : public ODFactoryBase
{
	virtual BaseDescriptor *CreateDescriptor () {return new T;}

public:
	ODFactory ()
	{
		SDLField :: StoreFactory (T :: GetCode (), this);
	}
};

#endif                                      // _sdl_h sentry.
