/***********************************************************************
    This software module  was  originally  developed by Zvi  Lifshitz
    (Triton R&D Ltd.) and edited by Mauro Rossi (CSELT),
    Guido Franceschini (CSELT), Michael Stepping (FTK) and Carsten 
    Herpel (THOMSON multimedia)
    in the course of development of the MPEG-4 DMIF (ISO/IEC 14496-6)
    standard.  This software module is an implementation of a part of
    one or more MPEG-4  DMIF  (ISO/IEC 14496-6)  tools  as  specified
    by  the  MPEG-4 DMIF (ISO/IEC 14496-6)  standard.  ISO/IEC  gives
    users of  the  MPEG-4  DMIF  (ISO/IEC 14496-6)  free  license  to
    this software module or modifications thereof for use in hardware
    or software products claiming  conformance  to  the  MPEG-4  DMIF 
    (ISO/IEC 14496-6). Those intending to use this software module in
    hardware or software  products  are  advised  that  its  use  may
    infringe   existing  patents.  The  original  developer  of  this
    software module and his/her company, the subsequent  editors  and
    their  companies,  and  ISO/IEC have no liability for use of this
    software module or modifications thereof  in  an  implementation.
    Copyright  is not released for non MPEG-4 DMIF  (ISO/IEC 14496-6)
    conforming  products. Triton R&D Ltd.
    retains full  right to use the code for  its  own purpose, assign
    or donate the code to a third party and to inhibit third  parties
    from  using  the  code  for  non  MPEG-4  DMIF  (ISO/IEC 14496-6)
    conforming products. This copyright notice must  be  included  in
    all copies or derivative works. Copyright  1998.
************************************************************************/

// Contact: zvil@csi.com

/**************************************************************************
 * Project Im1Player
 * INTERNATIONAL STANDARD ORGANIZATION
 * ISO/IEC JTC1/SC29/WG11
 * 
 * SUBSYSTEM: DMIF
 * FILE:      DMIFClientFilter.cpp
 * AUTHOR:    Zvi Lifshitz (Triton R&D Ltd.), Mauro Rossi (CSELT)
 * OVERVIEW:  Implementation of FileTrifInstance, the class that implements
 *			  local storage DMIF.
 * 
/***********************************************************************/

#include <string.h>
#include "FileTrifInstance.h"

LPCSTR FileTrifInstance::s_moduleID = "FileTrifInstance";
LPCSTR FileTrifInstance::s_myURLScheme = "d-trif:";
LPCSTR FileTrifInstance::s_myFileExtension = ".trif";

/*-------------------------------------------------------------
The exported function that instantiates the proper decoder
-------------------------------------------------------------*/
extern "C" Service *CreateInstance ()
{
	return new FileTrifInstance;
}

LPCSTR FileTrifInstance::GetModuleID() 
{
	return s_moduleID;
}

/*-----------------------------------------------------------------
Constructor
-----------------------------------------------------------------*/
FileTrifInstance :: FileTrifInstance (FileTrifInstance* pMain, int nID) :
	m_smpWaitForChannel (1, 0),
	m_fileThread (
// PORTING to CC 5.3
//		FileThread,
		& FileTrifInstance :: FileThread,
		this,
		THREAD_PRIORITY_BELOW_NORMAL
	)
{
	m_nID = nID;
	m_nChannelWaitedFor = -1;
	m_fInput = NULL;
	m_bRewinding = m_bTerminate = false;
	m_pMain = (pMain == NULL) ? this : pMain ;
}

/*-----------------------------------------------------------------
The object destructor makes sure the service is closed. Then it deletes
all auxiliary file objects.
-----------------------------------------------------------------*/
FileTrifInstance :: ~FileTrifInstance ()
{
	m_smpActive.P ();
	if (m_fInput)
		fclose (m_fInput);
	m_smpActive.V ();
	for (int i = m_auxFiles.GetLength (); --i >= 0;)
		delete m_auxFiles [i];
	m_fileThread.Wait ();
}

/*-----------------------------------------------------------------
Open the file, load the association table, start the file thread,
and notify the application
-----------------------------------------------------------------*/
// SOLARIS PORTING - Added "const"
//void FileTrifInstance :: DPI_ServiceAttach_Req(LPCSTR szURL, const BYTE*, int)
void FileTrifInstance :: DPI_ServiceAttach_Req(LPCSTR const szURL, const BYTE*, const int)
{
	ZString initialOD;

	RESPONSE response = Connect (szURL, &initialOD);
	GetSink () -> DPI_ServiceAttach_Cnf (
		m_pMain,
		(const BYTE *) (LPCSTR) initialOD,
		initialOD.GetLength (),
		response
	);
	return;
}

/*-----------------------------------------------------------------
Open the file, load the association table, and start the file thread.
Return a success code.
-----------------------------------------------------------------*/
RESPONSE FileTrifInstance :: Connect (
	LPCSTR szURL,
	ZString *initialOD
)
{
	int c, n;

	/*-------------------------------------------------------------
	If the URL starts with the appropriate scheme
	it is followed by the file name
	-------------------------------------------------------------*/
	int schemeLen = strlen( s_myURLScheme );
	if ( _strnicmp(szURL, s_myURLScheme, schemeLen) == 0 )
		szURL += schemeLen;
	m_url = szURL;

	/*-------------------------------------------------------------
	If the file was opened before, rewind it. Otherwise open it.
	-------------------------------------------------------------*/
	if (m_fInput)
	{	rewind (m_fInput);
		m_channelMap.Clear ();
	}
	else if ((m_fInput = fopen (szURL, "rb")) == NULL)
		return RESPONSE_URL_NOT_FOUND;

	try
	{
		/*-----------------------------------------------------------------
		Read the initial OD
		-----------------------------------------------------------------*/
		c = GetByte ();
		n = MAKEWORD (GetByte (), c);
		while (--n >= 0)
		{	c = GetByte ();
			if (initialOD != NULL)
				*initialOD += c;
		}
		
		/*-----------------------------------------------------------------
		Read the association map
		-----------------------------------------------------------------*/
		m_nNotReady = n = GetByte ();
		while (--n >= 0)
		{	
			int nESid = 0;
			c = GetByte ();
			nESid = MAKEWORD (GetByte (), c);
			/*-------------------------------------------------------------
			If the initialOD is requested, then this is the main file,
			and channel map entries consists of several entries each with
			its own ESid
			If the initialOD is not requested, then this is an
			auxiliary file, and channel map entries consists of several
			entries each with its own string of the form <filename>:<esid>
			-------------------------------------------------------------*/
			if (initialOD)
			{	
				ChannelAssociation aChannel(nESid);
				m_channelMap += aChannel;
			}
			else
			{	
				ChannelAssociation aChannel;
				
				// temporary build the new string
				char *szTmpURL = new char[ strlen( szURL ) + 7 ] ;
				sprintf(szTmpURL,"%s:%d", szURL, nESid);
				
				aChannel.m_pData = new BYTE[ strlen( szTmpURL ) + 1 ];
				strcpy( (char *)aChannel.m_pData, szTmpURL ) ;
				aChannel.m_nDataLen = strlen( szTmpURL ) + 1 ;

				m_channelMap += aChannel;
				delete szTmpURL;
			}
		}
	}
	catch (FileException)
	{
		fclose (m_fInput);
		return RESPONSE_REMOTE_ERROR;
	}

	/* ---------------------------------------------------------------
	Static mux code table read from file. Updates are not implemented. 
	--------------------------------------------------------------- */
	{ 
		char *szMuxCodeFile = new char[ strlen(szURL) + 5 ] ;
		sprintf (szMuxCodeFile,"%s.mct",szURL);
		if (ZFileExists (szMuxCodeFile))
			ReadMuxCodeTableFromFile(szMuxCodeFile, m_muxCodeTable);
		delete[] szMuxCodeFile;
		
	}

	m_bActive = TRUE;
  	if (ZThread<FileTrifInstance> (
// PORTING to CC 5.3
//			FileThread,
			& FileTrifInstance :: FileThread,
			this,
			THREAD_PRIORITY_BELOW_NORMAL
		).Begin () == -1
	)
	{	fclose (m_fInput);
		return RESPONSE_RESOURCES_EXHAUSTED;
	}
	return RESPONSE_OK;
}


/*-------------------------------------------------------------
A serviceDetach request closes the main file service and all
auxiliry services
-------------------------------------------------------------*/
// SOLARIS PORTING - Added "const"
//void FileTrifInstance :: DPI_ServiceDetach_Req (REASON)
void FileTrifInstance :: DPI_ServiceDetach_Req (const REASON)
{
	for (int i = m_auxFiles.GetLength (); --i >= 0;)
	{	m_auxFiles [i] -> Disconnect ();
		delete (m_auxFiles [i]);
		m_auxFiles.DeleteFrom (i);
	}
	Disconnect ();
	GetSink () -> DPI_ServiceDetach_Cnf (m_pMain,RESPONSE_OK);

	return;
}


/*-----------------------------------------------------------------
Signal the file thread to terminate
-----------------------------------------------------------------*/
void FileTrifInstance :: Disconnect()
{ 
	m_mtxChannels.P ();
	m_bTerminate = true;	// this way the FileThread will not scan the channelMap at the end
	m_bActive = FALSE;
	if (m_nChannelWaitedFor >= 0)
	{	
		m_nChannelWaitedFor = -1;
		m_smpWaitForChannel.V ();
	}
	for (int i = m_channelMap.GetLength (); --i >= 0;)
	{
		TRACE("ready=%d  ",m_channelMap[i].m_bReady);
		TRACE("needRew=%d  ",m_channelMap[i].m_bNeedRewind);
		TRACE("esid=%d  ",m_channelMap[i].m_nESid);
		TRACE("\n");

		m_channelMap [i].m_bReady = FALSE;
	}
	m_mtxChannels.V ();

	m_smpActive.P ();
	m_smpActive.V ();
}

/*-------------------------------------------------------------
ChannelsAdd requests to open a set of channels. For each channel,
we check whether is belongs to the main file (no URL) or to an auxiliary
file (URL supplied). In the later case, find the appropriate file service
or create a new one
-------------------------------------------------------------*/
void FileTrifInstance :: DPI_ChannelAdd_Req (const DAIChannelAddParam &inParam)
{
	for (int i = 0; i < inParam.GetLength (); i++)
	{	
		const DAIChannelAddElement &element = inParam [i];

		if(element.m_eDirection != DOWNSTREAM)
		{
			StreamMapElement rMapElement;
			rMapElement.m_nChid = element.m_nChid;
			rMapElement.m_nTmpId = element.m_nTmpId;
			rMapElement.m_pData = NULL;
			rMapElement.m_nDataLen = 0;
			rMapElement.m_nFmxInfo = 0;		// unused field
			GetSink () -> DPI_ChannelAdd_Cnf (m_pMain, rMapElement, RESPONSE_INVALID_DIRECTION);
			continue;
		}

		/*-------------------------------------------------------------
		No URL case relates to this object
		-------------------------------------------------------------*/
		if (element.m_nDataLen == 2)
		{	
			// ES id.
			// Take care that it is encoded in network byte order
			unsigned short t = ntohs(*(unsigned short*)element.m_pData);
			OpenChannel ((BYTE*)&t,element.m_nDataLen,element.m_nChid,element.m_nTmpId);
			continue;
		}
		
		/*-------------------------------------------------------------
		Look if container file already intialized
		-------------------------------------------------------------*/
		char *szTmpURL = new char[element.m_nDataLen];
		strcpy(szTmpURL, (char *)element.m_pData);
		char *szEnd = strrchr(szTmpURL, ':');
		// the last ':' is used to separate the ESID
		if (szEnd)
			memset(szEnd, 0, 1);

		char *szExpandedURL = NULL;

		// if there are no more ':' then this is assumed to be a relative URL,
		// and the expanded URL is computed using the basename of m_url
		// if there are more ':' then this is assumed to be an absolute URL,
		// and the expanded URL is straightforward

		if ((strchr(szTmpURL,':')==NULL) && (strchr(szTmpURL,'|')==NULL)) {
			char *t;
			// PORTING to CC 5.3
			//if (((t=strrchr(m_url,'\\'))==NULL) && ((t=strrchr(m_url,'/'))==NULL)) {
			if (((t=(char *)strrchr(m_url,'\\'))==NULL) && ((t=(char *)strrchr(m_url,'/'))==NULL)) {
				szExpandedURL = new char[strlen(szTmpURL)+strlen(szEnd+1)+2];
				sprintf(szExpandedURL,"%s %s",szTmpURL,szEnd+1);
			} else {
				// SOLARIS PORTING - Added typecast to prevent compiler error
				szExpandedURL = new char[((BYTE *)t - (BYTE *)m_url)+strlen(szTmpURL)+strlen(szEnd+1)+3];
				memcpy((void *)szExpandedURL,(void *)m_url, (BYTE *)t-(BYTE *)m_url+1);
				sprintf(szExpandedURL+((BYTE *)t-(BYTE *)m_url+1),"%s %s",szTmpURL,szEnd+1);
			}
		} else {
			szExpandedURL = new char[strlen(szTmpURL)+strlen(szEnd+1)+2];
			sprintf(szExpandedURL,"%s %s",szTmpURL,szEnd+1);
		}
		if (szEnd) {
			szEnd=szExpandedURL+strlen(szExpandedURL)-strlen(szEnd+1)-1;
			memset(szEnd, 0, 1);
		}

		// PORTING to CC 5.3
		//for (int j = m_auxFiles.GetLength (); --j >= 0;)
        int j = 0;
		for ( j = m_auxFiles.GetLength (); --j >= 0;)
		{
			if (m_auxFiles [j] -> IsMyFile (szExpandedURL))
			{	
				if (szEnd)	
					memset(szEnd, ':', 1);
				m_auxFiles [j] -> OpenChannel ((unsigned char*)szExpandedURL,strlen(szExpandedURL)+1,element.m_nChid,element.m_nTmpId);
				break;
			}
		}

		/*-------------------------------------------------------------
		Open a new file
		-------------------------------------------------------------*/
		if (j < 0)
		{	
			const int nID = m_auxFiles.GetLength () + 1;
			
			FileTrifInstance *pNewService = new FileTrifInstance(m_pMain, nID ) ;
				m_auxFiles += pNewService ;
			pNewService -> SetSink (GetSink ());
			
			const RESPONSE response = pNewService->Connect( szExpandedURL );
			if (response != RESPONSE_OK)
			{	
				StreamMapElement rMapElement;
				rMapElement.m_nChid = element.m_nChid;
				rMapElement.m_nTmpId = element.m_nTmpId;
				rMapElement.m_pData = NULL;
				rMapElement.m_nDataLen = 0;
				rMapElement.m_nFmxInfo = 0;		// unused field
				GetSink () -> DPI_ChannelAdd_Cnf (m_pMain, rMapElement, response);
			}
			else
			{
				if (szEnd)	
					memset(szEnd, ':', 1);
				pNewService -> OpenChannel ((unsigned char*)szExpandedURL,strlen(szExpandedURL)+1,element.m_nChid,element.m_nTmpId);
			}
		}
		delete szTmpURL;
//		delete szExpandedURL;
	}
	return;
}


/*-----------------------------------------------------------------
Mark that a channel is requested and confirm (or not) with DPI_TransmuxAdd_Cnf.
When channels are requested, the function calls
ServiceClientSink::DPI_TransmuxAdd_Cnf() for each requested channel.
-----------------------------------------------------------------*/
void FileTrifInstance :: OpenChannel (BYTE* pData, int nDataLen, UINT nChid, UINT nTmpId)
{
	DPI_TransmuxAdd_CnfParam smt;
	StreamMapElement &rMapElement = smt.ElementAt (0);
	rMapElement.m_pData = NULL;
	rMapElement.m_nDataLen = 0;
	rMapElement.m_nChid = 0;
	rMapElement.m_nTmpId = nTmpId;
	rMapElement.m_nFmxInfo = 0;		// unused field

	const int nTag = TagFromData (pData, nDataLen);

	if (nTag >= 0)
	{	
		rMapElement.m_nChid = nChid;
		m_channelMap[nTag].m_nChid = rMapElement.m_nChid;
		GetSink () -> DPI_ChannelAdd_Cnf (m_pMain, rMapElement, RESPONSE_OK);
	}
	else
	{
		GetSink () -> DPI_ChannelAdd_Cnf (m_pMain, rMapElement, RESPONSE_CHANNEL_NOT_FOUND);
	}
}


/*-----------------------------------------------------------------
Play a channel
-----------------------------------------------------------------*/
// SOLARIS PORTING - Removed "const"
//void FileTrifInstance :: PlayChannel (const int nTag)
void FileTrifInstance :: PlayChannel (int nTag)
{	
	m_mtxChannels.P ();
	/*-----------------------------------------------------------------
	If the stream has already played once, rewind the file and
	play everything again.
	-----------------------------------------------------------------*/
	if (!m_channelMap [nTag].m_bReady)
	{	
		if (m_channelMap [nTag].m_bNeedRewind)
			Rewind ();
		m_channelMap [nTag].m_bReady = TRUE;
//		m_channelMap [nTag].m_bNeedRewind = TRUE;

		/*-----------------------------------------------------------------
		If the file thread is waiting for this channel to be ready,
		signal it to go
		-----------------------------------------------------------------*/
		if (m_nChannelWaitedFor == nTag)
		{	
			m_smpWaitForChannel.V ();
			m_nChannelWaitedFor = -1;
		}
	}
	m_mtxChannels.V ();
}

/*-----------------------------------------------------------------
The file thread reads data from the file and pass it to ServiceClientSink
-----------------------------------------------------------------*/
void FileTrifInstance :: FileThread ()
{
	BYTE buffer [256];
	RESPONSE nTerminationCode = RESPONSE_OK;

	m_smpActive.P ();
	while (m_bActive)
	{
		try
		{	
			/*-----------------------------------------------------------------
			If stream transmission has not been requested yet, wait till it
			is requested
			-----------------------------------------------------------------*/
			const int nFmcNo = GetByte ();
			m_mtxChannels.P ();		// critical section
			if (!m_bActive)
			{
				m_mtxChannels.V ();
				break;
			}

			int chMapLen = m_channelMap.GetLength();

			if (nFmcNo < 240) {
				if (nFmcNo >= chMapLen)	// ignore packets with unknown FmcNo
					break;

				if (!m_channelMap [nFmcNo].m_bReady)
				{	
					m_nChannelWaitedFor = nFmcNo;
					m_mtxChannels.V ();
					m_smpWaitForChannel.P ();
					if (!m_bActive)
						break;
				}
				else
				{	m_channelMap [nFmcNo].m_bInProcess = true;
					m_mtxChannels.V ();
				}

				UINT nLength = GetByte ();
//				printf("FileThread: fmc=%d l=%d, file=%s\n",nFmcNo,(int)nLength,(LPCSTR)m_url);

				if (fread (buffer, 1, nLength, m_fInput) < nLength)
					throw FileException ();
				GetSink () -> DPI_DataReceive_Ind (
					m_pMain,
					m_channelMap [nFmcNo].m_nChid,
					buffer,
					nLength,
					NULL,
					0,
					0
				);
				m_mtxChannels.P ();
				m_channelMap [nFmcNo].m_bInProcess = false;
				bool bReady = m_channelMap [nFmcNo].m_bReady;
				m_mtxChannels.V ();
				if (!bReady)
					StopChannel (nFmcNo, RESPONSE_OK);
			}
			else // muxcode demultiplex
			{
				UINT nLength = GetByte ();
				int nRemain = nLength;
				UINT nVersion = (GetByte () & 0xF0)>>4;
				int nTableVersion = -1;
				BYTE *pBuffer = &buffer[0];
				BOOL bMuxCodeValid = m_muxCodeTable.GetLength() > nFmcNo-240;

				if (bMuxCodeValid) {
					nTableVersion = m_muxCodeTable[nFmcNo-240].version;
					bMuxCodeValid = (nTableVersion == (int)nVersion);
				}

				if (fread (buffer, 1, nLength, m_fInput) < nLength)
					throw FileException ();

				if (bMuxCodeValid) {
					LinearMuxCodeTableElement &mcte = m_muxCodeTable[nFmcNo-240];
					int nSlotCount = mcte.slots.GetLength();

//					printf("FileThread: fmc=%d l=%d, file=%s\n",nFmcNo,(int)nLength,(LPCSTR)m_url);

					while (nRemain > 0) {		// repeat the muxcode pattern as often as needed

						for (int iSlot = 0; iSlot < nSlotCount; iSlot++) 
						{
							int nNumBytes = mcte.slots[iSlot].numberOfBytes;
							int nFmxChId = mcte.slots[iSlot].flexMuxChannel;

							if (nFmxChId >= chMapLen) { //this channel not opened
								nRemain -= nNumBytes;
								pBuffer += nNumBytes;
								if (nRemain <= 0) {
									nRemain = 0;
									break;
								}
								continue;
							}
							if (!m_channelMap [nFmxChId].m_bReady)
							{	
								m_nChannelWaitedFor = nFmxChId;
								m_mtxChannels.V ();
								m_smpWaitForChannel.P ();
								if (!m_bActive)
									break;
							}
							else
							{	m_channelMap [nFmxChId].m_bInProcess = true;
								m_mtxChannels.V ();
							}
							int nActBytes = nRemain - nNumBytes > 0 ? nNumBytes : nRemain;
							GetSink () -> DPI_DataReceive_Ind (
								m_pMain,
								m_channelMap [nFmxChId].m_nChid,
								pBuffer,
								nActBytes,
								NULL,
								0,
								0
							);
							nRemain -= nActBytes;
							pBuffer += nActBytes;

							m_mtxChannels.P ();
							m_channelMap [nFmxChId].m_bInProcess = false;
							bool bReady = m_channelMap [nFmxChId].m_bReady;
							m_mtxChannels.V ();

							if (!bReady)
								StopChannel (nFmxChId, RESPONSE_OK);
							if (nRemain <= 0)
								break;
						}
					}
				}
			}
			if (m_bRewinding)
				RewindFile ();
		}
		catch (FileException except)
		{
			m_bActive = FALSE;
			if (except.m_nCode == FileException :: MEMORY_ERROR)
				nTerminationCode = RESPONSE_RESOURCES_EXHAUSTED;
		}
	}
	m_smpActive.V ();
	m_nChannelWaitedFor = -1;

	/*-------------------------------------------------------------
	If end of a file has reached (not by forced termination), notify
	the sink that all the contained channels have closed.
	-------------------------------------------------------------*/
	if (!m_bTerminate)
	{	
		for (int i = m_channelMap.GetLength (); --i >= 0;)
			if (m_channelMap [i].m_bReady ||
				m_channelMap [i].m_bInProcess
			)
			{	m_channelMap [i].m_bInProcess = false;
				m_channelMap [i].m_bNeedRewind = TRUE;
				StopChannel (i, nTerminationCode);
			}
	}
	else
		m_bTerminate = false;
}

/*-----------------------------------------------------------------
Read one byte and throw exception on end of file
-----------------------------------------------------------------*/
int FileTrifInstance :: GetByte ()
{
	const int c = fgetc (m_fInput);
	if (c == EOF)
		throw FileException ();
	return c;
}

/*-----------------------------------------------------------------
stop a channel from playing
-----------------------------------------------------------------*/
void FileTrifInstance :: StopChannel (
	int nTag,
	RESPONSE nTerminationCode
)
{
	m_mtxChannels.P ();		// critical section
	m_channelMap [nTag].m_bReady = false;
	bool bInProcess = m_channelMap [nTag].m_bInProcess;
	m_mtxChannels.V ();
	if (!bInProcess)
		GetSink () -> DPI_ChannelDelete_Cnf (
			m_pMain,
			m_channelMap [nTag].m_nChid,
			nTerminationCode
		);
}

/*-----------------------------------------------------------------
Rewind the file in order to replay the streams
-----------------------------------------------------------------*/
void FileTrifInstance :: Rewind ()
{
	/*-------------------------------------------------------------
	If file thread is active, signal it to rewind after it finishes its current loop
	otherwise rewind the file and start a new thread
	-------------------------------------------------------------*/
	if (m_bActive)
		m_bRewinding = TRUE;
	else
 	{	Disconnect();						
		RewindFile ();

		/*-----------------------------------------------------------------
		Restart the file thread
		-----------------------------------------------------------------*/
		m_bActive = TRUE;
		m_fileThread.Wait ();
		m_fileThread.Begin ();
	}
}

/*-------------------------------------------------------------
Rewind the file and skip the IOD and channel map
-------------------------------------------------------------*/
void FileTrifInstance :: RewindFile ()
{
	rewind (m_fInput);
	/*-----------------------------------------------------------------
	Skip the initial OD
	-----------------------------------------------------------------*/
	int c = GetByte ();
	int n = MAKEWORD (GetByte (), c);
	while (--n >= 0)
		GetByte ();
	
	/*-----------------------------------------------------------------
	Skip the association map
	-----------------------------------------------------------------*/
	m_nNotReady = n = GetByte ();
	while (--n >= 0)
	{	
		GetByte ();
		GetByte ();
	}

	for (int i = m_channelMap.GetLength (); --i >= 0;)
		m_channelMap [i].m_bNeedRewind = FALSE;
	m_bRewinding = false;
}

/*-------------------------------------------------------------
ChannelsDelete stops the channels in the list.
The function tries to perform this operation on all open files, since
we don't know which file contains this stream.
-------------------------------------------------------------*/
void FileTrifInstance :: DPI_ChannelDelete_Req (const DAIChannelDeleteParam &inParam)
{
	if ( ! CloseChannels( inParam ) )
	{
		int i = m_auxFiles.GetLength () ;
		for ( ; --i >= 0 && !m_auxFiles [i] -> CloseChannels (inParam) ; )
			;
	}
	return;
}

/*-------------------------------------------------------------
Close all the channels in the list and return true if managed to close
all of them (i.e. all are in this file
-------------------------------------------------------------*/
bool FileTrifInstance :: CloseChannels (const DAIChannelDeleteParam &inParam)
{
	bool bAllClosed = true;
	for (int i = inParam.GetLength (); --i >= 0;)
	{	int nTag = TagFromChid ( inParam[i].m_nChid ) ;
		if (nTag >= 0)
		{	
			StopChannel (nTag, RESPONSE_OK);
		}
		else
			bAllClosed = false;
	}
	return bAllClosed;
}			


/*-------------------------------------------------------------
This is my URL if it starts with the correct scheme or
is a valid file name with the correct extension
-------------------------------------------------------------*/
BOOL FileTrifInstance :: IsYourURL (LPCSTR szURL)
{
	return !_strnicmp( szURL, s_myURLScheme, strlen(s_myURLScheme) ) ||
		(ZFileExists( szURL ) && !_strnicmp( szURL + strlen(szURL) - strlen (s_myFileExtension), s_myFileExtension, strlen(s_myFileExtension)));
}

/*-------------------------------------------------------------
Parse and execute a user command
-------------------------------------------------------------*/
void FileTrifInstance :: DPI_UserCommand_Req(const DAIUserCmdParam& inParam)
{
	for(int i=0;i<inParam.m_nChLen;i++)
	{
		assert(inParam.m_nDataLen >= 1);

		int nTag = TagFromChid(inParam.m_pChid[i]);
		if (nTag >= 0)
		{
			if (strncmp(inParam.m_pData, "PLAY", 4)==0) {
				PlayChannel(nTag);
			} else if (strcmp(inParam.m_pData, "PAUSE")==0) {
				StopChannel(nTag, RESPONSE_OK);
			} else {
				;
			}
		}
		else // try with auxiliary files
		{	int j;
			for (j = m_auxFiles.GetLength (); --j >= 0;)
			{
				if ((nTag = m_auxFiles [j] -> TagFromChid(inParam.m_pChid[i])) >= 0)
				{
					if (strncmp(inParam.m_pData, "PLAY", 4)==0) {
						m_auxFiles [j] -> PlayChannel(nTag);
					} else if (strcmp(inParam.m_pData, "PAUSE")==0) {
						m_auxFiles [j] -> StopChannel(nTag, RESPONSE_OK);
					} else {
						;
					}
					break;
				}
			}
		}
	}
	return;
}

void FileTrifInstance :: DPI_UserCommandAck_Req(const DAIUserCmdParam& inParam)
{
	bool bFound = false;
	int	 nFound = 0; 
	DAIUserCmdParam outParam;

	for(int i=0;i<inParam.m_nChLen;i++)
	{
		assert(inParam.m_nDataLen >= 1);

		bFound = false;
		int nTag = TagFromChid(inParam.m_pChid[i]);
		if (nTag >= 0)
		{
			printf ("Channel %d: %s\n", inParam.m_pChid[i], inParam.m_pData);
			bFound = true;
			if (strncmp (inParam.m_pData, "PLAY", 4) == 0)
			{
				PlayChannel(nTag);
			}
			else if (
				strcmp (inParam.m_pData, "PAUSE") == 0 ||
				strcmp (inParam.m_pData, "STOP") == 0
			)
			{
				StopChannel(nTag, RESPONSE_OK);
			}
			else
			{
				;
			}
		}
		else // try with auxiliary files
		{	int j;
			for (j = m_auxFiles.GetLength (); --j >= 0;)
			{
				if ((nTag = m_auxFiles [j] -> TagFromChid(inParam.m_pChid[i])) >= 0)
				{
					bFound = true;
					printf ("Channel %d: %s\n", inParam.m_pChid[i], inParam.m_pData);
					if (strncmp (inParam.m_pData, "PLAY", 4) == 0)
					{
						m_auxFiles [j] -> PlayChannel(nTag);
					}
					else if (
						strcmp (inParam.m_pData, "PAUSE") == 0 ||
						strcmp (inParam.m_pData, "STOP") == 0
					)
						m_auxFiles [j] -> StopChannel(nTag, RESPONSE_OK);
					else
					{
						;
					}
					break;
				}
			}
		}
		if(bFound) nFound++;
	}
	char *string=(nFound < inParam.m_nChLen || inParam.m_nChLen==0) ? "ERROR" : "OK";
	outParam.m_nDataLen=strlen(string)+1;
	outParam.m_pData=new char[outParam.m_nDataLen];
	strcpy(outParam.m_pData,string);
	outParam.m_nChLen=inParam.m_nChLen;
	outParam.m_pChid=inParam.m_pChid;
	outParam.m_nCmdId=inParam.m_nCmdId;
	outParam.m_eResponse=RESPONSE_OK;
	GetSink () -> DPI_UserCommandAck_Cnf(m_pMain,outParam);
	delete outParam.m_pData;
}

int FileTrifInstance :: TagFromChid(UINT chid)
{
	for (int i = m_channelMap.GetLength (); --i >= 0;)
		if (m_channelMap[i].m_nChid == chid)
			return i;
	return -1;
}

int FileTrifInstance :: TagFromData(BYTE *pData, int nDataLen)
{
	for (int i = m_channelMap.GetLength (); --i >= 0;)
		if ((m_channelMap[i].m_nDataLen == nDataLen) && (!memcmp (pData, m_channelMap[i].m_pData,nDataLen)))
			return i;
	return -1;
}


///////////
void FileTrifInstance::DPI_ServiceAttach_Rsp(
	const BYTE*				pUuDataOutBuffer, 
	const int				nUuDataOutLen,
	const RESPONSE			eResponse)
{
	pUuDataOutBuffer; nUuDataOutLen; eResponse;
}

void FileTrifInstance::DPI_ServiceDetach_Rsp(
	const RESPONSE			eResponse)
{
	eResponse;
}

FlexMuxDataParam* FileTrifInstance::DPI_ChannelAdd_Rsp(
	const OnChannelsAddParam& outParam)
{
	outParam; return NULL;
}

FlexMuxDataParam* FileTrifInstance::DPI_TransmuxAdd_Rsp(
	const OnTransmuxAddParam& outParam)
{
	outParam; return NULL;
}

void FileTrifInstance::DPI_ChannelDelete_Rsp(
	const OnChannelsDeleteParam& outParam)
{
	outParam;
}

void FileTrifInstance :: DPI_UserCommandAck_Rsp(
	const DAIUserCmdParam&	outParam)
{
	outParam;
}
///////////


/*-------------------------------------------------------------
Compare the given URL to the file name of this object
-------------------------------------------------------------*/
bool FileTrifInstance :: IsMyFile (LPCSTR szFileName)
{
	/*-------------------------------------------------------------
	If the URL starts with the correct scheme
	it is followed by the file name
	-------------------------------------------------------------*/
	int schemeLen = strlen( s_myURLScheme );
	if ( _strnicmp(szFileName, s_myURLScheme, schemeLen) == 0 )
		szFileName += schemeLen;
	return ( stricmp (szFileName, m_url) == 0 ) ;
}

/*-------------------------------------------------------------
Dummy implementation of MuxCodeTable signaling.
Read the MuxCodeTable from a text file, translate it to a linear
list of slots ((fmxChId,noBytes);(fmxChId,noBytes);...)
and signal it to DMIF
-------------------------------------------------------------*/
void FileTrifInstance :: ReadMuxCodeTableFromFile(
	LPCSTR szFile,
	LinearMuxCodeTable &mcte)
{
	FILE *fInput;
	McteSlot aSlot;
	int nElementCount;

	if ((fInput = fopen (szFile, "rt")) == NULL)
		return;

	fscanf(fInput," %d //elementCount [0..16]",&nElementCount);

	for (int iEl = 0; iEl < nElementCount; iEl++) 
	{
		int nLength, nMuxCode, nSubStrucCount, nVersion;

		fscanf(fInput," %d //length [0..255]",&nLength);
		fscanf(fInput," %d //muxCode [0..15]",&nMuxCode);
		fscanf(fInput," %d //version [0..15]",&nVersion);
		fscanf(fInput," %d //substructureCount [1..255]",&nSubStrucCount);

		{
			LinearMuxCodeTableElement &pMcte = mcte.ElementAt(nMuxCode);
			pMcte.version = nVersion;

			for (int iSub = 0; iSub <  nSubStrucCount; iSub++)
			{
				int nSlotCount, nRepetitionCount;
				int iFirstSlot = pMcte.slots.GetLength();

				fscanf(fInput," %d //slotCount [1..31]",&nSlotCount);
				fscanf(fInput," %d //repetitionCount [1..7]",&nRepetitionCount);

				// store slot pattern for one substructure
				for (int iSlot = 0; iSlot < nSlotCount; iSlot++) 
				{
					fscanf(fInput," %d //fmxChannel [1..255]",&aSlot.flexMuxChannel);
					fscanf(fInput," %d //numberOfBytes [1..255]",&aSlot.numberOfBytes);
					pMcte.slots += aSlot;
				}
				// repeat slot pattern nRepetitionCount-1 times
				for (int iRep = 1; iRep < nRepetitionCount; iRep++)
					for (int iSlot = 0; iSlot < nSlotCount; iSlot++)
						pMcte.slots += (McteSlot) pMcte.slots [iFirstSlot+iSlot];
			}
		}

		fscanf(fInput," ");
	}
//	It does not make sense to cascade FlexMuxes (TRIF _is_ FlexMux!).
//	Other DMIFInstances, however, would signal to DMIF like this.
//	For code verification this call is made but ignored by the sink.

//	if (nElementCount)
//		GetSink () -> DPI_MuxCodeTableReceived_Ind (
//				m_pMain,
//				(HANDLE) 0,
//				mcte,
//				FALSE
//			);
}
