/**************************************************************************
 * Project Im1Player
 * INTERNATIONAL STANDARD ORGANIZATION
 * ISO/IEC JTC1/SC29/WG11
 * 
 * SUBSYSTEM: DMIF
 * FILE:      FlexDemux.cpp
 * AUTHOR:    Mauro Rossi (CSELT)
 * OVERVIEW:  Implementation of the class FlexDemux
 * 
 **************************************************************************/

#include <iostream.h>
#include "ztl.h"
#include "MList.h"
#include "mpg4dmif.h"
#include "FlexDemux.h"
#include "TDxI_Sink.h"
#include "DMXChannel.h"

DemuxInterface* CreateDemuxObject()
{
	return ((DemuxInterface*) new FlexDemux());
}


/*-------------------------------------------------------------
	Constructor
	Fills the m_channelTable with MAX_CHANNELS_PER_FLEXMUX
	NULL channels
-------------------------------------------------------------*/
FlexDemux :: FlexDemux ()
{
	for(int i=0;i<MAX_CHANNELS_PER_FLEXMUX;i++)
		m_channelTable += (DMXChannel*)0;
}


FlexDemux :: ~FlexDemux()
{
	for(int i=0;i<m_channelTable.GetLength();i++)
		delete m_channelTable[i];
	m_channelTable.Clear();
}


/*-----------------------------------------------------------------
A received data packet is demultiplexed and dispatched to the
proper data channel(s).
Note: Multiplex tables are not implemented yet.
-----------------------------------------------------------------*/
void FlexDemux :: DPIC_PacketReceived_Ind (
	const BYTE *pStreamData,
	int nStreamDataLength,
	const BOOL bErr)
{
	/*-------------------------------------------------------------
	If the packet is erronous, don't parse it, and acknowledge all
	FlexMux channels that their data might have been corrupted.
	-------------------------------------------------------------*/
	if (bErr)
	{	
		for (int i = m_channelTable.GetLength (); --i >= 0;)
			if (m_channelTable [i])
				m_channelTable [i] -> DPIC_PacketReceived_Ind (NULL, 0, TRUE);
		return;
	}

	while (nStreamDataLength > 0)
	{
		int nFmxChNum = (0xff /*RBA*/ & *pStreamData++);
		int nLength = (0xff /*RBA*/ & *pStreamData++);
		if (nFmxChNum == 238 || nFmxChNum ==255)	// padding stream
		{
		}
		else if (nFmxChNum == 239)	// CR stream
		{
		}
		else if (nFmxChNum > 239)	// muxcode mode
		{
			int bmxErr = 1;
			int nMCTentry = nFmxChNum-240;
			int nVersion = *pStreamData & 0xF0;

			if (m_muxCodeTable.GetLength() > nMCTentry)
				if (m_muxCodeTable [nMCTentry].version == nVersion)	// Correct version of table entry?
				{
					bmxErr = 0;
					DemuxMuxCodePacket (m_muxCodeTable [nMCTentry], pStreamData+1, nLength-1);
				}

			if (bmxErr) 
			{
			/*-------------------------------------------------------------
			if MuxCodeTable entry does not exist or is not up to date acknowledge 
			all	FlexMux channels that their data might have been corrupted.
			-------------------------------------------------------------*/
				for (int i = m_channelTable.GetLength (); --i >= 0;)
					if (m_channelTable [i])
						m_channelTable [i] -> DPIC_PacketReceived_Ind (NULL, 0, TRUE);
			}
		}
		else if ( nFmxChNum < m_channelTable.GetLength())
		{
			if (m_channelTable [nFmxChNum] != NULL)
			{
				m_channelTable [nFmxChNum] ->
					DPIC_PacketReceived_Ind (pStreamData, nLength, FALSE);
			}
		}
		nStreamDataLength -= nLength + 2;
		pStreamData += nLength;
	}
}

/*-------------------------------------------------------------
Demultiplex one MuxCode packet
-------------------------------------------------------------*/
void FlexDemux :: DemuxMuxCodePacket(
	LinearMuxCodeTableElement &mcte, 
	const BYTE *pStreamData,
	int nLength)
{
	int nSlotCount = mcte.slots.GetLength();

	while (nLength > 0)		// repeat the muxcode pattern as often as needed

		for (int iSlot = 1; iSlot < nSlotCount; iSlot++) 
		{
			int nNumBytes = mcte.slots[iSlot].numberOfBytes;
			int nFmxChId = mcte.slots[iSlot].flexMuxChannel;
			int nActBytes = nLength - nNumBytes > 0 ? nNumBytes : nLength;

			if (m_channelTable [nFmxChId] != NULL)
				m_channelTable [nFmxChId] ->
					DPIC_PacketReceived_Ind (pStreamData, nActBytes, FALSE);
			pStreamData += nActBytes;
			nLength -= nActBytes;
			if (nLength <= 0)
				return;
		}
}

/*-------------------------------------------------------------
Add or modify MuxCodeTableElements for this FlexMux stream
-------------------------------------------------------------*/	
void FlexDemux :: UpdateMuxCodeTableElements(const LinearMuxCodeTable &muxCodeTable)
{
	m_muxCodeTable = muxCodeTable;
	// update of individual MuxCode table entries is not yet supported
}

/*-------------------------------------------------------------
Add a FlexMux channel to the channel table.
-------------------------------------------------------------*/	
bool FlexDemux :: AddFlexmuxChannel (TDxI_Sink* r_pSnk,const UINT v_nChid,const QOSSTREAM r_qosStream,const int v_nFmxChNum,const BYTE* pAppInfo,const int nAppInfoLength)
{
	DMXChannel* pNewChannel =  new DMXChannel(r_pSnk,v_nChid);

	if(pNewChannel == NULL)
	{
		return false;
	}

	if (nAppInfoLength == sizeof(struct SLConfigDesc))
	{
		pNewChannel->m_slcd = *(struct SLConfigDesc*)pAppInfo;
		pNewChannel->m_bUseSlcd = true;
	}

	pNewChannel->SetQos(r_qosStream);

	if(m_channelTable[v_nFmxChNum] != (DMXChannel*)0)
		return false;

	m_channelTable.SetAt(v_nFmxChNum,pNewChannel);
	return true;
}

/*-------------------------------------------------------------
Delete a FlexMux channel from the table and notify the application.
-------------------------------------------------------------*/
void FlexDemux :: DeleteFlexmuxChannel(int nFmxInfo)
{
	Channel *pChannel = m_channelTable[nFmxInfo];
	if (pChannel)
	{	
		pChannel -> DPIC_ChannelDelete_Cnf(RESPONSE_OK);
		delete (m_channelTable[nFmxInfo]);
		m_channelTable[nFmxInfo] = NULL;
	}
}


/*-------------------------------------------------------------
Enables the channel
-------------------------------------------------------------*/
void FlexDemux :: PlayChannel(int nFmxInfo)
{
	Channel *pChannel = m_channelTable[nFmxInfo];
	if (pChannel)
	{	
		((DMXChannel*)pChannel) -> PlayChannel();
	}
}

/*-------------------------------------------------------------
Disables the channel
-------------------------------------------------------------*/
void FlexDemux :: PauseChannel(int nFmxInfo)
{
	Channel *pChannel = m_channelTable[nFmxInfo];
	if (pChannel)
	{	
		((DMXChannel*)pChannel) -> PauseChannel();
	}
}


/*---------------------------------------------
Check if there is at least one valid channel.
---------------------------------------------*/
bool FlexDemux :: isEmpty()
{
	for(int i=0;i<MAX_CHANNELS_PER_FLEXMUX;i++)
	{
		if(m_channelTable[i] != NULL)
			return false;
	}
	return true;
}

/*-------------------------------------------
Get the fmxInfo of the specified channel.
-------------------------------------------*/
int FlexDemux :: GetIndexOf(UINT p_nChId)
{
	int i;
	for(i=0;i<m_channelTable.GetLength();i++)
	{
		if((UINT)m_channelTable[i] == p_nChId)
		{
			return i;
		}
	}
	return -1;
}


UINT FlexDemux :: GetDMXChannelPtr(const int nFmxInfo)
{
	return ( (UINT) (m_channelTable[nFmxInfo]) );
}


unsigned __int64 FlexDemux :: GetTotalPayload() const
{
	int i;
	unsigned __int64 nRet = 0x0000000000000000;

	for(i=0;i<MAX_CHANNELS_PER_FLEXMUX;i++)
	{
		if(m_channelTable[i] != NULL)
			nRet += m_channelTable[i] -> GetTotalPayload();
	}
	return nRet;
}

DWORD FlexDemux :: GetTotalFmxPackets() const
{
	int i;
	DWORD nRet = 0x00000000;

	for(i=0;i<MAX_CHANNELS_PER_FLEXMUX;i++)
	{
		if(m_channelTable[i] != NULL)
			nRet += m_channelTable[i] -> GetTotalFmxPackets();
	}
	return nRet;
}



