/************************************************************************

This software module was originally developed by
        Pat Mulroy (pmulroy@visual.bt.co.uk) BT Labs
        Wei Wu (weiwu@stallion.risc.rockwell.com),  Rockwell Science Center
        Homer Chen (homer@jupiter.risc.rockwell.com),  Rockwell Science Center
        Karl Olav Lillevold (Karl.Lillevold@nta.no) Telenor
	Robert Danielsen (Robert.Danielsen@nta.no) Telenor
        Tolga K. Capin (capin@lig.di.epfl.ch), EPFL-LIG,
        Joaquim Esmerado (jle@lig.di.epfl.ch), EPFL-LIG.

in the course of development of the MPEG-4 Video (ISO/IEC 14496-2).
This software module is an implementation of a part of one or more MPEG-4 Video
tools as specified by the MPEG-4 Video. ISO/IEC gives users of the MPEG-4 Video
free license to this software module or modifications thereof for use in
hardware or software products claiming conformance to the MPEG-4 Video. Those
intending to use this software module in hardware or software products are advised
that its use may infringe existing patents. The original developer of this software
module and his/her company, the subsequent editors and their companies, and ISO/IEC
have no liability for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Video conforming
products. BT Labs, Rockwell, and Telenor retains full right to use the
code for his/her own purpose, assign or donate the code to a third party and to
inhibit third parties from using the code for non <MPEG standard> conforming
products. This copyright notice must be included in all copies or derivative
works. Copyright (c) 1997.

*************************************************************************/

#include <string.h>
#include "sim.h"

#define   q1    16384
#define   q2    32768
#define   q3    49152
#define   top   65535

/* local prototypes */
void bit_out_psc_layer(void);

/*********************************************************************
 *        SAC Decoder Algorithm as Specified in H26P Annex -E
 *
 *        Name:        decode_a_symbol
 *
 *	Description:	Decodes an Aritmetically Encoded Symbol
 *
 *	Input:        array holding cumulative freq. data
 *        also uses static data for decoding endpoints
 *        and code_value variable
 *
 *	Returns:	Index to relevant symbol model
 *
 *	Side Effects:	Modifies low, high, length, cum and code_value
 *
 *	Author:        Wayne Ellis <ellis_w_wayne@bt-web.bt.co.uk>
 *
 *********************************************************************/
 
static long low, high, code_value, bit, length, sacindex, cum, zerorun=0;

int decode_a_symbol(int cumul_freq[ ])
{

  length = high - low + 1;
  cum = (-1 + (code_value - low + 1) * cumul_freq[0]) / length;
  for (sacindex = 1; cumul_freq[sacindex] > cum; sacindex++);
  high = low - 1 + (length * cumul_freq[sacindex-1]) / cumul_freq[0];
  low += (length * cumul_freq[sacindex]) / cumul_freq[0];

  for ( ; ; ) {  
    if (high < q2) ;
    else if (low >= q2) {
      code_value -= q2; 
      low -= q2; 
      high -= q2;
    }
    else if (low >= q1 && high < q3) {
      code_value -= q1; 
      low -= q1; 
      high -= q1;
    }
    else {
      break;
    }
 
    low *= 2; 
    high = 2*high + 1;
    bit_out_psc_layer(); 
    code_value = 2*code_value + bit;
  }

  return (sacindex-1);
}
 
/*********************************************************************
 *
 *        Name:        decoder_reset
 *
 *	Description:	Fills Decoder FIFO after a fixed word length
 *        string has been detected.
 *
 *	Input:        None
 *
 *	Returns:	Nothing
 *
 *	Side Effects:	Fills Arithmetic Decoder FIFO
 *
 *	Author:        Wayne Ellis <ellis_w_wayne@bt-web.bt.co.uk>
 *
 *********************************************************************/

/* 
   Comment by Wei Wu: This function is called right before arithmetic 
   coding after fix length string has been outputed
*/

void decoder_reset( void )
{
  int i;
  zerorun = 0;        /* clear consecutive zero's counter */
  code_value = 0;
  low = 0;
  high = top;
  for (i = 1;   i <= 16;   i++) {
    bit_out_psc_layer(); 
    code_value = 2 * code_value + bit;
  }
/*
  if (trace)
    printf("Arithmetic Decoder Reset \n");
*/
}

/*********************************************************************
 *
 *        Name:        bit_out_psc_layer
 *
 *	Description:	Gets a bit from the Encoded Stream, Checks for
 *        and removes any PSC emulation prevention bits
 *        inserted at the decoder, provides 'zeros' to the
 *        Arithmetic Decoder FIFO to allow it to finish 
 *        data prior to the next PSC. (Garbage bits)
 *
 *	Input:        None
 *
 *	Returns:	Nothing
 *
 *	Side Effects:	Gets a bit from the Input Data Stream
 *
 *	Author:        Wayne Ellis <ellis_w_wayne@bt-web.bt.co.uk>
 *
 *********************************************************************/

void
bit_out_psc_layer(void) {

  bit = getbits(1);
  if(zerorun==22) {
	if(!bit) {
		fprintf(stderr, "Marker bit stuffing error!\n");
		exit(-1);
	}
	bit = getbits(1);
	zerorun = !bit;
	return;
  }

  if(bit==0) {
	zerorun++;
  } else {
	zerorun = 0;
  }

}

/*********************************************************************
 *
 * Code added for adaptive arithmetic coding
 *
 * Author: Wei Wu (weiwu@stallion.risc.rockwell.com)
 *
 *********************************************************************/

void
update(int symbol, int index) {
 
  int n;
  int *cumul_tmp, *freq_tmp;
 
  n = fop.fo.range_max[index]-fop.fo.range_min[index]+1;
  cumul_tmp = fop.fo.cumul_freq[index];
  freq_tmp = fop.fo.freq[index];

  if (cumul_tmp[0] == Max_frequency) {
    int i, cum;
    cum = 0;
    for(i=n-1; i>=0; i--) {
        freq_tmp[i] = (freq_tmp[i]+1)/2;
        cum += freq_tmp[i];
        cumul_tmp[i] = cum;
    }
    cumul_tmp[n] = 0;
  }
  freq_tmp[symbol] += 1;
  cumul_tmp[symbol] += 1;
  while(symbol>0) {
        symbol -= 1;
        cumul_tmp[symbol] += 1;
  }
 
}


void
bap_update(int symbol, int index) {
 
  int n;
  int *cumul_tmp, *freq_tmp;
 
  n = fop.bo.range_max[index]-fop.bo.range_min[index]+1;
  cumul_tmp = fop.bo.cumul_freq[index];
  freq_tmp = fop.bo.freq[index];

  if (cumul_tmp[0] == Max_frequency) {
    int i, cum;
    cum = 0;
    for(i=n-1; i>=0; i--) {
        freq_tmp[i] = (freq_tmp[i]+1)/2;
        cum += freq_tmp[i];
        cumul_tmp[i] = cum;
    }
    cumul_tmp[n] = 0;
  }
  freq_tmp[symbol] += 1;
  cumul_tmp[symbol] += 1;
  while(symbol>0) {
        symbol -= 1;
        cumul_tmp[symbol] += 1;
  }
 
}

void
update_viseme_expression(int symbol, int index) {

int n;
int *cumul_tmp, *freq_tmp;

  n = fop.fo.range_ve_max[index] - fop.fo.range_ve_min[index]+1;
  cumul_tmp = fop.fo.cumul_ve[index];
  freq_tmp = fop.fo.freq_ve[index];

  if (cumul_tmp[0] == Max_frequency) {
    int i, cum;
    cum = 0;
    for(i=n-1; i>=0; i--) {
        freq_tmp[i] = (freq_tmp[i]+1)/2;
        cum += freq_tmp[i];
        cumul_tmp[i] = cum;
    }
    cumul_tmp[n] = 0;
  }
  freq_tmp[symbol] += 1;
  cumul_tmp[symbol] += 1;
  while(symbol>0) {
        symbol -= 1;
        cumul_tmp[symbol] += 1;
  }
 
}

void
init_arith(/*int *fap_mask*/void) {
 
  int i, j, n;
 
  if(fop.fo.fap_mask[0]) {
        free(fop.fo.Ifreq_ve[0]);
        free(fop.fo.Icumul_ve[0]);
        free(fop.fo.Ifreq_ve[1]);
        free(fop.fo.Icumul_ve[1]);
        free(fop.fo.Ifreq_ve[2]);
        free(fop.fo.Icumul_ve[2]);
        n = fop.fo.Irange_ve_max[0] - fop.fo.Irange_ve_min[0] + 1;
        fop.fo.Ifreq_ve[0] = (int*) malloc(n*sizeof(int));
        fop.fo.Icumul_ve[0] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.Ifreq_ve[0][j] = 1;
                fop.fo.Icumul_ve[0][j] = n-j;
        }
        fop.fo.Icumul_ve[0][n] = 0;

        n = fop.fo.Irange_ve_max[1] - fop.fo.Irange_ve_min[1] + 1;
        fop.fo.Ifreq_ve[1] = (int*) malloc(n*sizeof(int));
        fop.fo.Icumul_ve[1] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.Ifreq_ve[1][j] = 1;
                fop.fo.Icumul_ve[1][j] = n-j;
        }
        fop.fo.Icumul_ve[1][n] = 0;
 
        n = fop.fo.Irange_ve_max[2] - fop.fo.Irange_ve_min[2] + 1;
        fop.fo.Ifreq_ve[2] = (int*) malloc(n*sizeof(int));
        fop.fo.Icumul_ve[2] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.Ifreq_ve[2][j] = 1;
                fop.fo.Icumul_ve[2][j] = n-j;
        }
        fop.fo.Icumul_ve[2][n] = 0;
  }

  if(fop.fo.fap_mask[1]) {
        free(fop.fo.Ifreq_ve[3]);
        free(fop.fo.Icumul_ve[3]);
        free(fop.fo.Ifreq_ve[4]);
        free(fop.fo.Icumul_ve[4]);
        free(fop.fo.Ifreq_ve[5]);
        free(fop.fo.Icumul_ve[5]);
        free(fop.fo.Ifreq_ve[6]);
        free(fop.fo.Icumul_ve[6]);
 
        n = fop.fo.Irange_ve_max[3] - fop.fo.Irange_ve_min[3] + 1;
        fop.fo.Ifreq_ve[3] = (int*) malloc(n*sizeof(int));
        fop.fo.Icumul_ve[3] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.Ifreq_ve[3][j] = 1;
                fop.fo.Icumul_ve[3][j] = n-j;
        }
        fop.fo.Icumul_ve[3][n] = 0;
 
        n = fop.fo.Irange_ve_max[4] - fop.fo.Irange_ve_min[4] + 1;
        fop.fo.Ifreq_ve[4] = (int*) malloc(n*sizeof(int));
        fop.fo.Icumul_ve[4] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.Ifreq_ve[4][j] = 1;
                fop.fo.Icumul_ve[4][j] = n-j;
        }
        fop.fo.Icumul_ve[4][n] = 0;
 
        n = fop.fo.Irange_ve_max[5] - fop.fo.Irange_ve_min[5] + 1;
        fop.fo.Ifreq_ve[5] = (int*) malloc(n*sizeof(int));
        fop.fo.Icumul_ve[5] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.Ifreq_ve[5][j] = 1;
                fop.fo.Icumul_ve[5][j] = n-j;
        }
        fop.fo.Icumul_ve[5][n] = 0;
 
        n = fop.fo.Irange_ve_max[6] - fop.fo.Irange_ve_min[6] + 1;
        fop.fo.Ifreq_ve[6] = (int*) malloc(n*sizeof(int));
        fop.fo.Icumul_ve[6] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.Ifreq_ve[6][j] = 1;
                fop.fo.Icumul_ve[6][j] = n-j;
        }
        fop.fo.Icumul_ve[6][n] = 0;
  }

  for(i=2; i<NFAP; i++) {
	if(fop.fo.fap_mask[i]) {
		free(fop.fo.Ifreq[i]);
		free(fop.fo.Icumul_freq[i]);
                n = fop.fo.Irange_max[i]-fop.fo.Irange_min[i]+1;
                fop.fo.Ifreq[i] = (int *) malloc(n*sizeof(int));
                fop.fo.Icumul_freq[i] = (int *) malloc((n+1)*sizeof(int));
                for(j=0; j<n; j++) {
                        fop.fo.Ifreq[i][j] = 1;
                        fop.fo.Icumul_freq[i][j] = n-j;
                }
                fop.fo.Icumul_freq[i][n] = 0;
	}
  }

  if(fop.fo.fap_mask[0]) {
        free(fop.fo.freq_ve[0]);
        free(fop.fo.cumul_ve[0]);
        free(fop.fo.freq_ve[1]);
        free(fop.fo.cumul_ve[1]);
        free(fop.fo.freq_ve[2]);
        free(fop.fo.cumul_ve[2]);
        n = fop.fo.range_ve_max[0] - fop.fo.range_ve_min[0] + 1;
        fop.fo.freq_ve[0] = (int*) malloc(n*sizeof(int));
        fop.fo.cumul_ve[0] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.freq_ve[0][j] = 1;
                fop.fo.cumul_ve[0][j] = n-j;
        }
        fop.fo.cumul_ve[0][n] = 0;
 
        n = fop.fo.range_ve_max[1] - fop.fo.range_ve_min[1] + 1;
        fop.fo.freq_ve[1] = (int*) malloc(n*sizeof(int));
        fop.fo.cumul_ve[1] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.freq_ve[1][j] = 1;
                fop.fo.cumul_ve[1][j] = n-j;
        }
        fop.fo.cumul_ve[1][n] = 0;
 
        n = fop.fo.range_ve_max[2] - fop.fo.range_ve_min[2] + 1;
        fop.fo.freq_ve[2] = (int*) malloc(n*sizeof(int));
        fop.fo.cumul_ve[2] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.freq_ve[2][j] = 1;
                fop.fo.cumul_ve[2][j] = n-j;
        }
        fop.fo.cumul_ve[2][n] = 0;
  }

  if(fop.fo.fap_mask[1]) {
        free(fop.fo.freq_ve[3]);
        free(fop.fo.cumul_ve[3]);
        free(fop.fo.freq_ve[4]);
        free(fop.fo.cumul_ve[4]);
        free(fop.fo.freq_ve[5]);
        free(fop.fo.cumul_ve[5]);
        free(fop.fo.freq_ve[6]);
        free(fop.fo.cumul_ve[6]);
        n = fop.fo.range_ve_max[3] - fop.fo.range_ve_min[3] + 1;
        fop.fo.freq_ve[3] = (int*) malloc(n*sizeof(int));
        fop.fo.cumul_ve[3] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.freq_ve[3][j] = 1;
                fop.fo.cumul_ve[3][j] = n-j;
        }
        fop.fo.cumul_ve[3][n] = 0;
 
        n = fop.fo.range_ve_max[4] - fop.fo.range_ve_min[4] + 1;
        fop.fo.freq_ve[4] = (int*) malloc(n*sizeof(int));
        fop.fo.cumul_ve[4] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.freq_ve[4][j] = 1;
                fop.fo.cumul_ve[4][j] = n-j;
        }
        fop.fo.cumul_ve[4][n] = 0;
 
        n = fop.fo.range_ve_max[5] - fop.fo.range_ve_min[5] + 1;
        fop.fo.freq_ve[5] = (int*) malloc(n*sizeof(int));
        fop.fo.cumul_ve[5] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.freq_ve[5][j] = 1;
                fop.fo.cumul_ve[5][j] = n-j;
        }
        fop.fo.cumul_ve[5][n] = 0;
 
        n = fop.fo.range_ve_max[6] - fop.fo.range_ve_min[6] + 1;
        fop.fo.freq_ve[6] = (int*) malloc(n*sizeof(int));
        fop.fo.cumul_ve[6] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.freq_ve[6][j] = 1;
                fop.fo.cumul_ve[6][j] = n-j;
        }
        fop.fo.cumul_ve[6][n] = 0;
  }

  for(i=2; i<NFAP; i++) {
	if(fop.fo.fap_mask[i]) {
		free(fop.fo.freq[i]);
		free(fop.fo.cumul_freq[i]);
                n = fop.fo.range_max[i]-fop.fo.range_min[i]+1;
                fop.fo.freq[i] = (int *) malloc(n*sizeof(int));
                fop.fo.cumul_freq[i] = (int *) malloc((n+1)*sizeof(int));
                for(j=0; j<n; j++) {
                        fop.fo.freq[i][j] = 1;
                        fop.fo.cumul_freq[i][j] = n-j;
                }
                fop.fo.cumul_freq[i][n] = 0;
	}
  }
}


void
bap_init_arith(/*int *fap_mask*/void) {
 
  int i, j, n;
 
  for(i=0; i<NUM_BAPS; i++) {
	if(fop.bo.bap_mask[i]) {
		free(fop.bo.Ifreq[i]);
		free(fop.bo.Icumul_freq[i]);
                n = fop.bo.Irange_max[i]-fop.bo.Irange_min[i]+1;
                fop.bo.Ifreq[i] = (int *) malloc(n*sizeof(int));
                fop.bo.Icumul_freq[i] = (int *) malloc((n+1)*sizeof(int));
                for(j=0; j<n; j++) {
                        fop.bo.Ifreq[i][j] = 1;
                        fop.bo.Icumul_freq[i][j] = n-j;
                }
                fop.bo.Icumul_freq[i][n] = 0;
	}
  }

  for(i=0; i<NUM_BAPS; i++) {
	if(fop.bo.bap_mask[i]) {
		free(fop.bo.freq[i]);
		free(fop.bo.cumul_freq[i]);
                n = fop.bo.range_max[i]-fop.bo.range_min[i]+1;
                fop.bo.freq[i] = (int *) malloc(n*sizeof(int));
                fop.bo.cumul_freq[i] = (int *) malloc((n+1)*sizeof(int));
                for(j=0; j<n; j++) {
                        fop.bo.freq[i][j] = 1;
                        fop.bo.cumul_freq[i][j] = n-j;
                }
                fop.bo.cumul_freq[i][n] = 0;
	}
  }
}

#if 0
void bit_out_psc_layer()
{
  if (showbits(17)!=1) { /* check for startcode in Arithmetic Decoder FIFO */

    bit = getbits(1);

    if(zerorun > 13) {	/* if number of consecutive zeros = 14 */	 
      if (!bit) {
        if (trace)
          printf("PSC/GBSC, Header Data, or Encoded Stream Error \n");
        zerorun = 1;        
      }
      else { /* if there is a 'stuffing bit present */
        if (trace)
          printf("Removing Startcode Emulation Prevention bit \n");
        bit = getbits(1);        /* overwrite the last bit */	
        zerorun = !bit;        /* zerorun=1 if bit is a '0' */
      }
    }

    else { /* if consecutive zero's not exceeded 14 */

      if (!bit)
        zerorun++;
      else
        zerorun = 0;
    }

  } /* end of if !(showbits(17)) */

  else {
    bit = 0;
    if (trace)
      printf("Startcode Found:Finishing Arithmetic Decoding using 'Garbage bits'\n");
  }

   /*	
   printf("lastbit = %ld bit = %ld zerorun = %ld \n", lastbit, bit, zerorun); 
   lastbit = bit;
   */
        /* latent diagnostics */
        
 }
#endif

