//=============================================================================
//  This software module was originally developed by IBM Research (T.J. Watson)
//  in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) 
//  standard. This software module is an implementation of a part of one or 
//  more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 
//  Systems (ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 
//  Systems (ISO/IEC 14496-1) free license to this software module or 
//  modifications thereof for use in hardware or software products claiming 
//  conformance to the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use
//  this software module in hardware or software products are advised that its
//  use may infringe existing patents. The original developer of this software 
//  module and his/her company, the subsequent editors and their companies, and
//  ISO/IEC have no liability for use of this software module or modifications 
//  thereof in an implementation. Copyright is not released for non MPEG-4 
//  Systems (ISO/IEC 14496-1) conforming products. IBM retain full right to use
//  the code for their own purpose, assign or donate the code to a third party 
//  and to inhibit third parties from using the code for non MPEG-4 Systems 
//  (ISO/IEC 14496-1) conforming products. This copyright notice must be 
//  included in all copies or derivative works. Copyright  1999-2000.
//=============================================================================
//  IBM  TJ Watson Research, USA
//-----------------------------------------------------------------------------
//  FLEXTIMECONNECTION.CPP
//
//  Contact   :
//    Dr. Michelle Kim ( mykim@us.ibm.com )
//
//  Overview  :
//    Implementation of FlexTimeConnection
//
//=============================================================================  


#include <stdio.h>
#include <stdlib.h>
#include "defs.h"
#include <assert.h>

//  FlexTime stuff
#include "FlexTimeBase.h"
#include "FlexTimeMgr.h"
//#include "Spring.h"
#include "spring.h" // SOLARIS PORTING - typo

#include "FlexTimeObject.h"
#include "FlexTimeConnection.h"


//-----------------------------------------------------------------------------
//  Construct a new empty FlexTime Connection.
//-----------------------------------------------------------------------------
FlexTimeConnection::FlexTimeConnection(int nId)
{
  //  We also want a valid id
  assert(nId>=0);
  m_nId = nId;
  
  m_bHasSwitched = false;

  //  This boolean means that the object hasn't been evaluated (with its method
  //  evaluateSwitchOver)
  m_bFirstEvalTimeNotSet = true;
  
  // Init with a meaningless value
  m_dLastEvalTime = -1.0;
}

FlexTimeConnection::~FlexTimeConnection()
{
  //  these two asserts are here to be sure that all connections have been 
  //  detached before we destroy all these objects
  assert(startingObjects.GetLength()==0);
  assert(endingObjects.GetLength()==0);
}


//-----------------------------------------------------------------------------
//  Add a starting object. This will never result into duplicate objects :
//  if the object already exists as an ending object, the object will not be 
//  added.
//-----------------------------------------------------------------------------
void FlexTimeConnection::AddStartingObject(FlexTimeObject *pObject)
{
  assert(pObject);
  
  //  Don't duplicate if the object already exists: just return
  if(pObject->GetStartConnection() == this) 
    return;
  
  pObject->SetStartConnection(this);
  
  //  Add it to the list of the starting objects
  startingObjects += pObject;
}


//-----------------------------------------------------------------------------
//  Add a ending object. This will never result into duplicate objects: if the 
//  object already exists as an ending object, it will not be added.
//-----------------------------------------------------------------------------
void FlexTimeConnection::AddEndingObject(FlexTimeObject *pObject)
{
  assert(pObject);
  
  // don't duplicate if the object already exists: just return
  if(pObject->GetEndConnection() == this) 
    return;
  
  pObject->SetEndConnection(this);
  
  //  Add it to the list of the starting objects
  endingObjects   +=  pObject;
}


//-----------------------------------------------------------------------------
//  Delete a starting object.
//-----------------------------------------------------------------------------
void FlexTimeConnection::DeleteStartingObject(FlexTimeObject *pObject)
{
  assert(pObject);
  
  //  Delete it from the list
  startingObjects -= pObject;
}


//-----------------------------------------------------------------------------
//  Delete a ending object. 
//-----------------------------------------------------------------------------
void FlexTimeConnection::DeleteEndingObject(FlexTimeObject *pObject)
{
  assert(pObject);
  
  //  Delete it from the list
  endingObjects   -=  pObject;
}


//-----------------------------------------------------------------------------
//  Get the number of "starting" objects.
//-----------------------------------------------------------------------------
int FlexTimeConnection::GetNumStartingObjects()
{
  return startingObjects.GetLength();
}


//-----------------------------------------------------------------------------
//  Get the number of "Ending" objects.
//-----------------------------------------------------------------------------
int FlexTimeConnection::GetNumEndingObjects()
{
  return endingObjects.GetLength();
}


//-----------------------------------------------------------------------------
//  Get the "starting" object pointer at given index.
//  @return The FlexTime Object pointer, or null if the index is out of bounds.
//-----------------------------------------------------------------------------
FlexTimeObject *FlexTimeConnection::GetStartingObject(int iIndex)
{
  return startingObjects.ElementAt(iIndex);
}


//-----------------------------------------------------------------------------
//  Get the "ending" object pointer at given index.
//  @return The FlexTime Object pointer, or null if the index is out of bounds.
//-----------------------------------------------------------------------------
FlexTimeObject *FlexTimeConnection::GetEndingObject(int iIndex)
{
  return endingObjects.ElementAt(iIndex);
}


//-----------------------------------------------------------------------------
//  Evaluate the switch-over for this Connection and execute it if appropriate.
//  @param curTime Current time.
//  @return True if the switch-over took place; false otherwise.
//-----------------------------------------------------------------------------
bool FlexTimeConnection::evaluate(double curTime)
{
  //this will be changed to the actual SwitchOverTime within the following call
  double dActualSwitchOverTime = curTime;
  if( evaluateSwitchOver(curTime, dActualSwitchOverTime) ) 
  {
    executeSwitchOver(dActualSwitchOverTime);
    return true;
  }
  else
    return false;
}


//-----------------------------------------------------------------------------
//  Evaluate whether this Connection should switch-over by ending the "ending"
//  objects and starting the "starting" objects.
//  @param curTime Current time.
//  @return True if the switch-over must take place; false otherwise.
//-----------------------------------------------------------------------------
bool FlexTimeConnection::evaluateSwitchOver(double dCurTime, 
                                            double &dActualSwitchOverTime)
{
  //  If it has already switched, returns
  if(m_bHasSwitched)
    return false;
  
  if(m_bFirstEvalTimeNotSet)
  {
    m_bFirstEvalTimeNotSet = false;
    m_dFirstEvalTime = dCurTime;
  }
  
  bool bTraceOn = true;
  
  m_dLastEvalTime = dCurTime;
  
  //the following will give back the time where it should have ideally stopped
  dActualSwitchOverTime = dCurTime;
  
  
  //  We DO switch when AT LEAST ONE ENDING object's MAXIMUM end time has 
  //  expired
  int iCount = GetNumEndingObjects();
  for(int iNum=0; iNum<iCount; iNum++) 
  {
    if(endingObjects[iNum]->MaxTimeReached(dCurTime, dActualSwitchOverTime)) 
    { 
      if(bTraceOn)
      {
        //  miniTrace();
        TRACE(" The following ENDING object's MAXIMUM end time has expired:\n");
        endingObjects[iNum] ->trace();
      }
      return true;
    }
  } 
  
  //  We DON'T switch if NOT all STARTING objects have their start AU received
  iCount = GetNumStartingObjects();
  for(iNum=0; iNum<iCount; iNum++) 
  {
    if( !startingObjects[iNum]->IsReadyToStart())
      return false;
    /*
    {
      if(bTraceOn) 
      {
        TRACE("\rFlextime> %.1f did NOT switch connection\n",
              dCurTime-m_dFirstEvalTime);
        miniTrace();
        TRACE(" because NOT all STARTING obj have their start AU received\n");
      }
      return false;
    }
    */
  }
  
  //  We DON'T switch when NOT ALL ENDING objects are ready to end 
  //  (MinTime + have started)
  iCount = GetNumEndingObjects();
  for(iNum=0; iNum<iCount; iNum++) 
  {
    if( !endingObjects[iNum]->IsReadyToEnd(dCurTime))
      return false;
    /*
    {
      if(bTraceOn) 
      {
        TRACE("\rFlextime> %.1f did NOT switch connection\n",
              dCurTime-m_dFirstEvalTime);
        miniTrace();
        TRACE(" because NOT all ENDING obj are ready to stop\n");
      }
      return false;
    }
    */
  }
  
  //  We switch when one of the HIGHEST PRIORITY ENDING object's OPTIMUM
  //  time has expired
  float fHighestPriorityValue = -1;
  int   iHighestPriorityIndex = 0;
  
  iCount = GetNumEndingObjects();
  //  Find the fHighestPriority value
  for(iNum=0; iNum<iCount; iNum++) 
  {
    if(endingObjects[iNum]->GetPriority() >= fHighestPriorityValue) 
    {
      fHighestPriorityValue = endingObjects[iNum]->GetPriority();
      iHighestPriorityIndex = iNum;
    }
  }
  
  //  We then check for OptTime Reached any of the objects that has the 
  //  fHighestPriorityValue 
  for(iNum=0; iNum<iCount; iNum++) 
  {
    if(endingObjects[iNum]->GetPriority() >= fHighestPriorityValue) 
    {
      //to bedone : we may want to update to the smallest dActualSwitchOverTime
      if (endingObjects[iNum]->OptTimeReached(dCurTime,dActualSwitchOverTime))
      {
        if(bTraceOn) 
        {
          TRACE("\nFlextime> %.3f switching connection %d \n",
                dActualSwitchOverTime-m_dFirstEvalTime, m_nId);
          //miniTrace();
          TRACE(" OPTIMUM time has expired for the following HIGHEST PRIORITY ENDING obj :\n");
          endingObjects[iNum]->trace();
        }
        return true;
      }
    }
  }
  
  // Still waiting for Optimal Time of the highestPriority Object
  return false;
}


//-----------------------------------------------------------------------------
//  Execute the switch-over for this Connection 
//-----------------------------------------------------------------------------
void FlexTimeConnection::executeSwitchOver(double dTime)
{
  FlexTimeObject *pFTObject;
  m_bHasSwitched = true;
  
  //PAB_NL: End objects first, start the others after 
  int iCount = GetNumEndingObjects();
  TRACE("\nFlexTimeConnection %2d>     end   %2d ending   objects\n", 
    m_nId, iCount);
  
  
  for(int i=0;i<iCount;i++) 
  {
    pFTObject = GetEndingObject(i);
    pFTObject->trace();
    pFTObject->End(dTime);
  }
  
  //PAB_NL: this has updated the feasibleMin and feasibleMax ==>
  // recompute the TCN.
  GetFlexTimeMgr()->UpdateFeasibleDurations();
  iCount = GetNumStartingObjects();
  TRACE("\nFlexTimeConnection %2d>     start %2d starting objects\n", 
    m_nId,iCount );
  
  for(i=0; i<iCount; i++) 
  {
   	pFTObject = GetStartingObject(i);
    pFTObject->trace();
    pFTObject->Start(dTime);
  }
  
}


//-----------------------------------------------------------------------------
//  Force the switch-over for this Connection 
//-----------------------------------------------------------------------------
void FlexTimeConnection::ForceSwitchOver(double dTime)
{
  executeSwitchOver(dTime);
}


//-----------------------------------------------------------------------------
//  Print the contents.
//-----------------------------------------------------------------------------
void FlexTimeConnection::trace()
{
  TRACE("    Connection Id: %d\n", GetId());
  
  int iCount = GetNumEndingObjects();
  if(iCount)
  {
    TRACE("    EndObjects   :\n");
    for(int iNum=0; iNum<iCount; iNum++) 
      endingObjects[iNum]->trace();
  }
  
  iCount = GetNumStartingObjects();  
  if(iCount)
  {
    TRACE("    StartObjects :\n");
    for(int iNum=0; iNum<iCount; iNum++) 
      startingObjects[iNum]->trace();
  }
}


//-----------------------------------------------------------------------------
//  Print the contents.
//-----------------------------------------------------------------------------
void FlexTimeConnection::miniTrace()
{
  int iCount = GetNumEndingObjects();
  for(int iNum=0; iNum<iCount; iNum++) 
  {
    TRACE("       end   ");
    endingObjects[iNum]->trace();
  }
  
  iCount = GetNumStartingObjects();
  for(iNum=0; iNum<iCount; iNum++) 
  {
    TRACE("       start ");
    startingObjects[iNum]->trace();
  }
}


bool FlexTimeConnection::IsClass(const char *pszClassName) const
{
  return ::stricmp("FlexTimeConnection",pszClassName)==0;
}
