//=============================================================================
//  This software module was originally developed by IBM Research (T.J. Watson)
//  in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) 
//  standard. This software module is an implementation of a part of one or 
//  more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 
//  Systems (ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 
//  Systems (ISO/IEC 14496-1) free license to this software module or 
//  modifications thereof for use in hardware or software products claiming 
//  conformance to the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use
//  this software module in hardware or software products are advised that its
//  use may infringe existing patents. The original developer of this software 
//  module and his/her company, the subsequent editors and their companies, and
//  ISO/IEC have no liability for use of this software module or modifications 
//  thereof in an implementation. Copyright is not released for non MPEG-4 
//  Systems (ISO/IEC 14496-1) conforming products. IBM retain full right to use
//  the code for their own purpose, assign or donate the code to a third party 
//  and to inhibit third parties from using the code for non MPEG-4 Systems 
//  (ISO/IEC 14496-1) conforming products. This copyright notice must be 
//  included in all copies or derivative works. Copyright  1999-2000.
//=============================================================================
//  IBM  TJ Watson Research, USA
//-----------------------------------------------------------------------------
//  FLEXTIMEOBJECT.CPP
//
//  Contact   :
//    Dr. Michelle Kim ( mykim@us.ibm.com )
//
//  Overview  :
//    Implementation of the FlexTimeObject 
//
//  Comment   :
//
//=============================================================================  

/******************************************************************
* FlexTime Object.
*
* @since November 1999.
* @version Februari 2000.
* @author Peter Westerink.
* @comment update by Patrice April 2000
*****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>


#include "defs.h"
#include "../core/Inline.h"
#include "../core/ODImp.h"
//#include "../core/stream.h"  
#include "../core/Stream.h"  // SOLARIS PORTING - typo


// needed to get the system time in the Evaluate() function
#include <sys/timeb.h>
#include <time.h>

//  Flextime stuff
#include "FlexTimeBase.h"
#include "FlexTimeMgr.h"
//#include "Spring.h"
#include "spring.h"  // SOLARIS PORTING - typo
#include "FlexTimeObject.h"
#include "TemporalNode.h"
#include "FlxTime.h"
#include "FlexTimeConnection.h"
#include "TemporalTransformImpFlex.h"


//  This sets how PlayBackDuration will be from the margin (minD or MaxD) 
//  when optDuration is off anyway
#define BOUND_MARGIN 0.10 

#ifdef _DEBUG
int FlexTimeObject::s_nIdCounter = 0;
#endif //_DEBUG


//=============================================================================
//  Construct a new FlexTime object.
//=============================================================================
FlexTimeObject::FlexTimeObject( double dMin, double dMax, double dOpt,float,
                               TemporalNode *pTemporalNode)
{
  m_bActive         = false;
  
  m_dStartTime  =  0.0;
  m_dEndTime    = -1.0;
  
  m_bStartAuReceived  = false;
  m_bEndAuReceived    = false;
  
  m_fPriority = -1.0;
  SetFeasibleMin(dMin);
  SetFeasibleMax(dMax);
  
  optDuration = dOpt;  
  m_dActualDuration = optDuration;
  
  m_pStartFTConnection = m_pEndFTConnection = 0;
  
  m_pTemporalNode = pTemporalNode;
  odId=0;
#ifdef _DEBUG
  m_nId = s_nIdCounter++;
  static int s_nIdCounter;
#endif //_DEBUG
}


FlexTimeObject::~FlexTimeObject()
{
  //  get the start and the end connection so we'll be able to detach this  
  //  object
  FlexTimeConnection *pftcSart = GetStartConnection();
  FlexTimeConnection *pftcEnd  = GetEndConnection();
  
  if(pftcSart)
    pftcSart->DeleteStartingObject(this);
  
  if(!pftcSart->IsLinked())
    GetFlexTimeMgr()->DeleteFTConnection(pftcSart);
  
  if(pftcEnd)
    pftcEnd->DeleteEndingObject(this);
  
  
  if(!pftcEnd->IsLinked())
    GetFlexTimeMgr()->DeleteFTConnection(pftcEnd);
  
}


//=============================================================================
//  Start (activate) this object. This is done by the FlexTimeManager when a 
//  FlexTimeConnection switches over and this object is a starting object of 
//  that connection.
//=============================================================================
void FlexTimeObject::Start(double dStartTime)
{
  m_dStartTime = dStartTime;
  
  float fSpeed = 1.0;
  FlexTimeObject *pFTSibling;
  float fHighestPriority = GetPriority();
  
  double dOptDuration = GetOptDuration();
  m_dActualDuration = dOptDuration;
  
  double dFeasMin = GetFeasibleMin();
  double dFeasMax = GetFeasibleMax();
  FlexTimeConnection *pEndCon = GetEndConnection();
  
  //  check how long we should play given the other objects which we will have
  //  to coend with:
  for(int i=0; i<pEndCon->GetNumEndingObjects(); i++)
  {
    pFTSibling = pEndCon->GetEndingObject(i);
    
    if( pFTSibling->IsActive() && pFTSibling->GetPriority()>=fHighestPriority )
    {
      fHighestPriority = pFTSibling->GetPriority();
      m_dActualDuration = pFTSibling->GetActualDuration() +
        pFTSibling->GetStartTime() - dStartTime;
    }
  }
  
  //  compute the playback speed, so that
  //  FeasMin <= ActualDuration <= FeasMax
  //  if OptDuration is already between the boundary, then we default to that
  //  ActualDuration of course.
  if( m_dActualDuration < dFeasMin )
  {
    m_dActualDuration = dFeasMin;
    //  If there is some margin between FeasMin and FeasMax, choose a 
    //  ActualDuration at BOUND_MARGIN of FeasMin
    if(dFeasMin < dFeasMax) 
      m_dActualDuration += BOUND_MARGIN * (dFeasMax-dFeasMin);
  }
  //else if(dOptDuration >= dFeasMax)
  else if(m_dActualDuration > dFeasMax)
  {
    m_dActualDuration = dFeasMax;
    
    //  if there is some margin between FeasMin and FeasMax, choose a 
    //  playbackDuration at BOUND_MARGIN of FeasMax
    if(dFeasMin < dFeasMax) 
      m_dActualDuration -= BOUND_MARGIN * (dFeasMax-dFeasMin);
    
  }
  if( m_dActualDuration != 0 )
    //  which should never happen anyway
    fSpeed = dOptDuration/m_dActualDuration;
  
  if(m_pTemporalNode)
  {
    m_pTemporalNode->Start(fSpeed);
  }
  m_bActive = true;
  
}


//=============================================================================
//  Returns true if the object is ready to play
//=============================================================================
bool FlexTimeObject::IsReadyToStart()
{
  //PAB: uses the Core 5.0 TemporalTransform helpers
  // to get the state of the media pointed by url.
  //currently have problem with video streams, ok with JPEG.
  if(m_bStartAuReceived)
    return true;
  
  m_bStartAuReceived = m_pTemporalNode->IsStreamActive();
  
#ifdef _DEBUG
  if(m_bStartAuReceived)
    TRACE ("Stream ready for FTObject %d.\n", m_nId);
#endif //_DEBUG
  
  return	m_bStartAuReceived ;  
}


//=============================================================================
//  Returns true if the object is ready to stop
//=============================================================================
bool FlexTimeObject::IsReadyToEnd(double dTime)
{
  //PAB_NL: we currently consider that the data is always there:
  // this will have to be removed once the stream input will be implemented
  //PAB_NL-: return (m_bActive && (dTime >= getMinEndTime()) && m_bEndAuReceived);
  return m_bActive && (dTime>=getMinEndTime()); //PAB_NL+:
}


//=============================================================================
//  Returns true if the object has played for more than optimal duration
//=============================================================================
bool FlexTimeObject::OptTimeReached(double dTime, double &/*dActualOptTime*/)
{
  //PAB-: dActualOptTime = getOptEndTime();
  //PAB-: this does not work if we have reached OptTime a while ago because we 
  //PAB-: were waiting for something else.
  //PAB-: return m_bActive && (dTime >= dActualOptTime);
  return m_bActive && (dTime >= getOptEndTime());
}


//=============================================================================
//  Returns true if the object has played for more than max Duration
//=============================================================================
bool FlexTimeObject::MaxTimeReached(double dTime, double &dActualSwitchOverTime)
{
  bool bResult = m_bActive && (dTime >= getMaxEndTime());
  if(bResult)
    dActualSwitchOverTime = getMaxEndTime();
  
  return bResult;
}


//=============================================================================
//  Get the start connection of this object.
//=============================================================================
void FlexTimeObject::putStartAu()
{
  m_bStartAuReceived = true;
}


//=============================================================================
//  Get the end connection of this object.
//=============================================================================
void FlexTimeObject::putEndAu()
{
  m_bEndAuReceived = true;
}


//=============================================================================
//  End (de-activate) this object. This is done by the FlexTimeManager when a 
//  FlexTimeConnection switches over and this object is an ending object of 
//  that connection.
//=============================================================================
void FlexTimeObject::End(double dEndTime)
{
  m_bActive = false;
  m_dEndTime =  dEndTime;
  
  m_pTemporalNode->End();
  //  TRACE("End OD: %d\n", odId);
  
  //  Fix the Duration it played for, in order to help the TCNSolver this is a 
  //  quick and easy way. A better way would be to replace the constraints set,
  //  in order to avoid to have played FTObjects in the TCN.
  double dActualPlaybackDuration = dEndTime - m_dStartTime;
  SetOptDuration(dActualPlaybackDuration);
  SetFeasibleMin(dActualPlaybackDuration);
  SetFeasibleMax(dActualPlaybackDuration);
}


//=============================================================================
//  Get the maximum end time. Note, that if the object is not active (it has 
//  not been started yet), the returned value is not valid.
//=============================================================================
double FlexTimeObject::getMaxEndTime()
{
  //PAB_NL: return m_dStartTime+maxDuration;
  return m_dStartTime + GetFeasibleMax();
}


//=============================================================================
//  Get the minimum end time. Note, that if the object is not active (it has 
//  not been started yet), the returned value is not valid.
//=============================================================================
double FlexTimeObject::getMinEndTime()
{
  //PAB_NL: return m_dStartTime+minDuration;
  return m_dStartTime + GetFeasibleMin();
}


//=============================================================================
//  Get the optimum end time. Note, that if the object is not active (it has 
//  not been started yet), the returned value is not valid.
//=============================================================================
double FlexTimeObject::getOptEndTime()
{
  //PAB_NL: I may move to have a feasible Optimal duration...
  return m_dStartTime + optDuration;
}


//=============================================================================
//  Print the contents.
//=============================================================================
void FlexTimeObject::trace()
{
  int nStartId = NOT_CONNECTED;
  int nEndId = NOT_CONNECTED;
  
  if(GetStartConnection()!=NULL)
    nStartId = GetStartConnection()->GetId();
  
  if(GetEndConnection()!=NULL) 
    nEndId = GetEndConnection()->GetId();
  
  
  //  If nStartId (/nEndId) is -1 it means that this vertex hasn't been 
  //  connected to the rest of the constraint network
#ifdef _DEBUG
  TRACE("      FTObjectId: %d", m_nId);
#endif //_DEBUG
  TRACE("      StartVxId:%d, EndVxId:%d, min:%0.3f, max:%0.3f, opt:%0.3f\n",
    nStartId,nEndId,GetFeasibleMin(),GetFeasibleMax(),optDuration);
  
    /*
    TRACE("         OdId= %2d    ", odId);
    TRACE(  m_bStartAuReceived ? "      (have AU)" : "      (no AU)  ");
    TRACE(  m_bEndAuReceived   ? "    (have AU)" : "    (no AU)");
    TRACE("\n");
    
      if(GetStartVertex() == 0)
      TRACE("  startConnection == 0");
      
        if(GetEndVertex() == 0)
        TRACE("  endConnection == 0");
        
          TRACE("\n");
  */
}


//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
void  FlexTimeObject::SetPriority(float fPriority)
{
  if(m_fPriority<fPriority)
    m_fPriority = fPriority;
}


//-----------------------------------------------------------------------------
//  Get the current system time in seconds (is time since 1 jan 1970)
//-----------------------------------------------------------------------------
double GetSysTime()
{
// SOLARIS PORTING
#ifdef WIN32
  struct _timeb timebuffer;
  _ftime(&timebuffer);
  return (double) timebuffer.time + (double) timebuffer.millitm / 1000.0;
#else
  struct timeval loc_time;
  ::gettimeofday( &loc_time, NULL );
  return (double) loc_time.tv_sec + (double) loc_time.tv_usec/1000000.0;
#endif
}


//-----------------------------------------------------------------------------
// evaluate whether the maxDelay has expired since this object should have started
//-----------------------------------------------------------------------------
bool FlexTimeObject::MaxDelayHasExpired()
{
  TemporalTransformImpFlex *ptt = GetTemporalNode()->GetTT();

  if(IsReadyToStart())
  {
    //  If the URL is active then it means that we don'n have to wait for the 
    //  stream.  
    return false;
  }
  
  double dStartTime   = GetStartTime();
  double dCurrentTime = GetSysTime();
  
  
  assert(ptt);
  double dMaxDelay = ptt->maxDelay;
  
  // if we have to wait for the stream to be ready, we check if the 
  return (dCurrentTime-dStartTime) >= dMaxDelay;
  
}

//-----------------------------------------------------------------------------
// Set the Start connection
//-----------------------------------------------------------------------------
void FlexTimeObject::SetStartConnection(FlexTimeConnection *pFTCon) 
{
  m_pStartFTConnection = pFTCon;
  if (pFTCon != NULL) m_nStartVertexId = pFTCon->GetId();
  else m_nStartVertexId = NOT_CONNECTED;
}

//-----------------------------------------------------------------------------
// Set the End connection
//-----------------------------------------------------------------------------
void FlexTimeObject::SetEndConnection(FlexTimeConnection *pFTCon) 
{
  m_pEndFTConnection = pFTCon;
  if (pFTCon != NULL) m_nEndVertexId = pFTCon->GetId();
  else m_nEndVertexId = NOT_CONNECTED;
  
}


bool FlexTimeObject::IsClass(const char *pszClassName) const
{
  return ::stricmp("FlexTimeObject",pszClassName)==0;
}
