//=============================================================================
//  This software module was originally developed by IBM Research (T.J. Watson)
//  in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) 
//  standard. This software module is an implementation of a part of one or 
//  more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 
//  Systems (ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 
//  Systems (ISO/IEC 14496-1) free license to this software module or 
//  modifications thereof for use in hardware or software products claiming 
//  conformance to the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use
//  this software module in hardware or software products are advised that its
//  use may infringe existing patents. The original developer of this software 
//  module and his/her company, the subsequent editors and their companies, and
//  ISO/IEC have no liability for use of this software module or modifications 
//  thereof in an implementation. Copyright is not released for non MPEG-4 
//  Systems (ISO/IEC 14496-1) conforming products. IBM retain full right to use
//  the code for their own purpose, assign or donate the code to a third party 
//  and to inhibit third parties from using the code for non MPEG-4 Systems 
//  (ISO/IEC 14496-1) conforming products. This copyright notice must be 
//  included in all copies or derivative works. Copyright  1999-2000.
//=============================================================================
//  IBM  TJ Watson Research, USA
//-----------------------------------------------------------------------------
//  TCNSIMPLE.CPP
//
//  Contact   :
//    Dr. Michelle Kim ( mykim@us.ibm.com )
//
//  Overview  :
//    Declaration of the Constraint Network
//
//  Comment   :
//    Port to C++ of Michelle Kim's TCNSimple
//
//=============================================================================  


#include "defs.h"
//#include "TCNSimple.h"
#include "TCNsimple.h" // SOLARIS PORTING - typo
#include "spring.h" 
#include <stdio.h>
#include <assert.h>
#define DEBUG true


//-----------------------------------------------------------------------------
//  constructor & destructor
//-----------------------------------------------------------------------------
TCNSimple::TCNSimple()
{
	m_pSpringArray = NULL;
}

TCNSimple::~TCNSimple()
{
	if(m_pSpringArray) 
		delete[] m_pSpringArray;
}


//-----------------------------------------------------------------------------
//  Reset with a new set of Arcs (maybe should be in the constructor)
//-----------------------------------------------------------------------------
bool TCNSimple::Reset(Spring** pSpringArray, int nArrayLength)
{
	TRACE("\nReset--------------------------------------------------------------"
        "------------------------------------------------------\n");
	// Check if the parameter is a valid buffer
	assert(pSpringArray);
	
	//	if(nArrayLength > ARCS_MAX_NUMBER) return false;
	if(m_pSpringArray) 
		delete[] m_pSpringArray;
	
	m_pSpringArray = NULL;
	
	//  Copy the pointers array in the local one
	m_nSpringCount = nArrayLength;
	
	//  Allocate a new block of memory
	m_pSpringArray = new Spring*[m_nSpringCount];
	
	//  Copy the pointers  
	::memcpy(m_pSpringArray,pSpringArray,sizeof(Spring*)*nArrayLength);
	
  // reset the vertex count to zero.
	m_nVertexCount = 0;

	ResetConstraintNetwork();
	return true;
}


//-----------------------------------------------------------------------------
//  Resets the constraint matrix
//-----------------------------------------------------------------------------
void TCNSimple::ResetConstraintNetwork()
{
	for(int i=0; i<VERTEX_MAX_NUMBER; i++)
		for(int j=0; j<VERTEX_MAX_NUMBER; j++)
			m_dConstraintNetwork[i][j] = MAX_BOUNDARY;
		
		for(i=0; i<VERTEX_MAX_NUMBER; i++)
			m_dConstraintNetwork[i][i]=0.0;
}


//-----------------------------------------------------------------------------
//  Evaluate the Constraint Network:
//  - Load in the ConstraintNetwork the constraints induced by the min and max 
//    values of the Springs
//  - Minimize the network
//  - Returns true if the Network is consistent.
//-----------------------------------------------------------------------------
bool TCNSimple::Evaluate()
{
  int startId, endId;
  
  //  Check if we really have a valid buffer
  assert(m_pSpringArray);
  
  //  The current implementation expect the arcs to have a startVertex and an 
  //  endVertex. The connectivity information could also go through another 
  //  medium
  
  //PAB: I could add a test so that we are sure that we redo the 
  //  evaluation only when something has changed.
  
  //  For each arc, try and merge the given constraints with the previous
  //  existing ones.
  
  for(int i=0; i<m_nSpringCount; i++) 
  {
    startId = m_pSpringArray[i]->GetStartVertexId();
    endId = m_pSpringArray[i]->GetEndVertexId();
    
    //  Only enter the constraint when there is both an end and a 
    //  start Vertices.
    if( (startId!=NOT_CONNECTED) && (endId!=NOT_CONNECTED) )
    {
      bool bOK = true;
      
      double dMax = m_pSpringArray[i]->GetFeasibleMax();
      double dMin = m_pSpringArray[i]->GetFeasibleMin();
      
      //  Enter the constraint induced by the minDuration
      bOK = AddConstraintToNetwork(endId,startId,dMax);
      if(!bOK)
        TRACE("TCN: pb entering constraint startId: %d, endId: %d, dMax: %f\n",
              startId, endId, dMax);
      
      //  Enter the constraint induced by the maxDuration
      bOK = AddConstraintToNetwork(startId,endId, -dMin);
      if(!bOK)
        TRACE("TCN: pb entering constraint startId: %d, endId: %d, dMin: %f\n",
              startId, endId, dMin);
    }
  }
  
  //  MinimizeNetwork() returns true if the network is consistant
  //  m_dConstraintNetwork now contains the minimal network.
  return MinimizeNetwork();
}


//-----------------------------------------------------------------------------
//  Minimize the ConstraintNetwork in order to generate the feasible Min and 
//  max returns false if the network is not consistent
//  (incompatible constraints)
//-----------------------------------------------------------------------------
bool TCNSimple::MinimizeNetwork() 
{
  //  Apply Floyd-Warshall's All-pairs-shortest-paths algorithm to the current
  //  constraint network.
  ComputeShortestPath();
  
  //  NegativeCycle() returns true if there is a Negative Cycle. In this case,
  //  the constraintNetwork is inconsistent.
  if( NegativeCycle() )
  {
    TRACE("inconsistency --> feasible durations not updated\n"	);
    return false;
  }
  return true;
}


//-----------------------------------------------------------------------------
//  Floyd-Warshall's All-pairs-shortest-paths algorithm
//  1. for i := 1 to n do d(ii) <- 0;
//  2. for i,j := 1 to n do d(ij) <- a(ij);
//    1. and 2. already done
//    ComputeShortestPath() does the last steps
//  3. for k := 1 to n do
//  4.    for i,j := 1 to n do
//  5.       d(ij) <- min {d(ij), d(ik) + d(kj)};
//-----------------------------------------------------------------------------
void TCNSimple::ComputeShortestPath()
{
  //  This implementation works directly on the original network for java
  //  performance issue, think to (from Jeff):  copy locally
  //  m_nVertexCount; avoid unnecessary dereferencing for the arrays...
  for(int k=0; k<m_nVertexCount; k++)
  {
    for(int j=0; j<m_nVertexCount ; j++)
    {
      for(int i=0; i<m_nVertexCount; i++)
      {
        double sum = m_dConstraintNetwork[i][k] + m_dConstraintNetwork[k][j];
        double dij = m_dConstraintNetwork[i][j];
        m_dConstraintNetwork[i][j] = (dij<sum) ? dij : sum ;
      }
    }
  }
}


//-----------------------------------------------------------------------------
//  Detects negative cycles by examining the sign of the diagonal elements
//  !! different from the Java implementation:
//  returns true if there is a Negative Cycle
//  (ie: if the network is inconsistent)
//-----------------------------------------------------------------------------
bool TCNSimple::NegativeCycle()
{
	for(int i=0; i<m_nVertexCount; i++)
	{
    // dii is an element of the diagonal of the constraint matrik
		double dii =  m_dConstraintNetwork[i][i];
		if(dii<0)
		{
			TRACE("Negative value in TCN %.3f at C[%d,%d]\n",dii,i,i);
			if(dii<-1)
				return true;
		}
	}
	return false;
}


//-----------------------------------------------------------------------------
//  Add a constraint of type:  TimeOfVertexA - TimeOfVertexB <= dBoundary
//  in the Network of constraints.
//-----------------------------------------------------------------------------
bool TCNSimple::AddConstraintToNetwork(int nVertexIdA,int nVertexIdB,double dBoundary)
{
	//  The constraint Network is a double dimension array with parameters being
	//  the Vertices. The VertexId has to be mapped to the array numbering
	int nIndexA = FindNetworkIndex(nVertexIdA);
	int nIndexB = FindNetworkIndex(nVertexIdB);

  //  If one of these nodes haven't a correct index, it means that it has no
  //  temporal constraint.
  if( nIndexA<0 || nIndexB<0 )
		return false;
	
	//  Check this constraint with the corresponding one:
	if (dBoundary < - m_dConstraintNetwork[nIndexB][nIndexA])
  {
		TRACE("TCNSimple: trying to enter a Constraint inconsistent with the rest"
          " of the TCN, Entering Maximal acceptable constraint instead\n");
		dBoundary = - m_dConstraintNetwork[nIndexB][nIndexA];
	}

	//  Merge this constraint with the previous one corresponding to this set
	//  of 2 vertices.
	//  If there was none, the original constraint is MAX_BOUNDARY
	if(dBoundary < m_dConstraintNetwork[nIndexA][nIndexB])
    m_dConstraintNetwork[nIndexA][nIndexB] = dBoundary;

	return true;
}


//-----------------------------------------------------------------------------
//  Find the Id of the first Connection (the one that should be started first)
//  Should only be called once the TCN has been minimized, and is consistent
//-----------------------------------------------------------------------------
int TCNSimple::FindFirstConnectionId() 
{
	int nStartVertexId,nEndVertexId;
	FindLongestPair(nStartVertexId, nEndVertexId);
	return nStartVertexId;
}


//-----------------------------------------------------------------------------
//  Finds a pair of nodes whose distance is the greatest
//  This is useful to identify the node with the earliest startTime
//-----------------------------------------------------------------------------
bool TCNSimple::FindLongestPair(int &iStartVertexId, int &iEndVertexId) 
{
	double dLongest = -1.0e6;
	iStartVertexId = NOT_CONNECTED;
	iEndVertexId   = NOT_CONNECTED;
	
	//  We could cache them in member variables....
	
	//  We will scan the whole constraint matrix to find the longuest path.
	//  so we'll have the 'first' and the 'last' object for this particular
	//  network
	for(int i=0;i<m_nVertexCount;i++) 
	{
		for(int j=0;j<m_nVertexCount;j++) 
		{
			if(m_dConstraintNetwork[i][j] > dLongest) 
			{
				dLongest = m_dConstraintNetwork[i][j];
				iEndVertexId = VertexIdMap[i];
				iStartVertexId = VertexIdMap[j];
			}
		}
	}
	
	#ifdef DEBUG
		TRACE("Longest: %f  StartVertex: %d  EndVertex: %d\n",
          dLongest,iStartVertexId,iEndVertexId);
	#endif

	return (iStartVertexId!=NOT_CONNECTED && iEndVertexId!=NOT_CONNECTED);
}


//-----------------------------------------------------------------------------
//  Would not be needed if I had a map template
//  Find our index in the ConstraintNetwork for this VertexId.
//  If it is not already mapped, add it
//-----------------------------------------------------------------------------
int TCNSimple::FindNetworkIndex(int nVertexId)
{
	for(int i=0;i<m_nVertexCount;i++) 
		if(VertexIdMap[i]==nVertexId) 
			return i;
	
	//  if this Id has not yet been mapped, map it into the next IdHolder
	// check if we have enough room to store this new vertex
	if(m_nVertexCount == VERTEX_MAX_NUMBER) 
	{
		TRACE("TCNSimple: critical error: too many vertices\n");
		return -1;
	}
	
	VertexIdMap[m_nVertexCount++] = nVertexId;
	return m_nVertexCount-1 ;
}


//-----------------------------------------------------------------------------
//  Update the feasible min and max from the Constraint Network to the Spring 
//  objects : (should be called only if the Network is consistent)
//-----------------------------------------------------------------------------
void TCNSimple::UpdateFeasibleLengths()
{
	//  return;
	int iStartVertex;
	int iEndVertex;
	Spring *pSpring;
	
	// check if we really have a valid buffer
	assert(m_pSpringArray);
	
	for(int i=0; i<m_nSpringCount; i++)
	{
		pSpring = m_pSpringArray[i];
		
		//  The indices (in our contraintNetwork) of the start/end vertices
		//  of this arc are:
		iStartVertex = FindNetworkIndex(pSpring->GetStartVertexId());
		iEndVertex   = FindNetworkIndex(pSpring->GetEndVertexId());
		
		//  The network has been minimized, we update the feasible min and max
		double dMin = - m_dConstraintNetwork[iStartVertex][iEndVertex];
		if(dMin > pSpring->GetFeasibleMin())
			pSpring->SetFeasibleMin(dMin);
		
		double dMax = m_dConstraintNetwork[iEndVertex][iStartVertex];
		if(dMax < pSpring->GetFeasibleMax())
			pSpring->SetFeasibleMax( m_dConstraintNetwork[iEndVertex][iStartVertex]);
	}
}


//-----------------------------------------------------------------------------
//  A helpful debug method used to dump the contents of the Constraint matrix
//  and also the contraint vector (link between the VertexId and the index of 
//  the constraint in the matrix.
//-----------------------------------------------------------------------------
void TCNSimple::DumpMatrix()
{
	TRACE("\n");
	
	for(int j=0;j<m_nVertexCount;j++) 
	{
		for(int i=0;i<m_nVertexCount;i++) 
		{
			TRACE("%3.2f\t", m_dConstraintNetwork[i][j]);
		}
		TRACE("\n");
	}
	TRACE("\n");
	
	for(int i=0;i<m_nVertexCount;i++)
		TRACE("%3d : %d\n",i,VertexIdMap[i]);
}