/******************************************************************************
 * This software module was originally developed by Agfa Monotype Corporation
 * in the course of development of the ISO/IEC 14996-18 standard for reference 
 * purposes and its performance may not have been optimized. This software 
 * module is an implementation of font compression technology as specified 
 * by the ISO/IEC 14996-18 standard.
 * 
 * ISO/IEC gives users a free copyright license to this software module or
 * modifications thereof for use in products claiming conformance to 
 * ISO/IEC 14996-18 International Standard. ISO/IEC gives users the same free
 * license to this software module or modifications thereof for research 
 * purposes and further ISO/IEC standardization.
 * 
 * Those intending to use this software module in products are advised that
 * its use may infringe existing patents and that this license does not grant 
 * rights under any patents. ISO/IEC have no liability for use of this 
 * software module or modifications thereof.
 * 
 * No license is granted and copyright is not released for products that do 
 * not conform to ISO/IEC 14996-18 International Standard.
 * 
 * Agfa Monotype Corporation retains full right to modify and use the code for
 * its own purpose, assign or donate the code to a third party and to inhibit 
 * third parties from using the code for products that do not conform to
 * audiovisual and image-coding related ITU recommendations and 
 * ISO/IEC International Standards. All rights not expressly granted to the 
 * software module or modifications thereof are reserved by 
 * Agfa Monotype Corporation.
 * 
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2003. All rights reserved.
 ******************************************************************************/

/*
 * File:						bitio.c
 * Revision history
 * $Log: BITIO.C,v $
 * Revision 1.2  2004/03/01 21:48:03  vladlev
 * minor optimization for performance
 *
 * Revision 1.1  2004/02/10 22:01:08  vladlev
 * Initial version
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>

#include "CONFIG.H"
#include "BITIO.H"
#include "ERRCODES.H"

const static BYTE mask[] = {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};

/****************************************************************/
void MTX_BITIO_output_bit( BITIO *t, int bit )
{
	int index = t->index;
	BYTE *array;

	/* grow the array as needed */
	if (index + 1>= t->max)
	{
		t->max += 1024;
		t->array = MTX_mem_realloc(t->mem, t->array, (t->max >> 3));
	}
	
	/* write the bit */
	array = t->array;
	if (bit)
		array[index >> 3] |= mask[index & 7];
	else
		array[index >> 3] &= ~mask[index & 7];


	/* maintain high water mark */
	if (index >= t->high)
		t->high = index;

}
/****************************************************************/
void MTX_BITIO_WriteValue( BITIO *t, ULONG value, LONG numberOfBits )
{
	BYTE *array;
	LONG index = t->index;
	ULONG m;

	/* grow the array as needed */
	if (index + numberOfBits >= t->max)
	{
		t->max += 1024;
		t->array = MTX_mem_realloc(t->mem, t->array, (t->max >> 3));
	}

	/* write the bits */
	array = t->array;
	t->index += numberOfBits;
	m = (1UL << (numberOfBits-1));
	while (m)
	{
		if (m & value)
			array[index >> 3] |= mask[index & 7];
		else
			array[index >> 3] &= ~mask[index & 7];
		index++;
		m >>= 1;
	}

	/* maintain high water mark */
	if (t->index >= t->high)
		t->high = t->index;
}

/* the file header is byte aligned, hence the following special cases */
/****************************************************************/
void MTX_BITIO_Write8( BITIO *t, ULONG value )
{
	BYTE *array;
	LONG index = t->index;

	/* grow the array as needed */
	if (index + 8 >= t->max)
	{
		t->max += 1024;
		t->array = MTX_mem_realloc(t->mem, t->array, (t->max >> 3));
	}

	/* write the bits */
	array = t->array;
	index = t->index;
	if (0 == (index & 7))
	{
		array[index >> 3] = value;
		t->index += 8;
	}
	else
		MTX_BITIO_WriteValue(t,value,8);
	
	/* maintain high water mark */
	if (t->index >= t->high)
		t->high = t->index;
}
/****************************************************************/
void MTX_BITIO_Write32( BITIO *t, ULONG value )
{
	BYTE *array;
	LONG index = t->index;

	/* grow the array as needed */
	if (index + 32 >= t->max)
	{
		t->max += 1024;
		t->array = MTX_mem_realloc(t->mem, t->array, (t->max >> 3));
	}

	/* write the bits */
	array = t->array;
	if (0 == (index & 7))
	{
		index >>= 3;
		array[index++] = value >> 24;
		array[index++] = 0xFF & (value >> 16);
		array[index++] = 0xFF & (value >> 8);
		array[index] = 0xFF & value;
		t->index+=32;
	}
	else
		MTX_BITIO_WriteValue(t,value,32);
	
	/* maintain high water mark */
	if (t->index > t->high)
		t->high = t->index;
}

/****************************************************************/

static ULONG mask32[] = {
	0x00000000,	
	0x00000001,0x00000003, 0x00000007, 0x0000000F, 
	0x0000001F,0x0000003F, 0x0000007F, 0x000000FF, 
	0x000001FF,0x000003FF, 0x000007FF, 0x00000FFF, 
	0x00001FFF,0x00003FFF, 0x00007FFF, 0x0000FFFF, 
	0x0001FFFF,0x0003FFFF, 0x0007FFFF, 0x000FFFFF, 
	0x001FFFFF,0x003FFFFF, 0x007FFFFF, 0x00FFFFFF, 
	0x01FFFFFF,0x03FFFFFF, 0x07FFFFFF, 0x0FFFFFFF, 
	0x1FFFFFFF,0x3FFFFFFF, 0x7FFFFFFF, 0xFFFFFFFF
	};
		
/****************************************************************/

/* get the next <numberOfBits> bits from the bitarray <t> */
ULONG MTX_BITIO_ReadValue( BITIO *t, LONG numberOfBits )
{
	ULONG first,last,r,mask;
	BYTE *p;
	int n,shift;

	first = t->index >> 3;
	r = t->index + numberOfBits - 1;
	last = r >> 3;
	n = 1 + last - first;

	shift = 7 - (r & 7);
	mask = mask32[numberOfBits];
	p = t->array + first;

	switch(n)
	{
		case 1:
			r = *p;
			r >>= shift;
			r &= mask;
			break;
		case 2:
			r = *p++;
			r = (r << 8) | *p;
			r >>= shift;
			r &= mask;
			break;
		case 3:
			r = *p++;
			r = (r << 8) | *p++;
			r = (r << 8) | *p;
			r >>= shift;
			r &= mask;
			break;
		case 4:
			r = *p++;
			r = (r << 8) | *p++;
			r = (r << 8) | *p++;
			r = (r << 8) | *p;
			r >>= shift;
			r &= mask;
			break;
		default: 
			/* We can never span more than 5 bytes with <= 32 bits, but
			* we can span 5 bytes with as few as 26 requested bits, eg:
			* [.......x][xxxxxxxx][xxxxxxxx][xxxxxxxx][x.......] */
#if SIZEOF_LONG >= 5
			/* which is no problem on machines with longs of 40+ bits */
			{
			unsigned long temp;

			temp = *p++;
			temp = (temp << 8) | *p++;
			temp = (temp << 8) | *p++;
			temp = (temp << 8) | *p++;
			temp = (temp << 8) | *p;
			temp >>= shift;
			r = (ULONG)temp & mask;
			}
#else 
			/* which is rather ugly on a 32 bit machine ... fortunately 
			* spans of 5 are rather uncommon ... only 28 of 256 possible
			* fetches span 5 bytes.  Moreover, they all occur for tokens
			* with very low frequency - a 26 bit code spanning 5 bytes
			* will occur once in every 2**29 fetches; a 32 bit code which
			* spans 5 bytes will occur once in every 7*2**35 fetches.
			*
			* By the way: a 32 bit code would signify a token which had a
			* frequency of 1 in 2**32 in the data ,,, but which occured
			* at least twice (in order to be chosen a token) and since the
			* minimum token size is 2 bytes, this will occur only for files
			* which exceed 8 gigabytes uncompressed.
			*
			* Like I said ... rather uncommon.
			*/
			first = numberOfBits/2;
			last = numberOfBits - first;
			r = MTX_BITIO_ReadValue(t,first);
			r = (r << last) | MTX_BITIO_ReadValue(t,last);
#endif /* SIZEOF_LONG >= 5  */
			
			break;
	}

	t->index += numberOfBits;
	return r;
}

/* the file header is byte aligned, hence the special cases */
/****************************************************************/
ULONG MTX_BITIO_Read8( BITIO *t )
{
	ULONG r;
	int index = t->index;
	BYTE *array = t->array;

	if (0 == (index & 7))
	{
		r = array[index >> 3];
		t->index += 8;
	}
	else
		r = MTX_BITIO_ReadValue(t,8);
	return r;
}

/****************************************************************/
ULONG MTX_BITIO_Read32( BITIO *t )
{
	ULONG r;
	int index = t->index;
	BYTE *array = t->array;

	if (0 == (index & 7))
	{
		index >>= 3;
		r = array[index++];
		r = (r << 8) | array[index++];
		r = (r << 8) | array[index++];
		r = (r << 8) | array[index];
		t->index += 32;
	}
	else
		r = MTX_BITIO_ReadValue(t,32);
	return r;
}

/****************************************************************/
/* turned these two functions into macros -- see bitio.h */
#if 0
ULONG MTX_BITIO_ftell( register BITIO *t )
	{
	return t->index;
	}

/****************************************************************/
VOID MTX_BITIO_fseek( register BITIO *t, ULONG pos )
	{
	t->index = pos;
	}
#endif /* 0 */

/****************************************************************/
void MTX_BITIO_flush_bits( BITIO *t )
{
	/* nada */
}

/****************************************************************/
BYTE *MTX_BITIO_GetMemoryPointer( BITIO *t )
{
	return t->array;
}

/****************************************************************/
LONG MTX_BITIO_GetBytesOut( BITIO *t )
{
	return (7 + t->high) >> 3;
}


/****************************************************************/
BITIO *MTX_BITIO_Create( MTX_MemHandler *mem, void *memPtr, LONG memSize, const CHAR param )
{
	BITIO *r;

	r = MTX_mem_malloc(mem, sizeof(BITIO));
	r->array = memPtr;
	r->index = 0;
	r->max = memSize<<3;
	r->mem = mem;
	r->high = 0;
	return r;
}

/****************************************************************/
void MTX_BITIO_Destroy(BITIO *t)
{
	MTX_mem_free(t->mem,t);
}
