/******************************************************************************
 * This software module was originally developed by Agfa Monotype Corporation
 * in the course of development of the ISO/IEC 14996-18 standard for reference 
 * purposes and its performance may not have been optimized. This software 
 * module is an implementation of font compression technology as specified 
 * by the ISO/IEC 14996-18 standard.
 * 
 * ISO/IEC gives users a free copyright license to this software module or
 * modifications thereof for use in products claiming conformance to 
 * ISO/IEC 14996-18 International Standard. ISO/IEC gives users the same free
 * license to this software module or modifications thereof for research 
 * purposes and further ISO/IEC standardization.
 * 
 * Those intending to use this software module in products are advised that
 * its use may infringe existing patents and that this license does not grant 
 * rights under any patents. ISO/IEC have no liability for use of this 
 * software module or modifications thereof.
 * 
 * No license is granted and copyright is not released for products that do 
 * not conform to ISO/IEC 14996-18 International Standard.
 * 
 * Agfa Monotype Corporation retains full right to modify and use the code for
 * its own purpose, assign or donate the code to a third party and to inhibit 
 * third parties from using the code for products that do not conform to
 * audiovisual and image-coding related ITU recommendations and 
 * ISO/IEC International Standards. All rights not expressly granted to the 
 * software module or modifications thereof are reserved by 
 * Agfa Monotype Corporation.
 * 
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2003. All rights reserved.
 ******************************************************************************/

/*
 * File:						MEMORY.C
 * Revision history
 * $Log: MEMORY.C,v $
 * Revision 1.1  2004/02/10 22:01:09  vladlev
 * Initial version
 *
 */
 
#include <stdio.h>
#include <assert.h>
#include "config.h"
#include "MEMORY.H"
#include "ERRCODES.H"


const LONG overWriteExtraSpace		= sizeof(LONG) + sizeof(LONG) + sizeof(CHAR); /* 4-byte Stamp, length, data, 1-byte Stamp */
const LONG overWriteOffsetToData	= sizeof(LONG) + sizeof(LONG); /* 4-byte Stamp, length */


/* Our malloc */
void *MTX_mem_malloc(MTX_MemHandler *t, size_t size)
{
	register LONG i, limit;
	void *babyObject;
	register void **pointers;
	
	if ( size == 0 ) return NULL;
	
	t->mem_numNewCalls++;
	babyObject = t->malloc( size );
	if ( babyObject == NULL )
	{ 
		t->errorCode = ERR_no_more_memory;
		return NULL;
	}
	if ( t->mem_numPointers >= t->mem_maxPointers ) {
		LONG oldMaxPointers = t->mem_maxPointers;
		/* We need to allocate more space for pointers */
		t->mem_maxPointers = t->mem_maxPointers + t->mem_maxPointers/4 + 128;
		if ( t->mem_pointers != NULL ) {
			t->mem_pointers = (void **)t->realloc( t->mem_pointers, t->mem_maxPointers * sizeof( void *) );
		} else { 
			t->mem_pointers = (void **)t->malloc( t->mem_maxPointers * sizeof( void *) );
		}
		if ( t->mem_pointers == NULL )
		{ 
			t->errorCode = ERR_no_more_memory;
			return NULL;
		}
		/* Zero out the new entries; */
		for ( i = oldMaxPointers; i < t->mem_maxPointers; i++ ) {
			t->mem_pointers[i] = NULL;
		}
	}
	/* OK, now we have a space to store the pointers, so go find it ! */
	limit = t->mem_maxPointers;
	pointers = t->mem_pointers;
	for ( i = 0; i < limit; i++ ) {
		if ( pointers[i] == NULL ) {
			pointers[i] = babyObject;
			t->mem_numPointers++;
			return babyObject; /******/
		}
	}

	t->errorCode = ERR_new_logical;
	return NULL; /*****/
}

/* Our realloc */
void *MTX_mem_realloc(MTX_MemHandler *t, void *p, size_t size)
{
	register LONG i, limit;
	register void **pointers = t->mem_pointers;
	void *new_p;

	if ( p == NULL ) {
		return MTX_mem_malloc( t, size ); /******/
	}
	if ( size == 0 ) {
		MTX_mem_free( t, p );
		return NULL;
	}
	
	/* See if we can find it */
	limit = t->mem_maxPointers;
	for ( i = 0; i < limit; i++ ) {
		if ( pointers[i] == p ) {
			new_p = t->realloc( p, size );
		    if ( new_p == NULL ) 
		    {
		    	t->errorCode = ERR_realloc_failed;
				return NULL;
			}
			pointers[i] = new_p;
			return new_p; /******/
		}
	}
	t->errorCode = ERR_realloc_bad_pointer;
	return NULL; /******/
}


/* Our free */
void MTX_mem_free(MTX_MemHandler *t, void *deadObject)
{
	register LONG i, limit;
	register void **pointers = t->mem_pointers;
	if ( deadObject == NULL ) return; /*****/ /* This is OK by definition */
	/* See if we can find it */
	limit = t->mem_maxPointers;
	for ( i = 0; i < limit; i++ ) {
		if ( pointers[i] == deadObject ) {
			t->free( deadObject );
			pointers[i] = NULL;
			t->mem_numPointers--;
			return; /******/
		}
	}
	/* this is a "shouldn't ever happen" error - calling exit at this point
		would also be an appropriate action */
	t->errorCode = ERR_delete_bad_pointer;
	/*** printf ("MTX_mem_free error = %d\n", ERR_delete_bad_pointer); ***/
	/*** exit(-6); ***/
}



/* Call mem_FreeAllMemory insted on an abnormal (exception) exit */
void MTX_mem_FreeAllMemory( MTX_MemHandler *t )
{
	register LONG i;

	if ( t->mem_pointers != NULL ) {
		for ( i = 0; i < t->mem_maxPointers; i++ ) {
			if ( t->mem_pointers[i] != NULL ) {
				t->free( t->mem_pointers[i] );
				t->mem_pointers[i] = NULL;
				t->mem_numPointers--; 
			}
		}
		t->free( t->mem_pointers );  t->mem_pointers = NULL;
	}
}

/* Call mem_CloseMemory on normal exit */
void MTX_mem_CloseMemory( MTX_MemHandler *t )
{
	if ( t->mem_numPointers != 0 ) 
	{
		/* this is a "shouldn't ever happen" error - calling exit at this point
			might be a bit of overkill, however, since we're shutting down */
		t->errorCode = ERR_mem_dangling_pointers;
		/*** printf ("MTX_mem_CloseMemory error = %d\n", ERR_mem_dangling_pointers); ***/
		/*** exit(-7); ***/
	}
	if ( t->mem_pointers != NULL  ) t->free( t->mem_pointers ); t->mem_pointers = NULL;
}


/* Constructor */
MTX_MemHandler *MTX_mem_Create(MTX_MALLOCPTR mptr, MTX_REALLOCPTR rptr, MTX_FREEPTR fptr)
{
	MTX_MemHandler *t	= (MTX_MemHandler *)mptr( sizeof( MTX_MemHandler ) );
	if ( t == NULL ) return NULL; /*****/
	
	t->mem_pointers		= NULL;
	t->mem_maxPointers	= 0;
	t->mem_numPointers 	= 0; /* Number of non-zero pointers */
	t->mem_numNewCalls	= 0;

	assert( mptr != NULL && rptr != NULL && fptr != NULL );
	t->malloc			= mptr;
	t->realloc			= rptr;
	t->free				= fptr;

	t->errorCode = 0;	/* no error, initially */
	return t; /******/
}

/* Destructor */
void MTX_mem_Destroy(MTX_MemHandler *t)
{
	assert ( t != NULL );
	t->free( t );
}

