/******************************************************************************
 * This software module was originally developed by Agfa Monotype Corporation
 * in the course of development of the ISO/IEC 14996-18 standard for reference 
 * purposes and its performance may not have been optimized. This software 
 * module is an implementation of font compression technology as specified 
 * by the ISO/IEC 14996-18 standard.
 * 
 * ISO/IEC gives users a free copyright license to this software module or
 * modifications thereof for use in products claiming conformance to 
 * ISO/IEC 14996-18 International Standard. ISO/IEC gives users the same free
 * license to this software module or modifications thereof for research 
 * purposes and further ISO/IEC standardization.
 * 
 * Those intending to use this software module in products are advised that
 * its use may infringe existing patents and that this license does not grant 
 * rights under any patents. ISO/IEC have no liability for use of this 
 * software module or modifications thereof.
 * 
 * No license is granted and copyright is not released for products that do 
 * not conform to ISO/IEC 14996-18 International Standard.
 * 
 * Agfa Monotype Corporation retains full right to modify and use the code for
 * its own purpose, assign or donate the code to a third party and to inhibit 
 * third parties from using the code for products that do not conform to
 * audiovisual and image-coding related ITU recommendations and 
 * ISO/IEC International Standards. All rights not expressly granted to the 
 * software module or modifications thereof are reserved by 
 * Agfa Monotype Corporation.
 * 
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2003. All rights reserved.
 ******************************************************************************/

/*
 * File:						RA_COMP.C
 * Revision history
 * $Log: Ra_comp.c,v $
 * Revision 1.1  2004/02/10 22:01:10  vladlev
 * Initial version
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "ERRCODES.H"
#include "CONFIG.h"


#include "MEMORY.H"
#include "BITIO.H"
#include "STRCTR.H"
#include "RA_COMP.H"
#include "RADECOMP.H"
#include "RA_TEST.H"
#include "RA_UTIL.H"

#include "canonical.h"


/*---------------------*/
const int minLength = 2;

/* 
 * Computes the offset into the string store for string p of length len.
 * Returns -1 if the string does not appear in the string store.
 */

/* Boyer Moore string search never uses more then M+N compares, and uses 
	about N/M if N is large and M is small ... like it usually is here.
	from profiling, seems to be about 4-5 times faster than brute force
*/
static LONG GetStringOffset( MTX_RA_COMP *t, BYTE *p, LONG m )
	{
	static skip[256];
	int i,j,s,n = t->stringStoreLength;
	BYTE *a = t->stringStore;

	/* no way there's a match if substring is longer than string */
	if (m > n)
		return -1;
	
	/* initialize <skip> */
	for (i=0; i<256; i++)
		skip[i] = m;
	for (i=0,j=m-1; i<m; i++,j--)
		skip[p[i]] = j;

	/* match the pattern */
	for (i=m-1, j=m-1; j>=0; i--,j--)
		while (a[i] != p[j])
			{
			s = skip[a[i]];
			i += (m-j > s) ? m-j : s;
			if (i >= n) 
				return -1;	/* not found */
			j = m-1;
			}
	return i+1;
	}

/*
 * Stores the string p in the String Store,
 * also computes and returns the approximate
 * byte cost for storing the string (from
 * the string store and node areas).
 * If 'store' is false then the function just returns the cost
 * without actually storing the string.
 */
static LONG StoreString( MTX_RA_COMP *t, BYTE *p, LONG len, ULONG count, int store, LONG *offsetPtr )
{
	LONG oldNodeLength = t->nodeLength;
	LONG oldstringStoreLength = t->stringStoreLength;
	LONG cost;
	LONG offset = GetStringOffset(t,  p, len );
	
	*offsetPtr = offset;
	
	t->nodeLength++; /* store flags */
	if ( count >= 64 ) {
		t->nodeLength += MTX_RA_ULongStorageCost( count/64 ); /* store count/64 */
	}
	if ( offset >= 0 ) {
		t->nodeLength++; /* store offset % 256 */
		t->nodeLength += MTX_RA_ULongStorageCost( offset/256 ); /* store offset/256 */
	} else {
		/* We have to store the string. */
		LONG i;
		
		if ( t->stringStoreLength + len > t->maxStringStoreLength ) {
			/* Grow the length exponentially. */
			t->maxStringStoreLength  = t->stringStoreLength + len;
			t->maxStringStoreLength += t->maxStringStoreLength/2 + 256;
			printf("*****Increased max string store length to %ld\n", t->maxStringStoreLength);
			t->stringStore = (BYTE *)MTX_mem_realloc(t->mem, t->stringStore, t->maxStringStoreLength);
		}
		for ( i = 0; i < len; i++ ) {
			t->stringStore[ t->stringStoreLength++] = p[i];
		}
		assert( t->stringStoreLength <= t->maxStringStoreLength );
	}
	cost = (t->stringStoreLength - oldstringStoreLength) + (t->nodeLength - oldNodeLength);
	if ( !store ) {
		/* Undo the size increase. */
		t->stringStoreLength	= oldstringStoreLength;
		t->nodeLength			= oldNodeLength;
	}
	return cost; /*****/
}



/* Allocates various memory areas. */
static int CompAllocate( MTX_RA_COMP *t )
{
	t->MaxSymbolCount = t->symbolCount * 2 + 10;
	t->dictSymbol = (CHAR **)MTX_mem_malloc( t->mem, t->MaxSymbolCount * sizeof( CHAR * ) );
	if (!t->dictSymbol && t->mem->errorCode)
		return (t->mem->errorCode);
	t->dictSymbolLength = (LONG *)MTX_mem_malloc( t->mem, t->MaxSymbolCount * sizeof( LONG ) );
	if (!t->dictSymbolLength && t->mem->errorCode)
		return (t->mem->errorCode);
	t->offsets = (LONG *)MTX_mem_malloc( t->mem, t->MaxSymbolCount * sizeof( LONG ) );
	if (!t->offsets && t->mem->errorCode)
		return (t->mem->errorCode);
	
	t->count	= (LONG *)MTX_mem_malloc( t->mem, 2 * t->MaxSymbolCount * sizeof( LONG ) );
	if (!t->count && t->mem->errorCode)
		return (t->mem->errorCode);
	t->parent	= (LONG *)MTX_mem_malloc( t->mem, 2 * t->MaxSymbolCount * sizeof( LONG ) );
	if (!t->parent && t->mem->errorCode)
		return (t->mem->errorCode);
	t->child_0	= (LONG *)MTX_mem_malloc( t->mem, 2 * t->MaxSymbolCount * sizeof( LONG ) );
	if (!t->child_0 && t->mem->errorCode)
		return (t->mem->errorCode);
	t->child_1	= (LONG *)MTX_mem_malloc( t->mem, 2 * t->MaxSymbolCount * sizeof( LONG  ) );
	if (!t->child_1 && t->mem->errorCode)
		return (t->mem->errorCode);

	return 0;	/* success */
}



/* Reallocates various memory areas. */
static void CompReAllocate( MTX_RA_COMP *t )
{
	if ( t->symbolCount >= t->MaxSymbolCount ) {
		LONG longSize, charPSize;
		assert( t->symbolCount == t->MaxSymbolCount );
		t->MaxSymbolCount += t->MaxSymbolCount/4 + 1;
		longSize  = t->MaxSymbolCount * sizeof( LONG );
		charPSize = t->MaxSymbolCount * sizeof( CHAR * );
		/* printf("** Increasing MaxSymbolCount to %ld ***\n",   t->MaxSymbolCount ); */
		t->dictSymbolLength	= (LONG *)MTX_mem_realloc( t->mem,  t->dictSymbolLength	, longSize );
		t->offsets			= (LONG *)MTX_mem_realloc( t->mem,  t->offsets			, longSize );
		t->dictSymbol		= (CHAR **)MTX_mem_realloc( t->mem,  t->dictSymbol		, charPSize );

		longSize   += longSize;
		t->count	= (LONG *)MTX_mem_realloc( t->mem,  t->count,	longSize );
		t->parent	= (LONG *)MTX_mem_realloc( t->mem,  t->parent,	longSize );
		t->child_0	= (LONG *)MTX_mem_realloc( t->mem,  t->child_0, longSize );
		t->child_1	= (LONG *)MTX_mem_realloc( t->mem,  t->child_1, longSize );
	}
}


/* Deallocates various memory areas. */
static void CompDeAllocate( MTX_RA_COMP *t )
{
	MTX_mem_free( t->mem, t->dictSymbol );
	MTX_mem_free( t->mem, t->dictSymbolLength );
	MTX_mem_free( t->mem, t->offsets );

	MTX_mem_free( t->mem, t->count );
	MTX_mem_free( t->mem, t->parent );
	MTX_mem_free( t->mem, t->child_0 );
	MTX_mem_free( t->mem, t->child_1 );
}


/* Returns the best symbolIndex for position 'pos' */
LONG FindBestMatch( MTX_RA_COMP *t, LONG pos )
{
	register LONG i, j, len;
	register CHAR *p1, *p2;
	
	len = t->taken[pos];
	if ( len == 0 ) len = 1;
	p1 = &t->dataIn[ pos ];
	for ( i = 0; i < t->symbolCount; i++ ) {
		if ( t->dictSymbolLength[i] != len ) continue;/***** AVOID A GREEDY PATH *****/
		p2 = t->dictSymbol[i];
		for ( j = 0; p1[j] == p2[j] && j < len; ) {
			if ( ++j == len ) return i; /***** We found a match! *****/
		}
	}

	/* We should never get here. catastrophic failure we EXIT */
	t->mem->errorCode = ERR_find_best_match_logical;
	printf("FindBestMatch error = %d\n", ERR_find_best_match_logical);
	exit(-8);
}

/* Recompute all the symbol counts */ 
static void SetAllCounts( MTX_RA_COMP *t )
{
	register LONG i, j, match;
	register LONG totalCount = 0;
	register LONG *count = t->count;
	
	/* Set all symbol counts to zero. */
	j = t->symbolCount;
	for ( i = 0; i < j; i++ ) {
		count[i] = 0;
	}
	/* Recompute all the symbol counts */
	j = t->lengthIn;
	for ( i = 0; i < j; ) {
		match = FindBestMatch(t, i);
		assert( match >= 0 && match < t->symbolCount );
		count[match]++;
		totalCount++;
		i += t->dictSymbolLength[match];
	}
	t->totalCount = totalCount;
	assert( i == j );
}

/*****************************************************************/

/* a strncmp which isn't confused by 0x00 bytes. */
int bstrncmp(BYTE *a, BYTE *b, int n)
{
	int i;
 
	for (i=0; i<n; i++)
	{
		if (a[i] < b[i])
			return -1;
		if (a[i] > b[i])
			return 1;
	}
	return 0;
}

/* form and sort a grade vector to speed up counting matches
 (matching strings will be contiguous in the grade vector)	*/
void create_grade_vector(MTX_RA_COMP *t)
{
	LONG n,gap,i,j,temp;
	BYTE *data;
	LONG *v;

	/* printf("create_grade_vector\n"); /* */

	/* allocate goodies */
	n = t->lengthIn;
	t->grade_vector = (LONG *) malloc(4*n);
	t->matches = (LONG *) malloc(4*n);
	t->match_length = (BYTE *) malloc(n);
	if (t->grade_vector==0 || t->matches==0 || t->match_length==0)
	{
		printf("** unable to allocate grade_vector");
		exit(-9);
	}

	/* make the grade vector (for search_grade_vector) */
	data = t->dataIn;
	v = t->grade_vector;
	for (i=0; i<n; i++)
		v[i] = i;

	for (gap=n/2; gap>0; gap/=2)
		for (i=gap; i<n; i++)
			for (j=i-gap; j>=0 && bstrncmp(data+v[j],data+v[j+gap],MAX_SYMBOL_LENGTH)>0; j-=gap)
			{
				temp = v[j];
				v[j]=v[j+gap];
				v[j+gap]=temp;
			}
	
	/* record match lengths (for walk_grade_vector)
		(recall <data> is oversize so don't worry about indexing off the end */
	for (i=0; i<t->lengthIn-1; i++)
	{
		for (j=0; j<MAX_SYMBOL_LENGTH; j++)
		{
			if (data[v[i]+j] != data[v[i+1]+j])
				break;
		}
		t->match_length[i] = (CHAR)j;
	}

	/* repair the last few "off the end" match lengths */
	for (i=0; i<MAX_SYMBOL_LENGTH; i++)
		if (t->match_length[t->lengthIn-(i+1)] > i)
			t->match_length[t->lengthIn-(i+1)] = (CHAR)i;

#ifdef DEBUG
	{
		int i;

		for (i=t->lengthIn-24; i<t->lengthIn; i++)
		{
			printf("%3d: %3d",i,t->match_length[i]);
			show_string(t->dataIn + t->grade_vector[i], 32);
		}
	}
#endif
}


/* binary search the sorted list for a string */
static LONG search_grade_vector( MTX_RA_COMP *t, BYTE *p, int n)
{
	LONG hi,lo,mid;
	int test;

	/* printf("search_grade_vector: ");	*/
	lo = 0;
	hi = t->lengthIn;
	while (lo <= hi)
	{
		mid = (lo + hi)/2;
		test = bstrncmp(p,t->dataIn+t->grade_vector[mid],n);
		if (test < 0)
			hi = mid-1;
		else if (test > 0)
			lo = mid+1;
		else
			return mid;
	}
	/* this shouldn't ever happen... */
	return -1;
}

/* find the first/last occurance of a string given AN occurance	*/
static LONG walk_grade_vector( MTX_RA_COMP *t, int n, int dir, LONG start)
{
	/* printf("walk_grade_vector: "); */
	if (dir>0)
	{
		while (start+1 < t->lengthIn && t->match_length[start]>=n)
			start++;
		return start;
	}
	else
	{
		while (start>0 && t->match_length[start-1]>=n)
			start--;
		return start;
	}
}

/* shell sort of a vector of LONG integers - straight from K&R */
static void shell(LONG *v, LONG n)
{
	LONG gap,i,j,temp;

	for (gap=n/2; gap>0; gap/=2)
		for (i=gap; i<n; i++)
			for (j=i-gap; j>=0 && v[j]>v[j+gap]; j-=gap)
			{
				temp = v[j];
				v[j]=v[j+gap];
				v[j+gap]=temp;
			}
}

/* runs in O(log(N)) time rather than O(N^^2) of the original */
static LONG MatchCount( MTX_RA_COMP *t, register CHAR *p, register LONG n, register CHAR mark, FILE *verbose )
{
	/* p == data we're trying to match n == length of p
	 * if (mark) set taken for the matching locations */
	LONG i,j,first,last,count;

	/* first time ?	*/
	if (t->grade_vector == 0)
		create_grade_vector(t);

	/* binary search to find an occurance of <p> */
	i = search_grade_vector(t,p,n);

	/* walk forward and back to find first & last occurance	*/
	first = walk_grade_vector(t,n,-1,i);
	last = walk_grade_vector(t,n,1,i);

	/* run through the occurances tossing the taken ones */
	count = 0;
	for (i=first; i<=last; i++)
	{
		BYTE *q;
		LONG m,min;
	
		/* is the i-th occurance taken?	*/
		q = t->taken + t->grade_vector[i];
		m = t->lengthIn - t->grade_vector[i];
		min = (m<n) ? m : n;
		for (j=0; j<min && q[j]==0; j++)
			{};
	
		/* not taken ?  add to list of possibles */
		if (j == n)
			t->matches[count++] = t->grade_vector[i];
	}
	if (count==0)
		return 0;

	/* sort the possibles and remove overlaps  */
	shell(t->matches,count);
	i=j=1;
	while (i<count)
	{
		if (t->matches[i] >= t->matches[j-1]+n)
			t->matches[j++] = t->matches[i++];
		else
			i++;
	}
	count = j;
	if (count==0)
		return 0;

	/* mark occurances  ? */
	if (mark)
	{
		for (i=0; i<count; i++)
		{
			memset(t->taken+t->matches[i],mark,n);
		}
	}

	return count;
}

/****************************************************************************/
/* 
 * This function resets the entire sequence of matches giving  
 * preference to longer matches.
 * It starts by matching the longest strings, and then works down.
 */
static void ResetControlSequence( MTX_RA_COMP *t, LONG maxSymLength )
{
	LONG pos, length, i;
	
	/* Initialize. */
	for ( pos = 0; pos < t->lengthIn; pos++ ) {
		t->taken[pos] = 0;
	}
	/* First look for the longest matches, and then work down. */
	for ( length = maxSymLength; length >= 2; length-- ) {
		for ( i = 0; i < t->symbolCount; i++ ) {
			if ( t->dictSymbolLength[i] == length ) {
				MatchCount( t, t->dictSymbol[i], length, (CHAR)length, false ); /* Mark all occurances */
			}
		}
	}
}


/* The natural log of 2.0 */
#define log2 0.6931471805599452862

/* Returns the bitcost of storing 'c' computed as being identical to the entropy in the information */
static double CharCost( MTX_RA_COMP *t, BYTE c )
{
	double pINV, cost;
	
	pINV = (double)t->totalCount / (double)t->count[c];
	cost = log(pINV) / log2; /* The base-2-log of pINV. */
	return cost; /*****/
}

/*
 * Returns the bitcost of storing a symbol with 'eventCount' occurances
 * computed as being identical to the entropy in the information
 */
static double SymbolCost( MTX_RA_COMP *t, LONG eventCount )
{
	double pINV, cost;
	
	pINV = (double)t->totalCount / (double)eventCount;
	cost = log(pINV) / log2; /* The base-2-log of pINV. */
	return cost; /*****/
}

/*
 * Returns the approximated gain in bytes from adding the string
 * 'string' with length 'len' and event count 'eventCount' as
 * a symbol.
 */
static LONG Gain( MTX_RA_COMP *t, BYTE *string, LONG len, LONG eventCount )
{
	LONG offset;
	LONG dictCost = StoreString( t, string, len, eventCount, false, &offset );
	LONG gain;

	if ( len > 2 ) {
		LONG newCost, oldCost;
		oldCost = eventCount * len; /* Approximation */
		newCost = eventCount * 2 + dictCost;
		gain = oldCost - newCost;
	} else {
		double newCost, oldCost;
		
		assert( len == 2 );
		oldCost = eventCount * ( CharCost(t,string[0]) + CharCost(t,string[1])  ); /* In bits */
		newCost = eventCount * SymbolCost(t,eventCount) + dictCost*8.0;
		gain = (LONG)(oldCost - newCost) / 8; /* Convert to bytes */
	
	}
	return gain; /*****/
}


#define strCount 32

/* Finds the string (potential symbol) with the biggest gain */
static CHAR *FindBestString( MTX_RA_COMP *t, MTX_RA_StringCounter *sc, LONG len, LONG *eventCount, LONG *gainPtr, LONG minGain )
{
	CHAR *best;
	register LONG i;
	CHAR *arr[strCount];
	LONG eventCounts1[strCount];
	LONG eventCounts2[strCount];
	LONG gain[strCount];
	LONG minEC = 1;
	
	assert(len > 1 );
	MTX_STRCTR_GetNMaxStrings( sc, strCount, arr, eventCounts1 );
	/* Bump up minEC for low lengths. */
	if ( len <= 3 ) {
		minEC++;
		if ( len <= 2 ) {
			minEC++;
		}
	}
	/* Compute the exact event counts. */
	for ( i = 0; i < strCount; i++ ) {
		assert(arr[i] != NULL || eventCounts1[i] == 0 );
		eventCounts2[i] = eventCounts1[i] <= minEC ? 0 : MatchCount( t, arr[i], len, false, false );
	}

	/* Compute the gains. */
	for ( i = 0; i < strCount; i++ ) {
		gain[i] = -1;
		if ( eventCounts2[i] > 0 ) {
			gain[i] = Gain( t, (BYTE *)arr[i], len, eventCounts2[i] );
		}
	}
	MTX_STRCTR_SortNMaxStrings2( strCount, arr, gain, eventCounts2 );

	for ( i = 1; i < strCount; i++ ) {
		if ( gain[i] <= minGain && arr[i] != NULL  ) {
			MTX_STRCTR_Clear( sc, arr[i] ); /* Clear them so we won't use them again. */
		}
	}

	*eventCount	= eventCounts2[0];
	*gainPtr	= gain[0];
	best		= arr[0];
	
	return best; /*****/
}


	
/*
 * This function creates our dictionary of symbols.
 * It uses a slow heuristic where is adds one symbol
 * at a time to the dictionary using a hill-climbing
 * search approach.
 */
static int GrowDictionary( MTX_RA_COMP *t, FILE *verbose)
{
	register LONG i, len, lastLen = -1;
	LONG eventCount;
	CHAR *string;
	LONG gain, minSavings, minLen;
	LONG savings = 0;
	LONG lastMaxLenforEC2 = MAX_SYMBOL_LENGTH;
	CHAR *p;
	MTX_RA_StringCounter *sc;
	
	sc = MTX_STRCTR_Create( t->mem, -1, t->lengthIn + t->lengthIn/2 + 17 );
	assert( sc != NULL );
	if (!sc)
		return (t->mem->errorCode);
	for ( minSavings = 64*2; minSavings > 0; ) { 
		minSavings = minSavings / 2;
		printf("*** growing dictionary ***\n"); /* */
		/* printf("*** minSavings *** = %ld\n", minSavings ); /* */
		minLen = minSavings > 0 ? 3 : 2;
		for ( len = lastMaxLenforEC2, lastMaxLenforEC2 = 0;len >= minLen && t->symbolCount < t->MaxSymbolCount;) {
			if ( len != lastLen ) {
				LONG limit = t->lengthIn-len;
				lastLen = len;
				
				MTX_STRCTR_ClearHashTable( sc, len );
				assert( sc != NULL );	
				for ( p = t->dataIn, i = 0; i < limit; ) {
					register LONG j, taken = 0;
					register CHAR *ptr = &t->taken[i];
					for ( j = 0; j < len; j++ ) {
						if ( ptr[j] ) {
							taken = j + ptr[j];
							break; /*****/
						}
					}
					if ( taken ) {
						i += taken;
					} else {
						MTX_STRCTR_Increment( sc, &p[i] );
						i++;
					}
				}
				
			}
			if ( len == 2 ) SetAllCounts(t);
			assert( len > 1 );
			string = FindBestString( t, sc, len, &eventCount, &gain, minSavings );
			if ( len > lastMaxLenforEC2 && eventCount >= 2 ) {
				lastMaxLenforEC2 = len;
			}
			if ( gain <= minSavings || ( len == 2 && gain <= 10 ) ) {
				len--;
				continue; /*****/
			}
			MTX_STRCTR_Clear( sc, string );
			StoreString( t, (BYTE *)string, len, eventCount, true, &t->offsets[t->symbolCount] );
			savings  += gain;

#ifdef DEBUG
			/* show the string */
			printf("new token: count=%d ",eventCount);
			show_string(string,len);
#endif

			if (verbose)
				printf("%ld:Len %ld, ec %ld, gain %ld, savings %ld, ",
					t->symbolCount, len, eventCount, gain, savings );

			assert( eventCount >= 1 );
			if  (verbose)
				printf("stringStore %ld, nodes %ld\n", t->stringStoreLength, t->nodeLength );
			i = t->symbolCount++;
			t->dictSymbol[i] = string;
			t->dictSymbolLength[i]  = len;
			i = MatchCount(t,  string, len, (CHAR)len, verbose ); /* mark them so we won't use them again. */
			assert( i == eventCount );
			CompReAllocate( t );
		}
	}
	ResetControlSequence( t, MAX_SYMBOL_LENGTH );

	assert( sc != NULL );
	MTX_STRCTR_Destroy( sc );
	return 0;	/* success */
}

			
/*  
 * Outputs one symbol with the index 'index'.
 * Returns number of bits written.
 */
LONG OutputSymbol( MTX_RA_COMP *t, BITIO *bio, LONG index )
{
	CHAR stack[64];
	LONG bitCount, sp, up, a, ROOT = t->symbolCount * 2 - 2;
	
	sp = 0;
	a = index;
	do {
		up = t->parent[a];
		stack[sp++] = (CHAR)(t->child_1[up] == a);
		a = up;
	} while (a != ROOT);
	
	assert( sp < 64 );

	bitCount = sp;

	if (bio)
	{
		do {
			MTX_BITIO_output_bit( bio, stack[--sp]);
		} while (sp);
	}
	return bitCount; /*****/
}


/* THE call to compress the date in the file called 'nameIn'. */
/* fileType is currently not used .*/
static int MTX_RA_Compress1( MTX_RA_COMP *t, int fileType, CHAR *nameIn,
						CHAR *nodesOut, CHAR *stringsOut, CHAR *bitsOut, FILE *verbose )
{
	register LONG i;
	CHAR *taken;
	int status;
	
	t->fileType = fileType;
	/* Initialize Symbols */
	t->symbolCount = 256;
	status = CompAllocate(t);
	if (status)
		return (status);
	
	/* Initialize the first simple (single byte!) 256 symbols */
	for ( i = 0; i < 256; i++ ) {
		t->dictSymbol[i]    	= &MTX_RA_AllCharArray[i];
		t->dictSymbol[i][0]		= (CHAR) i;
		t->dictSymbolLength[i]	= 1;
		t->offsets[i]			= 0;
	}

	/*  Read into memory.*/
	t->dataIn = (CHAR *)RA_ReadFileIntoMemory( t->mem, nameIn, &t->lengthIn );
	if (!t->dataIn)
		return (t->mem->errorCode);
	
	t->taken  = taken = (CHAR *)MTX_mem_malloc( t->mem, t->lengthIn * sizeof(CHAR) ); assert( t->taken != NULL );
	if (!t->taken && t->mem->errorCode)
		return (t->mem->errorCode);

	/* Initialize, nothing is taken. */
	for ( i = t->lengthIn; i > 0; i-- ) 
		*taken++ = 0;

	t->grade_vector = 0;

	/* Create our symbol set. */
	status = GrowDictionary(t, verbose);
	if (status)
		return (status);

	SetAllCounts(t);

	/*
	 * arrays are 2*n LONG
	 * input:  0..n-1 are set in the count array to >= 0
	 * output: child_0[], child_1[], parent[], and count[] contents 
	 *         starting at n
	 */

	status = MTX_RA_BuildHuffmanTree( t->mem, t->symbolCount, t->count, t->parent, t->child_0, t->child_1 );
	if (status == -1)
		return (t->mem->errorCode);

	assemble_tokens(t);
	make_canonical();

	status = EncodeData( t, t->lengthIn, bitsOut );
	if (status)
		return (t->mem->errorCode);
	
	MTX_mem_free( t->mem, t->dataIn );
	MTX_mem_free( t->mem, t->taken );
	CompDeAllocate(t);

	return 0;	/* success */
}


/* The Constructor. */
MTX_RA_COMP *MTX_COMP_Create( MTX_MemHandler *mem )
{
	MTX_RA_COMP *t		= (MTX_RA_COMP *)MTX_mem_malloc( mem, sizeof( MTX_RA_COMP ) );
	if (!t && mem->errorCode)
		return NULL;
	t->mem				= mem;
	t->symbolCount		= 0;
	t->dictSymbol		= NULL;
	t->dictSymbolLength	= NULL;
	
	t->count = t->parent = t->child_0 = t->child_1 = NULL;

	t->maxStringStoreLength = 50000L; 
	t->stringStoreLength = 0;
	t->stringStore = (BYTE *)MTX_mem_malloc( mem, t->maxStringStoreLength * sizeof(BYTE) ); 
	if (!t->stringStore && t->mem->errorCode)
		return NULL;
	t->nodeLength = 0;
	return t; /******/
}


/* The Destructor. */
void MTX_COMP_Destroy( MTX_RA_COMP *t )
{
	assert( t->stringStoreLength < t->maxStringStoreLength );
	MTX_mem_free( t->mem, t->stringStore);
	MTX_mem_free( t->mem, t );
}



static int SetOutFileNames( CHAR *inName, CHAR *nodesOut, CHAR *stringsOut, CHAR *bitsOut )
{
	int i, len = strlen( inName );
	register CHAR c;

	for ( i = 0; i < len; i++ ) {
		CHAR c = inName[i];
		if ( c == '.' ) break; /*****/
		nodesOut[i]		= c;
		stringsOut[i]	= c;
		bitsOut[i] 		= c;
	}
	c = '.';	
	nodesOut[i]		= c;
	stringsOut[i]	= c;
	bitsOut[i++] 	= c;
	
	nodesOut[i]		= 'N';
	stringsOut[i]	= 'S';
	bitsOut[i++] 	= 'B';
	
	nodesOut[i]		= 'D';
	stringsOut[i]	= 'T';
	bitsOut[i++] 	= 'T';
	
	nodesOut[i]		= 'E';
	stringsOut[i]	= 'R';
	bitsOut[i++] 	= 'S';

	nodesOut[i]		= 0;
	stringsOut[i]	= 0;
	bitsOut[i++] 	= 0;
	return i;
}


static int GuessFileType( CHAR *name )
{
	int i, len = strlen( name );
	int fileType = FILE_TYPE_ANY;
	
	i = len - 4;
	if ( i > 0 ) {
		if ( strncmp( &name[i], ".TTF", 4 ) == 0 ||
			 strncmp( &name[i], ".ttf", 4 ) == 0 ) {
			fileType = FILE_TYPE_TTF; 
			printf("FILE TYPE TTF!!!\n");
		} else if ( strncmp( &name[i], ".CTF", 4 ) == 0 ||
			 strncmp( &name[i], ".ctf", 4 ) == 0 ) {
			fileType = FILE_TYPE_CTF; 
			printf("FILE TYPE CTF!!!\n");
		}
	}
	return fileType;
}

/*
 * Merged File format2:
 *
 * ULongFormat: CURRENT_VERSION_NUMBER
 * ULongFormat: offset from start of file to start of the string-pool
 * ULongFormat: offset from start of file to start of the bitstream
 * ULongFormat: offset from start of file to start of the magic_number right after the bit-stream data
 *
 * sequence of bytes: The data from the nodes file
 * sequence of bytes: The data from the string-pool file
 * sequence of bytes: The data from the bitstream file
 *
 * ULongFormat : MAGIC_NUMBER
 *
 *
 */
/*
 * This function merges the 3 individual files into one file
 * and deletes them.
 */
static LONG MergeFiles( MTX_MemHandler *mem, const CHAR *nodeName, const CHAR *stringName, const CHAR *bitsName, const CHAR *outName )
{
	/* remember "bitsName" is the COMPLETE file */
	int n;
	remove(nodeName);
	remove(stringName);
	remove(outName);
	n = rename(bitsName,outName);
	if (n==0)  /* successfull rename */
		{
		MTX_RA_DECOMP *r;
		BYTE *data;
		LONG size;
		LONG crumbs;

		printf("*** Success ***\n");
		data = RA_ReadFileIntoMemory( mem, outName,	&size );
        if(!data)
        {
            printf("ERROR. RA_ReadFileIntoMemory() failed. errorCode %d\n", mem->errorCode);
            return 0;
        }
		r = MTX_RA_DECOMP_Create(mem, data );
        if(!r)
        {
            printf("ERROR. MTX_RA_DECOMP_Create() failed.\n");
            return 0;
        }
		printf("original size=%d, compressed size=%d\n",r->unCompressedByteCount, size);

		crumbs = r->unCompressedByteCount / BYTE_CACHE_INTERVAL + 1;
		printf("combined size of bitMarks and DeltaBytes = %d\n",5*crumbs);

		/* lengths + string (ignore calculated offsets) */
		printf("dictionary entries=%d, total dictionary size=%d\n",
			r->num_tokens, 4 + r->num_tokens + r->string_len);
		printf("number of huffman groups=%d, total size of huffman apparatus=%d\n",
			r->num_groups, 8+9*r->num_groups);

		MTX_mem_free( mem, data );
		MTX_RA_DECOMP_Destroy(r);
		return size;
		}
	return 0;
}

/* The function packs the file 'name' using MTX_RA_COMP */
int MTX_RA_Pack( CHAR *nameIn, CHAR *nameOut, FILE *verbose )
{
	MTX_MemHandler *mem;
	int bSuccess = false;
	int ierrorNumber = 0;
	LONG size;
    clock_t t1, t2;
    double dur1 = 0.0;
    CHAR nodesOut[256], stringsOut[256], bitsOut[256];

    t1 = t2 = 0; /* To avoid warnings about use without initialization */
	if ( mem = MTX_mem_Create( malloc, realloc, free) )
	{
		MTX_RA_COMP *ra_comp;
		if ( ra_comp = MTX_COMP_Create( mem ) )
		{
			t2 = clock();
			SetOutFileNames( nameIn, nodesOut, stringsOut, bitsOut );
			ierrorNumber = MTX_RA_Compress1( ra_comp, GuessFileType(nameIn), nameIn , nodesOut, stringsOut, bitsOut, verbose );
			if (!ierrorNumber)
			{
				t1 = clock(); 
				MTX_COMP_Destroy( ra_comp);
				dur1 = ((difftime(t1,t2))/ (CLOCKS_PER_SEC));
				
				printf("Compression time = %f\n", (double)dur1);
				
				if ( size = MergeFiles( mem, nodesOut, stringsOut, bitsOut, nameOut ) )
				{
					bSuccess = MTX_RA_Test( mem, nameIn, nameOut );
					MTX_mem_CloseMemory( mem );
					MTX_mem_Destroy( mem );
					return bSuccess; /*****/
				}
				else /* error encountered in MergeFiles */
				{
					RA_DisplayMessage( mem->errorCode );
					MTX_mem_FreeAllMemory(mem);
					MTX_mem_Destroy( mem );
					return bSuccess; /*****/
				}
			}
		}
		else ierrorNumber = mem->errorCode;
	}
	else ierrorNumber = ERR_mem_Create_Failed;

	/* error handling */
	if ( mem != NULL ) {
		MTX_mem_FreeAllMemory(mem);
		MTX_mem_Destroy( mem );
	}
	RA_DisplayMessage( ierrorNumber );
	return bSuccess;
}

int MTX_RA_PackToRAM( MTX_MemHandler *mem, LONG inSize, CHAR *inData,
                            LONG *compressedSize, CHAR ** compressedData)
{
    int bSuccess;

    RA_WriteDataToFile( mem, "intable.tmp", inData, inSize);
    bSuccess= MTX_RA_Pack( "intable.tmp", "tabletmp.tmp", NULL );
    if (!bSuccess)
        return 999;
	*compressedData= (CHAR *)RA_ReadFileIntoMemory( mem, "tabletmp.tmp", compressedSize );
    remove("intable.tmp");
    remove("tabletmp.tmp");
	if (!*compressedData)
		return (mem->errorCode);
    return 0;
}



