/******************************************************************************
 * This software module was originally developed by Agfa Monotype Corporation
 * in the course of development of the ISO/IEC 14996-18 standard for reference 
 * purposes and its performance may not have been optimized. This software 
 * module is an implementation of font compression technology as specified 
 * by the ISO/IEC 14996-18 standard.
 * 
 * ISO/IEC gives users a free copyright license to this software module or
 * modifications thereof for use in products claiming conformance to 
 * ISO/IEC 14996-18 International Standard. ISO/IEC gives users the same free
 * license to this software module or modifications thereof for research 
 * purposes and further ISO/IEC standardization.
 * 
 * Those intending to use this software module in products are advised that
 * its use may infringe existing patents and that this license does not grant 
 * rights under any patents. ISO/IEC have no liability for use of this 
 * software module or modifications thereof.
 * 
 * No license is granted and copyright is not released for products that do 
 * not conform to ISO/IEC 14996-18 International Standard.
 * 
 * Agfa Monotype Corporation retains full right to modify and use the code for
 * its own purpose, assign or donate the code to a third party and to inhibit 
 * third parties from using the code for products that do not conform to
 * audiovisual and image-coding related ITU recommendations and 
 * ISO/IEC International Standards. All rights not expressly granted to the 
 * software module or modifications thereof are reserved by 
 * Agfa Monotype Corporation.
 * 
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2003. All rights reserved.
 ******************************************************************************/

/*
 * File:						STRCTR.C
 * Revision history
 * $Log: Strctr.c,v $
 * Revision 1.1  2004/02/10 22:01:10  vladlev
 * Initial version
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "ERRCODES.H"
#include "CONFIG.h"


#include "MEMORY.H"
#include "STRCTR.H"


/* Computes the initial hash-key for string p. */
/*
 the old HashKey() showed VERY poor distribution.  The average number
 of probes in FindNode() while compressing the file h2sa1m.ctf was 144.
 with the new HashKey() the average number of probes fell to 2.39. This
 reduced FindNode() time from 598,001 tics to 179,730 tics, reducing its
 share (including children) from 34% to 13% of total.
*/
ULONG crc32_table[256];

void init_crc32()
{
	int i, j;
	ULONG c;

	for (i = 0; i < 256; ++i) {
		for (c = i << 24, j = 8; j > 0; --j)
			c = c & 0x80000000 ? (c << 1) ^ 0x04c11db7 : (c << 1);
		crc32_table[i] = c;
	}
}

ULONG crc32(BYTE *buf, int len)
{
	BYTE *p;
	ULONG  crc;

	if (!crc32_table[1])    /* if not already done, */
		init_crc32();   /* build table */
	crc = 0xffffffff;       /* preload shift register, per CRC-32 spec */
	for (p = buf; len > 0; ++p, --len)
		crc = (crc << 8) ^ crc32_table[(crc >> 24) ^ *p];
	return ~crc;            /* transmit complement, per CRC-32 spec */
}



ULONG Hashkey( MTX_RA_StringCounter *t, CHAR *p )
{
	return crc32(p,t->strLength) % t->hashTableSize;
}


/* Finds the hashtable index for string p as fast as possible. */
static LONG FindNode( register MTX_RA_StringCounter *t, register CHAR *p )
{
	register ULONG key = Hashkey( t, p );
	register CHAR **ht_string =  t->ht_string;
	BYTE *ht_xor = t->ht_xor;
	register CHAR *hStr;
	register int i, strLength = t->strLength;
	ULONG hashTableSize = t->hashTableSize;
	BYTE xor;

	/* compute parity byte	*/
	for (xor=0, i=0; i<strLength; i++)
		xor ^= (BYTE) p[i];
	
	while ( (hStr = ht_string[key]) != NULL ) 
	{

		/* different parity ? */
		if (xor != ht_xor[key]) goto tryAgain;

		/* different elements ?	*/
		for (i=0; i<strLength; i++)
			if (hStr[i] != p[i]) goto tryAgain;

		/* they must match ... */
		return key;
tryAgain:
		/* key = (key + 1 ) % hashTableSize;  */
		if (++key == hashTableSize)
			key = 0;

	}

	/* We found a NULL entry => go ahead and take it */
	ht_string[key] 		= p;
	t->ht_xor[key]		= xor;
	t->ht_count[key]	= 0;
	return key; /*****/
}



/* The function increments the event count for string p */
void MTX_STRCTR_Increment( MTX_RA_StringCounter *t, CHAR *p )
{
	LONG nodeIndex = FindNode( t, p );
	USHORT count = ++t->ht_count[ nodeIndex ];
	if ( count > t->maxCount ) {
		t->maxCount		= count;
		t->maxNodeIndex	= nodeIndex;
	}
}

/* The function zeroes the event count for string p */
void MTX_STRCTR_Clear( MTX_RA_StringCounter *t,  CHAR *p )
{
	LONG nodeIndex = FindNode( t, p );
	t->ht_count[ nodeIndex ] = 0;
}

/* Call this to get the most frequent string (highest event count). */
CHAR *MTX_STRCTR_GetMaxString( MTX_RA_StringCounter *t, USHORT *eventCount )
{
	assert( t->maxNodeIndex != -1 );
	*eventCount = t->ht_count[ t->maxNodeIndex ];
	return t->ht_string[ t->maxNodeIndex  ]; /*****/
}





/* the original -- despite my efforts, this has resisted attempts to reduce its runtime. SWP */
void MTX_STRCTR_GetNMaxStrings( MTX_RA_StringCounter *t, LONG n, CHAR **arr, LONG *eventCounts )
{
	register LONG i, j, minIndex = 0;
	register LONG tmp, minValue = 0;
	register USHORT *ht_count = t->ht_count;
	register CHAR *p;

	for ( j = 0; j < n; j++ ) {
		eventCounts[j]	= 0;
		arr[j]			= NULL;
	}
	for ( i = 0; i < (LONG)t->hashTableSize; i++ ) {
		if ( ( p = t->ht_string[i]) != NULL && (tmp = ht_count[i]) > minValue ) {
			eventCounts[minIndex]	= tmp;
			arr[minIndex]			= p;
			/* Find the smallest index; */
			minValue = *eventCounts;
			minIndex = 0;
			for ( j = n-1; j > 0; j-- ) {
				if ( (tmp = eventCounts[j]) < minValue ) {
					minIndex = j;
					minValue = tmp;
				}
			}
		}
	}
	MTX_STRCTR_SortNMaxStrings( n, arr, eventCounts );
}


void MTX_STRCTR_SortNMaxStrings( LONG n, CHAR **arr, LONG *eventCounts )
{
	register LONG max, i, j;
	
	/* Sort them by insertion sort */
	for ( max = 0; max < n; max++ ) {
		i = max;
		for ( j = max+1; j < n; j++ ) {
			if ( eventCounts[j] > eventCounts[i] ) {
				i = j;
			}
		}
		if ( i != max ) {
			register LONG tmp_l;
			register CHAR *tmp_cp;
			/* swap i and max */
			tmp_l = eventCounts[i];
			eventCounts[i] = eventCounts[max];
			eventCounts[max] = tmp_l;

			tmp_cp = arr[i];
			arr[i] = arr[max];
			arr[max] = tmp_cp;
		}
	}
}

/* Sorts the three arrays according to eventCounts. */
void MTX_STRCTR_SortNMaxStrings2( LONG n, CHAR **arr, LONG *eventCounts, LONG *arr2 )
{
	register LONG max, i, j;
	
	/* Sort them by insertion sort */
	for ( max = 0; max < n; max++ ) {
		i = max;
		for ( j = max+1; j < n; j++ ) {
			if ( eventCounts[j] > eventCounts[i] ) {
				i = j;
			}
		}
		if ( i != max ) {
			register LONG tmp_l;
			register CHAR *tmp_cp;
			/* swap i and max */
			tmp_l				= eventCounts[i];
			eventCounts[i]		= eventCounts[max];
			eventCounts[max]	= tmp_l;

			tmp_l 		= arr2[i];
			arr2[i] 	= arr2[max];
			arr2[max] 	= tmp_l;

			tmp_cp 		= arr[i];
			arr[i] 		= arr[max];
			arr[max]	= tmp_cp;
		}
	}
}

/* Call to reset the object */
void MTX_STRCTR_ClearHashTable( register MTX_RA_StringCounter *t, LONG strLength )
{
	register LONG i;
	register LONG hashTableSize = t->hashTableSize;
	t->strLength		= strLength;

	for ( i = 0; i < hashTableSize; i++ )
		{
		t->ht_string[i] = NULL;
		t->ht_xor[i] = 0x00;
		}
	t->maxNodeIndex = -1;
	t->maxCount = 0;
}


/* The constructor. */
MTX_RA_StringCounter *MTX_STRCTR_Create( MTX_MemHandler *mem, LONG strLength, LONG hashTableSize )
{
	MTX_RA_StringCounter *t	= (MTX_RA_StringCounter *)MTX_mem_malloc( mem, sizeof( MTX_RA_StringCounter ) );
	if (!t && mem->errorCode)
		return NULL;
	t->mem				    = mem;

	t->hashTableSize	= hashTableSize;
	t->ht_string = (CHAR **)MTX_mem_malloc( mem, t->hashTableSize * sizeof( CHAR *) ); 
	if (!t->ht_string && mem->errorCode)
		return NULL;
	t->ht_count = (USHORT *)MTX_mem_malloc( mem, t->hashTableSize * sizeof(USHORT ) ); 
	if (!t->ht_count && mem->errorCode)
		return NULL;
	t->ht_xor = (BYTE *)MTX_mem_malloc( mem, t->hashTableSize); 
	if (!t->ht_xor && mem->errorCode)
		return NULL;
	MTX_STRCTR_ClearHashTable( t, strLength );
	return t; /******/
}



/* The destructor. */
void MTX_STRCTR_Destroy (MTX_RA_StringCounter *t )
{
	MTX_mem_free( t->mem, t->ht_string );
	MTX_mem_free( t->mem, t->ht_count );
	MTX_mem_free( t->mem, t->ht_xor );
	MTX_mem_free( t->mem, t );
}

