/******************************************************************************
 * This software module was originally developed by Agfa Monotype Corporation
 * in the course of development of the ISO/IEC 14996-18 standard for reference 
 * purposes and its performance may not have been optimized. This software 
 * module is an implementation of font compression technology as specified 
 * by the ISO/IEC 14996-18 standard.
 * 
 * ISO/IEC gives users a free copyright license to this software module or
 * modifications thereof for use in products claiming conformance to 
 * ISO/IEC 14996-18 International Standard. ISO/IEC gives users the same free
 * license to this software module or modifications thereof for research 
 * purposes and further ISO/IEC standardization.
 * 
 * Those intending to use this software module in products are advised that
 * its use may infringe existing patents and that this license does not grant 
 * rights under any patents. ISO/IEC have no liability for use of this 
 * software module or modifications thereof.
 * 
 * No license is granted and copyright is not released for products that do 
 * not conform to ISO/IEC 14996-18 International Standard.
 * 
 * Agfa Monotype Corporation retains full right to modify and use the code for
 * its own purpose, assign or donate the code to a third party and to inhibit 
 * third parties from using the code for products that do not conform to
 * audiovisual and image-coding related ITU recommendations and 
 * ISO/IEC International Standards. All rights not expressly granted to the 
 * software module or modifications thereof are reserved by 
 * Agfa Monotype Corporation.
 * 
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2003. All rights reserved.
 ******************************************************************************/

/*
 * File:								ttf_conv.cp
 * Revision history
 * $Log: Ttf_conv.c,v $
 * Revision 1.1  2004/02/10 22:01:11  vladlev
 * Initial version
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>
#include <string.h>

#include "CONFIG.H"
#include "SFNT.H"
#include "TTF_CONV.H"
#include "MEMORY.H"
#include "ERRCODES.H"



/* Here we have both big and little endian versions of SWAPENDS. */
/* These routines are needed since the TrueType format is Big-Endian native. */
#ifndef BIG_ENDIAN
SHORT SWAPENDS_16( register SHORT w )
{
	register BYTE /* b0, */ b1;
	
	/* b0 = (CHAR)w; */
	b1 = (CHAR)(w >> 8);
	
	/* w = b0; */
	w <<= 8;
	w |= b1;
	
	return w; /*****/
}

USHORT SWAPENDS_U16( register USHORT w )
{
	register BYTE /* b0, */ b1;
	
	/* b0 = (CHAR)w; */
	b1 = (CHAR)(w >> 8);
	
	/* w = b0; */
	w <<= 8;
	w |= b1;
	
	return w; /*****/
}


LONG SWAPENDS_32( register LONG ww )
{
	register USHORT w0, w1;
	
	w0 = (SHORT)ww;
	w1 = (SHORT)(ww >> 16);
	w0 = SWAPENDS_U16( w0 );
	w1 = SWAPENDS_U16( w1 );
	
	ww = w0;
	ww <<= 16;
	ww |= w1;
	
	return ww; /*****/
}

ULONG SWAPENDS_U32( register ULONG ww )
{
	register USHORT w0, w1;
	
	w0 = (SHORT)ww;
	w1 = (SHORT)(ww >> 16);
	w0 = SWAPENDS_U16( w0 );
	w1 = SWAPENDS_U16( w1 );
	
	ww = w0;
	ww <<= 16;
	ww |= w1;
	
	return ww; /*****/
}
#endif /* BIG_ENDIAN */

/* Reads a word and increments the pointer */
/* It works for both big and little endian machines */
SHORT READWORD_INC( CHAR * *pRef )
{
	SHORT x;
	register CHAR *p = *pRef;
	
	x = (SHORT)((((SHORT)*p) << 8) | *(BYTE *)(p+1));

	*pRef += 2;
	return x; /*****/
}
ULONG READLONG_INC( BYTE * *pRef )
{
	ULONG x;
	register BYTE *p = *pRef;
	x = (ULONG)((((ULONG)*p) << 24) | (((ULONG)*(p+1)) << 16) | (((ULONG)*(p+2)) << 8)
                                                            | ((ULONG)*(p+3)));
	*pRef += 4;
	return x; /*****/
}


/* Writes a word and increments the pointer */
/* It works for both big and little endian machines */
void WRITEWORD_INC( CHAR * *pRef, SHORT x )
{
	register CHAR *p = *pRef;
	*p++ = (CHAR)((USHORT)x >> 8);
	*p++ = (CHAR)x;
	*pRef = p;
}


#define WRITEWORD_INC_MACRO( p, x ) (*(p++) = (CHAR)((USHORT)(x) >> 8), *(p++)	= (CHAR)(x))

/* Writes a LONG and increments the pointer */
/* It works for both big and little endian machines */
void WRITELONG_INC( CHAR * *p, LONG x )
{
	SHORT w0, w1;
	
	w0 = (SHORT)x;
	w1 = (SHORT)(x >> 16);

	WRITEWORD_INC( p, w1 );
	WRITEWORD_INC( p, w0 );
}


const BYTE flipSignCode		= 250;
const BYTE Hop3Code			= 251; /* A,X1,A,X2,A -> A,X1,Hop3Code,X2 - >A,X1,A,X2,A */
const BYTE Hop4Code			= 252; /* A,X1,A,X2,A,X3,A -> A,X1,Hop4Code,X2,X3 - >A,X1,A,X2,A,X3,A */
const BYTE wordCode			= 253;
const BYTE oneMoreByteCode2	= 254;
const BYTE oneMoreByteCode1	= 255;

const SHORT lowestUCode = 253; /* wordCode */

const SHORT lowestCode  = 250; /* flipSignCode */


/* Writes a SHORT in our 255UShort format */
static void Write255UShort( CHAR * *pRef, SHORT value )
{
	register CHAR *p = *pRef;
	
	assert( value >= 0 );
	if ( value >= lowestUCode*3 ) {
		*p++ = wordCode; /* Do a word */
		*p++ = (CHAR)((USHORT)value >> 8);
		*p++ = (CHAR)(value & 0xff);
	} else {
		if ( value >= lowestUCode ) {
			value = (SHORT)(value - lowestUCode);
			if ( value >= lowestUCode ) {
				value = (SHORT)(value - lowestUCode);
				*p++ = oneMoreByteCode2; /* one more byte */
			} else {
				*p++ = oneMoreByteCode1; /* one more byte */
			}
		}
		assert( value >= 0 && value < lowestUCode );
		*p++ = (CHAR)value;
	}
	*pRef = p;
}

/* Reads a SHORT in our 255UShort format */
static SHORT Read255UShort( CHAR * *pRef )
{
	BYTE code;
	SHORT value;
	register CHAR *p = *pRef;
	
	code = *((BYTE *)p++);
	if ( code == wordCode ) {
		value = *p++;
		value <<= 8;
		value |= (BYTE)*p++;
	} else if ( code == oneMoreByteCode1 ) {
		value = *((BYTE *)p++);
		assert( value >= 0 && value < lowestUCode );
		value = (SHORT)(value + lowestUCode);
	} else if ( code == oneMoreByteCode2 ) {
		value = *((BYTE *)p++);
		assert( value >= 0 && value < lowestUCode );
		value = (SHORT)(value + lowestUCode*2);
	} else {
		value = code;
	}
	assert( value >= 0 );
	*pRef = p;
	return value; /*****/
}


/* The Glyph class */
/* This class is internal to this file */
/* and not intended to be used by clients */
/* outside of this file */
typedef struct {
	SHORT xmax, xmin, ymax, ymin;
	SHORT leftSideBearing, advanceWidth;

	SHORT numberOfContours;
	SHORT numberOfPoints;
	
	USHORT *startPoint;
	USHORT *endPoint;
	SHORT *x;
	SHORT *y;
	BYTE *onCurve;
	SHORT *componentData;
	SHORT maxComponentSize;
	SHORT componentSize;
	SHORT componentVersionNumber;
	
	USHORT numberOfInstructions;
	BYTE *code;
	
	
	USHORT	pushCount;
	SHORT			*pushData;
	USHORT	remainingCodeSize;
	BYTE	*remainingCode;

	
	CHAR *inDataStart;		/* The start of the glyph table (ctf or ttf) */
	CHAR *inDataEnd;		/* Points at the first byte past the glyph table (ctf or ttf) */

	MTX_MemHandler *mem;
} Glyph;

static int GlyphsEqual( register Glyph *g1, register Glyph *g2 )
{
	register int i;

	if ( g1->numberOfContours != g2->numberOfContours )	return false; /*****/
	if ( g1->numberOfContours == 0 ) 					return true; /***** NULL GLYPH ******/
	if ( g1->xmax != g2->xmax ) 						return false; /*****/
	if ( g1->xmin != g2->xmin ) 						return false; /*****/
	if ( g1->ymax != g2->ymax ) 						return false; /*****/
	if ( g1->ymin != g2->ymin ) 						return false; /*****/
	
	if ( g1->numberOfContours < 0 ) {
		if ( g1->componentSize != g2->componentSize ) 			return false; /*****/
		for ( i = 0; i < g1->componentSize; i++ ) {
			if ( g1->componentData[i] != g2->componentData[i] )	return false; /*****/
		}
	} else {
		for ( i = 0; i < g1->numberOfContours; i++ ) {
			if ( g1->startPoint[i] != g2->startPoint[i] )		return false; /*****/
			if ( g1->endPoint[i] != g2->endPoint[i] )			return false; /*****/
		}
		
		if ( g1->numberOfPoints != g2->numberOfPoints )			return false; /*****/
		for ( i = 0; i < g1->numberOfPoints; i++ ) {
			if ( g1->x[i] != g2->x[i] )							return false; /*****/
			if ( g1->y[i] != g2->y[i] )							return false; /*****/
			if ( g1->onCurve[i] != g2->onCurve[i] )				return false; /*****/
		}
	}
	
	if ( g1->pushCount != g2->pushCount )						return false; /*****/
	for ( i = 0; i < g1->pushCount; i++ ) {
		if ( g1->pushData[i] != g2->pushData[i] )				return false; /*****/
	}

	if ( g1->remainingCodeSize != g2->remainingCodeSize )		return false; /*****/
	for ( i = 0; i < g1->remainingCodeSize; i++ ) {
		if ( g1->remainingCode[i] != g2->remainingCode[i] )		return false; /*****/
	}

	return true; /***** ***** *****/
}


static int AllocateGlyphSpace( Glyph *t, CHAR * *destSfnt, LONG pos, register LONG byteCount, register LONG *maxOutSize )
{
	byteCount += 4; /* Always keep a 4 byte margin to allow for the zero padding */
	if ( pos + byteCount > *maxOutSize ) {
		*maxOutSize = pos + byteCount + (*maxOutSize>>1) + 2; /* Allocate memory in exponentially increasing steps */
		*destSfnt = (CHAR *)MTX_mem_realloc( t->mem, *destSfnt, *maxOutSize );
		if (!(*destSfnt) && t->mem->errorCode)
			return (t->mem->errorCode);
	}
	return 0;	/* success */
}

/* Getter function for the bounding box */
static void MTX_Glyph_GetBBox( Glyph *t, SHORT *xmax, SHORT *xmin, SHORT *ymax, SHORT *ymin )
{
	*xmax = t->xmax;
	*xmin = t->xmin;
	*ymax = t->ymax;
	*ymin = t->ymin;
}

/* Computes the bounding box for this glyph, returns results in the arguments */
static void MTX_Glyph_ComputeBBox( Glyph *t, SHORT *xmax, SHORT *xmin, SHORT *ymax, SHORT *ymin )
{
	LONG i;
	
	*xmax = *xmin = t->x[0];
	*ymax = *ymin = t->y[0];
	
	for ( i = 1; i < t->numberOfPoints; i++ ) {
		if ( t->x[i] > *xmax ) {
			*xmax = t->x[i];
		} else if ( t->x[i] < *xmin ) {
			*xmin = t->x[i];
		}
		if ( t->y[i] > *ymax ) {
			*ymax = t->y[i];
		} else if ( t->y[i] < *ymin ) {
			*ymin = t->y[i];
		}
	}
}

/* Parses a sequence of TrueType code. The most important */
/* Function of this code to split the data into initally pushed data */
/* everything else */
static int ParseCode( register Glyph *t )
{
	BYTE uByte, argc_b, argc_w;
	SHORT stmp;
	LONG i, j;
	LONG pushBytes;
	
	MTX_mem_free( t->mem,t->pushData); 			t->pushData = NULL;
	MTX_mem_free( t->mem,t->remainingCode); 	t->remainingCode = NULL;

	t->pushData = (SHORT *)MTX_mem_malloc( t->mem, sizeof(SHORT) *  t->numberOfInstructions );
	if (!t->pushData && t->mem->errorCode)
		return (t->mem->errorCode);

	t->remainingCode = (BYTE *)MTX_mem_malloc( t->mem, sizeof( BYTE ) * t->numberOfInstructions );
	if (!t->remainingCode && t->mem->errorCode)
		return (t->mem->errorCode);

	pushBytes 				= 0;
	t->pushCount 			= 0;
	t->remainingCodeSize	= 0;
	
	for ( i = 0; i < t->numberOfInstructions;  ) {
		argc_b = argc_w = 0;
		uByte = t->code[i++];
		if ( uByte == 0x40 ) {
			/* NPUSHB */
			argc_b = t->code[i++];
		} else if ( uByte >= 0xB0 && uByte <= 0xB7 ) {
			/* PUSHB */
			argc_b = (BYTE)(uByte - 0xB0 + 1);
		} else if ( uByte == 0x41 ) {
			/* NPUSHW */
			argc_w = t->code[i++];
		} else if ( uByte >= 0xB8 && uByte <= 0xBF ) {
			/* PUSHW */
			argc_w = (BYTE)(uByte - 0xB8 + 1);
		} else {
			i--;
		}

		if ( argc_b > 0 ) {
			for ( j = 0; j < argc_b; j++ ) {
				t->pushData[t->pushCount++] = t->code[i++];
			}
		} else if ( argc_w > 0 ) {
			for ( j = 0; j < argc_w; j++ ) {
				stmp = t->code[i++];
				stmp <<= 8;
				stmp |= (BYTE)t->code[i++];
				t->pushData[t->pushCount++] = stmp;
			}
		} else {
			pushBytes = i;
			while ( i < t->numberOfInstructions ) {
				t->remainingCode[ t->remainingCodeSize++] = t->code[i++];
			}
			break; /****** not a push */
		}
	}
	assert( i == t->numberOfInstructions  );
		

	assert( pushBytes+t->remainingCodeSize == t->numberOfInstructions  );
	assert( t->pushCount+t->remainingCodeSize <= t->numberOfInstructions  );
	assert( t->pushCount >= 0 );
	assert( t->remainingCodeSize >= 0 );

	return 0;	/* success */
}


/* Computes the bounding box for this glyph */
static void SetBBox( register Glyph *t )
{
	register LONG i, limit;
	register SHORT coord, *xp, *yp;
	
	xp = t->x;
	yp = t->y;
	t->xmax = t->xmin = xp[0];
	t->ymax = t->ymin = yp[0];
	limit = t->numberOfPoints;
	for ( i = 1; i < limit; i++ ) {
		if ( (coord = xp[i]) > t->xmax ) {
			t->xmax = coord;
		} else if ( coord < t->xmin ) {
			t->xmin = coord;
		}
		if ( (coord = yp[i]) > t->ymax ) {
			t->ymax = coord;
		} else if ( coord < t->ymin ) {
			t->ymin = coord;
		}
	}
}


static void CheckReadBoundsForTable( TTF_Converter *t, CHAR *readStart, LONG byteCount )
{
	if ( (readStart+byteCount) > t->tableEnd || readStart < t->tableStart ) 
	{
		t->mem->errorCode = ERR_OutOfBoundsRead_In_InputFont;
		printf("CheckReadBoundsForTable error = %d\n",ERR_OutOfBoundsRead_In_InputFont);
	}
}

static void CheckReadBoundsGlyph( Glyph *t, CHAR *readStart, LONG byteCount )
{
	if ( (readStart+byteCount) > t->inDataEnd || readStart < t->inDataStart )
	{
		t->mem->errorCode = ERR_OutOfBoundsRead_In_InputFont;
		printf("CheckReadBoundsGlyph error = %d\n",ERR_OutOfBoundsRead_In_InputFont);
	}
}

/* Reads a glyph in the TrueType format */
static LONG ReadTTFSpline( Glyph *t, CHAR *sp)
{
	LONG i, j;
	SHORT stmp;
	BYTE bitflags;
	CHAR *spStart = sp;
	
	CheckReadBoundsGlyph( t, sp, 5 * 2 ); /* numberOfContours, + the bounding box */
	t->numberOfContours = READWORD_INC( &sp );
	t->xmin = READWORD_INC( &sp );
	t->ymin = READWORD_INC( &sp );
	t->xmax = READWORD_INC( &sp );
	t->ymax = READWORD_INC( &sp );
	
	MTX_mem_free( t->mem,t->code); 			t->code = NULL;
	MTX_mem_free( t->mem,t->startPoint);	t->startPoint = NULL;
	MTX_mem_free( t->mem,t->onCurve);		t->onCurve = NULL;
	MTX_mem_free( t->mem,t->x);				t->x = NULL;
	
	t->numberOfPoints = 0; /* Initialize */
	if ( t->numberOfContours < 0  ) {
 		USHORT flags, glyphIndex;
		int weHaveInstructions;
 		
 		t->componentVersionNumber = t->numberOfContours;
 		assert( t->componentData != NULL );
 		do {
 			CheckReadBoundsGlyph( t, sp, 2*2 ); /* flags, glyphIndex */
 			flags = READWORD_INC( &sp );
 			weHaveInstructions = flags & WE_HAVE_INSTRUCTIONS;
 			t->componentData[ t->componentSize++] = flags;
 			
 			t->componentData[ t->componentSize++] = glyphIndex = READWORD_INC( &sp );
 			if ( flags & ARG_1_AND_2_ARE_WORDS ) {
 				/* arg 1 and 2 as words */
 				CheckReadBoundsGlyph( t, sp, 2*2 );
	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
 			} else {
 				/* arg 1 and 2 as bytes */
  				CheckReadBoundsGlyph( t, sp, 2 );
	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
 			}
 			
 			if ( flags & WE_HAVE_A_SCALE ) {
 				/* scale */
  				CheckReadBoundsGlyph( t, sp, 2 );
  	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
 			} else if ( flags & WE_HAVE_AN_X_AND_Y_SCALE ) {
 				/* xscale, yscale */
  				CheckReadBoundsGlyph( t, sp, 2*2 );
   	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
  	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
 			} else if ( flags & WE_HAVE_A_TWO_BY_TWO ) {
 				/* xscale, scale01, scale10, yscale */
  				CheckReadBoundsGlyph( t, sp, 4*2 );
   	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
  	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
   	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
  	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
 			}
 			assert( t->componentSize < t->maxComponentSize );
 		} while ( flags & MORE_COMPONENTS );
 		t->numberOfInstructions  = 0;
 		if ( weHaveInstructions ) {
 			CheckReadBoundsGlyph( t, sp, 2 ); /* numberOfInstructions */
		 	t->numberOfInstructions  = READWORD_INC( &sp );
			/*code = new BYTE [ t->numberOfInstructions  ];assert( code != NULL ); */
			t->code = (BYTE *)MTX_mem_malloc( t->mem, sizeof( BYTE ) * t->numberOfInstructions  );
			if (!t->code && t->mem->errorCode)
				return -1;
			
 			CheckReadBoundsGlyph( t, sp, t->numberOfInstructions ); /* the instructions */
			for ( i = 0; i < t->numberOfInstructions ; i++ ) {
		        t->code[i] = (BYTE)*sp++;
		    }
 		}
	} else if ( t->numberOfContours > 0 ) {
		t->startPoint = (USHORT *)MTX_mem_malloc( t->mem, sizeof( USHORT ) * 2 * t->numberOfContours );
		if (!t->startPoint && t->mem->errorCode)
				return -1;
		t->endPoint = &t->startPoint[t->numberOfContours];
		
		/* Non Composite */
 		CheckReadBoundsGlyph( t, sp, t->numberOfContours * 2 ); /* the end points */
		stmp = 0;
		for ( i = 0; i < t->numberOfContours; i++ ) {
		    t->startPoint[i]	= stmp;
		    t->endPoint[i]	= READWORD_INC( &sp );
		 	assert( t->endPoint[i] >= 0  );
		 	stmp = (SHORT)(t->endPoint[i] + 1);
		}
		t->numberOfPoints = (SHORT)(t->endPoint[t->numberOfContours - 1] + 1);
		assert( t->numberOfPoints > 0 );

 		CheckReadBoundsGlyph( t, sp, 2 ); /* numberOfInstructions */
		t->numberOfInstructions  = READWORD_INC( &sp );
 		CheckReadBoundsGlyph( t, sp, t->numberOfInstructions ); /* the instructions */
		if ( t->numberOfInstructions  > 0 ) {
			t->code = (BYTE *)MTX_mem_malloc( t->mem, sizeof(BYTE) * t->numberOfInstructions  );
			if (!t->code && t->mem->errorCode)
				return -1;
			for ( i = 0; i < t->numberOfInstructions ; i++ ) {
		        t->code[i] = *sp++;
		    }
	    }
		t->onCurve = (BYTE *)MTX_mem_malloc( t->mem, sizeof(BYTE) * t->numberOfPoints );
		if (!t->onCurve && t->mem->errorCode)
			return -1;

	    for ( i = 0; i < t->numberOfPoints; ) {
 			CheckReadBoundsGlyph( t, sp, 1 );
	        t->onCurve[i++] = bitflags = *sp++;
			if ( bitflags & REPEAT_FLAGS ) {
	            for ( j = *sp++; j--; ) {
	            	t->onCurve[i++] = bitflags;
	            }
	        }
	    }
		t->x = (SHORT *)MTX_mem_malloc( t->mem, sizeof( SHORT ) * 2 * t->numberOfPoints );
		if (!t->x && t->mem->errorCode)
			return -1;
		t->y = &t->x[t->numberOfPoints];

		for ( stmp = 0, i = 0; i < t->numberOfPoints; i++ ) {
			bitflags = t->onCurve[i];
		 	if ( bitflags & XSHORT ) {
		 		if ( bitflags & SHORT_X_IS_POS ) {
 					CheckReadBoundsGlyph( t, sp, 1 );
		 			stmp = (SHORT)(stmp + (*(BYTE *)sp++));
		 		} else {
 					CheckReadBoundsGlyph( t, sp, 1 );
		 			stmp = (SHORT)(stmp - (*(BYTE *)sp++));
		 		}
		 	} else if ( ! ( bitflags & NEXT_X_IS_ZERO) ) {
 				CheckReadBoundsGlyph( t, sp, 2 );
		 		stmp = (SHORT)(stmp + READWORD_INC( &sp ));
		 	}
			t->x[i] = stmp;
		}
	
		for ( stmp = 0, i = 0; i < t->numberOfPoints; i++ ) {
			bitflags = t->onCurve[i];
		 	if ( bitflags & YSHORT ) {
		 		if ( bitflags & SHORT_Y_IS_POS ) {
 					CheckReadBoundsGlyph( t, sp, 1 );
		 			stmp = (SHORT)(stmp + (*(BYTE *)sp++));
		 		} else {
 					CheckReadBoundsGlyph( t, sp, 1 );
		 			stmp = (SHORT)(stmp - (*(BYTE *)sp++));
		 		}
		 	} else if ( ! ( bitflags & NEXT_Y_IS_ZERO) ) {
 				CheckReadBoundsGlyph( t, sp, 2 );
		 		stmp = (SHORT)(stmp + READWORD_INC( &sp ));
		 	}
			t->y[i] = stmp;
			t->onCurve[i] = (BYTE)(bitflags & ONCURVE); /* Clear other bitflags */
		}
	}
	return sp - spStart; /*****/
}


/* Writes out a glyph in the TrueType format */
static LONG MTX_Glyph_WriteTTFSpline( register Glyph *t, CHAR * *destSfnt, LONG pos, LONG *maxOutSize )
{
	register BYTE *f;
	register SHORT i, limit;
	register SHORT x, y;
	SHORT delta, j;
	BYTE bitFlags;
	LONG deltaFromStart;
	/* Pointers dependent on destSfnt */
	register CHAR *p, *pStart;

	
	if (t->numberOfContours == 0 ) {
		return 0; /******/
	}
	
	deltaFromStart = 0;
	if (AllocateGlyphSpace( t, destSfnt, pos, 10, maxOutSize ))
		return -1;
	pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
	
	if ( t->componentSize > 0 ) {
		assert( t->componentVersionNumber < 0 );
		WRITEWORD_INC_MACRO( p, t->componentVersionNumber );
	} else {
		WRITEWORD_INC_MACRO( p, t->numberOfContours );
	}

	/* Bounding box */
	x = t->xmin; WRITEWORD_INC_MACRO( p, x );
	x = t->ymin; WRITEWORD_INC_MACRO( p, x );
	x = t->xmax; WRITEWORD_INC_MACRO( p, x );
	x = t->ymax; WRITEWORD_INC_MACRO( p, x );
	
	if ( t->componentSize > 0 ) {
		deltaFromStart = p - pStart;
		if (AllocateGlyphSpace( t, destSfnt, pos, deltaFromStart+t->componentSize*2, maxOutSize ))
			return -1;
		pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
		limit = t->componentSize;
		for ( i = 0; i < limit; i++ ) {
			x = t->componentData[i];
			WRITEWORD_INC_MACRO( p, x );
		}
		deltaFromStart = p - pStart;
		if (AllocateGlyphSpace( t, destSfnt, pos, deltaFromStart+2+t->numberOfInstructions , maxOutSize ))
			return -1;
		pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
		if ( t->numberOfInstructions  > 0 ) {
			limit = t->numberOfInstructions;
			WRITEWORD_INC_MACRO( p, limit );
			f = t->code;
			for ( i = 0; i < limit ; i++ ) {
				*p++ = f[i];
			}
		}
		return ( p - pStart ); /******/
	}

	deltaFromStart = p - pStart;
	if (AllocateGlyphSpace( t, destSfnt, pos, deltaFromStart+t->numberOfContours*2, maxOutSize ))
		return -1;
	pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
	limit = t->numberOfContours;
	for ( i = 0; i < limit; i++ ) {
		USHORT ep = t->endPoint[i];
		WRITEWORD_INC_MACRO( p, ep );
	}

	deltaFromStart = p - pStart;
	if (AllocateGlyphSpace( t, destSfnt, pos, deltaFromStart+2+t->numberOfInstructions , maxOutSize ))
		return -1;
	pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
	limit = t->numberOfInstructions;
	WRITEWORD_INC_MACRO( p, limit );
	f = t->code;
	for ( i = 0; i < limit; i++ ) {
		*p++ = f[i];
	}
	
	f = (BYTE *)MTX_mem_malloc( t->mem, sizeof( BYTE ) * t->numberOfPoints );
	if (!f && t->mem->errorCode)
		return -1;
	/* Calculate flags */
	limit = t->numberOfPoints;
	for ( x = y = i = 0; i < limit; i++ ) {
		bitFlags = (BYTE)(t->onCurve[i] & ONCURVE);
		delta = (SHORT)(t->x[i] - x); x = t->x[i];
		if ( delta == 0 ) {
			bitFlags |= NEXT_X_IS_ZERO;
		} else if ( delta >= -255 && delta <= 255 ) {
			bitFlags |= XSHORT;
			if ( delta > 0 ) {
				bitFlags |= SHORT_X_IS_POS;
			}
		}
		delta = (SHORT)(t->y[i] - y); y = t->y[i];
		if ( delta == 0 ) {
			bitFlags |= NEXT_Y_IS_ZERO;
		} else if ( delta >= -255 && delta <= 255 ) {
			bitFlags |= YSHORT;
			if ( delta > 0 ) {
				bitFlags |= SHORT_Y_IS_POS;
			}
		}
		f[i] = bitFlags;
	}

	/* limit == t->numberOfPoints already set */
	deltaFromStart = p - pStart;
	if (AllocateGlyphSpace( t, destSfnt, pos, deltaFromStart+limit, maxOutSize ))
		return -1; /* At least big enough */
	pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
	/* Write out bitFlags */
	for ( i = 0; i < limit;) {
		BYTE repeat = 0;
		for ( j = (SHORT)(i+1); j < limit && f[i] == f[j] && repeat < 255; j++ )
			repeat++;
		if ( repeat > 1 ) {
			*p++ = (CHAR)(f[i] | REPEAT_FLAGS);
			*p++ = repeat;
			i = j;
		} else {
			*p++ = f[i++];
		}
	}

	/* limit == t->numberOfPoints already set */
	deltaFromStart = p - pStart;
	if (AllocateGlyphSpace( t, destSfnt, pos, deltaFromStart+limit+limit, maxOutSize ))
		return -1; /* At least big enough */
	pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
	/* Write out X */
	for ( x = i = 0; i < limit; i++ ) {
		delta = (SHORT)(t->x[i] - x); x = t->x[i];
		if ( f[i] & XSHORT ) {
			if ( !(f[i] & SHORT_X_IS_POS) ) delta = (SHORT)-delta;
			*p++ = (CHAR)delta;
		} else if ( !(f[i] & NEXT_X_IS_ZERO ) ) {
			WRITEWORD_INC_MACRO( p, delta );
		}
	}

	/* Write out Y */
	/* limit == t->numberOfPoints already set */
	deltaFromStart = p - pStart;
	if (AllocateGlyphSpace( t, destSfnt, pos, deltaFromStart+limit+limit, maxOutSize ))
		return -1; /* At least big enough */
	pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
	
	for ( y = i = 0; i < limit; i++ ) {
		delta = (SHORT)(t->y[i] - y); y = t->y[i];
		if ( f[i] & YSHORT ) {
			if ( !(f[i] & SHORT_Y_IS_POS) ) delta = (SHORT)-delta;
			*p++ = (CHAR)delta;
		} else if ( !(f[i] & NEXT_Y_IS_ZERO ) ) {
			WRITEWORD_INC_MACRO( p, delta );
		}
	}
	MTX_mem_free( t->mem,f);
	
	return ( p - pStart ); /*****/
}


typedef struct {
	SHORT			byteCount;
	SHORT			xBits;
	SHORT			yBits;
	SHORT			dxPlus;
	SHORT			dyPlus;
	CHAR 			xIsNegative; /* No meaning if xBits == 0 */
	CHAR 			yIsNegative; /* No meaning if yBits == 0 */
} coordEncodingType;

/* This data structure is read only, except when it is initialized once */
/* by InitCoordEncoding(), this means it can be shared in a preemptive */
/* multi-tasking environment. */
static coordEncodingType coordEncoding[128];
static CHAR coordEncodingIsInited = false;

/* Initializes the relative coordinate encoding technique used */
/* to store the coordinates in the CTF format. *
/* Is is designed to be more compact than TrueType,  */
/* but still allow compaction by traditional compression methods. */
void MTX_TTC_InitCoordEncoding( void )
{
	SHORT i, index = 0;
	SHORT dxPlus, dyPlus;

	if ( coordEncodingIsInited ) return; /******/
	for ( dyPlus = 0; dyPlus <= 1024; dyPlus += 256 ) {
		coordEncoding[index].byteCount = 2;
		coordEncoding[index].xBits 	= 0;
		coordEncoding[index].yBits 	= 8;
		coordEncoding[index].dxPlus = 0;
		coordEncoding[index].dyPlus = dyPlus;
		index++;
	}
	for ( dxPlus = 0; dxPlus <= 1024; dxPlus += 256 ) {
		coordEncoding[index].byteCount = 2;
		coordEncoding[index].xBits 	= 8;
		coordEncoding[index].yBits 	= 0;
		coordEncoding[index].dxPlus = dxPlus;
		coordEncoding[index].dyPlus = 0;
		index++;
	}
	
	/* No need to start at zero since the prior cases will have taken care of that case */
	for ( dxPlus = 1; dxPlus <= 49+1; dxPlus += 16 ) {
		for ( dyPlus = 1; dyPlus <= 49+1; dyPlus += 16 ) {
			coordEncoding[index].byteCount = 2;
			coordEncoding[index].xBits 	= 4;
			coordEncoding[index].yBits 	= 4;
			coordEncoding[index].dxPlus = dxPlus;
			coordEncoding[index].dyPlus = dyPlus;
			index++;
		}
	}

	/* No need to start at zero since the prior cases will have taken care of that case */
	for ( dxPlus = 1; dxPlus <= 513; dxPlus += 256 ) {
		for ( dyPlus = 1; dyPlus <= 513; dyPlus += 256 ) {
			coordEncoding[index].byteCount = 3;
			coordEncoding[index].xBits 	= 8;
			coordEncoding[index].yBits 	= 8;
			coordEncoding[index].dxPlus = dxPlus;
			coordEncoding[index].dyPlus = dyPlus;
			index++;
		}
	}

	
	coordEncoding[index].byteCount	= 4;
	coordEncoding[index].xBits 		= 12;
	coordEncoding[index].yBits 		= 12;
	coordEncoding[index].dxPlus 	= 0;
	coordEncoding[index].dyPlus 	= 0;
	index++;
	
	coordEncoding[index].byteCount	= 5;
	coordEncoding[index].xBits 		= 16;
	coordEncoding[index].yBits 		= 16;
	coordEncoding[index].dxPlus 	= 0;
	coordEncoding[index].dyPlus 	= 0;
	/*index++; */
	
	
	/*index--; */
	for ( i = index; i >= 0; i-- ) {
		coordEncoding[i].xIsNegative = false;
		coordEncoding[i].yIsNegative = false;
	}
	for ( i = 127; index >= 0;) {
		coordEncoding[i] = coordEncoding[index];
		i--;
		if ( coordEncoding[i+1].xBits == 0 ) {
			/* *2 */
			coordEncoding[i] = coordEncoding[index];
			coordEncoding[i].yIsNegative = true;
			i--;
		} else if ( coordEncoding[i+1].yBits == 0 ) {
			/* *2 */
			coordEncoding[i] = coordEncoding[index];
			coordEncoding[i].xIsNegative = true;
			i--;
		} else {
			/* *4 */
			coordEncoding[i] = coordEncoding[index];
			coordEncoding[i].xIsNegative = true;
			i--;
			coordEncoding[i] = coordEncoding[index];
			coordEncoding[i].yIsNegative = true;
			i--;
			coordEncoding[i] = coordEncoding[index];
			coordEncoding[i].xIsNegative = true;
			coordEncoding[i].yIsNegative = true;
			i--;
		}
		index--;
	}
	assert( i == -1 );
	assert( index == -1 );
	coordEncodingIsInited = true;
}



const BYTE cvt_pos8				= 255;
const BYTE cvt_pos1				= 255 - 7;					/* == cvt_pos8 - 7 */
const BYTE cvt_neg8				= 255 - 7 - 1;				/* == cvt_pos1 - 1 */
const BYTE cvt_neg1				= 255 - 7 - 1 - 7;			/* == cvt_neg8 - 7 */
const BYTE cvt_neg0				= 255 - 7 - 1 - 7 - 1;		/* == cvt_neg1 - 1 */
const BYTE cvt_wordCode			= 255 - 7 - 1 - 7 - 1 - 1;	/* == cvt_neg0 - 1 */

const SHORT cvt_lowestCode 					= 255 - 7 - 1 - 7 - 1 - 1;	/* == cvt_wordCode */




/* encode a SHORT in 1 to 3 bytes (255 SHORT format ); */
static void Write255Short( CHAR * *pRef, SHORT value )
{
	register CHAR *p = *pRef;
	
	if ( value >= lowestCode+512 || value < -256 ) {
		*p++ = wordCode; /* Do a word */
		*(BYTE *)p++ = (BYTE)(value >> 8);
		*(BYTE *)p++ = (BYTE)(value & 0xff);
	} else {
		if ( value < 0 ) {
			/* value <= -1 && value >= -256 */
			*p++ = flipSignCode; /* flip the sign */
			value = (SHORT)-value; /* value >= 1 && value <= 256 */
			value--; /* value >= 0 && value <= 255 */
			*p++ = (CHAR)value;
		} else if ( value >= lowestCode ) {
			value = (SHORT)(value - lowestCode);
			if ( value >= 256 ) {
				value = (SHORT)(value - 256 );
				assert( value < 256 );
				*p++ = oneMoreByteCode2; /* one more byte */
			} else {
				*p++ = oneMoreByteCode1; /* one more byte */
			}
			*p++ = (CHAR)value;
		} else {
			assert( value >= 0 && value < lowestCode );
			*p++ = (CHAR)value;
		}
	}
	*pRef = p;
}

/* Reads a SHORT in our 255 SHORT format */
static SHORT Read255Short( register CHAR * *pRef )
{
	register SHORT value, code;
	register CHAR *p = *pRef;
	
	value = *((BYTE *)p++);
	if ( value >= lowestCode ) {
		code	= value;
		value	= *((BYTE *)p++);	/* value >= 0 && value <= 255 */
		if ( code == wordCode ) {
			value <<= 8;
			value |= *(BYTE *)p++;
		} else if ( code == flipSignCode ) {
			value++;							/* value >= 1 && value <= 256 */
			value = (SHORT)-value;						/* value <= -1 && value >= -256 */
		} else {
			assert( code >= 254 );
			assert( oneMoreByteCode1 == 255 && oneMoreByteCode2 == 254 );
			value = (SHORT)(value + lowestCode + (255-code)*256 );
		}
	}
	*pRef = p;
	return value; /*****/
}



/* Stores the intial push sequence in CTF format */
static void PushInitialCTFStack( Glyph *t, CHAR * *pExternal )
{
	register SHORT i;
	CHAR *p, *p0;
	LONG byteCount;
	
	assert( t->pushCount >= 0 );
	p0 = *pExternal;
	Write255UShort( pExternal, t->pushCount );
	p = *pExternal;
	
	/* Assume worst case */
	byteCount = 3 + t->pushCount * 3;

	
	for ( i = 0; i < t->pushCount; ) {
		if ( i-2 >= 0 && t->pushData[i] == t->pushData[i-2] && i+2 < t->pushCount && t->pushData[i] == t->pushData[i+2] ) {
			if ( i+4 < t->pushCount && t->pushData[i] == t->pushData[i+4] ) {
				*p++ = Hop4Code; i++;
				Write255Short( &p, t->pushData[i++] );
				i++;
				Write255Short( &p, t->pushData[i++] );
				i++;
			} else {
				*p++ = Hop3Code; i++;
				Write255Short( &p, t->pushData[i++] );
				i++;
			}
		} else {
			Write255Short( &p, t->pushData[i++] );
		}
	}
	assert( i == t->pushCount );
	assert( p - p0 <= byteCount );
	*pExternal = p;
}

/* Writes out the remaining CTF code beyond the inital push sequence */
static void WriteRemainingCTFCode( Glyph *t, CHAR * *pExternal )
{
	LONG byteCount;
	assert( t->remainingCodeSize >= 0 );
	
	Write255UShort( pExternal, t->remainingCodeSize );
	byteCount = t->remainingCodeSize;
	
	memcpy( *pExternal, &t->remainingCode[0], byteCount );
	*pExternal += byteCount;
}

/* Emits a TrueType PUSHB[] or NPUSHB[] instruction */
static void BytePush( CHAR * *pRef, SHORT argStore[], LONG numberofArgs )
{
	register LONG i;
	register CHAR *p = *pRef;

	assert( numberofArgs > 0 );
	if ( numberofArgs <= 8 ) {
		*p++ = (CHAR)(0xB0 + numberofArgs -1);/* PUSHB */
	} else {
		*p++ = 0x40; /* NPUSHB */
		assert( numberofArgs <= 255 );
		*p++ = (CHAR)numberofArgs;
	}
	for ( i = 0; i < numberofArgs; i++ ) {
		assert( argStore[i] >= 0 && argStore[i] <= 255 );
		*p++ = (CHAR)argStore[i];
	}
	*pRef = p;
}

/* Emits a TrueType PUSHW[] or NPUSHW[] instruction */
static void WordPush( CHAR * *pRef, SHORT argStore[], LONG numberofArgs )
{
	register LONG i;
	register CHAR *p = *pRef;

	assert( numberofArgs > 0 );
	if ( numberofArgs <= 8 ) {
		*p++ = (CHAR)(0xB8 + numberofArgs -1);/* PUSHW */
	} else {
		*p++ = 0x41; /* NPUSHW */
		assert( numberofArgs <= 255 );
		*p++ = (CHAR)numberofArgs;
	}
	for ( i = 0; i < numberofArgs; i++ ) {
		*p++ = (CHAR)((argStore[i] >> 8) & 0x00ff);
		*p++ = (CHAR)(argStore[i] & 0x00ff);
	}
	*pRef = p;
}


/* Does an optimized sequence of TrueType push statements */
static void OptimizingPush( CHAR * *p, SHORT argStore[], LONG numberofArgs )
{
	register SHORT i, argument, n2;
	SHORT argCount, n, byteRun, runLength;
	SHORT doBytePush, limit;

	for ( n = 0; numberofArgs > 0;  ) {
		argCount = (SHORT)(numberofArgs > 255 ? 255 : numberofArgs);
		doBytePush = false;
		
		limit = 2;
		for ( runLength = 0, i = n; i < (n + argCount); i++ ) {
			argument = argStore[i];
			if ( argument >= 0  && argument <= 255  ) {
				/* byteRun = ByteRunLength( &argStore[i], (SHORT)(n + argCount - i) ); */
				/* set 'byteRun' = the number of bytes that all fit within an unsigned byte */
				n2 			= (SHORT)(n + argCount - i);
				for ( byteRun = 0; byteRun < n2; ) {
					argument = argStore[++byteRun + i];
					if ( argument > 255 || argument < 0 ) break; /*****/
				}
				
				/* if we have a run of bytes of length 3 (2 if first or last) or more it is more
				   optimal in terms of space to push them as bytes */
				if ( ((byteRun + i) >= (n + argCount)) ) {
					limit = 2;
				}
				if ( byteRun >= limit ) {
					if ( runLength > 0 ) {
						argCount	= runLength;
						doBytePush	= false;
					} else {
						argCount	= byteRun;
						doBytePush	= true;
					}
					break; /*****/
				}
			}
			runLength++;
			limit = 3;
		}
		if ( doBytePush ) {
			BytePush( p, &argStore[n], argCount );
		} else {
			WordPush( p, &argStore[n], argCount );
		}
		numberofArgs -= argCount;
		n = (SHORT)(n+argCount);
	}
	assert( numberofArgs == 0 );
}

/* Reads CTF instructions */
static int ReadAllCTFInstructions( Glyph *t, CHAR * *p )
{
	register SHORT i, limit, *pushData;
	register BYTE uc;
	BYTE *out;

	MTX_mem_free( t->mem,t->pushData); 		t->pushData			= NULL;

	t->pushCount	= limit = Read255UShort( p );
	t->pushData 	= (SHORT *)MTX_mem_malloc( t->mem, sizeof( SHORT ) * limit );
	if (!t->pushData && t->mem->errorCode)
		return (t->mem->errorCode);
	
	pushData = t->pushData;
	for ( i = 0; i < limit; ) {
		uc = **p;
		if ( uc < lowestCode ) {
			/* This is a speed optimization */
			pushData[i++] = uc;						/* Data */
			(*p)++;
		} else if ( uc == Hop3Code ) {
			register SHORT A  = pushData[i-2];
			(*p)++;									/* Hop3Code */
			pushData[i++] = A;   					/* A */
			pushData[i++] = Read255Short( p );		/* X2 */
			pushData[i++] = A;   					/* A */
		} else if ( uc == Hop4Code ) {
			register SHORT A  = pushData[i-2];
			(*p)++;									/* Hop4Code */
			pushData[i++] = A;  					/* A */
			pushData[i++] = Read255Short( p );		/* X2 */
			pushData[i++] = A;						/* A */
			pushData[i++] = Read255Short( p );		/* X3 */
			pushData[i++] = A;  					/* A */
		} else {
			pushData[i++] = Read255Short( p );		/* Data */
		}
	}
	assert( i == limit );									/* i 		== t->pushCount */
	t->remainingCodeSize	= limit = Read255UShort( p );	/* limit	== t->remainingCodeSize */

	MTX_mem_free( t->mem,t->code); t->code = NULL;
	t->code = out = (BYTE *)MTX_mem_malloc( t->mem, sizeof( BYTE ) * (i*3 + limit + 16) );
	if (!t->code && t->mem->errorCode)
		return (t->mem->errorCode);
	
	OptimizingPush( (CHAR **)&out, pushData, i ); /* Generate TrueType to set the intial stack */
	assert( out - t->code < i*2 + 16  );
	
	/* limit == t->remainingCodeSize */
	memcpy( out, *p, limit );
	*p += limit;
	out += limit;
	t->numberOfInstructions = (USHORT)(out - t->code);
	assert( t->numberOfInstructions < t->pushCount*3 + limit + 16  );
	return 0;	/* success */
}

/* Reads a glyph in the ctf format at the location pointed to by <p>. */
/* Returns the number of bytes read */
static LONG ReadCTFSpline( register Glyph *t, CHAR *sp )
{
	register LONG i, numberOfPoints;
	SHORT stmp, lastEndPoint;
	SHORT x, y;
	BYTE bitflags;
	CHAR *spStart = sp;
	int status;
	int setBBox = true; /* Initialize here to avoid compiler warnings */
	
	MTX_mem_free( t->mem,t->startPoint);	t->startPoint	= NULL;
	MTX_mem_free( t->mem,t->onCurve);		t->onCurve		= NULL;
	MTX_mem_free( t->mem,t->x);				t->x			= NULL;
	
	t->numberOfContours = READWORD_INC( &sp );
	if ( t->numberOfContours < 0 ) {
		t->xmin = READWORD_INC( &sp );
		t->ymin = READWORD_INC( &sp );
		t->xmax = READWORD_INC( &sp );
		t->ymax = READWORD_INC( &sp );
	} else {
		if ( t->numberOfContours == 0x7fff) { /* flag to read bbox */
			t->numberOfContours = READWORD_INC( &sp );
			t->xmin = READWORD_INC( &sp );
			t->ymin = READWORD_INC( &sp );
			t->xmax = READWORD_INC( &sp );
			t->ymax = READWORD_INC( &sp );
			setBBox = false;
		}
	}
	assert( t->numberOfContours != 0 );
	
	if ( t->numberOfContours < 0  ) {
 		USHORT flags, glyphIndex;
 		int weHaveInstructions;
 		
 		t->componentVersionNumber = t->numberOfContours;
 		assert( t->componentData != NULL );
 		do {
 			flags = READWORD_INC( &sp );
 			weHaveInstructions = flags & WE_HAVE_INSTRUCTIONS;
 			t->componentData[ t->componentSize++] = flags;
 			
 			t->componentData[ t->componentSize++] = glyphIndex = READWORD_INC( &sp );
 			if ( flags & ARG_1_AND_2_ARE_WORDS ) {
 				/* arg 1 and 2 */
	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
 			} else {
 				/* arg 1 and 2 as bytes */
 	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
 			} 
 			
 			if ( flags & WE_HAVE_A_SCALE ) {
 				/* scale */
  	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
 			} else if ( flags & WE_HAVE_AN_X_AND_Y_SCALE ) {
 				/* xscale, yscale */
   	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
  	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
 			} else if ( flags & WE_HAVE_A_TWO_BY_TWO ) {
 				/* xscale, scale01, scale10, yscale */
   	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
  	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
   	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
  	 			t->componentData[ t->componentSize++] = READWORD_INC( &sp );
 			}
 			assert( t->componentSize < t->maxComponentSize );
 		} while ( flags & MORE_COMPONENTS );
 		t->numberOfInstructions = 0;
 		if ( weHaveInstructions ) {
 			status = ReadAllCTFInstructions( t, &sp );
			if (status)
				return -1;
 		}
	} else {
		t->startPoint = (USHORT *)MTX_mem_malloc( t->mem, sizeof( USHORT ) * 2 * t->numberOfContours );
		if (!t->startPoint && t->mem->errorCode)
			return -1;
		t->endPoint   = &t->startPoint[t->numberOfContours];

		/* Non Composite */
		lastEndPoint = stmp = 0;
		for ( i = 0; i < t->numberOfContours; i++ ) {
		    t->startPoint[i]	= stmp;
		    t->endPoint[i]		= (USHORT)(Read255UShort( &sp ) + lastEndPoint);
		 	assert( t->endPoint[i] >= 0  );

		 	lastEndPoint = t->endPoint[i];
		 	stmp = (SHORT)(lastEndPoint + 1);
		}
		numberOfPoints = lastEndPoint + 1;
		t->numberOfPoints = (SHORT)numberOfPoints;
		assert( numberOfPoints > 0 );
		t->onCurve = (BYTE *)MTX_mem_malloc( t->mem, sizeof( BYTE ) * numberOfPoints );
		if (!t->onCurve && t->mem->errorCode)
			return -1;

	    for ( i = 0; i < numberOfPoints; ) {
	        t->onCurve[i++] = *sp++;
	    }
		t->x = (SHORT *)MTX_mem_malloc( t->mem, sizeof( SHORT ) * 2 * numberOfPoints );
		if (!t->x && t->mem->errorCode)
			return -1;
		t->y = &t->x[numberOfPoints];

		x = y = 0;
		for ( stmp = 0, i = 0; i < numberOfPoints; i++ ) {
			register SHORT xBits, yBits;
			USHORT index;
			LONG dx, dy;
			ULONG data, ultmp;
			CHAR xIsNegative, yIsNegative;
			
			bitflags = t->onCurve[i];
			
			t->onCurve[i] = (BYTE)(bitflags & 0x80 ? 0 : ONCURVE);
			index = (USHORT)(bitflags & (128-1));
			
			xIsNegative = coordEncoding[index].xIsNegative;
			yIsNegative = coordEncoding[index].yIsNegative;
			
			ultmp = *(BYTE *)sp;
			ultmp <<= 24;
			data = ultmp;
			ultmp = *((BYTE *)sp+1);
			ultmp <<= 16;
			data |= ultmp;
			ultmp = *((BYTE *)sp+2);
			ultmp <<= 8;
			data |= ultmp;
			ultmp = *((BYTE *)sp+3);
			data |= ultmp;
			sp += coordEncoding[index].byteCount-1; /* Since 1 byte is for the flags */
			
			xBits = coordEncoding[index].xBits;
			ultmp = data >> (32 - xBits);
			ultmp &= ((1L << xBits ) - 1);
			dx = ultmp; assert( dx >= 0 );
			yBits = coordEncoding[index].yBits;
			ultmp = data >> (32 - xBits - yBits);
			ultmp &= ((1L << yBits ) - 1);
			dy = ultmp; assert( dy >= 0 );
			dx += coordEncoding[index].dxPlus; assert( dx >= 0 );
			dy += coordEncoding[index].dyPlus; assert( dy >= 0 );
			if ( xIsNegative ) dx = -dx;
			if ( yIsNegative ) dy = -dy;
			x = (SHORT)(x+dx);
			y = (SHORT)(y+dy);

			t->x[i] = x;
			t->y[i] = y;
		}
		if (setBBox) SetBBox(t); /* Set it, since it is not stored in the CTF format */
 		status = ReadAllCTFInstructions( t, &sp );
		if (status)
			return -1;

	} /* t->numberOfContours >= 0 */
	return sp - spStart; /******/
}

	

/* Writes out the Glyph in ctf format at the location pointed to by <p>. */
/* Returns the number of bytes written. */
static LONG MTX_Glyph_WriteCTFSpline( Glyph *t, CHAR * *destSfnt, LONG pos, LONG *maxOutSize )
{
	register SHORT i, j, k;
	register SHORT x, y;
	BYTE bitFlags;
	LONG bitFlagOffset;
	LONG deltaFromStart;
	/* Harmless pointer */
	register BYTE *f;
	/* Pointers that depend on destSfnt */
	CHAR *pStart, *p;
	
	deltaFromStart = 0;
	pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart;
	
	if ( t->numberOfContours == 0 ) {
		deltaFromStart = p - pStart;
		if (AllocateGlyphSpace( t, destSfnt, pos, p-pStart+2, maxOutSize ))
			return -1;
		pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
		return p - pStart; /******/
	}
	
	if ( t->componentSize > 0 ) {
		assert( t->componentVersionNumber < 0 );
		deltaFromStart = p - pStart;
		if (AllocateGlyphSpace( t, destSfnt, pos, p-pStart+5*2, maxOutSize ))
			return -1;
		pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
		WRITEWORD_INC( &p, t->componentVersionNumber );
		/* Bounding box */
		WRITEWORD_INC( &p, t->xmin );
		WRITEWORD_INC( &p, t->ymin );
		WRITEWORD_INC( &p, t->xmax );
		WRITEWORD_INC( &p, t->ymax );
	} else {
 		SHORT xmax, xmin, ymax, ymin;
  		SHORT xmax2, xmin2, ymax2, ymin2;

		MTX_Glyph_GetBBox( t, &xmax, &xmin, &ymax, &ymin);
		MTX_Glyph_ComputeBBox( t, &xmax2, &xmin2, &ymax2, &ymin2);
		if ((xmin != xmin2) | (ymin != ymin2) | (xmax != xmax2) | (ymax != ymax2)){
			deltaFromStart = p - pStart;
			if (AllocateGlyphSpace( t, destSfnt, pos, p-pStart+6*2, maxOutSize ))
				return -1;
			pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
			WRITEWORD_INC( &p, 0x7fff );	   /* flag indicating BBox follows */
			WRITEWORD_INC( &p, t->numberOfContours );
 			/* Bounding box */
			WRITEWORD_INC( &p, xmin );
			WRITEWORD_INC( &p, ymin );
			WRITEWORD_INC( &p, xmax );
			WRITEWORD_INC( &p, ymax );
		} else {
			deltaFromStart = p - pStart;
			if (AllocateGlyphSpace( t, destSfnt, pos, p-pStart+2, maxOutSize ))
				return -1;
			pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
			WRITEWORD_INC( &p, t->numberOfContours );
		}
	}
	
	if ( t->componentSize > 0 ) {
		deltaFromStart = p - pStart;
		if (AllocateGlyphSpace( t, destSfnt, pos, p-pStart+t->componentSize*2, maxOutSize ))
			return -1;
		pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
		for ( i = 0; i < t->componentSize; i++ ) {
			WRITEWORD_INC( &p, t->componentData[i] );
		}
		if ( t->numberOfInstructions > 0 ) {
			deltaFromStart = p - pStart;
			if (AllocateGlyphSpace( t, destSfnt, pos, p-pStart + 3 + t->pushCount*3, maxOutSize ))
				return -1; /* Assume worst case */
			pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
			PushInitialCTFStack( t, &p );  /* goes to ctf1 again for random access  */
			
			deltaFromStart = p - pStart;
			if (AllocateGlyphSpace( t, destSfnt, pos, p-pStart + 3 + t->remainingCodeSize, maxOutSize ))
				return -1;
			pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
			WriteRemainingCTFCode( t, &p ); /* goes to ctf1 again for random access  */
		}
		return ( p - pStart ); /******/
	}
	f = (BYTE *)MTX_mem_malloc( t->mem, sizeof( BYTE ) * t->numberOfPoints );
	if (!f && t->mem->errorCode)
		return -1;

	deltaFromStart = p - pStart;
	if (AllocateGlyphSpace( t, destSfnt, pos, p-pStart+t->numberOfContours*3, maxOutSize ))
		return -1; /* At least big enough */
	pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */

	for ( i = j = 0; i < t->numberOfContours; i++ ) {
		k = (SHORT)(t->endPoint[i] - j);
		Write255UShort( &p, k );
		j = t->endPoint[i];
	}
	
	bitFlagOffset = p - pStart;
	p += t->numberOfPoints; /* skip the bitFlags */
	deltaFromStart = p - pStart;
	if (AllocateGlyphSpace( t, destSfnt, pos, p-pStart, maxOutSize ))
		return -1;
	pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
	x = y = 0;
	/* Calculate flags */
	for ( x = y = i = 0; i < t->numberOfPoints; i++ ) {
		CHAR xIsNegative, yIsNegative;
		LONG dx, dy;
		USHORT index, bCount;
		USHORT x0, x1, y0, y1;
		ULONG data, coord, shift;
		
		bitFlags = (BYTE)(t->onCurve[i] & ONCURVE ?  0 : 0x80); /* This comresses better ! */
		dx = t->x[i] - x; x = t->x[i];
		dy = t->y[i] - y; y = t->y[i];

		xIsNegative = false;
		if ( dx < 0 ) {
			xIsNegative = true;
			dx = -dx;
		}
		yIsNegative = false;
		if ( dy < 0 ) {
			yIsNegative = true;
			dy = -dy;
		}
		
		/* Pick the lowest index that works */
		for ( index = 0; index < 128; index++ ) {
			if ( coordEncoding[index].xBits != 0 && coordEncoding[index].xIsNegative != xIsNegative ) continue; /******/
			if ( coordEncoding[index].yBits != 0 && coordEncoding[index].yIsNegative != yIsNegative ) continue; /******/
			x0 = coordEncoding[index].dxPlus;
			x1 = (USHORT)(x0 + (1L << coordEncoding[index].xBits ) - 1);
			y0 = coordEncoding[index].dyPlus;
			y1 = (USHORT)(y0 + (1L << coordEncoding[index].yBits ) - 1);
			if ( dx >= x0 && dx <= x1 && dy >= y0 && dy <= y1 ) break; /******/
		}
		assert( index < 128 );
		bitFlags |= index;
		f[i] = bitFlags;
		deltaFromStart = p - pStart;
		if (AllocateGlyphSpace( t, destSfnt, pos, p-pStart+coordEncoding[index].byteCount-1, maxOutSize ))
			return -1;
		pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
		
		dx -= coordEncoding[index].dxPlus; assert( dx >= 0 );
		dy -= coordEncoding[index].dyPlus; assert( dy >= 0 );
		assert( dx < (1L << coordEncoding[index].xBits ) );
		assert( dy < (1L << coordEncoding[index].yBits ) );
		data = 0;
		if ( coordEncoding[index].xBits > 0 ) {
			coord = dx & ((1L << coordEncoding[index].xBits ) - 1);
			data |= coord << (32 - coordEncoding[index].xBits);
		}
		if ( coordEncoding[index].yBits > 0 ) {
			coord = dy & ((1L << coordEncoding[index].yBits ) - 1);
			data |= coord << (32 - coordEncoding[index].xBits - coordEncoding[index].yBits);
		}
		for ( shift = 24, bCount = 1; bCount < coordEncoding[index].byteCount; bCount++ ) {
			assert( shift >= 0 );
			*(BYTE *)p++ = (CHAR)((data >> shift) & 0xff);
			shift -= 8;
		}
	} /* for i */
	
	/* Write out bitFlags, at the old location upstream in the data */
	for ( i = 0; i < t->numberOfPoints; i++) {
		pStart[bitFlagOffset + i] = f[i];
	}

	deltaFromStart = p - pStart;
	if (AllocateGlyphSpace( t, destSfnt, pos, p-pStart+3 + t->pushCount*3, maxOutSize ))
		return -1; /* Assume worst case */
	pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
	PushInitialCTFStack( t, &p );  /* goes to ctf1 again for random access  */
	
	deltaFromStart = p - pStart;
	if (AllocateGlyphSpace( t, destSfnt, pos, p-pStart+3 + t->remainingCodeSize, maxOutSize ))
		return -1;
	pStart = (CHAR *)&(*destSfnt)[pos]; p = pStart + deltaFromStart; /* Reset pointers */
	WriteRemainingCTFCode( t, &p ); /* goes to ctf1 again for random access  */

	MTX_mem_free( t->mem, f );
	return p - pStart; /******/
}


/* The Glyph Constructor */
static Glyph *MTX_Glyph_Create( MTX_MemHandler *mem, CHAR isTTFSpline, CHAR *sp,
										LONG splineLength, SHORT lsb, SHORT aw,
										LONG *bytesRead, CHAR *inDataStart, CHAR *inDataEnd)
{
	Glyph *t	= (Glyph *)MTX_mem_malloc( mem, sizeof( Glyph ) );
	if (!t && mem->errorCode)
		return NULL;

	t->mem		= mem;
	
	t->advanceWidth			= aw;
	t->leftSideBearing		= lsb;
 	t->componentSize 		= 0;
	
	t->numberOfContours		= 0;
	t->numberOfPoints		= 0;
	t->numberOfInstructions	= 0;
	t->startPoint			= NULL;
	t->endPoint				= NULL;
	t->x					= NULL;
	t->y					= NULL;
	t->onCurve				= NULL;
	t->code					= NULL;
	t->componentData		= NULL;
	t->pushData				= NULL;
	t->remainingCode		= NULL;
	
	t->inDataStart			= inDataStart;
	t->inDataEnd			= inDataEnd;
	
 		
	t->maxComponentSize = 1024;
	t->componentData = (SHORT *)MTX_mem_malloc( t->mem, sizeof( SHORT ) * t->maxComponentSize );
	if (!t->componentData && t->mem->errorCode)
		return NULL;
 		
	if ( isTTFSpline ) {
		if ( splineLength ) {
				if ( sp ) {
					*bytesRead = ReadTTFSpline( t, sp );
					if (*bytesRead == -1)
						return NULL;
					if (ParseCode( t ))
						return NULL;
				} else {
					*bytesRead = 0;
					t->numberOfContours = 0;
				}
		} else { 
			*bytesRead = 0;
		}
	} else {
		if ( splineLength ) {
			*bytesRead = ReadCTFSpline( t, sp );
			if (*bytesRead == -1)
				return NULL;
		} else { 
			*bytesRead = 0;
			t->numberOfContours = 0;
		}
	}
	return t; /******/
}


/* Destructor */
static void MTX_Glyph_Destroy( Glyph *t )
{
	/* It is always ok to do delete on NULL */
	MTX_mem_free( t->mem, t->startPoint);
	/* MTX_mem_free( t->mem, t->endPoint); t->endPoint points into t->startPoint */
	MTX_mem_free( t->mem, t->x);
	/* MTX_mem_free( t->mem, t->y); t->y points into t->x */
	MTX_mem_free( t->mem, t->onCurve);
	MTX_mem_free( t->mem, t->code);
	MTX_mem_free( t->mem, t->componentData);
	MTX_mem_free( t->mem, t->pushData);
	MTX_mem_free( t->mem, t->remainingCode);
	MTX_mem_free( t->mem, t );
}

/* Returns true if they are equal. */
int MTX_TTF_GlyphsAreEqual( MTX_MemHandler *mem, CHAR *spline1, LONG len1, CHAR *spline2, LONG len2 )
{
	int result;
	Glyph *g1, *g2;
	LONG bytesRead1, bytesRead2;
	
	if ( len1 == 0 && len2 == 0 ) {
		result = true;
	} else {
		/* if either glyph couldn't be created, the compare will be FALSE */
		g1 = MTX_Glyph_Create( mem, true, spline1, len1, 0, 0, &bytesRead1, spline1,
						spline1+len1);
		if (!g1)
			return false;
		g2 = MTX_Glyph_Create( mem, true, spline2, len2, 0, 0, &bytesRead2, spline2,
																	spline2+len2);
		if (!g2)
			return false;
		result = GlyphsEqual( g1, g2 );
		
		MTX_Glyph_Destroy( g1 );
		MTX_Glyph_Destroy( g2 );
	}
	return result; /*****/
}


/* Returns the offset for the table with the tag <tag> */
static LONG GetTableOffset( CHAR *sfnt, sfnt_TableTag tag )
{
	register LONG offset, i, numOffsets;
	
	assert( sfnt != NULL );
	
	offset = 0;
	numOffsets = SWAPENDS_U16( GET_sfnt_numOffsets( sfnt ) );
	for ( i = 0; i < numOffsets; i++ ) {
		if ( SWAPENDS_32( GET_sfnt_table_tag( sfnt, i ) ) == tag ) {
			offset = SWAPENDS_U32( GET_sfnt_table_offset( sfnt, i ) );
			break; /******/
		}
	}
	return offset; /******/
}


/* Returns a pointer to the table with the tag <tag> */
CHAR *GetTablePtr( CHAR *sfnt, int fontOff, sfnt_TableTag tag )
{
	register LONG offset;

	assert( sfnt != NULL );
	offset = GetTableOffset( sfnt+fontOff, tag );
	return (sfnt + offset); /******/
}


/* Returns the length of the table with the tag <tag> */
/* returns zero it the table does not exist */
static LONG GetTableLength( CHAR *sfnt, sfnt_TableTag tag )
{
	register LONG length, i, numOffsets;
	
	assert( sfnt != NULL );
	length = 0;
	numOffsets = SWAPENDS_U16( GET_sfnt_numOffsets( sfnt ) );
	for ( i = 0; i < numOffsets; i++ ) {
		if ( SWAPENDS_32( GET_sfnt_table_tag( sfnt, i ) ) == tag ) {
			length = SWAPENDS_U32( GET_sfnt_table_length( sfnt, i ) );
			break; /******/
		}
	}
	return length; /******/
}


/* Returns number of glyphs in the sfnt file */
static LONG GetNumberOfGlyphs( CHAR *sfnt, int fontOff )
{
	LONG	numEntries;
	CHAR	*profile;
	
	profile = GetTablePtr( sfnt, fontOff, tag_MaxProfile );
	/* We do not need to check for out of bounds read since we have already
	  checked that the maxp table is big enough and inside the input data */
	numEntries = SWAPENDS_U16( GET_maxp_numGlyphs( profile ) );
	return numEntries; /*****/
}


/* Returns a pointer to the glyph with glyp index <glyphIndex> */
static CHAR *GetSpline( CHAR *sfnt, int fontOff, USHORT glyphIndex, LONG *length )
{
    ULONG		*longIndexToLoc;
    USHORT		*shortIndexToLoc;
    CHAR				*glyphData;
    LONG				glyphDataOffset, glyphDataLength;
	CHAR			 	*head = GetTablePtr( sfnt, fontOff, tag_FontHeader );
    
    longIndexToLoc  = (ULONG *)GetTablePtr( sfnt, fontOff, tag_IndexToLoc );
	if ( SWAPENDS_16( GET_head_indexToLocFormat( head ) ) == SHORT_INDEX_TO_LOC_FORMAT ) {
		shortIndexToLoc = (USHORT *)longIndexToLoc;
		glyphDataOffset = SWAPENDS_U16(shortIndexToLoc[glyphIndex]);
		glyphDataLength = SWAPENDS_U16(shortIndexToLoc[glyphIndex + 1]) - glyphDataOffset;
		glyphDataOffset *= 2L;
		glyphDataLength *= 2L;
	} else {
		glyphDataOffset = SWAPENDS_U32(longIndexToLoc[glyphIndex]);
		glyphDataLength = SWAPENDS_U32(longIndexToLoc[glyphIndex+1]) - glyphDataOffset;
	}
	glyphData = GetTablePtr( sfnt, fontOff, tag_GlyphData );
	if ( length != NULL ) {
		*length = glyphDataLength;
	}
	
    /* sfnt_GetMetrics( sfntH, glyphIndex, &wStuff->glyph->advanceWidth, &wStuff->glyph->leftSideBearing ); */
 	return glyphData + glyphDataOffset; /*****/
}

static void AllocateTTFSpace( TTF_Converter *t, CHAR * *destSfnt, LONG pos, LONG byteCount, LONG *maxOutSize )
{
	CHAR * newDestSfnt;
	byteCount += 4; /* Always keep a 4 byte margin to allow for the zero padding */
	if ( pos + byteCount > *maxOutSize ) {
		*maxOutSize = pos + byteCount + (*maxOutSize>>1) + 2; /* Allocate memory in exponentially increasing steps */
		newDestSfnt = (CHAR *)MTX_mem_realloc( t->mem, *destSfnt, *maxOutSize );
		if ( *destSfnt == t->ttf ) {
			t->ttf		= newDestSfnt;
			t->ttfSize	= *maxOutSize;
		} else if ( *destSfnt == (CHAR *)t->ctfData.ctf1 ) {
			t->ctfData.ctf1		= (BYTE *)newDestSfnt;
			t->ctfData.ctfSize1	= *maxOutSize;
		} else {
			assert( false );
		}
		*destSfnt = newDestSfnt;
	}
}


/* Writes out the 'loca' table appropriately using SHORT or LONG formats */
/* It saves the format picked into t->indexToLocFormat so */
/* that it can be later set correctly in the 'head' table */
static LONG Write_loca_Table( TTF_Converter *t, LONG posTable[], LONG numEntries, CHAR * *destSfnt, LONG pos, LONG *maxOutSize )
{
	CHAR *start, *p; /* Only assign these after AllocateSpace has been invoked */
	LONG i, maxVal;
	
	maxVal = posTable[numEntries-1];
	for ( i = numEntries-2; i >= 0; i-- ) {
		if ( posTable[i] > maxVal ) {
			maxVal = posTable[i];
		}
	}
	if ( maxVal <= 0x0000ffff * 2L ) {
		AllocateTTFSpace( t, destSfnt, pos, numEntries * 2, maxOutSize );
		start = p = &(*destSfnt)[pos];
		/* Do SHORT format */
		for ( i = 0; i < numEntries; i++ ) {
			USHORT value;
			assert( ( posTable[i] & 1 ) == 0 ); /* make sure it is word aligned */
			value = (USHORT)(posTable[i] / 2);
			WRITEWORD_INC( &p, value );
		}
		t->indexToLocFormat = 0;
		assert( p - start == numEntries * 2 );
	} else {
		AllocateTTFSpace( t, destSfnt, pos, numEntries * 4, maxOutSize );
		start = p = &(*destSfnt)[pos];
		/* Do LONG format */
		for ( i = 0; i < numEntries; i++ ) {
			LONG value;
			assert( ( posTable[i] & 1 ) == 0 ); /* make sure it is word aligned */
			value = posTable[i];
			WRITELONG_INC( &p, value );
		}
		t->indexToLocFormat = 1;
		assert( p - start == numEntries * 4 );
	}
	return p - start; /*****/
}


/* Converts the glyph table both ways between ttf and ctf formats */
static LONG ConvertGlyphTable( TTF_Converter *t, int TTF_To_CTF, CHAR *inFont, int fontOff, LONG posTable[], CHAR * *destSfnt, LONG pos1, LONG *maxOutSize )
{
	CHAR *spline, *glyphDataIn;
	USHORT glyphIndex;
	Glyph *glyph;
	LONG numEntries, thisLength, oldLength;
	LONG startPos, outPos, bytesRead, totalBytesRead;

	
	numEntries 		= GetNumberOfGlyphs( inFont, fontOff );
	glyphDataIn 	= GetTablePtr( inFont, fontOff, tag_GlyphData );
	startPos 		= pos1;
	outPos   		= pos1;
	totalBytesRead 	= 0;
	for ( glyphIndex = 0; glyphIndex < numEntries; glyphIndex++ ) {
		posTable[ glyphIndex ] = outPos - startPos;
		
		spline = GetSpline( inFont, fontOff, glyphIndex, &oldLength );
		CheckReadBoundsForTable( t, spline, oldLength ); /* >= tableStart && < tableEnd */
		
		assert( spline != NULL );
		assert( spline >= inFont );
		glyph = MTX_Glyph_Create( t->mem, (CHAR)TTF_To_CTF, spline,	oldLength, 0 ,0,
											&bytesRead , spline, spline+oldLength);
		if (!glyph)
			return -1;
		thisLength		 = 0;
		totalBytesRead	+= bytesRead;
		if ( bytesRead > 2 ) {
			thisLength = MTX_Glyph_WriteTTFSpline( glyph, destSfnt, outPos, maxOutSize ); /* Write to outPos, use as tmp memory */
			if (thisLength == -1)
				return -1;
			/* We need to do this here since the glyph object can not access these fields */
			if ( TTF_To_CTF ) {
				t->ctfData.ctf1 	= (BYTE *)*destSfnt;
				t->ctfData.ctfSize1	= *maxOutSize;
			} else if ( (*destSfnt) == (CHAR *)t->ctfData.ctf1 ) {
				t->ttf		= *destSfnt;
				t->ttfSize	= *maxOutSize;
			}
		}

		if ( TTF_To_CTF ) {
			/* UNDO and go to CTF intstead */
			thisLength = MTX_Glyph_WriteCTFSpline( glyph, destSfnt, outPos, maxOutSize );
			if (thisLength == -1)
				return -1;
			/* We need to do this here since the glyph object can not access these fields */
			t->ctfData.ctf1 	= (BYTE *)*destSfnt;
			t->ctfData.ctfSize1	= *maxOutSize;
			outPos += thisLength;
			if ( outPos & 1 ) { /* Odd offset */
				(*destSfnt)[outPos++] = 0; /* Zero Pad */
			}
		} else {
			outPos += thisLength;
			if ( outPos & 1 ) { /* Odd offset */
				(*destSfnt)[outPos++] = 0; /* Zero Pad */
			}
		}
		assert( outPos <= *maxOutSize );
		MTX_Glyph_Destroy( glyph );
	}
	posTable[ glyphIndex ] = outPos - startPos;
	assert( totalBytesRead > 0 );

	return outPos - startPos; /*****/
}




static ULONG CalcTableCheckSum( ULONG *table, ULONG length )
{
	ULONG sum = 0L;
	ULONG *endPtr = table + ((length+3) & ~3) / sizeof( ULONG );
	
	while ( table < endPtr ) {
		sum += SWAPENDS_U32(*table++);
	}
	return sum; /******/
}

static ULONG CalcFontCheckSum( ULONG *sfnt, ULONG length )
{
	ULONG sum = 0L;
	ULONG *endPtr = sfnt + ((length+3) & ~3) / sizeof( ULONG );
	
	while ( sfnt < endPtr ) {
		sum += SWAPENDS_U32(*sfnt++);
	}
	return 0xB1B0AFBA - sum; /******/
}

static int SanityCheck( TTF_Converter *t, int TTF_To_CTF, int fontOff0 )
{
	CHAR *sfnt0;
	int inSize, i;
	CHAR *head;

	if ( TTF_To_CTF ) 
	{
		sfnt0 = t->ttf;
		inSize  = t->ttfSize;
	} 
	else 
	{
		sfnt0 = (CHAR *)t->ctfData.ctf1;
		inSize = t->ctfData.ctfSize1;
	}

	/* Do some sanity checking on the data */
	if ( SWAPENDS_U16(GET_sfnt_numOffsets( sfnt0 + fontOff0)) * (ULONG)DirectoryEntrySize >= 
		(ULONG)inSize/2 )
	{
		return ERR_Convert_FONT_too_many_tables;
	}
	
	for ( i = 0; i < SWAPENDS_U16(GET_sfnt_numOffsets( sfnt0 + fontOff0 )); i++ ) 
	{
		unsigned int length = SWAPENDS_U32(GET_sfnt_table_length( sfnt0 + fontOff0, i ));
		if ( length != 0 ) 
		{
			unsigned int offset = SWAPENDS_U32(GET_sfnt_table_offset( sfnt0 + fontOff0, i ));
			if ( offset >= (ULONG)inSize )
			{ 
				return ERR_Convert_FONT_table_outside;
			}

			if ( offset + length > (ULONG)inSize )
			{ 
				return ERR_Convert_FONT_table_end_outside;
			}
		}
	}
	/* Now we know all font tables are inside the font. This means that we do not
	   need to check the result of GetTablePtr, since we know that all tables
	   are inside the font */
	/* tag_FontHeader 'head' table is 54 bytes LONG */
	if ( GetTableLength( sfnt0 + fontOff0, tag_FontHeader ) < 54 )
	{
		return ERR_head_table_gone_or_to_short;
	}

	head = GetTablePtr( sfnt0, fontOff0, tag_FontHeader );
	if ( SWAPENDS_U32(GET_head_magicNumber( head )) != 0x5F0F3CF5 )
	{
		return ERR_Convert_FONT_bad_magic;
	}
	if ( SWAPENDS_16(GET_head_glyphDataFormat( head )) != 0 ) /* VL */
		return ERR_Convert_FONT_unknown_glyphDataFormat;

	/* tag_HoriHeader 'hhea' table is 36 bytes LONG */
	if ( GetTableLength( sfnt0 + fontOff0, tag_HoriHeader ) < 36 )
	{
		return ERR_hhea_table_gone_or_to_short;
	}
	/* tag_MaxProfile 'maxp' table is 32 bytes LONG */
	if ( GetTableLength( sfnt0 + fontOff0, tag_MaxProfile ) < 32 )
	{
		return ERR_maxp_table_gone_or_to_short;
	}
	/* if we have a 'glyf' table then we need a 'loca' table */
	if ( TTF_To_CTF && GetTableLength( sfnt0 + fontOff0, tag_GlyphData ) > 0 && 
		GetTableLength( sfnt0 + fontOff0, tag_IndexToLoc ) == 0 )
	{
		return ERR_loca_table_gone_or_to_short;
	}

	/* successful check */
	return 0;
}

/* If table is repeated in source font, than make the directory entry in the destination
   font that repeats it and return "true". Else, return false */
int CheckForRepeatedTable( CHAR *sfnt0,     /* source font */
                           CHAR *sfnt1,     /* destination font */
                           int matchIndex,  /* index of table to check */
                           int matchTTC     /* index in ttc of font containing above table */
                         )
{
    sfnt_TableTag tag; 
    USHORT numTables; /* mumber of TrueType tables in current font */
    ULONG matchOffset;
    int fontOff0, fontOff1, fontOffCur0, fontOffCur1;
    int idxTTC;
    int isTTC;
    int i;

    /* check for TrueTypeCollection, set flag & #-fonts-in-file if so */
    tag = *((sfnt_TableTag *)((CHAR *)sfnt0 + TTC_TAG));
    if ( SWAPENDS_32(tag) == tag_TTCollection ) 
    {
        int numInTTC;
        numInTTC   = SWAPENDS_32(*((int *)((CHAR *)sfnt0 + TTC_NUMFONTS)));
        fontOffCur0 = SWAPENDS_32( GET_ttc_font_offset( sfnt0, matchTTC ) );
        fontOffCur1 = SWAPENDS_32( GET_ttc_font_offset( sfnt1, matchTTC ) );
        isTTC = 1;
        if(matchTTC>=numInTTC)
        {
            printf("Error: matchTTC>=numInTTC");
            return 1;
        }
    }
    else
    {
        fontOff0 = fontOff1 = fontOffCur0 = fontOffCur1 = 0;
        isTTC = 0;
    }
    matchOffset = (GET_sfnt_table_offset( sfnt0 + fontOffCur0, matchIndex ));
    for (idxTTC=0; idxTTC <= matchTTC; idxTTC++)
    {
        if (isTTC)
        {
            fontOff0 = SWAPENDS_32( GET_ttc_font_offset( sfnt0, idxTTC ) );
            fontOff1 = SWAPENDS_32( GET_ttc_font_offset( sfnt1, idxTTC ) );
        }
        
        /* Loop runs once for each table in the current font */
        numTables = SWAPENDS_U16(GET_sfnt_numOffsets( sfnt0 + fontOff0 ));
        for ( i = 0; (idxTTC  < matchTTC && i < numTables) ||
                     (idxTTC == matchTTC && i < matchIndex); i++ )
        {
            if ( GET_sfnt_table_length( sfnt0 + fontOff0, i ) == 0 ) 
                continue; /******/
            if ( GET_sfnt_table_offset( sfnt0 + fontOff0, i ) == matchOffset )
            {
                /* If there is a match just copy these 3 fields, and then skip the table */
                SET_sfnt_table_checkSum( sfnt1 + fontOffCur1, matchIndex, GET_sfnt_table_checkSum( sfnt1 + fontOff1, i ) );
                SET_sfnt_table_offset( sfnt1 + fontOffCur1, matchIndex, GET_sfnt_table_offset( sfnt1 + fontOff1, i ) );
                tag = /*SWAPENDS_32*/(GET_sfnt_table_tag( sfnt1 + fontOffCur1, matchIndex ));
                printf("=== REPEAT! setting table #%d (tag=%.4s) offset to %d\n",
                    i, (CHAR *)&tag, SWAPENDS_32(GET_sfnt_table_offset( sfnt1 + fontOff1, i )));
                SET_sfnt_table_length( sfnt1 + fontOffCur1, matchIndex, GET_sfnt_table_length( sfnt1 + fontOff1, i ) );
                return (true); /******/
            }
        }	/*** END for-tables-in-font ***/
    } /*** END for-fonts-in-TTC ***/
    return 0;
}

/* Insert a table entry in the font directory. It is assumed that the font
 * directory has already been expanded so that there is room for one more
 * entry at the end of the directory. The number of offsets for the table already
 * reflects this larger size.
 */
void insertTable(CHAR *sfnt, sfnt_TableTag tag, int offset, int length)
{
    USHORT numTables, i;
	sfnt_TableTag t;
    CHAR *tableStart;

	/* Find the table entry after which to store the new entry */
	numTables = SWAPENDS_U16(GET_sfnt_numOffsets( sfnt )) - 1;
	for ( i = 0; i < numTables; i++ )
	{
		t = SWAPENDS_32(GET_sfnt_table_tag( sfnt, i));
		if(t > tag)
			break;
	}
	tableStart = sfnt + OFFSETTABLESIZE + DirectoryEntrySize * i;
	/*  Move tail of directory one slot to the back. */
	memmove(tableStart+DirectoryEntrySize, tableStart, DirectoryEntrySize*(numTables - i));
	/*  Set the Directory entry i */
	*(sfnt_TableTag *)(tableStart)= SWAPENDS_32(tag);
	SET_sfnt_table_offset( sfnt, i, SWAPENDS_32(offset) );
	SET_sfnt_table_length( sfnt, i, SWAPENDS_32(length) );
}


/* The method to do the big conversion between TTF and CTF */
static int Convert_FONT( TTF_Converter *t, int TTF_To_CTF  )
{
	unsigned int tableLength;
	int fontOff0=0, fontOff1=0;
	int numInTTC=1, idxTTC;
	sfnt_TableTag tag; 
	int isTTC = 0;
	int numTables, status;

	int loca_pos, loca_len, loca_ck;
	unsigned int csumTemp, fsumTemp;
	CHAR *head;

	int i, oldPos, headerSize;
	int pos=0;
	const LONG zero = 0L;
	int maxOutSize;
	/* Harmless pointers */
	int *posTable = 0;
	CHAR *sfnt0;
	/* Pointer  which may be reallocated */
	CHAR *sfnt1;
	
	if ( TTF_To_CTF ) 
	{
		sfnt0 = t->ttf;
		sfnt1 = (CHAR *)t->ctfData.ctf1;
		maxOutSize = t->ctfData.ctfSize1;

		/* check for TrueTypeCollection, set flag & #-fonts-in-file if so */
		tag = *((sfnt_TableTag *)((CHAR *)sfnt0 + TTC_TAG));

		if ( SWAPENDS_32(tag) == tag_TTCollection ) 
		{
			numInTTC = SWAPENDS_32(*((int *)((CHAR *)sfnt0 + TTC_NUMFONTS)));
			isTTC = 1;

			/* copy TTC header to output file now - repair offsets later */
			headerSize = TTC_OFFSET + (numInTTC * 4);
			AllocateTTFSpace( t, &sfnt1, pos, headerSize, &maxOutSize );
			/* printf("+++ writing TT Collection header: # fonts = %d, old/new pos=%d,%d\n", 
				numInTTC, pos, pos + headerSize); /* */
			pos += headerSize;
			memcpy( (CHAR *)(sfnt1), (CHAR *)sfnt0, headerSize );
		}
	} 
	else	/* CTF_To_TTF */ 
	{
		sfnt0 = (CHAR *)t->ctfData.ctf1; 
		sfnt1 = t->ttf;
		maxOutSize = t->ttfSize;
	}
	
	for (idxTTC=0; idxTTC <	numInTTC; idxTTC++)
	{
		if (isTTC)
		{
			fontOff0 = SWAPENDS_32( GET_ttc_font_offset( sfnt0, idxTTC ) );

			/* fix up offset in TTC header in output file */
			SET_ttc_font_offset( sfnt1, idxTTC, SWAPENDS_32(pos) );
			fontOff1 = pos;
			printf("+++ processing font # %d: offset(in)=%d, (out)=%d\n", idxTTC, fontOff0, fontOff1);
		}

		/* validate TableDirectory and a few tables for current font */
		status = SanityCheck(t, TTF_To_CTF, fontOff0);
		if (status)
		{
			printf("===> ERROR RETURNED from SanityCheck=%d <+++\n", status);
			t->mem->errorCode = status;
			return status;
		}
	
		/* if TableDirectory (header) was valid, write it to output file */
		headerSize = OFFSETTABLESIZE + (SWAPENDS_U16(GET_sfnt_numOffsets( sfnt0 + fontOff0 )) * DirectoryEntrySize );
		AllocateTTFSpace( t, &sfnt1, pos, headerSize, &maxOutSize );
		/* printf("=== writing font header: old/new pos=%d,%d\n", pos, pos + headerSize); /* */
		pos += headerSize;
		memcpy( (CHAR *)sfnt1 + fontOff1, (CHAR *)sfnt0 + fontOff0, headerSize );
		/* don't allocate this twice! (possible if processing a TrueTypeCollection) */
		/* ASSUMED: there will only be one GLYF table in a TrueTypeCollection file */
		if (!posTable)
		{
			posTable = (int *)MTX_mem_malloc( t->mem, 
				sizeof( int ) * (GetNumberOfGlyphs( sfnt0, fontOff0) + 1));
			if (!posTable && t->mem->errorCode)
				return (t->mem->errorCode);

			printf("# of glyphs in font=%d\n",GetNumberOfGlyphs( sfnt0, fontOff0 ));
		}

		for ( i = 0; i < SWAPENDS_U16(GET_sfnt_numOffsets( sfnt0 + fontOff0 )); i++ )
		{
			
			tag = SWAPENDS_32(GET_sfnt_table_tag( sfnt0 + fontOff0, i ));
			oldPos = pos;
			
			SET_sfnt_table_length( sfnt1 + fontOff1, i, GET_sfnt_table_length( sfnt0 + fontOff0, i ) );

			/* Tolerate zero length tables */
			if ( TTF_To_CTF ) 
			{
				if ( GET_sfnt_table_length( sfnt1 + fontOff1, i ) == 0 ) 
					continue; /******/
			} 
			else if ( tag != tag_IndexToLoc )
			{
				/* The length of tag_IndexToLoc table in CTF is zero and we need to process it */
				if ( GET_sfnt_table_length( sfnt1 + fontOff1, i ) == 0 ) 
					continue; /******/
			}

			if (CheckForRepeatedTable(sfnt0, sfnt1, i, idxTTC))
				continue; /******/

			/* don't do this till AFTER repeat-check - correct, I think... */
			SET_sfnt_table_offset( sfnt1 + fontOff1, i, SWAPENDS_32(pos) );

			tableLength = SWAPENDS_U32(GET_sfnt_table_length( sfnt0 + fontOff0, i ));
			t->tableStart 	= (CHAR *)&sfnt0[SWAPENDS_U32(GET_sfnt_table_offset( sfnt0 + fontOff0, i ))];
			t->tableEnd		= t->tableStart + tableLength;

			switch( tag ) 
			{			
				case tag_GlyphData:
					printf("found GLYF table at idx %d: length(old)=%d\n",i,tableLength);
					tableLength = ConvertGlyphTable( t, TTF_To_CTF, sfnt0, fontOff0, posTable,
															  &sfnt1, pos, &maxOutSize );
					if (tableLength == -1)
						return (t->mem->errorCode);
					printf("=== writing GLYF table: old/new pos=%d,%d: length(new)=%d\n",
						pos, pos+tableLength, tableLength); /* */
					pos += tableLength;
					SET_sfnt_table_length( sfnt1 + fontOff1, i, SWAPENDS_32(tableLength) );
					/* checksum not repaired - presumably OK, because it's not a TrueType table now */
					break; /******/
				
				case tag_IndexToLoc:
					printf("found LOCA table at idx %d: length=%d\n",i,tableLength);
					assert( GET_sfnt_table_length( sfnt0 + fontOff0, i ) != 0 );
					SET_sfnt_table_offset( sfnt1 + fontOff1, i, 0 );	/* set to zero - invalid now */
					SET_sfnt_table_length( sfnt1 + fontOff1, i, 0 );	/* set to zero - invalid now */
					SET_sfnt_table_checkSum( sfnt1 + fontOff1, i, 0);	/* set to zero - invalid now */
					break; /*****/
				
				default:
                    {
                        int tag1;
                        AllocateTTFSpace( t, &sfnt1, pos, tableLength, &maxOutSize );
                        memcpy( (CHAR *)&sfnt1[pos], 
                            t->tableStart, 
                            tableLength );
                        tag1 = SWAPENDS_32(tag);
                        /* printf("=== writing table 0x%x(%.4s) at idx %d: old/new pos=%d,%d\n",
                            tag, (CHAR *)&tag1, i, pos, pos+tableLength); /* */
                        pos += tableLength;
                        /* length and offset set above; copied checksum presumably still OK */
                        break; /******/
                    }
			}
			memcpy( (CHAR *)&sfnt1[pos], (CHAR *)&zero, 4 );	/* Zero Pad */
			pos += 3; 
			pos &= ~3;	/* Long Word Align */
			
			if ( !TTF_To_CTF ) 
			{
				SET_sfnt_table_checkSum( sfnt1, i, SWAPENDS_U32(CalcTableCheckSum( (ULONG *)&sfnt1[oldPos], SWAPENDS_U32( GET_sfnt_table_length( sfnt1 + fontOff1, i  ) ) ) ) );
			}
			assert( pos < maxOutSize );

		}	/*** END for-tables-in-font ***/

	}	/*** END for-fonts-in-TTC ***/

	if (isTTC)
	{
		fontOff1 = SWAPENDS_32( GET_ttc_font_offset( sfnt1, 0 ) );
		/* printf("++++ New fontOff(0)=%d\n", fontOff1); /* */
	}
	else
	{
		fontOff1 = 0;
	}


	/* ASSUMING that # tables wil be the same for each font in TTC file - RISKY */ 
	numTables =  SWAPENDS_U16( GET_sfnt_numOffsets(sfnt1 + fontOff1));
	/* printf("++++ numTables (output)=%d\n",numTables); /* */
	

	/* Add loca table */
	for ( i = 0; i < numTables; i++ ) 
	{
		if ( SWAPENDS_32( GET_sfnt_table_tag( sfnt1 + fontOff1, i ) ) == tag_IndexToLoc ) 
		{
			loca_pos =  SWAPENDS_32(pos);
			SET_sfnt_table_offset( sfnt1 + fontOff1, i, loca_pos );
			/* printf("+++ setting LOCA table offset to %d\n", SWAPENDS_32(loca_pos)); /* */

			loca_len = SWAPENDS_32(Write_loca_Table( t, posTable, 
				GetNumberOfGlyphs( sfnt0, fontOff0 ) + 1 , &sfnt1, pos, &maxOutSize) );
			SET_sfnt_table_length( sfnt1 + fontOff1, i, loca_len );
			/* printf("+++ setting LOCA table length to %d\n", SWAPENDS_32(loca_len)); /* */

			oldPos = pos;
			pos +=  SWAPENDS_32(loca_len);
			/* printf("=== writing LOCA table: old/new pos=%d,%d\n",oldPos,pos); /* */
			memcpy( (CHAR *)&sfnt1[pos], (CHAR *)&zero, 4 );	/* Zero Pad */
			pos += 3; 
			pos &= ~3;	/* Long Word Align */
			assert( pos < maxOutSize );

			/* now set LOCA checksum */
			/* LATER - is the checksum VALID? */			 
			loca_ck = SWAPENDS_U32(CalcTableCheckSum( (ULONG *)&sfnt1[oldPos], 
						SWAPENDS_U32(GET_sfnt_table_length( sfnt1 + fontOff1, i )) ) );
			SET_sfnt_table_checkSum( sfnt1 + fontOff1, i, loca_ck);
			/**printf("+++ setting LOCA checksum to %d\n", loca_ck);**/
			break;/******/
		}
	}

	for (idxTTC=1; idxTTC <	numInTTC; idxTTC++)
	{
		fontOff1 = SWAPENDS_32( GET_ttc_font_offset( sfnt1, idxTTC ) );
		/* printf("++++ New fontOff(%d)=%d\n", idxTTC, fontOff1); /* */
		numTables =  SWAPENDS_U16( GET_sfnt_numOffsets(sfnt1 + fontOff1));
		/* printf("++++ numTables (output)=%d\n",numTables); /* */
		for ( i = 0; i < numTables; i++ ) 
		{
			if ( SWAPENDS_32( GET_sfnt_table_tag( sfnt1 + fontOff1, i ) ) == tag_IndexToLoc ) 
			{			 
				SET_sfnt_table_offset( sfnt1 + fontOff1, i, loca_pos );
				/* printf("+++ setting LOCA table offset to %d\n", SWAPENDS_32(loca_pos)); /* */

				SET_sfnt_table_length( sfnt1 + fontOff1, i, loca_len );
				/* printf("+++ setting LOCA table length to %d\n", SWAPENDS_32(loca_len)); /* */

				SET_sfnt_table_checkSum( sfnt1 + fontOff1, i, loca_ck);
				/**printf("+++ setting LOCA checksum to %d\n", loca_ck);**/
				break;/******/
			}
		}
	}
	
	t->ctfData.ctfSize1 = pos;

	/* If the "head" table is shared between the individual fonts in the TTC 
		file, then the checksumAdj value will only be correct for the last font;
		if there are individual "head" tables, then the checksumAdj values 
		will be correct. */
	for (idxTTC=0; idxTTC <	numInTTC; idxTTC++)
	{
		if (isTTC)
		{
			fontOff1 = SWAPENDS_32( GET_ttc_font_offset( sfnt1, idxTTC ) );
		}
		else
		{
			fontOff1 = 0;
		}
		/* printf("++++ New fontOff(%d)=%d\n", idxTTC, fontOff1); /* */
		numTables =  SWAPENDS_U16( GET_sfnt_numOffsets(sfnt1 + fontOff1));
		for ( i = 0; i < numTables; i++ ) 
		{
			if ( SWAPENDS_32( GET_sfnt_table_tag( sfnt1 + fontOff1, i ) ) == tag_FontHeader ) 
			{
				head = (CHAR *)&sfnt1[ SWAPENDS_U32(GET_sfnt_table_offset( sfnt1 + fontOff1, i )) ];
				SET_head_indexToLocFormat( head, SWAPENDS_16(t->indexToLocFormat) ); /* Set by Write_loca_Table, and it may have changed, so write it out */
				SET_head_checkSumAdjustment( head, 0 );

				csumTemp = 
					SWAPENDS_U32(CalcTableCheckSum( (ULONG *)head, 
					SWAPENDS_U32(GET_sfnt_table_length( sfnt1 + fontOff1, i )) ) ) ;
				SET_sfnt_table_checkSum( sfnt1 + fontOff1, i, csumTemp );
				/* printf("++++ setting checksum=0x%x for table-index %d\n", csumTemp, i); /* */

				fsumTemp = SWAPENDS_U32( CalcFontCheckSum( (ULONG *)sfnt1, pos ) );
				SET_head_checkSumAdjustment( head, fsumTemp );
				/* printf("++++ setting checksum-adj=0x%x at head=0x%x\n", fsumTemp, head); /* */
			}
		}
	}

	if ( TTF_To_CTF ) 
		t->ctfData.ctfSize1 = pos;
	else 
		t->ttfSize = pos;
	
	MTX_mem_free( t->mem, posTable );

	return 0;	/* success */
}


/* Converts the TTF data to CTF data. */
BYTE *MTX_TTC_TTF_To_CTF( TTF_Converter *t, BYTE *ttfData, LONG ttfSize, LONG *size1 )
{
	t->ttf					= (CHAR *)ttfData;
	t->ttfSize				= ttfSize;
	assert( t->ctfData.ctf1 == NULL );
	
	/* Allocate Memory */
	t->ctfData.ctfSize1 = t->ttfSize; /* Reduced to 100% of ttfSize instead of 150% Oct 2, 1996 */
	/* This just ensures that we do not grab too much memory and it also */
	/* ensures that most fonts will go through this code without requiring the */
	/* memory to be reallocated. If this should be necessary everything is still OK */
	/* since the reallocation is automatic, and we can even set ctfSize to 1, and this */
	/* code will still work ! */
	t->ctfData.ctf1 = (BYTE *)MTX_mem_malloc( t->mem, sizeof( CHAR ) * t->ctfData.ctfSize1 );
	if (!t->ctfData.ctf1 && t->mem->errorCode)
		return NULL;

	if (Convert_FONT( t, true ))
		return NULL;
	
	/* Realloc to free excess memory. */
	*size1	= t->ctfData.ctfSize1;
	t->ctfData.ctf1 = (BYTE *)MTX_mem_realloc( t->mem, t->ctfData.ctf1, *size1 );
	return t->ctfData.ctf1; /*****/
}



/* Converts the CTF data to TTF data. */
BYTE *MTX_TTC_CTF_To_TTF( TTF_Converter *t, BYTE *ctf1, LONG size1, LONG *ttSize)
{
	t->ctfData.ctf1     	= ctf1;
	t->ctfData.ctfSize1 	= size1;
	
	/* DeAllocate Memory */
	if ( t->ttf != NULL ) MTX_mem_free( t->mem, t->ttf );
	t->ttf = NULL;
	
	t->ttfSize = size1 + (size1>>1);
	t->ttf = (CHAR *)MTX_mem_malloc( t->mem, sizeof( CHAR ) * t->ttfSize );
	if (!t->ttf && t->mem->errorCode)
		return NULL;

	if (Convert_FONT( t, false ))
		return NULL;
		
	/* Realloc to free excess memory. */
	*ttSize = t->ttfSize;
	t->ttf = (CHAR *)MTX_mem_realloc( t->mem, t->ttf, t->ttfSize );
	return (BYTE *)t->ttf; /******/
}


/* It asssumes that spline does point at the beginning of glyph data (IsSpline() == true). */
/* Expands a glyph in the 'glyf' table from the CTF format to the native TTF format on the fly. */
CHAR *MTX_TTC_GetTTFGlyphMemory( MTX_MemHandler *mem, CHAR *spline, LONG oldLength, LONG *newLength)
{
	Glyph *glyph;
	LONG bytesRead, maxOutSize = oldLength + oldLength/2;
	CHAR *destPtr = MTX_mem_malloc(mem, maxOutSize);
	if (!destPtr && mem->errorCode)
		return NULL;

	glyph = MTX_Glyph_Create( mem, (CHAR)false, spline, oldLength, 0 ,0 , &bytesRead, 
														spline, spline+oldLength);
	if (!glyph)
		return NULL;

	assert( bytesRead <= oldLength  );
	assert( bytesRead == oldLength || bytesRead == oldLength-1 );

	*newLength = MTX_Glyph_WriteTTFSpline( glyph, &destPtr, 0, &maxOutSize ); /* Recreate TTF glyph data on the fly */
	if (*newLength == -1)
		return NULL;
	assert( *newLength <= maxOutSize );

	MTX_Glyph_Destroy( glyph );
	return destPtr; /*****/
}

/* Constructor */
TTF_Converter *MTX_TTC_Create( MTX_MemHandler *mem, SHORT version )
{
	register TTF_Converter *t	= (TTF_Converter *)MTX_mem_malloc( mem, sizeof( TTF_Converter ) );
	if (!t && mem->errorCode)
		return NULL;
	t->mem						= mem;
	
	MTX_TTC_InitCoordEncoding();
	t->ttf 						= NULL;
	t->ttfSize					= 0;
	t->ctfData.ctf1     		= NULL;
	t->ctfData.ctfSize1 		= 0;
	t->convertFormat 			= version;
	
#ifdef DEBUG
	{
		CHAR buffer[16];
		CHAR *p;
		SHORT i, v1, v2;
		
		for ( i = -1000; i <= 1000; i++ ) {
			p = buffer;
			Write255Short( &p, i );
			Write255Short( &p, i );
			
			p = buffer;
			v1 = Read255Short( &p);
			v2 = Read255Short( &p);
			assert( i == v1 );
			assert( i == v2 );
		}
	}
#endif /* DEBUG */
	

	return t; /******/
}

/* Destructor */
void MTX_TTC__Destroy( TTF_Converter *t )
{
	MTX_mem_free( t->mem, t );
}
