/******************************************************************************
 * This software module was originally developed by Agfa Monotype Corporation
 * in the course of development of the ISO/IEC 14996-18 standard for reference 
 * purposes and its performance may not have been optimized. This software 
 * module is an implementation of font compression technology as specified 
 * by the ISO/IEC 14996-18 standard.
 * 
 * ISO/IEC gives users a free copyright license to this software module or
 * modifications thereof for use in products claiming conformance to 
 * ISO/IEC 14996-18 International Standard. ISO/IEC gives users the same free
 * license to this software module or modifications thereof for research 
 * purposes and further ISO/IEC standardization.
 * 
 * Those intending to use this software module in products are advised that
 * its use may infringe existing patents and that this license does not grant 
 * rights under any patents. ISO/IEC have no liability for use of this 
 * software module or modifications thereof.
 * 
 * No license is granted and copyright is not released for products that do 
 * not conform to ISO/IEC 14996-18 International Standard.
 * 
 * Agfa Monotype Corporation retains full right to modify and use the code for
 * its own purpose, assign or donate the code to a third party and to inhibit 
 * third parties from using the code for products that do not conform to
 * audiovisual and image-coding related ITU recommendations and 
 * ISO/IEC International Standards. All rights not expressly granted to the 
 * software module or modifications thereof are reserved by 
 * Agfa Monotype Corporation.
 * 
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2003. All rights reserved.
 ******************************************************************************/

/*
 * File:						canonical.c
 * Revision history
 * $Log: canonical.c,v $
 * Revision 1.1  2004/02/10 22:01:11  vladlev
 * Initial version
 *
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "SFNT.H"
#include "BITIO.H"
#include "MEMORY.H"
#include "RA_COMP.H"
#include "RADECOMP.H"
#include "canonical.h"


LONG *grade_vector;
BYTE *match_length;
LONG *match_indices;

LONG *taken_tokens;
LONG *taken_indices;
LONG num_pairs;
LONG max_pairs;

BYTE *taken;
TOKENS **tokens;
LONG num_tokens;
LONG max_tokens;

LONG OutputSymbol( MTX_RA_COMP *t, BITIO *bio, LONG index );

int debug = 0;

#ifdef DEBUG
/****************************************************************/
/* print a number in binary ... a printf("%b") ... */
void print_code(ULONG code, int len)
{
	ULONG m;

	if (len)
		{
		m = 1 << (len-1);
		while (m)
			{
			putchar( (code & m) ? '1' : '0');
			m >>= 1;
			}
		putchar(' ');
		}
}
#endif

/****************************************************************/
static LONG find_string(BYTE *str, LONG strlen, BYTE *sub, LONG sublen)
	{
	LONG i,j;

	for (i=0; i<strlen; i++)
		{
		for (j=0; i+j<strlen && j<sublen; j++)
			{
			if (str[i+j] != sub[j])
				break;
			}
		if (j == sublen)
			return i;
		}
	return -1;	/* THIS SHOULD NEVER HAPPEN */
	}

/****************************************************************/
/* after dictionary has been made and huffman tree exists
* assemble into the tokens[] we need for the rest of process
*/
void assemble_tokens(MTX_RA_COMP *t)
{
	int i,offset_length;
	int size_a, size_b;
	LONG hcl;
	
	num_tokens = 0;
	max_tokens = 256;
	tokens = (TOKENS **) malloc(max_tokens * sizeof(TOKENS *));

	/* first thing see if we're going to use inferred or explicit offsets */

	/* explicit offsets need <all 256 bytes> + overlapping string */
	size_a = 256 + t->stringStoreLength;

	/* how many bytes we need for the offsets? */
	if (size_a < 0xFF)
		offset_length = 1;
	else if (size_a < 0xFFFF)
		offset_length = 2;
	else if (size_a < 0xFFFFFF)
		offset_length = 3;
	else
		offset_length = 4;
	/* add size of the offset vector */
	size_a += offset_length * t->symbolCount;

	/* inferred size -- just 256 + the sum of the token string lengths */
	for (size_b=256,i=0; i<t->symbolCount; i++)
		size_b += t->dictSymbolLength[i];

	/* if inferred offsets are smaller, set offset_length==0 */
#ifdef DEBUG
	printf("inferred size=%d explicit size=%d, therefore using ",size_b,size_a);
#endif

	if (size_b < size_a)
	{
		offset_length = 0;
#ifdef DEBUG
		printf("non-overlapping tokens with inferred offsets \n");
#endif
	}
	else
	{
#ifdef DEBUG
		printf("%d byte explicit offsets\n",offset_length);
#endif
	}
	t->offset_length = offset_length;

	/* assemble the tokens */
	for (i=0; i<t->symbolCount; i++)
	{
		if (t->count[i] > 0)
		{
			TOKENS *token;

			token = (TOKENS *) malloc(sizeof(TOKENS));
			token->num = i;
			token->string = t->dictSymbol[i];
			token->length = (BYTE)t->dictSymbolLength[i];
			token->offset = find_string(t->stringStore,t->stringStoreLength,t->dictSymbol[i],t->dictSymbolLength[i]);

			/* we always compute the real offsets -- even if we don't use them (?)
			* account for automatic 256 bytes up front of string data
			*/
			if (token->offset >= 0)
				token->offset += 256;
			else	
				token->offset = (BYTE)t->dictSymbol[i][0];

			token->count = t->count[i];
			token->code = 0;

			/* get the huffman code length */
			hcl = OutputSymbol(t,0,i);
			if (hcl > 32)
			{
#ifdef DEBUG
				printf("***** WHOA ... HUFFMAN CODE LONGER THAN 32 BITS *****\n");
				/*** this should never happen since it implies that one of our "common"
				*** tokens occurs less than 1 in 2^^32 occurances ... hardly common
				***/
#endif /* DEBUG */
				exit (-16);
			}
			token->bits = (BYTE)hcl;

			/* add to list of tokens */
			if (num_tokens == max_tokens)
			{
				max_tokens += 256;
				tokens = (TOKENS **) realloc(tokens, max_tokens * sizeof(TOKENS *));
			}
			tokens[num_tokens++] = token;
		}
	}
}
/****************************************************************/
/* armed with huffman code lengths in tokens[]->bits
* make the canonical huffman codes in tokens[]->code
*/
void make_canonical()
{
	LONG gap,i,j;
	TOKENS *nt;
	
	/* first shell sort tokens in order of ascending <bits> */
	for (gap=num_tokens/2; gap>0; gap/=2)
		for (i=gap; i<num_tokens; i++)
			for (j=i-gap; j>=0 && tokens[j]->bits > tokens[j+gap]->bits; j-=gap)
			{
				nt = tokens[j];
				tokens[j] = tokens[j+gap];
				tokens[j+gap] = nt;
			}

	/* now assign (new) token numbers */
	for (i=0; i<num_tokens; i++)
		tokens[i]->index = i;

	/* now assign/replace the huffman codes with the canonical codes */
	tokens[0]->code = 0;
	for (i=1; i<num_tokens; i++)
	{
		tokens[i]->code = tokens[i-1]->code + 1;
		if (tokens[i]->bits != tokens[i-1]->bits)
			tokens[i]->code <<= (tokens[i]->bits - tokens[i-1]->bits);
	}

	/* print things so we know all is cool */
#ifdef DEBUG
	if (debug)
	{
		printf("*** tokens[] with canonical huffman codes ***\n");
		for (i=0; i<num_tokens; i++)
		{
			printf("%3d: num=%5d count=%5d bits=%2d code=",i,tokens[i]->num,
				tokens[i]->count, tokens[i]->bits);
			print_code(tokens[i]->code, tokens[i]->bits);
			show_string(tokens[i]->string, tokens[i]->length);
		}
		fflush(stdout);
	}
#endif /* DEBUG */
}

/****************************************************************/
/* write the header (with fake bitMarks and daltaBytes) and return
* the bit offset to there these fakes live */
void write_canonical_header(MTX_RA_COMP *t, BITIO *out, LONG n_tokens, TOKENS **list)
{
	LONG temp,i;
	int n_groups;
	LONG fsize = t->lengthIn;
	LONG offset_length = t->offset_length;

	/* RAC version 3 */
	MTX_BITIO_WriteValue(out,'R',8);
	MTX_BITIO_WriteValue(out,'A',8);
	MTX_BITIO_WriteValue(out,'C',8);
	MTX_BITIO_WriteValue(out,3,8);

	/* compresed file size -- 0 for now */
	MTX_BITIO_Write32(out,0);

	/* uncompressed size */
	MTX_BITIO_Write32(out,fsize);
#ifdef DEBUG
	printf("unCompressedByteCount=%d\n",fsize);
#endif
	/* number of tokens */
	MTX_BITIO_Write32(out,n_tokens);
#ifdef DEBUG
	printf("r->num_tokens=%d\n",n_tokens);
#endif

	/* now the length bytes -- one per token */
	for (i=0; i<n_tokens; i++)
		MTX_BITIO_Write8(out,list[i]->length);

	/* write the offset_length -- number of bytes in each offset */
	MTX_BITIO_Write8(out,offset_length);
	
	/* and maybe the explicit offsets -- otherwise they are inferred from lengths[] */
	if (offset_length != 0)
	{
		int i, bits = offset_length << 3;
		
		for (i=0; i<n_tokens; i++)
			MTX_BITIO_WriteValue(out,list[i]->offset,bits);

	}
			
	/* now the string data */
	if (offset_length == 0)
	{
		/* total size of non_overlapping string data */
		for (temp=0, i=0; i<n_tokens; i++)
			temp += list[i]->length;
		MTX_BITIO_Write32(out,temp);

#ifdef DEBUG
		printf("r->string_len=%d\n",temp);
#endif

		/* now the individual non_overlapping strings */
		for (i=0; i<n_tokens; i++)
		{
			int j;

			temp = list[i]->length;
			for (j=0; j<temp; j++)
				MTX_BITIO_Write8(out,list[i]->string[j]);
		}
	}
	else
	{
		/* the overlapping data length */
		MTX_BITIO_Write32(out,t->stringStoreLength + 256);
#ifdef DEBUG
		printf("r->string_len = %d\n",t->stringStoreLength + 256);
#endif
        
		/* write the first 256 bytes */
		for (i=0; i<256; i++)
			MTX_BITIO_Write8(out,i);

		/* and the overlapping data itself */
		for (i=0; i<t->stringStoreLength; i++)
			MTX_BITIO_Write8(out,t->stringStore[i]);
	
	}


	/* number of huffman code length groups ... can be from 1 to 32 */
	for (n_groups=1, i=1; i<n_tokens; i++)
		if (list[i]->bits != list[i-1]->bits)
			n_groups++;
	MTX_BITIO_Write8(out,n_groups);

#ifdef DEBUG
	printf("r->num_groups=%d\n",n_groups);
#endif

	/* now the code lengths in each huffman group -- as bytes */
	MTX_BITIO_Write8(out,list[0]->bits);
#ifdef DEBUG
	printf("bits[] = %d ",list[0]->bits);
#endif
    for (i=1; i<n_tokens; i++)
	{
		if (list[i]->bits != list[i-1]->bits)
		{
			MTX_BITIO_Write8(out,list[i]->bits);
#ifdef DEBUG
			if (debug)
				printf("%d ",list[i]->bits);
#endif
		}
	}
#ifdef DEBUG
	printf("\n");
#endif

	/* now the first codes in each huffman group  */
#ifdef DEBUG
	if (debug)
		printf("firsts[] = ");
#endif

	MTX_BITIO_Write32(out,list[0]->code);	/* the first code */
#ifdef DEBUG
	if (debug)
		print_code(list[0]->code,list[0]->bits);
#endif

	for (i=1; i<n_tokens; i++)
	{
		if (list[i]->bits != list[i-1]->bits)
		{
			/* code of new group */
			MTX_BITIO_Write32(out,list[i]->code);
#ifdef DEBUG
			if (debug)
				print_code(list[i]->code,list[i]->bits);
#endif
		}
	}
#ifdef DEBUG
	if (debug)
		printf("\n");
#endif

	/* now the last codes in each huffman group */
#ifdef DEBUG
	if (debug)
		printf("lasts[]  = ");
#endif

	for (i=1; i<n_tokens; i++)
	{
		if (list[i]->bits != list[i-1]->bits)
		{
			/* code of last group */
			MTX_BITIO_Write32(out,list[i-1]->code);
#ifdef DEBUG
			if (debug)
				print_code(list[i-1]->code,list[i-1]->bits);
#endif
		}
	}
	MTX_BITIO_Write32(out,list[n_tokens-1]->code);
#ifdef DEBUG
	if (debug)
	{
		print_code(list[n_tokens-1]->code,list[n_tokens-1]->bits);
		printf("\n");
		fflush(stdout);
	}
#endif

	/* resort the list by ascending original token numbers to make
	* looking up the canonical code for an original token number easy */
	{
		LONG gap,i,j;
		TOKENS *temp;

		for (gap=n_tokens/2; gap>0; gap/=2)
			for (i=gap; i<n_tokens; i++)
				for (j=i-gap; j>=0 && list[j]->num > list[j+gap]->num; j-=gap)
				{
					temp = list[j];
					list[j] = list[j+gap];
					list[j+gap] = temp;
				}
	}
}

/****************************************************************/
/* encode the data using canonical codes -- also install the real
* bitMarks and DeltaBytes.  This COMPLETES the RAC3 file */
int EncodeData( MTX_RA_COMP *t, LONG byteCount, CHAR *bitStreamName )
{
	LONG crumbs,crumb_offset,*bitMarks;
	BYTE *deltaBytes;
	LONG next_pos, old_i, old_bitMark, token_offset;

	LONG i, match, bitCount = 0;
	BITIO *bio;
	LONG dataSize, memSize = 1024;
	FILE *bFile  = NULL; 
	void *memPtr = MTX_mem_malloc( t->mem, memSize );

	if (!memPtr && t->mem->errorCode)
		return (t->mem->errorCode);

	bio = MTX_BITIO_Create( t->mem, memPtr, memSize, 'w' );
	if (!bio)
		return (t->mem->errorCode);
	
	/* write the header, string tables, huffman code tables */
	write_canonical_header(t,bio, num_tokens, tokens);

	/* number of bread crumbs (bitMarks and deltaBytes) */
	crumbs = (t->lengthIn) / BYTE_CACHE_INTERVAL + 1;
	MTX_BITIO_Write32(bio,crumbs);

	if (debug)
		printf("crumbs=%d\n",crumbs);

	bitMarks = (LONG *)MTX_mem_malloc(t->mem, 4 * crumbs);
	deltaBytes = (BYTE *)MTX_mem_malloc(t->mem, crumbs);

	/* remember the offset, because we're writing junk now */
	crumb_offset = MTX_BITIO_ftell(bio);
	for (i=0; i<crumbs; i++)
		MTX_BITIO_Write32(bio,bitMarks[i]);	
	for (i=0; i<crumbs; i++)
		MTX_BITIO_Write8(bio,deltaBytes[i]);

	/* write token_offset ... which is 32 bits from HERE */
	token_offset = 32 + MTX_BITIO_ftell(bio);
	MTX_BITIO_Write32(bio,token_offset);

	if (debug)
		printf("token_offset = %d\n",token_offset);
	
	/* encode the data and calculate REAL bitMarks[] and deltaBytes[] */
	next_pos = 0;
	old_bitMark = MTX_BITIO_ftell(bio);

	old_i = 0;
	for ( i = 0; i < t->lengthIn; ) 
	{
		LONG lo,hi,mid;
		LONG FindBestMatch( MTX_RA_COMP *t, LONG pos );	/* used to be static */


        old_bitMark = MTX_BITIO_ftell(bio);
		old_i = i;

		/*** the following takes the place of <ReadSymbolLength> ***/
		/*** we're actually determining the best symbol, and then ***/
		/*** puting it to the BITIO, finally we grab the length ***/
		match = FindBestMatch(t,i);
		assert( match >= 0 && match < t->symbolCount );

		/* binary search to find the <match> in <tokens> */
		lo = 0;
		hi = num_tokens - 1;
		while (lo <= hi)
		{
			mid = (lo + hi)/2;
			if (match < tokens[mid]->num)
				hi = mid-1;
			else if (match > tokens[mid]->num)
				lo = mid+1;
			else
				break;
		}
		assert(tokens[mid]->num == match);

		/* output the bits and advance */
#ifdef DEBUG
		if (debug)
		{
			printf("i=%6d bitmark=%d token_num=%d ",i,MTX_BITIO_ftell(bio),tokens[mid]->index);
			print_code(tokens[mid]->code, tokens[mid]->bits);
			show_string(tokens[mid]->string, tokens[mid]->length);
		}
#endif /* DEBUG */

		MTX_BITIO_WriteValue(bio, tokens[mid]->code, tokens[mid]->bits);
		/** ok ,,, we have indeed found the next symbol length **/
		/** in addition to selecting and outputting it ... so **/

		i += tokens[mid]->length;
        /* time for a new bitMark and deltaByte ? */
		if (i > next_pos)
		{
			LONG new_interval, old_interval, n;
			
			/* A token may be longer then BYTE_CACHE_INTERVAL. If this is the case, we */
			/* could skip an interval (or more), if we jumped over BYTE_CACHE_INTERVAL */
			/* boundary. If this happened, we need to establish proper bitMarks[n] and */
			/* deltaBytes[n] for "skipped" 'n' values and (!) keep 'next_pos' current */
			new_interval = i / BYTE_CACHE_INTERVAL;
			old_interval = next_pos / BYTE_CACHE_INTERVAL;
			/* if we are "on track", then new and old interval counts are the same */
			/* and statements in the loop will be executed only once */
			for (n = old_interval; n <= new_interval; n++)
			{
				bitMarks[n] = old_bitMark;
				deltaBytes[n] = (BYTE)(next_pos - old_i);
				next_pos += BYTE_CACHE_INTERVAL;
			}
		}

	}

	/* write the real bitMarks[] */
	MTX_BITIO_fseek(bio, crumb_offset);
	for (i=0; i<crumbs; i++)
		MTX_BITIO_Write32(bio,bitMarks[i]);

	/* write the real deltaBytes[] */
	for (i=0; i<crumbs; i++)
		MTX_BITIO_Write8(bio,deltaBytes[i]);
	MTX_BITIO_flush_bits( bio );

	/* write the real compressed size */
	dataSize = MTX_BITIO_GetBytesOut( bio );
	MTX_BITIO_fseek(bio, 4<<3);
	MTX_BITIO_Write32(bio,dataSize);
	MTX_BITIO_flush_bits( bio );

	/* note ... <bio> and thus "bitStreamName" file is the ENTIRE compressed file */
	memPtr = (void *)MTX_BITIO_GetMemoryPointer( bio );
	bFile = fopen( bitStreamName,   "wb" ); 
	fwrite( memPtr, dataSize, 1, bFile );
	fclose( bFile );
	
	MTX_mem_free( t->mem, memPtr );
	MTX_mem_free( t->mem, bitMarks );
	MTX_mem_free( t->mem, deltaBytes );
	MTX_BITIO_Destroy( bio );


	return 0;	/* success */
}

/****************************************************************/
/* given a RAC3 file in memory, readin all the necessary glop 
* to facilitate decoding a file 
*
* code assumes that a BYTE pointer can be any address
* if NOT, you will need to allocate and readin the BYTES
* Also you'll need to free them in MTX_RA_DECOMP_Destroy
*
* NB: this handles BOTH inferred and explicit offsets
* even though <write_canonical_header> can only do inferred
* offsets at this time. 
*/
MTX_RA_DECOMP *MTX_RA_DECOMP_Create(MTX_MemHandler *mem, BYTE *data )
	{
	BITIO *bio;
	MTX_RA_DECOMP *r;
	int offset_length;	/* length in bytes of each offset entry. 0 => inferred offsets */

	LONG *p,i,n,offset,crumbs;
	LONG fsize = 0;		/*** bytes in the compressed file ... ***/

#ifdef DEBUG
	printf("***** reading *****\n");
#endif /* DEBUG */

	/* verify 'RAC3' signature */
	if (data[0] != 'R' || data[1] != 'A' || data[2] != 'C' || data[3] != 3)
	{
#ifdef DEBUG
		printf("*** not a RAC3 data image ***\n");
#endif /* DEBUG */
		return 0;
	}

	/* get compressed file size */
	fsize = data[4];
	fsize = (fsize << 8) | data[5];
	fsize = (fsize << 8) | data[6];
	fsize = (fsize << 8) | data[7];

	/* create BITIO and skip signature and compressed file size */
	bio = MTX_BITIO_Create(mem, data, fsize, 'r');
	MTX_BITIO_fseek(bio,8<<3);

	r = (MTX_RA_DECOMP*) MTX_mem_malloc(mem, sizeof(MTX_RA_DECOMP));
	r->mem = mem;
	r->bio = bio;

    /* Initialize the Huffman code lookup table */
    for (i=0; i<TABLE_SIZE; i++)
        r->token_bits_table[i] = 0;  /* means no entry for code i */
    r->token_num_table[0] = -1;      /* means table not built     */

	/* read uncompressed file size */
	r->unCompressedByteCount = MTX_BITIO_Read32(bio);

	/* read number of tokens */
	r->num_tokens = MTX_BITIO_Read32(bio);

	/* establish pointer to lengths[] -- and skip the data */
	offset = MTX_BITIO_ftell(bio);
	assert(0 == (offset & 7));
	r->lengths = data + (offset >> 3);
	offset += (r->num_tokens << 3);
	MTX_BITIO_fseek(bio,offset);

	/* read in the offset_length */
	offset_length = MTX_BITIO_Read8(bio);
	
	/* allocate and calculate or read the offsets */
	n = r->num_tokens;
	p = r->offsets = (LONG *)MTX_mem_malloc(mem, n * 4);

	if (offset_length == 0)
	{
		/* infer the offsets of the non_overlapping token strings from their lengths */
		p[0] = 0;
		for (i=1; i<r->num_tokens; i++)
			p[i] = p[i-1] + r->lengths[i-1];
	}
	else
	{
		/* read in the offsets of the overlapping token strings */
		offset_length <<= 3;
		for (i=0; i<r->num_tokens; i++)
			p[i] = MTX_BITIO_ReadValue(r->bio,offset_length);
	}

#ifdef DEBUG
	if (debug)
		{
		printf("offsets\n");
		for (i=0; i<r->num_tokens; i++)
			printf("%d: %d\n",i,r->offsets[i]);
		}
#endif /* DEBUG */

	/* read total string data length */
	n = r->string_len = MTX_BITIO_Read32(bio);
#ifdef DEBUG
	printf("r->string_len=%d\n",r->string_len);
#endif /* DEBUG */


	/* establish pointer to string[] -- and skip the data */
	offset = MTX_BITIO_ftell(bio);
	assert(0 == (offset & 7));
	r->string = data + (offset >> 3);
	offset += (r->string_len<<3);
	MTX_BITIO_fseek(bio,offset);

	/* read number of canonical huffman code groups */
	r->num_groups = MTX_BITIO_Read8(bio);
#ifdef DEBUG
	printf("r->num_groups=%d\n",r->num_groups);
#endif /* DEBUG */
	
	/* establish pointer to bits[] -- and skip the data */
	offset = MTX_BITIO_ftell(bio);
	assert(0 == (offset & 7));
	r->bits = data + (offset>>3);
	offset += (r->num_groups<<3);
	MTX_BITIO_fseek(bio,offset);

	/* allocate and read in firsts[]  */
	n = r->num_groups;
	p = r->firsts = (LONG *)MTX_mem_malloc(mem, n * 4);
	while (n--)
		*p++ = MTX_BITIO_Read32(bio);


#ifdef DEBUG
	if (debug)
		{
		int i;
		printf("firsts[] = ");
		for (i=0; i<r->num_groups; i++)
			print_code(r->firsts[i], r->bits[i]);
		printf("\n");
		}
#endif /* DEBUG */

	/* allocate and read in lasts[]  */
	n = r->num_groups;
	p = r->lasts = (LONG *)MTX_mem_malloc(mem, n * 4);
	while (n--)
		*p++ = MTX_BITIO_Read32(bio);

#ifdef DEBUG
	if (debug)
		{
		int i;
		printf("lasts[]  = ");
		for (i=0; i<r->num_groups; i++)
			print_code(r->lasts[i], r->bits[i]);
		printf("\n");
		}
#endif /* DEBUG */

	/* calculate next_mask to keep rightmost MAX bits */
	if(r->bits[r->num_groups-1] == 32)
		r->next_mask = 0xFFFFFFFF;
	else
		r->next_mask = (1 << r->bits[r->num_groups-1]) - 1;

	/* allocate and calculate tokens[]  */
	n = r->num_groups;
	p = r->tokens = (LONG *)MTX_mem_malloc(mem, n * 4);
	p[0] = 0;
	for (i=1; i<n; i++)
		p[i] = p[i-1] + 1 + (r->lasts[i-1] - r->firsts[i-1]);

#ifdef DEBUG
	if (debug)
		{
		int i;
		printf("tokens[] = ");
		for (i=0; i<r->num_groups; i++)
			printf("%d ",r->tokens[i]);
		printf("\n");
		}
#endif /* DEBUG */

	/* read crumbs (size of bitMarks and deltaBytes arrays) */
	crumbs = MTX_BITIO_Read32(bio);

	/* allocate and read in the bitMarks[] ... or establish the pointer if ROMbitMarks */
	r->bitMarks = &bio->array[bio->index>>3]; /* point at the ROM, we'll decode the LONGs on the fly */
	bio->index += 32*crumbs;

#ifdef DEBUG
	if (debug)
		{
		int i;
		printf("bitMarks[] =");
		for (i=0; i<crumbs; i++)
		{
			ULONG  bitMark;
			BYTE *p = &r->bitMarks[4*i];
			bitMark = *p++;
			bitMark = (bitMark<<8) | *p++;
			bitMark = (bitMark<<8) | *p++;
			bitMark = (bitMark<<8) | *p++;
			printf("%d ",bitMark);
		}

		printf("\n");
		}
#endif /* DEBUG */

	/* establish pointer to deltaBytes[] */
	n = crumbs;
	offset = MTX_BITIO_ftell(bio);
	assert(0 == (offset & 7));
	r->deltaBytes = data + (offset>>3);
	offset += n<<3;
	MTX_BITIO_fseek(bio,offset);

#ifdef DEBUG
	if (debug)
		{
		int i;
		printf("deltaBytes[] =");
		for (i=0; i<crumbs; i++) printf("%d ",r->deltaBytes[i]);
		printf("\n");
		}
#endif /* DEBUG */

	/* read token_offset */
	r->token_offset = MTX_BITIO_Read32(bio);
#ifdef DEBUG
	if (debug)
		{
		printf("token_offset = %d\n",r->token_offset);
		fflush(stdout);
		}
#endif /* DEBUG */
	return r;
	}
/****************************************************************/
void MTX_RA_DECOMP_Destroy(MTX_RA_DECOMP *t)
	{
	MTX_mem_free(t->mem,t->firsts);
	MTX_mem_free(t->mem,t->lasts);
	MTX_mem_free(t->mem,t->tokens);
	MTX_mem_free(t->mem,t->offsets);
	MTX_BITIO_Destroy( t->bio );
	MTX_mem_free(t->mem, t );
	}
/****************************************************************/
/* from current position (assuming someone else has done the
* MTX_BITIO_fseek() and read in <next>) get the next token
* number and refresh <next>.
*/

/****************************************************************/
#ifdef DEBUG
/* build the table by decoding codes  */
static old_build_token_table(MTX_RA_DECOMP *r)
	{
	int lo,hi,mid;
	ULONG first;
	ULONG last;
	ULONG next;
	int drop;
	LONG i,token_num;
	BYTE n;
	BYTE max = TABLE_BITS;


	for (i=0; i<TABLE_SIZE; i++)
		{
		/* binary search for <i> in first/last */
		lo = 0;
		hi = r->num_groups-1;
		while (lo <= hi)
			{
			mid = (lo + hi)/2;
			n = r->bits[mid];
			drop = max - n;	
			next = i >> drop;
			first = r->firsts[mid];
			last = r->lasts[mid];

			if (next < first)
				hi = mid-1;
			else if (next > last)
				lo = mid+1;
			else
				{
				token_num = r->tokens[mid] + (next - first);
				r->token_num_table[i] = token_num;
				r->token_bits_table[i] = n;
				break;
				}
			}
		}

	if (debug)
		{
		printf("code         bits token_num\n");
		for (i=0; i<TABLE_SIZE; i++)
			{
			print_code(i,TABLE_BITS);
			printf("%4d %9d\n",r->token_bits_table[i],r->token_num_table[i]);
			}
		fflush(stdout);
		}
	}
#endif /* DEBUG */
/****************************************************************/
/* here is the faster one ... always the same (I think) */
static build_token_table(MTX_RA_DECOMP *r)
	{
	int i,j,k,bits,count,num,token,index;

#ifdef DEBUG
	/* do it the old fashioned way */
	old_build_token_table(r);
	printf("*** checking for errors in (fast) build_token_table()\n");
#endif /* DEBUG */

	for (i=0,index=0;  (index < TABLE_SIZE) &&  (i < r->num_groups) && (r->bits[i] <= TABLE_BITS); i++)
		{
		bits = r->bits[i];
		count = 1 << (TABLE_BITS - bits);
		num = 1 + r->lasts[i] - r->firsts[i];
		token = r->tokens[i];
        for (j=0,token=r->tokens[i]; (index < TABLE_SIZE) &&  (j<num); j++,token++)
			{
			for (k=0; (index < TABLE_SIZE) &&  (k < count); k++)
				{
#ifdef DEBUG
				/* and compare with the fast way */
				if (r->token_num_table[index] != token)
					printf("*** token_num_table[%d] != %d\n",index,token);
				if (r->token_bits_table[index] != bits)
					printf("*** token_bits_table[%d] != %d\n",index,bits);
#endif /* DEBUG */

				r->token_num_table[index] = token;
				r->token_bits_table[index] = bits;
				index++;
				}
			}
		}
    printf("Build Table Size:index %d\n", index);
    r->slowGroup = i;   /* first group not completely contained in the table */
#ifdef DEBUG
	if (debug)
		{
		printf("code         bits token_num\n");
		for (i=0; i<TABLE_SIZE; i++)
			{
			print_code(i,TABLE_BITS);
			printf("%4d %9d\n",r->token_bits_table[i],r->token_num_table[i]);
			}
		fflush(stdout);
		}
#endif /* DEBUG */

	}


/****************************************************************/
static int get_next_token(MTX_RA_DECOMP *r)
{
	int lo,hi,mid;
	ULONG first;
	ULONG last;
	ULONG next;
	ULONG rest;
    int bitsNeeded;
	int drop;
	LONG token_num;
	BYTE n;
	BYTE max;

	/* do we need to build the token table ? */
	if (r->token_num_table[0] < 0)
		build_token_table(r);

    next = r->next;

    /* ? possible && legitimate entry in the table */
    if (r->token_bits_table[next])
	{
		n = r->token_bits_table[next];
		token_num = r->token_num_table[next];
		next <<= n;
		next |= MTX_BITIO_ReadValue(r->bio,n);
		next &= NEXT_MASK;
		r->next = next;
		return token_num;
	}
	else
	{
		/* not in table ... binary search the longer length groups */
	    max = r->bits[r->num_groups-1];
        bitsNeeded = max - TABLE_BITS;
        if (bitsNeeded > 0)
            r->next = (next << bitsNeeded) | MTX_BITIO_ReadValue(r->bio, bitsNeeded);
        else if (bitsNeeded < 0)
        {
            r->next >>= -bitsNeeded;
            r->bio->index += bitsNeeded;
        }
		lo = 0;
		while (r->bits[lo] <= TABLE_BITS)
			lo++;
        if(lo != r->slowGroup)
			printf("********found the bug   ********\n");
		hi = r->num_groups-1;
		while (lo <= hi)
		{
			mid = (lo + hi)/2;
			n = r->bits[mid];
			drop = max - n;	
			next = r->next >> drop;
			first = r->firsts[mid];
			last = r->lasts[mid];

			if (next < first)
				hi = mid-1;
			else if (next > last)
				lo = mid+1;
			else
			{
				token_num = r->tokens[mid] + (next - first);
                bitsNeeded = TABLE_BITS - max + n;  /* to get TABLE_BITS bits into next */
                if(bitsNeeded > 0)
                {
                    r->next <<= bitsNeeded;
	                rest = MTX_BITIO_ReadValue(r->bio,bitsNeeded);
                    r->next |= rest;
                }
                else if(bitsNeeded<0)
                {
                    r->next >>= -bitsNeeded;
                    r->bio->index += bitsNeeded ;
                }
				/* else we need 0 bits but in all cases clean off the used bits */

				r->next &= NEXT_MASK;

				return token_num;
			}
		}
#ifdef DEBUG
		printf("****** THIS CAN'T HAPPEN, WE GOT A BAD HUFFMAN CODE ******\n");
#endif /* DEBUG */
		exit(-17);
	}
}

/****************************************************************/
/* return the next token's string pointer and set its length */
CHAR *ReadSymbol(MTX_RA_DECOMP *t, LONG *len )
{
	int n;
	CHAR *r;
	
	n = get_next_token(t);
#ifdef DEBUG
	if (debug)
		printf("index=%d ",n);
#endif /* DEBUG */
	*len = t->lengths[n];
	r = t->string + t->offsets[n];
	return r;
}
/****************************************************************/

