/******************************************************************************
 * This software module was originally developed by Agfa Monotype Corporation
 * in the course of development of the ISO/IEC 14996-18 standard for reference 
 * purposes and its performance may not have been optimized. This software 
 * module is an implementation of font compression technology as specified 
 * by the ISO/IEC 14996-18 standard.
 * 
 * ISO/IEC gives users a free copyright license to this software module or
 * modifications thereof for use in products claiming conformance to 
 * ISO/IEC 14996-18 International Standard. ISO/IEC gives users the same free
 * license to this software module or modifications thereof for research 
 * purposes and further ISO/IEC standardization.
 * 
 * Those intending to use this software module in products are advised that
 * its use may infringe existing patents and that this license does not grant 
 * rights under any patents. ISO/IEC have no liability for use of this 
 * software module or modifications thereof.
 * 
 * No license is granted and copyright is not released for products that do 
 * not conform to ISO/IEC 14996-18 International Standard.
 * 
 * Agfa Monotype Corporation retains full right to modify and use the code for
 * its own purpose, assign or donate the code to a third party and to inhibit 
 * third parties from using the code for products that do not conform to
 * audiovisual and image-coding related ITU recommendations and 
 * ISO/IEC International Standards. All rights not expressly granted to the 
 * software module or modifications thereof are reserved by 
 * Agfa Monotype Corporation.
 * 
 * This copyright notice must be included in all copies or derivative works.
 * Copyright (c) ISO/IEC 2003. All rights reserved.
 ******************************************************************************/

/*
 * File:						tablecmp.c
 * Revision history
 * $Log: tablecmp.c,v $
 * Revision 1.1  2004/02/10 22:01:11  vladlev
 * Initial version
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>
#include <string.h>

#include "CONFIG.H"
#include "MEMORY.H"
#include "SFNT.H"
#include "BITIO.H"
#include "RADECOMP.H"
#include "tablecmp.h"
#include "ra_comp.h"
#include "TTF_CONV.H"
#include "ERRCODES.H"
#include "RA_UTIL.H"
#include "RA_TTF.H"


extern void insertTable(CHAR *sfnt, sfnt_TableTag tag, int offset, int length);

MTX_TABLE_COMPRESS *MTX_TABLE_COMPRESS_Create(MTX_MemHandler *mem)
{
    MTX_TABLE_COMPRESS *t;
    t = (MTX_TABLE_COMPRESS *)MTX_mem_malloc(mem, sizeof(MTX_TABLE_COMPRESS));
    if(!t)
        return NULL;
    t->mem              = mem;
    t->numEntries       = 0;
    t->numSlots         = 0;
    t->rb               = NULL;
    t->numTags          = 0;
    t->tagsToCompress   = NULL;
    t->sfnt0            = NULL;
    t->sfnt1            = NULL;
    return t;
}

MTX_TABLE_COMPRESS_Destroy(MTX_TABLE_COMPRESS *t)
{
    MTX_mem_free(t->mem, t->rb);
    MTX_mem_free(t->mem, t);
}

MTX_TABLE_COMPRESS *MTX_TABLE_COMPRESS_Add(MTX_TABLE_COMPRESS *t, ULONG start, ULONG length,
                                                             ULONG cmpStart, SHORT isCompressed)
{
    /* Make sure there's room */
    if(t->numEntries == t->numSlots)
    {
        if(!t->numSlots)                    /* First time through ...                       */
            t->numSlots = 20;               /* cough up 20 slots.                           */
        else                                /* later times through ...                      */
            t->numSlots += t->numSlots/2;   /* add half again as many as we already have    */
        t->rb = MTX_mem_realloc(t->mem, t->rb, t->numSlots*sizeof(MTX_ROM_BLOCK));
        if(!t->rb)
            return NULL;
    }
    t->rb[t->numEntries].start          = start;
    t->rb[t->numEntries].length         = length;
    t->rb[t->numEntries].cmpStart       = cmpStart;
    t->rb[t->numEntries].isCompressed   = isCompressed;
    t->numEntries++;
    return t;
}


void MTX_TABLE_COMPRESS_SortAndCombine(MTX_TABLE_COMPRESS *t)
{
    SHORT curRBindex, i;

    /* Combine uncompressed blocks */
    curRBindex=0;   /* index of output rom blocks- we try to combine as many future inexes
                       into this one */
    i=0;            /* index of input ROM blocks that we are trying to combine */
    while(i<t->numEntries)
    {
        /* rb[i] IS NOT compressed */
        t->rb[curRBindex].start    = t->rb[i].start;
        t->rb[curRBindex].cmpStart = t->rb[i].cmpStart;
        while(i<t->numEntries && !t->rb[i].isCompressed)
            i++;
        t->rb[curRBindex].length = t->rb[i-1].start + t->rb[i-1].length
                                            - t->rb[curRBindex].start;
        t->rb[curRBindex].isCompressed = 0;
        t->rb[curRBindex].decomp = NULL;
        curRBindex++;

        /* rb[i] IS compressed */
        while(i<t->numEntries && t->rb[i].isCompressed)
        {
            /* Each compressed table gets its own block */
            t->rb[curRBindex].start  = t->rb[i].start;
            t->rb[curRBindex].length = t->rb[i].length;
            t->rb[curRBindex].cmpStart = t->rb[i].cmpStart; 
            t->rb[curRBindex].isCompressed = (SHORT) BYTE_CACHE_INTERVAL;
            curRBindex++;
            i++;
        }
    }
    t->numEntries = curRBindex;
}

#define ACTheaderSize 20
#define ACTentrySize  14

static void MTX_TABLE_COMPRESS_WriteHeader(MTX_TABLE_COMPRESS *t, int pos )
{
    SHORT i, headerLength;
    CHAR *data;
    CHAR *dataStart;

    dataStart = data = &t->sfnt1[pos];

    WRITEWORD_INC(&data, 0);   /* header length, fix at end */
    WRITELONG_INC(&data, 0);   /* size out, fix later */
    WRITELONG_INC(&data, t->sizeIn);
    WRITELONG_INC(&data, t->glyf_OffsetStart);
    WRITELONG_INC(&data, t->glyf_OffsetEnd  );
    WRITEWORD_INC(&data, t->numEntries);
    for(i=0; i<t->numEntries; i++)
    {
        WRITELONG_INC(&data, t->rb[i].start);
        WRITELONG_INC(&data, t->rb[i].length);
        WRITELONG_INC(&data, t->rb[i].cmpStart);
        WRITEWORD_INC(&data, t->rb[i].isCompressed);
    }
    headerLength = ((SHORT)(data - dataStart) + 3) & ~3;
    t->sizeOut = pos + headerLength;
    WRITEWORD_INC(&dataStart,headerLength);
    WRITELONG_INC(&dataStart, t->sizeOut);
}

MTX_RA_TT_Decomp *MTX_RA_TT_ReadRom(MTX_RA_TT_Decomp *t, CHAR *data )
{
    SHORT i;
    ULONG start;
    ULONG length;
    ULONG cmpStart;
    SHORT isCompressed;
    MTX_TABLE_COMPRESS *tc;

    tc = t->tc;

    t->headerLength     = READWORD_INC(&data);
    t->cmpSize          = READLONG_INC(&data);
    t->uncmpSize        = READLONG_INC(&data);
    t->glyf_OffsetStart = READLONG_INC(&data);
    t->glyf_OffsetEnd   = READLONG_INC(&data);
    i                   = READWORD_INC(&data);
    while(i--)
    {
        start        = READLONG_INC(&data);
        length       = READLONG_INC(&data);
        cmpStart     = READLONG_INC(&data);
        isCompressed = READWORD_INC(&data);
        if(!MTX_TABLE_COMPRESS_Add(tc, start, length, cmpStart, isCompressed))
            return NULL;
    }
    return t;
}

static int MTX_TABLE_COMPRESS_ShouldCompress(MTX_TABLE_COMPRESS *t, sfnt_TableTag tag)
{
    int i;
    for(i=0; i < t->numTags; i++)
    {
        if(tag == t->tagsToCompress[i])
            return 1;
    }
    return 0;
}

static void AllocateSpace(MTX_MemHandler *mem,  CHAR * *destSfnt,
                                LONG pos, LONG byteCount, LONG *maxOutSize )
{
	CHAR * newDestSfnt;
	byteCount += 4; /* Always keep a 4 byte margin to allow for the zero padding */
	if ( pos + byteCount > *maxOutSize ) {
		*maxOutSize = pos + byteCount + (*maxOutSize>>1) + 2; /* Allocate memory in exponentially increasing steps */
		newDestSfnt = (CHAR *)MTX_mem_realloc( mem, *destSfnt, *maxOutSize );
		*destSfnt = newDestSfnt;
	}
}

static int CompressTables( MTX_TABLE_COMPRESS *t)
{
    USHORT numTables; /* mumber of TrueType tables in current font */
    ULONG tableOffset, tableLength;
    int fontOff0=0, fontOff1=0;
    int numInTTC=1, idxTTC;
    sfnt_TableTag tag; 
    int isTTC = 0;
    
    int i, headerSize;
    int pos=0;
    int uncPos=0;
    const LONG zero = 0L;
    int maxOutSize;
    CHAR *sfnt0;
    CHAR *sfnt1;
    CHAR *tableStart;
    int status;
    unsigned int packedLength;
                    CHAR* packedData;
    sfnt0 = t->sfnt0;
    sfnt1 = t->sfnt1;
    maxOutSize = t->sizeIn;

    /* check for TrueTypeCollection, set flag & #-fonts-in-file if so */
    tag = *((sfnt_TableTag *)((CHAR *)sfnt0 + TTC_TAG));
    if ( SWAPENDS_32(tag) == tag_TTCollection ) 
    {
        numInTTC = SWAPENDS_32(*((int *)((CHAR *)sfnt0 + TTC_NUMFONTS)));
        isTTC = 1;
        
        /* copy TTC header to output file now - repair offsets later */
        headerSize = TTC_OFFSET + (numInTTC * 4);
        AllocateSpace( t->mem, &sfnt1, pos, headerSize, &maxOutSize );
        printf("+++ writing TT Collection header: # fonts = %d, old/new pos=%d,%d\n", 
            numInTTC, pos, pos + headerSize);
        if(!MTX_TABLE_COMPRESS_Add(t, uncPos, headerSize, pos, 0))
            return t->mem->errorCode ;
        uncPos += headerSize;
        pos += headerSize;
        memcpy( (CHAR *)(sfnt1), (CHAR *)sfnt0, headerSize );
    }
    /* Make two passes through the fonts first for uncompressed blocks
	   (ttc directory, font direcories and uncompressed OpenType tables)
	   second for compressed blocks with OpneType/TrueType tables.
       We write all of the uncompressed data first, followed by the 
	   compressed data.
       The last table we write is "act3" which contains the machinery 
	   to decompress all of the TrueType tables that are compressed.
	   Loop runs once for each font in a TrueType collection.
	   It runs once if our TTF font is not a TrueType collection
     */
	for (idxTTC=0; idxTTC <	numInTTC; idxTTC++)
	{
	    if (isTTC)
	    {
	        fontOff0 = SWAPENDS_32( GET_ttc_font_offset( sfnt0, idxTTC ) );
	        
	        /* fix up offset in TTC header in output file */
	        SET_ttc_font_offset( sfnt1, idxTTC, SWAPENDS_32(uncPos) );
	        fontOff1 = pos;
	        printf("+++ processing font # %d: offset(in)=%d, p(out)=%d\n", idxTTC, fontOff0, fontOff1);
	    }
	    
	    /* TableDirectory --write it to output file */
	    headerSize = OFFSETTABLESIZE + (SWAPENDS_U16(GET_sfnt_numOffsets( sfnt0 + fontOff0 )) * DirectoryEntrySize );
	    if(!idxTTC)
	        headerSize += DirectoryEntrySize; /* 1st font gets an extra directory entry */
											  /* for the ACT table. */
	    AllocateSpace( t->mem, &sfnt1, pos, headerSize, &maxOutSize );
	    printf("=== writing font header: old/new pos=%d,%d\n", pos, pos + headerSize);
	    if(!MTX_TABLE_COMPRESS_Add(t, uncPos, headerSize, pos, 0))
	        return t->mem->errorCode;
	    uncPos += headerSize;
	    pos += headerSize;
	    memcpy( (CHAR *)sfnt1 + fontOff1, (CHAR *)sfnt0 + fontOff0, headerSize );
	    /* Loop runs once for each table in the current font */
	    numTables = SWAPENDS_U16(GET_sfnt_numOffsets( sfnt0 + fontOff0 ));
	    if(!idxTTC)
	        *((USHORT *)(sfnt1 + fontOff1 + 4)) = SWAPENDS_U16((USHORT)(numTables + 1)); //VL
	
	    for ( i = 0; i < numTables; i++ )
	    {
	        
	        tag = SWAPENDS_32(GET_sfnt_table_tag( sfnt0 + fontOff0, i ));
	        
	        
	        if(!MTX_TABLE_COMPRESS_ShouldCompress(t, tag))
	        {
	            SET_sfnt_table_length( sfnt1 + fontOff1, i, GET_sfnt_table_length( sfnt0 + fontOff0, i ) );
	            
	            /* Tolerate zero length tables */
	            if ( GET_sfnt_table_length( sfnt1 + fontOff1, i ) == 0 ) 
	                continue; /******/
	            
	            if (CheckForRepeatedTable(sfnt0, sfnt1, i, idxTTC))
	                continue; /******/
	            /* Get the input table's start, offset and length */
	            tableOffset = SWAPENDS_U32(GET_sfnt_table_offset( sfnt0 + fontOff0, i ));
	            tableLength = SWAPENDS_U32(GET_sfnt_table_length( sfnt0 + fontOff0, i ));
	            tableStart 	= (CHAR *)&sfnt0[tableOffset];
	            /* Remember where the glyf table was */
	            if(tag == tag_GlyphData)
	            {
	                t->glyf_OffsetStart = uncPos;
	                t->glyf_OffsetEnd   = uncPos + tableLength;
	            }
	            SET_sfnt_table_offset( sfnt1 + fontOff1, i, SWAPENDS_32(uncPos) );
	            AllocateSpace( t->mem, &sfnt1, pos, tableLength, &maxOutSize );
	            if(!MTX_TABLE_COMPRESS_Add(t, uncPos, tableLength, pos, 0))
	                return t->mem->errorCode;
	            uncPos += tableLength;
	            memcpy( (CHAR *)&sfnt1[pos], tableStart, tableLength );
	            /* printf("=== writing table 0x%x(%.4s) at idx %d: old/new pos=%d,%d\n",
	                tag, (CHAR *)&tag, i, pos, pos+tableLength); /* */
	            pos += tableLength;
	            
	            memcpy( (CHAR *)&sfnt1[pos], (CHAR *)&zero, 4 );	/* Zero Pad */
	            pos += 3; 
	            pos &= ~3;	/* Long Word Align */
	            uncPos += 3; 
	            uncPos &= ~3;	/* Long Word Align */
	            assert( pos < maxOutSize );
	        }
	    }	/*** END for-tables-in-font ***/
	}	/*** END for-fonts-in-TTC ***/

	/* Pass two through the fonts we process only compressed tables.*/
	for (idxTTC=0; idxTTC <	numInTTC; idxTTC++)
	{
	    if (isTTC)
	    {
	        fontOff0 = SWAPENDS_32( GET_ttc_font_offset( sfnt0, idxTTC ) );
	        fontOff1 = SWAPENDS_32(GET_ttc_font_offset(sfnt1, idxTTC));
	        printf("+++ processing font # %d: offset(in)=%d, (out)=%d\n", idxTTC, fontOff0, fontOff1);
	    }
	    
	    /* Loop runs once for each table in the current font */
	    numTables = SWAPENDS_U16(GET_sfnt_numOffsets( sfnt0 + fontOff0 ));
	    for ( i = 0; i < numTables; i++)
	    {
	        
	        tag = SWAPENDS_32(GET_sfnt_table_tag( sfnt0 + fontOff0, i));
	
	        
	        if(MTX_TABLE_COMPRESS_ShouldCompress(t, tag))
	        {
	            
	            /* Tolerate zero length tables */
	            if ( GET_sfnt_table_length( sfnt1 + fontOff1, i ) == 0 ) 
	                continue; /******/
	            
	            if (CheckForRepeatedTable(sfnt0, sfnt1, i, idxTTC))
	                continue; /******/
	            SET_sfnt_table_offset( sfnt1 + fontOff1, i, SWAPENDS_32(uncPos) );
	            /* Get the input table's start, offset and length */
	            tableOffset = SWAPENDS_U32(GET_sfnt_table_offset( sfnt0 + fontOff0, i ));
	            tableLength = SWAPENDS_U32(GET_sfnt_table_length( sfnt0 + fontOff0, i ));
	            tableStart 	= (CHAR *)&sfnt0[tableOffset];
	            /* Remember where the glyf table was */
	            if(tag == tag_GlyphData)
	            {
	                t->glyf_OffsetStart = uncPos;
	                t->glyf_OffsetEnd   = uncPos + tableLength;
	            }
				status = MTX_RA_PackToRAM( t->mem, tableLength, tableStart,
	                &packedLength, &packedData);
	            if(status)
	                return status;
	            
	            AllocateSpace( t->mem, &sfnt1, pos, packedLength, &maxOutSize );
	            if(!MTX_TABLE_COMPRESS_Add(t, uncPos, tableLength, pos, (SHORT)BYTE_CACHE_INTERVAL))
	                return t->mem->errorCode;
	            memcpy( (CHAR *)&sfnt1[pos], 
	                packedData, 
	                packedLength );
	            /* printf("=== writing table 0x%x(%.4s) at idx %d: old/new pos=%d,%d\n",
	                tag, (CHAR *)&tag, i, pos, pos+tableLength); /* */
	            pos += packedLength;
	            uncPos += tableLength;
	            MTX_mem_free(t->mem, packedData);
	            /* Put in pad bytes */
	            memcpy( (CHAR *)&sfnt1[pos], (CHAR *)&zero, 4 );	/* Zero Pad */
	            pos += 3; 
	            pos &= ~3;	/* Long Word Align */
	            uncPos += 3; 
	            uncPos &= ~3;	/* Long Word Align */
	            assert( pos < maxOutSize );
	        }
	    }	/*** END for-tables-in-font ***/
	}	/*** END for-fonts-in-TTC ***/

	MTX_TABLE_COMPRESS_SortAndCombine(t);

	/*------------------------ Write out ACT table -----------------*/
	tableLength = (ACTheaderSize + ACTentrySize*t->numEntries + 3) & ~3;
    AllocateSpace( t->mem, &sfnt1, pos, tableLength, &maxOutSize );
    t->sfnt1 = sfnt1;
	MTX_TABLE_COMPRESS_WriteHeader(t, pos);
	/* Add ACT table to the first font's directory */            
	if (isTTC)
		fontOff1 = SWAPENDS_32(GET_ttc_font_offset(sfnt1, 0));
	/* else fontOff1 is 0 */
	insertTable(sfnt1 + fontOff1, tag_ACT, pos, tableLength);
	pos += tableLength;
    
    return 0;	/* success */
}

int MTX_TABLE_COMPRESS_Compress(CHAR* nameIn, CHAR* nameOut)
{
	int bSuccess = false;
	int ierrorNumber = 0;
	static sfnt_TableTag tagsToCompress = tag_GlyphData;
	MTX_MemHandler *mem;

	/* try */
	if ( mem = MTX_mem_Create( malloc, realloc, free) )
	{
		MTX_TABLE_COMPRESS *t;
		mem->errorCode = 0;

		if ( t = MTX_TABLE_COMPRESS_Create( mem ) )
		{
	        t->numTags = 1;
	        t->tagsToCompress = &tagsToCompress;
	
			if ( t->sfnt0 = (CHAR *)RA_ReadFileIntoMemory( mem, nameIn, &t->sizeIn ) )
			{
				printf("Size IN = %ld\n", t->sizeIn );
		        t->sizeIn += 16;  /* to allow for the act3 table directory entry */
		        if (t->sfnt1 = MTX_mem_malloc(mem, t->sizeIn)) /* should be big enough but we grow it if not */
				{
			        if(!CompressTables(t))
					{
				        /* Write output file */
				        RA_WriteDataToFile( mem, nameOut, t->sfnt1, t->sizeOut );
				        if (!(mem->errorCode))
						{
							printf("Size OUT = %ld\n", t->sizeOut );
					
							MTX_mem_free( mem, t->sfnt0 );
							MTX_mem_free( mem, t->sfnt1 );
					        MTX_TABLE_COMPRESS_Destroy( t );
							MTX_mem_CloseMemory( mem );

					        if(mem->errorCode)
					            printf("******* Error: errorCode %d\n", mem->errorCode);

							MTX_mem_Destroy( mem );
							bSuccess = true;
						}
					}
				}
			}
		}
		if (!bSuccess) ierrorNumber = mem->errorCode;
	}
	else ierrorNumber = ERR_mem_Create_Failed;
	
	if (ierrorNumber)
	{
		if ( mem != NULL ) {
			MTX_mem_FreeAllMemory(mem);
			MTX_mem_Destroy( mem );
		}
		RA_DisplayMessage( ierrorNumber );
	}
	return bSuccess;
}

