/************************* MPEG-2 NBC Audio Decoder **************************
 *                                                                           *
"This software module was originally developed by 
Bernd Edler and Hendrik Fuchs, University of Hannover in the course of 
development of the MPEG-2 NBC/MPEG-4 Audio standard ISO/IEC 13818-7, 
14496-1,2 and 3. This software module is an implementation of a part of one or more 
MPEG-2 NBC/MPEG-4 Audio tools as specified by the MPEG-2 NBC/MPEG-4 
Audio standard. ISO/IEC  gives users of the MPEG-2 NBC/MPEG-4 Audio 
standards free license to this software module or modifications thereof for use in 
hardware or software products claiming conformance to the MPEG-2 NBC/MPEG-4
Audio  standards. Those intending to use this software module in hardware or 
software products are advised that this use may infringe existing patents. 
The original developer of this software module and his/her company, the subsequent 
editors and their companies, and ISO/IEC have no liability for use of this software 
module or modifications thereof in an implementation. Copyright is not released for 
non MPEG-2 NBC/MPEG-4 Audio conforming products.The original developer
retains full right to use the code for his/her  own purpose, assign or donate the 
code to a third party and to inhibit third party from using the code for non 
MPEG-2 NBC/MPEG-4 Audio conforming products. This copyright notice must
be included in all copies or derivative works." 
Copyright(c)1996.
 *                                                                           *
 ****************************************************************************/

/***********************************************************************************
 * MONOPRED									   *
 *										   *
 *	Contains the core functions for an intra channel (or mono) predictor	   *
 *	using a backward adaptive lattice predictor.				   *
 *										   *
 *	init_pred_stat():	initialisation of all predictor parameters	   *
 *	monopred():		calculation of a predicted value from		   *
 *				preceeding (quantised) samples		 	   *
 *	predict():		carry out prediction for all spectral lines	   *
 *	predict_reset():	carry out cyclic predictor reset mechanism	   *
 *				(long blocks) resp. full reset (short blocks)	   *
 *										   *
 *	Internal Functions:							   *
 *	  reset_pred_state():	reset the predictor state variables		   *
 *										   *
 **********************************************************************************/

#include "all.h"

#define GRAD	PRED_ORDER
static	float	ALPHA;
static	float	A;
static	float	B;


/********************************************************************************
 *** FUNCTION: reset_pred_state()						*
 ***										*
 ***    reset predictor state variables						*
 ***										*
 ********************************************************************************/
void
reset_pred_state(PRED_STATUS *psp)
{
    int m, grad;
    grad = GRAD;
    for(m=0;m<=grad;m++)	psp->r[m] = 0.;
    for(m=0;m<=grad;m++)	psp->kor[m] = 0.;
    for(m=0;m<=grad;m++)	psp->var[m] = 1.;
    for(m=0;m<=grad;m++)	psp->k[m] = 0.;
}


/********************************************************************************
 *** FUNCTION: init_pred_stat()							*
 ***										*
 ***    initialisation of all predictor parameter				*
 ***										*
 ********************************************************************************/
void
init_pred_stat(PRED_STATUS *psp, int grad, float alpha, float a, float b)
{
    /* Test of parameters */

    if(grad<0 || grad>MAX_PGRAD) {
	PRINT(SE,"\n\n ****** error in routine init_pred_stat ******\n");
	PRINT(SE,"\nwrong predictor order: %d\n",grad);
	PRINT(SE,"range of allowed values: 0 ... %d (=MAX_PGRAD)\n\n",MAX_PGRAD);
	exit(1);
    }
    if(alpha<0 || alpha>=1) {
	PRINT(SE,"\n\n ****** error in routine init_pred_stat ******\n");
	PRINT(SE,"\nwrong time constant alpha: %e\n",alpha);
	PRINT(SE,"range of allowed values: 0 ... 1\n\n");
	exit(1);
    }
    if(a<0 || a>1) {
	PRINT(SE,"\n\n ****** error in routine init_pred_stat ******\n");
	PRINT(SE,"\nwrong attenuation factor a: %e\n",a);
	PRINT(SE,"range of allowed values: 0 ... 1\n\n");
	exit(1);
    }
    if(b<0 || b>1) {
	PRINT(SE,"\n\n ****** error in routine init_pred_stat ******\n");
	PRINT(SE,"\nwrong attenuation factor b: %e\n",b);
	PRINT(SE,"range of allowed values: 0 ... 1\n\n");
	exit(1);
    }

    /* Initialisation */
    psp->grad = grad;
    psp->alpha = alpha;
    psp->a = a;
    psp->b = b;

    ALPHA = alpha;
    A = a;
    B = b;

    reset_pred_state(psp);
}




/********************************************************************************
 *** FUNCTION: monopred()							*
 ***										*
 ***    calculation of a predicted value from preceeding (quantised) samples	*
 ***	using a second order backward adaptive lattice predictor with full	*
 ***	LMS adaption algorithm for calculation of predictor coefficients	*
 ***										*
 ***    parameters:	lqs:	last quantised sample				*
 ***			psp:	pointer to structure with predictor status	*
 ***										*
 ***    returns:	predicted value						*
 ********************************************************************************/

Float
monopred(Float lx, PRED_STATUS *psp)
{
    float pv;		/* predicted value */
    float dr1;		/* difference in the R-branch */
    float e0,e1;		/* "partial" prediction errors (E-branch) */
    float r0,r1;		/* content of delay elements */
    float k1,k2;		/* predictor coefficients */

    float *R = psp->r;		/* content of delay elements */
    float *KOR = psp->kor;	/* estimates of correlations */
    float *VAR = psp->var;	/* estimates of variances */
    float *K = psp->k;		/* predictor coefficients */

    r0=R[0];
    r1=R[1];
    k1=K[1];
    k2=K[2];

    /* E-Branch:
     *	Calculate the partial prediction errors using the old predictor coefficients
     *	and the old r-values stored in the registers in order to reconstruct the
     *	predictor status of the previous step (instead of storage in state variables!)
     */

    e0 = lx;
    e1 = e0-k1*r0;


    /* Difference in the R-Branch:
     *	Calculate the difference in the R-Branch using the old predictor coefficients and
     *	the old partial prediction errors as calculated above in order to reconstruct the
   *	predictor status of the previous step (instead of storage in state variables!)
     */


    dr1 = k1*e0;

    /* Adaption of variances and correlations for predictor coefficients:
     *	These calculations are based on the predictor status of the previous step and give
     *	the new estimates of variances and correlations used for the calculations of the
     *	new predictor coefficients to be used for calculating the current predicted value
     */

    VAR[1] = ALPHA*VAR[1]+(0.5F)*(r0*r0 + e0*e0);   /* float const */
    KOR[1] = ALPHA*KOR[1] + r0*e0;
    VAR[2] = ALPHA*VAR[2]+(0.5F)*(r1*r1 + e1*e1);   /* float const */
    KOR[2] = ALPHA*KOR[2] + r1*e1;

    /* Summation and delay in the R-Branch => new R-values */

    r1 = A*(r0-dr1);
    r0 = A*e0;

    /* Calculation of new predictor coefficients to be used for the 
     * calculation of the current predicted value
     */
    k1 = (VAR[1]>MINVAR) ? KOR[1]/VAR[1]*B : 0.0F;
    k2 = (VAR[2]>MINVAR) ? KOR[2]/VAR[2]*B : 0.0F;

    /* Predicted values (using the new predictor coefficients!) */
    pv  = k1*r0 + k2*r1;

    K[1]=k1;
    K[2]=k2;
    R[0]=r0;
    R[1]=r1;

    return(pv);
}




/********************************************************************************
 *** FUNCTION: predict()							*
 ***										*
 ***    carry out prediction for all allowed spectral lines			*
 ***										*
 ********************************************************************************/


void
predict(Info* info, int profile, int *lpflag, PRED_STATUS *psp, Float *prev_quant,
Float *coef)
{
    int j, k, b, to, flag0;
    short *top;

    if (profile != Main_Profile) {
	if (*lpflag != 0) {
	    PRINT(SE, "Prediction not allowed in this profile!\n");
	    myexit("profile");
	}
	else {
	    /* prediction calculations not required */
	    return;
	}
    }
    
    if (info->islong) {
	b = 0;
	k = 0;
	top = info->sbk_sfb_top[b];
	flag0 = *lpflag++;
	for (j = 0; j < MAX_PRED_SFB; j++) {
	    to = *top++;
	    if (flag0 && *lpflag++) {
		for ( ; k < to; k++) {
		    coef[k] += monopred(prev_quant[k], &psp[k]);
		}
	    }
	    else {
		for ( ; k < to; k++) {
		    monopred(prev_quant[k], &psp[k]);
		}
	    }
	}
	fltcpy(prev_quant, coef, LN2);
    } 
}





/********************************************************************************
 *** FUNCTION: predict_reset()							*
 ***										*
 ***    carry out cyclic predictor reset mechanism (long blocks)		*
 ***    resp. full reset (short blocks)						*
 ***										*
 ********************************************************************************/
void
predict_reset(Info* info, int *prstflag, PRED_STATUS **psp, Float **prev_quant,
int firstCh, int lastCh)
{
    int j, prstflag0, prstgrp, ch;

    prstgrp = 0;
    if (info->islong) {
	prstflag0 = *prstflag++;
	if (prstflag0) {
	    for (j=LEN_PRED_RSTGRP-1; j>0; j--) {
		prstgrp |= prstflag[j];
		prstgrp <<= 1;
	    }
	    prstgrp |= prstflag[0];

	    if (debug['r']) {
		PRINT(SE,"PRST: prstgrp: %d  prstbits: ", prstgrp);
		for (j=LEN_PRED_RSTGRP-1; j>=0; j--)
		   PRINT(SE,"%d ", prstflag[j]);
		PRINT(SE,"FIRST: %d LAST %d\n", firstCh, lastCh);
	    }

	    if ( (prstgrp<1) || (prstgrp>30) ) {
		PRINT(SE, "ERROR in prediction reset pattern\n");
		return;
	    }

	    for (ch=firstCh; ch<=lastCh; ch++) {
		for (j=prstgrp-1; j<LN2; j+=30) {
		    reset_pred_state(&psp[ch][j]);
		    prev_quant[ch][j] = 0.0F;
		}
	    }
	} /* end predictor reset */
    } /* end islong */
    else { /* short blocks */
	/* complete prediction reset in all bins */
	for (ch=firstCh; ch<=lastCh; ch++) {
	    for (j=0; j<LN2; j++)
		reset_pred_state(&psp[ch][j]);
	    fltclr(prev_quant[ch], LN2);
	}
    }
}
