//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------

#pragma warning(disable:4786)

#include <stdio.h>
#include <stdlib.h>
#include "MPG_IfsQ_E.h"
#include "MPG_QFac.h"
#include "MPG_Global.h"
#include "MPG_VGraph.h"
#include "MPG_SPolygon.h"
#include "MPG_Stitcher.h"
#include "MPG_IfsMesh.h"
#include "MPG_MeshFs.h"

#ifndef M_PI
#define M_PI    3.14159265
#endif

#ifndef MPGELIB
extern char* baseFname;
extern char* inExt;
extern char* inDir;
extern char* outDir;
extern char* inFname;
extern char* outFname;
extern char* logFname;
#endif // MPGELIB

//---------------------------------------------------------------------------
//
//   MPG_IfsQ_E::expand
//
//---------------------------------------------------------------------------

bool MPG_IfsQ_E::expand(MPG_InputStream& istrm, bool saveWrlLevels)
{
  MPG_QFpm::_shift = 16;
	
  bool success = false;
	
  clear();
	
  // arrays to store quantized property values
	
  // AG: in 3D_Mesh_Object_Base_Layer
  // these arrays are global for per vertex properties
  // (containing all the values decoded so far),
  // and local otherwise
  // (containing only values corresponding to the current component)
	
  // GT: in 3D_Mesh_Object_Forest_Split
  // all these arrays are global
	
  vector <int> qCoord;
  vector <int> qNormal;
  vector <int> qColor;
  vector <int> qTexCoord;
	
  // create proper headers and quantizers
  // if not yet created by calling method
  MPG_Property pCoord(PT_COORD,PB_NONE);
  MPG_BBoxQ* coordQuantizer = getCoordQuantizer();
  if((MPG_BBoxQ*)0==coordQuantizer)
	{
		coordQuantizer    = new MPG_BBoxQ(3);
		setCoordQuantizer(coordQuantizer);
	}
  pCoord.setBBoxQ(coordQuantizer);
	
  MPG_Property pNormal(PT_NORMAL,PB_NONE);
  MPG_NormalQ* normalQuantizer = getNormalQuantizer();
  if((MPG_NormalQ*)0==normalQuantizer)
	{
		normalQuantizer   = new MPG_NormalQ();
		setNormalQuantizer(normalQuantizer);
	}
  pNormal.setNormalQ(normalQuantizer);
	
  MPG_Property pColor(PT_COLOR,PB_NONE);
  MPG_BBoxQ* colorQuantizer = getColorQuantizer();
  if((MPG_BBoxQ*)0==colorQuantizer)
	{
		colorQuantizer    = new MPG_BBoxQ(3);
		setColorQuantizer(colorQuantizer);
	}
  pColor.setBBoxQ(colorQuantizer);
	
  MPG_Property pTexCoord(PT_TEXCOORD,PB_NONE);
  MPG_BBoxQ* texCoordQuantizer = getTexCoordQuantizer();
  if((MPG_BBoxQ*)0==texCoordQuantizer)
	{
		texCoordQuantizer = new MPG_BBoxQ(2);
		setTexCoordQuantizer(texCoordQuantizer);
	}
  pTexCoord.setBBoxQ(texCoordQuantizer);
	
  // quantization parameters
  int bitsPerCoord    = 0;
  int bitsPerNormal   = 0;
  int bitsPerColor    = 0;
  int bitsPerTexCoord = 0;
  
  // input bitstream parsing --------------------------------------------------
	
  MPG_Stitcher stitcher;
  vector<int>  v_father;
  MPG_Stitches v_stitches;
  vector<int>  stk;  // stitch encoding stack
	
  vector<int>  v_visited;
  vector<int>  v_cluster;
  int          nV0    = 0; // number of vertices so far without current cc
  int          nV1    = 0; // number of vertices so far with current cc
  int          nVcls0 = 0; // AG: number of vertices after clustering
  int          nVcls1 = 0; // AG: number of vertices after clustering
	
	// open a bit stream
	MPG_InputBitStream ibstrm(istrm);
	
  // decode 3D_Mesh_Object() 
  //
  // 3D_Mesh_Object() 
  //   3D_MO_start_code
  //   3D_Mesh_Object_Header()
  //   do 
  //     3D_Mesh_Object_Layer()
  //   while(nextbits_bytealigned()==3D_MOL_start_code)

  // read 3D_MO_start_code
  if(ibstrm.getMOStartCodeBitInt(MPG_Global::MO_start_code)==false)
    return false;
	
  // decode 3D_Mesh_Object_Header()
  {
    // 3D_Mesh_Object_Header() 
    //   ccw
    //   convex
    //   solid
    //   creaseAngle
    //   coord_header()
    //   normal_header()
    //   color_header()
    //   texCoord_header()
    //   ce_SNHC
    //   if(ce_SNHC=='1')
    //     ce_SNHC_header()
		
	if(ibstrm.getMOBit(_ccw               )==false) return false;
    if(ibstrm.getMOBit(_convex            )==false) return false;
    if(ibstrm.getMOBit(_solid             )==false) return false;
    int cAng = 0;
	if(ibstrm.getMOBitInt(6,cAng)==false) return false;
    if(cAng<0 || cAng>=64) return false;
    _creaseAngle = M_PI*((float)cAng)/32;
		
    // decode coord_header()
    {
      // coord_header() 
      //   coord_binding       <--- must be == 1 (PER_VERTEX)
			
      int b = 0;
      if(ibstrm.getMBitInt(2,b)==false) return false;
      MPG_Binding binding =
        (b==0)?PB_NONE:(b==1)?PB_PER_VERTEX:(b==2)?PB_PER_FACE:PB_PER_CORNER;
      if(binding!=PB_NONE)
			{
				// coord can only be bound PER_VERTEX
				if(binding!=PB_PER_VERTEX) return false;
				pCoord.setBinding(binding);
				
				//   coord_bbox
				//   if(coord_bbox) 
				//     coord_xmin
				//     coord_ymin
				//     coord_zmin
				//     coord_size
				//   coord_quant
				//   coord_pred_type
				//   if (coord_pred_type=="tree_prediction" ||
				//       coord_pred_type=='"parallelogram_prediction") 
				//     coord_nlambda
				//     for(i=1;i<coord_nlambda;i++)
				//       coord_lambda[i]
				
				success = pCoord.decode(ibstrm);
				if(success==false) return false;
			}
    }
		
    // decode normal_header()
    {
      // normal_header() 
      //   normal_binding
      //   if(normal_binding!="not_bound") 
			
      int b = 0;
      if(ibstrm.getMBitInt(2,b)==false) return false;
      MPG_Binding binding =
        (b==0)?PB_NONE:(b==1)?PB_PER_VERTEX:(b==2)?PB_PER_FACE:PB_PER_CORNER;
      if(binding!=PB_NONE)
			{
				pNormal.setBinding(binding);
				
				//     normal_bbox       <--- must be == 0
				//     normal_quant
				//     normal_pred_type
				//     if (normal_pred_type=="tree_prediction"||
				//         normal_pred_type=="parallelogram_prediction") 
				//       normal_nlambda
				//       for(i=1;i<normal_nlambda;i++)
				//         normal_lambda[i]
				
				success = pNormal.decode(ibstrm);
				if(success==false) return false;
			}
      _normalPerVertex = (binding==PB_PER_FACE)?false:true;
    }
		
    // decode color_header()
    {
      // color_header()
      //   color_binding
      //   if(color_binding!="not_bound")
			
      int b = 0;
      if(ibstrm.getMBitInt(2,b)==false) return false;
      MPG_Binding binding =
        (b==0)?PB_NONE:(b==1)?PB_PER_VERTEX:(b==2)?PB_PER_FACE:PB_PER_CORNER;
      if(binding!=PB_NONE)
			{
				pColor.setBinding(binding);
				
				//     color_bbox
				//     if(color_bbox == '1') 
				//       color_rmin
				//       color_gmin
				//       color_bmin
				//       color_size
				//     color_quant
				//     color_pred_type
				//     if (color_pred_type=="tree_prediction" ||
				//         color_pred_type=='"parallelogram_prediction")
				//       color_nlambda
				//       for(i=1;i<color_nlambda;i++)
				//         color_lambda[i]
				
				success = pColor.decode(ibstrm);
				if(success==false) return false;
			}
      _colorPerVertex = (binding==PB_PER_FACE)?false:true;
    }
		
    // decode texture_header()
    {
      // texCoord_header() 
      //   texCoord_binding     <--- must be != 2 (PER_FACE)
      //   if(texCoord_binding!="not_bound")
			
      int b = 0;
      if(ibstrm.getMBitInt(2,b)==false) return false;
      MPG_Binding binding =
        (b==0)?PB_NONE:(b==1)?PB_PER_VERTEX:(b==2)?PB_PER_FACE:PB_PER_CORNER;
      if(binding!=PB_NONE)
			{
				// TexCoord cannot be bound per_face
				if(binding==PB_PER_FACE) return false;
				pTexCoord.setBinding(binding);
				
				//     texCoord_bbox
				//     if(texCoord_bbox=='1')
				//       texCoord_umin
				//       texCoord_vmin
				//       texCoord_size
				//     texCoord_quant
				//     texCoord_pred_type
				//     if (texCoord_pred_type=="tree_prediction" ||
				//         texCoord_pred_type=="parallelogram_prediction") 
				//       texCoord_nlambda
				//       for(i=1;i<texCoord_nlambda;i++)
				//         texCoord_lambda[i]
				
				success = pTexCoord.decode(ibstrm);
				if(success==false) return false;
			}
    }
		
    // decode cgd_data
    {
      _cgd_data = false;
      if(ibstrm.getBit(_cgd_data)==false) return false;
			
      if(_cgd_data==true)
		return false;	//MJH 000817: removed cgd part
    }
		
    // flush input buffer
    ibstrm.flush();
    
  } // end decode 3D_Mesh_Object_Header()
	
  bitsPerCoord    =  pCoord.getBBoxQ()->getBitsPerCoordinate();
  bitsPerNormal   =  pNormal.getNormalQ()->getBitsPerNormal();
  bitsPerColor    =  pColor.getBBoxQ()->getBitsPerCoordinate();
  bitsPerTexCoord =  pTexCoord.getBBoxQ()->getBitsPerCoordinate();
	
  // decode one or more 3D_Mesh_Object_Layer()
  int mol_id = 0;
  do {
    // decode 3D_Mesh_Object_Layer()
		
    // 3D_Mesh_Object_Layer()
		//   3D_MOL_start_code
    //   mol_id
    //   if(ce_SNHC =='1')
    //     ce_SNHC_n_vertices
    //     ce_SNHC_n_triangles
    //     ce_SNHC_n_edges
    //   if(mol_id==0)
    //     3D_Mesh_Object_Base_Layer()
    //   else
    //     3D_Mesh_Object_Refinement_Layer()
		
    // read 3D_MOL_start_code
    if(istrm.getStartCode(MPG_Global::MOL_start_code)==false)
      return false;
		
    {
      // open a bit stream
      MPG_InputBitStream ibstrm(istrm);
			
      // read mol_id
      if(ibstrm.getMBitInt(8,mol_id)==false) return false;
			
      // read cgd data
      if(_cgd_data==true)
		  return false; //MJH 000817: removed cgd part
      
      ibstrm.flush();
    }
		
    if(mol_id==0)
		{
			//
			// decode 3D_Mesh_Object_Base_Layer()
			//
			
			// 3D_Mesh_Object_Base_Layer() 
			//	do 
			//    3D_MOBL_start_code
			//    mobl_id
			//		while (!bytealigned())
			//			one_bit
			//		qf_start()
			//		if (3D_MOBL_start_code == "partition_type_0") 
			//			...
			//		else if (3D_MOBL_start_code == "partition_type_1") 
			//			...
			//		else if (3D_MOBL_start_code == "partition_type_2") 
			//			...
			//	while (nextbits_bytealigned() == 3D_MOBL_start_code)
			
			int previous_mobl_id = -1; // Index of previous MOBL; used to check correct sequencing
			int vg_number = 0;
			int nVcc;	
			
			typedef vector<int> vec_array;
			vector<vec_array> v_loop_cc;
			vector<int> nVtotal;	// for pt1
			MPG_SPolygon sp;	//MJH 001006 : moved here for library generation			

			do {  // 3D_Mesh_Object_Base_Layer()
				// read 3D_MOBL_start_code
				// get mobl start code and determine the partition type
				
synchronize:
			int partition_type;
			unsigned char mobl_sc = istrm.getMOBLStartCode(); 
			if (mobl_sc == MPG_Global::MOBL0_start_code)
				partition_type = 0;
			else if (mobl_sc == MPG_Global::MOBL1_start_code)
				partition_type = 1;
			else if (mobl_sc == MPG_Global::MOBL2_start_code)
				partition_type = 2;
			else
				break;
			
			// open a bit stream
			MPG_InputBitStream ibstrm(istrm);     
			
			// read mobl_id
			int mobl_id = 0;
			if(ibstrm.getMBitInt(8,mobl_id)==false) 
			{
#ifndef MPGELIB	//MJH 081800
				fprintf(stderr, "Error in mobl_id\n");
#endif
				goto synchronize;	
			}
//			else
//				printf("decoding mobl : %d\n", mobl_id);
			
			int correct_sequencing = (mobl_id == ((previous_mobl_id + 1) & 0xff));
			//Check for partition loss error
			if (!correct_sequencing) {
#ifndef MPGELIB //MJH 081800
				printf("Warning: partition number sequence incomplete. ");
				printf("(%d, %d)\n", previous_mobl_id, mobl_id);
#endif
			}
			previous_mobl_id = mobl_id;
			
			ibstrm.flush();
			
			//  start the arithmetic codec
			MPG_QFac codec(istrm);
			codec.start_decode();
			MPG_Global::reset();
			
			///////////////////////////////////////////////////////////////
			// partition type 0 ( one or more cc's)
			//	if (3D_MOBL_start_code == "partition_type_0") 
			//		do 
			//			connected_component()
			//			qf_decode(last_component, last_component_context)
			//		while (last_component == '0')
			//
			if (partition_type == 0) {	
				vg_number=0;	
				
				do {
					// decode connected_component()
					
					// connected_component() 
					//   vertex_graph()
					//   qf_decode(has_stitches, has_stitches_context)
					//   if(has_stitches=='1')
					//     stitches()
					//   triangle_tree()
					//   triangle_data()
					// 
					
					// decode vertex_graph() and construct vertex loop
					nV0 = nV1;	
					v_loop_cc.clear(); 
					vec_array v_loop_cc0;
					v_loop_cc.push_back(v_loop_cc0);
					{
						// decode vertex_graph()
						MPG_VGraph vg;
						if(vg.decode(codec) == false) return false;
						
						nVcc = vg.getNumberOfVertices();
						nV1 = nV0 + nVcc;
						while(v_visited.size()<nV1)
							v_visited.push_back(false);
						
						// convert to vertex loop look-up table
						MPG_Loops loop;
						vg.toLoop(loop);
						loop.getLoop(0, v_loop_cc[0]);
						
						if(vg.getFather(v_father)==false || v_father.size()!=nV1)
							return false;
					}
					
					v_stitches.decode(nV0,nV1,stk,codec);
					{
						// use stitches nV0<=i<nV1 to partially recover v_cluster
						
						if(stitcher.use(nV0,v_father,v_stitches,v_cluster)==false)
							return false;
					}
					
					// we should have here :
					// v_cluster.size()==nV1
					
					if(v_cluster.size()!=nV1)
						return false;
					
					// for i=0,..,nV1-1
					//   v_cluster[i]>=0 && v_cluster[i]<nV1
					// v_father.size()==nV1 
					// for i=0,..,nV1-1
					//   v_father[i]<=i
					// v_stitches.size()==nV1
					
					// decode triangle_tree()
//					MPG_SPolygon sp;	//MJH 001006 : moved to top for library generation
					sp.clear();			//MJH 001006 : added for library generation

					sp.codap_left_bloop_idx = 1;	
					sp.codap_right_bloop_idx = 0; 
					sp.codap_bdry_pred = 1;
					
					if(sp.decodeTriangleTree(codec,partition_type,v_loop_cc[0].size()) == false)	
						goto synchronize;
					
					// decode triangle_data()
					vector<int>   coordIndex;
					if(sp.decodeTriangleData(codec,
						nVcc,
						nV0,
						v_loop_cc[0],
						v_visited,
						v_cluster,
						coordIndex,
						// property headers and quantizers
						pCoord,
						pNormal,
						pColor,
						pTexCoord,
						// quantized values
						qCoord,
						qNormal,
						qColor,
						qTexCoord,
						partition_type	
						)==false) 
						goto synchronize;	
					
					// transfer coordIndex to Ifs
					{
						int j0,j1,j;
						for(j0=j1=0;j1<coordIndex.size();j1++)
							if(coordIndex[j1]==-1)	
							{
								int nf = j1-j0;
//								bool skip_poly = 0;		//MJH 000111 : skip_poly related not needed for pt0
								int* f = &(coordIndex[j0]);
								for(j=0;j<nf;j++)
								{
//									if (f[j] == -2)
//										skip_poly = 1;
//									if (skip_poly)
//										_coordIndex.push_back(-2);
//									else
										_coordIndex.push_back(f[j]);
								}
								_coordIndex.push_back(-1);
//								skip_poly = 0;
								j0 = j1+1;
							}
					}
					
			} while(codec.decode(MPG_Global::last_component) == false);	
			
			nVtotal.clear();
			nVtotal.push_back(nV0);
			nVtotal.push_back(nV1);	
		}
		
		///////////////////////////////////////////////////////////////
		// partition type 1 ( one or more vg's)
		//	else if (3D_MOBL_start_code == "partition_type_1") 
		//		vg_number=0
		//		do 
		//			vertex_graph()
		//			vg_number++
		//			qf_decode(has_stitches, has_stitches_context)
		//			if (has_stitches == '1')
		//				stitches()
		//			qf_decode(codap_last_vg, codap_last_vg_context)
		//		while (codap_last_vg == '0')
		//		
		else if (partition_type == 1) {	
			vg_number=0;	
			v_loop_cc.clear(); 
			nVtotal.clear();
			nVtotal.push_back(nV1);
			int codap_last_vg;
			
			do {
				nV0 = nV1;	
				
				vec_array v_loop_cc1;
				v_loop_cc.push_back(v_loop_cc1);
				
				// decode vertex_graph()
				MPG_VGraph vg;
				if(vg.decode(codec) == false) return false;
				
				nVcc = vg.getNumberOfVertices();
				nV1 = nV0 + nVcc;
				
				nVtotal.push_back(nV1);
				
				while(v_visited.size()<nV1)
					v_visited.push_back(false);
				
				// convert to vertex loop look-up table
				MPG_Loops loop;
				vg.toLoop(loop);
				loop.getLoop(0, v_loop_cc[vg_number]);
				
				if(vg.getFather(v_father)==false || v_father.size()!=nV1)
				{
					return false;
				}
				
				v_stitches.decode(nV0,nV1,stk,codec);
				{
					// use stitches nV0<=i<nV1 to partially recover v_cluster
					
					if(stitcher.use(nV0,v_father,v_stitches,v_cluster)==false)
						return false;
				}
				
				// we should have here :
				// v_cluster.size()==nV1
				
				if(v_cluster.size()!=nV1)
					return false;
				
				// for i=0,..,nV1-1
				//   v_cluster[i]>=0 && v_cluster[i]<nV1
				// v_father.size()==nV1 
				// for i=0,..,nV1-1
				//   v_father[i]<=i
				// v_stitches.size()==nV1
				
				vg_number++;	
				codap_last_vg = codec.decode(MPG_Global::codap_last_vg_context);
				
			} while (!codap_last_vg);
			
		} // if (partition_type == 1)
		
		///////////////////////////////////////////////////////////////
		// partition type 2 (one tt/td pair)
		//	else if (3D_MOBL_start_code == "partition_type_2") 
		//		if(vg_number > 1)
		//			qf_decode(codap_vg_id)
		//		qf_decode(codap_left_bloop_idx)
		//		qf_decode(codap_right_bloop_idx)
		//		qf_decode(codap_bdry_pred)
		//		triangle_tree()
		//		triangle_data()
		//		
		else { // if (partition_type == 2)
			
			// decode triangle_tree()
//			MPG_SPolygon sp;	//MJH 001006 : moved to top for library generation
			sp.clear();			//MJH 001006 : added for library generation
			
			int l,nbits;
			int codap_vg_id = 0;
			if (vg_number>1) {
				nbits = slog2(vg_number);
				for (l=0; l<nbits; l++)
					codap_vg_id += codap_vg_id + codec.decode();
			}
			nbits = slog2(v_loop_cc[codap_vg_id].size());
			sp.codap_left_bloop_idx = 0;	
			for (l=0; l<nbits; l++)
				sp.codap_left_bloop_idx += sp.codap_left_bloop_idx + codec.decode();
			sp.codap_right_bloop_idx = 0; 
			for (l=0; l<nbits; l++)
				sp.codap_right_bloop_idx += sp.codap_right_bloop_idx + codec.decode();
			sp.codap_bdry_pred = codec.decode();
			
			// decode triangle_tree()
			if(sp.decodeTriangleTree(codec,partition_type,v_loop_cc[codap_vg_id].size()) == false) //return false;
				goto synchronize;
			
			nV0 = nVtotal[codap_vg_id];
			nVcc = nVtotal[codap_vg_id+1]-nVtotal[codap_vg_id];
			
			// decode triangle_data()
			vector<int>   coordIndex;
			if(sp.decodeTriangleData(codec,
				nVcc,
				nV0,
				v_loop_cc[codap_vg_id],
				v_visited,
				v_cluster,
				coordIndex,
				// property headers and quantizers
				pCoord,
				pNormal,
				pColor,
				pTexCoord,
				// quantized values
				qCoord,
				qNormal,
				qColor,
				qTexCoord,
				partition_type	
				)==false) 
				goto synchronize;	
			
			// transfer coordIndex to Ifs
			{
				int j0,j1,j;
				for(j0=j1=0;j1<coordIndex.size();j1++)
					if(coordIndex[j1]==-1)	
					{
						int nf = j1-j0;
						bool skip_poly = 0;		
						int* f = &(coordIndex[j0]);
						for(j=0;j<nf;j++)
						{
							if (f[j] == -2)
								skip_poly = 1;
							_coordIndex.push_back(f[j]);
						}
						if (skip_poly)
						{
							int corner_cnt;
							//MJH 052500 : _coordIndex.size()>0 added in case the first triangle in _coordIndex has missing vertex
							for (corner_cnt=0; _coordIndex.back() != -1 && _coordIndex.size()>0 
								; corner_cnt++)
								_coordIndex.pop_back();
							for ( ; corner_cnt>0; corner_cnt--)
								_coordIndex.push_back(-2);
						}
						_coordIndex.push_back(-1);
						skip_poly = 0;
						j0 = j1+1;
					}
			}
				  
		} // if (partition_type == 2)
		///////////////////////////////////////////////////////////////
		
		codec.finish_decode();
		
	} while (1);	// 3D_Mesh_Object_Base_Layer()
	
	
	int i0 = 0; 
	nV0 = 0;	
	
	// determine the new number of vertices after clustering
	{
		int i;
		for (i=nV0; i < nV1; i++)
			if (v_cluster[i] > nVcls1) nVcls1 = v_cluster[i];
	}
	int nVcls11 = nVcls1+1;
	
	// unquantize vertices between nVcls0 and nVcls1
	
	// coord properties
	vector<float> coord((size_t) (3*(nVcls11-nVcls0)),0.0);
	MPG_BBoxQ*   coordQuantizer    = pCoord.getBBoxQ();
	coordQuantizer->unQuantize(qCoord.begin()+3*nVcls0,
		coord.begin(),
		nVcls11-nVcls0);
	
	//  normal properties
	vector<float> normal;
	MPG_NormalQ* normalQuantizer   = pNormal.getNormalQ();
	if(pNormal.getBinding() != PB_NONE)
	{
		// special case for per-vertex binding,
		// the qNormal coordinates are global to allow reuse
		// of some vertices by components decoded afterwards
		if(pNormal.getBinding() == PB_PER_VERTEX)
		{
			// take only the trailing part of qNormal
			vector <int> qNormal1;
			int i;
			for (i=nVcls0;i<nVcls11;i++)
				qNormal1.push_back(qNormal[i]);
			
			normalQuantizer->unQuantize(qNormal1, normal);
		}
		else normalQuantizer->unQuantize(qNormal, normal);
	}
	//  color properties
	vector<float> color;
	MPG_BBoxQ*   colorQuantizer    = pColor.getBBoxQ();
	if(pColor.getBinding() != PB_NONE)
	{
		if(pColor.getBinding() == PB_PER_VERTEX)
		{
			color.erase(color.begin(),color.end());
			color.insert(color.end(),(size_t)(3*(nVcls11-nVcls0)),0.0);
			colorQuantizer->unQuantize(qColor.begin()+3*nVcls0,
				color.begin(),
				nVcls11-nVcls0);
		}
		else colorQuantizer->unQuantize(qColor, color);
	}
	//  texture properties
	vector<float> texCoord;
	MPG_BBoxQ*   texCoordQuantizer = pTexCoord.getBBoxQ();
	if(pTexCoord.getBinding() != PB_NONE)
	{
		if(pTexCoord.getBinding() == PB_PER_VERTEX)
		{
			texCoord.erase(texCoord.begin(),texCoord.end());
			texCoord.insert(texCoord.end(),
				(size_t)(2*(nVcls11-nVcls0)),0.0);
			texCoordQuantizer->unQuantize(qTexCoord.begin()+2*nVcls0,
				texCoord.begin(),
				nVcls11-nVcls0);
		}
		else
			texCoordQuantizer->unQuantize(qTexCoord, texCoord);
	}
	
	// transfer coord to Ifs
	{
		int i;
		for(i=0;i<coord.size();i++) 
			_coord.push_back(coord[i]);
	}
	
	if(pNormal.getBinding()==PB_PER_CORNER)
	{
		// adjust _normalIndex
		// normal.size() should be equal to
		// the number of new corners
		int i = i0;
		int j = _normal.size()/3;
		for(;i<_coordIndex.size();i++)
			if(_coordIndex[i]!=-1)
			{ _normalIndex.push_back(j); j++; }
			else	
			{ 
				if (_normalIndex.back()==-2)
				{
					while (_normalIndex.back() != -1)
						_normalIndex.pop_back();
				}
				else
					_normalIndex.push_back(-1); 
				  }
	}
	
	// transfer normal to Ifs
	{
		int i;
		for(i=0;i<normal.size();i++)
			_normal.push_back(normal[i]);
	}
	
	if(pColor.getBinding()==PB_PER_CORNER)
	{
		// adjust _colorIndex
		// color.size() should be equal to
		// the number of new corners
		int i = i0;
		int j = _color.size()/3;
		for(;i<_coordIndex.size();i++)
			if(_coordIndex[i]!=-1)
			{ _colorIndex.push_back(j); j++; }
			else
			{ 
				if (_colorIndex.back()==-2)
				{
					while (_colorIndex.back() != -1)
						_colorIndex.pop_back();
				}
				else
					_colorIndex.push_back(-1); 
			}
	}
	
	// transfer color to Ifs
	{
		int i;
		for(i=0;i<color.size();i++) 
			_color.push_back(color[i]);
	}
	
	if(pTexCoord.getBinding()==PB_PER_CORNER)
	{
		// adjust _texCoordIndex
		// texCoord.size() should be equal to
		// the number of new corners
		int i = i0;
		int j = _texCoord.size()/2;
		for(;i<_coordIndex.size();i++)
			if(_coordIndex[i]!=-1)
			{ _texCoordIndex.push_back(j); j++; }
			else
			{ 
				if (_texCoordIndex.back()==-2)
				{
					while (_texCoordIndex.back() != -1)
						_texCoordIndex.pop_back();
				}
				else
					_texCoordIndex.push_back(-1); 
			}
	}
	
	// transfer texCoord to Ifs
	{
		int i;
		for(i=0;i<texCoord.size();i++)
			_texCoord.push_back(texCoord[i]);
	}

#ifdef MPGELIB
	//MJH 000112 : added to change every -2 index to 0
	{
		for(int i=0;i<_coordIndex.size();i++)
			if (_coordIndex[i] == -2)
				_coordIndex[i] = 0;
	}
#endif

	nVcls0 = nVcls11;
	nV0 = nV1;
    }
		
    else // if(mol_id!=0)
		return false;	//MJH 000817: removed cgd part
	
	
#ifndef MPGELIB
	if(saveWrlLevels==true)
      {
        int n = strlen(outDir)+strlen(baseFname)+16;
        outFname = new char[n];
        if(!strcmp(outDir,"."))
          sprintf(outFname,"%s-E%02d.debug.vrml",baseFname,mol_id);
        else
          sprintf(outFname,"%s/%s-E%02d.debug.vrml",outDir,baseFname,mol_id);
        bool success = write(outFname);
        delete [] outFname;
        outFname = NULL;
      }
#endif // MPGELIB    

  } while(istrm.lookingAtStartCode(MPG_Global::MOL_start_code)==true);
  
  success = true;
  return success;
}

