//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------


#ifndef _MPG_Mesh_h_
#define _MPG_Mesh_h_

#include <vector>
#include "MPG_Stream.h"
#include "MPG_Graph.h"
#include "MPG_Partition.h"
#include "MPG_SPolygon.h"
#include "MPG_VGraph.h"

#ifdef _AIX
inline void destroy(vector<int>*) {}
inline void destroy(vector<int>**) {}
#endif // _AIX

using namespace std;	//MJH 010911

class MPG_Mesh {

  // the coordIndexArray given as an argument to the constructor
  // is treated in this class as READ-ONLY
  //
  // use the MPG_MeshEdit class for operations which modify
  // the values of the coordIndex array

protected:

  // FACES AND CONNECTIVITY

  int                    _nV;
  vector<int>            _face;
  vector<int>&           _coordIndex;	// passed as argument by constructor
  MPG_Graph*             _graph;
  MPG_Graph::pEdge       _curr_e;
  int                    _curr_i;
  int                    _nE_isolated;
  int                    _nE_boundary;
  int                    _nE_regular;
  int                    _nE_singular;
  int                    _nV_isolated;
  int                    _nV_regular;
  int                    _nV_singular;

  // CONNECTED COMPONENTS

  int                    _n_cc;	// total number of connected components
  int                    _i_cc;	// current connected component

  // the vertex indices (with respect to the global Map)
  // of the i-th. connected component are
  // { _v_cc[j] : _v_cc_beg[i] <= j < _v_cc_beg[i+1] }
  vector<int>            _v_cc_beg; // [_n_cc+1]
  vector<int>            _v_cc;     // [getNumberOfVertices()]

  // the face indices
  // of the i-th. connected component are
  // { _f_cc[j] : _f_cc_beg[i] <= j < _f_cc_beg[i+1] }
  vector<int>            _f_cc_beg; // [_n_cc+1]
  vector<int>            _f_cc;     // [getTotalNumberOfFaceIndices()]

  int                    _curr_i_cc;
  MPG_Graph::pEdge       _curr_e_cc;

protected:

  // index within own connected component, i.e.,
  // _inv_v_cc[getCoordIndexCC(cc,i)]==i
  vector<int>            _inv_v_cc; // same size as _v_cc
  // _inv_f_cc[getFaceIndexCC(cc,i)]==i
  vector<int>            _inv_f_cc; // same size as _f_cc

public:

  enum PartitionMode {
    UNKNWN=-1,
    ORIENTED_MANIFOLD,
    MANIFOLD,
    MARKED_EDGES
  };

protected:

  MPG_Partition         *_partitionCorners(PartitionMode pm =ORIENTED_MANIFOLD,
					   vector<bool> *emk=NULL);

         void            _createEdges();
         void            _classifyEdges();
         void            _classifyVertices();

public:

  inline                  MPG_Mesh(vector<int>& coordIndex);
  inline                 ~MPG_Mesh();
  inline void             clear();
         void             set(vector<int>& coordIndex);
  inline vector<int>&     getCoordIndex();
  inline void             reset();
         const bool       isOrientedManifold();
  inline const int        getNumberOfIsolatedEdges();
  inline const int        getNumberOfBoundaryEdges();
  inline const int        getNumberOfRegularEdges();
  inline const int        getNumberOfSingularEdges();
  inline const bool       hasIsolatedEdges();
  inline const bool       hasBoundaryEdges();
  inline const bool       hasRegularEdges();
  inline const bool       hasSingularEdges();
  inline const int        getNumberOfIsolatedVertices();
  inline const int        getNumberOfRegularVertices();
  inline const int        getNumberOfSingularVertices();
  inline const bool       hasIsolatedVertices();
  inline const bool       hasRegularVertices();
  inline const bool       hasSingularVertices();
  inline bool             isTriangulated();
  inline void             setNumberOfVertices(int nV);
  inline const int        getNumberOfVertices();
  inline const int        getNumberOfFaces();
  inline const int        getNumberOfTriangles();
  inline void             getFace(int i, int *fi);
  inline bool             getFace(int i, vector<int>& fi);
  inline bool             getFace(int i, vector<int>& fi, vector<int>& map);
  inline int*             getFace(int i);
  inline void             pushBackFace(int nC, int *indx);
  inline int*             operator[](int i);
  inline const int        getCorner(int i, int iv);
  inline int              getFaceCornerIndex(int i, int j=0);
  inline const int        getNumberOfFaceCorners(int i);
  inline const int        getNumberOfCorners();
  inline int              getFaceCorner(int i, int j);
  inline int              getNextFaceCorner(int i, int v_prev, int v_curr);
  inline int              getFaceCornerCurrPos(int fi, int vj);
  inline int              getFaceCornerPrevPos(int fi, int vj);
  inline int              getFaceCornerNextPos(int fi, int vj);
  inline int              getFacePrevCorner(int fi, int vj);
  inline int              getFaceNextCorner(int fi, int vj);
  inline const bool       hasEdges();
  inline void             deleteEdges();
  inline const int        getNumberOfEdges();
  inline MPG_Graph::pEdge getEdge(int i, int j);
  inline MPG_Graph::pEdge getFirstEdge(int i);
  inline MPG_Graph::pEdge getNextEdge(MPG_Graph::pEdge e);
  inline int              getNumberOfEdgeFaces(int eIndx);
  inline int              getNumberOfEdgeFaces(MPG_Graph::Edge& e);
  inline int              getEdgeFace(int eIndx, int i);
  inline int              getEdgeFace(int eIndx);
  inline int              getEdgeFace(MPG_Graph::Edge& e, int i);
  inline int              getEdgeFace(MPG_Graph::Edge& e);
  inline int              getOtherEdgeFace(int eIndx, int fi);
  inline int              getOtherEdgeFace(MPG_Graph::Edge& e, int fi);
  inline bool             isEdgeFace(int eIndx, int fi);
  inline bool             isEdgeFace(MPG_Graph::Edge& e, int fi);
  inline int              getEdgeVertex(MPG_Graph::Edge& e, int i);
  inline bool             isEdgeVertex(MPG_Graph::Edge& e, int vi);
  inline const bool       isUnknownEdge(int eIndx);
  inline const bool       isUnknownEdge(MPG_Graph::Edge& e);
  inline const bool       isIsolatedEdge(int eIndx);
  inline const bool       isIsolatedEdge(MPG_Graph::Edge& e);
  inline const bool       isBoundaryEdge(int eIndx);
  inline const bool       isBoundaryEdge(MPG_Graph::Edge& e);
  inline const bool       isRegularEdge(int eIndx);
  inline const bool       isRegularEdge(MPG_Graph::Edge& e);
  inline const bool       isManifoldEdge(int eIndx);
  inline const bool       isManifoldEdge(MPG_Graph::Edge& e);
  inline const bool       isSingularEdge(int eIndx);
  inline const bool       isSingularEdge(MPG_Graph::Edge& e);

         bool             findEdgeInFace(MPG_Graph::Edge *e,
					 int if0, int& if0v0, int &if0v1);
         bool             hasFacesWithDuplicateVertices();
         int              markBoundaryVertices(vector<bool>& isBdryVertex);
         int              markFacesWithDuplicateVertices(vector<bool>& fmk);

protected:

         bool            _createConnectedComponents();
  inline void            _deleteConnectedComponents();

public:

         int              computeNumberOfConnectedComponents();
  inline int              getNumberOfConnectedComponents();
  inline bool             isConnected();
  inline int              getNumberOfVerticesCC(int icc);
  inline int              getNumberOfVerticesCC();
  inline int              getCoordIndexCC(int cc, int i);
  inline int              getCoordIndexCC(int i);
  inline int              getInvCoordIndexCC(int i);
  inline int              getNumberOfFacesCC(int cc);
  inline int              getNumberOfTrianglesCC(int icc);
  inline int              getNumberOfTrianglesCC();
  inline int              getNumberOfCornersCC(int icc);
  inline int              getNumberOfCornersCC();
  inline int              getNumberOfFacesCC();
  inline int              getFaceIndexCC(int cc, int i);
  inline int              getFaceIndexCC(int i);
  inline int              getNumberOfFaceCornersCC(int cc, int i);
  inline int              getNumberOfFaceCornersCC(int i);
  inline int              getIndexCC();
  inline bool             setIndexCC(int icc);
  inline void             ccBegin();
  inline bool             ccEnd();
  inline void             ccNext();
  inline MPG_Graph::pEdge getEdgeCC(int icc, int j, int k);
  inline MPG_Graph::pEdge getFirstEdgeCC(int icc);
  inline MPG_Graph::pEdge getFirstEdgeCC();
         MPG_Graph::pEdge getNextEdgeCC(int icc);
  inline MPG_Graph::pEdge getNextEdgeCC();
  inline int              getNumberOfEdgesCC(int icc);
  inline int              getNumberOfEdgesCC();

         void             partitionVertices(int nD,
					    vector<float>& coord,
					    vector<int>&   coordIndex,
					    int            x_band,
					    int            y_band,
					    int            z_band,
					    int            n_band,
					    vector<bool>&  e_marks);

public:

  inline bool        smooth(vector<float>& x,
			    vector<float>& sx,
			    float          a,
			    float          b,
			    int            n,
			    vector<bool>*  e_sharp = NULL,
			    vector<bool>*  v_fixed = NULL);

};

// FACES AND EDGES ------------------------------------------------------------

inline MPG_Mesh::MPG_Mesh(vector<int>& coordIndex):
  _coordIndex(coordIndex),
  _graph((MPG_Graph*)0)
{ set(coordIndex); }

inline MPG_Mesh::~MPG_Mesh()
{ if(_graph!=(MPG_Graph*)0) delete _graph; }

inline vector<int>& MPG_Mesh::getCoordIndex()
{ return _coordIndex; }

inline void MPG_Mesh::clear()
{
  _nV = 0;
  if(_graph!=(MPG_Graph*)0) { delete _graph; _graph=(MPG_Graph*)0; }
  _curr_i = 0;
  _curr_e = (MPG_Graph::pEdge)0;
  _face.erase(_face.begin(),_face.end());
  _face.push_back(0);
  _deleteConnectedComponents();
}

inline void MPG_Mesh::reset()
{ set(_coordIndex); }

inline const int MPG_Mesh::getNumberOfIsolatedEdges()
{ return _nE_isolated; }

inline const int MPG_Mesh::getNumberOfBoundaryEdges()
{ return _nE_boundary; }

inline const int MPG_Mesh::getNumberOfRegularEdges()
{ return _nE_regular;  }

inline const int MPG_Mesh::getNumberOfSingularEdges()
{ return _nE_singular; }

inline const bool MPG_Mesh::hasIsolatedEdges()
{ return (_nE_isolated>0)?1:0; }

inline const bool MPG_Mesh::hasBoundaryEdges()
{ return (_nE_boundary>0)?1:0; }

inline const bool MPG_Mesh::hasRegularEdges()
{ return (_nE_regular >0)?1:0; }

inline const bool MPG_Mesh::hasSingularEdges()
{ return (_nE_singular>0)?1:0; }

inline const int MPG_Mesh::getNumberOfIsolatedVertices()
{ return _nV_isolated; }

inline const int MPG_Mesh::getNumberOfRegularVertices()
{ return _nV_regular;  }

inline const int MPG_Mesh::getNumberOfSingularVertices()
{ return _nV_singular; }

inline const bool MPG_Mesh::hasIsolatedVertices()
{ return (_nV_isolated>0)?1:0; }

inline const bool MPG_Mesh::hasRegularVertices()
{ return (_nV_regular >0)?1:0; }

inline const bool MPG_Mesh::hasSingularVertices()
{ return (_nV_singular>0)?1:0; }

inline bool MPG_Mesh::isTriangulated()
{
  bool is_triangulated = true;

  int nF = getNumberOfFaces();
  for(int i=0;i<nF;i++)
    if(getNumberOfFaceCorners(i)!=3)
      { is_triangulated = false; break; }

  return is_triangulated;
}

inline void MPG_Mesh::setNumberOfVertices(int nV)
{
  if(nV>getNumberOfVertices())
    {
      _nV = nV;
      if(_graph!=(MPG_Graph*)0)	{ delete _graph; _graph = (MPG_Graph*)0; }
      _curr_i = 0;
      _curr_e = (MPG_Graph::pEdge)0;
    }
}

inline const int MPG_Mesh::getNumberOfVertices()
{
  if(_nV==0 && _coordIndex.size()>0)
    {
      int i=0;
      while(_coordIndex[i]<0 && i<_coordIndex.size()) i++;
      if(i<_coordIndex.size())
	{
	  int min = _coordIndex[i];
	  int max = _coordIndex[i];
	  for(i++;i<_coordIndex.size();i++)
	    if(_coordIndex[i]>=0)
	      {
		if(_coordIndex[i]<min) min = _coordIndex[i];
		if(_coordIndex[i]>max) max = _coordIndex[i];
	      }
	  _nV = max+1;
	}
      else
	_nV=0;
    }
  return _nV;
}

inline const int MPG_Mesh::getNumberOfFaces()
{ return _face.size()-1; }

inline const int MPG_Mesh::getNumberOfTriangles()
{
  int nT = 0;
  int nF = getNumberOfFaces();
  for(int i=0;i<nF;i++)
    {
      int nfi = getNumberOfFaceCorners(i);
      if(nfi>2) nT += (nfi-2);
    }
  return nT;
}

inline void MPG_Mesh::getFace(int i, int *fi)
{
  int nC = getNumberOfFaceCorners(i);
  for(int j=0;j<nC;i++)
    fi[j] = getFaceCorner(i,j);
}

inline bool MPG_Mesh::getFace(int i, vector<int>& fi)
{
  int nF = getNumberOfFaces();
  fi.erase(fi.begin(),fi.end());
  int j,nfi,fij;
  nfi = getNumberOfFaceCorners(i);
  for(j=0;j<nfi;j++)
    {
      fij = getFaceCorner(i,j);
      if(j==0 || fij!=fi.back())
	fi.push_back(fij);
      // still need to split face with
      // duplicate vertices into two or more faces !!!
      // OK for triangular face, though
    }
  if(fi.front()==fi.back()) fi.pop_back();
  if(fi.size()<3)
    { fi.erase(fi.begin(),fi.end()); }
  return (fi.size()>=3)?true:false;
}

inline bool MPG_Mesh::getFace(int i, vector<int>& fi, vector<int>& map)
{
  int nF = getNumberOfFaces();
  fi.erase(fi.begin(),fi.end());
  int j,nfi,fij;
  nfi = getNumberOfFaceCorners(i);
  for(j=0;j<nfi;j++)
    {
      fij = getFaceCorner(i,j);
      fij = map[fij];
      if(j==0 || fij!=fi.back())
	fi.push_back(fij);
      // still need to split face with
      // duplicate vertices into two or more faces !!!
      // OK for triangular face, though
    }
  if(fi.front()==fi.back()) fi.pop_back();
  if(fi.size()<3)
    { fi.erase(fi.begin(),fi.end()); }
  return (fi.size()>=3)?true:false;
}

inline int* MPG_Mesh::getFace(int i)
{ return _coordIndex.begin()+_face[i]; }

inline void MPG_Mesh::pushBackFace(int nC, int *indx)
{
  for(int j=0;j<nC;j++)
    _coordIndex.push_back(indx[j]);
  _coordIndex.push_back(-1);
}

inline int* MPG_Mesh::operator[](int i)
{ return getFace(i); }

inline int MPG_Mesh::getFaceCornerIndex(int i, int j)
{ return _face[i]+j-i; }

inline const int MPG_Mesh::getNumberOfFaceCorners(int i)
{ return _face[i+1]-_face[i]-1; }

inline const int MPG_Mesh::getNumberOfCorners()
{
  int nC = 0;
  for(int i=0;i<getNumberOfFaces();i++)
    nC += getNumberOfFaceCorners(i);
  return nC;
}

inline int MPG_Mesh::getFaceCorner(int i, int j)
{ return _coordIndex[_face[i]+j]; }

inline const int MPG_Mesh::getCorner(int i, int iv)
{
  int corner = -1;
  int nfi = getNumberOfFaceCorners(i);
  int *fi = getFace(i);
  for(int j=0;j<nfi;j++)
    if(fi[j]==iv)
      { corner = j; break; }
  return corner;
}

inline int MPG_Mesh::getNextFaceCorner(int i, int v_prev, int v_curr)
{
  int c_curr = getCorner(i,v_curr);
  if(c_curr<0) return -1;
  int n_i = getNumberOfFaceCorners(i);
  int c_next =
    (v_prev==getFaceCorner(i,(c_curr+n_i-1)%n_i))?(c_curr+    1)%n_i:
    (v_prev==getFaceCorner(i,(c_curr+    1)%n_i))?(c_curr+n_i-1)%n_i:-1;
  if(c_next<0) return -1;
  return getFaceCorner(i,c_next);
}

int MPG_Mesh::getFaceCornerCurrPos(int fi, int vj)
{
  int nfi = getNumberOfFaceCorners(fi);
  for(int j=0;j<nfi;j++)
    if(getFaceCorner(fi,j)==vj)
      return(j);
  return -1;
}

int MPG_Mesh::getFaceCornerPrevPos(int fi, int vj)
{
  int nf = getNumberOfFaceCorners(fi);
  int  i = getFaceCornerCurrPos(fi,vj);
  return (i>=0)?(i+nf-1)%nf:-1;
}

int MPG_Mesh::getFaceCornerNextPos(int fi, int vj)
{
  int nf = getNumberOfFaceCorners(fi);
  int  i = getFaceCornerCurrPos(fi,vj);
  return (i>=0)?(i+1)%nf:-1;
}

int MPG_Mesh::getFacePrevCorner(int fi, int vj)
{ return getFaceCorner(fi,getFaceCornerPrevPos(fi,vj)); }

int MPG_Mesh::getFaceNextCorner(int fi, int vj)
{ return getFaceCorner(fi,getFaceCornerNextPos(fi,vj)); }

inline const bool MPG_Mesh::hasEdges()
{ return (_graph!=(MPG_Graph*)0)?true:false; }

inline void MPG_Mesh::deleteEdges()
{
  if(_graph!=(MPG_Graph*)0) { delete _graph; _graph=(MPG_Graph*)0; }
  _curr_i      = 0;
  _curr_e      = (MPG_Graph::pEdge)0;
  _nE_isolated = 0;
  _nE_boundary = 0;
  _nE_regular  = 0;
  _nE_singular = 0;
  _nV_isolated = 0;
  _nV_regular  = 0;
  _nV_singular = 0;
  _nV          = 0;
}

inline const int MPG_Mesh::getNumberOfEdges()
{
  _createEdges();
  return _graph->getNumberOfEdges();
}

inline MPG_Graph::pEdge MPG_Mesh::getEdge(int i, int j)
{ return (_graph)?_graph->getEdge(i,j):(MPG_Graph::pEdge)0; }

inline MPG_Graph::pEdge MPG_Mesh::getFirstEdge(int i)
{ return (_graph)?_graph->getFirstEdge(i):(MPG_Graph::pEdge)0; }

inline MPG_Graph::pEdge MPG_Mesh::getNextEdge(MPG_Graph::pEdge e)
{ return (_graph)?_graph->getNextEdge(e):(MPG_Graph::pEdge)0; }

inline int MPG_Mesh::getNumberOfEdgeFaces(int eIndx)
{ return (_graph)?_graph->getNumberOfEdgeFaces(eIndx):0; }

inline int MPG_Mesh::getNumberOfEdgeFaces(MPG_Graph::Edge& e)
{ return getNumberOfEdgeFaces(e.getIndex()); }

inline int MPG_Mesh::getEdgeFace(int eIndx, int i)
{ return (_graph)?_graph->getEdgeFace(eIndx,i):-1; }

inline int MPG_Mesh::getEdgeFace(int eIndx)
{ return (_graph)?_graph->getEdgeFace(eIndx):-1; }

inline int MPG_Mesh::getEdgeFace(MPG_Graph::Edge& e, int i)
{ return getEdgeFace(e.getIndex(),i); }

inline int MPG_Mesh::getEdgeFace(MPG_Graph::Edge& e)
{ return getEdgeFace(e.getIndex()); }

inline int MPG_Mesh::getOtherEdgeFace(int eIndx, int fi)
{ return (_graph)?_graph->getOtherEdgeFace(eIndx,fi):-1; }

inline int MPG_Mesh::getOtherEdgeFace(MPG_Graph::Edge& e, int fi)
{ return getOtherEdgeFace(e.getIndex(),fi); }

inline bool MPG_Mesh::isEdgeFace(int eIndx, int fi)
{ return (_graph)?_graph->isEdgeFace(eIndx,fi):false; }

inline bool MPG_Mesh::isEdgeFace(MPG_Graph::Edge& e, int fi)
{ return isEdgeFace(e.getIndex(),fi); }

inline int MPG_Mesh::getEdgeVertex(MPG_Graph::Edge& e, int i)
{ return e.getVertex(i); }

inline bool MPG_Mesh::isEdgeVertex(MPG_Graph::Edge& e, int vi)
{ return e.isVertex(vi); }

inline const bool MPG_Mesh::isUnknownEdge(int eIndx)
{ return (_graph)?_graph->isUnknownEdge(eIndx):true; }

inline const bool MPG_Mesh::isUnknownEdge(MPG_Graph::Edge& e)
{ return isUnknownEdge(e.getIndex()); }

inline const bool MPG_Mesh::isIsolatedEdge(int eIndx)
{ return (_graph)?_graph->isIsolatedEdge(eIndx):false; }

inline const bool MPG_Mesh::isIsolatedEdge(MPG_Graph::Edge& e)
{ return isIsolatedEdge(e.getIndex()); }

inline const bool MPG_Mesh::isBoundaryEdge(int eIndx)
{ return (_graph)?_graph->isBoundaryEdge(eIndx):false; }

inline const bool MPG_Mesh::isBoundaryEdge(MPG_Graph::Edge& e)
{ return isBoundaryEdge(e.getIndex()); }

inline const bool MPG_Mesh::isRegularEdge(int eIndx)
{ return (_graph)?_graph->isRegularEdge(eIndx):false; }

inline const bool MPG_Mesh::isRegularEdge(MPG_Graph::Edge& e)
{ return isRegularEdge(e.getIndex()); }

inline const bool MPG_Mesh::isManifoldEdge(int eIndx)
{ return (_graph)?_graph->isManifoldEdge(eIndx):false; }

inline const bool MPG_Mesh::isManifoldEdge(MPG_Graph::Edge& e)
{ return isManifoldEdge(e.getIndex()); }

inline const bool MPG_Mesh::isSingularEdge(int eIndx)
{ return (_graph)?_graph->isSingularEdge(eIndx):false; }

inline const bool MPG_Mesh::isSingularEdge(MPG_Graph::Edge& e)
{ return isSingularEdge(e.getIndex()); }

// CONNECTED COMPONENTS -------------------------------------------------------

inline void MPG_Mesh::_deleteConnectedComponents()
{
  _n_cc = 0;
  _i_cc = 0;
  _curr_i_cc = 0;
  _curr_e_cc = (MPG_Graph::pEdge)0;
  _v_cc_beg.erase(_v_cc_beg.begin(),_v_cc_beg.end());
  _v_cc.erase(_v_cc.begin(),_v_cc.end());
  _f_cc_beg.erase(_f_cc_beg.begin(),_f_cc_beg.end());
  _f_cc.erase(_f_cc.begin(),_f_cc.end());
}

inline int MPG_Mesh::getNumberOfConnectedComponents()
{
  if(_n_cc<=0) _createConnectedComponents();
  return _n_cc ;
}

inline bool MPG_Mesh::isConnected()
{ return (getNumberOfConnectedComponents()==1)?true:false; }

inline int MPG_Mesh::getNumberOfVerticesCC(int icc)
{
  return (0<=icc&&icc<getNumberOfConnectedComponents())?
    _v_cc_beg[icc+1]-_v_cc_beg[icc]:0;
}

inline int MPG_Mesh::getNumberOfVerticesCC()
{ return getNumberOfVerticesCC(_i_cc); }

inline int  MPG_Mesh::getCoordIndexCC(int cc, int i)
{ return (0<=i&&i<getNumberOfVerticesCC(cc))?_v_cc[_v_cc_beg[cc]+i]:-1; }

inline int  MPG_Mesh::getCoordIndexCC(int i)
{ return getCoordIndexCC(_i_cc,i); }

inline int  MPG_Mesh::getInvCoordIndexCC(int i)
{ return (0<=i&&i<_inv_v_cc.size())?_inv_v_cc[i]:-1; }

inline int  MPG_Mesh::getNumberOfFacesCC(int cc)
{
  return (0<=cc&&cc<getNumberOfConnectedComponents())?
    _f_cc_beg[cc+1]-_f_cc_beg[cc]:0;
}

inline int  MPG_Mesh::getNumberOfTrianglesCC(int icc)
{
  int nFcc = getNumberOfFacesCC(icc);
  int nTcc = 0;
  int i;
  for(i=0;i<nFcc;i++)
    nTcc += (getNumberOfFaceCornersCC(icc,i)-2);
  return nTcc;
}

inline int  MPG_Mesh::getNumberOfTrianglesCC()
{ return getNumberOfTrianglesCC(_i_cc); }

inline int MPG_Mesh::getNumberOfCornersCC(int icc)
{
  int i;
  int nCcc = 0;
  for(i=0;i<getNumberOfFacesCC(icc);i++)
    nCcc += getNumberOfFaceCornersCC(icc,i);
  return nCcc;
}

inline int  MPG_Mesh::getNumberOfCornersCC()
{ return getNumberOfCornersCC(_i_cc); }

inline int  MPG_Mesh::getNumberOfFacesCC()
{ return getNumberOfFacesCC(_i_cc); }

inline int  MPG_Mesh::getFaceIndexCC(int cc, int i)
{ return (0<=i&&i<getNumberOfFacesCC(cc))?_f_cc[_f_cc_beg[cc]+i]:-1; }

inline int MPG_Mesh::getFaceIndexCC(int i)
{ return getFaceIndexCC(_i_cc,i); }

inline int  MPG_Mesh::getNumberOfFaceCornersCC(int cc, int i)
{ return getNumberOfFaceCorners(getFaceIndexCC(cc,i)); }

inline int  MPG_Mesh::getNumberOfFaceCornersCC(int i)
{ return getNumberOfFaceCornersCC(_i_cc,i); }

inline int  MPG_Mesh::getIndexCC()
{ return _i_cc ; }

inline bool  MPG_Mesh::setIndexCC(int icc)
{
  if(0<=icc&&icc<_n_cc) _i_cc=icc;
  return (icc==_i_cc)?true:false;
}

inline void  MPG_Mesh::ccBegin()
{ _i_cc = 0; }

inline bool  MPG_Mesh::ccEnd()
{ return (_i_cc==_n_cc)?true:false; }

inline void  MPG_Mesh::ccNext()
{ if(_i_cc!=_n_cc) _i_cc++; }

inline MPG_Graph::pEdge MPG_Mesh::getEdgeCC(int icc, int j, int k)
{
  if(icc<0 || icc>=_n_cc || j==k ||
     j<0 || j>=getNumberOfVerticesCC(icc) ||
     k<0 || k>=getNumberOfVerticesCC(icc)) return (MPG_Graph::pEdge)0;

  int jcc = getCoordIndexCC(icc,j);
  int kcc = getCoordIndexCC(icc,k);

  return getEdge(jcc,kcc);
}

inline MPG_Graph::pEdge MPG_Mesh::getFirstEdgeCC(int icc)
{
  _curr_i_cc = -1; _curr_e_cc = (MPG_Graph::pEdge)0;
  for(int i=0;i<getNumberOfVerticesCC();i++)
    {
      int k = getCoordIndexCC(icc,i);
      MPG_Graph::pEdge e = getFirstEdge(k);
      if(e!=(MPG_Graph::pEdge)0)
	{ _curr_i_cc = i; _curr_e_cc = e; break; }
    }
  return(_curr_e_cc);
}

inline MPG_Graph::pEdge  MPG_Mesh::getFirstEdgeCC()
{ return getFirstEdgeCC(_i_cc); }

inline MPG_Graph::pEdge MPG_Mesh::getNextEdgeCC()
{ return getNextEdgeCC(_i_cc); }

inline int  MPG_Mesh::getNumberOfEdgesCC(int icc)
{
  int nEcc = 0;
  MPG_Graph::Edge *e;
  for(e=getFirstEdgeCC(icc);e;e=getNextEdgeCC())
    nEcc++;
  return nEcc;
}

inline int  MPG_Mesh::getNumberOfEdgesCC()
{ return getNumberOfEdgesCC(_i_cc); }

inline bool MPG_Mesh::smooth(vector<float>& x,
			     vector<float>& sx,
			     float          a,
			     float          b,
			     int            n,
			     vector<bool>*  e_sharp,
			     vector<bool>*  v_fixed)
{ return ((_graph)?_graph->smooth(x,sx,a,b,n,e_sharp,v_fixed):false); }

#endif // _MPG_Mesh_h_
