//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------

#pragma warning(disable:4786)

#include <stdio.h>
#include <stdlib.h>
#include "MPG_MeshFs.h"
#include "MPG_Perm.h"

//#include "MPG_Vector.h"	//MJH 010911 : removed because MPG_Vector class is not used. Only vector_erase(x) is being used and this is added below
#define vector_erase(x) x.erase(x.begin(),x.end())	//MJH 010911 : added because MPG_Vector.h was removed.

//---------------------------------------------------------------------------
//
//   MPG_MeshFs::MPG_MeshFs
//
//---------------------------------------------------------------------------

MPG_MeshFs::MPG_MeshFs(vector<int>& coordIndex0, vector<int>& map,
                       vector<int>& coordIndex1):
  MPG_Mesh(coordIndex1)
{
  clear();
  coordIndex1.erase(coordIndex1.begin(),coordIndex1.end());

  vector<int> f;
  
  int h0,h1,h;
  for(h0=h1=0;h1<coordIndex0.size();h1++)
    if(coordIndex0[h1]<0)
      {
        for(h=h0;h<h1;h++)
          if(f.size()==0 || map[coordIndex0[h]]!=f.back())
            f.push_back(map[coordIndex0[h]]);
        while(f.size()>1 && f.front()==f.back())
          f.pop_back();
        // still need to remove repeated indices !!!
        if(f.size()>=3)
          {
            for(h=0;h<f.size();h++)
              coordIndex1.push_back(f[h]);           
            coordIndex1.push_back(-1);
          }
        f.erase(f.begin(),f.end());
        h0 = h1+1;
      }
  set(coordIndex1);
}

//---------------------------------------------------------------------------
//
//   MPG_MeshFs::_isForest
//
//---------------------------------------------------------------------------

bool MPG_MeshFs::_isForest(vector<bool>& emk, MPG_Partition& vp)
{
  int nE = getNumberOfEdges();
  if(emk.size()!=nE) return false;

  int nV = getNumberOfVertices();
  if(vp.n()!=nV) return false;

  MPG_Graph::Edge *e;
  int i;
  for(i=0;i<nV;i++)
    for(e=getFirstEdge(i);e;e=getNextEdge(e))
      {
        int ie  = e->getIndex();
        if(emk[ie]==true &&
           (getNumberOfEdgeFaces(ie)==2 || getNumberOfEdgeFaces(ie)==1))
          {
            int iv0 = e->getVertex(0);
            int iv1 = e->getVertex(1);
            if(vp.find(iv0)==vp.find(iv1))
              { return false; } // error
            else
              vp.join(iv0,iv1);
          }
      }
  return true;
}

//---------------------------------------------------------------------------
//
//   MPG_MeshFs::isConnectivityPreservingMap
//
//---------------------------------------------------------------------------

bool MPG_MeshFs::isConnectivityPreservingMap(vector<int>& map)
{
  if(!isValidMap(map)) return false;

  int i;

  int nV = map.size();
  int nF = getNumberOfFaces();
  MPG_Partition p(nV);
  MPG_Graph g(_coordIndex);
  int nE = g.getNumberOfEdges();
  for(i=0;i<nV;i++)
    for(MPG_Graph::Edge* e=g.getFirstEdge(i);e;e=g.getNextEdge(e))
      {
        int ev0 = e->getVertex(0);
        int ev1 = e->getVertex(1);

        if(map[ev0]==map[ev1])
          p.join(ev0,ev1);
      }

  int minValue = 0;
  int maxValue = p.parts();
  bool is_onto = true;
  {
    vector<int>::iterator i;
    for(i=map.begin();i!=map.end();i++)
      if((*i)<minValue || (*i)>=maxValue)
        { is_onto = false; break; }
    if(is_onto!=false)
      {
        bool *is_in_range = new bool[maxValue-minValue];
        int i;
        for(i=minValue;i<maxValue;i++)
          is_in_range[i-minValue] = false;
        for(i=0;i<map.size();i++)
          is_in_range[minValue+map[i]] = true;
        for(i=minValue;i<maxValue;i++)
          if(is_in_range[i-minValue]==false)
            { is_onto = false; break; }
        delete [] is_in_range;
      }
  }
  bool is_conn_pres = is_onto;
  return is_conn_pres;
}

//------------------------------------------------------------------------
//
//   MPG_MeshFs::representForestSplit
//
//------------------------------------------------------------------------

bool MPG_MeshFs::representForestSplit(// inputs
                                      vector<int>&           map,
                                      // outputs
                                      vector<bool>&          emk,
                                      vector<MPG_pSPolygon>& vp_sp,
                                      MPG_Property&          pCoord,
                                      MPG_Property&          pNormal,
                                      MPG_Property&          pColor,
                                      MPG_Property&          pTexCoord,
                                      vector<int>&           qCoord,
                                      vector<int>&           qNormal,
                                      vector<int>&           qColor,
                                      vector<int>&           qTexCoord,
                                      vector<int>&           triData,
                                      vector<int>&           v_traversal,
                                      vector<int>&           t_trav_relative,
                                      vector<MPG_pVecInt>&   vp_v_loop,
                                      vector<MPG_pVecInt>&   vp_vcf_loop)
{

  // GT Thu Jul 01 10:37:42 1999 TODO
  // collect triData as in MPG_MeshTs::representConnectivity() !!!

  // NEEDS WORK !!!
  // collect vp_vcf_loop

  vector_erase(emk);
  vector_erase(vp_sp);
  vector_erase(t_trav_relative);
  vector_erase(v_traversal);

  // START WITH SOME ERROR CHECKING -----------------------------------------

  if(!isConnectivityPreservingMap(map))
    {
      return false;
    }

  if(!isTriangulated())
    {
      return false;
    }

  if(!isOrientedManifold())
    {
      return false;
    }

  // PRINT SOME INFO --------------------------------------------------------

  int i,j,k;

  int nV = map.size();
  int nF = getNumberOfFaces();

  // MARK COLLAPSED TRIANGLES -----------------------------------------------

  int  n_t_collapsed = 0;
  vector<bool> t_is_collapsed((size_t) nF,false);
  vector<int> fi;
  for(i=0;i<nF;i++)
    if(getFace(i,fi,map)==false)
      { t_is_collapsed[i] = true; n_t_collapsed++; }

  // PARTITION TRIANGLES INTO CONNECTED COMPONENTS --------------------------

  MPG_Partition p(nF);
  int nE = getNumberOfEdges();
  for(i=0;i<nE;i++)
    if(getNumberOfEdgeFaces(i)==2)
      {
        int ef0 = getEdgeFace(i,0);
        int ef1 = getEdgeFace(i,1);

        if(t_is_collapsed[ef0]==t_is_collapsed[ef1])
          p.join(ef0,ef1);
      }
  p.end();
  
  // PUT CONNECTED COMPONENTS IN LIST FORM ----------------------------------

  int  nCC = 0;
  vector<int> ccF;
  vector<int> ccI;

  vector<int> part((size_t) nF,-1);
  {
    int nccI = 0;
    for(j=i=0;i<nF;i++)
      if(t_is_collapsed[i]==true && p.size(i)>0)
        { part[i] = j++; nccI += p.size(i); }

    nCC = j;
    int nccF = nCC+1;
    ccF.reserve(nccF);
    ccF.insert(ccF.end(),(size_t) nccF,0);
    for(i=0;i<nF;i++)
      if(part[i]>=0)
        ccF[part[i]+1] = p.size(i);
    for(i=0;i<nCC;i++)
      ccF[i+1] += ccF[i];

    vector<int> n((size_t) nCC,0);
    ccI.reserve(nccI);
    ccI.insert(ccI.end(),(size_t) nccI,0);
    for(i=0;i<nF;i++)
      if(t_is_collapsed[i]==true)
        {
          int icc = part[p.find(i)];
          ccI[ccF[icc]+n[icc]] = i;
          n[icc]++;
        }

    // DETERMINE ORDER OF TRIANGLES IN TRIANGLE TREES -----------------------

    vector<int> f1_order((size_t) nF,0);
    for(i=0;i<nV;i++)
      for(MPG_Graph::Edge* e=getFirstEdge(i);e;e=getNextEdge(e))
        if(getNumberOfEdgeFaces(e->getIndex())==2)
          {
            int ef0 = getEdgeFace(e->getIndex(),0);
            int ef1 = getEdgeFace(e->getIndex(),1);
            if(t_is_collapsed[ef0]==true && t_is_collapsed[ef1]==true)
              { f1_order[ef0]++; f1_order[ef1]++; }
          }
  }

  // faces belonging to connected component i are
  // ccI[j] with ccF[i]<=j<ccF[i+1]

  // MORE ERROR CHECKING ----------------------------------------------------

  {
    // determine vertices shared by 2 or more cc's
    // by counting number of incident edges
    // which are boundaries of cc's

    // SHOULD ALSO CHECK IF BOUNDARIES OF CC'S TOUCH BOUNDARIES OF
    // MESH AT ISOLATED VERTICES !!!

    vector<int> nCCv((size_t) nV,0);
    for(i=0;i<nV;i++)
      for(MPG_Graph::Edge* e=getFirstEdge(i);e;e=getNextEdge(e))
        if(getNumberOfEdgeFaces(e->getIndex())==2)
          {
            int ef0 = getEdgeFace(e->getIndex(),0);
            int ef1 = getEdgeFace(e->getIndex(),1);

            if(t_is_collapsed[ef0]!=t_is_collapsed[ef1])
              {
                nCCv[e->getVertex(0)]++;
                nCCv[e->getVertex(1)]++;
              }
          }
    int nVmultiple = 0;
    for(i=0;i<nV;i++)
      if(nCCv[i]>2)
        nVmultiple++;

    if(nVmultiple>0)
      {
        return false;
      }

    // determine if cc's have internal vertices
    for(i=0;i<nV;i++)
      for(MPG_Graph::Edge* e=getFirstEdge(i);e;e=getNextEdge(e))
        if(getNumberOfEdgeFaces(e->getIndex())==2)
          {
            int ef0 = getEdgeFace(e->getIndex(),0);
            int ef1 = getEdgeFace(e->getIndex(),1);
            if((t_is_collapsed[ef0]==true && t_is_collapsed[ef1]==true) &&
               (nCCv[e->getVertex(0)]==0||nCCv[e->getVertex(1)]==0))
              {
                return false;
              }
          }

  }

  // DETERMINE IF CC'S ARE SIMPLE POLYGONS -------------------------------------
  {
    // only need to count vertices, edges and faces of each component,
    // and verify that V-E+F = 1
    vector<int> V((size_t) nCC,0);
    {
      vector<int> nvcc((size_t) nV,0);
      vector<int> vcc((size_t) nV,-1);
      for(i=0;i<nF;i++)
        if(t_is_collapsed[i]==true)
          {
            int icc = part[p.find(i)];
            int nfi = getNumberOfFaceCorners(i);
            for(j=0;j<nfi;j++)
              {
                int vij = getFaceCorner(i,j);
                if(vcc[vij]<0)
                  { vcc[vij] = icc; nvcc[vij]++; V[icc]++; }
                else if(vcc[vij]!=icc) // should not happen !
                  {
                    // vertex belongs to several connected components
                    nvcc[vij]++; V[icc]++;
                  }
              }
          }
    }

    vector<int> E((size_t) nCC,0);
    for(i=0;i<nE;i++)
      {
        int nfi = getNumberOfEdgeFaces(i);
        if(nfi==1 || nfi==2)
          {
            // note that if nfe==2 and both faces are collapsed, then
            // ef0 and ef1 belong to the same connected component

            int icc = -1;
            if(t_is_collapsed[getEdgeFace(i,0)]==true)
              icc = part[p.find(getEdgeFace(i,0))];
            else if(nfi==2 &&
                    t_is_collapsed[getEdgeFace(i,1)]==true)
              icc = part[p.find(getEdgeFace(i,1))];
            if(icc>=0)
              E[icc]++;
          }
      }

    vector<int> F((size_t) nCC,0);
    for(i=0;i<nCC;i++)
      F[i] = ccF[i+1]-ccF[i];

    int  nCCisSPolygon = 0;
    vector<int> CCisSPolygon(nCC);
    for(i=0;i<nCC;i++)
      {
        CCisSPolygon[i] = (V[i]-E[i]+F[i]==1)?1:0;
        nCCisSPolygon += CCisSPolygon[i];

      }

    if(nCCisSPolygon<nCC) return false;
  }

  // COMPRESSING ------------------------------------------------------------
  {
    // CREATE COLLAPSED MESH ------------------------------------------------

    MPG_MeshFs& mesh0        = *this;

    vector<int>& coordIndex0 = mesh0.getCoordIndex();
    vector<int>& map01       = map;
    vector<int> coordIndex1;
    MPG_MeshFs mesh1(coordIndex0,map,coordIndex1); // <<< !!!

    int nV0 = mesh0.getNumberOfVertices();
    int nF0 = mesh0.getNumberOfFaces();
    int nE0 = mesh0.getNumberOfEdges();

    int nV1 = mesh1.getNumberOfVertices();
    int nF1 = mesh1.getNumberOfFaces();
    int nE1 = mesh1.getNumberOfEdges();

    // generate maps associating
    // non-collapsed triangles of mesh0 with triangles of mesh1

    t_trav_relative.reserve(nF1);
    for(i=0;i<nF0;i++)
      if(t_is_collapsed[i]==false)
        t_trav_relative.push_back(i);

    // at this point mesh1 should also be manifold and
    // should have the same Euler number as mesh0,
    // but we don't need to check ???

    int nE1_collapsed = 0;
    int nE1_not_found = 0;
    int nE1_marked    = 0;

    // MARK LOOP EDGES OF CCs IN mesh0 AND CORRESPONDING TREE EDGES IN mesh1

    vector<bool> emk0;
        // VC++ doesn't like this
        // MPG_Vector::init(emk0,nE0,false);
    emk0.erase(emk0.begin(),emk0.end());
    emk0.insert(emk0.end(),(size_t)nE0,false);
    
        vector<bool>& emk1 = emk;
        // VC++ doesn't like this
        // MPG_Vector::init(emk1,nE1,false);
    emk1.erase(emk1.begin(),emk1.end());
    emk1.insert(emk1.end(),(size_t)nE1,false);

    MPG_Graph::Edge *e0,*e1;
    for(i=0;i<nV0;i++)
      for(e0=mesh0.getFirstEdge(i);e0;e0=mesh0.getNextEdge(e0))
        {
          int e0i = e0->getIndex();

          if(mesh0.getNumberOfEdgeFaces(e0i)==1 &&
             t_is_collapsed[mesh0.getEdgeFace(e0i,0)]==true)
            emk0[e0i] = true;
          else if(mesh0.getNumberOfEdgeFaces(e0i)==2)
            {
              int e0f0 = mesh0.getEdgeFace(e0i,0);
              int e0f1 = mesh0.getEdgeFace(e0i,1);
              if((t_is_collapsed[e0f0]==true &&
                  t_is_collapsed[e0f1]==false) ||
                 (t_is_collapsed[e0f0]==false &&
                  t_is_collapsed[e0f1]==true))
                {
                  emk0[e0i] = true;

                  int e0v0 = e0->getVertex(0);
                  int e0v1 = e0->getVertex(1);
                  int e1v0 = map01[e0v0];
                  int e1v1 = map01[e0v1];

                  // e1v0 and e1v1 should be different here !!!

                  if(e1v0==e1v1)
                    {
                      // should never happen
                      nE1_collapsed++;
                    }
                  else if((e1=mesh1.getEdge(e1v0,e1v1))==(MPG_Graph::Edge*)0)
                    {
                      // should never happen
                      nE1_not_found++;
                    }
                  else
                    {
                      // should always happen
                      nE1_marked++;
                      emk1[e1->getIndex()] = true;
                    }
                }
            }
        }

    // ERROR CHECKING -----------------------------------------------------

    if(nE1_collapsed>0 || nE1_not_found>0)
      return(false);

    // GENERATE SIMPLE POLYGONS -------------------------------------------

    // faces belonging to connected component i are
    // ccI[j] with ccF[i]<=j<ccF[i+1]

    {
      // 1) sort cc's by min{ map01[fij] : fi face of cc[i] }
      vector<int> cc_stack; cc_stack.reserve(nCC);
      {
        vector<int> cc_order((size_t) nCC,-1);
        for(i=0;i<nCC;i++)
          {
            for(j=ccF[i];j<ccF[i+1];j++)
              {
                int nfj = getNumberOfFaceCorners(ccI[j]);
                for(k=0;k<nfj;k++)
                  {
                    int vjk_1 = getFaceCorner(ccI[j],k);
                    int vjk_2 = map01[vjk_1];
                    if(cc_order[i]<0 || vjk_2<cc_order[i])
                      cc_order[i] = vjk_2;
                  }
              }
          }

        vector<int> v_cc((size_t) nV,-1);
        for(i=0;i<nCC;i++)
          v_cc[cc_order[i]] = i;

        // for(j=nV-1;j>=0;j--)
        for(j=0;j<nV;j++)
          if(v_cc[j]>=0)
            cc_stack.push_back(v_cc[j]);
      }
      
      // 2) choose rootV, rootE and rootF in mesh1 using same method as in
      //    Mesh::compress()

      vector<int> rootVertex((size_t) nCC,-1);

      // for each cc
      //   {
      //     make a list of vertices in mesh1
      //     determine leafs as order 1 vertices
      //     choose min leaf as root
      //   }

      // determine order of marked vertices in mesh1
      vector<int> v2_order((size_t) nV1,0);
      {
        int i,j;
        MPG_Graph::Edge *e1;
        for(i=0;i<nV1;i++)
          for(e1=mesh1.getFirstEdge(i);e1;e1=mesh1.getNextEdge(e1))
            if(emk1[e1->getIndex()]==true)
              for(j=0;j<2;j++)
                v2_order[e1->getVertex(j)]++;
      }

      // choose min leaf of each cc as root vertex
      {
        int i,j;
        for(i=0;i<nF0;i++)
          if(t_is_collapsed[i]==true)
            {
              int icc = part[p.find(i)];
              int nfi = getNumberOfFaceCorners(i);
              for(j=0;j<nfi;j++)
                {
                  int v1ij = getFaceCorner(i,j);
                  int v2ij = map01[v1ij];
                  if(v2_order[v2ij]==1 &&
                     (rootVertex[icc]<0 || v2ij<rootVertex[icc]))
                    rootVertex[icc] = v2ij;
                }
            }
      }

      vector<MPG_Graph::Edge*> rootEdge((size_t) nCC,(MPG_Graph::Edge*)0);

      vector<int> rootFace((size_t) nCC,-1);
      {
        vector<int> rootVertexNumber((size_t) nV1,-1);
        int h,i,j,k,l;
        for(i=0;i<nCC;i++)
          rootVertexNumber[rootVertex[i]] = i;

        MPG_Graph::Edge *e1;
        for(h=0;h<nV1;h++)
          for(e1=mesh1.getFirstEdge(h);e1;e1=mesh1.getNextEdge(e1))
            if(emk1[e1->getIndex()]==true)
              for(j=0;j<2;j++)
                {
                  l = rootVertexNumber[e1->getVertex(j)];
                  if(l>=0)
                    {
                      int nef  = mesh1.getNumberOfEdgeFaces(*e1);
                      int minF = mesh1.getEdgeFace(*e1,0);
                      for(k=1;k<nef;k++)
                        {
                          int efk = mesh1.getEdgeFace(*e1,k);
                          if(efk<minF) minF = efk;
                        }
                      rootEdge[l] = e1;
                      rootFace[l] = minF;
                      break;
                    }
                }
        int nCCnoRoot = 0;
        for(i=0;i<nCC;i++)
          if(rootFace[i]==-1 || rootEdge[i]==(MPG_Graph::Edge*)0)
            nCCnoRoot++;
      }


      // 3) the corners of rootF[i] corresponding to
      //    rootV[i] and nextV[i] = rootE[i]->getOtherVertex(rootV[i])
      //    mesh1.getCorner(rootF[i],rootV[i]);
      //    mesh1.getCorner(rootF[i],rootE[i]->getOtherVertex(rootV[i]);
      //    determine entry edge of simple polygon by looking at the same
      //    corners of corresponding triangle in mesh0

      vector<int> loop_next((size_t) nV,-1);

      // 4) encode simple polygon and collect property binding indices

      {
        int cL,cR,cO,ifP,ifN,ifL,ifR,*fP,*fN;
        MPG_Graph::Edge *e,*eL,*eR;

        v_traversal.insert(v_traversal.end(),(size_t) nV1,-1);

        vector<int>::iterator icc;
        vector<int> sp_stack;

        for(icc=cc_stack.begin();icc!=cc_stack.end();icc++)
          {
            i = *icc;

            MPG_SPolygon* p_sp = new MPG_SPolygon();
            vp_sp.push_back(p_sp);

            MPG_pVecInt vl = new vector<int>();
            vp_v_loop.push_back(vl);
            vector<int>& v_loop = *vl;

            MPG_pVecInt vcfl = new vector<int>();
            vp_vcf_loop.push_back(vcfl);
            vector<int>& vcf_loop = *vcfl;

            // ---

            e1 = rootEdge[i];
            int j;
            for(j=0;j<2;j++)
              if(rootVertex[i]==e1->getVertex(j))
                break;
            int iv0 = e1->getVertex(j);
            int iv1 = e1->getVertex((j==0)?1:0);

            cR  = mesh1.getCorner(rootFace[i],iv0);
            cL  = mesh1.getCorner(rootFace[i],iv1);
            ifP = t_trav_relative[rootFace[i]];
            fP  = mesh0[ifP];

            int loop_root = fP[cR];
            loop_next[fP[cR]] = fP[cL];

            e   = mesh0.getEdge(fP[cR],fP[cL]);
            ifN = mesh0.getOtherEdgeFace(e->getIndex(),ifP);
            // if(fN<0 || t_is_collapsed[ifN]==false) ERROR();
            fN  = mesh0[ifN];
            cL  = mesh0.getCorner(ifN,fP[cL]);
            cR  = mesh0.getCorner(ifN,fP[cR]);

            sp_stack.push_back(ifN);
            sp_stack.push_back(cL);
            sp_stack.push_back(cR);

            while(sp_stack.size()>=3)
              {
                cR  = sp_stack.back(); sp_stack.pop_back();
                cL  = sp_stack.back(); sp_stack.pop_back();
                ifN = sp_stack.back(); sp_stack.pop_back();
                fN  = mesh0[ifN];
                e   = mesh0.getEdge(fN[cR],fN[cL]);

                t_trav_relative.push_back(ifN);

                switch(cR+cL)
                  {
                  case 1 : cO =  2; break;
                  case 2 : cO =  1; break;
                  case 3 : cO =  0; break;
                  default: cO = -1; break; // ERROR();
                  }

                eL  = mesh0.getEdge(fN[cL],fN[cO]);
                ifL = mesh0.getOtherEdgeFace(eL->getIndex(),ifN);
                eR  = mesh0.getEdge(fN[cR],fN[cO]);
                ifR = mesh0.getOtherEdgeFace(eR->getIndex(),ifN);

                int code = 0;
                if(ifL>=0 && t_is_collapsed[ifL]==true)
                  {
                    sp_stack.push_back(ifL);
                    sp_stack.push_back(mesh0.getCorner(ifL,fN[cL]));
                    sp_stack.push_back(mesh0.getCorner(ifL,fN[cO]));
                    code += 1;
                  }
                else
                  loop_next[fN[cL]] = fN[cO];

                if(ifR>=0 && t_is_collapsed[ifR]==true)
                  {
                    sp_stack.push_back(ifR);
                    sp_stack.push_back(mesh0.getCorner(ifR,fN[cO]));
                    sp_stack.push_back(mesh0.getCorner(ifR,fN[cR]));
                    code += 2;
                  }
                else
                  loop_next[fN[cO]] = fN[cR];

                switch(code)
                  {
                  case 0:       // sp_LEAF
                    p_sp->pushTri(MPG_SPolygon::sp_LEAF);
                    break;
                  case 1:       // sp_LEAFT
                    p_sp->pushTri(MPG_SPolygon::sp_LEFT);
                    break;
                  case 2:       // sp_RIGHT
                    p_sp->pushTri(MPG_SPolygon::sp_RIGHT);
                    break;
                  case 3:       // sp_BRANCH
                    p_sp->pushTri(MPG_SPolygon::sp_BRANCH);
                    break;
                  }
              }
            // fill in v_traversal

            int iv=loop_root;
            do {
              if(v_traversal[map01[iv]]<0)
                v_traversal[map01[iv]] = iv;
              else
                v_traversal.push_back(iv);
              v_loop.push_back(iv);
              iv=loop_next[iv];
            } while(iv>=0 && iv!=loop_root);

          }

        int nMinus = 0;

        for(i=0;i<nV0;i++)
          if(v_traversal[i]<0)
            nMinus++;

        for(i=0;i<nV0;i++)
          if(v_traversal[map01[i]]<0)
            v_traversal[map01[i]] = i;

        nMinus = 0;
        for(i=0;i<nV0;i++)
          if(v_traversal[i]<0)
            nMinus++;
      }
    }
  }

  return true;
}

//------------------------------------------------------------------------
//
//   LOOP_PUSH_BACK
//
//------------------------------------------------------------------------

#define LOOP_PUSH_BACK(v) \
{ \
  if(v_is_free[v]==true) { v_new = v; v_is_free[v] = false; }          \
  else                   { v_new = v_old.size(); v_old.push_back(v); } \
  v_loop.push_back(v_new); \
}

//------------------------------------------------------------------------
//
//   PUSH_BACK_CORNER
//
//------------------------------------------------------------------------

#define PUSH_BACK_CORNER() \
{ \
  int c_curr = getCorner(f_curr,v_curr); \
  vcf_loop.push_back(v_new);  \
  vcf_loop.push_back(c_curr); \
  vcf_loop.push_back(f_curr); \
}

//------------------------------------------------------------------------
//
//   MPG_MeshFs::_buildLoop
//
//------------------------------------------------------------------------

void MPG_MeshFs::_buildLoop(vector<bool>&    e_mark,
                            vector<bool>&    v_is_bdry,
                            vector<int>&     v_order,
                            int              nVtree,
                            int              rootV,
                            MPG_Graph::Edge& rootE,
                            int              rootF,
                            vector<int>&     v_loop,
                            vector<bool>&    v_is_free,
                            vector<int>&     v_old,
                            vector<int>&     vcf_loop)
{
  v_loop.erase(v_loop.begin(),v_loop.end());
  
  if(isEdgeVertex(rootE,rootV) && isEdgeFace(rootE,rootF) && v_order[rootV]==1)
    {
      int v_new=-1,v_prev=-1,v_curr=-1,v_next=-1,f_curr=-1,f_next=-1;
      MPG_Graph::Edge* e = (MPG_Graph::Edge*)0;

      LOOP_PUSH_BACK(rootV);

      // loop (clockwise) around root vertex
      // until a boundary or marked edge is found
      // to visit all the corners
      // of faces incident to the root vertex
      // on the side of the root face

      v_prev = rootE.getOtherVertex(rootV);
      v_curr = rootV;
      f_curr = rootF;
      while(1)
        {
          PUSH_BACK_CORNER();

          v_next = getNextFaceCorner(f_curr,v_prev,v_curr);
          e      = getEdge(v_curr,v_next);
          f_next = getOtherEdgeFace(*e,f_curr);
          v_prev = v_next;
          if(e_mark[e->getIndex()]==true || f_next<0)
            break;
          f_curr = f_next;
        }

      v_prev = -1;
      v_curr = rootV;
      v_next = rootE.getOtherVertex(rootV);
      f_curr = rootF;
      f_next = -1;

      LOOP_PUSH_BACK(v_next);

      while(1) // A
        {
          v_prev = v_curr; v_curr = v_next;

          // loop (counterclockwise) around v_curr
          // until a marked or boundary edge is found
          int f_start = f_curr;
          MPG_Graph::Edge* e_start = getEdge(v_prev,v_curr);
          bool done = false;
          while(done==false) // B
            {
              PUSH_BACK_CORNER();

              v_next = getNextFaceCorner(f_curr,v_prev,v_curr);
              e      = getEdge(v_curr,v_next);
              f_next = getOtherEdgeFace(*e,f_curr);

              if(e_mark[e->getIndex()]==true)
                {
                  if(f_next>=0)
                    break;
                  //

                  v_prev = v_curr;
                  v_curr = v_next;
                  LOOP_PUSH_BACK(v_curr);

                  continue;
                }
              else
                v_prev = v_next;

              if(f_next>=0)
                f_curr = f_next;
              else
                {
                  // find boundary edge on the other side
                  while(1) // C
                    {
                      bool prev_bdry_e_was_marked = e_mark[e->getIndex()];
                      while(1)
                        {
                          v_next = getNextFaceCorner(f_curr,v_prev,v_curr);
                          e      = getEdge(v_curr,v_next);
                          f_next = getOtherEdgeFace(*e,f_curr);
                          v_prev = v_next;
                          if(f_next<0)
                            break;
                          f_curr = f_next;
                        }

                      if(e_mark[e->getIndex()]==true)
                        {
                          if(v_next==rootV)
                            { done=true; break; }

                          v_prev = v_curr;
                          v_curr = v_next;
                          LOOP_PUSH_BACK(v_curr);
                        }
                      else
                        {
                          if(prev_bdry_e_was_marked==false)
                            {
                              LOOP_PUSH_BACK(v_curr);
                            }
                          break;
                        }
                    } // C
                }
            } // B

          // if reached root vertex ...
          if(v_next==rootV)
            {
              // ... and root vertex is not boundary we are done
              if(v_is_bdry[rootV]==true)
                {
                  // if root edge is not boundary, i.e.,
                  // if root vertex is isolated boundary,
                  // add edge to loop

                  if(getNumberOfEdgeFaces(rootE)>1)
                    { LOOP_PUSH_BACK(rootV); }
                  else
                    v_new = rootV;

                  // loop (counterclockwise) around root vertex until
                  // a boundary edge is found,
                  // to finish visiting all the corners
                  // of faces incident to the root vertex

                  v_prev = v_curr;
                  v_curr = v_next;
                  while(1)
                    {
                      PUSH_BACK_CORNER();

                      v_next = getNextFaceCorner(f_curr,v_prev,v_curr);
                      e      = getEdge(v_curr,v_next);
                      f_next = getOtherEdgeFace(*e,f_curr);

                      if(e_mark[e->getIndex()]==true || f_next<0)
                        break;

                      v_prev = v_next;
                      f_curr = f_next;
                    }
                }
              break;
            }
          if(f_next==-1)
            {
              v_loop.erase(v_loop.begin(),v_loop.end());
              break;
            }
          LOOP_PUSH_BACK(v_next);

        } // A

    }
}

#undef LOOP_PUSH_BACK
#undef PUSH_BACK_CORNER

//------------------------------------------------------------------------
//
//   MPG_MeshFs::decodeForestSplit
//
//------------------------------------------------------------------------

bool MPG_MeshFs::decodeForestSplit( // inputs
                                   vector<bool>&        emk,
                                   vector<MPG_pVecInt>& vp_coordIndex,
                                   // output
                                   vector<int>&         v_cluster,
                                   vector<MPG_pVecInt>& vp_vcf_loop)
{
  v_cluster.erase(v_cluster.begin(),v_cluster.end());

  int nV = getNumberOfVertices();
  int nF = getNumberOfFaces();
  int nE = getNumberOfEdges();
  int nC = getNumberOfCorners();

  if(emk.size()!=nE)
    {
      return false;
    }

  // check that edge marks form a forest
  vector<int> vOrder((size_t) nV,0);
  MPG_Partition vp(nV,vOrder.begin());
  if(_isForest(emk,vp)==false)
    {
      return false;
    }

  vp.makeParts();

  vector<bool> v_is_bdry((size_t) nV,false);
  int n_v_bdry = markBoundaryVertices(v_is_bdry);

  vector<int> rootVertex;
  vector<int> rootPart;
  
  // count useful connected components
  int nP     = vp.parts();
  int nTrees = 0;
  
  int i;
  for(i=0;i<nP;i++)
    {
      int nPi   = 0;
      int *part = NULL;
      vp.getPart(i,nPi,part);
      if((nPi>0 && part!=(int*)0) &&
			((nPi==2 && (v_is_bdry[part[0]]==true ||
                          v_is_bdry[part[1]]==true)) || nPi>2) )
        {
          nTrees++;
          
          int nBdryVertices = 0;
          int nLeafs        = 0;
          int nRegular      = 0;
          int nBranch       = 0;
          
          int minLeaf       = -1;

          int j;
          for(j=0;j<nPi;j++)
            {
              if(v_is_bdry[part[j]]==true)
                nBdryVertices++;
              switch(vOrder[part[j]])
                {
                case 0 :             break;
                case 1 :
                  if(minLeaf<0 || part[j]<minLeaf) minLeaf = part[j];
                  nLeafs++;   break;
                case 2 : nRegular++; break;
                default: nBranch++;  break;
                }
            }

          rootVertex.push_back(minLeaf);
          rootPart.push_back(i);
        }
    }
  // get first edge
  vector<MPG_Graph::Edge*> rootEdge((size_t) nTrees,(MPG_Graph::Edge*)0);

  vector<int> rootFace((size_t) nTrees,-1);
  {
    int i,j,k;

    vector<int> rootVertexNumber((size_t) nV,-1);
    for(i=0;i<rootVertex.size();i++)
      rootVertexNumber[rootVertex[i]] = i;

    MPG_Graph::Edge *e;
    for(i=0;i<nV;i++)
      for(e=getFirstEdge(i);e;e=getNextEdge(e))
        if(emk[e->getIndex()]==true)
          for(j=0;j<2;j++)
            {
              int ivn = rootVertexNumber[e->getVertex(j)];
              if(ivn>=0)
                {
                  rootEdge[ivn] = e;
                  int nef  = getNumberOfEdgeFaces(*e);
                  int minF = getEdgeFace(*e,0);
                  for(k=1;k<nef;k++)
                    {
                      int efk = getEdgeFace(*e,k);
                      if(efk<minF) minF = efk;
                    }
                  rootFace[ivn] = minF;
                  break;
                }
            }
  }

  vector<bool> v_is_free(nV,true);

  vector<int>& v_old = v_cluster;
  v_old.reserve(nV);
  for(i=0;i<nV;i++)
    v_old.push_back(i);

  for(i=0;i<nTrees;i++)
    {
      // one v_loop and one simple polygon per tree

      int nPi   = 0;
      int *part = NULL;
      vp.getPart(rootPart[i],nPi,part);

      vector<int> v_loop;
  
      MPG_pVecInt nc = new vector<int>();
      vp_vcf_loop.push_back(nc);
      vector<int>& vcf_loop = *nc;

      _buildLoop(emk,
                 v_is_bdry,
                 vOrder,
                 nPi,rootVertex[i],*(rootEdge[i]),rootFace[i],
                 // out
                 v_loop,
                 v_is_free,
                 v_old,
                 vcf_loop);
        
      // determine if polygon faces should be inverted or not
      bool invert_faces = true;
      {
        int  v0 = rootVertex[i];
        int  v1 = rootEdge[i]->getOtherVertex(v0);
        int  c0 = getCorner(rootFace[i],v0);
        int  c1 = getCorner(rootFace[i],v1);
        int  nc = getNumberOfFaceCorners(rootFace[i]);
        invert_faces = (c1==((c0+1)%nc))?true:false;
      }

      // polygon faces over loop indices
      // constructed by MPG_SPolygon::decodeTriangleData()
      // and passed as argument
      vector<int>& coordIndex = *(vp_coordIndex[i]);

      // convert from loop indices to mesh indices in place
      int j,j0,j1,k0,k1;
      for(j0=j1=0;j1<coordIndex.size();j1++)
        if(coordIndex[j1]<0)
          {
            // end of face

            // invert face if necessary
            if(invert_faces==true)
              for(k0=j0,k1=j1-1;k0<k1;k0++,k1--)
                {
                  int aux = coordIndex[k0];
                  coordIndex[k0] = coordIndex[k1];
                  coordIndex[k1] = aux;
                }

            // replace loop indices by mesh indices
            for(j=j0;j<j1;j++)
              coordIndex[j] = v_loop[coordIndex[j]];

            // advance to next face
            j0 = j1+1;
          }
    }

  // now we will start editing the mesh
  // delete edges which will become invalid
  // they will be reconstructed later 
  deleteEdges();

  // set the new number of vertices
  // after cutting through forest edges

  nV = v_old.size();
  setNumberOfVertices(nV);

  // cut mesh through forest edges by
  // replacing old corner values
  {
    int i,j;
    for(j=0;j<vp_vcf_loop.size();j++)
      {
        vector<int>& vcf_loop = *(vp_vcf_loop[j]);

        for(i=0;3*i+2<vcf_loop.size();i++)
          {
            int  v = vcf_loop[3*i  ];
            int  c = vcf_loop[3*i+1];
            int  k = vcf_loop[3*i+2];
            int* f = getFace(k);

            // setFaceCorner
            f[c] = v;
          }
      }
  }

  // do not erase the vcf_loop array:
  // it is used to update the
  // per face and per corner properties
  // in the calling method

  // add the internal faces of the simple polygons
  {
    for(i=0;i<nTrees;i++)
      {
        // polygon faces over mesh indices
        vector<int>& coordIndex = *(vp_coordIndex[i]);
        
        int j0,j1;
        for(j0=j1=0;j1<coordIndex.size();j1++)
          if(coordIndex[j1]<0)
            {
              // end of face
              int *f = &(coordIndex[j0]);
              int nf = j1-j0;
              
              // add new face
              pushBackFace(nf,f);
              
              // advance to next face
              j0 = j1+1;
            }
      }
    reset();
  }

  // reset();
  nF = getNumberOfFaces();
  nE = getNumberOfEdges();
  nC = getNumberOfCorners();
  
  return true;
}

//---------------------------------------------------------------------------
//
//   MPG_MeshFs::isValidMap
//
//---------------------------------------------------------------------------

bool MPG_MeshFs::isValidMap(vector<int>& map)
{
  bool is_valid_map = true;

  int nV = map.size();
  int nF = getNumberOfFaces();
  int i,j;
  for(i=0;i<nF;i++)
    {
      int nfi = getNumberOfFaceCorners(i);
      for(j=0;j<nfi;j++)
        {
          int fij = getFaceCorner(i,j);
          if(fij<0 || fij>=nV)
            { is_valid_map = false; break; }
        }
      if(is_valid_map==false)
        break;
    }

  return is_valid_map;
}

//---------------------------------------------------------------------------
//
//   MPG_MeshFs::isCollapsedFace
//
//---------------------------------------------------------------------------

bool MPG_MeshFs::isCollapsedFace(int i, vector<int>& map)
{
  int nfi  = getNumberOfFaceCorners(i);
  int fi_0 = map[getFaceCorner(i,0)];
  int fi_p = fi_0;
  int fi_j = -1;
  int nfi_map = 1;
  int j;
  for(j=1;j<nfi;j++)
    {
      fi_j = map[getFaceCorner(i,j)];
      if(fi_j!=fi_p)
        { nfi_map++; fi_p=fi_j; }
    }
  if(fi_j==fi_0)
    nfi_map--;
  return (nfi_map<3)?true:false;
}

//---------------------------------------------------------------------------
//
//   MPG_MeshFs::permuteFaces
//
//---------------------------------------------------------------------------

bool MPG_MeshFs::permuteFaces(vector<int>& f_traversal)
{
  int nF = getNumberOfFaces();
  // check that f_traversal is a permutation
  if(nF!=f_traversal.size()) return false;
  if(MPG_Perm::isPerm(f_traversal)==false) return false;

  vector<int> permCoordIndex;
  int h,i,j,nfi,*fi;
  for(h=0;h<nF;h++)
    {
      i   = f_traversal[h];
      nfi = getNumberOfFaceCorners(i);
      fi  = getFace(i);
      for(j=0;j<nfi;j++)
        permCoordIndex.push_back(fi[j]);
      permCoordIndex.push_back(-1);
    }
  set(permCoordIndex);

  return true;
}
