//
//   Simulation of packet loss in MPEG-4 3DMC bitstream

//   Copyright (C) 2000 Samsung Electronics Co. LTD. 
//   Mahnjin Han  (manjini@sait.samsung.co.kr)
//   Sungjin Kim  (sunjin@sait.samsung.co.kr)
//   Signal Processing Lab. Samsung A.I.T. 
//
//   Name:  plerror
//   
//   Usage: plerror [-id<inDir>] [-od<outDir>] [-rate<plRate>] <input_file_name>
//
//   Example usage:
//       - Simulate packet loss of rate 10% in ..\bits\venus.m3d bitstream file and store result in ..\rec directory
//         plerror -id..\bits -od..\rec -rate0.1 venus
//

// PLError.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "MPG_Stream.h"

FILE* logfp     = stderr;

#define not(b) ((b)==true)?false:true

const char MOBL0_start_code = 0x31;	
const char MOBL1_start_code = 0x33; 
const char MOBL2_start_code = 0x34;	
const char visual_object_sequence_end_code	= 0xB1;  

vector<unsigned char>	_buff;
int						_next=0;	
int						nBytesRead=0;

void usage()
{
	static char *mm[] = {
		"USAGE: plerror [options] baseFname ",
			"    -h           for help",
			"    -id<inDir>   tryes to read input files from 'inDir' directory",
			"    -od<outDir>  places output files in 'outDir' directory",
			"    -rate<plRate>	  the packet loss rate",
			NULL
	};
	for(char **m=mm;*m;m++)
		fprintf(stderr,"%s\n",*m);
	exit(1);
}

void error(const char *msg)
{
	fprintf(logfp,"error : %s\n",(msg)?msg:"");
	exit(-1);
}

unsigned char copy_header(FILE *fp)
{
	unsigned char tmpchar[2];
	while (_next+2<=_buff.size()) {
		if (_buff[_next]==0x00 && _buff[_next+1]!=0x00 && _buff[_next+1] < 0x60) {
			if (_buff[_next+1]==MOBL0_start_code ||
				_buff[_next+1]==MOBL1_start_code ||
				_buff[_next+1]==MOBL2_start_code) {
				return _buff[_next+1];
			}
			else {
				tmpchar[0] = _buff[_next];
				nBytesRead += fwrite(tmpchar,1,1,fp);
				_next++;
			}
		}
		else {
			tmpchar[0] = _buff[_next];
			nBytesRead += fwrite(tmpchar,1,1,fp);
			_next++;
		}
	}
	return 0;
}

unsigned char copy_partition(FILE *fp)
{
	unsigned char tmpchar[2];
	tmpchar[0] = _buff[_next];
	tmpchar[1] = _buff[_next+1];
	nBytesRead += fwrite(tmpchar,1,2,fp);
	_next+=2;
	
	while (_next+5<=_buff.size()) {
		if (_buff[_next]==0x00 && _buff[_next+1]!=0x00 && _buff[_next+1] < 0x60) {
			if (_buff[_next+1]==MOBL0_start_code ||
				_buff[_next+1]==MOBL1_start_code ||
				_buff[_next+1]==MOBL2_start_code) {
				return _buff[_next+1];
			}
			else {
				tmpchar[0] = _buff[_next];
				nBytesRead += fwrite(tmpchar,1,1,fp);
				_next++;
			}
		}
		else {
			tmpchar[0] = _buff[_next];
			nBytesRead += fwrite(tmpchar,1,1,fp);
			_next++;
		}
	}
	return 0;
}

unsigned char skip_partition(FILE *fp)
{
	_next+=2;
	while (_next+5<=_buff.size()) {
		if (_buff[_next]==0x00 && _buff[_next+1]!=0x00 && _buff[_next+1] < 0x60) {
			if (_buff[_next+1]==MOBL0_start_code ||
				_buff[_next+1]==MOBL1_start_code ||
				_buff[_next+1]==MOBL2_start_code ) {
				return _buff[_next+1];
			}
			else if (_buff[_next+1]==0x01 && _buff[_next+2]==visual_object_sequence_end_code) {
				_next--;
				return 0;
			}
			else
				_next++;
		}
		else
			_next++;
	}
	return 0;
}

int main(int argc, char* argv[])
{
	char* baseFname = NULL;
	char* inExt     = NULL;
	char* inDir     = NULL;
	char* outDir    = NULL;
	char* inFname   = NULL;
	char* outFname  = NULL;
	float plRate = 0;
	int i,j,success;
	
	if(argc==1)
		usage();
	else for(i=1;i<argc;i++)
	{
		float d;
		if(!strcmp(argv[i],"-h") || !strcmp(argv[i],"-help"))
			usage();
		else if(!strncmp(argv[i],"-id",3))
			inDir = strdup(argv[i]+3);
		else if(!strncmp(argv[i],"-od",3))
			outDir = strdup(argv[i]+3);
		else if(sscanf(argv[i],"-rate%f",&d)>0)
			plRate = d;
		else if(argv[i][0]=='-')
			error("unknown option");
		else if(!baseFname)
			baseFname = strdup(argv[i]);
		else
			error("too many input files");
	}		
	
	// check arguments
	if(!baseFname) {
		error("no input file");
	}
	
	// remove directory baseFname
	for(j=strlen(baseFname)-1;j>=0;j--) {
		if(baseFname[j]=='/') { baseFname[j]='\0'; j++; break; }
	}
	if(!inDir)
		inDir = (j>=0)?baseFname:strdup(".");
	else if(j>0 && strcmp(inDir,baseFname)!=0)
		error("inDir");
	if(j>0) baseFname += j;
	// remove extension
	for(j=strlen(baseFname)-1;j>0;j--) {
		if(baseFname[j]=='.') 
		{ baseFname[j]='\0'; inExt = baseFname+j+1; break; }
	}
	if(!baseFname)
		error("no input file");
	if(!outDir) outDir = strdup(inDir);
	if(!inExt)  inExt  = strdup("");
	
	// read from input file
    int n = strlen(inDir)+strlen(baseFname)+8;
    inFname = new char[n];
    // construct input file name
    if(!strcmp(inDir,"."))
		sprintf(inFname,"%s.m3d",baseFname);
//		sprintf(inFname,"%s.bif",baseFname);
    else
		sprintf(inFname,"%s/%s.m3d",inDir,baseFname);
//		sprintf(inFname,"%s/%s.bif",inDir,baseFname);
	
    n = strlen(outDir)+strlen(baseFname)+14;
    outFname = new char[n];
    if(!strcmp(outDir,"."))
		sprintf(outFname,"%s-PL.m3d",baseFname);
//		sprintf(outFname,"%s-PL.bif",baseFname);
    else
		sprintf(outFname,"%s/%s-PL.m3d",outDir,baseFname);
//		sprintf(outFname,"%s/%s-PL.bif",outDir,baseFname);
	
	_buff.erase(_buff.begin(),_buff.end());
	if(inFname!=NULL)
    {
		FILE *fp = fopen(inFname,"rb");
		if(fp!=(FILE*)0)
		{
			struct stat struct_stat;
			if(stat(inFname,&struct_stat)>=0 && struct_stat.st_size>0)
			{
				int nBytesStat = struct_stat.st_size;
				_buff.insert(_buff.end(),(size_t)nBytesStat,'\0');
				int nInBytesRead = fread(_buff.begin(),1,nBytesStat,fp);
				if(nBytesStat==nInBytesRead) success = true;
			}
			fclose(fp);
		}
		else 
			error("Cannot make input file");
    }
	else 
		error("Cannot make input file");
	
	// make the output file
	if(outFname!=NULL)
    {
		FILE *fp = fopen(outFname,"wb");
		if(fp!=(FILE*)0)
		{
			unsigned int seed = (unsigned)time( NULL );
//			unsigned int seed = 1; 
			srand( seed );
			plRate = plRate * RAND_MAX;
			unsigned char mobl_sc = copy_header(fp);
			while (mobl_sc) {
				if (mobl_sc == MOBL1_start_code) {
					mobl_sc = copy_partition(fp);
				}
				else if(rand() < plRate) {
					printf("skipping a partition!\n");
					mobl_sc = skip_partition(fp);
				}
				else {
					mobl_sc = copy_partition(fp);
				}
			}
			unsigned char tmpchar[1];
			while (_next<=_buff.size()) {
				tmpchar[0] = _buff[_next];
				nBytesRead += fwrite(tmpchar,1,1,fp);
				_next++;
			}
			fclose(fp);
		}
		else 
			error("Cannot make output file");
    }
	else 
		error("Cannot make output file");
	
    delete [] inFname;
    delete [] outFname;
	
	printf("Done.\n");
	
	return 0;
}
