//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------


#ifndef _MPG_Property_h_
#define _MPG_Property_h_

#include <float.h>
#include <math.h>
#include "MPG_BBoxQ.h"
#include "MPG_NormalQ.h"
#include "MPG_BranchTree.h"	

//ory DMLAB 2007 01 11
#ifndef _MPG_TYPE_
#define _MPG_TYPE_
enum MPG_Type {
  PT_NONE = 0,
  PT_COORD,
  PT_NORMAL,
  PT_COLOR,
  PT_TEXCOORD
};
#endif
#ifndef _MPG_Binding_
#define _MPG_Binding_
enum MPG_Binding {
  PB_NONE = 0,
  PB_PER_VERTEX,
  PB_PER_FACE,
  PB_PER_CORNER
};
#endif
#ifndef _MPG_Prediction_
#define _MPG_Prediction_
enum MPG_Prediction {
  PP_NONE = 0,
  PP_TREE,
  PP_PARALLELOGRAM,
  PP_RESERVED
};
#endif
//~ory DMLAB


class MPG_Property {

protected:

  MPG_Type             _type;
  MPG_Binding          _binding;
  MPG_Prediction       _prediction;
  vector<float>        _lambda;
  MPG_BBoxQ*           _bbq;
  MPG_NormalQ*         _nrq;
  int                  _bitsPerCoord;

public:

  inline               ~MPG_Property();
  inline                MPG_Property(MPG_Type       type       = PT_NONE,
				     MPG_Binding    binding    = PB_NONE,
				     MPG_Prediction prediction = PP_NONE);

  inline void           setType(MPG_Type type);
  inline MPG_Type       getType();

  inline void           setBinding(MPG_Binding binding);
  inline MPG_Binding    getBinding();

  inline void           setPrediction(MPG_Prediction prediction);
  inline MPG_Prediction getPrediction();

  inline int            getDimensionValue();
  inline int            getDimensionCode();
  inline int            getDimensionLambda();
  inline float*         getLambda();
  inline float          getLambda(int i);
  inline int            getNumberOfFathersPerValue();
  inline void           quantizeLambda();
  inline void           setDefaultLambda();

  inline void           setBBoxQ(MPG_BBoxQ* bbq=NULL);
  inline MPG_BBoxQ*     getBBoxQ();

  inline void           setNormalQ(MPG_NormalQ* nrq=NULL);
  inline MPG_NormalQ*   getNormalQ();

  inline void predictE(int *p, int *l, int *r, int *o);
  inline void predictE2(int *l, int *r, int *o);				  
  inline void predictE(vector<int>& father, vector<int>&qv, int i);

  inline void predictC(int *p, int *l, int *r, int *o);
  inline void predictC2(int *l, int *r, int *o);	             
  inline void predictC(vector<int>& father, vector<int>&qv, int i, int *o,
					vector<int>&index);	


         bool           decode(MPG_InputBitStream& ibstrm);
         bool           encode(MPG_OutputBitStream& obstrm);
};

inline MPG_Property::~MPG_Property()
{}

inline MPG_Property::MPG_Property(MPG_Type       type,
				  MPG_Binding    binding,
				  MPG_Prediction prediction):
  _type(type), _binding(binding), _prediction(prediction)
{
  _bbq = NULL;
  _nrq = NULL;
}

inline void MPG_Property::setType(MPG_Type type)
{ _type = type; }

inline MPG_Type MPG_Property::getType()
{ return _type; }

inline void MPG_Property::setBinding(MPG_Binding binding)
{ _binding = binding; }

inline MPG_Binding MPG_Property::getBinding()
{ return _binding; }

inline void MPG_Property::setPrediction(MPG_Prediction prediction)
{ _prediction = prediction; }

inline MPG_Prediction MPG_Property::getPrediction()
{ return _prediction; }

inline int MPG_Property::getDimensionValue()
{
  int dimV = 0;
  switch(_type) {
  case PT_COORD:    dimV = 3; break;
  case PT_NORMAL:   dimV = 3; break;
  case PT_COLOR:    dimV = 3; break;
  case PT_TEXCOORD: dimV = 2; break;
  default:          dimV = 0; break;
  }
  return dimV;
}

inline int MPG_Property::getDimensionCode()
{
  int dimC = 0;
  switch(_type) {
  case PT_COORD:    dimC = 3; break;
  case PT_NORMAL:   dimC = 1; break;
  case PT_COLOR:    dimC = 3; break;
  case PT_TEXCOORD: dimC = 2; break;
  default:          dimC = 0; break;
  }
  return dimC;
}

inline int MPG_Property::getDimensionLambda()
{ return _lambda.size(); }

inline float* MPG_Property::getLambda()
{ return (_lambda.size()>0)?_lambda.begin():(float*)0; }

inline float  MPG_Property::getLambda(int i)
{ return _lambda[i]; }

inline int MPG_Property::getNumberOfFathersPerValue()
{
  int nFatherPerValue = (_prediction==PP_NONE         )?0:
                        (_prediction==PP_TREE         )?1:
                        (_prediction==PP_PARALLELOGRAM)?3:0;
 return nFatherPerValue;
}

inline void MPG_Property::quantizeLambda()
{
  if(_lambda.size()>0)
    {
      int nLambda = _lambda.size();
      _lambda[nLambda-1] = 1.0;
      int i;
      for(i=0;i<nLambda-1;i++)
	{
	  int iLambda = (int)(ldexp(_lambda[i],_bitsPerCoord));
	  _lambda[i] = iLambda;
	  _lambda[i] = ldexp(_lambda[i],-_bitsPerCoord);
	  _lambda[nLambda-1] -= _lambda[i];
	}
    }
}

inline void MPG_Property::setDefaultLambda()
{
  _lambda.erase(_lambda.begin(),_lambda.end());
  if(_prediction==PP_TREE)
    {
      _lambda.insert(_lambda.end(),(size_t)1,1.0);
    }
  else if(_prediction==PP_PARALLELOGRAM)
    {
      _lambda.insert(_lambda.end(),(size_t)3,1.0);
      _lambda[2] = -1.0;
    }
  quantizeLambda();
}

inline void MPG_Property::setBBoxQ(MPG_BBoxQ* bbq)
{
  _bbq = bbq;
  _bitsPerCoord = _bbq->getBitsPerCoordinate();
}

inline MPG_BBoxQ* MPG_Property::getBBoxQ()
{ return _bbq; }

inline void MPG_Property::setNormalQ(MPG_NormalQ* nrq)
{
  _nrq = nrq;
  _bitsPerCoord = _nrq->getBitsPerNormal();
}

inline MPG_NormalQ* MPG_Property::getNormalQ()
{ return _nrq; }

inline void MPG_Property::predictE(int *p, int *l, int *r, int *o)
{
  if(_prediction==PP_PARALLELOGRAM)
    {
      int nD = getDimensionValue();
      int j;
      for(j=0;j<nD;j++)
	{
	  float predictor_j = _lambda[2]*p[j]+_lambda[1]*l[j]+_lambda[0]*r[j];

	  int xx;
	  if (predictor_j > 0.0)
	    xx = (int) (predictor_j+0.5);
	  else
	    xx = (int) (predictor_j-0.5);

	  o[j] += xx;
	  o[j] &= (1<<_bitsPerCoord)-1;
	}
    }
}

inline void MPG_Property::predictE(vector<int>& father, vector<int>&qv, int i)
{
  if(_prediction==PP_TREE)
    {
      int nD = getDimensionValue();
      int nL = getDimensionLambda();

      int l,j,k;
      for(j=0;j<nD;j++)
	{
	  float predictor_j = 0.0;
	  for(l=i,k=0;k<nL;k++)
	    { l=father[l]; predictor_j += _lambda[k]*qv[l*nD+j]; }

	  int xx;
	  if (predictor_j > 0.0)
	    xx = (int) (predictor_j+0.5);
	  else
	    xx = (int) (predictor_j-0.5);

	  qv[nD*i+j] += xx;
	  qv[nD*i+j] &= (1<<_bitsPerCoord)-1;
	}
    }
}

//for per vertex with two available ancestors
inline void MPG_Property::predictE2(int *l, int *r, int *o)
{
  if(_prediction==PP_PARALLELOGRAM)
    {
      int nD = getDimensionValue();
      int j;
      for(j=0;j<nD;j++)
	{
	  float predictor_j = (l[j]+r[j])/2;

	  int xx;
	  if (predictor_j > 0.0)
	    xx = (int) (predictor_j+0.5);
	  else
	    xx = (int) (predictor_j-0.5);

	  o[j] += xx;
	  o[j] &= (1<<_bitsPerCoord)-1;
	}
    }
}

inline void MPG_Property::predictC(int *p, int *l, int *r, int *o)
{
  if(_prediction==PP_PARALLELOGRAM)
    {
      int nD = getDimensionValue();
      int j;
      for(j=0;j<nD;j++)
	{
	  float predictor_j = _lambda[2]*p[j]+_lambda[1]*l[j]+_lambda[0]*r[j];

	  int xx;
	  if (predictor_j > 0.0)
	    xx = (int) (predictor_j+0.5);
	  else
	    xx = (int) (predictor_j-0.5);

	  o[j] -= xx;
	  o[j] <<= 32-_bitsPerCoord;
	  o[j] >>= 32-_bitsPerCoord;
	}
    }
}

inline void MPG_Property::predictC2(int *l, int *r, int *o)
{
  if(_prediction==PP_PARALLELOGRAM)
    {
      int nD = getDimensionValue();
      int j;
      for(j=0;j<nD;j++)
	{
	  float predictor_j = (l[j]+r[j])/2;

	  int xx;
	  if (predictor_j > 0.0)
	    xx = (int) (predictor_j+0.5);
	  else
	    xx = (int) (predictor_j-0.5);

	  o[j] -= xx;
	  o[j] <<= 32-_bitsPerCoord;
	  o[j] >>= 32-_bitsPerCoord;
	}
    }
}

inline void MPG_Property::predictC(vector<int>& father,
				   vector<int>&qv,
				   int i,
				   int* o,
				   vector<int>&index	
				   )
{
  if(_prediction==PP_TREE)
    {
      int nD = getDimensionValue();
      int nL = getDimensionLambda();
      int j,k,fk;
      for(j=0;j<nD;j++)
	{
	  float predictor_j = 0.0;
	  for(fk=i,k=0;k<nL;k++)
	    { fk=father[fk]; predictor_j += _lambda[k]*qv[index[fk]*nD+j]; }

	  int xx;
	  if (predictor_j > 0.0)
	    xx = (int) (predictor_j+0.5);
	  else
	    xx = (int) (predictor_j-0.5);

	  o[j] = qv[nD*index[i]+j] - xx;
	  o[j] <<= 32-_bitsPerCoord;
	  o[j] >>= 32-_bitsPerCoord;
	}
    }
}

#endif // _MPG_PROPERTY_H_
