//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------

#ifndef _MPG_Simplify_h_
#define _MPG_Simplify_h_

/* 

   apply edge collapses to a triangular mesh respecting the topological
   constraints described in Taubin et al, Siggraph 98 
   "Progressive Forest Split Compression"

 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <memory.h>
#include  <sys/types.h>
#include "MPG_Property.h"
#include "MPG_Heap.h"
#include "MPG_Lex.h"

#undef  bzero
#define bzero(pointer, size)         memset ((void *) (pointer), 0, (int) (size))

/* because they are not standard accross platforms
   (unless the include files were not completely right)
   I have had to re-typedef the following standard definitions at some point */

typedef unsigned short U_short;
typedef unsigned char  U_char;

/* define minimum and maximum of two numbers */

#undef  MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))

#undef  MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))


/* some vector operations: */

#undef  SQUARED_DISTANCE3
#define SQUARED_DISTANCE3(u,v) (((u)[0]-(v)[0])*((u)[0]-(v)[0]) + ((u)[1]-(v)[1])*((u)[1]-(v)[1]) +\
                                ((u)[2]-(v)[2])*((u)[2]-(v)[2]))


     /* At some point, I might wast to make sure that sqrt is called on non-zero input.
	I will have to replace the macro with a function call */
	
#undef  DIST3
#define DIST3(u,v)             ( sqrt ( SQUARED_DISTANCE3((u),(v))))

#undef  SCALPROD3
#define SCALPROD3(u,v)         ( (u)[0]*(v)[0] + (u)[1]*(v)[1] + (u)[2]*(v)[2] )

#undef  NORM3
#define NORM3(u)               ( sqrt ( SCALPROD3((u),(u)) ) )


#undef  VECPROD3
#define VECPROD3(u,v,w) { (w)[0] = (u)[1]*(v)[2]-(u)[2]*(v)[1]; (w)[1] = (u)[2]*(v)[0]-(u)[0]*(v)[2]; \
                          (w)[2] = (u)[0]*(v)[1]-(u)[1]*(v)[0];}

#undef  MAKEVEC3
#define MAKEVEC3(u,v,w) {(w)[0] = (v)[0] - (u)[0]; (w)[1] = (v)[1] - (u)[1]; (w)[2] = (v)[2] - (u)[2];}

#undef  ADDVEC3
#define ADDVEC3(u,v) {(u)[0] += (v)[0]; (u)[1] += (v)[1];  (u)[2] += (v)[2];}

#undef  TEXMAP_AREA
#define TEXMAP_AREA(uv, area) {\
                            register float u[2],v[2];\
                            u[0] = (uv)[2]-(uv)[0]; u[1] = (uv)[3]-(uv)[1]; v[0] = (uv)[4]-(uv)[0]; v[1] = (uv)[5]-(uv)[1];\
                            (area) = (u[0]*v[1]-u[1]*v[0])/2.;\
                              }

extern int _smpTopoSimplifyTriangleMesh(
				 int nV, float *v, int *new_nV,
				 int nT, int *t,   int *new_nT,
				 float *colors, float *normals, float *texCoords,
				 MPG_Binding colorBinding, MPG_Binding normalBinding, MPG_Binding texCoordBinding,
				 float tolerance, int *vertex_parents, 
				 float *error_volume, int move_vertices, int simplify_boundary);

extern void* _smpAllocateZero(int size);

				    
extern float _smpFloatDataBoundingBoxDiagonal(float *data, int n_data, int data_dim);
/* compute the bounding box of any multidimensional array of data */
    

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpComputeVertexValence                                               |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

/* using the set of triangles, compute the valence of each vertex */

#define _smpComputeVertexValence(nT, t, valence)                               \
{                                                                              \
  register int i, i1;                                                          \
  register U_char k;                                                           \
                                                                               \
  /* for each triangle, increment the valence of its three vertices */         \
                                                                               \
  for (i=i1=0; i<(nT); i++, i1+=3) for (k=0; k<3; k++) (valence)[(t)[i1+k]]++; \
										 }



/* data structures for the surface edges and associated routines */


typedef struct _EdgeS_ {

  short            label;               /* used for a flag or for
					   classifying edges among boundary, regular and singular edges */
  int		    v[2];		/* end points */
  int		    t[2];		/* incident triangles */
} EdgeS;

/*
  surface edges have an orientation:
  v[0] >= v[1] and t[0] is before t[1] in clockwise order
  when we rotate around v0.

                      /|\v1
                     / | \
                    /  |  \              
                   / t0|t1 \             
                   \   |   /
                    \  |  /
		     \ | /
		      \|/v0    v0 >= v1
*/

/* routines for indexing edges in a hash table */

extern int         _smpEdgeSKey(EdgeS *);
extern int         _smpEdgeSFilter(EdgeS *, EdgeS *);


#define SurfEdgeLabel(edg)           (edg)->label
#define SurfEdgeGetVertices(edg)     (edg)->v
#define SurfEdgeGetTriangles(edg)    (edg)->t


/*...........................................................................*/

/* Following are includes needed for hash-tables using linear probing.
   In case of a code conflict, data items are chained */

/*...........................................................................*/

typedef char  *RECTYPE ;
typedef char  *KEYTYPE ;

typedef struct _nodetype_{
  int   code;
  RECTYPE rec;
  struct _nodetype_ *next;
} Nodetype;

typedef struct _Block_{
  RECTYPE rec;
  struct _Block_ *next;
} Block;

typedef struct _htable_{
  Nodetype **bucket;
  Block     *block;
  Nodetype  *freenode;

  int nodes_per_block ;
  int size, mask;
  int        search_code;
  Nodetype  *search_node;
  KEYTYPE     search_key;

  int     (*fcode)(KEYTYPE);
  int     (*filter)(RECTYPE, RECTYPE);
  KEYTYPE (*compute_key)(RECTYPE);

} Htable;

/*...........................................................................*/

/* prototypes for H-table procedures 
*/
/*...........................................................................*/

/* add a record to the hash-table */

extern int         _smpAddRecord(Htable *, RECTYPE, KEYTYPE);
extern KEYTYPE     _smpIdentifyRecord2Key(RECTYPE rec);


/* initialize the hash-table */                                                            

extern int         _smpInitHashTable(Htable *, int, int (*fcode)(KEYTYPE), 
				     int (*filter)(RECTYPE,RECTYPE), 
				     KEYTYPE (*compute_key)(RECTYPE));

/* determine whether a particular record is in the hash-table and if yes, find that record 
   meaning, return a pointer to that record*/

extern char        *_smpFindRecord(Htable *, KEYTYPE);
extern void        _smpHashTableReset(Htable *);

extern int        _smpNewBlock(Htable *);

/* once _smpFindRecord() has been called, retrieve the next record with the same code and
   passing through the same filter() from the hash table */

extern char        *_smpNextRecord(Htable *);
extern Nodetype    *_smpgetnode(Htable *);
extern Nodetype    *_smpnewnode(Htable *,RECTYPE, int);

extern int          _smpFlog2(int n);/* log base two using integer arithmetic */

/* reset the hash table and free the buckets
   after that, either (1) inithashtable must be called again to refill the hash table with new records or
   completely different data
   
   or (2) free(table) to finish it off
   unless it was defined as Htable table[1] in which case nothing more has to be done */

#define       _smpFreeHashTable(table) {_smpHashTableReset(table);if ((table)->bucket) \
                                       {free((void*)(table)->bucket);(table)->bucket = NULL;}}


/* find a particular surface edge in a Hash table indexing edges */

extern EdgeS *_smpFindEdge(Htable *e, int vA, int vB);

/* build the edge table and detect some non-manifold situations */

extern int    _smpBuildEdgeTable(int nV,   int nT,   int *t,
				 int *n_edges, Htable *e_table, EdgeS **edges);

extern int    _smpNoDegenerateTriangles(int nV, int nT, int *t);

extern int    _smpNoStandaloneVertices(int nV, int nT, int *t);

extern int    _smpIndexManifoldEdges(int *nE, EdgeS **edges, Htable *e_table, int nT, int *t);

extern int    _smpFillEdgeTable(int *n_edges, EdgeS *edges, Htable *e_table, int nT, int *t);

extern void   _smpInitFather(int n, int *father);


/* perform path conpression on the representatives of a forest data structure
   the macro FATHER is more efficient than the _smpFather(i,f) procedure
   it calls the procedure only if necessary */

#define FATHER(i, f) (((i) == (f)[i]) ? (i) : \
                      ((f)[i] == (f)[(f)[i]]) ? (f)[i] : _smpFather(i,f))

extern int _smpFather(int i, int *father);

/*...........................................................................*/
/* 
	default parameters for simplification
*/
/*...........................................................................*/


#define SIMPLIFY_MAX_ANGLE_NOR      0.6     /* maximum angle between normals of corresponding triangles
					       before and after an edge-collapse */
#define SIMPLIFY_COMPACTNESS_RATIO  .5      /* maximum ratio between the minimum compactness
					       in the vertex star after simplification and the
					       minimum compactness in the edge star before simplification */
#define SIMPLIFY_VALENCE_MAX          30    /* maximum valence at a vertex after simplification */
#define SIMPLIFY_VALENCE_MAX1         31
#define SIMPLIFY_VALENCE_MAX4         34
#define SIMPLIFY_TILING_VERTICES_MAX  2051

#define FOUR_SQRT_3 6.928203230275508

/* typedef for some simple geometric entities: */

typedef float  Vertex[3];
typedef float  Plane[4];
typedef double VertexD[3];
typedef int    Face[3];

#define DIRECTION_COUNTER_CLOCKWISE 2
#define DIRECTION_CLOCKWISE         1




class  SimpData {
  
  /* data marked "to be allocated" is managed by the 

     _smpCreateSimplification....

     _smpFreeSimplification....

     routines.
     */

public:

  MPG_Heap   
        *edge_heap;                  /* to be allocated */

  int   
        nV, nT, nE, 
        *triangles,
        *edge2index,                 /* to be allocated */
        *index2edge,                 /* to be allocated */
        *vertex_father,
        *edge_father,                /* to be allocated */
        *triangle_father,
        *removed_component_father,
        move_vertices,
	simplify_boundary,
        valence_max,

    /* various counters: */

        num_edg_remaining_4_testing,
        num_edg_collapsed,
        num_edg_weights,
        edg_rejected_4_topology,
        edg_rejected_4_geometry,
        edg_rejected_4_tolerance,
        heap_outside_index;       /* special index used for counting the number of edges added to the heap
				     and making sure we aren't deleting them twice */
 
  EdgeS 
        *edge_array;

  Htable 
        *e_table;

  Vertex
        *vert,
        *face_normals,             /* to be allocated */
        simplified_vertex,
        simplified_vertex_alt;

  Vertex
         *vx_color,                /* vertex properties */
         *vx_normal,
         psv_normal,
         psv_color;
  float 
        *vx_texCoord,
        psv_texCoord[2],

        *vx_normal_error,
        *vx_color_error,
        *vx_texCoord_error,

        vx_normal_psve, /* potential simplified vertex error */
        vx_color_psve,
        vx_texCoord_psve,

        vx_normal_tol,
        vx_color_tol,
        vx_texCoord_tol,

        normal_error_penalty; /* penalty value used for dividing the normal error, 
				 accounting for its non-linear variation */

    float 
        *colors,
        *normals,
        *texCoords;

  MPG_Binding
        colorBinding,
        normalBinding,
        texCoordBinding;

    float 
        _texture_area,
        _collapsed_texture_area,
        _inc_collapsed_texture_area;
      
  U_char 
        *boundary_vert;            /* to be allocated */

  U_short 
        *valence,                  /* to be allocated */
        *red_valence,              /* to be allocated */
        *ori_valence;              /* to be allocated */
  
  float 
        tolerance,
        data_tolerance,
        *err_volume,               
        *tol_volume,               /* to be allocated */
        *face_area,                /* to be allocated */
        *compactness,              /* to be allocated */ 
        min_scalprod_nor,
        compactness_ratio;
        
  /* other buffers used for storing information relative to vertex and edge stars */

  int 
      *_vertex_star_buffer;

  void 
      *_edge_star_buffer_fl,
      *_edge_star_buffer_vx;

  // constructor and destructor

  SimpData() {}
  ~SimpData() {}


  // member functions

       
  int    _smpCreateSimplificationDataStructure();

  // free all the simplification data structures: 
  
  int    _smpFreeSimplificationDataStructure();

  /* free only the simplification data structures that are not exported by the driver routine,
     to avoid running out of memory in _smpBuildSimplifiedSurface() */

  int    _smpPartialFreeSimplificationDataStructure();

  int    _smpBuildSimplifiedSurface(int& new_nV, int& new_nT);

  int    _smpAddEdge2Heap(int the_edge);

  int    _smpReinstateEdgesInHeap(int *estar, U_short val, int *num_edg_added);

  int    _smpRemoveEdgesFromHeap(int *estar, U_short val);
 
  int    _smpRemoveLowestWeightEdgeFromHeap();

  int    _smpMarkEdgesAdjacent2Boundary();

  int    _smpBuildEdgeHeap();

  void   _smpBuildParentEdgeStars(EdgeS *edg, int v0, int v1, U_short val0, U_short val1, 
				  int *star0, int *star1);

  void   _smpParentVertexStar(int vf, int t0, U_short val, int *star);

  int    _smpDirectedParentVertexStar(int parent_vertex, int first_triangle,
				      int valence, int *star, int direction);

  void   _smpBuildEdgeStar(Vertex *star_vert, Face *star_face, Plane *star_plane, float *star_areas, 
			   float *star_comp, int v0, int v1, U_short val0, U_short val1, 
			   int *star0, int *star1);

  int    _smpCollapseEdge(int v0, int v1, int *star0, int *star1, U_short val0, U_short val1, 
			  Vertex *nor0, Vertex *nor1, float *area0, float *area1, 
			  float *comp0, float *comp1);

  int    _smpCollapseBoundaryEdge1stKnd(int edg_num, int v0, int v1, int val0, 
					int val1, int *star1, int *vstar1, int *estar1,
					Vertex *s_normal1, float *s_area1);

  int    _smpCollapseBoundaryEdge2ndKnd(int edg_num, int v0, int v1, int val0, 
					int val1, int *star0, int *star1, int *vstar0, int *vstar1, 
					int *estar0, int *estar1, Vertex *s_normal0, Vertex *s_normal1, 
					float *s_area0, float *s_area1, int parent, int child);

  int    _smpSimplifiedVertexLocation(int v0, int v1, U_short val0, U_short val1, 
				      U_short star_val, Vertex *v, Face   *f, Plane  *p, float *area, 
				      float *w, int method);

  int    _smpSimplifiedBoundaryVertexLocation(int v0, int v1, int *vstar0, int *vstar1, 
					      int val0, int val1, int method, int parent);

  int    _smpCollapsibilityTestsBoundaryEdge1stKind(int edge_num, int v0, int v1, int val0, int val1);

  int    _smpCollapsibilityTestsBoundaryEdge2ndKind(int edge_num, int v0, int v1, int val0, int val1, 
						    int move_simplified_vertex);

  int    _smpBoundaryCollapse2KndGeometricallyOk(int v0, int v1,
						 int *star0, int *vstar0, int val0, Vertex *s_normal0, float *s_area0,
						 int direction, float min_scalprod_nor, float compactness_ratio);

  int    _smpBoundaryCollapse1stKndGeometricallyOk(int v0, int *star0, int *vstar0, 
						   int val0, Vertex *s_normal0, float *s_area0,
						   float min_scalprod_nor, float compactness_ratio);

  int    _smpColorPerFaceOk(int fa, int fb, int n_adj0, int *neighboring_faces0, int n_adj1, int *neighboring_faces1);

  int    _smpTexCoordPerCornerOk(int fa, int fb, int n_adj0, int *neighboring_faces0, 
				 int n_adj1, int *neighboring_faces1);

  int    _smpManifoldLinkTest1stKnd(int v0, int val0, int *star0, int *star1, int *link1, int val1);

  int    _smpErrorWithinTolerance(int *star0, int *star1, int val0, int val1);

  int    _smpErrorWithinToleranceBoundary2ndKnd(int v0, int v1, int *vstar0, int *vstar1, int val0, int val1,
						int parent, int child);

  int    _smpErrorWithinToleranceBoundary1stKnd(int v0, int v1, int *vstar1, int val1);


  void   _smpErrorTestVertexPropertiesInit(int old_v1, int old_v2, int old_v3, 
					   float alpha_o_v1, float alpha_o_v2, float alpha_o_v3);

  int    _smpPenaltyFactor4ErrorOnNormals(int val0, int val1, int *vstar0, int *vstar1);

  int    _smpPenaltyFactor4ErrorOnNormalsBoundary2ndKind(int v0, int v1,
							 int val0, int val1, int *vstar0, int *vstar1);

  int    _smpPenaltyFactor4ErrorOnNormalsBoundary1stKind(int v1, int *vstar1, int val1);

  int    _smpErrorTestVertexPropertiesUpdate(int sv, int new_v2, int new_v3,
					     float alpha_simp_vertex, float alpha_n_v2, float alpha_n_v3,
					     int old_v1, int old_v2, int old_v3, 
					     float alpha_o_v1, float alpha_o_v2, float alpha_o_v3);

  void   _smpErrorTestVertexPropertiesEnd(int v0);

  int    _smpNoInteriorVerticesInRemovedFaceSetUpdateRedValence(int the_collapsed_edge);

  int    _smpManifoldRemovedFaceSet(int the_collapsed_edge);
  
  int    _smpManifoldVertexRemovedSet(int v, int *triangles, int first_tri);

  int    _smpRemovedFaceForest(int the_collapsed_edge, int verify_forest, int update_representative);

  int    _smpSimplifyManifoldSurface();

};


extern int   _smpTrianglesNormalAreaCompactness( int nT, Face *t, Vertex *v, Vertex *t_normal,
						float *t_area, float *t_compactness);

extern float _smpTriangleNormal(Vertex *tri, Vertex n);

extern float _smpNormalize3(float *vec);

extern int   _smpFlagBoundaryVertices(int nE, EdgeS *edges, U_char *boundary_vert);
 
extern void  _smpInitArray(char *array, char *data, int n, int size);

extern int   _smpRotateParentTriangle(int  parent_vertex, int *vertex_fathers, 
				      int *tri_v, int *tri_v_parents, int *tri_v_rotated);

extern int   _smpRotateParentEdge(int parent_vertex, int *vertex_fathers,
				  int *edge_v, int *edge_t, int *edge_v_parents, 
				  int *edge_v_rotated, int *edge_t_rotated);

/* macros used for finding the next edge when rotating around a vertex in
   on a simplified surface */

#define _smpNextParentEdge(oriented_tri, directionCW)                                                     \
	 _smpFather(                                                                                      \
	  (int) (((EdgeS *)_smpFindEdge(e_table,(oriented_tri)[0],(oriented_tri)[directionCW]))\
		 -edge_array), edge_father)



#define _smpNextParentTriangle(oriented_edge_t, directionCW)                     \
             ((directionCW) == 1)? FATHER((oriented_edge_t)[1], triangle_father):\
				   FATHER((oriented_edge_t)[0], triangle_father)

extern int   _smpManifoldLinkTest(int *link0, int *link1, int val0, int val1);

extern int   _smpCmpIntSmallFirst(char *s1, char *s2);

extern float _smpFastCompactness3DTriangle2(Vertex *tri, float *triangle_area);



/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpTriangleBasisVectors                                               |
  |                                                                          |
  +--------------------------------------------------------------------------+*/



#define  _smpTriangleBasisVectors(tri, x1, x2, origin)                         \
{                                                                              \
  /* compute two basis vectors of a triangle, such that the first basis vector \
     will correspond to the shortest edge,                                     \
     return the number of the vertex that is chosen as the triangle origin */  \
                                                                               \
                                                                               \
  int the_origin = 0;                                                          \
                                                                               \
  /* I- find the shortest edge of the triangle                                 \
                   2                                                           \
                  /\                                                           \
          edg1   /  \ edg0                                                     \
                /____\                                                         \
               0 edg2 1 */                                                     \
                                                                               \
  register U_char j;                                                           \
  double min_len = SQUARED_DISTANCE3((tri)[0],(tri)[2]);                       \
  U_char shortest_edge = 1;                                                    \
                                                                               \
  for (j=0;j<2;j++)                                                            \
    {                                                                          \
      double len = SQUARED_DISTANCE3((tri)[j],(tri)[j+1]);                     \
                                                                               \
      if (len < min_len)                                                       \
	{                                                                      \
	  min_len = len;                                                       \
	  shortest_edge = (j+2)%3;                                             \
	}                                                                      \
      }                                                                        \
                                                                               \
  the_origin = (shortest_edge + 1) %3;                                         \
                                                                               \
  (origin) = the_origin;                                                       \
                                                                               \
  /* II- assign the smallest edge to one of the vectors x1 and x2 */           \
  {                                                                            \
    U_char                                                                     \
      dest1 =  (shortest_edge+2)%3,                                            \
      dest2 =   shortest_edge;                                                 \
                                                                               \
    register U_char k;                                                         \
                                                                               \
    for(k=0;k<3;k++)                                                           \
      {                                                                        \
	(x1)[k] = (tri)[dest1][k] - (tri)[the_origin][k];                      \
	(x2)[k] = (tri)[dest2][k] - (tri)[the_origin][k];                      \
      }                                                                        \
  }                                                                            \
										 }


extern float  _smpClosestPointOnEdge(Vertex X, Vertex XP, Vertex A, Vertex B, float *t);

extern int    _smpClosestPointOnTriangle(Vertex *tri, Vertex w, Vertex wp, Vertex bary, float *dist);

extern int    _smpTriangle3DBarycentricCoordinates(Vertex *tri, Vertex w, Vertex wp, Vertex bary, float *residual);

extern float  _smpSegmentIntersection(Vertex A, Vertex B, Vertex C, Vertex D,
				      float *lambda, float *mu, int changed_AorB, int changed_C, int changed_D);

extern void   _smpCopyFaceNormalsAreasCompactness(Plane *plane, float *s_area, float *s_comp, 
						  int *star0, int *star1, int val0, int val013, 
						  Vertex *t_normal, float *t_area, float *t_comp);

extern float  _smpMinFloatArray(float *array, int n);


extern int     _smpPositionVertexLSConstraintsWArea(Plane *p, float *areas, U_short val, Vertex simpv);

extern void   _smpPlaneForIdenticalVolume(U_short val0, U_short val1, Plane p, Vertex *v, Face *f, float volume);

extern double _smpStarPlaneEquationsAndVolume(Vertex *star_vert,  Face *star_face, 
					      Plane  *star_plane, float *star_areas, 
					      int   star_val);


extern int _smpHouseholderPreMultiplication( float *A, int mA,  float *v, int m,   int n, float *w );

extern int _smpSolveQuadraticEquation(double A, double B, double C, 
				      double *sol1, double *sol2, double eps);

extern int _smpSolveSymmetric2x2Eqn(double a,  double b, double d, double e, double f,
				    double *x, double *y);

extern int _smpCollapseTopologicallyFeasible(int *vstar0, int *vstar1, 
					     U_short val0, U_short val1);

extern int _smpCollapseGeometricallyFeasible(Vertex *svert, Face *sface, 
					     Plane *splane, float *old_comp, float new_comp_min,
					     Vertex *snor0, Vertex *snor1,   
					     U_short val0, U_short sval, 
					     float min_scalprod, float compactness_ratio);

extern int    _smpNoFlipOverCheck(U_short first_face, U_short last_face, 
				  Vertex *nor, Plane *plane, float min_scalprod_nor);

extern void   _smpSimplifiedStarNormals(Face *sface, Vertex *svert, Vertex *snor, float *sarea, float *scomp, 
					Vertex simpvert, U_short first_face, U_short last_face);

extern void   _smpReplaceFaceNormals(int *star, Vertex *new_nor, float *new_area, float *new_comp, 
				     Vertex *nor, float *area, float *comp, U_short val);

extern int    _smpAreValuesAllDifferent(int n_elements, char *array_of_elements, 
					int element_size);

extern int    _smpNextTriangle(int *triangle, Htable *e_table, int theVertex, U_char ccw);


// defines and functions relative to simplifying data on surfaces

#define SIMPLIFY_VERTEX_NORMAL_TOL .75

/*  .75:  45 degrees of rotation tolerated for the normal 
    .50:  30 degrees 
    .25:  15 degrees */

#define SIMPLIFY_VERTEX_COLOR_TOL .4     /* sqrt (r^2 + g^2 + b^2) distance in rgb space */

#define SIMPLIFY_VERTEX_TEXCOORD_TOL .1  /* Euclidean distance in texture coordinate (u,v) space */

#define SIMPLIFY_FACE_COLOR_TOL   .2     /* distance in rgb space from the removed face to neighboring faces */ 

#define SIMPLIFY_TOT_COLLAPSED_TEXTURE_AREA_RATIO .1 /* max percentage of the texture area that is removed 
							with 20 levels, at the worst .9 ^20 = .12 area left */
#define SIMPLIFY_INC_COLLAPSED_TEXTURE_AREA_RATIO .1 /* allowed incremental removal percentage within an edge star */

extern void _smpMakeCentroid(int nV, Vertex *vv, Vertex centroid);
extern void _smpMakeBarycenter(int nV, Vertex *vv, Vertex bary, float *bary_coord);

extern void _smpRedistributePerFaceProperty(int nD, float *property, int nF, int new_nF, int *f_map, 
					    float *face_areas);

extern void _smpRedistributePerCornerProperty(int nD, float *property, int *t, int nT, int *new_t, int new_nT, 
					       int *v_map, int *t_map, float *face_areas);

#endif  /* _pfs_simplify_ */

