//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------

#ifndef _MPG_VGraph_h_
#define _MPG_VGraph_h_

#include <vector>
#include "MPG_Stream.h"
#include "MPG_Loops.h"
#include "MPG_Graph.h"
#include "MPG_QFac.h"

class MPG_VGraph {

protected:

  vector<char> _bits;
  vector<int>  _length;
  vector<int>  _jumpTo;

public:

  static const char LAST;
  static const char START;
  static const char LEAF;
  static const char JUMP;
  static const char J_INV_OR;

  static const char NO_LEAF;
  static const char NO_START;
  static const char NO_LAST;
  static const char NO_JUMP;
  static const char NO_J_INV_OR;

public:

  inline             ~MPG_VGraph();
  inline              MPG_VGraph();

  inline void         clear();

  inline vector<char>& getBitsData() { return _bits; } 
  inline vector<int>&  getLenData() { return _length; } 
  inline vector<int>&  getJumpData() { return _jumpTo; } 
  inline int          getNumberOfRuns();
  inline int          getNumberOfEdges();
  inline int          getNumberOfVertices();
  inline int          getNumberOfJumps();

  inline bool         setRun(int i,
                             int last   = 0,
                             int start  = 0,
                             int length = 0,
                             int leaf   = 0,
                             int jump   = 0,
                             int jInvOr = 0,
                             int jumpTo  = 0);

  inline bool         getRun(int i,
                             int& last,
                             int& start,
                             int& length,
                             int& leaf,
                             int& jump,
                             int& jInvOr,
                             int& jumpTo);

  inline void         pushBackRun(int last   = 0,
                                  int start  = 0,
                                  int length = 0,
                                  int leaf   = 0,
                                  int jump   = 0,
                                  int jInvOr = 0,
                                  int jumpTo = 0);

  inline bool         popBackRun(int& last,
                                 int& start,
                                 int& length,
                                 int& leaf,
                                 int& jump,
                                 int& jInvOr,
                                 int& jumpTo);

  // expand methods

  bool         getFather(vector<int>& v_father);

  bool         toLoop(MPG_Loops& loop);

  bool         encode(MPG_QFac &codec);

  // compress methods

  bool         fromLoop(MPG_Loops& loop);
  bool         fromLoop(MPG_Loops& loop,
                        vector<int>& v_traversal,
                        vector<int>& v_father);

  bool         decode(MPG_QFac &codec);

};

typedef MPG_VGraph* MPG_pVGraph;
#ifndef __GNUC__
// to make the compiler happy
inline void destroy(MPG_VGraph* pointer) {}
inline void destroy(MPG_pVGraph* pointer) {}
#endif // __GNUC__

inline MPG_VGraph::~MPG_VGraph()
{ }

inline MPG_VGraph::MPG_VGraph()
{ }

inline void MPG_VGraph::clear()
{
  _bits.erase(_bits.begin(),_bits.end());
  _length.erase(_length.begin(),_length.end());
  _jumpTo.erase(_jumpTo.begin(),_jumpTo.end());
}

inline int MPG_VGraph::getNumberOfRuns()
{ return _bits.size(); }

inline int MPG_VGraph::getNumberOfEdges()
{
  int nEdges = 0;
  vector<int>::iterator iLength;
  for (iLength=_length.begin();iLength!=_length.end(); iLength++)
    nEdges += *iLength;
  return nEdges;
}

inline int MPG_VGraph::getNumberOfVertices()
{ return 1+getNumberOfEdges()-getNumberOfJumps(); }

inline int MPG_VGraph::getNumberOfJumps()
{
  int nJumps = 0;
  vector<char>::iterator iBits;
  for (iBits=_bits.begin();iBits!=_bits.end(); iBits++)
    if(!((*iBits) & START)) nJumps++;
  return nJumps;
}

inline bool MPG_VGraph::
setRun(int i,
       int last, int start, int  length, int leaf,
       int jump, int jInvOr, int  jumpTo)
{
  if(i<0 || i>=getNumberOfRuns()) return false;
  char bits = 0x00;
  if(last)   bits |= LAST;
  if(start)  bits |= START;
  if(leaf)   bits |= LEAF;
  if(jump)   bits |= JUMP;
  if(jInvOr) bits |= J_INV_OR;
  _bits[i]   = bits;
  _length[i] = length;
  _jumpTo[i] = jumpTo;
  return true;
}

inline bool MPG_VGraph::
getRun(int i,
       int& last, int& start, int&  length, int& leaf,
       int& jump, int& jInvOr, int&  jumpTo)
{
  if(i<0 || i>=getNumberOfRuns()) return false;
  char bits = _bits[i];
  last   = (bits & LAST)?1:0;
  start  = (bits & START)?1:0;
  leaf   = (bits & LEAF)?1:0;
  jump   = (bits & JUMP)?1:0;
  jInvOr = (bits & J_INV_OR)?1:0;
  length = _length[i];
  jumpTo = _jumpTo[i];
  return true;
}

inline void MPG_VGraph::
pushBackRun(int last, int start, int  length, int leaf,
            int jump, int jInvOr, int  jumpTo)
{
  char bits = 0x00;
  if(last)   bits |= LAST;
  if(start)  bits |= START;
  if(leaf)   bits |= LEAF;
  if(jump)   bits |= JUMP;
  if(jInvOr) bits |= J_INV_OR;
  _bits.push_back(bits);
  _length.push_back(length);
  _jumpTo.push_back(jumpTo);
}

inline bool MPG_VGraph::
popBackRun(int& last, int& start, int&  length, int& leaf,
           int& jump, int& jInvOr, int&  jumpTo)
{
  if(0>=getNumberOfRuns()) return false;
  char bits = _bits.back(); _bits.pop_back();
  last   = (bits & LAST)?1:0;
  start  = (bits & START)?1:0;
  leaf   = (bits & LEAF)?1:0;
  jump   = (bits & JUMP)?1:0;
  jInvOr = (bits & J_INV_OR)?1:0;
  length = _length.back(); _length.pop_back();
  jumpTo = _jumpTo.back(); _jumpTo.pop_back();
  return true;
}

#endif // _MPG_VGraph_h_
