//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------


#include <stdio.h>
#include <float.h>
#include <math.h>
#include "MPG_BBoxQ.h"

const float MPG_BBoxQ::_epsilon = float(1e-20);

static inline int _quantize(float x, int n)
//{ return (int)(floor(ldexp(x,n))); }
{ return (int)(ldexp(x,n)-0.5); }

static inline float _unQuantize(int x, int n)
{ return ldexp((float)(x),-n); }

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ::MPG_BBoxQ
//
//---------------------------------------------------------------------------

MPG_BBoxQ::MPG_BBoxQ(int nD, int nBitsPerCoordinate)
{
  if(nD>0)
    {
      setDimension(nD);
      setBitsPerCoordinate(nBitsPerCoordinate);
      _sideLength.reserve(nD);
      _sideLength.insert(_sideLength.begin(),nD,0.0);
      _avgError = 0.0;
      _maxError = 0.0;
    }
}

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ::encode
//
//---------------------------------------------------------------------------

bool MPG_BBoxQ::encode(MPG_OutputBitStream& obstrm)
{
  float* minCorner   = getLowerCorner();
  float side         = getSideLength();

  // determine if bbox is default or not
  bool  non_default_bbox = false;
  int i;
  for(i=0;i<getDimension();i++)
    if(minCorner[i]!=0.0)
      { non_default_bbox = true; break; }
  if(side!=1.0) non_default_bbox = true;

  if(obstrm.putBit(non_default_bbox)==false) return false;

  // encode if not
  if(non_default_bbox==true)
    {
      for(i=0;i<getDimension();i++)
	if(obstrm.put32BitFloat(minCorner[i])==false) return false;
      if(obstrm.put32BitFloat(side)==false) return false;
    }

  // encode bits per coordinate
  int bitsPerCoord = getBitsPerCoordinate();
  if(obstrm.putMBitInt(5,bitsPerCoord)==false) return false;

  return true;
}

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ::decode
//
//---------------------------------------------------------------------------

bool MPG_BBoxQ::decode(MPG_InputBitStream& ibstrm)
{
  int dimValue = getDimension();

  bool non_default_bbox = false;
  if(ibstrm.getBit(non_default_bbox)==false) return false;
  if(non_default_bbox)
    {
      float minCorner_i = 0.0;
      int i;
      for(i=0;i<dimValue;i++)
	{
	  if(ibstrm.get32BitFloat(minCorner_i)==false) return false;
	  setLowerCorner(i,minCorner_i);
	}
      float side;
      if(ibstrm.get32BitFloat(side)==false) return false;
      setSideLength(side);
    }
  else
    {
      int i;
      for(i=0;i<dimValue;i++)
	setLowerCorner(i,0.0);
      setSideLength(1.0);
    }

  // decode bits per coordinate
  int bitsPerCoord = 0;
  if(ibstrm.getMBitInt(5,bitsPerCoord)==false) return false;
  if(bitsPerCoord<=0) return false;
  setBitsPerCoordinate(bitsPerCoord);

  return true;
}

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ::_update
//
//---------------------------------------------------------------------------

void MPG_BBoxQ::_update(int nValues, float* value)
{
  int   nD      = getDimension();
  while(--nValues>=0)
    {
      for(int i=0;i<nD;i++)
	{
	  float dvi = value[i] - _lowerCorner[i];
	  if(dvi<0.0)
	    {
	      _sideLength[i]  -= dvi;
	      _lowerCorner[i] += dvi;
	    }
	  else if(dvi>_sideLength[i])
	    _sideLength[i] = dvi;
	}
      value += nD;
    }
}

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ::_update
//
//---------------------------------------------------------------------------

void MPG_BBoxQ::_update(vector<float>& v)
{
  int   nValues = v.size()/getDimension();
  float *value  = v.begin();
  _update(nValues,value);
}

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ::quantize
//
//---------------------------------------------------------------------------

bool MPG_BBoxQ::quantize(float* value, int* iqValue, int nValues)
{
  int nD = getDimension();

  float maxsidelen=getSideLength();
  if(maxsidelen<=0.0) return false;

  int bpc = getBitsPerCoordinate();
  if(bpc<=0) return false;

  float delta = ldexp(maxsidelen,-bpc-1);
  maxsidelen += 2*delta;

  float* dqv = new float[nD];

  int nErr  = nValues;
  _avgError = 0.0;
  _maxError = 0.0;

  // int maxVal = 1<<bpc;
  while(--nValues>=0)
    {
    float nvi = 0.0;

      int i;
      for(i=0;i<nD;i++)
	{
	  // normalize to [0,1]
	  nvi = (value[i]-_lowerCorner[i]+delta)/maxsidelen;
	  if(nvi<0.0) nvi = 0.0;
          else if(nvi>1.0) nvi = 1.0;
	  iqValue[i] = _quantize(nvi,_bitsPerCoordinate);
	  // compute unQuantize value to evaluate error
	  nvi     = _unQuantize(iqValue[i],_bitsPerCoordinate);
	  dqv[i]  = nvi*maxsidelen+_lowerCorner[i]-delta;
	  dqv[i] -= value[i];
	}
      value   += nD;
      iqValue += nD;

  	  if(nvi<0.0 || nvi>1.0) continue;

      float nqv = 0.0;
      for(i=0;i<nD;i++)
	nqv += dqv[i]*dqv[i];
      if(nqv>0.0) nqv = sqrt(nqv);
      if(nqv>_maxError) _maxError = nqv;
      _avgError += nqv;

    }
  _avgError /= nErr;
  _maxError /= maxsidelen;
  _avgError /= maxsidelen;

  delete [] dqv;
  return true;
}

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ::quantize
//
//---------------------------------------------------------------------------

bool MPG_BBoxQ::quantize(vector<float>& v,
                         vector<int>&   iqv)
{
  int   nD       = getDimension();
  int   nValues  = v.size()/nD;
  iqv.erase(iqv.begin(),iqv.end());
  iqv.insert(iqv.end(),(size_t)(nD*nValues),0);
  float *value   = v.begin();
  int   *iqValue = iqv.begin();
  return quantize(value,iqValue,nValues);
}

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ::quantize
//
//---------------------------------------------------------------------------

bool MPG_BBoxQ::quantize(vector<float>& v,
                         vector<int>&   vIndx,
                         vector<int>&   iqv)
{
  int i,j,k;
  int nD  = getDimension();
  int nV  = vIndx.size();
  vector<float> v_flat;
  for(i=0;i<vIndx.size();i++)
    if((j=vIndx[i])>=0)
      for(k=0;k<nD;k++)
        v_flat.push_back(v[nD*j+k]);
  return quantize(v_flat,iqv);
}

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ::unQuantize
//
//---------------------------------------------------------------------------

bool MPG_BBoxQ::unQuantize(int* iqValue, float* uqValue, int nValues)
{
  int nD = getDimension();

  float maxsidelen=getSideLength();
  if(maxsidelen<=0.0) return false;

  int bpc = getBitsPerCoordinate();
  if(bpc<=0) return false;
  if(bpc>30) return false;

  float delta = ldexp(maxsidelen,-bpc-1);
  maxsidelen += 2*delta;

  int*   iqv = iqValue;
  float* uqv = uqValue;

  int maxVal = (1<<bpc)-1;
  while(--nValues>=0)
    {
      for(int i=0;i<nD;i++)
	{
          int iqvi = iqValue[i];
          if(iqvi<0) iqvi = 0;
          else if(iqvi>maxVal) iqvi=maxVal;
		  float nvi = _unQuantize(iqv[i],_bitsPerCoordinate);
		  uqv[i]  = nvi*maxsidelen+_lowerCorner[i]-delta;
	}
      uqv += nD;
      iqv += nD;
    }
  return true;
}

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ::unQuantize
//
//---------------------------------------------------------------------------

bool MPG_BBoxQ::unQuantize(vector<int>& iqv, vector<float>& qv)
{
  int nD      = getDimension();
  int nValues = iqv.size()/nD;
  qv.erase(qv.begin(),qv.end());
  qv.insert(qv.end(),iqv.size(),(float)0.0);
  float* uqValue   = qv.begin();
  int*   iqValue = iqv.begin();
  return unQuantize(iqValue,uqValue,nValues);
}
