//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "MPG_Perm.h"
#include "MPG_Ifs.h"
#include "MPG_ClustSurf.h"
#include "MPG_Simplify.h"

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::pfsGenerate
//
//---------------------------------------------------------------------------

bool MPG_ClustSurf::pfsGenerate(int move_vertices)
{
  bool success = false;

  // model to be simplified is the highest resolution level of detail
  // delete other levels of detail if necessary
  int nL = getNumberOfLevels();
  while(nL>2)
    { deleteLastLevel(); nL--; }

  if(nL==1)
    {
      MPG_Mesh mesh(_coordIndex);

      // so far only for triangular meshes
      if(mesh.isTriangulated()==false)
          return false;
      
      int nT0  = mesh.getNumberOfFaces();
      if (nT0<=0)
          return false;

      if(getDimension() !=3)
          return false;

      int nV0  = getNumberOfCoord();
      if (nV0<=0)
          return false;

      /* data for storing the clustered mesh */
      
      int new_nV   = 0;
      int new_nT   = 0;
      int *t       = NULL;
      int *new_t   = NULL;
      int nV       = nV0;
      int nT       = nT0;

      int simplify_boundary       = 1;

      int *vertex_parents         = NULL;
      
      float *v                    = NULL;
      float *error_volume         = NULL;
      float *colors               = NULL;
      float *normals              = NULL;
      float *texCoords            = NULL;
      float initial_tolerance     = 0.;
      float tolerance             = 0.;
      
      int   n_colors    = 0;
      int   n_normals   = 0;
      int   n_texCoords = 0;

      MPG_Binding colorBinding    = PB_NONE;
      MPG_Binding normalBinding   = PB_NONE;
      MPG_Binding texCoordBinding = PB_NONE;

      /* copy original triangles and vertices */

      t = (int *) malloc (nT0 * 3 * sizeof (int));
      if (!t) goto error;
      {
	int i,j;
	for(i=0;i<nT0;i++)
	  {
	    int* fi = mesh.getFace(i);
	    for(j=0;j<3;j++)
	      t[3*i+j] = fi[j];
	  }
      }
      
      v = (float *) malloc (nV0 * 3 * sizeof (float));
      if (!v) goto error;
      memcpy(v,getCoord(0), nV0 * 3 * sizeof (float));

      error_volume = (float *) malloc (nV0 * 3 * sizeof (float));
      if (!error_volume) goto error;

      // make sure to set initial errors to zero
      bzero(error_volume, nV0 * sizeof (float));
      
      // copy original colors if available

      colorBinding = getColorBinding();

      n_colors = 
	(colorBinding == PB_NONE) ? 0 :
	(colorBinding == PB_PER_VERTEX) ? nV*3 :
	(colorBinding == PB_PER_FACE)   ? nT*3 :
	nT*9 ;

      if (n_colors)
	{
	  colors = (float *) malloc (n_colors * sizeof(float));
	  if (!colors) goto error;
	  memcpy(colors, getColor(0), n_colors * sizeof (float));
	}
     

      // copy original normals if available
      
      normalBinding = getNormalBinding();

      n_normals = 
	(normalBinding == PB_NONE) ? 0 :
	(normalBinding == PB_PER_VERTEX) ? nV*3 :
	(normalBinding == PB_PER_FACE)   ? nT*3 :
	nT*9 ;

      if (n_normals)
	{
	  normals = (float *) malloc (n_normals * sizeof(float));
	  if (!normals) goto error;
	  memcpy(normals, getNormal(0), n_normals * sizeof (float));
	}

      // copy original texCoords if available

      texCoordBinding = getTexCoordBinding();

      n_texCoords = 
	(texCoordBinding == PB_NONE) ? 0 :
	(texCoordBinding == PB_PER_VERTEX) ? nV*2 :
	(texCoordBinding == PB_PER_FACE)   ? nT*2 :
	nT*6 ;

      if (n_texCoords)
	{
	  texCoords = (float *) malloc (n_texCoords * sizeof(float));
	  if (!texCoords) goto error;
	  memcpy(texCoords, getTexCoord(0), n_texCoords * sizeof (float));
	}
      
      vertex_parents = (int *) malloc (nV0 * sizeof (int));
      if (!vertex_parents) goto error;

      // increase the tolerance gradually from 1/20 of the bounding box diagonal to 1/2
	  
      initial_tolerance = _smpFloatDataBoundingBoxDiagonal(v,nV,3)/ 20.;
      tolerance         = initial_tolerance;
      
      /* loop on the simplification procedure,
	 until 10 levels were reached, or until the number of vertices
	 was reduced by 100 */
      
      while (nL < 10 && nV * 100 > nV0)
	{
	
	  // the v, t, colors, normals, texCoords arrays are copies of the originals
	  // they are overwritten by the following function call.
	  // and the new values are placed at the beginning of the arrays.

	  _smpTopoSimplifyTriangleMesh(nV, v, &new_nV, 
				       nT, t, &new_nT,
				       colors,       normals,       texCoords, 
				       colorBinding, normalBinding, texCoordBinding,
				       tolerance,
				       vertex_parents,
				       error_volume,
				       move_vertices, 
				       simplify_boundary);

	  tolerance += initial_tolerance;

	  if (!vertex_parents) 
	      goto error;

	  // determine how many new colors/normals/texCoords there are depending upon the binding
	    
	  if (n_colors > 0) 
	    n_colors = 
 	      (colorBinding == PB_NONE) ? 0 :           (colorBinding == PB_PER_VERTEX) ? new_nV*3 :
	      (colorBinding == PB_PER_FACE)? new_nT*3 : new_nT*9 ;

	  if (n_normals > 0) 
	    n_normals = 
 	      (normalBinding == PB_NONE) ? 0 :           (normalBinding == PB_PER_VERTEX) ? new_nV*3 :
	      (normalBinding == PB_PER_FACE)? new_nT*3 : new_nT*9 ;

	  if (n_texCoords > 0) 
	    n_texCoords = 
 	      (texCoordBinding == PB_NONE) ? 0 :           (texCoordBinding == PB_PER_VERTEX) ? new_nV*2 :
	      (texCoordBinding == PB_PER_FACE)? new_nT*2 : new_nT*6 ;

	  // append level to container data structure
	  {
	    register int i,j;

	    for(i=0;i<nV;i++) 
	      _coordMap.push_back(vertex_parents[i]); /* write the map */

	    float *v1 = v;
	    for(i=0;i<new_nV;i++)
	      for (j=0;j<3;j++)
		_coord.push_back(*v1++);

	    for(i=0;i<n_colors; i++)
	      _color.push_back(colors[i]);
	    
	    for(i=0;i<n_normals; i++)
	      _normal.push_back(normals[i]);
	    
	    for(i=0;i<n_texCoords; i++)
	      _texCoord.push_back(texCoords[i]);

	    closeLevel();
	  }

	  nL ++;
	  
	  /* update the vertices and triangles */
	  
	  nV     = new_nV; new_nV = 0;
	  
	  nT     = new_nT; new_nT = 0;

	}
      /* end while(nL) */
     
      success = true;

    error:
      
      /* free the various arrays if they have been allocated */
      
      if (v )
	{ free ((char *) v );                  v                    = NULL; }
      if (t )
	{ free ((char *) t );                  t                    = NULL; }
      if (colors )
	{ free ((char *) colors );             colors               = NULL; }
      if (normals )
	{ free ((char *) normals );            normals              = NULL; }
      if (texCoords )
	{ free ((char *) texCoords );          texCoords            = NULL; }
      if (vertex_parents )
	{ free ((char *) vertex_parents );     vertex_parents       = NULL; }
      if (error_volume) 
	{ free ((char *) error_volume);        error_volume         = NULL; }
    }

  return success;
}
