//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------

#pragma warning(disable:4786)

#include <stdio.h>
#include <stdlib.h>
#include "MPG_IfsQ_C.h"
#include "MPG_ClustSurf.h"
#include "MPG_MeshTs.h"
#include "MPG_MeshFs.h"
#include "MPG_IfsEdit.h"
#include "MPG_QFac.h"
#include "MPG_Global.h"
#include "MPG_VGraph.h"
#include "MPG_SPolygon.h"
#include "MPG_Stitcher.h"
#include "MPG_Perm.h"
//#include "MPG_Vector.h"
#include "MPG_ER.h"   

#ifndef M_PI
#define M_PI    3.14159265
#endif
#define boolStr(b) ((b)==true)?"true":"false"

#ifndef MPGCLIB
extern char* baseFname;
extern char* inExt;
extern char* inDir;
extern char* outDir;
extern char* inFname;
extern char* outFname;
#endif

//---------------------------------------------------------------------------
//
//   MPG_IfsQ_C::compress
//
//---------------------------------------------------------------------------
bool MPG_IfsQ_C::compress(MPG_OutputStream& ostrm,
						  MPG_OutputStream& t_ostrm,  
                          int  bitsPerCoord,
                          int  coordPrediction,
                          int  bitsPerNormal,
                          int  normalPrediction,
                          int  bitsPerColor,
                          int  colorPrediction,
                          int  bitsPerTexCoord,
                          int  texCoordPrediction,
						  int  erMode,    
						  int  packetSize,
						  int  boundaryPrediction,
                          bool hierarchicalMode,
                          bool saveClsFile,
                          bool saveWrlLevels)
{
  // determine if mesh has isolated vertices
  bool hasIsolatedVertices = false;
  {
    int nV = getNumberOfCoord();
    vector<bool> used((size_t)nV,false);
    vector<int>& coordIndex = getCoordIndex();
    int i;
    for(i=0;i<coordIndex.size();i++)
      if(coordIndex[i]>=0)
        {
          if(coordIndex[i]>=nV) return false;
          used[coordIndex[i]] = true;
        }
    for(i=0;i<nV;i++)
      if(used[i]==false)
        { hasIsolatedVertices = true; break; }
  }

  // remove isolated vertices
  if(hasIsolatedVertices)
    {
      MPG_IfsMesh mesh(*this);
      MPG_IfsEdit ifsEdit(mesh);
      if(ifsEdit.removeIsolatedVertices()==false) return false;
    }

  // deal with non-manifolds without isolated vertices
  vector<int>  v_cluster_enc;
  vector<int>  v_cluster_dec;
  {
	  MPG_IfsMesh mesh(*this);
	  if(hierarchicalMode==true &&
		  mesh.isTriangulated()==false)
		  return false;
	  if(mesh.isOrientedManifold()==false)
      {
		  if(hierarchicalMode==true)
			  return false;
		  MPG_IfsEdit ifsEdit(mesh);
		  if(ifsEdit.convertToOrientedManifold(v_cluster_enc)==false)
			  return false;
	  }
	  vector<bool> fmk;
	  if(mesh.markFacesWithDuplicateVertices(fmk))
		  return false;
  }

  // generate hierarchy of levels of detail
  int nLevelsOfDetail=1;
  MPG_ClustSurf csrf;

#ifndef MPGCLIB
  if(hierarchicalMode==true)
  {
      // decompose into a base mesh and one or more forest splits
      csrf = *this; // uses MPG_ClustSurf::operator=(MPG_IfsQ& ifs)
      // simplify
      if(csrf.pfsGenerate(true)==false)
          return false;
      nLevelsOfDetail = csrf.getNumberOfLevels();

      if(saveClsFile==true)
	  {
          int nL = nLevelsOfDetail; 
          int n = strlen(outDir)+strlen(baseFname)+12;
          outFname = new char[n];
          if(!strcmp(outDir,"."))
			  sprintf(outFname,"%s-C%02d.cls",baseFname,nL);
          else
			  sprintf(outFname,"%s/%s-C%02d.cls",outDir,baseFname,nL);
          bool success = csrf.write(outFname);
          delete [] outFname;
          outFname = NULL;
	  }
      if(saveWrlLevels==true)
	  {
          int n = strlen(outDir)+strlen(baseFname)+12;
          outFname = new char[n];
          MPG_Ifs ifs;
          int l,ll;
          for(l=0;l<nLevelsOfDetail;l++)
			  if(csrf.getIfs(ifs,l)==true)
              {
				  ll = nLevelsOfDetail-l-1;
				  if(!strcmp(outDir,"."))
					  sprintf(outFname,"%s-C%02d.wrl",baseFname,ll);
				  else
					  sprintf(outFname,"%s/%s-C%02d.wrl",outDir,baseFname,ll);
				  bool success = ifs.write(outFname);
              }
			  delete [] outFname;
			  outFname = NULL;
	  }
      // set *this to simplest model;
      if(nLevelsOfDetail>1)
	  {
          // note that MPG_ClustSurf::getIfs(MPG_Ifs&,int)
          // does NOT reset the quantizer pointers of *this
          // which is exactly what we want
          if(csrf.getIfs(*this,nLevelsOfDetail-1)==false)
              return false;
	  }
  }
#endif

  // compress base mesh
  MPG_MeshTs mesh(getCoordIndex());
  int nCC = mesh.getNumberOfConnectedComponents();
  // create property headers and quantizers if not yet created
  MPG_Prediction prediction = PP_NONE;

  // valid combinations of property/binding/prediction
  //
  //              | PER_VERTEX | PER_FACE   | PER_CORNER    |
  //   -----------+------------+------------+---------------+
  //   COORD      |     X      |            |               |
  //   NORMAL     |     X      |     X      |     X         |
  //   COLOR      |     X      |     X      |     X         |
  //   TEXCOORD   |     X      |            |     X         |
  //
  //              | NONE       | TREE       | PARALLELOGRAM |
  //   -----------+------------+------------+---------------+
  //   PER_VERTEX |     NS     |     NS     |     X         |
  //   PER_FACE   |     NS     |     X      |               |
  //   PER_CORNER |     NS     |     X      |               |

  MPG_Binding coordBinding = getCoordBinding();
  prediction = (MPG_Prediction)coordPrediction;
  MPG_Property pCoord(PT_COORD,coordBinding,prediction);
  MPG_BBoxQ*   coordQuant = getCoordQuantizer(bitsPerCoord);
  bitsPerCoord = coordQuant->getBitsPerCoordinate();
  pCoord.setBBoxQ(coordQuant);
  pCoord.setDefaultLambda();

  MPG_Binding normalBinding   = getNormalBinding();
  prediction = (MPG_Prediction)normalPrediction;
  MPG_Property pNormal(PT_NORMAL,normalBinding,prediction);
  MPG_NormalQ* normalQuant   = getNormalQuantizer(bitsPerNormal);
  bitsPerNormal = normalQuant->getBitsPerNormal();
  pNormal.setNormalQ(normalQuant);
  pNormal.setDefaultLambda();

  MPG_Binding colorBinding    = getColorBinding();
  prediction = (MPG_Prediction)colorPrediction;
  MPG_Property pColor(PT_COLOR,colorBinding,prediction);
  MPG_BBoxQ*   colorQuant    = getColorQuantizer(bitsPerColor);
  bitsPerColor = colorQuant->getBitsPerCoordinate();
  pColor.setBBoxQ(colorQuant);
  pColor.setDefaultLambda();

  MPG_Binding texCoordBinding = getTexCoordBinding();
  prediction = (MPG_Prediction)texCoordPrediction;
  MPG_Property pTexCoord(PT_TEXCOORD,texCoordBinding,prediction);
  MPG_BBoxQ*   texCoordQuant = getTexCoordQuantizer(bitsPerTexCoord);
  bitsPerTexCoord = texCoordQuant->getBitsPerCoordinate();
  pTexCoord.setBBoxQ(texCoordQuant);
  pTexCoord.setDefaultLambda();

  // decompose each connected component into a vertex graph
  // and a simple polygon, and collect quantized prediction
  // errors in the triData array in decoder order,
  // i.e., in the order of traversal of the simple polygon.
  // the collected quantized prediction errors are later
  // entropy encoded during a second simple polygon traversal
  // by sp.encodeTriangleData()

  vector<MPG_SPolygon*> vpsp;
  vector<MPG_VGraph*>   vpvg;
  vector<int>           v_visited;
  MPG_Stitches          v_stitches_dec;

  // these arrays are here because they are used in hiearchical mode
  vector<int> v_traversal;
  vector<int> v_traversal_inv;
  vector<int> f_traversal;
  vector<int> f_index;
  vector<int> c_index;
  MPG_BranchForest* root_btf = new MPG_BranchForest;
  vector <int> qCoord1;
  vector <int> qNormal1;
  vector <int> qColor1;
  vector <int> qTexCoord1;
  {
	  if(coordBinding!=PB_NONE)
		  coordQuant->quantize(getCoord(),qCoord1);
	  if(normalBinding!=PB_NONE)
      {
		  vector<int>& nIndex = getNormalIndex();
		  if(nIndex.size()>0)
			  normalQuant->quantize(getNormal(),nIndex,qNormal1);
		  else
			  normalQuant->quantize(getNormal(),qNormal1);
      }
	  if(colorBinding!=PB_NONE)
      {
		  vector<int>& cIndex = getColorIndex();
		  if(cIndex.size()>0)
			  colorQuant->quantize(getColor(),cIndex,qColor1);
		  else
			  colorQuant->quantize(getColor(),qColor1);
      }
	  if(texCoordBinding!=PB_NONE)
      {
		  vector<int>& tcIndex = getTexCoordIndex();
		  if(tcIndex.size()>0)
			  texCoordQuant->quantize(getTexCoord(),tcIndex,qTexCoord1);
		  else
			  texCoordQuant->quantize(getTexCoord(),qTexCoord1);
      }
	  
	  vector<int> v_father_enc; // used to build v_father_dec
	  if(mesh.representConnectivity(vpsp, vpvg, pCoord,	pNormal, 
						pColor, pTexCoord, qCoord1, qNormal1,
						qColor1, qTexCoord1, v_visited, v_cluster_enc,
						v_father_enc, v_traversal, f_traversal,
						f_index, c_index, root_btf)==false)
      		  return false;
	  if(MPG_Perm::makeInvPerm(v_traversal,v_traversal_inv)==false)
		  return false;
	  // construct v_cluster_dec 
	  int i;
	  for(i=0;i<v_cluster_enc.size();i++)
		  v_cluster_dec.push_back(v_cluster_enc[v_traversal[i]]);
	  
	  // AG 2/16/99: 
	  // if v_cluster_dec[1]   = 256 and 
	  //    v_cluster_dec[256] = 256 this is bad if 1 and 256 belong to
	  // different connected components: the coordinates of 1 will be encoded
	  // because 256 is not visited yet whereas they shouldn't (duplicate)
	  // and this will crash the decoder which expects no duplication
	  // so among all elements that cluster to the same location,
	  // the minimum vertex index must be used in v_cluster_dec
	  {
		  int n = v_cluster_dec.size();
		  vector <int> min_cluster_index((size_t) n);
		  for (i=n-1;i>=0;i--) 
			  min_cluster_index[v_cluster_dec[i]] = i;      
		  for (i=0;i<n;i++) 
			  v_cluster_dec[i] = min_cluster_index[v_cluster_dec[i]];

		  //MJH 091900 : output changed vertex indices
#ifdef INDEX_OUTPUT
		  if (outputIndex) {
			  FILE* fp = fopen("index.txt","w");
			  if(fp==(FILE*)0) return false;

			  fprintf(fp,"%d\n",n);
			  for (i=0;i<n;i++) 
				  fprintf(fp,"%d\n",v_traversal[i]);
			  fclose(fp);
		  }

#endif
		  //MJH 091900 end
	  }
	  // construct v_father_dec
	  vector<int> v_father_dec;
	  for(i=0;i<v_father_enc.size();i++)
		  v_father_dec.push_back(v_traversal_inv[v_father_enc[v_traversal[i]]]);
	  vector <int> nVertCC;

	  // we have to make sure that the components are in the decoder order
	  for(i=0;i<nCC;i++) 
      {
		  MPG_VGraph&   vg = *(vpvg[i]);
		  nVertCC.push_back(vg.getNumberOfVertices());
      }
	  for(i=1;i<nCC;i++) nVertCC[i] += nVertCC[i-1];

	  // represent the v_cluster_enc as a sequence of stitch commands
	  MPG_Stitcher mc;
	  if(mc.represent(nVertCC, v_father_dec, v_cluster_dec, v_stitches_dec)==false)
		  return false;
  }

  // VisualObjectSequence() {
  //	visual_object_sequence_start_code
  //	profile_and_level_indication
  //    VisualObject() {
  //	    visual_object_start_code
  //	    is_visual_object_identifier
  //	    if (is_visual_object_identifier) {
  //	       visual_object_verid
  //	        visual_object_priority
  //	    }
  //	    visual_object_type
  //	    if (visual_object_type == "3D mesh ID") 
  //           3D_Mesh_Object () {
  //               3D_MO_start_code
  //	           3D_Mesh_Object_Header()
  //	           do {
  //		           3D_Mesh_Object_Layer()
  //	           } while (nextbits_bytealigned() == 3D_MOL_start_code)
  //           }
  //    }
  //    visual_object_sequence_end_code
  // }

  // open an output bit stream
  MPG_OutputBitStream obstrm(ostrm);

  // write 3D_MO_start_code
  if(obstrm.putMOStartCodeBitInt(MPG_Global::MO_start_code)==false)
    return false;

  // begin encode 3D_Mesh_Object_Header()  
  {      
	  // 3D_Mesh_Object_Header() {
	  //   ccw
	  //   convex
	  //   solid
	  //   creaseAngle
	  //   coord_header()
	  //   normal_header()
	  //   color_header()
	  //   texCoord_header()
	  //   ce_SNHC
	  //   if(ce_SNHC==1)
	  //      ce_SNHC_header()
	  // }
	  if(obstrm.putMOBit(_ccw               )==false) return false;
	  if(obstrm.putMOBit(_convex            )==false) return false;
	  if(obstrm.putMOBit(_solid             )==false) return false;
	  int cAng = (int)((_creaseAngle*32.0)/M_PI);
	  if(cAng<0 || cAng>=64) return false;
	  if(obstrm.putMOBitInt(6,cAng)==false) return false;
	  
	  // coord_header() {
	  //   coord_binding       <--- must be == 1 (PER_VERTEX)
	  //   coord_bbox
	  //   if(coord_bbox) {
	  //     coord_xmin
	  //     coord_ymin
	  //     coord_zmin
	  //     coord_size
	  //   }
	  //   coord_quant
	  //   coord_pred_type
	  //   if(coord_pred_type==2) {
	  //     coord_nlambda
	  //     for(i=1;i<coord_nlambda;i++)
	  //       coord_lambda[i]
	  //   }
	  // }
	  if(obstrm.putMBitInt(2,coordBinding)==false) return false;
	  if(coordBinding!=PB_NONE) {
		  if(coordBinding!=PB_PER_VERTEX) return false;
		  if(pCoord.encode(obstrm)==false) return false;
	  }
 
	  // normal_header() {
	  //    normal_binding
	  //    if(normal_binding!=0) {
	  //       normal_bbox       <--- must be == 0
	  //       normal_quant
	  //       normal_pred_type
	  //       if(normal_pred_type==2) {
	  //          normal_nlambda
	  //          for(i=1;i<normal_nlambda;i++)
	  //              normal_lambda[i]
	  //       }
	  //    }
	  // }
	  if(obstrm.putMBitInt(2,normalBinding)==false) return false;
	  if(normalBinding!=PB_NONE)
	  {
		  if(pNormal.encode(obstrm)==false) return false;
	  }

	  // color_header() {
	  //    color_binding
	  //    if(color_binding!=0) {
	  //        color_bbox
	  //        if(color_bbox) {
	  //           color_rmin
	  //           color_gmin
	  //           color_bmin
	  //           color_size
	  //        }
	  //        color_quant
	  //        color_pred_type
	  //        if(color_pred_type==2) {
	  //            color_nlambda
	  //            for(i=1;i<color_nlambda;i++)
	  //                color_lambda[i]
	  //        }
	  //    }
	  // }
	  if(obstrm.putMBitInt(2,colorBinding)==false) return false;
	  if(colorBinding!=PB_NONE)
	  {
		  if(pColor.encode(obstrm)==false) return false;
	  }

	  // texCoord_header() {
	  //    texCoord_binding     <--- must be != 2 (PER_FACE)
	  //    if(texCoord_binding!=0) {
	  //       texCoord_bbox
	  //       if(texCoord_bbox) {
	  //          texCoord_umin
	  //          texCoord_vmin
	  //          texCoord_size
	  //        }
	  //        texCoord_quant
	  //        texCoord_pred_type
	  //        if(texCoord_pred_type==2) {
	  //           texCoord_nlambda
	  //           for(i=1;i<texCoord_nlambda;i++)
	  //               texCoord_lambda[i]
	  //        }
	  //    }
	  // }
	  if(obstrm.putMBitInt(2,texCoordBinding)==false) return false;
	  if(texCoordBinding!=PB_NONE)
	  {
		  if(texCoordBinding==PB_PER_FACE) return false;
		  if(pTexCoord.encode(obstrm)==false) return false;
	  }

	  // DO NOT encode ce_SNHC
	  if(obstrm.putBit(false)==false) return false;
	  obstrm.flush();  // flush output buffer
  } // end encode 3D_Mesh_Object_Header()

  // begin encoding 3D_Mesh_Object_Layer ()
  int mol_id = 0;
  {
	  // 3D_Mesh_Object_Layer () {
	  //	3D_MOL_start_code
	  //	mol_id
	  //	if (ce_SNHC == '1') {
	  //		ce_SNHC _n_vertices
	  //		ce_SNHC _n_triangles
	  //		ce_SNHC _n_edges
	  //	}
	  //	if (mol_id == 0)
	  //		3DMeshObject_Base_Layer()
	  //	else
	  //		3DMeshObject_Refinement_Layer()
	  // }
	  // write 3D_MOL_start_code
	  if(ostrm.putStartCode(MPG_Global::MOL_start_code)==false)
		  return false;
	  {
		  // open an output bit stream
		  MPG_OutputBitStream obstrm(ostrm);
		  // write mol_id
		  if(obstrm.putMBitInt(8,mol_id)==false) return false;
		  mol_id++;
		  // flush output buffer
		  obstrm.flush();
	  }

	  // DO NOT encode ce_SNHC
	  // a single 3D_Mesh_Object_Base_Layer() for now

	  // begin encoding 3D_Mesh_Object_Base_Layer() 
      {
		  // 3DmeshObject_Base_Layer() {
		  //	do {
		  //		3D_MOBL_start_code
		  //		mobl_id
		  //		while (!bytealigned())
		  //			one_bit
		  //		qf_start()
		  //		if (3D_MOBL_start_code == "partition_type_0") {
		  //			do {
		  //		        connected_component()
		  //		        qf_decode(last_component, last_component_context)
		  //	        } while (last_component == '0')
		  //		}
		  //		else if (3D_MOBL_start_code == "partition_type_1") {
		  //			vg_number=0
		  //			do {
		  //				vertex_graph()
		  //				vg_number++
		  //				qf_decode(has_stitches, has_stitches_context)
		  //				if (has_stitches == '1')
		  //					stitches()
		  //				qf_decode(codap_last_vg, codap_last_vg_context)
		  //			} while (codap_last_vg == '0')
		  //		}
		  //		else if (3D_MOBL_start_code == "partition_type_2") {
		  //			if(vg_number > 1)
		  //				qf_decode(codap_vg_id)
		  //			qf_decode(codap_left_bloop_idx)
		  //			qf_decode(codap_right_bloop_idx)
		  //			qf_decode(codap_bdry_pred)
		  //			triangle_tree()
		  //			triangle_data()
		  //		}
		  //	} while (nextbits_bytealigned() == 3D_MOBL_start_code)
		  // }
		  MPG_OutputBitStream t_obstrm(t_ostrm);  
		  MPG_QFac t_codec(t_ostrm, erMode);
		  MPG_OutputBitStream obstrm(ostrm);  
		  MPG_QFac codec(ostrm, erMode);  	
		  MPG_OutputStream vgostrm;  
		  MPG_OutputBitStream vgobstrm(vgostrm);  
		  MPG_QFac vgcodec(vgostrm, erMode);
		  MPG_OutputStream spostrm;  
		  MPG_OutputBitStream spobstrm(spostrm);  
		  MPG_QFac spcodec(spostrm, erMode);

		  int header_bits=0, totalbits = 0, restbits = packetSize * 8;
		  int cpass=0, rootbits=0, tribits=0;
		  cpass=(erMode==1) ? 0: 1;
		  if(coordBinding==PB_PER_VERTEX) {
			  rootbits = 100;
			  tribits = bitsPerCoord;
		  }
		  if(normalBinding!=PB_NONE) {
			  rootbits += 50;
			  tribits += bitsPerNormal;
		  }
		  if(colorBinding!=PB_NONE) {
			  rootbits += 50;
			  tribits += bitsPerColor;
		  }
		  if(texCoordBinding==PB_PER_CORNER || texCoordBinding==PB_PER_VERTEX) {
			  rootbits += 50;
			  tribits += bitsPerTexCoord;
		  }
		  header_bits = ostrm.size()*8; 
		  MPG_ER2 ccbit;               
		  		  
		  do {  
			  // stitch encoding stack
			  vector<int>  stk;
			  int icc = 0;
			  vector<int> nV0;
			  vector<int> nV1;
			  nV0.push_back(0);

			  typedef vector<int> vec_array;
		      vector<vec_array> v_loop_cc;
			  v_loop_cc.clear();

			  int GiC=0, GiF=-1;	//MJH 001011 : GiF changed from 0 to -1 for per face
			  int prevCR=0, prevCL=0;	//MJH 001013 : added in case of partitioning at non-polygon edge
			  totalbits = header_bits;  
			  restbits = (packetSize * 8)-totalbits;
			  MPG_PacketQ_C codap(erMode,packetSize,totalbits,restbits);
			  MPG_BranchForest* cur_btf=root_btf;	

			  // connected_component() {
			  //	  vertex_graph()
			  //	  qf_decode(has_stitches, has_stitches_context)
			  //	  if (has_stitches == '1')
			  //		  stitches()
			  //	  triangle_tree()
			  //	  triangle_data()
			  // } 
			  // decomposition is done at the beginning of the method,
			  // after creating the mesh
			  for(icc=0;icc<nCC;icc++)
			  {
				  if(!cpass) {  
					  //encoding connected_component() with partition_type_0
					  codap.erMode=0;
					  MPG_VGraph&   vg = *(vpvg[icc]);
					  nV1.push_back(nV0[icc] + vg.getNumberOfVertices());
					  MPG_SPolygon& sp = *(vpsp[icc]); 
					  
					  vec_array tmp_v_loop_cc;
					  {
						  MPG_Loops loop;
						  vg.toLoop(loop);
						  loop.getLoop(0, tmp_v_loop_cc);
					  }

					  // encode vertex_graph()
					  codap.vg_init(ccbit.partition_type);
					  codap.vg_sc_enable=1;
					  if(icc>0) codap.vg_sc_enable=0;
					  int vg_bits=0, vg_mobl_bits=0;
					  if(codap.vg_StartCode(vgostrm, t_ostrm, vgobstrm, t_obstrm,
						  vgcodec, t_codec, vg_mobl_bits)==false) return false;
					  MPG_Global::vg_restore();    
					  vgcodec.cbits=0;
					  if(vg.encode(vgcodec) == false) return false;
					  vg_bits+=vgcodec.cbits;
					  vgcodec.cbits=0;

					  // encoding Stitches()
					  v_stitches_dec.encode(nV0[icc],nV1[icc],stk,vgcodec);
					  vg_bits+=vgcodec.cbits; 
					  vgcodec.cbits=0;
					  ccbit.putVGbits(vg_bits);  
					  
					  //write VG data to output buffer and 
					  //initialize variables for GF & buffer 
					  codap.vg_packetization(vgostrm,t_ostrm,vgobstrm,t_obstrm,
						  vgcodec,t_codec, icc, ccbit);
					  
					  sp.DetermineDirection();   //define traversal order with td_orientation 
					  int next=0;
					  vector<int> Grstk((size_t)1, -1);  //stack run value for subtree partitioning
					  vector<int> Gistk((size_t)1, -1);  //stack loop index for subtree partitioning
					  codap.sp_init(tmp_v_loop_cc.size(), rootbits, tribits);
					  cur_btf = cur_btf->GetNextBTree();
					  MPG_BranchTree* cur_bt = cur_btf->GetRootBTree();	
					  cur_bt->SetCurDir(LEFT_BRANCH);
					  int sp_mobl_bits=0, Gdepth=0;
					  if(codap.sp_StartCode(spostrm, t_ostrm,
						  spobstrm, t_obstrm, spcodec, t_codec, sp_mobl_bits)==false) return false;
					  MPG_Global::sp_restore();
					  spcodec.cbits=0;
					  
					  // encode triangle_tree()
					  if(sp.encodeTriangleTree(spcodec, Grstk, Gdepth, codap) == false) return false;
					  ccbit.putTTbits(spcodec.cbits);
					  spcodec.cbits=0;
					  
					  // encode triangle_data()
					  MPG_BranchTree* tmp_bt = cur_bt;	
					  if(sp.encodeTriangleData(spcodec,
						  nV0[icc],  // first vertex index of cc
						  tmp_v_loop_cc,  // local  indices
						  v_visited,     // global decoder indices
						  v_cluster_dec, // global decoder indices
						  pCoord,
						  pNormal,
						  pColor,
						  pTexCoord,
						  qCoord1,
						  qNormal1,
						  qColor1,
						  qTexCoord1,
						  v_traversal, f_index, 
						  c_index,	cur_bt,	
						  Grstk, Gistk, Gdepth, GiC, GiF,
						  prevCR, prevCL,	//MJH 001013 : added in case of partitioning at non-polygon edge
						  restbits, codap  
						  ) == false) return false;
					  ccbit.putTDbits(spcodec.cbits);
					  spcodec.cbits=0;
					  
					  //write SP data to output buffer and 
					  //initialize variables for GF & buffer 
					  codap.sp_packetization(
					    vgostrm, spostrm, t_ostrm, vgobstrm, spobstrm, t_obstrm,
					    vgcodec, spcodec, t_codec, icc, nCC, next, Gdepth, totalbits, restbits, 
					    cur_bt, tmp_bt, ccbit);
					  nV0.push_back(nV1[icc]);  
					  if(icc==nCC-1) {
						  codap.erMode=1;
						  v_visited.erase(v_visited.begin(),v_visited.end());
                          v_visited.insert(v_visited.end(),(size_t)nV1[icc],AVAIL_PRED);  
					  }
				  } else {
					  //encoding connected_component() with partition_type_1 or 2
					  int tcc=icc;
					  if(codap.erMode) {
						  if(ccbit.next==icc) {
						     if(ccbit.determine_PT(icc, nCC, codap, ostrm, obstrm, codec)==false)
							    return false;
							 codap.vg_sc_enable=1;
						  } else {
							  ccbit.partition_type=0;
							  codap.vg_sc_enable=0;
						  }
					  } else {
						  ccbit.curr=0;
						  ccbit.next=nCC;
						  ccbit.partition_type=0;
						  codap.vg_sc_enable=1;
						  if(icc>0) codap.vg_sc_enable=0;
					  }
					  codap.vg_init(ccbit.partition_type);
					  
					  // encode vertex_graph() & Stitches()
					  do {
						  MPG_VGraph&   vg = *(vpvg[icc]);
						  nV1.push_back(nV0[icc]+vg.getNumberOfVertices());
   	                      
						  vec_array tmp_v_loop;
						  v_loop_cc.push_back(tmp_v_loop);
						  {
							  MPG_Loops loop;
							  vg.toLoop(loop);
							  loop.getLoop(0 ,v_loop_cc[icc]);
						  }
						  
						  int vg_bits=0, vg_mobl_bits=0;
						  if(codap.vg_StartCode(vgostrm, ostrm, vgobstrm, obstrm,
							  vgcodec, codec, vg_mobl_bits)==false) return false;
						  MPG_Global::vg_restore();        
						  vgcodec.init_bits(codap.totalbits, codap.restbits); 
						  
						  // encode vertex_graph()
						  if(vg.encode(vgcodec) == false) return false;
						  codap.totalbits=vgcodec.tbits;
						  codap.restbits=vgcodec.rbits;
						  vg_bits+=vgcodec.cbits;
						  
						  // encoding Stitches()
						  vgcodec.init_bits(codap.totalbits, codap.restbits); 
						  v_stitches_dec.encode(nV0[icc],nV1[icc],stk,vgcodec);
						  codap.totalbits=vgcodec.tbits;
						  codap.restbits=vgcodec.rbits;
						  vg_bits+=vgcodec.cbits;
						  
						  if(!codap.vg_pt0_enable) {
							  vgcodec.init_bits(codap.totalbits, codap.restbits); 
							  if(icc+1==ccbit.next) {
								  vgcodec.encode(1, MPG_Global::codap_last_vg_context);
								  vgcodec.finish_encode();
							  } else {
								  vgcodec.encode(0, MPG_Global::codap_last_vg_context);
							  }
							  codap.totalbits=vgcodec.tbits;
							  codap.restbits=vgcodec.rbits;
							  vg_bits+=vgcodec.cbits;
						  }
						  
						  //write VG data to output buffer and 
						  //initialize variables for GF & buffer
						  codap.vg_packetization(vgostrm,ostrm,vgobstrm,obstrm,
							  vgcodec, codec, icc, ccbit);
						  codap.vg_sc_enable=0;
						  delete vpvg[icc]; vpvg[icc] = NULL;
						  nV0.push_back(nV1[icc]);
						  icc++;
					  } while(!codap.vg_pt0_enable && (icc<ccbit.next));

					  if(codap.vg_pt0_enable) {
						  if((icc-1)==ccbit.curr)
							  codap.vg_sc_enable=1;
					  } else {
						  codap.vg_sc_enable=1;
					  }
					  icc=tcc;
					  codap.vg_id_len = (codap.nvg>1) ? slog2(codap.nvg): 0;  
					  codap.sp_vg_id=0;

					  // encode triangle_tree() & triangle_data()
					  do {
						  MPG_SPolygon& sp = *(vpsp[icc]); 
						  sp.DetermineDirection();   //define traversal order with td_orientation 
						  
						  int next=0, Gdepth=0;	
						  vector<int> Grstk((size_t)1, -1);  //stack run value for subtree partitioning
						  vector<int> Gistk((size_t)1, -1);  //stack loop index for subtree partitioning
						  codap.sp_init(v_loop_cc[icc].size(), rootbits, tribits);
						  totalbits = codap.totalbits;  
						  restbits = codap.restbits;
						  cur_btf = cur_btf->GetNextBTree();
						  MPG_BranchTree* cur_bt = cur_btf->GetRootBTree();	
						  cur_bt->SetCurDir(LEFT_BRANCH);
						  do {
							  int sp_mobl_bits=0;
							  if(codap.sp_StartCode(spostrm, ostrm,
								  spobstrm, obstrm, spcodec, codec, sp_mobl_bits)==false) return false;
							  
							  MPG_Global::sp_restore();
							  spcodec.init_bits(codap.totalbits, codap.restbits);  
							  
							  // encode triangle_tree()
							  if(sp.encodeTriangleTree(spcodec, Grstk, Gdepth, codap) == false) return false;
							  if(codap.sp_pt2_enable && codap.sp_sub_pt_enable) {
								  spcodec.encodePT2(codap.bdry_loopbits,codap.sp_branch_len,
									  MPG_Global::codap_branch_len_context);
							  }
							  codap.totalbits=spcodec.tbits;
							  codap.restbits=spcodec.rbits;
							  if(codap.sp_pass0 && codap.sp_pt2_enable) {
								  sp.tt_avgbits = (float)spcodec.cbits/(v_loop_cc[icc].size()-2);
							  }
							  
							  // encode triangle_data()
							  MPG_BranchTree* tmp_bt = cur_bt;	
							  spcodec.init_bits(codap.totalbits, codap.restbits);  
							  if(sp.encodeTriangleData(spcodec,
								  nV0[icc],   // first vertex index of cc
								  v_loop_cc[icc],     // local  indices
								  v_visited,     // global decoder indices
								  v_cluster_dec, // global decoder indices
								  pCoord,
								  pNormal,
								  pColor,
								  pTexCoord,
								  qCoord1,
								  qNormal1,
								  qColor1,
								  qTexCoord1,
								  v_traversal, f_index, 
								  c_index,	cur_bt,	
								  Grstk, Gistk, Gdepth, GiC, GiF,
		  						  prevCR, prevCL,	//MJH 001013 : added in case of partitioning at non-polygon edge
								  restbits, codap ) == false)
								  return false;

							  //write SP data to output buffer and 
							  //initialize variables for GF & buffer
							  codap.sp_packetization(
								  vgostrm, spostrm, ostrm, vgobstrm, spobstrm, obstrm,
								  vgcodec, spcodec, codec, icc, nCC, next, Gdepth, totalbits, restbits, 
								  cur_bt, tmp_bt, ccbit);
							  
						  } while(!next); 
						  
						  delete vpsp[icc]; vpsp[icc] = NULL;
						  icc++;
					  } while(!codap.vg_pt0_enable && (icc<ccbit.next));
					  
					  if(erMode>0 && icc==nCC) {
						  //to fill the last packet with dummy bits
//						  int rbytes=codap.stuffing(spostrm,spobstrm,spcodec);
						  spobstrm.init();
						  codap.writeStream(spostrm,ostrm);
						  codap.updateQFac(spobstrm,obstrm,spcodec,codec);
						  MPG_Global::sp_store();    
						  codap.totalbits=spcodec.tbits;//+rbytes*8;
						  codap.restbits=spcodec.rbits;//-(rbytes*8);
					  }
					  icc--; 
				  } //if(!cpass) 
              }  //for(icc)
          } while(++cpass < 2);  
      } // end encoding 3D_Mesh_Object_Base_Layer() 
  } //end encoding 3D_Mesh_Object_Layer() 

	return true;
}
