//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include "MPG_Mesh.h"

//---------------------------------------------------------------------------
//
//   MPG_Mesh::set
//
//---------------------------------------------------------------------------

void MPG_Mesh::set(vector<int>& coordIndex)
{
  clear();
  if(&_coordIndex != &coordIndex) _coordIndex = coordIndex;
  if(_coordIndex.size()>0)
    {
      if(_coordIndex.back()>=0) _coordIndex.push_back(-1);

      int i;
      for(i=0;i<_coordIndex.size();i++)
        if(_coordIndex[i]>_nV)
          _nV = _coordIndex[i];
      _nV++;

      _face.erase(_face.begin(),_face.end());
      for(i=0;i<_coordIndex.size();i++)
        if(i==0||_coordIndex[i-1]<0)
          _face.push_back(i);
      if(i==0||_coordIndex[i-1]<0)
        _face.push_back(i);
      
      _createEdges();
      _createConnectedComponents();
    }
}

//---------------------------------------------------------------------------
//
//   MPG_Mesh::_createConnectedComponents
//
//---------------------------------------------------------------------------

bool MPG_Mesh::_createConnectedComponents()
{
  if(_n_cc==0)
    {
      int nV = getNumberOfVertices();
      int nF = getNumberOfFaces();

      MPG_Partition p(nV);

      int nfi,i,j,k,icc;

      _deleteConnectedComponents();

      // partition vertices into connected components
      for(i=0;i<getNumberOfFaces();i++)
        {
          nfi = getNumberOfFaceCorners(i);
          for(j=0,k=1;j<nfi-1;j++,k=(++k)%nfi)
            p.join(getFaceCorner(i,j),getFaceCorner(i,k));
        }
      _n_cc = p.parts();

      _v_cc_beg.erase(_v_cc_beg.begin(),_v_cc_beg.end());
      _v_cc_beg.insert(_v_cc_beg.end(),(size_t)(_n_cc+1),0);

      _f_cc_beg.erase(_f_cc_beg.begin(),_f_cc_beg.end());
      _f_cc_beg.insert(_f_cc_beg.end(),(size_t)(_n_cc+1),0);

      // assign component numbers to the vertices so that
      // each ifs is partitioned into a range of consecutive components
      vector<int> v_comp((size_t) nV,-1);
      icc = 0;
      for(i=0;i<getNumberOfFaces();i++)
        {
          j = p.find(getFaceCorner(i,0));
          if(v_comp[j]<0)
            { _v_cc_beg[icc+1] = p.size(j); v_comp[j] = icc++; }
          _f_cc_beg[v_comp[j]+1]++;
        }
      for(i=0;i<nV;i++)
        if(p.size(i)==0)
          v_comp[i] = v_comp[p.find(i)];
      // gt @ 05/04/99
      for(i=0;i<nV;i++)
        if(v_comp[i]<0)
          v_comp[i] = icc++;

      // determine the sizes of the vertex and face components
      for(icc=0;icc<_n_cc;icc++)
        {
          _v_cc_beg[icc+1] += _v_cc_beg[icc];
          _f_cc_beg[icc+1] += _f_cc_beg[icc];
        }

      // note that unused vertices,
      // i.e., vertices not pointed to by any face,
      // correspond to connected components with one vertex and zero faces

      vector<int> ic((size_t) _n_cc,0);

      _v_cc.erase(_v_cc.begin(),_v_cc.end());
      _v_cc.insert(_v_cc.end(),(size_t)nV,0);
      _inv_v_cc.erase(_inv_v_cc.begin(),_inv_v_cc.end());
      _inv_v_cc.insert(_inv_v_cc.end(),(size_t)nV,-1);
      for(i=0;i<nV;i++)
        {
          icc = v_comp[i];      // connected component index of vertex i
          int j=_v_cc_beg[icc]+ic[icc]; // vertex index within component
          _v_cc[j] = i;
          _inv_v_cc[i] = ic[icc];
          ic[icc]++;
        }

      for(i=0;i<_n_cc;i++)
        ic[i] = 0;

      _f_cc.erase(_f_cc.begin(),_f_cc.end());
      _f_cc.insert(_f_cc.end(),(size_t)nF,0);
      _inv_f_cc.erase(_inv_f_cc.begin(),_inv_f_cc.end());
      _inv_f_cc.insert(_inv_f_cc.end(),(size_t)nF,-1);
      for(i=0;i<nF;i++)
        {
          icc = v_comp[getFaceCorner(i,0)];
          int j=_f_cc_beg[icc]+ic[icc];
          _f_cc[j] = i;
          _inv_f_cc[i] = ic[icc];
          ic[icc]++;
        }

    }

  return false;
}

//---------------------------------------------------------------------------
//
//   MPG_Mesh::_createEdges
//
//---------------------------------------------------------------------------

void MPG_Mesh::_createEdges()
{
  int nV = getNumberOfVertices();
  if(nV>0 && _graph==(MPG_Graph*)0)
    {
      _graph = new MPG_Graph(nV);
      int nF = getNumberOfFaces();
      int i;
      for(i=0;i<nF;i++)
        {
          int nfi = getNumberOfFaceCorners(i);
          int *fi = getFace(i);
          _graph->insertFace(nfi,fi);
        }
      _graph->enumerateEdges();


      vector<int> edgeF;
      vector<int> edgeI;
      {
        int nE = getNumberOfEdges();
        edgeF.reserve(nE+1);
        edgeF.insert(edgeF.end(),(size_t) nE+1,0);

        // count number of incident faces
        int i,j,k;
        for(i=0;i<nF;i++)
          {
            int nfi = getNumberOfFaceCorners(i);
            int *fi = getFace(i);
            for(j=0;j<nfi;j++)
              {
                int k = (j+1)%nfi;
                MPG_Graph::Edge* e = getEdge(fi[j],fi[k]);
                if(e!=(MPG_Graph::Edge*)0)
                  {
                    int  ie = e->getIndex();
                    edgeF[ie+1]++;
                  }
              }
          }
        for(i=0;i<nE;i++)
          edgeF[i+1] += edgeF[i];
        int nEF = edgeF[nE];
        
        edgeI.reserve(nEF);
        edgeI.insert(edgeI.end(),(size_t) nEF,-1);
        
        vector<int> nFe((size_t) nE,0);

        for(i=0;i<nF;i++)
          {
            int nfi = getNumberOfFaceCorners(i);
            int *fi = getFace(i);
            for(j=0;j<nfi;j++)
              {
                k = (j+1)%nfi;
                MPG_Graph::Edge* e = getEdge(fi[j],fi[k]);
                if(e!=(MPG_Graph::Edge*)0)
                  {
                    int  ie = e->getIndex();
                    edgeI[edgeF[ie]+nFe[ie]] = i;
                    nFe[ie]++;
                  }
              }
          }
      }
      _graph->setEdgeFaces(edgeF,edgeI);

      _classifyEdges();
      _classifyVertices();
    }
}

//---------------------------------------------------------------------------
//
//   MPG_Mesh::_partitionCorners
//
//---------------------------------------------------------------------------

MPG_Partition *MPG_Mesh::_partitionCorners(PartitionMode pm,
                                           vector<bool> *emk)
{
  MPG_Partition *p = new MPG_Partition(getNumberOfCorners());
  int nV = getNumberOfVertices();

  if( pm==ORIENTED_MANIFOLD ||
      pm==MANIFOLD          ||
      (pm==MARKED_EDGES && emk!=(vector<bool>*)0))
    {
      // make sure edge marks array is long enough
      int nE = getNumberOfEdges();
      if(emk!=(vector<bool>*)0 && emk->size()<nE)
        {
          int n = nE-emk->size();
          emk->insert(emk->end(),(size_t)n,false);
        }

      MPG_Graph::Edge *e;

      // faces must have no repeated vertices !!!
      // otherwise findEdgeInFace() makes mistakes

      // mesh is cut through marked edges by cutting through
      // all the non-boundary edges and then stitching together
      // the non-marked edges
      //
      // edge marks are implicit in the MANIFOLD and ORIENTED_MANIFOLD
      // cases

      int h,i;
      for(i=0;i<nV;i++)
        for(e=getFirstEdge(i);e;e=getNextEdge(e))
          if((pm==ORIENTED_MANIFOLD && isRegularEdge(*e)==true ) ||
             (pm==MANIFOLD          && isRegularEdge(*e)==true ) ||
             (pm==MARKED_EDGES      && (*emk)[e->getIndex()]==false))
            {
              int if0 = getEdgeFace(*e,0);
              int if1 = -1;
              for(h=1;h<getNumberOfEdgeFaces(*e);h++)
                if(if0!=(if1=getEdgeFace(*e,h)))
                  {
                    // find edge in first face
                    int if0v0,if0v1;
                    if(findEdgeInFace(e,if0,if0v0,if0v1)==false)
                      continue;

                    // find edge in second face
                    int if1v0,if1v1;
                    if(findEdgeInFace(e,if1,if1v0,if1v1)==false)
                      continue;

                    int if0c0 = getFaceCornerIndex(if0,if0v0);
                    int if0c1 = getFaceCornerIndex(if0,if0v1);
                    int if1c0 = getFaceCornerIndex(if1,if1v0);
                    int if1c1 = getFaceCornerIndex(if1,if1v1);

                    int nf0 = getNumberOfFaceCorners(if0);
                    int nf1 = getNumberOfFaceCorners(if1);
                    if(pm!=ORIENTED_MANIFOLD ||
                       // join only if the two faces are consistently oriented
                       ((if0v1==(if0v0+1)%nf0 && if1v0==(if1v1+1)%nf1) ||
                        (if0v0==(if0v1+1)%nf0 && if1v1==(if1v0+1)%nf1)))
                      {
                        p->join(if0c0,if1c0);
                        p->join(if0c1,if1c1);
                      }
                  }
            }
    }

  return p;
}

//---------------------------------------------------------------------------
//
//   MPG_Mesh::_classifyEdges
//
//---------------------------------------------------------------------------

void MPG_Mesh::_classifyEdges()
{
  if(_graph==(MPG_Graph*)0) _createEdges();
  _nE_isolated = 0;
  _nE_boundary = 0;
  _nE_regular  = 0;
  _nE_singular = 0;
  int i;
  for(i=0;i<_nV;i++)
    for(MPG_Graph::Edge *e=getFirstEdge(i);e;e=getNextEdge(e))
      switch(getNumberOfEdgeFaces(*e))
        {
        case 0:  _nE_isolated++; break;
        case 1:  _nE_boundary++; break;
        case 2:  _nE_regular++;  break;
        default: _nE_singular++; break;
        }
  _graph->enumerateEdges();
}

//---------------------------------------------------------------------------
//
//   MPG_Mesh::_classifyVertices
//
//---------------------------------------------------------------------------

void MPG_Mesh::_classifyVertices()
{
  MPG_Partition  *p = _partitionCorners(); // edges are used here
  int           nFV = p->n();
  int            nV = getNumberOfVertices();

  int i,if0,if0v0;

  vector<int> n_v_used((size_t) nV,0);

  for(i=if0=0;if0<getNumberOfFaces();if0++)
    for(if0v0=0;if0v0<getNumberOfFaceCorners(if0);if0v0++,i++)
      if(p->size(i)>0)
        n_v_used[getFaceCorner(if0,if0v0)]++;

  _nV_isolated = 0;
  _nV_regular  = 0;
  _nV_singular = 0;
  for(i=0;i<nV;i++)
    switch(n_v_used[i])
      {
      case 0 : _nV_isolated++; break;
      case 1 : _nV_regular++;  break;
      default: _nV_singular++; break;
      }

  delete p;
}

//---------------------------------------------------------------------------
//
// MPG_Mesh::isOrientedManifold
//
// note that:
//
// 1) for a regular edge
//    to be counted as such, both incident faces must
//    be consitently oriented !!!
//
// 2) the manifold may be orientable but not be oriented
//    in which case isOrientedManifold()==false
//
// 3) a surface composed of exactly two triangles
//    with opposite orientations is a manifold
//
// gives valid results only if called after
// _classifyEdges() and _classifyFaces()
//
//---------------------------------------------------------------------------

const bool MPG_Mesh::isOrientedManifold()
{
  _createEdges();

  bool is_oriented_manifold =
    (getNumberOfEdges()==_nE_boundary+_nE_regular &&
     getNumberOfVertices()==_nV_regular)?
    true:false;

  return is_oriented_manifold;
}

//---------------------------------------------------------------------------
//
//   MPG_Mesh::findEdgeInFace
//
//---------------------------------------------------------------------------

bool MPG_Mesh::findEdgeInFace(MPG_Graph::Edge *e,
                              int if0, int& if0v0, int &if0v1)
{
  bool found_edge = false;
  if0v0 = if0v1 = -1;
  if(e!=(MPG_Graph::pEdge)0 && 0<=if0 && if0<getNumberOfFaces())
    {
      int i,j;
      int iv0 = getEdgeVertex(*e,0);
      int iv1 = getEdgeVertex(*e,1);
      int nf0 = getNumberOfFaceCorners(if0);
      for(i=0,j=1;i<nf0;i++,j=(j+1)%nf0)
        if(getFaceCorner(if0,i)==iv0 &&
           getFaceCorner(if0,j)==iv1)
          { if0v0 = i; if0v1 = j; found_edge = true; break; }
        else if(getFaceCorner(if0,j)==iv0 &&
                getFaceCorner(if0,i)==iv1)
          { if0v0 = j; if0v1 = i; found_edge = true; break; }
    }
  return found_edge;
}

//---------------------------------------------------------------------------
//
//   MPG_Mesh::hasFacesWithDuplicateVertices
//
//---------------------------------------------------------------------------

bool MPG_Mesh::hasFacesWithDuplicateVertices()
{
  bool has_faces_with_multiple_vertices = false;

  int nV = getNumberOfVertices();
  int nE = getNumberOfEdges();
  int nF = getNumberOfFaces();

  int i,j,nfi,fij;

  vector<bool> v_is_in_f((size_t) nV,false);

  for(i=0;i<nF;i++)
    {
      nfi = getNumberOfFaceCorners(i);
      for(j=0;j<nfi;j++)
        {
          fij = getFaceCorner(i,j);
          if(v_is_in_f[fij]==true)
            { has_faces_with_multiple_vertices = true; break; }
          v_is_in_f[fij] = true;
        }
      for(j=0;j<nfi;j++)
        v_is_in_f[getFaceCorner(i,j)] = false;

      if(has_faces_with_multiple_vertices==true)
        break;
    }

  return has_faces_with_multiple_vertices;
}

//------------------------------------------------------------------------
//
//   MPG_Mesh::markBoundaryVertices
//
//------------------------------------------------------------------------

int MPG_Mesh::markBoundaryVertices(vector<bool>& v_is_bdry)
{
  int nV_bdry = 0;
  int nV = getNumberOfVertices();
  v_is_bdry.erase(v_is_bdry.begin(),v_is_bdry.end());
  v_is_bdry.insert(v_is_bdry.end(),(size_t) nV,false);
  int i;
  for(i=0;i<nV;i++)
    for(MPG_Graph::Edge *e=getFirstEdge(i);e;e=getNextEdge(e))
      if(getNumberOfEdgeFaces(e->getIndex())==1)
        {
          int iv0 = e->getVertex(0);
          if(v_is_bdry[iv0]==false) { v_is_bdry[iv0] = true; nV_bdry++; }
          int iv1 = e->getVertex(1);
          if(v_is_bdry[iv1]==false) { v_is_bdry[iv1] = true; nV_bdry++; }
        }
  return nV_bdry;
}

//---------------------------------------------------------------------------
//
//   MPG_Mesh::markFacesWithDuplicateVertices
//
//---------------------------------------------------------------------------

int MPG_Mesh::markFacesWithDuplicateVertices(vector<bool>& fmk)
{
  int n_faces_with_multiple_vertices = 0;

  int nV = getNumberOfVertices();
  int nE = getNumberOfEdges();
  int nF = getNumberOfFaces();

  int i,j,nfi,fij;

  vector<bool> v_is_in_f((size_t) nV,false);

  int iMax = (fmk.size()<nF)?fmk.size():nF;
  for(i=0;i<iMax;i++)
    if(fmk[i]==false)
      {
        nfi = getNumberOfFaceCorners(i);
        for(j=0;j<nfi;j++)
          {
            fij = getFaceCorner(i,j);
            if(v_is_in_f[fij]==true)
              { fmk[i] = true; n_faces_with_multiple_vertices++; break; }
            v_is_in_f[fij] = true;
          }
        // clean up for next face
        for(j=0;j<nfi;j++)
          {
            fij = getFaceCorner(i,j);
            v_is_in_f[fij] = false;
          }
      }
  return n_faces_with_multiple_vertices;
}

//---------------------------------------------------------------------------
//
//   MPG_Mesh::getNextEdgeCC
//
//---------------------------------------------------------------------------

MPG_Graph::pEdge MPG_Mesh::getNextEdgeCC(int icc)
{
  MPG_Graph::Edge *e = (MPG_Graph::pEdge)0;
  if(_curr_e_cc==(MPG_Graph::pEdge)0)
    e = getFirstEdgeCC(icc);
  else if((e=getNextEdge(_curr_e_cc))==(MPG_Graph::pEdge)0)
    {
      int nVcc = getNumberOfVerticesCC(icc);
      int i;
      for(i=_curr_i_cc+1;i<nVcc;i++)
        {
          int k = _v_cc[_v_cc_beg[icc]+i];
          if((e=getFirstEdge(k))!=(MPG_Graph::pEdge)0)
            { _curr_i_cc = i; _curr_e_cc = getFirstEdge(k); break; }
        }
      if(i>=nVcc)
        { _curr_i_cc = -1; _curr_e_cc = (MPG_Graph::pEdge)0; }
      e = _curr_e_cc;
    }
  _curr_e_cc = e;
  return e;
}

//---------------------------------------------------------------------------
//
//   MPG_Mesh::partitionVertices
//
//---------------------------------------------------------------------------

void MPG_Mesh::partitionVertices(int nD,
                                 vector<float>& coord,
                                 vector<int>&   coordIndex,
                                 int            x_band,
                                 int            y_band,
                                 int            z_band,
                                 int            n_band,
                                 vector<bool>&  e_marks)
{
  vector<int>       qCoord;
  vector<int>       qNormal;
  int               nF;
  {
    vector<float>     faceCentroids;
    vector<float>     faceNormals;
    {
      int i = 0;
      int t = 0;
      while(i < coordIndex.size())
        {
          int i0 = coordIndex[i];
          int i1 = coordIndex[i+1];
          int i2 = coordIndex[i+2];

          float v1[3], v2[3];
          v1[0] = coord[3*i1] - coord[3*i0];
          v1[1] = coord[3*i1+1] - coord[3*i0+1];
          v1[2] = coord[3*i1+2] - coord[3*i0+2];

          v2[0] = coord[3*i2] - coord[3*i0];
          v2[1] = coord[3*i2+1] - coord[3*i0+1];
          v2[2] = coord[3*i2+2] - coord[3*i0+2];

          float n[3];
          n[0] = v1[1] * v2[2] - v1[2] * v2[1];
          n[1] = v1[2] * v2[0] - v1[0] * v2[2];
          n[2] = v1[0] * v2[1] - v1[1] * v2[0];

          float n_length = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
          if(n_length != 0.0)
            {
              n[0] = n[0]/n_length;
              n[1] = n[1]/n_length;
              n[2] = n[2]/n_length;
            }

          faceNormals.push_back(n[0]);
          faceNormals.push_back(n[1]);
          faceNormals.push_back(n[2]);

          float c[3];
          c[0] = c[1] = c[2] = 0.;

          int nIndices = 0;
          do {

            if(coordIndex[i] != -1)
              {
                nIndices++;

                c[0] += coord[3*coordIndex[i]];
                c[1] += coord[3*coordIndex[i]+1];
                c[2] += coord[3*coordIndex[i]+2];
              }

            i++;

          } while(coordIndex[i] != -1);

          if(nIndices != 0)
            {
              c[0] /= nIndices;
              c[1] /= nIndices;
              c[2] /= nIndices;

              faceCentroids.push_back(c[0]);
              faceCentroids.push_back(c[1]);
              faceCentroids.push_back(c[2]);
            }

          i++;  t++;
        }
    }

#define MAX(a,b)        (((a)>(b))?(a):(b))

    if((x_band != 0) ||
       (y_band != 0) ||
       (z_band != 0))
      {
        int bitsPerCoord  = x_band;
        bitsPerCoord  = MAX(bitsPerCoord, MAX(y_band, z_band));

        MPG_BBoxQ *boxQuantizer = new MPG_BBoxQ(nD, bitsPerCoord);
        boxQuantizer->setBBox(faceCentroids);
        boxQuantizer->quantize(faceCentroids, qCoord);
      }

    if(n_band >= 3)
      {
        //    int bitsPerNormal = 3+2*n_band;
        int bitsPerNormal = n_band;
        MPG_NormalQ *normalQuantizer = new MPG_NormalQ(bitsPerNormal);
        normalQuantizer->quantize(faceNormals, qNormal);
      }

    nF = faceCentroids.size()/3;
  }

  MPG_Partition *p = new MPG_Partition(nF);

  {
    int nV = getNumberOfVertices();
    int nE = getNumberOfEdges();
    if(e_marks.size()<nE)
      {
        int n = nE-e_marks.size();
        e_marks.insert(e_marks.end(),(size_t)n,false);
      }

    int i;
    MPG_Graph::Edge *e;

    int changes = 1;
    while(changes)
      {
        changes = 0;

        for(i=0;i<nV;i++)
          for(e=getFirstEdge(i);e;e=getNextEdge(e))
            if(isRegularEdge(*e)==true)
              {
                int f0 = getEdgeFace(*e,0);
                int f1 = getEdgeFace(*e,1);

                if(p->find(f0) != p->find(f1))
                  {
                    bool join = false;

                    if(n_band != 0)
                      if(qNormal[f0] == qNormal[f1])
                        join = true;

                    if(x_band != 0)
                      if(qCoord[3*f0] == qCoord[3*f1])
                        join = true;

                    if(y_band != 0)
                      if(qCoord[3*f0+1] == qCoord[3*f1+1])
                        join = true;

                    if(z_band != 0)
                      if(qCoord[3*f0+2] == qCoord[3*f1+2])
                        join = true;

                    if((n_band == 0) &&
                       (x_band == 0) &&
                       (y_band == 0) &&
                       (z_band == 0))
                      {
                        if((p->size(p->find(f0)) + p->size(p->find(f1))) < 256)
                          join = true;
                      }

                    if(join)
                      {
                        changes = 1;
                        p->join(f0, f1);
                        p->join(f1, f0);
                      }
                  }
              }
      }

    for(i=0;i<nV;i++)
      for(e=getFirstEdge(i);e;e=getNextEdge(e))
        if(isRegularEdge(*e)==true)
          {
            int f0 = getEdgeFace(*e,0);
            int f1 = getEdgeFace(*e,1);
            
            if(p->find(f0) != p->find(f1))
              e_marks[e->getIndex()] = true;
          }
  }

}

