//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------

#pragma warning(disable:4786)

#include <stdio.h>
#include <stdlib.h>
#include "MPG_MeshTs.h"

#define NO_VERTEX -1
#define NO_EDGE   -1
#define NO_FACE   -1
#define NO_CORNER -1

#define ALREADY_CODED	0x01
#define AVAIL_PRED		0X02
#define AVAIL_RECON		0x04

#define TryToAddEdgeToFTree(e) \
{ \
  int ie = e->getIndex(); \
  if(e!=(MPG_Graph::Edge*)0 && isInVtree(ie)==false && isInFtree(ie)==false) \
    { \
      int i0 = getEdgeFace(*e,0); \
      int i1 = getEdgeFace(*e,1); \
      if(i0!=NO_FACE&&i1!=NO_FACE && \
         f_set.find(_inv_f_cc[i0])!=f_set.find(_inv_f_cc[i1])) \
        { \
          f_set.join(_inv_f_cc[i0],_inv_f_cc[i1]); \
          _is_ftree_edge[ie] = true; \
        } \
    } \
}

#define TryToAddEdgeToVTree(e) \
{ \
  int ie = e->getIndex(); \
  if(e!=(MPG_Graph::Edge*)0 && isInVtree(ie)==false && isInFtree(ie)==false) \
    { \
      int i0 = getEdgeVertex(*e,0); \
      int i1 = getEdgeVertex(*e,1); \
      if(i0!=NO_VERTEX&&i1!=NO_VERTEX && \
         v_set.find(_inv_v_cc[i0])!=v_set.find(_inv_v_cc[i1])) \
        { \
          v_set.join(_inv_v_cc[i0],_inv_v_cc[i1]); \
          _is_vtree_edge[ie] = true; \
        } \
    } \
}

#define EdgeIsInternalToLayer(e) \
((face_is_in_f_loop[getEdgeFace(*e,0)]==true && \
  face_is_in_f_loop[getEdgeFace(*e,1)]==true)?1:0)

//---------------------------------------------------------------------------
//
//   MPG_MeshTs::constructTrees
//
//---------------------------------------------------------------------------

bool MPG_MeshTs::_constructTrees()
{
  bool success = true;

  int nV = getNumberOfVertices();
  int nF = getNumberOfFaces();
  int nE = getNumberOfEdges();

  _is_vtree_edge.erase(_is_vtree_edge.begin(),_is_vtree_edge.end());
  _is_vtree_edge.insert(_is_vtree_edge.end(),(size_t)nE,false);

  _is_ftree_edge.erase(_is_ftree_edge.begin(),_is_ftree_edge.end());
  _is_ftree_edge.insert(_is_ftree_edge.end(),(size_t)nE,false);

  for(ccBegin();!ccEnd();ccNext())
    {
      // process one connected component ------------------------------------
      int i_cc = getIndexCC();

      // fill inverse maps used by
      // TryToAddEdgeToVTree(e) and TryToAddEdgeToFTree(e)
      int nVcc = getNumberOfVerticesCC();
      int nFcc = getNumberOfFacesCC();

      // data structures to mantain spanning trees --------------------------
      vector<int> v_order((size_t) nVcc,0);
      MPG_Partition v_set(nVcc,v_order.begin());
      vector<int> f_order((size_t) nFcc,0);
      MPG_Partition f_set((size_t) nFcc,f_order.begin());

      // construct face tree ------------------------------------------------
      {
        int fi,j,j0,k,fij,fik,nfi;
        MPG_Graph::Edge *e;

        vector<int> face_stack;

        // traversal starts at first edge of first face of CC
        fi  = getFaceIndexCC(0);
        nfi = getNumberOfFaceCorners(fi);
        j   = 0;
        k   = 1;
        face_stack.push_back(fi);
        face_stack.push_back(j);

        // enter face fi through edge (j,k)
        int nFccVisited = 0;
        do {
          nFccVisited++;
          if(nFccVisited==nFcc)
            break;

          j0 = j;

          // push all other edges onto face stack
          for(j=(j+1)%nfi;j!=j0;j=k)
            {
              k=(j+1)%nfi;
              fij = getFaceCorner(fi,j);
              fik = getFaceCorner(fi,k);

              // to reduce the size of the face stack
              // test if edge is regular and
              // opposite face has been visited)
              
              e = getEdge(fij,fik);
              
              if(isRegularEdge(e->getIndex())==true &&
                 isInFtree(e->getIndex())==false)
                {
                  face_stack.push_back(fi);
                  face_stack.push_back(j);
                }
            }

          // pop edges from stack until one is found
          // which can be added to the FTree
          int face_stack_size = face_stack.size();
          do {
            j   = face_stack.back(); face_stack.pop_back();
            fi  = face_stack.back(); face_stack.pop_back();
            nfi = getNumberOfFaceCorners(fi);
            k=(j+1)%nfi;

            fij = getFaceCorner(fi,j);
            fik = getFaceCorner(fi,k);
            e   = getEdge(fij,fik);

            TryToAddEdgeToFTree(e);
            // get face index on the other side of edge
            fi = (isInFtree(e->getIndex())==true)?
              getOtherEdgeFace(*e,fi):NO_FACE;

            face_stack_size -= 2;
          } while (fi==NO_FACE && face_stack_size>0);

          // last edge poped from the stack has been crossed
          // and fi is the index of the face accross the edge
          // edge (fik,fij) should have inverse orientation in fi :
          // find vertex fik in face fi
          j = getFaceCornerCurrPos(fi,fik);

          // next vertex in face fi should now be fij
          nfi = getNumberOfFaceCorners(fi);
          k = (j+1)%nfi;
        } while(getFaceCorner(fi,k)==fij);

        // error if number of connected components in face forest > 1
        if(f_set.parts()>1)
          {
            success = false;
          }
      }

      if(success==false) break;

      // construct vertex tree -------------------------------------------------
      {
        MPG_Graph::Edge *e;
        for(e=getFirstEdgeCC();e!=(MPG_Graph::Edge*)0;e=getNextEdgeCC())
          if(isInFtree(e->getIndex())==false)
            { TryToAddEdgeToVTree(e); }
      }
    } // for(i_cc=0,ccBegin();!ccEnd();i_cc++,ccNext())

  return(success);
}

#undef TryToAddEdgeToFTree
#undef TryToAddEdgeToVTree
#undef EdgeIsInternalToLayer
#undef EdgeIsInVTree
#undef EdgeIsInFTree
#undef EdgeIsJumpEdge

class Stack9 {
protected:
  vector<int> _VI;
public:
  ~Stack9() {}
   Stack9() {}

  void push(int ifN, int cL, int cR, int icLp, int icRp,
            int marchL, int loopCurr, int ivP, int iTp)
  {
    _VI.push_back(ifN);
    _VI.push_back(cL);
    _VI.push_back(cR);
    _VI.push_back(icLp);
    _VI.push_back(icRp);
    _VI.push_back(marchL);
    _VI.push_back(loopCurr);
    _VI.push_back(ivP);
    _VI.push_back(iTp);
  }

  bool pop(int& ifN, int& cL, int& cR, int& icLp, int& icRp,
           int& marchL, int& loopCurr, int& ivP, int iTp)
  {
    if(_VI.size()>=7)
      {
        iTp      = _VI.back(); _VI.pop_back();
        ivP      = _VI.back(); _VI.pop_back();
        loopCurr = _VI.back(); _VI.pop_back();
        marchL   = _VI.back(); _VI.pop_back();
        icRp     = _VI.back(); _VI.pop_back();
        icLp     = _VI.back(); _VI.pop_back();
        cR       = _VI.back(); _VI.pop_back();
        cL       = _VI.back(); _VI.pop_back();
        ifN      = _VI.back(); _VI.pop_back();
        return true;
      }
    return false;
  }

  int depth()
  { return _VI.size()/6; }
};

//---------------------------------------------------------------------------
//
//   pushBackSample
//
//---------------------------------------------------------------------------

inline int* pushBackSample(vector<int>& v, int d, int* nv)
{
  int s = v.size();
  int i;
  for(i=0;i<d;i++)
    v.push_back(nv[i]);
  nv = v.begin()+s;
  return nv;
}

//---------------------------------------------------------------------------
//
//   MPG_MeshTs::representConnectivity
//
//---------------------------------------------------------------------------

bool MPG_MeshTs::representConnectivity(vector<MPG_pSPolygon>& vpsp,
                                     vector<MPG_pVGraph>&  vpvg,
                                     MPG_Property&         pCoord,
                                     MPG_Property&         pNormal,
                                     MPG_Property&         pColor,
                                     MPG_Property&         pTexCoord,
                                     vector<int>&          qCoord,
                                     vector<int>&          qNormal,
                                     vector<int>&          qColor,
                                     vector<int>&          qTexCoord,
                                     vector<int>&          v_visited,
                                     vector<int>&          v_cluster,
                                     vector<int>&          v_father,
                                     vector<int>&          v_traversal,
                                     vector<int>&          f_traversal,
									 vector<int>&           f_index,
									 vector<int>&           c_index,
									 MPG_BranchForest*	   root_btf
									 )
{
  int nV  = getNumberOfVertices();
  int nF  = getNumberOfFaces();
  int nC  = getNumberOfCorners();
  int nT  = getNumberOfTriangles();
  int nE  = getNumberOfEdges();
  int nCC = getNumberOfConnectedComponents();

  vpsp.erase(vpsp.begin(),vpsp.end());
  vpvg.erase(vpvg.begin(),vpvg.end());
  v_visited.erase(v_visited.begin(),v_visited.end());
  v_visited.insert(v_visited.end(),(size_t) nV,AVAIL_PRED);  

  if(v_cluster.size()>0)
    {
      if(v_cluster.size()!=nV) return false;
      int i;
      for(i=0;i<nV;i++)
        if(v_cluster[i]<0) return false;
    }
  else
    {
      int i;
      for(i=0;i<nV;i++)
        v_cluster.push_back(i);
    }

  v_father.erase(v_father.begin(),v_father.end());
  v_father.insert(v_father.end(),(size_t) nV,NO_VERTEX);
  v_traversal.erase(v_traversal.begin(),v_traversal.end());
  f_traversal.erase(f_traversal.begin(),f_traversal.end());

  vector<int> f_father((size_t) nF,NO_FACE);
  vector<int> c_traversal;
  vector<int> c_father((size_t) nC,NO_CORNER);

  f_index.erase(f_index.begin(),f_index.end());
  c_index.erase(c_index.begin(),c_index.end());
  MPG_BranchForest*   cur_btf = root_btf;
  
  if(_constructTrees()==false) return false;

  int cL,cR,cO,cLO,cRO,cLL,cLR;
  int ifP,ifN,ifL,ifR,nfN;
  int ivL,ivR,ivO,ivP,ivRO,ivLO;
  int icL,icR,icO,icLp,icRp;
  int marchL,incL,incR,loopCurr;
  bool polyBit;
  MPG_Graph::pEdge e,eL,eR;

  Stack9 stack;

  vector<MPG_pLoop> vplp;
  vector<int>       nf_visits((size_t) nF,0);

  int iT = 0;
  for(ccBegin();!ccEnd();ccNext())
    { // process one connected component
      int i_cc = getIndexCC();
      MPG_SPolygon* p_sp = new MPG_SPolygon();
      vpsp.push_back(p_sp);
      MPG_Loops* p_loop = new MPG_Loops();
      vplp.push_back(p_loop);

      int nVcc = getNumberOfVerticesCC();
      int nFcc = getNumberOfFacesCC();
      int nTcc = 0;
      int i;
      for(i=0;i<nFcc;i++)
        nTcc += (getNumberOfFaceCornersCC(i)-2);

      // choose one vertex tree edge as the root edge
      MPG_Graph::Edge *eRoot = NULL;
      for(e=getFirstEdgeCC();e;e=getNextEdgeCC())
        if(isInVtree(e->getIndex()))
          { eRoot = e; break; }
      if(eRoot==NULL) return false;

      // choose one face incident to the root edge as the root face
      int ifRoot = getEdgeFace(eRoot->getIndex());

      // root triangle has no previous vertex
      ivP    = NO_VERTEX;
      marchL = 0;

      // traversal starts at root edge of root face
      ifP = ifRoot;
      ifN = ifRoot;
      nfN = getNumberOfFaceCorners(ifRoot);

      // find root edge endpoints in root face
      cL = getFaceCornerCurrPos(ifRoot,(*eRoot)[0]);
      cR = getFaceCornerCurrPos(ifRoot,(*eRoot)[1]);
      // traversal of face corners should be clockwise
      if(cL!=(cR+1)%nfN) { int k=cL; cL=cR; cR=k; }
      // test just in case anything goes wrong
      if(cL!=(cR+1)%nfN) return false;

      // root vertex is choosen to make orientation of edge
      // compatible with orientation of face
      int ivRoot = getFaceCorner(ifRoot,cR);

      ivR     = getFaceCorner(ifN,cR);
      ivL     = getFaceCorner(ifN,cL);
      p_loop->init(ivR,ivL);
      icLp = getFaceCornerIndex(ifN,cL);
      icRp = icLp; // getFaceCornerIndex(ifN,cR);

	  cur_btf->MakeNextBTree();
	  cur_btf = cur_btf->GetNextBTree();
	  MPG_BranchTree*   cur_bt = cur_btf->GetRootBTree();
	  cur_bt->InsertInfo(LEFT_BRANCH,f_index.size()-1,c_index.size());	//MJH 001010 : -1 added

      int iTp = iT;
      int iTcc;
      for(iTcc = 0;iTcc<nTcc;iTcc++,iT++)
        {
          // take care of reorienting faces later !!!
          nfN     = getNumberOfFaceCorners(ifN);
          incL    = 1;
          incR    = nfN-1;

          e = (cL==(cR+incL)%nfN)?getEdge(ivR,ivL):(MPG_Graph::Edge*)0;
          marchL  = (marchL)?0:1;
          polyBit = (e!=(MPG_Graph::Edge*)0)?true:false;

          icL = getFaceCornerIndex(ifN,cL);
          icR = getFaceCornerIndex(ifN,cR);
          if(polyBit)
            {
              c_father[icL] = icLp; c_traversal.push_back(icL);
              c_father[icR] = icRp; c_traversal.push_back(icR);
			  c_index.push_back(icR);	c_index.push_back(icL);	
            }

          if(nf_visits[ifN]==0)
            {
              f_traversal.push_back(ifN);
              f_father[ifN] = ifP;
			  f_index.push_back(ifN);
            }
          nf_visits[ifN]++;

          // determine left edge and left face
          cLO     = (cL+incL)%nfN;
          ivLO    = getFaceCorner(ifN,cLO);
          eL      = getEdge(ivL,ivLO);
          if(!eL) return false;
          ifL     = (isInFtree(eL)==true)?getOtherEdgeFace(*eL,ifN):NO_FACE;

          // determine right edge and right face
          cRO     = (cR+incR)%nfN;
          ivRO    = getFaceCorner(ifN,cRO);
          eR      = getEdge(ivR,ivRO);
          if(!eR) return false;
          ifR     = (isInFtree(eR)==true)?getOtherEdgeFace(*eR,ifN):NO_FACE;

          // if cannot march left try to march right
          if( marchL && ifL!=NO_FACE && ifR==NO_FACE) marchL = 0;
          // and viceversa
          if(!marchL && ifR!=NO_FACE && ifL==NO_FACE) marchL = 1;

          {
            // if prediction along the triangle tree is ussed,
            // predict ivO from ivP,ivL,ivR here

            ivO = (cLO==cRO && ifL==NO_FACE && ifR==NO_FACE)?ivLO:
                  ( marchL && ifL==NO_FACE)?                 ivLO:
                  (!marchL && ifR==NO_FACE)?                 ivRO:
                  (cLO==cRO)?                                ivLO:
                  (marchL)?                                  ivLO:
                                                             ivRO;

            //       O <----- next vertex in traversal order
            //     /   \      if not previously visited
            //   /   ^   \    (to be predicted)
            //  L ---|--- R
            //   \       /
            //     \   /
            //       P
            //
            //  For the root triangle of each connected component,
            //  there is no P vertex. So, the ancestors of the first
            //  3 vertices of each connected component
            //  are defined as follows
            //  R | R R R  (ignored because R is the root vertex)
            //  L | R R R
            //  O | L R R
            //
            // instead of storing the indices of vertices, faces, and corners,
            // quantize, predict, and store the prediction errors here !!!
            //      if(v_visited[v_cluster[ivO]]==false)
            //        {
            //          // predict per vertex properties here
            //          // tv_vO.push_back(ivO);
            //          // tv_vL.push_back(ivL);
            //          // tv_vR.push_back(ivR);
            //          // tv_vP.push_back(ivP);
            //          // tv_visited.push_back(1);
            //
            //          // v_visited[v_cluster[ivO]] = true;
            //        }

            cO = (cLO==cRO && ifL==NO_FACE && ifR==NO_FACE)?cLO:
                 ( marchL && ifL==NO_FACE)?                 cLO:
                 (!marchL && ifR==NO_FACE)?                 cRO:
                 (cLO==cRO)?                                cLO:
                 (marchL)?                                  cLO:
                                                            cRO;
            icO = getFaceCornerIndex(ifN,cO);
            c_father[icO] = icL; c_traversal.push_back(icO);
			c_index.push_back(icO);	
          }

          p_loop->insert(ivO);

          if(cLO==cRO && ifL==NO_FACE && ifR==NO_FACE)
            { // leaf triangle
              p_sp->pushTri(MPG_SPolygon::sp_LEAF,polyBit);
              if(stack.depth()>0)
                {
                  stack.pop(ifN,cL,cR,icLp,icRp,marchL,loopCurr,ivP,iTp);
                  ivL     = getFaceCorner(ifN,cL);
                  ivR     = getFaceCorner(ifN,cR);
                  e       = getEdge(ivR,ivL);
                  ifP     = (e)?getOtherEdgeFace(*e,ifN):ifN;
                  p_loop->setCurr(loopCurr);
                }
			  while (cur_bt->GetCurDir() == RIGHT_BRANCH) 
				  cur_bt = cur_bt->GetParent();
			  if (!cur_bt) {
				  printf("Error while making Branch Tree!!!");
				  return false;
			  }
			  cur_bt->InsertInfo(RIGHT_BRANCH,f_index.size()-1,c_index.size());	//MJH 001010 : -1 added
			  cur_bt->SetCurDir(RIGHT_BRANCH);
            }
          else if(marchL && ifL==NO_FACE)
            {
              // march left : next triangle on the right
              p_sp->pushTri(MPG_SPolygon::sp_RIGHT,polyBit);
              cL = cLO; ivP = ivL; ivL = ivLO;
              e  = (cR==(cL+incL)%nfN)?getEdge(ivR,ivL):(MPG_Graph::Edge*)0;
              if(e!=(MPG_Graph::Edge*)0 && isInFtree(e))
                {
                  // move to right face
                  ifP = ifN;
                  ifN = getOtherEdgeFace(*e,ifN);
                  cL  = getFaceCornerCurrPos(ifN,ivL);
                  cR  = getFaceCornerCurrPos(ifN,ivR);
                }
              p_loop->advance();
              icLp = icO;
              icRp = icR;
              // else stay in current face
              iTp = iT;
            }
          else if(!marchL && ifR==NO_FACE)
            {
              // march right : next triangle on the left
              p_sp->pushTri(MPG_SPolygon::sp_LEFT,polyBit);
              cR = cRO; ivP = ivR; ivR = ivRO;
              e  = (cR==(cL+incL)%nfN)?getEdge(ivR,ivL):(MPG_Graph::Edge*)0;
              if(e!=(MPG_Graph::Edge*)0 && isInFtree(e))
                {
                  // move to left face
                  ifP = ifN;
                  ifN = getOtherEdgeFace(*e,ifN);
                  cL  = getFaceCornerCurrPos(ifN,ivL);
                  cR  = getFaceCornerCurrPos(ifN,ivR);
                }
              icLp = icL;
              icRp = icO;
              // else stay in current face
              iTp = iT;
            }
          else // ifL!=NO_FACE && ifR!=NO_FACE
            {
              // branching triangle
              p_sp->pushTri(MPG_SPolygon::sp_BRANCH,polyBit);
			  if (cur_bt->GetCurDir() == LEFT_BRANCH) {
				  cur_bt->MakeBranch(LEFT_BRANCH);
				  cur_bt = cur_bt->GetBranch(LEFT_BRANCH);
				  cur_bt->SetCurDir(LEFT_BRANCH);
				  cur_bt->InsertInfo(LEFT_BRANCH,f_index.size()-1,c_index.size());	//MJH 001010 : -1 added
			  }
			  else {
				  cur_bt->MakeBranch(RIGHT_BRANCH);
				  cur_bt = cur_bt->GetBranch(RIGHT_BRANCH);
				  cur_bt->SetCurDir(LEFT_BRANCH);
				  cur_bt->InsertInfo(LEFT_BRANCH,f_index.size()-1,c_index.size());	//MJH 001010 : -1 added
			  }

              if(cLO==cRO)
                {
                  // push left next face
                  cLL  = getFaceCornerCurrPos(ifL,ivL);
                  cLR  = getFaceCornerCurrPos(ifL,ivLO);
                  stack.push(ifL,cLL,cLR,icL,icO,
                             marchL,p_loop->getCurr(),ivR,iT);
                  // move right to next face
                  e    = eR;
                  ifP  = ifN;
                  ifN  = ifR;
                  ivP  = ivL;
                  ivL  = ivLO;
                  cL   = getFaceCornerCurrPos(ifN,ivL);
                  cR   = getFaceCornerCurrPos(ifN,ivR);
                  icLp = icO;
                  icRp = icR;
                }
              else if(marchL)
                {
                  // push left next face
                  e = eL; // ==NULL
                  cLL = getFaceCornerCurrPos(ifL,ivL);
                  cLR = getFaceCornerCurrPos(ifL,ivLO);
                  stack.push(ifL,cLL,cLR,icL,icO,
                             marchL,p_loop->getCurr(),ivR,iT);
                  // stay in current face
                  ivP   = ivL;
                  ivL   = ivLO;
                  cL    = cLO;
                  icLp  = icO;
                  icRp  = icR;
                }
              else // !marchL
                {
                  // push left current face
                  stack.push(ifN,cL,cRO,icL,icO,
                             marchL,p_loop->getCurr(),ivR,iT);
                  // move right to next face
                  e    = eR;
                  ifP  = ifN;
                  ifN  = ifR;
                  ivP  = ivL;
                  ivL  = ivRO;
                  cL   = getFaceCornerCurrPos(ifN,ivL);
                  cR   = getFaceCornerCurrPos(ifN,ivR);
                  icLp = icO;
                  icRp = icR;
                }
              p_loop->advance();
              iTp = iT;
            }
        }

      // all the faces of the CC must have been visited
      // if(iFcc!=nFcc) return false;
      // all the triangles of the CC must have been constructed
      if(iTcc!=nTcc) return false;

      // make vertex graph from loop
      // returns vertex tree order of traversal and father array !!!
      MPG_VGraph* p_vg = new MPG_VGraph();
      p_vg->fromLoop(*p_loop,v_traversal,v_father);
      vpvg.push_back(p_vg);

    } // for(ccBegin();!ccEnd();ccNext())

  if(!(v_traversal.size()==nV)) return false;
  if(!(v_father.size()==nV)) return false;

  // reset v_visited
  v_visited.erase(v_visited.begin(),v_visited.end());
  v_visited.insert(v_visited.end(),(size_t) nV,AVAIL_PRED);  

  return true;
}
