//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------

#pragma warning(disable:4786)

#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>

#include "MPG_NormalQ.h"

//---------------------------------------------------------------------------
//
//   MPG_NormalQ::_normal2Code
//
//---------------------------------------------------------------------------

unsigned long MPG_NormalQ::_normal2Code(const float *normal)
{
  float vx = normal[0], vy = normal[1], vz = normal[2];
  unsigned long octantCode = 0;

  // Generate bits b30, b29, b28 of code (0: positive, 1: negative)
  if (vx < 0) { vx = -vx; octantCode  = 0x00000004; }
  if (vy < 0) { vy = -vy; octantCode |= 0x00000002; }
  if (vz < 0) { vz = -vz; octantCode |= 0x00000001; }

  // Convert coordinates to integer values
  unsigned long maxCoord = 1 << _nSubdivisionLevels;
  float s = maxCoord / (vx + vy + vz);
  unsigned long x = (unsigned long)(vx*s);
  unsigned long y = (unsigned long)(vy*s);
  unsigned long z = (unsigned long)(maxCoord-vx*s-vy*s); //  long z = vz * s;

  short upsideDown = 0;

  if (maxCoord - x - y - z == 2) upsideDown = 1;
  else if (x == maxCoord ) x -= 1;
  else if ( x + y == maxCoord ) y -= 1;

  unsigned long tricode = y * (2 * maxCoord - 1) - y * (y - 1) +
    2 * x + upsideDown;

  unsigned long code =  (octantCode << 2*_nSubdivisionLevels) | tricode;
 
  return code;
}

//---------------------------------------------------------------------------
//
//   MPG_NormalQ::_code2Normal
//
//---------------------------------------------------------------------------

void MPG_NormalQ::_code2Normal(const unsigned long code,
				       float *normal)
{
  // Find triangle number in [0, 2^L] range
  unsigned long mask =  (1 << (2 * _nSubdivisionLevels)) - 1;
  unsigned long tricode = code & mask;
  unsigned long tcode = tricode;

  // Find y coordinate by solving 2nd degree equation
  // long factor = 1 << (numLevels + 1);
  // long y = (factor - sqrt(factor * factor - 4 * tricode)) / 2.0;
  unsigned long factor = 1 << _nSubdivisionLevels;
  unsigned long y = (unsigned long)(factor-sqrt(factor*factor-tricode));
  tricode += y * ( y - 2 * factor);
  unsigned long x = tricode / 2;
  short upsideDown = tricode % 2;

  // Calculate coordinates for all vertices in triangle
  float v1x = float(x + upsideDown);
  float v1y = float(y + upsideDown);
  float v2x = float(x + 1);
  float v2y = float(y);
  float v3x = float(x);
  float v3y = float(y + 1);

  // Calculate coordinates of barycenter
  float invMaxCoord = 1.0 / factor;
  normal[0] = (v1x + v2x + v3x) * invMaxCoord;
  normal[1] = (v1y + v2y + v3y) * invMaxCoord;
  normal[2] = 3.0 - normal[0] - normal[1];

  // Flip component signs if necessary
  // unsigned long octantCode = (code & ~mask) >> 2*_nSubdivisionLevels;
  unsigned long octantCode = (code >> 2*_nSubdivisionLevels)&0x7;

  if (octantCode & 0x4) normal[0] = -normal[0];
  if (octantCode & 0x2) normal[1] = -normal[1];
  if (octantCode & 0x1) normal[2] = -normal[2];

  //Normalize vector
  float invNorm = 1.0 / sqrt(normal[0]*normal[0] +
			     normal[1]*normal[1] +
			     normal[2]*normal[2]);
  normal[0] *= invNorm; normal[1] *= invNorm; normal[2] *= invNorm;
}

//---------------------------------------------------------------------------
//
//   MPG_NormalQ::quantize
//
//---------------------------------------------------------------------------

bool MPG_NormalQ::quantize(float* n, int* iqn, int nNormals)
{
  _avgError = 0.0;
  _maxError = 0.0;

  float qni[3];
  float* ni = n;
  int i,j;
  for(i=0;i<nNormals;i++,ni+=3)
    {
      iqn[i] = _normal2Code(ni);
      _code2Normal(iqn[i],qni);

      float err_i = 1.0;
      for(j=0;j<3;j++)
	err_i -= qni[j]*ni[j];

      if(err_i>_maxError)
	_maxError = err_i;
      _avgError += err_i;

    }
  _avgError /= nNormals;

  return true;
}

//---------------------------------------------------------------------------
//
//   MPG_NormalQ::quantize
//
//---------------------------------------------------------------------------

bool MPG_NormalQ::quantize(vector<float>& n, vector<int>& iqn)
{
  int nNormals = n.size()/3;
  iqn.erase(iqn.begin(),iqn.end());
  iqn.insert(iqn.end(),(size_t)nNormals, 0);
  return quantize(n.begin(),iqn.begin(),nNormals);
}

//---------------------------------------------------------------------------
//
//   MPG_NormalQ::quantize
//
//---------------------------------------------------------------------------

bool MPG_NormalQ::quantize(vector<float>& n,
                           vector<int>&   nIndx,
                           vector<int>&   iqn)
{
  int i,j,k;
  int nN  = nIndx.size();
  vector<float> n_flat;
  for(i=0;i<nIndx.size();i++)
    if((j=nIndx[i])>=0)
      for(k=0;k<3;k++)
        n_flat.push_back(n[3*j+k]);

  return quantize(n_flat,iqn);
}
//---------------------------------------------------------------------------
//
//   MPG_NormalQ::unQuantize
//
//---------------------------------------------------------------------------

bool MPG_NormalQ::unQuantize(int* iqn, float* n, int nNormals)
{
  for(int i=0;i<nNormals;i++)
    _code2Normal(iqn[i],n+3*i);

  return true;
}

//---------------------------------------------------------------------------
//
//   MPG_NormalQ::unQuantize
//
//---------------------------------------------------------------------------

bool MPG_NormalQ::unQuantize(vector<int>& iqn, vector<float>& n)
{
  int nNormals = iqn.size();

  n.erase(n.begin(),n.end());
  n.insert(n.end(),(size_t)(nNormals*3), 0.0);

  return(unQuantize(iqn.begin(),n.begin(),nNormals));
}

