//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------


#include "MPG_Partition.h"

//---------------------------------------------------------------------------
//
//   MPG_Partition::_init
//
//---------------------------------------------------------------------------

void MPG_Partition::_init(int n, int *father, int *size, int *order)
{
  _zero();
  if(n>0)
    {
      _n_max     = n;
      _father    = (father!=(int*)0)?father:new int[_n_max];
      _size      = (  size!=(int*)0)?  size:new int[_n_max];
      _order     = order;
      _partFirst = (int*)0;
      _partIndex = (int*)0;

      restart(n);
    }
}

//---------------------------------------------------------------------------
//
//   MPG_Partition::restart
//
//---------------------------------------------------------------------------

void MPG_Partition::restart(int n)
{
  if(n>0 && n<=_n_max)
    {
      _parts  = _n = n;
      int i;
      for(i=0;i<_n;i++)
        { _size[i] = 1; _father[i] = i; }
      if(_order!=(int*)0)
        memset((char*)_order,'\0',_n*sizeof(int));
    }
}

//---------------------------------------------------------------------------
//
//   MPG_Partition::find
//
//---------------------------------------------------------------------------

int MPG_Partition::find(int i)
{
  int Ri,Fi;

  /*--- travers path ---*/
  for(Ri=i;_father[Ri]!=Ri;Ri=_father[Ri]);

  /*--- compress path ---*/
  while(i!=Ri)
    { Fi = _father[i]; _father[i] = Ri; i = Fi; }

  return(i);
}

//---------------------------------------------------------------------------
//
//   MPG_Partition::join
//
//---------------------------------------------------------------------------

int MPG_Partition::join(int i, int j)
{
  int Ci  = find(i);
  int Cj  = find(j);
  int Uij = Ci;

  if(Ci!=Cj)
    {
      _parts--;
      if(_size[Ci]>=_size[Cj])
	{ Uij = _father[Cj] = Ci; _size[Ci] += _size[Cj]; _size[Cj] = 0; }
      else
	{ Uij = _father[Ci] = Cj; _size[Cj] += _size[Ci]; _size[Ci] = 0; }
      if(_order!=(int*)0)
	{ _order[i]++; _order[j]++; }
    }

  return(Uij);
}

//---------------------------------------------------------------------------
//
//   MPG_Partition::end
//
//---------------------------------------------------------------------------

void MPG_Partition::end()
{
  int* min = new int[_n];
  int i;
  for(i=0;i<_n;i++)
    { min[i] = -1; min[find(i)] = find(i); }
  for(i=0;i<_n;i++)
    if(i<min[find(i)]) min[find(i)] = i;
  for(i=0;i<_n;i++)
    if(size(i)>0 && min[i]<find(i))
      {
	_father[i]    =  min[i];  _father[min[i]] = min[i];
	_size[min[i]] = _size[i]; _size[i]        = 0;
      }

  delete [] min;
}

//---------------------------------------------------------------------------
//
//   MPG_Partition::makeParts
//
//---------------------------------------------------------------------------

void MPG_Partition::makeParts()
{
  end();
  if(_partFirst) delete [] _partFirst;

  int nP = parts();

  _partFirst = new int[_n+nP+1];
  _partIndex = _partFirst+nP+1;
  int j=0;
  int *part = new int[_n];
  int i;
  for(i=0;i<_n;i++)
    part[i] = -1;
  for(i=0;i<_n;i++)
    if(_size[i]>0)
      part[i] = j++;

  for(i=0;i<=nP;i++)
    _partFirst[i] = 0;
  for(i=0;i<_n;i++)
    _partFirst[part[find(i)]+1]++;
  for(i=0;i<nP;i++)
    _partFirst[i+1] += _partFirst[i];

  int *indx = new int[nP];
  for(i=0;i<nP;i++)
    indx[i] = 0;

  for(i=0;i<_n;i++)
    {
      int ip = part[find(i)];
      _partIndex[_partFirst[ip]+indx[ip]] = i;
      indx[ip]++;
    }

  delete [] indx;
  delete [] part;
}

//---------------------------------------------------------------------------
//
//   MPG_Partition::getPart
//
//---------------------------------------------------------------------------

void MPG_Partition::getPart(int i, int& n, int*& part)
{
  if(i>=0 && i<parts())
    {
      n    = _partFirst[i+1]-_partFirst[i];
      part = _partIndex+_partFirst[i];
    }
  else
    {
      n    = 0;
      part = (int*)0;
    }
}
