//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------

#include "MPG_Global.h"
#include "MPG_SPolygon.h"

bool MPG_SPolygon::decodeTriangleTree(MPG_QFac& codec,
																			int partition_type, int v_loop_cc_size) 
{
  int depth        = 0;
	
  int branch_position = -2;	
  ntriangles = 0;
  codap_branch_len = 0;
	
  do
	{
		for(;codec.decode(MPG_Global::tt_run_length)!=1;ntriangles++) 
			pushTri(sp_UNDET);
		ntriangles++;	
		if(codec.decode(MPG_Global::tt_leaf))
		{ pushTri(sp_LEAF); depth--; }
		else
		{ pushTri(sp_BRANCH); depth++; 
		branch_position=ntriangles; 
		}
	} while (depth >= 0);
	
	if (partition_type != 1) {
		int tmp_right_bloop_idx = (codap_right_bloop_idx)?codap_right_bloop_idx:v_loop_cc_size;
		if (tmp_right_bloop_idx - codap_left_bloop_idx - 1 > ntriangles) {
			if (branch_position == ntriangles - 2) {
				int nbits = slog2(v_loop_cc_size);
				for (int i=0; i<nbits; i++)
					codap_branch_len += codap_branch_len + codec.decode(MPG_Global::codap_branch_len_context);
				popTri(); popTri();
			}
			else {
				popTri();
			}
		}
	}
	
  return true;
}


bool MPG_SPolygon::decodeTriangleData(MPG_QFac&      codec,
																			int            nVtree,
																			int            nV0,
																			vector<int>&   v_loop_sp,
																			vector<int>&   v_visited,
																			vector<int>&   v_cluster,
																			vector<int>&   coordIndex,
																			MPG_Property&  pCoord,
																			MPG_Property&  pNormal,
																			MPG_Property&  pColor,
																			MPG_Property&  pTexCoord,
																			vector <int>&  qCoord,
																			vector <int>&  qNormal,
																			vector <int>&  qColor,
																			vector <int>&  qTexCoord,
																			int partition_type	
																			)
{
  // determine if SPolygon has properties per vertex
  
  bool has_per_vertex = (pCoord.getBinding()    == PB_PER_VERTEX ||
			 pNormal.getBinding()   == PB_PER_VERTEX ||
			 pColor.getBinding()    == PB_PER_VERTEX ||
			 pTexCoord.getBinding() == PB_PER_VERTEX)?true:false;
	
  // v_loop_sp, v_visited and v_clustered
  // are not refered to if has_per_vertex==false
  
  // triangle_data() 
  //   triangulated
  //   root_triangle()
  //   for(i=1;i<ntriangles;i++)
  //     triangle()
	
  int bitsPerCoord    =  pCoord.getBBoxQ()->getBitsPerCoordinate();
  int bitsPerNormal   =  pNormal.getNormalQ()->getBitsPerNormal();
  int bitsPerColor    =  pColor.getBBoxQ()->getBitsPerCoordinate();
  int bitsPerTexCoord =  pTexCoord.getBBoxQ()->getBitsPerCoordinate();
	
  int prev_marching_pattern = 0;
  int prev_polygon_edge     = 1;
  int triangulated          = codec.decode(MPG_Global::tt_triangulated);
	
  int nT     = getNumberOfTriangles();
  int nVloop = v_loop_sp.size(); 
  int nRuns  = getNumberOfRuns();

  //MJH 001006 : moved to class declaration (header file) for library generation
  // face tree array
//  static vector<int>  f_father;		
  // corner tree arrays
//  static vector<int>  c_traversal;
//  static vector<int>  c_father;		
	
  // triangle tree arrays
  vector<int>  run_father((size_t)nRuns,-1);
	
  vector<int>  run_length((size_t)nRuns+1,0);	
	
  vector<bool> run_ends_in_leaf((size_t)nRuns,false);
	
  // traversal stack
  vector<int>  stk((size_t) 1,-1);
	
  int nPE    = 0;    //      number of {non_root polygon_edge==0}
  int nF     = nT;   // - nPE
  int nC     = 3*nT; // - 2*nPE
	
  int SnF = f_father.size();
  int SnC = c_father.size();
	
  // AG: only per vertex bindings involve a global array and global indexing to
  // allow reuse of coordinates/properties from earlier components
  // otherwise, the qNormal, qColor, and qTexCoord arrays are local
  // to the component
	
  int nCoord = 3*(nVtree + nV0) - qCoord.size();
	
  if(pCoord.getBinding() == PB_PER_VERTEX)
    qCoord.insert(qCoord.end(),(size_t)(nCoord), 0);
	
  int nNormal = 0;
  switch(pNormal.getBinding())
	{
	case PB_PER_VERTEX: nNormal = nVtree + nV0 - qNormal.size();   break;
		
	case PB_PER_FACE:   nNormal = nF;       break;
	case PB_PER_CORNER: nNormal = nC;       break;
	default:                                break;
	}
  if(nNormal>0)
    qNormal.insert(qNormal.end(),(size_t)(nNormal), 0);
	
  int nColor = 0;
  switch(pColor.getBinding())
	{
	case PB_PER_VERTEX: nColor = 3*(nVtree + nV0) - qColor.size();   break;
		
	case PB_PER_FACE:   nColor = 3*nF;       break;
	case PB_PER_CORNER: nColor = 3*nC;       break;
	default:                                 break;
	}
  if(nColor>0)
    qColor.insert(qColor.end(),(size_t)(nColor), 0);
	
  int nTexCoord = 0;
  switch(pTexCoord.getBinding())
	{
	case PB_PER_VERTEX: nTexCoord = 2*(nVtree + nV0) - qTexCoord.size();   break;
		
	case PB_PER_CORNER: nTexCoord = 2*nC;     break;
	default:                                  break;
	}
  if(nTexCoord>0)
    qTexCoord.insert(qTexCoord.end(),(size_t)(nTexCoord), 0);
	
  // construct the father array
  int iT,iRun;
  for(iT=iRun=0;iT<nT;iT++)
	{
		run_length[iRun]++;
		switch(getTri(iT))
		{
		case sp_LEAF:
			run_father[iRun] = stk.back(); stk.pop_back();
			run_ends_in_leaf[iRun] = true;
			iRun++;
			break;
			
		case sp_UNDET:
			break;
			
		case sp_LEFT:
		case sp_RIGHT:
			break;
			
		case sp_BRANCH:
			run_father[iRun] = stk.back(); stk.pop_back();
			run_ends_in_leaf[iRun] = false;
			stk.push_back(iRun);
			stk.push_back(iRun);
			iRun++;
			break;
		}
	}
	
  // determine the L->R length of each run
  for(iRun=nRuns-1;iRun>0;iRun--)
	{
		if(run_ends_in_leaf[iRun]==true)
			run_length[iRun]++;
		else
			run_length[iRun]--;
		run_length[run_father[iRun]] += run_length[iRun];
	}
  if (!iRun)	
		run_length[iRun]--;		// iRun==0
	
  if (ntriangles == nT + 2) {
		run_length[nRuns] = codap_branch_len+1;
  }
  poly_part = -1;	
	
  int iP = -1;
  int iR =  codap_right_bloop_idx; 
  int iL =  codap_left_bloop_idx; 
  int iO = -1;
	
  vector<int> faceStack;
  int iF  = f_father.size();	
  faceStack.push_back(iF);
  int iFp = iF;
	
  int nC0  =  c_father.size();	// next free corner index
  int iC   = nC0;
  //MJH 062600 : swapped L and R
  int cRp  =  iC;
  int cR   =  iC;
  int cLp  =  iC;
  int cL   =  iC+1;
  int cO   = -1;
	
  c_father.insert(c_father.end(),(size_t)nC,-1);
	
  MPG_Loops loop;
  MPG_Loops c_loop;
	
  int depth=0;
  int td_orientation=1;	//0=right, 1=left first
  for(iT=0; iT<nVtree; iT++) 	// initialize already_coded and avail_pred
		v_visited[v_cluster[nV0+iT]] &= AVAIL_RECON;
  if (partition_type==2 && codap_bdry_pred==0) {
		if (codap_right_bloop_idx != 0 || codap_left_bloop_idx != 1) {
      for (iT=0; iT<=codap_left_bloop_idx; iT++)
				v_visited[v_cluster[nV0+v_loop_sp[iT]]] |= ALREADY_CODED;
      if (codap_right_bloop_idx != 0)
				for (iT=codap_right_bloop_idx; iT<v_loop_sp.size(); iT++)
					v_visited[v_cluster[nV0+v_loop_sp[iT]]] |= ALREADY_CODED;
    }
  }
	
  int npe1 = 0;	//MJH inserted for checking if first edge is non-polygon edge
  for(iT=0,iRun=0;iT<nT;iT++)
	{
		// if(iT==0)
		//   root_triangle()
		// else
		//   triangle()
		
		// get marching bit
		switch(getTri(iT))
		{
		case sp_UNDET:
			{
				int marching =
					codec.decode(MPG_Global::td_marching_pattern[prev_marching_pattern]);
				if(marching != prev_marching_pattern)
					prev_marching_pattern = !prev_marching_pattern;
				
				if(marching == 0)
					setTri(iT,sp_RIGHT,1);
				else
					setTri(iT,sp_LEFT,1);
			}
			break;
			
		case sp_LEFT:
		case sp_RIGHT:
			// should never occur !!!
			break;
			
		case sp_LEAF:
			depth--;	
			break;
			
		case sp_BRANCH:
			if (partition_type==2)
				if (depth==0)
					td_orientation = codec.decode(MPG_Global::td_orientation_context);
				depth++;
				break;
				
		default:
			break;
		}
		
		// if neither root not triangulated get polygon_edge bit
		
		int polygon_edge = ((iT>0||(iT==0&&partition_type==2))&&triangulated==0)?	
			codec.decode(MPG_Global::td_polygon_edge[prev_polygon_edge]):1;
		setPolyBit(iT,(polygon_edge==1)?true:false);
		prev_polygon_edge = polygon_edge;
		
		if(polygon_edge==0) 
			nPE++;	
		
		// if necessary start a new face
		// iT==0||triangulated==1 => polygon_edge==1
		
		if(polygon_edge==1 || iT==0)	
		{
			
			if (iT==0 && partition_type==2 && polygon_edge==0) {
				// partitioned inside a polygon
				// send a flag so that it can be attached to the rest of the polygon in the previous partition
				poly_part = iL;	
			}
			
			loop.start(iR,iL);
			
			// add face to the face tree
			iF = f_father.size();
			f_father.push_back(iFp);
			if (polygon_edge==0)	
				npe1=1;
			
			// add corners to the corner tree
		    //MJH 062600 : swapped L and R
			cR = iC++; c_father[cR] = cRp; c_traversal.push_back(cR);
			cL = iC++; c_father[cL] = cLp; c_traversal.push_back(cL);
			
			c_loop.start(cR,cL);
		}
		else // GT @ Tue Nov 30 1999 to fix per face properties
			iF = iFp;
		
		// add corner to the corner tree
		cO = iC++; c_father[cO] = cL; c_traversal.push_back(cO);
		
		// corners
		//
		// if(polygon_edge==1) 
		//
		//         cO          c_father[cR] = cRp
		//        /   \        c_father[cL] = cLp
		//      /       \      c_father[cO] = cL
		//    cL         cR
		//    -------------
		//    cLP        cRp
		//
		// else 
		//
		//         cO          c_father[cO] = cL
		//        /   \
		//      /       \
		//    cL         cR
		//    .............
		//    cL         cR
		//
		
		// determine iO
		switch(getTri(iT))
		{
		case sp_LEAF:   iO  = iL+1;                                 break;
		case sp_LEFT:   iO  = (iR+nVloop-1)%nVloop;                 break;
		case sp_RIGHT:  iO  = iL+1;                                 break;
			
		case sp_BRANCH: iO = (td_orientation)?(iR+nVloop-run_length[iRun+1])%nVloop:
			(iL+run_length[iRun+1])%nVloop; break;
		}
		loop.insert(iO);
		
		// decode properties
		
		if(iT == 0)
		{
			// root_triangle()
			//
			//       O
			//     / F  \
			//   /   ^   \
			//  L ---|--- R
			//       Fp
			
			// decode root_coord()
			//
			// root_coord() 
			//   root_coord_sample()    <--- R
			//   coord_sample()         <--- L
			//   coord_sample()         <--- O
			//
			if(pCoord.getBinding() == PB_PER_VERTEX)
			{
				int* nR = &(qCoord[3*v_cluster[nV0+v_loop_sp[iR]]]);
				int* nL = &(qCoord[3*v_cluster[nV0+v_loop_sp[iL]]]);
				int* nO = &(qCoord[3*v_cluster[nV0+v_loop_sp[iO]]]);
				
				if (partition_type == 2) {
					if( !(v_visited[v_cluster[nV0+v_loop_sp[iR]]] & ALREADY_CODED))
					{
						decodeRootCoordSample(codec,bitsPerCoord,nR);
						if( !(v_visited[v_cluster[nV0+v_loop_sp[iL]]] & ALREADY_CODED))
						{
							decodeCoordSample(codec,bitsPerCoord,nL);
							pCoord.predictE(nR,nR,nR,nL);
							decodeCoordSample(codec,bitsPerCoord,nO);
							pCoord.predictE(nR,nR,nL,nO);
						}
					}
					else if( !(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED))
					{
						decodeRootCoordSample(codec,bitsPerCoord,nO);
					}
					else if( !(v_visited[v_cluster[nV0+v_loop_sp[iL]]] & ALREADY_CODED))
					{
						// this should not happen
						decodeRootCoordSample(codec,bitsPerCoord,nL);
#ifndef MPGELIB	//MJH 081800
						printf("Something's strange in root decoding of coord in partition type 2.\n");
#endif
					}
				}
				else {
					
					// AG 1/27/99 added the "v_visited" tests
					// for PER_VERTEX properties
					// of vertices of the root triangle
					
					if(!(v_visited[v_cluster[nV0+v_loop_sp[iR]]] & ALREADY_CODED))	
					{
						decodeRootCoordSample(codec,bitsPerCoord,nR);
					}
					if(!(v_visited[v_cluster[nV0+v_loop_sp[iL]]] & ALREADY_CODED))	
					{
						decodeCoordSample(codec,bitsPerCoord,nL);
						pCoord.predictE(nR,nR,nR,nL);
					}
					if(!(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED))	
					{
						decodeCoordSample(codec,bitsPerCoord,nO);
						pCoord.predictE(nR,nR,nL,nO);
					}
				}	//else 
			}
			
			// decode root_normal()
			//
			// root_normal() 
			//   if(normal_binding!=0)    <--- PB_NONE
			//     root_normal_sample()
			//     if(normal_binding!=2)  <--- PB_PER_FACE
			//       normal_sample()
			//       normal_sample()
			//
			if(pNormal.getBinding() == PB_PER_VERTEX)
			{
				int* nR = & qNormal[v_cluster[nV0+v_loop_sp[iR]]];
				int* nL = & qNormal[v_cluster[nV0+v_loop_sp[iL]]];
				int* nO = & qNormal[v_cluster[nV0+v_loop_sp[iO]]];
				
				if (partition_type == 2) {
					if( !(v_visited[v_cluster[nV0+v_loop_sp[iR]]] & ALREADY_CODED))
					{
						decodeRootNormalSample(codec,bitsPerNormal,nR);
						if( !(v_visited[v_cluster[nV0+v_loop_sp[iL]]] & ALREADY_CODED))
						{
							decodeNormalSample(codec,bitsPerNormal,nL);
							pNormal.predictE(nR,nR,nR,nL);
							decodeNormalSample(codec,bitsPerNormal,nO);
							pNormal.predictE(nR,nR,nL,nO);
						}
					}
					else if( !(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED))
					{
						decodeRootNormalSample(codec,bitsPerNormal,nO);
					}
					else if( !(v_visited[v_cluster[nV0+v_loop_sp[iL]]] & ALREADY_CODED))
					{
						// this should not happen
						decodeRootNormalSample(codec,bitsPerNormal,nL);
#ifndef MPGELIB	//MJH 081800
						printf("Something's strange in root decoding of normal in partition type 2.\n");
#endif
					}
				}
				else {
					
					if(!(v_visited[v_cluster[nV0+v_loop_sp[iR]]] & ALREADY_CODED))
						decodeRootNormalSample(codec,bitsPerNormal,nR);
					
					if(!(v_visited[v_cluster[nV0+v_loop_sp[iL]]] & ALREADY_CODED)){
						decodeNormalSample(codec,bitsPerNormal,nL);
						pNormal.predictE(nR,nR,nR,nL);
					}
					
					if(!(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED)){
						decodeNormalSample(codec,bitsPerNormal,nO);
						pNormal.predictE(nR,nR,nL,nO);
					}
				}
			}
			else if(pNormal.getBinding() == PB_PER_FACE)
			{
				int* nF = & qNormal[iF];
				decodeRootNormalSample(codec,bitsPerNormal,nF);
			}
			else if(pNormal.getBinding() == PB_PER_CORNER)
			{
			    //MJH 062600 : swapped L and R
				int* nR = & qNormal[cR];
				int* nL = & qNormal[cL];
				int* nO = & qNormal[cO];
				
				decodeRootNormalSample(codec,bitsPerNormal,nR);
				
				decodeNormalSample(codec,bitsPerNormal,nL);
				pNormal.predictE(c_father,qNormal,cL);
				
				decodeNormalSample(codec,bitsPerNormal,nO);
				pNormal.predictE(c_father,qNormal,cO);
			}
			
			// decode root_color()
			//
			// root_color() 
			//   if(color_binding!=0)    <--- PB_NONE
			//     root_color_sample()
			//     if(color_binding!=2)  <--- PB_PER_FACE
			//       color_sample()
			//       color_sample()
			//
			if(pColor.getBinding() == PB_PER_VERTEX)
			{
				int* nR = & qColor[3*v_cluster[nV0+v_loop_sp[iR]]];
				int* nL = & qColor[3*v_cluster[nV0+v_loop_sp[iL]]];
				int* nO = & qColor[3*v_cluster[nV0+v_loop_sp[iO]]];
				
				if (partition_type == 2) {
					if( !(v_visited[v_cluster[nV0+v_loop_sp[iR]]] & ALREADY_CODED))
					{
						decodeRootColorSample(codec,bitsPerColor,nR);
						if( !(v_visited[v_cluster[nV0+v_loop_sp[iL]]] & ALREADY_CODED))
						{
							decodeColorSample(codec,bitsPerColor,nL);
							pColor.predictE(nR,nR,nR,nL);
							decodeColorSample(codec,bitsPerColor,nO);
							pColor.predictE(nR,nR,nL,nO);
						}
					}
					else if( !(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED))
					{
						decodeRootColorSample(codec,bitsPerColor,nO);
					}
					else if( !(v_visited[v_cluster[nV0+v_loop_sp[iL]]] & ALREADY_CODED))
					{
						// this should not happen
						decodeRootColorSample(codec,bitsPerColor,nL);
#ifndef MPGELIB	//MJH 081800
						printf("Something's strange in root decoding of color in partition type 2.\n");
#endif
					}
				}
				else {
					
					if(!(v_visited[v_cluster[nV0+v_loop_sp[iR]]] & ALREADY_CODED))
						decodeRootColorSample(codec,bitsPerColor,nR);
					
					if(!(v_visited[v_cluster[nV0+v_loop_sp[iL]]] & ALREADY_CODED)){
						decodeColorSample(codec,bitsPerColor,nL);
						pColor.predictE(nR,nR,nR,nL);
					}
					
					if(!(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED)){
						decodeColorSample(codec,bitsPerColor,nO);
						pColor.predictE(nR,nR,nL,nO);
					}
				}
			}
			else if(pColor.getBinding() == PB_PER_FACE)
			{
				int* nF = & qColor[3*iF];
				decodeRootColorSample(codec,bitsPerColor,nF);
			}
			else if(pColor.getBinding() == PB_PER_CORNER)
			{
			    //MJH 062600 : swapped L and R
				int* nR = & qColor[3*cR];
				int* nL = & qColor[3*cL];
				int* nO = & qColor[3*cO];
				
				decodeRootColorSample(codec,bitsPerColor,nR);
				
				decodeColorSample(codec,bitsPerColor,nL);
				pColor.predictE(c_father,qColor,cL);
				
				decodeColorSample(codec,bitsPerColor,nO);
				pColor.predictE(c_father,qColor,cO);
			}
			
			// decode root_texCoord()
			//
			// root_texCoord() 
			//   if(texCoord_binding!=0)    <--- PB_NONE
			//     root_texCoord_sample()
			//     if(texCoord_binding!=2)  <--- PB_PER_FACE
			//       texCoord_sample()
			//       texCoord_sample()
			
			if(pTexCoord.getBinding() == PB_PER_VERTEX)
			{
				int* nR = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iR]]];
				int* nL = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iL]]];
				int* nO = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iO]]];
				
				if (partition_type == 2) {
					if( !(v_visited[v_cluster[nV0+v_loop_sp[iR]]] & ALREADY_CODED))
					{
						decodeRootTexCoordSample(codec,bitsPerTexCoord,nR);
						if( !(v_visited[v_cluster[nV0+v_loop_sp[iL]]] & ALREADY_CODED))
						{
							decodeTexCoordSample(codec,bitsPerTexCoord,nL);
							pTexCoord.predictE(nR,nR,nR,nL);
							decodeTexCoordSample(codec,bitsPerTexCoord,nO);
							pTexCoord.predictE(nR,nR,nL,nO);
						}
					}
					else if( !(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED))
					{
						decodeRootTexCoordSample(codec,bitsPerTexCoord,nO);
					}
					else if( !(v_visited[v_cluster[nV0+v_loop_sp[iL]]] & ALREADY_CODED))
					{
						// this should not happen
						decodeRootTexCoordSample(codec,bitsPerTexCoord,nL);
#ifndef MPGELIB	//MJH 081800
						printf("Something's strange in root decoding of texCoord in partition type 2.\n");
#endif
					}
				}
				else {
					
					if(!(v_visited[v_cluster[nV0+v_loop_sp[iR]]] & ALREADY_CODED))
						decodeRootTexCoordSample(codec,bitsPerTexCoord,nR);
					
					if(!(v_visited[v_cluster[nV0+v_loop_sp[iL]]] & ALREADY_CODED)){
						decodeTexCoordSample(codec,bitsPerTexCoord,nL);
						pTexCoord.predictE(nR,nR,nR,nL);
					}
					
					if(!(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED)){
						decodeTexCoordSample(codec,bitsPerTexCoord,nO);
						pTexCoord.predictE(nR,nR,nL,nO);
					}
				}
			}
			else if(pTexCoord.getBinding() == PB_PER_CORNER)
			{
			    //MJH 062600 : swapped L and R
				int* nR = & qTexCoord[2*cR];
				int* nL = & qTexCoord[2*cL];
				int* nO = & qTexCoord[2*cO];
				
				decodeRootTexCoordSample(codec,bitsPerTexCoord,nR);
				
				decodeTexCoordSample(codec,bitsPerTexCoord,nL);
				pTexCoord.predictE(c_father,qTexCoord,cL);
				
				decodeTexCoordSample(codec,bitsPerTexCoord,nO);
				pTexCoord.predictE(c_father,qTexCoord,cO);
			}
			
			if(has_per_vertex==true)
			{
				if ( !(v_visited[v_cluster[nV0+v_loop_sp[iR]]] & ALREADY_CODED))
					v_visited[v_cluster[nV0+v_loop_sp[iR]]] |= AVAIL_PRED+ALREADY_CODED;
				if ( !(v_visited[v_cluster[nV0+v_loop_sp[iL]]] & ALREADY_CODED))
					v_visited[v_cluster[nV0+v_loop_sp[iL]]] |= AVAIL_PRED+ALREADY_CODED;
				if ( !(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED))
					v_visited[v_cluster[nV0+v_loop_sp[iO]]] |= AVAIL_PRED+ALREADY_CODED;
			}
	}
	else
	{
		// triangle()
		//
		
		//       O
		//     / F  \
		//   /   ^   \
		//  L ---|--- R
		//   \   Fp  /
		//     \   /
		//       P
		
		// coord() {
		//   if(!visited)
		//     coord_sample()
		// }
		//
		
		// coord_sample() per vertex (always)
		if(pCoord.getBinding() == PB_PER_VERTEX)
		{
			int* nO = & qCoord[3*v_cluster[nV0+v_loop_sp[iO]]];
			
			if (partition_type == 2) {
				if( !(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED))
				{
					int iPavail = v_visited[v_cluster[nV0+v_loop_sp[iP]]] & AVAIL_PRED;
					int iRavail = v_visited[v_cluster[nV0+v_loop_sp[iR]]] & AVAIL_PRED;
					int iLavail = v_visited[v_cluster[nV0+v_loop_sp[iL]]] & AVAIL_PRED;
					
					if (iPavail + iRavail + iLavail == 0) {
						decodeRootCoordSample(codec,bitsPerCoord,nO);
					}
					else {
						decodeCoordSample(codec,bitsPerCoord,nO);
						if (iPavail && iRavail && iLavail) {
							int* nP = & qCoord[3*v_cluster[nV0+v_loop_sp[iP]]];
							int* nR = & qCoord[3*v_cluster[nV0+v_loop_sp[iR]]];
							int* nL = & qCoord[3*v_cluster[nV0+v_loop_sp[iL]]];
							pCoord.predictE(nP,nL,nR,nO);
						}
						else if (iPavail+iRavail+iLavail == AVAIL_PRED*2)
						{
							if (iPavail) {
								int* nRL;
								if (iRavail)
									nRL = & qCoord[3*v_cluster[nV0+v_loop_sp[iR]]];
								else
									nRL = & qCoord[3*v_cluster[nV0+v_loop_sp[iL]]];
								pCoord.predictE(nRL,nRL,nRL,nO);
							}
							else {
								int* nR = & qCoord[3*v_cluster[nV0+v_loop_sp[iR]]];
								int* nL = & qCoord[3*v_cluster[nV0+v_loop_sp[iL]]];
								pCoord.predictE2(nR,nL,nO);
							}
						}
						else {
							int* nPRL;
							if (iPavail)
								nPRL = & qCoord[3*v_cluster[nV0+v_loop_sp[iP]]];
							else if (iRavail)
								nPRL = & qCoord[3*v_cluster[nV0+v_loop_sp[iR]]];
							else
								nPRL = & qCoord[3*v_cluster[nV0+v_loop_sp[iL]]];
							pCoord.predictE(nPRL,nPRL,nPRL,nO);
						}
					}
				}
			}
			else {
				
				if(!(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED))
				{
					int* nP = & qCoord[3*v_cluster[nV0+v_loop_sp[iP]]];
					int* nR = & qCoord[3*v_cluster[nV0+v_loop_sp[iR]]];
					int* nL = & qCoord[3*v_cluster[nV0+v_loop_sp[iL]]];
					
					decodeCoordSample(codec,bitsPerCoord,nO);
					pCoord.predictE(nP,nL,nR,nO);
				}
			}
		}
		
		// normal_sample()
		//
		if(pNormal.getBinding() == PB_PER_VERTEX)
		{
			int* nO = & qNormal[v_cluster[nV0+v_loop_sp[iO]]];
			
			if (partition_type == 2) {
				if( !(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED))
				{
					int iPavail = v_visited[v_cluster[nV0+v_loop_sp[iP]]] & AVAIL_PRED;
					int iRavail = v_visited[v_cluster[nV0+v_loop_sp[iR]]] & AVAIL_PRED;
					int iLavail = v_visited[v_cluster[nV0+v_loop_sp[iL]]] & AVAIL_PRED;
					
					if (iPavail + iRavail + iLavail == 0) {
						decodeRootNormalSample(codec,bitsPerNormal,nO);
					}
					else {
						decodeNormalSample(codec,bitsPerNormal,nO);
						if (iPavail && iRavail && iLavail) {
							int* nP = & qNormal[v_cluster[nV0+v_loop_sp[iP]]];
							int* nR = & qNormal[v_cluster[nV0+v_loop_sp[iR]]];
							int* nL = & qNormal[v_cluster[nV0+v_loop_sp[iL]]];
							pNormal.predictE(nP,nL,nR,nO);
						}
						else if (iPavail+iRavail+iLavail == AVAIL_PRED*2)
						{
							if (iPavail) {
								int* nRL;
								if (iRavail)
									nRL = & qNormal[v_cluster[nV0+v_loop_sp[iR]]];
								else
									nRL = & qNormal[v_cluster[nV0+v_loop_sp[iL]]];
								pNormal.predictE(nRL,nRL,nRL,nO);
							}
							else {
								int* nR = & qNormal[v_cluster[nV0+v_loop_sp[iR]]];
								int* nL = & qNormal[v_cluster[nV0+v_loop_sp[iL]]];
								pNormal.predictE2(nR,nL,nO);
							}
						}
						else {
							int* nPRL;
							if (iPavail)
								nPRL = & qNormal[v_cluster[nV0+v_loop_sp[iP]]];
							else if (iRavail)
								nPRL = & qNormal[v_cluster[nV0+v_loop_sp[iR]]];
							else
								nPRL = & qNormal[v_cluster[nV0+v_loop_sp[iL]]];
							pNormal.predictE(nPRL,nPRL,nPRL,nO);
						}
					}
				}
			}
			else {
				
				if(!(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED))
				{
					int* nP = & qNormal[v_cluster[nV0+v_loop_sp[iP]]];
					int* nR = & qNormal[v_cluster[nV0+v_loop_sp[iR]]];
					int* nL = & qNormal[v_cluster[nV0+v_loop_sp[iL]]];
					
					decodeNormalSample(codec,bitsPerNormal,nO);
					pNormal.predictE(nP,nL,nR,nO);
				}
			}
		}
		else if(pNormal.getBinding() == PB_PER_FACE)
		{
			if(polygon_edge==1)
			{
				int* nF  = & qNormal[iF];
				decodeNormalSample(codec,bitsPerNormal,nF);
				pNormal.predictE(f_father,qNormal,iF);
			}
		}
		else if(pNormal.getBinding() == PB_PER_CORNER)
		{
			if(polygon_edge==1)
			{
			    //MJH 062600 : swapped L and R
				int* nR = & qNormal[cR];
				decodeNormalSample(codec,bitsPerNormal,nR);
				pNormal.predictE(c_father,qNormal,cR);
				
				int* nL = & qNormal[cL];
				decodeNormalSample(codec,bitsPerNormal,nL);
				pNormal.predictE(c_father,qNormal,cL);
			}
			
			int* nO = & qNormal[cO];
			decodeNormalSample(codec,bitsPerNormal,nO);
			pNormal.predictE(c_father,qNormal,cO);
		}
		
		// color_sample()
		//
		if(pColor.getBinding() == PB_PER_VERTEX)
		{
			int* nO = & qColor[3*v_cluster[nV0+v_loop_sp[iO]]];
			
			if (partition_type == 2) {
				if( !(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED))
				{
					int iPavail = v_visited[v_cluster[nV0+v_loop_sp[iP]]] & AVAIL_PRED;
					int iRavail = v_visited[v_cluster[nV0+v_loop_sp[iR]]] & AVAIL_PRED;
					int iLavail = v_visited[v_cluster[nV0+v_loop_sp[iL]]] & AVAIL_PRED;
					
					if (iPavail + iRavail + iLavail == 0) {
						decodeRootColorSample(codec,bitsPerColor,nO);
					}
					else {
						decodeColorSample(codec,bitsPerColor,nO);
						if (iPavail && iRavail && iLavail) {
							int* nP = & qColor[3*v_cluster[nV0+v_loop_sp[iP]]];
							int* nR = & qColor[3*v_cluster[nV0+v_loop_sp[iR]]];
							int* nL = & qColor[3*v_cluster[nV0+v_loop_sp[iL]]];
							pColor.predictE(nP,nL,nR,nO);
						}
						else if (iPavail+iRavail+iLavail == AVAIL_PRED*2)
						{
							if (iPavail) {
								int* nRL;
								if (iRavail)
									nRL = & qColor[3*v_cluster[nV0+v_loop_sp[iR]]];
								else
									nRL = & qColor[3*v_cluster[nV0+v_loop_sp[iL]]];
								pColor.predictE(nRL,nRL,nRL,nO);
							}
							else {
								int* nR = & qColor[3*v_cluster[nV0+v_loop_sp[iR]]];
								int* nL = & qColor[3*v_cluster[nV0+v_loop_sp[iL]]];
								pColor.predictE2(nR,nL,nO);
							}
						}
						else {
							int* nPRL;
							if (iPavail)
								nPRL = & qColor[3*v_cluster[nV0+v_loop_sp[iP]]];
							else if (iRavail)
								nPRL = & qColor[3*v_cluster[nV0+v_loop_sp[iR]]];
							else
								nPRL = & qColor[3*v_cluster[nV0+v_loop_sp[iL]]];
							pColor.predictE(nPRL,nPRL,nPRL,nO);
						}
					}
				}
			}
			else {
				
				if(!(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED))
				{
					int* nP = & qColor[3*v_cluster[nV0+v_loop_sp[iP]]];
					int* nR = & qColor[3*v_cluster[nV0+v_loop_sp[iR]]];
					int* nL = & qColor[3*v_cluster[nV0+v_loop_sp[iL]]];
					decodeColorSample(codec,bitsPerColor,nO);
					pColor.predictE(nP,nL,nR,nO);
				}
			}
		}
		else if(pColor.getBinding() == PB_PER_FACE)
		{
			if(polygon_edge==1)
			{
				int* nF  = & qColor[3*iF];
				decodeColorSample(codec,bitsPerColor,nF);
				pColor.predictE(f_father,qColor,iF);
			}
		}
		else if(pColor.getBinding() == PB_PER_CORNER)
		{
			if(polygon_edge==1)
			{
			    //MJH 062600 : swapped L and R
				int* nR = & qColor[3*cR];
				decodeColorSample(codec,bitsPerColor,nR);
				pColor.predictE(c_father,qColor,cR);
				
				int* nL = & qColor[3*cL];
				decodeColorSample(codec,bitsPerColor,nL);
				pColor.predictE(c_father,qColor,cL);
			}
			
			int* nO = & qColor[3*cO];
			decodeColorSample(codec,bitsPerColor,nO);
			pColor.predictE(c_father,qColor,cO);
		}
		
		// texCoord_sample()
		//
		if(pTexCoord.getBinding() == PB_PER_VERTEX)
		{
			int* nO = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iO]]];
			
			if (partition_type == 2) {
				if( !(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED))
				{
					int iPavail = v_visited[v_cluster[nV0+v_loop_sp[iP]]] & AVAIL_PRED;
					int iRavail = v_visited[v_cluster[nV0+v_loop_sp[iR]]] & AVAIL_PRED;
					int iLavail = v_visited[v_cluster[nV0+v_loop_sp[iL]]] & AVAIL_PRED;
					
					if (iPavail + iRavail + iLavail == 0) {
						decodeRootTexCoordSample(codec,bitsPerTexCoord,nO);
					}
					else {
						decodeTexCoordSample(codec,bitsPerTexCoord,nO);
						if (iPavail && iRavail && iLavail) {
							int* nP = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iP]]];
							int* nR = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iR]]];
							int* nL = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iL]]];
							pTexCoord.predictE(nP,nL,nR,nO);
						}
						else if (iPavail+iRavail+iLavail == AVAIL_PRED*2)
						{
							if (iPavail) {
								int* nRL;
								if (iRavail)
									nRL = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iR]]];
								else
									nRL = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iL]]];
								pTexCoord.predictE(nRL,nRL,nRL,nO);
							}
							else {
								int* nR = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iR]]];
								int* nL = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iL]]];
								pTexCoord.predictE2(nR,nL,nO);
							}
						}
						else {
							int* nPRL;
							if (iPavail)
								nPRL = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iP]]];
							else if (iRavail)
								nPRL = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iR]]];
							else
								nPRL = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iL]]];
							pTexCoord.predictE(nPRL,nPRL,nPRL,nO);
						}
					}
				}
			}
			else {
				
				if(!(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED))
				{
					int* nP = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iP]]];
					int* nR = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iR]]];
					int* nL = & qTexCoord[2*v_cluster[nV0+v_loop_sp[iL]]];
					decodeTexCoordSample(codec,bitsPerTexCoord,nO);
					pTexCoord.predictE(nP,nL,nR,nO);
				}
			}
		}
		else if(pTexCoord.getBinding() == PB_PER_CORNER)
		{
			if(polygon_edge==1)
			{
			    //MJH 062600 : swapped L and R
				int* nR = & qTexCoord[2*cR];
				decodeTexCoordSample(codec,bitsPerTexCoord,nR);
				pTexCoord.predictE(c_father,qTexCoord,cR);
				
				int* nL = & qTexCoord[2*cL];
				decodeTexCoordSample(codec,bitsPerTexCoord,nL);
				pTexCoord.predictE(c_father,qTexCoord,cL);
			}
			
			int* nO = & qTexCoord[2*cO];
			decodeTexCoordSample(codec,bitsPerTexCoord,nO);
			pTexCoord.predictE(c_father,qTexCoord,cO);
		}
		
		if(has_per_vertex==true)
			if ( !(v_visited[v_cluster[nV0+v_loop_sp[iO]]] & ALREADY_CODED))
		      v_visited[v_cluster[nV0+v_loop_sp[iO]]] |= AVAIL_PRED+ALREADY_CODED;
	}
	
	// advance to next triangle
	switch(getTri(iT))
	{
		
	case sp_LEAF:
		
		// pop an edge from traversal stack
		//       ^
		//  L ---|--- R
		//   \       /
		//     \   /
		//       P
		
		loop.pop(iR,iL);
		c_loop.insert(cO);
		c_loop.pop(cR,cL);
		
		if (iT < nT-1) {	//MJH 20010221 : if statement added (to fix the crash problem in NT and W2k)
			cRp  = stk.back(); 
			stk.pop_back();
			cLp  = stk.back(); 
			stk.pop_back();
			iP   = stk.back(); 
			stk.pop_back();
			iFp  = stk.back(); 
			stk.pop_back();
		}
		
		iRun++;		// end of run
		break;
		
	case sp_LEFT:
		
		// turn left
		//
		//       O = R-1
		//     /   \
		//   /   ^   \
		//  L ---|--- R
		//   \       /
		//     \   /
		//       P
		
		iP  = iR;
		iR  = iO;
		iFp = iF;
		c_loop.insert(cO);
		cRp = cO;
		cLp = cL;
		cR  = cRp;
		cL  = cLp;
		break;
		
	case sp_RIGHT:
		
		// turn right
		//
		//       O = L+1
		//     /   \
		//   /   ^   \
		//  L ---|--- R
		//   \       /
		//     \   /
		//       P
		
		loop.advance();
		iP  = iL;
		iL  = iO;
		iFp = iF;
		c_loop.insert(cO);
		c_loop.advance();
		cRp = cR;
		cLp = cO;
		cR  = cRp;
		cL  = cLp;
		break;
		
	case sp_BRANCH:
		
		//       O
		//     /   \
		//   /   ^   \
		//  L ---|--- R
		//   \       /
		//     \   /
		//       P
		
		// push left edge onto traversal stack if td_orientation==1 (default)
		//       ^
		//  L ---|--- O
		//   \       /
		//     \   /
		//       R
		
		// push right edge onto traversal stack if td_orientation==0
		//       ^
		//  O ---|--- R
		//   \       /
		//     \   /
		//       L
		
		stk.push_back(iF);
		
		if (td_orientation) {
			loop.pushCurr();
			stk.push_back(iR);
			loop.advance();	
		}
		else {
			loop.pushNext();
			stk.push_back(iL);
		}
		
		// advance to right edge if td_orientation==1 (default)
		//       ^
		//  O ---|--- R
		//   \       /
		//     \   /
		//       L
		
		// advance to left edge if td_orientation==0
		//       ^
		//  L ---|--- O
		//   \       /
		//     \   /
		//       R
		
		iFp = iF;
		
		c_loop.insert(cO);
		
		if (td_orientation) {
			c_loop.pushCurr();
			c_loop.advance();
			
			iP = iL;
			iL = iO;
			
			stk.push_back(cL);
			stk.push_back(cO);
			
			cRp = cR;
			cLp = cO;
		}
		else {	
			c_loop.pushNext();
			
			iP = iR;
			iR = iO;
			
			stk.push_back(cO);
			stk.push_back(cR);
			
			cRp = cO;
			cLp = cL;
			td_orientation=1;
		}
		
		cR  = cRp;
		cL  = cLp;
		iRun++;		// end of run
		break;
	}
    }
		
		// now that we know the total number of internal polygon edges
		// we can adjust number of faces and corners
		nF -=   nPE-npe1;
		nC -= 2*(nPE-npe1);
		
		nF += SnF;	
		nC += SnC;	

		c_father.erase(c_father.begin()+nC,c_father.end());	//MJH 000111 : added

		// and the sizes of property arrays
		if(nPE>0)
    {
      if(pNormal.getBinding()==PB_PER_FACE)
				qNormal.erase(qNormal.begin()+nF,qNormal.end());
      else if(pNormal.getBinding()==PB_PER_CORNER)
				qNormal.erase(qNormal.begin()+nC,qNormal.end());
			
      if(pColor.getBinding()==PB_PER_FACE)
				qColor.erase(qColor.begin()+3*nF,qColor.end());
      else if(pColor.getBinding()==PB_PER_CORNER)
				qColor.erase(qColor.begin()+3*nC,qColor.end());
			
      if(pTexCoord.getBinding()==PB_PER_CORNER)
				qTexCoord.erase(qTexCoord.begin()+2*nC,qTexCoord.end());
    }
		
		// if per_corner properties, we need to permute corner properties
		if(pNormal.getBinding()==PB_PER_CORNER ||
			pColor.getBinding()==PB_PER_CORNER ||
			pTexCoord.getBinding()==PB_PER_CORNER)
    {
      vector<int> face;
      vector<int> indx;
      c_loop.getAllLoops(face,indx);
      
      vector<int> permQnormal;
      vector<int> permQcolor;
      vector<int> permQtexCoord;
			
      int h,i,j,k;
      for(i = 0; i < face.size()-1; i++)
			{
				int  nf = face[i+1]-face[i];
				int*  f = indx.begin()+face[i];
				
				for(j = 0; j < nf; j++)
				{
					k = f[j];
					
					if(pNormal.getBinding()==PB_PER_CORNER)
					{
						permQnormal.push_back(qNormal[k]);
					}
					if(pColor.getBinding()==PB_PER_CORNER)
					{
						for(h=0;h<3;h++)
							permQcolor.push_back(qColor[3*k+h]);
					}
					if(pTexCoord.getBinding()==PB_PER_CORNER)
					{
						for(h=0;h<2;h++)
							permQtexCoord.push_back(qTexCoord[2*k+h]);
					}
				}
				
			}
      
      if(pNormal.getBinding()==PB_PER_CORNER)
			{
				qNormal.erase(qNormal.end()-permQnormal.size(), qNormal.end());
				qNormal.insert(qNormal.end(), permQnormal.begin(), permQnormal.end());
			}
      if(pColor.getBinding()==PB_PER_CORNER)
			{
				qColor.erase(qColor.end()-permQcolor.size(), qColor.end());
				qColor.insert(qColor.end(), permQcolor.begin(), permQcolor.end());
			}
      if(pTexCoord.getBinding()==PB_PER_CORNER)
			{
				qTexCoord.erase(qTexCoord.end()-permQtexCoord.size(), qTexCoord.end());
				qTexCoord.insert(qTexCoord.end(), permQtexCoord.begin(), permQtexCoord.end());
			}
		}
		
		//update available for reconstruction
		for(iT=0; iT<nVtree; iT++)	
			if (v_visited[v_cluster[nV0+iT]] & AVAIL_PRED)
				v_visited[v_cluster[nV0+iT]] |= AVAIL_RECON;
			
			// convert connectivity
			vector<int> face;
			vector<int> indx;
			loop.getAllLoops(face,indx);
			
			int i, j;
			for(i = 0; i < face.size()-1; i++)
			{
				int  nf = face[i+1]-face[i];
				int*  f = indx.begin()+face[i];
				
				if (codap_bdry_pred) 
					for(j = 0; j < nf; j++)
					{
						coordIndex.push_back(v_cluster[nV0+v_loop_sp[f[j]]]);
					}
					else 
						for(j = 0; j < nf; j++)
						{
							if (v_visited[v_cluster[nV0+v_loop_sp[f[j]]]] & AVAIL_RECON)
								coordIndex.push_back(v_cluster[nV0+v_loop_sp[f[j]]]);
							else
								coordIndex.push_back(-2);	
						} 
						coordIndex.push_back(-1);
			}
			
			return true;
} 
// end of decodeTriangleData

