//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------


#pragma warning(disable:4786)

#include <stdio.h>
#include <stdlib.h>
#include "MPG_Graph.h"
#include "MPG_Stitcher.h"
#include "MPG_Partition.h"
#include "MPG_Stitches.h"

#ifndef MPGELIB
extern FILE* logfp;
#endif // MPGELIB

//MJH : not needed #define CHECK_ENCODER_INPUT_ARGUMENTS

#undef  MAX
#define MAX(a,b)        (((a)>(b))?(a):(b)) 
#undef  MIN
#define MIN(a,b)        (((a)<(b))?(a):(b)) 

/*
  this file contains a method to take a v_cluster clustering map
  (vector) and represents it with a series of stitching commands, one
  per vertex

  another method starts with a series of stitching commands and
  produces a v_cluster clustering map

 */

    
//---------------------------------------------------------------------------
//
//   longestStitchStartingAt
//
//---------------------------------------------------------------------------

// what is the longest stitch starting at i ? --> there can be several
// input : starting vertex i, multiplicity
//         other vertices with the same label
//         v_father directed graph
// output: --> other vertex j to stitch to
//         --> length of the stitch

static bool longestStitchStartingAt(int start_vertex, 
				    const vector <int> & v_multiplicity,
				    const vector <int> & v,
				    const vector <int> & v_label_start,
				    const vector <int> & v_label,
				    const vector <int> & v_father,
				    vector<int>& stitch_to, 
				    int  &stitch_length,
				    bool &stitch_backwards)
{
  bool success = true;

  int k, l;
  int i = start_vertex;

  int new_multiplicity;

  // retrieve all vertices that have the same label as i
				    
  const int *v_same_label = &v[v_label_start[v_label[i]]];

  int multiplicity  = v_multiplicity[v_label[i]];

  stitch_length = 0; // initialize the length of the stitch

  stitch_to.erase(stitch_to.begin(),stitch_to.end()); // safer

  // initialize the vertices we can stitch to with length zero

  for (k=0;k<multiplicity; k++)
    if (v_same_label[k] != start_vertex)
      stitch_to.push_back(v_same_label[k]);

  // to discover backward stitches, create a vector containing all the
  // labels of start_vertex and its fathers

  vector <int> s_labels;

  multiplicity --;

  // gradually increase the length
  
  while (multiplicity > 0)
    {
      vector <int> stitch_longer;

      new_multiplicity = multiplicity;

      s_labels.push_back(v_label[i]);
     
      int i_father = v_father[i];

      /* note that this algorithm can be later improved by
	 checking whether a longest-stitch-pair (i,j,l1) was already stored
	 for a given i,j, pair, in which case we can add length to l1
	 and store the new value */

      for (k=0; k < multiplicity; k++)
	{
	  int j0 = stitch_to[k];
	  int j  = j0;

	  // follow the path up to i 
	  for (l = 0; l < stitch_length; l++)
	    j = v_father[j];

	  int j_father = v_father[j];

	  // first problem: if both i and j have the same father 
	  // there is no need to stitch further and j
	  // should be removed from the stitch

	  if (j_father == i_father) new_multiplicity--;
	  else
	    {
	      // second problem: if both i and j are entering a self-loop
	      // there is no reason to stitch further.
	      if (i == i_father && j == j_father)
		new_multiplicity--;
	      
	      else // now the main point is: do the fathers have the same label
		{
		  // (remark: in all other cases, j is removed from the list)
		  if (v_label[j_father] == v_label[i_father])
		    stitch_longer.push_back(j0);
		  else                                    
		    new_multiplicity--;
		}
	    }
	}
     
      if (new_multiplicity > 0) 
	{
	  stitch_length++;
	  // overwrite stitch_to with stitch_longer
	  stitch_to = stitch_longer;

	  i         = i_father;
	} 
      multiplicity = new_multiplicity;
      
    }

  // stitch_length and stitch_to are properly assigned,
  // so we could just return them now

  // except that we wish to discover backward stitches
  // that are potentially longer
    
  i = v_father[i];
  bool i_self_loop = (v_father[i] == i);


  {
    // try to increase the length using backward stitches

    // at this point, the vertex i is the father of the tip vertex of the longest
    // stitch 

    vector <int> stitch_longer;

    new_multiplicity = 0;

    v_same_label =  &v[v_label_start[v_label[i]]];

    // the first time we do this, we must check whether the sequence of labels 
    // matches the stored sequence

    for (l = 0; l < v_multiplicity[v_label[i]]; l++)
      if (v_same_label[l] != i && v_same_label[l] != start_vertex)
      {
	int j        = v_same_label[l];
	int j_father = v_father[j];

	// we cannot have both i and j enter into self-loops

	if (!i_self_loop || (j!=j_father))
	  {
	    // take in turn each vertex with the same label 
	    // if the sequence of labels of its fathers matches the
	    // stored sequence we have found one stitch of length
	    // (stitch_length+1)

	    for (k= stitch_length; k >=0; k--, j_father = v_father[j_father])
	      {
		if (v_label[j_father] != s_labels[k] ||
		    j_father == start_vertex) break;
	      }
	    
	    if (k == -1) // the sequence of labels matched
	      {
		stitch_longer.push_back(j);
		new_multiplicity++;
	      }
	  }
      }
    if (new_multiplicity>0)
      {
	stitch_to = stitch_longer;
	stitch_length++;
	
	i         = v_father[i];
      }
    
  }
  // indicate that the stitch is backwards
  stitch_backwards = (new_multiplicity > 0);
  


  // continue increasing the length of potential backward stitches

  while (new_multiplicity> 0)
    {
      
      vector <int> stitch_longer;

      new_multiplicity = 0;

      i_self_loop = (v_father[i] == i);

      v_same_label =  &v[v_label_start[v_label[i]]];

      for (l = 0; l < v_multiplicity[v_label[i]]; l++)
	if (v_same_label[l] != i && v_same_label[l] != start_vertex)
	  {
	    // the subsequent times we do this, we must only check
	    // that the fathers of these vertices belong to the stitch_to list
	    // (we just made sure that these match the sequence of labels)

	    int j        = v_same_label[l];
	    int j_father = v_father[j];

	    if (!i_self_loop || (j!=j_father))
	      {
		for (k=0; k< stitch_to.size(); k++)
		  {
		    if (j_father == stitch_to[k])
		      {
			stitch_longer.push_back(j);
			new_multiplicity++;
			// of course, no need to explore
			// the rest of the stitch_to list
			break;
		      }
		  }
	      }
	  }

      if (new_multiplicity>0)
	{
	  stitch_to = stitch_longer;
	  stitch_length++;
	
	  i         = v_father[i];
	}
    }


  return success;
}

//---------------------------------------------------------------------------
//
//   ConnectedStitchingCommands
//
//---------------------------------------------------------------------------

// generate all the stitching commands corresponding to a given vertex label

static int ConnectedStitchingCommands(int nV,
				     vector<int>&  v,
				     vector<int>&  stitch_to, 
				     vector<int>&  s_length, 
				     vector<bool>& s_backwards,
				     vector<int>&  v_refCounts, 
				     MPG_Stitches& s_command)
{
  int numCommands = 0;
  int i;

  // now sort the vertices involved 
  {
    // bubble sort seems ok here: nV should be low: couple of vertices or so

    int j;

    for (i = nV-1; --i>=0;)
      {
	bool flipped = false;
	for (j=0;j<=i;j++)
	  {
	    int a = v[j];
	    int b = v[j+1];
	    if (a > b)
	      {
		v[j+1] = a;
		v[j]   = b;
		flipped = true;
	      }
	    if (!flipped) break;
	  }
      }
  }

  // initialize the first PUSH length

  s_length[v[0]] = s_length[v[1]];

  for (i = 0; i < nV; i++)
    {
      int j = stitch_to[v[i]];
      if (j != -1 && j != v[i])
	{
	  // AG 1/27/99 fixed bug here s_length[j] and not s_length[v[j]]
	  int differential_length = s_length[v[i]] - s_length[j];
 
	  // is this a pop or a get
	  if (v_refCounts[j]    == 1)

	    {
	      s_command.set(v[i], MPG_Stitches::ST_POP, j,
			    differential_length, 
			    (v_refCounts[v[i]] > 0), // push bit
			    s_backwards[v[i]]); // fix this later
	    }
	  else
	    s_command.set(v[i], MPG_Stitches::ST_GET, j,
			  differential_length, 
			  (v_refCounts[v[i]] > 0), // push bit
			  s_backwards[v[i]]);  // fix this later
	  
	  
	  v_refCounts[j]--;
	  
	  numCommands++;
	}
      else 
	if (v_refCounts[v[i]] > 0) // this is a simple push
	  {
	    s_command.set(v[i], MPG_Stitches::ST_PUSH, -1, s_length[v[i]], 
			  true, // push bit
			  s_backwards[v[i]]);
	    numCommands++;
	  }
    }
    
  return numCommands;
}
	
//---------------------------------------------------------------------------
//
//   path_compress
//
//---------------------------------------------------------------------------


static int path_compress(int i, int *father)
{
  static int Ci, Fi;

  /*--- path traversal ---*/
  for(Ci=i;(Fi=father[Ci])!=Ci;Ci=Fi);

  /*--- path compression ---*/
  for(;(Fi=father[i])!=Ci;i=Fi) father[i] = Ci;

  return(Ci);
}
  
//---------------------------------------------------------------------------
//
//   MPG_Stitcher::represent
//
//---------------------------------------------------------------------------

bool MPG_Stitcher::represent(// inputs
			     const vector<int>&        nVcc,
			     const vector<int>&        v_father,
			           vector<int>&        v_label0,
			     MPG_Stitches&             s_command
			     // output
			     //MPG_OutputBitStream& ostrm
			     )
{
  bool success = false;

  int nCC = nVcc.size();

  int nV  = v_father.size();

  s_command.reset(nV);

  int i, j;

#ifdef CHECK_ENCODER_INPUT_ARGUMENTS
  // check input arguments
  {
    
    // v_father is a map [0,...,nV) -> [0,...,nV) such that
    // (i,v_father[i]) is an edge of MPG_Mesh 
    // AG: but mesh is not an input of this function
   
    // v_label is any map [0,...,nV) -> [0,...,nV) to be encoded
    for(i=0;i<nV;i++)
      {
	j = v_label0[i];
	if(j<0 || j>=nV) 
	  {
#ifndef MPGELIB
	    fprintf(logfp,"v_label[%d] = %d is not between 0 and %d\n",
		    i, v_label0[i], nV);
#endif // MPGELIB
	    return false;
	  }
      }
  }
#endif // CHECK_ENCODER_INPUT_ARGUMENTS
  
  vector <int> v_label((size_t)nV);

  v_label = v_label0; /* make a local copy because we are doing a few things
			 on it */

  success = true;
  
  vector <int>  v_stitch((size_t) nV, -1); // temporarily record stitches 
  vector <int>  s_length((size_t) nV, -1);  // before producing actual stitch commands
  // compute the reference counts for each vertex
  // this will tell us when it is suitable to pop vertices
  vector <int>  v_refCounts((size_t) nV, 0); 
  // record the stitch direction
  vector <bool> s_backwards((size_t) nV, false);
  
  
  { MPG_Partition v_merge((size_t) nV); // keep track of which
    // vertices are merged when applying successive stitches
    
    // array where vertices with the same label are consecutive
    vector <int> v_sorted_label((size_t) nV, 0); 
    
    // where to find vertices of a given label in v_sorted_label
    vector <int> v_label_start((size_t) nV, 0);  

    {
      vector <int> v_multiplicity ((size_t) nV, 0);

      // 1) initialize: do connected components on the labels
      
      // AG: I have tried to use MPG_Vector::indexLabel but it does not
      // operate properly, so here is minimal connected components code:
 
      for(i=0; i<nV; i++) path_compress (i, v_label.begin());
    
      // compute the label multiplicities
      for(i=0; i<nV; i++) v_multiplicity[v_label[i]]++;

      // then sort the labels
      for(v_label_start[0]=0,i=0;i<nV-1;i++) { 
	v_label_start[i+1] = v_label_start[i]+v_multiplicity[i]; 
	v_multiplicity[i]  = 0; 
      }
      v_multiplicity[nV-1] = 0;
    
      for(i=0; i < nV; i++) {
	j = v_label[i];
	v_sorted_label[v_label_start[j]+v_multiplicity[j]] = i;
	v_multiplicity[j] += 1;
      }
    }
    // fresh v_multiplicity array

    vector <int> v_multiplicity ((size_t) nV, 0);

    //define a graph to avoid visiting vertex pairs twice
      
    MPG_Graph pot_stitches(nV, true); 
    // AG: the order of vertex pairs matters, so we need a directed graph

    // define a priority queue

    multiset< MPG_PotentialStitch, less<MPG_PotentialStitch> >  pot_stitch_q;
    
    int nV0 = 0, nV1 = 0, c;
    int nVjoined = 0;
    int nVcls = 0; // number of clusters
    
    for (nV0 = 0, c = 0; c < nCC; c++, nV0 = nV1) { // loop on the
      // first through last components

      // the number of vertices depends on how many components we have
      // incorporated

      nV1       = nVcc[c];

      // we must have v_label[i] <= i, but we checked for that before 
    
      // re-compute the multiplicity of each vertex
      for(i=nV0; i < nV1; i++) v_multiplicity[v_label[i]]++;
      
      ///additional clustered vertices
      for(i=nV0; i < nV1; i++) if (v_multiplicity[i] > 0)  nVcls++;
      
      // 2) discover the stitches
      
      /* For each vertex i with multiplicity > 1, identify a longest
	 stitch starting at that vertex and add it to a priority queue */
	
      /* how many stitches are we going to put in it?
	 for each label, we must count the vertices: n_l
	 and count how many pairs of vertices can be done
	 n_l * n_l-1 / 2 */
	
      // loop on the vertices 

      for (i = nV0; i < nV1; i ++) 
	// discover new stitches only for the current component
	{
	  int the_multiplicity = v_multiplicity[v_label[i]];

	  // each time the multiplicity is larger than 1
	  if (the_multiplicity > 1)
	    {
	      vector <int> stitch_to;
	     
	      int length;

	      // whether the stitch is a forward or backward stitch
	      bool backwards = false; 
	  
	      // find the longest stitch starting at i 
	  
	      longestStitchStartingAt(i, 
				      v_multiplicity, 
				      v_sorted_label,
				      v_label_start,
				      v_label,   
				      v_father,  
				      stitch_to, 
				      length,
				      backwards);
	    
	      // and add all the stitches in stitch_to to the heap
	      // indexed by its length
	  
	      for (j=0; j < stitch_to.size(); j++)
		{
		  // was this edge already inserted, this is only possible if
		  // stitch_to[j] was visited before i
		  int k = stitch_to[j];

		  if (pot_stitches.getEdge(k, i) == (MPG_Graph::pEdge)0)
		    {
		      MPG_PotentialStitch s(i, k, -length, backwards);
		      pot_stitches.insertEdge(i, k, length);
		      
		      pot_stitch_q.insert(s);
		    }
		}
	    }
	}
      
      // 3) remove stitches from the queue one at a time in order of
      // decreasing length

      // the target number of vertices to join is exactly nV1 - nVcls

	  //MJH 010911 : changed for removal of stl3.13
//      set< MPG_PotentialStitch, less<MPG_PotentialStitch> >::iterator top;
      multiset< MPG_PotentialStitch, less<MPG_PotentialStitch> >::iterator top;

      while (nVcls+nVjoined < nV1 && pot_stitch_q.empty() == false)
	{
	  top            = pot_stitch_q.begin();
	  i              =   (*top).origin();
	  j              =   (*top).dest();
	  int length     = - (*top).length();
	  bool backwards =   (*top).direction();
	  pot_stitch_q.erase(top);

	  int stitchTo = j;
	  if (backwards)
	    for (int k = 0;k<length;k++)
	      stitchTo = v_father[stitchTo];
	
	  // make sure the stitch was not done before
	  if (v_merge.find(i) != v_merge.find(stitchTo)) {
	    int max_endpt = MAX(i,j);
	    // AG, 2/14/99 and that the stitch can be encoded
	    // meaning the MAX(i,j) is not previously stitched
	    // this is an important design change: the longest stitch
	    // may be impossible
	    if (v_stitch[max_endpt] != -1) {
	      // 3/1/99 reinsert a stitch shorter by one in the queue
	      if (length > 0) {
		int new_length = length -1;
		int first_vertex  = i;
		int second_vertex = j;
		if (backwards) {
		  
		  if (i == max_endpt) first_vertex = v_father[i];
		  
		  else  // if (j == max_endpt)
		    second_vertex = v_father[j];
		  
		}
		else { //  forward stitch
		  first_vertex = v_father[i];
		  second_vertex = v_father[j];
		}
		if (pot_stitches.getEdge(first_vertex, second_vertex) ==
		    (MPG_Graph::pEdge)0) {
		  {
		    MPG_PotentialStitch s(first_vertex, second_vertex,
					  -new_length, backwards);
		    pot_stitch_q.insert(s);
		  }
		}
	      }
	    }
	     
	    else{ // process the stitch
	      // record the stitch in a temporary structure before encoding

	      if (i < j)
		{
		  v_stitch[j]    = i; 
		  s_length[j]    = length;
		  s_backwards[j] = backwards; 
		  v_refCounts[i]++;
		}
	      else // j < i: they can't be equal
		{
		  v_stitch[i]    = j; 
		  s_length[i]    = length;
		  s_backwards[i] = backwards; 
		  v_refCounts[j]++;
		}

	      // mark the vertices composing the stitch as merged
	      
	      // recall that "length" is the number of edges composing
	      // the stitch the number of vertices is length+1
	      if (backwards)
		{
		  int k;

		  vector <int> j_father;
		
		  for (k = 0; k < length+1; k ++, j = v_father[j])
		    j_father.push_back(j);
		  for (k = 0; k < length+1; k ++, i = v_father[i]){
		    int v = j_father[length-k];
		    if (v_merge.find(i) != v_merge.find(v)){
		      v_merge.join(i,v);
		      nVjoined++;
		    }
		  }
		}
	      else
		for (int k=0;k<length+1;k++,i=v_father[i],j=v_father[j]){
		  if (v_merge.find(i) != v_merge.find(j)){
		    v_merge.join(i,j); 
		    nVjoined++;
		  }
		}
	    }
	  }
	}
      // we can dispose of the priority queue, the graph and the partition
      // but we need to keep the v_stitch, s_length, etc. vectors 
    }
  }
 
  /*
    4) produce the stitching commands:
   */

  vector <int> v_mapsTo((size_t)nV,0);
  vector <int> v_stitched;  
          
  int nVstitched = 0;

  v_stitched.erase(v_stitched.begin(),v_stitched.end());

  for (i = 0 ; i < nV; i++)
    {
      
      // if vertex i is affected by a stitching command
      if (v_stitch[i] != -1 || v_refCounts[i] > 0)
	{
	  v_mapsTo[i] = nVstitched++;
	  v_stitched.push_back(i);
	}
    }
  // now among the vertices of v_stitched, we want to find out 
  // which subsets are connected by stitches (connected components)

  if (nVstitched>0)
    {
      int k;

      MPG_Partition p(nVstitched);

      for (k=0;k < nVstitched; k++)
	{
	  int v = v_stitched[k];
	  if ((v_stitch[v] != -1) && (v_stitch[v] != v))
	    {
	      p.join(k, v_mapsTo[v_stitch[v]]);
	    }
	}

      p.makeParts();

      int nParts = p.parts();

      for (k = 0; k < nParts; k ++)
	{
	  int *vk;
	  int nk;
	  p.getPart(k,nk,vk);

	  vector <int> component((size_t)nk,0);
	  
	  for (int l=0;l < nk;l++) component[l] = v_stitched[vk[l]];
	      
	  ConnectedStitchingCommands(nk,
				     component,
				     v_stitch,
				     s_length,
				     s_backwards,
				     v_refCounts,
				     s_command);
	}
	 
	
      
    }
  
  return success;
}
  
//---------------------------------------------------------------------------
//
//   join_lowest
//
//---------------------------------------------------------------------------


static int join_lowest(int i, int j, int *father)
{
  int Ci,Cj,Uij;

  if((Ci=path_compress(i,father))!=(Cj=path_compress(j,father)))
    {
      // the lowest index is always the representative
      if(Ci <= Cj)
	{ Uij = father[Cj] = Ci;  }
      else
	{ Uij = father[Ci] = Cj;  }
    }
  else
    Uij = Ci;

  return(Uij);
}

//---------------------------------------------------------------------------
//
//   MPG_Stitcher::use
//
//---------------------------------------------------------------------------

bool MPG_Stitcher::use(// inputs
		       int nV0,
		       const vector<int>&      v_father,
		       MPG_Stitches&           s_command,
		       // output
		       vector<int>&            v_label)

  // without nV0, the labels start from zero
{
  bool success               = true;

  int nV1 = v_father.size(); 
  int nVcc = nV1-nV0;

  int i;

  // insert new elements in the end of the arrays keeping track of the stitches
  
  _s_length.insert(_s_length.end(), (size_t) nVcc, 0);
  _stitch_to.insert(_stitch_to.end(), (size_t) nVcc, -1);
  _rev_stitch.insert(_rev_stitch.end(),(size_t) nVcc, false);

  
  // keep track of which vertices are merged when applying successive stitches

  // default "cluster to" for the newly added vertices

  for (i = nV0;i < nV1; i++) _v_cluster.push_back(i);
  

  // "interpret" the s_command entries starting from nV0 to nV1
  // to fill the s_length and stitch_to arrays

  for (i = nV0; i < nV1; i++)
    {
      MPG_Stitches::CommandType  type;
      int                        length;
      bool                       push;
      bool                       reverse;
      int                        anchor;
      
      s_command.get(i, type, anchor, length, push, reverse);

      if (type != MPG_Stitches::ST_NONE)
	{
	  _s_length[i] = length; // anchor = i
	  if (type != MPG_Stitches::ST_PUSH) 
	  // if (type == ST_GET || type == ST_POP)
	    {
	      _stitch_to[i] = anchor;
	      // the following was already done in MPG_Stitches.C
	      // _s_length[i]  = length+_s_length[anchor];

	      // if (push) _s_length[i = new anchor] is exacly
	      // length + _s_length[anchor], which is what we already have
	    }
	  _rev_stitch[i] = reverse;
	}
    }
  // apply all the stitches as specified in the s_length and stitch_to arrays

  // make sure to treat the reverse stitches properly
  for (i = nV0; i < nV1; i++)
    {
      int j = _stitch_to[i];
      if (j != -1)
	{
	  int length = _s_length[i];
	  int v      = i;
	  
	  // apply the (v,j, length) stitch
	  
	  if (_rev_stitch[i])
	    {
	      // build a j_father vector of all the fathers of j

	      int k;

	      vector <int> j_father;

	      for (k = 0; k < length+1; k ++, j = v_father[j])
		j_father.push_back(j);
	      for (k = 0; k < length+1; k ++, v = v_father[v])
		{
		  int w = j_father[length-k];
		  join_lowest(v,w,_v_cluster.begin());
		  
		}
	    }
	  else
	    for (int k=0;k<length+1;k++,v=v_father[v],j=v_father[j]) 
	      {
		join_lowest(v,j,_v_cluster.begin());
	      }
	}
    }

  // after all the stitches were applied, use the _v_cluster information to
  // create consecutive labels;

  {
    // path compress the v_merge array

    for (i = nV0; i < nV1; i++) path_compress(i, _v_cluster.begin());

    // assign consecutive new labels

    for (i = nV0; i < nV1; i++)  // if i < nV0 the labels were not affected
      {
	if (i == _v_cluster[i]) 
	  v_label.push_back(_nClustered++);
	else
	  v_label.push_back(v_label[_v_cluster[i]]);
      }

  }
  return success;
}

//---------------------------------------------------------------------------
//
//   bug fixes
//
//---------------------------------------------------------------------------

/*

  3/1/99 

  in bool longestStitchStartingAt(), when discovering backward
  stitches, make sure that candidate "stitch from" vertices aren't the
  same as the " stitch to" start_vertex

  similarly, when comparing the (reverse) sequence of labels to the
  stored sequence of labels, we also want to avoid going through a
  vertex that is the same as "start_vertex"

  3/1/99 

  in represent() method:

  in case a stitch cannot be encoded, reinsert a stitch with length-1
  in the priority queue to have more possibilities

  3/2/99 

  in use() method:

  the join_lowest() call replaces _v_cluster (MAX(a,b)) = MIN (a,b)
  we have to do connected components properly 

  (there was a core dump for enterprise.wrl and superfemur.wrl)

  
  3/3/99  (pb discoverted by FB)

  in use() method: comment out 
  _s_length[i]  = length+_s_length[anchor];
  
  also in MPG_stitches make 2 changes

 */
