//--------------------------------------------------------------------------
//
//  ANG_Arcball.cpp
//
//  Copyright (C) IBM Corporation, 1999
//  Program Property of IBM
//
//  AUTHORS: 
//		Ioana M. Martin (ioana@us.ibm.com)
//
//  IBM T. J. Watson Research Center
//  30 Saw Mill River Road
//  Hawthorne, New York 10532
//
//  Reference: Ken Shoemake, "Arcball Rotation Control", Graphics Gems iv
//--------------------------------------------------------------------------

#include <math.h>
#include "ANG_Arcball.h"

/*
 *------------------------------------------------------------------
 * Arcball -- initilaize the arcball
 *------------------------------------------------------------------
 */

Arcball::Arcball()
{
	vZero.x = vZero.y = vZero.z = 0.0;
	qOne.v.x = qOne.v.y = qOne.v.z = 0.0;
	qOne.w = 1.0;
	sLeft = sBottom = -1.0;
	sRight = sTop = 1.0;

	rotAngle = 0.0;
    center = vZero;
    radius = 1.0;
    v0 = vZero;
    v1 = vZero;
    qDown = qOne;
    qNow  = qOne;
    dragging = 0;
}

/*
 *------------------------------------------------------------------
 * ballPlace -- change arcball params
 *------------------------------------------------------------------
 */

void Arcball::ballPlace(WVECTOR c, float r)
{
	center = c;
	radius = r;
}

/*
 *------------------------------------------------------------------
 * ballBeginDrag -- begin drag sequence
 *------------------------------------------------------------------
 */

void Arcball::ballBeginDrag(WVECTOR here)
{
	dragging = 1;
	v0.x = here.x;
	v0.y = here.y;
	v0.z = here.z;
}

/*
 *------------------------------------------------------------------
 * ballEndDrag -- end drag sequence
 *------------------------------------------------------------------
 */

void Arcball::ballEndDrag()
{
	dragging = 0;
	qDown.v.x = qNow.v.x;
	qDown.v.y = qNow.v.y;
	qDown.v.z = qNow.v.z;
	qDown.w = qNow.w;
}

/*
 *------------------------------------------------------------------
 * worldToBall -- convert from world coordinates to ball coords
 *------------------------------------------------------------------
 */

void Arcball::worldToBall(float x, float y, WVECTOR& v)
{

	float mag, scale;

	v.x = (x - center.x) / radius; 
	v.y = (y - center.y) / radius; 

	mag = v.x * v.x + v.y * v.y;
	if(mag > 1.0)
	{
		 scale = 1.0 / sqrt(mag);
		 v.x *= scale; v.y *= scale;
		 v.z = 0.0;
	}
	else
	{
		 v.z = sqrt(1.0 - mag);
	}
} 

   
/*
 *----------------------------------------------------------------
 * getRotAngle -- extract angle of rotation from quaternion
 *----------------------------------------------------------------
 */

float Arcball::GetRotAngle(QUAT q)
{
	return(2.0 * acos(q.w));
}

float Arcball::GetRotAngle()
{
	return rotAngle;
}

void Arcball::SetRotAngle(float angle)
{
	rotAngle = angle;
}

/*
 *----------------------------------------------------------------
 * getRotAxis -- extract axis of rotation from quaternion
 *----------------------------------------------------------------
 */

void Arcball::GetRotAxis(QUAT q, WVECTOR& v)
{
	v.x = q.v.x; 
	v.y = q.v.y; 
	v.z = q.v.z;
}

float Arcball::GetXRotAxis()
{
	return rotAxis.x;
}

float Arcball::GetYRotAxis()
{
	return rotAxis.y;
}

float Arcball::GetZRotAxis()
{
	return rotAxis.z;
}

void Arcball::SetRotAxis(WVECTOR a)
{
	rotAxis = a; 
}


/*
 *----------------------------------------------------------------
 * quatMult -- multiply two quaternions
 *----------------------------------------------------------------
 */

void Arcball::quatMult(QUAT q1, QUAT q2, QUAT& q)
{
	WVECTOR vp;
	float dp;

	vectProd(q1.v, q2.v, vp);
	dotProd(q1.v, q2.v, dp);
	q.v.x = q2.w * q1.v.x + q1.w * q2.v.x + vp.x;
	q.v.y = q2.w * q1.v.y + q1.w * q2.v.y + vp.y;
	q.v.z = q2.w * q1.v.z + q1.w * q2.v.z + vp.z;
	q.w   = q1.w * q2.w - dp;

}

/*
 *----------------------------------------------------------------
 * quatInv -- take the inverse of a quaternion
 *----------------------------------------------------------------
 */

void Arcball::quatInv(QUAT q1, QUAT& q)
{
	q.v.x = q1.v.x; q.v.y = q1.v.y; q.v.z = q1.v.z;
	q.w = (-q1.w);
}

/*
 *----------------------------------------------------------------
 * vectToQuat -- convert a vector to a quaternion
 *----------------------------------------------------------------
 */

void Arcball::vectToQuat(WVECTOR v, QUAT& q)
{
	float norm;

	norm = sqrt(v.x*v.x + v.y*v.y + v.z*v.z);
	if(norm != 0.)
	{
		q.v.x = v.x / norm;
		q.v.y = v.y / norm;
		q.v.z = v.z / norm;
		q.w = 0.;
	}
	else
	{
		q.v.x = 0.;
		q.v.y = 0.;
		q.v.z = 0.;
		q.w = 1.;
	}
}

/*
 *--------------------------------------------------------------
 * vectProd -- vector product of two vectors
 *--------------------------------------------------------------
 */

void Arcball::vectProd(WVECTOR v1, WVECTOR v2, WVECTOR& v)
{
	v.x = v1.y * v2.z - v1.z * v2.y;
	v.y = v1.z * v2.x - v1.x * v2.z;
	v.z = v1.x * v2.y - v1.y * v2.x;
}

/*
 *--------------------------------------------------------------
 * dotProd -- dot product of two vectors
 *--------------------------------------------------------------
 */

void Arcball::dotProd(WVECTOR v1, WVECTOR v2, float& s)
{
	s = v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}

float Arcball::GetSLeft()
{
	return sLeft;
}

float Arcball::GetSRight()
{
	return sRight;
}

float Arcball::GetSTop()
{
	return sTop;
}

float Arcball::GetSBottom()
{
	return sBottom;
}





  
 
