//--------------------------------------------------------------------------
//
//  P_MPG_Ifs.cpp
//
//  Copyright (C) IBM Corporation, 1999
//  Program Property of IBM
//
//  AUTHORS: 
//		Ioana M. Martin (ioana@us.ibm.com)
//
//  IBM T. J. Watson Research Center
//  30 Saw Mill River Road
//  Hawthorne, New York 10532
//
//--------------------------------------------------------------------------

#include "afx.h"
#include <windows.h>
#include <GL/gl.h>
#include "P_MPG_IfsQ_E.h"

//------------------------------------------------------
// Render
//------------------------------------------------------
void P_MPG_IfsQ_E::Render()
{
	int nV = _coord.size()/3;
	int nF = _coordIndex.size();

	{
		glBegin(GL_POLYGON);
		for(int j=0;j<nF;j++)
		{
			if(_coordIndex[j] == -1)
			{
				glEnd();
				if(j != nF-1)
					glBegin(GL_POLYGON);
			}
			else
			{
				float nvec[3];
				nvec[0] = _normal[3*_coordIndex[j]];
				nvec[1] = _normal[3*_coordIndex[j] + 1];
				nvec[2] = _normal[3*_coordIndex[j] + 2];
				
				float vvec[3];
				vvec[0] = _coord[3*_coordIndex[j]];
				vvec[1] = _coord[3*_coordIndex[j] + 1];
				vvec[2] = _coord[3*_coordIndex[j] + 2];
				glNormal3fv(nvec);
				glVertex3fv(vvec);
			}
		}
	}
}


//------------------------------------------------------
// GetBBox
//------------------------------------------------------
void P_MPG_IfsQ_E::GetBBox(float& width, float& height, float& depth,
		float& cx, float& cy, float& cz)
{
	const float FACTOR = 1.2f;
	float minx, miny, minz;
	float maxx, maxy, maxz;

	minx = _coord[0]; miny = _coord[1]; minz = _coord[2];
	maxx = _coord[0]; maxy = _coord[1]; maxz = _coord[2];
	int nV = _coord.size()/3;
	for(int j=0;j<nV;j++)
	{
		if(minx > _coord[3*j]) 
			minx = _coord[3*j];
		if(miny > _coord[3*j+1]) 
			miny = _coord[3*j+1];
		if(minz > _coord[3*j+2]) 
			minz = _coord[3*j+2];

		if(maxx < _coord[3*j]) 
			maxx = _coord[3*j];
		if(maxy < _coord[3*j+1]) 
			maxy = _coord[3*j+1];
		if(maxz < _coord[3*j+2]) 
			maxz = _coord[3*j+2];
	}

	// box dimensions
	width = (maxx - minx > 0) ? FACTOR*(maxx - minx) : 1.0;
	height = (maxy - miny > 0) ? FACTOR*(maxy - miny) : 1.0;
	depth = (maxz - minz > 0) ? FACTOR*(maxz - minz) : 1.0;

	// box center
	cx = (maxx + minx)*0.5;
	cy = (maxy + miny)*0.5;
	cz = (maxz + minz)*0.5;
}

//------------------------------------------------------
// ComputeNormals
//------------------------------------------------------
void P_MPG_IfsQ_E::ComputeNormals()
{
	float v0[3], v1[3], v2[3];
	int nF = _coordIndex.size();
	int nV = _coord.size()/3;
	float *n = new float[3*nV];
	memset(n, '\0', 3*nV*sizeof(float));

	_normal.clear();
	int i = 0;
	while(i<nF)
	{
		// compute unit normal for this face
		v1[0] = _coord[3*_coordIndex[i+1]]   - _coord[3*_coordIndex[i]];
		v1[1] = _coord[3*_coordIndex[i+1]+1] - _coord[3*_coordIndex[i]+1];
		v1[2] = _coord[3*_coordIndex[i+1]+2] - _coord[3*_coordIndex[i]+2];

		v2[0] = _coord[3*_coordIndex[i+2]]   - _coord[3*_coordIndex[i]];
		v2[1] = _coord[3*_coordIndex[i+2]+1] - _coord[3*_coordIndex[i]+1];
		v2[2] = _coord[3*_coordIndex[i+2]+2] - _coord[3*_coordIndex[i]+2];

		
		CrossProd(v0, v1, v2); // triangle

		int j = i;
		while(_coordIndex[j] != -1)
		{
			int k = 3*_coordIndex[j];
			n[k]   += n[k]   + v0[0];
			n[k+1] += n[k+1] + v0[1];
			n[k+2] += n[k+2] + v0[2];
			j++;
		}
		
		// skip the -1
		i = j+1;
	}

	// normalize normals
	int k = 0;
	for (i=0; i<nV; i++)
	{
		float sum = n[3*i]*n[3*i] + 
			n[3*i+1]*n[3*i+1] + n[3*i+2]*n[3*i+2];
		for(k=0;k<3;k++)
		{
			if(sum)
				n[3*i+k] /= sqrt(sum);
			else
				n[3*i+k] = 0.0;
			_normal.push_back(n[3*i+k]);
		}
	}
}

//------------------------------------------------------
// CrossProd(c, a, b) - compute cross product: a x b -> c
//------------------------------------------------------
void P_MPG_IfsQ_E::CrossProd (float c[], float a[], float b[])
{
	double mag;
	const int X = 0, Y = 1, Z = 2;

	c[X] = a[Y] * b[Z] - a[Z] * b[Y];
	c[Y] = a[Z] * b[X] - a[X] * b[Z];
	c[Z] = a[X] * b[Y] - a[Y] * b[X];

	// normalize result 
	mag = sqrt(c[X]*c[X] + c[Y]*c[Y] + c[Z]*c[Z]);
	c[X] /= mag;
	c[Y] /= mag;
	c[Z] /= mag;
}  
