//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------


#include <stdio.h>
#include <float.h>
#include <math.h>
#include <fstream.h>
#include "MPG_BBoxQ.h"

const float MPG_BBoxQ_EXT::_epsilon = float(1e-20);

static inline int _quantize(float x, int n)
//{ return (int)(floor(ldexp(x,n))); }
{ return (int)(ldexp(x,n)-0.5); }	//MJH 001218

static inline float _unQuantize(int x, int n)
{ return ldexp((float)(x),-n); }

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ_EXT::MPG_BBoxQ_EXT
//
//---------------------------------------------------------------------------

MPG_BBoxQ_EXT::MPG_BBoxQ_EXT(int nD, int nBitsPerCoordinate)
{
  if(nD>0)
    {
      setDimension(nD);
      setBitsPerCoordinate(nBitsPerCoordinate);
      _sideLength.reserve(nD);
      _sideLength.insert(_sideLength.begin(),nD,0.0);
      _avgError = 0.0;
      _maxError = 0.0;
    }
}

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ_EXT::encode
//
//---------------------------------------------------------------------------

bool MPG_BBoxQ_EXT::encode(MPG_OutputBitStream_EXT& obstrm)
{
  float* minCorner   = getLowerCorner();
  float side         = getSideLength();

  // determine if bbox is default or not
  bool  non_default_bbox = false;
  int i;
  for(i=0;i<getDimension();i++)
    if(minCorner[i]!=0.0)
      { non_default_bbox = true; break; }
  if(side!=1.0) non_default_bbox = true;

  if(obstrm.putBit(non_default_bbox)==false) return false;

  // encode if not
  if(non_default_bbox==true)
    {
      for(i=0;i<getDimension();i++)
	if(obstrm.put32BitFloat(minCorner[i])==false) return false;
      if(obstrm.put32BitFloat(side)==false) return false;
    }

  // encode bits per coordinate
  int bitsPerCoord = getBitsPerCoordinate();
  if(obstrm.putMBitInt(5,bitsPerCoord)==false) return false;

  return true;
}

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ_EXT::decode
//
//---------------------------------------------------------------------------

bool MPG_BBoxQ_EXT::decode(MPG_InputBitStream_EXT& ibstrm)
{
  int dimValue = getDimension();

  bool non_default_bbox = false;
  if(ibstrm.getBit(non_default_bbox)==false) return false;
  if(non_default_bbox)
    {
      float minCorner_i = 0.0;
      int i;
      for(i=0;i<dimValue;i++)
	{
	  if(ibstrm.get32BitFloat(minCorner_i)==false) return false;
	  setLowerCorner(i,minCorner_i);
	}
      float side;
      if(ibstrm.get32BitFloat(side)==false) return false;
      setSideLength(side);
    }
  else
    {
      int i;
      for(i=0;i<dimValue;i++)
	setLowerCorner(i,0.0);
      setSideLength(1.0);
    }

  // decode bits per coordinate
  int bitsPerCoord = 0;
  if(ibstrm.getMBitInt(5,bitsPerCoord)==false) return false;
  if(bitsPerCoord<=0) return false;
  setBitsPerCoordinate(bitsPerCoord);

  return true;
}

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ_EXT::_update
//
//---------------------------------------------------------------------------

void MPG_BBoxQ_EXT::_update(int nValues, float* value)
{
  int   nD      = getDimension();
  while(--nValues>=0)
    {
      for(int i=0;i<nD;i++)
	{
	  float dvi = value[i] - _lowerCorner[i];
	  if(dvi<0.0)
	    {
	      _sideLength[i]  -= dvi;
	      _lowerCorner[i] += dvi;
	    }
	  else if(dvi>_sideLength[i])
	    _sideLength[i] = dvi;
	}
      value += nD;
    }
}

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ_EXT::_update
//
//---------------------------------------------------------------------------

void MPG_BBoxQ_EXT::_update(vector<float>& v)
{
  int   nValues = v.size()/getDimension();
  float *value  = v.begin();
  _update(nValues,value);
}

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ_EXT::quantize
//
//---------------------------------------------------------------------------

bool MPG_BBoxQ_EXT::quantize(float* value, int* iqValue, int nValues)
{
  //const char *file = "guest.dat";
  //ofstream fout(file, ios::out, ios::app);
  int nD = getDimension();

  float maxsidelen=getSideLength();
  if(maxsidelen<=0.0) return false;

  int bpc = getBitsPerCoordinate();
  if(bpc<=0) return false;

  float delta = ldexp(maxsidelen,-bpc-1);
  maxsidelen += 2*delta;

  float* dqv = new float[nD];

  int nErr  = nValues;
  _avgError = 0.0;
  _maxError = 0.0;

  // int maxVal = 1<<bpc;
  while(--nValues>=0)
    {
    float nvi = 0.0;

      int i;
      for(i=0;i<nD;i++)
	{
	  // normalize to [0,1]
	  nvi = (value[i]-_lowerCorner[i]+delta)/maxsidelen;
	  if(nvi<0.0) nvi = 0.0;
          else if(nvi>1.0) nvi = 1.0;
	  iqValue[i] = _quantize(nvi,_bitsPerCoordinate);
	 // fout<<iqValue[i];
	  // compute unQuantize value to evaluate error
	  nvi     = _unQuantize(iqValue[i],_bitsPerCoordinate);
	  dqv[i]  = nvi*maxsidelen+_lowerCorner[i]-delta;
	  dqv[i] -= value[i];
	}
      value   += nD;
      iqValue += nD;

  	  if(nvi<0.0 || nvi>1.0) continue;

      float nqv = 0.0;
      for(i=0;i<nD;i++)
	nqv += dqv[i]*dqv[i];
      if(nqv>0.0) nqv = sqrt(nqv);
      if(nqv>_maxError) _maxError = nqv;
      _avgError += nqv;

    }
  _avgError /= nErr;
  _maxError /= maxsidelen;
  _avgError /= maxsidelen;

  delete [] dqv;
  return true;
}

//DMLAB (200501) - ADD
bool MPG_BBoxQ_EXT::texquantize(float* value, int* iqValue, int nValues, int img_flag, int img_u, int img_v)
{
  
  int nD = getDimension();

  float maxsidelen=getSideLength();
  if(maxsidelen<=0.0) return false;

  int bpc = getBitsPerCoordinate();
  if(bpc<=0) return false;
  
  float delta = ldexp(maxsidelen,-bpc-1); 
    
  maxsidelen += 2*delta;

  float* dqv = new float[nD];

  int nErr  = nValues;
  _avgError = 0.0;
  _maxError = 0.0;

  int delta1_u, delta1_v, delta2_u, delta2_v;

	if(img_flag) {   
		delta1_u = (img_u-1);
		delta1_v = (img_v-1);
	}
  
		float* tex_u = new float[nValues];	  
		float* tex_v = new float[nValues];

		int nV = nValues, SFV = 1;
		int i, j, cnt=0;
		float* start_value = value;

		while(--nV>=0)
		{
			for(i=0; i<nD; i++) {
					
				if(i==0)  tex_u[cnt]=value[i];
				else if(i==1)  tex_v[cnt]=value[i];
			}
			cnt++;
			value   += nD;
		}
		value = start_value;

	// sorting tex_u, tex_v
		for(i=0;i<(nValues-1);i++)
			for(j=(i+1);j<nValues;j++)
			{
				if(tex_u[i]>tex_u[j])
				{float tmp=tex_u[i]; tex_u[i]=tex_u[j]; tex_u[j]=tmp;}
			
				if(tex_v[i]>tex_v[j])
				{float tmp=tex_v[i]; tex_v[i]=tex_v[j]; tex_v[j]=tmp;}
			}

	// error estimation
		int* error_u = new int[nValues-1];
		int* error_v = new int[nValues-1];

		int max_u=-1000 ,max_v=-1000;

		for(i=0;i<(nValues-1);i++)
		{
			if((tex_u[i+1]-tex_u[i]) != 0)
				error_u[i] = int(1/(tex_u[i+1] - tex_u[i]) + 0.5); // int delta
			else
				error_u[i] = 0;

			if((tex_v[i+1]-tex_v[i]) != 0)
				error_v[i] = int(1/(tex_v[i+1] - tex_v[i]) + 0.5); // int delta
			else
				error_v[i] = 0;

	   }

	   // error sort
	   for(i=0;i<(nValues-2);i++)
		for(int j=(i+1);j<(nValues-1);j++)
		{
			if(error_u[i]>error_u[j])
				{int tmp=error_u[i]; error_u[i]=error_u[j]; error_u[j]=tmp;}
			
			if(error_v[i]>error_v[j])
				{int tmp=error_v[i]; error_v[i]=error_v[j]; error_v[j]=tmp;}
		}
		
		// error sampling
	  int* key_u = new int[nValues-1];
	  int* key_v = new int[nValues-1];
	  int* key_u_num = new int[nValues-1];
	  int* key_v_num = new int[nValues-1];

		int cnt_u = 0, cnt_v = 0;

		key_u[0] = error_u[0];
		key_v[0] = error_v[0];
		key_u_num[0] = key_v_num[0] =0;

		for(i=0;i<(nValues-1);i++)
		{
			if(key_u[cnt_u] == error_u[i]) 
				key_u_num[cnt_u]+=1; 
			else {
				cnt_u++;
				key_u[cnt_u] = error_u[i];			
				key_u_num[cnt_u] = 1;
			}

			if(key_v[cnt_v] == error_v[i]) 
				key_v_num[cnt_v]+=1; 
			else {
				cnt_v++;
				key_v[cnt_v] = error_v[i];			
				key_v_num[cnt_v] = 1;
			}
		}

// Mode value
		int mode_u, mode_v, unit_u, unit_v;

		int max_key_u = -100, max_key_v = -100;
		int pos_u = 0, pos_v = 0;

		for(i=1;i<=cnt_u;i++) {
			if(key_u_num[i] >= max_key_u) {
				max_key_u = key_u_num[i];
				pos_u = i;
			}
		}

		for(i=1;i<=cnt_v;i++) {
			if(key_v_num[i] >= max_key_v) {
				max_key_v = key_v_num[i];
				pos_v = i;
			}
		}

		mode_u = unit_u = key_u[pos_u];	  
		mode_v = unit_v = key_v[pos_v];	  

		if(pos_u+1 <= cnt_u)
			for(i=pos_u+1;i<=cnt_u;i++) {
				if((mode_u + unit_u) == key_u[i])
					mode_u = key_u[i];
			}
		if(pos_v+1 <= cnt_v)
			for(i=pos_v+1;i<=cnt_v;i++) {
				if((mode_v + unit_v) == key_v[i])
					mode_v = key_v[i];
			}

	delta2_u = mode_u;
	delta2_v = mode_v;

	if(img_flag) {   
		if((img_u > delta2_u) && (img_u % delta2_u)==0)
			delta_u = delta2_u;
		else
			delta_u = delta1_u;

		if((img_v > delta2_v) && (img_v % delta2_v)==0)
			delta_v = delta2_v;
		else
			delta_v = delta1_v;
	}
	else {
		delta_u = delta2_u;
		delta_v = delta2_v;
	}

	int delta_bit;
	if(delta_u>delta_v) delta_bit = (int)ceil(log(delta_u)/log(2));
	else delta_bit = (int)ceil(log(delta_v)/log(2));

  if(_bitsPerCoordinate < delta_bit) {
	  printf("tex:1\n");
	  float intMSE=0, sumError=0;

    while(--nValues>=0)
    {
	    float nvi = 0.0;
		sumError=0;

		  int i;
	      for(i=0;i<nD;i++)
		{
		  // normalize to [0,1]
		  nvi = (value[i]-_lowerCorner[i]+delta)/maxsidelen;
		  if(nvi<0.0) nvi = 0.0;
		  else if(nvi>1.0) nvi = 1.0;
		  iqValue[i] = _quantize(nvi,_bitsPerCoordinate);

		  // compute unQuantize value to evaluate error
		  nvi     = _unQuantize(iqValue[i],_bitsPerCoordinate);
		  dqv[i]  = nvi*maxsidelen+_lowerCorner[i]-delta;

		  dqv[i] -= value[i];
		}
	  
	      value   += nD;
		  iqValue += nD;

	  	  if(nvi<0.0 || nvi>1.0) continue;

		  float nqv = 0.0;
	      for(i=0;i<nD;i++)
			nqv += dqv[i]*dqv[i];
	      if(nqv>0.0) nqv = sqrt(nqv);
		  if(nqv>_maxError) _maxError = nqv;
		  _avgError += nqv;

	}
	_avgError /= nErr;
	_maxError /= maxsidelen;
	_avgError /= maxsidelen;
	
	delta_flag = false;
  }

  else {
	  printf("%d %d\n",delta_u,delta_v);

	  while(--nValues>=0)
	  {
		float nvi = 0.0;
		int delta_UV;

		int i;
		for(i=0;i<nD;i++)
		{
			if(i==0) delta_UV = delta_u;
			else if(i==1) delta_UV = delta_v;

			// normalize to [0,1]
			if(value[i]<0.0) value[i] = 0.0;
			else if(value[i]>1.0) value[i] = 1.0;

			if(value[i]  > 0)
	 			iqValue[i] = (int)(value[i] * (float)delta_UV + 0.5);
			else
 				iqValue[i] = (int)(value[i] * (float)delta_UV - 0.5); // float and int (1/unit) 

			// compute unQuantize value to evaluate error
			dqv[i] = (float)iqValue[i]/delta_UV; // int (1/unit)

			dqv[i] -= value[i];
		}

		value   += nD;
		iqValue += nD;

	  	if(nvi<0.0 || nvi>1.0) continue;

		float nqv = 0.0;
		for(i=0;i<nD;i++)
			nqv += dqv[i]*dqv[i];
		if(nqv>0.0) nqv = sqrt(nqv);
	    if(nqv>_maxError) _maxError = nqv;
		_avgError += nqv;

	  }
	_avgError /= nErr;
	
	delta_flag = true;
	_bitsPerCoordinate = delta_bit;

  }
  delete [] key_u;
  delete [] key_v;
  delete [] key_u_num;
  delete [] key_v_num;
  delete [] dqv;
  return true;
}
//~DMLAB (200501)

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ_EXT::quantize
//
//---------------------------------------------------------------------------

bool MPG_BBoxQ_EXT::quantize(vector<float>& v,
                         vector<int>&   iqv)
{
  int   nD       = getDimension();
  int   nValues  = v.size()/nD;
  iqv.erase(iqv.begin(),iqv.end());
  iqv.insert(iqv.end(),(size_t)(nD*nValues),0);
  float *value   = v.begin();
  int   *iqValue = iqv.begin();
  return quantize(value,iqValue,nValues);
}
//DMLAB (200501) - ADD
bool MPG_BBoxQ_EXT::texquantize(vector<float>& v,
                         vector<int>&   iqv,
						 int img_flag,
						 int img_u,
						 int img_v)
{
  int   nD       = getDimension();
  int   nValues  = v.size()/nD;
  iqv.erase(iqv.begin(),iqv.end());
  iqv.insert(iqv.end(),(size_t)(nD*nValues),0);
  float *value   = v.begin();
  int   *iqValue = iqv.begin();
  return texquantize(value,iqValue,nValues,img_flag,img_u,img_v);
}
//~DMLAB (200501) 

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ_EXT::quantize
//
//---------------------------------------------------------------------------

bool MPG_BBoxQ_EXT::quantize(vector<float>& v,
                         vector<int>&   vIndx,
                         vector<int>&   iqv)
{
  int i,j,k;
  int nD  = getDimension();
  int nV  = vIndx.size();
  vector<float> v_flat;
  for(i=0;i<vIndx.size();i++)
    if((j=vIndx[i])>=0)
      for(k=0;k<nD;k++)
        v_flat.push_back(v[nD*j+k]);
  return quantize(v_flat,iqv);
}
//******************************************************************************
//DMLAB (200501) - ADD
bool MPG_BBoxQ_EXT::texquantize(vector<float>& v,
                         vector<int>&   vIndx,
                         vector<int>&   iqv,
						 int img_flag,
						 int img_u,
						 int img_v)
{
  int i,j,k;
  int nD  = getDimension();
  int nV  = vIndx.size();
  vector<float> v_flat;
  for(i=0;i<vIndx.size();i++)
    if((j=vIndx[i])>=0)
      for(k=0;k<nD;k++)
        v_flat.push_back(v[nD*j+k]);
  return texquantize(v_flat,iqv,img_flag,img_u,img_v);
}
//~DMLAB (200501)
//---------------------------------------------------------------------------
//
//   MPG_BBoxQ_EXT::unQuantize
//
//---------------------------------------------------------------------------

bool MPG_BBoxQ_EXT::unQuantize(int* iqValue, float* uqValue, int nValues)
{
  int nD = getDimension();

  float maxsidelen=getSideLength();
  if(maxsidelen<=0.0) return false;

  int bpc = getBitsPerCoordinate();
  if(bpc<=0) return false;
  if(bpc>30) return false;

  float delta = ldexp(maxsidelen,-bpc-1);
  maxsidelen += 2*delta;

  int*   iqv = iqValue;
  float* uqv = uqValue;

  int maxVal = (1<<bpc)-1;
  while(--nValues>=0)
    {
      for(int i=0;i<nD;i++)
	{
          int iqvi = iqValue[i];
          if(iqvi<0) iqvi = 0;
          else if(iqvi>maxVal) iqvi=maxVal;
	  float nvi = _unQuantize(iqv[i],_bitsPerCoordinate);
	  uqv[i]  = nvi*maxsidelen+_lowerCorner[i]-delta;
	}
      uqv += nD;
      iqv += nD;
    }
  return true;
}

//DMLAB (200501) - ADD
bool MPG_BBoxQ_EXT::texunQuantize(int* iqValue, float* uqValue, int nValues)
{
  int nD = getDimension();
  int min_u=delta_u;
  int min_v=delta_v;

  int bpc = getBitsPerCoordinate();
  if(bpc<=0) return false;
  if(bpc>30) return false;

  int delta_UV;

  int*   iqv = iqValue;
  float* uqv = uqValue;

  int maxVal = (1<<bpc)-1;
  
  while(--nValues>=0)
    {
      for(int i=0;i<nD;i++)
	{

	 if(i==0) delta_UV=min_u;
	 else  delta_UV=min_v;
     
     int iqvi = iqValue[i];
     if(iqvi<0) iqvi = 0;
     else if(iqvi>maxVal) iqvi=maxVal;

		float nvi = (float)iqv[i]/delta_UV;

	  uqv[i]  = nvi;
	}
      uqv += nD;
      iqv += nD;
    }
  return true;
}
//~DMLAB (200501) 

//---------------------------------------------------------------------------
//
//   MPG_BBoxQ_EXT::unQuantize
//
//---------------------------------------------------------------------------

bool MPG_BBoxQ_EXT::unQuantize(vector<int>& iqv, vector<float>& qv)
{
  int nD      = getDimension();
  int nValues = iqv.size()/nD;
  qv.erase(qv.begin(),qv.end());
  qv.insert(qv.end(),iqv.size(),(float)0.0);
  float* uqValue   = qv.begin();
  int*   iqValue = iqv.begin();
  return unQuantize(iqValue,uqValue,nValues);
}
//DMLAB (200501) - ADD
bool MPG_BBoxQ_EXT::texunQuantize(vector<int>& iqv, vector<float>& qv)
{
  int nD      = getDimension();
  int nValues = iqv.size()/nD;
  qv.erase(qv.begin(),qv.end());
  qv.insert(qv.end(),iqv.size(),(float)0.0);
  float* uqValue   = qv.begin();
  int*   iqValue = iqv.begin();
  return texunQuantize(iqValue,uqValue,nValues);
}
//~DMLAB (200501)