//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------


#include <stdio.h>
#include <stdlib.h>
#include "MPG_ClustSurf.h"
extern FILE* logfp;

int COM_Dbg_on = 0;

// #define COLOR_COLLAPSED_FACES

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::MPG_ClustSurf
//
//---------------------------------------------------------------------------

MPG_ClustSurf::MPG_ClustSurf(int nD)
#ifdef WITH_MESH
  :mesh(_coordIndex)
#endif // WITH_MESH
{
  _bbqCoord    = (MPG_BBoxQ*)0;
  _nrqNormal   = (MPG_NormalQ*)0;
  _bbqColor    = (MPG_BBoxQ*)0;
  _bbqTexCoord = (MPG_BBoxQ*)0;
  clear();
  _nD = (nD<1)?1:nD;
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::clear
//
//---------------------------------------------------------------------------

void MPG_ClustSurf::clear()
{
  _nD                = 3;

  _ccw               = true;
  _convex            = true;
  _creaseAngle       = 0.0;
  _solid             = true;

  _bitsPerCoord      = 10;
  _bitsPerNormal     =  9;
  _bitsPerColor      =  8;
  _bitsPerTexCoord   = 10;

  _coordBinding      = PB_PER_VERTEX;
  _normalBinding     = PB_NONE;
  _colorBinding      = PB_NONE;
  _texCoordBinding   = PB_NONE;

  _coordIndex.erase(_coordIndex.begin(),_coordIndex.end());
#ifdef WITH_MESH
  mesh.reset();
#endif // WITH_MESH

  _coordMap.erase(_coordMap.begin(),_coordMap.end());

  _coord.erase(_coord.begin(),_coord.end());
  _normal.erase(_normal.begin(),_normal.end());
  _color.erase(_color.begin(),_color.end());
  _texCoord.erase(_texCoord.begin(),_texCoord.end());

  _coordFirst.erase(_coordFirst.begin(),_coordFirst.end());
  _normalFirst.erase(_normalFirst.begin(),_normalFirst.end());
  _colorFirst.erase(_colorFirst.begin(),_colorFirst.end());
  _texCoordFirst.erase(_texCoordFirst.begin(),_texCoordFirst.end());

  _coordFirst.push_back(0);
  _normalFirst.push_back(0);
  _colorFirst.push_back(0);
  _texCoordFirst.push_back(0);

  if(_bbqCoord!=(MPG_BBoxQ*)0)
    { delete _bbqCoord;    _bbqCoord   = ((MPG_BBoxQ*)0);   }
  if(_nrqNormal!=(MPG_NormalQ*)0)
    { delete _nrqNormal;   _nrqNormal  = ((MPG_NormalQ*)0); }
  if(_bbqColor!=(MPG_BBoxQ*)0)
    { delete _bbqColor;    _bbqColor   = ((MPG_BBoxQ*)0);   }
  if(_bbqTexCoord!=(MPG_BBoxQ*)0)
    { delete _bbqTexCoord; _bbqTexCoord= ((MPG_BBoxQ*)0);   }

}

//---------------------------------------------------------------------------
//
//   getToken
//
//---------------------------------------------------------------------------

static char* getToken(vector<char>& str, int& i, int& j)
{
  char* tkn = "";
  int n = str.size();
  while(i<n &&  (str[i]==' ' || str[i]=='\t')) i++;
  if(i<n)
    {
      for(j=i+1;j<n && !(str[j]==' ' || str[j]=='\t');j++);
      if(j<=n)
	{ str[j] = '\0'; tkn = &(str[i]); i=j+1; }
    }
  return tkn;
}

//---------------------------------------------------------------------------
//
//   setBinding
//
//---------------------------------------------------------------------------

static void setBinding(char* token, MPG_Binding& binding)
{
  if(!strcmp(token,"perVertex"))
    binding = PB_PER_VERTEX;
  else if(!strcmp(token,"perFace"))
    binding = PB_PER_FACE;
  else if(!strcmp(token,"perCorner"))
    binding = PB_PER_CORNER;
  else
    binding = PB_NONE;
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::readFaces()
//
//---------------------------------------------------------------------------

int MPG_ClustSurf::readFaces(FILE *fp)
{
  int err = 0;
  if(fp==(FILE*)0)
    err = -2;
  else
    {
      _coordIndex.erase(_coordIndex.begin(),_coordIndex.end());
      int  nF=0;
      if(fscanf(fp,"%d",&nF)<1 || nF<=0)
	err = -3;
      else
	{
	  int  i,j,nfi,fij;
	  for(i=0;i<nF;i++)
	    {
	      if(fscanf(fp,"%d",&nfi)<1 || nfi<3)
		{ err = -3; break; }
	      for(j=0;j<nfi;j++)
		if(fscanf(fp,"%d",&fij)<1 || fij<0)
		  { err = -4; break; }
	      else
		_coordIndex.push_back(fij);
	      _coordIndex.push_back(-1);
	    }
#ifdef WITH_MESH
	  mesh.set(_coordIndex);
#endif // WITH_MESH
	}
    }
  return err;
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::readTriangles()
//
//---------------------------------------------------------------------------

int MPG_ClustSurf::readTriangles(FILE *fp)
{
  int err = 0;
  if(fp==(FILE*)0)
    err = -2;
  else
    {
      _coordIndex.erase(_coordIndex.begin(),_coordIndex.end());
      int  nT = 0;
      if(fscanf(fp,"%d",&nT)<1 || nT<=0)
	err = -3;
      else
	{
	  int  i,j,nfi,fij;
	  nfi = 3;
	  for(i=0;i<nT;i++)
	    {
	      for(j=0;j<nfi;j++)
		if(fscanf(fp,"%d",&fij)<1 || fij<0)
		  { err = -4; break; }
	      else
		_coordIndex.push_back(fij);
	      _coordIndex.push_back(-1);
	    }
#ifdef WITH_MESH
	  mesh.set(_coordIndex);
#endif // WITH_MESH
	}
    }
  return err;
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::writeFaces()
//
//---------------------------------------------------------------------------

int MPG_ClustSurf::writeFaces(FILE *fp)
{
  int err = 0;
  if(fp==(FILE*)0)
    err = -1;
  else
    {
      MPG_Mesh mesh(_coordIndex);
      int nF = mesh.getNumberOfFaces();
      if(fprintf(fp,"%d\n",nF)<1)
	err = -2;
      else
	{
	  int i,j;
	  for(i=0;i<nF;i++)
	    {
	      int nfi = mesh.getNumberOfFaceCorners(i);
	      if(fprintf(fp,"%d ",nfi)<1)
		{ err = -3; break; }
	      for(j=0;j<nfi;j++)
		if(fprintf(fp,"%d ",mesh.getFaceCorner(i,j))<1)
		  { err = -4; break; }
	      fprintf(fp,"\n");
	    }
	}
    }
  return err;
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::writeTriangles()
//
//---------------------------------------------------------------------------

int MPG_ClustSurf::writeTriangles(FILE *fp)
{
  int err = 0;
  if(fp==(FILE*)0)
    err = -1;
  else
    {
      MPG_Mesh mesh(_coordIndex);
      if(mesh.isTriangulated())
	{
	  int nT = mesh.getNumberOfFaces();
	  if(fprintf(fp,"%d\n",nT)<1)
	    err = -2;
	  else
	    {
	      int i,j;
	      for(i=0;i<nT;i++)
		{
		  for(j=0;j<3;j++)
		    if(fprintf(fp,"%d ",mesh.getFaceCorner(i,j))<1)
		      { err = -4; break; }
		  fprintf(fp,"\n");
		}
	    }
	}
      else
	err = -5;
    }

  return err;
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::read
//
//---------------------------------------------------------------------------

bool MPG_ClustSurf::read(char *fname)
{
  clear();

  int err = 0;
  if(fname==(char*)0)
    err = -1;
  else
    {
      FILE *fp = fopen(fname,"r");
      if(fp==(FILE*)0)
	{ err = -2; _error("MPG_ClustSurf::read(-2)"); }
      else
	{
	  int h,i,j,k,l;
	  float x;

	  // READ HEADER
	  err = fscanf(fp,"# multi-level clustered mesh");

	  // READ PROPERTY OPTIONS
	  // while((i=getc(fp))!='\n' && i!=EOF);
	  _coordBinding    = PB_PER_VERTEX;
	  _normalBinding   = PB_NONE;
	  _colorBinding    = PB_NONE;
	  _texCoordBinding = PB_NONE;
	  {
	    // get rest of comment line
	    vector<char> str;
	    char c;
	    while(fscanf(fp,"%c",&c)==1 && c!='\n')
	      str.push_back(c);
	    str.push_back('\0');
	    // parse collected text
	    int n=str.size()-1,i=0,j=0;
	    char* token;
	    do {
	      token = getToken(str,i,j);
	      if(!strcmp(token,"coord"))
		{
		  token = getToken(str,i,j);
		  setBinding(token,_coordBinding);
		}
	      else if(!strcmp(token,"normal"))
		{
		  token = getToken(str,i,j);
		  setBinding(token,_normalBinding);
		}
	      else if(!strcmp(token,"color"))
		{
		  token = getToken(str,i,j);
		  setBinding(token,_colorBinding);
		}
	      else if(!strcmp(token,"texCoord"))
		{
		  token = getToken(str,i,j);
		  setBinding(token,_texCoordBinding);
		}
	    } while(i<n);
	  }

	  if(_coordBinding!=PB_PER_VERTEX || _texCoordBinding==PB_PER_FACE)
	    { err = -3; _error("MPG_ClustSurf::read(bindings)"); }

	  // READ NUMBER OF LEVELS
	  int nL = 0;
	  if(fscanf(fp,"nL=%d",&nL)<1 || nL<1)
	    err = -4;

	  // READ VERTEX DIMENSION
	  int nD = 0;
	  while((i=getc(fp))!='\n' && i!=EOF);
	  if(fscanf(fp,"nD=%d",&nD)<1 || nD<1)
	    err = -5;
	  _nD = nD;

	  // READ HIGHEST RESOLUTION MESH
	  char fType[4];
	  while((i=getc(fp))!='\n' && i!=EOF);
	  if(fscanf(fp,"%3c",fType)<3)
	    err = -6;
	  fType[3] = '\0';
	  if(!strcmp(fType,"nT="))
	    err = readTriangles(fp);
	  else if(!strcmp(fType,"nF="))
	    err = readFaces(fp);
	  else
	    err = -7;

	  // make a copy of the mesh to keep track of
	  // collapsed faces and number of faces
	  vector<int> coordIndex;

	  bool hasPropertiesPerFace =
	    ( _normalBinding  ==PB_PER_FACE   ||
	      _normalBinding  ==PB_PER_CORNER ||
	      _colorBinding   ==PB_PER_FACE   ||
	      _colorBinding   ==PB_PER_CORNER ||
	      _texCoordBinding==PB_PER_FACE   ||
	      _texCoordBinding==PB_PER_CORNER )?true:false;

	  if(hasPropertiesPerFace==true)
	    coordIndex = _coordIndex;

	  // read vertex coordinates and clustering maps
	  for(l=0;l<nL;l++)
	    {
	      int nV = 0;
	      while((i=getc(fp))!='\n' && i!=EOF);
	      if(fscanf(fp,"nV=%d",&nV)<1 || nV<=0)
		{ err = -8; break; }
	      else
		{
		  int mapsTo;
		  for(h=i=0;i<nV;i++)
		    {
		      // coord
		      for(j=0;j<nD;j++)
			{
			  if(fscanf(fp,"%f",&x)<1)
			    { err = -90; break; }
			  _coord.push_back(x);
			}

		      // read normal[i] if per vertex
		      if(_normalBinding==PB_PER_VERTEX)
			for(j=0;j<nD;j++)
			  {
			    if(fscanf(fp,"%f",&x)<1)
			      { err = -91; break; }
			    _normal.push_back(x);
			  }

		      // read color[i] if per vertex
		      if(_colorBinding==PB_PER_VERTEX)
			for(j=0;j<nD;j++)
			  {
			    if(fscanf(fp,"%f",&x)<1)
			      { err = -92; break; }
			    _color.push_back(x);
			  }

		      // read texCoord[i] if per vertex
		      if(_texCoordBinding==PB_PER_VERTEX)
			for(j=0;j<nD;j++)
			  {
			    if(fscanf(fp,"%f",&x)<1)
			      { err = -93; break; }
			    _texCoord.push_back(x);
			  }

		      // read coordMap[i] if not last level
		      if(l<nL-1)
			{
			  if(!err && fscanf(fp,"%d",&mapsTo)<1)
			    err = -94;
			  _coordMap.push_back(mapsTo);
			}
		    }
    		}

	      // read properties per face and per corner

	      if(hasPropertiesPerFace==true)
		{
		  MPG_Mesh m(coordIndex);
		  vector<int> nextCoordIndex;

		  int nF = m.getNumberOfFaces();
		  vector<int> fi;
		  for(h=i=0;i<nF;i++)
		    {
		      int nfi = m.getNumberOfFaceCorners(i);

		      // read normal[i] if per vertex
		      if(_normalBinding==PB_PER_FACE)
			for(j=0;j<3;j++)
			  {
			    if(fscanf(fp,"%f",&x)<1)
			      { err = -100; break; }
			    _normal.push_back(x);
			  }
		      else if(_normalBinding==PB_PER_CORNER)
			for(k=0;k<nfi;k++)
			  for(j=0;j<3;j++)
			    {
			      if(fscanf(fp,"%f",&x)<1)
				{ err = -101; break; }
			      _normal.push_back(x);
			    }

		      // read color[i] if per vertex
		      if(_colorBinding==PB_PER_FACE)
			for(j=0;j<3;j++)
			  {
			    if(fscanf(fp,"%f",&x)<1)
			      { err = -8; break; }
			    _color.push_back(x);
			  }
		      else if(_colorBinding==PB_PER_CORNER)
			for(k=0;k<nfi;k++)
			  for(j=0;j<3;j++)
			    {
			      if(fscanf(fp,"%f",&x)<1)
				{ err = -102; break; }
			      _color.push_back(x);
			    }

		      // read texCoord[i] if per vertex
		      if(_texCoordBinding==PB_PER_FACE)
			for(j=0;j<2;j++)
			  {
			    if(fscanf(fp,"%f",&x)<1)
			      { err = -103; break; }
			    _texCoord.push_back(x);
			  }
		      else if(_texCoordBinding==PB_PER_CORNER)
			for(k=0;k<nfi;k++)
			  for(j=0;j<2;j++)
			    {
			      if(fscanf(fp,"%f",&x)<1)
				{ err = -104; break; }
			      _texCoord.push_back(x);
			    }

		      // apply map to face and determine if face collapses
		      {
			for(j=0;j<nfi;j++)
			  {
			    int fij    = m.getFaceCorner(i,j);
			    fij = getMap(l,fij);
			    if(j==0 || fi.back()!=fij)
			      fi.push_back(fij);
			  }
			if(fi.front()==fi.back())
			  fi.pop_back();
		      }
		      // still need to find other repetitions !!!
		      if(fi.size()>=3)
			{
			  for(j=0;j<fi.size();j++)
			    nextCoordIndex.push_back(fi[j]);
			  nextCoordIndex.push_back(-1);
			}
		    }

		  coordIndex.swap(nextCoordIndex);
		  nextCoordIndex.erase(nextCoordIndex.begin(),nextCoordIndex.end());
		}
	      closeLevel();

	    }

	  int nV0 = getNumberOfCoord(0);
	}
    }

  return (err==0)?true:false;
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::write
//
//---------------------------------------------------------------------------

bool MPG_ClustSurf::write(char *fname)
{
  if(fname==(char*)0) return false;

  FILE *fp = fopen(fname,"w");
  if(fp==(FILE*)0) return false;

  int h,i,j,l;

  // WRITE HEADER
  fprintf(fp,"# multi-level clustered mesh");

  // WRITE BINDINGS
  {
    if(_coordBinding!=PB_NONE)
      fprintf(fp," coord %s",
	      (_coordBinding==PB_PER_VERTEX)?"perVertex":
	      (_coordBinding==PB_PER_FACE  )?"perFace":
	      (_coordBinding==PB_PER_CORNER)?"perCorner":
	      "none");
    if(_normalBinding!=PB_NONE)
      fprintf(fp," normal %s",
	      (_normalBinding==PB_PER_VERTEX)?"perVertex":
	      (_normalBinding==PB_PER_FACE  )?"perFace":
	      (_normalBinding==PB_PER_CORNER)?"perCorner":
	      "NONE");
    if(_colorBinding!=PB_NONE)
      fprintf(fp," color %s",
	      (_colorBinding==PB_PER_VERTEX)?"perVertex":
	      (_colorBinding==PB_PER_FACE  )?"perFace":
	      (_colorBinding==PB_PER_CORNER)?"perCorner":
	      "NONE");
    if(_texCoordBinding!=PB_NONE)
      fprintf(fp," texCoord %s",
	      (_texCoordBinding==PB_PER_VERTEX)?"perVertex":
	      (_texCoordBinding==PB_PER_CORNER)?"perCorner":
	      "NONE");
  }
  fprintf(fp,"\n");

  // WRITE TOTAL NUMBER OF LEVELS
  int nL = getNumberOfLevels();
  fprintf(fp,"nL=%d\n",nL);

  // WRITE VERTEX DIMENSION
  int nD = getDimension();
  fprintf(fp,"nD=%d\n",nD);

  // WRITE HIGHEST RESOLUTION MESH
  MPG_Mesh mesh(_coordIndex);
  if(mesh.isTriangulated())
    {
      int nT = mesh.getNumberOfFaces();
      fprintf(fp,"nT=%d\n",nT);
      for(i=0;i<nT;i++)
	for(j=0;j<3;j++)
	  fprintf(fp,"%d%c",mesh.getFaceCorner(i,j),(j<2)?' ':'\n');
    }
  else
    {
      int nF = mesh.getNumberOfFaces();
      fprintf(fp,"nF=%d\n",nF);
      for(i=0;i<nF;i++)
	{
	  int nfi = mesh.getNumberOfFaceCorners(i);
	  fprintf(fp,"%d ",nfi);
	  for(j=0;j<nfi;j++)
	    fprintf(fp,"%d%c",mesh.getFaceCorner(i,j),(j<nfi-1)?' ':'\n');
	}
    }

  // make a copy of the mesh to keep track of
  // collapsed faces and number of faces
  vector<int> coordIndex;

  bool hasPropertiesPerFace =
    ( _normalBinding  ==PB_PER_FACE   ||
      _normalBinding  ==PB_PER_CORNER ||
      _colorBinding   ==PB_PER_FACE   ||
      _colorBinding   ==PB_PER_CORNER ||
      _texCoordBinding==PB_PER_FACE   ||
      _texCoordBinding==PB_PER_CORNER )?true:false;

  if(hasPropertiesPerFace==true)
    coordIndex = _coordIndex;

  int nV = 0;
  for(l=0;l<nL;l++)
    {
      nV = getNumberOfCoord(l);

      fprintf(fp,"nV=%d\n",nV);
      for(i=0;i<nV;i++)
	{
	  for(j=0;j<nD;j++)
	    fprintf(fp,"%10.4f ",getCoord(l,i,j));

	  if(_normalBinding==PB_PER_VERTEX)
	  for(j=0;j<3;j++)
	    fprintf(fp,"%10.4f ",getNormal(l,i,j));

	  if(_colorBinding==PB_PER_VERTEX)
	  for(j=0;j<3;j++)
	    fprintf(fp,"%10.4f ",getColor(l,i,j));

	  if(_texCoordBinding==PB_PER_VERTEX)
	  for(j=0;j<2;j++)
	    fprintf(fp,"%10.4f ",getTexCoord(l,i,j));

	  if(l<nL-1)
	    fprintf(fp,"%d",getMap(l,i));

	  fprintf(fp,"\n");
	}

      if(hasPropertiesPerFace==true)
	{
	  MPG_Mesh m(coordIndex);
	  vector<int> nextCoordIndex;

	  int nF = m.getNumberOfFaces();
	  vector<int> fi;

	  int iN  = 0;
	  int iC  = 0;
	  int iTC = 0;

	  for(i=0;i<nF;i++)
	    {
	      int nfi = m.getNumberOfFaceCorners(i);

	      // write normal[i] if per vertex
	      if(_normalBinding==PB_PER_FACE)
		{
		  for(j=0;j<3;j++)
		    fprintf(fp,"%10.4f ",getNormal(l,iN,j));
		  iN++;
		}
	      else if(_normalBinding==PB_PER_CORNER)
		for(h=0;h<nfi;h++)
		  {
		    for(j=0;j<3;j++)
		      fprintf(fp,"%10.4f ",getNormal(l,iN,j));
		    iN++;
		  }

	      // write color[i] if per vertex
	      if(_colorBinding==PB_PER_FACE)
		{
		  for(j=0;j<3;j++)
		    fprintf(fp,"%10.4f ",getColor(l,iC,j));
		  iC++;
		}
	      else if(_colorBinding==PB_PER_CORNER)
		for(h=0;h<nfi;h++)
		  {
		    for(j=0;j<3;j++)
		      fprintf(fp,"%10.4f ",getColor(l,iC,j));
		    iC++;
		  }

	      // write texCoord[i] if per vertex
	      if(_texCoordBinding==PB_PER_FACE)
		{
		  for(j=0;j<2;j++)
		    fprintf(fp,"%10.4f ",getTexCoord(l,iTC,j));
		  iTC++;
		}
	      else if(_texCoordBinding==PB_PER_CORNER)
		for(h=0;h<nfi;h++)
		  {
		    for(j=0;j<2;j++)
		      fprintf(fp,"%10.4f ",getTexCoord(l,iTC,j));
		    iTC++;
		  }

	      fprintf(fp,"\n");

	      // apply map to face and determine if face collapses
	      {
		for(j=0;j<nfi;j++)
		  {
		    int fij = m.getFaceCorner(i,j);
		    fij = getMap(l,fij);
		    if(j==0 || fi.back()!=fij)
		      fi.push_back(fij);
		  }
		if(fi.front()==fi.back())
		  fi.pop_back();
	      }
	      // still need to find other repetitions !!!
	      if(fi.size()>=3)
		{
		  for(j=0;j<fi.size();j++)
		    nextCoordIndex.push_back(fi[j]);
		  nextCoordIndex.push_back(-1);
		}
	    }

	  coordIndex.swap(nextCoordIndex);
	  nextCoordIndex.erase(nextCoordIndex.begin(),nextCoordIndex.end());
	}
    }

  fclose(fp);

  return true;			// no error
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::write
//
//---------------------------------------------------------------------------

bool MPG_ClustSurf::write(int level, const char *bn)
{
  int err = 0;

  char* fname = new char[strlen(bn)+16];
  sprintf(fname,"%s-%d.wrl",bn,level);

  //  PWL_Surf  srf(3);
  //  err = getSurf(srf,level);
  //  if(!err)
  //    {
  //      FILE *fp = fopen(fname,"w");
  //      if(fp==(FILE*)0)
  //	err = -1;
  //      else
  //	err = srf.writeWRL(fname);
  //    }

  delete [] fname;

  return (!err)?true:false;
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::_error
//
//---------------------------------------------------------------------------

void MPG_ClustSurf::_error(char *msg)
{
  fprintf(logfp,"MPG_ClustSurf::_error(\"");
  if(msg!=(char*)0)
    fprintf(logfp,"%s",msg);
  fprintf(logfp,"\")\n");
}

//---------------------------------------------------------------------------
//
//   *MPG_ClustSurf::getCoordIndex
//
//---------------------------------------------------------------------------

bool MPG_ClustSurf::getCoordIndex(int level, vector<int>& coordIndex)
{
  coordIndex.erase(coordIndex.begin(),coordIndex.end());
  int nL = getNumberOfLevels();
  if(!(0<=level && level<nL)) return false;

  coordIndex = _coordIndex;

  vector<int> newCoordIndex;
  vector<int> fi;

  // apply map
  int i,j,k,l;
  for(l=0;l<level;l++)
    {
      int* cmap = getMap(l);
      for(k=i=0;i<coordIndex.size();i++)
	if(coordIndex[i]>=0)
	  {
	    int fij = cmap[coordIndex[i]];
	    if(i==k || fi.back()!=fij)
	      fi.push_back(fij);
	  }
	else
	  {
	    if(fi.size()>0 && (fi.front()==fi.back()))
	      fi.pop_back();
	    if(fi.size()>=3)
	      {
		for(j=0;j<fi.size();j++)
		  newCoordIndex.push_back(fi[j]);
		newCoordIndex.push_back(-1);
	      }
	    k = i+1;
	    fi.erase(fi.begin(),fi.end());
	  }
      coordIndex.swap(newCoordIndex);
      newCoordIndex.erase(newCoordIndex.begin(),newCoordIndex.end());
    }

  return true;
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::getIfs
//
//---------------------------------------------------------------------------

bool MPG_ClustSurf::getIfs(MPG_Ifs& ifs, int level)
{
  bool success = false;

  ifs.clear();

  int nL = getNumberOfLevels();

  if(nL>0)
    {
      float* x;
      int h,i,j;

      if(level< 0 ) level = 0;
      if(level>=nL) level = nL-1;

      ifs.setCcw(_ccw);
      ifs.setConvex(_convex);
      ifs.setCreaseAngle(_creaseAngle);
      ifs.setSolid(_solid);

      // set ifs._coordIndex
      int nF = 0;		// number of faces
      int nC = 0;		// number of corners
      vector<int>& coordIndex = ifs.getCoordIndex();

#ifdef COLOR_COLLAPSED_FACES
      vector<bool> faceCollapses;
#endif // COLOR_COLLAPSED_FACES

      {
	int h,i,nfi,fij,j,k;

	vector<int> fi;
	for(h=i=0;i<_coordIndex.size();i++)
	  if(_coordIndex[i]<0)
	    {
	      fi.erase(fi.begin(),fi.end());
	      nfi = i-h;

	      for(j=0;j<nfi;j++)
		{
		  // fij = mesh.getFaceCorner(i,j);
		  fij = _coordIndex[h+j];
		  for(k=0;k<level;k++)
		    fij = getMap(k,fij);
		  if(j==0 || fij!=fi.back())
		    fi.push_back(fij);
		  // still need to split face with
		  // duplicate vertices into two or more faces !!!
		  // OK for triangular face, though
		}
	      if(fi.front()==fi.back())
		fi.pop_back();
	      if(fi.size()>=3)
		{
                  nF++;
		  nC += fi.size();
		  for(j=0;j<fi.size();j++)
		    coordIndex.push_back(fi[j]);
		  coordIndex.push_back(-1);

#ifdef COLOR_COLLAPSED_FACES

                  bool thisFaceCollapses = false;
                  if(level<nL)
                    {
                      vector<int> fi_next;
                      for(j=0;j<fi.size();j++)
                        {
                          fij = getMap(level,fi[j]);
                          if(j==0 || fij!=fi_next.back())
                            fi_next.push_back(fij);
                        }
                      if(fi_next.front()==fi_next.back())
                        fi_next.pop_back();
                      if(fi_next.size()<fi.size())
                        thisFaceCollapses = true;
                    }
                  faceCollapses.push_back(thisFaceCollapses);

#endif // COLOR_COLLAPSED_FACES

                }

	      h = i+1;
	    }
      }

      int nV = getNumberOfCoord(level);
      int nD = getDimension();

      // always COORD_PER_VERTEX
      vector<float>& coord         = ifs.getCoord();
      x = getCoord(level);
      for(i=0;i<nV;i++)
	for(j=0;j<nD;j++)
	  coord.push_back(*x++);

      // if(normal.size()==0)
      //   NO_NORMALS
      // else if(normalPerVertex==FALSE)
      //   if(normalIndex.size()>0)
      //     NORMAL_PER_FACE_INDEXED;
      //     normalIndex.size()==getNumberOfFaces()
      //   else
      //     NORMAL_PER_FACE;
      //     normal.size()/3==getNumberOfFaces()
      // else // if(normalPerVertex==TRUE)
      //   if(normalIndex.size()>0)
      //     NORMAL_PER_CORNER_INDEXED;
      //     normalIndex.size()==coordIndex.size()
      //   else
      //     NORMAL_PER_VERTEX;
      //     normal.size()/3==coord.size()/3

      vector<float>& normal        = ifs.getNormal();
      vector<int>&   normalIndex   = ifs.getNormalIndex();
      int            nNormal       = getNumberOfNormal(level);
      switch(getNormalBinding())
	{
	case PB_PER_VERTEX:
	  ifs.setNormalPerVertex(true);
	  x = getNormal(level);
	  for(i=0;i<nV;i++)
	    for(j=0;j<3;j++)
	      normal.push_back(*x++);
	  break;

	case PB_PER_FACE:
	  ifs.setNormalPerVertex(false);
	  x = getNormal(level);
	  for(i=0;i<nF;i++)
	    for(j=0;j<3;j++)
	      normal.push_back(*x++);

	  // try to make it indexed to get rid of repetition ?

	  break;

	case PB_PER_CORNER:
	  ifs.setNormalPerVertex(true);
	  x = getNormal(level);
	  for(i=0;i<nC;i++)
	    for(j=0;j<3;j++)
	      normal.push_back(*x++);
	  h = 0;
	  for(i=0;i<coordIndex.size();i++)
	    if(coordIndex[i]>=0)
	      normalIndex.push_back(h++);
	    else
	      normalIndex.push_back(-1);

	  // remap to try to get rid of repetition ?

	  break;

	case PB_NONE:
	default:
	  break;
	}

#ifdef COLOR_COLLAPSED_FACES

      vector<float>& color         = ifs.getColor();
      vector<int>&   colorIndex    = ifs.getColorIndex();

      // white for not collapsed faces
      color.push_back(1.0);
      color.push_back(1.0);
      color.push_back(1.0);
      // red for collapsed faces
      color.push_back(1.0);
      color.push_back(0.0);
      color.push_back(0.0);

      // color per face
      ifs.setColorPerVertex(false);

      for(i=0;i<nF;i++)
        {
          j = (faceCollapses[i]==true)?1:0;
          colorIndex.push_back(j);
        }

#else // COLOR_COLLAPSED_FACES

      // if(color.size()==0)
      //   NO_COLORS
      // else if(colorPerVertex==FALSE)
      //   if(colorIndex.size()>0)
      //     COLOR_PER_FACE_INDEXED;
      //     colorIndex.size()==getNumberOfFaces()
      //   else
      //     COLOR_PER_FACE;
      //     color.size()/3==getNumberOfFaces()
      // else // if(colorPerVertex==TRUE)
      //   if(colorIndex.size()>0)
      //     COLOR_PER_CORNER_INDEXED;
      //     colorIndex.size()==coordIndex.size()
      //   else
      //     COLOR_PER_VERTEX;
      //     color.size()/3==coord.size()/3

      vector<float>& color         = ifs.getColor();
      vector<int>&   colorIndex    = ifs.getColorIndex();
      int            nColor        = getNumberOfColor(level);
      switch(getColorBinding())
	{
	case PB_PER_VERTEX:
	  ifs.setColorPerVertex(true);
	  x = getColor(level);
	  for(i=0;i<nV;i++)
	    for(j=0;j<3;j++)
	      color.push_back(*x++);
	  break;

	case PB_PER_FACE:
	  ifs.setColorPerVertex(false);
	  x = getColor(level);
	  for(i=0;i<nF;i++)
	    for(j=0;j<3;j++)
	      color.push_back(*x++);

	  // try to make it indexed to get rid of repetition ?

	  break;

	case PB_PER_CORNER:
	  ifs.setColorPerVertex(true);
	  x = getColor(level);
	  for(i=0;i<nC;i++)
	    for(j=0;j<3;j++)
	      color.push_back(*x++);
	  h = 0;
	  for(i=0;i<coordIndex.size();i++)
	    if(coordIndex[i]>=0)
	      colorIndex.push_back(h++);
	    else
	      colorIndex.push_back(-1);

	  // remap to try to get rid of repetition ?

	  break;

	case PB_NONE:
	default:
	  break;
	}

#endif // COLOR_COLLAPSED_FACES

      // if(texCoord.size()==0)
      //   NO_TEX_COORD
      // else if(texCoordIndex.size()>0)
      //   TEX_COORD_PER_CORNER_INDEXED;
      //   texCoordIndex.size()==coordIndex.size()
      // else
      //   TEX_COORD_PER_VERTEX;
      //   texCoord.size()/3==coord.size()/3

      vector<float>& texCoord      = ifs.getTexCoord();
      vector<int>&   texCoordIndex = ifs.getTexCoordIndex();
      int            nTexCoord     = getNumberOfTexCoord(level);
      switch(getTexCoordBinding())
	{
	case PB_PER_VERTEX:
	  x = getTexCoord(level);
	  for(i=0;i<nV;i++)
	    for(j=0;j<3;j++)
	      texCoord.push_back(*x++);
	  break;

	case PB_PER_FACE:
	  break;

	case PB_PER_CORNER:
	  x = getTexCoord(level);
	  for(i=0;i<nC;i++)
	    for(j=0;j<3;j++)
	      texCoord.push_back(*x++);
	  h = 0;
	  for(i=0;i<coordIndex.size();i++)
	    if(coordIndex[i]>=0)
	      texCoordIndex.push_back(h++);
	    else
	      texCoordIndex.push_back(-1);

	  // remap to try to get rid of repetition ?

	  break;

	case PB_NONE:
	default:
	  break;
	}

      success = true;
    }

  return success;
}


//---------------------------------------------------------------------------
//
//   appendValues
//
//---------------------------------------------------------------------------

static void appendValues(int dimension,
                         vector<float>& _value,
			 vector<int>&   _index,
			 vector<float>&  val)
{
  int i,j,k;
  int nV = (_index.size()>0)?_index.size():_value.size()/dimension;
  for(i=0;i<nV;i++)
    {
      j = (_index.size()>0) ? _index[i]:i;
      if(j>=0)
	{
	  float* vi = &(_value[j*dimension]);
	  for(k=0;k<dimension;k++)
	    val.push_back(vi[k]);
	}
    }
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::operator=
//
//---------------------------------------------------------------------------

MPG_ClustSurf& MPG_ClustSurf::operator=(MPG_Ifs& ifs)
{
  if(COM_Dbg_on>0)
    fprintf(logfp,"MPG_ClustSurf::operator= {\n");

  clear();

  // IFS variabales

  _ccw               = ifs.getCcw();
  _convex            = ifs.getConvex();
  _creaseAngle       = ifs.getCreaseAngle();
  _solid             = ifs.getSolid();

  // FACES
  _coordIndex = ifs.getCoordIndex();
#ifdef WITH_MESH
  mesh.set(_coordIndex);
#endif // WITH_MESH

  // COORD
  _coordBinding = ifs.getCoordBinding();
  vector<float>& coord = ifs.getCoord();
  if(_coordBinding!=PB_NONE)
  {
    int nV = coord.size()/3;
    int i,j;
    for(i=0;i<nV;i++)
    for(j=0;j<3;j++)
    _coord.push_back(coord[3*i+j]);
  }

  // NORMAL
  vector<float>& normal      = ifs.getNormal();
  vector<int>&   normalIndex = ifs.getNormalIndex();
  _normalBinding             = ifs.getNormalBinding();
  if(_normalBinding!=PB_NONE)
    appendValues(3,normal,normalIndex,_normal);

  // COLOR
  vector<float>& color      = ifs.getColor();
  vector<int>&   colorIndex = ifs.getColorIndex();
  _colorBinding             = ifs.getColorBinding();
  if(_colorBinding!=PB_NONE)
    appendValues(3,color,colorIndex,_color);

  // TEXCOORD
  vector<float>& texCoord      = ifs.getTexCoord();
  vector<int>&   texCoordIndex = ifs.getTexCoordIndex();
  _texCoordBinding             = ifs.getTexCoordBinding();
  if(_texCoordBinding!=PB_NONE)
    appendValues(2,texCoord,texCoordIndex,_texCoord);

  closeLevel();

  if(COM_Dbg_on>0)
    fprintf(logfp,"}\n");

  return *this;
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::operator=
//
//---------------------------------------------------------------------------

MPG_ClustSurf& MPG_ClustSurf::operator=(MPG_IfsQ_EXT& ifsq)
{
  MPG_Ifs& ifs = ifsq;
  *this = ifs;
  setCoordQuantizer(ifsq.getCoordQuantizer());
  setNormalQuantizer(ifsq.getNormalQuantizer());
  setColorQuantizer(ifsq.getColorQuantizer());
  setTexCoordQuantizer(ifsq.getTexCoordQuantizer());
  return *this;
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::prependLevel
//
//---------------------------------------------------------------------------

bool MPG_ClustSurf::prependLevel(vector<int>&   coordIndex,
				 vector<int>&   coordMap,
				 vector<float>& coord,
				 vector<float>& normal,
				 vector<float>& color,
				 vector<float>& texCoord)
{
  bool success = false;

  // WARNING !!!
  // neither sizes nor bindings are checked
  // if they are not correct, something else may crash later

  _coordIndex = coordIndex;
#ifdef WITH_MESH
  mesh.set(_coordIndex);
#endif // WITH_MESH

  int nD        = getDimension();
  int nCoord    = coord.size()/nD;
  int nNormal   = normal.size()/3;
  int nColor    = color.size()/3;
  int nTexCoord = texCoord.size()/2;

  int i;

  int n1 = nCoord;
  _coordMap.insert(_coordMap.end(),(size_t)n1,-1);
  for(i=_coordMap.size();i>=n1;i--)
    _coordMap[i] = _coordMap[i-n1];
  for(;i>=0;i--)
    _coordMap[i] = coordMap[i];

  {
    int n1=nCoord*nD;
    _coord.insert(_coord.end(),(size_t)n1,0.0);
    for(i=_coord.size();i>=n1;i--)
      _coord[i] = _coord[i-n1];
    for(;i>=0;i--)
      _coord[i] = coord[i];
  }

  _coordFirst.push_back(0);
  for(i=_coordFirst.size()-1;i>0;i--)
    { _coordFirst[i] = _coordFirst[i-1]+nCoord; }

  if(nNormal>0)
    {
      int n1=nNormal*nD;
      _normal.insert(_normal.end(),(size_t)n1,0.0);
      for(i=_normal.size();i>=n1;i--)
	_normal[i] = _normal[i-n1];
      for(;i>=0;i--)
	_normal[i] = normal[i];
    }

  _normalFirst.push_back(0);
  for(i=_normalFirst.size()-1;i>0;i--)
    { _normalFirst[i] = _normalFirst[i-1]+nNormal; }

  if(nColor>0)
    {
      int n1=nColor*nD;
      _color.insert(_color.end(),(size_t)n1,0.0);
      for(i=_color.size();i>=n1;i--)
	_color[i] = _color[i-n1];
      for(;i>=0;i--)
	_color[i] = color[i];
    }

  _colorFirst.push_back(0);
  for(i=_colorFirst.size()-1;i>0;i--)
    { _colorFirst[i] = _colorFirst[i-1]+nColor; }

  if(nTexCoord>0)
    {
      int n1=nTexCoord*nD;
      _texCoord.insert(_texCoord.end(),(size_t)n1,0.0);
      for(i=_texCoord.size();i>=n1;i--)
	_texCoord[i] = _texCoord[i-n1];
      for(;i>=0;i--)
	_texCoord[i] = texCoord[i];
    }

  _texCoordFirst.push_back(0);
  for(i=_texCoordFirst.size()-1;i>0;i--)
    { _texCoordFirst[i] = _texCoordFirst[i-1]+nTexCoord; }

  success = true;

  return success;
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::closeLevel
//
//---------------------------------------------------------------------------

void MPG_ClustSurf::closeLevel()
{
  // at this point
  // vertex coordinates of next level have been pushed onto _coord
  //
  // we should check if the lengths of the properties are ok
  // but we are not doing it for now

  _coordFirst.push_back(_coord.size()/_nD);
  _normalFirst.push_back(_normal.size()/3);
  _colorFirst.push_back(_color.size()/3);
  _texCoordFirst.push_back(_texCoord.size()/2);
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::appendLevel
//
//---------------------------------------------------------------------------

void MPG_ClustSurf::appendLevel(vector<int>&   connInfo,
				vector<float>& coord,
				vector<float>& normal,
				vector<float>& color,
				vector<float>& texCoord)
{
  // WARNING !!!
  // neither sizes nor bindings are checked
  // if they are not correct, something else may crash later

  // connectivity
  if(getNumberOfLevels()==0)
    {
      if(connInfo.size()>0)
	_coordIndex.insert(_coordIndex.end(),connInfo.begin(),connInfo.end());
    }
  else
    {
      if(connInfo.size()>0)
	_coordMap.insert(_coordMap.end(),connInfo.begin(),connInfo.end());
    }

  // properties
  if(coord.size()>0)
    _coord.insert(_coord.end(),coord.begin(),coord.end());
  if(normal.size()>0)
    _normal.insert(_normal.end(),normal.begin(),normal.end());
  if(color.size()>0)
    _color.insert(_color.end(),color.begin(),color.end());
  if(texCoord.size()>0)
    _texCoord.insert(_texCoord.end(),texCoord.begin(),texCoord.end());
  closeLevel();
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::appendLevel
//
//---------------------------------------------------------------------------

void MPG_ClustSurf::appendLevel(vector<int>& connInfo,
				int          nCoord,
				int          nNormal,
				int          nColor,
				int          nTexCoord)
{
  // WARNING !!!
  // neither sizes nor bindings are checked
  // if they are not correct, something else may crash later

  // connectivity
  if(getNumberOfLevels()==0)
    {
      if(connInfo.size()>0)
	_coordIndex.insert(_coordIndex.end(),connInfo.begin(),connInfo.end());
    }
  else
    {
      if(connInfo.size()>0)
	_coordMap.insert(_coordMap.end(),connInfo.begin(),connInfo.end());
    }

  // properties
  if(nCoord>0)
    _coord.insert(_coord.end(),(size_t)(_nD*nCoord),0.0);
  if(nNormal>0)
    _normal.insert(_normal.end(),(size_t)(3*nNormal),0.0);
  if(nColor>0)
    _color.insert(_color.end(),(size_t)(3*nColor),0.0);
  if(nTexCoord>0)
    _texCoord.insert(_texCoord.end(),(size_t)(2*nTexCoord),0.0);

  closeLevel();
}

//---------------------------------------------------------------------------
//
//   MPG_ClustSurf::deleteLastLevel
//
//---------------------------------------------------------------------------

void MPG_ClustSurf::deleteLastLevel()
{
  int nL = getNumberOfLevels();
  if(nL==1)
    clear();
  else if(nL>1)
    {
      int last;

      _coordFirst.pop_back();
      last = _coordFirst.back();
      _coord.erase(_coord.begin()+last,_coord.end());

      _normalFirst.pop_back();
      last = _normalFirst.back();
      _normal.erase(_normal.begin(),_normal.end());

      _colorFirst.pop_back();
      last = _colorFirst.back();
      _color.erase(_color.begin()+last,_color.end());

      _texCoordFirst.pop_back();
      last = _texCoordFirst.back();
      _texCoord.erase(_texCoord.begin()+last,_texCoord.end());

      if(--nL>0)
	{
	  last = _coordFirst[nL-1];
	  _coordMap.erase(_coordMap.begin()+last,_coordMap.end());
	}
    }
}
