//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------


#include <stdio.h>
#include <stdlib.h>
#include "MPG_Ifs.h"

#define VRML_HEADER "#VRML V2.0 utf8"

//---------------------------------------------------------------------------
//
//   getToken
//
//---------------------------------------------------------------------------

static inline int getToken(FILE* fp, vector<char>& tkn, int skipComments=1)
{
  char c='\0';
  do {
    tkn.erase(tkn.begin(),tkn.end());
    // skip blank space
    while(fscanf(fp,"%c",&c)==1)
      if(!(c==' ' || c=='\t' || c=='\n' || c==',' || c=='\015'))  // \015=="^M"
	{ tkn.push_back(c); break; }
    // collect token characters
	  if (!(c==']' || c=='}' || c=='{' || c=='['))	//MJH 081800 : crashes when closing bracket is attached to other characters
	  {
		while((fscanf(fp,"%c",&c)==1) &&
			!(c==' ' || c=='\t' || c=='\n' || c==',' || c=='\015') )// c=="^M"
		{
			//MJH 081800 : if the next character is a closing bracket, this should be in the next token
			if ((c==']' || c=='}'))	
			{	(fp->_cnt)++;
				(fp->_ptr)--;
				break;
			}
			tkn.push_back(c);
		}
	  }
    // if comment, get the rest of the line, including blank spaces
    if(tkn.size()>0 && tkn[0]=='#')
      {
	// last blank space character read is not in tkn yet
	if(c!='\n')
	  tkn.push_back(c);
	while(fscanf(fp,"%c",&c)==1 && c!='\n')
	  tkn.push_back(c);
      }
    // terminate the string
    tkn.push_back('\0');
  } while(skipComments && tkn[0]=='#');
  return strlen(tkn.begin());
}

//---------------------------------------------------------------------------
//
//   MPG_Ifs::read
//
//---------------------------------------------------------------------------

bool MPG_Ifs::read(char* fname)
{
  clear();
  bool success = false;
  FILE* fp = fopen(fname,"r");
  if(fp!=(FILE*)0)
    {
      char header[16];
      memset(header,'\0',16);
      fscanf(fp,"%15c",header);
      if(!strcmp(header,VRML_HEADER))
	{
	  vector<char> tkn;

	  // get next token
	  int  nc=0;
	  while((nc=getToken(fp,tkn))>0) {

	    // until the first IndexedFaceSet is found
	    if(!strcmp(tkn.begin(),"IndexedFaceSet"))
	      {
		// next token must be "{"
		nc=getToken(fp,tkn);
		if(nc<=0 || strcmp(tkn.begin(),"{"))
		  break;

		while((nc=getToken(fp,tkn))>0) {

		  if(!strcmp(tkn.begin(),"ccw") ||
		     !strcmp(tkn.begin(),"convex") ||
		     !strcmp(tkn.begin(),"solid") ||
		     !strcmp(tkn.begin(),"colorPerVertex") ||
		     !strcmp(tkn.begin(),"normalPerVertex"))
		    {
		      int type =
			(!strcmp(tkn.begin(),"ccw"))?1:
			(!strcmp(tkn.begin(),"convex"))?2:
			(!strcmp(tkn.begin(),"solid"))?3:
			(!strcmp(tkn.begin(),"colorPerVertex"))?4:
			(!strcmp(tkn.begin(),"normalPerVertex"))?5:
			0;

		      bool value = true;
		      // next token should be TRUE or FALSE
		      nc=getToken(fp,tkn);
		      if(nc<=0)
			break;
		      else if(!strcmp(tkn.begin(),"TRUE"))
			value = true;
		      else if(!strcmp(tkn.begin(),"FALSE"))
			value = false;
		      else
			break;
		      switch(type)
			{
			case 1: _ccw             = value; break;
			case 2: _convex          = value; break;
			case 3: _solid           = value; break;
			case 4: _colorPerVertex  = value; break;
			case 5: _normalPerVertex = value; break;
			}
		    }
		  else if(!strcmp(tkn.begin(),"creaseAngle"))
		    {
		      float value = 0.0;
		      // next token should be a non-negative float
		      nc=getToken(fp,tkn);
		      if(nc>0 && sscanf(tkn.begin(),"%f",&value)==1 && value>=0)
			_creaseAngle = value;
		      else
			break;
		    }
		  else if(!strcmp(tkn.begin(),"color")  ||
			  !strcmp(tkn.begin(),"coord")  ||
			  !strcmp(tkn.begin(),"normal") ||
			  !strcmp(tkn.begin(),"texCoord"))
		    {
		      int type =
			(!strcmp(tkn.begin(),"color"))?1:
			(!strcmp(tkn.begin(),"coord"))?2:
			(!strcmp(tkn.begin(),"normal"))?3:
			(!strcmp(tkn.begin(),"texCoord"))?4:
			0;

		      nc=getToken(fp,tkn);
		      // skip DEF
		      if(!strcmp(tkn.begin(),"DEF"))
			{
			  // skip DEF name
			  nc=getToken(fp,tkn);
			  // get next token
			  nc=getToken(fp,tkn);
			}
		      if(nc<=0                                              ||
			 (type==1 && strcmp(tkn.begin(),"Color"))           ||
			 (type==2 && strcmp(tkn.begin(),"Coordinate"))      ||
			 (type==3 && strcmp(tkn.begin(),"Normal"))          ||
			 (type==4 && strcmp(tkn.begin(),"TextureCoordinate")))
			break;

		      nc=getToken(fp,tkn);
		      if(nc<=0 || strcmp(tkn.begin(),"{"))
			break;

		      nc=getToken(fp,tkn);
		      if(nc<=0                                     ||
			 (type==1 && strcmp(tkn.begin(),"color"))  ||
			 (type==2 && strcmp(tkn.begin(),"point"))  ||
			 (type==3 && strcmp(tkn.begin(),"vector")) ||
			 (type==4 && strcmp(tkn.begin(),"point")))
			break;

		      // next token must be "["
		      nc=getToken(fp,tkn);
		      if(nc<=0 || strcmp(tkn.begin(),"["))
			break;

		      // MFFloat
		      float value;
		      while((nc=getToken(fp,tkn))>0)
			if(sscanf(tkn.begin(),"%f",&value)<=0)
			  break;
			else switch(type)
			  {
			  case 1: _color.push_back(value);    break;
			  case 2: _coord.push_back(value);    break;
			  case 3: _normal.push_back(value);   break;
			  case 4: _texCoord.push_back(value); break;
			  }

		      // next token must be "]"
		      if(nc<=0 || strcmp(tkn.begin(),"]"))
			break;

		      nc=getToken(fp,tkn);
		      if(nc<=0 || strcmp(tkn.begin(),"}"))
			break;

		    }
		  else if(!strcmp(tkn.begin(),"colorIndex") ||
			  !strcmp(tkn.begin(),"coordIndex") ||
			  !strcmp(tkn.begin(),"normalIndex") ||
			  !strcmp(tkn.begin(),"texCoordIndex"))
		    {
		      int type =
			(!strcmp(tkn.begin(),"colorIndex"))?1:
			(!strcmp(tkn.begin(),"coordIndex"))?2:
			(!strcmp(tkn.begin(),"normalIndex"))?3:
			(!strcmp(tkn.begin(),"texCoordIndex"))?4:
			0;

		      // next token must be "["
		      nc=getToken(fp,tkn);
		      if(nc<=0 || strcmp(tkn.begin(),"["))
			break;

		      // MFInt32
		      int value;
		      while((nc=getToken(fp,tkn))>0)
			if(sscanf(tkn.begin(),"%d",&value)<=0)
			  break;
			else switch(type)
			  {
			  case 1: 
			    if(value != -1)
			      _colorIndex.push_back(value);    
			    else
			      if(_colorIndex.size() > 0)
				if(_colorIndex[_colorIndex.size()-1] != -1)
				  _colorIndex.push_back(value);    
			    break;
			  case 2: 
			    if(value != -1)
			      _coordIndex.push_back(value);    
			    else
			      if(_coordIndex.size() > 0)
				if(_coordIndex[_coordIndex.size()-1] != -1)
				  _coordIndex.push_back(value);    
			    break;
			  case 3: 
			    if(value != -1)
			      _normalIndex.push_back(value);    
			    else
			      if(_normalIndex.size() > 0)
				if(_normalIndex[_normalIndex.size()-1] != -1)
				  _normalIndex.push_back(value);    
			    break;
			  case 4: 
			    if(value != -1)
			      _texCoordIndex.push_back(value);    
			    else
			      if(_texCoordIndex.size() > 0)
				if(_texCoordIndex[_texCoordIndex.size()-1] != -1)
				  _texCoordIndex.push_back(value);    
			    break;
			  }

		      // next token must be "]"
		      if(nc<=0 || strcmp(tkn.begin(),"]"))
			break;
		    }
		  else
		    break;
		}

		// next token must be "}"
		nc=getToken(fp,tkn);
		if(nc==1 && !strcmp(tkn.begin(),"}"))
		  success = true;

		break;
	      }
	  }
	}
    }

  return success;
}



//MJH 000922 : added for library generation
bool MPG_Ifs::read(FILE* fp) 
{
	clear();
	bool success = false;
	
	if(fp!=(FILE*)0)
    {
		//     char header[16];
		//    memset(header,'\0',16);
		//    fscanf(fp,"%15c",header);
		//    if(!strcmp(header,VRML_HEADER))
		vector<char> tkn;
		
		// get next token
		int  nc=0;
		do {
			while((nc=getToken(fp,tkn))>0) {
				
				if(!strcmp(tkn.begin(),"ccw") ||
					!strcmp(tkn.begin(),"convex") ||
					!strcmp(tkn.begin(),"solid") ||
					!strcmp(tkn.begin(),"colorPerVertex") ||
					!strcmp(tkn.begin(),"normalPerVertex"))
				{
					int type =
								(!strcmp(tkn.begin(),"ccw"))?1:
							(!strcmp(tkn.begin(),"convex"))?2:
							(!strcmp(tkn.begin(),"solid"))?3:
							(!strcmp(tkn.begin(),"colorPerVertex"))?4:
							(!strcmp(tkn.begin(),"normalPerVertex"))?5:
							0;
					
					bool value = true;
					// next token should be TRUE or FALSE
					nc=getToken(fp,tkn);
					if(nc<=0)
						break;
					else if(!strcmp(tkn.begin(),"TRUE"))
						value = true;
					else if(!strcmp(tkn.begin(),"FALSE"))
						value = false;
					else
						break;
					switch(type)
					{
					case 1: _ccw             = value; break;
					case 2: _convex          = value; break;
					case 3: _solid           = value; break;
					case 4: _colorPerVertex  = value; break;
					case 5: _normalPerVertex = value; break;
					}
				}
				else if(!strcmp(tkn.begin(),"creaseAngle"))
				{
					float value = 0.0;
					// next token should be a non-negative float
					nc=getToken(fp,tkn);
					if(nc>0 && sscanf(tkn.begin(),"%f",&value)==1 && value>=0)
						_creaseAngle = value;
					else
						break;
				}
				else if(!strcmp(tkn.begin(),"color")  ||
					!strcmp(tkn.begin(),"coord")  ||
					!strcmp(tkn.begin(),"normal") ||
					!strcmp(tkn.begin(),"texCoord"))
				{
					int type =
								(!strcmp(tkn.begin(),"color"))?1:
							(!strcmp(tkn.begin(),"coord"))?2:
							(!strcmp(tkn.begin(),"normal"))?3:
							(!strcmp(tkn.begin(),"texCoord"))?4:
							0;
					
					nc=getToken(fp,tkn);
					// skip DEF
					if(!strcmp(tkn.begin(),"DEF"))
					{
						// skip DEF name
						nc=getToken(fp,tkn);
						// get next token
						nc=getToken(fp,tkn);
					}
					if(nc<=0                                              ||
						(type==1 && strcmp(tkn.begin(),"Color"))           ||
						(type==2 && strcmp(tkn.begin(),"Coordinate"))      ||
						(type==3 && strcmp(tkn.begin(),"Normal"))          ||
						(type==4 && strcmp(tkn.begin(),"TextureCoordinate")))
						break;
					
					nc=getToken(fp,tkn);
					if(nc<=0 || strcmp(tkn.begin(),"{"))
						break;
					
					nc=getToken(fp,tkn);
					if(nc<=0                                     ||
						(type==1 && strcmp(tkn.begin(),"color"))  ||
						(type==2 && strcmp(tkn.begin(),"point"))  ||
						(type==3 && strcmp(tkn.begin(),"vector")) ||
						(type==4 && strcmp(tkn.begin(),"point")))
						break;
					
					// next token must be "["
					nc=getToken(fp,tkn);
					if(nc<=0 || strcmp(tkn.begin(),"["))
						break;
					
					// MFFloat
					float value;
					while((nc=getToken(fp,tkn))>0)
						if(sscanf(tkn.begin(),"%f",&value)<=0)
							break;
						else switch(type)
						{
							  case 1: _color.push_back(value);    break;
							  case 2: _coord.push_back(value);    break;
							  case 3: _normal.push_back(value);   break;
							  case 4: _texCoord.push_back(value); break;
						}
						
						// next token must be "]"
						if(nc<=0 || strcmp(tkn.begin(),"]"))
							break;
						
						nc=getToken(fp,tkn);
						if(nc<=0 || strcmp(tkn.begin(),"}"))
							break;
						
				}
				else if(!strcmp(tkn.begin(),"colorIndex") ||
					!strcmp(tkn.begin(),"coordIndex") ||
					!strcmp(tkn.begin(),"normalIndex") ||
					!strcmp(tkn.begin(),"texCoordIndex"))
				{
					int type =
								(!strcmp(tkn.begin(),"colorIndex"))?1:
							(!strcmp(tkn.begin(),"coordIndex"))?2:
							(!strcmp(tkn.begin(),"normalIndex"))?3:
							(!strcmp(tkn.begin(),"texCoordIndex"))?4:
							0;
					
					// next token must be "["
					nc=getToken(fp,tkn);
					if(nc<=0 || strcmp(tkn.begin(),"["))
						break;
					
					// MFInt32
					int value;
					while((nc=getToken(fp,tkn))>0) {
						if(sscanf(tkn.begin(),"%d",&value)<=0)
							break;
						else switch(type)
						{
							  case 1: 
								  if(value != -1)
									  _colorIndex.push_back(value);    
								  else
									  if(_colorIndex.size() > 0)
										  if(_colorIndex[_colorIndex.size()-1] != -1)
											  _colorIndex.push_back(value);    
										  break;
							  case 2: 
								  if(value != -1)
									  _coordIndex.push_back(value);    
								  else
									  if(_coordIndex.size() > 0)
										  if(_coordIndex[_coordIndex.size()-1] != -1)
											  _coordIndex.push_back(value);    
										  break;
							  case 3: 
								  if(value != -1)
									  _normalIndex.push_back(value);    
								  else
									  if(_normalIndex.size() > 0)
										  if(_normalIndex[_normalIndex.size()-1] != -1)
											  _normalIndex.push_back(value);    
										  break;
							  case 4: 
								  if(value != -1)
									  _texCoordIndex.push_back(value);    
								  else
									  if(_texCoordIndex.size() > 0)
										  if(_texCoordIndex[_texCoordIndex.size()-1] != -1)
											  _texCoordIndex.push_back(value);    
										  break;
						}  //switch(type)
					} //while((nc=getToken(fp,tkn))>0)
					
					// next token must be "]"
					if(nc<=0 || strcmp(tkn.begin(),"]"))
						break;
				}
				else
					break;
		} // while((nc=getToken(fp,tkn))>0) 
		
		//"{"
		//nc=getToken(fp,tkn);
		if(nc==1 && !strcmp(tkn.begin(),"}"))
			success = true;
		break;
		
	  }  while((nc=getToken(fp,tkn))>0);
	  
    }  //if(fp!=(FILE*)0)
	
	return success;
}

