//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------


#include "MPG_Property.h"

//---------------------------------------------------------------------------
//
//   MPG_Property::encode
//
//---------------------------------------------------------------------------

bool MPG_Property::encode(MPG_OutputBitStream& obstrm)
{
  bool success = false;

  // check for incompatible arguments
  if(_type==PT_NONE) return false;

  int dimValue = getDimensionValue();

  vector<float> qv;

  // bbox
  // if(bbox) {
  //   min[0..dimValue-1]
  //   size
  // }
  // quant
  // pred_type
  // if(pred_type==2) {
  //   nlambda
  //   for(i=1;i<nlambda;i++)
  //   lambda[i]
  // }

  bool bbox_quantization = (_type!=PT_NORMAL)?true:false;
  if(bbox_quantization)
    {
      if(_bbq==(MPG_BBoxQ*)0 || _bbq->getDimension()!=dimValue) return false;

      // encode bbox
      float* minCorner   = _bbq->getLowerCorner();
      float side         = _bbq->getSideLength();

      bool bbox = true;
      if(side==1.0)
	{
	  int i;
	  for(i=0;i<dimValue;i++)
	    if(minCorner[i]!=0.0)
	      { break; }
	  if(i==dimValue) bbox = false;
	}
      if(obstrm.putBit(bbox)==false) return false;
      if(bbox)
	{
	  int i;
	  for(i=0;i<dimValue;i++)
	    { if(obstrm.put32BitFloat(minCorner[i])==false) return false; }
	  if(obstrm.put32BitFloat(side)==false) return false;
	}

      // encode bitsPerCoord
      int bitsPerCoord = _bbq->getBitsPerCoordinate();
      int precision    = bitsPerCoord;
      if(obstrm.putMBitInt(5,precision)==false) return false;

      // binding determines the prediction scheme
      int prediction = (_prediction==PP_NONE         )?0:
	               (_prediction==PP_TREE         )?1:
	               (_prediction==PP_PARALLELOGRAM)?2:
                                                       3;
      if(prediction==3) return false;
      if(obstrm.putMBitInt(2,prediction)==false) return false;

      // encode lambdas
      if(_prediction==PP_TREE ||_prediction==PP_PARALLELOGRAM)
	{
	  int   nLambda = getDimensionLambda();

	  if(nLambda==0 || nLambda>3) return false;
	  if(_prediction==PP_PARALLELOGRAM && nLambda!=3) return false;

	  float *lambda = getLambda();

	  if(_prediction==PP_TREE && (nLambda<=0 || nLambda>3)) return false;
	  if(_prediction==PP_PARALLELOGRAM && nLambda!=3      ) return false;

	  if(obstrm.putMBitInt(2,nLambda)==false) return false;
	  if(nLambda==0) return false;
	  if(_prediction==PP_PARALLELOGRAM && nLambda!=3) return false;

	  // encode quantized lambdas
	  // do not encode _lambda[nLambda-1] because sum of lambdas == 1
	  int i;
	  for(i=0;i<nLambda-1;i++)
	    {
	      int iLambda = (int) ldexp(_lambda[i],bitsPerCoord);
	      if(obstrm.putMBitSInt(3+precision,iLambda)==false) return false;
	    }
	}
      success = true;
    }
  else // if(normal quantization)
    {
      if(_nrq==(MPG_NormalQ*)0) return false;

      // bbox
      if(obstrm.putBit(false)==false) return false;

      // encode number of subdivision levels
      int bitsPerNormal = _nrq->getBitsPerNormal();
      int subdivLevels = (bitsPerNormal-3)/2;
      if(subdivLevels<0) return false;

      if(obstrm.putMBitInt(5,subdivLevels)==false) return false;

      // encode prediction type
      int prediction = (_prediction==PP_NONE         )?0:
	               (_prediction==PP_TREE         )?1:
	               (_prediction==PP_PARALLELOGRAM)?2:
                                                       3;
      if(prediction!=0) return false;	// no other prediction ??????
      if(obstrm.putMBitInt(2,prediction)==false) return false;

      success = true;
    }

  return success;
}

bool MPG_Property::texencode(MPG_OutputBitStream& obstrm)
{
  bool success = false;

  // check for incompatible arguments
  if(_type==PT_NONE) return false;

  int dimValue = getDimensionValue();

  vector<float> qv;

  bool bbox_quantization = (_type!=PT_NORMAL)?true:false;
  if(bbox_quantization)
    {
      if(_bbq==(MPG_BBoxQ*)0 || _bbq->getDimension()!=dimValue) return false;

      // encode bbox
      float* minCorner   = _bbq->getLowerCorner();
      float side         = _bbq->getSideLength();

      bool bbox = true;
      if(side==1.0)
	{
	  int i;
	  for(i=0;i<dimValue;i++)
	    if(minCorner[i]!=0.0)
	      { break; }
	  if(i==dimValue) bbox = false;
	}
      if(obstrm.putBit(bbox)==false) return false;
      if(bbox)
	{
	  int i;
	  for(i=0;i<dimValue;i++)
	    { if(obstrm.put32BitFloat(minCorner[i])==false) return false; }
	  if(obstrm.put32BitFloat(side)==false) return false;
	}

      // encode bitsPerCoord
      int bitsPerCoord = _bbq->getBitsPerCoordinate();
      int precision    = bitsPerCoord;
      if(obstrm.putMBitInt(5,precision)==false) return false;


      // binding determines the prediction scheme
      int prediction = (_prediction==PP_NONE         )?0:
	               (_prediction==PP_TREE         )?1:
	               (_prediction==PP_PARALLELOGRAM)?2:
                                                       3;
      if(prediction==3) return false;
      if(obstrm.putMBitInt(2,prediction)==false) return false;

      // encode lambdas
      if(_prediction==PP_TREE ||_prediction==PP_PARALLELOGRAM)
	{
	  int   nLambda = getDimensionLambda();

	  if(nLambda==0 || nLambda>3) return false;
	  if(_prediction==PP_PARALLELOGRAM && nLambda!=3) return false;

	  float *lambda = getLambda();

	  if(_prediction==PP_TREE && (nLambda<=0 || nLambda>3)) return false;
	  if(_prediction==PP_PARALLELOGRAM && nLambda!=3      ) return false;

	  if(obstrm.putMBitInt(2,nLambda)==false) return false;
	  if(nLambda==0) return false;
	  if(_prediction==PP_PARALLELOGRAM && nLambda!=3) return false;

	  // encode quantized lambdas
	  // do not encode _lambda[nLambda-1] because sum of lambdas == 1
	  int i;
	  for(i=0;i<nLambda-1;i++)
	    {
	      int iLambda = (int) ldexp(_lambda[i],bitsPerCoord);
	      if(obstrm.putMBitSInt(3+precision,iLambda)==false) return false;

	    }
	}
      success = true;
    }
  return success;
}

