//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------


#include <stdio.h>
#include <stdlib.h>
#include "MPG_Global.h"
#include "MPG_VGraph.h"

//---------------------------------------------------------------------------
//
//   MPG_VGraph::fromLoop
//
//   used for compression
//
//---------------------------------------------------------------------------

bool MPG_VGraph::fromLoop(MPG_Loops& loop)
{
  bool success = true;

  int i,j,l,vCurr,vNext,ie,ieo;
  int length,start,leaf,last,jump,jumpTo,iRunStart,vRunStart;
  MPG_Graph::pEdge e,eo;

  // convert loop from linked list to array ...
  vector<int> iloop;
  loop.getLoop(0,iloop);
  int nL    = loop.size();
  int vRoot = iloop[0];

  int nV = 0;
  for(j=1;j<nL;j++)
    if(iloop[j]>nV)
      nV = iloop[j];
  nV++;

  // these may also be passed as arguments
  vector<int> v_traversal;
  vector<int> v_father((size_t)nV,-1);

  // ... and back into linked list
  vector<int> next;
  next.reserve(nL);
  for(j=1;j<nL;j++)
    next.push_back(j);
  next.push_back(0);

  // get number of boundary edges and order of vertices in the vertex graph
  int nEbdry = 0;
  vector<int> v_order((size_t) nV,0);
  {
    // construct loop graph
    MPG_Graph gLoop(nV);
    for(i=0;i<nL;i++)
      gLoop.insertEdge(iloop[i],iloop[next[i]]);
    gLoop.enumerateEdges();

    // number of loop graph edges
    int nE = gLoop.getNumberOfEdges();

    // determine vertex graph order of edges and vertices
    vector<int> e_order((size_t) nE,0);
    for(i=0;i<nL;i++)
      {
	e  = gLoop.getEdge(iloop[i],iloop[next[i]]);
	if(e_order[e->getIndex()]==0)
	  { v_order[(*e)[0]]++; v_order[(*e)[1]]++; }
	e_order[e->getIndex()]++;
	// at most two loop edges per vertex graph edge are allowed
	// COM_ASSERT(e_order[e->getIndex()]<=2);
      }

    // count number of boundary edges
    nEbdry = 0;
    for(i=0;i<nE;i++)
      if(e_order[i]==1) nEbdry++;
  }

  // determine number of vertex graph runs
  int nVGruns = v_order[iloop[0]];
  for(i=0;i<nV;i++)
    if(v_order[i]>2 && i!=iloop[0])
      nVGruns += (v_order[i]-1);

  // construct oriented runs graph
  MPG_Graph gLoopOr(nV,true);
  for(i=0;i<nL;i++)
    gLoopOr.insertEdge(iloop[i],iloop[next[i]],i);
  int nGLoopOr = gLoopOr.getNumberOfEdges();

  // append boundary loops to iloop
  vector<int> first;
  first.push_back(0);
  if(nEbdry>0)
    {
      vector<bool> wasVisited((size_t) nGLoopOr,false);
      vector<int> loopStack;
      for(i=j=0;i<nV;i++)
	for(e=gLoopOr.getFirstEdge(i);e;e=gLoopOr.getNextEdge(e))
	  if(gLoopOr.hasInverseEdge(e)==false &&
	     wasVisited[e->getIndex()]==false)
	    {
	      // new boundary loop starting at vertex i
	      eo = e;
	      do {
		wasVisited[eo->getIndex()] = true;

		// push vertex values onto loopStack
		int eo0=(*eo)[0];
		loopStack.push_back(eo0);

		// find next bdry run starting where eo ends
		for(eo=gLoopOr.getFirstEdge((*eo)[1]);
		    eo && !(gLoopOr.hasInverseEdge(eo)==false && (*eo)[1]!=eo0);
		    eo=gLoopOr.getNextEdge(eo));

	      } while(eo!=e);

	      // pop vertex values from loopStack and push onto iloop
	      // so that the boundary loop is properly oriented
	      start = iloop.size();
	      first.push_back(start);
	      while(loopStack.size()>0)
		{
		  iloop.push_back(loopStack.back()); loopStack.pop_back();
		  next.push_back((loopStack.size()>0)?iloop.size():start);
		}
	    }
    }
  first.push_back(iloop.size());
  int nLoops = first.size()-1;

  // add boundary edges to gLoopOr
  for(l=1;l<nLoops;l++)
    {
      int i0 = i = first[l];
      do {
	j = next[i]; gLoopOr.insertEdge(iloop[i],iloop[j],i);
      } while((i=j)!=i0);
    }
  // gLoopOr.enumerateEdges();
  nGLoopOr = gLoopOr.getNumberOfEdges();

  // construction loop --------------------------------------------------------

  vRoot = iloop[0];
  vector<int> used((size_t) nGLoopOr,0);
  vector<int> travStack;
  travStack.push_back(vRoot);

  for(i=0;i<nL;i++)
    v_father[iloop[i]] = -1;
  v_father[vRoot] = vRoot;
  v_traversal.push_back(vRoot);
  int nVt = 1;

  int jInvOr = 0;
  int iRun   = 0;
  int iJump  = 0;

  length = i = 0; j = 1;

  v_order[vRoot]++;
  vector<int> v_n_visits((size_t) nV,0);

  while(iRun<nVGruns)
    {
      vCurr = iloop[i];
      vNext = iloop[j];

      start = 1;

      // COM_ASSERT(vCurr>=0);
      if(vNext<0) // take care of jump- runs
	{
	  v_n_visits[vCurr]++;
	  last=(v_n_visits[vCurr]==v_order[vCurr]-1)?true:false;

	  start  = 0;
	  leaf   = 1;
	  jump   = 1;
	  jumpTo = -vNext-1;

	  pushBackRun(last,start,length,leaf,jump,jInvOr,jumpTo);

	  iRun++; i = next[j]; j = next[i];
	  continue;
	}

      e  = gLoopOr.getEdge(vCurr,vNext); ie = e->getIndex();

      if(used[ie]>0)
	{
	  i = j; j = next[i];
	  continue;
	}

      eo = gLoopOr.getInverseEdge(e) ; ieo = eo->getIndex();

      used[ie]++;
      used[ieo]++;

      if(length==0)
	{ // begining of run
	  iRunStart = i;
	  vRunStart = vCurr;
	  v_n_visits[vCurr]++;
	  last=(v_n_visits[vCurr]==v_order[vCurr]-1)?true:false;
	}

      length++;

      if(v_father[vNext]<0)
	{
	  v_traversal.push_back(vNext); nVt++;
	  v_father[vNext] = vCurr;
	}
	
      if(vNext==vRoot || v_order[vNext]!=2)
	{ // end of run

	  // determine end-of-run type
	  start  = 1;
	  jumpTo = -1;
	  switch((vNext==vRoot)?3:v_order[vNext])
	    {
	    case 0: // impossible
	    case 2:
	      break;
	    case 1:		         // run ends in leaf
	      { leaf = true; jump = false; }
	    break;
	    default: // >2
	      if(v_father[vNext]==vCurr) // run ends in branch
		{ leaf = false; jump = false; }
	      else		         // run ends in jump+
		{ leaf =  true; jump = true; }
	      break;
	    }

	  pushBackRun(last,start,length,leaf,jump,jInvOr,jumpTo);

	  // process end-of-run
	  if(!leaf)		// run ends in branch
	    {
	      // j = i;
	    }
	  else if(jump)		// run ends in jump+
	    {
	      // convert the run in a leaf ending run
	      int iO = ieo; // index of (*eo)[0]  in iloop of
	      int jO = next[iO];

	      // reorient the end of run to make it go back on the
	      // other side so that i -> j0
	      next[i] = jO;

	      // insert a new vertex as a jump- run marker so that
	      // i0 -> next[i0] -> j

	      next[iO] = iloop.size();
	      iJump++;
	      iloop.push_back(-iJump);
	      next.push_back(j);

	      // move on
	      j  = next[i];
	    }
	  else			// run ends in leaf
	    {
	      // j = i;
	    }

	  length = 0;
	  iRun++;
	}

      i = j; j = next[i];
    }
  return success;
}

//---------------------------------------------------------------------------
//
//   MRS_VertexGraph::fromLoop
//
//   used by MPG_IfsMesh_EXT::representConnectivity()
//
//---------------------------------------------------------------------------

bool MPG_VGraph::fromLoop(MPG_Loops& loop,
			  vector<int>& v_traversal,
			  vector<int>& v_father)
{
  bool no_err = true;

  int i,j,l,vCurr,vNext,ie,ieo;
  int length,start,leaf,last,jump,jumpTo,iRunStart,vRunStart;
  MPG_Graph::pEdge e,eo;

  int nV = v_father.size();

  // convert loop from linked list to array ...
  vector<int> iloop;
  loop.getLoop(0,iloop);
  int nL    = loop.size();
  int vRoot = iloop[0];

  // ... and back into linked list
  vector<int> next;
  next.reserve(nL);
  for(j=1;j<nL;j++)
    next.push_back(j);
  next.push_back(0);

  // get number of boundary edges and order of vertices in the vertex graph
  int nEbdry = 0;
  vector<int> v_order((size_t) nV,0);
  {
    // construct loop graph
    MPG_Graph gLoop(nV);
    for(i=0;i<nL;i++)
      gLoop.insertEdge(iloop[i],iloop[next[i]]);
    gLoop.enumerateEdges();

    // number of loop graph edges
    int nE = gLoop.getNumberOfEdges();

    // determine vertex graph order of edges and vertices
    vector<int> e_order((size_t) nE,0);
    for(i=0;i<nL;i++)
      {
	e  = gLoop.getEdge(iloop[i],iloop[next[i]]);
	if(e_order[e->getIndex()]==0)
	  { v_order[(*e)[0]]++; v_order[(*e)[1]]++; }
	e_order[e->getIndex()]++;
	// at most two loop edges per vertex graph edge are allowed
	if(!(e_order[e->getIndex()]<=2)) return false;
      }

    // count number of boundary edges
    nEbdry = 0;
    for(i=0;i<nE;i++)
      if(e_order[i]==1) nEbdry++;
  }

  // determine number of vertex graph runs
  int nVGruns = v_order[iloop[0]];
  for(i=0;i<nV;i++)
    if(v_order[i]>2 && i!=iloop[0])
      nVGruns += (v_order[i]-1);

  // construct oriented runs graph
  MPG_Graph gLoopOr(nV,true);
  for(i=0;i<nL;i++)
    gLoopOr.insertEdge(iloop[i],iloop[next[i]],i);
  int nGLoopOr = gLoopOr.getNumberOfEdges();

  // append boundary loops to iloop
  vector<int> first;
  first.push_back(0);
  if(nEbdry>0)
    {
      vector<bool> wasVisited((size_t) nGLoopOr,false);
      vector<int> loopStack;
      for(i=j=0;i<nV;i++)
	for(e=gLoopOr.getFirstEdge(i);e;e=gLoopOr.getNextEdge(e))
	  if(gLoopOr.hasInverseEdge(e)==false &&
	     wasVisited[e->getIndex()]==false)
	    {
	      // new boundary loop starting at vertex i
	      eo = e;
	      do {
		wasVisited[eo->getIndex()] = true;

		// push vertex values onto loopStack
		int eo0=(*eo)[0];
		loopStack.push_back(eo0);

		// find next bdry run starting where eo ends
		for(eo=gLoopOr.getFirstEdge((*eo)[1]);
		    eo && !(gLoopOr.hasInverseEdge(eo)==false && (*eo)[1]!=eo0);
		    eo=gLoopOr.getNextEdge(eo));

	      } while(eo!=e);

	      // pop vertex values from loopStack and push onto iloop
	      // so that the boundary loop is properly oriented
	      start = iloop.size();
	      first.push_back(start);
	      while(loopStack.size()>0)
		{
		  iloop.push_back(loopStack.back()); loopStack.pop_back();
		  next.push_back((loopStack.size()>0)?iloop.size():start);
		}
	    }
    }
  first.push_back(iloop.size());
  int nLoops = first.size()-1;

  // add boundary edges to gLoopOr
  for(l=1;l<nLoops;l++)
    {
      int i0 = i = first[l];
      do {
	j = next[i]; gLoopOr.insertEdge(iloop[i],iloop[j],i);
      } while((i=j)!=i0);
    }
  // gLoopOr.enumerateEdges();
  nGLoopOr = gLoopOr.getNumberOfEdges();

  // construction loop --------------------------------------------------------

  vRoot = iloop[0];
  vector<int> used((size_t) nGLoopOr,0);
  vector<int> travStack;
  travStack.push_back(vRoot);

  for(i=0;i<nL;i++)
    v_father[iloop[i]] = -1;
  v_father[vRoot] = vRoot;
  v_traversal.push_back(vRoot);
  int nVt = 1;

  int jInvOr = 0;
  int iRun   = 0;
  int iJump  = 0;

  length = i = 0; j = 1;

  v_order[vRoot]++;
  vector<int> v_n_visits((size_t) nV,0);

  while(iRun<nVGruns)
    {
      vCurr = iloop[i];
      vNext = iloop[j];

      start = 1;

      if(!(vCurr>=0)) return false;
      if(vNext<0) // take care of jump- runs
	{
	  v_n_visits[vCurr]++;
	  last=(v_n_visits[vCurr]==v_order[vCurr]-1)?true:false;

	  start  = 0;
	  leaf   = 1;
	  jump   = 1;
	  jumpTo = -vNext-1;

	  pushBackRun(last,start,length,leaf,jump,jInvOr,jumpTo);

	  iRun++; i = next[j]; j = next[i];
	  continue;
	}

      e  = gLoopOr.getEdge(vCurr,vNext); ie = e->getIndex();

      if(used[ie]>0)
	{
	  i = j; j = next[i];
	  continue;
	}

      eo = gLoopOr.getInverseEdge(e) ; ieo = eo->getIndex();

      used[ie]++;
      used[ieo]++;

      if(length==0)
	{ // begining of run
	  // if(!last) travStack.push_back(vCurr);
	  iRunStart = i;
	  vRunStart = vCurr;
	  v_n_visits[vCurr]++;
	  last=(v_n_visits[vCurr]==v_order[vCurr]-1)?true:false;
	}

      length++;

      if(v_father[vNext]<0)
	{
	  v_traversal.push_back(vNext); nVt++;
	  v_father[vNext] = vCurr;
	}

      if(vNext==vRoot || v_order[vNext]!=2)
	{ // end of run

	  // determine end-of-run type
	  start  = 1;
	  jumpTo = -1;
	  switch((vNext==vRoot)?3:v_order[vNext])
	    {
	    case 0: // impossible
	    case 2:
	      break;
	    case 1:		         // run ends in leaf
	      { leaf = true; jump = false; }
	    break;
	    default: // >2
	      if(v_father[vNext]==vCurr) // run ends in branch
		{ leaf = false; jump = false; }
	      else		         // run ends in jump+
		{ leaf =  true; jump = true; }
	      break;
	    }

	  pushBackRun(last,start,length,leaf,jump,jInvOr,jumpTo);

	  // process end-of-run
	  if(!leaf)		// run ends in branch
	    {
	      // j = i;
	    }
	  else if(jump)		// run ends in jump+
	    {
	      // convert the run in a leaf ending run
	      int iO = ieo; // index of (*eo)[0]  in iloop of
	      int jO = next[iO];

	      // reorient the end of run to make it go back on the
	      // other side so that i -> j0
	      next[i] = jO;

	      // insert a new vertex as a jump- run marker so that
	      // i0 -> next[i0] -> j

	      next[iO] = iloop.size();
	      iJump++;
	      iloop.push_back(-iJump);
	      next.push_back(j);

	      // move on
	      j  = next[i];
	    }
	  else			// run ends in leaf
	    {
	      // j = i;
	    }

	  length = 0;
	  iRun++;
	}

      i = j; j = next[i];
    }

  return no_err;
}

//---------------------------------------------------------------------------
//
//   MPG_VGraph::encode
//
//---------------------------------------------------------------------------
// vertex_graph() {
//	  qf_decode(vg_simple, vg_simple_context)
//	  depth = 0
//	  code_last = '1'
//	  openloops = 0
//	  do {
//	  	 do {
//			if (code_last == '1') {
//				qf_decode(vg_last, vg_last_context)
//				if (openloops > 0) {
//					qf_decode(vg_forward_run, vg_forward_run_context)
//					if (vg_forward_run == '0') {
//						openloops--
//						if (openloops > 0)
//							qf_decode(vg_loop_index,vg_loop_index_context)
//						break
//					}
//				}
//			}
//			qf_decode(vg_run_length, vg_run_length_context)
//			qf_decode(vg_leaf, vg_leaf_context)
//			if (vg_leaf == '1' && vg_simple == '0') {
//				qf_decode(vg_loop, vg_loop_context)
//				if (vg_loop == '1')
//					openloops++
//			}
//		 } while (0)
//		 if (vg_leaf == '1' && vg_last == '1' && code_last == '1')
//			 depth--
//		 if (vg_leaf == '0' && (vg_last == '0' || code_last == '0'))
//			 depth++
//		 code_last = vg_leaf
//	  } while (depth >= 0)
// }

bool MPG_VGraph::encode(MPG_QFac &codec)
{
	bool success = false;
	
	// determine if all the graphs are simple or not
	bool isSimple = (getNumberOfJumps()==0)?true:false;
	codec.encode(isSimple,MPG_Global::vg_simple);
	vector<int> jStack;
	
	// vertex run record fields
	int last   =  0;
	int start  =  1;
	int length =  0;
	int leaf   =  0;
	int jump   =  0;
	int jumpTo = -1;
	int jInvOr =  0;
	
    // traversal variables
	int  nRuns          = getNumberOfRuns();
	int  iRun           = 0;     // within current VertexGraph
	int  vg_run_depth   = 1;     // depth in traversal stack
	bool after_branch   = false; // true if first run after branch ending run
	bool start_of_graph = true;  // true if first run of a VertexGraph
	
	for(iRun=0;iRun<nRuns;iRun++)
    {
		getRun(iRun,last,start,length,leaf,jump,jInvOr,jumpTo);
		if(after_branch==false) codec.encode(last,MPG_Global::vg_last);
		if(!last) vg_run_depth++;
		
		// start = (isSimple||start_of_graph==true||after_branch)?
		//         true : codec.decode(bpmStart);
		int nS = 0;
		int i;
		for(i=0;i<jStack.size();i++)
			if(jStack[i]>=0)
				nS++;
			
			if(after_branch==false && nS>0)
				codec.encode(start,MPG_Global::vg_forward_run);
			// else start == true;
			
			// if(start_of_graph==true)
			//   start_of_graph = false;
			
			if(start)
			{
				int j;
				for(j=0;j<length-1;j++)
					codec.encode(false,MPG_Global::vg_run_length);
				codec.encode(true,MPG_Global::vg_run_length);
				
				codec.encode(leaf,MPG_Global::vg_leaf);
				
				if(leaf)
				{
					if(!isSimple)
					{
						codec.encode(jump,MPG_Global::vg_loop);
						if(jump) jStack.push_back(iRun);
					}
					vg_run_depth--;
				}
			}
			else // !start
			{
				// if(!isOriented) codec.encode(jInvOr,bpmJInvOr);
				{
					// jumpTo must point to an element of jStack !!!
					// COM_ASSERT(0<=jumpTo && jumpTo<jStack.size());
					jStack[jumpTo] = -1;
					
					// determine depth skipping previously used jumps
					int depth = 0;
					int i;
					for(i=0;i<jumpTo;i++)
						if(jStack[i]>=0)
							depth++;
						int nS = 1;
						for(i=0;i<jStack.size();i++)
							if(jStack[i]>=0)
								nS++;
							int j;
							for(j=0;j<=depth;j++)
							{
								if(j==nS-1) break;
								bool val=(j<depth)?false:true;
								codec.encode(val,MPG_Global::vg_loop_index);
							}
				}
				
				vg_run_depth--;
				
				// set ignored fields
				leaf   = true;
			}
			
			after_branch = (leaf)?false:true;
    }
	
	return true;
}
