//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------


#include <stdio.h>
#include <stdlib.h>
#include "MPG_Vector.h"
#include "MPG_Heap.h"
#include "MPG_Lex.h"

//---------------------------------------------------------------------------
//
//   MPG_Vector_EXT::setAverageVector
//
//---------------------------------------------------------------------------

void MPG_Vector_EXT::setAverageVector(int nD, float* v0, vector<float*>& v)
{
  // accumulate on a separate array
  // because v0 could be part of v
  static vector<float> vv0;
  vv0.erase(vv0.begin(),vv0.end());
  vv0.insert(vv0.end(),(size_t)nD,0.0);

  int i,j;
  int nV = v.size();
  if(nV>0)
    {
      vector<float*>::iterator vi = v.begin();
      for(i=0;i<nV;i++,vi++)
	for(j=0;j<nD;j++)
	  vv0[j] += (*vi)[j];
      for(j=0;j<nD;j++)
	vv0[j] /= nV;
    }
  for(j=0;j<nD;j++)
    v0[j] = vv0[j];
}

//---------------------------------------------------------------------------
//
//   MPG_Vector_EXT::setAverageNormal
//
//---------------------------------------------------------------------------

void MPG_Vector_EXT::setAverageNormal(float* n0, vector<float*>& n)
{
  setAverageVector(3,n0,n);
  // set current face normal to the
  // average normal for the three faces
  float nrm = 0;
  int j;
  for(j=0;j<3;j++)
    nrm += n0[j]*n0[j];
  // normalize
  if(nrm>0.0)
    {
      nrm = 1.0/sqrt(nrm);
      for(j=0;j<3;j++)
	n0[j] *= nrm;
    }
}

//---------------------------------------------------------------------------
//
//   MPG_Vector_EXT::computeNormal
//
//---------------------------------------------------------------------------

void MPG_Vector_EXT::computeNormal(float* p0, float* p1, float* p2,
			       float* n, bool normalize)
{
  float v1[3], v2[3];

  v1[0] = p1[0] - p0[0];
  v1[1] = p1[1] - p0[1];
  v1[2] = p1[2] - p0[2];

  v2[0] = p2[0] - p0[0];
  v2[1] = p2[1] - p0[1];
  v2[2] = p2[2] - p0[2];

  n[0] = v1[1] * v2[2] - v1[2] * v2[1];
  n[1] = v1[2] * v2[0] - v1[0] * v2[2];
  n[2] = v1[0] * v2[1] - v1[1] * v2[0];

  if(normalize==true)
    normalizeNormal(n);
}

//---------------------------------------------------------------------------
//
//   MPG_Vector_EXT::computeFaceCentroid
//
//---------------------------------------------------------------------------

void MPG_Vector_EXT::computeFaceCentroid(vector<float>& coord,
				     int nfi, int* fi, float* fc)
{
  int j,k;
  for(k=0;k<3;k++)
    fc[k] = 0.0;
  for(j=0;j<nfi;j++)
    {
      float* vj = &(coord[3*fi[j]]);
      for(k=0;k<3;k++)
	fc[k] += vj[k];
    }
  for(k=0;k<3;k++)
    fc[k] /= nfi;
}

//---------------------------------------------------------------------------
//
//   MPG_Vector_EXT::computeFaceNormal
//
//---------------------------------------------------------------------------

void MPG_Vector_EXT::computeFaceNormal(vector<float>& coord,
				   int nfi, int* fi,float* fn,
				   bool normalize)
{
  if(nfi==3)
    computeNormal(&(coord[3*fi[0]]),
		  &(coord[3*fi[1]]),
		  &(coord[3*fi[2]]),
		  fn,
		  normalize);
  else if(nfi>3)
    {
      // compute face centroid
      float fc[3];
      computeFaceCentroid(coord,nfi,fi,fc);

      // accumulate vector products
      int j,k;
      float vpj[3];
      float* vj0 = &(coord[fi[0]]);
      for(j=0;j<nfi;j++)
	{
	  float* vj1 = &(coord[fi[(j+1)%nfi]]);
	  computeNormal(fc,vj0,vj1,vpj,false);
	  for(k=0;k<3;k++)
	    fn[k] += vpj[k];
	  vj0 = vj1;
	}
      if(normalize==true)
	normalizeNormal(fn);
    }
}

//---------------------------------------------------------------------------
//
//   MPG_Vector_EXT::clusterPropertyValues
//
//---------------------------------------------------------------------------

void MPG_Vector_EXT::clusterPropertyValues(vector<int>& cmap, vector<float>& p,
				       int nD, bool normalize)
{
  int nV0 = cmap.size();
  int nV1 = p.size();
  // it is assumed that 0<=cmap[i]<nV1

  vector<float> newP((size_t)(nV1*nD),0.0);
  vector<int>   newPn((size_t)nV1,0);
  int i,j,k;

  // accumulate
  for(i=0;i<nV0;i++)
    {
      j = cmap[i];
      if(0<=j && j<nV1)
	for(newPn[j]++,k=0;k<nD;k++)
	  newP[j*nD+k] += p[i*nD+k];
    }
  // normalize
  for(j=0;j<nV1;j++)
    {
      if(newPn[j]>0)
	{
	  float* vj = &(newP[j*nD]);
	  for(k=0;k<nD;k++)
	    vj[k] /= newPn[j];
	  if(normalize==true)
	    normalizeNormal(vj);
	}
    }

  p.swap(newP);
}

//---------------------------------------------------------------------------
//
//   MPG_Vector_EXT::indexLabel
//
//---------------------------------------------------------------------------

bool MPG_Vector_EXT::indexLabel(int          nV,
			    vector<int>& v_label,
			    vector<int>& v_labelIndex,
			    vector<int>& v_labelValue,
			    vector<int>* v_sortOrder,
			    vector<int>* v_partStart)
{
  if(nV<=0) return false;
  int dimLabel = v_label.size()/nV;
  if(dimLabel<1) return false;

  v_labelIndex.erase(v_labelIndex.begin(),v_labelIndex.end());
  v_labelValue.erase(v_labelValue.begin(),v_labelValue.end());
  if(v_sortOrder!=(vector<int>*)0)
    v_sortOrder->erase(v_sortOrder->begin(),v_sortOrder->end());
  if(v_partStart!=(vector<int>*)0)
    v_partStart->erase(v_partStart->begin(),v_partStart->end());

  int i,j,k,l;

  // sort v_label lexicographically
  MPG_Lex_EXT::D = dimLabel;
  MPG_Heap_EXT h(MPG_Lex_EXT::LT_I,MPG_Lex_EXT::EQ_I,nV);
  for(i=0;i<nV;i++)
    h.add((void*)(&(v_label[i*dimLabel])));

  // allocate array of indices instead of reusing v_labelIndex
  // to make it work when &v_label==&v_labelIndex
  vector<int> index((size_t)nV,-1);

  // traverse in increasing lexicographical order
  int  iprev      = -1;
  int  jprev      = -1;
  int* vlj        = NULL;
  int* vlj_prev   = NULL;
  int  nPart      = 0;
  int  last_start = 0;

  for(l=j=0;j<nV;j++)
    {
      // get original index
      void* ki;
      i  = h.delMin(ki);
      vlj = (int*)ki;

      // save the sorted order
      if(v_sortOrder!=(vector<int>*)0)
	v_sortOrder->push_back(i);

      if(iprev<0 || MPG_Lex_EXT::EQ_I((void*)vlj,(void*)vlj_prev)==false)
	{
	  // save start position of new subset in v_sortOrder
	  // (note that for the first one nPart==0)
	  if(v_partStart!=(vector<int>*)0)
	    {
	      last_start += nPart;
	      v_partStart->push_back(last_start);
	    }
	  nPart = 0;

	  // start a new subset assign new value
	  index[i] = l++; nPart++;
	  for(k=0;k<dimLabel;k++)
	    v_labelValue.push_back(vlj[k]);
	  
	}
      else // (iprev>=0 && MPG_Lex_EXT::EQ_I((void*)vlj,(void*)vlj_prev)==true)
	{
	  // assigned previously assigned value
	  index[i] = index[iprev]; nPart++;
	}

      iprev  = i;
      jprev  = j; vlj_prev = vlj;
    }

  // save upper limit of last subset in v_sortOrder;
  // after this we should have v_partStart->back==nV
  if(v_partStart!=(vector<int>*)0)
    {
      last_start += nPart;
      v_partStart->push_back(last_start);
    }

  v_labelIndex.swap(index);

  return true;
}

//---------------------------------------------------------------------------
//
//   MPG_Vector_EXT::cluster
//
//---------------------------------------------------------------------------

bool MPG_Vector_EXT::cluster(vector<int>&   v_order,
			 vector<int>&   v_partStart, // cMap
			 vector<float>& coord,
			 vector<float>& coordCluster,
			 bool closest)
{
  int nV = v_order.size();
  if(nV<=0) return false;
  int nParts = v_partStart.size()-1;
  if(nParts<1 || nParts>nV) return false;
  int dimCoord = coord.size()/nV;
  if(dimCoord<1) return false;

  // use a new array to accumulate coordSimplify
  // to handle the case when &coord==&coordCluster
  vector<float> newCoord((size_t)nParts*dimCoord,0.0);
  int i,j,k,ni,indx_ij,j_min;
  float *ci,*cij,di,dij;
  for(i=0;i<nParts;i++)
    {
      ni = v_partStart[i+1]-v_partStart[i];
      ci = &(newCoord[i*dimCoord]);
      for(j=0;j<ni;j++)
	{
	  indx_ij = v_order[v_partStart[i]+j];
	  cij = &(coord[dimCoord*indx_ij]);
	  for(k=0;k<dimCoord;k++)
	    ci[k] += cij[k];
	}
      for(k=0;k<dimCoord;k++)
	ci[k] /= ni;
      if(closest==true)
	{
	  // replace ci with the closest cij to the average
	  di    = -1.0;
	  j_min = -1;
	  for(j=0;j<ni;j++)
	    {
	      indx_ij = v_order[v_partStart[i]+j];
	      cij = &(coord[dimCoord*indx_ij]);

	      dij = 0;
	      for(k=0;k<dimCoord;k++)
		dij += (ci[k]-cij[k])*(ci[k]-cij[k]);
	      if(j_min<0 || dij<di)
		{ j_min = j; di = dij; }
	    }
	  indx_ij = v_order[v_partStart[i]+j_min];
	  cij     = &(coord[dimCoord*indx_ij]);
	  for(k=0;k<dimCoord;k++)
	    ci[k] = cij[k];
	}
    }
  coordCluster.swap(newCoord);

  return true;
}

//---------------------------------------------------------------------------
//
//   MPG_Vector_EXT::clusterLabel
//
//---------------------------------------------------------------------------

bool MPG_Vector_EXT::clusterLabel(int            nV,
			      vector<int>&   v_label,
			      vector<int>&   v_labelIndex, // cMap
			      vector<float>& coord,
			      vector<float>& coordCluster,
			      bool closest)
{
  if(nV<=0) return false;
  int dimCoord = coord.size()/nV;
  if(dimCoord<1) return false;
  int dimLabel = v_label.size()/nV;
  if(dimLabel<1) return false;

  vector<int> v_labelValue;
  vector<int> v_sortOrder;
  vector<int> v_partStart;

  if(indexLabel(nV,v_label,v_labelIndex,v_labelValue,
		&v_sortOrder, &v_partStart)==false) return false;

  return cluster(v_sortOrder,v_partStart,coord,coordCluster,closest);
}

//---------------------------------------------------------------------------
//
//   MPG_Vector_EXT::makeColorPalette
//
//---------------------------------------------------------------------------

void MPG_Vector_EXT::makeColorPalette(int nColor, vector<float>& color)
{
  color.erase(color.begin(),color.end());
  if(nColor<2) return;
  color.insert(color.end(),(size_t)(3*nColor),0.0);

  int i,j,k[4];
  float r0,g0,b0,r1,g1,b1;

  k[0] = i = 0;			// yellow
  color[3*i  ]	= 1.00;
  color[3*i+1]	= 1.00;
  color[3*i+2]	= 0.00;

  k[1] = i = (nColor-1)/3;	// red
  color[3*i  ]	= 1.00;
  color[3*i+1]	= 0.00;
  color[3*i+2]	= 0.00;

  k[2] = i = 2*(nColor-1)/3;	// green
  color[3*i  ]	= 0.00;
  color[3*i+1]	= 0.75;
  color[3*i+2]	= 0.00;

  k[3] = i = nColor-1;		// blue
  color[3*i  ]	= 0.00;
  color[3*i+1]	= 0.00;
  color[3*i+2]	= 1.00;

  for(j=0;j<3;j++)
    {
      r0 = color[3*k[j]  ]; r1 = color[3*k[j+1]  ];
      g0 = color[3*k[j]+1]; g1 = color[3*k[j+1]+1];
      b0 = color[3*k[j]+2]; b1 = color[3*k[j+1]+2];
      for(i=k[j]+1;i<k[j+1];i++)
	{
	  color[3*i  ] = ((i-k[j])*r1+(k[j+1]-i)*r0)/(k[j+1]-k[j]);
	  color[3*i+1] = ((i-k[j])*g1+(k[j+1]-i)*g0)/(k[j+1]-k[j]);
	  color[3*i+2] = ((i-k[j])*b1+(k[j+1]-i)*b0)/(k[j+1]-k[j]);
	}
    }
}
