/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player/ Mux
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  OD Decoder/ Encoder / Mux
FILE:       ODEncDec.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of parsing/encoding of descriptors of OD stream
-------------------------------------------------------------------------*/

#include "defs.h"
#include "ODP.h" // SOLARIS PORTING - typo
#include "ODCoder.h"
#include "sdl.h"
#include "math.h"

void MeshGridDecoderSpecificInfo :: EncDecVarSize(ODCoder *pCoder, OPTYPE op, SDLArray<32> &value, int size)
{
	switch (op)
	{
	case ENCODE:
	case CALC_LENGTH:
		{
			// compute number of bits to encode/decode value
			int maxValue = max(value[0], max(value[1], value[2]));
			nBits = (int) (log10(maxValue) / log10(2) + 1.5);
			nBits.EncDec(pCoder, this, op);

			value.EncDecArrayVarSize(pCoder, this, op, size, nBits);
		}
		break;
	case DECODE:
		{
			nBits.EncDec(pCoder, this, op);
			value.EncDecArrayVarSize(pCoder, this, op, size, nBits);
		}
		break;
	}
}

void MeshGridDecoderSpecificInfo :: DescEncDec (ODCoder *pCoder, OPTYPE op)
{
	bool isSingleLayered = false;

	EncDecVarSize(pCoder, op, nLevels, 3);
	EncDecVarSize(pCoder, op, nSlices, 3);
	EncDecVarSize(pCoder, op, nROIs, 3);

	hasConnectivityInfo.EncDec(pCoder, this, op);
	hasRefineInfo.EncDec(pCoder, this, op);
	hasRepositionInfo.EncDec(pCoder, this, op);
	hasGridInfo.EncDec(pCoder, this, op);
	attributes.EncDec(pCoder, this, op);
	hasMultiLayer.EncDec(pCoder, this, op);

	if (hasMultiLayer)
		nLayers.EncDec(pCoder, this, op);

	meshType.EncDec(pCoder, this, op);
	if (meshType == 2)
	{
		if (hasConnectivityInfo)
		{
			sameBorderOrientation.EncDec(pCoder, this, op);
			uniformSplit.EncDec(pCoder, this, op);
		}

		ASSERT(nSlices[0] > 1 && nSlices[1] > 1);
		isSingleLayered = (nSlices[2] == 1);
		if (isSingleLayered)
		{
			offsetAmplitude.EncDec(pCoder, this, op);
		}
	}

	cyclicMode.EncDec(pCoder, this, op);

	if (hasRefineInfo)
	{
		fullRefine.EncDec(pCoder, this, op);
		nRefineBits.EncDec(pCoder, this, op);
	}

	filterType.EncDec(pCoder, this, op);

	EncDecVarSize(pCoder, op, gridScale, 3);

	// how many grid corners are needed?
	int nrGridCorners = 8;

	if (isSingleLayered)				nrGridCorners /= 2;
	if (cyclicMode & CYCLIC_BITVAL_U)	nrGridCorners /= 2;
	if (cyclicMode == CYCLIC_UV)		nrGridCorners /= 2;

	ASSERT(nrGridCorners >= 1);
	int nrGridValues = nrGridCorners * 3;

	// the values of the grid corners are available
	// as a variable lenght field
	switch (op)
	{
	case ENCODE:
		{
			nQuantBits.EncDec(pCoder, this, op);

			for (int k = 0; k < nrGridValues; k++)
			{
				if (gridCorners[k] < 0)
				{
					sign = 1; corner = -gridCorners[k];
				}
				else
				{
					sign = 0; corner = gridCorners[k];
				}
				ASSERT(corner < (1 << nQuantBits));

				sign.EncDec(pCoder, this, op);
				corner.EncDecVarSize(pCoder, this, op, nQuantBits);
			}
		}
		break;
	case DECODE:
		{
			nQuantBits.EncDec(pCoder, this, op);

			for (int k = 0; k < nrGridValues; k++)
			{
				sign.EncDec(pCoder, this, op);
				corner.EncDecVarSize(pCoder, this, op, nQuantBits);

				if (sign == 1)
					gridCorners.InsertElement(-corner, k);
				else
					gridCorners.InsertElement((int)corner, k);
			}
		}
		break;
	case CALC_LENGTH:
		{
			nQuantBits.EncDec(pCoder, this, op);

			for (int k = 0; k < nrGridValues; k++)
			{
				sign.EncDec(pCoder, this, op);
				corner.EncDecVarSize(pCoder, this, op, nQuantBits);
			}
		}
	}
}

void WMDecoderSpecificInfo :: DescEncDec (ODCoder *pCoder, OPTYPE op)
{
	hasScaleCoeff.EncDec(pCoder, this, op);
	if (hasScaleCoeff)
		NbBpSC.EncDec(pCoder, this, op);
	NbBPX.EncDec(pCoder, this, op);
  NbBPY.EncDec(pCoder, this, op);
  NbBPZ.EncDec(pCoder, this, op);

  Wtype.EncDec(pCoder, this, op);
  lift.EncDec(pCoder, this, op);
  NumberOfLevels.EncDec(pCoder, this, op);
  Xmax.EncDec(pCoder, this, op);
  Ymax.EncDec(pCoder, this, op);
  Zmax.EncDec(pCoder, this, op);
}

void FPSetDecoderSpecificInfo :: DescEncDec (ODCoder *pCoder, OPTYPE op)
{
	FPObjectType.EncDec(pCoder, this, op);
	MaxNbFootPrints.EncDec(pCoder, this, op);
	FootPrintNbBits.EncDec(pCoder, this, op);
	Step.EncDec(pCoder, this, op);
	NbBitsMetricError.EncDec(pCoder, this, op);
	MinX.EncDec(pCoder, this, op);
	MaxX.EncDec(pCoder, this, op);
	MinY.EncDec(pCoder, this, op);
	MaxY.EncDec(pCoder, this, op);
	DEFIdUsed.EncDec(pCoder, this, op);
	if (FPObjectType==1)
	{
		MinAltitude.EncDec(pCoder, this, op);
		MaxHeight.EncDec(pCoder, this, op);
		NbBitsZBuilding.EncDec(pCoder, this, op);
		NbBitsNbStories.EncDec(pCoder, this, op);
		NbBitsStoryHeight.EncDec(pCoder, this, op);
		NbBitsFacadeWidth.EncDec(pCoder, this, op);
		NbBitsNbFacadeCellsByStorey.EncDec(pCoder, this, op);
	}
}
