/***********************************************************************
This software module was originally developed by Alexander Zhirkov (
Moscow State University), Joo Yeon Han (Samsung Advance Institute of Technology)
in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1)
standard. This software module is an implementation of a part of one or
more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4
Systems (ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 
Systems (ISO/IEC 14496-1) free license to this software module or 
modifications thereof for use in hardware or software products claiming
conformance to the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to 
use this software module in hardware or software products are advised that
its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Samsung Advance Institute of Technology
retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using
the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
This copyright notice must be included in all copies or derivative works.
Copyright ?2001
************************************************************************

For more information or to receive an up to date version of this module,
you may call Joo Yeon Han at jyhan@sait.samsung.co.kr
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Depth Image-based Representation
FILE:       TBVOCompress.cpp
AUTHOR:     Joo Yeon Han (SAMSUNG AIT)
OVERVIEW:   Implementation of Depth Image-based Representation
-------------------------------------------------------------------------*/

/*\   Programmed by Alexander Zhirkov <azh@graphics.cs.msu.su>
\*/
#ifdef SAIT_CONFORMANCE_2003

#include "Common.h"
#include "Array.h"
#include "Entropy.h"
#include "BinaryArithmetic.h"
#include "TBVOCompress.h"

#include <assert.h>

//------------------------------------------------------------------------------
typedef EntropyModelContext         Ppm;
typedef EntropyProbabilisticContext Prb;
//------------------------------------------------------------------------------

#define ALLVOXELBYTEGROUP 22
#define ALLTRANSFORMS 48

struct VoxelByte
{
	int Group;  // number of group
	int Trans;  // transform
}
VoxelByteArray[256];

struct VoxelGroup
{
	int VoxelBytes; // number of voxel bytes in group
	int VoxelBits;  // number of bits in each voxel byte
	int Base;       // base voxel
}
VoxelGroupArray[ALLVOXELBYTEGROUP];

struct VoxelTransform
{
	uhr Transform[ALLTRANSFORMS];
};

Array<VoxelTransform> VoxelTransformArray;
Array<VoxelTransform> VoxelRevTransformArray;

int OrthogonalMatrices[ALLTRANSFORMS][9]=
{
	{ 1, 0, 0, 0, 1, 0, 0, 0, 1}, //3
	{ 0, 1, 0, 1, 0, 0, 0, 0, 1}, //0
	{ 1, 0, 0, 0, 0, 1, 0, 1, 0}, //1
	{-1, 0, 0, 0, 1, 0, 0, 0, 1}, //2
	{ 0, 1, 0, 0, 0, 1, 1, 0, 0}, //4
	{ 0,-1, 0, 1, 0, 0, 0, 0, 1}, //5
	{ 0, 0, 1, 1, 0, 0, 0, 1, 0}, //6
	{-1, 0, 0, 0, 0, 1, 0, 1, 0}, //7
	{ 0, 0, 1, 0, 1, 0, 1, 0, 0}, //8
	{ 0, 0,-1, 1, 0, 0, 0, 1, 0}, //9
	{ 0, 1, 0,-1, 0, 0, 0, 0, 1}, //10
	{ 0, 1, 0, 0, 0, 1,-1, 0, 0}, //11
	{ 0,-1, 0,-1, 0, 0, 0, 0, 1}, //12
	{ 0, 0, 1,-1, 0, 0, 0, 1, 0}, //13
	{ 0, 0, 1, 0, 1, 0,-1, 0, 0}, //14
	{ 0, 0,-1,-1, 0, 0, 0, 1, 0}, //15
	{ 0,-1, 0, 0, 0, 1, 1, 0, 0}, //16
	{ 0, 0,-1, 0, 1, 0, 1, 0, 0}, //17
	{ 0, 0, 1, 0,-1, 0, 1, 0, 0}, //18
	{ 0, 0, 1, 1, 0, 0, 0,-1, 0}, //19
	{ 0, 0,-1, 0,-1, 0, 1, 0, 0}, //20
	{ 1, 0, 0, 0,-1, 0, 0, 0, 1}, //21
	{ 1, 0, 0, 0, 0, 1, 0,-1, 0}, //22
	{-1, 0, 0, 0,-1, 0, 0, 0, 1}, //23
	{ 1, 0, 0, 0, 0,-1, 0, 1, 0}, //24
	{ 1, 0, 0, 0, 1, 0, 0, 0,-1}, //25
	{-1, 0, 0, 0, 0,-1, 0, 1, 0}, //26
	{ 0, 1, 0, 0, 0,-1, 1, 0, 0}, //27
	{ 0, 1, 0, 1, 0, 0, 0, 0,-1}, //28
	{ 0,-1, 0, 0, 0,-1, 1, 0, 0}, //29
	{-1, 0, 0, 0, 0, 1, 0,-1, 0}, //30
	{ 0, 0,-1, 1, 0, 0, 0,-1, 0}, //31
	{ 0, 0, 1,-1, 0, 0, 0,-1, 0}, //32
	{ 0, 0, 1, 0,-1, 0,-1, 0, 0}, //33
	{ 0, 0,-1,-1, 0, 0, 0,-1, 0}, //34
	{ 0,-1, 0, 0, 0, 1,-1, 0, 0}, //35
	{ 0,-1, 0, 1, 0, 0, 0, 0,-1}, //36
	{ 1, 0, 0, 0, 0,-1, 0,-1, 0}, //37
	{ 1, 0, 0, 0,-1, 0, 0, 0,-1}, //38
	{-1, 0, 0, 0, 0,-1, 0,-1, 0}, //39
	{ 0, 0,-1, 0,-1, 0,-1, 0, 0}, //40
	{ 0, 0,-1, 0, 1, 0,-1, 0, 0}, //41
	{-1, 0, 0, 0,-1, 0, 0, 0,-1}, //42
	{ 0,-1, 0, 0, 0,-1,-1, 0, 0}, //43
	{ 0,-1, 0,-1, 0, 0, 0, 0,-1}, //44
	{ 0, 1, 0, 0, 0,-1,-1, 0, 0}, //45
	{ 0, 1, 0,-1, 0, 0, 0, 0,-1}, //46
	{-1, 0, 0, 0, 1, 0, 0, 0,-1}  //47
};

int MatrixTransformVoxelByte(int b1,int* m,int isfor)
{
	int i,b2,x,y,z,x1,y1,z1;
	b2=0;
	forinc(i,8)
	if(b1&(1<<i))
	{
		x=i&1?2:-2,y=i&2?2:-2,z=i&4?2:-2;
		if(isfor){
			x1=m[0]*x+m[1]*y+m[2]*z;
			y1=m[3]*x+m[4]*y+m[5]*z;
			z1=m[6]*x+m[7]*y+m[8]*z;}
		else{
			x1=m[0]*x+m[3]*y+m[6]*z;
			y1=m[1]*x+m[4]*y+m[7]*z;
			z1=m[2]*x+m[5]*y+m[8]*z;}
		b2|=1<<((x1<0?0:1)|(y1<0?0:2)|(z1<0?0:4));
	}
	return b2;
}

int TransformVoxelByte(int b1,int t)
{
	return MatrixTransformVoxelByte(b1,OrthogonalMatrices[t],1);
}

int RevTransformVoxelByte(int b1,int t)
{
	return MatrixTransformVoxelByte(b1,OrthogonalMatrices[t],0);
}

void VolumeSimilarityInit()
{
	int i,j,t,a=0;
	forinc(i,256)
		VoxelByteArray[i].Group=-1;
	forinc(i,256)
	if(VoxelByteArray[i].Group<0)
	{
		VoxelGroupArray[a].Base=i;
		VoxelGroupArray[a].VoxelBytes=0;
		VoxelGroupArray[a].VoxelBits=0;
		forinc(j,8)
			VoxelGroupArray[a].VoxelBits+=i&(1<<j)?1:0;
		forinc(t,ALLTRANSFORMS)
		{
			j=TransformVoxelByte(i,t);
			if(!t || VoxelByteArray[j].Group<0)
			{
				VoxelByteArray[j].Group=a;
				VoxelByteArray[j].Trans=t;
				VoxelGroupArray[a].VoxelBytes++;
			}
		}
		a++;
	}

	VoxelTransformArray.AllocSet(256);
	VoxelRevTransformArray.AllocSet(256);
	forinc(i,256)
	{
		forinc(t,ALLTRANSFORMS)
			VoxelTransformArray[i].Transform[t]=TransformVoxelByte(i,t),
			VoxelRevTransformArray[i].Transform[t]=RevTransformVoxelByte(i,t);
	}

	assert(a==ALLVOXELBYTEGROUP);
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

void Entropy_Model_CodeSymbol
(
	BinaryArithmetic* rc,
	Ppm* pm,
	int *sym,
	int c,
	int isdec
)
{
	int lowint,value;
	Prb *p=&(pm->PDis[c]);

	assert(!pm->UseHash);

	if(!isdec)
	{
		ProbabilisticTreeFromIndex(p,*sym,&lowint);
		BinaryArithmeticEncode(rc,lowint,p->Prob[*sym],p->Tree[0]);
	}
	else
	{
		value = BinaryArithmeticValue(rc,p->Tree[0]);
		ProbabilisticTreeFromValue(p,value,sym,&lowint);
		assert(*sym<p->AllSym);
		BinaryArithmeticDecode(rc,lowint,p->Prob[*sym],p->Tree[0]);
	}
	ProbabilisticUpdate(p,*sym);
}

//------------------------------------------------------------------------------

typedef EntropyProbabilisticContext Prb;

struct ContextEntropyBVO
{
	BinaryArithmetic *RC;
	EntropyModelContext* PM;
	uhr* BVO;
	int Pos;   // recursion bvo position
	int IsDecode;
	int Model; // 0,1,2 if <0 => number
	int ACam;
};

#define DEF_AN (1)
#define DEF_K  (1<<3) 
#define DEF_DL (1<<8)
#define DEF_EDL (4)

#define HModel0_ADAPT 2
#define HModel1_ADAPT 3
#define HModel2_ADAPT 4

#define HModel0_DL (DEF_DL)
#define HModel1_DL (DEF_DL)
#define HModel2_DL (DEF_DL)

#define HModel0_MC (8192)
#define HModel1_MC (8192)
#define HModel2_MC (8192)

#define HModel0_LEN (1<<9)
#define HModel1_LEN (1<<11)

#define ADAPT(adaptlevel)  ((dbl)(DEF_K)*((dbl)((adaptlevel)>=0)?(adaptlevel):(dbl)1./((-adaptlevel))))

#define CAM_AN (1)
#define CAM_DL (1<<3) 
#define CAM_MC (1<<6) 
#define CAM_CONTEXT 256*9+4
#define CAM_ADAPT_STEP (1<<3)

#define MAXCUM 

int EntropyModelBVO(ContextEntropyBVO *p)
// return correct model, if needed than update mode
{
	int a=p->Model,i,j;
	Assert(a>=0,"<EntropyModelBVO>");
	Assert(p->PM->MaxCnt>=(1<<11)+256+1+1+1+1+1+1,"Too small Model model");
	if(a<0) return -a-1;

	if(a==0)
	{
		VolumeSimilarityInit();

		Entropy_ProbabilisticInitMaxCumul(
			&(p->PM->PDis[1]),
			0,256,
			ADAPT(HModel0_ADAPT),
			HModel0_MC);
		//-- Camera initialization
		Entropy_ProbabilisticInitMaxCumul(
			&(p->PM->PDis[CAM_CONTEXT]),
			0,p->ACam+1,
			CAM_ADAPT_STEP,
			CAM_MC);
		return 0;
	}
	if(a<HModel0_LEN)
		return 0;
	if(a==HModel0_LEN)
	{
		Array<int> Prob;
		Prob.AllocSet(256);
		int *prb=p->PM->PDis[1].Prob;
		forinc(i,ALLVOXELBYTEGROUP) Prob[i]=0;
		forinc(i,256) 				Prob[VoxelByteArray[i].Group]+=prb[i];
		forinc(i,ALLVOXELBYTEGROUP) Prob[i]/=VoxelGroupArray[i].VoxelBytes;
		forinc(i,256) 				prb[i]=Prob[VoxelByteArray[i].Group];
		forinc(i,256)
			Entropy_ProbabilisticInitMaxCumul(
				&(p->PM->PDis[i+1+1]),
				p->PM->PDis[1].Prob,256,
				ADAPT(HModel1_ADAPT),
				HModel1_MC);
		return 1;
	}
	if(a<HModel1_LEN)
		return 1;
	if(a==HModel1_LEN)
	{
		forinc2(i,j,256,8)
			Entropy_ProbabilisticInitMaxCumul(
				&(p->PM->PDis[(i<<3)+j+256+1+1]),
				p->PM->PDis[i+1+1].Prob,256,
				ADAPT(HModel2_ADAPT),
				HModel2_MC);
		return 2;
	}
	return 2;
}

int CodeRecursBVO
(
	ContextEntropyBVO *p,
	int cntxt,
	int n
)
{
	int i;

	int sym,sym1;
	unt c[3];
	int model=EntropyModelBVO(p);

	int b1=cntxt>>3,b2,b;
	b2=cntxt&7;
	b=VoxelByteArray[b1].Group;

	i=RevTransformVoxelByte(1<<b2,VoxelByteArray[b1].Trans);
	if(i&15) if(i&3) b2=i==1?0:1;
				else b2=i==4?2:3;
	else     if(i&48)b2=i==16?4:5;
				else b2=i==64?6:7;

	cntxt=(b<<3)|(b2);

	switch(model)
	{
		case 0: c[0]=1; break;
		case 1: c[0]=(cntxt>>3)+1+1; break;
		case 2: c[0]=cntxt+256+1+1; break;
		default: Assert(0,"<CodeRecursBO> bad model");
	}

	if(p->Model>=0) p->Model++;

	if(!p->IsDecode)
	{
		sym=p->BVO[p->Pos++];

		sym1=VoxelRevTransformArray[sym].Transform[VoxelByteArray[b1].Trans];

		Entropy_Model_CodeSymbol(p->RC,p->PM,&sym1,c[0],0);
	}
	else
	{
		Entropy_Model_CodeSymbol(p->RC,p->PM,&sym1,c[0],1);

		sym=VoxelTransformArray[sym1].Transform[VoxelByteArray[b1].Trans];

		p->BVO[p->Pos++]=sym;
	}
	if(n==1)
		return sym;
	for(i=0;i<8;i++)
	if(sym&(1<<i))
		CodeRecursBVO(p,(sym<<3)|i,n-1);
	return sym;
}

int Entropy_CodeFrameBVO // return number of readed frames
(
	BinaryArithmetic *RC,
	EntropyModelContext* PM,
	uhr *bvo,
	int n,
	int isdecode,
	int model
)
{
	ContextEntropyBVO bv;
	bv.RC=RC;
	bv.PM=PM;
	bv.IsDecode=isdecode;
	bv.Pos=0;
	bv.BVO=bvo;
	bv.ACam=1;
	bv.Model=model;
	CodeRecursBVO(&bv,255,n);
	return bv.Pos; //BVO-bvo;
}

//======================================================
// TBVO Compression
//------------------------------------------------------

struct TBVOCodeNode
{
	int ParentSym,NodeNum,NeedCam;
	TBVOCodeNode(){Set(0,0,1);}
	int Context(){ return (ParentSym<<3)|NodeNum;}
	TBVOCodeNode(int ps,int nn,int needcam=1){Set(ps,nn,needcam);}
	void Set(int ps,int nn,int needcam){ParentSym=ps;NodeNum=nn;NeedCam=needcam;}
};

struct TBVOCodeNodeList:
	public Array<TBVOCodeNode>
{
};

int Entropy_CodeFrameSeparatedTBVO // return number of readed bytes
(
	BinaryArithmetic *RC_t,
	BinaryArithmetic *RC_bvo,
	EntropyModelContext* PM,

	int *array_t,
	int *array_bvo,
	int &pos_t,
	int &pos_bvo,

	int n,
	int acam,
	int isencode,
	int len
)
{
	Array<TBVOCodeNodeList> nodes;
	ContextEntropyBVO bv,*p=&bv;
	int cam;

	Exit(n>=0 && acam>=0 && len>=0,"<Entropy_CodeFrameTBVO> bad params");

	bv.PM=PM;

	bv.IsDecode=!isencode;

	bv.Pos=0;
	bv.Model=len;
	bv.ACam=acam;

	nodes.AllocSet(n+1);

	int i;
	int i0, // down node index
		curlev;
	int sym,sym1;
	unt c[3];

	forinc(curlev,n+1)
		nodes[curlev].SetPoint(0),
		nodes[curlev].AutoSize(true);

	nodes[0].Add(TBVOCodeNode(0,0,acam==0?0:1));

	forinc(curlev,n)
	{
		forinc(i0,nodes[curlev].All())
		{
			TBVOCodeNode &node=nodes[curlev][i0];
			
			int cntxt=node.Context();
			int model=EntropyModelBVO(p);
			int needcam=node.NeedCam;

			int b1=cntxt>>3,b2,b;
			b2=cntxt&7;
			b=VoxelByteArray[b1].Group;

			i=RevTransformVoxelByte(1<<b2,VoxelByteArray[b1].Trans);
			if(i&15) if(i&3) b2=i==1?0:1;
					 else b2=i==4?2:3;
			else     if(i&48)b2=i==16?4:5;
					 else b2=i==64?6:7;

			cntxt=(b<<3)|(b2);

			switch(model)
			{
				case 0: c[0]=1; break;
				case 1: c[0]=(cntxt>>3)+1+1; break;
				case 2: c[0]=cntxt+256+1+1; break;
				default: Assert(0,"<CodeRecursBVO> bad model");
			}

			if(bv.Model>=0) bv.Model++; ///

			if(isencode)
			{
				sym=array_bvo[pos_bvo++];  //f_byte(f_tbvo); //tbvo[pos++];
				sym1=VoxelRevTransformArray[sym].Transform[VoxelByteArray[b1].Trans];
				Entropy_Model_CodeSymbol(RC_bvo,PM,&sym1,c[0],0);

				if(needcam)
				{
					cam=array_t[pos_t++];  //f_byte(f_tbvo);
					Assert(cam>=0 && (cam<=acam || !acam),"Source TBVO-stream is corrupted "+cam+", at: "+curlev+","+i0);
					if(acam)
  						Entropy_Model_CodeSymbol(RC_t,PM,&cam,CAM_CONTEXT,0);
					needcam=!cam;
					Exit(!needcam || curlev!=n-1,
						"<Entropy_CodeFrameTBVO> Current compression method doesn't support cameras 1x1x1 at cam_pos= "+pos_t+" of "+array_t.All()+", and bvo_pos="+pos_bvo+" of "+array_bvo.All());
				}
			}
			else
			{
				Entropy_Model_CodeSymbol(RC_bvo,PM,&sym1,c[0],1);

				sym=VoxelTransformArray[sym1].Transform[VoxelByteArray[b1].Trans];

				array_bvo[pos_bvo++]=sym;
				
				if(needcam)
				{
					Entropy_Model_CodeSymbol(RC_t,PM,&cam,CAM_CONTEXT,1),

					array_t[pos_t++]=cam;  //f_byte(f_tbvo);
					needcam=!cam;
				}
			}

			if(curlev!=n-1)
			forinc(i,8)
			if(sym&(1<<i))
				nodes[curlev+1].Add(TBVOCodeNode(sym,i,needcam));

		}
	}

	return bv.Model;//Pos;
}

#endif