/*!
 ************************************************************************
 *  \file
 *     LiftCodec.cpp
 *  \brief
 *     LiftCodec class.
 *  \author
 *     Copyright (C) 2007  ARTEMIS Department INT/GET, Paris, France.
 *
 *     Khaled MAMOU               <khaled.mamou@int-evry.fr>
 *
 *	   Institut National des Telecommunications		tel.: +33 (0)1 60 76 40 94
 *     9, Rue Charles Fourier,						fax.: +33 (0)1 60 76 43 81
 *     91011 Evry Cedex France
 *
 ************************************************************************
 */
#include "LiftCodec.h"
#include "LiftScheme.h"
#include "CABAC.h"
#include <math.h>

LiftCodec::LiftCodec(int len, int nbr){
	len_ = len;
	nbr_ = nbr;
	function_ = new double*[nbr_];
	int v = 0;
	for (v = 0; v < nbr_; v++){	
		function_[v] = new double[len_];
	}
	for (v = 0; v < nbr_; v++){	
		for (int k = 0; k < len_; k++){
			function_[v][k] = 0.0;
		}
	}
}

LiftCodec::~LiftCodec(void){
	// cleraning the structure
	for (int v = 0; v < nbr_; v++){	
		delete[] function_[v];
	}
	delete[] function_;
}
void LiftCodec::LiftLS(bool directTransform) {
	LiftScheme lift(len_);
	int k = 0;
	int v = 0;
	for (v = 0; v < nbr_; v++){
		for (k = 0; k < len_; k++) {
			lift.Set(k, function_[v][k]);
		}
		lift.LiftLS(directTransform);
		for (k = 0; k < len_; k++) {
			lift.Get(k, function_[v][k]);
		}
	}
}
int LiftCodec::Round(double a){
	if (a >= 0.0) return (int) (a+0.5);
	return -(int) (-a+0.5);
}




int LiftCodec::Quantize(unsigned char nbits, Vec3D<int> *pcIntVertices, int iStride, int iCoord, float *pfMax, float *pfMin){

	int k = 0;
	int v = 0;
	// we compute the max for abs(DCT)
	*pfMax = (float) function_[0][0];
	*pfMin = (float) function_[0][0];
	for (k = 0; k < len_; k++) {
		for (v = 0; v < nbr_; v++){
			if ( function_[v][k] >  *pfMax)  *pfMax = (float) function_[v][k];
			if ( function_[v][k] <  *pfMin)  *pfMin = (float) function_[v][k];
		}
	}

	// Constant for uniform quantization
	double delta = 1.0;
	if ((*pfMax-*pfMin) != 0.0) delta = (double) ((1 << nbits) - 1) / (*pfMax-*pfMin);

	// Quantize function
	for (k = 0; k < len_; k++) {
		for (v = 0; v < nbr_; v++){
			function_[v][k] = (double) Round((function_[v][k] - *pfMin) * delta);
		}
	}

	// we apply the lifting scheme
	LiftLS(true);

	for (k = 0; k < len_; k++) {
		for (v = 0; v < nbr_; v++){
			pcIntVertices[v * iStride + k].data[iCoord] = (int) function_[v][k];
		}
	}

	// we apply the unlifting scheme
	LiftLS(false);	

	// unquantize function
	for (k = 0; k < len_; k++) {
		for (v = 0; v < nbr_; v++){
			function_[v][k] = function_[v][k] / delta + *pfMin;
		}
	}
	return 0;
}

int LiftCodec::Unquantize(unsigned char nbits, Vec3D<int> *pcIntVertices, int iStride, int iCoord, float *pfMax, float *pfMin){
	// we compute the max for abs(DCT)
	int k = 0;
	int v = 0;
	if (nbits > 0) {
		for (k = 0; k < len_; k++) {
			for (v = 0; v < nbr_; v++){
				function_[v][k]=0.0f;
			}
		}

		for (k = 0; k < len_; k++) {
			for (v = 0; v < nbr_; v++){
				function_[v][k] = (double) pcIntVertices[v * iStride + k].data[iCoord];
			}
		}
		
		// we apply the unlifting scheme
		LiftLS(false);	


		// Constant for uniform quantization
		double delta = 1.0;
		if ((*pfMax-*pfMin) != 0.0) delta = (double) ((1 << nbits) - 1) / (*pfMax-*pfMin);

		// unquantize function
		for (k = 0; k < len_; k++) {
			for (v = 0; v < nbr_; v++){
				function_[v][k] = function_[v][k] / delta + *pfMin;
			}
		}
	}
	return 0;
}