#include "FP.h"


//#define SHOW_LINES		// build and display cell boundary


enum DISPLAY_TYPE {
	EMPTY=0,
	QUAD,
	TRIANGLES,
	MODEL3D
};


class FPFacade;


class ModelGeometry
{
public:
	ModelGeometry() : m_nVtx(0), m_aVtx(NULL), m_nIdx(0), m_aIdx(NULL) {};
	ModelGeometry(int nVtx, float* aVtx, int nIdx, int* aIdx) : m_nVtx(nVtx), m_aVtx(aVtx), m_nIdx(nIdx), m_aIdx(aIdx) {};
	~ModelGeometry() {};

public:
	int		m_nVtx;		// number of vertices
	float*	m_aVtx;		// coord of vertices
	int		m_nIdx;		// size of indexfacette
	int*	m_aIdx;		// index facette
};



/*****************************************************************************
		CLASS:	FacadeBuilder
		computes triangles/quads for OPENGL display
		computes model3D section + positionning info
		uses facade.lib
*****************************************************************************/
class FacadeBuilder
{
public:
	FacadeBuilder(FPFacade* pFacade);
	~FacadeBuilder(void);

	void Build(void);

public:
	// to retrieve OPENGL info
	float* GetVertices() const { return m_Vertices; };
	int GetNbVertices() const { return m_NbVertices; };
	int GetNbTriangles() const { return m_NbTriangles; };
	int GetNbQuads() const { return m_NbQuads; };
	unsigned int* GetIndicesQuads() { return m_IndicesQuads; };
	unsigned int** GetIndicesLines() { return m_IndicesLines; };
	unsigned int* GetIndicesTriangles() { return m_IndicesTris; };
	DISPLAY_TYPE GetDisplayType(void) { return m_DisplayType; };

public:
	// to retrieve model info for positionning in scene
	int GetNbModels(void) { return m_NbModels; };
	float* GetModelTranslations() const { return m_Translations; };
	float GetScaleModelX() const { return m_ScaleModelX; };
	float GetScaleModelY() const { return m_ScaleModelY; };
	float GetScaleModelZ() const { return m_ScaleModelZ; };
	ZArray<ModelGeometry*>* GetModel() { return m_Model; };
	
public:
	// to build model
	void AddModelGeometry(int nVtx, float* aVtx, int nIdx, int* aIdx);

private:
	// to compute facade geometry
	void BuildQuadVertexArray();
	void BuildLines();
	void BuildQuads();
	void BuildTriangles();
	void BuildModel();


private:
	// OPENGL
	DISPLAY_TYPE	m_DisplayType;		// quad or triangle?
	int				m_NbVertices;		// nb of vertices
	int				m_NbQuads;			// nb of quads to display
	int				m_NbTriangles;		// nb of triangles to display
	int				m_NbLines;			// number of lines
	float*			m_Vertices;			// array of vertices coordinates (for GLInterleaved Array)
	unsigned int*	m_IndicesQuads;		// array of vertex forming quads
	unsigned int*	m_IndicesTris;		// array of vertex indices forming the display element (triangles)
	unsigned int**	m_IndicesLines;		// array of vertex forming lines

	// model3D info
	int				m_NbModels;			// number of models (because model can be repeated)
	float*			m_Translations;		// array of translations (because model can be repeated)
	float			m_ScaleModelX;		// model scale in X to fit facade element
	float			m_ScaleModelY;		// model scale in Y to fit facade element
	float			m_ScaleModelZ;		// model scale in Z to resize model correctly
	ZArray<ModelGeometry*>* m_Model;	// model is an array of geometries

	// facade element
	FPFacade*		m_pFacade;			// corresponding facade element
};





enum PRIM_TYPE {
	NO_PRIMITIVE=0,
	PRIM_TEXTURE,
	PRIM_3DMODEL
};

enum CELL_STATE {
	STATE_WAITING=0,
	STATE_WAITING_MODEL,
	STATE_WAITING_CHILD_MODEL,
	STATE_READY,
	STATE_DONE
};

/*****************************************************************************
		CLASS:	FPFacade
		Describes decoded object for bitstream
		calls FacadeBuilder to compute OPENGL display + position 3D models
*****************************************************************************/
class FPFacade
{
public:
	FPFacade(void);
	FPFacade(float x, float y, float z, float xfacade, float yfacade, float zfacade, float width, float height);
	~FPFacade();

	FPFacade(const FPFacade & in);
	FPFacade & operator =(const FPFacade &in);
	
public:
	// to retrieve bitstream info
	float GetWidthRatio() const { return m_WidthRatio; };
	float GetHeightRatio() const { return m_HeightRatio; };
	int	  GetNbStories() const;
	float GetXScale() const { return m_XScale; };
	float GetYScale() const { return m_YScale; };
	float GetXPosition() const { return m_XPosition; };
	float GetYPosition() const { return m_YPosition; };
	float GetXRepeatInterval() const { return m_XRepeatInterval; };
	float GetYRepeatInterval() const { return m_YRepeatInterval; };
	bool GetRepeat() const { return m_bRepeat; };
	PRIM_TYPE GetPrimitiveType() const { return m_PrimitiveType; };
	char* GetURL() const { return (char*)m_URL; };
	ZArray<float>* GetStoriesHeight() const { return m_StoriesHeight; };
	ZArray<ZArray<FPFacade*>* >* GetElements() const { return m_Elements; };

public:
	// to set bitstream info
	void SetWidthRatio(float width);
	void SetHeightRatio(float height);
	void SetXScale(float xscale) { m_XScale=xscale; };
	void SetYScale(float yscale) { m_YScale=yscale; };
	void SetXPosition(float xposition) { m_XPosition=xposition; };
	void SetYPosition(float yposition) { m_YPosition=yposition; };
	void SetXRepeat(float xrepeatinterval) { m_XRepeatInterval=xrepeatinterval; };
	void SetYRepeat(float yrepeatinterval) { m_YRepeatInterval=yrepeatinterval; };
	void SetRepeat(bool repeat) { m_bRepeat=repeat; };
	void SetPrimitiveType(int type) { m_PrimitiveType = (PRIM_TYPE)type; };
	void SetURL(char* Url, int nUrlLen);
	void SetStoriesHeight(ZArray<float>* sheight) { m_StoriesHeight=sheight; };
	void SetElements(ZArray<ZArray<FPFacade*>* >* elements) { m_Elements=elements; };

public:
	// test functions
	bool HasChild(void) const;
	bool HasTexture(void) const;
	bool HasModel3D(void) const;
//	bool HasFilledElements(void) const;

	// 3d construct
	void InitFacade(void);
	void ComputeFacade(void);
	void ComputeElement(void) { m_FacadeBuilder->Build(); };

public:
	// to retrieve OPENGL result
	float* GetVertices() const { return m_FacadeBuilder->GetVertices(); };
	int GetNbVertices() const { return m_FacadeBuilder->GetNbVertices(); };
	int GetNbTriangles() const { return m_FacadeBuilder->GetNbTriangles(); };
	int GetNbQuads() const { return m_FacadeBuilder->GetNbQuads(); };
	unsigned int* GetIndicesQuads() const { return m_FacadeBuilder->GetIndicesQuads(); };
	unsigned int** GetIndicesLines() const { return m_FacadeBuilder->GetIndicesLines(); };
	unsigned int* GetIndicesTriangles() const { return m_FacadeBuilder->GetIndicesTriangles(); };
	DISPLAY_TYPE GetDisplayType(void) { return m_FacadeBuilder->GetDisplayType(); };

public:
	// to retrieve geometry info
	float GetX() { return m_xBottomLeft; };
	float GetY() { return m_yBottomLeft; };
	float GetZ() { return m_zBottomLeft; };
	float GetXFacade() { return m_xFacade; };
	float GetYFacade() { return m_yFacade; };
	float GetZFacade() { return m_zFacade; };
	float GetXNormal() { return m_xNormal; };
	float GetYNormal() { return m_yNormal; };
	float GetZNormal() { return m_zNormal; };
	float GetWidth() { return m_elementwidth; };
	float GetHeight() { return m_elementheight; };
	float GetParentWidth() { return m_parentwidth; };
	float GetParentHeight() { return m_parentheight; };

	// to set geometry info
	void SetX(float x) { m_xBottomLeft=x; };
	void SetY(float y) { m_yBottomLeft=y; };
	void SetZ(float z) { m_zBottomLeft=z; };
	void SetXFacade(float x) { m_xFacade=x; };
	void SetYFacade(float y) { m_yFacade=y; };
	void SetZFacade(float z) { m_zFacade=z; };
	void SetXNormal(float x) { m_xNormal=x; };
	void SetYNormal(float y) { m_yNormal=y; };
	void SetZNormal(float z) { m_zNormal=z; };
	void SetParentWidth(float width) { m_parentwidth=width; };
	void SetParentHeight(float height) { m_parentheight=height; };
	
	// to update geometry info (at decoding)
	void ShiftRight(float posH);
	void ShiftUp(float posV);

private:
	// to build list of filled children
	void BuildFilledElements(ZArray<FPFacade*>** pElements);

public:
	// to access list of filled children
	int GetNbFilledChildren(void);
	ZArray<FPFacade*>* GetFilledChildren(void) const { return m_FilledChildren; };

public:
	// to retrieve model 3d info
	int GetNbModels() const { return m_FacadeBuilder->GetNbModels(); };
	float* GetModelTranslations() const { return m_FacadeBuilder->GetModelTranslations(); };
	float GetScaleModelX() const { return m_FacadeBuilder->GetScaleModelX(); };
	float GetScaleModelY() const { return m_FacadeBuilder->GetScaleModelY(); };
	float GetScaleModelZ() const { return m_FacadeBuilder->GetScaleModelZ(); };
	ZArray<ModelGeometry*>* GetModel(void) { return m_FacadeBuilder->GetModel(); };

public:
	// to build model3D
	void AddModelGeometry(int nVtx, float* aVtx, int nIdx, int* aIdx) { m_FacadeBuilder->AddModelGeometry(nVtx, aVtx, nIdx, aIdx);};

public:
	// to set/update state
	int  GetState(void) const { return m_State; };
	void SetState(CELL_STATE state) { m_State = state; };

	inline MediaObject* GetMediaObject() const { return m_pNode; };

	inline void SetMediaObject ( MediaObject* mo) { m_pNode=mo; };


private:
	// bitstream info
	float	m_WidthRatio;
	float	m_HeightRatio;
	float	m_XScale;
	float	m_YScale;
	float	m_XPosition;
	float	m_YPosition;
	float	m_XRepeatInterval;
	float	m_YRepeatInterval;
	bool	m_bRepeat;
	char*	m_URL;
	PRIM_TYPE	m_PrimitiveType;
	ZArray<float>* m_StoriesHeight;
	ZArray<ZArray<FPFacade*>* >* m_Elements;

	MediaObject* m_pNode;

	// geometry info
	float   m_xBottomLeft, m_yBottomLeft, m_zBottomLeft;	// coordinates of bottom left point
	float   m_xFacade, m_yFacade, m_zFacade;		// vector along facade
	float   m_xNormal, m_yNormal, m_zNormal;		// normal to this facade
	float	m_elementwidth, m_elementheight;
	float	m_parentwidth, m_parentheight;

	// filled children
	ZArray<FPFacade*>*	m_FilledChildren;

	FacadeBuilder* m_FacadeBuilder;
	CELL_STATE	m_State;
};