// CompressedCoordinateInterpolator.h: coordinate interpolator encoder
// Author: Shinjun Lee, Samsung AIT
// Date: 2001-12-27
// 4th revision: 2002-04-25

#ifndef _COMPRESSED_COORDINATE_INTERPOLATOR_H_
#define _COMPRESSED_COORDINATE_INTERPOLATOR_H_

#include <fstream.h>

#include "../include/KeyHeader.h"
#include "../include/CoordIKeyValueHeader.h"
#include "../include/Key.h"
#include "../include/CoordIKeyValue.h"
#include "../include/CCommandLineProcessor.h"
#include "../include/CParser.h"

#define MACRO_FILE_OF_TOTAL_BITS "bits_total.txt"			// file name of total bits

#define AACTotal() (codec._ostrm->Outbuf.nbytes) * 8 + (codec._ostrm->Outbuf.nbits)

const int CMAX_DIGIT = 15;

class CompressedCoordinateInterpolator
	{
  private:
	KeyHeader				kHeader;						// class of key header bitstream
	CoordIKeyValueHeader	coordIKVHeader;					// class of key value header bitstream
	Key						k;								// class of key bitstream
	CoordIKeyValue			coordIKeyValue;					// class of key value bitstream

	char			m_szInputCIFile[_MAX_PATH];				// input text file
	char			m_szBitstreamFile[_MAX_PATH];			// bitstream file name

	// variables: bits of bitstream
	int				m_nTotalBits;							// total bits of coordinate interpolator

  public:
	CompressedCoordinateInterpolator(void)
		{
		memset((void*)m_szInputCIFile, 0, _MAX_PATH);
		memset((void*)m_szBitstreamFile, 0, _MAX_PATH);
		}

	CompressedCoordinateInterpolator(int argc, char* argv[])
		{
		memset((void*)m_szInputCIFile, 0, _MAX_PATH);
		memset((void*)m_szBitstreamFile, 0, _MAX_PATH);

		// command line processor
		CCommandLineProcessor oCLProcessor;
		oCLProcessor.setEncoding();
		oCLProcessor.setArguments(argc, argv);

		strcpy(m_szInputCIFile, oCLProcessor.getInputCIFile());
		strcpy(m_szBitstreamFile, oCLProcessor.getBitstreamFile());

		// print: file name
		cout << "Input CI file = " << m_szInputCIFile << endl;
		cout << "Output bitstream file = " << m_szBitstreamFile << endl;
		cout << endl;

		// parser
		CParser oParser;
		oParser.setInputCIFile(oCLProcessor.getInputCIFile());
		
		// read: key, keyValue data of input file
		oParser.readKey(k.dKey, kHeader.nNumberOfKey, kHeader.nKeyDigit);
		oParser.readKeyValue(coordIKeyValue.dKeyValue, kHeader.nNumberOfKey, coordIKVHeader.nNumberOfCoord, coordIKVHeader.nKVDigit);

		kHeader.nKeyQBit = oCLProcessor.getQStep_key();
		coordIKVHeader.nKVQBit = oCLProcessor.getQStep_keyValue();
		coordIKVHeader.bTranspose = oCLProcessor.IsTransposed();
		kHeader.bIsLinearKeyUsed = oCLProcessor.IsLinearKeyUsed();
		kHeader.bIsSADUsed = oCLProcessor.IsSADUsed();
		}

	~CompressedCoordinateInterpolator()
		{
		}

	// encode: data of coordinate interpolator
	void encode(void)
		{
		// key header encoding
		kHeader.encodingProcess();

		coordIKVHeader.encodingProcess();

		// key encoding
		k.encodingProcess(kHeader);

		coordIKeyValue.encodingProcess(coordIKVHeader, kHeader.nNumberOfKey);
		
		// make: bitstream of key header, keyValue header, key, keyValue
		makeBitstream(m_szBitstreamFile);

		// dump: bitstream data
		//dump(m_szInputCIFile);
		//dump_real("Real");
		}

	// make: data of bitstream
	void makeBitstream(char* szBitstreamFile)
		{
		int			nBitsOfKeyHeader = 0;					// bits of key header
		int			nBitsOfAACKey = 0;						// bits of key in aac coding
		int			nBitsOfKeyValueHeader = 0;				// bits of keyValue header
		int			nBitsOfAACKeyValue = 0;					// bits of keyValue in aac coding
		int			i = 0;									// index

		// aac codec
		CAC::_shift = 0; 
		COutStream ostrm;
		CAC_QF codec(ostrm);
		bool bSuccess = ostrm.openenc(szBitstreamFile);

		// make: bitstream of header
		kHeader.makeBitstream(codec, nBitsOfKeyHeader);
		coordIKVHeader.makeBitstream(codec, nBitsOfKeyValueHeader);

		// byte align
		int nRest = 8-codec._ostrm->Outbuf.nbits;
		for(i = 0; i < nRest; i++)
			codec._ostrm->putbit(1);

		// start: aac encoding
		codec.start_encode();
		CQF_Status::reset();
		k.InitializeModel();
		coordIKeyValue.InitializeModel();

		// make: bitstream of aac data
		k.makeBitstream(codec, kHeader, nBitsOfAACKey);
		coordIKeyValue.makeBitstream(codec, coordIKVHeader, kHeader.nNumberOfKey, nBitsOfAACKeyValue);

		// finish: aac encoding
		codec.finish_encode();
		ostrm.close_enc();

		// print: bits of key coding
		cout << "Bits of key header = " << nBitsOfKeyHeader << endl;
		cout << "Bits of keys = " << nBitsOfAACKey << endl;
		cout << "Total key bits = " << nBitsOfKeyHeader + nBitsOfAACKey << endl;
		cout << endl;

		// print: bits of keyValue coding
		cout << "Bits of keyValue header = " << nBitsOfKeyValueHeader << endl;
		cout << "Bits of keyValues = " << nBitsOfAACKeyValue << endl;
		cout << "Total keyValue bits = " << nBitsOfKeyValueHeader + nBitsOfAACKeyValue << endl;
		cout << "-----------------------------------------------------------------------" << endl;
		cout << endl;

		// print: total bits of coordinate interpolator
		m_nTotalBits = (codec._ostrm->Outbuf.nbytes)*8 + (codec._ostrm->Outbuf.nbits);
		cout << "Total bits = " << m_nTotalBits << endl;
		cout << "-----------------------------------------------------------------------" << endl;

		// write: dump files
		//*//
		ofstream fout_bits;
		fout_bits.open("bits_k.txt", ios::app);
		fout_bits << nBitsOfKeyHeader + nBitsOfAACKey << endl;
		fout_bits.close();

		fout_bits.open("bits_kv.txt", ios::app);
		fout_bits << nBitsOfKeyValueHeader + nBitsOfAACKeyValue << endl;
		fout_bits.close();
		
		ofstream fout_q;
		fout_q.open("q_k.txt", ios::app);
		fout_q << kHeader.nKeyQBit << endl;
		fout_q.close();
		
		fout_q.open("q_kv.txt", ios::app);
		fout_q << coordIKVHeader.nKVQBit << endl;
		fout_q.close();
		
		fout_bits.open("bits_total.txt", ios::app);
		fout_bits << m_nTotalBits << endl;
		fout_bits.close();
		//*/

		ostrm.bfree();
		}

  private:
	// dump: each data of key, keyValue
	void dump(char* szInputCIFile)
		{
		int			j = 0;									// index

		// write: DPCM mode
		ofstream fout_dm;
		fout_dm.open("DPCMMode.txt");
		fout_dm << szInputCIFile << endl;
		for(j = 0; j < coordIKVHeader.nNumberOfCoord; j++)
			{
			fout_dm << j << "\t";
			fout_dm << coordIKeyValue.nDPCMMode[j*3] << "\t";
			fout_dm << coordIKeyValue.nDPCMMode[j*3+1] << "\t";
			fout_dm << coordIKeyValue.nDPCMMode[j*3+2] << endl;
			}
		fout_dm.close();

		// write: quantization selection flag
		ofstream fout_flag("QSelFlag_kv.txt");
		fout_flag << szInputCIFile << endl;
		for(j = 0; j < coordIKVHeader.nNumberOfCoord; j++)
			{
			fout_flag << j << "\t";
			fout_flag << coordIKeyValue.bSelFlag[j*3] << "\t";
			fout_flag << coordIKeyValue.bSelFlag[j*3+1] << "\t";
			fout_flag << coordIKeyValue.bSelFlag[j*3+2] << endl;
			}
		fout_flag.close();

		// write: AQStep
		ofstream fout_aqstep("AQStep_kv.txt");
		fout_aqstep << szInputCIFile << endl;
		for(j = 0; j < coordIKVHeader.nNumberOfCoord; j++)
			{
			fout_aqstep << j << "\t";
			if(coordIKeyValue.bSelFlag[j*3] == true)
				fout_aqstep << coordIKeyValue.nKVACodingBit[j*3] << "\t";
			else
				fout_aqstep << 0 << "\t";
			if(coordIKeyValue.bSelFlag[j*3+1] == true)
				fout_aqstep << coordIKeyValue.nKVACodingBit[j*3+1] << "\t";
			else
				fout_aqstep << 0 << "\t";
			if(coordIKeyValue.bSelFlag[j*3+2] == true)
				fout_aqstep << coordIKeyValue.nKVACodingBit[j*3+2] << endl;
			else
				fout_aqstep << 0 << endl;
			}
		fout_aqstep.close();
		
		// write: reference vertex
		ofstream fout_base("RefVertex_kv.txt");
		fout_base << szInputCIFile << endl;
		for(j = 0; j < coordIKVHeader.nNumberOfCoord; j++)
			{
			fout_base << j << "\t";
			if(coordIKeyValue.bSelFlag[j*3] == true)
				fout_base << coordIKeyValue.nRefVertex[j*3] << "\t";
			else
				fout_base << 0 << "\t";
			if(coordIKeyValue.bSelFlag[j*3+1] == true)
				fout_base << coordIKeyValue.nRefVertex[j*3+1] << "\t";
			else
				fout_base << 0 << "\t";
			if(coordIKeyValue.bSelFlag[j*3+2] == true)
				fout_base << coordIKeyValue.nRefVertex[j*3+2] << endl;
			else
				fout_base << 0 << endl;
			}
		fout_base << endl;
		fout_base.close();

		// write: qmin, qmax
		ofstream fout_qmm("QMinQMax.txt");
		fout_qmm << szInputCIFile << endl;
		for(j = 0; j < coordIKVHeader.nNumberOfCoord; j++)
			{
			fout_qmm << j << "\t";
			fout_qmm << coordIKeyValue.nQMin[j*3] << "\t";
			fout_qmm << coordIKeyValue.nQMax[j*3] << "\t";
			fout_qmm << coordIKeyValue.nQMin[j*3+1] << "\t";
			fout_qmm << coordIKeyValue.nQMax[j*3+1] << "\t";
			fout_qmm << coordIKeyValue.nQMin[j*3+2] << "\t";
			fout_qmm << coordIKeyValue.nQMax[j*3+2] << endl;
			}
		fout_qmm.close();

		// write: qmin, qmax after dpcm & cq
		fout_qmm.open("QMinQMax-aac.txt");
		fout_qmm << szInputCIFile << endl;
		for(j = 0; j < coordIKVHeader.nNumberOfCoord; j++)
			{
			fout_qmm << j << "\t";
			fout_qmm << coordIKeyValue.oDPCM_X_Min.getData_dpcm(j) << "\t";
			fout_qmm << coordIKeyValue.oDPCM_X_Max.getData_dpcm(j) << "\t";
			fout_qmm << coordIKeyValue.oDPCM_Y_Min.getData_dpcm(j) << "\t";
			fout_qmm << coordIKeyValue.oDPCM_Y_Max.getData_dpcm(j) << "\t";
			fout_qmm << coordIKeyValue.oDPCM_Z_Min.getData_dpcm(j) << "\t";
			fout_qmm << coordIKeyValue.oDPCM_Z_Max.getData_dpcm(j) << endl;
			}
		fout_qmm.close();
		}

	// Dump: real data File
	void dump_real(char* szString)
		{
		char szFile[_MAX_FNAME];				// File name
		int i = 0;
		int j = 0;

		ofstream fout;
		strcpy(szFile, szString);
		strcat(szFile, ".txt");
		fout.open(szFile);

		fout << szString << "(x)\t";
		for(i = 0; i < kHeader.nNumberOfKey; i++)
			fout << i << "\t";
		fout << endl;
		for(j = 0; j < coordIKVHeader.nNumberOfCoord; j++)
			{
			fout << j;
			if(coordIKeyValue.bSelFlag[j*3] == true)
				{
				for(i = 0; i < kHeader.nNumberOfKey; i++)
 					fout << "\t" << coordIKeyValue.dKeyValue[i*coordIKVHeader.nNumberOfCoord*3+j*3];
				}
			else
				{
				i = 0;
				fout << "\t" << coordIKeyValue.dKeyValue[i*coordIKVHeader.nNumberOfCoord*3+j*3];
				}
			fout << endl;
			}
		fout << endl;

		fout << szString << "(y)\t";
		for(i = 0; i < kHeader.nNumberOfKey; i++)
			fout << i << "\t";
		fout << endl;
		for(j = 0; j < coordIKVHeader.nNumberOfCoord; j++)
			{
			fout << j;
			if(coordIKeyValue.bSelFlag[j*3+1] == true)
				{
				for(i = 0; i < kHeader.nNumberOfKey; i++)
					fout << "\t" << coordIKeyValue.dKeyValue[i*coordIKVHeader.nNumberOfCoord*3+j*3+1];
				}
			else
				{
				i = 0;
				fout << "\t" << coordIKeyValue.dKeyValue[i*coordIKVHeader.nNumberOfCoord*3+j*3+1];
				}
			fout << endl;
			}
		fout << endl;

		fout << szString << "(z)\t";
		for(i = 0; i < kHeader.nNumberOfKey; i++)
			fout << i << "\t";
		fout << endl;
		for(j = 0; j < coordIKVHeader.nNumberOfCoord; j++)
			{
			fout << j;
			if(coordIKeyValue.bSelFlag[j*3+2] == true)
				{
				for(i = 0; i < kHeader.nNumberOfKey; i++)
					fout << "\t" << coordIKeyValue.dKeyValue[i*coordIKVHeader.nNumberOfCoord*3+j*3+2];
				}
			else
				{
				i = 0;
				fout << "\t" << coordIKeyValue.dKeyValue[i*coordIKVHeader.nNumberOfCoord*3+j*3+2];
				}
			fout << endl;
			}
		fout.close();
		}

	int getNbBits(int k)
		{
		int j=1;
		while (k >= (1<<j) ) 
			j++;
		
		return j;
		}
	};

#endif // _COMPRESSED_COORDINATE_INTERPOLATOR_H_
