// KeyHeader.h: key header decoder
// Author: Shinjun Lee, Samsung AIT
// 4th revision: 2002-04-18

#ifndef _KEYHEADER_H_
#define _KEYHEADER_H_

#include <limits.h>
#include <stdio.h>

#include "LinearKey.h"
#include "KeyMinMax.h"
#include "functionsforEncoding.h"


#if !defined AACTotal_ENC
#define AACTotal_ENC() (codec._ostrm->Outbuf.nbytes) * 8 + (codec._ostrm->Outbuf.nbits)
#endif

class KeyHeader
	{
  public:
	// variables: bitstream data of key
	int				nKeyQBit;								// quantization bit for key
	int				nNumberOfKey;							// the number of keys
	int				nKeyDigit;								// key digit

	bool			bIsLinearKeySubRegion;					// flag of linear key decoding
	LinearKey		lKey;									// class of linear key decoder

	bool			bRangeFlag;								// flag of key range
	KeyMinMax		keyMinMax;								// class of key min and key max decoder

	int				nBitSize;								// bit size
	int				nKDPCMOrder;							// order of key DPCM
	int*			nQIntraKeySign;							// sign array for intra keys
	int*			nQIntraKey;								// intra key array

	bool			bShiftFlag;								// flag of shift
	int				nKeyShiftSign;							// sign of key shift
	int				nKeyShift;								// the value of key shift

	int				nDNDOrder;								// order of DND

	int*			nKeyMaxSign;							// sign array for intra DND values
	int*			nKeyMax;								// intra DND values

	int				nKeyCodingBit;							// coding bit size for key
	bool			bKeyInvertDownFlag;						// flag of key invert down
	int				nKeyInvertDown;							// the value of key invert down

	// variables: nonparsable data for key decoding
	bool			bSignedAACFlag;							// flag of signed aac decoding

	double			fKeyMin;								// key min
	double			fKeyMax;								// key max


	// for encoding
	bool			bKeyInvertUpFlag;						// flag of key invert up
	int				nKeyInvertUp;							// the value of key invert up

  public:
	KeyHeader(void)
		{
		nKeyQBit = 0;
		nNumberOfKey = 0;
		nKeyDigit = 0;

		bIsLinearKeySubRegion = false;

		bRangeFlag = false;

		nBitSize = 0;
		nKDPCMOrder = 0;
		nQIntraKeySign = NULL;
		nQIntraKey = NULL;

		bShiftFlag = false;
		nKeyShiftSign = 0;
		nKeyShift = 0;

		nDNDOrder = 0;

		nKeyMaxSign = NULL;
		nKeyMax = NULL;

		nKeyCodingBit = 0;
		bKeyInvertDownFlag = false;
		nKeyInvertDown = 0;

		bSignedAACFlag = false;
		}

	~KeyHeader()
		{
		delete[] nQIntraKeySign;
		delete[] nQIntraKey;

		delete[] nKeyMaxSign;
		delete[] nKeyMax;
		}

	void readBitstream(CAC_QF& codec)
		{
		int			i = 0;									// index

		// get: nKeyQBit
		nKeyQBit = codec._istrm->getbits(5);
		
		// get: the number of keys
		int	nNumKeyCodingBit = codec._istrm->getbits(5);
		nNumberOfKey = codec._istrm->getbits(nNumKeyCodingBit);

		// get: key digit
		nKeyDigit = codec._istrm->getbits(4);

		// get: linear key
		if(codec._istrm->getbits(1) == 1)
			{
			bIsLinearKeySubRegion = true;
			lKey.readBitstream(codec, nKeyDigit);
			}
		else
			bIsLinearKeySubRegion = false;

		// get: fKeyMin, fKeyMax
		if(codec._istrm->getbits(1) == 1)
			{
			bRangeFlag = true;
			keyMinMax.readBitstream(codec, nKeyDigit);
			}
		else
			bRangeFlag = false;

		// get: nBitSize
		nBitSize = codec._istrm->getbits(5);

		// get: nKDPCMOrder
		nKDPCMOrder = codec._istrm->getbits(2);
		nQIntraKeySign = new int[nKDPCMOrder+1];
		
		////
		nQIntraKey = new int[nKDPCMOrder+1];

		// get: intra key data
		for(i = 0; i < nKDPCMOrder+1; i++)
			{
			nQIntraKeySign[i] = codec._istrm->getbits(1);
			if(i == 0 && nQIntraKeySign[i] == 1)
				continue;
			nQIntraKey[i] = codec._istrm->getbits(nBitSize);
			}

		// get: shift key data
		// Jameskim 2002427
		bShiftFlag = codec._istrm->getbits(1) == 1 ? true : false;
		if(bShiftFlag == true)
			{
			nKeyShiftSign = codec._istrm->getbits(1);
			nKeyShift = codec._istrm->getbits(nBitSize);
			}

		// get: nDNDOrder
		nDNDOrder = codec._istrm->getbits(3);
		if(nDNDOrder == 7)
			{
			int bNoDND = codec._istrm->getbits(1);
			if(bNoDND == 1)
				nDNDOrder = -1;
			}

		// get: nKeyMax
		int nMaxQBit = nBitSize;
		if(nDNDOrder > 0)
			{
			nKeyMaxSign = new int[nDNDOrder];
			nKeyMax = new int[nDNDOrder];
			}
		for(i = 0; i < nDNDOrder; i++)
			{
			nKeyMaxSign[i] = codec._istrm->getbits(1);
			nKeyMax[i] = codec._istrm->getbits(nMaxQBit);

			nMaxQBit = (int)(log10(abs(nKeyMax[i]))/log10(2))+1;
			if(nMaxQBit+1 < nBitSize)
				nMaxQBit += 1;
			else
				nMaxQBit = nBitSize;
			}

		// get: nKeyCodingBit
		int nKeyCodingBitQBit = (int)(log10(abs(nKeyQBit))/log10(2))+1;
		nKeyCodingBit = codec._istrm->getbits(nKeyCodingBitQBit); // 2001-11-20

		// get: nKeyInvertDown
		if(nDNDOrder != -1 && nDNDOrder != 0)
			{
			if(codec._istrm->getbits(1) == 1)
				{
				bKeyInvertDownFlag = true;
				nKeyInvertDown = codec._istrm->getbits(nKeyCodingBit);
				}
			else
				bKeyInvertDownFlag = false;
			}
		}

	void decodingProcess(void)
		{
		int			i = 0;									// index

		// get: linear key
		if(bIsLinearKeySubRegion == true)
			lKey.decodingProcess();

		// get: fKeyMin, fKeyMax
		if(bRangeFlag == true)
			keyMinMax.decodingProcess();
		else
			{
			fKeyMin = 0.0;
			fKeyMax = 1.0;
			}

		// get: intra key data
		for(i = 0; i < nKDPCMOrder+1; i++)
			{
			if(i == 0 && nQIntraKeySign[i] == 1)
				{
				nQIntraKey[i] = 0;
				continue;
				}

			if(nQIntraKeySign[i] == 1)
				nQIntraKey[i] *= -1;
			}

		// get: shift key data
		if(bShiftFlag == 1)
			{
			if(nKeyShiftSign == 1)
				nKeyShift *= -1;
			}

		// get: nKeyMax
		for(i = 0; i < nDNDOrder; i++)
			{
			if(nKeyMaxSign[i] == 1)
				nKeyMax[i] *= -1;
			}

		// get: nKeyShiftDown
		if(nDNDOrder != -1 && nDNDOrder != 0)
			{
			if(bKeyInvertDownFlag == true)
				{
				bSignedAACFlag = false;
				}
			else
				{
				nKeyInvertDown = -1;
				bSignedAACFlag = true;
				}
			}
		else
			{
			nKeyInvertDown = -1;
			bSignedAACFlag = true;
			}
		}


	// for encoding
	//void encodingProcess(bool bPreserveKey, int nKQBit, CSelectedKeyFrame selectedKeyFrame)
	void encodingProcess(int nKQBit, JWArray<double>& key)
	{
		int maxDigit = -INT_MAX;

		// kHeader.nNumberOfKey
		//nNumberOfKey = selectedKeyFrame.GetNumberOfOriginalKeys();
		nNumberOfKey = key.GetSize();

		for(int h = 0;h < nNumberOfKey;h++)
		{
			int nDigit =  GetDigit(key[h]);
			if(maxDigit < nDigit)
				maxDigit = nDigit;
		}
		
		// kHeader.nKeyDigit
		nKeyDigit = maxDigit;
		
		// kHeader.nKeyQBit
		nKeyQBit = nKQBit;
	}

	void makeBitstream(CAC_QF& codec, int& nTotalBitsOfKey)
		{
		int			i = 0;									// index

		ofstream fout("fout-encoding.txt");
		fout << "[Start: KeyHeader]" << endl;

		// put: nKeyQBit
		codec._ostrm->putbits(nKeyQBit, 5);
		fout << "nQStep_Key: " << nKeyQBit << endl;
		nTotalBitsOfKey += 5;

		// put: the number of keys
		int	nNumKeyCodingBit = getNbBits(nNumberOfKey);
		codec._ostrm->putbits(nNumKeyCodingBit, 5);
		fout << "nQStep_Num: " << nNumKeyCodingBit << endl;
		nTotalBitsOfKey += 5;
		codec._ostrm->putbits(nNumberOfKey, nNumKeyCodingBit);
		fout << "nNumberOfKey: " << nNumberOfKey << endl;
		nTotalBitsOfKey += nNumKeyCodingBit;

		// put: key digit
		codec._ostrm->putbits(nKeyDigit, 4);
		fout << "nKey_Digit: " << nKeyDigit << endl;
		nTotalBitsOfKey += 4;


		/////////////////////////////////////////////////////////////////
		/////////////////////////////////////////////////////////////////
		// 2002-04-17, Shinjun
		// put: linear key
		if(bIsLinearKeySubRegion == true)
			{
			// subset Ѵ.
			int	nIsColinear = 1;
			codec._ostrm->putbits(nIsColinear, 1);
			nTotalBitsOfKey += 1;

			lKey.makeBitstream(codec, nKeyDigit, nTotalBitsOfKey);
			}
		else
			{
			// subset  ʴ´.
			int nIsColinear = 0;
			codec._ostrm->putbits(nIsColinear, 1);
			nTotalBitsOfKey += 1;
			}

		// Jameskim - TEST
		int bit = AACTotal_ENC();


		// put: fKeyMin, fKeyMax
		if(fKeyMin < 0 || fKeyMax > 1)
			{
			codec._ostrm->putbits(1, 1);
			fout << "bRangeFlag: " << 1 << endl;
			nTotalBitsOfKey += 1;

			keyMinMax.makeBitstream(codec, nKeyDigit, nTotalBitsOfKey);
			}
		else
			{
			codec._ostrm->putbits(0, 1);
			fout << "bRangeFlag: " << 0 << endl;
			nTotalBitsOfKey += 1;
			}

		// put: m_nQStep_key_shift
		codec._ostrm->putbits(nBitSize, 5);
		fout << "nQStep_Shift: " << nBitSize << endl;
		nTotalBitsOfKey += 5;

		// 2002-04-18-02, Shinjun
		// put: nKDPCMOrder
		if(nKDPCMOrder == 0)
			{
			nQIntraKey = new int[1];
			nQIntraKey[0] = 0;
			nKDPCMOrder++;
			}
		codec._ostrm->putbits(nKDPCMOrder-1, 2);
		fout << "nOrder_DPCM: " << nKDPCMOrder << endl;
		nTotalBitsOfKey += 2;

		// put: intra key data
		for(i = 0; i < nKDPCMOrder; i++)
			{
			if(i == 0 && nQIntraKey[i] == 0)
				{
				codec._ostrm->putbits(1, 1);
				fout << "nSign_nQKey: " << 1 << endl;
				nTotalBitsOfKey += 1;
				continue;
				}
			if(nQIntraKey[i] >= 0)
				{
				codec._ostrm->putbits(0, 1);
				fout << "nSign_nQKey: " << 0 << endl;
				}
			else
				{
				codec._ostrm->putbits(1, 1);
				fout << "nSign_nQKey: " << 1 << endl;
				}
			nTotalBitsOfKey += 1;
			codec._ostrm->putbits(abs(nQIntraKey[i]), nBitSize);
			fout << "a_nQIntraKey: " << abs(nQIntraKey[i]) << endl;
			nTotalBitsOfKey += nBitSize;
			}

		// put: shift key data
		if(nKeyShift != 0)
			{
			codec._ostrm->putbits(1, 1);
			fout << "nShift: " << 1 << endl;
			nTotalBitsOfKey += 1;
			if(nKeyShift >= 0)
				{
				codec._ostrm->putbits(0, 1);
				fout << "nSign_Shift: " << 0 << endl;
				}
			else
				{
				codec._ostrm->putbits(1, 1);
				fout << "nSign_Shift: " << 1 << endl;
				}
			nTotalBitsOfKey += 1;
			codec._ostrm->putbits(abs(nKeyShift), nBitSize);
			fout << "nKey_Shift: " << abs(nKeyShift) << endl;
			nTotalBitsOfKey += nBitSize;
			}
		else
			{
			codec._ostrm->putbits(0, 1);
			fout << "nShift: " << 0 << endl;
			nTotalBitsOfKey += 1;
			}

		// put: nDNDOrder
		if(nDNDOrder != -1)
			{
			codec._ostrm->putbits(nDNDOrder, 3);
			fout << "unOrder_DND: " << nDNDOrder << endl;
			nTotalBitsOfKey += 3;
			if(nDNDOrder == 7)			// if 7, sign = 0(positive)
				{
				codec._ostrm->putbits(0, 1);
				fout << "nSign_Order: " << 0 << endl;
				nTotalBitsOfKey += 1;
				}
			}
		else
			{
			codec._ostrm->putbits(7, 3);	// if -1, put 7(binary: 111)
			fout << "unOrder_DND: " << 7 << endl;
			nTotalBitsOfKey += 3;
			codec._ostrm->putbits(1, 1);	// sign = 1(negative)
			fout << "nSign_Order: " << 1 << endl;
			nTotalBitsOfKey += 1;
			}

		// put: nKeyMax
		int nMaxQBit = nBitSize;
		for(i = 0; i < nDNDOrder; i++)
			{
			if(nKeyMax[i] >= 0)
				{
				codec._ostrm->putbits(0, 1);
				fout << "nSign_Max: " << 0 << endl;
				}
			else
				{
				codec._ostrm->putbits(1, 1);
				fout << "nSign_Max: " << 1 << endl;
				}
			nTotalBitsOfKey += 1;
			codec._ostrm->putbits(abs(nKeyMax[i]), nMaxQBit);
			fout << "a_nKey_Max: " << abs(nKeyMax[i]) << endl;
			nTotalBitsOfKey += nMaxQBit;

			nMaxQBit = getNbBits(abs(nKeyMax[i]));
			if(nMaxQBit+1 < nBitSize)
				nMaxQBit += 1;
			else
				nMaxQBit = nBitSize;
			}

		// put: nKeyCodingBit
		int nKeyCodingBitQBit = (int)(log10(abs(nKeyQBit))/log10(2))+1;
		codec._ostrm->putbits(nKeyCodingBit, nKeyCodingBitQBit);
		fout << "nQStep_Enc: " << nKeyCodingBit << endl;
		nTotalBitsOfKey += nKeyCodingBitQBit;

		// put: nKeyInvertUp
		if(nDNDOrder != -1 && nDNDOrder != 0)
			{
			if(nKeyInvertUp != -1)
				{
				codec._ostrm->putbits(1, 1);
				fout << "nShift_Up: " << 1 << endl;
				nTotalBitsOfKey += 1;

				codec._ostrm->putbits(nKeyInvertUp, nKeyCodingBit);
				fout << "nKey_Shift_Up: " << nKeyInvertUp << endl;
				nTotalBitsOfKey += nKeyCodingBit;
				bSignedAACFlag = false;
				}
			else
				{
				codec._ostrm->putbits(0, 1);
				fout << "nShift_Up: " << 0 << endl;
				nTotalBitsOfKey += 1;
				bSignedAACFlag = true;
				}
			}
		else
			bSignedAACFlag = false;

		fout << "[End: KeyHeader]" << endl;
		fout << endl;
		fout.close();

		// Jameskim - TEST
		bit = AACTotal_ENC() -bit;
		//printf("KEY CODER HEADER BIT %d\n", bit);
		bit = AACTotal_ENC();
		}
	};

#endif // _KEYHEADER_H_