/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2001
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Decoder
FILE:       MshGrd.cpp
AUTHOR:     Ioan Alexandru Salomie (VUB)
OVERVIEW:   Implementation of MGDecoder - an object that servers as
a template for devlopment of decoder plug-ins.
-------------------------------------------------------------------------*/
#include "StdAfx.h"
//#include "Defs.h"
#include "MeshGridDecoder.h"
//#include "Nodes.h"
#include "ODP.h"
//#include "Executiv.h"
#include "MshGrd\DataDefs.h"
#include "MshGrd\MGMesh.h"
#include "MshGrd\MGGrid.h"
#include "../../OpenGL/AFX/AfxStreamProxy.h"

// export as VRML file
static void WriteIFSHeader(CStdioFile &file)
{
	file.WriteString("#VRML V2.0 utf8 (Converted to ASCII)\n");
	file.WriteString("Shape {\n");
	file.WriteString("  appearance Appearance {\n");
	file.WriteString("    material Material {  # blue\n");
	file.WriteString("      ambientIntensity  0.5\n");
	file.WriteString("      diffuseColor      0.3 0.4 1.0\n");
	file.WriteString("    }\n");
	file.WriteString("  }\n");
	file.WriteString("  geometry IndexedFaceSet {\n");
}

static void Export(int noPoints, int noTriangles, ZArray<Vec3f> &verts, CIntArray &tris)
{
	CString error;
	bool bSuccess = true;

	try
	{
		CFileException e;
		CStdioFile vrmlFile;

		if (!vrmlFile.Open("MG_Export_IFS.debug.vrml", CFile::modeCreate | CFile::modeWrite, &e))
		{
			bSuccess = false;
			GetErrorMessage(&e, error);
		}
		else
		{
			CString str;

			// write the MG header
			WriteIFSHeader(vrmlFile);

			// write the vertices' coordinates
			{
				vrmlFile.WriteString("\tcoord Coordinate {\n");
				vrmlFile.WriteString("\t  point [\n");

				const Vec3f *coords = verts.GetBuffer();
				for (int k = 0; k < noPoints; k++)
				{
					str.Format("\t\t%f %f %f,\n", coords[k].x, coords[k].y, coords[k].z);
					vrmlFile.WriteString(str);
				}

				// end of vertices
				vrmlFile.WriteString("\t  ]\n");
				vrmlFile.WriteString("\t}\n");
			}

			// write the triangles
			{
				vrmlFile.WriteString("\tcoordIndex [\n");

				int *index = tris.GetData();
				for (int k = 0; k < noTriangles; k++)
				{
					str.Format("\t\t%d %d %d %d,\n", index[0], index[1], index[2], index[3]);
					index += 4;
					vrmlFile.WriteString(str);
				}

				// end of triangles
				vrmlFile.WriteString("\t]\n");
			}

			// end of the description
			vrmlFile.WriteString("  }\n}");

			vrmlFile.Close();
		}
	}
	catch (CFileException *e)
	{
		bSuccess = false;
		GetErrorMessage(e, error);
	}

	if (!bSuccess)
	{
		TRACE1("Cannot export MG to IFS: %s\n", error);
	}
}

static bool WriteToStream(int noGridPoints, int noPoints, int noTriEntries,
						  ZArray<Vec3f> &gridPoints,  ZArray<Vec3f> &points, CIntArray &tris,
						  CFPoint3DTArray &normal, CFltArray &offsets, int *nVertices, int nLevels,
						  BYTE *pOutput, int &len)
{
	MGDataExchange dataExchange(pOutput, false, len);

	if (!dataExchange.StoreNoTags(6))
		return false;
	if (!dataExchange.StoreNext(MGDataExchange::TAG_GRID_COORD, 3*sizeof(float)*noGridPoints, (void *)gridPoints.GetBuffer()))
		return false;
	if (!dataExchange.StoreNext(MGDataExchange::TAG_VERTEX_COORD, 3*sizeof(float)*noPoints, (void *)points.GetBuffer()))
		return false;
	if (!dataExchange.StoreNext(MGDataExchange::TAG_TRIANGLES, sizeof(int)*noTriEntries, (void *)tris.GetData()))
		return false;
	if (!dataExchange.StoreNext(MGDataExchange::TAG_NORMALS, 3*sizeof(float)*noPoints, (void *)normal.GetData()))
		return false;
	if (!dataExchange.StoreNext(MGDataExchange::TAG_OFFSETS, sizeof(float)*noPoints, (void *)offsets.GetData()))
		return false;
	if (!dataExchange.StoreNext(MGDataExchange::TAG_NO_VERTICES, sizeof(int)*nLevels, (void *)nVertices))
		return false;
	len = dataExchange.Finish();

	return true;
}

/*-----------------------------------------------------------------
Constructor, destructor
-----------------------------------------------------------------*/
MGDecoder :: MGDecoder()
{
	m_pGrid = NULL;
	m_pMesh = NULL;
	m_pParser = NULL;
}

MGDecoder :: ~MGDecoder()
{
	DeleteMGObjects();
	delete m_pParser;  m_pParser = NULL;
}

/*-----------------------------------------------------------------
MeshGrid object
-----------------------------------------------------------------*/
void MGDecoder :: UnrefMGObjects (CMGMesh *&pMesh, CMGGrid *&pGrid)
{
	pMesh = m_pMesh;    m_pMesh = NULL;
	pGrid = m_pGrid;    m_pGrid = NULL;
}

// MeshGrid object create and destroy
bool MGDecoder::CreateMGObjects()
{
	try
	{
		ASSERT(m_pMesh == NULL && m_pGrid == NULL);

		m_pGrid = new CMGGrid;
		m_pMesh = new CMGMesh;
		m_pMesh->SetGrid(m_pGrid);
	}
	catch (CMemoryException *)
	{
		TRACE0("Failed to allocatate the MeshGrid objects\n");
		return false;
	}

	return true;
}

void MGDecoder::DeleteMGObjects()
{
	delete m_pGrid;   m_pGrid = NULL;
	delete m_pMesh;   m_pMesh = NULL;
}

/*-----------------------------------------------------------------
Calculate the output size in bytes
-----------------------------------------------------------------*/
int MGDecoder :: GetOptimalOutputSize ()
{
	// decoding buffer
	return 10000000;
}

/*-----------------------------------------------------------------
Parse the decoder specific info
-----------------------------------------------------------------*/

BOOL MGDecoder :: Setup (AFXExtDescriptor *pSpecificInfo)
{
	CString error;
	m_pParser = CMGStreamParser::InitialiseStreamParser((MeshGridDecoderSpecificInfo *)pSpecificInfo, error);
	if (m_pParser == NULL)
	{
		TRACE1("Decoding error: %s\n", error);
		return FALSE;
	}

	return TRUE;
}

/*-------------------------------------------------------------
Decode one unit.
-------------------------------------------------------------*/

BOOL MGDecoder :: Decode (LPBYTE pInput, int nInputLength, LPBYTE pOutput, int &rnOutputLength)
{
	if (rnOutputLength < nInputLength)
	{
		rnOutputLength = nInputLength;  return FALSE;
	}

	try
	{
		// decode the stream
		ZArray<Vec3f> gridPoints, points;
		if (!Decode(pInput, nInputLength, gridPoints, points)) return FALSE;

		int maxNrLevels = m_pMesh->GetMaxNrLevels();
		int level = maxNrLevels - 1;

		// update the list of offsets of the VRML node
		// such that they are in the same order as the allocated vertices
		int noVertices = m_pMesh->GetTotalNrVertices();
		CFltArray vertexOffset;
		vertexOffset.SetSize(noVertices);
		m_pMesh->GetVertexOffset(vertexOffset.GetData());

		// get the triangle list
		int noPolygons = m_pMesh->GetNrPolygons(level);
		CIntArray coordIndex;
		coordIndex.SetSize(noPolygons * 4);

		m_pMesh->GetTriangles(coordIndex.GetData(), level);

		// get the normal vectors
		CFPoint3DTArray normals;
		normals.SetSize(noVertices);
		m_pMesh->GetVertexNormals(normals.GetData(), level);

		// dump the mesh as an IndexedFaceSet
		Export(noVertices, noPolygons, points, coordIndex);

		// get the number of vertices for each level
		int *nVerts = new int[maxNrLevels];
		for (int k = 0; k < maxNrLevels; k++)
			nVerts[k] = m_pMesh->GetNrVertices(k);

		// write to the output buffer
		const CIPointUVW &noSlices = m_pMesh->GetNrSlices();
		int noGridPoints = noSlices.u * noSlices.v * noSlices.w;
		if (!WriteToStream(noGridPoints, noVertices, noPolygons * 4,
			gridPoints, points, coordIndex, normals, vertexOffset,
			nVerts, maxNrLevels, pOutput, rnOutputLength))
			throw 0;
		delete []nVerts;
	}
	catch (CMemoryException *)
	{
		TRACE0("Memory allocation error\n");
	}

	return TRUE;
}

/*-------------------------------------------------------------
Attach to input data.
-------------------------------------------------------------*/

bool MGDecoder :: AttachToData(CMemFile *pFile, LPBYTE pInput, int nInputLength, int &dataSize)
{
	// map the buffer to a memory file
	pFile->Attach(pInput, nInputLength);

	// specify the input stream to the parser
	dataSize = nInputLength;
	if (m_pParser == NULL)
	{
		// decode the coding info of the MeshGrid stream
		CString error;    int headerSize;
		m_pParser = CMGStreamParser::ParseMeshGridHeader(pFile, error, headerSize);
		if (m_pParser == NULL)
		{
			TRACE1("Error %s\n", error);  return false;
		}
		dataSize -= headerSize;
	}
	else
	{
		// coding info of the MeshGrid stream has been retreived in Setup(...)
		m_pParser->AttachToStream(pFile, CBitStream::OP_READ);
	}

	return true;
}

/*-------------------------------------------------------------
Decode the buffer.
-------------------------------------------------------------*/
bool MGDecoder :: Decode (LPBYTE pInput, int nInputLength, ZArray<Vec3f> &gridPoints, ZArray<Vec3f> &points)
{
	initMesh = false;

	int dataSize;
	CMemFile memFile;

	if (!CreateMGObjects() || !AttachToData(&memFile, pInput, nInputLength, dataSize))
		return false;

	bool res = true;
	CString error;
	try
	{
		// store the number of levels
		// const CIPointUVW & nLevels = m_pParser->GetNrLevelsMesh();
		const CIPointUVW & nSlices = m_pParser->GetMaxNrSlices();

		// allocate the grid and vertices buffer large enough
		int nr_elem = nSlices.u * nSlices.v * nSlices.w;
		gridPoints.Expand(nr_elem);
		points.Expand(nr_elem);

		if (!m_pGrid->Init(gridPoints.GetBuffer(), false, m_pParser, error) ||
			!m_pMesh->Init(m_pGrid, m_pParser, points.GetBuffer(), error))
		{
			//		  pExecutive->Error(DECODER_ERROR, error);
			throw 0;
		}

		bool dump = false;
		if (!m_pMesh->LoadCompressed(error, dataSize, dump))
		{
			//		  pExecutive->Error(DECODER_ERROR, error);
			throw 0;
		}

		// the vertices buffer has been allocated too large, shrink it
		int noVertices = m_pMesh->GetTotalNrVertices();
		points.SetSize(noVertices);
	}
	catch (int)
	{
		res = false;
		TRACE1("Decoding error: %s\n", error);
	}

	initMesh = res;
	return res;
}

/*-------------------------------------------------------------
Decode the node fields.
-------------------------------------------------------------*/
bool MGDecoder :: Decode (ZArray<Vec3f> &gridPoints, ZArray<Vec3f> &points,
						  ZArray<int> &vertexLink, ZArray<float> &vertexOffset, 
						  const CIPointUVW &nrSlices, const CIPointUVW &nrLevels)
{
	initMesh = false;

	if (!CreateMGObjects())
		return FALSE;

	// retreive information about the model
	bool hasConnectivityInfo = false;
	bool hasRefineInfo = false;
	bool hasRepositionInfo = false;
	m_pMesh->RecoverEncodingParameters(vertexLink.GetBuffer(), vertexOffset.GetBuffer(),
									   vertexLink.GetLength(), vertexOffset.GetLength(),
									   hasConnectivityInfo, hasRefineInfo, hasRepositionInfo);

	// check the grid coordinates
	bool defaultInit;
	if (!CheckGridCoord(gridPoints, nrSlices, m_pMesh->GetCyclicMode(), defaultInit))
		return false;

	// prepare the vertices buffer large enough
	points.Expand(nrSlices.u * nrSlices.v * nrSlices.w);

	// initialise the grid and the mesh
	CString error;
	if (!m_pGrid->Init(gridPoints.GetBuffer(), defaultInit, nrSlices, nrLevels, error) ||
		!m_pMesh->Init(m_pGrid, points.GetBuffer(), error))
	{
		TRACE1("Error %s\n", error);
		return false;
	}

	// load the mesh
	bool dump = false;
	if (!m_pMesh->Load(hasConnectivityInfo, hasRefineInfo, hasRepositionInfo, dump, error))
	{
		TRACE1("Error %s\n", error);
		return false;
	}

	// the vertices buffer has been allocated too large, shrink it
	int noVertices = m_pMesh->GetTotalNrVertices();
	points.SetSize(noVertices);

	return true;
}

/*-------------------------------------------------------------
Decode the node fields.
-------------------------------------------------------------*/
bool MGDecoder::CheckGridCoord(ZArray<Vec3f> &gridPoints, const CIPointUVW &nrSlices, int cyclicMode, bool &defaultInit)
{
	int nrPts = gridPoints.GetLength();
	int totalPoints = nrSlices.u * nrSlices.v * nrSlices.w;

	// which grid corners are stored in the stream
	int uMax = 2, vMax = 2, wMax = 2;

	// check for single layered
	if (nrSlices.w == 1) wMax = 1;  // only layer 0

	// check for cyclic
	if (cyclicMode & CYCLIC_BITVAL_U)	uMax = 1;					// only one corner in the "U" direction
	if (cyclicMode == CYCLIC_UV)		vMax = 1;					// one corner in the "U" and "V" directions

	int nrCorners = uMax * vMax * wMax;

	defaultInit = false;
	if (nrPts < nrCorners)
	{
		TRACE0("The grid description should contain at least the eight corners");
		return false;
	}
	else if (nrPts > nrCorners && nrPts < totalPoints)
	{
		TRACE0("Incomplete grid description");
		return false;
	}
	else if (nrPts < totalPoints)
	{
		ASSERT(nrPts == nrCorners);

		if (nrPts < 8)
			FillInCorners(gridPoints, uMax, vMax, wMax);

		// only the corner description: resize the buffer
		defaultInit = true;
		gridPoints.Expand(totalPoints);
	}

	return true;
}

void MGDecoder::FillInCorners(ZArray<Vec3f> &gridPoints, int uMax, int vMax, int wMax)
{
	// initialise the remaining grid corners
	if (uMax == 1)
	{
		// copy point 0 to point 1 in both lines from the two layers
		for (int k = 0; k < 8; k += 2)
		{
			gridPoints[k + 1] = gridPoints[k + 0];
		}
	}

	if (vMax == 1)
	{
		// copy line 0 to line 1 in both layers
		for (int k = 0; k < 8; k += 4)
		{
			gridPoints[k + 2] = gridPoints[k + 0];
			gridPoints[k + 3] = gridPoints[k + 1];
		}
	}

	if (wMax == 1)
	{
		// copy layer 0 to layer 1 such that both are similar
		for (int k = 0; k < 8; k += 8)
		{
			gridPoints[k + 4] = gridPoints[k + 0];
			gridPoints[k + 5] = gridPoints[k + 1];
			gridPoints[k + 6] = gridPoints[k + 2];
			gridPoints[k + 7] = gridPoints[k + 3];
		}
	}
}