/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/
#ifndef _BASIC_TYPES
#define _BASIC_TYPES

#include "AfxTempl.h"
#include "Queue.h"
#include "List.h"
#include "../../../core/ztypes.h"

template<class DATA_TYPE>
bool FloatEq(const DATA_TYPE &val1, const DATA_TYPE &val2)
{
	return (((val1 - FLT_EPSILON) < val2) &&
					((val1 + FLT_EPSILON) > val2));
}

template<class DATA_TYPE>
bool DoubleEq(const DATA_TYPE &val1, const DATA_TYPE &val2)
{
	return (((val1 - DBL_EPSILON) < val2) &&
					((val1 + DBL_EPSILON) > val2));
}

template<class DATA_TYPE>
void IndexBuffer(DATA_TYPE *** &index, DATA_TYPE * buffer, int x_sz, int y_sz, int z_sz)
{
  index = (DATA_TYPE ***)new char[z_sz * sizeof(DATA_TYPE **) + z_sz * y_sz * sizeof(DATA_TYPE *)];

	DATA_TYPE **q = (DATA_TYPE **) (index + z_sz);
	DATA_TYPE  *p = buffer;
	
	for (int i = 0; i < z_sz; i++)
	{
		index[i] = q;		q += y_sz;
		for (int j = 0; j < y_sz; j++)
		{
			index[i][j] = p;		p += x_sz;
		}
	}
}

template<class DATA_TYPE>
void AllocBuffer(DATA_TYPE *** &buffer, int x_sz, int y_sz, int z_sz)
{
	buffer = (DATA_TYPE ***) new char[z_sz * sizeof(DATA_TYPE **) +
																		z_sz * y_sz * sizeof(DATA_TYPE *) +
																		z_sz * y_sz * x_sz * sizeof(DATA_TYPE)];
	
	DATA_TYPE **q = (DATA_TYPE **) (buffer + z_sz);
	DATA_TYPE  *p = (DATA_TYPE *)  (q + z_sz * y_sz);
	
	for (int i = 0; i < z_sz; i++)
	{
		buffer[i] = q;		q += y_sz;
		for (int j = 0; j < y_sz; j++)
		{
			buffer[i][j] = p;		p += x_sz;
		}
	}

	memset (&buffer[0][0][0], 0, x_sz * y_sz * z_sz * sizeof(DATA_TYPE));
}

////////////////////////////////////////////////////////////////////////
//class Cube
////////////////////////////////////////////////////////////////////////

struct Cube
{
	int u, v, w;
	int width, height, depth;
	
	Cube() { u = v = w = 0; width = height = depth = 0; }
	
	Cube(int uu, int vv, int ww, int w, int h, int d)
	{
		Set(uu, vv, ww, w, h, d);
	}
	
	void Set(int uu,int vv,int ww,int wi,int hi,int de)
	{
		u = uu;	v = vv;	w = ww;
		width = wi;	height = hi;		depth = de;
	}

	void SetDefault()
	{
		u = v = w = 0;
		width = height = depth = 1;
	}

	bool IsPixel() const
	{
		return (width == 1 && height == 1 && depth == 1);
	}

	bool IsValid() const
	{
		return (width != 0 && height != 0 && depth != 0);
	}
};

struct CubeI : public Cube
{
  int index;

  CubeI() { index = 0; }
	
	CubeI(int uu, int vv, int ww, int w, int h, int d, int i)
	{
    index = i;
		Set(uu, vv, ww, w, h, d);
	}
};

////////////////////////////////////////////////////////////////////////
//class CPoint3D
////////////////////////////////////////////////////////////////////////

template <class DATA_TYPE>
class CPoint3D
{
public:
	CPoint3D() { x = y = z = (DATA_TYPE)0; };
	CPoint3D(DATA_TYPE val) { x = y = z = val; };
	CPoint3D(DATA_TYPE x_val, DATA_TYPE y_val, DATA_TYPE z_val)
	{
		x = x_val; y = y_val;	z = z_val;
	};

// Member variables
public:
  DATA_TYPE x, y, z;

// Operations
public:
	friend bool ErrorEQ(const CPoint3D<DATA_TYPE> &p1, const CPoint3D<DATA_TYPE> &p2)
	{
		return (FloatEq<DATA_TYPE>(p1.x, p2.x) &&
						FloatEq<DATA_TYPE>(p1.y, p2.y) &&
						FloatEq<DATA_TYPE>(p1.z, p2.z));
	};
	friend bool ErrorNE(const CPoint3D<DATA_TYPE> &p1, const CPoint3D<DATA_TYPE> &p2)
	{
		return (!FloatEq<DATA_TYPE>(p1.x, p2.x) ||
						!FloatEq<DATA_TYPE>(p1.y, p2.y) ||
						!FloatEq<DATA_TYPE>(p1.z, p2.z));
	};
	bool ErrorEQ(const CPoint3D<DATA_TYPE> &p) const
	{
		return (FloatEq<DATA_TYPE>(x, p.x) &&
						FloatEq<DATA_TYPE>(y, p.y) &&
						FloatEq<DATA_TYPE>(z, p.z));
	};
	bool ErrorNE(const CPoint3D<DATA_TYPE> &p) const
	{
		return (!FloatEq<DATA_TYPE>(x, p.x) ||
						!FloatEq<DATA_TYPE>(y, p.y) ||
						!FloatEq<DATA_TYPE>(z, p.z));
	};
	bool ErrorEQ(DATA_TYPE val) const
	{
		return (FloatEq<DATA_TYPE>(x, val) &&
						FloatEq<DATA_TYPE>(y, val) &&
						FloatEq<DATA_TYPE>(z, val));
	};
	bool ErrorNE(DATA_TYPE val) const
	{
		return (!FloatEq<DATA_TYPE>(x, val) ||
						!FloatEq<DATA_TYPE>(y, val) ||
						!FloatEq<DATA_TYPE>(z, val));
	};
	bool operator==(const CPoint3D<DATA_TYPE> &p) const
	{
		return (x == p.x && y == p.y && z == p.z);
	};
	bool operator==(const DATA_TYPE &p) const
	{
		return (x == p && y == p && z == p);
	};
	bool operator!=(const CPoint3D<DATA_TYPE> &p) const
	{
		return (x != p.x || y != p.y || z != p.z);
	};
	bool operator!=(const DATA_TYPE &p) const
	{
		return (x != p && y != p && z != p);
	};
	bool operator>(const DATA_TYPE &p) const
	{
		return (x > p && y > p && z > p);
	};
	bool operator>=(const DATA_TYPE &p) const
	{
		return (x >= p && y >= p && z >= p);
	};
	bool operator<(const DATA_TYPE &p) const
	{
		return (x < p && y < p && z < p);
	};
	bool operator<=(const DATA_TYPE &p) const
	{
		return (x <= p && y <= p && z <= p);
	};
	const CPoint3D<DATA_TYPE>& operator=(const CPoint3D<DATA_TYPE> &p)
	{
		x = p.x; y = p.y; z = p.z;

		return p;
	};
	const DATA_TYPE & operator=(const DATA_TYPE &val)
	{
		x = y = z = val;

		return val;
	};
	CPoint3D<DATA_TYPE> operator+(const CPoint3D<DATA_TYPE> &p) const
	{
		CPoint3D<DATA_TYPE> res;
		res.x = x + p.x;	res.y = y + p.y;	res.z = z + p.z;
		return res;
	};
	CPoint3D<DATA_TYPE> operator-(const CPoint3D<DATA_TYPE> &p) const
	{
		CPoint3D<DATA_TYPE> res;
		res.x = x - p.x;	res.y = y - p.y;	res.z = z - p.z;
		return res;
	};
	CPoint3D<DATA_TYPE> operator*(const CPoint3D<DATA_TYPE> &p) const
	{
		CPoint3D<DATA_TYPE> res;
		res.x = x * p.x;	res.y = y * p.y;	res.z = z * p.z;
		return res;
	};
	CPoint3D<DATA_TYPE> operator/(const CPoint3D<DATA_TYPE> &p) const
	{
		CPoint3D<DATA_TYPE> res;
		res.x = x / p.x;	res.y = y / p.y;	res.z = z / p.z;
		return res;
	};
	CPoint3D<DATA_TYPE> operator+(DATA_TYPE val) const
	{
		CPoint3D<DATA_TYPE> res;
		res.x = x + val;	res.y = y + val;	res.z = z + val;
		return res;
	};
	CPoint3D<DATA_TYPE> operator-(DATA_TYPE val) const
	{
		CPoint3D<DATA_TYPE> res;
		res.x = x - val;	res.y = y - val;	res.z = z - val;
		return res;
	};
	CPoint3D<DATA_TYPE> operator*(DATA_TYPE val) const
	{
		CPoint3D<DATA_TYPE> res;
		res.x = x * val;	res.y = y * val;	res.z = z * val;
		return res;
	};
	CPoint3D<DATA_TYPE> operator/(DATA_TYPE val) const
	{
		CPoint3D<DATA_TYPE> res;
		res.x = x / val;	res.y = y / val;	res.z = z / val;
		return res;
	};
	CPoint3D<DATA_TYPE> operator>>(DATA_TYPE val) const
	{
		CPoint3D<DATA_TYPE> res;
		res.x = x >> val;	res.y = y >> val;	res.z = z >> val;
		return res;
	};
	CPoint3D<DATA_TYPE> operator<<(DATA_TYPE val) const
	{
		CPoint3D<DATA_TYPE> res;
		res.x = x << val;	res.y = y << val;	res.z = z << val;
		return res;
	};
	const CPoint3D<DATA_TYPE> &operator+=(const CPoint3D<DATA_TYPE> &p)
	{
		x += p.x;	y += p.y;	z += p.z;
		return *this;
	};
	const CPoint3D<DATA_TYPE> &operator-=(const CPoint3D<DATA_TYPE> &p)
	{
		x -= p.x;	y -= p.y;	z -= p.z;
		return *this;
	};
	const CPoint3D<DATA_TYPE> &operator*=(const CPoint3D<DATA_TYPE> &p)
	{
		x *= p.x;	y *= p.y;	z *= p.z;
		return *this;
	};
	const CPoint3D<DATA_TYPE> &operator/=(const CPoint3D<DATA_TYPE> &p)
	{
		x /= p.x;	y /= p.y;	z /= p.z;
		return *this;
	};
	const CPoint3D<DATA_TYPE> &operator+=(DATA_TYPE val)
	{
		x += val;	y += val;	z += val;
		return *this;
	};
	const CPoint3D<DATA_TYPE> &operator-=(DATA_TYPE val)
	{
		x -= val;	y -= val;	z -= val;
		return *this;
	};
	const CPoint3D<DATA_TYPE> &operator*=(DATA_TYPE val)
	{
		x *= val;	y *= val;	z *= val;
		return *this;
	};
	const CPoint3D<DATA_TYPE> &operator/=(DATA_TYPE val)
	{
		x /= val;	y /= val;	z /= val;
		return *this;
	};
	const CPoint3D<DATA_TYPE> &operator>>=(DATA_TYPE val)
	{
		x >>= val;	y >>= val;	z >>= val;
		return *this;
	};
	const CPoint3D<DATA_TYPE> &operator<<=(DATA_TYPE val)
	{
		x <<= val;	y <<= val;	z <<= val;
		return *this;
	};

// Implementation
public:
	DATA_TYPE		GetX() {return x;};
	DATA_TYPE		GetY() {return y;};
	DATA_TYPE		GetZ() {return z;};

	DATA_TYPE		GetLength() const { return (DATA_TYPE) sqrt(x*x + y*y + z*z); };
	DATA_TYPE		SquareLength() const { return x*x + y*y + z*z; };
	double			ScalarProduct (const CPoint3D<DATA_TYPE> &p) const { return x*p.x + y*p.y + z*p.z; };
	double			Dist(const CPoint3D<DATA_TYPE> &p) const { return sqrt(SquareDist(p)); };
	DATA_TYPE	  SquareDist(const CPoint3D<DATA_TYPE> &p) const
	{
		register DATA_TYPE x_dist = p.x - x;
		register DATA_TYPE y_dist = p.y - y;
		register DATA_TYPE z_dist = p.z - z;

		return (x_dist * x_dist +
						y_dist * y_dist + 
						z_dist * z_dist);
	};
	DATA_TYPE	  SquareSum(const CPoint3D<DATA_TYPE> &p) const
	{
		register DATA_TYPE x_dist = p.x + x;
		register DATA_TYPE y_dist = p.y + y;
		register DATA_TYPE z_dist = p.z + z;

		return (x_dist * x_dist +
						y_dist * y_dist + 
						z_dist * z_dist);
	};
};

////////////////////////////////////////////////////////////////////////
//class CDataArray
////////////////////////////////////////////////////////////////////////

template<class TYPE, class ARG_TYPE>
class CDataArray : public CArray<TYPE, ARG_TYPE>
{
// construct, destruct
public:
	CDataArray() { };
	CDataArray(const CDataArray &array) { Copy(array); };
	const CDataArray &operator=(const CDataArray &array)
	{
		Copy(array);
		return *this;
	};
	virtual ~CDataArray() { };

// Member functions
public:
  TYPE  &AddEmpty()
	{
		ASSERT_VALID(this);

		SetSize(m_nSize+1, -1);
		return m_pData[m_nSize-1];
	};

  void ResetCounter()
	{
		// destroy the elements but keep the buffer
		if (m_pData != NULL)
			DestructElements<TYPE>(m_pData, m_nSize);

		m_nSize = 0;
	};

	void RemoveRange(int index, int nrElements)
	{
		ASSERT (index >= 0 && nrElements > 0);
		if (index < m_nSize)
		{
			int val = min (nrElements, m_nSize - index - 1);
			DestructElements<TYPE>(&m_pData[index], nrElements);

			// move data in buffer byte by byte
			int nrBytes = (m_nSize - index - val) * sizeof(TYPE);
			for (int k = 0; k < nrBytes; k++)
				*((char *)&m_pData[index + k]) = *((char *)&m_pData[index + val + k]);

			m_nSize -= val;
		}
	};

// Member variables
public:
};

////////////////////////////////////////////////////////////////////////
//class CDynaArray
////////////////////////////////////////////////////////////////////////
typedef double	DYNA_POSITION;

template<class TYPE, class ARG_TYPE>
class CDynaArray
{
// construct, destruct
public:
	struct DynaPos
	{
		int idx, level;
	};

  CDynaArray(int levelSize = 10000)
	{
		ASSERT(levelSize > 0);
		m_levelSize = levelSize;

		m_dataBuf = NULL;

		m_count = 0;

		// force an allocation for the first added element
		m_idx = m_levelSize - 1;
		m_level = -1;
	};

// Member functions
public:
  TYPE *GetHeadItem(DYNA_POSITION *pos)
	{
		if (m_count == 0)	return NULL;

		DynaPos *dynaPos = (DynaPos *) pos;
		ASSERT(AfxIsValidAddress(dynaPos, sizeof(DynaPos *)));

		dynaPos->idx = 0;	dynaPos->level = 0;

		return &m_dynaArray[0][0];
	};

  TYPE *GetTailItem(DYNA_POSITION *pos)
	{
		if (m_count == 0)	return NULL;

		DynaPos *dynaPos = (DynaPos *) pos;
		ASSERT(AfxIsValidAddress(dynaPos, sizeof(DynaPos *)));

		dynaPos->idx = m_idx;	dynaPos->level = m_level;

		ASSERT(m_dataBuf == &m_dynaArray[m_level][m_idx]);
		return m_dataBuf;
	};

  TYPE *GetNextItem(DYNA_POSITION *pos)
	{
		DynaPos *dynaPos = (DynaPos *) pos;
		ASSERT(AfxIsValidAddress(dynaPos, sizeof(DynaPos *)));

		if (dynaPos->level == m_level)
		{
			ASSERT(dynaPos->idx <= m_idx);
			if (++dynaPos->idx > m_idx)
				return NULL;
		}
		else if (++dynaPos->idx == m_levelSize)
		{
			if (++dynaPos->level > m_level)
				return NULL;

			dynaPos->idx = 0;
		}

		return &m_dynaArray[dynaPos->level][dynaPos->idx];
  };

  TYPE *GetPrevItem(DYNA_POSITION *pos)
	{
		DynaPos *dynaPos = (DynaPos *) pos;
		ASSERT(AfxIsValidAddress(dynaPos, sizeof(DynaPos *)));

		if (--dynaPos->idx == -1)
		{
			if (--dynaPos->level == -1)
				return NULL;

			dynaPos->idx = m_levelSize - 1;
		}

    return &m_dynaArray[dynaPos->level][dynaPos->idx];
	};

  TYPE &GetItem(int pos)
	{
		ASSERT(pos >= 0);

		int level = pos / m_levelSize;
		int idx = pos - level * m_levelSize;
		ASSERT(level <= m_level);
		ASSERT(level < m_level || idx <= m_idx);

		return m_dynaArray[level][idx];
	};

  TYPE *GetItemPtr(int pos)
	{
		ASSERT(pos >= 0);

		int level = pos / m_levelSize;
		int idx = pos - level * m_levelSize;
		ASSERT(level <= m_level);
		ASSERT(level < m_level || idx <= m_idx);

		return &m_dynaArray[level][idx];
	};

  int  GetLevelSize() const { return m_levelSize; };
  int  GetSize() const { return m_count; };
  int  GetNrLevels() const { return m_level; };
  int  GetIndex() const { return m_idx; };
  void FreeExtra() { m_dynaArray.FreeExtra(); };

  CDataArray<TYPE, ARG_TYPE> &operator[](int idx) { return m_dynaArray[idx]; };
  CDataArray<TYPE, ARG_TYPE> *GetDynaBuffer() { return m_dynaArray.GetData(); };

  void RemoveAll()
	{
		m_dynaArray.RemoveAll();
		m_count = 0;
		// force an allocation for the first added element
		m_idx = m_levelSize - 1;
		m_level = -1;
	};

  TYPE *Add(ARG_TYPE p)
	{
		// once a data buffer is full increment the level
		// and allocate a new buffer of level size
		if (++m_idx == m_levelSize)
		{
			m_idx = 0; ++m_level;

			// allocate at least m_level + 1
			m_dynaArray.SetSize(m_level + 1, 0);
			CDataArray<TYPE, ARG_TYPE> &array = m_dynaArray[m_level];
			array.SetSize(m_levelSize);

			// get the pointer to the new buffer - 1 because it will be incremented
			m_dataBuf = array.GetData() - 1;
		}

		*(++m_dataBuf) = p;
		++m_count;

		return m_dataBuf;
	};

  TYPE *AddEmptyItem()
	{
		// once a data buffer is full increment the level
		// and allocate a new buffer of level size
		if (++m_idx == m_levelSize)
		{
			m_idx = 0; ++m_level;

			// allocate at least m_level + 1
			m_dynaArray.SetSize(m_level + 1, 0);
			CDataArray<TYPE, ARG_TYPE> &array = m_dynaArray[m_level];
			array.SetSize(m_levelSize);

			// get the pointer to the new buffer - 1 because it will be incremented
			m_dataBuf = array.GetData() - 1;
		}

		++m_count;

		return ++m_dataBuf;
	};


// Member variables
protected:
  int m_idx;						// last stored cell index into the CDataArray
  int m_level;					// last allocated level of CDataArray
  int m_count;					// total nr of cells
  int m_levelSize;			// size of a CDataArray

  TYPE *m_dataBuf;      // pointer to the next free cell in the CDataArray buffer

  // the array is growing with m_levelSize elements at once
  // for each new entry into the m_dynaArray.
  CArray<CDataArray<TYPE, ARG_TYPE>, CDataArray<TYPE, ARG_TYPE>&> m_dynaArray;
};

// example of accessing the data
template<class TYPE, class ARG_TYPE>
inline void LoopData(CDynaArray<TYPE, ARG_TYPE> &dataArr)
{
  register int levelSz = dataArr.GetLevelSize();
  register int count = dataArr.GetNrItems();
  register int level = -1;
  register int pos = levelSz;

  TYPE *dataBuf = NULL;
  while (count--)
  {
    if (pos == levelSz)
    {
      pos = 0; ++level;
      CDataArray *buf = dataArr.GetDynaBuffer();
      dataBuf = buf[level].GetDataBuffer();
    }

		// do something with the dataBuf

    ++dataBuf; ++pos;
  }
}

class CIPointUVW
{
public:
  CIPointUVW() { u = v = w = 0; };
  CIPointUVW(int i, int j, int k) { u = i; v = j; w = k; };
  CIPointUVW(int val) { u = v = w = val; };
  int u, v, w;

	const int & operator=(const int &val)
	{
		u = v = w = val;

		return val;
	};
	bool operator==(const CIPointUVW &p) const
	{
		return (u == p.u && v == p.v && w == p.w);
	};
	bool operator==(const int &p) const
	{
		return (u == p && v == p && w == p);
	};
	bool operator!=(const CIPointUVW &p) const
	{
		return (u != p.u || v != p.v || w != p.w);
	};
	bool operator!=(const int &p) const
	{
		return (u != p && v != p && w != p);
	};
	bool operator>(const int &p) const
	{
		return (u > p && v > p && w > p);
	};
	bool operator>=(const int &p) const
	{
		return (u >= p && v >= p && w >= p);
	};
	bool operator<(const int &p) const
	{
		return (u < p && v < p && w < p);
	};
	bool operator<=(const int &p) const
	{
		return (u <= p && v <= p && w <= p);
	};
	CIPointUVW operator+(const CIPointUVW &p) const
	{
		CIPointUVW res;
		res.u = u + p.u;	res.v = v + p.v;	res.w = w + p.w;
		return res;
	};
	CIPointUVW operator-(const CIPointUVW &p) const
	{
		CIPointUVW res;
		res.u = u - p.u;	res.v = v - p.v;	res.w = w - p.w;
		return res;
	};
	CIPointUVW operator*(const CIPointUVW &p) const
	{
		CIPointUVW res;
		res.u = u * p.u;	res.v = v * p.v;	res.w = w * p.w;
		return res;
	};
	CIPointUVW operator/(const CIPointUVW &p) const
	{
		CIPointUVW res;
		res.u = u / p.u;	res.v = v / p.v;	res.w = w / p.w;
		return res;
	};
	CIPointUVW operator+(int val) const
	{
		CIPointUVW res;
		res.u = u + val;	res.v = v + val;	res.w = w + val;
		return res;
	};
	CIPointUVW operator-(int val) const
	{
		CIPointUVW res;
		res.u = u - val;	res.v = v - val;	res.w = w - val;
		return res;
	};
	CIPointUVW operator*(int val) const
	{
		CIPointUVW res;
		res.u = u * val;	res.v = v * val;	res.w = w * val;
		return res;
	};
	CIPointUVW operator/(int val) const
	{
		CIPointUVW res;
		res.u = u / val;	res.v = v / val;	res.w = w / val;
		return res;
	};
	CIPointUVW operator>>(int val) const
	{
		CIPointUVW res;
		res.u = u >> val;	res.v = v >> val;	res.w = w >> val;
		return res;
	};
	CIPointUVW operator<<(int val) const
	{
		CIPointUVW res;
		res.u = u << val;	res.v = v << val;	res.w = w << val;
		return res;
	};
	const CIPointUVW &operator+=(const CIPointUVW &p)
	{
		u += p.u;	v += p.v;	w += p.w;
		return *this;
	};
	const CIPointUVW &operator-=(const CIPointUVW &p)
	{
		u -= p.u;	v -= p.v;	w -= p.w;
		return *this;
	};
	const CIPointUVW &operator*=(const CIPointUVW &p)
	{
		u *= p.u;	v *= p.v;	w *= p.w;
		return *this;
	};
	const CIPointUVW &operator/=(const CIPointUVW &p)
	{
		u /= p.u;	v /= p.v;	w /= p.w;
		return *this;
	};
	const CIPointUVW &operator+=(int val)
	{
		u += val;	v += val;	w += val;
		return *this;
	};
	const CIPointUVW &operator-=(int val)
	{
		u -= val;	v -= val;	w -= val;
		return *this;
	};
	const CIPointUVW &operator*=(int val)
	{
		u *= val;	v *= val;	w *= val;
		return *this;
	};
	const CIPointUVW &operator/=(int val)
	{
		u /= val;	v /= val;	w /= val;
		return *this;
	};
	const CIPointUVW &operator>>=(int val)
	{
		u >>= val;	v >>= val;	w >>= val;
		return *this;
	};
	const CIPointUVW &operator<<=(int val)
	{
		u <<= val;	v <<= val;	w <<= val;
		return *this;
	};
};

/***************************************************************************************/

typedef CPoint3D<int>								CIPoint3DT;
//typedef CPoint3D<float>							CFPoint3DT;
typedef Vec3f         								CFPoint3DT;

typedef CDataArray<bool, bool>					    CBoolArray;
typedef CDataArray<int, int>						CIntArray;
typedef CDataArray<float, float>					CFltArray;
typedef CDataArray<CFPoint3DT, CFPoint3DT &>		CFPoint3DTArray;
typedef CDataArray<CubeI, const CubeI&>	            CCubeIArray;
typedef CDataArray<Cube, const Cube&>	            CCubeArray;
typedef	CList<Cube, const Cube&>	                CCubeList;
typedef CDataArray<CCubeArray, const CCubeArray &>	CTileArray;
typedef	cQueue<Cube, const Cube&>	                CCubeQueue;
typedef cList<double, const double>					CDoubleList;

/***************************************************************************************/

class CSurrBoxT
{
public:
	CSurrBoxT() { };

	// Member variables
public:
	CFPoint3DT  top;
	CFPoint3DT  bottom;

	// Implementation
public:
	float GetXDim() const { return top.x - bottom.x; };
	float GetYDim() const { return top.y - bottom.y; };
	float GetZDim() const { return top.z - bottom.z; };
	void  Init(const CIPointUVW &dim)
	{
		bottom.x = bottom.y = bottom.z = 0;
		top.x = (float)dim.u;	top.y = (float)dim.v;	top.z = (float)dim.w;	
	};

	bool operator!=(const CSurrBoxT &box) const;
	bool operator==(const CSurrBoxT &box) const;

	float GetWidth() const { return top.x - bottom.x; };
	float GetHeight() const { return top.y - bottom.y; };
	float GetDepth() const { return top.z - bottom.z; };
};

inline bool CSurrBoxT::operator !=(const CSurrBoxT &box) const
{
	return (top != box.top || bottom != box.bottom);
}

inline bool CSurrBoxT::operator ==(const CSurrBoxT &box) const
{
	return (top == box.top && bottom == box.bottom);
}

/***************************************************************************************/

#endif