/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// BitStream.cpp: implementation of the CBitStream class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "BitStream.h"
#include "MGErrMsg.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CBitStream::CBitStream()
{
	m_operation = OP_NA;
	m_bitCount = 0;
	m_stream = NULL;
	m_limit = -1;

	m_bit = 0;
	m_buffer = 0;
}

CBitStream::CBitStream(BYTE* fileBuffer, UINT fileSize, BITSTREAM_OPERATIONS op, UINT lim)
{
	ASSERT(op != OP_NA);
	m_operation = op;
	m_bitCount = 0;
	m_stream = new CMemFile(fileBuffer, fileSize);
	m_limit = lim;
	m_bAllocInternally = true;

	if (m_operation == OP_READ)
		m_bit = 0;
	else
		m_bit = 8;
	m_buffer = 0;
}

CBitStream::CBitStream(CFile* stream, BITSTREAM_OPERATIONS op, UINT lim)
{
	ASSERT(op != OP_NA);
	m_operation = op;
	m_bitCount = 0;
	m_stream = stream;
	m_limit = lim;
	m_bAllocInternally = false;

	if (m_operation == OP_READ)
		m_bit = 0;
	else
		m_bit = 8;
	m_buffer = 0;
}

CBitStream::~CBitStream()
{
	if (m_operation != OP_NA)
		Flush();

	if (m_bAllocInternally)
	{
		m_stream->m_hFile = NULL;
		delete m_stream;
	}
}

void CBitStream::Init(CFile* stream, BITSTREAM_OPERATIONS op, UINT lim)
{
	ASSERT(m_stream == NULL);
	m_stream = stream;
	m_limit = lim;

	SetOperationMode(op);
}

void CBitStream::SetOperationMode(BITSTREAM_OPERATIONS op)
{
	m_operation = op;
	m_bitCount = 0;

	if (m_operation == OP_READ)
		m_bit = 0;
	else
		m_bit = 8;
	m_buffer = 0;
}

void CBitStream::Flush()
{
	// warning: Flush will alter the bitcount
	// if you would like to know the exact bitcount
	// you should call GetBitCount() before Flush()
	if (m_operation == OP_WRITE)
	{
		if (m_bit < 8)
		{
			m_bitCount += m_bit;
			try
			{
				m_stream->Write(&m_buffer, sizeof(m_buffer));
			}
			catch (CFileException *e)
			{
				throw CMGDataException(e);
			}
			m_buffer = 0;		m_bit = 8;
		}
		m_stream->Flush();
	}
	else
	{
		ASSERT(m_operation == OP_READ);
		m_bitCount += m_bit;
		m_buffer = 0;		m_bit = 0;
		m_limit = -1;
	}
}

void CBitStream::Reset()
{
	Flush();
	m_bitCount = 0;
}

void CBitStream::Close()
{
	Flush();
	m_operation = OP_NA;
	m_bitCount = 0;
	m_stream = NULL;
	m_limit = -1;
	m_bit = 0;
	m_buffer = 0;
}

int CBitStream::GetFilePosition() const
{
	return m_stream->GetPosition();
}

void CBitStream::SetFilePosition(int pos)
{
	m_stream->Seek(pos, CFile::begin);
}

int CBitStream::ReadBytes(BYTE *buffer, int count)
{
	return m_stream->Read(buffer, count);
}

BYTE CBitStream::GetRemainingValue() const
{
	if (m_operation == OP_WRITE)
	{
		BYTE mask = 0xff - (1 << m_bit) - 1;
		return m_buffer & mask;
	}
	else
	{
		ASSERT(m_operation == OP_READ);
		BYTE mask = (1 << m_bit) - 1;
		return m_buffer & mask;
	}
}

BYTE CBitStream::ReadBit()
{
	if (LimitReached())
		throw EEC_ReadingLimitReached();

	++m_bitCount;

	if (m_bit == 0)
	{
		try
		{
			UINT nLen = m_stream->Read(&m_buffer, sizeof(m_buffer));
			ASSERT(nLen == sizeof(m_buffer));
		}
		catch (CFileException *e)
		{
			throw CMGDataException(e);
		}
		m_bit = 8;
	}
	return (m_buffer & (1 << --m_bit));
}

void CBitStream::WriteBit(BYTE bit)
{
	if (LimitReached())
	{
		Flush();
		throw EEC_WritingLimitReached();
	}

	m_buffer |= (bit << --m_bit);

	if (m_bit == 0)
	{
		try
		{
			m_stream->Write(&m_buffer, sizeof(m_buffer));
		}
		catch (CFileException *e)
		{
			throw CMGDataException(e);
		}
		m_buffer = 0;		m_bit = 8;
	}

	m_bitCount++;
}

void CBitStream::ReadBits(BYTE *res, BYTE count)
{
	ASSERT(count <= 8 && count > 0);

	*res = 0;
	for (count--; count<0xff; count--)
	{
		if (ReadBit())
			*res |= (1 << count);
	}
}

void CBitStream::WriteBits(BYTE value, BYTE count)
{
	ASSERT(count <= 8 && count > 0);

	BYTE mask = 1 << (count - 1);
	BYTE shift = count - 1;

	for (BYTE bit = 0; bit < count; bit++)
	{
		WriteBit((value & mask) >> shift);
		mask >>= 1;		shift--;
	}
}

void CBitStream::ReadBits(WORD *res, BYTE count)
{
	ASSERT(count <= 16 && count > 0);

	*res = 0;
	for (count--; count<0xff; count--)
	{
		if (ReadBit())
			*res |= (1 << count);
	}
}

void CBitStream::WriteBits(WORD value, BYTE count)
{
	ASSERT(count <= 16 && count > 0);

	WORD mask = 1 << (count - 1);
	WORD shift = count - 1;

	for (BYTE bit = 0; bit < count; bit++)
	{
		WriteBit((value & mask) >> shift);
		mask >>= 1;		shift--;
	}
}

void CBitStream::ReadBits(DWORD *res, BYTE count)
{
	ASSERT(count <= 32 && count > 0);

	*res = 0;
	for (count--; count<0xff; count--)
	{
		if (ReadBit())
			*res |= (1 << count);
	}
}

void CBitStream::WriteBits(DWORD value, BYTE count)
{
	ASSERT(count <= 32 && count > 0);

	DWORD mask = 1 << (count - 1);
	DWORD shift = count - 1;

	for (BYTE bit = 0; bit < count; bit++)
	{
		WriteBit((BYTE)((value & mask) >> shift));
		mask >>= 1;		shift--;
	}
}

int CBitStream::ReadDynData(DWORD *res)
{
	ASSERT(res != NULL);

	BYTE bNextByte, value;
	int length = 0;
	(*res) = 0;

	do
	{
		ReadBit(&bNextByte);
		(*res) <<= 7;
		ReadBits(&value, 7);
		(*res) |= value;
		length += 8;
	}
	while (bNextByte);

	return length;
}

void CBitStream::WriteDynData(DWORD value)
{
	BYTE tmp;

	value = (value + 7) / 8;
	for (DWORD i = 21; i > 0; i -= 7)
	{
		if (value >= (((DWORD)1) << i))
		{
			WriteBit(1);
			tmp = (BYTE)((value >> i) & 0x0000007f);
			WriteBits(tmp, 7);
		}
	}
	WriteBit(0);
	tmp = (BYTE)(value & 0x0000007f);
	WriteBits(tmp, 7);
}
