/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2001
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// Codec3D.h: interface for the CCodec3D class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_CODEC3D_H__26A17B90_BBEF_4A66_8833_93501906C6AB__INCLUDED_)
#define AFX_CODEC3D_H__26A17B90_BBEF_4A66_8833_93501906C6AB__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "CMatrix.h"
#include "BitStream.h"
#include "MGStreamParse.h"

#define CODEC_TYPE				float
#define CODEC_INTERNAL_TYPE		TYPE

#define DECODE_AHEAD
#define addChildren		AddChildrenToQueue_DepthFirst3D	//AddChildrenToQueue_BreadFirst3D
#define INT_SIGN(val)	((val) & 0x80000000)

//////////////////////////////////////////////////////////////////////
// CCodec3D Class
//////////////////////////////////////////////////////////////////////

template <class TYPE>
class CCodec3D  
{
public:
	CCodec3D(const CIPointUVW &dim, int nrLevels = 0);
	virtual ~CCodec3D();

	// Attributes
protected:
	int					m_nrLevels;
	CIPointUVW			m_dim;
	CIPointUVW			m_dimWave;

	CMatrix<int>		*m_data;
	CBitStream			*m_pStream;

	// Implementation
public:
	void	InitializeCoder();
	void	InitializeDecoder();
	void	RetrieveDC(CMatrix<TYPE> &result, RGCoding &header);
	void	PreProcess(CTileArray *subbandTiles, CMatrix<TYPE> &input, CMatrix<int> &output, RGCoding &header, CIntArray &threshold, int nrBits);
	void	WriteROIs(CBitStream *pStream, const CCubeArray &cubes, CMatrix<int> &input, int threshold, int stopThreshold = 1);
	void	ReadROIs(CBitStream *pStream, const CCubeArray &cubes, CMatrix<TYPE> &result, int threshold, double scale, bool bShared);

	bool	IsEmpty(const Cube &cube) const;
	bool	SignificantCube(const Cube& cube, int threshold);

	CMatrix<int>*&	GetData()					{ return m_data; }

protected:
	inline int	GetSymbol();
	inline void	WriteSymbol(BYTE symbol);

	void	ClearSubband(const CCubeArray &cubes);
	void	ProcessStoreAndClearDC(CMatrix<TYPE> &input, CMatrix<int> &output, RGCoding &header);

	int		PreProcessCube(const Cube& cube, CMatrix<TYPE> &input, CMatrix<int> &output, double scale);
	void	InversePreProcess(const CCubeArray &cubes, CMatrix<TYPE> &result, double scale);

	void	SignificancePass(const Cube &initial_cube, int threshold);
	void	InverseSignificancePass(const Cube &initial_cube, int threshold);
	void	InverseSignificanceLastPass(const Cube &initial_cube);

public:
	static int		ComputeThreshold(int value);
	static double	ComputeScale(double maxVal, int nrBits);
};

template <class TYPE>
CCodec3D<TYPE>::CCodec3D(const CIPointUVW &dim, int nrLevels)
{
	m_pStream = NULL;

	m_dim = dim;
	// the dimension of the buffers should not be power of 2
	CWaveletCodec<TYPE>::GetEfficientDim(m_dim, m_dimWave);

	m_data = NULL;

	// nr of resolution levels is equal to the number of decomposition levels + 1
	m_nrLevels = nrLevels;
}

template <class TYPE>
CCodec3D<TYPE>::~CCodec3D()
{
	delete m_data; m_data = NULL;
}

template <class TYPE>
inline int CCodec3D<TYPE>::GetSymbol()
{
	return (m_pStream->ReadBit() ? 0x80000000 : 0x00000000);
}

template <class TYPE>
inline void CCodec3D<TYPE>::WriteSymbol(BYTE symbol)
{
	m_pStream->WriteBit(symbol);
}

template <class TYPE>
void CCodec3D<TYPE>::InitializeCoder()
{
}

template <class TYPE>
void CCodec3D<TYPE>::InitializeDecoder()
{
	m_data = new CMatrix<int>(m_dimWave.u, m_dimWave.v, m_dimWave.w);
	m_data->MemSet(0);
}

template <class TYPE>
int CCodec3D<TYPE>::ComputeThreshold(int value)
{
	int power2 = 0;
	while (value)
	{
		value >>= 1;	power2++;
	}
	if (power2 == 0) return 0;
	return 1 << (power2 - 1);
}

template <class TYPE>
double CCodec3D<TYPE>::ComputeScale(double maxVal, int nrBits)
{
	double scale = 1.0;
	// scale to maxValue - 1
	if (!FloatEq(maxVal, 0.0))
		scale = ((1 << nrBits) - 2) / maxVal;

	double valLim = (double)(1 << (nrBits + (nrBits >> 2)));
	if (scale > valLim)		scale = valLim;

	return scale;
}

inline Cube GetChildCube(const Cube& cube, int child_no)
{
	int middle_x = (cube.width + 1)		>> 1;
	int middle_y = (cube.height + 1)	>> 1;
	int middle_z = (cube.depth + 1)		>> 1;

	Cube result;
	switch(child_no)
	{
		// front-top-left child cube
	case 0: result.Set(cube.u,cube.v,cube.w,middle_x,middle_y,middle_z); break;
		// front-top-right child cube
	case 1: result.Set(cube.u+middle_x,cube.v,cube.w,cube.width-middle_x,middle_y,middle_z); break;
		// front-bottom-left child cube
	case 2: result.Set(cube.u,cube.v+middle_y,cube.w,middle_x,cube.height-middle_y,middle_z); break;
		// front-bottom-right child cube
	case 3: result.Set(cube.u+middle_x,cube.v+middle_y,cube.w,cube.width-middle_x,cube.height-middle_y,middle_z); break;

		// rear-top-left child cube
	case 4: result.Set(cube.u,cube.v,cube.w+middle_z,middle_x,middle_y,cube.depth-middle_z); break;
		// rear-top-right child cube
	case 5: result.Set(cube.u+middle_x,cube.v,cube.w+middle_z,cube.width-middle_x,middle_y,cube.depth-middle_z); break;
		// rear-bottom-left child cube
	case 6: result.Set(cube.u,cube.v+middle_y,cube.w+middle_z,middle_x,cube.height-middle_y,cube.depth-middle_z); break;
		// rear-bottom-right child cube
	case 7: result.Set(cube.u+middle_x,cube.v+middle_y,cube.w+middle_z,cube.width-middle_x,cube.height-middle_y,cube.depth-middle_z); break;
	}
	return result;
}

inline void AddChildrenToQueue_BreadFirst3D(CCubeQueue& queue, const Cube cube)
{
	for(int i = 0; i < 8; i++)
		queue.AddTail(GetChildCube(cube, i));
}

inline void AddChildrenToQueue_DepthFirst3D(CCubeQueue& queue, const Cube cube)
{
	for(int i = 7; i >= 0; i--)
		queue.AddHead(GetChildCube(cube, i));
}

template <class TYPE>
bool CCodec3D<TYPE>::SignificantCube(const Cube& cube, int threshold)
{
	for (int k = 0; k < cube.depth; k++)
	{
		int z = cube.w + k;
		int **slice = (*m_data)[z];

		for (int j = 0; j < cube.height; j++)
		{
			int y = cube.v + j;
			int *line = slice[y];

			for (int i = 0; i < cube.width; i++)
			{
				int x = cube.u + i;
				if (abs(line[x]) >= threshold)
					return true;
			}
		}
	}
	return false;
}

template <class TYPE>
void CCodec3D<TYPE>::SignificancePass(const Cube &initial_cube, int threshold)
{
	int previous_threshold = threshold << 1;

	CCubeQueue queue;
	queue.AddHead(initial_cube);

	while (!queue.IsEmpty())
	{
		const Cube& cube = queue.RemoveHead();
		if (cube.IsValid())
		{	
			if (!SignificantCube(cube, previous_threshold))
			{
				if (SignificantCube(cube, threshold))
				{
					WriteSymbol(1);
					if (cube.IsPixel())
					{
						if ((*m_data)(cube.w, cube.v, cube.u) > 0)
							WriteSymbol(1); // "+"
						else
							WriteSymbol(0); // "-"
					}
					else
					{
						addChildren(queue, cube);
					}
				}
				else
				{
					WriteSymbol(0);
				}
			}
			else
			{
				// the block is already marked as significant from a previous threshold test
				if (cube.IsPixel())
				{
					if (abs((*m_data)(cube.w, cube.v, cube.u)) & threshold)
						WriteSymbol(1);
					else
						WriteSymbol(0);
				}
				else
				{
					addChildren(queue, cube);
				}
			}
		}
	}
}

template <class TYPE>
void CCodec3D<TYPE>::InverseSignificancePass(const Cube &initial_cube, int threshold)
{
	int previous_threshold = threshold << 1;

	CCubeQueue queue;
	queue.AddHead(initial_cube);

	while (!queue.IsEmpty())
	{
		const Cube& cube = queue.RemoveHead();
		if (cube.IsValid())
		{	
			if (!SignificantCube(cube, 1))
			{
				if (GetSymbol())
				{
					if (cube.IsPixel())
					{
						if (GetSymbol())
#ifdef DECODE_AHEAD
							(*m_data)(cube.w,cube.v,cube.u) =  ((threshold + previous_threshold) >> 1);
						else
							(*m_data)(cube.w,cube.v,cube.u) = -((threshold + previous_threshold) >> 1);
#else
							(*m_data)(cube.w,cube.v,cube.u) =  threshold;
						else
							(*m_data)(cube.w,cube.v,cube.u) = -threshold;
#endif
					}
					else
					{
						addChildren(queue, cube);
					}
				}
			}
			else
			{
				if (cube.IsPixel())
				{
#ifdef DECODE_AHEAD
					register int& val = (*m_data)(cube.w,cube.v,cube.u);
					if ((GetSymbol() ^ val) < 0)
						val += (threshold >> 1);
					else
						val -= (threshold >> 1);
#else
					if (GetSymbol())
					{
						register int& val = (*m_data)(cube.w,cube.v,cube.u);
						if (INT_SIGN(val))
							val -= threshold;
						else
							val += threshold;
					}
#endif
				}
				else
				{
					addChildren(queue, cube);
				}
			}
		}
	}
}

template <class TYPE>
void CCodec3D<TYPE>::InverseSignificanceLastPass(const Cube &initial_cube)
{
	CCubeQueue queue;
	queue.AddHead(initial_cube);

	while (!queue.IsEmpty())
	{
		const Cube& cube = queue.RemoveHead();
		if (cube.IsValid())
		{	
			if (!SignificantCube(cube, 1))
			{
				if (GetSymbol())
				{
					if (cube.IsPixel())
					{
						if (GetSymbol())
							(*m_data)(cube.w,cube.v,cube.u) = 1;
						else
							(*m_data)(cube.w,cube.v,cube.u) = -1;
					}
					else
					{
						addChildren(queue, cube);
					}
				}
			}
			else
			{
				if (cube.IsPixel())
				{
					if (GetSymbol())
					{
						register int& val = (*m_data)(cube.w,cube.v,cube.u);
						if (INT_SIGN(val))
#ifdef DECODE_AHEAD
							val++;
						else
							val--;
#else
							val--;
						else
							val++;
#endif
					}
				}
				else
				{
					addChildren(queue, cube);
				}
			}
		}
	}
}

template <class TYPE>
void CCodec3D<TYPE>::PreProcess(CTileArray *subbandTiles, CMatrix<TYPE> &input, CMatrix<int> &output, RGCoding &header, CIntArray &threshold, int nrBits)
{
	double mean = 0.0;
	const double round = 0.499;

	// find the max value for the entire cube in order to compute the scale
	TYPE maxVal = (TYPE)0;
	for (int z = 0; z < m_dim.w; z++)
	{
		TYPE **slice = input[z];
		for (int y = 0; y < m_dim.v; y++)
		{
			TYPE *line = slice[y];
			for (int x = 0; x < m_dim.u; x++)
			{
				TYPE value = (TYPE)fabs(line[x]);
				if (value > maxVal)
					maxVal = value;
			}
		}
	}

	// compute the scale
	double scale = floor(max(ComputeScale(maxVal, nrBits), 1.0));
	header.scale = (int) scale;

	// process, write the DC to the stream and clear the DC values
	// it should not affect the computation of the threshold
	ProcessStoreAndClearDC(input, output, header);

	// compute subbands threshold
	for(int i = 0; i < m_nrLevels; i++)
	{
		int maxVal2 = 0;
		const CTileArray &tiles = subbandTiles[i];
		for (int tileIdx = 0; tileIdx < tiles.GetSize(); tileIdx++)
		{
			const CCubeArray &cubes = tiles[tileIdx];
			for (int k = 0; k < cubes.GetSize(); k++)
			{
				const Cube &cube = cubes[k];
				int value = PreProcessCube(cube, input, output, header.scale);
				if (value > maxVal2) maxVal2 = value;
			}
		}
		threshold[i] = ComputeThreshold(maxVal2);
	}
}

template <class TYPE>
void CCodec3D<TYPE>::ProcessStoreAndClearDC(CMatrix<TYPE> &input, CMatrix<int> &output, RGCoding &header)
{
	int count = 0;
	const double round = 0.499;
	double scale = header.scale;
	int *gridCorners = header.gridCorners;

	int maxX = min(m_dim.u - 1, 1);
	int maxY = min(m_dim.v - 1, 1);
	int maxZ = min(m_dim.w - 1, 1);

	for (int z = 0; z <= maxZ; z++)
	{
		TYPE **sliceIn  = input[z];
		int  **sliceOut = output[z];

		for (int y = 0; y <= maxY; y++)
		{
			TYPE *lineIn  = sliceIn[y];
			int  *lineOut = sliceOut[y];

			for (int x = 0; x <= maxX; x++)
			{
				double val = lineIn[x];
				if (val >= 0.0)
				{
					gridCorners[count++] = (int)(val * scale + round);
				}
				else
				{
					gridCorners[count++] = (int)(val * scale - round);
				}
				lineIn[x] = (TYPE) 0; lineOut[x] = 0;
			}
		}
	}
}

template <class TYPE>
void CCodec3D<TYPE>::RetrieveDC(CMatrix<TYPE> &result, RGCoding &header)
{
	int count = 0;
	double scale = header.scale;
	int *gridCorners = header.gridCorners;

	int maxX = min(m_dim.u - 1, 1);
	int maxY = min(m_dim.v - 1, 1);
	int maxZ = min(m_dim.w - 1, 1);

	for (int z = 0; z <= maxZ; z++)
	{
		TYPE **slice = result[z];
		for (int y = 0; y <= maxY; y++)
		{
			TYPE  *line  = slice[y];
			for (int x = 0; x <= maxX; x++)
			{
				line[x] = (TYPE)gridCorners[count++] / (TYPE) scale;
			}
		}
	}
}

template <class TYPE>
void CCodec3D<TYPE>::ClearSubband(const CCubeArray &cubes)
{
	for (int l = 0; l < cubes.GetSize(); l++)
	{
		const Cube &cube = cubes[l];
		for (int k = 0; k < cube.depth; k++)
		{
			int z = cube.w + k;
			int **slice = (*m_data)[z];
			for (int j = 0; j < cube.height; j++)
			{
				int y = cube.v + j;
				int  *line  = slice[y];
				for (int i = 0; i < cube.width; i++)
				{
					int x = cube.u + i;
					line[x] = 0;
				}
			}
		}
	}
}

template <class TYPE>
int CCodec3D<TYPE>::PreProcessCube(const Cube& cube, CMatrix<TYPE> &input, CMatrix<int> &output, double scale)
{
	int maxVal = 0;
	const double round = 0.499;

	for (int k = 0; k < cube.depth; k++)
	{
		int z = cube.w + k;
		TYPE **sliceIn  = input[z];
		int  **sliceOut = output[z];

		for (int j = 0; j < cube.height; j++)
		{
			int y = cube.v + j;
			TYPE *lineIn  = sliceIn[y];
			int  *lineOut = sliceOut[y];

			for (int i = 0; i < cube.width; i++)
			{
				int x = cube.u + i;

				int value;
				double val = lineIn[x];
				if (val >= 0.0)
				{
					value = (int)(val * scale + round);
					if (value > maxVal) maxVal = value;
				}
				else
				{
					value = (int)(val * scale - round);
					if (-value > maxVal) maxVal = -value;
				}
				lineOut[x] = value;
			}
		}
	}
	return maxVal;
}

template <class TYPE>
void CCodec3D<TYPE>::InversePreProcess(const CCubeArray &cubes, CMatrix<TYPE> &result, double scale)
{
	for (int l = 0; l < cubes.GetSize(); l++)
	{
		const Cube &cube = cubes[l];
		for (int k = 0; k < cube.depth; k++)
		{
			int z = cube.w + k;
			int  **sliceIn = (*m_data)[z];
			TYPE **sliceOut = result[z];

			for (int j = 0; j < cube.height; j++)
			{
				int y = cube.v + j;
				int  *lineIn  = sliceIn[y];
				TYPE *lineOut = sliceOut[y];

				for (int i = 0; i < cube.width; i++)
				{
					int x = cube.u + i;
					lineOut[x] = (TYPE)(lineIn[x] / scale);
				}
			}
		}
	}
}

template <class TYPE>
bool CCodec3D<TYPE>::IsEmpty(const Cube &cube) const
{
	for (int k = 0; k < cube.depth; k++)
	{
		int z = cube.w + k;
		int  **slice = (*m_data)[z];

		for (int j = 0; j < cube.height; j++)
		{
			int y = cube.v + j;
			int  *line  = slice[y];

			for (int i = 0; i < cube.width; i++)
			{
				int x = cube.u + i;
				if (line[x] != 0)  return false;
			}
		}
	}
	return true;
}

template <class TYPE>
void CCodec3D<TYPE>::WriteROIs(CBitStream *pStream, const CCubeArray &cubes, CMatrix<int> &input, int threshold, int stopThreshold)
{
	m_data = &input;
	m_pStream = pStream;

	bool bEmpty = true;
	for (int i = 0; i < cubes.GetSize(); i++)
	{
		if (!IsEmpty(cubes[i]))
		{
			bEmpty = false; break;
		}
	}

	if (!bEmpty)
	{
		try
		{
			while (threshold >= stopThreshold)
			{
				for (int i = 0; i < cubes.GetSize(); i++)
				{
					const Cube &cube = cubes[i];
					SignificancePass(cube, threshold);
				}
				threshold >>= 1;
			}
		}
		catch(EEC_WritingLimitReached&)
		{
		}
	}
	m_data = NULL;
}

template <class TYPE>
void CCodec3D<TYPE>::ReadROIs(CBitStream *pStream, const CCubeArray &cubes, CMatrix<TYPE> &result, int threshold, double scale, bool bShared)
{
	m_pStream = pStream;

	// clean the buffer
	if (bShared)  ClearSubband(cubes);

	// do the inverse significance pass
	try
	{
		while (threshold > 1)
		{
			for (int i = 0; i < cubes.GetSize(); i++)
			{
				const Cube &cube = cubes[i];
				InverseSignificancePass(cube, threshold);
			}
			threshold >>= 1;
		}
		ASSERT(threshold == 1);
		for (int i = 0; i < cubes.GetSize(); i++)
		{
			const Cube &cube = cubes[i];
			InverseSignificanceLastPass(cube);
		}
	}
	catch(EEC_ReadingLimitReached&)
	{
	}

	// add the sign, rescale it back and write it out
	InversePreProcess(cubes, result, scale);
}

#endif // !defined(AFX_CODEC3D_H__26A17B90_BBEF_4A66_8833_93501906C6AB__INCLUDED_)
