/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGGrid.cpp: implementation of the CMGGrid class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGGrid.h"
#include "MGStreamParse.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMGGrid::CMGGrid()
{
	m_filterType = 0;
	m_maxNrSlices = 0;
	m_maxNrLevels = 0;
	m_nrLevels = 0;
	m_nrSlices = NULL;
	m_levelU = m_levelV = m_levelW = NULL;
	m_cyclicMode = CYCLIC_NONE;

	m_gridUVW = NULL;
	m_detailUVW = NULL;
	m_backupUVW = NULL;
}

CMGGrid::~CMGGrid()
{
	DeleteContent();
}

void CMGGrid::ComputeNrSlicesLevel(const CIPointUVW &nrLevels)
{
	CIPointUVW nrSlices = m_maxNrSlices;

	m_nrLevels = 1;
	m_nrSlices[m_maxNrLevels - 1] = nrSlices;

	for (int level = m_maxNrLevels - 2; level >= 0; level--)
	{
		if (m_nrLevels.u < nrLevels.u)
		{
			++m_nrLevels.u;

			ASSERT(nrSlices.u >= MIN_DEC_SIZE);
			nrSlices.u = (nrSlices.u + 1) >> 1;
		}

		if (m_nrLevels.v < nrLevels.v)
		{ 
			++m_nrLevels.v;

			ASSERT(nrSlices.v >= MIN_DEC_SIZE);
			nrSlices.v = (nrSlices.v + 1) >> 1;
		}

		if (m_nrLevels.w < nrLevels.w)
		{ 
			++m_nrLevels.w;

			ASSERT(nrSlices.w >= MIN_DEC_SIZE);
			nrSlices.w = (nrSlices.w + 1) >> 1;
		}

		m_nrSlices[level] = nrSlices;
	}
}


void CMGGrid::ComputeNrSlicesLevel(int level, const CIPointUVW &nrLevels, CIPointUVW &nrSlices)
{
	CIPointUVW levelDiff;
	ComputeNrSlicesLevel(level, nrLevels, nrSlices, levelDiff);
}

void CMGGrid::ComputeNrSlicesLevel(int level, const CIPointUVW &nrLevels, CIPointUVW &nrSlices, CIPointUVW &levelDiff)
{
	ComputeLevelDiffDown(levelDiff, nrLevels, level);

	nrSlices.u = DevideBy2Odd(nrSlices.u, levelDiff.u);
	nrSlices.v = DevideBy2Odd(nrSlices.v, levelDiff.v);
	nrSlices.w = DevideBy2Odd(nrSlices.w, levelDiff.w);
}

void CMGGrid::ComputeNrSlicesROIsLevel(int level, const CIPointUVW &nrLevels, CIPointUVW &nrSlices, CIPointUVW &nrROIs, CIPointUVW &levelDiff)
{
	ComputeNrSlicesLevel(level, nrLevels, nrSlices, levelDiff);

	nrROIs.u = DevideBy2EvenMin1(nrROIs.u, levelDiff.u);
	nrROIs.v = DevideBy2EvenMin1(nrROIs.v, levelDiff.v);
	nrROIs.w = DevideBy2EvenMin1(nrROIs.w, levelDiff.w);
}

int CMGGrid::GetNrSlicesDir(int level, int dir)
{
	ASSERT(level >= 0 && level < m_maxNrLevels);
	const CIPointUVW &nrSlices = m_nrSlices[level];

	switch (dir)
	{
	case DIR_U: return nrSlices.u;
	case DIR_V: return nrSlices.v;
	case DIR_W: return nrSlices.w;
	default: ASSERT(0); return -1;
	}
}

int CMGGrid::GetNrLevelsDir(int dir)
{
	switch (dir)
	{
	case DIR_U: return m_nrLevels.u;
	case DIR_V: return m_nrLevels.v;
	case DIR_W: return m_nrLevels.w;
	default: ASSERT(0); return -1;
	}
}

void CMGGrid::ComputeLevelSlices(BYTE *levelSlices, int dir)
{
	int skip = 1;
	int level = m_maxNrLevels - 1;
	int maxLevel = m_maxNrLevels - GetNrLevelsDir(dir);
	int maxNrSlices = GetNrSlicesDir(level, dir);

	// the last index is computed differently if a previous size was even
	// the last index is always the last slice
	do
	{
		for (int k = 0; k < maxNrSlices - skip; k += skip)
			levelSlices[k] = level;

		levelSlices[maxNrSlices - 1] = level;

		if (level > maxLevel)
			skip <<= 1;
	}
	while (--level >= 0);
}

void CMGGrid::FillSlicesLevel(const CIPointUVW &nrLevels)
{
	// compute the number of levels for each direction {U,V,W}
	// and the corresponding number of slices
	ComputeNrSlicesLevel(nrLevels);

	// initialise the level for each reference surface position
	ComputeLevelSlices(m_levelU, DIR_U);
	ComputeLevelSlices(m_levelV, DIR_V);
	ComputeLevelSlices(m_levelW, DIR_W);
}

void CMGGrid::DeleteContent()
{
	m_maxNrLevels = 0;
	m_nrLevels = 0;
	m_maxNrSlices = 0;

	delete []m_nrSlices;  m_nrSlices = NULL;

	delete []m_levelU;	m_levelU = NULL;
	delete []m_levelV;	m_levelV = NULL;
	delete []m_levelW;	m_levelW = NULL;
	delete []m_gridUVW; m_gridUVW = NULL;

	DeleteDetails();
	DeleteBackup();
}

bool CMGGrid::Init(CFPoint3DT *pts, bool defaultInit, const CIPointUVW &nrSlices, const CIPointUVW &nrLevels, CString &error)
{
	DeleteContent();

	m_maxNrLevels = max(nrLevels.u, max(nrLevels.v, nrLevels.w));
	m_maxNrSlices = nrSlices;

	try
	{
		m_levelU = new BYTE[m_maxNrSlices.u];
		m_levelV = new BYTE[m_maxNrSlices.v];
		m_levelW = new BYTE[m_maxNrSlices.w];

		m_nrSlices = new CIPointUVW[m_maxNrLevels];

		FillSlicesLevel(nrLevels);

		// initialise the grid pointers in the input buffer
		IndexBuffer(m_gridUVW, pts, m_maxNrSlices.u, m_maxNrSlices.v, m_maxNrSlices.w);

		if (defaultInit)
		{
			// the first eight values in the grid represent the corners
			// copy them in the right positions starting from the end not to override values
			pts += 7;
			int incU = max(m_maxNrSlices.u - 1, 1);
			int incV = max(m_maxNrSlices.v - 1, 1);
			int incW = max(m_maxNrSlices.w - 1, 1);

			for (int w = m_maxNrSlices.w - 1; w >= 0; w -= incW)
			{
				GridUV &gridUV = m_gridUVW[w];
				for (int v = m_maxNrSlices.v - 1; v >= 0; v -= incV)
				{
					GridU &gridU = gridUV[v];
					for (int u = m_maxNrSlices.u - 1; u >= 0; u -= incU)
						gridU[u] = *pts--;
				}
			}
			// interpolate the values of the grid based on the 8 corners
			// update the surrounding box
			InterpolBox();
		}
	}
	catch (CMemoryException *e)
	{
		GetErrorMessage(e, error);
		DeleteContent();	return false;
	}

	return true;
}

bool CMGGrid::Init(CFPoint3DT *pts, bool defaultInit, CMGStreamParser *parser, CString &error)
{
	DeleteContent();

	m_maxNrLevels = parser->GetMaxNrLevelsMesh();
	m_maxNrSlices = parser->GetMaxNrSlices();

	SetCyclicMode(parser->m_streamDescr.GetCyclicMode());

	try
	{
		m_levelU = new BYTE[m_maxNrSlices.u];
		m_levelV = new BYTE[m_maxNrSlices.v];
		m_levelW = new BYTE[m_maxNrSlices.w];

		m_nrSlices = new CIPointUVW[m_maxNrLevels];

		FillSlicesLevel(parser->GetNrLevelsMesh());

		// initialise the grid pointers in the input buffer
		IndexBuffer(m_gridUVW, pts, m_maxNrSlices.u, m_maxNrSlices.v, m_maxNrSlices.w);

		if (defaultInit)
		{
			// the first eight values in the grid represent the corners
			// copy them in the right positions starting from the end not to override values
			pts += 7;
			int incU = max(m_maxNrSlices.u - 1, 1);
			int incV = max(m_maxNrSlices.v - 1, 1);
			int incW = max(m_maxNrSlices.w - 1, 1);

			for (int w = m_maxNrSlices.w - 1; w >= 0; w -= incW)
			{
				GridUV &gridUV = m_gridUVW[w];
				for (int v = m_maxNrSlices.v - 1; v >= 0; v -= incV)
				{
					GridU &gridU = gridUV[v];
					for (int u = m_maxNrSlices.u - 1; u >= 0; u -= incU)
						gridU[u] = *pts--;
				}
			}
			// interpolate the values of the grid based on the 8 corners
			// update the surrounding box
			InterpolBox();
		}
	}
	catch (CMemoryException *e)
	{
		GetErrorMessage(e, error);
		DeleteContent();	return false;
	}

	return true;
}

void CMGGrid::RetrieveGridCorners(RGCoding *codingRG, bool differential, BYTE &nrBits)
{
	const double round = 0.499;

	int incU = max(m_maxNrSlices.u - 1, 1);
	int incV = max(m_maxNrSlices.v - 1, 1);
	int incW = max(m_maxNrSlices.w - 1, 1);

	int maxVal = 0;
	CIPoint3DT val;

	for (int w = 0, wc = 0; w < m_maxNrSlices.w; w += incW)
	{
		const GridUV &gridUV = m_gridUVW[w];

		for (int v = 0, vc = 0; v < m_maxNrSlices.v; v += incV)
		{
			const GridU &gridU = gridUV[v];

			for (int u = 0, uc = 0; u < m_maxNrSlices.u; u += incU)
			{
				const CFPoint3DT &p = gridU[u];
				val.x =  (p.x < 0) ? (int)(p.x * codingRG[X_GRID].scale + round) : (int)(p.x * codingRG[X_GRID].scale - round);
				val.y =  (p.y < 0) ? (int)(p.y * codingRG[Y_GRID].scale + round) : (int)(p.y * codingRG[Y_GRID].scale - round);
				val.z =  (p.z < 0) ? (int)(p.z * codingRG[Z_GRID].scale + round) : (int)(p.z * codingRG[Z_GRID].scale - round);

				int idx = 4*wc + 2*vc + uc;
				if (differential)
				{
					codingRG[X_GRID].gridCorners[idx] = val.x - codingRG[X_GRID].gridCorners[idx];
					codingRG[Y_GRID].gridCorners[idx] = val.y - codingRG[Y_GRID].gridCorners[idx];
					codingRG[Z_GRID].gridCorners[idx] = val.z - codingRG[Z_GRID].gridCorners[idx];
				}
				else
				{
					codingRG[X_GRID].gridCorners[idx] = val.x;
					codingRG[Y_GRID].gridCorners[idx] = val.y;
					codingRG[Z_GRID].gridCorners[idx] = val.z;
				}
				
				if (maxVal < abs(val.x))	maxVal = abs(val.x);
				if (maxVal < abs(val.y))	maxVal = abs(val.y);
				if (maxVal < abs(val.z))	maxVal = abs(val.z);
			}
		}
	}

	nrBits = CMGStreamParser::GetNrBits(maxVal);
}

void CMGGrid::ExportCorners(CStdioFile &file)
{
	// header
	file.WriteString("\t\tgridCoord Coordinate {\n");
	file.WriteString("\t\t\tpoint [\n");

	// contents
	CString coord;

	int incU = max(m_maxNrSlices.u - 1, 1);
	int incV = max(m_maxNrSlices.v - 1, 1);
	int incW = max(m_maxNrSlices.w - 1, 1);

	for (int w = 0; w < m_maxNrSlices.w; w += incW)
	{
		const GridUV &gridUV = m_gridUVW[w];

		for (int v = 0; v < m_maxNrSlices.v; v += incV)
		{
			const GridU &gridU = gridUV[v];

			for (int u = 0; u < m_maxNrSlices.u; u += incU)
			{
				const CFPoint3DT &p = gridU[u];

				//				coord.Format("\t   %f %f %f,\n", (p.x - offsetX) / sz, (p.y - offsetY) / sz, (p.z - offsetZ) / sz);
				coord.Format("\t\t\t\t%f %f %f,\n", p.x, p.y, p.z);
				file.WriteString(coord);
			}
		}
	}

	// finish
	file.WriteString("\t\t\t]\n\t\t}\n");
}

void CMGGrid::Export(CStdioFile &file, int level)
{
	int nrSlicesU = GetNrSlicesU(level);
	int nrSlicesV = GetNrSlicesV(level);
	int nrSlicesW = GetNrSlicesW(level);

	// header
	file.WriteString("\t\tgridCoord Coordinate {\n");
	file.WriteString("\t\t\tpoint [\n");

	// contents
	CString coord;

	int w = -1;
	for (int z = 0; z < nrSlicesW; z++)
	{
		w = GetNextSliceW(w, level);
		const GridUV &gridUV = m_gridUVW[w];

		int v = -1;
		for (int y = 0; y < nrSlicesV; y++)
		{
			v = GetNextSliceV(v, level);
			const GridU &gridU = gridUV[v];

			int u = -1;
			for (int x = 0; x < nrSlicesU; x++)
			{
				u = GetNextSliceU(u, level);
				const CFPoint3DT &p = gridU[u];

				coord.Format("\t   %f %f %f,\n", p.x, p.y, p.z);
				file.WriteString(coord);
			}
		}
	}

	// finish
	file.WriteString("\t\t\t]\n\t\t}\n");
}

void CMGGrid::GetSurrBox(CSurrBoxT &gridBox)
{
	CFPoint3DT pb = gridBox.bottom;
	CFPoint3DT pt = gridBox.top;

	pb.x = pb.y = pb.z =  FLT_MAX;
	pt.x = pt.y = pt.z = -FLT_MAX;

	for (int w = 0; w < m_maxNrSlices.w; w++)
	{
		GridUV &gridUV = m_gridUVW[w];

		for (int v = 0; v < m_maxNrSlices.v; v++)
		{
			GridU &gridU = gridUV[v];

			for (int u = 0; u < m_maxNrSlices.u; u++)
			{
				if (pb.x > gridU[u].x)	pb.x = gridU[u].x;
				if (pb.y > gridU[u].y)	pb.y = gridU[u].y;
				if (pb.z > gridU[u].z)	pb.z = gridU[u].z;
				if (pt.x < gridU[u].x)	pt.x = gridU[u].x;
				if (pt.y < gridU[u].y)	pt.y = gridU[u].y;
				if (pt.z < gridU[u].z)	pt.z = gridU[u].z;
			}
		}
	}
}
