/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGGridCodec.h: interface for the CMGGridCodec class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_MGGRIDCODEC_H__85E0A823_97E8_4B6D_8F77_A9974CF76CD7__INCLUDED_)
#define AFX_MGGRIDCODEC_H__85E0A823_97E8_4B6D_8F77_A9974CF76CD7__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Codec3D.h"
#include "Wavelet3D.h"
#include "WaveletShort3D.h"
#include "WaveletSmooth3D.h"
#include "MGGrid.h"
#include "ROISizePos.h"

class CBitStream;
class MGStreamHeader;

///////////////////////////////////////////////////////////////////////////////////////////////
// class CGridCodec

// the notations of the 8 subbands of the wavelet decomposition
//
//       W         HLL -------- HLH  
//     /           /|           /|   
//   /           /  |         /  |   
//   ----> U   LLL -------- LLH  |   
//  |           |   |        |   |   
//  |           |  HHL ------|- HHH  
//  V           |  /         |  /    
//              |/           |/
//             LHL -------- LHH

enum CodecSpecific {          // the 8 subbands of the wavelet decomposition
	LLL = 0,
	LLH = 1,
	LHL = 2,
	LHH = 3,
	HLL = 4,
	HLH = 5,
	HHL = 6,
	HHH = 7
};

template <class TYPE>
class CMGGridCodec
{
public:
	virtual ~CMGGridCodec() {};

	static int  GetMaxNrLevelsG(const CIPointUVW &nrSlices);

	static	int*	CreateSubbands(const CIPointUVW &nrSlices, CIPointUVW &sizeL, CIPointUVW &sizeH, int &nrSubbands);
	static	void	MapROIToTiles(const Cube &roi, const CIPointUVW &sizeL, const CIPointUVW &sizeH, int *subbands, int nrSubbands, CCubeArray &cubes);

	void  CreateSubbandTiles(int levelG, const ROIDescription &roiDescr);
	int	  GetMaxNrLevelsG() const	{ return m_maxNrLevelsG; }

	CCodec3D<TYPE>*			GetCodec() const		{ return m_pCodec; }
	CWaveletCodec<TYPE>*	GetPrecode() const		{ return m_pPrecode; }
	CTileArray*				GetSubbandTiles() const	{ return m_subbandTiles; }

protected:
	CIPointUVW		m_nrSlices;             // size of the grid
	CIPointUVW		m_dimWave;              // dimension of the wavelet buffer to optimise memory access
	int             m_maxNrLevelsM;         // number of resolution levels of the mesh
	CIPointUVW      m_nrDecLevels;          // number of decomposition levels of the grid for each (u,v,w) direction
	CIPointUVW      m_nrLevelsG;			// number of resolution levels for the grid for each (u,v,w) direction
	int             m_maxNrLevelsG;         // maximum number of resolution levels of the grid
	CMGGrid		  * m_pGrid;                // pointer to the grid object
	CTileArray    * m_subbandTiles;         // array of tiles for each subband of the wavelet image

	MGStreamDescr         * m_pStreamDescr; // pointer to the stream description object
	CMGStreamParser       * m_pParser;      // pointer to the parser object
	CCodec3D<TYPE>		  * m_pCodec;       // pointer to the grid codec object
	CWaveletCodec<TYPE>	  * m_pPrecode;     // pointer to the wavelet transform object


#ifdef DEBUG_LOOKUP
	void  CheckLookup(Cube &cube);
	BYTE ***m_lookup;
#endif
};

template <class TYPE>
int CMGGridCodec<TYPE>::GetMaxNrLevelsG(const CIPointUVW &nrSlices)
{
	CIPointUVW decLevels;
	CWaveletCodec<TYPE>::GetNrDecLevels(nrSlices, decLevels);

	// The minimum grid size that will be encoded as subbands is MIN_SPLIT_SIZE = 9
	// The number of levels = 1 as long as the grid size is less than MIN_SPLIT_SIZE
	// independent of the number of decomposition levels.
	// In the case levels = 1 the grid will be encoded as one Tile
	return max(CWaveletCodec<TYPE>::GetMaxDecLevel(decLevels) - 1, 1);
}

template <class TYPE>
int *CMGGridCodec<TYPE>::CreateSubbands(const CIPointUVW &nrSlices, CIPointUVW &sizeL, CIPointUVW &sizeH, int &nrSubbands)
{
	const int Z_SPLIT = 4;
	const int Y_SPLIT = 2;
	const int X_SPLIT = 1;

	sizeL = nrSlices, sizeH = 0;

	// temporary stacks of cubes that are subdivided
	int idx1 = 0, idx2 = 0;
	int * stack1 = new int[8];
	int * stack2 = new int[8];

	stack1[idx1++] = 0;

	int countSplit = 0;
	if (nrSlices.w >= MIN_DEC_SIZE)
	{
		stack2[idx2++] = 0;
		stack2[idx2++] = Z_SPLIT;

		sizeH.w = sizeL.w / 2;  sizeL.w -= sizeH.w;

		countSplit = 1;
		idx1 = idx2;		idx2 = 0;
		int *swap = stack1;		stack1 = stack2;		stack2 = swap;
	}

	if (nrSlices.v >= MIN_DEC_SIZE)
	{
		for (int i = 0; i < idx1; i++)
		{
			stack2[idx2++] = stack1[i];
			stack2[idx2++] = stack1[i] + Y_SPLIT;
		}
		sizeH.v = sizeL.v / 2;  sizeL.v -= sizeH.v;

		++countSplit;
		idx1 = idx2;		idx2 = 0;
		int *swap = stack1;		stack1 = stack2;		stack2 = swap;
	}

	if (nrSlices.u >= MIN_DEC_SIZE)
	{
		for (int i = 0; i < idx1; i++)
		{
			stack2[idx2++] = stack1[i];
			stack2[idx2++] = stack1[i] + X_SPLIT;
		}
		sizeH.u = sizeL.u / 2;  sizeL.u -= sizeH.u;

		++countSplit;
		idx1 = idx2;		idx2 = 0;
		int *swap = stack1;		stack1 = stack2;		stack2 = swap;
	}

	delete []stack2;

	// since the LLL subband will not be tiled we remove it from the list
	ASSERT(idx1 >= 2);
	for (int idx = 1; idx < idx1; idx++)
		stack1[idx - 1] = stack1[idx];

	nrSubbands = idx1 - 1;

	if (countSplit == 3)
	{
		// the subbands have to be arranged in the following order
		// in order to have efficient lossy encoding
		// LLH-LHL-HLL-LHH-HLH-HHL-HHH
		int tmp = stack1[2];  stack1[2] = stack1[3];  stack1[3] = tmp;
	}

	return stack1;
}

#ifdef DEBUG_LOOKUP
template <class TYPE>
void CMGGridCodec<TYPE>::CheckLookup(Cube & cube)
{
	for (int w = cube.w; w < cube.w + cube.depth; w++)
	{
		for (int v = cube.v; v < cube.v + cube.height; v++)
		{
			for (int u = cube.u; u < cube.u + cube.width; u++)
			{
				if (m_lookup[w][v][u] != 0) ASSERT(0);
				m_lookup[w][v][u] = 1;
			}
		}
	}
}
#endif

template <class TYPE>
void CMGGridCodec<TYPE>::CreateSubbandTiles(int levelG, const ROIDescription &roiDescr)
{
	GStreamLevel &gridLevel = m_pStreamDescr->GetGridLevel(levelG);
	const CIPointUVW &nrSlices = gridLevel.GetNrSlices();

	// allocate the Tiles array
	CTileArray &tiles = m_subbandTiles[levelG];

	if (!gridLevel.CanSplit())
	{
		// if nrSlices is less or equalt to MIN_SPLIT_SIZE than
		// the cube level will not be split into subbands, it will be coded as it is
		ASSERT(gridLevel.GetTotalNrTiles() == 1);

		Cube cube(0, 0, 0, nrSlices.u, nrSlices.v, nrSlices.w);
		tiles.SetSize(1);   tiles[0].SetSize(1);
		tiles[0][0] = cube;

#ifdef DEBUG_LOOKUP
		CheckLookup(cube);
#endif

		return;
	}
	else
	{
		// the cube level is split into the following subbands
		int nrSubbands;
		CIPointUVW sizeL, sizeH;
		int *subbands = CreateSubbands(nrSlices, sizeL, sizeH, nrSubbands);

		// each subband is split into tiles
		tiles.SetSize(gridLevel.GetTotalNrTiles());

		// take the Tiles one by one and determine the cubes
		// from every subband belonging to each Tile
		for (int index = 0; index < gridLevel.GetTotalNrTiles(); index++)
		{
			Cube roi;
			roiDescr.ComputePositionAndSize(roi, index, gridLevel.GetNrTiles(), gridLevel.GetLevelDiff());

			CCubeArray &cubes = tiles[index];
			cubes.SetSize(nrSubbands);

			MapROIToTiles(roi, sizeL, sizeH, subbands, nrSubbands, cubes);

#ifdef DEBUG_LOOKUP
			for (int k = 0; k < nrSubbands; k++)
				CheckLookup(cubes[k]);
#endif
		}

		delete []subbands;
	}
}

template <class TYPE>
void CMGGridCodec<TYPE>::MapROIToTiles(const Cube &roi, const CIPointUVW &sizeL, const CIPointUVW &sizeH, int *subbands, int nrSubbands, CCubeArray &cubes)
{
	for (int index = 0; index < nrSubbands; index++)
	{
		Cube tile;

		switch (subbands[index])
		{
		case LLL: AfxMessageBox("error");
		case LLH: // filter on X
			tile.u      =  sizeL.u       + ((roi.u % 2 == 0) ? roi.u / 2: (roi.u - 1) / 2);
			tile.v      = (sizeH.v != 0) ? ((roi.v % 2 == 0) ? roi.v / 2: (roi.v + 1) / 2) : roi.v;
			tile.w      = (sizeH.w != 0) ? ((roi.w % 2 == 0) ? roi.w / 2: (roi.w + 1) / 2) : roi.w; 

			tile.width  =                   (roi.u % 2 == 0) ?  roi.width / 2      : (roi.width + 1) / 2;
			tile.height = (sizeH.v != 0) ? ((roi.v % 2 == 0) ? (roi.height + 1) / 2:  roi.height / 2) : roi.height;
			tile.depth  = (sizeH.w != 0) ? ((roi.w % 2 == 0) ? (roi.depth  + 1) / 2:  roi.depth  / 2) : roi.depth;
			break;
		case LHL: // filter on Y
			tile.u      = (sizeH.u != 0) ? ((roi.u % 2 == 0) ? roi.u / 2: (roi.u + 1) / 2) : roi.u;
			tile.v      =  sizeL.v       + ((roi.v % 2 == 0) ? roi.v / 2: (roi.v - 1) / 2); 
			tile.w      = (sizeH.w != 0) ? ((roi.w % 2 == 0) ? roi.w / 2: (roi.w + 1) / 2) : roi.w; 

			tile.width  = (sizeH.u != 0) ? ((roi.u % 2 == 0) ? (roi.width + 1)  / 2:  roi.width  / 2) : roi.width;
			tile.height =                   (roi.v % 2 == 0) ?  roi.height / 2     : (roi.height + 1) / 2;
			tile.depth  = (sizeH.w != 0) ? ((roi.w % 2 == 0) ? (roi.depth  + 1) / 2:  roi.depth  / 2) : roi.depth;
			break;
		case HLL: // filter on Z
			tile.u      = (sizeH.u != 0) ? ((roi.u % 2 == 0) ? roi.u / 2: (roi.u + 1) / 2) : roi.u;
			tile.v      = (sizeH.v != 0) ? ((roi.v % 2 == 0) ? roi.v / 2: (roi.v + 1) / 2) : roi.v; 
			tile.w      =  sizeL.w       + ((roi.w % 2 == 0) ? roi.w / 2: (roi.w - 1) / 2); 

			tile.width  = (sizeH.u != 0) ? ((roi.u % 2 == 0) ? (roi.width + 1)  / 2:  roi.width  / 2) : roi.width;
			tile.height = (sizeH.v != 0) ? ((roi.v % 2 == 0) ? (roi.height + 1) / 2:  roi.height / 2) : roi.height;
			tile.depth  =                   (roi.w % 2 == 0) ?  roi.depth / 2      : (roi.depth  + 1) / 2;
			break;
		case LHH: // filter on X and Y
			tile.u      =  sizeL.u       + ((roi.u % 2 == 0) ? roi.u / 2: (roi.u - 1) / 2);
			tile.v      =  sizeL.v       + ((roi.v % 2 == 0) ? roi.v / 2: (roi.v - 1) / 2); 
			tile.w      = (sizeH.w != 0) ? ((roi.w % 2 == 0) ? roi.w / 2: (roi.w + 1) / 2) : roi.w; 

			tile.width  =                   (roi.u % 2 == 0) ?  roi.width  / 2     : (roi.width  + 1) / 2;
			tile.height =                   (roi.v % 2 == 0) ?  roi.height / 2     : (roi.height + 1) / 2;
			tile.depth  = (sizeH.w != 0) ? ((roi.w % 2 == 0) ? (roi.depth  + 1) / 2:  roi.depth  / 2) : roi.depth;
			break;
		case HLH: // filter on X and Z
			tile.u      =  sizeL.u       + ((roi.u % 2 == 0) ? roi.u / 2: (roi.u - 1) / 2);
			tile.v      = (sizeH.v != 0) ? ((roi.v % 2 == 0) ? roi.v / 2: (roi.v + 1) / 2) : roi.v; 
			tile.w      =  sizeL.w       + ((roi.w % 2 == 0) ? roi.w / 2: (roi.w - 1) / 2); 

			tile.width  =                   (roi.u % 2 == 0) ?  roi.width  / 2     : (roi.width  + 1) / 2;
			tile.height = (sizeH.v != 0) ? ((roi.v % 2 == 0) ? (roi.height + 1) / 2:  roi.height / 2) : roi.height;
			tile.depth  =                   (roi.w % 2 == 0) ?  roi.depth / 2      : (roi.depth  + 1) / 2;
			break;
		case HHL: // filter on Y and Z
			tile.u      = (sizeH.u != 0) ? ((roi.u % 2 == 0) ? roi.u / 2: (roi.u + 1) / 2) : roi.u;
			tile.v      =  sizeL.v       + ((roi.v % 2 == 0) ? roi.v / 2: (roi.v - 1) / 2); 
			tile.w      =  sizeL.w       + ((roi.w % 2 == 0) ? roi.w / 2: (roi.w - 1) / 2); 

			tile.width  = (sizeH.u != 0) ? ((roi.u % 2 == 0) ? (roi.width + 1)  / 2:  roi.width  / 2) : roi.width;
			tile.height =                   (roi.v % 2 == 0) ?  roi.height / 2     : (roi.height + 1) / 2;
			tile.depth  =                   (roi.w % 2 == 0) ?  roi.depth  / 2     : (roi.depth  + 1) / 2;
			break;
		case HHH: // filter on X and Y and Z
			tile.u      = sizeL.u + ((roi.u % 2 == 0) ? roi.u / 2: (roi.u - 1) / 2);
			tile.v      = sizeL.v + ((roi.v % 2 == 0) ? roi.v / 2: (roi.v - 1) / 2); 
			tile.w      = sizeL.w + ((roi.w % 2 == 0) ? roi.w / 2: (roi.w - 1) / 2); 

			tile.width  =           (roi.u % 2 == 0) ?  roi.width  / 2     : (roi.width  + 1) / 2;
			tile.height =           (roi.v % 2 == 0) ?  roi.height / 2     : (roi.height + 1) / 2;
			tile.depth  =           (roi.w % 2 == 0) ?  roi.depth  / 2     : (roi.depth  + 1) / 2;
			break;
		default: ASSERT(0);
		}

		cubes[index] = tile;
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////
// class CMGGridDecoder

template <class TYPE>
class CMGGridDecoder : public CMGGridCodec<TYPE>
{
public:
	CMGGridDecoder(CMGStreamParser *pParser, CMGGrid *grid);
	virtual ~CMGGridDecoder();

	// implementation
public:
	void	RecoverLevel(int level, bool progressiveFrame, bool &bModified);	            // read and decompress a resolution level
	void	UpdateGridCorners();
	bool	RecoverTile(int tileIdx, int levelGrid);
	void	ApplyTransform(bool progressiveFrame);
	void	ApplyGridCorners();

protected:
	void	DecodeSubband(int subLevel, bool &bModified, bool progressiveFrame, bool lastLevel);
	void	DecodeTile(int dir, int subLevel, int tileIdx, bool &bModified, bool progressiveFrame);
	void	InitialiseMeshGrid();
	void	UpdateMeshGrid();

public:
	CMatrix<TYPE>     m_wavelet[3];
	CMatrix<TYPE>     m_image[3];
};

// Construction/Destruction

template <class TYPE>
CMGGridDecoder<TYPE>::CMGGridDecoder(CMGStreamParser *pParser, CMGGrid *grid)
{
	m_pGrid = grid;
	m_pParser = pParser;
	m_subbandTiles = NULL;
	m_pStreamDescr = &pParser->m_streamDescr;

	m_nrSlices = m_pStreamDescr->GetMaxNrSlices();
	m_nrLevelsG = m_pStreamDescr->GetNrLevelsGrid();
	m_maxNrLevelsM = m_pStreamDescr->GetMaxNrLevelsMesh();
	m_maxNrLevelsG = m_pStreamDescr->GetMaxNrLevelsGrid();

#ifdef DEBUG_LOOKUP
	m_lookup = (BYTE ***)Alloc3DTypedBuffer<BYTE>(m_nrSlices.u, m_nrSlices.v, m_nrSlices.w, 0);
#endif

	// initialise the precoder
	switch (m_pStreamDescr->GetFilterType())
	{
	case CMGGrid::LINEAR_FILTER:
		m_pPrecode = new CWaveletCodecShort<TYPE>(m_nrSlices, pParser->m_streamDescr.GetCyclicMode(), true);
		break;
	case CMGGrid::SMOOTH_FILTER:
		m_pPrecode = new CWaveletCodecSmooth<TYPE>(m_nrSlices, pParser->m_streamDescr.GetCyclicMode(), true);
		break;
	}
	m_pPrecode->GetEfficientDim(m_nrSlices, m_dimWave);

	// Subtract 1 since the first and second decomposition level are coded together.
	// They are too small to be coded separately or as tiles.
	m_nrDecLevels = m_pPrecode->GetDecLevels();
	ASSERT(m_maxNrLevelsG == max(m_pPrecode->GetMaxDecLevel(m_nrDecLevels) - 1, 1));

	// initialise of the stream
	m_pCodec = new CCodec3D<TYPE>(m_nrSlices, m_maxNrLevelsG);

	// initialise the reference grid decoder
	m_pCodec->InitializeDecoder();

	for (int dir = X_GRID; dir <= Z_GRID; dir++)
	{
		m_wavelet[dir].Init(m_dimWave.u, m_dimWave.v, m_dimWave.w);
		m_image[dir].Init(m_dimWave.u, m_dimWave.v, m_dimWave.w);
	}

	// create the tiles array
	m_subbandTiles = new CTileArray[m_maxNrLevelsG];
}

template <class TYPE>
CMGGridDecoder<TYPE>::~CMGGridDecoder()
{
	delete m_pCodec;	        m_pCodec = NULL;
	delete m_pPrecode;		    m_pPrecode = NULL;
	delete []m_subbandTiles;  m_subbandTiles = NULL;

#ifdef DEBUG_LOOKUP
	for (int w = 0; w < m_nrSlices.w; w++)
	{
		for (int v = 0; v < m_nrSlices.v; v++)
		{
			for (int u = 0; u < m_nrSlices.u; u++)
			{
				if (m_lookup[w][v][u] != 1) ASSERT(0);
			}
		}
	}
	delete []m_lookup;

	FILE *file1 = fopen("DecDump.txt", "wt");
	FILE *file2 = fopen("WavDecDump.txt", "wt");
	for (int k = 0; k < 3; k++)
	{
		for (int w = 0; w < m_dimWave.w; w++)
		{
			for (int v = 0; v < m_dimWave.v; v++)
			{
				for (int u = 0; u < m_dimWave.u; u++)
				{
					fprintf(file1, "%f\n", m_wavelet[k][w][v][u]);
					fprintf(file2, "%f\n", m_image[k][w][v][u]);
				}
			}
		}
	}
	fclose(file1);
	fclose(file2);
#endif
}

template <class TYPE>
void CMGGridDecoder<TYPE>::DecodeSubband(int subLevel, bool &bModified, bool progressiveFrame, bool lastLevel)
{
	// decode the (x, y, z) grid coordinates
	for (int dir = X_GRID; dir <= Z_GRID; dir++)
	{
		const CTileArray &tiles = m_subbandTiles[subLevel];
		for (int tileIdx = 0; tileIdx < tiles.GetSize(); tileIdx++)
			DecodeTile(dir, subLevel, tileIdx, bModified, progressiveFrame);
	}
}

template <class TYPE>
void CMGGridDecoder<TYPE>::DecodeTile(int dir, int subLevel, int tileIdx, bool &bModified, bool progressiveFrame)
{
	CMGStreamParser::QueryType query;
	switch (dir)
	{
	case X_GRID:  query = CMGStreamParser::Q_GRIDX;   break;
	case Y_GRID:  query = CMGStreamParser::Q_GRIDY;   break;
	case Z_GRID:  query = CMGStreamParser::Q_GRIDZ;   break;
	}

	GStreamLevel &gridLevel = m_pStreamDescr->GetGridLevel(subLevel);
	int threshold = gridLevel.GetThreshold(dir);
	RGCoding &paramRG  = m_pStreamDescr->GetGridCodingParam(dir);
	const CCubeArray &cubes = m_subbandTiles[subLevel][tileIdx];

	// tell the parser that the decoding of a new grid tile starts
	// the parser should set the stream position for for each tile
	bool bTileExists = m_pParser->OnBeginGridDecodingTile(query, subLevel, tileIdx);
	if (bTileExists)
	{
		m_pCodec->ReadROIs(m_pParser->GetStream(), cubes, m_wavelet[dir],
			threshold, paramRG.scale, true);

		// if the Tile is the entire grid, there was no splitting because the 
		// size of the grid is too low, write back the corners before applying the wavelet transform
		if (!gridLevel.CanSplit() && !progressiveFrame)
		{
			m_pCodec->RetrieveDC(m_wavelet[dir], paramRG);
		}

		bModified = true;
	}

	gridLevel.SetStreamedTile(dir, tileIdx, true);

	// tell the parser that the decoding of a new grid tile ends
	// the parser should update the bit counter for the decoded tile
	m_pParser->OnEndGridDecodingTile(subLevel);
}

template <class TYPE>
bool CMGGridDecoder<TYPE>::RecoverTile(int tileIdx, int levelGrid)
{
	// decode the (x, y, z) grid coordinates
	ASSERT(levelGrid < m_maxNrLevelsG);
	bool bModified = false;

	for (int dir = X_GRID; dir <= Z_GRID; dir++)
		DecodeTile(dir, levelGrid, tileIdx, bModified);

	return bModified;
}

template <class TYPE>
void CMGGridDecoder<TYPE>::RecoverLevel(int levelG, bool progressiveFrame, bool &bModified)
{
	ASSERT(levelG < m_maxNrLevelsG);
	DecodeSubband(levelG, bModified, progressiveFrame, levelG == (m_maxNrLevelsG - 1));
}

template <class TYPE>
void CMGGridDecoder<TYPE>::ApplyTransform(bool progressiveFrame)
{
	m_pPrecode->WaveletRec3D(m_wavelet[0], m_wavelet[1], m_wavelet[2],
		m_image[0]  , m_image[1]  , m_image[2]);

	m_wavelet[0] = 0;	m_wavelet[1] = 0; m_wavelet[2] = 0;

	if (progressiveFrame)	UpdateMeshGrid();
	else					InitialiseMeshGrid();
}

template <class TYPE>
void CMGGridDecoder<TYPE>::ApplyGridCorners()
{
	for (int dir = X_GRID; dir <= Z_GRID; dir++)
	{
		RGCoding &paramRG  = m_pStreamDescr->GetGridCodingParam(dir);
		m_pCodec->RetrieveDC(m_wavelet[dir], paramRG);
	}

	m_pPrecode->WaveletRec3D(m_wavelet[0], m_wavelet[1], m_wavelet[2],
		m_image[0]  , m_image[1]  , m_image[2]);

	m_wavelet[0] = 0;	m_wavelet[1] = 0; m_wavelet[2] = 0;

	InitialiseMeshGrid();
}

template <class TYPE>
void CMGGridDecoder<TYPE>::UpdateGridCorners()
{
	RGCoding *paramRG  = m_pStreamDescr->GetGridCodingParam();
	for (int dir = X_GRID; dir <= Z_GRID; dir++)
	{
		m_wavelet[dir] = 0;
		m_pCodec->RetrieveDC(m_wavelet[dir], paramRG[dir]);
	}
	m_pPrecode->WaveletRec3D(m_wavelet[0], m_wavelet[1], m_wavelet[2],
		m_image[0]  , m_image[1]  , m_image[2]);

	m_wavelet[0] = 0;	m_wavelet[1] = 0; m_wavelet[2] = 0;

	// update the entire volume
	InitialiseMeshGrid();
}

template <class TYPE>
void CMGGridDecoder<TYPE>::UpdateMeshGrid()
{
	for (int z = 0; z < m_nrSlices.w; z++)
	{
		GridUV &gridUV = (*m_pGrid)[z];
		TYPE **sliceX = m_image[X_GRID][z];
		TYPE **sliceY = m_image[Y_GRID][z];
		TYPE **sliceZ = m_image[Z_GRID][z];

		for (int y = 0; y < m_nrSlices.v; y++)
		{
			GridU &gridU = gridUV[y];
			TYPE *elemX = sliceX[y];
			TYPE *elemY = sliceY[y];
			TYPE *elemZ = sliceZ[y];

			for (int x = 0; x < m_nrSlices.u; x++)
			{
				CFPoint3DT &p = gridU[x];
				p.x += (float)*elemX++;
				p.y += (float)*elemY++;
				p.z += (float)*elemZ++;
			}
		}
	}
}

template <class TYPE>
void CMGGridDecoder<TYPE>::InitialiseMeshGrid()
{
	for (int z = 0; z < m_nrSlices.w; z++)
	{
		GridUV &gridUV = (*m_pGrid)[z];
		TYPE **sliceX = m_image[X_GRID][z];
		TYPE **sliceY = m_image[Y_GRID][z];
		TYPE **sliceZ = m_image[Z_GRID][z];

		for (int y = 0; y < m_nrSlices.v; y++)
		{
			GridU &gridU = gridUV[y];
			TYPE *elemX = sliceX[y];
			TYPE *elemY = sliceY[y];
			TYPE *elemZ = sliceZ[y];

			for (int x = 0; x < m_nrSlices.u; x++)
			{
				CFPoint3DT &p = gridU[x];
				p.x = (float)*elemX++;
				p.y = (float)*elemY++;
				p.z = (float)*elemZ++;
			}
		}
	}	
}


///////////////////////////////////////////////////////////////////////////////////////////////
// class CMGGridCoder

template <class TYPE>
class CMGGridCoder : public CMGGridCodec<TYPE>
{
public:
	CMGGridCoder(CMGStreamParser *pParser, CMGGrid *grid, int nrQuantBits, int filterType);
	virtual ~CMGGridCoder();

	// implementation
public:
	void  StreamHeader();							                // preprocess the input data and call StreamRGHeader
	void  StreamLevel(int level, BYTE update_replace);              // compress and write a resolution level
	void  InitExchangeObjectGrid(const ROIDescription &roiDescr);
	void  ExtractCoordinates(CMatrix<TYPE> &coordX, CMatrix<TYPE> &coordY, CMatrix<TYPE> &coordZ);

	CMatrix<int>&	GetCoeff(int dir)		{ return m_exchange[dir]; }
	CMatrix<int>*	GetCoeff()				{ return m_exchange; }

protected:
	void  EncodeSubband(int subLevel, BYTE update_replace);
	void  EncodeTile(int dir, int subLevel, int tileIdx);

protected:
	int				m_nrBits;                       // number of quantisation bits for TYPE to int conversion
	CMatrix<int>	m_exchange[3];                  // buffers holding the quantised grid coefficients (x, y, z)
};

// Construction/Destruction

template <class TYPE>
CMGGridCoder<TYPE>::CMGGridCoder(CMGStreamParser *pParser, CMGGrid *grid, int nrQuantBits, int filterType)
{
	m_pGrid = grid;
	m_pParser = pParser;
	m_subbandTiles = NULL;
	m_pStreamDescr = &pParser->m_streamDescr;

	m_nrBits = nrQuantBits;
	m_nrSlices = m_pStreamDescr->GetMaxNrSlices();
	m_nrLevelsG = m_pStreamDescr->GetNrLevelsGrid();
	m_maxNrLevelsM = m_pStreamDescr->GetMaxNrLevelsMesh();
	m_maxNrLevelsG = m_pStreamDescr->GetMaxNrLevelsGrid();

#ifdef DEBUG_LOOKUP
	m_lookup = (BYTE ***)Alloc3DTypedBuffer<BYTE>(m_nrSlices.u, m_nrSlices.v, m_nrSlices.w, 0);
#endif

	m_pStreamDescr->SetFilterType(filterType);

	// initialise the precoder
	switch (m_pStreamDescr->GetFilterType())
	{
	case CMGGrid::LINEAR_FILTER:
		m_pPrecode = new CWaveletCodecShort<TYPE>(m_nrSlices, pParser->m_streamDescr.GetCyclicMode(), true);
		break;
	case CMGGrid::SMOOTH_FILTER:
		m_pPrecode = new CWaveletCodecSmooth<TYPE>(m_nrSlices, pParser->m_streamDescr.GetCyclicMode(), true);
		break;
	}
	m_pPrecode->GetEfficientDim(m_nrSlices, m_dimWave);

	// Subtract 1 since the first and second decomposition level are coded together.
	// They are too small to be coded separately or as tiles.
	m_nrDecLevels = m_pPrecode->GetDecLevels();
	ASSERT(m_maxNrLevelsG == max(m_pPrecode->GetMaxDecLevel(m_nrDecLevels) - 1, 1));

	// initialise of the stream
	m_pCodec = new CCodec3D<TYPE>(m_nrSlices, m_maxNrLevelsG);

	// initialise the reference grid encoder
	m_pCodec->InitializeCoder();

	for (int dir = X_GRID; dir <= Z_GRID; dir++)
		m_exchange[dir].Init(m_dimWave.u, m_dimWave.v, m_dimWave.w);

	// create the tiles array
	m_subbandTiles = new CTileArray[m_maxNrLevelsG];

	// initialise the exchange object with the grid coding details
	InitExchangeObjectGrid(m_pStreamDescr->GetROIDescr());
}

template <class TYPE>
CMGGridCoder<TYPE>::~CMGGridCoder()
{
	delete m_pCodec;			    m_pCodec = NULL;
	delete m_pPrecode;		    m_pPrecode = NULL;
	delete []m_subbandTiles;  m_subbandTiles = NULL;

#ifdef DEBUG_LOOKUP
	for (int w = 0; w < m_nrSlices.w; w++)
	{
		for (int v = 0; v < m_nrSlices.v; v++)
		{
			for (int u = 0; u < m_nrSlices.u; u++)
			{
				if (m_lookup[w][v][u] != 1) ASSERT(0);
			}
		}
	}
	delete []m_lookup;
#endif
}

template <class TYPE>
void CMGGridCoder<TYPE>::InitExchangeObjectGrid(const ROIDescription &roiDescr)
{
	TRY
	{
		m_pStreamDescr->SetNrQuantBitsGrid(m_nrBits);

		// intialise the levels information
		CIPointUVW nrTiles, levelDiff;
		for (int level = 0; level < m_maxNrLevelsG; level++)
		{
			// compute the number of slices corresponding to sublevel
			CIPointUVW nrSlices = m_nrSlices;
			CIPointUVW nrTiles = roiDescr.GetNrROIs();
			CMGGrid::ComputeNrSlicesROIsLevel(level, m_nrLevelsG, nrSlices, nrTiles, levelDiff);

			// initialize the level information
			GStreamLevel &gridLevel = m_pStreamDescr->GetGridLevel(level);
			gridLevel.Init(nrSlices, nrTiles, levelDiff, -1);

			// initalize the tiles for mesh level
			CreateSubbandTiles(level, roiDescr);
		}
	}
	CATCH_ALL(e)
	{
		throw CMGDataException(e);
	}
	END_CATCH_ALL
}

template <class TYPE>
void CMGGridCoder<TYPE>::StreamHeader()
{
	// preproces the input buffers and compute the thresholds and mean
	// extract coordinates from the mesh
	int dir;
	CMatrix<TYPE>coord[3];
	for (dir = X_GRID; dir <= Z_GRID; dir++)
		coord[dir].Init(m_dimWave.u, m_dimWave.v, m_dimWave.w);

	ExtractCoordinates(coord[X_GRID], coord[Y_GRID], coord[Z_GRID]);

	CIntArray thresholds;  thresholds.SetSize(m_maxNrLevelsG);
	RGCoding *paramRG  = m_pStreamDescr->GetGridCodingParam();

	m_pPrecode->WaveletDec3D(coord[0], coord[1], coord[2]);

	for (dir = X_GRID; dir <= Z_GRID; dir++)
	{
		m_pCodec->PreProcess(m_subbandTiles, coord[dir], m_exchange[dir], paramRG[dir], thresholds, m_nrBits);
		m_pStreamDescr->SetThresholds(dir, thresholds);
	}
}

template <class TYPE>
void CMGGridCoder<TYPE>::EncodeSubband(int subLevel, BYTE update_replace)
{
	// check the threshold first
	GStreamLevel &gridLevel = m_pStreamDescr->GetGridLevel(subLevel);
	if (gridLevel.GetThreshold(X_GRID) < gridLevel.GetStopThreshold(X_GRID) &&
		gridLevel.GetThreshold(Y_GRID) < gridLevel.GetStopThreshold(Y_GRID) &&
		gridLevel.GetThreshold(Z_GRID) < gridLevel.GetStopThreshold(Z_GRID))
		return;

	// tell the parser that the coding of a new grid level starts
	// the parser should initialise a stream for each tile
	m_pParser->OnBeginGridCodingLevel(subLevel);

	const CTileArray &tiles = m_subbandTiles[subLevel];

	// encode the (x, y, z) grid coordinates
	for (int dir = X_GRID;  dir <= Z_GRID; dir++)
	{
		// tell the parser that the coding of a new grid coordinate starts
		m_pParser->OnBeginGridCodingDir(dir, subLevel);

		for (int tileIdx = 0; tileIdx < tiles.GetSize(); tileIdx++)
			EncodeTile(dir, subLevel, tileIdx);

		// tell the parser that the coding of a new grid coordinate ends
		m_pParser->OnEndGridCodingDir(dir);
	}

	// tell the parser that the coding of a new grid level ends
	// the parser should write the counters for each stream, and finally
	// the streams to the output stream
	m_pParser->OnEndGridCodingLevel(subLevel, update_replace);
}

template <class TYPE>
void CMGGridCoder<TYPE>::EncodeTile(int dir, int subLevel, int tileIdx)
{
	GStreamLevel &gridLevel = m_pStreamDescr->GetGridLevel(subLevel);
	const CCubeArray &cubes = m_subbandTiles[subLevel][tileIdx];
	ASSERT(cubes.GetSize() >= 1);

	// tell the parser that the coding of a new grid tile starts
	// the parser should supply a new stream for each tile
	m_pParser->OnBeginGridCodingTile(dir);

	// perform the coding of the Tile
	m_pCodec->WriteROIs(m_pParser->GetStream(), cubes, m_exchange[dir],
		gridLevel.GetThreshold(dir), gridLevel.GetStopThreshold(dir));

	// tell the parser that the coding of a new grid tile ends
	// the parser should update the bit counter for the coded tile
	m_pParser->OnEndGridCodingTile(dir, subLevel, tileIdx);
}

template <class TYPE>
void CMGGridCoder<TYPE>::StreamLevel(int level, BYTE update_replace)
{
	// Each mesh resolution has its corresponding grid resolution, with the exception
	// that the first resultion level mesh (depending on the number of slices for that level)
	// may have several grid resolutions. For efficiency reasons first and second resolution level
	// of the grid are coded as one tile. For this reason we particular cases as follows:

	// TO DO the parser should reset the file pointers
	ASSERT(level < m_maxNrLevelsM);

	int endLevel = m_maxNrLevelsG - m_maxNrLevelsM;
	if (level == 0)
	{
		for (int subLevel = 0; subLevel <= endLevel; subLevel++)
			EncodeSubband(subLevel, update_replace);
	}
	else
	{  
		int subLevel = level + endLevel;
		if (subLevel >= 0)
			EncodeSubband(subLevel, update_replace);
	}
}

template <class TYPE>
void CMGGridCoder<TYPE>::ExtractCoordinates(CMatrix<TYPE> &coordX, CMatrix<TYPE> &coordY, CMatrix<TYPE> &coordZ)
{
	for (int z = 0; z < m_nrSlices.w; z++)
	{
		const GridUV &gridUV = (*m_pGrid)[z];
		TYPE **sliceX = coordX[z];
		TYPE **sliceY = coordY[z];
		TYPE **sliceZ = coordZ[z];

		for (int y = 0; y < m_nrSlices.v; y++)
		{
			const GridU &gridU = gridUV[y];
			TYPE *elemX = sliceX[y];
			TYPE *elemY = sliceY[y];
			TYPE *elemZ = sliceZ[y];

			for (int x  = 0; x < m_nrSlices.u; x++)
			{
				const CFPoint3DT &p = gridU[x];
				*elemX++ = p.x;
				*elemY++ = p.y;
				*elemZ++ = p.z;
			}
		}
	}
}


#endif // !defined(AFX_MGGRIDCODEC_H__85E0A823_97E8_4B6D_8F77_A9974CF76CD7__INCLUDED_)
