/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGGridInterpol.cpp: implementation of the CMGGrid class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGGrid.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

void CMGGrid::InterpolBox()
{
	const CIPointUVW &nrSlices = GetMaxNrSlices();

	int incU = max(nrSlices.u - 1, 1);
	int incV = max(nrSlices.v - 1, 1);
	int incW = max(nrSlices.w - 1, 1);

	// interploate the corner grid lines in the w direction
	if (nrSlices.w > 1)
	{
		for (int v = 0; v < nrSlices.v; v += incV)
		{
			for (int u = 0; u < nrSlices.u; u += incU)
			{
				CFPoint3DT &p1 = m_gridUVW[0][v][u];
				CFPoint3DT &p2 = m_gridUVW[nrSlices.w - 1][v][u];

				CFPoint3DT step = (p2 - p1) / (float)(nrSlices.w - 1);

				for (int w = 1; w < nrSlices.w - 1; w++)
					m_gridUVW[w][v][u] = p1 + step * (float)w;
			}
		}
	}

	// interpolate the grid faces in the v direction
	if (nrSlices.v > 1)
	{
		for (int w = 0; w < nrSlices.w; w++)
		{
			GridUV &gridUV = m_gridUVW[w];

			GridU &gridU1 = gridUV[0];
			GridU &gridU2 = gridUV[nrSlices.v - 1];

			for (int u = 0; u < nrSlices.u; u += incU)
			{
				CFPoint3DT &p1 = gridU1[u];
				CFPoint3DT &p2 = gridU2[u];

				CFPoint3DT step = (p2 - p1) / (float)(nrSlices.v - 1);

				for (int v = 1; v < nrSlices.v - 1; v++)
					gridUV[v][u] = p1 + step * (float)v;
			}
		}
	}

	// interpolate remaining grid points
	if (nrSlices.u > 1)
	{
		for (int w = 0; w < nrSlices.w; w++)
		{
			GridUV &gridUV = m_gridUVW[w];
			for (int v = 0; v < nrSlices.v; v++)
			{
				GridU &gridU = gridUV[v];

				CFPoint3DT &p1 = gridU[0];
				CFPoint3DT &p2 = gridU[nrSlices.u - 1];

				CFPoint3DT step = (p2 - p1) / (float)(nrSlices.u - 1);

				for (int u = 1; u < nrSlices.u - 1; u++)
					gridU[u] = p1 + step * (float)u;
			}
		}
	}
}

bool CMGGrid::SetFilterType(int filterType)
{
	if (filterType != LINEAR_FILTER && filterType != SMOOTH_FILTER)
	{
		ASSERT(0);  return false;
	}

	m_filterType = filterType;
	if (m_detailUVW != 0)
		ComputeDetails();

	return true;
}

bool CMGGrid::SetHierarchicalMode(int hierarchicallevel)
{
	if (hierarchicallevel >= 0 && hierarchicallevel < m_maxNrLevels)
	{
		// initialise the details buffer
		if (m_detailUVW == NULL)
		{
			// allocate the  buffers
			if (!InitDetails())   return false;

			// compute the details
			ComputeDetails();
		}
		DeleteBackup();
	}
	else if (hierarchicallevel == -1)
	{
		// initialise the details buffer
		if (m_detailUVW == NULL)
		{
			// allocate the  buffers
			if (!InitDetails())   return false;

			// compute the details
			ComputeDetails();
		}

		if (m_backupUVW == NULL)
		{
			// allocate the  buffers
			if (!InitBackup())   return false;

			// copy the grid points
			BackupGrid();
		}
	}
	else
	{
		// delete the details buffer
		DeleteDetails();
		DeleteBackup();
	}

	return true;
}

bool CMGGrid::InitDetails()
{
	try
	{
		AllocBuffer(m_detailUVW, m_maxNrSlices.u, m_maxNrSlices.v, m_maxNrSlices.w);
	}
	catch (CMemoryException *e)
	{
		e->ReportError();   return false;
	}

	return true;
}

bool CMGGrid::InitBackup()
{
	try
	{
		AllocBuffer(m_backupUVW, m_maxNrSlices.u, m_maxNrSlices.v, m_maxNrSlices.w);
	}
	catch (CMemoryException *e)
	{
		e->ReportError();   return false;
	}

	return true;
}

void CMGGrid::DeleteDetails()
{
	delete []m_detailUVW; m_detailUVW = NULL;
}

void CMGGrid::DeleteBackup()
{
	delete []m_backupUVW; m_backupUVW = NULL;
}

void CMGGrid::BackupGrid()
{
	for (int w = 0; w < m_maxNrSlices.w; w++)
	{
		const GridUV &gridUV1 = m_gridUVW[w];
		GridUV &gridUV2 = m_backupUVW[w];

		for (int v = 0; v < m_maxNrSlices.v; v++)
		{
			const GridU &gridU1 = gridUV1[v];
			GridU &gridU2 = gridUV2[v];

			for (int u = 0; u < m_maxNrSlices.u; u++)
			{
				gridU2[u] = gridU1[u];
			}
		}
	}
}

void CMGGrid::Distribute(int baseLevel)
{
	if (m_detailUVW == NULL)
	{
		// interpolate without details
		ASSERT(m_backupUVW == NULL);
		Interpolate(baseLevel);
	}
	else if (m_backupUVW != NULL)
	{
		ASSERT(m_detailUVW != NULL && baseLevel == -1);

		// interpolate with details by tracking
		// the changes via the backup buffer
		InterpolateWithDetailBackup();
	}
	else
	{
		// interpolate with details
		InterpolateWithDetail(baseLevel);
	}
}

void CMGGrid::Interpolate(int baseLevel)
{
	for (int level = baseLevel; level < GetMaxNrLevels() - 1; level++)
	{
		if (m_filterType == LINEAR_FILTER)
			InterpolateLinear(level);
		else
			InterpolateSmooth(level);
	}
}

void CMGGrid::InterpolateWithDetail(int baseLevel)
{
	for (int level = baseLevel; level < GetMaxNrLevels() - 1; level++)
	{
		if (m_filterType == LINEAR_FILTER)
		{
			InterpolateLinearWithDetail(level);
			ComputeDetailsLinear(level);
		}
		else
		{
			InterpolateSmoothWithDetail(level);
			ComputeDetailsSmooth(level);
		}
	}
}

void CMGGrid::ComputeDetails()
{
	for (int level = 0; level < GetMaxNrLevels() - 1; level++)
	{
		if (m_filterType == LINEAR_FILTER)
			ComputeDetailsLinear(level);
		else
			ComputeDetailsSmooth(level);
	}
}

void CMGGrid::InterpolateWithDetailBackup()
{
	// compare the current grid with the previous grid
	int level;
	int maxLevel = GetMaxNrLevels() - 1;
	for (level = 0; level < maxLevel; level++)
	{
		if (CheckModifiedLevel(level))
			break;
	}

	// from this lowestLevel on interpolate, check for modified and backup modified values
	for (; level < maxLevel; level++)
	{
		if (m_filterType == LINEAR_FILTER)
			InterpolateLinearWithDetailBackup(level);
		else
			InterpolateSmoothWithDetailBackup(level);

		// update the details
		if (m_filterType == LINEAR_FILTER)
			ComputeDetailsLinear(level);
		else
			ComputeDetailsSmooth(level);
	}
}

void CMGGrid::ComputeHomogeneousSteps(CIPointUVW &step, CIPointUVW &step_2, CIPointUVW &levelDiff, int baseLevel)
{
	const CIPointUVW &nrLevels = GetNrLevelsUVW();

	levelDiff.u = nrLevels.u - GetMaxNrLevels() + baseLevel;
	levelDiff.v = nrLevels.v - GetMaxNrLevels() + baseLevel;
	levelDiff.w = nrLevels.w - GetMaxNrLevels() + baseLevel;

	int maxLevel = GetMaxNrLevels() - 1;
	if (levelDiff.u > 0)
	{
		step.u = 1 << (maxLevel - baseLevel);    step_2.u = step.u >> 1;
	}
	else
	{
		step.u = 1 << (nrLevels.u - 1);      step_2.u = max(step.u >> 1, 1);
	}

	if (levelDiff.v > 0)
	{
		step.v = 1 << (maxLevel - baseLevel);    step_2.v = step.v >> 1;
	}
	else
	{
		step.v = 1 << (nrLevels.v - 1);      step_2.v = max(step.v >> 1, 1);
	}

	if (levelDiff.w > 0)
	{
		step.w = 1 << (maxLevel - baseLevel);    step_2.w = step.w >> 1;
	}
	else
	{
		step.w = 1 << (nrLevels.w - 1);      step_2.w = max(step.w >> 1, 1);
	}
}

void CMGGrid::ComputeHomogeneousSteps(CIPointUVW &step, int baseLevel)
{
	CIPointUVW levelDiff;
	const CIPointUVW &nrLevels = GetNrLevelsUVW();

	levelDiff.u = nrLevels.u - GetMaxNrLevels() + baseLevel;
	levelDiff.v = nrLevels.v - GetMaxNrLevels() + baseLevel;
	levelDiff.w = nrLevels.w - GetMaxNrLevels() + baseLevel;

	int maxLevel = GetMaxNrLevels() - 1;
	if (levelDiff.u > 0)
	{
		step.u = 1 << (maxLevel - baseLevel);
	}
	else
	{
		step.u = 1 << (nrLevels.u - 1);
	}

	if (levelDiff.v > 0)
	{
		step.v = 1 << (maxLevel - baseLevel);
	}
	else
	{
		step.v = 1 << (nrLevels.v - 1);
	}

	if (levelDiff.w > 0)
	{
		step.w = 1 << (maxLevel - baseLevel);
	}
	else
	{
		step.w = 1 << (nrLevels.w - 1);
	}
}

bool CMGGrid::CheckModifiedLevel(int level)
{
	// compute the steps between the grid lines of same resolution level
	CIPointUVW step;
	ComputeHomogeneousSteps(step, level);

	for (int w = 0; w < m_maxNrSlices.w; w += step.u)
	{
		const GridUV &gridUV1 = m_gridUVW[w];
		GridUV &gridUV2 = m_backupUVW[w];

		for (int v = 0; v < m_maxNrSlices.v; v += step.v)
		{
			const GridU &gridU1 = gridUV1[v];
			GridU &gridU2 = gridUV2[v];

			for (int u = 0; u < m_maxNrSlices.u; u += step.u)
			{
				if (gridU2[u] != gridU1[u])
					return true;
			}
		}
	}

	return false;
}