/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGMesh.cpp: implementation of the CMGMesh class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGMesh.h"
#include "MGGrid.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif



//////////////////////////////////////////////////////////////////////
// CSurfNodeT
//////////////////////////////////////////////////////////////////////

CSurfNodeT::CSurfNodeT(const CIPointUVW &pos, int dir, int nrLevels, int index, int level, BYTE layer, float offset)
{
	SetOffset(offset);
	SetLevel(level);
	m_layer = 0;
	m_pos = pos;
	m_count = dir;	m_index = index;		m_flag = 0;
	m_upList = m_downList = m_hierarchyNode = NULL;

	typedef CSurfNodeT *	LPCSurfNodeT;
	m_prevNode1 = new LPCSurfNodeT[nrLevels];
	m_prevNode2 = new LPCSurfNodeT[nrLevels];
	m_nextNode1 = new LPCSurfNodeT[nrLevels];
	m_nextNode2 = new LPCSurfNodeT[nrLevels];

	for (int l = 0; l < nrLevels; l++)
	{
		m_prevNode1[l] = NULL;	m_prevNode2[l] = NULL;
		m_nextNode1[l] = NULL;	m_nextNode2[l] = NULL;
	}
}


CSurfNodeT::~CSurfNodeT()
{
	delete []m_prevNode1;
	delete []m_prevNode2;
	delete []m_nextNode1;
	delete []m_nextNode2;
}

//////////////////////////////////////////////////////////////////////
// CSurfNodeList
//////////////////////////////////////////////////////////////////////

void CSurfNodeList::RemoveAll()
{
	while (m_pList != NULL)
	{
		CSurfNodeT *pSN = m_pList;
		m_pList = m_pList->GetDownList();
		delete pSN;
	}
}

CSurfNodeT *CSurfNodeList::AddNode(const CIPointUVW &pos, int dir, int nrLevels, int index, int level, BYTE layer, float offset)
{
	// check that the vertex doesn't exist yet
	ASSERT(GetNode(pos, dir) == NULL);

	CSurfNodeT *pSN = new CSurfNodeT(pos, dir, nrLevels, index, level, layer, offset);

	if (m_pList)
	{
		ASSERT(m_pList->GetUpList() == NULL);
		m_pList->SetUpList(pSN);
	}
	// pSN->m_upList is already NULL
	pSN->SetDownList(m_pList);
	m_pList = pSN;

	return pSN;
}

CSurfNodeT* CSurfNodeList::GetNode(const CIPointUVW &pos, int dir)
{
	register CSurfNodeT *pSN = m_pList;
	while (pSN != NULL)
	{
		if (pSN->GetDir() == dir && pSN->GetPos().w == pos.w)
		{
			ASSERT(pSN->GetPos().u == pos.u && pSN->GetPos().v == pos.v);
			return pSN;
		}

		pSN = pSN->GetDownList();
	}

	// node not found
	return NULL;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMGMesh::CMGMesh()
{
	m_grid = NULL;

	m_level = -1;
	m_nrLayers = 0;
	m_defaultSizeROI    = 2;
	m_defaultNoROIs		= 1;
	m_nrRefBitsVert     = 0;
	m_refGridCR         = 0.0f;
	m_refGridQuant      = 0;
	m_refGridMAXAD		= 0.1f;
	m_bInterleaveRefine = false;

	m_hasConnectivityInfo = false;	
	m_hasGridInfo = false;	
	m_hasRefineInfo = false;		
	m_hasRepositionInfo = false;
	m_bSingleLayered = false;
	m_meshType = GENERIC_MESH;
	m_bSameConnectivity = false;
	m_bAutoResolution = false;
	m_offsetAmp = 0;

	m_selDir = -1;
	m_selPos = -1;
	m_localROI.SetDefault();
	m_globalROI.SetDefault();
	m_posVertMin = m_posVertMax = 0;

	m_countLink       = 0;
	m_countOffset     = 0;
	m_nrVertexLinks   = 0;
	m_nrVertexOffsets = 0;
	m_totalNrVertices = 0;
	m_nrTriangles     = 0;
	m_nrVerticesROI   = 0;

	m_vertexLink    = NULL;
	m_vertexOffset  = NULL;
	m_vertices      = NULL;
	m_nodeStack     = NULL;
	m_nrVertices    = NULL;
	m_startVertsROI = NULL;

	m_pMGIN         = NULL;
	m_pMGOUT        = NULL;
	m_pParser       = NULL;
	m_pStreamDescr  = NULL;
}

CMGMesh::~CMGMesh()
{
	DeleteContent();
}

bool CMGMesh::Init(CMGGrid *grid, CFPoint3DT *vertices, CString & error)
{
	ASSERT(grid != 0);
	typedef CSurfNodeList * LPSurfNodeList;

	// clean the buffers
	DeleteContent();

	// volume dimension
	m_grid = grid;
	m_pParser = NULL;
	m_vertices = vertices;
	m_nrLayers = 1;
	m_nrSlices = grid->GetMaxNrSlices();
	m_maxNrLevels = grid->GetMaxNrLevels();

	if (m_nrSlices.w == 1)
		SetSingleLayerGrid(true);

	try
	{
		m_polyLevels.SetSize(m_maxNrLevels);
		m_nrVertices = new int[m_maxNrLevels];
		memset(m_nrVertices, 0, m_maxNrLevels * sizeof(int));

		// allocate the x vector and initialise to NULL
		m_nodeStack = new LPSurfNodeList[m_nrSlices.u];
		memset(m_nodeStack, 0, m_nrSlices.u * sizeof (LPSurfNodeList));
	}
	catch (CMemoryException *)
	{
		error = CMGErrMsg::GetMessage(CMGErrMsg::MEMORY_EXCEPTION);
		DeleteContent();		return false;
	}

	return true;
}

bool CMGMesh::Init(CMGGrid *grid, CMGStreamParser *parser, CFPoint3DT *vertices, CString & error)
{
	ASSERT(grid != 0);
	typedef CSurfNodeList * LPSurfNodeList;

	// clean the buffers
	DeleteContent();

	// volume dimension
	m_grid = grid;
	m_pParser = parser;
	m_vertices = vertices;
	m_pStreamDescr = &m_pParser->m_streamDescr;

	m_nrSlices = m_pStreamDescr->GetMaxNrSlices();
	ASSERT(m_nrSlices == grid->GetMaxNrSlices());

	m_maxNrLevels = parser->GetMaxNrLevelsMesh();
	ASSERT(m_maxNrLevels == grid->GetMaxNrLevels());

	m_maxNrROIs = m_pStreamDescr->GetNrROIs();
	m_nrLayers = m_pStreamDescr->GetNrLayers();

	if (m_nrSlices.w == 1)
		SetSingleLayerGrid(true);

	try
	{
		m_polyLevels.SetSize(m_maxNrLevels);
		m_nrVertices = new int[m_maxNrLevels];
		memset(m_nrVertices, 0, m_maxNrLevels * sizeof(int));

		// retrieve the refine bits
		m_nrRefBitsVert = m_pStreamDescr->GetNrRefineBitsVertex();
		m_bInterleaveRefine = m_pStreamDescr->InterleaveRefine();

		// mesh parameters
		SetAutoResolution(m_pStreamDescr->IsAutoResolution());
		SetMeshType(m_pStreamDescr->GetMeshType());
		SetCyclicMode(m_pStreamDescr->GetCyclicMode());

		// allocate the x vector and initialise to NULL
		m_nodeStack = new LPSurfNodeList[m_nrSlices.u];
		memset(m_nodeStack, 0, m_nrSlices.u * sizeof (LPSurfNodeList));

		// allocate the array holding the starting vertices for each patch
		ASSERT(m_startVertsROI == NULL);
		m_startVertsROI = new CPtrList**[m_maxNrLevels];

		for (int level = 0; level < m_maxNrLevels; level++)
		{
			m_startVertsROI[level] = new CPtrList*[m_nrLayers];
			memset(m_startVertsROI[level], 0, m_nrLayers * sizeof(void *));
		}
	}
	catch (CMemoryException *)
	{
		error = CMGErrMsg::GetMessage(CMGErrMsg::MEMORY_EXCEPTION);
		DeleteContent();		return false;
	}

	return true;
}

void CMGMesh::DeleteContent()
{
	m_polyLevels.RemoveAll();
	delete []m_nrVertices;	m_nrVertices = NULL;

	if (m_nodeStack)
	{
		for (int k = 0; k < m_nrSlices.u; k++)
		{
			CSurfNodeList *pList = m_nodeStack[k];
			delete []pList;
		}

		delete []m_nodeStack; m_nodeStack = NULL;
	}

	// reset the node counter
	m_nrTriangles = 0;
}

/**********************************************************************************/

CSurfNodeT *CMGMesh::GetNode(const CIPointUVW &pos, int dir)
{
	CSurfNodeList *pNodeList = m_nodeStack[pos.u];
	if (pNodeList == NULL)	return NULL;

	return pNodeList[pos.v].GetNode(pos, dir);
}

CSurfNodeT *CMGMesh::GetAddNode(CSurfNodeList *pNodeList, const CIPointUVW &pos, int dir, int level, BYTE layer)
{
	ASSERT(dir >= BACK_FACE);
	CSurfNodeT *pSN = pNodeList[pos.v].GetNode(pos, dir);
	if (!pSN)
	{
		// update the vertex counter
		++m_totalNrVertices;

		// add new entry
		pSN = pNodeList[pos.v].AddNode(pos, dir, m_maxNrLevels, m_totalNrVertices - 1, level, layer, 0.0f);
	}
	ASSERT(pSN->GetDir() == dir);

	return pSN;
}

CSurfNodeT *CMGMesh::GetAddNode(CSurfNodeList *pNodeList, const CIPointUVW &pos, int dir, int level, BYTE layer, float offset)
{
	ASSERT(dir >= BACK_FACE);
	CSurfNodeT *pSN = pNodeList[pos.v].GetNode(pos, dir);
	if (!pSN)
	{
		// update the vertex counter
		++m_totalNrVertices;

		// add new entry
		pSN = pNodeList[pos.v].AddNode(pos, dir, m_maxNrLevels, m_totalNrVertices - 1, level, layer, offset);
		UpdateVertexCoord(level, pSN);
	}
	ASSERT(pSN->GetDir() == dir);

	return pSN;
}

CSurfNodeList *CMGMesh::GetColumn(int u)
{
	ASSERT(u >= 0 && u < m_nrSlices.u);
	if (m_nodeStack[u] == NULL)
	{
		// voxel y can have point y and (y+1)
		m_nodeStack[u] = new CSurfNodeList[m_nrSlices.v];
	}

	return m_nodeStack[u];
}

CSurfNodeT *CMGMesh::StoreVertex(CIPointUVW &pos3D, int level, BYTE layer, int dir)
{
	// allocate row if it does not exist
	CSurfNodeList *pList = GetColumn(pos3D.u);
	if (!pList) return NULL;

	// allocate the Link cell
	return GetAddNode(pList, pos3D, dir, level, layer);
}

CSurfNodeT *CMGMesh::StoreVertex(CIPointUVW &pos3D, int level, BYTE layer, int dir, float offset)
{
	// allocate row if it does not exist
	CSurfNodeList *pList = GetColumn(pos3D.u);
	if (!pList) return NULL;

	// allocate the Link cell
	return GetAddNode(pList, pos3D, dir, level, layer, offset);
}

/**********************************************************************************/

void CMGMesh::ComputeNormal(const CFPoint3DT &l, const CFPoint3DT &r, const CFPoint3DT &d, const CFPoint3DT &u, CFPoint3DT &res, float sgn)
{
	// Calculate the Normal direction

	//	N      = -(Xu x Xv)/(||Xu x Xv||)
	//	a x b  =  i(aybz-azby)+j(azbx-axbz)+k(axby-aybx)

	CFPoint3DT a, b;
	a.x = r.x - l.x;		b.x = u.x - d.x;
	a.y = r.y - l.y;		b.y = u.y - d.y;
	a.z = r.z - l.z;		b.z = u.z - d.z;

	CFPoint3DT c;
	c.x = a.y * b.z - a.z * b.y;
	c.y = a.z * b.x - a.x * b.z;
	c.z = a.x * b.y - a.y * b.x;

	float fTemp = sgn * (float)sqrt(c.x * c.x + c.y * c.y + c.z * c.z);

	if (fTemp == 0.0f) fTemp = 0.00001f;

	res.x = c.x / fTemp;		res.y = c.y / fTemp;		res.z = c.z / fTemp;
}

void CMGMesh::ComputeSmoothNormals(CFPoint3DT *normals, int level)
{
	if (IsEmpty() || normals == NULL) return;

	for (int u = 0; u < m_nrSlices.u; u++)
	{
		CSurfNodeList *pList = m_nodeStack[u];
		if (pList == NULL) continue;

		for (int v = 0; v < m_nrSlices.v; v++)
		{
			CSurfNodeT *pSN = pList[v].GetFirstNode();
			while (pSN != NULL)
			{
				if (pSN->GetLevel() <= level)
				{
					CSurfNodeT *pSNLeft = pSN->GetPrevNode1(level);
					if (pSNLeft == NULL)		pSNLeft = pSN;

					CSurfNodeT *pSNRight = pSN->GetNextNode1(level);
					if (pSNRight == NULL)		pSNRight = pSN;

					CSurfNodeT *pSNDown = pSN->GetPrevNode2(level);
					if (pSNDown == NULL)		pSNDown = pSN;

					CSurfNodeT *pSNUp = pSN->GetNextNode2(level);
					if (pSNUp == NULL)			pSNUp = pSN;

					ComputeNormal(m_vertices[pSNLeft->GetIndex()], m_vertices[pSNRight->GetIndex()],
						m_vertices[pSNDown->GetIndex()], m_vertices[pSNUp->GetIndex()], normals[pSN->GetIndex()], 1);
					//        GetOrthoNormal(pSN->GetNormal(), pSN->GetDir());
				}
				pSN = pSN->GetDownList();
			}
		}
	}
}

void CMGMesh::UpdateVertexCoordSingleLayer(int level, CSurfNodeT *pSN)
{
	int index = pSN->GetIndex();
	ASSERT(index >= 0 && index < m_totalNrVertices);
	const CIPointUVW &pos = pSN->GetPos();

	m_vertices[index] = (*m_grid)[pos.w][pos.v][pos.u];

	if (pSN->GetOffset() != 0.5f && GetOffsetAmplitude() != 0)
	{
		// get the four neighbour grid positions and compute the normal vector
		const CIPointUVW *posLeft = &pos;
		CSurfNodeT *pSNLeft = pSN->GetPrevNode1(level);
		if (pSNLeft != NULL)		posLeft = &pSNLeft->GetPos();

		const CIPointUVW *posRight = &pos;
		CSurfNodeT *pSNRight = pSN->GetNextNode1(level);
		if (pSNRight != NULL)		posRight = &pSNRight->GetPos();

		const CIPointUVW *posDown = &pos;
		CSurfNodeT *pSNDown = pSN->GetPrevNode2(level);
		if (pSNDown != NULL)		posDown = &pSNDown->GetPos();

		const CIPointUVW *posUp = &pos;
		CSurfNodeT *pSNUp = pSN->GetNextNode2(level);
		if (pSNUp != NULL)			posUp = &pSNUp->GetPos();

		CFPoint3DT n;
		ComputeNormal((*m_grid)[posLeft->w][posLeft->v][posLeft->u], (*m_grid)[posRight->w][posRight->v][posRight->u],
			(*m_grid)[posDown->w][posDown->v][posDown->u], (*m_grid)[posUp->w][posUp->v][posUp->u], n, 1.0f);

		m_vertices[index] += n * ((float)GetOffsetAmplitude() * (pSN->GetOffset() - 0.5f));
	}
}

void CMGMesh::UpdateVertexCoord(CSurfNodeT *pSN)
{
	if (IsSingleLayerGrid())
	{
		UpdateVertexCoordSingleLayer(GetMaxNrLevels() - 1, pSN);
	}
	else
	{
		int index = pSN->GetIndex();
		ASSERT(index >= 0 && index < m_totalNrVertices);
		const CIPointUVW &pos = pSN->GetPos();

		switch (pSN->GetDir())
		{
		case BACK_FACE:		GetBackPos	(pos, m_vertices[index], pSN->GetOffset());	break;
		case FRONT_FACE:	GetFrontPos	(pos, m_vertices[index], pSN->GetOffset());	break;
		case LEFT_FACE:		GetLeftPos	(pos, m_vertices[index], pSN->GetOffset());	break;
		case RIGHT_FACE:	GetRightPos	(pos, m_vertices[index], pSN->GetOffset());	break;
		case BOTTOM_FACE:	GetBottomPos(pos, m_vertices[index], pSN->GetOffset());	break;
		case TOP_FACE:		GetTopPos		(pos, m_vertices[index], pSN->GetOffset());	break;
		}
	}
}

void CMGMesh::UpdateVertexCoord(int level, CSurfNodeT *pSN)
{
	if (IsSingleLayerGrid())
	{
		UpdateVertexCoordSingleLayer(level, pSN);
	}
	else
	{
		int index = pSN->GetIndex();
		ASSERT(index >= 0 && index < m_totalNrVertices);
		const CIPointUVW &pos = pSN->GetPos();

		switch (pSN->GetDir())
		{
		case BACK_FACE:		GetBackPos	(pos, m_vertices[index], pSN->GetOffset(), level);	break;
		case FRONT_FACE:	GetFrontPos	(pos, m_vertices[index], pSN->GetOffset(), level);	break;
		case LEFT_FACE:		GetLeftPos	(pos, m_vertices[index], pSN->GetOffset(), level);	break;
		case RIGHT_FACE:	GetRightPos	(pos, m_vertices[index], pSN->GetOffset(), level);	break;
		case BOTTOM_FACE:	GetBottomPos(pos, m_vertices[index], pSN->GetOffset(), level);	break;
		case TOP_FACE:		GetTopPos		(pos, m_vertices[index], pSN->GetOffset(), level);	break;
		}
	}
}

int	CMGMesh::GetNrVertices(int level)
{
	ASSERT(level >= 0 && level < m_maxNrLevels);
	return m_nrVertices[level];
}

DynaSurfPolyT &CMGMesh::GetTrianglesBuffer(int level)
{
	ASSERT(level >= 0 && level < m_maxNrLevels);
	return m_polyLevels[level];
}

void CMGMesh::GetTriangles(int *f, int level) 
{
	//  FILE *fp = fopen("DumpTriangles.txt", "wt");

	DynaSurfPolyT &poly = m_polyLevels[level];
	register int levelSz = poly.GetLevelSize();
	register int count = poly.GetSize();
	register int dynaLevel = -1;
	register int pos = levelSz;
	int counter = GetTotalNrVertices();

	register CSurfPolyT *polyBuf = NULL;
	while (count--)
	{
		if (pos == levelSz)
		{
			pos = 0; ++dynaLevel;
			polyBuf = poly[dynaLevel].GetData();
		}

		// we should find at least a triangle
		CSurfNodeT *pSN1 = polyBuf->GetRef1();
		ASSERT (pSN1 != NULL);

		CSurfNodeT *pSN2 = polyBuf->GetRef2();
		ASSERT (pSN2 != NULL);

		CSurfNodeT *pSN3 = polyBuf->GetRef3();
		ASSERT (pSN3 != NULL);

		ASSERT(pSN1->GetIndex() >= 0 && pSN1->GetIndex() < counter);
		ASSERT(pSN2->GetIndex() >= 0 && pSN2->GetIndex() < counter);
		ASSERT(pSN3->GetIndex() >= 0 && pSN3->GetIndex() < counter);

		*f++ = pSN1->GetIndex();    // fprintf(fp, "%d\n", pSN1->GetIndex());
		*f++ = pSN2->GetIndex();    // fprintf(fp, "%d\n", pSN2->GetIndex());
		*f++ = pSN3->GetIndex();    // fprintf(fp, "%d\n", pSN3->GetIndex());
		*f++ = -1;                  // fprintf(fp, "%d\n", -1);

		++polyBuf; ++pos;	
	}

	//  fclose(fp);
}

void CMGMesh::GetTextureIndices(int *ti, int level) 
{
	DynaSurfPolyT &poly = m_polyLevels[level];
	register int levelSz = poly.GetLevelSize();
	register int count = poly.GetSize();
	register int dynaLevel = -1;
	register int pos = levelSz;

	register CSurfPolyT *polyBuf = NULL;
	while (count--)
	{
		if (pos == levelSz)
		{
			pos = 0; ++dynaLevel;
			polyBuf = poly[dynaLevel].GetData();
		}

		// we should find at least a triangle
		CSurfNodeT *pSN1 = polyBuf->GetRef1();
		ASSERT (pSN1 != NULL);

		CSurfNodeT *pSN2 = polyBuf->GetRef2();
		ASSERT (pSN2 != NULL);

		CSurfNodeT *pSN3 = polyBuf->GetRef3();
		ASSERT (pSN3 != NULL);

		const CIPointUVW &pos1 = pSN1->GetPos();
		const CIPointUVW &pos2 = pSN2->GetPos();
		const CIPointUVW &pos3 = pSN3->GetPos();

		*ti++ = pos1.u + pos1.v * m_nrSlices.u;
		*ti++ = pos2.u + pos2.v * m_nrSlices.u;
		*ti++ = pos3.u + pos3.v * m_nrSlices.u;
		*ti++ = -1;

		++polyBuf; ++pos;	
	}
}

void CMGMesh::GetTextureCoordinates(float *tc) 
{
	float stepU = 1.0f / (m_nrSlices.u - 1);
	float stepV = 1.0f / (m_nrSlices.v - 1);

	for (int v = 0; v < m_nrSlices.v; v++)
	{
		float v_coord = (float)v * stepV;
		for (int u = 0; u < m_nrSlices.u; u++)
		{
			float u_coord = (float)u * stepU;
			*tc++ = u_coord;
			*tc++ = v_coord;
		}
	}
}

void CMGMesh::GetVertexOffset(float *offset)
{
	m_vertexOffset = offset;
	if (offset == NULL)	return;

	// update the list of offsets of the VRML node
	// such that they are in the same order as the allocated vertices
	int count = 0;
	for (int u = 0; u < m_nrSlices.u; u++)
	{
		CSurfNodeList *pList = m_nodeStack[u];
		if (pList == NULL) continue;

		for (int v = 0; v < m_nrSlices.v; v++)
		{
			register CSurfNodeT *pSN = pList[v].GetFirstNode();
			while (pSN != NULL)
			{
				m_vertexOffset[count++] = pSN->GetOffset();
				pSN = pSN->GetDownList();
			}
		}
	}
}

void CMGMesh::GetVertexNormals(CFPoint3DT *normals, int level)
{
	ComputeSmoothNormals(normals, level);
}

void CMGMesh::Update(float *verts, int nrPts, CString &error)
{
	int count = 0;
	for (int u = 0; u < m_nrSlices.u; u++)
	{
		CSurfNodeList *pList = m_nodeStack[u];
		if (pList == NULL) continue;

		for (int v = 0; v < m_nrSlices.v; v++)
		{
			register CSurfNodeT *pSN = pList[v].GetFirstNode();
			while (pSN != NULL)
			{
				pSN->SetOffset(verts[count]);
				UpdateVertexCoord(pSN);
				pSN = pSN->GetDownList();

				if (++count == nrPts)	return;
			}
		}
	}
}

void CMGMesh::ResetConnectivity()
{
	if (m_nodeStack)
	{
		for (int k = 0; k < m_nrSlices.u; k++)
		{
			CSurfNodeList *pList = m_nodeStack[k];
			delete []pList;		m_nodeStack[k] = NULL;
		}
	}

	for (int k = 0; k < GetMaxNrLevels(); k++)
		m_nrVertices[k] = 0;
}

