/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGMeshCodec.cpp: implementation of the CMGMesh class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGMesh.h"
#include "MGGrid.h"
#include "Globals.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


/**********************************************************************************/

CSurfNodeT *CMGMesh::CheckStartingPoint(int level, BYTE layer)
{
	if (m_selPos == -1 || m_selDir == -1)
		throw CMGDataException("No starting point selected");

	CSurfNodeT *pSNStart = GetNode(m_selPos, m_selDir);
	if (pSNStart == NULL || pSNStart->GetLevel() > level)
		throw CMGDataException("Invalid starting point");

	if (!InROI(pSNStart, level, layer))
		throw CMGDataException("Starting point not in active box");

	ASSERT(pSNStart->GetPos() == m_selPos);
	ASSERT(pSNStart->GetDir() == m_selDir);

	if (pSNStart->IsVisited() != 0)
		throw CMGDataException("Starting point already streamed");

	return pSNStart;
}

bool CMGMesh::IsVertexInROI(CSurfNodeT *pSN, int level, BYTE layer)
{
	return (pSN->GetLevel() <= level && pSN->GetLayer() == layer &&
			pSN->GetWPos() >= m_posVertMin.w && pSN->GetWPos() <= m_posVertMax.w &&
			InROI(pSN, level));
}

bool CMGMesh::InROI(CSurfNodeT *pSN, int level, BYTE layer)
{
	return (pSN->GetLevel() <= level && pSN->GetLayer() == layer &&
			InROI(pSN, level));
}

bool CMGMesh::InROI(CSurfNodeT *pSN, int level)
{
	CIPointUVW pos = pSN->GetPos();
	FindPosNodeGlobal(pos, pSN->GetDir(), level);

	return (pos.u >= m_globalROI.u && pos.u < m_globalROI.u + m_globalROI.width &&
			pos.v >= m_globalROI.v && pos.v < m_globalROI.v + m_globalROI.height &&
			pos.w >= m_globalROI.w && pos.w < m_globalROI.w + m_globalROI.depth);
}

bool CMGMesh::IsROIEmpty(int level, BYTE layer)
{
#ifdef _DEBUG
	bool bEmpty = true;
#endif

	for (int x = m_posVertMin.u; x <= m_posVertMax.u; x++)
	{
		CSurfNodeList *pList = m_nodeStack[x];
		if (pList == NULL) continue;

		for (int y = m_posVertMin.v; y <= m_posVertMax.v; y++)
		{
			CSurfNodeT *pSN = pList[y].GetFirstNode();
			while (pSN != NULL)
			{
#ifdef _DEBUG
				if (IsVertexInROI(pSN, level, layer))
				{
					ASSERT (pSN->IsVisited() == 0);
					bEmpty = false;
				}
#else
				if (IsVertexInROI(pSN, level, layer))
					return false;
#endif
				pSN = pSN->GetDownList();
			}
		}
	}

#ifdef _DEBUG
	return bEmpty;
#else
	return true;
#endif
}

bool CMGMesh::IsROIStreamed(int level, BYTE layer)
{
	for (int x = m_posVertMin.u; x <= m_posVertMax.u; x++)
	{
		CSurfNodeList *pList = m_nodeStack[x];
		if (pList == NULL) continue;

		for (int y = m_posVertMin.v; y <= m_posVertMax.v; y++)
		{
			CSurfNodeT *pSN = pList[y].GetFirstNode();
			while (pSN != NULL)
			{
				if (pSN->IsVisited() == 0 && IsVertexInROI(pSN, level, layer))
					return false;

				pSN = pSN->GetDownList();
			}
		}
	}

	return true;
}

/**********************************************************************************/

void CMGMesh::GetNextStartingPoint(int level, BYTE layer)
{
	int x;
	m_selPos = -1;    m_selDir = -1;

	// find that vertex in the box, which has no incoming connectivity
	for (x = m_posVertMin.u; x <= m_posVertMax.u; x++)
	{
		CSurfNodeList *pList = m_nodeStack[x];
		if (pList == NULL) continue;

		for (int y = m_posVertMin.v; y <= m_posVertMax.v; y++)
		{
			CSurfNodeT *pSN = pList[y].GetFirstNode();
			while (pSN != NULL)
			{
				if (pSN->IsVisited() == 0 && IsVertexInROI(pSN, level, layer))
				{
					CSurfNodeT *pSNPrev1 = pSN->GetPrevNode1(level);
					CSurfNodeT *pSNPrev2 = pSN->GetPrevNode2(level);

					if ((pSNPrev1 == NULL || !InROI(pSNPrev1, level, layer)) &&
						(pSNPrev2 == NULL || !InROI(pSNPrev2, level, layer)))
					{
						m_selDir = pSN->GetDir();   m_selPos = pSN->GetPos();
						return;
					}
				}

				pSN = pSN->GetDownList();
			}
		}
	}

	for (x = m_posVertMin.u; x <= m_posVertMax.u; x++)
	{
		CSurfNodeList *pList = m_nodeStack[x];
		if (pList == NULL) continue;

		for (int y = m_posVertMin.v; y <= m_posVertMax.v; y++)
		{
			CSurfNodeT *pSN = pList[y].GetFirstNode();
			while (pSN != NULL)
			{
				if (pSN->IsVisited() == 0 && IsVertexInROI(pSN, level, layer))
				{
					CSurfNodeT *pSNPrev1 = pSN->GetPrevNode1(level);
					TrackAppropriateEdge(pSN, true, "Rule2 Link1", level, layer);
					ASSERT(pSN->IsVisited() == 0);
					return;
				}

				pSN = pSN->GetDownList();
			}
		}
	}
}

CSurfNodeT *CMGMesh::GetFirstPrevious(CSurfNodeT *pSN, bool &bLink1, int level, BYTE layer)
{
	CSurfNodeT *pSNStart = pSN, *pSNPrev = NULL;
	do
	{
		if (bLink1)
			pSNPrev = pSN->GetPrevNode1(level);
		else
			pSNPrev = pSN->GetPrevNode2(level);

		if (pSNPrev == NULL || !InROI(pSNPrev, level, layer))
			return pSN;

		// link 1 and link2 may switch depending on the
		// border face and scanning direction
		bLink1 = pSNPrev->GetNextNode1(level) == pSN;
		pSN = pSNPrev;
	}
	while (pSNPrev != pSNStart);

	return NULL;
}

bool CMGMesh::TrackAppropriateEdge(CSurfNodeT *pSN, bool bLink1, LPCTSTR name, int level, BYTE layer)
{
	// get till the edge of Link1/Link2
	CSurfNodeT *pSNFirst = GetFirstPrevious(pSN, bLink1, level, layer);
	if (pSNFirst == NULL)   pSNFirst = pSN;

	// get till the edge on Link2/Link1
	pSN = pSNFirst;
	bLink1 = !bLink1;

	pSNFirst = GetFirstPrevious(pSN, bLink1, level, layer);
	if (pSNFirst == NULL)   pSNFirst = pSN;

	// the connectivity at the edge on Link1/Link2 should be towards inside the box
	bLink1 = !bLink1;
	CSurfNodeT *pSNPrev = bLink1 ? pSNFirst->GetPrevNode1(level) :
		pSNFirst->GetPrevNode2(level);

	if (pSNPrev != NULL && InROI(pSNPrev, level, layer))
	{
		pSN = pSNFirst;

		pSNFirst = GetFirstPrevious(pSN, bLink1, level, layer);
		if (pSNFirst == NULL)   pSNFirst = pSN;
	}

	m_selDir = pSNFirst->GetDir();   m_selPos = pSNFirst->GetPos();
	return true;
}

/**********************************************************************************/

void CMGMesh::CacheStartPoint(CSurfNodeT *pSN, int level, BYTE layer, int roiIdx)
{
	m_startVertsROI[level][layer][roiIdx].AddTail(pSN);
}
/**********************************************************************************/

void CMGMesh::AddNodesToList(CSurfNodeT *pSNNext1, CSurfNodeT *pSNNext2, int level, bool bCount)
{
	// mark the nodes and put them in the list
	if (pSNNext1 != NULL)
	{
		if (pSNNext1->IsVisited() == 0)
		{
			if (bCount) ++m_nrVerticesROI;
			pSNNext1->SetVisited();
			m_nodeList.AddTail(pSNNext1);
		}
	}

	if (pSNNext2 != NULL)
	{
		if (pSNNext2->IsVisited() == 0)
		{
			if (bCount) ++m_nrVerticesROI;
			pSNNext2->SetVisited();
			m_nodeList.AddTail(pSNNext2);
		}
	}
}

void CMGMesh::AddNodesToListBox(CSurfNodeT *pSNNext1, CSurfNodeT *pSNNext2, int level, BYTE layer, bool bCount)
{
	// mark the nodes and put them in the list
	if (pSNNext1 != NULL)
	{
		if (InROI(pSNNext1, level, layer))
		{
			if (pSNNext1->IsVisited() == 0)
			{
				if (bCount) ++m_nrVerticesROI;
				pSNNext1->SetVisited();
				m_nodeList.AddTail(pSNNext1);
			}
		}
	}

	if (pSNNext2 != NULL)
	{
		if (InROI(pSNNext2, level, layer))
		{
			if (pSNNext2->IsVisited() == 0)
			{
				if (bCount) ++m_nrVerticesROI;
				pSNNext2->SetVisited();
				m_nodeList.AddTail(pSNNext2);
			}
		}
	}
}

void CMGMesh::AddStartNodeBox(CSurfNodeT *pSNStart, bool bCount)
{
	ASSERT (pSNStart != NULL);
	ASSERT (m_nodeList.GetCount() == 0);

	ASSERT(pSNStart->IsVisited() == 0);

	if (bCount) ++m_nrVerticesROI;

	pSNStart->SetVisited();
	m_nodeList.AddHead(pSNStart);
}

/**********************************************************************************/

CSurfNodeT * CMGMesh::ResetVisitedVertices(int level)
{
	// reset the node flags
	CSurfNodeT *pSNStart = NULL;
	for (int x = 0; x < m_nrSlices.u; x++)
	{
		// check row
		CSurfNodeList *pList = m_nodeStack[x];
		if (pList == NULL) continue;

		// check colums
		for (int y = 0; y < m_nrSlices.v; y++)
		{
			CSurfNodeT *pSN = pList[y].GetFirstNode();
			while(pSN != NULL)
			{
				if (pSN->GetLevel() <= level)
				{
					pSNStart = pSN;
					pSN->ResetVisited();
				}
				pSN = pSN->GetDownList();
			}
		}
	}

	return pSNStart;
}

void CMGMesh::ResetVisitedVertices()
{
	// reset the node flags
	for (int x = 0; x < m_nrSlices.u; x++)
	{
		// check row
		CSurfNodeList *pList = m_nodeStack[x];
		if (pList == NULL) continue;

		// check colums
		for (int y = 0; y < m_nrSlices.v; y++)
		{
			CSurfNodeT *pSN = pList[y].GetFirstNode();
			while(pSN != NULL)
			{
				pSN->ResetVisited();
				pSN = pSN->GetDownList();
			}
		}
	}
}

void CMGMesh::ResetVisitedVerticesGrid(int level, BYTE layer)
{
	// reset the node flags
	for (int x = 0; x < m_nrSlices.u; x++)
	{
		// check row
		CSurfNodeList *pList = m_nodeStack[x];
		if (pList == NULL) continue;

		// check colums
		for (int y = 0; y < m_nrSlices.v; y++)
		{
			CSurfNodeT *pSN = pList[y].GetFirstNode();
			while(pSN != NULL)
			{
				if (pSN->GetLevel() <= level && pSN->GetLayer() == layer)
				{
					pSN->ResetVisited();
				}
				pSN = pSN->GetDownList();
			}
		}
	}
}

void CMGMesh::ResetVisitedVerticesBox(int level, BYTE layer)
{
	// reset the node flags
	for (int x = m_posVertMin.u; x <= m_posVertMax.u; x++)
	{
		CSurfNodeList *pList = m_nodeStack[x];
		if (pList == NULL) continue;

		for (int y = m_posVertMin.v; y <= m_posVertMax.v; y++)
		{
			CSurfNodeT *pSN = pList[y].GetFirstNode();
			while(pSN != NULL)
			{
				if (IsVertexInROI(pSN, level, layer))
					pSN->ResetVisited();

				pSN = pSN->GetDownList();
			}
		}
	}
}

/**********************************************************************************/

void CMGMesh::RestorePosNode(CIPointUVW &pos, int level, BYTE layer)
{
	const MStreamLayer &meshLevel = m_pStreamDescr->GetMeshLayer(level, layer);
	const CIPointUVW &levelDiff = meshLevel.GetLevelDiff();

	// map the position (given with respect to the current level) to highest resolution level
	m_grid->ComputePositionUp(pos, level, levelDiff);

	if (pos.u < 0 || pos.u >= m_nrSlices.u ||
		pos.v < 0 || pos.v >= m_nrSlices.v ||
		pos.w < 0 || pos.w >= m_nrSlices.w)
		throw CMGDataException(CMGErrMsg::INCONSIST_DESCR);
}

/**********************************************************************************/

void CMGMesh::RepositionVertex(int nextLevel, CSurfNodeT *pSN)
{
	CIPointUVW &pos3D = pSN->GetPos();
	int posX = pos3D.u, posY = pos3D.v;

	// inc or dec the reference pos
	switch (pSN->GetDir())
	{
	case BACK_FACE:		pos3D.u = GetPrevSliceU(pos3D.u, nextLevel);	break;
	case FRONT_FACE:	pos3D.u = GetNextSliceU(pos3D.u, nextLevel);	break;
	case LEFT_FACE:		pos3D.v = GetPrevSliceV(pos3D.v, nextLevel);	break;
	case RIGHT_FACE:	pos3D.v = GetNextSliceV(pos3D.v, nextLevel);	break;
	case BOTTOM_FACE:	pos3D.w = GetPrevSliceW(pos3D.w, nextLevel);	break;
	case TOP_FACE:		pos3D.w = GetNextSliceW(pos3D.w, nextLevel);	break;
	}

	// recompute the (x,y,z) coordinate
	UpdateVertexCoord(nextLevel, pSN);

	if (posX == pos3D.u && posY == pos3D.v)
		return;

	// move the node into the right position
	CSurfNodeList *pList = m_nodeStack[posX];
	ASSERT (pList != NULL);

#ifdef _DEBUG
	CSurfNodeT *pSNTemp = pList[posY].GetFirstNode();
	ASSERT (pSNTemp != NULL);

	while(pSNTemp != NULL)
	{
		if (pSNTemp == pSN)		break;

		pSNTemp = pSNTemp->GetDownList();
	}
	ASSERT(pSNTemp != NULL);
#endif

	// move the node into the corresponding position
	RemoveNodeInList(pSN, pList[posY]);

	pList = GetColumn(pos3D.u);
	InsertNodeInList(pSN, pList[pos3D.v]);
}

/**********************************************************************************/

CSurfNodeT * CMGMesh::RemoveNodeInList(CSurfNodeT *pSN, CSurfNodeList &nodeList)
{
	CSurfNodeT *pPrev = pSN->GetUpList();
	CSurfNodeT *pNext = pSN->GetDownList();

	// restore the list connectivity
	if (pPrev != NULL)
	{
		ASSERT(pPrev->GetDownList() == pSN);
		pPrev->SetDownList(pNext);
	}
	else
	{
		// removing the first node
		ASSERT(nodeList.GetFirstNode() == pSN);
		nodeList.SetFirstNode(pNext);
	}

	if (pNext != NULL)
	{
		ASSERT(pNext->GetUpList() == pSN);
		pNext->SetUpList(pPrev);
	}

	return pNext;
}

void CMGMesh::InsertNodeInList(CSurfNodeT *pSN, CSurfNodeList &nodeList)
{
	CSurfNodeT *pNext = nodeList.GetFirstNode();

	// insert it as the first node
	nodeList.SetFirstNode(pSN);
	pSN->SetUpList(NULL);
	pSN->SetDownList(pNext);

	if (pNext != NULL)
	{
		ASSERT(pNext->GetUpList() == NULL);
		pNext->SetUpList(pSN);
	}
}

/**********************************************************************************/

void CMGMesh::FindPosNodeLocal(CIPointUVW &pos, BYTE dir, int level, BYTE layer)
{
	// due to the reposition bits the position of a vertex might have migrated to other positions
	// belonging to higher resolution levels.
	// here we move (if necessary) the position back to current level
	FindPosNodeGlobal(pos, dir, level);

	const MStreamLayer &meshLevel = m_pStreamDescr->GetMeshLayer(level, layer);
	const CIPointUVW &levelDiff = meshLevel.GetLevelDiff();

	// map the position (given with respect to the highest resolution level) to current level
	pos.u = DevideBy2Even(pos.u, levelDiff.u);
	pos.v = DevideBy2Even(pos.v, levelDiff.v);
	pos.w = DevideBy2Even(pos.w, levelDiff.w);

	ASSERT(pos.u >= 0 && pos.u < GetNrSlicesU(level) &&
		   pos.v >= 0 && pos.v < GetNrSlicesV(level) &&
		   pos.w >= 0 && pos.w < GetNrSlicesW(level));
}

void CMGMesh::FindPosNodeGlobal(CIPointUVW &pos, BYTE dir, int level)
{
	switch (dir)
	{
	case BACK_FACE:	
		ASSERT(pos.u > 0);
		pos.u = GetNextSliceU(pos.u - 1, level);
		break;
	case FRONT_FACE:
		ASSERT(pos.u < m_nrSlices.u - 1);
		pos.u = GetPrevSliceU(pos.u + 1, level);
		break;
	case LEFT_FACE:
		ASSERT(pos.v > 0);
		pos.v = GetNextSliceV(pos.v - 1, level);
		break;
	case RIGHT_FACE:
		ASSERT(pos.v < m_nrSlices.v - 1);
		pos.v = GetPrevSliceV(pos.v + 1, level);
		break;
	case BOTTOM_FACE:
		ASSERT(pos.w > 0);
		pos.w = GetNextSliceW(pos.w - 1, level);
		break;
	case TOP_FACE:
		//		ASSERT(pos.w < m_nrSlices.w - 1);
		pos.w = GetPrevSliceW(pos.w + 1, level);
		break;
	default: ASSERT(0);
	}
}

/**********************************************************************************/

float CMGMesh::GetLevelOffset(int dir, const CIPointUVW &pos, float offset, int level)
{
	CFPoint3DT p1, p2;
	switch (dir)
	{
	case BACK_FACE:
		{
			ASSERT(pos.u > 0);
			int low		= GetNextSliceU(pos.u - 1, level);
			int high	= GetPrevSliceU(pos.u, level);

			GetBackPos(pos, p1, offset);
			const GridU &gridU = (*m_grid)[pos.w][pos.v];
			p1 -= gridU[low];
			p2 =  gridU[high]  - gridU[low];
			return p1.GetLength() / p2.GetLength();
		}
	case FRONT_FACE:
		{
			ASSERT(pos.u < m_nrSlices.u - 1);
			int low		= GetPrevSliceU(pos.u + 1, level);
			int high	= GetNextSliceU(pos.u, level);

			GetFrontPos(pos, p1, offset);
			const GridU &gridU = (*m_grid)[pos.w][pos.v];
			p1 -= gridU[low];
			p2 =  gridU[high]  - gridU[low];
			return p1.GetLength() / p2.GetLength();
		}
	case LEFT_FACE:
		{
			ASSERT(pos.v > 0);
			int low		= GetNextSliceV(pos.v - 1, level);
			int high	= GetPrevSliceV(pos.v, level);

			GetLeftPos(pos, p1, offset);
			const GridUV &gridUV = (*m_grid)[pos.w];
			p1 -= gridUV[low][pos.u];
			p2 =  gridUV[high][pos.u]  - gridUV[low][pos.u];
			return p1.GetLength() / p2.GetLength();
		}
	case RIGHT_FACE:
		{
			ASSERT(pos.v < m_nrSlices.v - 1);
			int low		= GetPrevSliceV(pos.v + 1, level);
			int high	= GetNextSliceV(pos.v, level);

			GetRightPos(pos, p1, offset);
			const GridUV &gridUV = (*m_grid)[pos.w];
			p1 -= gridUV[low][pos.u];
			p2 =  gridUV[high][pos.u]  - gridUV[low][pos.u];
			return p1.GetLength() / p2.GetLength();
		}
	case BOTTOM_FACE:
		{
			ASSERT(pos.w > 0);
			int low		= GetNextSliceW(pos.w - 1, level);
			int high	= GetPrevSliceW(pos.w, level);

			GetBottomPos(pos, p1, offset);
			p1 -= (*m_grid)[low][pos.v][pos.u];
			p2 =  (*m_grid)[high][pos.v][pos.u]  - (*m_grid)[low][pos.v][pos.u];
			return p1.GetLength() / p2.GetLength();
		}
	case TOP_FACE:
		{
			ASSERT(pos.w < m_nrSlices.w - 1);
			int low		= GetPrevSliceW(pos.w + 1, level);
			int high	= GetNextSliceW(pos.w, level);

			GetTopPos(pos, p1, offset);
			p1 -= (*m_grid)[low][pos.v][pos.u];
			p2 =  (*m_grid)[high][pos.v][pos.u]  - (*m_grid)[low][pos.v][pos.u];
			return p1.GetLength() / p2.GetLength();
		}
	default: ASSERT(0);	return 0.0f;
	}
}

bool CMGMesh::IsSpecialCase(int nextLevel, CSurfNodeT *pSN)
{
	bool evenLength, bLast;
	const	CIPointUVW &pos = pSN->GetPos();

	switch (pSN->GetDir())
	{
	case BACK_FACE:
		{
			evenLength = (GetNrSlicesU(nextLevel) % 2) == 0;
			if (!evenLength)  return false;

			bLast = GetPrevSliceU(pos.u, nextLevel - 1) == 0;
			return bLast == true;
		}
		break;
	case FRONT_FACE:
		{
			evenLength = (GetNrSlicesU(nextLevel) % 2) == 0;
			if (!evenLength)  return false;

			bLast = GetNextSliceU(pos.u, nextLevel - 1) == (m_nrSlices.u - 1);
			return bLast == true;
		}
		break;
	case LEFT_FACE:
		{
			evenLength = (GetNrSlicesV(nextLevel) % 2) == 0;
			if (!evenLength)  return false;

			bLast = GetPrevSliceV(pos.v, nextLevel - 1) == 0;
			return bLast == true;
		}
		break;
	case RIGHT_FACE:
		{
			evenLength = (GetNrSlicesV(nextLevel) % 2) == 0;
			if (!evenLength)  return false;

			bLast = GetNextSliceV(pos.v, nextLevel - 1) == (m_nrSlices.v - 1);
			return bLast == true;
		}
		break;
	case BOTTOM_FACE:
		{
			evenLength = (GetNrSlicesW(nextLevel) % 2) == 0;
			if (!evenLength)  return false;

			bLast = GetPrevSliceW(pos.w, nextLevel - 1) == 0;
			return bLast == true;
		}
		break;
	case TOP_FACE:
		{
			evenLength = (GetNrSlicesW(nextLevel) % 2) == 0;
			if (!evenLength)  return false;

			bLast = GetNextSliceW(pos.w, nextLevel - 1) == (m_nrSlices.w - 1);
			return bLast == true;
		}
		break;
	default: ASSERT(0);	return false;
	}
}

int CMGMesh::IsSpecialSplit(int nextLevel, int dir, const CIPointUVW &pos1, const CIPointUVW &pos2, int &dirSplit)
{  
	switch (dir)
	{
	case BACK_FACE:
	case FRONT_FACE:
		ASSERT(pos1.u == pos2.u);
		if (pos1.v == pos2.v)
		{
			dirSplit = DIR_W;

			if (GetNrSlicesW(nextLevel - 1) == GetNrSlicesW(nextLevel))
				return NO_SPLIT;

			// even length, check if located on the last grid position
			bool evenLength = (GetNrSlicesW(nextLevel) % 2) == 0;
			if (evenLength)
			{
				bool lastGridLine = pos1.w == m_nrSlices.w - 1 || pos2.w == m_nrSlices.w - 1;
				return lastGridLine ? NON_UNIFORM: UNIFORM;
			}

			return UNIFORM;
		}
		else
		{
			ASSERT(pos1.w == pos2.w);
			dirSplit = DIR_V;

			if (GetNrSlicesV(nextLevel - 1) == GetNrSlicesV(nextLevel))
				return NO_SPLIT;

			// even length, check if located on the last grid position
			bool evenLength = (GetNrSlicesV(nextLevel) % 2) == 0;
			if (evenLength)
			{
				bool lastGridLine = pos1.v == m_nrSlices.v - 1 || pos2.v == m_nrSlices.v - 1;
				return lastGridLine ? NON_UNIFORM: UNIFORM;
			}

			return UNIFORM;
		}
		break;
	case LEFT_FACE:
	case RIGHT_FACE:
		ASSERT(pos1.v == pos2.v);
		if (pos1.u == pos2.u)
		{
			dirSplit = DIR_W;
			if (GetNrSlicesW(nextLevel - 1) == GetNrSlicesW(nextLevel))
				return NO_SPLIT;

			// even length, check if located on the last grid position
			bool evenLength = (GetNrSlicesW(nextLevel) % 2) == 0;
			if (evenLength)
			{
				bool lastGridLine = pos1.w == m_nrSlices.w - 1 || pos2.w == m_nrSlices.w - 1;
				return lastGridLine ? NON_UNIFORM: UNIFORM;
			}

			return UNIFORM;
		}
		else
		{
			ASSERT(pos1.w == pos2.w);
			dirSplit = DIR_U;

			if (GetNrSlicesU(nextLevel - 1) == GetNrSlicesU(nextLevel))
				return NO_SPLIT;

			// even length, check if located on the last grid position
			bool evenLength = (GetNrSlicesU(nextLevel) % 2) == 0;
			if (evenLength)
			{
				bool lastGridLine = pos1.u == m_nrSlices.u - 1 || pos2.u == m_nrSlices.u - 1;
				return lastGridLine ? NON_UNIFORM: UNIFORM;
			}

			return UNIFORM;
		}
		break;
	case BOTTOM_FACE:
	case TOP_FACE:
		ASSERT(pos1.w == pos2.w);
		if (pos1.u == pos2.u)
		{
			dirSplit = DIR_V;
			if (GetNrSlicesV(nextLevel - 1) == GetNrSlicesV(nextLevel))
				return NO_SPLIT;

			// even length, check if located on the last grid position
			bool evenLength = (GetNrSlicesV(nextLevel) % 2) == 0;
			if (evenLength)
			{
				bool lastGridLine = pos1.v == m_nrSlices.v - 1 || pos2.v == m_nrSlices.v - 1;
				return lastGridLine ? NON_UNIFORM: UNIFORM;
			}

			return UNIFORM;
		}
		else
		{
			ASSERT(pos1.v == pos2.v);
			dirSplit = DIR_U;
			if (GetNrSlicesU(nextLevel - 1) == GetNrSlicesU(nextLevel))
				return NO_SPLIT;

			// even length, check if located on the last grid position
			bool evenLength = (GetNrSlicesU(nextLevel) % 2) == 0;
			if (evenLength)
			{
				bool lastGridLine = pos1.u == m_nrSlices.u - 1 || pos2.u == m_nrSlices.u - 1;
				return lastGridLine ? NON_UNIFORM: UNIFORM;
			}

			return UNIFORM;
		}
		break;
	default: ASSERT(0);	return 0;
	}
}

BYTE CMGMesh::VertexRepositionBit(int nextLevel, CSurfNodeT *pSN, bool &bSpecialCase)
{
	bSpecialCase = false;
	const	CIPointUVW &pos = pSN->GetPos();

	int low, high;
	bool evenLength, bLast;

	switch (pSN->GetDir())
	{
	case BACK_FACE:
		{
			ASSERT(pos.u > 0);
			low		= GetNextSliceU(pos.u - 1, nextLevel - 1);
			high	= GetNextSliceU(pos.u - 1, nextLevel);
			ASSERT(low >= high);

			if (low == high)    return 0;   // 0 -> 1-bit symbol
			else
			{
				evenLength = (GetNrSlicesU(nextLevel) % 2) == 0;
				if (!evenLength)  return 1;   // 1 -> 1-bit symbol

				bLast = GetPrevSliceU(pos.u, nextLevel - 1) == 0;
				if (!bLast)       return 1;   // 1 -> 1-bit symbol

				// two reference-surfaces will be inserted at the next level
				bSpecialCase = true;

				high = GetPrevSliceU(pos.u, nextLevel);
				if (high > 0)     return 2;   // 10 ->2-bit symbol
				else              return 3;   // 11 ->2-bit symbol
			}
		}
		break;
	case FRONT_FACE:
		{
			ASSERT(pos.u < m_nrSlices.u - 1);
			low		= GetPrevSliceU(pos.u + 1, nextLevel - 1);
			high	= GetPrevSliceU(pos.u + 1, nextLevel);
			ASSERT(low <= high);

			if (low == high)        return 0;   // 0 -> 1-bit symbol
			else
			{
				evenLength = (GetNrSlicesU(nextLevel) % 2) == 0;
				if (!evenLength)      return 1;   // 1 -> 1-bit symbol

				int lastSlice = m_nrSlices.u - 1;
				bLast = GetNextSliceU(pos.u, nextLevel - 1) == lastSlice;
				if (!bLast)           return 1;   // 1 -> 1-bit symbol

				// two reference-surfaces will be inserted at the next level
				bSpecialCase = true;

				high = GetNextSliceU(pos.u, nextLevel);
				if (high < lastSlice) return 2;   // 10 ->2-bit symbol
				else                  return 3;   // 11 ->2-bit symbol
			}
		}
		break;
	case LEFT_FACE:
		{
			ASSERT(pos.v > 0);
			low		= GetNextSliceV(pos.v - 1, nextLevel - 1);
			high	= GetNextSliceV(pos.v - 1, nextLevel);
			ASSERT(low >= high);

			if (low == high)    return 0;   // 0 -> 1-bit symbol
			else
			{
				evenLength = (GetNrSlicesV(nextLevel) % 2) == 0;
				if (!evenLength)  return 1;   // 1 -> 1-bit symbol

				bLast = GetPrevSliceV(pos.v, nextLevel - 1) == 0;
				if (!bLast)       return 1;   // 1 -> 1-bit symbol

				// two reference-surfaces will be inserted at the next level
				bSpecialCase = true;

				high = GetPrevSliceV(pos.v, nextLevel);
				if (high > 0)     return 2;   // 10 ->2-bit symbol
				else              return 3;   // 11 ->2-bit symbol
			}
		}
		break;
	case RIGHT_FACE:
		{
			ASSERT(pos.v < m_nrSlices.v - 1);
			low		= GetPrevSliceV(pos.v + 1, nextLevel - 1);
			high	= GetPrevSliceV(pos.v + 1, nextLevel);
			ASSERT(low <= high);

			if (low == high)        return 0;   // 0 -> 1-bit symbol
			else
			{
				evenLength = (GetNrSlicesV(nextLevel) % 2) == 0;
				if (!evenLength)      return 1;   // 1 -> 1-bit symbol

				int lastSlice = m_nrSlices.v - 1;
				bLast = GetNextSliceV(pos.v, nextLevel - 1) == lastSlice;
				if (!bLast)           return 1;   // 1 -> 1-bit symbol

				// two reference-surfaces will be inserted at the next level
				bSpecialCase = true;

				high = GetNextSliceV(pos.v, nextLevel);
				if (high < lastSlice) return 2;   // 10 ->2-bit symbol
				else                  return 3;   // 11 ->2-bit symbol
			}
		}
		break;
	case BOTTOM_FACE:
		{
			ASSERT(pos.w > 0);
			low		= GetNextSliceW(pos.w - 1, nextLevel - 1);
			high	= GetNextSliceW(pos.w - 1, nextLevel);
			ASSERT(low >= high);

			if (low == high)    return 0;   // 0 -> 1-bit symbol
			else
			{
				evenLength = (GetNrSlicesW(nextLevel) % 2) == 0;
				if (!evenLength)  return 1;   // 1 -> 1-bit symbol

				bLast = GetPrevSliceW(pos.w, nextLevel - 1) == 0;
				if (!bLast)       return 1;   // 1 -> 1-bit symbol

				// two reference-surfaces will be inserted at the next level
				bSpecialCase = true;

				high = GetPrevSliceW(pos.w, nextLevel);
				if (high > 0)     return 2;   // 10 ->2-bit symbol
				else              return 3;   // 11 ->2-bit symbol
			}
		}
		break;
	case TOP_FACE:
		{
			ASSERT(pos.w < m_nrSlices.w - 1);
			low		= GetPrevSliceW(pos.w + 1, nextLevel - 1);
			high	= GetPrevSliceW(pos.w + 1, nextLevel);
			ASSERT(low <= high);

			if (low == high)        return 0;   // 0 -> 1-bit symbol
			else
			{
				evenLength = (GetNrSlicesW(nextLevel) % 2) == 0;
				if (!evenLength)      return 1;   // 1 -> 1-bit symbol

				int lastSlice = m_nrSlices.w - 1;
				bLast = GetNextSliceW(pos.w, nextLevel - 1) == lastSlice;
				if (!bLast)           return 1;   // 1 -> 1-bit symbol

				// two reference-surfaces will be inserted at the next level
				bSpecialCase = true;

				high = GetNextSliceW(pos.w, nextLevel);
				if (high < lastSlice) return 2;   // 10 ->2-bit symbol
				else                  return 3;   // 11 ->2-bit symbol
			}
		}
		break;
	default: ASSERT(0);	return 0;
	}
}

/**********************************************************************************/

void CMGMesh::GetDirectionDifference(int dir, int dirNext, BYTE &val)
{
	switch (dir)
	{
	case BACK_FACE:
		switch (dirNext)
		{
		case BACK_FACE:		val = SAME_DIR;		break;
		case FRONT_FACE:	ASSERT(0);				break;
		case RIGHT_FACE:	// continue with next
		case BOTTOM_FACE:	val = CW90_DIR;		break;
		case LEFT_FACE:		// continue with next
		case TOP_FACE:		val = CCW90_DIR;  break;
		}
		break;
	case FRONT_FACE:
		switch (dirNext)
		{
		case FRONT_FACE:	val = SAME_DIR;		break;
		case BACK_FACE:		ASSERT(0);				break;
		case LEFT_FACE:		// continue with next
		case TOP_FACE:		val = CW90_DIR;	  break;
		case BOTTOM_FACE:	// continue with next
		case RIGHT_FACE:	val = CCW90_DIR;	break;
		}
		break;
	case LEFT_FACE:
		switch (dirNext)
		{
		case LEFT_FACE:		val = SAME_DIR;		break;
		case RIGHT_FACE:	ASSERT(0);				break;
		case TOP_FACE:		// continue with next
		case BACK_FACE:		val = CW90_DIR;	  break;
		case BOTTOM_FACE:	// continue with next
		case FRONT_FACE:	val = CCW90_DIR;	break;
		}
		break;
	case RIGHT_FACE:
		switch (dirNext)
		{
		case RIGHT_FACE:	val = SAME_DIR; 	break;
		case LEFT_FACE:		ASSERT(0);				break;
		case BOTTOM_FACE:	// continue with next
		case FRONT_FACE:	val = CW90_DIR;	  break;
		case BACK_FACE:		// continue with next
		case TOP_FACE:		val = CCW90_DIR;	break;
		}
		break;
	case BOTTOM_FACE:
		switch (dirNext)
		{
		case BOTTOM_FACE:	val = SAME_DIR;		break;
		case TOP_FACE:		ASSERT(0);				break;
		case FRONT_FACE:	// continue with next
		case LEFT_FACE:		val = CW90_DIR;	  break;
		case BACK_FACE:		// continue with next
		case RIGHT_FACE:	val = CCW90_DIR;	break;
		}
		break;
	case TOP_FACE:
		switch (dirNext)
		{
		case TOP_FACE:		val = SAME_DIR;		break;
		case BOTTOM_FACE:	ASSERT(0);				break;
		case BACK_FACE:		// continue with next
		case RIGHT_FACE:	val = CW90_DIR;	  break;
		case LEFT_FACE:		// continue with next
		case FRONT_FACE:	val = CCW90_DIR;	break;
		}
		break;
	}
}

/**********************************************************************************/

void CMGMesh::UpdateBackFaceCoordinates(BYTE &dirNext, int flag, bool &dirSwitch, CIPointUVW &pos, int level)
{
	switch (dirNext)
	{
	case SAME_DIR:
		dirNext = BACK_FACE;	 dirSwitch = false;
		if (flag == 1) pos.v = GetNextSliceV(pos.v, level);
		else		   pos.w = GetPrevSliceW(pos.w, level);
		break;
	case CW90_DIR:
		if (flag == 1)
		{
			dirNext = RIGHT_FACE;	dirSwitch = false;
		}
		else
		{
			dirNext = BOTTOM_FACE; dirSwitch = true;
		}
		break;
	case CCW90_DIR:
		if (flag == 1)
		{
			dirNext = LEFT_FACE;	dirSwitch = true;
			pos.u = GetPrevSliceU(pos.u, level);
			pos.v = GetNextSliceV(pos.v, level);
		}
		else
		{
			dirNext = TOP_FACE;	 dirSwitch = false;
			pos.u = GetPrevSliceU(pos.u, level);
			pos.w = GetPrevSliceW(pos.w, level);
		}
		break;
	default: ASSERT(0);
	}
}

void CMGMesh::UpdateFrontFaceCoordinates(BYTE &dirNext, int flag, bool &dirSwitch, CIPointUVW &pos, int level)
{
	switch (dirNext)
	{
	case SAME_DIR:
		dirNext = FRONT_FACE;		dirSwitch = false;
		if (flag == 1) pos.w = GetNextSliceW(pos.w, level);
		else		   pos.v = GetPrevSliceV(pos.v, level);
		break;
	case CW90_DIR:
		if (flag == 1)
		{
			dirNext = TOP_FACE;		dirSwitch = true;
		}
		else
		{
			dirNext = LEFT_FACE; dirSwitch = false;
		}
		break;
	case CCW90_DIR:
		if (flag == 1)
		{
			dirNext = BOTTOM_FACE;	dirSwitch = false;
			pos.u = GetNextSliceU(pos.u, level);
			pos.w = GetNextSliceW(pos.w, level);
		}
		else
		{
			dirNext = RIGHT_FACE;	 dirSwitch = true;
			pos.u = GetNextSliceU(pos.u, level);
			pos.v = GetPrevSliceV(pos.v, level);
		}
		break;
	default: ASSERT(0);
	}
}

void CMGMesh::UpdateLeftFaceCoordinates(BYTE &dirNext, int flag, bool &dirSwitch, CIPointUVW &pos, int level)
{
	switch (dirNext)
	{
	case SAME_DIR:
		dirNext = LEFT_FACE;	dirSwitch = false;
		if (flag == 1) pos.w = GetNextSliceW(pos.w, level);
		else		   pos.u = GetPrevSliceU(pos.u, level);
		break;
	case CW90_DIR:
		if (flag == 1)
		{
			dirNext = TOP_FACE;		dirSwitch = false;
		}
		else
		{
			dirNext = BACK_FACE; dirSwitch = true;
		}
		break;
	case CCW90_DIR:
		if (flag == 1)
		{
			dirNext = BOTTOM_FACE;	dirSwitch = true;
			pos.v = GetPrevSliceV(pos.v, level);
			pos.w = GetNextSliceW(pos.w, level);
		}
		else
		{
			dirNext = FRONT_FACE;	 dirSwitch = false;
			pos.v = GetPrevSliceV(pos.v, level);
			pos.u = GetPrevSliceU(pos.u, level);
		}
		break;
	default: ASSERT(0);
	}
}

void CMGMesh::UpdateRightFaceCoordinates(BYTE &dirNext, int flag, bool &dirSwitch, CIPointUVW &pos, int level)
{
	switch (dirNext)
	{
	case SAME_DIR:
		dirNext = RIGHT_FACE;	dirSwitch = false;
		if (flag == 1) pos.u = GetNextSliceU(pos.u, level);
		else		   pos.w = GetPrevSliceW(pos.w, level);
		break;
	case CW90_DIR:
		if (flag == 1)
		{
			dirNext = FRONT_FACE;		dirSwitch = true;
		}
		else
		{
			dirNext = BOTTOM_FACE;	dirSwitch = false;
		}
		break;
	case CCW90_DIR:
		if (flag == 1)
		{
			dirNext = BACK_FACE;	 dirSwitch = false;
			pos.v = GetNextSliceV(pos.v, level);
			pos.u = GetNextSliceU(pos.u, level);
		}
		else
		{
			dirNext = TOP_FACE;		 dirSwitch = true;
			pos.v = GetNextSliceV(pos.v, level);
			pos.w = GetPrevSliceW(pos.w, level);
		}
		break;
	default: ASSERT(0);
	}
}

void CMGMesh::UpdateBottomFaceCoordinates(BYTE &dirNext, int flag, bool &dirSwitch, CIPointUVW &pos, int level)
{
	switch (dirNext)
	{
	case SAME_DIR:
		dirNext = BOTTOM_FACE;	dirSwitch = false;
		if (flag == 1) pos.u = GetNextSliceU(pos.u, level);
		else		   pos.v = GetPrevSliceV(pos.v, level);
		break;
	case CW90_DIR:
		if (flag == 1)
		{
			dirNext = FRONT_FACE;	 dirSwitch = false;
		}
		else
		{
			dirNext = LEFT_FACE;	 dirSwitch = true;
		}
		break;
	case CCW90_DIR:
		if (flag == 1)
		{
			dirNext = BACK_FACE;	 dirSwitch = true;
			pos.w = GetPrevSliceW(pos.w, level);
			pos.u = GetNextSliceU(pos.u, level);
		}
		else
		{
			dirNext = RIGHT_FACE;	 dirSwitch = false;
			pos.w = GetPrevSliceW(pos.w, level);
			pos.v = GetPrevSliceV(pos.v, level);
		}
		break;
	default: ASSERT(0);
	}
}

void CMGMesh::UpdateTopFaceCoordinates(BYTE &dirNext, int flag, bool &dirSwitch, CIPointUVW &pos, int level)
{
	switch (dirNext)
	{
	case SAME_DIR:
		dirNext = TOP_FACE;	dirSwitch = false;
		if (flag == 1) pos.v = GetNextSliceV(pos.v, level);
		else		   pos.u = GetPrevSliceU(pos.u, level);
		break;
	case CW90_DIR:
		if (flag == 1)
		{
			dirNext = RIGHT_FACE;	 dirSwitch = true;
		}
		else
		{
			dirNext = BACK_FACE;	 dirSwitch = false;
		}
		break;
	case CCW90_DIR:
		if (flag == 1)
		{
			dirNext = LEFT_FACE;	 dirSwitch = false;
			pos.w = GetNextSliceW(pos.w, level);
			pos.v = GetNextSliceV(pos.v, level);
		}
		else
		{
			dirNext = FRONT_FACE;	 dirSwitch = true;
			pos.w = GetNextSliceW(pos.w, level);
			pos.u = GetPrevSliceU(pos.u, level);
		}
		break;
	default: ASSERT(0);
	}
}

void CMGMesh::UpdateCoordinateFromConnectivity(int dir, BYTE &dirNext, int flag, bool &dirSwitch, CIPointUVW &pos, int level)
{
	switch (dir)
	{
	case BACK_FACE:		UpdateBackFaceCoordinates(dirNext, flag, dirSwitch, pos, level);		break;
	case FRONT_FACE:	UpdateFrontFaceCoordinates(dirNext, flag, dirSwitch, pos, level);		break;
	case LEFT_FACE:		UpdateLeftFaceCoordinates(dirNext, flag, dirSwitch, pos, level);		break;
	case RIGHT_FACE:	UpdateRightFaceCoordinates(dirNext, flag, dirSwitch, pos, level);		break;
	case BOTTOM_FACE:	UpdateBottomFaceCoordinates(dirNext, flag, dirSwitch, pos, level);	break;
	case TOP_FACE:		UpdateTopFaceCoordinates(dirNext, flag, dirSwitch, pos, level);			break;
	default: ASSERT(0);
	}

	ASSERT(pos.u >= 0 && pos.u < m_nrSlices.u &&
		pos.v >= 0 && pos.v < m_nrSlices.v &&
		pos.w >= 0 && pos.w < m_nrSlices.w);
}

/**********************************************************************************/

//  The following is an example of a hierarchical multi-resolution grid with 4 levels
//  Notice in the following explanation that it is easier to think in terms of distances
//  between the points instead of the number of points (size of the ROI).

//  The first level (Level0) consists of 3 points (2^1 = 2 distances, size = 3)
//  The second level (Level1) consists of 5 points (2^2 = 4 distances, size = 5)
//  The third level (Level2) consists of 9 points (2^3 = 8 distances, size = 9)
//  The fourth level (Level3) consists of 17 points (2^4 = 16 distances, size = 17)
//  The number of distances d(l+1) of level (l+1) (between two consecutive points of the level l+1)
//  are the double of the number distances d(l) of level l (between two consecutive points of level l).
//  The relation between the number of distances d(l) and nr of points g(l) = d(l) + 1

//  The ROIs for a resolution level l will contain only points of level l
//  The ROIs for consecutive resolution levels are defined in a hierarchical way
//  such that the ROIs from a resolution level (l+1) consist of a multiple of the number of
//  distances dr(l+1) = k * dr(l) of level l, and the number of grid points gr(l+1) of level (l+1)
//  can be defined in terms of the samples gr(l) of level l as gr(l+1) = k * (gr(l) - 1) + 1
//  with k <= 2. k can be less than 1 such that the ratio k = dr(i) / dr(l+1) yields an integer value.

//  For k = 2 the ROIs from successive levels have the same limit grid points.
//  Theoretically k can be larger than two but it has been restricted k <= 2 in order to exploit
//  the hierarchical nature of the grid

//  It is more intuitive to define the ROI (position and size) for a resolution level l and map it to the highest resolution
//  of the mesh, containig all the grid points.

//  Position  0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15  16
//  Level0    *                               *                               *
//  Level1    *               -               *               -               *
//  Level2    *       |       -       |       *       |       -       |       *
//  Level3    *   +   |   +   -   +   |   +   *   +   |   +   -   +   |   +   *


void CMGMesh::InitROI(int level, BYTE layer, int roiIdx)
{
	MStreamLayer &meshLevel = m_pStreamDescr->GetMeshLayer(level, layer);
	const CIPointUVW &levelDiff = meshLevel.GetLevelDiff();

	ROIDescription &roiDescr = m_pStreamDescr->GetROIDescr();
	roiDescr.ComputePositionAndSize(m_localROI, roiIdx, meshLevel.GetNrROIs(), levelDiff);

	// the minimum position of a vertex (with respect to the highest resolution level)
	// that may reside within the limits of m_localROI
	m_posVertMin.u = (m_localROI.u == 0) ? 0 : MapPositionUp(m_localROI.u - 1, levelDiff.u) + 1;
	m_posVertMin.v = (m_localROI.v == 0) ? 0 : MapPositionUp(m_localROI.v - 1, levelDiff.v) + 1;
	m_posVertMin.w = (m_localROI.w == 0) ? 0 : MapPositionUp(m_localROI.w - 1, levelDiff.w) + 1;

	// the maximum position of a vertex (with respect to the highest resolution level)
	// that may reside within the limits of m_localROI
	CIPointUVW  posLocalMax(m_localROI.u + m_localROI.width  - 1,
							m_localROI.v + m_localROI.height - 1,
							m_localROI.w + m_localROI.depth  - 1);
	m_posVertMax.u = (posLocalMax.u == GetNrSlicesU(level) - 1) ? GetNrSlicesU() - 1 : MapPositionUp(posLocalMax.u + 1, levelDiff.u) - 1;
	m_posVertMax.v = (posLocalMax.v == GetNrSlicesV(level) - 1) ? GetNrSlicesV() - 1 : MapPositionUp(posLocalMax.v + 1, levelDiff.v) - 1;
	m_posVertMax.w = (posLocalMax.w == GetNrSlicesW(level) - 1) ? GetNrSlicesW() - 1 : MapPositionUp(posLocalMax.w + 1, levelDiff.w) - 1;

	// the mapping of the local ROI at the highest resolution level
	m_globalROI.u = MapPositionUp(m_localROI.u, levelDiff.u);
	m_globalROI.v = MapPositionUp(m_localROI.v, levelDiff.v);
	m_globalROI.w = MapPositionUp(m_localROI.w, levelDiff.w);


	if (m_localROI.u + m_localROI.width == GetNrSlicesU(level))
		m_globalROI.width  = GetNrSlicesU() - m_globalROI.u;
	else
		m_globalROI.width  = MapSizeUp(m_localROI.width,  levelDiff.u);
	if (m_localROI.v + m_localROI.height == GetNrSlicesV(level))
		m_globalROI.height  = GetNrSlicesV() - m_globalROI.v;
	else
		m_globalROI.height = MapSizeUp(m_localROI.height, levelDiff.v);
	if (m_localROI.w + m_localROI.depth == GetNrSlicesW(level))
		m_globalROI.depth  = GetNrSlicesW() - m_globalROI.w;
	else
		m_globalROI.depth  = MapSizeUp(m_localROI.depth,  levelDiff.w);
}
