/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGMeshCoder.cpp: implementation of the CMGMesh class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGMesh.h"

// export as VRML file
static void WriteMGHeader(CStdioFile &file)
{
	file.WriteString("#VRML V2.0 utf8 (Converted to ASCII)\n");
	file.WriteString("Shape {\n");
	file.WriteString("  appearance Appearance {\n");
	file.WriteString("    material Material {  # blue\n");
	file.WriteString("      ambientIntensity  0.5\n");
	file.WriteString("      diffuseColor      0.3 0.4 1.0\n");
	file.WriteString("    }\n");
	file.WriteString("  }\n");
	file.WriteString("  geometry MeshGrid {\n");
}

bool CMGMesh::Export(CStdioFile &file, int level, bool bCompact,
					 bool hasConnectivityInfo, bool hasRefineInfo, bool hasRepositionInfo, bool interleaveRefine)
{
	CString str;

	// initialise the exchange object with the mesh coding details
	InitExchangeObjectMesh(hasConnectivityInfo, hasRefineInfo, hasRepositionInfo, interleaveRefine);

	// write the MG header
	WriteMGHeader(file);

	// write the number of slices and number of levels
	int nrSlicesU = GetNrSlicesU(level);
	int nrSlicesV = GetNrSlicesV(level);
	int nrSlicesW = GetNrSlicesW(level);

	str.Format("\t\tnSlices [ %d %d %d ]\n", nrSlicesU, nrSlicesV, nrSlicesW);
	file.WriteString(str);

	str.Format("\t\tnLevels %d\n", 1);
	file.WriteString(str);

	str.Format("\t\tdisplayLevel %d\n", 0);
	file.WriteString(str);

	// write the Reference Grid
	if (bCompact)
		m_grid->ExportCorners(file);
	else
		m_grid->Export(file, level);

	// write the vertices' connectivity
	{
		file.WriteString("\t\tvertexLink [\n");

		ExportMeshConnectivity(level, file);

		// end of connectivity
		file.WriteString("\t\t]\n");
	}

	// write the vertices' offsets
	{
		file.WriteString("\t\tvertexOffset [\n");

		ExportMeshRefineBits(level, file);

		// end of refine bits
		file.WriteString("\t\t]\n");
	}

	// end of the description
	file.WriteString("\t}\n}");

	delete m_pStreamDescr;  m_pStreamDescr = NULL;
	for (int k = 0; k < GetMaxNrLevels(); k++)
	{
		delete m_startVertsROI[k][0];
		delete []m_startVertsROI[k];
	}
	delete []m_startVertsROI;   m_startVertsROI = NULL;

	return true;
}

void CMGMesh::ExportMeshGridFields(CStdioFile &file)
{
	CString str;

	const CIPointUVW &nrSlices = GetNrSlices();
	str.Format("\t\tnSlices [ %d %d %d ]\n", nrSlices.u, nrSlices.v, nrSlices.w);
	file.WriteString(str);

	const CIPointUVW &nrLevels = m_grid->GetNrLevelsUVW();
	str.Format("\t\tnLevels [ %d %d %d ]\n", nrLevels.u, nrLevels.v, nrLevels.w);
	file.WriteString(str);

	str.Format("\t\tdisplayLevel %d\n", GetMaxNrLevels() - 1);
	file.WriteString(str);
}

void CMGMesh::ExportMeshParameters(CStdioFile &file)
{
	CString str;

	str.Format("\t\t\t%d", m_pStreamDescr->HasConnectivityInfo() ? 1 : 0);
	file.WriteString(str);

	int meshType = GetMeshType();
	str.Format(", %d", meshType);
	file.WriteString(str);

	if (meshType == QUADRI_MESH)
	{
		if (m_pStreamDescr->HasConnectivityInfo())
		{
			str.Format(", %d", IsSameConnectivity() ? 1 : 0);
			file.WriteString(str);

			str.Format(", %d", IsAutoResolution() ? 1 : 0);
			file.WriteString(str);
		}
		// the offset amplitude will be saved in the vertexOffset field
	}

	str.Format(", %d\n", GetCyclicMode());
	file.WriteString(str);
}

void CMGMesh::ExportAbsoluteOffset(CStdioFile &file)
{
	float offset = (float) m_pStreamDescr->GetOffsetAmplitude();

	// store the sign
	if (offset < 0.0f)
	{
		offset *= -1;
		file.WriteString(", 1");
	}
	else
		file.WriteString(", 0");

	// write the power of 10; offset = 10^n * offset;
	while (offset > 1.0f)
	{
		file.WriteString(", 1");
		offset /= 10.0f;
	}

	// the offset is less than 1
	file.WriteString(", 0");

	// write the offset < 1
	CString str;
	str.Format(", %f", offset);
	file.WriteString(str);
}

void CMGMesh::ExportRefineParameters(CStdioFile &file)
{
	CString str;

	str.Format("\t\t\t%d", m_pStreamDescr->HasRefineInfo() ? 1 : 0);
	file.WriteString(str);

	// are there offsets ?
	if (m_pStreamDescr->HasRefineInfo())
	{
		// the offsetAmplitude
		if (m_pStreamDescr->IsSingleLayered())
			ExportAbsoluteOffset(file);

		// are the offsets interleaved ?
		str.Format(", %d", InterleaveRefine() ? 1 : 0);
		file.WriteString(str);
	}

	str.Format(", %d", m_pStreamDescr->HasRepositionInfo() ? 1 : 0);
	file.WriteString(str);

	str.Format("\n");
	file.WriteString(str);
}

bool CMGMesh::Export(CStdioFile &file, bool bCompact,
					 bool hasConnectivityInfo, bool hasRefineInfo, bool hasRepositionInfo, bool interleaveRefine)
{
	CString str;

	// initialise the exchange object with the mesh coding details
	InitExchangeObjectMesh(hasConnectivityInfo, hasRefineInfo, hasRepositionInfo, interleaveRefine);

	// write the MG header
	WriteMGHeader(file);

	// write the number of slices and number of levels
	ExportMeshGridFields(file);

	// write the Reference Grid
	if (bCompact)
		m_grid->ExportCorners(file);
	else
		m_grid->Export(file, GetMaxNrLevels() - 1);

	// write the vertices' connectivity for all the levels
	int maxLevel = IsAutoResolution() ? 1 : GetMaxNrLevels();
	{
		file.WriteString("\t\tvertexLink [\n");

		// store mesh parameters
		ExportMeshParameters(file);

		if (hasConnectivityInfo)
		{
			for (int level = 0; level < maxLevel; level++)
				ExportMeshConnectivity(level, file);
		}
		// end of connectivity
		file.WriteString("\t\t]\n");
	}
	{
		// write the refine bits
		file.WriteString("\t\tvertexOffset [\n");

		// store refine parameters
		ExportRefineParameters(file);

		for (int level = 0; level < maxLevel; level++)
		{
			bool lastLevel = (level == maxLevel - 1);

			if (hasRefineInfo)
			{
				if (lastLevel || InterleaveRefine())
					ExportMeshRefineBits(level, file);
			}

			if (hasRepositionInfo && !lastLevel)
			{
				int refineMode = ExportRefineMode(level, file);
				if (refineMode == REPOSITION_ANY)
					ExportMeshRepositionBits(level, file);
			}
		}

		// end of refine bits
		file.WriteString("\t\t]\n");
	}

	// end of the description
	file.WriteString("\t}\n}");

	// dump the description of the MeshGrid
	bool dump = false;
	if (dump)
	{
		CStdioFile output;
		CFileException e;
		if (output.Open("dump.xml", CFile::modeCreate | CFile::modeWrite, &e))
		{
			Dump(output);
			output.Close();
		}
		else
		{
			e.ReportError();
		}
	}

	CleanUpVerticesList();
	delete m_pStreamDescr;  m_pStreamDescr = NULL;

	return true;
}

void CMGMesh::ExportField(unsigned int number, CStdioFile &file)
{
	// only positive numbers
	CString str;
	BYTE tmp1, tmp2;

	for (int i = 28; i > 0; i -= 7)
	{
		if (number >= (unsigned int) (1 << i))
		{
			tmp1 = (number >> i) & 0x7f;
			tmp1 |= 0x80;
			for (int j = 6; j >= 0; j -= 2)
			{
				tmp2 = (tmp1 >> j) & 0x3;
				str.Format("%d ", tmp2);
				file.WriteString(str);
			}
		}
	}

	tmp1 = number & 0x7f;
	for (int j = 6; j >= 0; j -= 2)
	{
		tmp2 = (tmp1 >> j) & 0x3;
		str.Format("%d ", tmp2);
		file.WriteString(str);
	}
}

void CMGMesh::ExportByteField(BYTE number, CStdioFile &file)
{
	CString str;
	for (int j = 6; j >= 0; j -= 2)
	{
		int tmp = (number >> j) & 0x3;
		str.Format(" %d", tmp);
		file.WriteString(str);
	}
}

void CMGMesh::ExportMeshConnectivity(int level, CStdioFile &file)
{
	// do not encode the mesh part if the automatic multi-resolution is possible
	if (level > 0 && m_pStreamDescr->IsAutoResolution())
		return;

	// start a new level description
	OnBeginMeshExportLevel(level);

	for (int roiIdx = 0; roiIdx < m_pStreamDescr->GetTotalNrROIs(); roiIdx++)
	{
		InitROI(level, 0, roiIdx);

		// store the number of vertices for ROI
		file.WriteString("\t\t\t");

		int nrVertices = ComputeNrVertices(level, 0);
		ExportField(nrVertices, file);

		file.WriteString("\n");

		if (nrVertices > 0)
		{
			// a ROI may consist of several sub-patches
			bool bFinish = false;
			while (!bFinish)
			{
				// find an appropriate starting point
				GetNextStartingPoint(level, 0);

				// first save the vertices' connectivities
				SaveMeshConnectivityROI(bFinish, file, level, roiIdx);
			}
		}
	}
}

int CMGMesh::ExportRefineMode(int level, CStdioFile &file)
{
	const MStreamLayer &meshLevel = m_pStreamDescr->GetMeshLayer(level, 0);
	int refineMode = meshLevel.GetRepositionBitsMode();

	CString str;
	str.Format("\t\t\t%d\n", refineMode);
	file.WriteString(str);

	return refineMode;
}

void CMGMesh::ExportMeshRefineBits(int level, CStdioFile &file)
{
	// reset the visited vertices
	ResetVisitedVerticesGrid(level, 0);

	// loop for all ROIs and store the refine bits
	for (int roiIdx = 0; roiIdx < m_pStreamDescr->GetTotalNrROIs(); roiIdx++)
	{
		InitROI(level, 0, roiIdx);

		// second save the refine bits
		SaveMeshRefine(level, roiIdx, file);
	}
}

void CMGMesh::ExportMeshRepositionBits(int level, CStdioFile &file)
{
	// reset the visited vertices
	ResetVisitedVerticesGrid(level, 0);

	// loop for all ROIs and store the refine bits
	for (int roiIdx = 0; roiIdx < m_pStreamDescr->GetTotalNrROIs(); roiIdx++)
	{
		InitROI(level, 0, roiIdx);

		// second save the refine bits
		SaveMeshReposition(level, roiIdx, file);
	}
}

/**********************************************************************************/

void CMGMesh::SaveMeshConnectivityROI(bool &bFinish, CStdioFile &file, int level, int roiIdx)
{
	CSurfNodeT *pSNStart = CheckStartingPoint(level, 0);
	if (!pSNStart)
	{
		bFinish = false;  return;
	}

	// we need the starting point when we refine the vertices
	CacheStartPoint(pSNStart, level, 0, roiIdx);

	// stream the position of the first node
	StreamStartNode(level, pSNStart, file);

	// write the node
	StreamNodeBox(level, pSNStart, file); 

	// check whether the entire box is streamed
	bFinish = IsROIStreamed(level, 0);
}

void CMGMesh::SaveMeshRefine(int level, int roiIdx, CStdioFile &file)
{
	// loop for all start vertices
	CPtrList &startNodeList = m_startVertsROI[level][0][roiIdx];
	while (startNodeList.GetCount() != 0)
	{
		CSurfNodeT *pSNStart = (CSurfNodeT *)startNodeList.RemoveHead();

		// write the offset
		StreamVerticesRefineBits(level, pSNStart, file);
	}
}

void CMGMesh::SaveMeshReposition(int level, int roiIdx, CStdioFile &file)
{
	// loop for all start vertices
	CPtrList &startNodeList = m_startVertsROI[level][0][roiIdx];
	while (startNodeList.GetCount() != 0)
	{
		CSurfNodeT *pSNStart = (CSurfNodeT *)startNodeList.RemoveHead();

		// write the reposition bit for the next level
		StreamVerticesRepositionBit(level, pSNStart, file);
	}
}

/**********************************************************************************/

void CMGMesh::StreamStartNode(int level, CSurfNodeT *pSNStart, CStdioFile &file)
{
	CIPointUVW pos3D = pSNStart->GetPos();
	FindPosNodeLocal(pos3D, pSNStart->GetDir(), level, 0);

	StreamAbsoluteNode(pos3D, pSNStart->GetDir(), file);
}

void CMGMesh::StreamNodeBox(int level, CSurfNodeT *pSNStart, CStdioFile &file)
{
	// add the starting vertex to the list
	AddStartNodeBox(pSNStart, true);

	while (m_nodeList.GetCount() != 0)
	{
		CSurfNodeT *pSN = (CSurfNodeT *)m_nodeList.RemoveHead();
		ASSERT(pSN->GetLevel() <= level);

		BYTE link1, link2;
		CSurfNodeT *pSNNext1 = pSN->GetNextNode1(level);
		CSurfNodeT *pSNNext2 = pSN->GetNextNode2(level);

		// write the Link1 connectivity vector
		if (pSNNext1 != NULL)
			GetDirectionDifference(pSN->GetDir(), pSNNext1->GetDir(), link1);
		else
			link1 = NO_LINK;

		// write the Link2 connectivity vector
		if (pSNNext2 != NULL)
			GetDirectionDifference(pSN->GetDir(), pSNNext2->GetDir(), link2);
		else
			link2 = NO_LINK;

		WriteConnectivitySymbol(link1, link2, file);

		AddNodesToList(pSNNext1, pSNNext2, level, true);
	}
}

void CMGMesh::StreamVerticesRefineBits(int level, CSurfNodeT *pSNStart, CStdioFile &file)
{
	// add the starting vertex to the list
	AddStartNodeBox(pSNStart);

	while (m_nodeList.GetCount() != 0)
	{
		CSurfNodeT *pSN = (CSurfNodeT *)m_nodeList.RemoveHead();
		ASSERT(pSN->GetLevel() <= level);

		// retrieve the offset
		float offset;
		if (level < GetMaxNrLevels() - 1 && !IsAutoResolution())
			offset = GetLevelOffset(pSN->GetDir(), pSN->GetPos(), pSN->GetOffset(), level);
		else
			offset = pSN->GetOffset();

		WriteOffset(offset, file);

		AddNodesToList(pSN->GetNextNode1(level), pSN->GetNextNode2(level), level, false);
	}
}

void CMGMesh::StreamVerticesRepositionBit(int level, CSurfNodeT *pSNStart, CStdioFile &file)
{
	// add the starting vertex to the list
	AddStartNodeBox(pSNStart);

	while (m_nodeList.GetCount() != 0)
	{
		CSurfNodeT *pSN = (CSurfNodeT *)m_nodeList.RemoveHead();
		ASSERT(pSN->GetLevel() <= level);

		bool bSpecialCase;
		float val = (float)VertexRepositionBit(level + 1, pSN, bSpecialCase);
		WriteOffset(val, file);

		AddNodesToList(pSN->GetNextNode1(level), pSN->GetNextNode2(level), level, false);
	}
}

/**********************************************************************************/

void CMGMesh::StreamAbsoluteNode(const CIPointUVW &pos3D, BYTE dir, CStdioFile &file)
{
	CString str;
	file.WriteString("\t\t\t");

	ExportField(pos3D.u, file);
	file.WriteString(",");

	ExportField(pos3D.v, file);
	file.WriteString(",");

	ExportField(pos3D.w, file);
	file.WriteString(",");

	ExportByteField(dir, file);
	file.WriteString("\n");
}

void CMGMesh::WriteConnectivitySymbol(int link1, int link2, CStdioFile &file)
{
	ASSERT (link1 >= 0 || link1 <= 2);
	ASSERT (link2 >= 0 || link2 <= 2);

	CString str;
	str.Format("\t\t\t%d %d\n", link1, link2);
	file.WriteString(str);
}

void CMGMesh::WriteOffset(float offset, CStdioFile &file)
{
	ASSERT(offset >= 0 && offset <= 1);

	CString str;
	str.Format("\t\t\t%f\n", offset);
	file.WriteString(str);
}
