/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGMeshCoderBinary.cpp: implementation of the CMGMesh class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGMesh.h"
#include "MGGridCodec.h"
#include "BitStream.h"
#include "MGStreamParse.h"
#include "MgErrMsg.h"

#define IINFINITY				0x7FFFFFFF

// Save as binary file
bool CMGMesh::Save(CFile *file, const CIPointUVW &nrROIs, bool hasHeader, bool hasConnectivityInfo, bool hasRefineInfo, bool hasRepositionInfo,
				   bool hasGridInfo, bool encodeMultiLayer, int nrRefinementBits, bool interleaveRefine, int nrQuabtizationBits, int filterType,
				   float fMAXAD, CString &error)
{
	bool res = true;
	try
	{
		// NOTE: This flag must be set according to the encoding mode!!!
		BYTE update_replace1 = REPLACE_DATA;
		BYTE update_replace2 = UPDATE_DATA;

		// create the parser object
		//CreateParser(NULL, CBitStream::OP_WRITE);
		CreateParser(file, CBitStream::OP_WRITE);

		// initialise the exchange object with the mesh coding details
		InitExchangeObjectMeshBinary(nrROIs, hasConnectivityInfo, hasRefineInfo, hasRepositionInfo, hasGridInfo, nrRefinementBits, interleaveRefine, encodeMultiLayer);

		// create and start the encoder object
		CreateAndStartEncoder(nrQuabtizationBits, filterType);

		// Find the best combination of truncation points
		// (the min. stream length which remains under the given MAXAD).
		if (hasGridInfo)
			ComputeOptimalTruncation(fMAXAD, filterType);

		// choose which layer to save (should be given as parameter)
		CIntArray layers;	layers.SetSize(GetNrLayers());
		for (int k = 0; k < GetNrLayers(); k++)
			layers[k] = k;

		m_pParser->SetActiveLayers(layers);

		// initialise the parser and store the header of the stream
		StoreMeshGridHeader(hasHeader);

		// stream the levels
		for (int level = 0; level < GetMaxNrLevels(); level++)
		{
			// store a new frame
			StoreBeginFrame();
 
			// store the mesh coding for a level
			StoreMGMesh(level, update_replace1);

			// store the grid coding for a level
			StoreMGGrid(level, update_replace2);
		}
		StopEncoder();
	}
	catch (CMGDataException &e)
	{
		res = false;    error = e.GetErrorMessage();
	}

	return res;
}

/**********************************************************************************/

void CMGMesh::StoreMeshGridHeader(bool keepHeader)
{
	m_pParser->StoreMeshGridHeader(keepHeader);
}

void CMGMesh::StoreBeginFrame()
{
	m_pParser->StoreBeginFrame();
}

void CMGMesh::StoreGridCorners(bool differential)
{
	// get the grid corners, scale snd store them and compute the quanitisation bits
	BYTE nrQuantBits = 0;
	m_grid->RetrieveGridCorners(m_pStreamDescr->GetGridCodingParam(), differential, nrQuantBits);

	m_pParser->StoreGridCorners(differential, nrQuantBits);
}

void CMGMesh::WriteOutput(CFile *file)
{
	m_pParser->WriteOutput(file);
}

void CMGMesh::DoConnectivityCoding(int level, BYTE layer, int roiIdx)
{
	if (!m_pStreamDescr->IsConnectivityInfoAvailable(level))
		return;

	// begin of connectivity coding
	OnBeginMeshCodingConnectivityROI(layer);

	// a ROI may consist of several sub-patches
	bool bFinish = false;
	while (!bFinish)
	{
		// find an appropriate starting point
		GetNextStartingPoint(level, layer);

		// first save the vertices' connectivities
		SaveMeshConnectivityROI(bFinish, level, layer, roiIdx);
	}

	// end of mesh connectivity coding
	OnEndMeshCodingConnectivityROI(level, layer, roiIdx);
}

void CMGMesh::DoVertsReposCoding(int level, BYTE layer, int roiIdx)
{
	if (!m_pStreamDescr->IsRepositionInfoAvailable(level, layer))
		return;

	// begin of vertices' reposition bits coding
	OnBeginMeshCodingVertsReposROI(layer);

	// second save the reposition bits
	SaveMeshReposROI(level, layer, roiIdx);

	// end of vertices' reposition bits coding
	OnEndMeshCodingVertsReposROI(level, layer, roiIdx);
}

void CMGMesh::DoVertsRefineCoding(int level, BYTE layer, int roiIdx)
{
	if (!m_pStreamDescr->IsRefinementInfoAvailable(level))
		return;

	// begin of vertices' reposition bits coding
	OnBeginMeshCodingVertsRefineROI(layer);

	// second save the refine bits
	SaveMeshRefineROI(level, layer, roiIdx);

	// end of vertices' reposition bits coding
	OnEndMeshCodingVertsRefineROI(level, layer, roiIdx);
}

void CMGMesh::StoreMGMesh(int level, BYTE update_replace)
{
	// start a new level description
	OnBeginMeshCodingLevel(level);

	const CIntArray &layers = m_pParser->GetLayers();
	for (int k = 0; k < layers.GetSize(); k++)
	{
		int layer = layers[k];

		// start a new layer description
		OnBeginMeshCodingLayer(level, layer);

		MStreamLayer &meshLevel = m_pStreamDescr->GetMeshLayer(level, layer);
		for (int roiIdx = 0; roiIdx < meshLevel.GetTotalNrROIs(); roiIdx++)
		{
			InitROI(level, layer, roiIdx);

			// start a new ROI description
			OnBeginMeshCodingROI(level, layer, roiIdx);

			bool bEmpty = IsROIEmpty(level, layer);
			if (!bEmpty)
			{
				// do the connectivity, vertices' reposition/refine coding
				DoConnectivityCoding(level, layer, roiIdx);
				DoVertsRefineCoding(level, layer, roiIdx);
				DoVertsReposCoding(level, layer, roiIdx);

				CleanUpVerticesList(level, layer, roiIdx);
			}

			// end of ROI description
			OnEndMeshCodingROI(bEmpty, level, layer, roiIdx);
		}

		// end of layer description
		OnEndMeshCodingLayer(level, layer);
	}

	// end of level description
	OnEndMeshCodingLevel(level, update_replace);
}

void CMGMesh::StoreMGGrid(int level, BYTE update_replace)
{
	if (m_pStreamDescr->IsGridInfoAvailable(level))
		m_pMGOUT->StreamLevel(level, update_replace);
}

void CMGMesh::ComputeOptimalTruncation(double MAXAD, int nFilterType)
{
	int dir;
	if (MAXAD <= 0)
		return; // store the grid loss-less

	CCodec3D<CODEC_TYPE>		*pCodec = m_pMGOUT->GetCodec();
	CWaveletCodec<CODEC_TYPE>	*pPrecode = m_pMGOUT->GetPrecode();
	CMatrix<int>* orgCoeffQ = m_pMGOUT->GetCoeff();
	CMatrix<int> modCoeffQ[3];
	CMatrix<CODEC_TYPE> modCoeffW[3];
	CMatrix<CODEC_TYPE> orgCoeff[3];
	CMatrix<CODEC_TYPE> modCoeff[3];
	for (dir = X_GRID; dir <= Z_GRID; dir++)
	{
		modCoeffQ[dir].Init(orgCoeffQ[dir].GetDim1(), orgCoeffQ[dir].GetDim2(), orgCoeffQ[dir].GetDim3());
		modCoeffW[dir].Init(orgCoeffQ[dir].GetDim1(), orgCoeffQ[dir].GetDim2(), orgCoeffQ[dir].GetDim3());
		orgCoeff[dir].Init(orgCoeffQ[dir].GetDim1(), orgCoeffQ[dir].GetDim2(), orgCoeffQ[dir].GetDim3());
		modCoeff[dir].Init(orgCoeffQ[dir].GetDim1(), orgCoeffQ[dir].GetDim2(), orgCoeffQ[dir].GetDim3());
	}
	m_pMGOUT->ExtractCoordinates(orgCoeff[X_GRID], orgCoeff[Y_GRID], orgCoeff[Z_GRID]);

	// Compute absolut value of MAXAD
	CSurrBoxT gridBox;
	m_grid->GetSurrBox(gridBox);
	double dDiagonal = sqrt(gridBox.GetWidth()*gridBox.GetWidth() + gridBox.GetHeight()*gridBox.GetHeight() + gridBox.GetDepth()*gridBox.GetDepth());
	MAXAD *= dDiagonal / 100.0;

	// Initialize the lists
	CDoubleList arrLambdas;
	CTruncations* arrTruncationPointsList[3];
	for (dir = X_GRID; dir <= Z_GRID; dir++)
		arrTruncationPointsList[dir] = new CTruncations[m_pMGOUT->GetMaxNrLevelsG()];

	//Search the optimal cutting thresholds based on the theoretical MAXAD
	SearchTruncationThresholds(orgCoeff, orgCoeffQ, modCoeff, modCoeffQ, modCoeffW, MAXAD, dDiagonal, nFilterType, arrTruncationPointsList, arrLambdas);

	//Set the stream sizes
	for (int level = 0; level < m_pMGOUT->GetMaxNrLevelsG(); level++)
	{
		GStreamLevel &gridLevel = m_pStreamDescr->GetGridLevel(level);

		for (int dir = X_GRID;  dir <= Z_GRID; dir++)
		{
			gridLevel.SetStopThreshold(dir, GetStopThreshold(arrTruncationPointsList[dir][level], gridLevel.GetThreshold(dir)));
		}
	}

	//Clean mem.
	for (dir = X_GRID; dir <= Z_GRID; dir++)
		delete[] arrTruncationPointsList[dir];
}

void CMGMesh::ComputeTruncationPointsList(int nFilterType, CTruncations* arrTruncationPointsList[], CDoubleList &arrLambdas)
{
	CCodec3D<CODEC_TYPE> *pCodec = m_pMGOUT->GetCodec();

	arrLambdas.Empty();

	// Compute the list of {treshold, size} for each subband (level)
	for (int level = 0; level < m_pMGOUT->GetMaxNrLevelsG(); level++)
	{
		if (!m_pStreamDescr->IsGridInfoAvailable(level))
		{
			ASSERT(0);
			continue;
		}

		const CTileArray &tiles = (m_pMGOUT->GetSubbandTiles())[level];
		GStreamLevel &gridLevel = m_pStreamDescr->GetGridLevel(level);

		// encode the (x, y, z) grid coordinates
		for (int dir = X_GRID;  dir <= Z_GRID; dir++)
		{
			CTruncationPoints &arrTruncationPoints = arrTruncationPointsList[dir][level].points;
			arrTruncationPoints.RemoveAll();
			pCodec->GetData() = &(m_pMGOUT->GetCoeff(dir));
			for (int tileIdx = 0; tileIdx < tiles.GetSize(); tileIdx++)
			{
				const CCubeArray &cubes = tiles[tileIdx];

				// perform the coding of the Tile
				bool bEmpty = true;
				for (int i = 0; i < cubes.GetSize(); i++)
				{
					if (!pCodec->IsEmpty(cubes[i]))
					{
						bEmpty = false;
						break;
					}
				}
				if (bEmpty) continue;

				int threshold = gridLevel.GetThreshold(dir);
				while (threshold >= 1)
				{
					STruncationPoint &trPoint = arrTruncationPoints.GetPoint(threshold);
					// Compute the length of the stream added by this threshold
					for (int i = 0; i < cubes.GetSize(); i++)
					{
						int previous_threshold = threshold << 1;

						CCubeQueue queue;
						queue.AddHead(cubes[i]);

						while (!queue.IsEmpty())
						{
							const Cube &cube = queue.RemoveHead();
							if (cube.width != 0 && cube.height != 0 && cube.depth != 0)
							{	
								if (!pCodec->SignificantCube(cube, previous_threshold))
								{
									trPoint.length++;
									if (pCodec->SignificantCube(cube, threshold))
									{
										if (cube.IsPixel())
										{
											trPoint.length++;
										}
										else
										{
											addChildren(queue, cube);
										}
									}
								}
								else
								{
									// the block is already marked as significant from a previous threshold test
									if (cube.IsPixel())
									{
										trPoint.length++;
									}
									else
									{
										addChildren(queue, cube);
									}
								}
							}
						}
					}

					// Compute the distortion introduced by this truncation point
					if (trPoint.distortion == -1.0) //Only if it was not already computed at an earlier tile
						trPoint.distortion = GetDDistortion(level, m_pMGOUT->GetMaxNrLevelsG(), threshold, m_pStreamDescr->GetGridCodingParam(dir).scale, nFilterType);

					threshold >>= 1;
				}
			}
			pCodec->GetData() = NULL;

			// Compute the list of feasible truncation points
			ComputeFeasibleTruncationPoints(arrTruncationPoints);

			//Add to the list of lambdas
			for (int i=arrTruncationPoints.GetSize()-1; i>=0; i--)
			{
				arrLambdas.InsertSort(arrTruncationPoints[i].lambda);
			}

			arrTruncationPointsList[dir][level].truncation = arrTruncationPointsList[dir][level].points.GetSize()-1;
		}
	}
}

void CMGMesh::SearchTruncationThresholds(CMatrix<CODEC_TYPE> orgCoeff[], CMatrix<int> orgCoeffQ[], CMatrix<CODEC_TYPE> modCoeff[], CMatrix<int> modCoeffQ[], CMatrix<CODEC_TYPE> modCoeffW[], double MAXAD, double scaleMAXAD, int nFilterType, CTruncations* arrTruncationPointsList[], CDoubleList &arrLambdas)
{
	//Search the optimal cutting thresholds based on the theoretical MAXAD.
	// Compute the list of {treshold, size} for each subband (level)
	// and the ordered list with all lambdas.
	ComputeTruncationPointsList(nFilterType, arrTruncationPointsList, arrLambdas);

	//Find the optimal point for truncation in order to fulfill the MAXAD limitation
	double MAXADs[3] = {MAXAD, MAXAD, MAXAD};

	//Search the optimal points
	for (int search=0; search<arrLambdas.GetSize(); search++)
	{
		double lambda = arrLambdas[search];
		for (int dir=0; dir<3; dir++)
		{
			if (MAXADs[dir] <= -0.0)
				continue;
			for (int level = m_pMGOUT->GetMaxNrLevelsG()-1; level >= 0; level--)
			{
				CTruncations &arrTruncations = arrTruncationPointsList[dir][level];
				for (int i=arrTruncations.truncation; i>=0; i--)
				{
					//Exclud all the points till the first lambda bigger then the searched one.
					if (arrTruncations.points[i].lambda > lambda)
						break;
					arrTruncations.truncation = i-1;
					MAXADs[dir] -= arrTruncations.points[i].distortion;
					if (MAXADs[dir] <= -0.0)
					{
						arrTruncations.truncation = i;
						break;
					}
				}
			}
		}
	}

	//Continue the search computing the real MAXADat eac step to check is the limit was reached.
	FindTruncationPoints(orgCoeff, orgCoeffQ, modCoeff, modCoeffQ, modCoeffW, MAXAD, scaleMAXAD, arrTruncationPointsList, arrLambdas);
}

// "level" should be between 0 and N-1;
// "maxLevels" should be between 1 and N;
// "threshold" is the last threshold that is encoded;
// returns the delta distortion corresponding to not encoding this threshold.
double CMGMesh::GetDDistortion(int level, int maxLevels, int threshold, int scale, int filterType) const
{
	double distortion;
#ifdef DECODE_AHEAD
	threshold = (threshold+1) / 2;
#endif
	// If this is the fisrt level that includes also the LLL subband,
	// then add also its contribution.
	if (filterType == CMGGrid::SMOOTH_FILTER)
	{
		distortion = (604.0/64.0) * pow(125.0/64.0, maxLevels-1-level) * (double)threshold;
		if (level == 0)
			distortion += pow(125.0/64.0, maxLevels) * (double)threshold;
	}
	else //CMGGrid::SHORT_FILTER
	{
		if (level == 0)
			distortion = 8.0 * (double)threshold;
		else
			distortion = 7.0 * (double)threshold;
	}
	return distortion / scale;
}

void CMGMesh::ComputeFeasibleTruncationPoints(CTruncationPoints &arrTruncationPoints)
{
	int i;
	for (i = 0; i < arrTruncationPoints.GetSize(); i++)
		if (arrTruncationPoints[i].distortion > 0)
			break;
	if (i == arrTruncationPoints.GetSize())
	{
		arrTruncationPoints.RemoveAll();
		return;
	}
	CList<int, int&> arrFeasiblePoints;
	arrFeasiblePoints.AddHead(i);
	arrTruncationPoints[i].lambda = arrTruncationPoints[i].distortion / arrTruncationPoints[i].length;
	for (i++; i < arrTruncationPoints.GetSize(); i++)
	{
		STruncationPoint &trPoint = arrTruncationPoints[i];
		if (trPoint.distortion <= 0)
		{
			//add the length and distortion to the previous threshold
			int index = (i+2 < arrTruncationPoints.GetSize() ? i+1 : arrFeasiblePoints.GetTail());
			STruncationPoint &trLastPoint = arrTruncationPoints[index];
			trLastPoint.distortion += trPoint.distortion;
			trLastPoint.length += trPoint.length;
			trLastPoint.lambda = trLastPoint.distortion / trLastPoint.length;
			continue;
		}
		while (arrFeasiblePoints.GetCount() &&
			trPoint.distortion >= arrTruncationPoints[arrFeasiblePoints.GetTail()].lambda * trPoint.length)
		{
			STruncationPoint &trLastPoint = arrTruncationPoints[arrFeasiblePoints.GetTail()];
			trPoint.distortion += trLastPoint.distortion;
			trPoint.length += trLastPoint.length;
			arrFeasiblePoints.RemoveTail();
		}
		arrFeasiblePoints.AddTail(i);
		trPoint.lambda = trPoint.distortion/trPoint.length;
	}

	// Remove the not feasible points
	POSITION pos = arrFeasiblePoints.GetTailPosition();
	int lastPoint = arrFeasiblePoints.GetPrev(pos);
	if (arrTruncationPoints.GetSize() > lastPoint+1)
		arrTruncationPoints.RemoveAt(lastPoint+1, arrTruncationPoints.GetSize() - lastPoint-1);
	while (pos)
	{
		int currentPoint = arrFeasiblePoints.GetPrev(pos);
		if (lastPoint - currentPoint > 1)
			arrTruncationPoints.RemoveAt(currentPoint+1, lastPoint - currentPoint - 1);
		lastPoint = currentPoint;
	}
	if (lastPoint > 0)
		arrTruncationPoints.RemoveAt(0, lastPoint);
}

void CMGMesh::FindTruncationPoints(CMatrix<CODEC_TYPE> orgCoeff[], CMatrix<int> orgCoeffQ[], CMatrix<CODEC_TYPE> modCoeff[], CMatrix<int> modCoeffQ[], CMatrix<CODEC_TYPE> modCoeffW[], double MAXAD, double scaleMAXAD, CTruncations* arrTruncationPointsList[], CDoubleList &arrLambdas)
{
	//Search the optimal points
	int dir;
	double realMAXAD[3], std[3], MSE[3];
	int* nEnd[3];
	for (dir=0; dir<3; dir++)
	{
		nEnd[dir] = new int[m_pMGOUT->GetMaxNrLevelsG()];
		for (int level = m_pMGOUT->GetMaxNrLevelsG()-1; level >= 0; level--)
		{
			nEnd[dir][level] = 1;
		}
	}
	for (int search=0; search<arrLambdas.GetSize(); search++)
	{
		double lambda = arrLambdas[search];
		for (int dir=0; dir<3; dir++)
		{
			for (int level = m_pMGOUT->GetMaxNrLevelsG()-1; level >= 0; level--)
			{
				if (nEnd[dir][level] <= 0)
					continue;
				CTruncations &arrTruncation = arrTruncationPointsList[dir][level];
				for (int i=arrTruncation.truncation; i>=0; i--)
				{
					//Exclud all the points till the first lambda bigger then the searched one.
					if (arrTruncation.points[i].lambda > lambda)
						break;
					arrTruncation.truncation = i-1;
					GetRealMAXAD(orgCoeff, orgCoeffQ, modCoeff, modCoeffQ, modCoeffW, arrTruncationPointsList, std, MSE, realMAXAD);
					if (MAXAD < realMAXAD[dir])
					{
						arrTruncation.truncation = i;
						nEnd[dir][level]--;
						break;
					}
				}
			}
		}
	}
	for (dir=0; dir<3; dir++)
	{
		delete[] nEnd[dir];
	}

	// Extend the search by trying all the left points (it's posible that cutting a level not to introduce additional distortion or even to decrease it).
	ExtraSearch(orgCoeff, orgCoeffQ, modCoeff, modCoeffQ, modCoeffW, MAXAD, scaleMAXAD, arrTruncationPointsList);
}

void CMGMesh::ExtraSearch(CMatrix<CODEC_TYPE> orgCoeff[], CMatrix<int> orgCoeffQ[], CMatrix<CODEC_TYPE> modCoeff[], CMatrix<int> modCoeffQ[], CMatrix<CODEC_TYPE> modCoeffW[], double MAXAD, double scaleMAXAD, CTruncations* arrTruncationPointsList[])
{
	CCodec3D<CODEC_TYPE>		*pCodec = m_pMGOUT->GetCodec();
	CWaveletCodec<CODEC_TYPE>	*pPrecode = m_pMGOUT->GetPrecode();

	//Search the optimal points
	int dir;
	for (int levelTry = m_pMGOUT->GetMaxNrLevelsG()-1; levelTry >= 0; levelTry--)
	{
		int* maxThresholds = m_pStreamDescr->GetGridLevel(levelTry).GetThreshold();
		int minThresholds[3];
		bool orgThresholds[3] = {true, true, true};
		int maxThreshold = 0;
		int minThreshold = IINFINITY;
		for (dir = X_GRID; dir <= Z_GRID; dir++)
		{
			if (maxThresholds[dir] > maxThreshold)
				maxThreshold = maxThresholds[dir];
			minThresholds[dir] = GetStopThreshold(arrTruncationPointsList[dir][levelTry], maxThresholds[dir]);
			if (minThresholds[dir] < minThreshold)
				minThreshold = minThresholds[dir];
		}
		for (int threshold = maxThreshold; threshold >= minThreshold; threshold>>=1)
		{
			// Compute the real MAXAD.
			for (dir = X_GRID; dir <= Z_GRID; dir++)
			{
				modCoeffQ[dir] = orgCoeffQ[dir];

				// Add the possible distortion of the threshold cutting
				for (int level = 0; level < m_pMGOUT->GetMaxNrLevelsG(); level++)
				{
					if (!m_pStreamDescr->IsGridInfoAvailable(level))
						continue;

					GStreamLevel &gridLevel = m_pStreamDescr->GetGridLevel(level);
					const CTileArray &tiles = (m_pMGOUT->GetSubbandTiles())[level];
					int nTruncation;
					if (level == levelTry)
					{
						if (threshold < minThresholds[dir] ||
							threshold > maxThresholds[dir])
							break;
						nTruncation = threshold << 1;
					}
					else
					{
						nTruncation = GetStopThreshold(arrTruncationPointsList[dir][level], gridLevel.GetThreshold(dir));
						if (nTruncation <= 1)
							continue;
					}
					for (int tileIdx = 0; tileIdx < tiles.GetSize(); tileIdx++)
					{
						const CCubeArray &cubes = tiles[tileIdx];

						for (int i = 0; i < cubes.GetSize(); i++)
						{
							EstimateError(modCoeffQ[dir], cubes[i], nTruncation);
						}
					}
				}
			}

			// Reverse quantize transform.
			InversePreProcess(modCoeffQ, modCoeffW);

			// Reverse wavelet transform.
			for (dir = X_GRID;  dir <= Z_GRID; dir++)
				pCodec->RetrieveDC(modCoeffW[dir], m_pStreamDescr->GetGridCodingParam(dir));
			pPrecode->WaveletRec3D(modCoeffW[X_GRID], modCoeffW[Y_GRID], modCoeffW[Z_GRID], modCoeff[X_GRID], modCoeff[Y_GRID], modCoeff[Z_GRID]);

			// Compute the standard deviation.
			double std[3], MSE[3], mmm[3];
			StandardDeviation(orgCoeff, modCoeff, std, MSE, mmm);

			for (dir=0; dir<3; dir++)
			{
				if (threshold < minThresholds[dir] ||
					threshold > maxThresholds[dir] ||
					mmm[dir] > MAXAD)
					continue;
				orgThresholds[dir] = false;
				minThresholds[dir] = threshold;
				minThreshold = IINFINITY;
				if (minThresholds[0] < minThreshold)
					minThreshold = minThresholds[0];
				if (minThresholds[1] < minThreshold)
					minThreshold = minThresholds[1];
				if (minThresholds[2] < minThreshold)
					minThreshold = minThresholds[2];
			}
		}

		// Set the new thresholds
		for (dir=0; dir<3; dir++)
		{
			if (orgThresholds[dir])
				continue;
			if (minThresholds[dir] == maxThresholds[dir])
			{
				arrTruncationPointsList[dir][levelTry].truncation = -1;
				continue;
			}
			minThresholds[dir] <<= 1;
			CTruncations &arrTruncations = arrTruncationPointsList[dir][levelTry];
			int i;
			for (i=0; i < arrTruncations.points.GetSize(); i++)
			{
				if (arrTruncations.points[i].threshold <= minThresholds[dir])
					break;
			}
			if (i == arrTruncations.points.GetSize() || arrTruncations.points[i].threshold < minThresholds[dir])
				arrTruncations.points[arrTruncations.truncation].threshold = minThresholds[dir];
			else
				arrTruncations.truncation = i;
		}
	}
}

int CMGMesh::GetStopThreshold(CTruncations &arrTruncations, int gridLevelThreshold, int deltaDomain)
{
	int nStopThreshold = arrTruncations.truncation;
	if (nStopThreshold >= 0) //return the stop threshold (power of 2)
	{
		nStopThreshold -= deltaDomain;
		if (nStopThreshold >= 0 && nStopThreshold < arrTruncations.points.GetSize())
		{
			nStopThreshold = arrTruncations.points[nStopThreshold].threshold;
		}
		else if (nStopThreshold == -1) //cut all
		{
			ASSERT(gridLevelThreshold > 0);
			nStopThreshold = gridLevelThreshold << 1;
		}
		else
		{
			nStopThreshold = IINFINITY;
		}
	}
	else if (gridLevelThreshold > 0 && deltaDomain == 0) //cut all
	{
		nStopThreshold = gridLevelThreshold << 1;
	}
	else
	{
		nStopThreshold = IINFINITY;
	}
	return nStopThreshold;
}

void CMGMesh::InversePreProcess(CMatrix<int> matIn[], CMatrix<CODEC_TYPE> matOut[])
{
	for (int level = 0; level < m_pMGOUT->GetMaxNrLevelsG(); level++)
	{
		if (!m_pStreamDescr->IsGridInfoAvailable(level))
			continue;

		const CTileArray &tiles = (m_pMGOUT->GetSubbandTiles())[level];

		// Reverse quantize
		for (int dir = X_GRID;  dir <= Z_GRID; dir++)
		{
			const int nScale = m_pStreamDescr->GetGridCodingParam(dir).scale;
			for (int tileIdx = 0; tileIdx < tiles.GetSize(); tileIdx++)
			{
				const CCubeArray &cubes = tiles[tileIdx];
				for (int l = 0; l < cubes.GetSize(); l++)
				{
					const Cube &cube = cubes[l];
					for (int k = 0; k < cube.depth; k++)
					{
						int z = cube.w + k;
						int  **sliceIn = matIn[dir][z];
						CODEC_TYPE **sliceOut = matOut[dir][z];

						for (int j = 0; j < cube.height; j++)
						{
							int y = cube.v + j;
							int  *lineIn  = sliceIn[y];
							CODEC_TYPE *lineOut = sliceOut[y];

							for (int i = 0; i < cube.width; i++)
							{
								int x = cube.u + i;
								lineOut[x] = (CODEC_TYPE)(lineIn[x]) / nScale;
							}
						}
					}
				}
			}
		}
	}
}

void CMGMesh::StandardDeviation(CMatrix<CODEC_TYPE> mats1[], CMatrix<CODEC_TYPE> mats2[], double std[], double MSE[], double mmm[])
{
	for (int dir = X_GRID;  dir <= Z_GRID; dir++)
	{
		CMatrix<CODEC_TYPE> mat1 = mats1[dir];
		CMatrix<CODEC_TYPE> mat2 = mats2[dir];
		double sumDiff = 0;
		double &sumDiffSquare = MSE[dir] = 0;
		double &MAXAD = mmm[dir] = 0;
		for (int z = 0; z < mat1.GetDim3(); z++)
		{
			CODEC_TYPE **slice1 = mat1[z];
			CODEC_TYPE **slice2 = mat2[z];
			for (int y = 0; y < mat1.GetDim2(); y++)
			{
				CODEC_TYPE *line1  = slice1[y];
				CODEC_TYPE *line2  = slice2[y];
				for (int x = 0; x < mat1.GetDim1(); x++)
				{
					double diff = line1[x] - line2[x];
					sumDiff += diff;
					sumDiffSquare += diff*diff;
					diff = fabs(diff);
					if (MAXAD < diff)
						MAXAD = diff;
				}
			}
		}
		const double N = mat1.GetDim1()*mat1.GetDim2()*mat1.GetDim3();
		std[dir] = sqrt((sumDiffSquare-(sumDiff*sumDiff)/N)/(N-1));
		sumDiffSquare /= N;
	}
}

void CMGMesh::GetRealMAXAD(CMatrix<CODEC_TYPE> orgCoeff[], CMatrix<int> orgCoeffQ[], CMatrix<CODEC_TYPE> modCoeff[], CMatrix<int> modCoeffQ[], CMatrix<CODEC_TYPE> modCoeffW[], CTruncations* arrTruncationPointsList[], double std[], double MSE[], double realMAXAD[], int compLevel)
{
	CCodec3D<CODEC_TYPE>		*pCodec = m_pMGOUT->GetCodec();
	CWaveletCodec<CODEC_TYPE>	*pPrecode = m_pMGOUT->GetPrecode();

	// Compute the real MAXAD.
	int dir;
	for (dir = X_GRID;  dir <= Z_GRID; dir++)
	{
		modCoeffQ[dir] = orgCoeffQ[dir];

		// Add the possible distortion of the threshold cutting
		for (int level = 0; level < m_pMGOUT->GetMaxNrLevelsG(); level++)
		{
			if (!m_pStreamDescr->IsGridInfoAvailable(level) ||
				(compLevel >= 0 && compLevel != level))
				continue;

			GStreamLevel &gridLevel = m_pStreamDescr->GetGridLevel(level);
			const CTileArray &tiles = (m_pMGOUT->GetSubbandTiles())[level];
			const int nTruncation = GetStopThreshold(arrTruncationPointsList[dir][level], gridLevel.GetThreshold(dir));
			if (nTruncation <= 1)
				continue;
			for (int tileIdx = 0; tileIdx < tiles.GetSize(); tileIdx++)
			{
				const CCubeArray &cubes = tiles[tileIdx];

				for (int i = 0; i < cubes.GetSize(); i++)
				{
					EstimateError(modCoeffQ[dir], cubes[i], nTruncation);
				}
			}
		}
	}

	// Reverse quantize transform.
	InversePreProcess(modCoeffQ, modCoeffW);

	// Reverse wavelet transform.
	for (dir = X_GRID;  dir <= Z_GRID; dir++)
		pCodec->RetrieveDC(modCoeffW[dir], m_pStreamDescr->GetGridCodingParam(dir));
	pPrecode->WaveletRec3D(modCoeffW[X_GRID], modCoeffW[Y_GRID], modCoeffW[Z_GRID], modCoeff[X_GRID], modCoeff[Y_GRID], modCoeff[Z_GRID]);

	// Compute the standard deviation.
	StandardDeviation(orgCoeff, modCoeff, std, MSE, realMAXAD);
}

void CMGMesh::GetLength(CTruncations* arrTruncationPointsList[], int length[])
{
	//Sum up the lengths.
	for (int dir = X_GRID;  dir <= Z_GRID; dir++)
	{
		length[dir] = 0;

		// Add the length till the threshold cutting
		for (int level = 0; level < m_pMGOUT->GetMaxNrLevelsG(); level++)
		{
			if (!m_pStreamDescr->IsGridInfoAvailable(level))
				continue;

			GStreamLevel &gridLevel = m_pStreamDescr->GetGridLevel(level);
			CTruncations &arrTruncations = arrTruncationPointsList[dir][level];
			for (int threshold = arrTruncations.truncation; threshold >= 0; threshold--)
			{
				length[dir] += arrTruncations.points[threshold].length;
			}
		}
	}
}

void CMGMesh::EstimateError(CMatrix<int> &matData, const Cube &cube, int stopThreshold)
{
	for (int k = 0; k < cube.depth; k++)
	{
		int **slice = matData[cube.w + k];
		for (int j = 0; j < cube.height; j++)
		{
			int *line = slice[cube.v + j];
			for (int i = 0; i < cube.width; i++)
			{
				int &val = line[cube.u + i];
#ifdef DECODE_AHEAD
				const int sign = (val < 0 ? -1 : 1);
				if ((val *= sign) < stopThreshold)
				{
					val = 0;
				}
				else
				{
					for (int threshold = 1; threshold < stopThreshold; threshold <<= 1)
					{
						if (val & threshold)
							val -= threshold >> 1;
						else
							val += (threshold + 1) >> 1;
					}
					val *= sign;
				}
#else
				val -= val%stopThreshold;
#endif
			}
		}
	}
}

/**********************************************************************************/

void CMGMesh::CreateParser(CFile *file, CBitStream::BITSTREAM_OPERATIONS op)
{
	TRY
	{
		m_pParser = new CMGStreamParser(file, op);
	}
	CATCH_ALL(e)
	{
		throw CMGDataException(e);
	}
	END_CATCH_ALL
}

void CMGMesh::CreateAndStartEncoder(int quantizationBits, int filterType)
{
	TRY
	{
		m_pMGOUT = new CMGGridCoder<CODEC_TYPE>(m_pParser, m_grid, quantizationBits, filterType);
		m_pMGOUT->StreamHeader();
	}
	CATCH_ALL(e)
	{
		throw CMGDataException(e);
	}
	END_CATCH_ALL
}

void CMGMesh::StopEncoder()
{
	delete []m_startVertsROI;   m_startVertsROI = NULL;

	delete m_pParser;   m_pParser = NULL;
	delete m_pMGOUT;	m_pMGOUT = NULL;
}

/**********************************************************************************/

void CMGMesh::InitExchangeObjectMeshBinary(const CIPointUVW &nrROIs, bool hasConnectivityInfo,
										   bool hasRefineInfo, bool hasRepositionInfo, bool hasGridInfo,
										   int refinmentBits, bool interleaveRefine, bool encodeMultiLayer)
{
	TRY
	{
		int nrLevels = GetMaxNrLevels();

		// start vertex lookup
		ASSERT(m_startVertsROI == NULL);
		m_startVertsROI = new CPtrList**[nrLevels];

		// initialise the exchange objects
		m_pStreamDescr = &m_pParser->m_streamDescr;
		m_pStreamDescr->Init(m_grid->GetNrLevelsUVW(), m_nrSlices, nrROIs); 

		// mesh type
		m_pStreamDescr->SetMeshType(GetMeshType());
		if (GetMeshType() == QUADRI_MESH)
		{
			if (hasConnectivityInfo)
			{
				m_pStreamDescr->SetSameConnectivity(IsSameConnectivity());
				m_pStreamDescr->SetAutoResolution(IsAutoResolution());
			}
			else
			{
				m_pStreamDescr->SetSameConnectivity(false);
				m_pStreamDescr->SetAutoResolution(false);
			}

			const CIPointUVW & nrSlices = GetNrSlices();
			if (m_pStreamDescr->IsSingleLayered())
			{
				m_pStreamDescr->SetOffsetAmplitude(GetOffsetAmplitude());
				if (GetOffsetAmplitude() == 0)
				{
					hasRefineInfo = 0; // no offsets
				}
			}

			if (IsAutoResolution())
			{
				// check for consistency
				ASSERT(IsSameConnectivity());
				if (hasRefineInfo)  
					interleaveRefine = true;       // need to set it because only the first
			}                                  // resolution level is available
		}

		// store the flags
		m_pStreamDescr->HasConnectivityInfo(hasConnectivityInfo);
		m_pStreamDescr->HasRefineInfo(hasRefineInfo);
		m_pStreamDescr->HasRepositionInfo(hasRepositionInfo);
		m_pStreamDescr->HasGridInfo(hasGridInfo);

		// prepare the layers
		BYTE nrBits = (GetNrLayers() > 1) ? CMGStreamParser::GetPower2(GetNrLayers()): 0;
		m_pStreamDescr->SetNrBitsLayer(nrBits);
		m_pStreamDescr->InitMeshLayers(GetNrLayers());
		m_pParser->InitStreams(GetNrLayers());

		// refinement info
		if (hasRefineInfo)
		{
			ASSERT(refinmentBits > 0);

			m_bInterleaveRefine = interleaveRefine;
			m_nrRefBitsVert = refinmentBits;
		}
		else
		{
			ASSERT(interleaveRefine == false);
			ASSERT(refinmentBits == 0);

			m_bInterleaveRefine = false;
			m_nrRefBitsVert = 0;
		}

		m_pStreamDescr->SetInterleaveRefine(m_bInterleaveRefine);
		m_pStreamDescr->SetNrRefineBitsVertex(m_nrRefBitsVert);
		m_pStreamDescr->SetCyclicMode(GetCyclicMode());

		// intialise the level information
		for (int level = 0; level < nrLevels; level++)
		{
			CIPointUVW levelDiff;
			m_grid->ComputeLevelDiffDown(levelDiff, level);

			CIPointUVW nrROIs;
			ROIDescription &roiDescr = m_pStreamDescr->GetROIDescr();
			roiDescr.ComputeNrROIs(nrROIs, levelDiff);

			for (int k = 0; k < m_pStreamDescr->GetNrLayers(); k++)
			{
				MStreamLayer &meshLevel = m_pStreamDescr->GetMeshLayer(level, k);
				meshLevel.Init(GetNrSlices(level), nrROIs, roiDescr.GetMaxROISize(), levelDiff);
			}
		}
	}
	CATCH_ALL(e)
	{
		throw CMGDataException(e);
	}
	END_CATCH_ALL
}

void CMGMesh::InitExchangeObjectMesh(bool hasConnectivityInfo, bool hasRefineInfo, bool hasRepositionInfo, bool interleaveRefine)
{
	TRY
	{
		int level;
		int nrLevels = GetMaxNrLevels();

		ASSERT(m_startVertsROI == NULL);
		m_startVertsROI = new CPtrList**[nrLevels];

		for (level = 0; level < nrLevels; level++)
		{
			m_startVertsROI[level] = new CPtrList*[1];
			m_startVertsROI[level][0] = NULL;
		}

		// initialise the exchange objects
		m_pStreamDescr = new MGStreamDescr;
		m_pStreamDescr->Init(m_grid->GetNrLevelsUVW(), m_nrSlices, 1);
		m_pStreamDescr->InitMeshLayers(1);

		// mesh type
		m_pStreamDescr->SetMeshType(GetMeshType());
		if (GetMeshType() == QUADRI_MESH)
		{
			if (hasConnectivityInfo)
			{
				m_pStreamDescr->SetSameConnectivity(IsSameConnectivity());
				m_pStreamDescr->SetAutoResolution(IsAutoResolution());
			}
			else
			{
				m_pStreamDescr->SetSameConnectivity(false);
				m_pStreamDescr->SetAutoResolution(false);
			}

			const CIPointUVW & nrSlices = GetNrSlices();
			if (m_pStreamDescr->IsSingleLayered())
			{
				m_pStreamDescr->SetOffsetAmplitude(GetOffsetAmplitude());
				if (GetOffsetAmplitude() == 0)
				{
					hasRefineInfo = 0; // no offsets
				}
			}

			if (IsAutoResolution())
			{
				// check for consistency
				ASSERT(IsSameConnectivity());
				if (hasRefineInfo)
					interleaveRefine = true;       // need to set it because only the first
			}                                  // resolution level is available
		}

		// store the flags
		m_pStreamDescr->HasConnectivityInfo(hasConnectivityInfo);
		m_pStreamDescr->HasRefineInfo(hasRefineInfo);
		m_pStreamDescr->HasRepositionInfo(hasRepositionInfo);
		m_pStreamDescr->HasGridInfo(true);

		// refinement info
		if (hasRefineInfo)
		{
			m_bInterleaveRefine = interleaveRefine;
			m_nrRefBitsVert = 1;
		}
		else
		{
			ASSERT(interleaveRefine == false);

			m_bInterleaveRefine = false;
			m_nrRefBitsVert = 0;
		}

		m_pStreamDescr->SetInterleaveRefine(m_bInterleaveRefine);
		m_pStreamDescr->SetNrRefineBitsVertex(m_nrRefBitsVert);
		m_pStreamDescr->SetCyclicMode(GetCyclicMode());

		// intialise the level information
		for (level = 0; level < nrLevels; level++)
		{
			CIPointUVW levelDiff;
			m_grid->ComputeLevelDiffDown(levelDiff, level);

			// store the number of slices and nr of ROIs (1)
			MStreamLayer &meshLevel = m_pStreamDescr->GetMeshLayer(level, 0);
			meshLevel.Init(GetNrSlices(level), 1, GetNrSlices(level), levelDiff);
		}
	}
	CATCH_ALL(e)
	{
		throw CMGDataException(e);
	}
	END_CATCH_ALL
}

/******************************************************************************/

int CMGMesh::GetNrVerticesROI(int level, BYTE layer)
{
	int count = 0;
	for (int x = m_posVertMin.u; x <= m_posVertMax.u; x++)
	{
		CSurfNodeList *pList = m_nodeStack[x];
		if (pList == NULL) continue;

		for (int y = m_posVertMin.v; y <= m_posVertMax.v; y++)
		{
			CSurfNodeT *pSN = pList[y].GetFirstNode();
			while (pSN != NULL)
			{
				if (IsVertexInROI(pSN, level, layer))
				{
					if (pSN->IsVisited())
						throw CMGDataException("Overlaping vertices");
					pSN->SetVisited();
					++count;
				}

				pSN = pSN->GetDownList();
			}
		}
	}

	return count;
}

void CMGMesh::TraceNonVisitedVertices(int level, BYTE layer)
{
	ASSERT(level >= 0 && level < GetMaxNrLevels());

	for (int x = 0; x < m_nrSlices.u; x++)
	{
		// check row
		CSurfNodeList *pList = m_nodeStack[x];
		if (pList == NULL) continue;

		// check colums
		for (int y = 0; y < m_nrSlices.v; y++)
		{
			CSurfNodeT *pSN = pList[y].GetFirstNode();
			while(pSN != NULL)
			{
				if (pSN->GetLevel() <= level && pSN->GetLayer() == layer)
				{
					if (!pSN->IsVisited())
						ASSERT(0);
				}
				pSN = pSN->GetDownList();
			}
		}
	}
}

int CMGMesh::ComputeNrVertices(int level, BYTE layer)
{
	ASSERT(level >= 0 && level < GetMaxNrLevels());

	int count = 0;
	for (int x = 0; x < m_nrSlices.u; x++)
	{
		// check row
		CSurfNodeList *pList = m_nodeStack[x];
		if (pList == NULL) continue;

		// check colums
		for (int y = 0; y < m_nrSlices.v; y++)
		{
			CSurfNodeT *pSN = pList[y].GetFirstNode();
			while(pSN != NULL)
			{
				if (pSN->GetLevel() <= level && pSN->GetLayer() == layer)
				{
					++count;
				}
				pSN = pSN->GetDownList();
			}
		}
	}
	return count;
}

void CMGMesh::CheckValidVerticesNumber(int nrROIs, int level, BYTE layer)
{
	// reset the streamed flag for the vertices
	ResetVisitedVerticesGrid(level, layer);

	// reset the position of the ROI
	int nrVerts = 0, count = 0;
	for (int roiIdx = 0; roiIdx < nrROIs; roiIdx++)
	{
		InitROI(level, layer, roiIdx);
		++count;
		nrVerts += GetNrVerticesROI(level, layer);
	}

	int nrVerLevel = ComputeNrVertices(level, layer);
	ASSERT(nrVerts == nrVerLevel);

	if (nrVerts != nrVerLevel)
	{
		TraceNonVisitedVertices(level, layer);
	}
}

void CMGMesh::OnBeginMeshCodingLevel(int level)
{
	// allocate the lookup
	m_startVertsROI[level] = new CPtrList*[m_pStreamDescr->GetNrLayers()];
}

void CMGMesh::OnBeginMeshCodingLayer(int level, BYTE layer)
{
	const MStreamLayer &meshLevel = m_pStreamDescr->GetMeshLayer(level, layer);
	int nrROIs = meshLevel.GetTotalNrROIs();
	m_startVertsROI[level][layer] = new CPtrList[nrROIs];

#ifdef _DEBUG
	CheckValidVerticesNumber(nrROIs, level, layer);
#endif

	// reset the streamed flag for the vertices
	ResetVisitedVerticesGrid(level, layer);

	// mark the beginning of a mesh level
	m_pParser->OnBeginMeshCodingLayer(level, layer);
}

void CMGMesh::OnBeginMeshExportLevel(int level)
{
	// update the level for display
	const MStreamLayer &meshLevel = m_pStreamDescr->GetMeshLayer(level, 0);
	int nrROIs = meshLevel.GetTotalNrROIs();
	m_startVertsROI[level][0] = new CPtrList[nrROIs];

#ifdef _DEBUG
	CheckValidVerticesNumber(nrROIs, level, 0);
#endif

	// reset the streamed flag for the vertices
	ResetVisitedVerticesGrid(level, 0);
}

void CMGMesh::OnEndMeshCodingLayer(int level, BYTE layer)
{
	// delete the layer list
	delete []m_startVertsROI[level][layer];
	m_startVertsROI[level][layer] = NULL;
}

void CMGMesh::OnEndMeshCodingLevel(int level, BYTE update_replace)
{
	// mark the end of a mesh level
	m_pParser->OnEndMeshCodingLevel(level, update_replace);

	// clean up the loookup
	delete []m_startVertsROI[level];
	m_startVertsROI[level] = NULL;
}

void CMGMesh::OnBeginMeshCodingROI(int level, BYTE layer, int roiIdx)
{
	m_pParser->OnBeginMeshCodingROI(level, layer, roiIdx);
}

void CMGMesh::OnEndMeshCodingROI(bool bEmpty, int level, BYTE layer, int roiIdx)
{
	m_pParser->OnEndMeshCodingROI(bEmpty, level, layer, roiIdx);
}

void CMGMesh::OnBeginMeshCodingConnectivityROI(BYTE layer)
{
	m_pParser->OnBeginMeshCodingConnectivityROI(layer);
}

void CMGMesh::OnEndMeshCodingConnectivityROI(int level, BYTE layer, int roiIdx)
{
	m_pParser->OnEndMeshCodingConnectivityROI(level, layer, roiIdx);
}

void CMGMesh::OnBeginMeshCodingVertsReposROI(BYTE layer)
{
	m_pParser->OnBeginMeshCodingVertsReposROI(layer);
}

void CMGMesh::OnEndMeshCodingVertsReposROI(int level, BYTE layer, int roiIdx)
{
	m_pParser->OnEndMeshCodingVertsReposROI(level, layer, roiIdx);
}

void CMGMesh::OnBeginMeshCodingVertsRefineROI(BYTE layer)
{
	m_pParser->OnBeginMeshCodingVertsRefineROI(layer);
}

void CMGMesh::OnEndMeshCodingVertsRefineROI(int level, BYTE layer, int roiIdx)
{
	m_pParser->OnEndMeshCodingVertsRefineROI(level, layer, roiIdx);
}

/**********************************************************************************/

void CMGMesh::SaveMeshConnectivityROI(bool &bFinish, int level, BYTE layer, int roiIdx)
{
	CSurfNodeT *pSNStart = CheckStartingPoint(level, layer);
	if (!pSNStart)
	{
		bFinish = false;  return;
	}

	// we need the starting point when we refine the vertices
	CacheStartPoint(pSNStart, level, layer, roiIdx);

	// stream the position of the first node
	StreamStartNode(level, layer, pSNStart);

	// write the node
	StreamNodeBox(level, layer, pSNStart); 

	// check whether the entire box is streamed
	bFinish = IsROIStreamed(level, layer);
}

void CMGMesh::SaveMeshRefineROI(int level, BYTE layer, int roiIdx)
{
	CPtrList &startNodeList = m_startVertsROI[level][layer][roiIdx];
	if (startNodeList.IsEmpty())  return;

	// write the vertices' refine bits
	int nrBits = GetNrRefBitsVert();
	int t = 1 << (nrBits - 1);
	int maxVal = (1 << nrBits) - 1;

	for (int k = nrBits; k >= 1; k--, t >>= 1)
	{
		// reset the visited vertices
		ResetVisitedVerticesBox(level, layer);

		// loop for all start vertices
		POSITION pos = startNodeList.GetHeadPosition();
		while (pos)
		{
			CSurfNodeT *pSNStart = (CSurfNodeT *)startNodeList.GetNext(pos);

			StreamVerticesRefineBits(level, t, maxVal, layer, pSNStart);
		}
	}
}

void CMGMesh::SaveMeshReposROI(int level, BYTE layer, int roiIdx)
{
	int maxLevel = GetMaxNrLevels() - 1;
	ASSERT(level < maxLevel);

	CPtrList &startNodeList = m_startVertsROI[level][layer][roiIdx];
	if (startNodeList.IsEmpty())  return;

	// reset the visited vertices
	ResetVisitedVerticesBox(level, layer);

	// loop for all start vertices
	POSITION pos = startNodeList.GetHeadPosition();
	while (pos)
	{
		CSurfNodeT *pSNStart = (CSurfNodeT *)startNodeList.GetNext(pos);

		// write the reposition bit
		StreamVerticesRepositionBit(level, layer, pSNStart);
	}
}

void CMGMesh::CleanUpVerticesList()
{
	// clear the list
	if (m_pStreamDescr && m_pStreamDescr->HasConnectivityInfo())
	{
		int maxLevel = IsAutoResolution() ? 1 : GetMaxNrLevels();
		for (int level = 0; level < maxLevel; level++)
			CleanUpVerticesList(level);
	}
	delete []m_startVertsROI;   m_startVertsROI = NULL;
}

void CMGMesh::CleanUpVerticesList(int level)
{
	// clear the list
	for (int layer = 0; layer < m_pStreamDescr->GetNrLayers(); layer++)
		CleanUpVerticesList(level, layer);

	delete []m_startVertsROI[level];  m_startVertsROI[level] = NULL;
}

void CMGMesh::CleanUpVerticesList(int level, BYTE layer)
{
	// clear the list
	const MStreamLayer &meshLevel = m_pStreamDescr->GetMeshLayer(level, layer);
	int maxNrROIs = meshLevel.GetTotalNrROIs();

	for (int boxIndex = 0; boxIndex < maxNrROIs; boxIndex++)
		CleanUpVerticesList(level, layer, boxIndex);

	delete []m_startVertsROI[level][layer];  m_startVertsROI[level][layer] = NULL;
}

void CMGMesh::CleanUpVerticesList(int level, BYTE layer, int boxIndex)
{
	// clear the list
	CPtrList &startNodeList = m_startVertsROI[level][layer][boxIndex];
	startNodeList.RemoveAll();
}

/**********************************************************************************/

void CMGMesh::StreamStartNode(int level, BYTE layer, CSurfNodeT *pSNStart)
{
	CIPointUVW pos3D = pSNStart->GetPos();
	FindPosNodeLocal(pos3D, pSNStart->GetDir(), level, layer);

	// make the position relative to the ROI
	pos3D.u -= m_localROI.u;
	pos3D.v -= m_localROI.v;
	pos3D.w -= m_localROI.w;

	BYTE dir = pSNStart->GetDir() >> OFFSET_DIR;
	StreamAbsoluteNode(level, layer, pos3D, dir);
}

void CMGMesh::StreamAbsoluteNode(int level, BYTE layer, const CIPointUVW &pos3D, BYTE dir)
{
	m_pParser->SetAbsoluteVertex(pos3D, dir, level, layer);
}

void CMGMesh::StreamNodeBox(int level, BYTE layer, CSurfNodeT *pSNStart)
{
	// add the starting vertex to the list
	AddStartNodeBox(pSNStart, true);

	while (m_nodeList.GetCount() != 0)
	{
		CSurfNodeT *pSN = (CSurfNodeT *)m_nodeList.RemoveHead();
		ASSERT(pSN->GetLevel() <= level && pSN->GetLayer() == layer);

#ifdef TRACE_CONNECTIVITY_SAVE
		AfxTrace("Pos Root (%d, %d, %d), %d\n", pSN->GetUPos(), pSN->GetVPos(), pSN->GetWPos(), pSN->GetDir());
#endif

		BYTE symbol;
		CSurfNodeT *pSNNext1 = pSN->GetNextNode1(level);
		CSurfNodeT *pSNNext2 = pSN->GetNextNode2(level);

		// write the Link1 connectivity vector
		if (pSNNext1 != NULL)
			GetDirectionDifference(pSN->GetDir(), pSNNext1->GetDir(), symbol);
		else
			symbol = NO_LINK;
		WriteConnectivitySymbol(symbol);

#ifdef TRACE_CONNECTIVITY_SAVE
		if (pSNNext1 != NULL)
			AfxTrace("Pos Link1 (%d, %d, %d), %d\n", pSNNext1->GetUPos(), pSNNext1->GetVPos(), pSNNext1->GetWPos(), pSNNext1->GetDir());
#endif

		// write the Link2 connectivity vector
		if (pSNNext2 != NULL)
			GetDirectionDifference(pSN->GetDir(), pSNNext2->GetDir(), symbol);
		else
			symbol = NO_LINK;
		WriteConnectivitySymbol(symbol);

#ifdef TRACE_CONNECTIVITY_SAVE
		if (pSNNext2 != NULL)
			AfxTrace("Pos Link2 (%d, %d, %d), %d\n", pSNNext2->GetUPos(), pSNNext2->GetVPos(), pSNNext2->GetWPos(), pSNNext2->GetDir());
#endif

		AddNodesToListBox(pSNNext1, pSNNext2, level, layer, true);
	}
}

/**********************************************************************************/

void CMGMesh::StreamVerticesRepositionBit(int level, BYTE layer, CSurfNodeT *pSNStart)
{
	ASSERT(level < GetMaxNrLevels() - 1);

	// add the starting vertex to the list
	AddStartNodeBox(pSNStart);

	while (m_nodeList.GetCount() != 0)
	{
		CSurfNodeT *pSN = (CSurfNodeT *)m_nodeList.RemoveHead();
		ASSERT(pSN->GetLevel() <= level && pSN->GetLayer() == layer);

		bool bSpecialCase;
		BYTE val = VertexRepositionBit(level + 1, pSN, bSpecialCase);
		WriteRepositionSymbol(val, bSpecialCase);

		AddNodesToListBox(pSN->GetNextNode1(level), pSN->GetNextNode2(level), level, layer, false);
	}
}

void CMGMesh::StreamVerticesRefineBits(int level, int mask, int maxVal, BYTE layer, CSurfNodeT *pSNStart)
{
	// add the starting vertex to the list
	AddStartNodeBox(pSNStart);

	while (m_nodeList.GetCount() != 0)
	{
		CSurfNodeT *pSN = (CSurfNodeT *)m_nodeList.RemoveHead();
		ASSERT(pSN->GetLevel() <= level && pSN->GetLayer() == layer);

		float offset;
		if (level < GetMaxNrLevels() - 1)
			offset = GetLevelOffset(pSN->GetDir(), pSN->GetPos(), pSN->GetOffset(), level);
		else
			offset = pSN->GetOffset();
		ASSERT(offset >= 0 && offset < 1);

		int bitVal = (int) floor(offset * maxVal + 0.5f);
		if ((bitVal & mask) != 0)
			WriteRefineSymbol(1);
		else
			WriteRefineSymbol(0);

		AddNodesToListBox(pSN->GetNextNode1(level), pSN->GetNextNode2(level), level, layer, false);
	}
}

/**********************************************************************************/

void CMGMesh::WriteConnectivitySymbol(BYTE symbol)
{
	m_pParser->SetVertexConnectivity(symbol);
}

void CMGMesh::WriteRepositionSymbol(BYTE symbol, bool bSpecialCase)
{
	m_pParser->SetVertexReposition(symbol, bSpecialCase);
}

void CMGMesh::WriteRefineSymbol(BYTE symbol)
{
	m_pParser->SetVertexRefineBit(symbol);
}

/**********************************************************************************/