/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGMeshDecoder.cpp: implementation of the CMGMesh class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGMesh.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


/**********************************************************************************/

bool CMGMesh::Load(bool hasConnectivityInfo, bool hasRefineInfo, bool hasRepositionInfo, bool dump, CString & error)
{
	m_totalNrVertices = 0;

	bool success = true;
	try
	{
		// initialise internal data
		InitialiseDecoderData(hasConnectivityInfo, hasRefineInfo, hasRepositionInfo);

		// load all the levels
		for (int level = 0; level < GetMaxNrLevels(); level++)
		{
			if (level == 0 || !IsAutoResolution())
			{
				if (hasConnectivityInfo)
				{
					// initialise internal level data
					InitialiseLevelData(level);

					// recover the mesh connectivity
					RecoverMeshConnectivity(level, InterleaveRefine());
				}
				else
				{
					BYTE cyclicMode = m_pStreamDescr->GetCyclicMode();
					DefineConectivityUV(hasRefineInfo);
				}

				if (hasRepositionInfo && level < GetMaxNrLevels() - 1)
				{
					// recover the reposition mode
					int reposMode = RecoverRepositionMode();
					RecoverVerticesRepositionBit(level);
				}

				CleanUpVerticesList(level);

				// display the result
				BuildSmoothSurface(level);
			}
			else
			{
				// uniform split the connectivity wireframe
				DefaultMeshConnectivity(level - 1, 0);
				BuildQuadrilateralSurface(level, 0);
			}
		}
	}
	catch (CMGDataException &e)
	{
		error = e.GetErrorMessage();		success = false;
	}

	// Output the MeshGrid description
	if (dump)
	{
		CStdioFile output;
		CFileException e;
		if (output.Open("c:\\dump.xml", CFile::modeCreate | CFile::modeWrite, &e))
		{
			Dump(output);     output.Close();
		}
		else
		{
			TRACE0("Error creating the output file.\n");
		}
	}

	for (int level = 0; level < GetMaxNrLevels(); level++)
		delete []m_startVertsROI[level];

	delete []m_startVertsROI;   m_startVertsROI = NULL;
	delete m_pStreamDescr;  m_pStreamDescr = NULL;

	return success;
}

/**********************************************************************************/

void CMGMesh::RecoverEncodingParameters(const int *vertexLink, float *vertexOffset, int nrVertexLinks, int nrVertexOffsets,
										bool &hasConnectivityInfo, bool &hasRefineInfo, bool &hasRepositionInfo)
{
	m_vertexLink = vertexLink;
	m_vertexOffset = vertexOffset;
	m_nrVertexLinks = nrVertexLinks;
	m_nrVertexOffsets = nrVertexOffsets;

	// read the mesh coding parameters
	RecoverMeshParameters(hasConnectivityInfo);

	// read the refine parameters
	RecoverRefineParameters(hasRefineInfo, hasRepositionInfo);
}

void CMGMesh::RecoverMeshParameters(bool &hasConnectivityInfo)
{
	// parse mesh type, auto resolution flag and offset amplitude
	hasConnectivityInfo = GetNextMeshFlag() == 1 ? true : false;

	int meshType = GetNextMeshFlag();
	SetMeshType(meshType);

	if (meshType == QUADRI_MESH)
	{
		if (hasConnectivityInfo)
		{
			bool flag = GetNextMeshFlag() == 1 ? true : false;
			SetSameConnectivity(flag);

			flag = GetNextMeshFlag() == 1 ? true : false;
			SetAutoResolution(flag);
		}
		else
		{
			SetSameConnectivity(true);
			SetAutoResolution(true);
		}
		// the offset amplitude will be recovered from the vertexOffset field
	}

	// parse the cyclic mode
	int cyclicMode = GetNextMeshFlag();
	SetCyclicMode(cyclicMode);
	m_grid->SetCyclicMode(cyclicMode);
}

int CMGMesh::RecoverAbsoluteOffset()
{
	int sgn = GetNextRefineFlag();

	float factor = 1.0f;
	while(GetNextRefineFlag() == 1)
		factor *= 10.0f;

	float offset = factor * GetNextRefineFlag();
	int offsetAmplitde = (int) offset;
	ASSERT((float)offsetAmplitde == offset);

	if (sgn == 1) offsetAmplitde *= -1;

	return offsetAmplitde;
}

void CMGMesh::RecoverRefineParameters(bool &hasRefineInfo, bool &hasRepositionInfo)
{
	hasRefineInfo = GetNextRefineFlag() == 1 ? true : false;

	if (hasRefineInfo)
	{
		// parse the offset amplitude
		const CIPointUVW & nrSlices = GetNrSlices();
		if (nrSlices.w == 1)
		{
			int offsetAmplitude = RecoverAbsoluteOffset();
			SetOffsetAmplitude(offsetAmplitude);
		}

		SetNrRefBitsVert(1);

		// parse the interleave refine flag
		bool interleaveRefine = GetNextRefineFlag() == 1 ? true : false;
		SetInterleaveRefine(interleaveRefine);
	}
	else
	{
		SetNrRefBitsVert(0);
		SetInterleaveRefine(false);
	}

	hasRepositionInfo = GetNextRefineFlag() == 1 ? true : false;
}

void CMGMesh::InitialiseDecoderData(bool hasConnectivityInfo, bool hasRefineInfo, bool hasRepositionInfo)
{
	try
	{
		// the description of the stream
		m_pStreamDescr = new MGStreamDescr;

		// initialise the description of the stream
		m_pStreamDescr->Init(m_grid->GetNrLevelsUVW(), GetNrSlices(), 1);
		m_pStreamDescr->InitMeshLayers(1);
		m_pStreamDescr->HasConnectivityInfo(hasConnectivityInfo);
		m_pStreamDescr->HasRefineInfo(hasRefineInfo);
		m_pStreamDescr->HasRepositionInfo(hasRepositionInfo);
		m_pStreamDescr->SetSameConnectivity(IsSameConnectivity());
		m_pStreamDescr->SetMeshType(GetMeshType());
		m_pStreamDescr->SetAutoResolution(IsAutoResolution());
		m_pStreamDescr->SetOffsetAmplitude(GetOffsetAmplitude());
		m_pStreamDescr->SetCyclicMode(GetCyclicMode());
		m_pStreamDescr->SetNrRefineBitsVertex(GetNrRefBitsVert());
		m_pStreamDescr->SetInterleaveRefine(InterleaveRefine());

		if (GetNrRefBitsVert() == 0 && InterleaveRefine())
			throw CMGDataException(CMGErrMsg::WRONG_PARAMETERS);

		if (GetMeshType() == QUADRI_MESH)
		{
			const CIPointUVW &nrSlices = GetNrSlices();
			if (nrSlices.w == 1)
			{
				if (GetOffsetAmplitude() == 0 && GetNrRefBitsVert() > 0)
					throw CMGDataException(CMGErrMsg::WRONG_PARAMETERS);
			}

			if (IsAutoResolution())
			{
				if (!IsSameConnectivity() || (GetNrRefBitsVert() > 0 && !InterleaveRefine()))
					throw CMGDataException(CMGErrMsg::WRONG_PARAMETERS);
			}
		}

		// allocate the array holding the starting vertices for each patch
		int maxNrLevels = GetMaxNrLevels();
		m_startVertsROI = new CPtrList**[maxNrLevels];
		memset(m_startVertsROI, 0, maxNrLevels * sizeof(void *));
	}
	catch (CMemoryException *e)
	{
		throw CMGDataException(e);
	}
}

/**********************************************************************************/

void CMGMesh::InitialiseLevelData(int level)
{
	MStreamLayer &meshLevel = m_pStreamDescr->GetMeshLayer(level, 0);

	// store the number of slices
	CIPointUVW nrSlices = m_pStreamDescr->GetMaxNrSlices();
	const CIPointUVW &nrLevels = m_pStreamDescr->GetNrLevelsMesh();

	// initialise the level
	CIPointUVW levelDiff;
	CMGGrid::ComputeNrSlicesLevel(level, nrLevels, nrSlices, levelDiff);

	meshLevel.Init(nrSlices, 1, nrSlices, levelDiff);

	// allocate the start vertices array for each ROI
	m_startVertsROI[level] = new CPtrList*[1];
	m_startVertsROI[level][0] = new CPtrList[1];
}

int CMGMesh::RecoverField()
{
	int byteValue, tmp, number = 0;

	do
	{
		byteValue = 0;
		for (int j = 4; j > 0; j--)
		{
			tmp = GetNextMeshFlag();
			byteValue <<= 2;
			byteValue |= tmp;
		}

		number <<= 7;
		number |= (byteValue & 0x7F);
	}
	while (byteValue & 0x80);

	return number;
}

int CMGMesh::RecoverByteField()
{
	int byteValue = 0;
	for (int j = 4; j > 0; j--)
	{
		int tmp = GetNextMeshFlag();
		byteValue <<= 2;
		byteValue |= tmp;
	}

	return byteValue;
}

int CMGMesh::RecoverRepositionMode()
{
	return (int) GetNextOffset();
}

int CMGMesh::RecoverNumberVertices()
{
	return RecoverField();
}

/**********************************************************************************/

void CMGMesh::BuildMeshUV(int u1, int u2, int v1, int v2, bool bUpdate)
{
	CIPointUVW	pos3D;
	float offset;
	CSurfNodeList *pNodeList = NULL;

	// the w position is always 0
	pos3D.w = 0;

	// top left corner
	pos3D.u = u1;
	pos3D.v = v1;
	pNodeList = GetColumn(pos3D.u);
	if (bUpdate)
		offset = GetNextOffset();
	else
		offset = 0.5f;
	CSurfNodeT *pSN1 = GetAddNode(pNodeList, pos3D, TOP_FACE, 0, 0, offset);

	// top right corner
	pos3D.u = u1;
	pos3D.v = v2;
	pNodeList = GetColumn(pos3D.u);
	if (bUpdate)
		offset = GetNextOffset();
	else
		offset = 0.5f;
	CSurfNodeT *pSN2 = GetAddNode(pNodeList, pos3D, TOP_FACE, 0, 0, offset);

	// bottom left corner
	pos3D.u = u2;
	pos3D.v = v1;
	pNodeList = GetColumn(pos3D.u);
	if (bUpdate)
		offset = GetNextOffset();
	else
		offset = 0.5f;
	CSurfNodeT *pSN3 = GetAddNode(pNodeList, pos3D, TOP_FACE, 0, 0, offset);

	// bottom right corner
	pos3D.u = u2;
	pos3D.v = v2;
	pNodeList = GetColumn(pos3D.u);
	if (bUpdate)
		offset = GetNextOffset();
	else
		offset = 0.5f;
	CSurfNodeT *pSN4 = GetAddNode(pNodeList, pos3D, TOP_FACE, 0, 0, offset);

	// set the connectivity in the UV plane
	pSN1->SetPrevNodeWU(pSN3, TOP_FACE);
	pSN3->SetNextNodeWU(pSN1, TOP_FACE);
	pSN2->SetPrevNodeWU(pSN4, TOP_FACE);
	pSN4->SetNextNodeWU(pSN2, TOP_FACE);

	pSN1->SetNextNodeVW(pSN2, TOP_FACE);
	pSN2->SetPrevNodeVW(pSN1, TOP_FACE);
	pSN3->SetNextNodeVW(pSN4, TOP_FACE);
	pSN4->SetPrevNodeVW(pSN3, TOP_FACE);
}

bool CMGMesh::DefineConectivityUV(bool bUpdate)
{
	const CIPointUVW &nrSlices = GetNrSlices(0);
	try
	{
		// (u,v) = position of the top left corner of a quad
		int v0 = GetNextSliceV(-1, 0);
		for (int v = 0; v < nrSlices.v - 1; v++)
		{
			int v1 = GetNextSliceV(v0, 0);
			int u0 = GetNextSliceU(-1, 0);

			for (int u = 0; u < nrSlices.u - 1; u++)
			{
				int u1 = GetNextSliceU(u0, 0);
				BuildMeshUV(u0, u1, v0, v1, bUpdate);
				u0 = u1;
			}

			v0 = v1;
		}
	}
	catch (CMemoryException *e)
	{
		e->ReportError();   return false;
	}

	return true;
}

void CMGMesh::RecoverMeshConnectivity(int level, bool fullRefine)
{
	// reset the streamed flag for the vertices
	if (level > 0) ResetVisitedVerticesGrid(level - 1, 0);

	m_nrVertices[level] = 0;

	// reset the counter
	m_nrVerticesROI = 0;

	// retrieve the number of vertices for ROI
	int nrVertices = RecoverNumberVertices();

	// store whether there is coding information for ROI
	MStreamLayer &meshLevel = m_pStreamDescr->GetMeshLayer(level, 0);
	meshLevel.SetCodingLookup(0, nrVertices > 0);

	if (nrVertices > 0)
	{
		// a ROI may consist of several sub-patches
		while (m_nrVerticesROI < nrVertices)
		{
			// recover the position of the starting vertex for the current patch
			CSurfNodeT *pSNStart = RecoverNextStartingPoint(level);

			// we need the starting point when we refine the vertices
			CacheStartPoint(pSNStart, level, 0, 0);

			// recover the vertices' connectivity from the starting vertex
			bool lastLevel = (level == GetMaxNrLevels() - 1);
			RecoverMeshConnectivity(level, pSNStart, (lastLevel || fullRefine));
		}

		m_nrVertices[level] += nrVertices;
	}
}

void CMGMesh::RecoverMeshConnectivity(int level, CSurfNodeT *pSNStart, bool bUpdate)
{
	BYTE dirNext;
	bool dirSwitch;
	CIPointUVW pos3D;

	// add the starting vertex to the list
	AddStartNodeBox(pSNStart, true);

	while (m_nodeList.GetCount() != 0)
	{
		CSurfNodeT *pSN = (CSurfNodeT *)m_nodeList.RemoveHead();

		if (bUpdate)
			pSN->SetOffset(GetNextOffset());
		else
			pSN->SetOffset(0.5);
		UpdateVertexCoord(pSN);

		// read the Link1 connectivity vector
		CSurfNodeT *pSNNext1 = NULL;
		dirNext = GetNextLink();
		if (dirNext != NO_LINK)
		{
			pos3D = pSN->GetPos();
			UpdateCoordinateFromConnectivity(pSN->GetDir(), dirNext, 1, dirSwitch, pos3D, level);

			// store the vertex
			pSNNext1 = StoreVertex(pos3D, level, 0, dirNext);
			pSN->SetNextNode1(pSNNext1, level);

			if (dirSwitch == 1)
				pSNNext1->SetPrevNode2(pSN, level);
			else
				pSNNext1->SetPrevNode1(pSN, level);
		}

		// read the Link2 connectivity vector
		CSurfNodeT *pSNNext2 = NULL;
		dirNext = GetNextLink();
		if (dirNext != NO_LINK)
		{
			pos3D = pSN->GetPos();
			UpdateCoordinateFromConnectivity(pSN->GetDir(), dirNext, 2, dirSwitch, pos3D, level);

			// store the vertex
			pSNNext2 = StoreVertex(pos3D, level, 0, dirNext);
			pSN->SetNextNode2(pSNNext2, level);

			if (dirSwitch == 1)
				pSNNext2->SetPrevNode1(pSN, level);
			else
				pSNNext2->SetPrevNode2(pSN, level);
		}

		// add to the list
		AddNodesToList(pSNNext1, pSNNext2, level, true);
	}
}

/**********************************************************************************/

void CMGMesh::RecoverVerticesRepositionBit(int level)
{
	// reset the visited vertices
	ResetVisitedVerticesGrid(level, 0);

	// loop for all start vertices
	CPtrList &startNodeList = m_startVertsROI[level][0][0];
	while (startNodeList.GetCount() != 0)
	{
		CSurfNodeT *pSNStart = (CSurfNodeT *)startNodeList.RemoveHead();

		// write the reposition bit for the next level
		RecoverRepositionBits(level, pSNStart);
	}
}

void CMGMesh::RecoverRepositionBits(int level, CSurfNodeT *pSNStart)
{
	ASSERT(level < GetMaxNrLevels() - 1);

	// add the starting vertex to the list
	AddStartNodeBox(pSNStart);

	while (m_nodeList.GetCount() != 0)
	{
		CSurfNodeT *pSN = (CSurfNodeT *)m_nodeList.RemoveHead();

		// update the reference planes for each vertex
		if (GetNextReposition() == 1.0f)
			RepositionVertex(level + 1, pSN);
		else
			UpdateVertexCoord(level + 1, pSN);

		CSurfNodeT *pSNNext1 = pSN->GetNextNode1(level);	ASSERT(pSNNext1->GetLevel() <= level);
		CSurfNodeT *pSNNext2 = pSN->GetNextNode2(level);	ASSERT(pSNNext2->GetLevel() <= level);

		AddNodesToList(pSNNext1, pSNNext2, level, false);
	}
}

/**********************************************************************************/

CSurfNodeT * CMGMesh::RecoverNextStartingPoint(int level)
{
	// get the absolute position in the reference grid
	CIPointUVW pos3D;
	pos3D.u = RecoverField();
	pos3D.v = RecoverField();
	pos3D.w = RecoverField();

	// get the absolute direction
	BYTE dir = RecoverByteField();

	// bring the position from level to the max level
	RestorePosNode(pos3D, level, 0);

	return StoreVertex(pos3D, level, 0, dir);
}

/**********************************************************************************/

int CMGMesh::GetNextLink()
{
	if (m_countLink == m_nrVertexLinks)
		throw CMGDataException(CMGErrMsg::NOT_ENOUGH_CONNECTIVITY);

	return m_vertexLink[m_countLink++];
}

float CMGMesh::GetNextOffset()
{
	if (m_countOffset == m_nrVertexOffsets)
		throw CMGDataException(CMGErrMsg::NOT_ENOUGH_OFFSETS);

	return m_vertexOffset[m_countOffset++];
}

float CMGMesh::GetNextReposition()
{
	if (m_countOffset == m_nrVertexOffsets)
		throw CMGDataException(CMGErrMsg::NOT_ENOUGH_OFFSETS);

	float val = GetNextOffset();
	if (val != 0.0f && val != 1.0f)
		throw CMGDataException(CMGErrMsg::WRONG_REPOSITION_FLAG);

	return val;
}

int CMGMesh::GetNextMeshFlag()
{
	if (m_countLink == m_nrVertexLinks)
		throw CMGDataException(CMGErrMsg::NOT_ENOUGH_FLAGS);

	return m_vertexLink[m_countLink++];
}

int CMGMesh::GetNextRefineFlag()
{
	if (m_countOffset == m_nrVertexOffsets)
		throw CMGDataException(CMGErrMsg::NOT_ENOUGH_FLAGS);

	return (int) m_vertexOffset[m_countOffset++];
}
