/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGStreamDecoder.cpp: implementation of the CMGStreamParser class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGStreamParse.h"
#include "MGGridCodec.h"
#include "BitStream.h"
#include "MGErrMsg.h"
#include "MGMesh.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


// parsing the main header of the stream
int CMGStreamParser::ParseMeshGridHeader()
{
	// parse the tag and size
	BYTE tag;   int partSize;
	int nrBytes = ParseTagAndSize(tag, partSize);

	if (tag != MGInfoTag)
		throw CMGDataException(CMGErrMsg::EXPECT_HEADER);

#ifdef _DEBUG
	int startPos = m_streamMG.GetPosition();
#endif

	// parse the mesh description
	ParseROIDescription();
	ParseFlags();
	ParseReservedBits();
	ParseNrLayers();
	ParseMeshType();
	ParseCyclicMode();
	ParseNrRefineBitsVertex();

	// parse the grid description
	ParseFilterType();
	ParseGridCodingParam();
	ParseGridCorners(false);

	// byte allign the data
	FlushBitStream();

	// store the header size
#ifdef _DEBUG
	int endPos = m_streamMG.GetPosition();
	ASSERT(endPos - startPos == partSize);
#endif
	m_streamDescr.SetHeaderSize(partSize);

	return partSize + nrBytes;
}

// parse the size of the stream
int CMGStreamParser::ParseStreamSize(int &counterLength)
{
	int streamSize;
	counterLength = ParseValue(streamSize);

	// byte allign the data
	FlushBitStream();

	m_streamDescr.SetStreamSize(streamSize);
	return streamSize;
}

// parse the parts identifier and resolution level
int CMGStreamParser::ParseTagAndSize(BYTE &tag, int &size)
{
	int nrBits = TAG_BITS;
	m_streamMG.ReadBits(&tag, TAG_BITS);
	nrBits += ParseValue(size);

	// byte allign the data
	FlushBitStream();

	return AlignInBytes(nrBits);
}

// parse the resolution level and flags
void CMGStreamParser::ParseLevelDescription(int &level, BYTE &update_replace)
{
	// read the level
	m_streamMG.ReadBits((DWORD *)&level, LEVEL_BITS);

	// read the update/replace flag
	m_streamMG.ReadBits(&update_replace, UPDATE_REPLACE_BITS);
}

// parse the layer description
void CMGStreamParser::ParseLayerDescription(int &level, BYTE &update_replace, CIntArray &layers)
{
	// read the level description
	ParseLevelDescription(level, update_replace);

	// read the layer description
	BYTE nBitsLayer = m_streamDescr.GetNrBitsLayer();
	if (nBitsLayer > 0)
		ParseIndices(nBitsLayer, layers);
}

void CMGStreamParser::ParseIndices(int nBitsIndex, CIntArray &layers)
{	
	// the number of elements in the indices list.
	DWORD number;
	m_streamMG.ReadBits(&number, nBitsIndex);

	BYTE isIndexed;
	for (DWORD i = 0; i < number; i++)
	{
		m_streamMG.ReadBit(&isIndexed);

		// check if indexed or list
		if (isIndexed)
		{
			// read the index
			DWORD index;
			m_streamMG.ReadBits(&index, nBitsIndex);
			layers.Add(index);
		}
		else
		{
			// read the start index and count from the list
			DWORD start, count;
			m_streamMG.ReadBits(&start, nBitsIndex);
			m_streamMG.ReadBits(&count, nBitsIndex);

			for (DWORD j = 0; j < count; j++)
			{
				layers.Add(start + j);
			}
		}
	}
}

int CMGStreamParser::ParseValue(int &number)
{
	BYTE bNextByte, value;
	int length = 0;
	number = 0;

	do
	{
		m_streamMG.ReadBit(&bNextByte);
		number <<= 7;
		m_streamMG.ReadBits(&value, 7);
		number |= value;
		length += 8;
	}
	while (bNextByte);

	return length;
}

// parsing mesh related information of the main header of the stream
void CMGStreamParser::ParseNrLevels(CIPointUVW &nrLevels)
{
	m_streamMG.ReadBits((DWORD *)&nrLevels.u, LEVEL_BITS);
	m_streamMG.ReadBits((DWORD *)&nrLevels.v, LEVEL_BITS);
	m_streamMG.ReadBits((DWORD *)&nrLevels.w, LEVEL_BITS);

	if (nrLevels.u < MIN_LEVEL || 
		nrLevels.v < MIN_LEVEL ||
		nrLevels.w < MIN_LEVEL)
	{
		throw CMGDataException(CMGErrMsg::WRONG_NR_LEVELS);
	}
}

void CMGStreamParser::ParseNrSlices(CIPointUVW &nrSlices)
{
	// read the number of slices
	ParseValue(nrSlices.u);
	ParseValue(nrSlices.v);
	ParseValue(nrSlices.w);

	if (nrSlices.u <= 0 || nrSlices.v <= 0 || nrSlices.w <= 0)
		throw CMGDataException(CMGErrMsg::WRONG_NR_SLICES);
}

void CMGStreamParser::ParseNrROIs(CIPointUVW &nrROIs)
{
	// read the number of slices
	ParseValue(nrROIs.u);
	ParseValue(nrROIs.v);
	ParseValue(nrROIs.w);

	if (nrROIs.u <= 0 || nrROIs.v <= 0 || nrROIs.w <= 0)
		throw CMGDataException(CMGErrMsg::WRONG_NR_SLICES);
}

void CMGStreamParser::ParseROIDescription()
{
	CIPointUVW nrLevels, nrSlices, nrROIs;

	ParseNrLevels(nrLevels);
	ParseNrSlices(nrSlices);
	ParseNrROIs(nrROIs);

	// allocate internal arrays
	m_streamDescr.Init(nrLevels, nrSlices, nrROIs);
}

void CMGStreamParser::ParseFlags()
{
	BYTE flag;
	m_streamMG.ReadBit(&flag);    m_streamDescr.HasConnectivityInfo(flag == 1 ? true: false);
	m_streamMG.ReadBit(&flag);    m_streamDescr.HasRefineInfo(flag == 1 ? true: false);
	m_streamMG.ReadBit(&flag);    m_streamDescr.HasRepositionInfo(flag == 1 ? true: false);
	m_streamMG.ReadBit(&flag);    m_streamDescr.HasGridInfo(flag == 1 ? true: false);
	m_streamDescr.HasNormalsInfo(false);
}

void CMGStreamParser::ParseMeshType()
{
	BYTE meshType;
	m_streamMG.ReadBits(&meshType, MESH_TYPE_BITS);    m_streamDescr.SetMeshType(meshType);

	if (meshType == QUADRI_MESH)
	{
		if (m_streamDescr.HasConnectivityInfo())
		{
			BYTE flag;
			m_streamMG.ReadBit(&flag);     m_streamDescr.SetSameConnectivity(flag == 1 ? true: false);
			m_streamMG.ReadBit(&flag);     m_streamDescr.SetAutoResolution(flag == 1 ? true : false);
		}
		else
		{
			m_streamDescr.SetSameConnectivity(true);
			m_streamDescr.SetAutoResolution(true);
		}

		if (m_streamDescr.IsSingleLayered())
		{
			int offsetAmplitude;		ParseValue(offsetAmplitude);
			m_streamDescr.SetOffsetAmplitude((int)offsetAmplitude);
		}
	}
}

void CMGStreamParser::ParseNrRefineBitsVertex()
{
	// parse whether the refine bits are interleaved in the stream or not
	if (m_streamDescr.HasRefineInfo())
	{
		BYTE flag;
		m_streamMG.ReadBit(&flag);  m_streamDescr.SetInterleaveRefine(flag == TRUE);

		// parse the number of refine bits
		BYTE nrBits;
		m_streamMG.ReadBits(&nrBits, REFINE_BITS);   m_streamDescr.SetNrRefineBitsVertex(nrBits);

		if (nrBits == 0 || nrBits > MAX_REFINE_BITS)
		{
			throw CMGDataException(CMGErrMsg::WRONG_VERTEX_REFINE_BITS);
		}
	}
	else
	{
		m_streamDescr.SetInterleaveRefine(false);
		m_streamDescr.SetNrRefineBitsVertex(0);
	}
}

void CMGStreamParser::ParseCyclicMode()
{
	BYTE flag;
	m_streamMG.ReadBits(&flag, CYCLIC_MODE_BITS);   m_streamDescr.SetCyclicMode(flag);
}

void CMGStreamParser::ParseReservedBits()
{
	WORD bits;
	m_streamMG.ReadBits(&bits, RESERVED_BITS);
}

void CMGStreamParser::ParseNrLayers()
{
	BYTE hasLayers = 0;
	m_streamMG.ReadBit(&hasLayers);

	int nrLayers = 0;
	if (hasLayers != 0)
		ParseValue(nrLayers);

	m_streamDescr.InitMeshLayers(max(nrLayers, 1));
}

void CMGStreamParser::InitStreams(BYTE nrLayers)
{
	m_streamConnectivity = new MGStream[nrLayers];
	m_streamVertsRepos = new MGStream[nrLayers];
	m_streamVertsRefine = new MGStream[nrLayers];
}

// parsing grid related information of the main header of the stream
void CMGStreamParser::ParseFilterType()
{
	BYTE bValue;
	m_streamMG.ReadBits(&bValue, FILTER_BITS);   m_streamDescr.SetFilterType(bValue);
}

void CMGStreamParser::ParseGridCodingParam()
{
	// read the number of bits allocated for the variables
	BYTE nrBits;
	m_streamMG.ReadBits(&nrBits, FIELD_BITS);   m_streamDescr.SetNrBitsGridParam(nrBits);

	if (nrBits == 0 || nrBits > MAX_FIELD_BITS)
	{
		throw CMGDataException(CMGErrMsg::WRONG_NR_SCALE_BITS);
	}

	RGCoding * codingRG = m_streamDescr.GetGridCodingParam();

	// read the scale
	for (int dir = X_GRID; dir <= Z_GRID; dir++)
	{
		DWORD wValue;
		m_streamMG.ReadBits(&wValue, nrBits);   codingRG[dir].scale = wValue;
	}

	if (codingRG[X_GRID].scale < MIN_SCALE ||
		codingRG[Y_GRID].scale < MIN_SCALE ||
		codingRG[Z_GRID].scale < MIN_SCALE)
	{
		throw CMGDataException(CMGErrMsg::WRONG_SCALE_VALUE);
	}

	// prepare the header description of the levels
	InitGridCodingHeader();
}

void CMGStreamParser::ProcessGridCorners(BYTE &update_replace)
{
	// check if the update/replace flag is present
	m_streamMG.ReadBits(&update_replace, UPDATE_REPLACE_BITS);
	
	// read the grid corners
	ParseGridCorners(update_replace == UPDATE_DATA);
}

void CMGStreamParser::ParseGridCorners(bool update)
{
	// retrieve the number of quantization bits of the grid values
	BYTE nrQuantBits;
	m_streamMG.ReadBits(&nrQuantBits, QUANT_BITS);

	if (nrQuantBits == 0 || nrQuantBits > MAX_QUANT_BITS)
	{
		throw CMGDataException(CMGErrMsg::WRONG_NR_QUANT_BITS);
	}

	m_streamDescr.SetNrQuantBitsGrid(nrQuantBits);

	// which grid corners are stored in the stream
	int uMax = 1, vMax = 1, wMax = 1;
	const CIPointUVW &nrSlices = m_streamDescr.GetMaxNrSlices();

	// check for single layered
	if (nrSlices.w == 1) wMax = 0;  // only layer 0

	// check for cyclic
	if (m_streamDescr.IsCyclicU()) uMax = 0;  // only one corner in the u direction
	if (m_streamDescr.IsCyclicV()) vMax = 0;  // only one corner in the v direction

	// retrieve the grid corner values
	for (int w = 0; w <= wMax; w++)
	{
		for (int v = 0; v <= vMax; v++)
		{
			for (int u = 0; u <= uMax; u++)
			{
				for (int dir = X_GRID; dir <= Z_GRID; dir++)
				{
					RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);

					BYTE sgn;     m_streamMG.ReadBit(&sgn);
					DWORD value;  m_streamMG.ReadBits(&value, nrQuantBits);

					if (update)
						codingRG.gridCorners[4*w + 2*v + u] += (sgn ? -((int)value) : value);
					else
						codingRG.gridCorners[4*w + 2*v + u] = (sgn ? -((int)value) : value);
				}
			}
		}
	}

	// initialise the remaining grid corners
	if (uMax == 0)
	{
		// copy point 0 to point 1 in both lines from the two layers
		for (int k = 0; k < 8; k += 2)
		{
			for (int dir = X_GRID; dir <= Z_GRID; dir++)
			{
				RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
				codingRG.gridCorners[k + 1] = codingRG.gridCorners[k + 0];
			}
		}
	}

	if (vMax == 0)
	{
		// copy line 0 to line 1 in both layers
		for (int k = 0; k < 8; k += 4)
		{
			for (int dir = X_GRID; dir <= Z_GRID; dir++)
			{
				RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
				codingRG.gridCorners[k + 2] = codingRG.gridCorners[k + 0];
				codingRG.gridCorners[k + 3] = codingRG.gridCorners[k + 1];
			}
		}
	}

	if (wMax == 0)
	{
		// copy layer 0 to layer 1 such that both are similar
		for (int k = 0; k < 8; k += 8)
		{
			for (int dir = X_GRID; dir <= Z_GRID; dir++)
			{
				RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
				codingRG.gridCorners[k + 4] = codingRG.gridCorners[k + 0];
				codingRG.gridCorners[k + 5] = codingRG.gridCorners[k + 1];
				codingRG.gridCorners[k + 6] = codingRG.gridCorners[k + 2];
				codingRG.gridCorners[k + 7] = codingRG.gridCorners[k + 3];
			}
		}
	}
}

// parse the stream headers for each level
void CMGStreamParser::ParseMeshGridStream(int streamSize)
{
	int pos;
	while (streamSize > 0)
	{
		// parse the part tag and size of part
		BYTE tag;   int partSize;
		int nrBytes = ParseTagAndSize(tag, partSize);
		streamSize -= (nrBytes + partSize);

#ifdef _DEBUG
		int startPos = GetStreamPositon();
#endif

		int level, k;
		BYTE update_replace;
		CIntArray layers;

		// parse the resolution level
		switch (tag)
		{
		case MGMeshInfoTag:
			ParseLayerDescription(level, update_replace, layers);
			if (layers.GetSize() == 0)
			{
				MeshDescriptor(level, 0);
				//for (k = 1; k < m_streamDescr.GetNrLayers(); k++)
				//	CopyMeshDescriptor(level, update_replace, 0, k);
			}
			else
			{
				for (k = 0; k < layers.GetSize(); k++)
					MeshDescriptor(level, k);
			}
			break;
		case MGGridInfoTag:
			ParseLevelDescription(level, update_replace);
			GridDescriptor(level);
			break;
		case MGMeshConnectivityROIInfoTag:
			ParseLayerDescription(level, update_replace, layers);
			for (k = 0; k < layers.GetSize(); k++)
			{
				pos = MeshConnectivityROIHeader(level, k);
				m_streamMG.SetPosition(pos);
			}
			break;
		case MGMeshConnectivityInfoTag:
			ParseLayerDescription(level, update_replace, layers);
			for (k = 0; k < layers.GetSize(); k++)
			{
				pos = MeshConnectivityHeader(level, k);
				m_streamMG.SetPosition(pos);
			}
			break;
		case MGVerticesRepositionROIInfoTag:
			ParseLayerDescription(level, update_replace, layers);
			for (k = 0; k < layers.GetSize(); k++)
			{
				pos = VertexRepositionROIHeader(level, k);
				m_streamMG.SetPosition(pos);
			}
			break;
		case MGVerticesRepositionInfoTag:
			ParseLayerDescription(level, update_replace, layers);
			for (k = 0; k < layers.GetSize(); k++)
			{
				pos = VertexRepositionHeader(level, k);
				m_streamMG.SetPosition(pos);
			}
			break;
		case MGVerticesRefinementInfoTag:
			ParseLayerDescription(level, update_replace, layers);
			for (k = 0; k < layers.GetSize(); k++)
			{
				pos = VertexRefineHeader(level, k);
				m_streamMG.SetPosition(pos);
			}
			break;
		case MGVerticesRefinementROIInfoTag:
			ParseLayerDescription(level, update_replace, layers);
			for (k = 0; k < layers.GetSize(); k++)
			{
				pos = VertexRefineROIHeader(level, k);
				m_streamMG.SetPosition(pos);
			}
			break;
		case MGGridCoefficientsROIInfoTag:
			ParseLevelDescription(level, update_replace);
			pos = GridCoefficientsROIHeader(level);
			m_streamMG.SetPosition(pos);
			break;
		case MGGridCoefficientsInfoTag:
			ParseLevelDescription(level, update_replace);
			pos = GridCoefficientsHeader(level);
			m_streamMG.SetPosition(pos);
			break;
		case MGGridCornersInfoTag:
			m_streamMG.MovePosition(partSize);
			break;
		case MGBeginFrameTag:
			break;
		default:
			// skip the part
			TRACE0(CMGErrMsg::GetMessage(CMGErrMsg::WRONG_PART_IDENTIFIER));
			m_streamMG.MovePosition(partSize);
		}

#ifdef _DEBUG
		int finalPos = GetStreamPositon();
		ASSERT(finalPos == startPos + partSize);
#endif

		EndOfPart();
	}
}

// parsing mesh related information of the level header of the stream
void CMGStreamParser::MeshDescriptor(int level, BYTE layer)
{
	// get the current position, needed to update the counters
	int pos = m_streamMG.GetPosition();

	// store the position of the header
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	meshLevel.SetHeaderPosition(pos);

	// store the number of slices and number of ROIs
	CIPointUVW levelDiff;
	CIPointUVW nrROIs = m_streamDescr.GetNrROIs();
	CIPointUVW nrSlices = m_streamDescr.GetMaxNrSlices();
	const CIPointUVW &nrLevels = m_streamDescr.GetNrLevelsMesh();
	CMGGrid::ComputeNrSlicesROIsLevel(level, nrLevels, nrSlices, nrROIs, levelDiff);

	meshLevel.Init(nrSlices, nrROIs, m_streamDescr.GetMaxROISize(), levelDiff);

	// retrieve the mesh header for the level & byte allign the data
	ParseMeshCodingHeader(level, layer);

	// byte allign the data
	FlushBitStream();

	// store the size of the header
	meshLevel.SetHeaderSize(m_streamMG.GetPosition() - pos);
}

void CMGStreamParser::DefaultMeshDescriptor(int level, BYTE layer)
{
	ASSERT(level > 0 && m_streamDescr.IsAutoResolution());
	MStreamLayer &meshLevel0 = m_streamDescr.GetMeshLayer(0, layer);

	// the number of ROIs increases with a factor of two
	CIPointUVW nrROIs = meshLevel0.GetNrROIs();
	nrROIs.u <<= level;		nrROIs.v <<= level;		nrROIs.w <<= level;

	// store the number of slices and nr ROIs
	CIPointUVW levelDiff;
	CIPointUVW nrSlices = m_streamDescr.GetMaxNrSlices();
	const CIPointUVW &nrLevels = m_streamDescr.GetNrLevelsMesh();

	CMGGrid::ComputeNrSlicesLevel(level, nrLevels, nrSlices, levelDiff);

	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	meshLevel.Init(nrSlices, nrROIs, m_streamDescr.GetMaxROISize(), levelDiff);
}

void CMGStreamParser::GridDescriptor(int levelG)
{
	// get the current position, needed to update the counters
	int pos = m_streamMG.GetPosition();

	// store the position of the header
	GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(levelG);
	gridLevel.SetHeaderPosition(pos);

	// initialise the grid level
	CIPointUVW levelDiff;
	CIPointUVW nrLevels = m_streamDescr.GetNrLevelsGrid();
	CIPointUVW nrTiles = m_streamDescr.GetNrROIs();
	CIPointUVW nrSlices = m_streamDescr.GetMaxNrSlices();
	CMGGrid::ComputeNrSlicesROIsLevel(levelG, nrLevels, nrSlices, nrTiles, levelDiff);

	gridLevel.Init(nrSlices, nrTiles, levelDiff, 0);

	// parse coding header 
	ParseThreshold(levelG);
	ParseGridCodingHeader(levelG);

	// byte allign the data
	FlushBitStream();

	// store the size of the header
	gridLevel.SetHeaderSize(m_streamMG.GetPosition() - pos);
}

int CMGStreamParser::MeshConnectivityHeader(int level, BYTE layer)
{
	ParseROICodingLookup(level, layer);               FlushBitStream();
	ParseConnectivityCodingHeader(level, layer);      FlushBitStream();

	return UpdateConnectivityFilePointers(level, layer);
}

int CMGStreamParser::VertexRepositionHeader(int level, BYTE layer)
{
	ParseRepositionCodingHeader(level, layer);     FlushBitStream();

	return UpdateRepositionFilePointers(level, layer);
}

int CMGStreamParser::VertexRefineHeader(int level, BYTE layer)
{
	ParseRefinementCodingHeader(level, layer);     FlushBitStream();

	return UpdateRefineFilePointers(level, layer);
}

int CMGStreamParser::GridCoefficientsHeader(int level)
{
	ParseTilesCodingHeader(level);          FlushBitStream();

	return UpdateTilesFilePointers(level);
}

void CMGStreamParser::ParseMeshCodingHeader(int level, BYTE layer)
{
	ASSERT(level >= 0 && level < m_streamDescr.GetMaxNrLevelsMesh());

	MStreamLayer &meshLevel  = m_streamDescr.GetMeshLayer(level, layer);

	if (m_streamDescr.IsRepositionInfoAvailable(level))
	{
		BYTE repositionFlag;
		m_streamMG.ReadBits(&repositionFlag, REPOS_MODE_BITS);

		meshLevel.SetRepositionBitsMode(repositionFlag);
	}

	// read the number of bits for storing the counters: can be 0 bits
	BYTE nrBits;

	// for the connectivity counters
	if (m_streamDescr.IsConnectivityInfoAvailable(level))
	{
		m_streamMG.ReadBits(&nrBits, VERTS_BITS);
		meshLevel.SetNrBitsConnectivityCounterField(nrBits);

		if (nrBits > MAX_VERTS_BITS)
		{
			throw CMGDataException(CMGErrMsg::WRONG_NR_VERTEX_BITS);
		}
	}

	if (m_streamDescr.IsRepositionInfoAvailable(level, layer))
	{
		// for the reposition counters
		m_streamMG.ReadBits(&nrBits, VERTS_BITS);
		meshLevel.SetNrBitsVertsReposCounterField(nrBits);

		if (nrBits > MAX_VERTS_BITS)
		{
			throw CMGDataException(CMGErrMsg::WRONG_NR_VERTEX_BITS);
		}
	}

	if (m_streamDescr.IsRefinementInfoAvailable(level))
	{
		// for the refinement counters
		m_streamMG.ReadBits(&nrBits, VERTS_BITS);
		meshLevel.SetNrBitsVertsRefineCounterField(nrBits);

		if (nrBits > MAX_VERTS_BITS)
		{
			throw CMGDataException(CMGErrMsg::WRONG_NR_VERTEX_BITS);
		}
	}
}

void CMGStreamParser::ParseROICodingLookup(int level, BYTE layer)
{
	ASSERT(level >= 0 && level < m_streamDescr.GetMaxNrLevelsMesh());

	MStreamLayer &meshLevel  = m_streamDescr.GetMeshLayer(level, layer);
	CBoolArray &codingLookup = meshLevel.GetCodingLookup();

	// read the ROI coding lookup
	int nrROIs = meshLevel.GetTotalNrROIs();
	for (int k = 0; k < nrROIs; k++)
	{
		BYTE bValue;  m_streamMG.ReadBit(&bValue);
		codingLookup[k] = bValue != 0 ? true: false;
	}
}

void CMGStreamParser::ParseConnectivityCodingHeader(int level, BYTE layer)
{
	ASSERT(level >= 0 && level < m_streamDescr.GetMaxNrLevelsMesh());

	MStreamLayer &meshLevel  = m_streamDescr.GetMeshLayer(level, layer);
	CBoolArray &codingLookup = meshLevel.GetCodingLookup();

	// read the number of bits allocated for the variables
	BYTE nrBits = meshLevel.GetNrBitsConnectivityCounterField();

	// read the counters
	int nrROIs = meshLevel.GetTotalNrROIs();
	for (int k = 0; k < nrROIs; k++)
	{
		if (nrBits > 0 && codingLookup[k])
		{
			DWORD value;      m_streamMG.ReadBits(&value, nrBits);
			meshLevel.SetNrBytesConnectivityCoding(k, value);
		}
		else
		{
			meshLevel.SetNrBytesConnectivityCoding(k, 0);
		}
	}
}

void CMGStreamParser::ParseRepositionCodingHeader(int level, BYTE layer)
{
	ASSERT(level >= 0 && level < m_streamDescr.GetMaxNrLevelsMesh());

	if (!m_streamDescr.IsRepositionInfoAvailable(level, layer)) return;

	MStreamLayer &meshLevel  = m_streamDescr.GetMeshLayer(level, layer);
	CBoolArray &codingLookup = meshLevel.GetCodingLookup();

	// read the number of bits allocated for the variables
	BYTE nrBits = meshLevel.GetNrBitsVertsReposCounterField();

	// read the counters
	int nrROIs = meshLevel.GetTotalNrROIs();
	for (int k = 0; k < nrROIs; k++)
	{
		if (nrBits > 0 && codingLookup[k])
		{
			DWORD value;      m_streamMG.ReadBits(&value, nrBits);
			meshLevel.SetNrBytesVertsReposCoding(k, value);
		}
		else
		{
			meshLevel.SetNrBytesVertsReposCoding(k, 0);
		}
	}
}

void CMGStreamParser::ParseRefinementCodingHeader(int level, BYTE layer)
{
	ASSERT(level >= 0 && level < m_streamDescr.GetMaxNrLevelsMesh());

	if (!m_streamDescr.IsRefinementInfoAvailable(level)) return;

	MStreamLayer &meshLevel  = m_streamDescr.GetMeshLayer(level, layer);
	CBoolArray &codingLookup = meshLevel.GetCodingLookup();

	// read the number of bits allocated for the variables
	BYTE nrBits = meshLevel.GetNrBitsVertsRefineCounterField();

	// read the counters
	int nrROIs = meshLevel.GetTotalNrROIs();
	for (int k = 0; k < nrROIs; k++)
	{
		if (nrBits > 0 && codingLookup[k])
		{
			DWORD value;      m_streamMG.ReadBits(&value, nrBits);
			meshLevel.SetNrBytesVertsRefineCoding(k, value);
		}
		else
		{
			meshLevel.SetNrBytesVertsRefineCoding(k, 0);
		}
	}
}

void CMGStreamParser::ParseThreshold(int level)
{
	GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(level);
	for (int dir = X_GRID; dir <= Z_GRID; dir++)
	{
		BYTE bValue;
		m_streamMG.ReadBits(&bValue, FIELD_BITS);
		gridLevel.SetThreshold(dir, bValue != 0 ? (1 << (bValue - 1)) : 0);
	}
}

void CMGStreamParser::ParseGridCodingHeader(int levelG)
{
	ASSERT(levelG >= 0 && levelG < m_streamDescr.GetMaxNrLevelsGrid());

	GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(levelG);

	// read the number of bits allocated for the variables
	BYTE nrBits;
	m_streamMG.ReadBits(&nrBits, GRID_BITS);   gridLevel.SetNrBitsGridCounterField(nrBits);

	if (nrBits > MAX_GRID_BITS)
	{
		throw CMGDataException(CMGErrMsg::WRONG_NR_GRID_BITS);
	}
}

void CMGStreamParser::ParseTilesCodingHeader(int levelG)
{
	ASSERT(levelG >= 0 && levelG < m_streamDescr.GetMaxNrLevelsGrid());

	GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(levelG);

	// the number of bits allocated for the counters
	BYTE nrBits = gridLevel.GetNrBitsGridCounterField();

	// read the counters for the tiles
	for (int k = 0; k < gridLevel.GetTotalNrTiles(); k++)
	{
		for (int dir = X_GRID; dir <= Z_GRID; dir++)
		{
			if (nrBits > 0)
			{
				BYTE bValue;        m_streamMG.ReadBit(&bValue);
				if (bValue == 0)
				{
					gridLevel.SetNrBytesTileCoding(dir, k, 0);
				}
				else
				{
					DWORD value;
					m_streamMG.ReadBits(&value, nrBits);     gridLevel.SetNrBytesTileCoding(dir, k, value);
				}
			}
			else
			{
				gridLevel.SetNrBytesTileCoding(dir, k, 0);
			}
		}
	}
}

/*******************************************************************************/
// Mesh decoding stream access

void CMGStreamParser::GetAbsoluteVertex(CIPointUVW &pos3D, BYTE &dir, int level, BYTE layer)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	const CIPointUVW &nrBits = meshLevel.GetNrBitsVertPosField();

	DWORD value;
	m_streamMG.ReadBits(&value, nrBits.u);     pos3D.u = value;
	m_streamMG.ReadBits(&value, nrBits.v);     pos3D.v = value;
	m_streamMG.ReadBits(&value, nrBits.w);     pos3D.w = value;

	m_streamMG.ReadBits(&dir, DIR_BITS);
}

void CMGStreamParser::GetVertexConnectivity(BYTE &symbol)
{
	if (m_streamDescr.IsSameConnectivity())
	{
		m_streamMG.ReadBits(&symbol, 1);
		if (symbol == 1) symbol = NO_LINK;
	}
	else
	{
		m_streamMG.ReadBits(&symbol, BITS_PER_CONNECT);
	}
}

void CMGStreamParser::GetVertexRefineBit(BYTE &symbol)
{
	m_streamMG.ReadBit(&symbol);
}

void CMGStreamParser::GetVertexReposition(BYTE &symbol, bool bSpecialCase)
{
	m_streamMG.ReadBit(&symbol);

	if (symbol == 1 && bSpecialCase)
	{
		BYTE bit;
		m_streamMG.ReadBit(&bit);
		symbol = 2 | bit;
	}
}
