/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGStreamDescription.cpp: implementation of the CMGStreamDescription class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGStreamDescription.h"
#include "MGErrMsg.h"
#include "MGStreamParse.h"
#include "MGMesh.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MStreamLayer::MStreamLayer()
{
	m_nrROIs = -1;
	m_nrSlices = -1;
	m_levelDiff = -1;
	m_headerSz = -1;
	m_headerPos = -1;
	m_reposMode = REPOSITION_ANY;

	m_nrBitsROIIndex = -1;
	m_nrBitsVertPos = -1;
	m_nrBitsConnetivityCounter = -1;
	m_nrBitsVertsReposCounter = -1;
	m_nrBitsVertsRefineCounter = -1;

	m_msgs = NULL;
}

int MStreamLayer::Init(const CIPointUVW &nrSlices, const CIPointUVW &nrROIs, const CIPointUVW &maxROISz, const CIPointUVW &levelDiff)
{
	m_levelDiff = levelDiff;
	m_nrSlices = nrSlices;
	m_nrROIs = nrROIs;
	int totalNrROIs = nrROIs.u * nrROIs.v * nrROIs.w;

	// number of bits for the ROI indices fields
	int nBitsIndex = CMGStreamParser::GetNrBits(totalNrROIs);
	SetNrBitsROIIndex(nBitsIndex);

	CIPointUVW nrBits;
	nrBits.u = CMGStreamParser::GetPower2(maxROISz.u);
	nrBits.v = CMGStreamParser::GetPower2(maxROISz.v);
	nrBits.w = CMGStreamParser::GetPower2(maxROISz.w);

	// number of bits for the vertex position field
	SetNrBitsVertPosField(nrBits);

	try
	{
		m_codingLookup.SetSize(totalNrROIs);
		m_connectivityCount.SetSize(totalNrROIs);
		m_vertsReposCount.SetSize(totalNrROIs);
		m_vertsRefineCount.SetSize(totalNrROIs);
		m_posStreamConnectivity.SetSize(totalNrROIs);
		m_posStreamRefine.SetSize(totalNrROIs);
		m_posStreamReposition.SetSize(totalNrROIs);
		m_cacheROI.SetSize(totalNrROIs);
		m_cacheRefine.SetSize(totalNrROIs);
		m_cacheReposition.SetSize(totalNrROIs);

		for (int roiIdx = 0; roiIdx < totalNrROIs; roiIdx++)
		{
			m_codingLookup[roiIdx] = false;
			m_connectivityCount[roiIdx] = 0;
			m_vertsReposCount[roiIdx] = -1;
			m_vertsRefineCount[roiIdx] = -1;
			m_posStreamConnectivity[roiIdx] = -1;
			m_posStreamRefine[roiIdx] = -1;
			m_posStreamReposition[roiIdx] = -1;
			m_cacheROI[roiIdx] = false;
			m_cacheRefine[roiIdx] = false;
			m_cacheReposition[roiIdx] = false;
		}
	}
	catch (CMemoryException *e)
	{
		throw CMGDataException(e);
	}

	return totalNrROIs;
}

int MStreamLayer::AddMsg(LPCTSTR lpszFormat, ...)
{
	ASSERT(m_msgs != NULL);
	if (!m_msgs)  return -1;

	CString str;

	va_list argList;
	va_start(argList, lpszFormat);
	str.FormatV(lpszFormat, argList);
	va_end(argList);

	return AddMsg(str);
}

void MStreamLayer::DumpStream(CStringArray &msgs, const ROIDescription &roiDescr)
{
	m_msgs = &msgs;

	// mesh header information
	AddMsg("Nr of Slices = (%d, %d, %d)", m_nrSlices.u, m_nrSlices.v, m_nrSlices.w);

	AddMsg("Nr of bits per <ROI index> field = %d", m_nrBitsROIIndex);
	AddMsg("Nr of ROIs = (%d, %d, %d)", m_nrROIs.u, m_nrROIs.v, m_nrROIs.w);

	int repositionFlag = GetRepositionBitsMode();
	switch (repositionFlag)
	{
	case REPOSITION_NO:   AddMsg("No reposition bits are all '0'.");   break;
	case REPOSITION_YES:  AddMsg("No reposition bits are all '1'.");   break;
	case REPOSITION_ANY:  AddMsg("Reposition bits are available.");    break;
	}

	AddMsg("Nr of bits per <Connectivity Coding Size> field = %d", m_nrBitsConnetivityCounter);
	AddMsg("Nr of bits per <Vertices' Reposition Coding Size> field = %d", m_nrBitsVertsReposCounter);
	AddMsg("Nr of bits per <Vertices' Refinement Coding Size> field = %d", m_nrBitsVertsRefineCounter);

	// dump the description for each ROI
	Cube roi;
	for (int index = 0; index < GetTotalNrROIs(); index++)
	{
		roiDescr.ComputePositionAndSize(roi, index, m_nrROIs, m_levelDiff);

		AddMsg("ROI %d at Pos = (%d, %d, %d), Size = (%d, %d, %d), Connectivity (Stream Pos = %d, nr bytes = %d), Reposition (Stream Pos = %d, nr bytes = %d), Refinement (Stream Pos = %d, nr bytes = %d)",
			index, roi.u, roi.v, roi.w, roi.width, roi.height, roi.depth,
			m_posStreamConnectivity[index], m_connectivityCount[index],
			m_posStreamReposition[index], m_vertsReposCount[index],
			m_posStreamRefine[index], m_vertsRefineCount[index]);
	}

	AddMsg("Nr of bits per <Start Vertex position> field = (%d, %d, %d)",
		m_nrBitsVertPos.u, m_nrBitsVertPos.v, m_nrBitsVertPos.w);
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

GStreamLevel::GStreamLevel()
{
	m_headerSz = -1;
	m_headerPos = -1;
	m_bSplit = false;
	m_nrTiles = -1;
	m_levelDiff = -1;
	m_nrSlices = -1;
	m_threshold[0] = m_threshold[1] = m_threshold[2] = -1;
	m_stopThreshold[0] = m_stopThreshold[1] = m_stopThreshold[2] = 1; //full coding

	m_bRequested = false;
	m_bDecoded = false;

	m_nrBitsTileIndex = -1;
	m_nrBitsGridCounter = -1;

	m_msgs = NULL;
}

int GStreamLevel::Init(const CIPointUVW &nrSlices, const CIPointUVW &nrTiles, const CIPointUVW &levelDiff, int initValue)
{
	m_nrSlices = nrSlices;
	m_levelDiff = levelDiff;
	m_nrTiles = nrTiles;

	// when the nr of slices is less than MIN_SPLIT_SIZE we do not apply the tiling
	// since it is not efficient; we encode it a single cube of size equal to nr slices
	SetSplit(nrSlices.u >= MIN_SPLIT_SIZE || nrSlices.v >= MIN_SPLIT_SIZE || nrSlices.w >= MIN_SPLIT_SIZE);

#ifdef _DEBUG
	if (!CanSplit())
		ASSERT(GetTotalNrTiles() == 1);
#endif

	int totalNrTiles = GetTotalNrTiles();
	ASSERT (totalNrTiles != 0);

	BYTE nrBits = (BYTE) CMGStreamParser::GetPower2(totalNrTiles);
	SetNrBitsTileIndex(nrBits);

	try
	{
		for (int dir = 0; dir <= 2; dir++)
		{
			CIntArray  &tileCount = m_tilesCount[dir];
			CIntArray  &posStream = m_posStreamTile[dir];
			CBoolArray &cacheTile = m_cacheTile[dir];

			tileCount.SetSize(totalNrTiles);
			posStream.SetSize(totalNrTiles);
			cacheTile.SetSize(totalNrTiles);
			for (int tileIdx = 0; tileIdx < totalNrTiles; tileIdx++)
			{
				cacheTile[tileIdx] = false;
				tileCount[tileIdx] = initValue;
				posStream[tileIdx] = -1;
			}
		}
	}
	catch (CMemoryException *e)
	{
		throw CMGDataException(e);
	}

	return totalNrTiles;
}

int GStreamLevel::AddMsg(LPCTSTR lpszFormat, ...)
{
	ASSERT(m_msgs != NULL);
	if (!m_msgs)  return -1;

	CString str;

	va_list argList;
	va_start(argList, lpszFormat);
	str.FormatV(lpszFormat, argList);
	va_end(argList);

	return AddMsg(str);
}

void GStreamLevel::DumpStream(const CIPointUVW &nrSlices, CStringArray &msgs, ROIDescription &roiDescr)
{
	m_msgs = &msgs;

	// grid header information
	AddMsg("Nr of Slices = (%d, %d, %d)", m_nrSlices.u, m_nrSlices.v, m_nrSlices.w);
	const CIPointUVW &nrTiles = GetNrTiles();
	int totalNumTiles = nrTiles.u*nrTiles.v*nrTiles.w;
	AddMsg("Nr of WROIs = (%d, %d, %d), total = %d", nrTiles.u, nrTiles.v, nrTiles.w, totalNumTiles);
	AddMsg("Nr of bits per <Tile index> field = %d", m_nrBitsTileIndex);
	AddMsg("Nr of bits per <Grid Size> field = %d", m_nrBitsGridCounter);

	AddMsg("Threshold X = %d, Y = %d, Z = %d", m_threshold[0], m_threshold[1], m_threshold[2]);

	if (!CanSplit())
	{
		ASSERT(totalNumTiles == 1);

		AddMsg("No splitting possible");

		AddMsg("WROI 0, X (Stream Pos = %d, nr bytes = %d), Y (Stream Pos = %d, nr bytes = %d), Z (Stream Pos = %d, nr bytes = %d)",
			m_posStreamTile[0][0], m_tilesCount[0][0],
			m_posStreamTile[1][0], m_tilesCount[1][0],
			m_posStreamTile[2][0], m_tilesCount[2][0]);

		AddMsg("Tile 0 at Pos = (%d, %d, %d), Size = (%d, %d, %d)",
			0, 0, 0, nrSlices.u, nrSlices.v, nrSlices.w);
	}
	else
	{
		// the cube level is split into the following subbands
		int nrSubbands;
		CIPointUVW sizeL, sizeH;
		int *subbands = CMGGridCodec<float>::CreateSubbands(nrSlices, sizeL, sizeH, nrSubbands);

		// take the Tiles one by one and determine the cubes
		// from every subband belonging to each Tile
		for (int tileIdx = 0; tileIdx < totalNumTiles; tileIdx++)
		{
			Cube roi;
			roiDescr.ComputePositionAndSize(roi, tileIdx, nrTiles, m_levelDiff);

			// position in the bitstream of the WROI
			AddMsg("WROI %d, X (Stream Pos = %d, nr bytes = %d), Y (Stream Pos = %d, nr bytes = %d), Z (Stream Pos = %d, nr bytes = %d)", tileIdx,
				m_posStreamTile[0][tileIdx], m_tilesCount[0][tileIdx],
				m_posStreamTile[1][tileIdx], m_tilesCount[1][tileIdx],
				m_posStreamTile[2][tileIdx], m_tilesCount[2][tileIdx]);

			// map the WROI (or ROI) into tiles
			CCubeArray cubes;
			cubes.SetSize(nrSubbands);

			CMGGridCodec<float>::MapROIToTiles(roi, sizeL, sizeH, subbands, nrSubbands, cubes);

			for (int k = 0; k < nrSubbands; k++)
			{
				AddMsg("Tile %d at Pos = (%d, %d, %d), Size = (%d, %d, %d)",
					k, cubes[k].u, cubes[k].v, cubes[k].w,
					cubes[k].width, cubes[k].height, cubes[k].depth);
			}
		}

		delete []subbands;
	}
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MGStreamDescr::MGStreamDescr()
{
	m_headerSz = -1;
	m_maxNrLevelsM = -1;
	m_maxNrLevelsG = -1;
	m_nrLevelsM = -1;
	m_maxNrSlices = -1;
	m_streamSize = -1;
	m_bSingleLayered = false;
	m_bHasConnectivityInfo = false;
	m_bHasRefineInfo = false;  
	m_bHasRepositionInfo = false;  
	m_bHasGridInfo = false;
	m_bHasNormalsInfo = false;
	m_bInterleaveRefine = false;
	m_nrRefBitsVert = 0;
	m_nrBitsLayer = 0;
	m_nrLayers = 0;
	m_nrQuantBits = 0;
	m_meshType = GENERIC_MESH;
	m_bSameConnect = false;
	m_bAutoResolution = false;
	m_offsetAmp = 0;
	m_cyclicMode = CYCLIC_NONE;

	m_nrBitsFieldGrid = -1;

	m_msgs = NULL;
}

bool MGStreamDescr::IsConnectivityInfoAvailable(int level) const
{
	return (HasConnectivityInfo() &&
		(level == 0 ||
		(level > 0 && !IsAutoResolution())));
}

bool MGStreamDescr::IsRefinementInfoAvailable(int level) const
{
	int maxLevel = GetMaxNrLevelsMesh() - 1;
	if (HasRefineInfo())
	{
		ASSERT(GetNrRefineBitsVertex() > 0);

		return ((level == 0 && IsAutoResolution()) ||
			((level == maxLevel || InterleaveRefine()) &&
			!IsAutoResolution())); 
	}
	else
	{
		ASSERT(GetNrRefineBitsVertex() == 0);
		ASSERT(InterleaveRefine() == false);
		return false;
	}
}

bool MGStreamDescr::IsRepositionInfoAvailable(int level, BYTE layer) const
{
	if (IsRepositionInfoAvailable(level))
	{
		const MStreamLayer &meshLevel  = ((MStreamLevels&)(m_streamLevelM))[level][layer];
		BYTE repositionFlag = meshLevel.GetRepositionBitsMode();

		return (repositionFlag == REPOSITION_ANY);
	}
	else
	{
		return false;
	}
}

bool MGStreamDescr::IsGridInfoAvailable(int level) const
{
	return HasGridInfo();
}

void MGStreamDescr::Init(const CIPointUVW &nrLevels, const CIPointUVW &nrSlices, const CIPointUVW &nrROIs)
{
	// levels Mesh
	m_nrLevelsM = nrLevels;
	m_maxNrLevelsM = max(nrLevels.u, max(nrLevels.v, nrLevels.w));
	m_streamLevelM.SetSize(m_maxNrLevelsM);

	ASSERT((nrSlices.u >> (nrLevels.u - 1)) >= 1 &&
		   (nrSlices.v >> (nrLevels.v - 1)) >= 1 &&
		   (nrSlices.w >> (nrLevels.w - 1)) >= 1);

	// levels Grid
	CWaveletCodec<float>::GetNrResolutionLevels(nrSlices, m_nrLevelsG, m_maxNrLevelsG);
	m_streamLevelG.SetSize(m_maxNrLevelsG);

	// slices
	ASSERT(nrSlices.u > 1 && nrSlices.v > 1);
	m_maxNrSlices = nrSlices;

	bool singleLayered = (nrSlices.w == 1);
	SetSingleLayered(singleLayered);

	// rois
	m_roiDescr.Init(nrSlices, nrROIs);
}

void MGStreamDescr::InitMeshLayers(BYTE nrLayers)
{
	m_nrLayers = nrLayers;

	BYTE bits = CMGStreamParser::GetPower2(nrLayers);
	SetNrBitsLayer(bits);

	for (int k = 0; k < GetMaxNrLevelsMesh(); k++)
	{
		MStreamLayers &meshLevel = GetMeshLevel(k);
		meshLevel.SetSize(nrLayers);
	}
}

void MGStreamDescr::InitGridLevels(BYTE nrLevels)
{
	m_maxNrLevelsG = nrLevels;
	m_streamLevelG.SetSize(nrLevels);
}

void MGStreamDescr::SetThresholds(int dir, const CIntArray &thresholds)
{
	for (int level = 0; level < m_maxNrLevelsG; level++)
	{
		GStreamLevel &gridLevel = GetGridLevel(level);
		gridLevel.SetThreshold(dir, thresholds[level]);
	}
}

int MGStreamDescr::AddMsg(LPCTSTR lpszFormat, ...)
{
	ASSERT(m_msgs != NULL);
	if (!m_msgs)  return -1;

	CString str;

	va_list argList;
	va_start(argList, lpszFormat);
	str.FormatV(lpszFormat, argList);
	va_end(argList);

	return AddMsg(str);
}

void MGStreamDescr::DumpStream(CStringArray &msgs)
{
	m_msgs = &msgs;

	// mesh header
	AddMsg(" ");  AddMsg("GLOBAL MESH DESCRIPTION");    AddMsg(" ");

	AddMsg("Nr of Resolution Levels Mesh = (%d, %d, %d)", m_nrLevelsM.u, m_nrLevelsM.v, m_nrLevelsM.w);
	AddMsg("Nr of Resolution Levels Grid = %d", m_maxNrLevelsG);

	AddMsg("Total Nr of Slices = (%d, %d, %d)", m_maxNrSlices.u, m_maxNrSlices.v, m_maxNrSlices.w);

	const CIPointUVW &nrROIs = m_roiDescr.GetNrROIs();
	AddMsg("Total Nr of ROIs = (%d, %d, %d)", nrROIs.u, nrROIs.v, nrROIs.w);
	const CIPointUVW &szROI = m_roiDescr.GetROISize();
	AddMsg("Size of ROI = (%d, %d, %d)", szROI.u, szROI.v, szROI.w);

	AddMsg("Total Nr of Layers = %d", m_nrLayers);

	// the flags
	AddMsg("Has Connectivity Info = %s", HasConnectivityInfo() ? "true" : "false");
	AddMsg("Has Refine Info = %s", HasRefineInfo() ? "true" : "false");
	AddMsg("Has Reposition Info = %s", HasRepositionInfo() ? "true" : "false");
	AddMsg("Has Grid Info = %s", HasGridInfo() ? "true" : "false");

	// mesh type
	int meshType = GetMeshType();
	switch (meshType)
	{
	case GENERIC_MESH:  AddMsg("Mesh type is generic");      break;
	case TRI_MESH:      AddMsg("MeshType is triangular");    break;
	case QUADRI_MESH:   AddMsg("MeshType is quadrilateral"); break;
	case HEXA_MESH:     AddMsg("MeshType is hexagonal");     break;
	default: ASSERT(0);
	}

	// cyclic mode
	int cyclicMode = GetCyclicMode();
	switch (cyclicMode)
	{
	case CYCLIC_NONE:	AddMsg("Mesh is not cyclic");							break;
	case CYCLIC_U:		AddMsg("Mesh is cyclic in 'U' direction");				break;
	case CYCLIC_UV:		AddMsg("Mesh is cyclic in 'U' and 'V' directions");		break;
	case FOLDED_SINGLE: AddMsg("Mesh is cyclic in 'U' direction and folded in the 'V' direction at the end corresponding to the first index");		break;
	case FOLDED_BOTH:   AddMsg("Mesh is cyclic in 'U' direction and folded in the 'V' direction at both ends");		break;
	default: ASSERT(0);
	}

	AddMsg("Interlevel Full Refine = %s", m_bInterleaveRefine ? "true" : "false");
	AddMsg("Nr of Refine Bits Vertices = %d", m_nrRefBitsVert);

	// grid header
	AddMsg(" ");  AddMsg("GLOBAL GRID DESCRIPTION");    AddMsg(" ");
	switch (m_filterType)
	{
	case CMGGrid::LINEAR_FILTER:  AddMsg("Interpolation Filter type: LINEAR");  break;
	case CMGGrid::SMOOTH_FILTER:  AddMsg("Interpolation Filter type: SMOOTH");  break;
	default:                      AddMsg("Interpolation Filter type: UNKNOWN");  break;
	}
	AddMsg("Nr of bits per <Grid Scale> field = %d", m_nrBitsFieldGrid);
	AddMsg("Reference Grid X: Scale = %d", m_gridCoding[0].scale);
	AddMsg("Reference Grid Y: Scale = %d", m_gridCoding[1].scale);
	AddMsg("Reference Grid Z: Scale = %d", m_gridCoding[2].scale);

	AddMsg("Nr of bits per <Grid Corner> field = %d", m_nrQuantBits);
	for (int nrCorner = 0; nrCorner < 8; nrCorner++)
	{
		AddMsg("Grid Corner %d = (%d, %d, %d)", nrCorner,
			m_gridCoding[0].gridCorners[nrCorner],
			m_gridCoding[1].gridCorners[nrCorner], 
			m_gridCoding[2].gridCorners[nrCorner]);
	}

	// mesh level information
	int endLevel = m_maxNrLevelsG - m_maxNrLevelsM;
	for (int level = 0; level < m_maxNrLevelsM; level++)
	{
		if (HasConnectivityInfo() && (level == 0 || !IsAutoResolution()))
		{
			AddMsg(" ");  AddMsg("MESH DESCRIPTION LEVEL %d", level);   AddMsg(" ");
			for (int layer = 0; layer < m_nrLayers; layer++)
			{
				AddMsg(" ");  AddMsg("MESH DESCRIPTION LAYER %d", layer);   AddMsg(" ");
				m_streamLevelM[level][layer].DumpStream(msgs, m_roiDescr);
			}
		}

		// grid level information
		if (HasGridInfo())
		{

			if (level == 0)
			{
				for (int subLevel = 0; subLevel <= endLevel; subLevel++)
				{
					AddMsg(" ");  AddMsg("GRID DESCRIPTION LEVEL %d", subLevel);   AddMsg(" ");

					CIPointUVW nrSlices = GetMaxNrSlices();
					CMGGrid::ComputeNrSlicesDown(nrSlices, m_maxNrLevelsG - 1 - subLevel);
					m_streamLevelG[subLevel].DumpStream(nrSlices, msgs, GetROIDescr());
				}
			}
			else
			{
				int subLevel = level + endLevel;
				if (subLevel < 0) continue;

				AddMsg(" ");  AddMsg("GRID DESCRIPTION LEVEL %d", subLevel);   AddMsg(" ");

				CIPointUVW nrSlices = GetMaxNrSlices();
				CMGGrid::ComputeNrSlicesDown(nrSlices, m_maxNrLevelsG - 1 - subLevel);
				m_streamLevelG[subLevel].DumpStream(nrSlices, msgs, GetROIDescr());
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

RGCoding::RGCoding()
{
	scale = -1;

	for (int k = 0; k < 8; k++)
		gridCorners[k] = 0;
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

MGStream::MGStream()
{
	m_pFile = NULL;
	m_pStream = NULL;
	m_bAllocInternal = false;
}

MGStream::~MGStream()
{
	ASSERT(m_pStream == NULL);
}

void MGStream::DeleteStream()
{
	delete m_pStream;   m_pStream = NULL;
	if (m_bAllocInternal)
	{
		m_bAllocInternal = false;
		delete m_pFile;		m_pFile = NULL;
	}
}

void MGStream::Initialise(CFile *file, CBitStream::BITSTREAM_OPERATIONS op)
{
	ASSERT(m_pStream == NULL);
	ASSERT(m_pFile == NULL);

	m_bAllocInternal = false;
	m_pFile = file;
	m_pStream = new CBitStream(file, op);
}

CBitStream * MGStream::InitStream(int limit)
{
	ASSERT(m_pStream == 0);

	if (m_pFile != NULL)
	{
		m_pFile->SeekToBegin();
	}
	else
	{
		m_bAllocInternal = true;
		m_pFile = new CMemFile;
	}

	m_pStream = new CBitStream(m_pFile, CBitStream::OP_WRITE, limit);
	return m_pStream;
}

void MGStream::SetPosition(int pos)
{
	m_pFile->Seek(pos, CFile::begin);
	m_pStream->Flush();
}

void MGStream::MovePosition(int pos)
{
	m_pFile->Seek(pos, CFile::current);
	m_pStream->Flush();
}

void MGStream::SetPositionAndLimit(int pos, int limit, CBitStream::BITSTREAM_OPERATIONS op)
{
	m_pFile->Seek(pos, CFile::begin);
	delete m_pStream;   m_pStream = NULL;
	m_pStream = new CBitStream(m_pFile, op, limit);
}

bool MGStream::RawRead(BYTE *buf, int pos, int count)
{
	if (!m_pFile) return false;

	try
	{
		m_pFile->Seek(pos, CFile::begin);
		m_pFile->Read(buf, count);
	}
	catch (CFileException *)
	{
		return false;
	}
	return true;
}