/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGStreamEncoder.cpp: implementation of the CMGStreamParser class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGStreamParse.h"
#include "MGGridCodec.h"
#include "BitStream.h"
#include "MGErrMsg.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


// storing the main header of the stream
void CMGStreamParser::StoreMeshGridHeader(bool keepHeader)
{
	m_decSpecInfoFile = fopen("MGDecSpecInfo.txt", "wt");
	if (m_decSpecInfoFile == NULL)
	{
		AfxMessageBox("Cannot open decoder specific info file");	return;
	}

	int partSize = 0;

	// store the mesh description
	StoreNrLevels(keepHeader, false, &partSize);
	StoreNrSlices(keepHeader, false, &partSize);
	StoreNrROIs(keepHeader, false, &partSize);
	StoreFlags(keepHeader, false, &partSize);
	StoreReservedBits(keepHeader, false, &partSize);
	StoreNrLayers(keepHeader, false, &partSize);
	StoreMeshType(keepHeader, false, &partSize);
	StoreCyclicMode(keepHeader, false, &partSize);
	StoreNrRefineBitsVertex(keepHeader, false, &partSize);

	// store the grid description
	StoreFilterType(keepHeader, false, &partSize);
	StoreGridCodingParam(keepHeader, false, &partSize);
	StoreGridCorners(keepHeader, false, &partSize);

	if (keepHeader)
	{
		partSize = AlignInBits(partSize);
		StoreTagAndSize(MGInfoTag, partSize);
	}

	// store the mesh description
	StoreNrLevels(keepHeader, true, NULL);
	StoreNrSlices(keepHeader, true, NULL);
	StoreNrROIs(keepHeader, false, NULL);
	StoreFlags(keepHeader, true, NULL);
	StoreReservedBits(keepHeader, true, NULL);
	StoreNrLayers(keepHeader, true, NULL);
	StoreMeshType(keepHeader, true, NULL);
	StoreCyclicMode(keepHeader, true, NULL);
	StoreNrRefineBitsVertex(keepHeader, true, NULL);

	// store the grid description
	StoreFilterType(keepHeader, true, NULL);
	StoreGridCodingParam(keepHeader, true, NULL);
	StoreGridCorners(keepHeader, true, NULL);

	fclose(m_decSpecInfoFile);
	m_decSpecInfoFile = NULL;

	// copy the stream to file
	FlushBitStream();
}

// copy the stream to the output file
void CMGStreamParser::WriteOutput(CFile *file)
{
	BYTE buffer[1024];

	// get the size of the stream
	int nrBytes = m_streamMG.GetPosition();
	m_streamMG.SeekToBegin();

	// create output stream
	MGStream stream;
	stream.Initialise(file, CBitStream::OP_WRITE);

	StoreValue(stream, nrBytes, NULL);

	// copy the stream data
	while (nrBytes > 0)
	{
		int sz = min(1024, nrBytes);
		int count = m_streamMG.Read(buffer, sz);
		ASSERT(count == sz);

		stream.Write(buffer, sz);
		nrBytes -= sz;
	}

	// flush the buffers and destroy
	stream.DeleteStream();
}

void CMGStreamParser::StoreValue(MGStream &stream, int number, int *partSize)
{
	// only positive numbers
	BYTE tmp;
	ASSERT(number >= 0);

	for (int i = 21; i > 0; i -= 7)
	{
		if (number >= (1 << i))
		{
			if (!partSize)
				stream.WriteBit(1);

			tmp = (number >> i) & 0x7f;

			if (!partSize)
				stream.WriteBits(tmp, 7);
			else
				*partSize += 8;
		}
	}
	if (!partSize)
	{
		stream.WriteBit(0);
		tmp = number & 0x7f;
		stream.WriteBits(tmp, 7);
	}
	else 
		*partSize += 8;
}

// storing the tag identifier and size
void CMGStreamParser::StoreTagAndSize(BYTE tag, int size)
{
	int sizeBytes = (size + 7) / 8;
	ASSERT((8 * sizeBytes) == size);
	m_streamMG.WriteBits(tag, TAG_BITS);
	StoreValue(m_streamMG, sizeBytes, NULL);

	TRACE3("Tag = %d, Size = (%d, %d)\n", tag, size, sizeBytes);

	FlushBitStream();
}

// storing the begin frame tag
void CMGStreamParser::StoreBeginFrame()
{
	m_streamMG.WriteBits((BYTE)MGBeginFrameTag, TAG_BITS);
	StoreValue(m_streamMG, 0, NULL);

	TRACE2("Tag = %d, Size = (%d)\n", MGBeginFrameTag, 0);

	FlushBitStream();
}

// storing the level description
void CMGStreamParser::StoreLevelDescription(int level, BYTE update_replace, int *partSize)
{
	if (partSize == NULL)
	{
		// write the level and flag
		m_streamMG.WriteBits((DWORD)level, LEVEL_BITS);
		m_streamMG.WriteBits(update_replace, UPDATE_REPLACE_BITS);
	}
	else
	{
		*partSize += (LEVEL_BITS + UPDATE_REPLACE_BITS);
	}
}

// storing the layer description
void CMGStreamParser::StoreLayerDescription(int level, BYTE update_replace, bool allLayers, int *partSize)
{
	StoreLevelDescription(level, update_replace, partSize);

	BYTE nBitsLayer = m_streamDescr.GetNrBitsLayer();

	if (nBitsLayer > 0)
	{
		if (allLayers)
			StoreDefaultIndex(nBitsLayer, partSize);
		else
			StoreIndices(nBitsLayer, partSize);
	}
}

void CMGStreamParser::StoreIndices(BYTE nBitsLayer, int *partSize)
{
	if (partSize != NULL)
	{
		*partSize += (nBitsLayer + (nBitsLayer * m_layerGroups.GetSize()) + m_isIndexed.GetSize());
	}
	else
	{
		// write the number of elements in the indices list.
		DWORD number = m_isIndexed.GetSize();
		m_streamMG.WriteBits(number, nBitsLayer);

		int counter = 0;
		for (DWORD i = 0; i < number; i++)
		{
			BYTE isIndexed = m_isIndexed[i] ? 1 :  0;
			m_streamMG.WriteBit(isIndexed);

			// check if indexed or list
			if (isIndexed == 1)
			{
				// write the index
				m_streamMG.WriteBits((DWORD)m_layerGroups[counter++], nBitsLayer);
			}
			else
			{
				// write the start index and count from the list
				m_streamMG.WriteBits((DWORD)m_layerGroups[counter++], nBitsLayer);
				m_streamMG.WriteBits((DWORD)m_layerGroups[counter++], nBitsLayer);
			}
		}

		ASSERT(counter == m_layerGroups.GetSize());
	}
}

void CMGStreamParser::StoreDefaultIndex(BYTE nBitsLayer, int *partSize)
{
	if (partSize != NULL)
	{
		*partSize += nBitsLayer;
	}
	else
	{
		// no indices in the list.
		DWORD number = 0;
		m_streamMG.WriteBits(number, nBitsLayer);
	}
}

void CMGStreamParser::SetActiveLayers(const CIntArray &layers)
{
	int k;
	int nrLayers = layers.GetSize();

	// the actual size might be smaller, shrink at the end
	m_layers.SetSize(nrLayers);
	m_layerGroups.SetSize(nrLayers);
	m_isIndexed.SetSize(nrLayers);

	// find the number of layer groups
	if (nrLayers == 0)		return;

	// copy the layers
	for (k = 0; k < nrLayers; k++)
		m_layers[k] = layers[k];

	// group the layers in ascending and consecutive order
	// it is assumed that the input layers are already sorted in ascending order
	int nextIndex = 0, nextElement = 0, nextGroup = 0;

	int layer1 = layers[nextIndex];
	m_layerGroups[nextElement++] = layers[nextIndex++];
	m_isIndexed[nextGroup++] = true;

	for (k = 1; k < nrLayers; k++)
	{
		nextIndex = k;
		layer1 = layers[nextIndex - 1];

		while (nextIndex < nrLayers && (layer1 + 1) == layers[nextIndex])
			layer1 = layers[nextIndex++];

		if (nextIndex != k)
		{
			// sequential mode
			m_isIndexed[nextGroup-1] = false;
			m_layerGroups[nextElement++] = nextIndex - k + 1;
			
			k = nextIndex;
		}

		if (nextIndex < nrLayers)
		{
			m_isIndexed[nextGroup++] = true;
			m_layerGroups[nextElement++] = layers[nextIndex++];
		}
	}

	// shrink the arrays
	m_isIndexed.SetSize(nextGroup);
	m_layerGroups.SetSize(nextElement);
}

// storing mesh related information of the main header of the stream
void CMGStreamParser::StoreNrLevels(bool inBinary, bool inASCII, int *partSize)
{
	const CIPointUVW &nrLevels = m_streamDescr.GetNrLevelsMesh();

	if (inBinary)
	{
		if (partSize == NULL)
		{
			// write the fields
			m_streamMG.WriteBits((DWORD)nrLevels.u, LEVEL_BITS);
			m_streamMG.WriteBits((DWORD)nrLevels.v, LEVEL_BITS);
			m_streamMG.WriteBits((DWORD)nrLevels.w, LEVEL_BITS);
		}
		else
		{
			*partSize += (3 * LEVEL_BITS);
		}
	}

	if (inASCII)
		fprintf(m_decSpecInfoFile, "nLevels [ %d %d %d ]\n",
		nrLevels.u, nrLevels.v, nrLevels.w);
}

void CMGStreamParser::StoreNrSlices(bool inBinary, bool inASCII, int *partSize)
{
	const CIPointUVW &nrSlices = m_streamDescr.GetMaxNrSlices();

	if (inBinary)
	{
		StoreValue(m_streamMG, nrSlices.u, partSize);
		StoreValue(m_streamMG, nrSlices.v, partSize);
		StoreValue(m_streamMG, nrSlices.w, partSize);
	}

	if (inASCII)
		fprintf(m_decSpecInfoFile, "nSlices [ %d %d %d ]\n",
		nrSlices.u, nrSlices.v, nrSlices.w);
}

void CMGStreamParser::StoreNrROIs(bool inBinary, bool inASCII, int *partSize)
{
	const CIPointUVW &nrROIs = m_streamDescr.GetNrROIs();

	if (inBinary)
	{
		StoreValue(m_streamMG, nrROIs.u, partSize);
		StoreValue(m_streamMG, nrROIs.v, partSize);
		StoreValue(m_streamMG, nrROIs.w, partSize);
	}

	if (inASCII)
		fprintf(m_decSpecInfoFile, "nROIS [ %d %d %d ]\n",
		nrROIs.u, nrROIs.v, nrROIs.w);
}

void CMGStreamParser::StoreFlags(bool inBinary, bool inASCII, int *partSize)
{
	if (inBinary)
	{
		if (partSize == NULL)
		{
			m_streamMG.WriteBit(m_streamDescr.HasConnectivityInfo() ? 1: 0);
			m_streamMG.WriteBit(m_streamDescr.HasRefineInfo() ? 1: 0);
			m_streamMG.WriteBit(m_streamDescr.HasRepositionInfo() ? 1: 0);
			m_streamMG.WriteBit(m_streamDescr.HasGridInfo() ? 1: 0);
		}
		else
			*partSize += 4;
	}

	if (inASCII)
	{
		fprintf(m_decSpecInfoFile, "hasConnectivityInfo %d\n",
			m_streamDescr.HasConnectivityInfo() ? 1: 0);
		fprintf(m_decSpecInfoFile, "hasRefineInfo %d\n",
			m_streamDescr.HasRefineInfo() ? 1: 0);
		fprintf(m_decSpecInfoFile, "hasRepositionInfo %d\n",
			m_streamDescr.HasRepositionInfo() ? 1: 0);
		fprintf(m_decSpecInfoFile, "hasGridInfo %d\n",
			m_streamDescr.HasGridInfo() ? 1: 0);
	}
}

void CMGStreamParser::StoreMeshType(bool inBinary, bool inASCII, int *partSize)
{
	BYTE meshType = m_streamDescr.GetMeshType();

	if (inBinary)
	{
		if (partSize == NULL)
			m_streamMG.WriteBits(meshType, MESH_TYPE_BITS);
		else
			*partSize += MESH_TYPE_BITS;
	}

	if (inASCII)
		fprintf(m_decSpecInfoFile, "meshType %d\n", meshType);

	if (meshType == QUADRI_MESH)
	{
		if (m_streamDescr.HasConnectivityInfo())
		{
			bool flag;
			flag = m_streamDescr.IsSameConnectivity();

			if (inBinary)
			{
				if (partSize == NULL)
					m_streamMG.WriteBit(flag ? 1 : 0);
				else
					*partSize += 1;
			}

			if (inASCII)
				fprintf(m_decSpecInfoFile, "sameBorderOrientation %d\n", flag ? 1 : 0);

			flag = m_streamDescr.IsAutoResolution();

			if (inBinary)
			{
				if (partSize == NULL)
					m_streamMG.WriteBit(flag ? 1 : 0);
				else
					*partSize += 1;
			}

			if (inASCII)
				fprintf(m_decSpecInfoFile, "uniformSplit %d\n", flag ? 1 : 0);
		}

		if (m_streamDescr.IsSingleLayered())
		{
			int offsetAmplitude = m_streamDescr.GetOffsetAmplitude();

			if (inBinary)
			{
				StoreValue(m_streamMG, offsetAmplitude, partSize);
			}

			if (inASCII)
				fprintf(m_decSpecInfoFile, "offsetAmplitude %d\n", offsetAmplitude);
		}
	}
}

void CMGStreamParser::StoreNrRefineBitsVertex(bool inBinary, bool inASCII, int *partSize)
{
	// store whether the refine bits are interleaved in the stream or not
	if (m_streamDescr.HasRefineInfo())
	{
		BYTE flag = m_streamDescr.InterleaveRefine() ? TRUE : FALSE;

		if (inBinary)
		{
			if (partSize == NULL)
				m_streamMG.WriteBit(flag);
			else
				*partSize += 1;
		}

		if (inASCII)
			fprintf(m_decSpecInfoFile, "fullRefine %d\n", flag);

		// store the number of refine bits
		BYTE nrBits = m_streamDescr.GetNrRefineBitsVertex();
		ASSERT(nrBits >= 0 && nrBits <= MAX_REFINE_BITS);

		if (inBinary)
		{
			if (partSize == NULL)
				m_streamMG.WriteBits(nrBits, REFINE_BITS);
			else
				*partSize += REFINE_BITS;
		}

		if (inASCII)
			fprintf(m_decSpecInfoFile, "nRefineBits %d\n", nrBits);
	}
}

void CMGStreamParser::StoreCyclicMode(bool inBinary, bool inASCII, int *partSize)
{
	BYTE flag = (BYTE) m_streamDescr.GetCyclicMode();

	if (inBinary)
	{
		if (partSize == NULL)
			m_streamMG.WriteBits(flag, CYCLIC_MODE_BITS);
		else
			*partSize += CYCLIC_MODE_BITS;
	}

	if (inASCII)
		fprintf(m_decSpecInfoFile, "cyclicMode %d\n", flag);
}

void CMGStreamParser::StoreReservedBits(bool inBinary, bool inASCII, int *partSize)
{
	WORD bits = 0;

	if (inBinary)
	{
		if (partSize == NULL)
			m_streamMG.WriteBits(bits, RESERVED_BITS);
		else
			*partSize += RESERVED_BITS;
	}

	if (inASCII)
		fprintf(m_decSpecInfoFile, "attributes %d\n", bits);
}

void CMGStreamParser::StoreNrLayers(bool inBinary, bool inASCII, int *partSize)
{
	BYTE nrBits = m_streamDescr.GetNrBitsLayer();
	int nrLayers = m_streamDescr.GetNrLayers();

	if (inBinary)
	{
		if (partSize == NULL)
		{
			m_streamMG.WriteBit(nrLayers > 1 ? 1 : 0);
			if (nrLayers > 1)
				StoreValue(m_streamMG, nrLayers, NULL);
		}
		else
		{
			*partSize += 1;
			if (nrLayers > 1)
				StoreValue(m_streamMG, nrLayers, partSize);
		}
	}

	if (inASCII)
		fprintf(m_decSpecInfoFile, "nLayers %d\n", nrLayers);
}

// storing grid related information of the main header of the stream
void CMGStreamParser::StoreFilterType(bool inBinary, bool inASCII, int *partSize)
{
	BYTE bValue = m_streamDescr.GetFilterType();

	if (inBinary)
	{
		if (partSize == NULL)
			m_streamMG.WriteBits(bValue, FILTER_BITS);
		else
			*partSize += FILTER_BITS;
	}

	if (inASCII)
		fprintf(m_decSpecInfoFile, "filterType %d\n", bValue);
}

void CMGStreamParser::StoreGridCodingParam(bool inBinary, bool inASCII, int *partSize)
{
	RGCoding * codingRG = m_streamDescr.GetGridCodingParam();

	int maxVal = max(codingRG[X_GRID].scale, max(codingRG[Y_GRID].scale, codingRG[Z_GRID].scale));
	BYTE nrBits = (BYTE)GetNrBits(maxVal);
	ASSERT(nrBits <= MAX_FIELD_BITS);

	m_streamDescr.SetNrBitsGridParam(nrBits);

	// write the number of bits allocated for the variables
	if (inBinary)
	{
		if (partSize == NULL)
		{
			m_streamMG.WriteBits(nrBits, FIELD_BITS);

			// write the scale
			for (int dir = X_GRID; dir <= Z_GRID; dir++)
			{
				DWORD wValue;
				wValue = codingRG[dir].scale;   m_streamMG.WriteBits(wValue, nrBits);
			}
		}
		else
			*partSize += (FIELD_BITS + 3 * nrBits);
	}

	if (inASCII)
		fprintf(m_decSpecInfoFile, "gridScale [ %d %d %d ]\n",
		codingRG[0].scale, codingRG[1].scale, codingRG[2].scale);
}

void CMGStreamParser::StoreGridCorners(bool inBinary, bool inASCII, int *partSize)
{
	// store the number of quantization bits of the grid values
	BYTE nrQuantBits = m_streamDescr.GetNrQuantBitsGrid();
	ASSERT(nrQuantBits >= 1);

	if (inBinary)
	{
		if (partSize == NULL)
			m_streamMG.WriteBits(nrQuantBits, QUANT_BITS);
		else
			*partSize += QUANT_BITS;
	}

	if (inASCII)
	{
		fprintf(m_decSpecInfoFile, "nQuantBits %d\n",nrQuantBits);

		// store the grid corner values
		fprintf(m_decSpecInfoFile, "gridCorners [\n");
	}

	// which grid corners to write to stream
	int uMax = 1, vMax = 1, wMax = 1;
	const CIPointUVW &nrSlices = m_streamDescr.GetMaxNrSlices();

	// check for single layered
	if (nrSlices.w == 1) wMax = 0;  // only layer 0

	// check for cyclic
	if (m_streamDescr.IsCyclicU())	uMax = 0;		// only one corner in the u direction
	if (m_streamDescr.IsCyclicV())	vMax = 0;		// only one corner in the v direction

	int corners[3];
	for (int w = 0; w <= wMax; w++)
	{
		for (int v = 0; v <= vMax; v++)
		{
			for (int u = 0; u <= uMax; u++)
			{
				for (int dir = X_GRID; dir <= Z_GRID; dir++)
				{
					const RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
					corners[dir] = codingRG.gridCorners[4*w + 2*v + u];

					if (inBinary)
					{
						if (partSize == NULL)
						{
							if (corners[dir] < 0)
							{
								m_streamMG.WriteBit(1);    m_streamMG.WriteBits((DWORD)-corners[dir], nrQuantBits);
							}
							else
							{
								m_streamMG.WriteBit(0);    m_streamMG.WriteBits((DWORD)corners[dir], nrQuantBits);
							}
						}
						else
							*partSize += (1 + nrQuantBits);
					}
				}

				if (inASCII)
					fprintf(m_decSpecInfoFile, "\t%d %d %d\n", corners[0], corners[1], corners[2]);
			}
		}
	}

	if (inASCII)
		fprintf(m_decSpecInfoFile, "\t]");
}

void CMGStreamParser::StoreGridCorners(bool differential, BYTE nrBits)
{
	// align data
	FlushBitStream();

	// compute size of the bitstream
	int partSize = 0;
	StoreGridCornersBits(differential, nrBits, &partSize);
	partSize = AlignInBits(partSize);

	// store the tag and size
	StoreTagAndSize(MGGridCornersInfoTag, partSize);

	// store the corners bits to stream
	StoreGridCornersBits(differential, nrBits, NULL);

	FlushBitStream();
}

void CMGStreamParser::StoreGridCornersBits(bool differential, BYTE nrQuantBits, int *partSize)
{
	if (partSize == NULL)
	{
		// store the update/replace flag
		BYTE update_replace = differential ? UPDATE_DATA : REPLACE_DATA;
		m_pMemStream->WriteBits(update_replace, UPDATE_REPLACE_BITS);

		// store the number of quantization bits of the grid values
		ASSERT(nrQuantBits >= 1);
		m_pMemStream->WriteBits(nrQuantBits, QUANT_BITS);
	}
	else
	{
		// store the update/replace flag
		*partSize += UPDATE_REPLACE_BITS;

		// store the number of quantization bits of the grid values
		*partSize += QUANT_BITS;
	}

	// which grid corners to write to stream
	int uMax = 1, vMax = 1, wMax = 1;
	const CIPointUVW &nrSlices = m_streamDescr.GetMaxNrSlices();

	// check for single layered
	if (nrSlices.w == 1) wMax = 0;  // only layer 0

	// check for cyclic
	if (m_streamDescr.IsCyclicU())	uMax = 0;		// only one corner in the u direction
	if (m_streamDescr.IsCyclicV())	vMax = 0;		// only one corner in the v direction

	int corners[3];
	for (int w = 0; w <= wMax; w++)
	{
		for (int v = 0; v <= vMax; v++)
		{
			for (int u = 0; u <= uMax; u++)
			{
				for (int dir = X_GRID; dir <= Z_GRID; dir++)
				{
					const RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
					corners[dir] = codingRG.gridCorners[4*w + 2*v + u];

					if (partSize == NULL)
					{
						if (corners[dir] < 0)
						{
							m_pMemStream->WriteBit(1);    m_pMemStream->WriteBits((DWORD)-corners[dir], nrQuantBits);
						}
						else
						{
							m_pMemStream->WriteBit(0);    m_pMemStream->WriteBits((DWORD)corners[dir], nrQuantBits);
						}
					}
					else
					{
						*partSize += (1 + nrQuantBits);
					}
				}
			}
		}
	}
}

// storing mesh related information of the level header of the stream

void CMGStreamParser::StoreMeshHeaderLevelDetails(int level, BYTE layer, int *partSize)
{
	// store the number of bits for storing the coding counters for
	// connectivity, reposition and refine
	StoreMeshCodingHeader(level, layer, partSize);

	if (partSize == NULL)
		FlushBitStream();
	else
		*partSize = AlignInBits(*partSize);
}

void CMGStreamParser::ComputeMaxCounter(int level, BYTE layer,
										 int &maxCountConnectivity, int &maxCountReposition, int &maxCountRefinement)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);

	CIntArray &connectivityCount = meshLevel.GetNrBytesConnectivityCoding();
	CIntArray &repositionCount   = meshLevel.GetNrBytesVertsReposCoding();
	CIntArray &refinementCount   = meshLevel.GetNrBytesVertsRefineCoding();

	// get the max count
	int nrROIs = meshLevel.GetTotalNrROIs();

	for (int k = 0; k < nrROIs; k++)
	{
		if (m_streamDescr.IsConnectivityInfoAvailable(level) && connectivityCount[k] > maxCountConnectivity)   
			maxCountConnectivity = connectivityCount[k];

		if (m_streamDescr.IsRepositionInfoAvailable(level, layer) && repositionCount[k] > maxCountReposition)   
			maxCountReposition = repositionCount[k];

		if (m_streamDescr.IsRefinementInfoAvailable(level) && refinementCount[k] > maxCountRefinement)   
			maxCountRefinement = refinementCount[k];
	}
}

void CMGStreamParser::ComputeCounterSize(int level, BYTE layer,
										 int maxCountConnectivity, int maxCountReposition, int maxCountRefinement)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);

	// number of bits for storing the counters
	BYTE nrBits;
	if (m_streamDescr.IsConnectivityInfoAvailable(level))
	{
		nrBits = GetPower2(maxCountConnectivity);
		ASSERT(nrBits <= MAX_VERTS_BITS);

		meshLevel.SetNrBitsConnectivityCounterField(nrBits);
	}

	if (m_streamDescr.IsRepositionInfoAvailable(level, layer))
	{
		// for the reposition counters
		nrBits = GetPower2(maxCountReposition);
		ASSERT(nrBits <= MAX_VERTS_BITS);

		meshLevel.SetNrBitsVertsReposCounterField(nrBits);
	}

	if (m_streamDescr.IsRefinementInfoAvailable(level))
	{
		// for the refinement counters
		nrBits = GetPower2(maxCountRefinement);
		ASSERT(nrBits <= MAX_VERTS_BITS);

		meshLevel.SetNrBitsVertsRefineCounterField(nrBits);
	}
}

void CMGStreamParser::StoreMeshCodingHeader(int level, BYTE layer, int *partSize)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);

	// store the reposition flag
	if (m_streamDescr.IsRepositionInfoAvailable(level, layer))
	{
		BYTE repositionFlag = meshLevel.GetRepositionBitsMode();

		if (partSize == NULL)
			m_streamMG.WriteBits(repositionFlag, REPOS_MODE_BITS);
		else
			*partSize += REPOS_MODE_BITS;
	}

	// number of bits for storing the counters
	BYTE nrBits;

	// for the connectivity counters
	if (m_streamDescr.IsConnectivityInfoAvailable(level))
	{
		nrBits = meshLevel.GetNrBitsConnectivityCounterField();
		if (partSize == NULL)
			m_streamMG.WriteBits(nrBits, VERTS_BITS);
		else
			*partSize += VERTS_BITS;
	}

	if (m_streamDescr.IsRepositionInfoAvailable(level, layer))
	{
		// for the reposition counters
		nrBits = meshLevel.GetNrBitsVertsReposCounterField();
		if (partSize == NULL)
			m_streamMG.WriteBits(nrBits, VERTS_BITS);
		else
			*partSize += VERTS_BITS;
	}

	if (m_streamDescr.IsRefinementInfoAvailable(level))
	{
		// for the refinement counters
		nrBits = meshLevel.GetNrBitsVertsRefineCounterField();
		if (partSize == NULL)
			m_streamMG.WriteBits(nrBits, VERTS_BITS);
		else
			*partSize += VERTS_BITS;
	}
}

// storing grid related information of the level header of the stream
void CMGStreamParser::StoreGridHeaderLevelDetails(int levelG, int *partSize)
{
	// store the grid coding header for the level
	StoreThreshold(levelG, partSize);
	StoreGridCodingHeader(levelG, partSize);

	if (partSize == NULL)
		FlushBitStream();
	else
		*partSize = AlignInBits(*partSize);
}

void CMGStreamParser::StoreThreshold(int level, int *partSize)
{
	const GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(level);
	for (int dir = X_GRID; dir <= Z_GRID; dir++)
	{
		BYTE bValue;
		bValue = GetPower2(gridLevel.GetThreshold(dir));
		if (partSize == NULL)
			m_streamMG.WriteBits(bValue, FIELD_BITS);
		else
			*partSize += FIELD_BITS;
	}  
}

void CMGStreamParser::StoreGridCodingHeader(int levelG, int *partSize)
{
	// write the counters for the tiles
	ASSERT(levelG >= 0 && levelG < m_streamDescr.GetMaxNrLevelsGrid());

	GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(levelG);

	int maxSize = 0;
	for (int dir = X_GRID; dir <= Z_GRID; dir++)
	{
		const CIntArray &size = gridLevel.GetNrBytesTileCoding(dir);
		for (int k = 0; k < gridLevel.GetTotalNrTiles(); k++)
			maxSize = max(maxSize, size[k]);
	}

	// can be 0 bits
	BYTE nrBits = (BYTE)GetPower2(maxSize);
	ASSERT(nrBits <= MAX_GRID_BITS);

	// store the number of bits allocated for the variables
	if (partSize == NULL)
	{
		m_streamMG.WriteBits(nrBits, GRID_BITS);
		gridLevel.SetNrBitsGridCounterField(nrBits);
	}
	else
	{
		*partSize += GRID_BITS;
	}
}

void CMGStreamParser::StoreTilesCodingHeader(int levelG, int *partSize)
{
	// write the counters for the tiles
	ASSERT(levelG >= 0 && levelG < m_streamDescr.GetMaxNrLevelsGrid());

	// store the number of bits allocated for the variables
	GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(levelG);
	BYTE nrBits = gridLevel.GetNrBitsGridCounterField();

	// write the counters
	if (nrBits > 0)
	{
		int nrTiles = gridLevel.GetTotalNrTiles();
		for (int k = 0; k < nrTiles; k++)
		{
			for (int dir = X_GRID; dir <= Z_GRID; dir++)
			{
				int nrBytesTile = gridLevel.GetNrBytesTileCoding(dir, k);
				BYTE bValue = nrBytesTile != 0 ? 1: 0;

				if (partSize == NULL)
				{
					m_streamMG.WriteBit(bValue);

					if (bValue != 0)
					{
						DWORD value = nrBytesTile;    m_streamMG.WriteBits(value, nrBits);
					}
				}
				else
				{
					++(*partSize);
					if (bValue != 0)
						*partSize += nrBits;
				}
			}
		}

		if (partSize == NULL)
			FlushBitStream();
		else
			*partSize = AlignInBits(*partSize);
	}
}

void CMGStreamParser::StoreROICodingLookup(int level, BYTE layer, int *partSize)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	CBoolArray &codingLookup = meshLevel.GetCodingLookup();

	// write the ROI coding lookup
	int nrROIs = meshLevel.GetTotalNrROIs();
	for (int k = 0; k < nrROIs; k++)
	{
		BYTE bValue = codingLookup[k] ? 1: 0;
		if (partSize == NULL)
			m_streamMG.WriteBit(bValue);
		else
			++(*partSize);
	}

	if (partSize == NULL)
		FlushBitStream();
	else
		*partSize = AlignInBits(*partSize);
}

void CMGStreamParser::StoreConnectivityCodingHeader(int level, BYTE layer, int *partSize)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	BYTE nrBits = meshLevel.GetNrBitsConnectivityCounterField();
	CIntArray &connectivityCount = meshLevel.GetNrBytesConnectivityCoding();
	CBoolArray &codingLookup     = meshLevel.GetCodingLookup();

	// write the counters
	int nrROIs = meshLevel.GetTotalNrROIs();
	for (int k = 0; k < nrROIs; k++)
	{
		if (codingLookup[k])
		{
			ASSERT(connectivityCount[k] != 0);
			if (partSize == NULL)
				m_streamMG.WriteBits((DWORD)connectivityCount[k], nrBits);
			else
				*partSize += nrBits;
		}
#ifdef _DEBUG
		else
			ASSERT(connectivityCount[k] == 0);
#endif
	}

	if (partSize == NULL)
		FlushBitStream();
	else
		*partSize = AlignInBits(*partSize);
}

void CMGStreamParser::StoreCodedConnectivity(int level, BYTE layer, int *partSize)
{
	BYTE buffer[1024];
	MStreamLayer & meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	CIntArray & connectivityCoding = meshLevel.GetNrBytesConnectivityCoding();
	int nrROIs = meshLevel.GetTotalNrROIs();

	int nrBytes = 0;
	for (int roiIdx = 0; roiIdx < nrROIs; roiIdx++)
		nrBytes += connectivityCoding[roiIdx];

	m_streamConnectivity[layer].SeekToBegin();

	if (partSize == NULL)
	{
		while (nrBytes > 0)
		{
			int sz = min(1024, nrBytes);
			int count = m_streamConnectivity[layer].Read(buffer, sz);
			ASSERT(count == sz);

			m_streamMG.Write(buffer, sz);
			nrBytes -= sz;
		}

		FlushBitStream();
	}
	else
	{
		*partSize = AlignInBits(*partSize + 8*nrBytes);
	}
}

void CMGStreamParser::StoreRepositionCodingHeader(int level, BYTE layer, int *partSize)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	BYTE nrBits = meshLevel.GetNrBitsVertsReposCounterField();
	CIntArray &repositionCount = meshLevel.GetNrBytesVertsReposCoding();
	CBoolArray &codingLookup   = meshLevel.GetCodingLookup();

	// write the counters
	ASSERT (nrBits > 0);
	int nrROIs = meshLevel.GetTotalNrROIs();
	for (int k = 0; k < nrROIs; k++)
	{
		if (codingLookup[k])
		{
			ASSERT(repositionCount[k] != 0);
			if (partSize == NULL)
				m_streamMG.WriteBits((DWORD)repositionCount[k], nrBits);
			else
				*partSize += nrBits;
		}
#ifdef _DEBUG
		else
			ASSERT(repositionCount[k] == 0);
#endif
	}

	if (partSize == NULL)
		FlushBitStream();
	else
		*partSize = AlignInBits(*partSize);
}

void CMGStreamParser::StoreCodedVertsReposBits(int level, BYTE layer, int *partSize)
{
	BYTE buffer[1024];
	MStreamLayer & meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	CIntArray & reposCoding  = meshLevel.GetNrBytesVertsReposCoding();
	int nrROIs = meshLevel.GetTotalNrROIs();

	int nrBytes = 0;
	for (int roiIdx = 0; roiIdx < nrROIs; roiIdx++)
		nrBytes += reposCoding[roiIdx];

	ASSERT(nrBytes > 0);
	m_streamVertsRepos[layer].SeekToBegin();

	if (partSize == NULL)
	{
		while (nrBytes > 0)
		{
			int sz = min(1024, nrBytes);
			int count = m_streamVertsRepos[layer].Read(buffer, sz);
			ASSERT(count == sz);

			m_streamMG.Write(buffer, sz);
			nrBytes -= sz;
		}

		FlushBitStream();
	}
	else
	{
		*partSize = AlignInBits(*partSize + 8*nrBytes);
	}
}

void CMGStreamParser::StoreRefinementCodingHeader(int level, BYTE layer, int *partSize)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	BYTE nrBits = meshLevel.GetNrBitsVertsRefineCounterField();
	CIntArray &refinementCount = meshLevel.GetNrBytesVertsRefineCoding();
	CBoolArray &codingLookup   = meshLevel.GetCodingLookup();

	// write the counters
	ASSERT (nrBits > 0);
	int nrROIs = meshLevel.GetTotalNrROIs();
	for (int k = 0; k < nrROIs; k++)
	{
		if (codingLookup[k])
		{
			ASSERT(refinementCount[k] != 0);
			if (partSize == NULL)
				m_streamMG.WriteBits((DWORD)refinementCount[k], nrBits);
			else
				*partSize += nrBits;
		}
#ifdef _DEBUG
		else
			ASSERT(refinementCount[k] == 0);
#endif
	}

	if (partSize == NULL)
		FlushBitStream();
	else
		*partSize = AlignInBits(*partSize);
}

void CMGStreamParser::StoreCodedVertsRefineBits(int level, BYTE layer, int *partSize)
{
	BYTE buffer[1024];
	MStreamLayer & meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	CIntArray & refineCoding = meshLevel.GetNrBytesVertsRefineCoding();
	int nrROIs = meshLevel.GetTotalNrROIs();

	int nrBytes = 0;
	for (int roiIdx = 0; roiIdx < nrROIs; roiIdx++)
		nrBytes += refineCoding[roiIdx];

	ASSERT(nrBytes > 0);
	m_streamVertsRefine[layer].SeekToBegin();

	if (partSize == NULL)
	{
		while (nrBytes > 0)
		{
			int sz = min(1024, nrBytes);
			int count = m_streamVertsRefine[layer].Read(buffer, sz);
			ASSERT(count == sz);

			m_streamMG.Write(buffer, sz);
			nrBytes -= sz;
		}

		FlushBitStream();
	}
	else
	{
		*partSize = AlignInBits(*partSize + 8*nrBytes);
	}
}

// store the coded grid of a level
void CMGStreamParser::StoreCodedTiles(int level, int *partSize)
{
	GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(level);
	int nrTiles = gridLevel.GetTotalNrTiles();

	// rewind the stream first
	for (int dir = X_GRID; dir <= Z_GRID; dir++)
		m_streamGrid[dir].SeekToBegin();

	for (int tileIdx = 0; tileIdx < nrTiles; tileIdx++)
	{
		for (int dir = X_GRID; dir <= Z_GRID; dir++)
		{
			RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
			int nrBytes = gridLevel.GetNrBytesTileCoding(dir, tileIdx);
			if (nrBytes == 0) continue;

			if (partSize == NULL)
			{
				BYTE buffer[1024];
				while (nrBytes > 0)
				{
					int sz = min(1024, nrBytes);
					int count = m_streamGrid[dir].Read(buffer, sz);
					ASSERT(count == sz);

					m_streamMG.Write(buffer, sz);
					nrBytes -= sz;
				}
			}
			else
			{
				*partSize += (nrBytes << 3);
			}
		}
	}

	if (partSize == NULL)
		FlushBitStream();
	else
		*partSize = AlignInBits(*partSize);
}

/*******************************************************************************/
// Mesh encoding stream access

void CMGStreamParser::SetAbsoluteVertex(const CIPointUVW &pos3D, BYTE dir, int level, BYTE layer)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	const CIPointUVW &nrBits = meshLevel.GetNrBitsVertPosField();

	DWORD wValue;
	wValue = pos3D.u;       m_pMemStream->WriteBits(wValue, nrBits.u);
	wValue = pos3D.v;       m_pMemStream->WriteBits(wValue, nrBits.v);
	wValue = pos3D.w;       m_pMemStream->WriteBits(wValue, nrBits.w);

	m_pMemStream->WriteBits(dir, DIR_BITS);
}

void CMGStreamParser::SetVertexConnectivity(BYTE symbol)
{
	if (m_streamDescr.IsSameConnectivity())
	{
		ASSERT (symbol == 0 || symbol == NO_LINK);
		m_pMemStream->WriteBits(symbol, 1);
	}
	else
	{
		ASSERT (symbol >= 0 && symbol <= NO_LINK);
		m_pMemStream->WriteBits(symbol, BITS_PER_CONNECT);
	}
}

void CMGStreamParser::SetVertexRefineBit(BYTE symbol)
{
	m_pMemStream->WriteBit(symbol);
}

void CMGStreamParser::SetVertexReposition(BYTE symbol, bool bSpecialCase)
{
	if (!bSpecialCase)
	{
		m_pMemStream->WriteBit(symbol);
	}
	else
	{
		m_pMemStream->WriteBit(symbol & 0x2);
		m_pMemStream->WriteBit(symbol & 0x1);
	}
}
