/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGStreamParse.cpp: implementation of the CMGStreamParser class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MGStreamParse.h"
#include "MGGridCodec.h"
#include "BitStream.h"
#include "MGErrMsg.h"
#include "ODP.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMGStreamParser::CMGStreamParser(CFile *file, CBitStream::BITSTREAM_OPERATIONS op)
{
	ASSERT(op == CBitStream::OP_WRITE || op == CBitStream::OP_READ);

	m_streamConnectivity = NULL;
	m_streamVertsRepos = NULL;
	m_streamVertsRefine = NULL;
	m_pMemStream = NULL;
	m_nrVertRoot = 0;
	m_decSpecInfoFile = NULL;

	// create the stream object
	CreateStream(file, op);
}

CMGStreamParser::CMGStreamParser()
{
	m_streamConnectivity = NULL;
	m_streamVertsRepos = NULL;
	m_streamVertsRefine = NULL;
	m_pMemStream = NULL;
	m_nrVertRoot = 0;
}

CMGStreamParser::~CMGStreamParser()
{
	delete []m_streamConnectivity;
	delete []m_streamVertsRepos;
	delete []m_streamVertsRefine;

	CloseStream();
}

void CMGStreamParser::AttachToStream(CFile *file, CBitStream::BITSTREAM_OPERATIONS op)
{
	ASSERT(op == CBitStream::OP_WRITE || op == CBitStream::OP_READ);

	// create the stream object
	CreateStream(file, op);
}

void CMGStreamParser::Initialise(MeshGridDecoderSpecificInfo *pinfo)
{
	// init the mesh description
	InitROIDescription(pinfo);
	InitFlags(pinfo);
	InitReservedBits(pinfo);
	InitNrLayers(pinfo);
	InitMeshType(pinfo);
	InitCyclicMode(pinfo);
	InitNrRefineBitsVertex(pinfo);

	// init the grid description
	InitFilterType(pinfo);
	InitGridCodingParam(pinfo);
	InitGridCorners(pinfo);
}

// parsing mesh related information from the decoder specific info

void CMGStreamParser::InitROIDescription(MeshGridDecoderSpecificInfo *pinfo)
{
	CIPointUVW nrLevels, nrSlices, nrROIs;

	InitNrLevels(pinfo, nrLevels);
	InitNrSlices(pinfo, nrSlices);
	InitNrROIs(pinfo, nrROIs);

	// allocate internal arrays
	m_streamDescr.Init(nrLevels, nrSlices, nrROIs);
}

void CMGStreamParser::InitNrLevels(MeshGridDecoderSpecificInfo *pinfo, CIPointUVW &nrLevels)
{
	nrLevels.u = pinfo->nLevels[0];
	nrLevels.v = pinfo->nLevels[1];
	nrLevels.w = pinfo->nLevels[2];

	if (nrLevels.u < MIN_LEVEL || 
		nrLevels.v < MIN_LEVEL ||
		nrLevels.w < MIN_LEVEL)
	{
		throw CMGDataException(CMGErrMsg::WRONG_NR_LEVELS);
	}
}

void CMGStreamParser::InitNrSlices(MeshGridDecoderSpecificInfo *pinfo, CIPointUVW &nrSlices)
{
	// read the number of slices
	nrSlices.u = pinfo->nSlices[0];
	nrSlices.v = pinfo->nSlices[1];
	nrSlices.w = pinfo->nSlices[2];
}

void CMGStreamParser::InitNrROIs(MeshGridDecoderSpecificInfo *pinfo, CIPointUVW &nrROIs)
{
	// read the number of slices
	nrROIs.u = pinfo->nROIs[0];
	nrROIs.v = pinfo->nROIs[1];
	nrROIs.w = pinfo->nROIs[2];
}

// parsing mesh related information from the decoder specific info
void CMGStreamParser::InitNrLayers(MeshGridDecoderSpecificInfo *pinfo)
{
	int nrLayers = max(pinfo->nLayers, 1);

	// allocate internal arrays
	m_streamDescr.InitMeshLayers(nrLayers);
}

void CMGStreamParser::InitFlags(MeshGridDecoderSpecificInfo *pinfo)
{
	m_streamDescr.HasConnectivityInfo(pinfo->hasConnectivityInfo);
	m_streamDescr.HasRefineInfo(pinfo->hasRefineInfo);
	m_streamDescr.HasRepositionInfo(pinfo->hasRepositionInfo);
	m_streamDescr.HasGridInfo(pinfo->hasGridInfo);
	m_streamDescr.HasNormalsInfo(false);
}

void CMGStreamParser::InitMeshType(MeshGridDecoderSpecificInfo *pinfo)
{
	BYTE meshType = pinfo->meshType;
	m_streamDescr.SetMeshType(meshType);

	if (meshType == QUADRI_MESH)
	{
		if (m_streamDescr.HasConnectivityInfo())
		{
			BYTE flag;
			flag = pinfo->sameBorderOrientation;   m_streamDescr.SetSameConnectivity(flag == 1 ? true: false);
			flag = pinfo->uniformSplit;            m_streamDescr.SetAutoResolution(flag == 1 ? true : false);
		}
		else
		{
			m_streamDescr.SetSameConnectivity(true);
			m_streamDescr.SetAutoResolution(true);
		}

		const CIPointUVW & nrSlices = m_streamDescr.GetMaxNrSlices();
		if (m_streamDescr.IsSingleLayered())
		{
			int offsetAmplitude = pinfo->offsetAmplitude;
			m_streamDescr.SetOffsetAmplitude(offsetAmplitude);
		}
	}
}

void CMGStreamParser::InitNrRefineBitsVertex(MeshGridDecoderSpecificInfo *pinfo)
{
	// check whether the refine bits are interleaved in the stream or not
	// init the number of refine bits
	if (m_streamDescr.HasRefineInfo())
	{
		BYTE flag = pinfo->fullRefine;        m_streamDescr.SetInterleaveRefine(flag == TRUE);
		BYTE nrBits = pinfo->nRefineBits;     m_streamDescr.SetNrRefineBitsVertex(nrBits);

		if (nrBits > MAX_REFINE_BITS)
		{
			throw CMGDataException(CMGErrMsg::WRONG_VERTEX_REFINE_BITS);
		}
	}
	else
	{
		m_streamDescr.SetInterleaveRefine(false);
		m_streamDescr.SetNrRefineBitsVertex(0);
	}
}

void CMGStreamParser::InitCyclicMode(MeshGridDecoderSpecificInfo *pinfo)
{
	BYTE flag = pinfo->cyclicMode;   m_streamDescr.SetCyclicMode(flag);
}

void CMGStreamParser::InitReservedBits(MeshGridDecoderSpecificInfo *pinfo)
{
	ASSERT(pinfo->attributes == 0);
}

// init grid related information from the decoder specific info
void CMGStreamParser::InitFilterType(MeshGridDecoderSpecificInfo *pinfo)
{
	BYTE bValue = pinfo->filterType;    m_streamDescr.SetFilterType(bValue);
}

void CMGStreamParser::InitGridCodingParam(MeshGridDecoderSpecificInfo *pinfo)
{
	// read the scale
	RGCoding * codingRG = m_streamDescr.GetGridCodingParam();
	for (int dir = X_GRID; dir <= Z_GRID; dir++)
	{
		DWORD wValue = pinfo->gridScale[dir];   codingRG[dir].scale = wValue;
	}

	if (codingRG[X_GRID].scale < MIN_SCALE ||
		codingRG[Y_GRID].scale < MIN_SCALE ||
		codingRG[Z_GRID].scale < MIN_SCALE)
	{
		throw CMGDataException(CMGErrMsg::WRONG_SCALE_VALUE);
	}

	// prepare the header description of the levels
	InitGridCodingHeader();
}

void CMGStreamParser::InitGridCorners(MeshGridDecoderSpecificInfo *pinfo)
{
	// retrieve the number of quantization bits of the grid values
	BYTE nrQuantBits = pinfo->nQuantBits;
	if (nrQuantBits == 0 || nrQuantBits > MAX_QUANT_BITS)
	{
		throw CMGDataException(CMGErrMsg::WRONG_NR_QUANT_BITS);
	}

	m_streamDescr.SetNrQuantBitsGrid(nrQuantBits);

	// which grid corners are stored in the stream
	int uMax = 1, vMax = 1, wMax = 1;
	const CIPointUVW &nrSlices = m_streamDescr.GetMaxNrSlices();

	// check for single layered
	if (nrSlices.w == 1) wMax = 0;  // only layer 0

	// check for cyclic
	if (m_streamDescr.IsCyclicU())	uMax = 0;		// only one corner in the u direction
	if (m_streamDescr.IsCyclicV())	vMax = 0;		// only one corner in the v direction

	// retrieve the grid corner values
	int idx = 0;
	for (int w = 0; w <= wMax; w++)
	{
		for (int v = 0; v <= vMax; v++)
		{
			for (int u = 0; u <= uMax; u++)
			{
				for (int dir = X_GRID; dir <= Z_GRID; dir++)
				{
					RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
					codingRG.gridCorners[4*w + 2*v + u] = pinfo->gridCorners[idx++];
				}
			}
		}
	}

	// initialise the remaining grid corners
	if (wMax == 0)
	{
		// copy layer 0 to layer 1 such that both are similar
		for (int k = 0; k < 8; k += 8)
		{
			for (int dir = X_GRID; dir <= Z_GRID; dir++)
			{
				RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
				codingRG.gridCorners[k + 4] = codingRG.gridCorners[k + 0];
				codingRG.gridCorners[k + 5] = codingRG.gridCorners[k + 1];
				codingRG.gridCorners[k + 6] = codingRG.gridCorners[k + 2];
				codingRG.gridCorners[k + 7] = codingRG.gridCorners[k + 3];
			}
		}
	}

	if (vMax == 0)
	{
		// copy line 0 to line 1 in both layers
		for (int k = 0; k < 8; k += 4)
		{
			for (int dir = X_GRID; dir <= Z_GRID; dir++)
			{
				RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
				codingRG.gridCorners[k + 2] = codingRG.gridCorners[k + 0];
				codingRG.gridCorners[k + 3] = codingRG.gridCorners[k + 1];
			}
		}
	}

	if (uMax == 0)
	{
		// copy point 0 to point 1 in both lines from the two layers
		for (int k = 0; k < 8; k += 2)
		{
			for (int dir = X_GRID; dir <= Z_GRID; dir++)
			{
				RGCoding &codingRG = m_streamDescr.GetGridCodingParam(dir);
				codingRG.gridCorners[k + 1] = codingRG.gridCorners[k + 0];
			}
		}
	}
}

void CMGStreamParser::EndOfPatch(int nrVertices, bool &bFinish, int level, BYTE layer, int roiIdx)
{
	ASSERT(level >= 0 && level < m_streamDescr.GetMaxNrLevelsMesh());

	++m_nrVertRoot;

	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	const CIPointUVW &nrBits = meshLevel.GetNrBitsVertPosField();
	int nrBitsConnectivity = m_streamDescr.IsSameConnectivity() ? 1 : BITS_PER_CONNECT;
	int nrBytesRead = (m_nrVertRoot * (nrBits.u + nrBits.v + nrBits.w + DIR_BITS) +
		nrVertices * 2 * nrBitsConnectivity + 7) / 8;

	int nrBytesROICoding = meshLevel.GetNrBytesConnectivityCoding(roiIdx);

	ASSERT(nrBytesRead <= nrBytesROICoding);
	bFinish = nrBytesRead == nrBytesROICoding;
}

bool CMGStreamParser::IsRemainingEmpty()
{
	BYTE val = m_streamMG.GetRemainingValue();
	ASSERT(val < 128);
	return val == 0;
}

const CIPointUVW &CMGStreamParser::GetMaxNrSlices() const
{
	return m_streamDescr.GetMaxNrSlices();
}

int CMGStreamParser::GetMaxNrLevelsMesh() const
{
	return m_streamDescr.GetMaxNrLevelsMesh();
}

const CIPointUVW &CMGStreamParser::GetNrLevelsMesh() const
{
	return m_streamDescr.GetNrLevelsMesh();
}

int CMGStreamParser::GetMaxNrLevelsGrid() const
{
	return m_streamDescr.GetMaxNrLevelsGrid();
}

void CMGStreamParser::InitGridCodingHeader()
{
	// compute the number of levels of the grid
	const CIPointUVW &nrSlices = m_streamDescr.GetMaxNrSlices();
	int nrLevelsG = CMGGridCodec<CODEC_TYPE>::GetMaxNrLevelsG(nrSlices);

	// initialise the grid codec levels
	m_streamDescr.InitGridLevels(nrLevelsG);
}

int CMGStreamParser::UpdateConnectivityFilePointers(int level, BYTE layer)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);

	int pos = m_streamMG.GetPosition();
	int nrROIs = meshLevel.GetTotalNrROIs();

	// stream position of the connectivity bits
	for (int k = 0; k < nrROIs; k++)
	{
		meshLevel.SetPosStreamConnectivity(k, pos);
		pos += meshLevel.GetNrBytesConnectivityCoding(k);
	}

	return pos;
}

int CMGStreamParser::UpdateRepositionFilePointers(int level, BYTE layer)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);

	int pos = m_streamMG.GetPosition();
	int nrROIs = meshLevel.GetTotalNrROIs();

	if (m_streamDescr.IsRepositionInfoAvailable(level, layer))
	{
		// stream position of the reposition bits
		for (int k = 0; k < nrROIs; k++)
		{
			meshLevel.SetPosStreamReposition(k, pos);
			pos += meshLevel.GetNrBytesVertsReposCoding(k);
		}
	}

	return pos;
}

int CMGStreamParser::UpdateRefineFilePointers(int level, BYTE layer)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);

	int pos = m_streamMG.GetPosition();
	int nrROIs = meshLevel.GetTotalNrROIs();

	if (m_streamDescr.IsRefinementInfoAvailable(level))
	{
		// stream position of the refinement bits
		for (int k = 0; k < nrROIs; k++)
		{
			meshLevel.SetPosStreamRefine(k, pos);
			pos += meshLevel.GetNrBytesVertsRefineCoding(k);
		}
	}

	return pos;
}

int CMGStreamParser::UpdateTilesFilePointers(int level)
{
	GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(level);

	int filePos = m_streamMG.GetPosition();
	int nrTiles = gridLevel.GetTotalNrTiles();

	for (int tileIdx = 0; tileIdx < nrTiles; tileIdx++)
	{
		for (int dir = X_GRID; dir <= Z_GRID; dir++)
		{
			gridLevel.SetPosStreamTile(dir, tileIdx, filePos);

			filePos += gridLevel.GetNrBytesTileCoding(dir, tileIdx);
		}
	}

	return filePos;
}

/******************************************************************************/
// Mesh Coding

void CMGStreamParser::OnBeginMeshCodingLayer(int level, BYTE layer)
{
	ASSERT(level >= 0 && level < m_streamDescr.GetMaxNrLevelsMesh());

	if (m_streamDescr.IsConnectivityInfoAvailable(level))
		m_streamConnectivity[layer].InitStream(-1);

	if (m_streamDescr.IsRefinementInfoAvailable(level))
		m_streamVertsRefine[layer].InitStream(-1);

	if (m_streamDescr.IsRepositionInfoAvailable(level, layer))
		m_streamVertsRepos[layer].InitStream(-1);
}

void CMGStreamParser::OnBeginMeshCodingROI(int level, BYTE layer, int roiIdx)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);

	meshLevel.SetCodingLookup(roiIdx, false);
	meshLevel.SetNrBytesConnectivityCoding(roiIdx, 0);
	meshLevel.SetNrBytesVertsReposCoding(roiIdx, 0);
	meshLevel.SetNrBytesVertsRefineCoding(roiIdx, 0);
}

void CMGStreamParser::OnBeginMeshCodingConnectivityROI(BYTE layer)
{
	m_streamConnectivity[layer].ResetStream();

	ASSERT(m_pMemStream == NULL);
	m_pMemStream = m_streamConnectivity[layer].GetStream();
}

void CMGStreamParser::OnBeginMeshCodingVertsReposROI(BYTE layer)
{
	m_streamVertsRepos[layer].ResetStream();

	ASSERT(m_pMemStream == NULL);
	m_pMemStream = m_streamVertsRepos[layer].GetStream();
}

void CMGStreamParser::OnBeginMeshCodingVertsRefineROI(BYTE layer)
{
	m_streamVertsRefine[layer].ResetStream();

	ASSERT(m_pMemStream == NULL);
	m_pMemStream = m_streamVertsRefine[layer].GetStream();
}

void CMGStreamParser::OnEndMeshCodingConnectivityROI(int level, BYTE layer, int roiIdx)
{
	// flush the bitstream buffers and byte align the stream data
	// after calling flush, the bit count will be rounded upwards
	// such that the stream size in bits is a multiple of bytes
	ASSERT(m_pMemStream == m_streamConnectivity[layer].GetStream());
	m_pMemStream->Flush();

	int nrBytes = m_pMemStream->GetBitCount() / 8;
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	meshLevel.SetNrBytesConnectivityCoding(roiIdx, nrBytes);

	m_pMemStream = NULL;
}

void CMGStreamParser::OnEndMeshCodingVertsReposROI(int level, BYTE layer, int roiIdx)
{
	// flush the bitstream buffers and byte align the stream data
	// after calling flush, the bit count will be rounded upwards
	// such that the stream size in bits is a multiple of bytes
	ASSERT(m_pMemStream == m_streamVertsRepos[layer].GetStream());
	m_pMemStream->Flush();

	int nrBytes = m_pMemStream->GetBitCount() / 8;
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	meshLevel.SetNrBytesVertsReposCoding(roiIdx, nrBytes);

	m_pMemStream = NULL;
}

void CMGStreamParser::OnEndMeshCodingVertsRefineROI(int level, BYTE layer, int roiIdx)
{
	// flush the bitstream buffers and byte align the stream data
	// after calling flush, the bit count will be rounded upwards
	// such that the stream size in bits is a multiple of bytes
	ASSERT(m_pMemStream == m_streamVertsRefine[layer].GetStream());
	m_pMemStream->Flush();

	int nrBytes = m_pMemStream->GetBitCount() / 8;
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	meshLevel.SetNrBytesVertsRefineCoding(roiIdx, nrBytes);

	m_pMemStream = NULL;
}

void CMGStreamParser::OnEndMeshCodingROI(bool bEmpty, int level, BYTE layer, int roiIdx)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);
	meshLevel.SetCodingLookup(roiIdx, !bEmpty);
}

void CMGStreamParser::StoreMeshHeaderPart(int level, BYTE update_replace, bool allLayers)
{
	// compute the number of bits for the counters
	int k;
	if (allLayers)
	{
		// compute the parameters the same for all layers
		int maxCountConnectivity = 0;	int maxCountReposition = 0;	int maxCountRefinement = 0;
		for (k = 0; k < m_layers.GetSize(); k++)
			ComputeMaxCounter(level, m_layers[k], maxCountConnectivity, maxCountReposition, maxCountRefinement);

		for (k = 0; k < m_layers.GetSize(); k++)
			ComputeCounterSize(level, m_layers[k], maxCountConnectivity, maxCountReposition, maxCountRefinement);
	}
	else
	{
		// compute the parameters for each layer
		for (k = 0; k < m_layers.GetSize(); k++)
		{
			int maxCountConnectivity = 0;	int maxCountReposition = 0;	int maxCountRefinement = 0;
			ComputeMaxCounter(level, m_layers[k], maxCountConnectivity, maxCountReposition, maxCountRefinement);
			ComputeCounterSize(level, m_layers[k], maxCountConnectivity, maxCountReposition, maxCountRefinement);
		}
	}

	// store the mesh header for the level & byte allign the data
	int partSize = 0;
	{
		// compute the size of the part, store tag, size, level and data
		StoreLayerDescription(level, update_replace, allLayers, &partSize);
		if (allLayers)
			StoreMeshHeaderLevelDetails(level, 0, &partSize);
		else
			for (int k = 0; k < m_layers.GetSize(); k++)
				StoreMeshHeaderLevelDetails(level, m_layers[k], &partSize);
	}

	StoreTagAndSize(MGMeshInfoTag, partSize);

	{
		StoreLayerDescription(level, update_replace, allLayers, NULL);
		if (allLayers)
			StoreMeshHeaderLevelDetails(level, 0, NULL);
		else
			for (int k = 0; k < m_layers.GetSize(); k++)
				StoreMeshHeaderLevelDetails(level, m_layers[k], NULL);
	}
}

void CMGStreamParser::StoreMeshConnectivityPart(int level, BYTE update_replace)
{
	int partSize = 0;
	{
		// compute the size of the part, store tag, size, level and data
		StoreLayerDescription(level, update_replace, false, &partSize);

		for (int k = 0; k < m_layers.GetSize(); k++)
		{
			BYTE layer = m_layers[k];
			// first the headers
			StoreROICodingLookup(level, layer, &partSize);
			StoreConnectivityCodingHeader(level, layer, &partSize);
			// follows the data
			StoreCodedConnectivity(level, layer, &partSize);
		}
	}

	StoreTagAndSize(MGMeshConnectivityInfoTag, partSize);

	{
		StoreLayerDescription(level, update_replace, false, NULL);

		for (int k = 0; k < m_layers.GetSize(); k++)
		{
			BYTE layer = m_layers[k];
			// first the headers
			StoreROICodingLookup(level, layer, NULL);
			StoreConnectivityCodingHeader(level, layer, NULL);
			// follows the data
			StoreCodedConnectivity(level, layer, NULL);
			// delete the streams
			m_streamConnectivity[layer].DeleteStream();
		}
	}
}

void CMGStreamParser::StoreMeshRefinementPart(int level, BYTE update_replace)
{
	// store the coded vertices' refinement bits for the level & byte allign the data
	int partSize = 0;

	// compute the size of the part, store tag, size, level and data
	if (m_streamDescr.IsRefinementInfoAvailable(level))
	{
		{
			// store the part identifier and level
			StoreLayerDescription(level, update_replace, false, &partSize);

			for (int k = 0; k < m_layers.GetSize(); k++)
			{
				BYTE layer = m_layers[k];
				// first the header
				StoreRefinementCodingHeader(level, layer, &partSize);
				// follows the data
				StoreCodedVertsRefineBits(level, layer, &partSize); 
			}
		}

		StoreTagAndSize(MGVerticesRefinementInfoTag, partSize);

		{
			StoreLayerDescription(level, update_replace, false, NULL);

			for (int k = 0; k < m_layers.GetSize(); k++)
			{
				BYTE layer = m_layers[k];
				// first the header
				StoreRefinementCodingHeader(level, layer, NULL);
				// follows the data
				StoreCodedVertsRefineBits(level, layer, NULL); 
				// delete the streams
				m_streamVertsRefine[layer].DeleteStream();
			}
		}
	}
}

void CMGStreamParser::StoreMeshRepositionPart(int level, BYTE update_replace)
{
	// store the coded vertices' reposition bits for the level & byte allign the data
	int partSize = 0;

	// compute the size of the part, store tag, size, level and data
	if (m_streamDescr.IsRepositionInfoAvailable(level, 0))	// take layer 0
	{
		{
			// store the part identifier and level
			StoreLayerDescription(level, update_replace, false, &partSize);

			for (int k = 0; k < m_layers.GetSize(); k++)
			{
				BYTE layer = m_layers[k];
				// first the header
				StoreRepositionCodingHeader(level, layer, &partSize);
				// follows the data
				StoreCodedVertsReposBits(level, layer, &partSize);
			}
		}

		StoreTagAndSize(MGVerticesRepositionInfoTag, partSize);

		{
			StoreLayerDescription(level, update_replace, false, NULL);

			for (int k = 0; k < m_layers.GetSize(); k++)
			{
				BYTE layer = m_layers[k];
				// first the header
				StoreRepositionCodingHeader(level, layer, NULL);
				// follows the data
				StoreCodedVertsReposBits(level, layer, NULL);
				// delete the streams
				m_streamVertsRepos[layer].DeleteStream();
			}
		}
	}
}

void CMGStreamParser::OnEndMeshCodingLevel(int level, BYTE update_replace)
{
	// do we need to encode something?
	if (!m_streamDescr.IsConnectivityInfoAvailable(level) &&
		!m_streamDescr.IsRefinementInfoAvailable(level) &&
		!m_streamDescr.IsRepositionInfoAvailable(level, 0))	// take layer 0
		return;

	// store the commun header
	StoreMeshHeaderPart(level, update_replace, true);

	if (m_streamDescr.IsConnectivityInfoAvailable(level))
		StoreMeshConnectivityPart(level, update_replace);

	if (m_streamDescr.IsRefinementInfoAvailable(level))
		StoreMeshRefinementPart(level, update_replace);

	if (m_streamDescr.IsRepositionInfoAvailable(level, 0))
		StoreMeshRepositionPart(level, update_replace);

	// remove the reference to the stream
	m_pMemStream = NULL;  
}

bool CMGStreamParser::OnMeshDecodingROI(int level, BYTE layer, int roiIdx)
{
	// return false if there is no mesh inside the ROI
	if (!SetStreamPosition(Q_VERTEX, level, layer, roiIdx))
		return false;

	// clear the stream buffer
	FlushBitStream();

	// reset the counters that keep track of the amount of vertices
	// that have been recovered
	m_nrVertRoot = 0;

	return true;
}

bool CMGStreamParser::OnMeshDecodingROIRefine(int level, BYTE layer, int roiIdx)
{
	// return false if there is no mesh inside the ROI
	if (!SetStreamPosition(Q_REFINE, level, layer, roiIdx))
		return false;

	// clear the stream buffer
	FlushBitStream();

	return true;
}

bool CMGStreamParser::OnMeshDecodingROIReposition(int level, BYTE layer, int roiIdx)
{
	// return false if there is no mesh inside the ROI
	if (!SetStreamPosition(Q_REPOS, level, layer, roiIdx))
		return false;

	// clear the stream buffer
	FlushBitStream();

	return true;
}

/******************************************************************************/
// Grid Coding

void CMGStreamParser::OnBeginGridCodingDir(int dir, int level)
{
	ASSERT(dir >= X_GRID && dir <= Z_GRID);

	// initialise the stream to encode the grid
	GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(level);
	m_pMemStream = m_streamGrid[dir].InitStream(gridLevel.GetNrBytesTileCoding(dir, 0) * 8);
}

void CMGStreamParser::OnEndGridCodingDir(int dir)
{
	// delete the stream
	m_pMemStream = NULL;
}

void CMGStreamParser::OnBeginGridCodingLevel(int levelG)
{
	ASSERT(levelG >= 0 && levelG < m_streamDescr.GetMaxNrLevelsGrid());
}

void CMGStreamParser::OnEndGridCodingLevel(int levelG, BYTE update_replace)
{
	ASSERT(levelG >= 0 && levelG < m_streamDescr.GetMaxNrLevelsGrid());

	// store the part identifier and level, store the header & byte allign the data
	int partSize = 0;
	StoreLevelDescription(levelG, update_replace, &partSize);
	StoreGridHeaderLevelDetails(levelG, &partSize);

	StoreTagAndSize(MGGridInfoTag, partSize);      

	StoreLevelDescription(levelG, update_replace, NULL);
	StoreGridHeaderLevelDetails(levelG, NULL);     

	// store the part identifier and level, store each of the X, Y, Z coded buffers & byte allign the data
	partSize = 0;
	StoreLevelDescription(levelG, update_replace, &partSize);
	StoreTilesCodingHeader(levelG, &partSize);
	StoreCodedTiles(levelG, &partSize);       

	StoreTagAndSize(MGGridCoefficientsInfoTag, partSize);      

	StoreLevelDescription(levelG, update_replace, NULL);
	StoreTilesCodingHeader(levelG, NULL);
	StoreCodedTiles(levelG, NULL);       

	for (int dir = X_GRID;  dir <= Z_GRID; dir++)
		m_streamGrid[dir].DeleteStream();
}

void CMGStreamParser::OnBeginGridCodingTile(int dir)
{
	ASSERT(dir >= X_GRID && dir <= Z_GRID);

	m_streamGrid[dir].ResetStream();
}

void CMGStreamParser::OnEndGridCodingTile(int dir, int levelG, int tileIdx)
{
	ASSERT(dir >= X_GRID && dir <= Z_GRID);
	ASSERT(levelG >= 0 && levelG < m_streamDescr.GetMaxNrLevelsGrid());

	GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(levelG);

	m_pMemStream->Flush();
	int nrBytes = m_pMemStream->GetBitCount() / 8;

	gridLevel.SetNrBytesTileCoding(dir, tileIdx, nrBytes);
}

bool CMGStreamParser::OnBeginGridDecodingTile(QueryType qType, int level, int tileIdx)
{
	return SetStreamPosition(qType, level, 0, tileIdx);
}

void CMGStreamParser::OnEndGridDecodingTile(int tileIdx)
{
	m_pMemStream = NULL;
}

/******************************************************************************/

bool CMGStreamParser::SetStreamPosition(QueryType qType, int level, BYTE layer, int roiIdx)
{
	switch (qType)
	{
	case Q_VERTEX:  return SetStreamPositionConnectivity(level, layer, roiIdx);
	case Q_REFINE:  return SetStreamPositionRefine(level, layer, roiIdx);
	case Q_REPOS:   return SetStreamPositionReposition(level, layer, roiIdx);
	case Q_GRIDX:   return SetStreamPositionGrid(X_GRID, level, roiIdx);
	case Q_GRIDY:   return SetStreamPositionGrid(Y_GRID, level, roiIdx);
	case Q_GRIDZ:   return SetStreamPositionGrid(Z_GRID, level, roiIdx);
	default: ASSERT(0); return false;
	}
}

bool CMGStreamParser::SetStreamPositionGrid(int dir, int level, int tileidx)
{
	GStreamLevel &gridLevel = m_streamDescr.GetGridLevel(level);

	int nrBits = gridLevel.GetNrBytesTileCoding(dir, tileidx) * 8;
	if (nrBits == 0) return false;

	m_streamMG.SetPositionAndLimit(gridLevel.GetPosStreamTile(dir, tileidx),
		nrBits, CBitStream::OP_READ);
	m_pMemStream = m_streamMG.GetStream();
	return true;
}

bool CMGStreamParser::SetStreamPositionConnectivity(int level, BYTE layer, int roiIdx)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);

	if (meshLevel.GetNrBytesConnectivityCoding(roiIdx) == 0)
		return false;

	m_streamMG.SetPositionAndLimit(meshLevel.GetPosStreamConnectivity(roiIdx),
		-1, CBitStream::OP_READ);
	m_pMemStream = m_streamMG.GetStream();

	return true;
}

bool CMGStreamParser::SetStreamPositionReposition(int level, BYTE layer, int roiIdx)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);

	if (meshLevel.GetNrBytesVertsReposCoding(roiIdx) == 0)
		return false;

	m_streamMG.SetPositionAndLimit(meshLevel.GetPosStreamReposition(roiIdx),
		-1, CBitStream::OP_READ);
	m_pMemStream = m_streamMG.GetStream();

	return true;
}

bool CMGStreamParser::SetStreamPositionRefine(int level, BYTE layer, int roiIdx)
{
	MStreamLayer &meshLevel = m_streamDescr.GetMeshLayer(level, layer);

	if (meshLevel.GetNrBytesVertsRefineCoding(roiIdx) == 0)
		return false;

	m_streamMG.SetPositionAndLimit(meshLevel.GetPosStreamRefine(roiIdx),
		-1, CBitStream::OP_READ);
	m_pMemStream = m_streamMG.GetStream();

	return true;
}

/******************************************************************************/

void CMGStreamParser::CloseStream()
{
	m_streamMG.DeleteStream();
}

void CMGStreamParser::CreateStream(CFile *file, CBitStream::BITSTREAM_OPERATIONS op)
{
	// initialize bitStream
	if (file == NULL)
	{
		ASSERT(op == CBitStream::OP_WRITE);

		// write to temporary mem file and only
		// at the end copy the size and content to the file
		m_streamMG.Initialise(&m_memFile, op);
	}
	else
	{
		m_streamMG.Initialise(file, op);
	}
}

/******************************************************************************/

int CMGStreamParser::GetPower2(int value)
{
	int count = 0;
	while(value >= 1) { ++count;    value >>= 1; }
	return count;
}

int CMGStreamParser::GetNrBits(int value)
{
	int count = 1;
	while(value > 1) { ++count;    value >>= 1; }
	return count;
}

int CMGStreamParser::GetNrBits(const CIPointUVW &p)
{
	return GetNrBits(max(p.u, max(p.v, p.w)));
}

/******************************************************************************/

void CMGStreamParser::FlushBitStream()
{
	m_streamMG.Flush();
}

/******************************************************************************/

void CMGStreamParser::DumpStream(CStringArray &msgs)
{
	m_streamDescr.DumpStream(msgs);
}

