/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGStreamParse.h: interface for the CMGStreamParser class.
//
//////////////////////////////////////////////////////////////////////

class MeshGridDecoderSpecificInfo;

#if !defined(AFX_MGSTREAMPARSE_H__8973651B_CBD9_4473_BC92_F390AD903F76__INCLUDED_)
#define AFX_MGSTREAMPARSE_H__8973651B_CBD9_4473_BC92_F390AD903F76__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "MGErrMsg.h"
#include "MGStreamDescription.h"

class CMGMesh;
class CMGGrid;
class CBitStream;

enum RepositionMode {
	REPOSITION_NO = 0,                                        // the reposition bits are all 0
	REPOSITION_YES = 1,                                       // the reposition bits are all 1
	REPOSITION_ANY = 2                                        // no default reposition bits value
};

enum {                                                      // connectivity links values 
	SAME_DIR  = 0,		                                        // 00
	CCW90_DIR	= 1,		                                        // 01
	CW90_DIR  = 2,		                                        // 10
	NO_LINK	  =	3	                                            // 11
};

enum {
	GENERIC_MESH = 0,                                         // The encoded mesh is generic, meaning that the triangulation
	// may contain triangles, quadrilaterals, pentagons, hexagons and heptagons
	TRI_MESH = 1,                                             // The encoded mesh contains only triangles
	QUADRI_MESH = 2,                                          // The encoded mesh contains only quadrilaterals
	HEXA_MESH = 3,                                            // The encoded mesh contains only hexagons
};

enum {
	TAG_BITS = 8,                                             // number of bits to encode the tag of encoded part
	ForbiddenTag = 0,                                         // not allowed
	MGInfoTag = 1,                                            // mesh grid header
	MGMeshInfoTag = 2,                                        // mesh coding header
	MGGridInfoTag = 3,                                        // grid coding header
	MGMeshConnectivityROIInfoTag = 4,                         // mesh connectivity ROI part 
	MGMeshConnectivityInfoTag = 5,                            // mesh connectivity part
	MGVerticesRepositionROIInfoTag = 6,                       // vertices reposition bits ROI part
	MGVerticesRepositionInfoTag = 7,                          // vertices reposition bits part
	MGVerticesRefinementROIInfoTag = 8,                       // vertices refinement bits ROI part
	MGVerticesRefinementInfoTag = 9,                          // vertices refinement bits part
	MGGridCoefficientsROIInfoTag = 10,                        // grid coefficients ROI bitplanes
	MGGridCoefficientsInfoTag = 11,                           // grid coefficients bitplanes
	MGGridCornersInfoTag = 12,								  // grid corners part
	MGBeginFrameTag = 255									  // start a new frame that may consist of several parts
};

enum GridCoord {
	X_GRID = 0,                                               // the index of the buffer containing the X coordinates
	Y_GRID = 1,                                               // the index of the buffer containing the Y coordinates
	Z_GRID = 2,                                               // the index of the buffer containing the Z coordinates
};

enum UPDATE_REPLACE {
	UPDATE_DATA = 0,										  // the data is a differential update
	REPLACE_DATA = 1										  // the data is standalone 
};

class CMGStreamParser  
{
public:
	CMGStreamParser(CFile *file, CBitStream::BITSTREAM_OPERATIONS op);
	CMGStreamParser();
	~CMGStreamParser();

	enum {
		LEVEL_BITS = 6,
		MIN_LEVEL = 1,                  
		MIN_SLICES = 1,                                           // min 1 slice
		REFINE_BITS = 6,
		MAX_REFINE_BITS = (1 << REFINE_BITS) - 1,
		FILTER_BITS = 4,                                          // different choices of the filter
		ROI_BITS = 6,                                             // max value 31
		MIN_ROI_SIZE = 5,                                         // min value of the ROI size
		TILE_BITS = 5,
		MAX_TILE_BITS = (1 << TILE_BITS) - 1,
		MAX_TILE_SIZE = UINT_MAX,
		FIELD_BITS = 5,
		MAX_FIELD_BITS = (1 << FIELD_BITS) - 1,
		MIN_SCALE = 1,
		MAX_SCALE = UINT_MAX,
		QUANT_BITS = 5,
		MAX_QUANT_BITS = (1 << QUANT_BITS) - 1,                   // max value 31
		MIN_DC_FIELD = INT_MIN,
		MAX_DC_FIELD = INT_MAX,
		GRID_BITS =  5,
		MAX_GRID_BITS = (1 << GRID_BITS) - 1,
		VERTS_BITS = 5,
		MAX_VERTS_BITS = (1 << VERTS_BITS) - 1,
		DIR_BITS = 3,                                             // 6 choices for the absolute direction
		BITS_PER_CONNECT = 2,                                     // nr of bits per vertex for a connectivity
		BITS_PER_REPOSITION  = 1,                                 // nr of bits per vertex inter-level refine (reposition)
		RESERVED_BITS = 8,                                        // reserved bits
		CYCLIC_MODE_BITS = 3,                                     // to store the the cyclic mode
		MESH_TYPE_BITS = 2,                                       // to store the type of mesh
		REPOS_MODE_BITS = 2,                                      // to store the default reposition mode
		UPDATE_REPLACE_BITS = 4									  // to store the update replace flag
	};

	enum QueryType {                                                  
		Q_VERTEX = 1,                                             // query vertex connectivity bits
		Q_REFINE = 2,                                             // query vertex refine bits
		Q_REPOS  = 3,                                             // query vertex reposition bits
		Q_GRIDX  = 4,                                             // query grid stream
		Q_GRIDY  = 5,                                             // query grid stream
		Q_GRIDZ  = 6,                                             // query grid stream
	};

public:
	static CMGStreamParser *InitialiseStreamParser(MeshGridDecoderSpecificInfo *pinfo,
		CString &error);            // create and initialise the pasers object
	void Initialise(MeshGridDecoderSpecificInfo *pinfo);                       // initialise the parser
	static CMGStreamParser *ParseMeshGridHeader(CFile *file, CString &error, int &headerSize);  // construct the parser object and do the parsing
	void AttachToStream(CFile *file, CBitStream::BITSTREAM_OPERATIONS op);     // attach the paser to the stream

	void  StoreMeshGridHeader(bool keepHeader);
	void  WriteOutput(CFile *file);
	void  StoreValue(MGStream &stream, int number, int *partSize);
	void  StoreTagAndSize(BYTE tag, int size);
	void  StoreBeginFrame();
	void  StoreLevelDescription(int level, BYTE update_replace, int *partSize);
	void  StoreLayerDescription(int level, BYTE update_replace, bool allLayers, int *partSize);
	void  StoreIndices(BYTE nBitsLayer, int *partSize);
	void  StoreDefaultIndex(BYTE nBitsLayer, int *partSize);
	void  SetActiveLayers(const CIntArray &layers);
	const CIntArray &GetLayers()	{ return m_layers; };
	int   ParseMeshGridHeader();                                                // parse stream header description
	void  ParseMeshGridStream(int streamSize);                                  // parse the description for each level
	int   ParseStreamSize(int &counterLength);
	int   ParseValue(int &number);
	int   ParseTagAndSize(BYTE &tag, int &size);
	void  ParseLevelDescription(int &level, BYTE &update_replace);
	void  ParseLayerDescription(int &level, BYTE &update_replace, CIntArray &layers);
	void  ParseIndices(int nBitsIndex, CIntArray &layers);
	void  MeshDescriptor(int level, BYTE layer);                                // retrieve mesh header level coding details
	void  DefaultMeshDescriptor(int level, BYTE layer);                         // initialise mesh header level for auto resolution
	void  GridDescriptor(int levelG);                                           // retrieve the grid coding level header
	int   MeshConnectivityROIHeader(int, BYTE) { return 0; };			        // retrieve the connectivity coding header
	int   MeshConnectivityHeader(int level, BYTE layer);                        // retrieve the connectivity coding header
	int   VertexRepositionROIHeader(int, BYTE) { return 0; };			        // retrieve the reposition coding header
	int   VertexRepositionHeader(int level, BYTE layer);                        // retrieve the reposition coding header
	int   VertexRefineROIHeader(int, BYTE) { return 0; };			            // retrieve the refine coding header
	int   VertexRefineHeader(int level, BYTE layer);                            // retrieve the refine coding header
	int   GridCoefficientsROIHeader(int) { return 0; };		                    // retrieve the tiles coding header
	int   GridCoefficientsHeader(int level);                                    // retrieve the tiles coding header
	void  ParseROICodingLookup(int level, BYTE layer);							// retrieve the ROI coding lookup
	void  ParseConnectivityCodingHeader(int level, BYTE layer);                 // retrieve the nr of bytes/ROI for the connectivity coding
	void  ParseRepositionCodingHeader(int level, BYTE layer);                   // retrieve the nr of bytes/ROI for the reposition bits coding
	void  ParseRefinementCodingHeader(int level, BYTE layer);                   // retrieve the nr of bytes/ROI for the refinment bits coding
	void  ParseTilesCodingHeader(int level);                                    // retrieve the counters for each Tile per level

	void  ParseMeshGridHeader(BYTE *data, int nrBytes);                         // parse the stream header description from the buffer
	void  ParseMeshLevelHeader(BYTE *data, int nrBytes, int level);             // parse the mesh level header description from the buffer
	void  ParseGridLevelHeader(BYTE *data, int nrBytes, int level);             // parse the grid level header description from the buffer

	// retrive the stream header information
	const CIPointUVW &GetMaxNrSlices() const;                                   // get the max number of slices
	int   GetMaxNrLevelsMesh() const;                                           // get the max number of resolution levels of the mesh
	const CIPointUVW &GetNrLevelsMesh() const;                                  // get the number of resolution levels of the mesh
	int   GetMaxNrLevelsGrid() const;                                           // get the number of resolution levels of the grid

	void  InitStreams(BYTE nrLayers);											// allocate the temporary bitstreams, one for each layer

	CBitStream *GetStream() { return m_pMemStream; };
	MGStream &GetStreamMG() { return m_streamMG; };
	void  EndOfPatch(int nrVertices, bool &bFinish, int level, BYTE layer, int roiIdx);     // mark for the end of a connectivity patch in ROI
	void  EndOfPart() { FlushBitStream(); };                                    // mark for the end of a stream part
	void  EndOfPartPos(int pos) { m_streamMG.SetPosition(pos); };               // set the stream position

	// streming mesh                                                            
	void  OnBeginMeshCodingLayer(int level, BYTE layer);                        // the coding of a new mesh level starts
	void  OnBeginMeshCodingROI(int level, BYTE layer, int roiIdx);              // the coding of a new mesh ROI starts
	void  OnBeginMeshCodingConnectivityROI(BYTE layer);							// the coding of the connectivity description starts
	void  OnBeginMeshCodingVertsReposROI(BYTE layer);							// the coding of the vertices' reposition bits starts
	void  OnBeginMeshCodingVertsRefineROI(BYTE layer);							// the coding of the vertices' refinement bits starts
	bool  OnMeshDecodingROI(int level, BYTE layer, int roiIdx);                 // position the stream for ROI decoding
	bool  OnMeshDecodingROIRefine(int level, BYTE layer, int roiIdx);           // position the stream for ROI Refine decoding
	bool  OnMeshDecodingROIReposition(int level, BYTE layer, int roiIdx);       // position the stream for ROI Reposition decoding

	void  OnEndMeshCodingLevel(int level, BYTE update_replace);					// the coding of a new mesh level ends
	void  OnEndMeshCodingROI(bool bEmpty, int level, BYTE layer, int roiIdx);   // the coding of a new mesh ROI ends
	void  OnEndMeshCodingConnectivityROI(int level, BYTE layer, int roiIdx);    // the coding of the connectivity description ends
	void  OnEndMeshCodingVertsReposROI(int level, BYTE layer, int roiIdx);      // the coding of the vertices' reposition bits ends
	void  OnEndMeshCodingVertsRefineROI(int level, BYTE layer, int roiIdx);     // the coding of the vertices' refinement bits ends

	// streaming grid                                                           
	void  ProcessGridCorners(BYTE &update_replace);								// process the grid corners bits
	void  StoreGridCorners(bool differential, BYTE nrBits);						// store the grid corners
	void  OnBeginGridCodingDir(int dir, int leve);                              // the coding of a new grid direction starts
	void  OnBeginGridCodingLevel(int level);                                    // the coding of a new grid resolution level starts
	void  OnBeginGridCodingTile(int dir);                                       // the coding of a new grid tile start
	bool  OnBeginGridDecodingTile(QueryType qType, int level, int tileIdx);     // the decoding of a new grid tile start

	void  OnEndGridCodingDir(int dir);                                          // the coding of a new grid direction ends
	void  OnEndGridCodingLevel(int level, BYTE update_replace);                 // the coding of a new grid resolution level ends
	void  OnEndGridCodingTile(int dir, int level, int tileIdx);                 // the coding of a new grid tile ends
	void  OnEndGridDecodingTile(int tileIdx);                                   // the decoding of a new grid tile ends

	// position stream Mesh and Grid
	bool  SetStreamPosition(QueryType qType, int level, BYTE layer, int roi);   // set the stream position
	bool  SetStreamPositionConnectivity(int level, BYTE layer, int roi);        // set the stream position for mesh connectivity in ROI
	bool  SetStreamPositionRefine(int level, BYTE layer, int roi);              // set the stream position for refining the vertices in ROI
	bool  SetStreamPositionReposition(int level, BYTE layer, int roi);          // set the stream position for repositioning the vertices in ROI
	bool  SetStreamPositionGrid(int dir, int level, int tileidx);               // set the stream position for a grid tile
	void  SetStreamPositon(int pos) { m_streamMG.SetPosition(pos); };           // set the position in the stream
	void  MoveStreamPosition(int size) { m_streamMG.MovePosition(size); };      // move the stream position
	int   GetStreamPositon() const { return m_streamMG.GetPosition(); };        // retrieve the position in the stream

	// read / write mesh coding symbols                                         
	void  GetAbsoluteVertex(CIPointUVW &pos3D, BYTE &dir, int level, BYTE layer);	// get the absolute position/direction of the starting vertex
	void  GetVertexConnectivity(BYTE &symbol);                                  // get a connectivity symbol for a vertex
	void  GetVertexRefineBit(BYTE &symbol);                                     // get a refine bit for a vertex
	void  GetVertexReposition(BYTE &symbol, bool bSpecialCase);                 // get a refine bit for a vertex

	void  SetAbsoluteVertex(const CIPointUVW &pos3D, BYTE dir, int level, BYTE layer);      
	void  SetVertexConnectivity(BYTE symbol);                                   
	void  SetVertexRefineBit(BYTE symbol);                                      
	void  SetVertexReposition(BYTE symbol, bool bSpecialCase);                                      

	void  DumpStream(CStringArray &msgs);                                       // display the fields from the stream

	static int   GetNrBits(int value);
	static int   GetPower2(int value);
	static int   GetNrBits(const CIPointUVW &value);

protected:
	void  CreateStream(CFile *file, CBitStream::BITSTREAM_OPERATIONS op);
	void  CloseStream();
	void  InitGridCodingHeader();

	int   AlignInBits(int value) const { return AlignInBytes(value) * 8; };
	int   AlignInBytes(int value) const { return (value + 7) / 8; };

	// store / retrieve main MeshGrid information in / from the stream
	void  StoreNrLevels(bool inBinary, bool inASCII, int *partSize);            // store the number of levels mesh
	void  StoreNrSlices(bool inBinary, bool inASCII, int *partSize);            // store the number of slices
	void  StoreNrROIs(bool inBinary, bool inASCII, int *partSize);              // store the number of slices
	void  StoreMeshType(bool inBinary, bool inASCII, int *partSize);            // store the type of mesh, homogeneous or not
	void  StoreCyclicMode(bool inBinary, bool inASCII, int *partSize);          // store the cyclic mode flag
	void  StoreNrRefineBitsVertex(bool inBinary, bool inASCII, int *partSize);  // store the number of refine bits for the vertices' positions
	void  StoreReservedBits(bool inBinary, bool inASCII, int *partSize);        // store reserved bits
	void  StoreNrLayers(bool inBinary, bool inASCII, int *partSize);			// store layer bits
	void  StoreFilterType(bool inBinary, bool inASCII, int *partSize);          // store the filter type
	void  StoreFlags(bool inBinary, bool inASCII, int *partSize);               // store connectivity, refine and grid availability flags
	void  StoreGridCodingParam(bool inBinary, bool inASCII, int *partSize);     // store the grid coding parameters
	void  StoreGridCorners(bool inBinary, bool inASCII, int *partSize);         // store the grid corners bits

	// parse the mesh and grid description from the stream
	void  ParseROIDescription();
	void  ParseNrLevels(CIPointUVW &nrLevels);                                  // retrieve the number of levels mesh
	void  ParseNrSlices(CIPointUVW &nrSlices);                                  // retrieve the number of slices
	void  ParseNrROIs(CIPointUVW &nrROIs);                                      // retrieve the number of ROIs
	void  ParseMeshType();                                                      // retrieve the type of mesh, homogeneous or not
	void  ParseCyclicMode();                                                    // retrieve cyclic mode of the mesh
	void  ParseNrRefineBitsVertex();                                            // retrieve the number of refine bits for the vertices' positions
	void  ParseReservedBits();                                                  // retrieve reserved bits
	void  ParseNrLayers();														// retreive the nr of bits for storing the layer identifier
	void  ParseFilterType();                                                    // retrieve the filter type
	void  ParseFlags();                                                         // retrieve connectivity, refine and grid availability flags
	void  ParseGridCodingParam();                                               // retrieve the grid coding parameters
	void  ParseGridCorners(bool update);	                                    // retrieve the grid corners bits

	// init the mesh and grid description from the decoder specific info
	void  InitROIDescription(MeshGridDecoderSpecificInfo *pinfo);               
	void  InitNrLayers(MeshGridDecoderSpecificInfo *pinfo);                     // retrieve the number of layers mesh
	void  InitNrSlices(MeshGridDecoderSpecificInfo *pinfo, CIPointUVW &nrSlices);  // retrieve the number of slices
	void  InitNrROIs(MeshGridDecoderSpecificInfo *pinfo, CIPointUVW &nrROIs);      // retrieve the number of ROIs
	void  InitNrLevels(MeshGridDecoderSpecificInfo *pinfo, CIPointUVW &nrLevels);  // retrieve the number of levels mesh
	void  InitFlags(MeshGridDecoderSpecificInfo *pinfo);                        // retrieve connectivity, refine and grid availability flags
	void  InitMeshType(MeshGridDecoderSpecificInfo *pinfo);                     // retrieve the type of mesh, homogeneous or not
	void  InitCyclicMode(MeshGridDecoderSpecificInfo *pinfo);                   // retrieve cyclic mode of the mesh
	void  InitNrRefineBitsVertex(MeshGridDecoderSpecificInfo *pinfo);           // retrieve the number of refine bits for the vertices' positions
	void  InitReservedBits(MeshGridDecoderSpecificInfo *pinfo);                 // retrieve reserved bits
	void  InitFilterType(MeshGridDecoderSpecificInfo *pinfo);                   // retrieve the filter type
	void  InitGridCodingParam(MeshGridDecoderSpecificInfo *pinfo);              // retrieve the grid coding parameters
	void  InitGridCorners(MeshGridDecoderSpecificInfo *pinfo);                  // retrieve the grid corners bits

	// store / retrieve resolution level mesh information in / from the stream
	void  StoreMeshHeaderPart(int level, BYTE update_replace, bool allLayers);	// compute size & store mesh header level coding details
	void  StoreMeshConnectivityPart(int level, BYTE update_replace);			// compute size & store connectivity coding details
	void  StoreMeshRefinementPart(int level, BYTE update_replace);              // compute size & store refinement coding details
	void  StoreMeshRepositionPart(int level, BYTE update_replace);              // compute size & store reposition coding details

	void  StoreMeshHeaderLevelDetails(int level, BYTE layer, int *partSize);    // store mesh header level coding details
	void  StoreMeshCodingHeader(int level, BYTE layer, int *partSize);          // store a lookup of coded ROIs
	void  StoreROICodingLookup(int level, BYTE layer, int *partSize);           // store the ROI coding lookup
	void  StoreConnectivityCodingHeader(int level, BYTE layer, int *partSize);  // store the nr of bytes/ROI for the connectivity coding
	void  StoreRepositionCodingHeader(int level, BYTE layer, int *partSize);    // store the nr of bytes/ROI for the reposition bits coding
	void  StoreRefinementCodingHeader(int level, BYTE layer, int *partSize);    // store the nr of bytes/ROI for the refinrment bits coding
	void  StoreCodedConnectivity(int level, BYTE layer, int *partSize);         // store the coded connectivity bits
	void  StoreCodedVertsReposBits(int level, BYTE layer, int *partSize);       // store the coded vertices' reposition bits
	void  StoreCodedVertsRefineBits(int level, BYTE layer, int *partSize);      // store the coded vertices' refinement bits
	void  StoreGridCornersBits(bool differential, BYTE nrQuantBits, int *partSize);	// store the grid corners bits
	void  ComputeMaxCounter(int level, BYTE layer, int &maxCountConnectivity, int &maxCountReposition, int &maxCountRefinement);
	void  ComputeCounterSize(int level, BYTE layer, int maxCountConnectivity, int maxCountReposition, int maxCountRefinement);

	void  ParseMeshCodingHeader(int level, BYTE layer);                         // retrieve the coded mesh bits

	// store / retrieve resolution level grid information in / from the stream
	void  StoreGridHeaderLevelDetails(int level, int *partSize);                // store grid header level coding details
	void  StoreThreshold(int level, int *partSize);                             // store the threshold for the coded grid
	void  StoreGridCodingHeader(int level, int *partSize);                      // store the nr of bits for each Tile per level
	void  StoreTilesCodingHeader(int level, int *partSize);                     // store the lookup and counters for each Tile per level
	void  StoreCodedTiles(int level, int *partSize);                            // store the coded grid bits

	void  ParseThreshold(int level);                                            // retrieve the threshold for the coded grid
	void  ParseGridCodingHeader(int level);                                     // retrieve the nr of bits for each Tile per level

	// updating references to the stream
	int   UpdateConnectivityFilePointers(int level, BYTE layer);
	int   UpdateRepositionFilePointers(int level, BYTE layer);
	int   UpdateRefineFilePointers(int level, BYTE layer);
	int   UpdateTilesFilePointers(int level);

	// process the ROIs
	void  FlushBitStream();
	bool  IsRemainingEmpty();

public:
	MGStreamDescr   m_streamDescr;        // description of the MeshGrid stream

protected:
	CMemFile        m_memFile;            // memory file for view-dependent decoding
	CBitStream    * m_pMemStream;         // temporary stream pointing to the currently selected stream

	MGStream        m_streamMG;           // the stream containing the coded MeshGrid
	MGStream      * m_streamConnectivity; // temporary stream for storing the coded mesh
	MGStream      * m_streamVertsRepos;   // temporary stream for storing the coded mesh
	MGStream      * m_streamVertsRefine;  // temporary stream for storing the coded mesh
	MGStream        m_streamGrid[3];      // temporary stream for storing the coded grid

	int             m_nrVertRoot;         // keep the number of root vertices in a ROI

	FILE          * m_decSpecInfoFile;    // dump the decoder specific info

	CIntArray		m_layers;			  // the list of layers
	CIntArray		m_layerGroups;		  // the list of layer groups
	CBoolArray		m_isIndexed;		  // the lookup for layers
};

inline CMGStreamParser *CMGStreamParser::InitialiseStreamParser(MeshGridDecoderSpecificInfo *pinfo, CString &error)
{
	CMGStreamParser *pParser = NULL;

	// create the parser object
	try
	{
		pParser = new CMGStreamParser();
	}
	catch (CMemoryException *e)
	{
		GetErrorMessage(e, error);  return NULL;
	}

	try
	{
		// do the parsing of the stream headers
		pParser->Initialise(pinfo);
	}
	catch (CMGDataException &e)
	{
		error = e.GetErrorMessage();  return NULL;
	}

	return pParser;
}

inline CMGStreamParser *CMGStreamParser::ParseMeshGridHeader(CFile *file, CString &error, int &headerSize)
{
	CMGStreamParser *pParser = NULL;

	// create the parser object
	TRY
	{
		pParser = new CMGStreamParser(file, CBitStream::OP_READ);
	}
	CATCH_ALL(e)
	{
		GetErrorMessage(e, error);  return NULL;
	}
	END_CATCH_ALL

		try
	{
		// do the parsing of the stream headers
		headerSize = pParser->ParseMeshGridHeader();
	}
	catch (CMGDataException &e)
	{
		error = e.GetErrorMessage();
		delete pParser;   return NULL;
	}

	return pParser;
}

#endif // !defined(AFX_MGSTREAMPARSE_H__8973651B_CBD9_4473_BC92_F390AD903F76__INCLUDED_)
