/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

// MGStreamDescription.h: interface for the CMGStreamDescription class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(_ROI_SIZE_POS)
#define _ROI_SIZE_POS

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "BasicTypes.h"
#include "Globals.h"

class ROIDescription
{
public:
	CIPointUVW m_nrSlices;
	CIPointUVW m_szROI;
	CIPointUVW m_maxSzROI;
	CIPointUVW m_nrROIs;
	CIPointUVW m_reminder;

public:
	ROIDescription() { m_nrSlices = m_nrROIs = -1;	m_szROI = 0; }
	const CIPointUVW &GetNrSlices() const { return m_nrSlices; }
	const CIPointUVW &GetROISize() const { return m_szROI; }
	const CIPointUVW &GetMaxROISize() const { return m_maxSzROI; }
	const CIPointUVW &GetNrROIs() const { return m_nrROIs; }
	int	  GetTotalNrROIs() const { return m_nrROIs.u * m_nrROIs.v * m_nrROIs.w; }

	void Init(const CIPointUVW &nrSlices, const CIPointUVW &nrROIs)
	{
		m_nrROIs = nrROIs;
		m_nrSlices = nrSlices;
		m_szROI = nrSlices / nrROIs;

		m_reminder.u = nrSlices.u - m_szROI.u*nrROIs.u;
		m_reminder.v = nrSlices.v - m_szROI.v*nrROIs.v;
		m_reminder.w = nrSlices.w - m_szROI.w*nrROIs.w;

		m_maxSzROI.u = (m_reminder.u > 0) ? m_szROI.u + 1 : m_szROI.u;
		m_maxSzROI.v = (m_reminder.v > 0) ? m_szROI.v + 1 : m_szROI.v;
		m_maxSzROI.w = (m_reminder.w > 0) ? m_szROI.w + 1 : m_szROI.w;
	}

	void ComputeNrROIs(CIPointUVW &nrROIs, const CIPointUVW &levelDiff)
	{
		// estimate nrROIs, at least one ROI
		nrROIs.u = DevideBy2EvenMin1(m_nrROIs.u, levelDiff.u);
		nrROIs.v = DevideBy2EvenMin1(m_nrROIs.v, levelDiff.v);
		nrROIs.w = DevideBy2EvenMin1(m_nrROIs.w, levelDiff.w);
	}

	void ComputePositionAndSize(int &pos, int &sz, int szROI, int nrROIs, int r, int idx) const
	{
		int halfNrROIs = nrROIs >> 1;
		int prevIncr = ( idx   *r + halfNrROIs) / nrROIs;
		int currIncr = ((idx+1)*r + halfNrROIs) / nrROIs;

		pos = (idx*szROI) + prevIncr;
		sz  = szROI + currIncr - prevIncr;
	}

	void ComputePositionAndSize(CIPointUVW& pos, CIPointUVW& sz, const CIPointUVW& idx) const
	{
		ComputePositionAndSize(pos.u, sz.u, m_szROI.u, m_nrROIs.u, m_reminder.u, idx.u);
		ComputePositionAndSize(pos.v, sz.v, m_szROI.v, m_nrROIs.v, m_reminder.v, idx.v);
		ComputePositionAndSize(pos.w, sz.w, m_szROI.w, m_nrROIs.w, m_reminder.w, idx.w);
	}

	void ComputePosition(int &pos, int szROI, int nrROIs, int r, int idx) const
	{
		int prevIncr = (idx*r + (nrROIs >> 1)) / nrROIs;
		pos = (idx*szROI) + prevIncr;
	}

	void ComputePosition(CIPointUVW& pos, const CIPointUVW& idx) const
	{
		ComputePosition(pos.u, m_szROI.u, m_nrROIs.u, m_reminder.u, idx.u);
		ComputePosition(pos.v, m_szROI.v, m_nrROIs.v, m_reminder.v, idx.v);
		ComputePosition(pos.w, m_szROI.w, m_nrROIs.w, m_reminder.w, idx.w);
	}

	void ComputePositionAndSize(Cube &roi, const CIPointUVW &idxUVW, const CIPointUVW &ratio, const CIPointUVW &levelDiff) const
	{
		// find the pos and size of the corresponding left add right ROIs at the last level				
		CIPointUVW idxTopLeft  = idxUVW * ratio;
		CIPointUVW idxTopRight = idxTopLeft + ratio - 1;

		CIPointUVW posTopLeft, posTopRight, szTopRight;
		ComputePosition(posTopLeft, idxTopLeft);
		ComputePositionAndSize(posTopRight, szTopRight, idxTopRight);

		// map the position and size of the highest resolution ROIs to current level
		MapPositionAndSize(posTopLeft, posTopRight, szTopRight, roi, levelDiff);
	}

	void ComputePositionAndSize(Cube &roi, int index, const CIPointUVW &nrROIs, const CIPointUVW &levelDiff) const
	{
		CIPointUVW idxUVW;
		Index2Position(index, nrROIs, idxUVW);

		CIPointUVW ratio = m_nrROIs / nrROIs;
		ComputePositionAndSize(roi, idxUVW, ratio, levelDiff);
	}

	void MapPositionAndSize(int posT1, int posTp, int szTp, int &pos1, int &sz, int levelDiff) const
	{
		// compute the position and size of the ROIs
		//
		// [posT1-----szT-----posT2]...(ratio)...[posTp-----szT-----posTq][posTr-----szT-----posTs]...(ratio)...[posTu-----szT-----posTv]
		//    [pos1-----------------------sz-----------------------pos2]     [pos3-----------------------sz-----------------------pos4]
		//
		//		pos1  = (posT1 + 2^levelDiff - 1) / 2^levelDiff
		//		posTq = posTp + szT - 1
		//		pos2  = posTq / 2^levelDiff
		//		sz    = pos2 - pos1 + 1
		//

		pos1	  = DevideBy2Odd(posT1, levelDiff);
		int pos2  = DevideBy2Even(posTp + szTp - 1, levelDiff);
		sz		  = pos2 - pos1 + 1;				
	}

	void MapPositionAndSize(const CIPointUVW& posT1, const CIPointUVW& posTp, const CIPointUVW& szTp, Cube &roi, const CIPointUVW& levelDiff) const
	{
		MapPositionAndSize(posT1.u, posTp.u, szTp.u, roi.u, roi.width,  levelDiff.u);
		MapPositionAndSize(posT1.v, posTp.v, szTp.v, roi.v, roi.height, levelDiff.v);
		MapPositionAndSize(posT1.w, posTp.w, szTp.w, roi.w, roi.depth,  levelDiff.w);
	}
};

#endif