/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2001
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

#ifndef __cmatrix_h
#define __cmatrix_h

//---------------------------------------------------------------------------

template <class TYPE> class CMatrix
{
public:
	CMatrix();
	CMatrix(int dim1,int dim2,int dim3) { Init(dim1, dim2, dim3, false); };
	CMatrix(const CMatrix<TYPE>&);
	~CMatrix() { Delete(); };

protected:
	int m_d1, m_d2, m_d3;
	TYPE *m_elements;
	TYPE ***m_area;

public:
	void Init(int dim1,int dim2,int dim3, bool clear = true);
	void Delete();

	int GetSize() const { return m_d1*m_d2*m_d3; };;
	TYPE ***GetMatrix3D() { return m_area; };

	int GetDim1() const { return m_d1; }
	__declspec(property(get = GetDim1)) int dim1;
	int GetDim2() const { return m_d2; }
	__declspec(property(get = GetDim2)) int dim2;
	int GetDim3() const { return m_d3; }
	__declspec(property(get = GetDim3)) int dim3;

	TYPE *GetElements() { return m_elements; }
	__declspec(property(get = GetElements)) TYPE *elements;

	TYPE & operator()(int e) { return m_elements[e]; };
	TYPE & operator()(int z, int y, int x) { return m_area[z][y][x]; };
	TYPE **operator[](int z) { return m_area[z]; };

	CMatrix<TYPE>& operator =(const CMatrix<TYPE>& a);
	CMatrix<TYPE>& operator =(TYPE value);
	void MemSet(int value);

/*
	CMatrix<TYPE> operator +(CMatrix<TYPE>& a);
	CMatrix<TYPE> operator +(TYPE value);
	CMatrix<TYPE> operator -(CMatrix<TYPE>& a);
	CMatrix<TYPE> operator -(TYPE value);
	CMatrix<TYPE>& operator +=(CMatrix<TYPE>& a);
	CMatrix<TYPE>& operator +=(TYPE value);
	CMatrix<TYPE>& operator -=(CMatrix<TYPE>& a);
	CMatrix<TYPE>& operator -=(TYPE value);
	CMatrix<TYPE> operator *(CMatrix<TYPE>& a);
	CMatrix<TYPE> operator *(TYPE value);
	CMatrix<TYPE> operator *=(CMatrix<TYPE>& a);
	CMatrix<TYPE>& operator *=(TYPE value);
	CMatrix<TYPE> operator /(CMatrix<TYPE>& a);
	CMatrix<TYPE> operator /=(CMatrix<TYPE>& a);
	CMatrix<TYPE> operator /(TYPE value);
	CMatrix<TYPE>& operator /=(TYPE value);
*/
protected:
	void BuildRow();
	void BuildRowColumn();
};

// global functions
void PrintBuffer3D(CMatrix<int> &img1);
void PrintBuffer3D(CMatrix<double> &img1);

//---------------------------------------------------------------------------

template <class TYPE>
CMatrix<TYPE>::CMatrix()
{
	m_d1 = m_d2 = m_d3 = 0;
	m_elements = NULL;
	m_area = NULL;
}

template <class TYPE>
CMatrix<TYPE>::CMatrix(const CMatrix<TYPE>& a)
{
	m_d1 = m_d2 = m_d3 = 0;
	m_elements = NULL;
	m_area = NULL;

	*this = a;
}

template <class TYPE>
void CMatrix<TYPE>::Init(int dim1, int dim2, int dim3, bool clear)
{
	ASSERT(dim1 > 0 && dim2 > 0 && dim3 > 0);

	if (clear) Delete();

	m_d1 = dim1; m_d2 = dim2; m_d3 = dim3;

	m_area = NULL;
	m_elements = new TYPE[m_d1*m_d2*m_d3];
	BuildRowColumn();
}

template <class TYPE>
void CMatrix<TYPE>::Delete()
{
	delete [] m_elements;	m_elements = NULL;

	if (m_area != NULL)
	{
		for (int j = 0; j < m_d3; j++)
			delete []m_area[j];
		delete [] m_area;		m_area = NULL;
	}
}

template <class TYPE>
void CMatrix<TYPE>::BuildRowColumn()
{
	if (m_area != NULL)
	{
		for (int j = 0; j < m_d3; j++)
			delete m_area[j];
		delete [] m_area;		m_area = NULL;
	}

	m_area=new TYPE**[m_d3];
	for (int j = 0; j < m_d3; j++)
	{
		m_area[j] = new TYPE*[m_d2];
		for (int i = 0; i < m_d2; i++)
			m_area[j][i] = &m_elements[j*m_d2*m_d1 + i*m_d1];
	}
}

template <class TYPE>
CMatrix<TYPE>& CMatrix<TYPE>::operator=(const CMatrix<TYPE>& a)
{
	if (m_d1 != a.m_d1 || m_d2 != a.m_d2 || m_d3 != a.m_d3)
		Init(a.m_d1, a.m_d2, a.m_d3, true);

	TYPE *dest = m_elements;
	TYPE *src = a.m_elements;
	int size = GetSize();
	for (int i = 0; i < size; i++)	*dest++ = *src++;

	return *this;
}

template <class TYPE>
CMatrix<TYPE>& CMatrix<TYPE>::operator=(TYPE value)
{
	int size = GetSize();
	TYPE *dest = m_elements;
	for(int i = 0; i < size; i++)		*dest++ = value;

	return *this;
}

template <class TYPE>
void CMatrix<TYPE>::MemSet(int value)
{
	memset(m_elements, value, GetSize()*sizeof(TYPE));
}

#endif