/***********************************************************************
This software module was originally developed by Shinjun Lee
(SAMSUNG AIT) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. SAMSUNG AIT retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright ?2002.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Shinjun Lee at sjlee@sait.samsung.co.kr.
----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include "arith.h"
#include "bitio.h"

/* Define coder_desc[] string, which describes the coder */

#define MULT_STR  "Arithmetic coding (multiply with "
#define FIXED_STR "fixed bits)"

char *coder_desc =	MULT_STR FIXED_STR;

#define		Half		((code_value) 1 << (B_bits-1))
#define		Quarter		((code_value) 1 << (B_bits-2))

/* Separate input and output */
/* Input decoding state */
static code_value	in_R;				/* code range */
static code_value	in_D;				/* = V-L (V offset)*/
static div_value	in_r;				/* normalized range */

/* Output encoding state */
static code_value	out_L;				/* lower bound */
static code_value	out_R;				/* code range */
static unsigned long	out_bits_outstanding;		/* follow bit count */

/*
 * BIT_PLUS_FOLLOW(b)
 * responsible for outputting the bit passed to it and an opposite number of
 * bit equal to the value stored in bits_outstanding
 *
 */
void BIT_PLUS_FOLLOW(int b, cOutStream &Out)		      
{ 	  			                        
  Out.putbit(b);           		    
  while (out_bits_outstanding > 0)	
  { 					                      
		Out.putbit(1-b);      		      
		out_bits_outstanding--;    	    
  } 	                		          
} 

/*
 * ENCODE_RENORMALISE
 * output code bits until the range has been expanded
 * to above QUARTER
 * (a redundant zero will otherwise be emitted every time the encoder is
 * started)
 */
void ENCODE_RENORMALISE(cOutStream &Out)				
{														
	while (out_R <= Quarter)				
	{																
		if (out_L >= Half)						
		{															
			BIT_PLUS_FOLLOW(1, Out);					
			out_L -= Half;							
		}															
		else if (out_L+out_R <= Half)	
		{															
			BIT_PLUS_FOLLOW(0, Out);					
		}															
		else 													
		{															
			out_bits_outstanding++;			
			out_L -= Quarter;						
		}															
		out_L <<= 1;									
		out_R <<= 1;									
	}																
}


/*
 * DECODE_RENORMALISE
 * input code bits until range has been expanded to
 * more than QUARTER. Mimics encoder.
 */
void DECODE_RENORMALISE(cInStream& In)
{																	
  while (in_R <= Quarter)						  
  {																	  
    in_R <<= 1;
		in_D <<= 1;
		if(In.getbit()) 
			in_D++;											  
  }																	  
} 

/*
 *
 * encode a symbol given its low, high and total frequencies
 *
 */
void arithmetic_encode(freq_value low, freq_value high, freq_value total, cOutStream &Out)
{ 
  /* The following pseudocode is a concise (but slow due to arithmetic
   * calculations) description of what is calculated in this function.
   * Note that the division is done before the multiplication.  This is
   * one of the key points of this version of arithmetic coding.  This means
   * that much larger frequency values can be accomodated, but that the integer
   * ratio R/total (code range / frequency range) 
   * becomes a worse approximation as the total frequency count
   * is increased.  Therefore less than the whole code range will be
   * allocated to the frequency range.  The whole code range is used by
   * assigning the symbol at the end of the frequency range (where high==total)
   * this excess code range above and beyond its normal code range.  This
   * of course distorts the probabilities slightly, 
   * see the paper connected with this program for details and compression
   * results.
   *
   * Restricting the total frequency range means that the division and
   * multiplications can be done to low precision with shifts and adds.
   *
   * The following code describes this function:
   * (The actual code is only written less legibly to improve speed)
   *
   *    L += low*(R/total);			* Adjust low bound	*
   *
   *    if (high < total)			* 			*
   *    	R = (high-low) * (R/total);	* Restrict range	*
   *    else					* If symbol at end of	*
   *		R = R   -  low * (R/total);	*   range.  Restrict &	*
   *						*   allocate excess 	*
   * 						*   codelength to it.	*
   *    ENCODE_RENORMALISE;			* Expand code range	*
   *						*   and output bits	*
   *
   *    if (bits_outstanding > MAX_BITS_OUTSTANDING)
   *		abort();			* EXTREMELY improbable	*
   *						* (see comments below) *
   */

	code_value temp; 

	{
		div_value out_r;			
		out_r = out_R/total;		/* Calc range:freq ratio */
		temp = out_r*low;			/* Calc low increment */
		out_L += temp;			/* Increase L */
		if (high < total)
			out_R = out_r*(high-low);	/* Restrict R */
		else
			out_R -= temp;			/* If at end of freq range */
		/* Give symbol excess code range */
	}

  ENCODE_RENORMALISE(Out);

  if (out_bits_outstanding > MAX_BITS_OUTSTANDING)
  {
			/* 
			 * For MAX_BITS_OUTSTANDING to be exceeded is extremely improbable, but
			 * it is possible.  For this to occur the COMPRESSED file would need to
			 * contain a sequence MAX_BITS_OUTSTANDING bits long (eg: 2^31 bits, or
			 * 256 megabytes) of all 1 bits or all 0 bits.  This possibility is
			 * extremely remote (especially with an adaptive model), and can only
			 * occur if the compressed file is greater than MAX_BITS_OUTSTANDING
			 * long.  Assuming the compressed file is effectively random,
			 * the probability for any 256 megabyte section causing an overflow
			 * would be 1 in 2^(2^31).  This is a number approximately 600 million
			 * digits long (decimal).
			 * 
			 */
	
		fprintf(stderr,"Bits_outstanding limit reached - File too large\n");
		exit(1);
  }
}

/*
 * arithmetic decode_target()
 * decode the target value using the current total frequency
 * and the coder's state variables
 *
 * The following code describes this function:
 * The actual code is only written less legibly to improve speed,
 * including storing the ratio in_r = R/total for use by arithmetic_decode()
 *
 *  target = D / (R/total);	* D = V-L.  (Old terminology) 		*
 *				* D is the location within the range R	*
 *				* that the code value is located	*
 *				* Dividing by (R/total) translates it	*
 *				* to its correspoding frequency value	*
 *
 *				
 *  if (target < total) 	* The approximate calculations mean the *
 *	  return target;		  * encoder may have coded outside the	*
 *  else			            * valid frequency range (in the excess	*
 *	  return total-1;		  * code range).  This indicates that the	*
 *												* symbol at the end of the frequency	*
 *												* range was coded.  Hence return end of *
 *												* range (total-1)			*
 *
 */
freq_value arithmetic_decode_target(freq_value total)
{
  freq_value target;
    
  in_r = in_R/total;
  target = (in_D)/in_r;

  return (target >= total ? total-1 : target);
}



/*
 * arithmetic_decode()
 *
 * decode the next input symbol
 *
 * The following code describes this function:
 * The actual code is only written less legibly to improve speed,
 * See the comments for arithmetic_encode() which is essentially the
 * same process.
 *
 * D -= low * (R/total);		* Adjust current code value 	   *
 *
 * if (high < total)
 *	R = (high-low) * (R/total);	* Adjust range 			   *
 * else
 *	R -= low * (R/total);		* End of range is a special case   *
 *
 * DECODE_RENORMALISE;			* Expand code range and input bits *
 *
 */

void arithmetic_decode(freq_value low, freq_value high, freq_value total, cInStream &In)
{     
  code_value temp;

  /* assume r has been set by decode_target */
  temp = in_r*low;
  in_D -= temp;
  if (high < total)
		in_R = in_r*(high-low);
  else
		in_R -= temp;

  DECODE_RENORMALISE(In);
}

/*
 *
 * start the encoder
 *
 */
void start_arithmetic_encode(void)
{
  out_L = 0;				/* Set initial coding range to	*/
  out_R = Half;			/* [0,Half)			*/
  out_bits_outstanding = 0;
}

/*
 * finish_encode() 
 * finish encoding by outputting follow bits and all the bits in L
 * (B_bits long) to make the last symbol unambiguous.  This also means the
 * decoder can read them harmlessly as it reads beyond end of what would
 * have been valid input.
 *
 */
void finish_arithmetic_encode(cOutStream &Out)
{
  int nbits, i;
  code_value bits;

  nbits = B_bits;
  bits  = out_L;
  for (i = 1; i <= nbits; i++)        /* output the nbits integer bits */
    BIT_PLUS_FOLLOW(((bits >> (nbits-i)) & 1), Out);

	// Out.flushbit();
}

/* 
 * start_decode() start the decoder Fills the decode value in_D from the
 * bitstream.  
 */
void start_arithmetic_decode(cInStream &In)
{
  int i;

  in_D = 0;			    /* Initial offset in range is 0 */
  in_R = Half;			/* Range = Half */

  for (i = 0; i<B_bits; i++){			/* Fill D */
		in_D <<= 1;
		if(In.getbit()) 
			in_D++;											  
	}

  if (in_D >= Half)
	{
	  fprintf(stderr,"Corrupt input file (start_arithmetic_decode())\n");
	  exit(1);
	}
}

void finish_arithmetic_decode(cInStream &In)
{
	/* No action */

	// In.flushbit();
}
