/***********************************************************************
This software module was originally developed by Sebastian Amengual
(Universidad Politecnica de Madrid, UPM) in the course of development of 
the MPEG-4 Systems (ISO/IEC 14496-1) standard. This software module is
an implementation of a part of one or more MPEG-4 Systems 
(ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems
(ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 Systems 
(ISO/IEC 14496-1) free license to this software module or modifications
thereof for use in hardware or software products claiming conformance to
the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use this 
software module in hardware or software products are advised that its 
use may infringe existing patents. The original developer of this 
software module and his/her company, the subsequent editors and their
companies, and ISO/IEC have no liability for use of this software module
or modifications thereof in an implementation. Copyright is not 
released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
UPM retains full right to use the code for his/her own purpose, assign 
or donate the code to a third party and to inhibit third parties from 
using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
products. This copyright notice must be included in all copies or 
derivative works. Copyright  2005
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Francisco Moran at fmb@gti.ssr.upm.es.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Player3d
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  WaveSurfDecoder
FILE:       WaveSurf.cpp
AUTHOR:     Sebastian Amengual (UPM)
OVERVIEW:   The WSS tool uses an external library: the sdslib. All the 
			interaction with it is performed here.
-------------------------------------------------------------------------*/

#include <fstream>
#include "sdslib/WSSNode.h"
#include "WaveSurf.h"
#include "OpenGL/AFX/WaveSurfProxy.h"


//-----------------------------------------------------------------------------
// Name: WaveSurf()
// Desc: The constructor creates the SSNode and the WSSNode, that are needed to 
//       decode the details and add them to the base mesh.
//-----------------------------------------------------------------------------

WaveSurf :: WaveSurf()
{
	da_geom = new SSNode();			//This SSNode will represent the base mesh
	da_w_geom = new WSSNode();
}


//-----------------------------------------------------------------------------
// Name: ~WaveSurf()
// Desc: Destructor.
//-----------------------------------------------------------------------------

WaveSurf :: ~WaveSurf()
{
	delete da_geom;
	delete da_w_geom;
}


//-----------------------------------------------------------------------------
// Name: Setup()
// Desc: All the decoder specific info is given to the WaveSurf object here.
//-----------------------------------------------------------------------------

bool WaveSurf :: Setup(unsigned nbx, unsigned nby, unsigned nbz, unsigned nol, unsigned lnb,
					   float xmx, float ymx, float zmx,
					   bool hsc, int wty, bool lif, bool ipt, bool ilc)
{
	nbBPX=nbx;
	nbBPY=nby;
	nbBPZ=nbz;
	numberOfLevels=nol;
	lengthNbBits=lnb;
	xmax=xmx;
	ymax=ymx;
	zmax=zmx;
	hasScaleCoeff=hsc;
	wtype=wty;
	lift=lif;
	isPartial=ipt;
	isInLocalCoordinates=ilc;
	
	return true;
}


//-----------------------------------------------------------------------------
// Name: SetMesh()
// Desc: The information about the geometry of the base mesh is given. Then
//       all the configuration info is given to the SS and WSS  Nodes and the 
//       bitstream is read.
//-----------------------------------------------------------------------------

bool WaveSurf :: SetMesh(float* xarray, float* yarray, float* zarray, int* cindex, int clength,int ilength) 
{

	//The coordinates and the indexes have to be into a vector for the sdslib to read them.

	vector<int> coord_i(ilength - 1);
	vector<coord3D> coord(clength);

	for (int i = 0 ; i < ilength -1; i++)
		coord_i[i] = cindex[i];
    
    for (i = 0 ; i < clength ; i++)
	{
		coord[i].x = xarray[i];
		coord[i].y = yarray[i];
		coord[i].z = zarray[i];
	}

	//Initialization of the WSSNode.

	da_geom->subdivisionLevel = 0;
    da_geom->setCoordId(coord_i);
    da_geom->pcoordNode = &coord;
    da_geom->onCoordChange();
    da_geom->subdivisionType = 2;
    da_w_geom->pbaseMesh = da_geom;

	//Configuration of the WSSNode.

	da_w_geom->_nbitSc = hasScaleCoeff;
    da_w_geom->_nbBp[X] = nbBPX;
    da_w_geom->_nbBp[Y] = nbBPY;
    da_w_geom->_nbBp[Z] = nbBPZ;
    switch (wtype)
	{
	case 0:
		da_w_geom->_ssType = MID_POINT;
		break;
	case 1:
		da_w_geom->_ssType = LOOP;
		break;
	case 2:
		da_w_geom->_ssType = BUTTERFLY;
	}
	da_w_geom->_lift =lift;
    da_w_geom->_isPartial = isPartial;
    da_w_geom->_localCoord = isInLocalCoordinates;
	da_w_geom->_nbLvl = numberOfLevels;
		
	da_w_geom->_max[X] = xmax;
	da_w_geom->_max[Y] = ymax;
	da_w_geom->_max[Z] = zmax;
	
	//Decode

	da_w_geom->initMeshDecoder(strm);
    da_w_geom->decodeAllBitplanes();

/*
	int nbitread=0;
	do 
		nbitread+=da_w_geom->bpMeshDecoder();
	while(nbitread<strmlength*8);
*/

	return true;
	}


//-----------------------------------------------------------------------------
// Name: SetStream()
// Desc: A pointer to the beginning of the bitstream and its length are given.
//-----------------------------------------------------------------------------

bool WaveSurf :: SetStream(char* stream, int length)
{
	strm=stream;
	strmlength=length;
	return true;
}


//-----------------------------------------------------------------------------
// Name: GetCoord()
// Desc: Returns a pointer to the beginning of the array that contains the
//       coordinates of the vertices of the final mesh.
//-----------------------------------------------------------------------------

const float* WaveSurf :: GetCoord() const
{
	return &da_w_geom->pgetCoord()->begin()->x;
}


//-----------------------------------------------------------------------------
// Name: GetCoordLength()
// Desc: Returns the length of the array that contains the
//       coordinates of the vertices of the final mesh.
//-----------------------------------------------------------------------------

int WaveSurf :: GetCoordLength() const
{
	return da_w_geom->pgetCoord()->size();
}


//-----------------------------------------------------------------------------
// Name: GetCoordIndex()
// Desc: Returns a pointer to the beginning of the array that contains the
//       indices of the triangles of the final mesh.
//-----------------------------------------------------------------------------

const int* WaveSurf :: GetCoordIndex() const
{
	return &(*da_w_geom->pgetCoordId())[0];
}


//-----------------------------------------------------------------------------
// Name: GetCoordIndexLength()
// Desc: Returns the length of the array that contains the
//       coordinates of the vertices of the final mesh.
//-----------------------------------------------------------------------------

int WaveSurf :: GetCoordIndexLength() const
{
	return da_w_geom->pgetCoordId()->size();
}