/***********************************************************************
This software module was originally developed by Yannick Maret (from the
Ecole Polytechnique Federale de Lausanne, EPFL, during a stay at the
Universidad Politecnica de Madrid, UPM) in the course of development of 
the MPEG-4 Systems (ISO/IEC 14496-1) standard. This software module is
an implementation of a part of one or more MPEG-4 Systems 
(ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems
(ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 Systems 
(ISO/IEC 14496-1) free license to this software module or modifications
thereof for use in hardware or software products claiming conformance to
the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use this 
software module in hardware or software products are advised that its 
use may infringe existing patents. The original developer of this 
software module and his/her company, the subsequent editors and their
companies, and ISO/IEC have no liability for use of this software module
or modifications thereof in an implementation. Copyright is not 
released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
EPFL and UPM retain full right to use the code for his/her own purpose,
assign or donate the code to a third party and to inhibit third parties 
from using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
products. This copyright notice must be included in all copies or 
derivative works. Copyright  2003-2005
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Yannick Maret at Yannick.Maret@a3.epfl.ch or Francisco
Moran at fmb@gti.ssr.upm.es.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Player3d
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  WaveSurfDecoder
FILE:       _SSsTopo.cpp
AUTHOR:     Yannick Maret (EPFL, UPM)
OVERVIEW:   Implementation of the topological subdivision surfaces.
-------------------------------------------------------------------------*/

// *** INCLUDES ***
#include "_SSsTopo.h"
// *** CONSTRUCTION/DESTRUCTION ***
_SSsTopo::_SSsTopo():
_currMeshType(OTHER)
#if _TIME_SSsTopoLIB_
    ,
    _split1Time(0), _split2Time(0),
    _ring1Time(0), _ring2Time(0),
    _LUTTime(0)
#endif
{
}
// *** SET THE BASE TOPOLOGY ***
// (and eventually the meshing type)
void _SSsTopo::psetBaseCoordId(const vector<int>* rpCoordId, const enum etMeshType rMeshType)
{
   _pbaseCoordId = _pcurrCoordId = rpCoordId;
   _currNbVert = (*max_element(rpCoordId->begin(), rpCoordId->end())) + 1;
   _currMeshType = rMeshType;
}
// *** TOPOLOGICAL SUBDIVISION, MAIN FUNCTION ***
void _SSsTopo::createTopoLUT(const unsigned rNbSub, const bool rCompleteLastRing)
{
   // *** SUBDIVISE TOPOLOGY
   _buildRings(true);
   _updateLUT();
   for (unsigned i = 0 ; i < rNbSub ; i++)
   {
      _splitStep(i == 0);
      if (!(i == rNbSub - 1 && !rCompleteLastRing))
         _buildRings(false);
      _updateLUT();
   }
#if _TIME_SSsTopoLIB_
   cout << "_SSsTopo Timings (in [s]):\n";
   cout << "---------------------\n";
   cout << "Total Ring Building   *" << (_ring1Time + _ring2Time)/(float)CLK_TCK;
   cout << '*' << endl;
   cout << "  Edge Listing         " << _ring1Time/(float)CLK_TCK << endl;
   cout << "  Ring Order           " << _ring2Time/(float)CLK_TCK << endl;
   cout << "Total Topology subdiv *" << (_split1Time + _split2Time)/(float)CLK_TCK;
   cout << '*' << endl;
   cout << "  Edge Splitting       " << _split1Time/(float)CLK_TCK << endl;
   cout << "  CoordId update       " << _split2Time/(float)CLK_TCK << endl;
   cout << "Total Caching         *" << _LUTTime/(float)CLK_TCK << '*' << endl;
   _split1Time = 0; _split1Time = 0;
   _ring1Time = 0; _ring2Time = 0;
   _LUTTime = 0;
#endif
}
// *** TOPOLOGY SPLITTING, ONE STEP ***
void _SSsTopo::_splitStep(const bool rFirstSlplit)
{
#if _TIME_SSsTopoLIB_
   _startTime = clock();
#endif
   // *** STACK VARIABLES
   const unsigned old_nb_vert = _nbVertLUT.back();
   ring::const_iterator i_old_ring;
   vector<ring>::const_iterator i_LUT_ring = _ringsLUT.back().begin();
   unsigned i, j, ring_size;
   int v0, v1;
   // *** BUILD THE NEW RINGS (FOR THE EVEN VERTICES ONLY)
   // i.e. split edges
   _currRings.reserve(old_nb_vert + _nbEdgeLUT.back());
   _currRings.resize(old_nb_vert);
   for (i = 0; i < old_nb_vert ; i++)
      _currRings[i].resize(_ringsLUT.back()[i].size());
   for (i = 0; i < old_nb_vert ; i++, i_LUT_ring++)
   {
      i_old_ring = i_LUT_ring->begin();
      ring_size = i_LUT_ring->size();
      for (j = 0 ; j < ring_size ; j++, i_old_ring++)
         if (*i_old_ring >= i)
         {
            // this edge is not yet subdivided
            v1 = *i_old_ring;
            v0 = _ringsLUT.back()[v1].getPosInRing(i);
            _currRings[i][j] = _currNbVert;
            _currRings[v1][v0] = _currNbVert;
            // next vertex to be added
            _currNbVert++;
         }
   }
#if _TIME_SSsTopoLIB_
   _endTime = clock();
   _split1Time += _endTime - _startTime;
   _startTime = clock();
#endif
   // *** STACK VARIABLES
   const unsigned old_nb_poly = _nbPolyLUT.back();
   int v2, v3, v4, v5;
   vector<int>::iterator i_cid;
   vector<int>::const_iterator i_old_cid, i_old_cid_end;
   // *** UPDATE THE COORDINATE INDEX
   _currCoordId.clear();
   if (_currMeshType == TRI)
   {
      if (rFirstSlplit)
      {
         i_old_cid = _pbaseCoordId->begin();
         i_old_cid_end = _pbaseCoordId->end();
         _currCoordId.resize(_pbaseCoordId->size()*4);
      }
      else
      {
         i_old_cid = _coordIdLUT.back().begin();
         i_old_cid_end = _coordIdLUT.back().end();
         _currCoordId.resize(_coordIdLUT.back().size()*4);
      }
      i_cid = _currCoordId.begin();
      for (i_old_cid ; i_old_cid < i_old_cid_end ; i_old_cid++)
      {
         // # GET THE EVEN VERTICES
         v0 = *(i_old_cid++);
         v1 = *(i_old_cid++);
         v2 = *(i_old_cid++);
         // # GET THE ODD VERTICES
         // get the middle edge of v0 & v1
         v3 = _currRings[v0][_ringsLUT.back()[v0].getPosInRing(v1)];
         // get the middle edge of v2 & v1
         v4 = _currRings[v1][_ringsLUT.back()[v1].getPosInRing(v2)];
         // get the middle edge of v0 & v2
         v5 = _currRings[v2][_ringsLUT.back()[v2].getPosInRing(v0)]; 
         // # CREATE THE FOUR NEW TRIANGLES
         (*i_cid++) = v0; (*i_cid++) = v3; (*i_cid++) = v5; (*i_cid++) = -1; // T0
         (*i_cid++) = v1; (*i_cid++) = v4; (*i_cid++) = v3; (*i_cid++) = -1; // T1
         (*i_cid++) = v2; (*i_cid++) = v5; (*i_cid++) = v4; (*i_cid++) = -1; // T2
         (*i_cid++) = v3; (*i_cid++) = v4; (*i_cid++) = v5; (*i_cid++) = -1; // T3
      }
   }
   else if (_currMeshType == QUAD)
   {
      // to be implemented
   }
   else
   {
      // to be implemented
   }
   /// *** CURR. UPDATE ***
   _pcurrCoordId = &_currCoordId;
   _currNbPoly = _nbPolyLUT.back()*4;
   _currMeshType = TRI;
#if _TIME_SSsTopoLIB_
   _endTime = clock();
   _split2Time += _endTime - _startTime;
#endif
}
// *** LUT UPDATE ***
void _SSsTopo::_updateLUT()
{
#if _TIME_SSsTopoLIB_
   _startTime = clock();
#endif
   // *** STACK VARIABLES
   // use dummy variables not to whole stuff, just the reference
   vector<int> dummy_cid;
   vector<ring> dummy_rings;
   // # THE RELEVANT INFORMATION AND CLEAR THE _CURR STUFF
   _nbPolyLUT.push_back(_currNbPoly);
   _nbEdgeLUT.push_back(_currNbEdge);
   _nbVertLUT.push_back(_currNbVert);
   // coord id (there will be nothing in the first one)
   _coordIdLUT.push_back(dummy_cid);
   _coordIdLUT.back().swap(_currCoordId);
   // rings
   _ringsLUT.push_back(dummy_rings);
   _ringsLUT.back().swap(_currRings);
#if _TIME_SSsTopoLIB_
   _endTime = clock();
   _LUTTime += _endTime - _startTime;
#endif
}
// *** 1-RING BUILDING ***
// build the ordered 1-rings, work for any type of meshing
void _SSsTopo::_buildRings(const bool rFirstBuild)
{
#if _TIME_SSsTopoLIB_
   _startTime = clock();
#endif
   // *** STACK VARIABLES
   vector<int>::const_iterator i_cid, i_p, i_end = (*_pcurrCoordId).end();
   // *** LIST ALL PAIRS OF CONSECUTIVE VERTICES ON THE 1-RING OF EACH VERTEX
   //   (oposite edge for triangular mesh
   vector<edgeList> edge_lists;
   edge_lists.resize(_currNbVert);
   _currNbPoly = 0;
   for (i_cid = (*_pcurrCoordId).begin() ; i_cid < i_end ; i_cid++)
   {
      i_p = i_cid++;
      while (*(i_cid + 1) != -1)
      {
         edge_lists[*i_cid].push_back(*(i_cid + 1), *(i_cid - 1));
         i_cid++;
      }
      edge_lists[*i_p].push_back(*(i_p + 1), *i_cid);
      edge_lists[*i_cid].push_back(*i_p, *(i_cid - 1));
      _currNbPoly++;
      i_cid++;
   }
#if _TIME_SSsTopoLIB_
   _endTime = clock();
   _ring1Time += _endTime - _startTime;
   _startTime = clock();
#endif
   // *** STACK VARIABLES
   unsigned nb_edge, old_nb_vert, i, l, last_id, first_id;
   vector<char> done;  // trick: bool vector element are stored in a bit
                       // -----  and access is done trough a proxy
                       //        --> very slow to access
   vector<tEdge>::const_iterator i_edge_list;
   vector<char>::iterator i_done;
   bool edge_added;
   // *** ORDER THE 1-RINGS
   if (rFirstBuild)
   {
      old_nb_vert = 0;
      _currNbEdge = 0;
   }
   else
   {
      // the first part of the ring (corresponding to old vertices)
      // was already built during _splitStep() phase...
      old_nb_vert = _nbVertLUT.back();
      _currNbEdge *= 2;
   }
   _currRings.resize(_currNbVert);
   _vertTypeLUT.resize(_currNbVert);
   for (i = old_nb_vert ; i < _currNbVert ; i++) 
   {
      nb_edge = edge_lists[i].size();
      if(nb_edge == 0)		//if the vertex isn't part of any triangle
      {
         throw OrphanVertex();
         continue;
      }
      done.clear();
      done.resize(nb_edge, false);
      // # exact memory reservation
      _currRings[i].clear();
      _currRings[i].reserve(nb_edge + 1);
      // # put 1st two elements in the star
      _currRings[i].push_back(edge_lists[i].front().e0);
      _currRings[i].push_back(edge_lists[i].front().e1);
      if (edge_lists[i].front().e0 > (int)i)
         _currNbEdge++;
      if (edge_lists[i].front().e1 > (int)i)
         _currNbEdge++;
      done[0] = true;
      l = 1;

      // # add the other elements
      while (l < nb_edge)
      {
         edge_added = false;
         i_edge_list = edge_lists[i].begin();
         i_done = done.begin();
         for (i_done = done.begin() ; i_done < done.end() ; i_edge_list++, i_done++)
         {
            if (*i_done)
               continue;
            last_id = _currRings[i].back();
            first_id = _currRings[i].front();
			//
            if (i_edge_list->e0 == first_id) 
            {
               if (i_edge_list->e1 > i)
                  _currNbEdge++;
               // push v1 front
               _currRings[i].insert(_currRings[i].begin(), i_edge_list->e1);
               edge_added = true;
               *i_done = true;
               l++;
            }
            else if (i_edge_list->e1 == first_id)
            {
               if (i_edge_list->e0 > i)
                  _currNbEdge++;
               // push v0 front
               _currRings[i].insert(_currRings[i].begin(), i_edge_list->e0);
               edge_added = true;
               *i_done = true;
               l++;
            }
            else if (i_edge_list->e0 == last_id)
            {
               if (i_edge_list->e1 > i)
                  _currNbEdge++;
               // push v1 back
               _currRings[i].push_back(i_edge_list->e1);
               edge_added = true;
               *i_done = true;
               l++;
            }
            else if (i_edge_list->e1 == last_id)
            {
               if (i_edge_list->e0 > i)
                  _currNbEdge++;
               // push v0 back
               _currRings[i].push_back(i_edge_list->e0);
               edge_added = true;
               *i_done = true;
               l++;      
            }
         }
         if (!edge_added)
         {
            throw NonManifoldVertex();
            _vertTypeLUT[i] = NON_MANIFOLD;
            _currRings[i].clear();
         }
      }
      if (_currRings[i].front() == _currRings[i].back())
      {
         if (_currRings[i].back() > (int)i)
            _currNbEdge--;
         _currRings[i].pop_back();
         nb_edge--;
         _vertTypeLUT[i] = INTERIOR;
      }
      else
         _vertTypeLUT[i] = BOUNDARY;

   }
#if _SSsTopo_DBG_VERBOSE_BASE
   _printRings();
#endif
#if _TIME_SSsTopoLIB_
   _endTime = clock();
   _ring2Time += _endTime - _startTime;
#endif
}
// *** DEBUGING FUNCTIONS ***
#if _SSsTopo_DBG_VERBOSE_BASE
void _SSsTopo::_printRings()
{
   unsigned i, j;
   for (i = 0 ; i < _currNbVert ; i++)
   {
      cout << "Vertex " << i << ", valence = " << _currRings[i].size();
      switch (_vertTypeLUT[i])
      {
         case INTERIOR: cout << ", interior"; break;
         case BOUNDARY: cout << ", boundary"; break;
         case NON_MANIFOLD: cout << ", non-manifold"; break;
      }
      cout << "\n1-ring = ";
      for (j = 0 ; j < _currRings[i].size() ; j++)
         cout << _currRings[i][j] << ' ';
      cout << '\n';
   }
}
void _SSsTopo::_printCoordId()
{
   unsigned i;
   cout << "Coordinate Index : \n";
   for (i = 0; i < (*_pcurrCoordId).size() ; i++) 
         cout << (*_pcurrCoordId)[i]<< ' ';
   cout << '\n';
}
#endif