/***********************************************************************
This software module was originally developed by Yannick Maret (from the
Ecole Polytechnique Federale de Lausanne, EPFL, during a stay at the
Universidad Politecnica de Madrid, UPM) in the course of development of 
the MPEG-4 Systems (ISO/IEC 14496-1) standard. This software module is
an implementation of a part of one or more MPEG-4 Systems 
(ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems
(ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 Systems 
(ISO/IEC 14496-1) free license to this software module or modifications
thereof for use in hardware or software products claiming conformance to
the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use this 
software module in hardware or software products are advised that its 
use may infringe existing patents. The original developer of this 
software module and his/her company, the subsequent editors and their
companies, and ISO/IEC have no liability for use of this software module
or modifications thereof in an implementation. Copyright is not 
released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
EPFL and UPM retain full right to use the code for his/her own purpose,
assign or donate the code to a third party and to inhibit third parties 
from using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
products. This copyright notice must be included in all copies or 
derivative works. Copyright  2003-2005
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Yannick Maret at Yannick.Maret@a3.epfl.ch or Francisco
Moran at fmb@gti.ssr.upm.es.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Player3d
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  WaveSurfDecoder
FILE:       _SSsTopo.h
AUTHOR:     Yannick Maret (EPFL, UPM)
OVERVIEW:   Interface of the topological subdivision surfaces.
-------------------------------------------------------------------------*/

#ifndef _SS_LIB_SUBDIVISION_SURFACE_INCLUDED_
#define _SS_LIB_SUBDIVISION_SURFACE_INCLUDED_
// *** INCLUDES ***
#include "_SSsCommon.h"
#include "_SSsConfig.h"
// *** EDGE ***
struct tEdge
{
   unsigned e0;
   unsigned e1;
};
class edgeList: public vector<tEdge>
{
public:
   void push_back(const unsigned, const unsigned);
};
inline void edgeList::push_back(const unsigned rE0, const unsigned rE1)
{
   static tEdge elem;
   elem.e0 = rE0;
   elem.e1 = rE1;
   vector<tEdge>::push_back(elem);
}
// *** RING ***
class ring: public vector<unsigned>
{
public:
   unsigned getPosInRing(const unsigned) const;
   unsigned getNextRidx() const;
   unsigned getPrevRidx() const;
   unsigned getNext() const;
   unsigned getPrev() const;
   void setCurrRidx(const unsigned) const;
private:
   mutable unsigned _currRidx;
};
inline unsigned ring::getPosInRing(const unsigned rToFind) const
{
   unsigned i = 0;
   while ((*this)[i] != rToFind)
      i++;
   return i;
}
inline unsigned ring::getNextRidx() const
{
   if (_currRidx < (this->size() - 1))
      return ++_currRidx;   
   else
      return _currRidx = 0;
}
inline unsigned ring::getNext() const
{
   return (*this)[getNextRidx()];
}
inline unsigned ring::getPrevRidx() const
{
   if (_currRidx > 0)
      return --_currRidx;   
   else
      return _currRidx = this->size() - 1;
}
inline unsigned ring::getPrev() const
{
   return (*this)[getPrevRidx()];
}
inline void ring::setCurrRidx(const unsigned ridx) const
{
   _currRidx =  ridx;
}
// *** VERTICES TYPE ***
enum etVertType {INTERIOR, BOUNDARY, NON_MANIFOLD};
// *** MESH TYPE ***
enum etMeshType {TRI, QUAD, OTHER};
// *** TOPOLOGY SSs CLASS ***
class _SSsTopo
{
public:
   // # PUBLIC INTERFACE
   // constructor/destructor
   _SSsTopo();
   // SSs topology creation
   void psetBaseCoordId(const vector<int>*, const enum etMeshType = OTHER);
   void createTopoLUT(const unsigned, const bool);
   // data access
   const vector<ring>* pgetRingsAtLvl(const unsigned) const;
   const ring* pgetRingsAtLvl(const unsigned, const unsigned) const;
   const vector<vector<ring> >* pgetRingsLUT() const;
   // coordId
   const vector<vector<int> >* pgetCoordIdLUT() const;
   const vector<int>* pgetCoordIdAtLvl(const unsigned) const;
   const vector<int>* pgetFinalCoordId() const;
   // vertex info
   const vector<enum etVertType>* pgetVertTypeLUT() const;
   enum etVertType getVertType(const unsigned) const;
   unsigned getNbAdjFaces(const unsigned) const;
   unsigned getNbAdjVert(const unsigned) const;
   // mesh info
   const vector<unsigned>* pgetNbEdgeLUT() const;
   const vector<unsigned>* pgetNbPolyLUT() const;
   const vector<unsigned>* pgetNbVertLUT() const;
   const unsigned getNbEdgeAtLvl(const unsigned) const;
   const unsigned getNbPolyAtLvl(const unsigned) const;
   const unsigned getNbVertAtLvl(const unsigned) const;
private:
   // # PRIVATE DATA & FUNCTIONS
   // Look-Up Tables
   vector<vector<ring> > _ringsLUT;
   vector<vector<int> > _coordIdLUT;
   const vector<int>* _pbaseCoordId;
   vector<unsigned> _nbPolyLUT, _nbVertLUT, _nbEdgeLUT;
   vector<enum etVertType> _vertTypeLUT;
   // current topology data
   vector<ring> _currRings;
   vector<int> _currCoordId;
   const vector<int>* _pcurrCoordId;
   enum etMeshType _currMeshType;
   unsigned _currNbPoly, _currNbEdge, _currNbVert;
   // functions
   void _buildRings(const bool);
   void _splitStep(const bool);;
   void _updateLUT();
   // timing data
#if _TIME_SSsTopoLIB_
   clock_t _startTime, _endTime;
   clock_t _split1Time, _split2Time;
   clock_t _ring1Time, _ring2Time;
   clock_t _LUTTime;
#endif
   // debuging functions
#if _DEBUG_SSsTopoLIB_ && _SSsTopo_DBG_VERBOSE_BASE
   void _printRings();
   void _printCoordId();
#endif
};
// # INLINED FUNCTIONS DEFINITIONS
// LUT access
inline const vector<int>* _SSsTopo::pgetCoordIdAtLvl(const unsigned r) const
{
   if (r != 0)
      return &_coordIdLUT[r];
   else
      return _pbaseCoordId;
}
inline const vector<vector<int> >* _SSsTopo::pgetCoordIdLUT() const
{
   return &_coordIdLUT;
}
inline const vector<unsigned>* _SSsTopo::pgetNbEdgeLUT() const
{
   return &_nbEdgeLUT;
}
inline const vector<unsigned>* _SSsTopo::pgetNbPolyLUT() const
{
   return &_nbPolyLUT;
}
inline const vector<enum etVertType>* _SSsTopo::pgetVertTypeLUT() const
{
   return &_vertTypeLUT;
}
inline enum etVertType _SSsTopo::getVertType(const unsigned r) const
{
   return _vertTypeLUT[r];
}
inline unsigned _SSsTopo::getNbAdjFaces(const unsigned r) const
{
   return _ringsLUT.back()[r].size() - (_vertTypeLUT[r] == BOUNDARY);
}
inline unsigned _SSsTopo::getNbAdjVert(const unsigned r) const
{
   return _ringsLUT.back()[r].size();
}
inline const vector<ring>* _SSsTopo::pgetRingsAtLvl(const unsigned r) const
{
   return &_ringsLUT[r];
}
inline const ring* _SSsTopo::pgetRingsAtLvl(const unsigned rlvl, const unsigned rvidxx) const
{
   return &_ringsLUT[rlvl][rvidxx];
}
inline const vector<vector<ring> >* _SSsTopo::pgetRingsLUT() const
{
   return &_ringsLUT;
}
inline const vector<unsigned>* _SSsTopo::pgetNbVertLUT() const
{
   return &_nbVertLUT;
}
inline const unsigned _SSsTopo::getNbEdgeAtLvl(const unsigned r) const
{
   return _nbEdgeLUT[r];
}
inline const unsigned _SSsTopo::getNbPolyAtLvl(const unsigned r) const
{
   return _nbPolyLUT[r];
}
inline const unsigned _SSsTopo::getNbVertAtLvl(const unsigned r) const
{
   return _nbVertLUT[r];
}
inline const vector<int>*  _SSsTopo::pgetFinalCoordId() const
{
   return &_coordIdLUT.back();
}
#endif//ndef _SS_LIB_SUBDIVISION_SURFACE_INCLUDED_
