/************************************************************************
 *  
 *  H263.cpp, TMN specific defines for libh263 (H.263 decoder)
 *  Copyright (C) 1995, 1996, 1997  Telenor R&D, Norway
 *        Karl Olav Lillevold <Karl.Lillevold@nta.no>
 *        Robert Danielsen <Robert.Danielsen@nta.no>
 *        Karl Anders ygard <Karl.Oygard@nta.no>
 *  
 *  Contacts: 
 *  Robert Danielsen                  <Robert.Danielsen@nta.no>
 *
 *  Telenor Research and Development  http://www.nta.no/brukere/DVC/
 *  P.O.Box 83                        tel.:   +47 63 84 84 00
 *  N-2007 Kjeller, Norway            fax.:   +47 63 81 00 76
 *  
 ************************************************************************/

/*
 * Disclaimer of Warranty
 *
 * These software programs are available to the user without any
 * license fee or royalty on an "as is" basis.  Telenor Research and
 * Development disclaims any and all warranties, whether express,
 * implied, or statuary, including any implied warranties or
 * merchantability or of fitness for a particular purpose.  In no
 * event shall the copyright-holder be liable for any incidental,
 * punitive, or consequential damages of any kind whatsoever arising
 * from the use of these programs.
 *
 * This disclaimer of warranty extends to the user of these programs
 * and user's customers, employees, agents, transferees, successors,
 * and assigns.
 *
 * Telenor Research and Development does not represent or warrant that
 * the programs furnished hereunder are free of infringement of any
 * third-party patents.
 *
 * Commercial implementations of H.263, including shareware, are
 * subject to royalty fees to patent holders.  Many of these patents
 * are general enough such that they are unavoidable regardless of
 * implementation design.
 * */
 
/*
 * based on mpeg2decode, (C) 1994, MPEG Software Simulation Group
 * and mpeg2play, (C) 1994 Stefan Eckart
 *                         <stefan@lis.e-technik.tu-muenchen.de>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#define GLOBAL
#include "config.h"
#include "H263.h"
#include "global.h"
#include "bitio.h"

const unsigned char H263Dec::zig_zag_scan[64] = {
	0,1,8,16,9,2,3,10,17,24,32,25,18,11,4,5,
	12,19,26,33,40,48,41,34,27,20,13,6,7,14,21,28,
	35,42,49,56,57,50,43,36,29,22,15,23,30,37,44,51,
	58,59,52,45,38,31,39,46,53,60,61,54,47,55,62,63
};

/* color space conversion coefficients
 *
 * entries are {crv,cbu,cgu,cgv}
 *
 * crv=(255/224)*65536*(1-cr)/0.5
 * cbu=(255/224)*65536*(1-cb)/0.5
 * cgu=(255/224)*65536*(cb/cg)*(1-cb)/0.5
 * cgv=(255/224)*65536*(cr/cg)*(1-cr)/0.5
 *
 * where Y=cr*R+cg*G+cb*B (cr+cg+cb=1)
 */

const int H263Dec::convmat[8][4] = 
{
	{117504, 138453, 13954, 34903}, /* no sequence_display_extension */
	{117504, 138453, 13954, 34903}, /* ITU-R Rec. 709 (1990) */
	{104597, 132201, 25675, 53279}, /* unspecified */
	{104597, 132201, 25675, 53279}, /* reserved */
	{104448, 132798, 24759, 53109}, /* FCC */
	{104597, 132201, 25675, 53279}, /* ITU-R Rec. 624-4 System B, G */
	{104597, 132201, 25675, 53279}, /* SMPTE 170M */
	{117579, 136230, 16907, 35559}  /* SMPTE 240M (1987) */
};

const int H263Dec::bscan_tab[4] = {2,4,6,8};
const int H263Dec::bquant_tab[4] = {5,6,7,8};

const int H263Dec::roundtab[16] =  {0,0,0,1,1,1,1,1,1,1,1,1,1,1,2,2};

// constructor

H263Dec::H263Dec()
  : fault(0),
    first(1),
    framenum(0),
    clp(NULL),
    horizontal_size(0),
    vertical_size(0),
    mb_width(0),
    mb_height(0),
    coded_picture_width(0),
    coded_picture_height(0),
    chrom_width(0),
    chrom_height(0),
    source_format(0),
    pict_type(0),
    mv_outside_frame(0),
    long_vectors(0),
    syntax_arith_coding(0),
    adv_pred_mode(0),
    pb_frame(0),
    quant(0),
    trace(0)
{
  for (int cc = 0; cc < 3; cc++) {
    exnewframe[cc] = NULL;
    edgeframeorig[cc] = NULL;
    refframe[cc] = NULL;
    oldrefframe[cc] = NULL;
    bframe[cc] = NULL;
  }
}

// destructor

H263Dec::~H263Dec()
{
  if (clp) {
    clp -= 384;
    delete clp;
  }

  for (int cc = 0; cc < 3; cc++) {
    if (exnewframe[cc])
      delete exnewframe[cc];
    if (edgeframeorig[cc])
      delete edgeframeorig[cc];
    if (refframe[cc])
      delete refframe[cc];
    if (oldrefframe[cc])
      delete oldrefframe[cc];
    if (bframe[cc])
      delete bframe[cc];
  }
}

int
H263Dec::Decode(bitiob& inbuf, unsigned char *frame[3])
{
  if (GetHeader(inbuf)) {
    if (first) {
      initdecoder();
      first = 0;
    }
    
    getpicture(inbuf, &framenum, frame, this);
    
    framenum++;

    return 1;
  } else
    return 0;
}

void H263Dec::initdecoder()
{
  int cc, size;
	
  /* clip table */
  clp = new unsigned char[1024];
  if (!clp)
    error("memory exhausted in H263Dec::initdecoder()\n");
	
  clp += 384;
	
  for (int i = -384; i < 640; ++i)
    clp[i] = (i < 0) ? 0 : ((i > 255) ? 255 : i);
	
  /* MPEG-1 = TMN parameters */
  matrix_coefficients = 5;
	
  switch (source_format) {
  case SF_SQCIF:
    horizontal_size = 128;
    vertical_size = 96;
    break;

  case SF_QCIF:
    horizontal_size = 176;
    vertical_size = 144;
    break;

  case SF_CIF:
    horizontal_size = 352;
    vertical_size = 288;
    break;

  case SF_4CIF:
    horizontal_size = 704;
    vertical_size = 576;
    break;

  case SF_16CIF:
    horizontal_size = 1408;
    vertical_size = 1152;
    break;

  default:
    printf("ERROR: Illegal input format\n");
    exit(-1);
  }
	
  mb_width = horizontal_size / 16;
  mb_height = vertical_size / 16;
  coded_picture_width = horizontal_size;
  coded_picture_height = vertical_size;
  chrom_width =  coded_picture_width >> 1;
  chrom_height = coded_picture_height >> 1;
  blk_cnt = 6;
	
  for (cc=0; cc<3; cc++) {
    if (cc==0)
      size = coded_picture_width*coded_picture_height;
    else
      size = chrom_width*chrom_height;
		
    if (!(refframe[cc] = new unsigned char[size]))
      error("memory exhausted in H263Dec::initdecoder()\n");
		
    if (!(oldrefframe[cc] = new unsigned char[size]))
      error("memory exhausted in H263Dec::initdecoder()\n");
		
    if (!(bframe[cc] = new unsigned char[size]))
      error("memory exhausted in H263Dec::initdecoder()\n");
  }
	
  for (cc=0; cc<3; cc++)
    if (cc==0) {
      size = (coded_picture_width + 64)*(coded_picture_height + 64);
      if (!(edgeframeorig[cc] = new unsigned char[size]))
	error("memory exhausted in H263Dec::initdecoder()\n");
      edgeframe[cc] = edgeframeorig[cc] + (coded_picture_width + 64) * 32 + 32;
    }
    else {
      size = (chrom_width + 32) * (chrom_height + 32);
      if (!(edgeframeorig[cc] = new unsigned char[size]))
	error("memory exhausted in H263Dec::initdecoder()\n");
      edgeframe[cc] = edgeframeorig[cc] + (chrom_width + 32) * 16 + 16;
    }
	
  if (expand)
    for (cc=0; cc<3; cc++) {
      if (cc==0)
	size = coded_picture_width * coded_picture_height * 4;
      else
	size = chrom_width * chrom_height * 4;
      
      if (!(exnewframe[cc] = new unsigned char[size]))
	error("memory exhausted in H263Dec::initdecoder()\n");
    }
  
  /* IDCT */
  if (refidct)
    init_idctref();
  else
    init_idct();
	
}

void error(char *text)
{
  fprintf(stderr,text);
  exit(1);
}

/* trace output */
void printbits(int code, int bits, int len)
{
  for (int i = 0; i < len; i++)
    printf("%d", (code >> (bits - 1 - i)) & 1);
}

int
H263Dec::FindStartcode(bitiob& inbuf)
{
  while (inbuf.peek(17) != 1 && !inbuf.eof())
    inbuf.get(1);

  return !inbuf.eof();
}

int 
H263Dec::GetHeader(bitiob& inbuf)
{
  if (FindStartcode(inbuf)) {
    inbuf.get(17);					// PSC_LENGTH = 17

    int gob = inbuf.get(5);	

    if (gob == 31)				// SE_CODE = 31
      return 0;

    if (gob == 0) {
      GetPictureHeader(inbuf);
      
      if (syntax_arith_coding)        /* reset decoder after receiving */
	decoder_reset(inbuf, this);	        /* fixed length PSC string */
    }

    return gob + 1;
  } else
    return 0;
}

void 
H263Dec::GetPictureHeader(bitiob& inbuf)
{
  int pei, tmp;

  prev_temp_ref = temp_ref;
  temp_ref = inbuf.get(8);
  trd = temp_ref - prev_temp_ref;

  if (trd < 0)
    trd += 256;

  tmp = inbuf.get(1); /* always "1" */
  if (!tmp)
    if (!quiet)
      printf("warning: spare in picture header should be \"1\"\n");
  tmp = inbuf.get(1); /* always "0" */
  if (tmp)
    if (!quiet)
      printf("warning: H.261 distinction bit should be \"0\"\n");
  tmp = inbuf.get(1); /* split_screen_indicator */
  if (tmp) {
    if (!quiet)
      printf("error: split-screen not supported in this version\n");
    exit(-1);
  }
  tmp = inbuf.get(1); /* document_camera_indicator */
  if (tmp)
    if (!quiet)
      printf("warning: document camera indicator not supported in this version\n");

  tmp = inbuf.get(1); /* freeze_picture_release */
  if (tmp)
    if (!quiet)
      printf("warning: frozen picture not supported in this version\n");

  source_format = inbuf.get(3);
  pict_type = inbuf.get(1);
  mv_outside_frame = inbuf.get(1);
  long_vectors = (mv_outside_frame ? 1 : 0);
  syntax_arith_coding = inbuf.get(1);
  adv_pred_mode = inbuf.get(1);
  mv_outside_frame = (adv_pred_mode ? 1 : mv_outside_frame);
  pb_frame = inbuf.get(1);
  quant = inbuf.get(5);
  tmp = inbuf.get(1);
  if (tmp) {
    if (!quiet)
      printf("error: CPM not supported in this version\n");
    exit(-1);
  }

  if (pb_frame) {
    trb = inbuf.get(3);
    bquant = inbuf.get(2);
  }
  else {
    trb = 0;
  }

  pei = inbuf.get(1);
pspare:
  if (pei) {
    /* extra info for possible future backward compatible additions */
    inbuf.get(8);  /* not used */
    pei = inbuf.get(1);
    if (pei) goto pspare; /* keep on reading pspare until pei=0 */
  }


  if (verbose>0)
    if (verbose>1) {
      printf("  temp_ref=%d\n", temp_ref);

      if (pb_frame)
        printf("  trb=%d\n", trb);
    }
}

/* horizontal 1:2 interpolation filter */
void conv422to444(unsigned char *src, unsigned char *dst, H263Dec* dec)
{
  int i, i2, w, j, im3, im2, im1, ip1, ip2, ip3;

  w = dec->coded_picture_width>>1;

  for (j=0; j<dec->coded_picture_height; j++) {
    for (i=0; i<w; i++) {

      i2 = i<<1;
      im3 = (i<3) ? 0 : i-3;
      im2 = (i<2) ? 0 : i-2;
      im1 = (i<1) ? 0 : i-1;
      ip1 = (i<w-1) ? i+1 : w-1;
      ip2 = (i<w-2) ? i+2 : w-1;
      ip3 = (i<w-3) ? i+3 : w-1;

      /* FIR filter coefficients (*256): 5 -21 70 228 -37 11 */
      dst[i2] =   dec->clp[(int)(   5 * src[im3]
				 - 21 * src[im2]
				 + 70 * src[im1]
				 +228 * src[i]
				 - 37 * src[ip1]
				 + 11 * src[ip2] + 128) >> 8];

      dst[i2+1] = dec->clp[(int)(   5 * src[ip3]
				 - 21 * src[ip2]
				 + 70 * src[ip1]
				 +228 * src[i]
				 - 37 * src[im1]
				 + 11 * src[im2] + 128) >> 8];
    }
    src+= w;
    dst+= dec->coded_picture_width;
  }
}

/* vertical 1:2 interpolation filter */
void conv420to422(unsigned char *src, unsigned char *dst, H263Dec* dec)
{
  int w, h, i, j, j2;
  int jm3, jm2, jm1, jp1, jp2, jp3;

  w = dec->coded_picture_width>>1;
  h = dec->coded_picture_height>>1;

  /* intra frame */
  for (i=0; i<w; i++) {
    for (j=0; j<h; j++) {
      j2 = j<<1;
      jm3 = (j<3) ? 0 : j-3;
      jm2 = (j<2) ? 0 : j-2;
      jm1 = (j<1) ? 0 : j-1;
      jp1 = (j<h-1) ? j+1 : h-1;
      jp2 = (j<h-2) ? j+2 : h-1;
      jp3 = (j<h-3) ? j+3 : h-1;
      
      /* FIR filter coefficients (*256): 5 -21 70 228 -37 11 */
      /* New FIR filter coefficients (*256): 3 -16 67 227 -32 7 */
      dst[w*j2] =     dec->clp[(int)(    3 * src[w*jm3]
				      - 16 * src[w*jm2]
				      + 67 * src[w*jm1]
				      +227 * src[w*j]
				      - 32 * src[w*jp1]
				      +  7 * src[w*jp2] + 128) >> 8];

      dst[w*(j2+1)] = dec->clp[(int)(    3 * src[w*jp3]
				      - 16 * src[w*jp2]
				      + 67 * src[w*jp1]
				      +227 * src[w*j]
				      - 32 * src[w*jm1]
				      +  7 * src[w*jm2] + 128) >> 8];
    }
    src++;
    dst++;
  }
}

#ifdef TEST_DEC

int
main(int argc, char* argv[])
{
    unsigned char *frame[3];
    bitiof infile(argv[1], 0);
    unsigned char inbuffer[38016];
    bitiob inbuf(inbuffer, 1, 38016);
    H263Dec decoder;
        
    decoder.trace = 1;
    decoder.framenum = 0;
    decoder.temp_ref = 0;
    decoder.prev_temp_ref = -1; 
        
    int step = 1;
    int dummy;
    unsigned int word;
    FILE *outfile;
        
    if ((outfile = fopen("outfile", "a")) == NULL) 
    {
                fprintf(stdout,"Couldn't append to outfile\n");
    }
    while (step != 0 && !infile.reof())
    {
                fprintf(stdout, "Press <1,RET> to decode the next frame. Anything else quits.\n");
                fscanf(stdin, "%i", &dummy);
                
                if (dummy == 1)
                {
                        int i;

                        inbuf.reset(1);
                        infile.peek(17, word);
                        if (word != 1)
                        {
                                fprintf(stderr, "No start code at start of file, exiting...\n");
                                exit(-1);
                        }
                        do {
                                infile.get(1, word);
                                inbuf.put(1, word);
                                infile.peek(17, word);
                        } while (word != 1);

                        inbuf.put(17, 0);
                        inbuf.reset(0);

                        decoder.Decode(inbuf, frame);
                        
                        fprintf(stdout,"saving to outfile\n");
                        
                        unsigned char *outbuf_u = (unsigned char *) malloc(sizeof(unsigned char)*88*144);
                        unsigned char *outbuf_v = (unsigned char *) malloc(sizeof(unsigned char)*88*144);
                        conv420to422(frame[1], outbuf_u, &decoder);
                        conv420to422(frame[2], outbuf_v, &decoder);
                        
                        for (i=0; i<144; i++) {
                                for (int j=0; j<176; j++)
                                        fputc(frame[0][i*176+j], outfile);
                        }
                        for (i=0; i<72; i++) {
                                for (int j=0; j<176; j++)
                                        fputc(outbuf_u[i*176+j], outfile);
                        }
                        for (i=0; i<72; i++) {
                                for (int j=0; j<176; j++)
                                        fputc(outbuf_v[i*176+j], outfile);
                        }
                                                
                        free(outbuf_u);
                        free(outbuf_v);
                                                
                }
                else step = 0;
    }
    fclose(outfile);
    return 0;
}

#endif

