/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  IPMP
FILE:       IPMP.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of IPMPManagerImp - base class for all
			IPMP sub-systems.
-------------------------------------------------------------------------*/
#include "defs.h"
#include "IPMPImp.h"
#include "MediaStream.h"

/*-----------------------------------------------------------------
IPMPManagerImp constructor
-----------------------------------------------------------------*/
IPMPManagerImp :: IPMPManagerImp () :
	m_ipmpThread ( & IPMPManagerImp::IPMPThread, this)
// SOLARIS PORTING
{
	m_pOutStream = m_pInStream = m_pIPMPStream = NULL;
	m_bActive = false;
}

/*-----------------------------------------------------------------
IPMP destructor
-----------------------------------------------------------------*/
IPMPManagerImp :: ~IPMPManagerImp ()
{	
	/*-----------------------------------------------------------------
	Make sure the IPMP thread terminates before the object is deleted.
	-----------------------------------------------------------------*/
	ASSERT (m_ipmpThread.GetThread () == NULL);
}

/*-----------------------------------------------------------------
Stop() closes all the streams
-----------------------------------------------------------------*/
void IPMPManagerImp :: Stop ()
{
	m_bActive = false;
	if (m_pOutStream)
		m_pOutStream -> Close ();
	if (m_pInStream)
	{	m_pInStream -> Abort ();
		m_pInStream -> Terminate ();
	}
	if (m_pIPMPStream)
	{	m_pIPMPStream -> Abort ();
		m_pIPMPStream -> Terminate ();
	}
	WaitForThread ();
}

/*-----------------------------------------------------------------
Start() opens the streams and creates the IPMP
thread.
-----------------------------------------------------------------*/
void IPMPManagerImp :: Start ()
{
	ASSERT (m_ipmpThread.GetThread () == NULL);
	m_bActive = true;
	if (m_pIPMPStream)
		m_pIPMPStream -> Start ();
	if (m_pInStream)
		m_pInStream -> Start ();
	if (m_pOutStream)
		m_pOutStream -> Open ();
	if (StartIPMPThread () == -1)
    	throw;
}

/*-----------------------------------------------------------------
Run() retrieves units from the input stream, synchronizes them with units
from the IPMP stream, decrypt the input units, and dispatches the decrypted
units to the output stream, till the input stream's end is exceeded.
-----------------------------------------------------------------*/
void IPMPManagerImp :: Run ()
{
	LPBYTE pInData, pOutData;
	int nInSize, nOutSize;
	DWORD dwTime;
	BOOL bEOU;
	LPBYTE m_pIPMPData = NULL, m_pNextIPMPData = NULL;
	int m_nIPMPDataSize = 0, m_nNextIPMPDataSize = 0;
	DWORD m_dwNextIPMPTime = 0;

	while (
		m_bActive &&
		(pInData =
			(LPBYTE) m_pInStream -> FetchNext (nInSize, dwTime, bEOU, TRUE)
		) !=
		END_OF_STREAM
	)
	{
		if (m_bActive)
		{	/*-------------------------------------------------------------
			Peform sync of input unit with IPMP unit
			-------------------------------------------------------------*/
			while (m_dwNextIPMPTime <= dwTime)
			{	if (m_pNextIPMPData != NULL) // only discard the current IPMPData if a valid NextIPMPData is arrived
				{	if (m_pIPMPData != NULL)
						m_pIPMPStream -> Release (true);
					m_pIPMPData	= m_pNextIPMPData;
					m_nIPMPDataSize = m_nNextIPMPDataSize;
				}
				m_pNextIPMPData = (LPBYTE) m_pIPMPStream -> FetchNext (
					m_nNextIPMPDataSize,
					m_dwNextIPMPTime,
					bEOU,
					(m_pIPMPData == NULL ? true : false) // block until the first data is retrieved
				);
				if (m_pNextIPMPData == NULL)
				{	m_dwNextIPMPTime = 0;
					break;
				}
				else if (m_pNextIPMPData == END_OF_STREAM)
					m_dwNextIPMPTime = ULONG_MAX;
			}

			if (m_pIPMPData) // skip the unit if no valid IPMPData is available
			{
				/*-------------------------------------------------------------
				Decrypt the unit
				-------------------------------------------------------------*/
				nOutSize = nInSize;
				pOutData = (LPBYTE) m_pOutStream -> Allocate (nOutSize, TRUE, FALSE);
				if (pOutData == NULL)
					throw;
				if (pOutData == END_OF_STREAM)
					return;
				while (!Decrypt (
							pInData,
							nInSize,
							pOutData,
							nOutSize,
							m_pIPMPData,
							m_nIPMPDataSize
						)
				)
				{	pOutData = (LPBYTE) m_pOutStream -> Reallocate (nOutSize);
					if (pOutData == NULL)
						throw;
					if (pOutData == END_OF_STREAM)
						return;
				}
				m_pOutStream -> Dispatch (nOutSize, dwTime);
			}
		}
		m_pInStream -> Release (TRUE);
	}
}

/*-----------------------------------------------------------------
The IPMP thread function calls the virtual Run() function and 
release the flag semaphore when finishes.
-----------------------------------------------------------------*/
void IPMPManagerImp :: IPMPThread ()
{
	if (Init ())
		Run ();
	Terminate ();
}
