
//	This software module was originally developed by Giovanni Caire 
//	(Telecom Italia Lab S.p.A.) and edited by Massimo Balestri (Telecom 
//  Italia Lab S.p.A.), in the course of development of the MPEG-4 Systems 
//  (ISO/IEC 14496-1) standard. 
//  This software module is an implementation of a part of one or more 
//  MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems 
//  (ISO/IEC 14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	Telecom Italia Lab S.p.A retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2000. 


#include "BitField.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

USING_MOSES_NAMESPACE

// constructor
BitField::BitField()
{
	rbuf = wbuf = NULL;			// last read/write buffer
	rindex = windex = 0;		// last read/write index
}

// copy constructor - must RESET the status var instead of copying!
BitField::BitField(const BitField & bf)
{
	rbuf = wbuf = NULL;			// last read/write buffer
	rindex = windex = 0;		// last read/write index
}

// destructor
BitField::~BitField()
{
}

// set the read buffer
void BitField::setReadBuffer(const byte *buf)
{
	rbuf = buf;
	rindex = 0;
}

// set the write buffer
void BitField::setWriteBuffer(byte *buf)
{
	wbuf = buf;
	windex = 0;
}

// get - read bits from a buffer
word32 BitField::get(const byte *buf,
           long  first,
           long  len)

{
long          last,first_byte,last_byte,standing,i;
word32        value,temp;
static word32 mask[9] = { 0x00000000,
                          0x00000001,
                          0x00000003,
                          0x00000007,
                          0x0000000f,
                          0x0000001f,
                          0x0000003f,
                          0x0000007f,
                          0x000000ff };

  rbuf = buf;					// save buffer for next "getNext" operation
  rindex = first + len;			// next read index


  last = first + len - 1;
  first_byte = first / 8;
  last_byte = last / 8;
  first %= 8;
  last %= 8;

  if (first_byte == last_byte)
      value = (buf[first_byte] >> (7 - last)) & mask[len];
  else
  {
      value = buf[first_byte] & mask[8 - first];
      for (i = 1;i < last_byte - first_byte;++i)
          cat(&value,buf[first_byte + i],8);
      standing = last + 1;
      temp = (buf[last_byte] >> (8 - standing)) & mask[standing];
      cat(&value,temp,standing);
  }
  return(value);
}


void BitField::getBuffer(void * destination,
			const byte *buf,
			long  first,
			long  len)
{
	rbuf = buf;						// save buffer for next "getNext" operation
	rindex = first + len;			// next read index
	assert((first % 8) == 0);		// check if byte-aligned
	assert((len % 8) == 0);			// check if byte-aligned

	if (len >= 8)
		memcpy(destination, buf + (first/8), len/8);
}


void BitField::getBuffer(void * destination, long  len)
{
	if (rbuf != NULL)
		getBuffer(destination, rbuf, rindex, len);
}

void BitField::putBuffer(void * source, byte *buf, long first, long len)
{
	wbuf = buf;						// save buffer for next "append" operation
	windex = first + len;			// next write index
	assert((first % 8) == 0);		// check if byte-aligned
	assert((len % 8) == 0);			// check if byte-aligned

	if (len >= 8)
		memcpy(buf + (first/8), source, len/8);
}
	
void BitField::putBuffer(void * source, long  len)
{
	if (wbuf != NULL)
		putBuffer(source, wbuf, windex, len);
}

word32 BitField::get(long nbits)
{
	if (rbuf != NULL)
		return get(rbuf, rindex, nbits);
	else
		return 0;
}

long BitField::getReadCounter()
{
	return rindex;
}

void BitField::skipReading(long nbits)
{
	rindex += nbits;
}

void BitField::skipWriting(long nbits)
{
	windex += nbits;
}


void BitField::cat(word32 *dest,
         word32  value,
         long    len)

{
  *dest <<= len;
  *dest |= value;
}


void BitField::put(word32  val,
         byte   *buf,
         long    first,
         long    len)

{
long i_bit,i_byte,offset,shift;
bit  b;

  wbuf = buf;					// save buffer for next "append" operation
  windex = first + len;			// next write index

  for (i_bit = 0;i_bit < len;++i_bit)
  {
      i_byte = (first + i_bit) / 8;
      offset = (first + i_bit) % 8;
      b = (bit) ((val >> (len - 1 - i_bit)) & 0x00000001);
      shift = 7 - offset;
      buf[i_byte] &= ~(0x01 << shift);
      buf[i_byte] |= (b << shift);
  }
}


void BitField::put(word32 val, long nbit)
{
	if (wbuf != NULL)						// at least one absolute "put" call must be made! 
		put(val, wbuf, windex, nbit);
}


long BitField::getWriteCounter()
{
	return windex;
}

            
/* compute_CRC: CRC is computed as the remainder of the division N(x) / g(x) where:
   - N(x) is the polynomial whose coefficient are the bits stored in buf (first
     bit = highest degree coeff.) 
   - g(x) is the generator polynomial whose coefficient are the bits of the 
     parateter pol_gen (since highest degree coeff. is always 1 it musten't be
     specified)
   - n is the degree of g(x) and is specified by the parameter CRC_len

   The remainder is computed according to the following scheeme.

       b...................b---------------------------- p --->-- 
     last                first                                   |
                                                                 |
           ------<--------------back--- ... ---------------<----(+)
          |               |                      |               | q
          |               |                      |               |
     g0-->|          g1-->|               gn-1-->|          gn-->|
          |               |                      |         (=1)  |
          |               |                      |               ^
          |               |                      |               |
           ----->[R0]----(+)---->[R1]-- ... ----(+)---->[Rn-1]-----------> q
                 LSB                                     MSB

   Where - gi is the i-th bit of pol_gen (g0 = LSB)
         - the sums are |2| (binary XOR)
         - at the end of computation the remainder is stored in the R */

word32 BitField::compute_CRC(byte  *buf,
                   long   len,       /* Number of bit to process */
                   word32 pol_gen,
                   long   CRC_len,   /* CRC length = generator polynomial degree */
                   word32 reg0)      /* Initial shift register value */  
{
word32 R,tmp,mask1,mask2;
bit    q,p,back;
long   i_bit;

  R = reg0;
  mask1 = 1;
  if (CRC_len < 32)
      mask1 = (mask1 << CRC_len) - 1;
  else
      mask1 = 0xffffffff;
  mask2 = mask1 & 0xfffffffe;

  for (i_bit = 0;i_bit < len;++ i_bit)
  {
      p = (unsigned char)((buf[i_bit / 8] >> (7 - i_bit % 8)) & 0x01); /* get bit # i_bit 
                                                         from buffer */
      q = (bit) ((R >> (CRC_len - 1)) & 0x00000001);
      back = (unsigned char)((p ^ q) & 0x01);

      /** Update R **/
      tmp = (R << 1) & mask2;
      R = (tmp ^ (back * pol_gen)) &  mask1;
  }
  return(R);
}



bit BitField::shift(word32 *shift_reg, 
          long    shift_reg_len,
          word32  mask,
          bit     in)

{
bit out;

  out = (bit) ((*shift_reg >> (shift_reg_len - 1)) & 0x00000001);
  *shift_reg = ((*shift_reg << 1) & mask) | in;
  return (out);
}



const byte * BitField::getReadPtr()
{
	assert((rindex % 8) == 0);			// check if byte-aligned
	return (rbuf + (rindex/8));
}


byte * BitField::getWritePtr()
{
	assert((windex % 8) == 0);			// check if byte-aligned
	return (wbuf + (windex/8));
}


bool BitField::toFile(const char *pathname)
{
	if (! checkSyntax())
		return false;							// wrong syntax!
	FILE* fout;
	if ((fout = fopen(pathname, "wb")) == NULL) 
		return false;
	
	unsigned int size = getSize();				// size = size in bytes of this (derived) object
	byte * buffer = new byte[size];			
	memset(buffer, 0, size);
	unsigned int bitsize = toBuffer(buffer);	// dump data into a buffer
	assert((bitsize % 8) == 0);					// check size
	assert((bitsize/8) == size);				// check size
	fwrite(buffer, 1, size, fout);				// write n bytes into the file
	fclose(fout);
	delete[] buffer;							// destroy the temporary buffer

	return true;
}

bool BitField::fromFile(const char *pathname)
{
	FILE* fin;
	if ((fin = fopen(pathname, "rb")) == NULL)
		return false;

	fseek(fin,0L,SEEK_END);
	long size = ftell(fin);					// get the size of the file
	fseek(fin, 0L, SEEK_SET);

	if (size > 0)
	{
		byte * buffer = new byte[size];			// create the temporary buffer
		memset(buffer,0,size);					// cleanup the buffer
		fread(buffer,1,	size, fin);				// read from file
		fromBuffer(buffer);						// create an instance of this class
		delete[] buffer;						// destroy the temporary buffer
	}

	fclose(fin);
	return checkSyntax();					// check the syntax
}

void BitField::putObject(BitField & object)
{
	skipWriting(object.toBuffer( getWritePtr() ));
}

void BitField::getObject(BitField & object)
{
	skipReading(object.fromBuffer( getReadPtr() ));
}

bool BitField::operator == (BitField & second)
{
	unsigned int size = getSize();				// size = size in bytes of this object
	byte * firstBuffer = new byte[size];			
	unsigned int bitsize = toBuffer(firstBuffer);	// dump data into a buffer
	assert((bitsize % 8) == 0);					// check size
	assert((bitsize/8) == size);				// check size

	unsigned int size2 = second.getSize();				// size = size in bytes of this object
	byte * secondBuffer = new byte[size2];			
	bitsize = second.toBuffer(secondBuffer);	// dump data into a buffer
	assert((bitsize % 8) == 0);					// check size
	assert((bitsize/8) == size2);				// check size

	int result;

	if (size == size2)
		result = memcmp(firstBuffer, secondBuffer, size);
	else
		result = -1;

	delete[] firstBuffer;
	delete[] secondBuffer;

	return (result == 0);
}