
//	This software module was originally developed by Giovanni Caire 
//	(Telecom Italia Lab S.p.A.) and edited by Massimo Balestri (Telecom 
//  Italia Lab S.p.A.), in the course of development of the MPEG-4 Systems 
//  (ISO/IEC 14496-1) standard. 
//  This software module is an implementation of a part of one or more 
//  MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems 
//  (ISO/IEC 14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	Telecom Italia Lab S.p.A retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2000. 


#ifndef __BITFIELD__
#define __BITFIELD__
#include "namespace.h"

#define RESERVED	0	// all reserved bits have this value

typedef unsigned char  byte;
typedef unsigned long  word32;
typedef unsigned char  bit;

BEGIN_MOSES_NAMESPACE

/**
 *  @class BitField
 *  Abstract class representing a sequence of bits.
 *	All derived classes use the bit handling methods of this class.
 *  All bits are stored and read in big endian format, independently of the endianness of the CPU.
 *  This class is not thread-safe, a single instance of it must be used by a single thread. 
 *  @author Giovanni Caire, Massimo Balestri
 *  @date 1995
 */
class BitField
{
private:
	const byte * rbuf;		// last read buffer
	byte * wbuf;			// last write buffer
	long   rindex;			// last read index
	long   windex;			// last write index

	void   cat(word32 *dest, word32 val, long len);
	bit    shift(word32 *shift_reg, long shift_reg_len, word32 mask, bit p);

protected:
	/**
	 * The constructor is protected since only derived objects of this type can be created.
	 */
	BitField();
	
	/**	
	 * The copy constructor is protected since only derived objects of this type can be created.
	 */
	BitField(const BitField & bf);

	/**
	 * set the read buffer and initialize the read pointer to 0;
	 * this can be used to use relative "get" calls instead of absolute ones.
	 * Any subsequent call to an absolute get supercedes this method.
	 * @param buf the buffer that will be read by subsequent "get" calls.
	 */
	void setReadBuffer(const byte * buf);

	/**
	 * set the write buffer and initialize the write pointer to 0;
	 * this can be used to use relative "put" calls instead of absolute ones.
	 * Any subsequent call to an absolute put supercedes this method.
	 * @param buf the buffer that will be written by subsequent "put" calls.
	 */
	void setWriteBuffer(byte * buf);	

	/**
	 * skip the next n bits while reading from the current position.
	 * @param nbits number of bits to skip
	 */
	void skipReading(long nbits);

	/**
	 * skip the next n bits while writing from the current position.
	 * @param nbits number of bits to skip
	 */
	void skipWriting(long nbits);
	
	/**
	 * reads a maximum of 32 bits from a buffer.
	 * @param buf the buffer from which to read
	 * @param first the offset in the buffer of the first bit to read, normally 0.
	 * @param len the number of bits to read.
	 * @return the value of selected bits read from the buffer.
	 */
	word32 get(const byte *buf, long first, long len);

	/**
	 * reads an entire block of bits from a buffer, assuming that the block is byte-aligned.
	 * @param destination the destination buffer
	 * @param buf the buffer from which to read
	 * @param first the offset in the buffer of the first bit to read, normally 0.
	 * @param len the number of bits to read.
	 */
	void getBuffer(void * destination, const byte *buf, long first, long len);
	
	/**
	 * Incremental version of getBuffer. This version reads from the same buffer used by the last read operation.
	 * @param destination the destination buffer
	 * @param len the number of bits to read.
	 */
	void getBuffer(void * destination, long  len);
	
	/**
	 * reads the next n bits (max. 32) from the same buffer of the last read operation.
	 * The first time, an absolute "get" must be used. Then, all subsequent consecutive reads from the same buffer 
	 * can be done using this method.
	 * @param nbits the number of bits to read.
	 * @return the value of selected bits read from the buffer.
	 */
	word32 get(long nbits);
	
	/**
	 * writes a maximum of 32 bits into a buffer.
	 * @param val the value to store into the buffer
	 * @param buf the destination buffer
	 * @param first the offset in the buffer of the first bit to write, normally 0.
	 * @param len the number of bits to write
	 */
	void put(word32 val, byte *buf, long first, long len);

	/**
	 * writes the next n bits (max. 32) into the same buffer of the last write operation.
	 * The first time, the absolute "put" must be used. Then, all subsequent consecutive writes into the same buffer 
	 * can be done using this method.
	 * @param val the value to store into the buffer
	 * @param nbit the number of bits to write.
	 */
	void put(word32 val, long nbit);

	/**
	 * writes an entire block of bits from a buffer, assuming that the block is byte-aligned.
	 * @param source the source buffer
	 * @param buf the buffer into which to write
	 * @param first the offset in the buffer of the first bit to write, normally 0.
	 * @param len the number of bits to write.
	 */
	void putBuffer(void * source, byte *buf, long first, long len);

	/**
	 * writes an object of type BitField into the same buffer of the last write operation.
	 * @param object the BitField object to serialize into the buffer
	 */
	void putObject(BitField & object);
	
	/**
	 * reads an object of type BitField from the same buffer of the last read operation.
	 * @param object the BitField object to read from the buffer
	 */
	void getObject(BitField & object);

	
	/**
	 * Incremental version of putBuffer. This version writes into the same buffer used by the last write operation.
	 * @param source the source buffer
	 * @param len the number of bits to write.
	 */
	void putBuffer(void * source, long  len);

	/**
	 * returns the read index that will be used in the next getNext() operation.
	 * This is also a counter of the number of bits read so far.
	 * @return the number of bits read from the buffer so far.
	 */
	long getReadCounter();

	/**
	 * returns the write index that will be used in the next append() operation.
	 * This is also a counter of the number of bits written so far.
	 * @return the number of bits written into the buffer so far.
	 */
	long getWriteCounter();

	/**
	 * Compute the CRC of a data chunck.
	 * CRC is computed as the remainder of the division N(x) / g(x) where:
     * - N(x) is the polynomial whose coefficient are the bits stored in buf (first bit = highest degree coeff.) 
     * - g(x) is the generator polynomial whose coefficient are the bits of the 
     *  parateter pol_gen (since highest degree coefficient is always 1 it need not be specified)
     * - n is the degree of g(x) and is specified by the parameter CRC_len
	 * @param buf the input buffer
	 * @param len Number of bit to process
	 * @param pol_gen the generator polynomial 
	 * @param CRC_len the lenght of the CRC = generator polynomial degree 
	 * @param reg0 Initial shift register value
	 * @return the remainder R
	 */
	word32 compute_CRC(byte *buf, long len, word32 pol_gen, long CRC_len, word32 reg0);

	/**
	 * returns the current read pointer, assuming it is byte-aligned.
	 * @return the current read pointer
	 */
	const byte * getReadPtr();

	/**
	 * returns the current write pointer, assuming it is byte-aligned.
	 * @return the current write pointer
	 */
	byte * getWritePtr();

public:
	
	/**
	 * clean up the allocated instance.
	 */
	virtual ~BitField();
	
	/**
	 *	reads the data from the buffer and stores the fields internally.
	 *  Must be implemented in derived classes.
	 *  @param buf the buffer to read
	 *  @return the number of bits that have been read from the buffer 
	 */
	virtual int fromBuffer(const byte* buf) = 0;
	
	/**
	 *	writes the internal fields into the buffer
	 *  @param buf the buffer to write
	 *  @return the number of bits that have been written into the buffer 
	 */
	virtual int toBuffer(byte* buf) = 0;
	
	/**
	 *	checks if the syntax is correct.
	 *  @return true if the syntax is correct. 
	 */
	virtual bool checkSyntax() const = 0;

	/**
	 *  this method must be implemented in derived classes so that it returns the 
	 *  size in bytes of the binary message generated by the toBuffer method.
	 *  @return the size (in bytes) of this class.
	 */
	virtual unsigned int getSize() const = 0;
	
	/**
	 * reads the message from a binary file. This can be used for conformance testing,
	 * e.g. in order to exchange bitstreams among different implementers of the MPEG specifications.
	 * @param pathname the path name of the file to be read
	 * @return true if successful
	 */
	bool fromFile(const char * pathname);

	/**
	 * writes the internal fields into a binary file. This can be used for conformance testing,
	 * e.g. in order to exchange bitstreams among different implementers of the MPEG specifications.
	 * @param pathname the path name of the file that will store the result
	 * @return true if successful
	 */
	bool toFile(const char * pathname);
	
	/**
	 * Compare the bit field with another bit field. 
	 * The two are equal if they have the same size, and they produce the same bitstream.
	 * @param msg the message to compare with.
	 * @return true if the two bit fields are equal, false otherwise.
	 */
	bool operator == (BitField & msg);
};

END_MOSES_NAMESPACE
#endif
