
//	This software module was originally developed by Massimo Balestri (TILAB)
//	in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) standard. 
//	This software module is an implementation of a part of one or more MPEG-4 
//	Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 
//	14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	Telecom Italia Lab retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2002.

// Container.h

#ifndef	__CONTAINER_H__
#define	__CONTAINER_H__

#include "namespace.h"
#include <stdlib.h>

BEGIN_MOSES_NAMESPACE


/** 
 *  @class Container
 *  A simple container for objects.
 *  The Container is expandable, so it never fails adding an item;
 *  however, it can run out of memory.
 *  
 *  @author Massimo Balestri
 *  @date 2003
 */
template <class Object>
class Container
{
private:
	enum {
		BASE_SIZE = 16,				// the size of the initial vector
		INCREMENT =  8				// the increment used to inflate the vector  
	};

	Object** vector;				// the vector of pointers to the IPMP Tool descriptors
	unsigned int num;				// number of elements in the vector
	unsigned int size;				// size of the array

	void checksize()				// check size before adding a new object
	{
		if (num == 0)
		{
			vector = new Object * [BASE_SIZE];
			if (vector == NULL)
				throw;
			size = BASE_SIZE;						// set current vector size
		}

		if (num == size)								// inflate the vector
		{
			Object ** newVector = new Object * [size + INCREMENT];
			if (newVector == NULL)
				throw;
			memcpy(newVector, vector, size * sizeof(Object *));
			delete[] vector;
			vector = newVector;
			size += INCREMENT;
		}
	}

public:

	/**
	 * constructor.
	 */
	Container()
	{
		vector = NULL;
		num = size = 0;
	}

	/**
	 * copy constructor.
	 */
	Container(const Container & obj)
	{
		vector = NULL;
		num = size = 0;
		for (unsigned int i = 0; i < obj.getObjectNum(); i++)
			addObject( obj.getObject(i));
	}

	/**
	 * destructor
	 */
	virtual ~Container()	
	{
		clear();			// delete all
	}

	/**
	 * Remove all objects from the Container.
	 */
	void clear()					
	{
		if (num > 0)
		{
			for (unsigned int i=0; i<num; i++)
				delete vector[i];

			delete[] vector;
		}
		vector = NULL;
		num = size = 0;
	}

	/**
	 * add an object to the Container.
	 * @param obj the Object to be copied into the list. 
	 * @return the index of the object in the list if successful, otherwise -1.
	 */
	int addObject(const Object & obj)
	{
		checksize();
		vector[num++] = new Object(obj);		// use copy constructor
		return (num - 1);
	}

	/**
	 * Get the number of Objects contained in the Container.
	 * @return the object count.
	 */
	unsigned int getObjectNum() const
	{
		return num;
	}	

	/**
	 * Get the n-th Object from the Container.
	 * @param index the index of the Object in the vector.
	 * @return if index is in range, a pointer to the Object.
	 */
	Object & getObject(unsigned int index) const
	{
		if (index < num)
			return *vector[index];
		else
			throw;		// index out of range.
	}
};


END_MOSES_NAMESPACE

#endif
