
//	This software module was originally developed by Antonio Romeo (EPFL)
//  and modified by Massimo Balestri	
//	in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) standard. 
//	This software module is an implementation of a part of one or more MPEG-4 
//	Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 
//	14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	EPFL retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2002.

// IPMP_AudioWatermarkingInit.h

#ifndef	IPMP_AUDIOWATERMARKINGINIT_H
#define	IPMP_AUDIOWATERMARKINGINIT_H

#include "IPMP_Data_BaseClass.h"
#include "namespace.h"

BEGIN_MOSES_NAMESPACE

/** 
 * @class IPMP_AudioWatermarkingInit
 * Delivers to an audio watermerking tool all the information about the characteristics of audio content.
 * @author Antonio Romeo, Massimo Balestri
 * @date 2002
 */
class IPMP_AudioWatermarkingInit: public IPMP_Data_BaseClass 
{
private:
	word32 inputFormat;
	word32 requiredOp;
	bool   b_hasOpaqueData;

	word32 nChannels;
	word32 bitPerSample;
	word32 frequency;

	word32 wmPayloadLen;
	byte*  wmPayload;
	word32 wmRecipientId;
	word32 opaqueDataSize;
	byte*  opaqueData ;

	void clearUp() ;
	void init() ;

public:
	/**
	 * Table of possible values of the inputFormat parameter. 
	 * This table will be maintained by a registration authority.
	 * The current values have been assigned by MOSES.
	 */
	enum InputFormatTable {
		PCM					= 0x01,		///< uncompressed PCM audio
		MPEG4_AUDIO			= 0x40,		///< compressed MPEG-4 audio
		MPEG2_AUDIO_MAIN	= 0x66,		///< compressed MPEG-2 audio, main profile
		MPEG2_AUDIO_LOW		= 0x67,		///< compressed MPEG-2 audio, low profile
		MPEG2_AUDIO_SCAL	= 0x68,		///< compressed MPEG-2 audio, scalable profile
		MPEG2_AUDIO			= 0x69,		///< compressed MPEG-2 audio
		MPEG1_AUDIO			= 0x6B		///< compressed MPEG-1 audio
	};
	
	/**
	 * Table of possible values of the requiredOp parameter. 
	 * It indicates the operation that the watermarking tool is required to perform on the
	 * audio stream.
	 */
	enum RequiredOpTable {
		INSERT_WM			= 0,		///< insert watermark 
		EXTRACT_WM			= 1,		///< extract watermark
		REMARK_WM			= 2,		///< remark watermark
		DETECT_COMPRESSION	= 3,		///< detect compression
		USER_DEFINED		= 11		///< first user defined entry 
	};

	IPMP_AudioWatermarkingInit();										///< constructor
	IPMP_AudioWatermarkingInit(const IPMP_AudioWatermarkingInit & obj); ///< copy constructor
	virtual ~IPMP_AudioWatermarkingInit();								///< destructor	
	virtual int toBuffer(byte * buf);				// reimplemented
	virtual int fromBuffer(const byte * buf);		// reimplemented
	virtual unsigned int sizeOfInstance() const;	// reimplemented

	/**
	 * Get the number of audio channels (1=mono, 2=stereo...)
	 * @return the number of audio channels.
	 */
	word32 getnChannels() const;

	/**
	 * Get the number of bit per sample (e.g. 8, 16) of the input stream.
	 * @return the number of bit per sample.
	 */
	word32 getbitPerSample() const;

	/**
	 * Get the number of samples per second (in Hz, e.g. 44100) of the input stream.
	 * @return the frequency in samples/s.
	 */
	word32 getfrequency() const;
	
	/**
	 * Get the input format, which can be one of the values defined in InputFormatTable.
	 * @return 
	 */
	enum InputFormatTable getInputFormat() const;

	/**
	 * Get the required operation, which can be one of the values defined in RequiredOpTable.
	 * @return the operation code.
	 */
	enum RequiredOpTable getRequiredOp() const;
	
	/**
	 * Set the watermarking payload 
	 * @param buf the buffer containing the new watermarking payload
	 * @param len the length of the buffer 
	 * @return true if successful
	 */
	bool   setwmPayload(const byte* buf, word32 len);
	/**
	 * Get the  size of the watermark payload.
	 * @return the size in bytes of the watermark payload. 
	 */
	word32 getwmPayloadLen() const;

	/**
	 * Get the  watermark payload.
	 * @return a pointer to the watermark payload. The pointer is valid as long as the instance is not deleted. 
	 */
	byte * getwmPayload() const;

	/**
	 * Get the destination tool identified by the IPMP_Descriptor ID, to
	 * which the watermarking payload and compression information must be delivered.
	 * A value of 0x00 indicates the terminal.
	 * @return the watermark recipient ID. 
	 */
	word32 getwmRecipientId() const;
	
	/**
	 * Get the opaque data size.
	 * @return the opaque data size. 
	 */
	word32 getOpaqueDataSize() const;

	/**
	 * Get the opaque data.
	 * @return a pointer to the opaque data. The pointer is valid as long as the instance is not deleted. 
	 */
	byte * getOpaqueData() const;

	/**
	 * Indicates if some opaque data is available.
	 * @return TRUE if some opaque data is available. 
	 */
	bool hasOpaqueData() const;

	/**
	 * Set the audio input format.
	 * @param format the input format(can be PCM, MPEG1_AUDIO, etc.)
	 * @param nchannels the number of audio channels (1=mono, 2=stereo...)
	 * @param bitpersample the number of bit per sample.
	 * @param frequency get the number of samples per second (in Hz, e.g. 44100) of the input stream.
	 * @return TRUE if successful.
	 */
	bool setInputFormat(InputFormatTable format, word32 nchannels = 0, word32 bitpersample = 0, word32 frequency = 0);

	/**
	 * Require the watermarking tool to insert the specified data as watermarking payload.
	 * @param payload the watermark payload
	 * @param size the size in bytes of the watermark payload
	 * @return TRUE if successful.
	 */
	bool insertWatermark(const byte *payload, word32 size);

	/**
	 * Require the watermarking tool to remark the specified data as watermarking payload.
	 * @param payload the watermark payload
	 * @param size the size in bytes of the watermark payload
	 * @return TRUE if successful.
	 */
	bool remarkWatermark(const byte *payload, word32 size);	

	/**
	 * Require the watermarking tool to extract a watermark payload from the input audio stream.
	 * @param recipientId the destination tool identified by the IPMP_Tool_Descriptor_ID, to
	 * which the watermarking payload must be delivered.
	 * A value of 0x00 indicates the terminal.
	 * @return TRUE if successful.
	 */
	bool extractWatermark(word32 recipientId);

	/**
	 * Require the watermarking tool to detect if the input audio stream has been compressed.
	 * @param recipientId the destination tool identified by the IPMP_Tool_Descriptor_ID, to
	 * which the compression information must be delivered.
	 * A value of 0x00 indicates the terminal.
	 * @return TRUE if successful.
	 */
	bool detectCompression(word32 recipientId);

	/**
	 * Set the private data field of the message.
	 * @param buf the buffer containing private data
	 * @param len the length of the buffer
	 */
	bool setOpaqueData(const byte* buf, word32 len);

};

END_MOSES_NAMESPACE

#endif
