
//	This software module was originally developed by Massimo Balestri (TILAB)
//	and edited by James King (CRL), in the course of development of the MPEG-4 Systems 
//	(ISO/IEC 14496-1) standard. 
//	This software module is an implementation of a part of one or more MPEG-4 
//	Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 
//	14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	Telecom Italia Lab S.p.A. retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2002.

// IPMP_Descriptor.h

#ifndef	__IPMP_DESCRIPTOR_H__
#define	__IPMP_DESCRIPTOR_H__

#include "namespace.h"
#include "BaseDescriptor.h"
#include "container.h"
#include <stdlib.h>

BEGIN_MOSES_NAMESPACE

class IPMP_Data_BaseClass ;

/** 
 *  @class IPMP_Descriptor
 *  The IPMP_Descriptor carries IPMP information for one or more IPMP Tool instances. 
 *  It shall also contain optional instantiation information for one or more IPMP Tool 
 *  instances.  IPMP_Descriptors are conveyed in object descriptor streams via IPMP_DescriptorUpdate 
 *  commands or as part of the decoder specific information of an IPMP stream.  
 *  Multiple definitions of the same IPMP_Descriptor within a single IPMP_DescriptorUpdate
 *  command or a single decoder specific information structure for an IPMP stream are not 
 *  allowed.  The behavior in such a situation is undefined.  Note that, however, an IPMP_Descriptor 
 *  may be modified/updated through subsequent IPMP_DescriptorUpdate commands received 
 *  in the OD stream.  IPMP_Descriptors shall be referenced by object descriptors or ES_Descriptors, 
 *  using IPMP_DescriptorPointer. 
 *  @author Massimo Balestri
 *  @date 2002
 */
class IPMP_Descriptor: public BaseDescriptor 
{
public:
	enum DescrTypeEnum {
		HOOKS_TYPE		= 0,		///< indicates that this descriptor contains "hooks" data
		IPMPX_TYPE		= 1,		///< indicates that this descriptor contains IPMPX data
		URL_TYPE		= 2			///< indicates that this descriptor contains just an URL to another descriptor
	};

private:
	enum LocalConstants {
		MAX_IPMPX_DATA_MESSAGES = 255
	};
	
	DescrTypeEnum		descrType;					// identifies the type of this IPMP descriptor
	unsigned int		IPMP_DescriptorID;
	unsigned int		IPMPS_Type ;
	unsigned int		controlPointCode ;
	unsigned int		sequenceCode ;
	unsigned int		sizeOfData ;				// contains either the size of IPMP data or the number of IPMPX messages, depending on descrType (HOOKS or IPMPX)
	byte				IPMP_ToolId[16];			// the 128 bit id of the Tool

	Container<IPMP_Data_BaseClass*> IPMPX_data;

	char *				URLString ;
	byte *				IPMP_data ;

	void cleanUp() ;			// free all allocated memory
	void init();				// initialize all vars
	void clearIPMPX_data();		// free all mamory allocated to IPMPX messages

public:
	IPMP_Descriptor();								///< constructor
	IPMP_Descriptor(const IPMP_Descriptor & descriptor);	///< copy constructor
	virtual ~IPMP_Descriptor();						///< destructor
	virtual int toBuffer(byte * buf);				// reimplemented (same documentation)
	virtual int fromBuffer(const byte * buf);		// reimplemented
	virtual unsigned int sizeOfInstance() const;	// reimplemented
	
	/**
	 * Get the current type of the descriptor.
	 * @return an enum indicating the type. Can be IPMPX_TYPE, HOOKS_TYPE, URL_TYPE.
	 */
	DescrTypeEnum getType() const;

	/**
	 * Set this descritpor as an IPMPX descriptor.
	 * @param descrId a unique ID for this IPMP_Descriptor within its name scope. Values of 0x0000 and 0xFFFF are forbidden.  The scope of the IPMP_DescriptorID is the same as the OD, or IOD in which is it contained.
	 * @param toolID the globally unique id of the IPMP Tool described by this IPMP_Descriptor.
	 * @param controlPointCode The control point code specifies the IPMP control point at which the IPMP Tool 
	 * resides, and is one of the following values:
	 *            <table>
	 *                <tr><td><b>controlPointCode</b></td><td><b>Description</b></td></tr>	
	 *                <tr><td>0x00</td><td>No control point.</td></tr>
	 *                <tr><td>0x01</td><td>Control Point between the decode buffer and the decoder. This is between the decode buffer and class loader for MPEG-J streams.</td></tr>
	 *                <tr><td>0x02</td><td>Control Point between the decoder and the composition buffer.</td></tr>
	 *                <tr><td>0x03</td><td>Control Point between the composition buffer and the compositor.</td></tr>
	 *                <tr><td>0x04</td><td>BIFS Tree</td></tr>
	 *                <tr><td>0x05-0xDF</td><td>ISO Reserved.</td></tr>
	 *                <tr><td>0xE0-0xFE</td><td>User defined</td></tr>
	 *                <tr><td>0xFF</td><td>Forbidden</td></tr>
	 *            </table>
	 * @param sequenceCode indicates the sequencing priority of the IPMP Tool instance at the given control point. The tool with the highest sequenceCode for a given control point on a given stream shall process data first for that control point for that stream. Two tools shall not have the same sequence number at the same control point for the same stream.
	 * @return true if successful.
	 * @note The only difference between receiving composition units before the CB and after the CB in the MPEG-4 Systems decoder model is the order in which the units are received when the associated DTS is different from the CTS; in this case the decoding order is different from the composition order. For example, suppose that a watermark payload is embedded in more than a single video frame; if the watermark payload was embedded in decoding order, it has to be extracted before the CB; instead, if it was embedded in composition order, it has to be extracted after the CB.
	 * @note For streams of type 0x01, ObjectDescriptor and of type 0x02, ClockReferenceStream, only a controlPointCode of 0x00, 0x01 or the range 0xE0-0xFE are meaningful.
	 */
	bool setIPMPXType(unsigned short descrId, byte * toolID, byte controlPointCode, byte sequenceCode = 0);
	
	/**
	 * Set this descriptor as an URL type container.
	 * @param descrId a unique 8-bit ID for this IPMP descriptor within its namescope.
	 * @param URL contains a UTF-8 encoded URL that shall point to the location of a remote IPMP_Descriptor. If the IPMP_ToolID of this  IPMP_Descriptor is 0, another URL is referenced.  This process continues until an IPMP_Descriptor with a non-zero IPMP_ToolID is accessed. 
	 * @return true if successful.
	 */
	bool setUrlType(byte descrId, const char * URL);
	
	/**
	 * Set this descriptor as an "hooks" type descriptor, i.e. compliant to the old version of
	 * the MPEG-4 System specification.
	 * @param descrId a unique 8-bit ID for this IPMP descriptor within its namescope.
	 * @param IPMPS_Type the ID of the IPMP System ("hooks" version), assigned by a registration authority
	 * @param IPMP_data opaque data of unspecified format
	 * @param size lenght in bytes of the IPMP_data array
	 * @return true if successful.
	 * @deprecated This method is provided for backward compatibility only.
	 */
	bool setHooksType(byte descrId, unsigned short IPMPS_Type, byte * IPMP_data = NULL, int size = 0);
	
	/**
	 * Add an IPMPX message to the list of messages. 
	 * This method can only be called when creating IPMPX type descriptors.
	 * @param msg the message to add
	 * @return true if successful
	 */
	bool addIPMPXMessage( IPMP_Data_BaseClass &msg );

	/** 
	 * Get the URL contained in a URL type descriptor.
	 * @return a pointer to the URL string.
	 */
	char * getURLString() const;

	/** 
	 * Get the size of internal data. This has a different meaning depending on the descriptor type.
	 * For IPMPX descriptor, this is the number of IPMPX messages contained in the internal array.
	 * For URL descriptor, this is the size (in bytes) of the URL string.
	 * For HOOKS descriptors, this is the size (in bytes) of the opaque data.
	 * @return the size of internal data.
	 */
	int getSizeOfData() const;

	/** 
	 * Get the sequence code (only meaningful in IPMPX mode).
	 * The sequenceCode indicates the sequencing priority of the IPMP Tool instance at the given control point. The tool with the highest sequenceCode for a given control point on a given stream shall process data first for that control point for that stream. Two tools shall not have the same sequence number at the same control point for the same stream.
	 * @return the sequence code
	 */
	byte getSequenceCode() const;

	/** 
	 * Get the control point code (only meaningful in IPMPX mode).
	 * @return the control point code.
	 * @see setIPMPXType
	 */
	byte getControlPointCode() const;

	/** 
	 * Get the IPMP Tool ID (only meaningful in IPMPX mode).
	 * @return a pointer to the 128-bit tool id.
	 */
	byte * getToolID() const;

	/** 
	 * Get the IPMP Descriptor ID, regardless of the descriptor type.
	 * @return the ID of this IPMP descriptor.
	 */
	unsigned short getIPMP_DescriptorID() const;
	
	/** 
	 * Get the IPMPS_Type (only meaningful in HOOKS mode). 
	 * This is a 16-bits unique ID of the IPMP System 
	 * ("hooks" version), assigned by a registration authority.
	 * @return the IPMPS_Type
	 */
	unsigned short getIPMPS_Type( ) const ;

	/**
	 * Get the IPMPX messages (only meaningful in IPMPX mode).
	 * @param index the index of the message in the internal array.
	 * @return the object containing a message derived from IPMP_Data_BaseClass.
	 */
	IPMP_Data_BaseClass* getIPMPXMessage( unsigned int index ) const;

	/**
	 * Get the opaque IPMP data (only meaningful in HOOKS mode).
	 * @return a pointer to the opaque IPMP data (private data).
	 */
	byte * getIPMP_data() const;

	IPMP_Descriptor & operator=(const IPMP_Descriptor &rval);		///< assignment operator

};


END_MOSES_NAMESPACE

#endif
