
//	This software module was originally developed by Massimo Balestri
//	(Telecom Italia Lab S.p.A.), and edited by Claudio Alberti (EPFL) 
//  in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) 
//  standard. This software module is an implementation of a part of one 
//  or more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified by the 
//  MPEG-4 Systems (ISO/IEC 14496-1) standard.
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free
//	license to this software module or modifications thereof for use in hardware
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC
//	14496-1).
//
//	Those intending to use this software module in hardware or software
//	products are advised that its use may infringe existing patents.
//
//	The original developer of this software module and his/her company, the
//	subsequent editors and their companies, and ISO/IEC have no liability for
//	use of this software module or modifications thereof.
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming
//	products.
//
//	Telecom Italia Lab S.p.A retains full right to use the code for its own purpose,
//	assign or donate the code to a third party and to inhibit third parties from
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
//
//	This copyright notice must be included in all copies or derivative works.
//
//	Copyright (c) 2000.



#ifndef	__IPMP_Message_H__
#define	__IPMP_Message_H__

#include "ExpandableBaseClass.h"
#include "Container.h"
#include "namespace.h"

BEGIN_MOSES_NAMESPACE

class IPMP_Data_BaseClass;			// forward 


/**
 *  @class IPMP_Message
 *  Class to convey time-varying IPMP information for associated IPMP Tool instances.
 *  @author Massimo Balestri
 *  @date 2002
 */
class IPMP_Message : public ExpandableBaseClass
{
public:
	enum MessageTypeEnum {
		HOOKS_TYPE		= 0,		///< indicates that this message contains "hooks" data
		IPMPX_TYPE		= 1,		///< indicates that this message contains IPMPX data
		URL_TYPE		= 2			///< indicates that this message contains a URL to another IPMP_Message
	};

private:
	MessageTypeEnum		messageType;					// identifies the type of this IPMP message
	unsigned int		IPMP_DescriptorID;
	unsigned int		IPMPS_Type ;
	unsigned int		sizeOfData ;				// contains the number of IPMPX messages in IPMP_ExtendedData
	char *				URLString ;
	byte *				IPMP_data ;
	Container<IPMP_Data_BaseClass*> IPMPX_data ;

	void init() ;		// local initialization 
	void cleanUp() ;	// free local allocated memory

public:
	IPMP_Message();									///< constructor
	IPMP_Message(const IPMP_Message & message);		///< copy constructor
	virtual ~IPMP_Message();						///< destructor
	virtual int toBuffer(byte * buf);				// reimplemented
	virtual int fromBuffer(const byte * buf);		// reimplemented
	virtual unsigned int sizeOfInstance() const;	// reimplemented	
	
	/**
	 * Get the type of IPMP message.
	 * @return an enum indicating the type. Can be IPMPX_TYPE, HOOKS_TYPE, URL_TYPE.
	 */
	MessageTypeEnum getType() const;
	
	/**
	 * Set this IPMP message as an URL type container and initialize it with a URL.
	 * @param URL contains a UTF-8 encoded URL that shall point to the location of a remote IPMP_Message.
	 * @return true if successful.
	 */
	bool setUrlType(const char *URL);

	/**
	 * Set this IPMP message as an IPMP Extension (IPMPX) message.
	 * @param descrId one IPMP_DescriptorID's in the scope of service of this IPMP stream; it identifies the recipient(s) of the message. Values of 0x0000 and 0xFFFF are forbidden.
	 * @return true if successful.
	 */
	bool setIPMPXType(unsigned short descrId);
	
	/**
	 * Set this descriptor as an "hooks" type descriptor, i.e. compliant to the old version of
	 * the MPEG-4 System specification.
	 * @param IPMPS_Type the ID of the IPMP System ("hooks" version), assigned by a registration authority
	 * @param IPMP_data opaque data of unspecified format
	 * @param size lenght in bytes of the IPMP_data array
	 * @return true if successful.
	 * @deprecated This method is provided for backward compatibility only
	 */
	bool setHooksType(unsigned short IPMPS_Type, byte * IPMP_data, int size);

	/**
	 * Add an IPMPX message to the list of messages. 
	 * This method can only be called when creating IPMPX type messages.
	 * @param msg the message to add
	 * @return true if successful
	 */
	bool addIPMPXMessage( IPMP_Data_BaseClass &msg );
	
	/** 
	 * Get the URL contained in a URL type message.
	 * @return a pointer to the URL string (or NULL if the message does not contain an URL).
	 */
	char * getURLString() const;

	/** 
	 * Get the size of internal data. This has a different meaning depending on the message type.
	 * For IPMPX messages, this is the number of IPMPX messages contained in the internal array.
	 * For URL messages, this is the size (in bytes) of the URL string.
	 * For HOOKS messages, this is the size (in bytes) of the opaque data.
	 * @return the size of internal data.
	 */
	int getSizeOfData() const;

	/** 
	 * Get the IPMP Descriptor ID of the recipientIPMP Tool (only meaningful in IPMPX messages).
	 * @return the ID of the recipient IPMP Tool.
	 */
	unsigned short getIPMP_DescriptorID() const;
	
	/** 
	 * Get the IPMPS_Type (only meaningful in HOOKS mode). 
	 * This is a 16-bits unique ID of the IPMP System 
	 * ("hooks" version), assigned by a registration authority.
	 * @return the IPMPS_Type
	 */
	unsigned short getIPMPS_Type( ) const ;

	/**
	 * Get the IPMPX messages (only meaningful in IPMPX mode).
	 * @param index the index of the message in the internal array.
	 * @return the object containing a message derived from IPMP_Data_BaseClass.
	 */
	IPMP_Data_BaseClass* getIPMPXMessage( unsigned int index ) const;

	/**
	 * Get the opaque IPMP data (only meaningful in HOOKS mode).
	 * @return a pointer to the opaque IPMP data (private data).
	 */
	byte * getIPMP_data() const;
};


END_MOSES_NAMESPACE

#endif


