
//	This software module was originally developed by Claudio Alberti (EPFL)	
//	in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) standard. 
//	This software module is an implementation of a part of one or more MPEG-4 
//	Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 
//	14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	EPFL retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2002.


// IPMP_SecureContainer.cpp

#include "IPMP_SecureContainer.h"
#include <memory.h>
#include <string.h>
#include <assert.h>

USING_MOSES_NAMESPACE

IPMP_SecureContainer::IPMP_SecureContainer():
	IPMP_Data_BaseClass(IPMP_Secure_Container_tag)
{
	hasEncryption = false;
	hasMAC = false;
	isMACEncrypted = false;
	protectedMsg = NULL;
}	

IPMP_SecureContainer::IPMP_SecureContainer(const IPMP_SecureContainer & sc)			///< copy constructor
:	IPMP_Data_BaseClass(sc)
{
	hasEncryption = false;
	hasMAC = false;
	isMACEncrypted = false;
	protectedMsg = NULL;

	if (sc.hasEncryptedData())
	{
		setEncryptedData( sc.getEncryptedData(), (sc.hasMac() && sc.isMacEncrypted()));
		if (sc.hasMac() && ! sc.isMacEncrypted())
			setMAC( sc.getMAC());
	}
	else	// cleartext
	{
		setProtectedMsg( sc.getProtectedMsg() );
		if (sc.hasMac())
			setMAC( sc.getMAC());
	}
}


IPMP_SecureContainer::~IPMP_SecureContainer() 
{
	if(protectedMsg)
		delete protectedMsg;

	protectedMsg = NULL;
}

int IPMP_SecureContainer::fromBuffer(const byte * buf)
{
	IPMP_Data_BaseClass::fromBuffer(buf);
	hasEncryption = (1 == get(1));		// 1 bit
	hasMAC = (1 == get(1));				// 1 bit
	isMACEncrypted = (1 == get(1));		// 1 bit 
	skipReading(5);						// 5 reserved bits

	if(hasEncryption)
	{
		skipReading( encryptedData.fromBuffer(getReadPtr()));

		if(hasMAC && (!isMACEncrypted))
		{
			skipReading( MAC.fromBuffer(getReadPtr()));
		}
	}
	else
	{
		unsigned int nbit;
		protectedMsg = createMessage(getReadPtr(), nbit);
		skipReading( nbit );
		
		if(hasMAC)
		{
			skipReading( MAC.fromBuffer(getReadPtr()));
		}
	}
	return (getReadCounter());
}


int IPMP_SecureContainer::toBuffer(byte * buf)
{
	IPMP_Data_BaseClass::toBuffer(buf);
	put(hasEncryption, 1); 
	put(hasMAC, 1);
	put(isMACEncrypted, 1);
	put(0, 5);							// 5 reserved bits 

	if(hasEncryption)
	{
		skipWriting( encryptedData.toBuffer(getWritePtr()));

		if(hasMAC && (!isMACEncrypted))
			skipWriting( MAC.toBuffer(getWritePtr()));
	}
	else{
		skipWriting( protectedMsg->toBuffer(getWritePtr()));

		if(hasMAC)
			skipWriting( MAC.toBuffer(getWritePtr()));
	}

	return (getWriteCounter());
}


unsigned int IPMP_SecureContainer::sizeOfInstance() const
{

	unsigned int size = IPMP_Data_BaseClass::sizeOfInstance();		// size of the base class 

	size++;

	if(hasEncryption)
	{
		size += encryptedData.getSize();

		if(hasMAC && (!isMACEncrypted))
		{
			size += MAC.getSize();
		}
	}
	else{

		size += protectedMsg->getSize();

		if(hasMAC)
		{
			size += MAC.getSize();
		}
	}
	return size;									// size of the ByteArray
}

void IPMP_SecureContainer::setMAC(const IPMP_ByteArray &mac)
{
	MAC = mac;					// use the assignment operator
	isMACEncrypted = false;		// since it is assigned here, it is not encrypted (if it was encrypted, would be in encryptedData)
	hasMAC = true;				// here is the MAC!
}

void IPMP_SecureContainer::setEncryptedData(const IPMP_ByteArray &data, bool includesMAC)
{
	encryptedData = data;		// use the assignment operator
	hasEncryption = true;		// the whole data chunk is encrypted
	hasMAC = includesMAC;		// encrypted MAC 
	isMACEncrypted = includesMAC;
}


void IPMP_SecureContainer::setProtectedMsg(IPMP_Data_BaseClass &msg)
{
	if(protectedMsg)
		delete protectedMsg;

	protectedMsg = createMessage(msg);			// copy the message
	hasEncryption = false;						// the message is in cleartext
}

const IPMP_ByteArray & IPMP_SecureContainer::getEncryptedData() const
{
	return encryptedData;
}

bool IPMP_SecureContainer::hasMac() const
{
	return hasMAC;
}

const IPMP_ByteArray & IPMP_SecureContainer::getMAC() const
{
	return MAC;
}

IPMP_Data_BaseClass & IPMP_SecureContainer::getProtectedMsg() const
{
	return *protectedMsg;
}

bool IPMP_SecureContainer::isMacEncrypted() const
{
	return isMACEncrypted;
}

bool IPMP_SecureContainer::hasEncryptedData() const
{
	return hasEncryption;
}
