
//	This software module was originally developed by Claudio Alberti (EPFL)	
//	in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) standard. 
//	This software module is an implementation of a part of one or more MPEG-4 
//	Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 
//	14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	EPFL retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2002.


// IPMP_SecureContainer.h


#ifndef	IPMP_SECURECONTAINER_H
#define	IPMP_SECURECONTAINER_H

#include "IPMP_Data_BaseClass.h"
#include "IPMP_ByteArray.h"
#include "namespace.h"

BEGIN_MOSES_NAMESPACE

/** @class IPMP_SecureContainer
 *	This message forms a secure container for any message extending the IPMP_Data_BaseClass. 
 *  The use of this message is optional and is defined for transmitting any message in a secure 
 *  manner when needed by the application. The payload message is optionally encrypted. 
 *  The MAC allows verification of the integrity of the protectedMsg, and is optionally encrypted.
 *  The key and algorithm for creating and verifying the MAC, and decrypting the encrypted payload
 *  is not explicitly carried in the secure message, but is negotiated as part of an external 
 *  authentication mechanism.
 *  @author Claudio Alberti
 *  @date 2002
 */
class IPMP_SecureContainer: public IPMP_Data_BaseClass 
{
private:

	bool hasEncryption;
	bool hasMAC;
	bool isMACEncrypted;
	IPMP_ByteArray encryptedData;
	IPMP_ByteArray MAC;
	IPMP_Data_BaseClass *protectedMsg;

public:
	IPMP_SecureContainer();							///< constructor
	IPMP_SecureContainer(const IPMP_SecureContainer & sc);			///< copy constructor
	virtual ~IPMP_SecureContainer();				///< destructor
	virtual int toBuffer(byte * buf);				// reimplemented
	virtual int fromBuffer(const byte * buf);		// reimplemented
	virtual unsigned int sizeOfInstance() const;	// reimplemented

	/**
	 * Set the encrypted message.
	 * In this case the payload message is encrypted and can contain an encrypted MAC.
	 * @param data the ecrypted message (extending the IPMP_Data_BaseClass)
	 * @param includesMAC indicates if the encrypted data includes a MAC or not.
	 */
	void setEncryptedData(const IPMP_ByteArray &data, bool includesMAC);

	/**
	 * Set a cleartext MAC for the payload message.
	 * @param mac the Message Authentication Code that can be used to verify the message.
	 */
	void setMAC(const IPMP_ByteArray & mac);
	
	/**
	 * Set the cleartext message. 
	 * Although in cleartext, the message can be protected by a Message Authentication Code.
	 * @param msg the message to be stored in the secure container.
	 */
	void setProtectedMsg(IPMP_Data_BaseClass & msg);

	/**
	 * Get the encrypted message.
	 * In this case the payload message is encrypted and can contain an encrypted MAC.
	 * @return the ecrypted message (extending the IPMP_Data_BaseClass)
	 */
	const IPMP_ByteArray & getEncryptedData() const;

	/**
	 * Indicates if the object contains a MAC (either cleartext of encrypted).
	 * @return true if a MAC is present.
	 */
	bool hasMac() const;
	
	/**
	 * Indicates if the message is encrypted.
	 * @return true if the message is encrypted.
	 */
	bool hasEncryptedData() const;

	/**
	 * Indicates if the MAC is encrypted. 
	 * If this is the case, and the MAC is present, it can be found at the end of the encrypted data.
	 * Instead, if the MAC is present and it is not encrypted, it can be read using getMAC().
	 * @return true if the MAC is encrypted.
	 */
	bool isMacEncrypted() const;

	/**
	 * Get a cleartext MAC for the payload message.
	 * @return the Message Authentication Code that can be used to verify the message.
	 */
	const IPMP_ByteArray & getMAC() const;
	
	/**
	 * Get the cleartext message. 
	 * Although in cleartext, the message can be protected by a Message Authentication Code.
	 * @return the message retrieved from the secure container.
	 */
	IPMP_Data_BaseClass & getProtectedMsg() const;
};

END_MOSES_NAMESPACE

#endif

