
//	This software module was originally developed by Claudio Alberti (EPFL)	and
//  modified by Massimo Balestri (Telecom Italia Lab)
//	in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) standard. 
//	This software module is an implementation of a part of one or more MPEG-4 
//	Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 
//	14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	EPFL retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2002.


// IPMP_SelectiveDecryptionInit.cpp

#include "IPMP_SelectiveDecryptionInit.h"
#include <memory.h>
#include <string.h>
#include <assert.h>
#include "namespace.h"

USING_MOSES_NAMESPACE

IPMP_SelectiveDecryptionInit::IPMP_SelectiveDecryptionInit():
	IPMP_Data_BaseClass(IPMP_SelectiveDecryptionInit_tag)
{
	// set defaults
	mediaTypeExtension = 0x00;						// ISO/IEC
	mediaTypeIndication = MPEG1_AUDIO;				// suppose is MPEG-1 Audio
	profileLevelIndication = 0x05;					// Layer III
	compliance = 0;									// fully compliant
	isContentSpecificFlag = false;
	nSegments = 0;
	RLE_Data = NULL;
}

IPMP_SelectiveDecryptionInit::IPMP_SelectiveDecryptionInit(const IPMP_SelectiveDecryptionInit & sd)	
:	IPMP_Data_BaseClass(sd),
	bufInfoStruct(sd.bufInfoStruct),
	fieldStruct(sd.fieldStruct)
{
	mediaTypeExtension = sd.mediaTypeExtension;
	mediaTypeIndication = sd.mediaTypeIndication;
	profileLevelIndication = sd.profileLevelIndication;
	compliance = sd.compliance;
	nSegments = 0;
	RLE_Data = NULL;
	if (sd.isContentSpecific())
	{
		isContentSpecificFlag = true;
		// fieldStruc is already copied; see above.
	}
	else
	{
		bool b = setDecryptionSegments(sd.getDecryptionSegments(), sd.getDecryptionSegmentsNum());
		assert(b);	
	}
}

IPMP_SelectiveDecryptionInit::~IPMP_SelectiveDecryptionInit() 
{
	if (RLE_Data != NULL)
		delete[] RLE_Data;					// free memory!
}

int IPMP_SelectiveDecryptionInit::fromBuffer(const byte * buf)
{
	IPMP_Data_BaseClass::fromBuffer(buf);
	mediaTypeExtension = (byte) get(8);
	mediaTypeIndication = (byte) get(8);
	profileLevelIndication = (byte) get(8);
	compliance = (byte) get(8);
	unsigned int numBufs = get(8);
	for (unsigned int i = 0; i < numBufs; i++)
	{
		BufInfoStruct myBufInfo;
		skipReading(myBufInfo.fromBuffer(getReadPtr()));
		bufInfoStruct.addObject(myBufInfo);
	}

	isContentSpecificFlag = (get(1) == 1);
	skipReading(7);					// reserved

	if (isContentSpecificFlag)
	{
		unsigned int numFields = get(8);
		for(unsigned int i=0; i < numFields; i++)
		{
			FieldStruct myField;
			skipReading(myField.fromBuffer(getReadPtr()));
			fieldStruct.addObject(myField);
		}
	}
	else
	{
		nSegments = get(16);
		if (RLE_Data != NULL)
			delete[] RLE_Data;					// free memory before allocating new one!
		RLE_Data = new unsigned short[nSegments];		// allocate memory
		for (word32 k=0; k<nSegments; k++)
		{
			RLE_Data[k] = (short) get(16);
		}
	}
	return getReadCounter();
}



int IPMP_SelectiveDecryptionInit::toBuffer(byte * buf)
{
	IPMP_Data_BaseClass::toBuffer(buf);
	put(mediaTypeExtension, 8);
	put( mediaTypeIndication, 8);
	put(profileLevelIndication, 8);
	put(compliance, 8);
	unsigned int numBufs = bufInfoStruct.getObjectNum();
	put(numBufs, 8);
	for (unsigned int i = 0; i < numBufs; i++)
	{
		skipWriting(bufInfoStruct.getObject(i).toBuffer(getWritePtr()));
	}

	put(isContentSpecificFlag, 1);
	put(0, 7);					// reserved

	if (isContentSpecificFlag)
	{
		unsigned int numFields = fieldStruct.getObjectNum();
		put(numFields, 8);
		for(unsigned int i=0; i < numFields; i++)
		{
			skipWriting(fieldStruct.getObject(i).toBuffer(getWritePtr()));
		}
	}
	else
	{
		put(nSegments, 16);
		for (word32 k=0; k<nSegments; k++)
		{
			put(RLE_Data[k], 16);
		}
	}
	return getWriteCounter();
}


// get the size, in bytes!
unsigned int IPMP_SelectiveDecryptionInit::sizeOfInstance() const
{
	unsigned int size = IPMP_Data_BaseClass::sizeOfInstance();		// size of the base class

	size += 5;
	for (unsigned int i=0; i < bufInfoStruct.getObjectNum(); i++)
		size += bufInfoStruct.getObject(i).getSize();
	
	size+=1;		// isContentSpecific + reserved

	if (isContentSpecificFlag)
	{
		size += 1;
		for (unsigned int i=0; i < fieldStruct.getObjectNum(); i++)
			size += fieldStruct.getObject(i).getSize();
	}
	else
		size += 2 + 2*nSegments;		// nSegments + RLE_data
	
	return size;
}


FieldStruct & IPMP_SelectiveDecryptionInit::addContentSpecificField()
{
	isContentSpecificFlag = true;
	if (RLE_Data != NULL)
	{
		delete[] RLE_Data;
		RLE_Data = NULL;
	}
	
	// create a new object into the list and return it
	FieldStruct field;
	int index = fieldStruct.addObject(field);
	return fieldStruct.getObject(index);		// return the object so that the caller can modify it
}

bool IPMP_SelectiveDecryptionInit::setDecryptionSegments(const unsigned short *rleData, word32 numsegments)
{
	isContentSpecificFlag = false;				// this is content-independent!
	nSegments = numsegments;

	if (RLE_Data != NULL)
		delete[] RLE_Data;						// free memory before allocating new one!

	RLE_Data = new unsigned short[nSegments];		// allocate memory

	if (RLE_Data != NULL)
	{
		memcpy(RLE_Data, rleData, sizeof(short) * numsegments);
		return true;
	}
	else
		return false;
}

word32 IPMP_SelectiveDecryptionInit::getDecryptionSegmentsNum() const
{
	if (isContentSpecificFlag)
		return 0;
	return nSegments;
}

const unsigned short * IPMP_SelectiveDecryptionInit::getDecryptionSegments() const
{
	if (isContentSpecificFlag)
		return NULL;
	return RLE_Data;
}


void IPMP_SelectiveDecryptionInit::setMediaTypeExtension(enum MediaTypeExtensionTable mte)
{
	mediaTypeExtension = (byte) mte;
}

enum IPMP_SelectiveDecryptionInit::MediaTypeExtensionTable IPMP_SelectiveDecryptionInit::getMediaTypeExtension() const
{
	return (enum MediaTypeExtensionTable) mediaTypeExtension;
}

void IPMP_SelectiveDecryptionInit::setMediaTypeIndication(enum MediaTypeIndicationTable mti)
{
	mediaTypeIndication = (byte) mti;
}

enum IPMP_SelectiveDecryptionInit::MediaTypeIndicationTable IPMP_SelectiveDecryptionInit::getMediaTypeIndication() const
{
	return (enum MediaTypeIndicationTable) mediaTypeIndication;
}

void IPMP_SelectiveDecryptionInit::setProfileLevelIndication(byte pli)
{
	profileLevelIndication = pli;
}

byte IPMP_SelectiveDecryptionInit::getProfileLevelIndication() const
{
	return profileLevelIndication;
}

void IPMP_SelectiveDecryptionInit::setCompliance(enum ComplianceTable compl)
{
	compliance = (byte) compl;
}

enum IPMP_SelectiveDecryptionInit::ComplianceTable IPMP_SelectiveDecryptionInit::getCompliance() const
{
	return (enum ComplianceTable) compliance;
}

bool IPMP_SelectiveDecryptionInit::isContentSpecific() const
{
	return isContentSpecificFlag;
}



///////////////////////////// BufInfoStruct ////////////////////////////////////
BufInfoStruct::BufInfoStruct()
{}

BufInfoStruct::~BufInfoStruct()
{}

void BufInfoStruct::setBlockCipher(const byte * _cipher_id, enum SyncBoundaryTable _syncBoundary, enum ModeTable _mode, unsigned short _blockSize, unsigned short _keySize)
{
	isBlock = true;							// this is a block cipher!
	memcpy(cipher_id, _cipher_id, 16);		// copy the ID
	syncBoundary = _syncBoundary;			// boundary
	mode = _mode;							// mode
	blockSize = _blockSize;					// block size
	keySize = _keySize;						// key size
}

void BufInfoStruct::setStreamCipher(const byte * _cipher_id, enum SyncBoundaryTable _syncBoundary, const IPMP_ByteArray & streamCipherInit)
{
	isBlock = false;									// this is a stream cipher!
	memcpy(cipher_id, _cipher_id, 16);					// copy the ID
	syncBoundary = _syncBoundary;						// boundary
	streamCipherSpecificInitInfo = streamCipherInit;	// copy the init info.
}

bool BufInfoStruct::isBlockCipher() const
{
	return isBlock;
}

const byte * BufInfoStruct::getCipher_id() const
{
	return cipher_id;
}

enum BufInfoStruct::SyncBoundaryTable BufInfoStruct::getSyncBoundary() const
{
	return (enum SyncBoundaryTable) syncBoundary;
}

enum BufInfoStruct::ModeTable BufInfoStruct::getMode() const
{
	return (enum ModeTable) mode;
}

unsigned short BufInfoStruct::getBlockSize() const
{
	return (unsigned short) blockSize;
}

unsigned short BufInfoStruct::getKeySize() const
{
	return (unsigned short) keySize;
}

// reimplemented methods
int BufInfoStruct::fromBuffer(const byte* buf)	
{
	setReadBuffer(buf);
	getBuffer(cipher_id, 128);
	syncBoundary = get(8);
	isBlock = (get(1) == 1);
	skipReading(7);				// reserved
	if (isBlock)
	{
		mode = get(8);
		blockSize = get(16);
		keySize = get(16);
	}
	else
	{
		skipReading(streamCipherSpecificInitInfo.fromBuffer(getReadPtr()));
	}
	return getReadCounter();
}

int BufInfoStruct::toBuffer(byte* buf)
{	
	setWriteBuffer(buf);
	putBuffer(cipher_id, 128);
	put(syncBoundary, 8);
	put(isBlock, 1);
	put(0, 7);					// reserved
	if (isBlock)
	{
		put(mode, 8);
		put(blockSize, 16);
		put(keySize, 16);
	}
	else
	{
		skipWriting(streamCipherSpecificInitInfo.toBuffer(getWritePtr()));
	}
	return getWriteCounter();
}

bool BufInfoStruct::checkSyntax() const
{
	return true;
}

// get the size in bytes!
unsigned int BufInfoStruct::getSize() const
{
	unsigned int size = 16 + 2;  //	id + syncBoundary + isBlock
	if (isBlock)
		size += 5;
	else
		size += streamCipherSpecificInitInfo.getSize();

	return size;
}

//////////////////////////// FieldStruct ////////////////////////////////////
FieldStruct::FieldStruct()
{
	buf				= 0;
	field_Id		= 0;
	field_Scope		= 0;
	isMapped		= false;
	isShuffled		= false;
	mappingTable	= NULL;
	sendMapTable	= false;
	sizeMapTable	= 0;
}

FieldStruct::FieldStruct(const FieldStruct &fs)
{
	buf				= fs.buf;
	field_Id		= fs.field_Id;
	field_Scope		= fs.field_Scope;
	isMapped		= fs.isMapped;
	isShuffled		= fs.isShuffled;
	sendMapTable	= fs.sendMapTable;
	sizeMapTable	= fs.sizeMapTable;
	mappingTable	= NULL;
	
	if (fs.hasMapTable())
	{
		setMappingTable( fs.getMappingTable(), fs.getMappingTableSize() );
	}
	
	if (fs.hasShuffleTable())
	{
		setShuffleTable( fs.getShuffleTable() );
	}
}

FieldStruct::~FieldStruct()
{
	if (mappingTable != NULL)
		delete[] mappingTable;
}

// reimplemented methods
int FieldStruct::fromBuffer(const byte* _buf)	
{
	setReadBuffer(_buf);
	field_Id = (byte) get(8);
	field_Scope = (byte) get(3);
	skipReading(5);					// reserved
	buf = (byte) get(8);
	isMapped = (get(1) == 1);
	skipReading(7);					// reserved
	if (isMapped)
	{
		sendMapTable = (get(1) == 1);
		isShuffled = (get(1) == 1);
		skipReading(6);					// reserved
		if (sendMapTable)
		{
			sizeMapTable = (unsigned short) get(16);
			if (mappingTable != NULL)
				delete[] mappingTable;
			mappingTable = new unsigned short [sizeMapTable];
			for (unsigned int i = 0; i< sizeMapTable; i++)
				mappingTable[i] = (unsigned short) get(16);
		}
		if (isShuffled)
			skipReading(shuffleSpecificInfo.fromBuffer(getReadPtr()));
	}
	return getReadCounter();
}

int FieldStruct::toBuffer(byte* _buf)
{	
	setWriteBuffer(_buf);
	put(field_Id,8);
	put(field_Scope, 3);
	put(0,5);					// reserved
	put(buf, 8);
	put(isMapped, 1);
	put(0, 7);					// reserved
	if (isMapped)
	{
		put(sendMapTable, 1);
		put(isShuffled, 1);
		put(0, 6);					// reserved
		if (sendMapTable)
		{
			put(sizeMapTable, 16);
			for (unsigned int i = 0; i< sizeMapTable; i++)
				put(mappingTable[i], 16);
		}
		if (isShuffled)
			skipWriting(shuffleSpecificInfo.toBuffer(getWritePtr()));
	}
	return getWriteCounter();
}

bool FieldStruct::checkSyntax() const
{
	return true;		// tbd
}

// get the size in bytes!
unsigned int FieldStruct::getSize() const
{
	unsigned int size = 4;  //	id + scope + buf + isMapped
	if (isMapped)
	{
		size += 1;

		if (sendMapTable)
			size += 2 + 2*sizeMapTable;
		
		if (isShuffled)
			size += shuffleSpecificInfo.getSize();
	}

	return size;
}


void FieldStruct::setMappingTable(const unsigned short *mapTable, unsigned short size)
{
	isMapped = true;				// because this method is setting the map....
	sendMapTable = true;				// because there is no reason to set it without sending it...

	if (mappingTable != NULL)
		delete[] mappingTable;

	mappingTable = new unsigned short[ size ];
	memcpy(mappingTable, mapTable, sizeof(unsigned short) * size);
	sizeMapTable = size;
}

const unsigned short * FieldStruct::getMappingTable() const
{
	return mappingTable;
}

bool FieldStruct::hasMapTable() const
{
	return sendMapTable;
}

unsigned short FieldStruct::getMappingTableSize() const
{
	return sizeMapTable;
}

bool FieldStruct::hasShuffleTable() const
{
	return isShuffled;
}

const IPMP_ByteArray & FieldStruct::getShuffleTable() const
{
	return shuffleSpecificInfo;
}

void FieldStruct::setShuffleTable(const IPMP_ByteArray &table)
{
	isShuffled = true;
	isMapped = true;
	shuffleSpecificInfo = table;
}
