//	This software module was originally developed by Claudio Alberti (EPFL)	
//  and modified by Massimo Balestri (TILAB)
//	in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) standard. 
//	This software module is an implementation of a part of one or more MPEG-4 
//	Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 
//	14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	EPFL retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2002.

// IPMP_SelectiveDecryptionInit.h

#ifndef	IPMP_SELECTIVEDECRYPTIONINIT_H
#define	IPMP_SELECTIVEDECRYPTIONINIT_H

#include "IPMP_Data_BaseClass.h"
#include "Container.h"
#include "IPMP_ByteArray.h"
#include "namespace.h"
#include "BitField.h"	// Added by ClassView

BEGIN_MOSES_NAMESPACE

class BufInfoStruct;
class FieldStruct;

/** @class IPMP_SelectiveDecryptionInit
 *  This message initialize a decryptor Tool.
 *  The selective decryption configuration data is used to communicate to the decryptor, 
 *	how the encryption of the received content bitstream is encrypted, whether all bits 
 *	are encrypted or only portions of the received bits are encrypted. 
 *	In the case when only portions of the received bits are encrypted, what portions of 
 *	the received are encrypted and therefore need to be decrypted.
 *
 *  @author Claudio Alberti, Massimo Balestri
 *  @date 2002
 */
class IPMP_SelectiveDecryptionInit: public IPMP_Data_BaseClass 
{
private:
	byte mediaTypeExtension;
	byte mediaTypeIndication;
	byte profileLevelIndication;
	byte compliance;

	bool isContentSpecificFlag;
	word32 nSegments;
	unsigned short *RLE_Data;
	Container<FieldStruct> fieldStruct;			///< information needed for each selective decryption operation.

public:
	Container<BufInfoStruct> bufInfoStruct;		///< information needed for each decryption buffer.

	/**
	 * Defines the acceptable values for the mediaTypeExtension parameter.
	 */
	enum MediaTypeExtensionTable {
		ISO_IEC				= 0x00,		///< registration authority is ISO/IEC
		ITU					= 0x01		///< registration authority is ITU
	};

	/**
	 * Defines the acceptable values for the mediaTypeIndication parameter,
	 * when the mediaType extension is ISO_IEC.
	 */
	enum MediaTypeIndicationTable {
		MPEG4_AUDIO			= 0x40,			///< MPEG-4 audio
		MPEG2_AUDIO_MAIN	= 0x66,			///< MPEG-2 audio
		MPEG2_AUDIO_LOW		= 0x67,			///< MPEG-2 audio
		MPEG2_AUDIO_SCAL	= 0x68,			///< MPEG-2 audio
		MPEG2_AUDIO			= 0x69,			///< MPEG-2 audio
		MPEG1_AUDIO			= 0x6B,			///< MPEG-1 audio
		MPEG4_VIDEO			= 0x20,			///< MPEG-4 video
		MPEG2_VIDEO_SIMPLE	= 0x60,			///< MPEG-2 video
		MPEG2_VIDEO_MAIN	= 0x61,			///< MPEG-2 video
		MPEG2_VIDEO_SNR 	= 0x62,			///< MPEG-2 video
		MPEG2_VIDEO_SPATIAL	= 0x63,			///< MPEG-2 video
		MPEG2_VIDEO_HIGH	= 0x64,			///< MPEG-2 video
		MPEG2_VIDEO_422		= 0x65,			///< MPEG-2 video
		MPEG1_VIDEO			= 0x6A,			///< MPEG-1 video
		JPEG				= 0x6C			///< JPEG still pictures
	};

	/**
	 * Defines the acceptable values for the "compliance" parameter.
	 */
	enum ComplianceTable {
		FULLY_COMPLIANT		= 0x00,		///< Fully compliant.
		VIDEO_PACKET		= 0x01,		///< Compliant to video packet level for video.
		VOP					= 0x02,		///< Compliant to VOP level for video.
		VIDEO_NOT_COMPLIANT = 0x03,		///< Non-compliant for video.
		GOB					= 0x04,		///< Compliant to GOB level for H.263 baseline
		AAC_DATA_FRAME		= 0x30,		///< Compliant to data frame level for AAC audio
		AAC_NOT_COMPLIANT	= 0x31		///< Non-compliant for AAC audio.
	};


	IPMP_SelectiveDecryptionInit();				///< constructor.
	IPMP_SelectiveDecryptionInit(const IPMP_SelectiveDecryptionInit & sd);	///< copy constructor.
	virtual ~IPMP_SelectiveDecryptionInit();	///< destructor
	
	virtual int toBuffer(byte * buf);			// reimplemented methods
	virtual int fromBuffer(const byte * buf);
	virtual unsigned int sizeOfInstance() const;

	/**
	 * Get the mediaTypeExtension, which indicates the registration authority
	 * assigning values to the mediaTypeIndication table.
	 * @return the autority assigning identifiers for media types.
	 */
	enum MediaTypeExtensionTable getMediaTypeExtension() const;
	
	/**
	 * Set the mediaTypeExtension, which indicates the registration authority
	 * assigning values to the mediaTypeIndication table.
	 * @param mte the autority assigning identifiers for media types.
	 */
	void setMediaTypeExtension(enum MediaTypeExtensionTable mte);
	
	/**
	 * Get the mediaTypeIndication, which indicates the format of the data stream
	 * (for example, MPEG-2 Audio).
	 * @return the media type identifier.
	 */
	enum MediaTypeIndicationTable getMediaTypeIndication() const;

	/**
	 * Set the mediaTypeIndication, which indicates the format of the data stream
	 * (for example, MPEG-2 Audio).
	 * @param mti the media type identifier.
	 */
	void setMediaTypeIndication(enum MediaTypeIndicationTable mti);

	/**
	 * Get the media-specific profile and level indication.
	 * @return a value which identifies the profile and level of the media.
	 */
	byte getProfileLevelIndication() const;
	
	/**
	 * Set the media-specific profile and level indication.
	 * @return a value which identifies the profile and level of the media.
	 */
	void setProfileLevelIndication(byte pli);

	/**
	 * Set the level of compliance to the complession format, i.e. the smallest logical 
	 * unit in the encrypted bitstream that is correctly parsable by the media decoder.
	 * @param compliance the compliance level.
	 */
	void setCompliance(enum ComplianceTable compliance);

	/**
	 * Get the level of compliance to the complession format, i.e. the smallest logical 
	 * unit in the encrypted bitstream that is correctly parsable by the media decoder.
	 * @return the compliance level.
	 */
	enum ComplianceTable getCompliance() const;

	/**
	 * Indicates if the object instance contains content specific decryption information.
	 * @return true if content specific information is present.
	 */
	bool isContentSpecific() const;

	/**
	 * Add a content specific field to the list contained in this object (multiple calls of this method create a list of fields).
	 * As a side effect, a call to this method removes any information stored in decryption segments, 
	 * since the two informations are mutually exclusive.
	 * @returns a field to be filled with content specific selective decryption configuration information.
	 */
	FieldStruct & addContentSpecificField();

	/**
	 * Set the decryption segments (this is used in case of content-independent decryption).
	 * @param rleData the number of bits that are to be decrypted or skipped interleavingly,
	 * starting from the first segment that is to be decrypted. The value of the first entry can be zero.
	 * @param nSegments the number of segments 
	 * @return TRUE if successful
	 */
	bool setDecryptionSegments(const unsigned short * rleData, word32 nSegments);	
	
	/**
	 * Get the number of decryption segments.
	 * @return the number of segments (in case of content-independent decryption)
	 */
	word32 getDecryptionSegmentsNum() const;

	/**
	 * Get the decryption segments (this is used in case of content-independent decryption).
	 * @return rleData the number of bits that are to be decrypted or skipped interleavingly,
	 * starting from the first segment that is to be decrypted. The value of the first entry can be zero.
	 */
	const unsigned short * getDecryptionSegments() const;
};

/** 
 * @class BufInfoStruct
 * A class holding information needed for each decryption buffer.
 * @see IPMP_SelectiveDecryptionInit
 * @author Claudio Alberti, Massimo Balestri
 * @date 2002
 */
class BufInfoStruct : public BitField
{
private:
	byte cipher_id[16];
	word32 syncBoundary;
	bool isBlock;
	word32 mode;  
	word32 blockSize;
	word32 keySize;

	IPMP_ByteArray streamCipherSpecificInitInfo; 

public:
	BufInfoStruct();									///< constructor.
	virtual ~BufInfoStruct();							///< destructor.

	virtual int fromBuffer(const byte* buf);	// reimplemented methods
	virtual int toBuffer(byte* buf);
	virtual bool checkSyntax() const;
	virtual unsigned int getSize() const;

	/**
	 * Table defining acceptable values for the syncBoundary field.
	 */
	enum SyncBoundaryTable {
		VIDEO_PACKET			= 0x00,		///< sync at video packet level
		VIDEO_AU				= 0x01,		///< sync at vide access unit level
		GOV						= 0x02,		///< sync at GOV level
		AAC_DATA_FRAME			= 0x30		///< sync at AAC audio data frame level
	};

	/**
	 * Table defining modes of block cipher operations.
	 * These should be registered by a registration anthority 
	 * (current values have been assigned by MOSES).
	 */
	enum ModeTable {
		ECB	= 0x00,				///< Electronic Codebook mode
		CBC	= 0x01,				///< Cipher Block Chaining mode
		CTR	= 0x02,				///< Counter mode
		CFB	= 0x03,				///< Cipher Feedback mode
		OFB	= 0x04				///< Output Feedback mode
	};

	/**
	 * Specify the decryption to be performed on this buffer, using a block cipher.
	 * @param cipher_id the type of decryption algorithm used, e.g. DES, 3DES, AES, registered by a Registration Authority.
	 * @param syncBoundary information on what shall be considered a unit of cipher-text (bits from different cipher-text units are decrypted separately).
	 * @param mode mode of operation of the block cipher, e.g. CBC, ECB, CTR, etc.
	 * @param blockSize the block size, in bytes, used for the block cipher.
	 * @param keySize the key size, in bytes, used for the block cipher.
	 */
	void setBlockCipher(
		const byte * cipher_id, 
		enum SyncBoundaryTable syncBoundary, 
		enum ModeTable mode, 
		unsigned short blockSize, 
		unsigned short keySize);
	
	/**
	 * Specify the decryption to be performed on this buffer, using a stream cipher.
	 * @param cipher_id the type of decryption algorithm used, registered by a Registration Authority.
	 * @param syncBoundary information on what shall be considered a unit of cipher-text (bits from different cipher-text units are decrypted separately).
	 * @param streamCipherInit the initialization information for this stream cipher.
	 */
	void setStreamCipher(
		const byte * cipher_id, 
		enum SyncBoundaryTable syncBoundary, 
		const IPMP_ByteArray & streamCipherInit);
	
	/**
	 * Indicates if the cipher is a block cipher or a stream cipher.
	 * @return true if the cipher is a block cipher.
	 */
	bool isBlockCipher() const;

	/**
	 * Get the 128-bit ID of the algorithm that must be used to decrypt this buffer.
	 * @return the ID.
	 */
	const byte * getCipher_id() const;

	/**
	 * Get the sync boundary for this buffer.
	 * @return information on what shall be considered a unit of cipher-text (bits from different cipher-text units are decrypted separately).
	 */
	enum SyncBoundaryTable getSyncBoundary() const;

	/**
	 * Get the mode of operation of the block cipher.
	 * @return the mode (e.g. ECB)
	 */
	enum ModeTable getMode() const;

	/**
	 * Get the block size.
	 * @return the block size, in bytes, used for the block cipher.
	 */
	unsigned short getBlockSize() const;

	/**
	 * Get the key size.
	 * @return the key size, in bytes, used for the block cipher.
	 */
	unsigned short getKeySize() const;
};

/**
 * @class FieldStruct
 * A class holding content specific information for selective decryption.
 * @see IPMP_SelectiveDecryptionInit
 * @author Claudio Alberti, Massimo Balestri
 * @date 2002
 * 
 */
class FieldStruct : public BitField
{
private:
	bool isMapped;							///< a value of 0b1 indicates that the codewords of a specific field will be mapped, using a mapping table, to an index which is then subject to decryption.		
	bool isShuffled;						///< a value of 0b1 indicates the mapped index sequence will be shuffled using a shuffling table specified in Shuffle_Specific_Info.		
	bool sendMapTable;						///< a value of 0b1 indicates that the mapping table mappingTable will follow. Otherwise, the default mapping of the codeword table defined in the media format definition (e.g. MPEG4 video specification) is used.
	unsigned short *mappingTable;			///< entries of the mapping table. MappingTable[i] indicates that the ith codeword in the codeword table defined in the media format definition is mapped to the index of MappingTable[i].
	unsigned short sizeMapTable;			///< size, number of entries of the mapping table.
	IPMP_ByteArray shuffleSpecificInfo;		///< the shuffling table.

public:
	
	/**
	 * Table of acceptable values for field_id.
	 */
	enum FieldIdTable {
		MOTION_VECTOR		= 0x00,			///< Motion vector for video
		DC_COEFFICIENT		= 0x01,			///< DC coefficient for video
		DCT_SIGN			= 0x02,			///< DCT sign bit for video
		QUANTIZATION		= 0x03,			///< Quantization parameter Dquant for video
		DCT_COEFFICIENT		= 0x04,			///< DCT coefficients for video
		ALL					= 0x05,			///< All bits in a unit of cipher text.
		AAC_SIGN			= 0x30,			///< Sign bits for AAC audio
		AAC_RUNLENGTH		= 0x31,			///< Run-length codewords for AAC audio
		AAC_SCALE			= 0x32			///< Scale factors for AAC audio
	};

	byte field_Id;							///< index of the field based on a predefined syntax (see FieldIdTable).
	byte field_Scope;						///< represented in three bits, i.e., b2b1b0. A value of 0b1 for b2, b1, and b0 indicates that this field in I, P, and B VOPs, respectively, is selected and put into the buffer that it is associated with.
	byte buf;								///< which buffer this field will be put into. 		
	
	// ----- methods ----- //
	
	FieldStruct();								///< constructor
	FieldStruct(const FieldStruct & fs);		///< copy constructor
	virtual ~FieldStruct();						///< destructor
	virtual int fromBuffer(const byte* buf);	// reimplemented methods
	virtual int toBuffer(byte* buf);
	virtual bool checkSyntax() const;
	virtual unsigned int getSize() const;
	
	/**
	 * Get the number of entries of the mapping table.
	 * @return the number of entries.
	 */
	unsigned short getMappingTableSize() const;

	/**
	 * Indicates if a mapping table is present.
	 * @return true if the mapping table is present.
	 */
	bool hasMapTable() const;
		
	/**
	 * Get the mapping table.
	 * MappingTable[i] indicates that the ith codeword in the codeword table 
	 * defined in the media format definition is mapped to the index of MappingTable[i].
	 * @return the mapping table.
	 */
	const unsigned short * getMappingTable() const;

	/**
	 * Set the mapping table.
	 * MappingTable[i] indicates that the ith codeword in the codeword table 
	 * defined in the media format definition is mapped to the index of MappingTable[i].
	 * @param mapTable the mapping table.
	 * @param size the number of entries in the mapping table.
	 */
	void setMappingTable(const unsigned short * mapTable, unsigned short size);
	
	/**
	 * Set the shuffle table.
	 * @param table the shuffle table.
	 */
	void setShuffleTable(const IPMP_ByteArray & table);
	
	/**
	 * Get the shuffle table.
	 * @return table the shuffle table.
	 */
	const IPMP_ByteArray & getShuffleTable() const;

	/**
	 * Indicates if a shuffle table is present.
	 * @return true if a shuffle is present. 
	 */
	bool hasShuffleTable() const;
};

END_MOSES_NAMESPACE

#endif


	