
//	This software module was originally developed by Massimo Balestri (TILAB)
//	in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) standard. 
//	This software module is an implementation of a part of one or more MPEG-4 
//	Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 
//	14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	Telecom Italia Lab S.p.A. retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2002.

// IPMP_Tool.h

#ifndef	IPMP_TOOL_H
#define	IPMP_TOOL_H
#include "namespace.h"
#include "BaseDescriptor.h"
#include "IPMP_ParametricDescription.h"
#include "IPMP_ByteArray.h"
#include "IPMP_ToolId.h"

BEGIN_MOSES_NAMESPACE

/** 
 *  @class IPMP_Tool
 *  This class identifies one IPMP Tool that is required by the Terminal to consume the content.
 *  The IPMP Toolcan be specified either as a unique implementation, as one of a list of alternatives, or through
 *  a parametric description.
 *  
 *  @author Massimo Balestri
 *  @date 2002
 */
class IPMP_Tool: public BaseDescriptor 
{
private:
	unsigned int m_isAltGroup ;						// flags
	unsigned int m_isParametric ;

	IPMP_ParametricDescription toolParamDesc;		// the parametric description of this tool
	Container <IPMP_ToolId> specificToolID;			// a list of specific alternative IPMP tools that can be used in place of this tool. 

	void init() ;									// local initialization 

public:
	IPMP_ToolId  toolID;							///< the 128 bit id of the Tool described by this descriptor.
	Container <IPMP_ByteArray>  toolURL ;			///< a list of informative URLs from which the IPMP tool may be obtained.

	/**
	 * constructor.
	 */
	IPMP_Tool();	

	/**
	 * destructor.
	 */
	virtual ~IPMP_Tool();		

	virtual int toBuffer(byte * buf);				// reimplemented
	virtual int fromBuffer(const byte * buf);		// reimplemented
	virtual unsigned int sizeOfInstance() const;	// reimplemented

	/**
	 * Add an ID to the list of alternative tool IDs.
	 * @param toolId the 128 bit id of this alternate tool
	 */
	void addSpecificToolID(const IPMP_ToolId & toolId);

	/**
	 * get the Tool ID of one of the alternative tools.
	 * @param index the index of this tool in the list of alternate tools
	 * @return the 128 bit tool ID
	 */
	IPMP_ToolId & getSpecificToolID(unsigned int index) const;
	
	/**
	 * get the flag indicating if alternative tools exist for this tool.
	 * @return the value of the flag. TRUE indicates that alternative tools exist.
	 */
	bool isAltGroup() const;

	/**
	 * get the flag indicating if the tool is a parametric tool.
	 * @return the value of the flag. TRUE indicates that the tool is parametric.
	 */
	bool isParametric() const;

	/**
	 * get the number of alternate tools.
	 * @return the number of alternate tools. 
	 */
	unsigned int getNumAlternates() const;

	/**
	 * get the number of URLs associated to this tool.
	 * @return the number of URLs 
	 */
	unsigned int getNumURLs() const;

	/**
	 * set the parametric description associated to this tool.
	 * @param descr the parametric description
	 */
	void setToolParamDesc(const IPMP_ParametricDescription & descr );

	/**
	 * get the parametric description associated to this tool.
	 * @return the parametric description
	 */
	const IPMP_ParametricDescription & getToolParamDesc() const;

	/**
	 * get the size in bytes of the parametric description associated to this tool.
	 * @return the length of the parametric description
	 */
	unsigned int getToolParamDescSize() const;

	/**
	 * Add one URL to the list of URLs.
	 * This is one informative URL from which the tool specified in the 
	 * message may be obtained.
	 * @param url the URL
	 * @param size the URL length in bytes
	 */
	void addToolURL(byte *url, int size);

	/**
	 * get one of the informative URLs from which the tool specified in the 
	 * message may be obtained.
	 * @param index the position of this URL in the list of URLs
	 * @return a pointer to the URL
	 */
	byte * getToolURL(unsigned int index) const;
	
	/**
	 * get the size of one of the informative URLs from which the tool specified in the 
	 * message may be obtained.
	 * @param index the position of this URL in the list of URLs
	 * @return the URL length in bytes
	 */
	unsigned int getToolURLsize(unsigned int index) const;
};


END_MOSES_NAMESPACE

#endif
