
//	This software module was originally developed by Claudio Alberti (EPFL)	
//  and modified by Massimo Balestri (Telecom Italia Lab)
//	in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) standard. 
//	This software module is an implementation of a part of one or more MPEG-4 
//	Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 
//	14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	EPFL retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2002.

// IPMP_ToolNotificationListener.h

#ifndef	IPMP_TOOLNOTIFICATIONLISTENER_H
#define	IPMP_TOOLNOTIFICATIONLISTENER_H

#include "IPMP_Data_BaseClass.h"
#include "namespace.h"


BEGIN_MOSES_NAMESPACE

/**
 * @class Constants
 * Class containing constants for other classes.
 * This avoids the problem of conflicts with other constants definitions.
 * @author Massimo Balestri
 * @date 2003
 */
class Constants
{
public:
	enum EventType {
		EVENT_CONNECTED				= 0x00,		///< notify when connected successfully
		EVENT_CONNECTIONFAILED		= 0x01,		///< notify when failed to connect
		EVENT_DISCONNECTED			= 0x02,		///< notify when disconnected successfully
		EVENT_DISCONNECTIONFAILED	= 0x03,		///< notify when disconnection failed
		EVENT_WATERMARKDETECTED		= 0x04,		///< notify when watermarking is detected
		EVENT_USER_DEFINED_BASE		= 0xC1		///< base for user-defined values
	};

	enum ScopeType {
		SCOPE_ESD	= 0x00,	///< limit notification to tools protecting the same elementary stream
		SCOPE_OD	= 0x01,	///< limit notification to tools protecting the same OD
		SCOPE_LOCAL = 0x02	///< limit notification to all tools in the same name scope
	};
};

/** 
 *  @class IPMP_AddToolNotificationListener
 *  This message is sent from a Tool to the Terminal to request notification of certain events.
 *	There are five event types defined by the standard, plus a number of user defined events.
 *  In any case the total number cannot exceed 256.
 *  Please note that the default scope is SCOPE_LOCAL, this can be changed with setScope()
 *  @author Claudio Alberti, Massimo Balestri
 *  @date 2002
 */
class IPMP_AddToolNotificationListener: public IPMP_Data_BaseClass 
{
private:
	unsigned int eventTypeCount;
	byte eventArray[256];					// there cannot be more than 256 event types!
	Constants::ScopeType scope;				// there is a single scope

public:
	IPMP_AddToolNotificationListener();				///< constructor.
	virtual ~IPMP_AddToolNotificationListener();	///< destructor.
	
	// default copy constructor is OK for this class
	
	virtual int toBuffer(byte * buf);				// reimplemented	
	virtual int fromBuffer(const byte * buf);		// reimplemented	
	virtual unsigned int sizeOfInstance() const;	// reimplemented

	/**
	 * get the current scope
	 * @return the current scope
	 */
	Constants::ScopeType getScope() const;
	
	/**
	 * set the scope according to the given ScopeType.
	 * @param scope the new scope 
	 */
	void setScope(Constants::ScopeType scope);

	/**
	 * get the i-th event type
	 * @param et the returned event type
	 * @param index the index of the required event type
	 * @return TRUE is successful, FALSE otherwise
	 */
	bool getEventType(Constants::EventType & et, unsigned int index) const;

	/**
	 * add an event type to the list of required events.
	 * if the event type is already in the list, no action is performed and the method returns false.
	 * @param et the event type to add
	 * @return true if successfull
	 */
	bool addEventType(Constants::EventType et);

	/**
	 * get the number of event types currently stored
	 * @return the number of different event types
	 */
	unsigned int getEventTypeCount() const;
};

/** 
 *  @class IPMP_RemoveToolNotificationListener
 *  This message is sent from a Tool to the Terminal to request stop notification of certain events
 *	There are five event types defined by the standard, plus a number of user defined events.
 *  In any case the total number cannot exceed 256.
 *  @author Massimo Balestri
 *  @date 2002
 */
class IPMP_RemoveToolNotificationListener: public IPMP_Data_BaseClass 
{
private:
	unsigned int eventTypeCount;
	byte eventArray[256];		// there cannot be more than 256 event types!

public:
	IPMP_RemoveToolNotificationListener();				///< constructor.
	virtual ~IPMP_RemoveToolNotificationListener();		///< destructor.
	virtual int toBuffer(byte * buf);					// reimplemented
	virtual int fromBuffer(const byte * buf);			// reimplemented
	virtual unsigned int sizeOfInstance() const ;		// reimplemented

	/**
	 * get the i-th event type
	 * @param et the returned event type
	 * @param index the index of the required event type
	 * @return TRUE is successful, FALSE otherwise
	 */
	bool getEventType(Constants::EventType & et, unsigned int index) const;

	/**
	 * add an event type to the list of required events.
	 * if the event type is already in the list, no action is performed and the method returns false.
	 * @param et the event type to add
	 * @return true if successfull
	 */
	bool addEventType(Constants::EventType et);

	/**
	 * get the number of event types currently stored
	 * @return the number of different event types
	 */
	unsigned int getEventTypeCount() const;
};


/** 
 *  @class IPMP_NotifyToolEvent
 *  This message notifies an IPMP Tool of an event for which it had previous registered as a listener.
 *	There are five event types defined by the MPEG-4 System standard, plus a number of user defined events.
 *  @author Massimo Balestri
 *  @date 2002
 */
class IPMP_NotifyToolEvent: public IPMP_Data_BaseClass 
{
private:
	bool m_hasID;
	unsigned int m_OD_ID;
	unsigned int m_ESD_ID;
	Constants::EventType m_eventType;
	unsigned int m_toolContextID;

public:
	IPMP_NotifyToolEvent();						///< constructor.
	virtual ~IPMP_NotifyToolEvent();			///< destructor
	virtual int toBuffer(byte * buf);				// reimplemented
	virtual int fromBuffer(const byte * buf);		// reimplemented
	virtual unsigned int sizeOfInstance() const ;	// reimplemented

	/**
	 * get the tool context ID
	 * @return the context ID of the tool either connected or disconnected, or the failed state of either, as indicated by the event type.
	 */
	unsigned int getToolContextID() const;

	/**
	 *  set the tool context ID
	 *  @param id the id of the tool that has been either connected or disconnected.
	 */
	void setToolContextID(unsigned int id);

	/**
	 *  get the scope under which the event occurred
     */
	bool getScope(unsigned int &OD_ID, unsigned int & ESD_ID) const;
	
	/**
	 * check if this message has the scope indication (OD_ID and ESD_ID)
	 */
	bool hasID() const;
	
	/**
	 * set the scope under which the event occurred
	 */
	void setScope(unsigned int OD_ID, unsigned int ESD_ID);

	/**
	 * get the event type
	 * @return the event type which has occurred
	 */
	 Constants::EventType getEventType() const;

	/**
	 * set the event type that must be signalled.
	 * @param et the event type to signal
	 */
	void setEventType(Constants::EventType et);
};

END_MOSES_NAMESPACE

#endif
