
//	This software module was originally developed by Claudio Alberti (EPFL)	
//  and modified by Massimo Balestri (Telecom Italia Lab)
//	in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) standard. 
//	This software module is an implementation of a part of one or more MPEG-4 
//	Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 
//	14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	EPFL retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2002.

// IPMP_VideoWatermarkingInit.h

#ifndef	IPMP_VIDEOWATERMARKINGINIT_H
#define	IPMP_VIDEOWATERMARKINGINIT_H

#include "IPMP_Data_BaseClass.h"
#include "namespace.h"

BEGIN_MOSES_NAMESPACE

/** 
 * @class IPMP_VideoWatermarkingInit
 * This message delivers to a watermarking tool all the information about 
 * the characteristics of the video content, the type of action to be performed on it, and 
 * possibly other related proprietary data required by the watermarking tool.
 *  
 * @author Claudio Alberti, Massimo Balestri
 * @date 2002
 */
class IPMP_VideoWatermarkingInit: public IPMP_Data_BaseClass 
{
private:
	word32 inputFormat;
	word32 requiredOp;
	bool b_hasOpaqueData;

	word32 frame_horizontal_size;
	word32 frame_vertical_size;
	word32 chroma_format;

	word32 wmPayloadLen;
	byte*  wmPayload;
	word32 wmRecipientId;
	word32 opaqueDataSize;
	byte*  opaqueData ;

	bool   setwmPayload(const byte* buf, word32 len);

public:
	/**
	 * Table of possible values of the chromaFormat parameter. 
	 * It indicates the ratio between luminance and chrominance samples.
	 */
	enum ChromaFormatTable {
		FORBIDDEN			= 0x00,		///< this value is forbidden
		CHROMA_420			= 0x01,		///< format is 4:2:0
		CHROMA_422			= 0x02,		///< format is 4:2:2
		CHROMA_444			= 0x03		///< format is 4:4:4 
	};

	/**
	 * Table of possible values of the inputFormat parameter. 
	 * This table will be maintained by a registration authority.
	 * The current values have been assigned by MOSES.
	 */
	enum InputFormatTable {
		YUV					= 0x01,		///< uncompressed YUV samples
		MPEG4_VIDEO			= 0x20,		///< compressed MPEG-4 video
		MPEG2_VIDEO_SIMPLE	= 0x60,		///< compressed MPEG-2 video, simple profile 
		MPEG2_VIDEO_MAIN	= 0x61,		///< compressed MPEG-2 video, main profile 
		MPEG2_VIDEO_SNR 	= 0x62,		///< compressed MPEG-2 video, SNR profile 
		MPEG2_VIDEO_SPATIAL	= 0x63,		///< compressed MPEG-2 video, spatially scalable profile 
		MPEG2_VIDEO_HIGH	= 0x64,		///< compressed MPEG-2 video, high profile 
		MPEG2_VIDEO_422		= 0x65,		///< compressed MPEG-2 video, 4:2:2 profile (studio)
		MPEG1_VIDEO			= 0x6A,		///< MPEG-1 video
		JPEG				= 0x6C		///< JPEG
	};
	
	/**
	 * Table of possible values of the requiredOp parameter. 
	 * It indicates the operation that the watermarking tool is required to perform on the
	 * audio stream.
	 */
	enum RequiredOpTable {
		INSERT_WM			= 0,		///< insert watermark 
		EXTRACT_WM			= 1,		///< extract watermark
		REMARK_WM			= 2,		///< remark watermark
		DETECT_COMPRESSION	= 3,		///< detect compression
		USER_DEFINED		= 11		///< first user defined entry 
	};

	IPMP_VideoWatermarkingInit();											///< constructor
	IPMP_VideoWatermarkingInit(const IPMP_VideoWatermarkingInit & msg);		///< copy constructor
	virtual ~IPMP_VideoWatermarkingInit();									///< destructor

	virtual int toBuffer(byte * buf);				// reimplemented methods		
	virtual int fromBuffer(const byte * buf);
	virtual unsigned int sizeOfInstance() const;

	/**
	 * Get the chrominance format (4:2:0, 4:2:2, etc.)
	 * @return the chrominance format id
	 */
	enum ChromaFormatTable getChromaFormat() const;

	/**
	 * Get the vertical size of a frame.
	 * @return the vertical size in pixels.
	 */
	word32 getVerticalSize() const;

	/**
	 * Get the horizontal size of a frame.
	 * @return the horizontal size in pixels.
	 */
	word32 getHorizontalSize() const;

	/**
	 * Get the required operation, which can be one of the values defined in RequiredOpTable.
	 * @return the operation code.
	 */
	RequiredOpTable getRequiredOp() const;
	
	/**
	 * Get the size of the watermark payload.
	 * @return the size in bytes of the payload. 
	 */
	word32 getwmPayloadLen() const;

	/**
	 * Get the watermark payload.
	 * @return a pointer to the watermark payload. The pointer is valid as long as the instance is not deleted. 
	 */
	byte*  getwmPayload() const;

	/**
	 * Get the destination tool identified by the IPMP_Descriptor ID, to
	 * which the watermarking payload and compression information must be delivered.
	 * A value of 0x00 indicates the terminal.
	 * @return the Tool descriptor ID of the destination Tool. 
	 */
	word32 getwmRecipientId() const;
	
	/**
	 * Get the size of opaque data.
	 * @return the opaque data size. 
	 */
	word32 getOpaqueDataSize() const;

	/**
	 * Get the opaque data.
	 * @return a pointer to the opaque data. The pointer is valid as long as the instance is not deleted. 
	 */
	byte*  getOpaqueData() const;

	/**
	 * Indicates if some opaque data is present.
	 * @return TRUE if some opaque data is available. 
	 */
	bool   hasOpaqueData() const;

	/**
	 * Set the Video input format.
	 * @param format the input format(can be YUV, MPEG1_Video, etc.)
	 * @param nchannels the number of Video channels (1=mono, 2=stereo...)
	 * @param bitpersample the number of bit per sample.
	 * @param frequency get the number of samples per second (in Hz, e.g. 44100) of the input stream.
	 * @return TRUE if successful.
	 */
	bool setInputFormat(enum InputFormatTable format, word32 nchannels = 0, word32 bitpersample = 0, word32 frequency = 0);

	/**
	 * require the watermarking tool to insert the specified data as watermarking payload.
	 * @param payload the watermark payload
	 * @param size the size in bytes of the watermark payload
	 * @return TRUE if successful.
	 */
	bool insertWatermark(const byte *payload, word32 size);

	/**
	 * require the watermarking tool to remark the specified data as watermarking payload.
	 * @param payload the watermark payload
	 * @param size the size in bytes of the watermark payload
	 * @return TRUE if successful.
	 */
	bool remarkWatermark(const byte *payload, word32 size);	

	/**
	 * require the watermarking tool to extract a watermark payload from the input Video stream.
	 * @param recipientId the destination tool identified by the IPMP_Tool_Descriptor_ID, to
	 * which the watermarking payload must be delivered.
	 * A value of 0x00 indicates the terminal.
	 * @return TRUE if successful.
	 */
	bool extractWatermark(word32 recipientId);

	/**
	 * require the watermarking tool to detect if the input Video stream has been compressed.
	 * @param recipientId the destination tool identified by the IPMP_Tool_Descriptor_ID, to
	 * which the compression information must be delivered.
	 * A value of 0x00 indicates the terminal.
	 * @return TRUE if successful.
	 */
	bool detectCompression(word32 recipientId);

	/**
	 * set the private data field of the message.
	 * @param buf the buffer containing private data
	 * @param len the length of the buffer
	 */
	bool setOpaqueData(const byte* buf, word32 len);

};

END_MOSES_NAMESPACE

#endif
