//	This software module was originally developed by Claudio Alberti (EPFL)	
//  and edited by Massimo Balestri (Telecom Italia Lab)
//	in the course of development of the MPEG-4 Systems (ISO/IEC 14496-1) standard. 
//	This software module is an implementation of a part of one or more MPEG-4 
//	Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 
//	14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	EPFL retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2002.

// TrustSecurityMetadata.h

#ifndef	TRUSTSECURITYMETADATA_H
#define	TRUSTSECURITYMETADATA_H

#include "IPMP_Data_BaseClass.h"
#include "IPMP_ByteArray.h"
#include "Container.h"
#include "namespace.h"

BEGIN_MOSES_NAMESPACE

typedef byte	ToolIdType[16];						///< 128 bit Tool id type
typedef byte	DateType[5];						///< 40 bit date type


/** 
 *  @class DateClass
 *  Contains the audit date of IPMP tool in question, in Universal Time, 
 *  Co-ordinated (UTC) and Modified Julian Date (MJD). This field is coded 
 *  as 16 bits giving the 16 least significant bits of MJD followed by 24 
 *  bits coded as 6 digits in 4-bit Binary Coded Decimal (BCD). 
 *  If the audit date is undefined all bits of the field are set to 1.
 *  @author Claudio Alberti
 *  @date September 2002
 */

class DateClass  : public BitField
{
private:

	DateType Date;

public:
	DateClass();
	~DateClass();		
	virtual int toBuffer(byte * buf);		
	virtual int fromBuffer(const byte * buf);
	virtual unsigned int getSize() const;
	virtual bool checkSyntax() const;
	
	/**
	 * method for setting the date
	 * @param buf buffer containing the 5 byte long new value for the date
	 */
	void setDate(const DateType & buf);

	/**
	 * method for getting the current value for the date
	 * @return the date in binary form
	 */
	const DateType & DateClass::getDate() const;
};


/**
 *  @class  TrustSpecification
 *  For each tool referenced in TrustSecurityMetadata, one or more
 *  trust specification can be provided.
 *  The trust metadata shall consist of  either:
 *  The date of audit, AuditDate by the trust auditor
 *  For each StartDate, the tool, shall be trusted to protect 
 *  against the specified AttackerProfile for the TrustedDuration minutes. 
 *  Attacker profile is defined as class 1, 2, or 3.
 *  
 *  @author Claudio Alberti
 *  @date September 2002
 */
class TrustSpecification : public BitField
{
private:

	bool			hasCCBasedTrustMetadata;
	DateClass		startDate;
	word32			attackerProfile;
	word32			trustedDuration;
	IPMP_ByteArray	CCTrustMetadata;

public:
	TrustSpecification();				///< constructor
	virtual ~TrustSpecification();		///< destructor

	virtual int toBuffer(byte * buf);		
	virtual int fromBuffer(const byte * buf);
	virtual unsigned int getSize() const;
	virtual bool checkSyntax() const;


	/**
	 * Get the starting date.
	 * @return a date object.
	 */
	const DateClass & getStartDate() const;
	
	/**
	 * Get the attacker's profile.
	 * @return the attacker profile.
	 */
	word32 getAttackerProfile() const;

	/**
	 * Get the duration of trust.
	 * @return a 4 byte value for the duration of trust.
	 */
	word32 getTrustedDuration() const;

	/**
	 * Set the Trust metadata values.
	 * The tool shall be trusted as of startDate , to protect against the specified AttackerProfile for the TrustedDuration minutes. 
	 * @param date the start date from which the tool is trusted.
	 * @param ap the attacker profile, defined as class I, II, or III.
	 * @param td the trusted duration.
	 */
	void setTrustMetadata(const DateType & date,  word32 ap,  word32 td);

	/**
	 * An alternative, though opaque, means of specifying the trust metadata using Common Criteria.
	 * @param data buffer containing the trust metadata
	 * @param size size of the buffer in bytes
	 */
	void setCCTrustMetadata(const byte *data, unsigned int size);

};

/** 
 *  @class  TrustedTool
 *  Contains the TrustSecurityMetadata information of one trusted Tool.
 *
 *  @author Claudio Alberti
 *  @date September 2002
 */
class TrustedTool : public BitField
{	
private:
	ToolIdType			toolID;
	DateClass			auditDate;

public:
	Container<TrustSpecification> trustSpecification;	///< the list of trust specifications

	TrustedTool();						///< constructor
	virtual ~TrustedTool();				///< destructor
	virtual int toBuffer(byte * buf);		
	virtual int fromBuffer(const byte * buf);
	virtual bool checkSyntax() const;
	virtual unsigned int getSize() const;

	/**
	 * Set the toolID value
	 * @param toolId buffer containing the 128 bit long tool ID
	 */
	void setToolID(const ToolIdType & toolId);
	
	/**
	 * Set the toolID value
	 * @return the 128 bit long tool ID
	 */
	const ToolIdType & getToolID() const;

	/**
	 * Set the audit date
	 * @param date buffer containing the date of the last audit
	 */
	void setAuditDate(const DateClass & date);

	/**
	 * Get the audit date
	 * @return the date of the last audit
	 */
	const DateClass & getAuditDate() const;

};

/** 
 *  @class IPMP_TrustSecurityMetadata
 *  Message carrying metadata for the verification of trust between two tools.
 *  Trust is established by out of band relationships between 
 *  the different entities involved in protecting and managing the content. 
 *  The trust metadata that results from such relationships needs to be made 
 *  available to permit static and dynamic verification of trust.  
 *  This data structure encapsulates such information.
 *  
 *  @author Claudio Alberti
 *  @date September 2002
 */
class IPMP_TrustSecurityMetadata : public IPMP_Data_BaseClass
{

public:
	Container<TrustedTool> trustedTool;			///< the list of trusted tools.

	IPMP_TrustSecurityMetadata();									///< constructor.
	IPMP_TrustSecurityMetadata(const IPMP_TrustSecurityMetadata & tsm);	///< copy constructor.
	virtual ~IPMP_TrustSecurityMetadata();							///< destructor
	virtual int toBuffer(byte * buf);		
	virtual int fromBuffer(const byte * buf);
	virtual unsigned int sizeOfInstance() const;
};

END_MOSES_NAMESPACE

#endif
