//	This software module was originally developed by Craig A. Schultz 
//	(AccessTicket Systems Inc.)
//
//	in the course of development of the MPEG-4 Systems 
//	(ISO/IEC 14496-1) standard. This software module is an implementation 
//	of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as specified 
//	by the MPEG-4 Systems (ISO/IEC 14496-1) standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	AccessTicket Systems Inc. retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2000. 

#include "ztl.h"
#include "sdl.h"
#include "ControlPointObject.h"

/****************************************************************************
 *  ControlPointObject::ControlPointObject()			  ref:ContolPointObject.h
 *
 *  Constructor
 */
ControlPointObject::ControlPointObject()
	: m_MyThread			(Run, this)
	, IPMP_ControlPointCode	(0)
	, m_pFirstTool			((IPMPX_Tool*)this)
	, m_bActive				(false)
{
}

/****************************************************************************
 *  ControlPointObject::~ControlPointObject()			  ref:ContolPointObject.h
 *
 *  Destructor
 */
ControlPointObject::~ControlPointObject()
{
	ASSERT (m_MyThread.GetThread () == NULL);
}

/*-----------------------------------------------------------------
Stop() closes the stream
-----------------------------------------------------------------*/
void
ControlPointObject:: Stop ()
{	accessLock.P();

	DisconnectAllTools();

	m_bActive = false;
	if (m_pInStream)
	{	m_pInStream -> Abort ();
		m_pInStream -> Terminate ();
	}
	if (m_pOutStream)
		m_pOutStream -> Close ();
	WaitForThread ();
	accessLock.V();
}

/*-----------------------------------------------------------------
Start() connects all the tools and then opens the streams and 
creates the stream processing thread.
-----------------------------------------------------------------*/
void ControlPointObject :: Start ()
{	accessLock.P();
	SetUpTools();
	ASSERT (m_MyThread.GetThread () == NULL);
	m_bActive = true;
	if (m_pInStream)
		m_pInStream -> Start ();
	if (m_pOutStream)
		m_pOutStream -> Open ();
	if (StartControlPointObjectThread () == -1)
		throw;
	accessLock.V();
}

/*-----------------------------------------------------------------
Run() retrieves units from the input stream, passes them to the first 
tool in the tool sequence, and dispatches the processed units to 
the output stream, till the input stream's end is exceeded.
-----------------------------------------------------------------*/
void ControlPointObject :: Run ()
{
	LPBYTE pInData;
	int nInSize;
	DWORD dwTime;
	BOOL bEOU;

	while ( m_bActive &&
			(pInData = (LPBYTE) m_pInStream -> FetchNext (nInSize, dwTime, bEOU, TRUE) 
			) != END_OF_STREAM )
	{
		if (m_bActive)
		{	accessLock.P();
			/*-------------------------------------------------------------
			Process the unit
			-------------------------------------------------------------*/
			m_pFirstTool->ProcessData( pInData, nInSize, dwTime );

			accessLock.V();
		}
		m_pInStream -> Release (TRUE);
	}
	m_pOutStream -> Close ();
}

bool
ControlPointObject :: AddTool ( IPMPX_Tool * ptrTool, SDLInt<8> sequenceID )
{	accessLock.P();
	
	SequencedTool * newTool = new SequencedTool;
	newTool->ptrTool = ptrTool;
	newTool->sequenceID = sequenceID;

	tools += newTool;

	accessLock.V();

	return true;
}

bool
ControlPointObject :: RemoveTool ( IPMPX_Tool * ptrTool )
{	accessLock.P();

	SequencedTool * thisEntry = NULL;
	SequencedTool * previousEntry = NULL;
	SequencedTool * postEntry = NULL;
	int				thisIndex (0);
	int				previousIndex (0);
	int				postIndex (0);

	GetToolFromPtr ( ptrTool, thisEntry, thisIndex );

	if ( m_bActive )
	{	// Re-connect the previous and post tools to each other.
		// Tool will be deleted by the message router.
		GetToolFromPtr ( thisEntry->ptrNextTool, postEntry, thisIndex );
		GetConnectedToolFromPtr ( ptrTool, previousEntry, previousIndex );

		if ( postEntry != NULL && previousEntry != NULL )
			previousEntry->ptrTool->SetOutput ( postEntry->ptrTool );
		else if ( previousEntry == NULL && postEntry != NULL )
			m_pFirstTool = postEntry->ptrTool;
		else if ( previousEntry != NULL && postEntry == NULL )
			postEntry->ptrTool->SetOutput ( (IPMPX_Tool*)this );
		else
			m_pFirstTool = (IPMPX_Tool*)this;
	}
	
	tools.DeleteFrom(thisIndex);

	// by Liao Xuan
	if (thisEntry!=NULL)
	{
		delete thisEntry;
		thisEntry =  NULL;
	}
	// by Liao Xuan

	accessLock.V();

	return false;
}

bool
ControlPointObject :: ProcessData ( LPBYTE	pInput
								  , int		nInputLength 
								  , DWORD	dwTime )
{	LPBYTE pOutData;

	pOutData = (LPBYTE) m_pOutStream -> Allocate (nInputLength, TRUE, FALSE);

	if (pOutData == END_OF_STREAM){
		m_bActive = false;
		accessLock.V();
		return false;
	}

	memcpy ( pOutData
		   , pInput
		   , nInputLength );

	m_pOutStream -> Dispatch (nInputLength, dwTime);

	accessLock.V();

	return true;
}

bool
ControlPointObject :: GetToolFromID ( const SDLInt<8> sequenceID
									, SequencedTool * &ptrEntry
									, int & index  )
{	for ( int i = tools.GetLength(); --i >= 0; )
	{
		if ( tools[i]->sequenceID == sequenceID )
		{	ptrEntry = tools[i];
			index = i;
			return true;
		}
	}

	ptrEntry = NULL;
	index = -1;

	return false;
}

bool
ControlPointObject :: GetToolFromPtr ( const IPMPX_Tool * ptrTool
									 , SequencedTool * &ptrEntry
									 , int & index  )
{	for ( int i = tools.GetLength(); --i >= 0; )
	{
		if ( tools[i]->ptrTool == ptrTool )
		{	ptrEntry = tools[i];
			index = i;
			return true;
		}
	}

	ptrEntry = NULL;
	index = -1;

	return false;
}

bool
ControlPointObject :: GetConnectedToolFromID( const SDLInt<8> sequenceID, SequencedTool * &ptrEntry, int & index )
{	for ( int i = tools.GetLength(); --i >= 0; )
	 	if ( tools[i]->sequenceID == sequenceID )
		 	for ( int j = tools.GetLength(); --j >= 0; )
			 	if ( tools[j]->ptrNextTool == tools[i]->ptrTool )
				{	ptrEntry = tools[j];
					index = i;
					return true;
				}

	ptrEntry = NULL;
	index = -1;

	return false;
}

bool
ControlPointObject :: GetConnectedToolFromPtr( const IPMPX_Tool * ptrTool, SequencedTool * &ptrEntry, int & index  )
{	for ( int i = tools.GetLength(); --i >= 0; )
		if ( tools[i]->ptrNextTool == ptrTool )
		{	index = i;
			ptrEntry = tools[i];
			return true;
		}

	ptrEntry = NULL;
	index = -1;

	return false;
}

void
ControlPointObject :: SetUpTools()
{	SequencedTool * lastEntry = NULL;
	SequencedTool * previousEntry = NULL;
	SDLInt<8> currentSequenceID;

	GetLastTool( lastEntry );
	if ( lastEntry != NULL )
	{	lastEntry->ptrNextTool = (IPMPX_Tool*)this;
		lastEntry->ptrTool->SetOutput((IPMPX_Tool*)this);
		currentSequenceID = lastEntry->sequenceID;
	}
	
	if ( lastEntry == NULL )
		return;

	while ( GetPreviousTool( currentSequenceID, previousEntry ) )
	{	currentSequenceID = previousEntry->sequenceID;
		previousEntry->ptrNextTool = lastEntry->ptrTool;
		previousEntry->ptrTool->SetOutput( lastEntry->ptrTool );
		lastEntry = previousEntry;
	}

	m_pFirstTool = lastEntry->ptrTool;
}

bool
ControlPointObject :: GetLastTool( SequencedTool * &ptrEntry )
{	SDLInt<8>	sequenceID (INT_MAX);

	for ( int i = tools.GetLength(); --i >= 0; )
		if ( tools[i]->sequenceID < sequenceID )
		{	sequenceID = tools[i]->sequenceID;
			ptrEntry = tools[i];
		}
	
	if ( sequenceID != INT_MAX )
		return true;

	return false;
}

bool
ControlPointObject :: GetPreviousTool( const int currentSequenceID, SequencedTool * &ptrEntry  )
{	
	SequencedTool * targetEntry = NULL;

	for ( int i = tools.GetLength(); --i >= 0; )
		if ( tools[i]->sequenceID > currentSequenceID )
		{	targetEntry = tools[i];
			for ( int j = tools.GetLength(); --j >= 0; )
				if ( tools[j]->sequenceID < targetEntry->sequenceID
				&& tools[j]->sequenceID > currentSequenceID )
					targetEntry = tools[j];
		}

	if ( targetEntry == NULL )
		return false;

	ptrEntry = targetEntry;

	return true;
}

void
ControlPointObject :: DisconnectAllTools()
{	for ( int i = tools.GetLength(); --i >= 0; )
	{	tools[i]->ptrTool->SetOutput(NULL);
		// by Liao Xuan
		if(tools[i]!=NULL)
		{
			delete tools[i];
			tools[i] = NULL;
		}
		// by Liao Xuan
		tools.DeleteFrom(i);
	}
}
