
//	This software module was originally developed by Massimo Balestri 
//	(Telecom Italia Lab S.p.A.) in the course of development 
//	of the MPEG-4 Systems (ISO/IEC 14496-1) standard. This software module 
//	is an implementation of a part of one or more MPEG-4 Systems (ISO/IEC 
//	14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
//	standard. 
//
//	ISO/IEC gives users of the  the MPEG-4 Systems (ISO/IEC 14496-1) free 
//	license to this software module or modifications thereof for use in hardware 
//	or software products claiming conformance to the MPEG-4 Systems (ISO/IEC 
//	14496-1). 
//
//	Those intending to use this software module in hardware or software 
//	products are advised that its use may infringe existing patents. 
//
//	The original developer of this software module and his/her company, the 
//	subsequent editors and their companies, and ISO/IEC have no liability for 
//	use of this software module or modifications thereof. 
//
//	Copyright is not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
//	products. 
//
//	Telecom Italia Lab S.p.A retains full right to use the code for its own purpose, 
//	assign or donate the code to a third party and to inhibit third parties from 
//	using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products.
// 
//	This copyright notice must be included in all copies or derivative works. 
//
//	Copyright (c) 2000. 


#ifndef __EXPANDABLEBASECLASS_H__
#define __EXPANDABLEBASECLASS_H__

#include "BitField.h"
#include "Length.h"
//#include "namespace.h"

//BEGIN_MOSES_NAMESPACE

/** 
 * @class ExpandableBaseClass
 * Class which encodes its own size in bytes explicitly.
 * This class implements the expandable base class as defined in the Syntactic 
 * Description Language of the MPEG-4 System specification (ISO/IEC 14496-1:2001).
 * Depending on whether the expandable class has an object_id (a tag) or not, either the protected constructor 
 * or the default constructor must be called, respectively.
 * @author Massimo Balestri.
 * @date 2001 -2002
*/
class ExpandableBaseClass : public BitField
{
private:
	Length			len;
	bool			syntaxOk;
	byte			object_id;				///< the object ID of this class (0x00 = reserved)

protected:

	/**
	 * Use this constructor to create an expandable base class with an object_id.
	 * This allows to derive various classes from this base class and be able to distinguish them.
	 * @param objectId the tag to be used, normally set by the derived class.
	 */
	ExpandableBaseClass(byte objectId);

	/**
	 * Use this constructor to create an expandable base class with no object_id (that is, no tag).
	 */
	ExpandableBaseClass();

	/**
	 * Compute the size (in bytes) of this object when serialized into a bitstream, according
	 * to the current content of the member variables. 
	 * This method must be overloaded by any derived class to indicate the correct size.
	 * This size does NOT include the object_id and the length of the ExpandableBaseClass. 
	 * @return the size in bytes. 
	 */
	virtual unsigned int sizeOfInstance() const = 0;	

	/**
	 * Set the object ID of this class. This id is used to identify the derived classes.
	 * @param tag the object Id.
	 */
	void setObjectId(byte tag);
	 
public:

	
	/**
	 * destructor.
	 */
	virtual ~ExpandableBaseClass();
	
	/**
	 * reads member variables from a buffer
	 * @param buf the buffer to read
	 * @return the number of bits read from the buffer
	 */
	virtual int fromBuffer(const byte* buf);
	
	/**
	 * this method stores member variables into a buffer.
	 * @param buf the buffer to write
	 * @return the number of bit stored into the buffer
	 */
	virtual int toBuffer(byte* buf);

	/**
	 * check the syntax of the message.
	 * @return true if successful.
	 */
	virtual bool checkSyntax() const;

	/**
	 * get the object ID of this class. This id is used to identify the derived classes.
	 * @return the object Id 
	 */
	byte getObjectId() const;
	
	/**
	 * compute the total size (in bytes) of this object when serialized into a bitstream, according
	 * to the current content of the member variables. The returned size includes the 
	 * the size of the length and object_id fields, which range is 1..5 depending on the length value
	 * and on the presence of the object_id.
	 * This method must not be reimplemented in derived classes.
	 * @return the total size in bytes
	 */
	unsigned int getSize() const;

	/**
	 * returns the size (in bytes) of this object when serialized, as read from the bitstream.
	 * This method is usually called after a fromBuffer() call to know the size of the payload of
	 * the expandable base class.
	 * This size does NOT include the object_id and the length of the ExpandableBaseClass itself. 
	 * @return the size in bytes. 
	 */
	unsigned int getSizeOfInstance() const;	

	/**
	 * compute and store the size (in bytes) of the class according to the content of the class.
	 * This size does NOT include the object_id and the length of the ExpandableBaseClass. 
	 */
	 void setSizeOfInstance();

};


//END_MOSES_NAMESPACE

#endif
