/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* The University of Rochester retains full right to use the code for */
/* their own purposes, assign or donate the code to a third party and */
/* to inhibit third parties from using the code for non MPEG-4 Visual */
/* (ISO/IEC 14496-2) conforming products.                             */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1996, 1997                                           */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         delaunay.cpp
   AUTHOR       Peter Van Beek, Thu Apr 30 18:16:46 1998
   DESCRIPTION  Contains interface to external 'triangle' Delaunay 
				triangulation algorithm from J. R. Shewchuk (Carnegie 
				Mellon University). Interface runs through external data 
				files and system program calls. 

   SYNOPSIS     Poly2dTess* triangulate (Pt2dList &mpoints, Poly2d &p, 
   					string fname, int cflag)

   ASSUMPTIONS  -  The external Delaunay triangulator is evoked using
   				   system calls. Proper working of this depends on
				   a Unix environment. Also the name of the external
				   program is hardwired into this code. 
                -  
*********************************************************************-*/

#include <iostream.h>
#include <fstream.h>
#include <time.h>
#include <math.h>
//#define  DEBUGGING
#include "my_portab.h"
#include "listClass.h"
#include "pointClass.h"
#include "polyClass.h"
#include "polyTessClass.h"
#include "stringClass.h"
#include "general.h"
#include "delaunay.h"


#define LINESIZE 200


			/* ===== global variables ===== */


// name of external triangulator program and data files extension
static string program  = "triangle -Q -I ";
static string node_ext = ".node";
static string poly_ext = ".poly";
static string ele_ext  = ".ele";
static string rmstr    = "/usr/bin/rm -f ";


			/* ===== local function declarations ===== */

static Poly2dTess* read_triangles (Pt2dList *mp, string fname);
static Pt2dList* read_triangle_points(string fname);
static int write_triangle_points(Pt2dList &mp, string fname);
static int write_triangle_edges(Poly2d &p, Pt2dList &mp, string fname);
static int write_triangle_holes(Poly2d &p, string fname);
static int next_valid_line(ifstream &ifile);
static void ignore_line(ifstream &ifile);


			/* ===== function definitions ===== */


/*+==========================================================================
	NAME		triangulate
	VERSION		1.0
	PURPOSE		Given a set of points, create a Delaunay triangulation
				whose vertices are the point set.
				Optionally, a constrained triangulation is computed. If so,
				a set of constraining segment edges must be given as well.
	METHOD		Call external program, use files to communicate data.
	INPUT		mpoints			list of points
				poly			set of constraining segment edges, stored
								in a Poly2d, which must be a closed 
								polygonal boundary
				fname			basename of data files
				cflag			signals constrained triangulation
	OUTPUT		
	RETURNS		Poly2dTess		the triangular mesh produced
	ERROR HANDLING
		Return	NOT_OK if:
				- files could not be opened for input/output
				- errors occurred during reading
				- triangulator program exited bad
	NOTES		Be sure that the input bounding polygon is closed! 
				Otherwise, your triangulation disappears.
				The vertices for each triangle are assumed to order in
				counter-clockwise order in the output of the external
				program, which assumes an ordinary x-y coordinate system. 
				Thus, they will be in clockwise order in our image 
				coordinate system (origin at top-left). 
				Assumes there are no four or more cocircular node points 
				that are mutually visible. One should make sure of this
				during mesh generation.
==========================================================================-*/
Poly2dTess* triangulate (Pt2dList &mpoints, Poly2d &poly, string fname, int cflag)
{
	Poly2dTess* mesh=NULL;				// mesh to be built
	Pt2dList *newpoints=NULL;			// point list after triangulation
	string command, poly_fname, node_fname, ele_fname;


	node_fname = fname+node_ext;
	poly_fname = fname+poly_ext;
	ele_fname  = fname+ele_ext;

	if (!cflag)							// not a constrained triangulation
	{
		// write the node points to a file
		if (write_triangle_points(mpoints, node_fname) != OK)
		{
			printf("triangulate: error writing node points\n");
			return(NULL);
		}
	}
	else								// constrained triangulation 
	{
		// write the node points to a file
		if (write_triangle_points(mpoints, poly_fname) != OK)
		{
			printf("triangulate: error writing node points\n");
			return(NULL);
		}
		// append constraining segment edges
		if (write_triangle_edges(poly, mpoints, poly_fname) != OK)
		{
			printf("triangulate: error writing edge segments\n");
			return(NULL);
		}
		// append holes outside of constraining polygons
		if (write_triangle_holes(poly, poly_fname) != OK)
		{
			printf("triangulate: error writing holes\n");
			return(NULL);
		}
	}
		
	// call the external delaunay triangulator
	if (!cflag)
		command = program+node_fname;
	else
		command = program+poly_fname;
	if (system(command))
	{
		printf("triangulate: external triangulator program failed\n");
		return(NULL);
	}

	// read the nodes from the triangular mesh from a file
	newpoints = (Pt2dList *) read_triangle_points(node_fname);
	if (newpoints == NULL)
	{
		printf("triangulate: error reading nodes\n");
		return(NULL);
	}

	// read the triangle topology of the triangular mesh from a file
	mesh = (Poly2dTess *) read_triangles (newpoints, ele_fname);
	if (mesh == NULL)
	{
		printf("triangulate: error reading triangles\n");
		return(NULL);
	}
	// do internal mesh pointer reordering
	mesh->reOrderOwners();
	// cleanup and return mesh
	delete newpoints;
	command = rmstr + node_fname;
	system(command);
	command = rmstr + ele_fname;
	system(command);
	if (cflag) {
		command = rmstr + poly_fname;
		system(command);
	}
	return(mesh);

} // triangulate


// Read the triangles from a file and, given the points, create the mesh
// Assumes a fixed ordering of vertices for each triangle is already
// present in the file data.
Poly2dTess* read_triangles (Pt2dList *mpoints, string fname)
{
	Poly2dTess *mesh=NULL;				// triangular mesh to be built
	int nt, i, cnt;						// # of triangles, counters
	ifstream ifile;						// input data file
	int  	i1, i2, i3;					// vertex indices
	Tri2d	tri;						// triangle


	// open file and alloc mesh
	ifile.open(fname, ios::in);
	if (!ifile)
	{
		printf ("read_triangles: cannot open file for reading\n");
		return(NULL);
	}
	mesh = new Poly2dTess();

	DBUG(("----- list of triangles ---------------\n"));

	// read number of triangles - ignore comment lines
	if (next_valid_line (ifile) != OK)
		return(NULL);
	ifile >> nt;						// read # of triangles
	ifile >> i1;						// read # vertices/triangle
	if (i1 != 3)
	{
		printf("read_triangles: # vertices per triangle != 3\n");
		mesh->destroy();
		return(NULL);
	}
	ignore_line(ifile);
	DBUG(("# of triangles: %d\n", nt));
	// read vertex indices for each triangle - ignore comment lines
	for (i=1; i<=nt; i++)
	{
		if (next_valid_line (ifile) != OK)
		{
			mesh->destroy();
			return(NULL);
		}
		ifile >> cnt >> i1 >> i2 >> i3;
		ignore_line(ifile);
		DBUG(("triangle: %d %d %d\n", i1, i2, i3));

		// get vertex points and add the triangle to mesh
		tri = Tri2d((*mpoints)[i1-1], (*mpoints)[i2-1], (*mpoints)[i3-1]);
		mesh->append(tri);
	}

	// close file and return list of points
	ifile.close();
	return(mesh);

} // read_triangles


// Read the list of points in the mesh from a file
Pt2dList* read_triangle_points(string fname)
{
	Pt2dList *plist=NULL;				// point list to be built
	int np, i, cnt;						// # of points, counters
	ifstream ifile;						// input data file
	double  x, y;						// point coordinates


	// open file and alloc point list
	ifile.open(fname, ios::in);
	if (!ifile)
	{
		printf("read_triangle_points: cannot open file for reading\n");
		return(NULL);
	}

	plist = new Pt2dList;

	DBUG(("----- list of points ---------------\n"));
	// read # of points - ignore comment lines
	if (next_valid_line (ifile) != OK)
	{
		plist->empty();
		return(NULL);
	}
	ifile >> np;
	ignore_line(ifile);
	DBUG(("# of points: %d\n", np));

	// read coordinates for each point - ignore comment lines
	for (i=1; i<=np; i++)
	{
		if (next_valid_line (ifile) != OK)
		{
			plist->empty();
			return(NULL);
		}
		ifile >> cnt >> x >> y;
		ignore_line(ifile);
		DBUG(("point: %i %lf %lf\n", cnt, x, y));

		// add point to list
		plist->append(Point2d(x, y));
	}

	// close file and return list of points
	ifile.close();
	return(plist);

} // read_triangle_points


// Go to the next valid line (not a comment or empty line)
int next_valid_line(ifstream &ifile)
{
	char c;

	ifile.get(c);						// read first char
	while (ifile.good() && (c=='#' || c=='\n'))
	{									// comment line or empty line
		ifile.ignore(LINESIZE, '\n');	// ignore rest of line
		ifile.get(c);					// read next char
	}
	if (ifile.good())
	{
		ifile.putback(c);				// put non-comment char back
		return(OK);
	}
	else
	{
		printf("read_triangle_points: reading failed\n");
		return(NOT_OK);
	}

} // next_valid_line


// Ignore the rest of the line
void ignore_line(ifstream &ifile)
{
	ifile.ignore(LINESIZE, '\n');		// ignore rest of line
} // ignore_line



// Write node point coordinates into a file
int write_triangle_points(Pt2dList &mpoints, string fname)
{
	unsigned int np, i;					// number of points
	fstream  ofile;						// output data file
	string   comments;
	time_t   timeval;
	Point2d  pt;

	// open file and write some comments
	ofile.open(fname, ios::out);
	if (!ofile)
	{
		printf ("write_triangle_points: cannot open file for writing\n");
		return(NOT_OK);
	}
	timeval = time(NULL);
	comments = "# Created ";
	comments += ctime(&timeval);
	ofile << comments;
	comments = "# Triangular mesh data file: node points";
	ofile << comments << endl << "#" << endl;

	// write number of points, point coordinates and close file
	np = mpoints.getSize();
	ofile << np << " 2 0 0 " << endl;
	for (i=1; i<=np; i++)
	{
		pt = mpoints[i-1];
		ofile << i << "  " << pt.X() << "  " << pt.Y() << endl;
	}
	ofile << "# " << endl;
	ofile.close();
	
	return(OK);

} // write_triangle_points


// Write edges of constraining polygon into a file
int write_triangle_edges(Poly2d &pol, Pt2dList &mpoints, string fname)
{
	unsigned int j, n, cnt;				// number of edges and counters
	fstream  ofile;						// output data file
	Point2d  pt1, pt2;
	int	     ind1, ind2;


	// open file and write some comments
	ofile.open(fname, ios::app);
	if (!ofile)
	{
		printf ("write_triangle_edges: cannot open file for writing\n");
		return(NOT_OK);
	}
	ofile << "# Triangular mesh data file: edge segments" << endl;
	ofile << "#" << endl;

	// get and write number of edges
	n = pol.getSize();					// closed polygonal poly2d
	ofile << n << " 0 " << endl;

	// write endpoint indices
	cnt = 0;
	for (j=0; j<n; j++)
	{
		pt1 = pol[j];
		pt2 = pol[j+1];
		ind1 = mpoints.where(pt1)+1;
		ind2 = mpoints.where(pt2)+1;
		if (ind1==0 || ind2==0)
		{
			printf ("write_triangle_edges: cannot find point in list!\n");
			return(NOT_OK);
		}
		cnt++;
		ofile << cnt << "  " << ind1 << "  " << ind2 << endl;
	}
	if (n != cnt)
	{
		printf ("write_triangle_edges: counting error\n");
		return(NOT_OK);
	}

	// close file and return
	ofile << "# " << endl;
	ofile.close();
	return(OK);

} // write_triangle_edges


// For writing holes into a file
// Currently writes no holes
int write_triangle_holes(Poly2d &p, string fname)
{
	fstream  ofile;						// output data file

	// open file and write some comments
	ofile.open(fname, ios::app);
	if (!ofile)
	{
		printf ("write_triangle_holes: cannot open file for writing\n");
		return(NOT_OK);
	}
	ofile << "# Triangular mesh data file: holes" << endl;
	ofile << "#" << endl;

	// currently no holes are written
	ofile << "0" << endl;
	
	// close file and return
	ofile << "# " << endl;
	ofile.close();
	return(OK);

} // write_triangle_holes



#undef LINESIZE // 200


// end of file
