/*************************************************************************

This software module was originally developed by 

    Ming-Chieh Lee (Microsoft Corporation)
    Wei-ge Chen (Microsoft Corporation)
    Bruce Lin (Microsoft Corporation)
    Chuang Gu (Microsoft Corporation)

and edited by

    Peter van Beek (University of Rochester)

in the course of development of the MPEG-4 Video (ISO/IEC 14496-2). 
This software module is an implementation of a part of one or more 
MPEG-4 Video (ISO/IEC 14496-2) tools as specified by MPEG-4 Video
(ISO/IEC 14496-2). ISO/IEC gives users of the MPEG-4 Video (ISO/IEC 
14496-2) free license to this software module or modifications thereof 
for use in hardware or software products claiming conformance to the 
MPEG-4 Video (ISO/IEC 14496-2). 
Those intending to use this software module in hardware or software 
products are advised that its use may infringe existing patents. The 
original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability 
for use of this software module or modifications thereof in an 
implementation. 
Copyright is not released for non MPEG-4 Video (ISO/IEC 14496-2) 
conforming products. Microsoft Corporation retains full right to use 
the code for his/her own purpose, assign or donate the code to a third 
party and to inhibit third parties from using the code for non MPEG 
Video (ISO/IEC 14496-2) conforming products. 
This copyright notice must be included in all copies or derivative works. 

Copyright (c) 1996, 1997.

Module Name:
	
	huffman.cpp

Abstract:

	Classes for Huffman encode/decode.

Revision History:

	Peter Van Beek, Fri Feb 28 12:52:53 1997
	- Removed unnecessary entropy encoders/decoders

	Peter Van Beek, Thu Oct 16 15:15:06 1997
	- Added attachStream methods to CEntropyEncoderSet and CEntropyDecoderSet

	Peter Van Beek, Thu May 14 16:46:41 1998
	- added new loadTable methods to CHuffmanDecoder and CHuffmanEncoder

*************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <search.h>
#include <iostream.h>
#include <fstream.h>
#include <math.h>

#include "basic.h"
#include "entropy.h"
#include "huffman.h"
#include "bitstrm.h"

// Peter Van Beek, Thu May 14 13:38:15 1998
// Definition of VLC tables
#include "vlcdef.h"


#ifdef __MFC_
#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

#define new DEBUG_NEW				   
#endif // __MFC_

class CNode
{
friend class CHuffmanTree;
friend Int huffmanNodeCompare (const Void* Element1,const Void* Element2);
private:
    Char m_cCode;
    Int m_lNodeIndex;
    Int m_lFrequency;
    Int m_lBalancer;
    CNode () {m_cCode = 0; m_lNodeIndex = -1; m_lFrequency = 0; m_lBalancer = 1;};
};

Class CHuffmanDecoderNode
{
friend class CHuffmanDecoder;
    Char m_c0End;
    Char m_c1End;
    Int m_l0NextNodeOrSymbol;
    Int m_l1NextNodeOrSymbol;
    CHuffmanDecoderNode () {m_c0End=0;m_c1End=0;
            m_l0NextNodeOrSymbol=-1;m_l1NextNodeOrSymbol=-1;};
    Bool is0Valid() {if ((m_c0End==0)&&(m_l0NextNodeOrSymbol==-1))
            return FALSE;return TRUE;};
    Bool is1Valid() {if ((m_c1End==0)&&(m_l1NextNodeOrSymbol==-1))
            return FALSE;return TRUE;};
};

CHuffmanTree::CHuffmanTree (Int lNOfSymbols, Int *lpFrequencies)
{
    assert(lNOfSymbols>1);
    m_lNOfSymbols=lNOfSymbols;
    m_pNodes = new CNode[2*m_lNOfSymbols-1];
    if(lpFrequencies)
        setFrequencies (lpFrequencies);
}

CHuffmanTree::~CHuffmanTree()
{
    delete[]m_pNodes;
}

Void CHuffmanTree::writeSymbol (Int symbolNo,ostream &stream)
{
    stream<<symbolNo<<" ";
}

Void CHuffmanTree::setFrequencies (Int* lpFrequencies)
{
    for (Int lIndex=0;lIndex<m_lNOfSymbols;lIndex++)
        setFrequency(lpFrequencies[lIndex],lIndex);
}

Void CHuffmanTree::setFrequency (Int lFrequency, Int lIndex)
{
    m_pNodes[lIndex].m_lFrequency=lFrequency;
}

Int huffmanNodeCompare (const Void* Element1, const Void* Element2)
{
    if((*((CNode**)Element1))->m_lFrequency<
            (*((CNode**)Element2))->m_lFrequency)return  1;   
    if((*((CNode**)Element1))->m_lFrequency>
            (*((CNode**)Element2))->m_lFrequency)return -1;
    if((*((CNode**)Element1))->m_lBalancer<
            (*((CNode**)Element2))->m_lBalancer)return  1;   
    if((*((CNode**)Element1))->m_lBalancer>
            (*((CNode**)Element2))->m_lBalancer)return -1;
    return 0;
}

Void CHuffmanTree::buildTree ()
{
    assert(m_lNOfSymbols>1);
    Int NOfNodesForSorting=m_lNOfSymbols;
    Int NextFreeNode=m_lNOfSymbols;
    CNode **pSortingArray;
    pSortingArray=new CNode*[m_lNOfSymbols];
    for (Int i=0;i<m_lNOfSymbols;i++)
        pSortingArray [i] = &m_pNodes[i];
    
    while (NOfNodesForSorting>1)
    {
        qsort(&pSortingArray[0],NOfNodesForSorting,
                sizeof(pSortingArray[0]),huffmanNodeCompare);
        pSortingArray[NOfNodesForSorting-2]->m_lNodeIndex=NextFreeNode;
        pSortingArray[NOfNodesForSorting-1]->m_lNodeIndex=NextFreeNode;
        pSortingArray[NOfNodesForSorting-2]->m_cCode=0;
        pSortingArray[NOfNodesForSorting-1]->m_cCode=1;
        m_pNodes[NextFreeNode].m_lFrequency=
                pSortingArray[NOfNodesForSorting-2]->m_lFrequency+
                pSortingArray[NOfNodesForSorting-1]->m_lFrequency;
        m_pNodes[NextFreeNode].m_lBalancer=
                pSortingArray[NOfNodesForSorting-2]->m_lBalancer+
                pSortingArray[NOfNodesForSorting-1]->m_lBalancer;
        pSortingArray[NOfNodesForSorting-2]=&m_pNodes[NextFreeNode];
        NOfNodesForSorting--;
        NextFreeNode++;
    };
    delete pSortingArray;
}

void CHuffmanTree::writeOneTableEntry(ostream &stream, Int entryNo,
                                    Double lTotalFrequency, Double &dNOfBits)
{
    Double dP=m_pNodes[entryNo].m_lFrequency/(Double)lTotalFrequency;
    Char *pCodeArray=new Char[m_lNOfSymbols-1];
    Int lNode=entryNo;
    Int lCodeNo=0;
    while(lNode!=2*m_lNOfSymbols-2)
    {
        pCodeArray[lCodeNo++]=m_pNodes[lNode].m_cCode;
        lNode=m_pNodes[lNode].m_lNodeIndex;
    };
    writeSymbol(entryNo,stream);
    dNOfBits+=dP*lCodeNo;
    while(lCodeNo>0)
    {
        lCodeNo--;
        stream<<(Int)pCodeArray[lCodeNo];
    };
    stream<<endl;
    delete pCodeArray;
}

Void CHuffmanTree::writeTable (ostream &stream)
{
    Int lTotalFrequency=0;
    Double dEntropy=0;
    Double dNOfBits=0;
    statistics (lTotalFrequency, dEntropy);
    for (Int i = 0; i < m_lNOfSymbols; i++)
        writeOneTableEntry (stream, i, lTotalFrequency, dNOfBits);
    printStatistics (dEntropy, dNOfBits, stream);
}

Void CHuffmanTree::writeTableSorted(ostream &stream)
{
    Int lTotalFrequency=0;
    Double dEntropy=0;
    Double dNOfBits=0;
    statistics(lTotalFrequency,dEntropy);
    CNode **pSortingArray;
    pSortingArray = new CNode*[m_lNOfSymbols];
	Int i;
    for (i = 0; i < m_lNOfSymbols; i++)
        pSortingArray [i] = &m_pNodes[i];
    qsort(&pSortingArray[0],m_lNOfSymbols,
            sizeof (pSortingArray [0]), huffmanNodeCompare);
    for(i=0;i<m_lNOfSymbols;i++)
        writeOneTableEntry (stream, pSortingArray [i] - m_pNodes,
                lTotalFrequency,dNOfBits);
    delete pSortingArray;
    printStatistics (dEntropy, dNOfBits, stream);
}

Void CHuffmanTree::statistics (Int& lTotalFrequency,Double &dEntropy)
{
	Int i;
    for (i=0;i<m_lNOfSymbols;i++)
        lTotalFrequency+=m_pNodes[i].m_lFrequency;
    for(i=0;i<m_lNOfSymbols;i++)
    {
        double dP=m_pNodes[i].m_lFrequency/(double)lTotalFrequency;
        if(dP!=0)
            dEntropy+=dP*log(1/dP)/log(2);    
    };
}

Void CHuffmanTree::printStatistics (Double dEntropy, Double dNOfBits, ostream &stream)
{
    stream<<endl<<endl;
    stream<<"//Entropy Per Symbol : "<<dEntropy<<endl;
    stream<<"//Bits Per Symbol    : "<<dNOfBits<<endl;
    stream<<"//Table Efficiency   : "<<dEntropy/dNOfBits<<endl;
}

Int CHuffmanCoDec::makeIndexFromSymbolInTable(istream &huffmanTable)  
{
    Int lR;
	huffmanTable >> lR;
	return lR;
}

Void CHuffmanCoDec::trashRestOfLine (istream &str)
{
    Int iC;
    do
    {
        iC = str.get ();
    }
    while ((iC!='\n') && (iC!=EOF));
}

Bool CHuffmanCoDec::processOneLine (istream &huffmanTable, Int &lSymbol,
                    Int &lCodeSize, Char *pCode)
{
	huffmanTable >> ws;
    while(huffmanTable.peek()=='/')
    {
		trashRestOfLine (huffmanTable);
		huffmanTable >> ws;
    }
    if(huffmanTable.peek()==EOF)
        return FALSE;

	lSymbol=makeIndexFromSymbolInTable(huffmanTable);      
	huffmanTable >> ws;
	int iC=huffmanTable.get();                            
	lCodeSize=0;                                       
	while((iC=='0')||(iC=='1'))                             
    {                                                       
        if(pCode)                                           
        {                                                   
            if(iC=='0')                                     
                pCode[lCodeSize]=0;                         
            else                                            
                pCode[lCodeSize]=1;                         
        };                                                  
        lCodeSize++;                                        
        iC=huffmanTable.get();                            
    };
    if((iC!='\n')&&(iC!=EOF))
        trashRestOfLine(huffmanTable);                         
    assert(lCodeSize);
    return TRUE;
}

Void CHuffmanCoDec::profileTable (istream &huffmanTable,
            Int &lNOfSymbols, Int &lMaxCodeSize)
{
    huffmanTable.clear();
    huffmanTable.seekg(0,ios::beg);
    lNOfSymbols=0;
    lMaxCodeSize=0;
    while(huffmanTable.peek()!=EOF)
    {
        Int lCodeSize;
        Int lSymbol;
        if (processOneLine(huffmanTable,lSymbol,lCodeSize,NULL))
        {
            lNOfSymbols++;
            if(lCodeSize>lMaxCodeSize)
                lMaxCodeSize=lCodeSize;
            assert(lCodeSize);
        };
    };
    assert(lNOfSymbols>1);
    assert(lMaxCodeSize);
}

CHuffmanDecoder::CHuffmanDecoder (CInBitStream &bitStream)
{
    attachStream (bitStream);
}

CHuffmanDecoder::~CHuffmanDecoder()
{
    delete []m_pTree;
}

Void CHuffmanDecoder::realloc(Int lOldSize,Int lNewSize)
{
    CHuffmanDecoderNode *pNewTree=new CHuffmanDecoderNode[lNewSize];
    Int i;
    for(i=0;i<lOldSize;i++)
    {
        pNewTree[i].m_c0End=m_pTree[i].m_c0End;
        pNewTree[i].m_c1End=m_pTree[i].m_c1End;
        pNewTree[i].m_l0NextNodeOrSymbol=m_pTree[i].m_l0NextNodeOrSymbol;
        pNewTree[i].m_l1NextNodeOrSymbol=m_pTree[i].m_l1NextNodeOrSymbol;
    };
    delete []m_pTree;
    m_pTree=pNewTree;
}

void CHuffmanDecoder::loadTable(istream &huffmanTable,Bool bIncompleteTree)
{
    Int lTableSize;
    Int lNOfSymbols;
    Int lMaxCodeSize;
    Int lNextFreeNode=1;
    profileTable (huffmanTable,lNOfSymbols,lMaxCodeSize);
    assert(lNOfSymbols>1);
    assert(lMaxCodeSize);
    m_pTree = new CHuffmanDecoderNode [lNOfSymbols - 1];
    lTableSize=lNOfSymbols-1;
    Char *pCode = new Char[lMaxCodeSize];
    huffmanTable.clear();
    huffmanTable.seekg(0,ios::beg);
    while(huffmanTable.peek()!=EOF)
    {
        Int lCodeSize;
        Int lSymbol;
        Int lCurrentNode=0;
        if (processOneLine (huffmanTable,lSymbol,lCodeSize,pCode))
        {
            assert((lSymbol<lNOfSymbols)||bIncompleteTree);
            assert(lCodeSize);
            for(Int i=0;i<lCodeSize;i++)
            {
				assert ((lCurrentNode<lNOfSymbols-1)||bIncompleteTree);
                int iBit=pCode[i];
				assert((iBit==0)||(iBit==1));
                if(i==(lCodeSize-1))
                {
                    if(iBit==0)
                    {
                        assert(!m_pTree[lCurrentNode].is0Valid());
                        m_pTree[lCurrentNode].m_c0End=1;
                        m_pTree[lCurrentNode].m_l0NextNodeOrSymbol=lSymbol;
                    }
                    else
                    {
                        assert(!m_pTree[lCurrentNode].is1Valid());
                        m_pTree[lCurrentNode].m_c1End=1;
                        m_pTree[lCurrentNode].m_l1NextNodeOrSymbol=lSymbol;
                    }
                }
                else
                {
                    if(iBit==0)
                    {
                        if(!m_pTree[lCurrentNode].is0Valid())
                        {
							if(bIncompleteTree)
                                if(lNextFreeNode>=lTableSize)
                                {
                                    realloc(lTableSize,lTableSize+10);
                                    lTableSize+=10;
                                };
                            assert((lNextFreeNode<lNOfSymbols-1)||bIncompleteTree);
                            m_pTree[lCurrentNode].m_c0End=0;
                            m_pTree[lCurrentNode].m_l0NextNodeOrSymbol=lNextFreeNode; 
                            lNextFreeNode++;
                        };
                        assert(m_pTree[lCurrentNode].m_c0End==0);
                        lCurrentNode=m_pTree[lCurrentNode].m_l0NextNodeOrSymbol;
                    }
                    else
                    {
                        if(!m_pTree[lCurrentNode].is1Valid())
                        {
							if(bIncompleteTree)
                                if(lNextFreeNode>=lTableSize)
                                {
                                    realloc(lTableSize,lTableSize+10);
                                    lTableSize+=10;
                                };
                            assert((lNextFreeNode<lNOfSymbols-1)||bIncompleteTree);                    
                            m_pTree[lCurrentNode].m_c1End=0;
                            m_pTree[lCurrentNode].m_l1NextNodeOrSymbol=lNextFreeNode; 
                            lNextFreeNode++;
                        };
                        assert(m_pTree[lCurrentNode].m_c1End==0);
                        lCurrentNode=m_pTree[lCurrentNode].m_l1NextNodeOrSymbol;      
                    }
                }
            }
        }
    }
    for(Int i=0;i<lTableSize;i++)
    {
        assert((m_pTree[i].is0Valid())||bIncompleteTree);
        assert((m_pTree[i].is1Valid())||bIncompleteTree);
    };
    delete pCode;
}

// Peter Van Beek, Thu May 14 13:44:15 1998
// Added to allow use of predefined vlc-table instead of reading from a file
void CHuffmanDecoder::loadTable(VlcSymbol vlcTable[], Int lNOfSymbols, Bool bIncompleteTree)
{
    assert(lNOfSymbols>1);
    Int lTableSize = lNOfSymbols-1;
    Int lNextFreeNode = 1;
    m_pTree = new CHuffmanDecoderNode [lNOfSymbols - 1];
    Char *pCode;
	for (Int n = 0; n < lNOfSymbols; n++)
    {
        Int lCodeSize;
        Int lSymbol;
        Int lCurrentNode = 0;
		pCode = vlcTable[n].code;
		lSymbol = vlcTable[n].val;
		lCodeSize = strlen(pCode);
		
		assert((lSymbol<lNOfSymbols)||bIncompleteTree);
		assert(lCodeSize);
		for (Int i=0; i<lCodeSize; i++)
		{
			assert ((lCurrentNode<lNOfSymbols-1)||bIncompleteTree);
			Int iBit;
			if (pCode[i] == '0')       iBit = 0;
			else if (pCode[i] == '1')  iBit = 1;
			else assert(FALSE);

			if(i==(lCodeSize-1))
			{
				if(iBit==0)
				{
					assert(!m_pTree[lCurrentNode].is0Valid());
					m_pTree[lCurrentNode].m_c0End=1;
					m_pTree[lCurrentNode].m_l0NextNodeOrSymbol=lSymbol;
				}
				else
				{
					assert(!m_pTree[lCurrentNode].is1Valid());
					m_pTree[lCurrentNode].m_c1End=1;
					m_pTree[lCurrentNode].m_l1NextNodeOrSymbol=lSymbol;
				}
			}
			else
			{
				if(iBit==0)
				{
					if(!m_pTree[lCurrentNode].is0Valid())
					{
						if(bIncompleteTree)
							if(lNextFreeNode>=lTableSize)
							{
								realloc(lTableSize,lTableSize+10);
								lTableSize+=10;
							};
						assert((lNextFreeNode<lNOfSymbols-1)||bIncompleteTree);
						m_pTree[lCurrentNode].m_c0End=0;
						m_pTree[lCurrentNode].m_l0NextNodeOrSymbol=lNextFreeNode; 
						lNextFreeNode++;
					};
					assert(m_pTree[lCurrentNode].m_c0End==0);
					lCurrentNode=m_pTree[lCurrentNode].m_l0NextNodeOrSymbol;
				}
				else
				{
					if(!m_pTree[lCurrentNode].is1Valid())
					{
						if(bIncompleteTree)
							if(lNextFreeNode>=lTableSize)
							{
								realloc(lTableSize,lTableSize+10);
								lTableSize+=10;
							};
						assert((lNextFreeNode<lNOfSymbols-1)||bIncompleteTree);
						m_pTree[lCurrentNode].m_c1End=0;
						m_pTree[lCurrentNode].m_l1NextNodeOrSymbol=lNextFreeNode; 
						lNextFreeNode++;
					};
					assert(m_pTree[lCurrentNode].m_c1End==0);
					lCurrentNode=m_pTree[lCurrentNode].m_l1NextNodeOrSymbol;      
				}
			}
		}
    }
    for (Int i=0; i<lTableSize; i++)
    {
        assert((m_pTree[i].is0Valid())||bIncompleteTree);
        assert((m_pTree[i].is1Valid())||bIncompleteTree);
    };
}

Void CHuffmanDecoder::attachStream(CInBitStream &bitStream)
{
    m_pBitStream=&bitStream;
}

Int CHuffmanDecoder::decodeSymbol()
{
    Char cEnd=0;
    Int lNextNodeOrSymbol=0;
    do
    {
        Int iBit;
        iBit=m_pBitStream->getBits(1);
        if(iBit==0)
        {
            if(m_pTree[lNextNodeOrSymbol].is0Valid())
            {
                cEnd=m_pTree[lNextNodeOrSymbol].m_c0End;
                lNextNodeOrSymbol=m_pTree[lNextNodeOrSymbol].m_l0NextNodeOrSymbol;
            }
            else
            {
                cEnd=TRUE;
                lNextNodeOrSymbol=-1;
            }
        }
        else
        {
            if(m_pTree[lNextNodeOrSymbol].is1Valid())
            {
                cEnd=m_pTree[lNextNodeOrSymbol].m_c1End;
                lNextNodeOrSymbol=m_pTree[lNextNodeOrSymbol].m_l1NextNodeOrSymbol;
            }
            else
            {
                cEnd=TRUE;
                lNextNodeOrSymbol=-1;
            }
        };
    }
    while(cEnd==0);
    return lNextNodeOrSymbol;
}

CHuffmanEncoder::~CHuffmanEncoder ()
{
	delete m_pCodeTable; 
	delete m_pSizeTable;
}

CHuffmanEncoder::CHuffmanEncoder(COutBitStream &bitStream)
{
    attachStream(bitStream);
}

Void CHuffmanEncoder::loadTable(istream &huffmanTable)
{
    Int lNOfSymbols;
    Int lMaxCodeSize;
    profileTable(huffmanTable,lNOfSymbols,lMaxCodeSize);
    assert(lNOfSymbols>1);
    assert(lMaxCodeSize);
    m_lCodeTableEntrySize=lMaxCodeSize/8;
    if(lMaxCodeSize%8)
        m_lCodeTableEntrySize++;
    m_pSizeTable = new Int [lNOfSymbols];
//  m_pCodeTable = new U8 [lNOfSymbols*m_lCodeTableEntrySize];
    m_pCodeTable = new Int [lNOfSymbols];
    Char *pCode=new Char[lMaxCodeSize];
    huffmanTable.clear();
    huffmanTable.seekg(0,ios::beg);
    while(huffmanTable.peek()!=EOF)
    {
        Int lCodeSize;
        Int lSymbol;
        Int iBitPosition=0;
        if(processOneLine(huffmanTable,lSymbol,lCodeSize,pCode))
        {
            assert(lSymbol<lNOfSymbols);
            assert(lCodeSize >=0 && lCodeSize <= sizeof (Int) * 8);
            m_pSizeTable[lSymbol]=lCodeSize;
            Int* pCodeTableEntry=&m_pCodeTable[lSymbol]; 
            for(Int i=0;i<lCodeSize;i++)
            {
                if(iBitPosition==0)
                    *pCodeTableEntry=0;
                assert((pCode[lCodeSize - i - 1]==0)||(pCode[lCodeSize - i - 1]==1));                                    
                if(pCode[lCodeSize - i - 1]==0)
                    *pCodeTableEntry&=~(0x01<<iBitPosition);
                else
                    *pCodeTableEntry|=0x01<<iBitPosition;
                iBitPosition++;
				/*
                if(iBitPosition>=8)
                {
                    iBitPosition=0;
                    pCodeTableEntry++;
                } 
				*/
            }
        }
    }
    delete pCode;
}

// Peter Van Beek, Thu May 14 13:44:15 1998
// Added to allow use of predefined vlc-table instead of reading from a file
Void CHuffmanEncoder::loadTable(VlcSymbol vlcTable[], Int lNOfSymbols)
{
    assert(lNOfSymbols>1);
    m_pSizeTable = new Int [lNOfSymbols];
    m_pCodeTable = new Int [lNOfSymbols];
    Char *pCode;
	for (Int n = 0; n < lNOfSymbols; n++)
    {
        Int lSymbol;
        Int lCodeSize;
        Int iBitPosition=0;
		pCode = vlcTable[n].code;
		lSymbol = vlcTable[n].val;
		lCodeSize = strlen(pCode);

		assert(lSymbol<lNOfSymbols);
		assert(lCodeSize >=0 && lCodeSize <= sizeof (Int) * 8);

		m_pSizeTable[lSymbol]=lCodeSize;
		Int* pCodeTableEntry=&m_pCodeTable[lSymbol]; 
		for (Int i=0;i<lCodeSize;i++)
		{
			if (iBitPosition==0)
				*pCodeTableEntry=0;
			assert((pCode[lCodeSize - i - 1] == '0')||(pCode[lCodeSize - i - 1] == '1'));                                    
			if (pCode[lCodeSize - i - 1] == '0')
				*pCodeTableEntry&=~(0x01<<iBitPosition);
			else
				*pCodeTableEntry|=0x01<<iBitPosition;
			iBitPosition++;
		}
    }
}

void CHuffmanEncoder::attachStream(COutBitStream &bitStream)
{
    m_pBitStream=&bitStream;
}

UInt CHuffmanEncoder::encodeSymbol (Int lSymbol, Char* rgchSymbolName, Bool bDontSendBits)
{
    Int lSize=m_pSizeTable[lSymbol];
	if (bDontSendBits == TRUE)					//for counting bits
		return (lSize);
    Int iCode=m_pCodeTable[lSymbol];
	Int lPrevCounter = m_pBitStream -> getCounter ();
    m_pBitStream -> putBits (iCode,lSize, rgchSymbolName);
	return (m_pBitStream -> getCounter () - lPrevCounter);
}


//!! Peter Van Beek, Wed Feb  5 15:00:46 1997
//!! Adapted for mesh geometry and motion compression
//!! Peter Van Beek, Thu May 14 16:24:25 1998
//!! Adapted to use defined VLC instead of reading from file
CEntropyEncoderSet::CEntropyEncoderSet (COutBitStream &bitStream)
{
	m_pentrencMV = new CHuffmanEncoder (bitStream);
 	m_pentrencWrpPnt = new CHuffmanEncoder (bitStream);

    m_pentrencMV -> loadTable (mvDiffTable, 65);
    m_pentrencWrpPnt->loadTable (coDiffTable, 12);

/*
	ifstream istrmTableMV;
	ifstream istrmWrpPnt;
#ifdef __PC_COMPILER_
	// get the vm_home value
	Char *rgchVmHome = getenv ( "VM_HOME" );
	assert (rgchVmHome != NULL );
	Char rgchFile [100];

	strcpy (rgchFile, rgchVmHome);
	strcat (rgchFile, "\\vlcmvd.txt");
	istrmTableMV.open (rgchFile);
	strcpy (rgchFile, rgchVmHome);
	strcat (rgchFile, "\\wrppnt.txt");
	istrmWrpPnt.open (rgchFile);
#else
	istrmTableMV.open (VM_HOME_DIR"/vlcmvd.txt");
	istrmWrpPnt.open (VM_HOME_DIR"/wrppnt.txt");
#endif // __PC_COMPILER_

    m_pentrencMV -> loadTable (istrmTableMV);
    m_pentrencWrpPnt->loadTable (istrmWrpPnt);
*/
}

//!! Peter Van Beek, Wed Feb  5 15:00:46 1997
//!! Adapted for mesh geometry and motion compression
CEntropyEncoderSet::~CEntropyEncoderSet()
{
	delete m_pentrencMV;
	delete m_pentrencWrpPnt;
}

//!! Peter Van Beek, Thu Oct 16 15:23:59 1997
//!! Added for player integration
Void CEntropyEncoderSet::attachStream (COutBitStream &bitStream)
{
    m_pentrencMV->attachStream (bitStream);
	m_pentrencWrpPnt->attachStream (bitStream);
}

//!! Peter Van Beek, Wed Feb  5 15:00:46 1997
//!! Adapted for mesh geometry and motion compression
//!! Peter Van Beek, Thu May 14 16:24:25 1998
//!! Adapted to use defined VLC instead of reading from file
CEntropyDecoderSet::CEntropyDecoderSet (CInBitStream &bitStream)
{
	m_pentrdecMV = new CHuffmanDecoder (bitStream);
 	m_pentrdecWrpPnt = new CHuffmanDecoder (bitStream);

    m_pentrdecMV -> loadTable (mvDiffTable, 65);
    m_pentrdecWrpPnt->loadTable (coDiffTable, 12);

/*
	ifstream istrmTableMV;
	ifstream istrmWrpPnt;
#ifdef __PC_COMPILER_
	// get the vm_home value
	Char *rgchVmHome = getenv ( "VM_HOME" );
	assert (rgchVmHome != NULL );
	Char rgchFile [100];

	strcpy (rgchFile, rgchVmHome);
	strcat (rgchFile, "\\vlcmvd.txt");
	istrmTableMV.open (rgchFile);
	strcpy (rgchFile, rgchVmHome);
	strcat (rgchFile, "\\wrppnt.txt");
	istrmWrpPnt.open (rgchFile);
#else
	istrmTableMV.open (VM_HOME_DIR"/vlcmvd.txt");
	istrmWrpPnt.open (VM_HOME_DIR"/wrppnt.txt");
#endif // __PC_COMPILER_

    m_pentrdecMV -> loadTable (istrmTableMV);
	m_pentrdecWrpPnt  -> loadTable (istrmWrpPnt);
*/
}

//!! Peter Van Beek, Wed Feb  5 15:00:46 1997
//!! Adapted for mesh geometry and motion compression
CEntropyDecoderSet::~CEntropyDecoderSet()
{
	delete m_pentrdecMV;
	delete m_pentrdecWrpPnt;
}

//!! Peter Van Beek, Thu Oct 16 15:23:59 1997
//!! Added for player integration
Void CEntropyDecoderSet::attachStream (CInBitStream &bitStream)
{
    m_pentrdecMV->attachStream (bitStream);
	m_pentrdecWrpPnt->attachStream (bitStream);
}


// end of file
