/*************************************************************************

This software module was originally developed by 

    Ming-Chieh Lee (Microsoft Corporation)
    Wei-ge Chen (Microsoft Corporation)
    Bruce Lin (Microsoft Corporation)
    Chuang Gu (Microsoft Corporation)

and edited by

    Peter van Beek (University of Rochester)

in the course of development of the MPEG-4 Video (ISO/IEC 14496-2). 
This software module is an implementation of a part of one or more 
MPEG-4 Video (ISO/IEC 14496-2) tools as specified by MPEG-4 Video
(ISO/IEC 14496-2). ISO/IEC gives users of the MPEG-4 Video (ISO/IEC 
14496-2) free license to this software module or modifications thereof 
for use in hardware or software products claiming conformance to the 
MPEG-4 Video (ISO/IEC 14496-2). 
Those intending to use this software module in hardware or software 
products are advised that its use may infringe existing patents. The 
original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability 
for use of this software module or modifications thereof in an 
implementation. 
Copyright is not released for non MPEG-4 Video (ISO/IEC 14496-2) 
conforming products. Microsoft Corporation retains full right to use 
the code for his/her own purpose, assign or donate the code to a third 
party and to inhibit third parties from using the code for non MPEG 
Video (ISO/IEC 14496-2) conforming products. 
This copyright notice must be included in all copies or derivative works. 

Copyright (c) 1996, 1997.

Module Name:
	
	huffman.hpp

Abstract:

	Classes for Huffman encode/decode.

Revision History:

	Peter van Beek,
	- Renamed huffman.h

	Peter Van Beek, Thu Oct 16 15:01:51 1997
	- Removed unnecessary entropy encoders/decoders

	Peter Van Beek, Thu Oct 16 15:15:06 1997
	- Added attachStream methods to CEntropyEncoderSet and CEntropyDecoderSet

	Peter Van Beek, Thu May 14 16:46:41 1998
	- added new loadTable methods to CHuffmanDecoder and CHuffmanEncoder

*************************************************************************/

#ifndef _HUFFMAN_HPP_
#define _HUFFMAN_HPP_

#define VM_HOME_DIR		"."

class istream;
class ostream;
class CInBitStream;
class COutBitStream;
class CEntropyEncoder;
class CEntropyDecoder;
class CNode;

Class CHuffmanTree
{
public:
	// Constructors
	// create tree and load with frequencies
    CHuffmanTree (Int lNOfSymbols, Int* lpFrequencies = NULL);
	// destroy tree
    ~CHuffmanTree ();

	// Operations
	// load frequencies of all symbols
    Void setFrequencies (Int* lpFrequencies);
	// set frequency of one symbol
    Void setFrequency (Int lFrequency, Int lIndex);
	// generate huffman table
    Void buildTree ();
	// write table into text stream
    Void writeTable (ostream &stream);
	// write table sorted by frequencies into text stream
    Void writeTableSorted (ostream &stream);

private:
    CNode* m_pNodes;
    Int m_lNOfSymbols;
    virtual Void writeSymbol (Int symbolNo,ostream &Stream);
    Void writeOneTableEntry (
		ostream &Stream,
		Int entryNo,
        Double lTotalFrequency,
		Double &dNOfBits
	);
    Void statistics (Int &lTotalFrequency, Double &dEntropy);
    Void printStatistics (Double dEntropy, Double dNOfBits, ostream &stream);
};

Class CHuffmanCoDec
{
protected:
    virtual Int makeIndexFromSymbolInTable (istream &huffmanTable);
    Void profileTable (
		istream &huffmanTable,
		Int& lNOfSymbols, 
		Int& lMaxCodeSize
	);
    Bool processOneLine (
		istream &huffmanTable,
		Int& lSymbol,
		Int& lCodeSize,
		Char* pCode
	);
    Void trashRestOfLine (istream &str);
};

class CHuffmanDecoderNode;
Class CHuffmanDecoder:public CHuffmanCoDec, public CEntropyDecoder
{
public:
	// Constructors
	// create unattached decoder
    CHuffmanDecoder () {};
	// create huffman decoder and attach to bitStream
    CHuffmanDecoder (CInBitStream &BitStream);
	// destroy decoder
    ~CHuffmanDecoder ();

	// Operations
	// load huffman table from the text istream
    Void loadTable (istream &HuffmanTable,Bool bIncompleteTree=TRUE);
	// load VLC table from global variable
	void loadTable(VlcSymbol vlcTable[], Int lNOfSymbols, Bool bIncompleteTree=TRUE);
	// return one decoded symbol from the attached CInBitStream
    Int decodeSymbol ();
	// attach bitStream to decoder
    Void attachStream (CInBitStream &bitStream);
	// return attached CInBitSteam
	CInBitStream* bitstream () {return m_pBitStream;};

private:
    Void realloc (Int lOldSize,Int lNewSize);
    CHuffmanDecoderNode* m_pTree;
    CInBitStream* m_pBitStream;
};

Class CHuffmanEncoder:public CHuffmanCoDec, public CEntropyEncoder
{
public:
	// Constructors
	// create and attach encoder
    CHuffmanEncoder (COutBitStream &bitStream);
	// destroy encoder
    ~CHuffmanEncoder ();

	// Operations
	// load huffman table from the text istream
    Void loadTable (istream &HuffmanTable);
	// load VLC table from global variable
	Void loadTable (VlcSymbol vlcTable[], Int lNOfSymbols);
	// attach bitStream to the decoder
    Void attachStream (COutBitStream &BitStream);
	// encode one symbol and put into the attached COutBitStream
    UInt encodeSymbol (Int lSymbol, Char* rgchSymbolName = NULL, Bool bDontSendBits = FALSE);
	// return attached COutBitStream
	COutBitStream* bitstream () {return m_pBitStream;};

private:
	Int m_lCodeTableEntrySize;
    Int* m_pCodeTable;
    Int* m_pSizeTable;
    COutBitStream* m_pBitStream;
};

Class CEntropyEncoderSet{
public:
	// Constructors
    CEntropyEncoderSet (COutBitStream &bitStream);	// create encoders and attach bitstream
    ~CEntropyEncoderSet ();							// destroy encoders

    Void attachStream (COutBitStream &bitStream); 	// attach bitStream to encoders

	CEntropyEncoder* m_pentrencMV;					// encoder for MVs
	CEntropyEncoder* m_pentrencWrpPnt;				// encoder for Points
};

Class CEntropyDecoderSet{
public:
	// Constructors
    CEntropyDecoderSet (CInBitStream &bitStream);	// create decoders and attach bitstream
    ~CEntropyDecoderSet ();							// destroy encoders

    Void attachStream (CInBitStream &bitStream);	// attach bitStream to decoders

	CEntropyDecoder* m_pentrdecMV;					// decoder for MVs
	CEntropyDecoder* m_pentrdecWrpPnt;				// decoder for Points
};

#endif

