/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* The University of Rochester retains full right to use the code for */
/* their own purposes, assign or donate the code to a third party and */
/* to inhibit third parties from using the code for non MPEG-4 Visual */
/* (ISO/IEC 14496-2) conforming products.                             */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1997, 1998                                           */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         mesh_decode.cpp
   AUTHOR       P. J. L. van Beek, University of Rochester
   DATE         Peter Van Beek, Wed May  6 15:35:36 1998

   DESCRIPTION  2D Mesh geometry and motion decoding.

				In general, the decoding program reads tracking data
				of a moving video-object from a bitstream and generates
				a sequence of 2-D triangular meshes. Each mesh describes
				the geometry of a 2-D mesh at a certain point in time.
				This tracking data may be used by other systems to
				render animated video object planes. 
				Note that all the nodes in the mesh are present from
				the first to the last time instant (i.e. no nodes are ever
				added or deleted). 
				The mesh node positions have 0.5 pixel accuracy,
				while motion vectors (and thus later mesh node positions)
				have 0.5 pixel accuracy. Motion vector difference coding 
				is similar to that of the Video VM (but not prediction).

				Some overhead is present around the core syntax. These 
				are "mesh object" start codes, "mesh object" end codes, 
				and "mesh object plane" start codes.

				Note that in this code, one assumes a coordinate system
				with origin at the top-left, such that the x-axis points
				to the right and the y-axis points down. Definitions of
				above/below, clockwise/counterclockwise etc. all depend
				on this assumption.

   NOTES		This file can be used both integrated into the MPEG-4
				Systems player or as a stand-alone Visual Object decoder. 

				In the latter case, each mesh object plane is saved in a 
				data-file having the extension ".tess" (for tesselation). 
				The place where the mesh is to be placed for composition
				is given as arguments on the command-line, in pixels wrt
				the origin of the output space (assumed at top-left). 

				Subroutines called from this code may attempt to call an 
				external program (triangle) to compute the Delaunay 
				triangulation of a set of node points. Communication is 
				done through files and system calls. This option is not 
				currently supported when the decoder is integrated into 
				the Systems player.
				
*********************************************************************-*/

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <math.h>
#include <float.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "my_portab.h"					// my portability
#include "linkClass.h"					// Pt2dLink
#include "listClass.h"					// Pt2dList
#include "pointClass.h"					// Point2d
#include "polyClass.h"					// Poly2d
#include "polyRefClass.h"				// Poly2dRef
#include "polyTessClass.h"				// Poly2dTess
#include "queueClass.h"					// TripletQueue
#include "stringClass.h"				// string
#include "general.h"					// various
#include "delaunay.h"					// triangulator interface 

#include "basic.h"						// Microsoft VM definitions
#include "entropy.h"					// Microsoft VM entropy coding
#include "huffman.h"					// Microsoft VM huffman coding 
#include "bitstrm.h"					// Microsoft VM bitstream functions

#include "mesh_decode.h"				// MeshDecoder



////////// COMPILER SWITCHES /////////////////////////////////////////////////

// define VISUAL_DECODER to run the decoder on a stand-alone bitstream
// conforming to the Visual part of the MPEG-4 standard;
// do not define in case of integration into "Systems" compositor/player
#ifdef VISUAL_DECODER
// Define to get running time message
#define TIMER
#endif


// Define printing macro which allows variable number of arguments
#if defined(DPRINTING) || defined(_DEBUG)
#define DPRINT(vargs)				((void) printf vargs)
#else
#define DPRINT(vargs)
#endif

// Define printing macro which allows variable number of arguments
#if defined(DEBUGGING) || defined(_DEBUG)
#define DBUG(vargs)					((void) printf vargs)
#else
#define DBUG(vargs)
#endif




////////// CONSTANTS /////////////////////////////////////////////////////////


// Mesh Object (MO) overhead
#define SC							1 			// start-code prefix
#define NUMBITS_SC					24
#define MO_START_CODE				188			// 0xbc
#define MO_END_CODE					177			// 0xb1 (video endcode)
#define NUMBITS_MO_START			8
#define NUMBITS_MO_END				8

// Mesh Object Plane (MOP) overhead
#define MOP_START_CODE				189			// 0xbd
#define NUMBITS_MOP_START			8

// Geometry and motion coding
#define NUMBITS_MESH_TYPE			2
#define NUMBITS_NUMMESHNODES		16
#define NUMBITS_NUMBOUNNODES		10
#define NUMBITS_FIRSTNODEPOS		10
#define NUMBITS_NUMNODE_UNIF		10
#define NUMBITS_RECTSIZE_UNIF		8
#define NUMBITS_SPLIT_UNIF			2
#define NUMBITS_MOP_FCODE			3

// Node location and motion vector accuracy
#define COORD_FACTOR				2.0
#define MOTVEC_FACTOR				2.0




////////// CLASS MEMBER FUNCTIONS ////////////////////////////////////////////

// constructor
MeshDecoder::MeshDecoder()
{
	startStrm = endStrm = 0;
	nBoun = 0;
	bounIndex = NULL;
	xmin = xmax = ymin = ymax = 0.0;
	// creates input stream not attached to any input buffer
	bitstrmIn    = new CInBitStream ();
	// attach input stream to entropy decoders
	entrdecSet   = new CEntropyDecoderSet (*bitstrmIn);
}


// constructor with given input buffer and its size
MeshDecoder::MeshDecoder(BYTE *pInBuffer, const int inSize)
{
	startStrm = endStrm = 0;
	nBoun = 0;
	bounIndex = NULL;
	xmin = xmax = ymin = ymax = 0.0;
	// creates input stream attached to input buffer
	bitstrmIn    = new CInBitStream (pInBuffer, inSize);
	// attach input stream to entropy decoders
	entrdecSet   = new CEntropyDecoderSet (*bitstrmIn);
}


// (re-)attach a given input buffer
void MeshDecoder::attachBuf(BYTE *pInBuffer, const int inSize, const int bitPos)
{
	// (re-)attach input buffer to stream
	bitstrmIn->attach (pInBuffer, inSize);
	// read some bits away if necessary
	if (bitPos > 0)
		bitstrmIn->getBits(bitPos);
}


// destructor
MeshDecoder::~MeshDecoder()
{
	delete entrdecSet;
	delete bitstrmIn;
	delete [] bounIndex;
}


// return pointer to current location in stream
BYTE *MeshDecoder::getStreamPtr()   
{
	return bitstrmIn->getStreamPtr(); 
}


// return current number of bits in buffer
UINT MeshDecoder::getBitsLeft()
{ 
	return bitstrmIn->getNumBitsInBuf(); 
}


/*+==========================================================================
	NAME		MeshDecoder::decodeMesh
	VERSION		2.0
	PURPOSE		Decode one Mesh Object Plane (MOP) from a mesh stream.
	METHOD		Generally, either an I-MOP is decoded or a P-MOP. An
				I-MOP is an MOP with the initial geometry of a mesh; a
				P-MOP is an MOP with motion of a mesh with respect to the
				previous instance.
	IN/OUTPUT	mType			type of decoded mesh (I_MOP or P_MOP)
	RETURNS		0: OK
				1: End of stream detected or error detected
	NOTES		Warning: assumes 'Int's are at least 32 bits !
==========================================================================-*/
int MeshDecoder::decodeMesh (int &mType)
{
	// a return code or code read from the bitstream
	int code;

	// check to see if mesh object has already ended
	if (endStrm == 1)
		return_errn("decodeMesh: end of stream\n", 1);

	// check to see if mesh object has already started
	if (startStrm == 0) {

#ifdef VISUAL_DECODER
		// decode mesh object start-code
		code = bitstrmIn->getBits (NUMBITS_SC + NUMBITS_MO_START);
		if (code != ((((Int)SC) << NUMBITS_MO_START) + MO_START_CODE)) {
			return_errn("meshdecoder: no mesh-object-start-code\n", 1);
		}
#endif
		startStrm = 1;
	}

#ifdef VISUAL_DECODER
	// search for the next start-code in the bitstream, or return EOF
	if ((code = findStartCode()) == EOF) {
		endStrm = 1;
		return 1;
	}
	// read/decode the next (mesh object plane) start-code
	code = bitstrmIn->getBits(NUMBITS_SC + NUMBITS_MOP_START);
	// check if mesh object has just ended (MO end-code detected) 
	if (code == 0 || code == ((((int)SC) << NUMBITS_MO_END) + MO_END_CODE)) {
		endStrm = 1;
		return 1;
	}

	// decode the isIntra flag
	if ((code = (int) bitstrmIn->getBits (1)) == 1)
		mType = I_MOP;				// Intra-type Mesh Object Plane
	else
		mType = P_MOP;				// Prediction-type Mesh Object Plane
#endif

	if (mType == I_MOP) {
		DPRINT(("------- Decoding mesh I-plane ----------\n"));

		// decode the mesh type code; this flag is '00' or '01'
		code = (int) bitstrmIn->getBits (NUMBITS_MESH_TYPE);
		if (code == 0)
			// decode uniform mesh geometry and boundary
			decUniformMesh();
		else
			// decode Delaunay mesh geometry and boundary
			decDelaunayMesh();

		// compute mesh bounding box attibutes
		setBBox();
		DPRINT(("xmin=%.2f, ymin=%.2f, xmax=%.2f, ymax=%.2f\n", xmin, ymin, xmax, ymax));

	}
	else {
		DPRINT(("------- Decoding mesh P-plane ----------\n"));

		// decode motion from previous mesh to current mesh
		decMeshMotion ();
	}

#ifndef VISUAL_DECODER
	// go to the next byte-aligned position in the bitstream
	if ((code = nextByteAlign()) != 0)
		return 1;
#endif

	// return OK
	return 0;

} // decodeMesh



/*+==========================================================================
	NAME		MeshDecoder::decDelaunayMesh
	VERSION		3.0
	AUTHOR		Peter Van Beek, Thu Aug 14 18:36:19 1997
	PURPOSE		Decode initial geometry of a Delaunay mesh and specify
				its boundary nodes.
	METHOD		First decode the node positions. To this end, the number
				of nodes and boundary nodes is decoded; then each node
				is decoded and stored in a list (boundary nodes are kept
				in a polygon as well). Then, this data is used as input
				for the triangulation routine, which creates the mesh.
				Register the boundary nodes in an array. 
	PRE			meshCurr		empty mesh
				bitstrmIn		pointer to input stream
				entrdecSet		pointer to entropy decoders
	POST		meshCurr		decoded mesh
				numBounNodes	the number of boundary nodes
				bounIndex		the array with indices of boundary nodes
	RETURNS		none
	NOTES		Assumes 0.5 pixel accuracy of initial node coordinates.
				Assumes there are no two nodes with the same coordinates.
==========================================================================-*/
void MeshDecoder::decDelaunayMesh ()
{
	int n;

	DBUG(("\tDecoding Delaunay mesh ...\n")); 

	// decode the number of nodes in the mesh and a marker bit
	int numMeshNodes = (int) bitstrmIn->getBits (NUMBITS_NUMMESHNODES);
	DBUG(("\tNumber of mesh nodes    : %d \n", numMeshNodes));
	bitstrmIn->getBits (1);
	// decode the number of nodes on the boundary and a marker bit
	nBoun = (int) bitstrmIn->getBits (NUMBITS_NUMBOUNNODES);
	DBUG(("\tNumber of boundary nodes: %d \n", nBoun));
	bitstrmIn->getBits (1);
	if (numMeshNodes <= 0  ||  nBoun <= 0)
		return_err("meshdecoder: number of nodes invalid\n");

	// decode the first (boundary) node coordinates and a marker bit
	double diffx = (double) bitstrmIn->getBits (NUMBITS_FIRSTNODEPOS);
	double diffy = (double) bitstrmIn->getBits (NUMBITS_FIRSTNODEPOS);
	bitstrmIn->getBits (1);

	Point2d pt = Point2d(diffx/COORD_FACTOR, diffy/COORD_FACTOR);
	DBUG(("\tDecoded first (boundary) node: %.1lf, %.1lf\n", pt.X(), pt.Y()));

	// initialize boundary point list and node list with first node
	Poly2d bounCurr(pt);
	Pt2dList nodesCurr(pt);

	// decode all other nodes
	for (n=1; n<numMeshNodes; n++)
	{
		// previously decoded node
		Point2d ptlast = pt;

		// decode difference value for current node 
		CInBitStream* pstrm = entrdecSet->m_pentrdecWrpPnt->bitstream();
		int diff;
		for (UINT XorY = 0; XorY < 2; XorY++) {
			long lSz = entrdecSet->m_pentrdecWrpPnt->decodeSymbol();
			int iSign = pstrm->peekBits (1);	// peek at the sign
			if (iSign == 1)	
				diff = pstrm->getBits (lSz);
			else {
				// first get the magnitude
				int iAbsDiff = ~(pstrm->getBits(lSz));
				// mask out the irrelevent bits and apply sign
				int iMask = ~(0xFFFFFFFF << lSz);	
				diff = -1 * (iAbsDiff & iMask);
			}

			if (XorY == 0)	diffx = (double) diff;
			else			diffy = (double) diff;
		}

		// compute coordinates of current node 
		pt = ptlast + Point2d(diffx/COORD_FACTOR, diffy/COORD_FACTOR);
		// store node in node list 
		nodesCurr.append(pt);
		// store node in boundary polygon if among first boundary nodes
		if (n < nBoun)
			bounCurr.append(pt);
		
		DBUG(("\tDecoded node %d: %.1lf, %.1lf\n", n, pt.X(), pt.Y()));
	}

	// constrained Delaunay triangulation of the set of decoded nodes
	string fname = "meshdec";
	Poly2dTess* meshdec = triangulate(nodesCurr, bounCurr, fname, 1);
	if (meshdec == NULL)
		return_err("meshdecoder: Delaunay triangulation failed\n");

	meshCurr = *meshdec;
	DBUG(("\tDelaunay-triangulated decoded nodes\n")); 

	// store indices of boundary points in an array 
	bounIndex = new int [nBoun];
	// make sure nodes are counter-clockwise ordered
	if (bounCurr.isCCW() == 1) {
		for (n=0; n<nBoun; n++)
			bounIndex[n] = meshCurr.where(bounCurr[n]);
	}
	else {
		for (n=0; n<nBoun; n++)
			bounIndex[n] = meshCurr.where(bounCurr[nBoun-n-1]);
	}

	// finish up
	nodesCurr.empty();

} // decDelaunayMesh


/*+==========================================================================
	NAME		MeshDecoder::decUniformMesh
	VERSION		3.0
	AUTHOR		Peter Van Beek, Tue May  5 16:44:26 1998
	PURPOSE		Decode initial geometry of a uniform mesh and specify its
				boundary nodes.
	METHOD		A uniform mesh is coded by only 5 parameters. Based on the
				parameters, one can construct the mesh simply by enumerating
				all triangles. 
	PRE			meshCurr		empty mesh
				bitstrmIn		pointer to input stream
	POST		meshCurr		decoded mesh
				nBoun			the number of boundary nodes
				bounIndex		the array with indices of boundary nodes
	RETURNS		none
	NOTES		
==========================================================================-*/
void MeshDecoder::decUniformMesh ()
{
	int    code;						// value decoded from bitstream
	int    n, b, nhori, nverti;			// number of hor/vert nodes 
	double xsize, ysize;				// rectangle sizes


	DBUG(("\tDecoding uniform mesh ... \n"));

	// decode the number of nodes in the mesh (horizontal/vertical)
	nhori = (int) bitstrmIn->getBits (NUMBITS_NUMNODE_UNIF);
	nverti = (int) bitstrmIn->getBits (NUMBITS_NUMNODE_UNIF);
	DBUG(("\tNumber of nodes horizontally: %d \n", nhori));
	DBUG(("\tNumber of nodes vertically  : %d \n", nverti));
	// marker bit
	bitstrmIn->getBits (1);

	// decode the sizes of the rectangles in the mesh
	code = (int) bitstrmIn->getBits (NUMBITS_RECTSIZE_UNIF);
	xsize = ((double) code) / 2.0;
	code = (int) bitstrmIn->getBits (NUMBITS_RECTSIZE_UNIF);
	ysize = ((double) code) / 2.0;
	DBUG(("\tHorizontal rectangle size   : %.2lf \n", xsize));
	DBUG(("\tVertical rectangle size     : %.2lf \n", ysize));

	// decode the triangle split code
	code = (int) bitstrmIn->getBits (NUMBITS_SPLIT_UNIF);
	DBUG(("\tTriangle split code         : %d\n", code));

	// create the mesh from top-left to bottom-right, rectangle by rectangle;
	// append two triangles for each rectangle; the orientation of these
	// triangles is determined by the split code.
	meshCurr.makeUniform(nhori, nverti, xsize, ysize, code);

	// reorder owner triangle pointers of nodes (just for bookkeeping)
	meshCurr.reOrderOwners();

	// store indices of boundary points in an array 
	nBoun = (nhori+nverti)*2 - 4;
	bounIndex = new int [nBoun];
	// store the nodes in counter-clockwise ordering
	Point2d bpt(0.0, 0.0);
	for (n=1, b=0; n<nverti; n++, b++) {
		bpt.move(0.0, ysize);
		bounIndex[b] = meshCurr.where(bpt);
	}
	for (n=1; n<nhori; n++, b++) {
		bpt.move(xsize, 0.0);
		bounIndex[b] = meshCurr.where(bpt);
	}
	for (n=1; n<nverti; n++, b++) {
		bpt.move(0.0, -ysize);
		bounIndex[b] = meshCurr.where(bpt);
	}
	for (n=1; n<nhori; n++, b++) {
		bpt.move(-xsize, 0.0);
		bounIndex[b] = meshCurr.where(bpt);
	}

} // decUniformMesh



/*+==========================================================================
	NAME		MeshDecoder::decMeshMotion
	VERSION		4.0
	AUTHOR		Peter Van Beek, Thu Feb 19 12:17:51 1998
	PURPOSE		Decoding of motion vectors of nodes in a triangular mesh. 
				Thereby computing the state of the mesh at the current time.
				A motion vector is a vector between a node in a mesh and 
				the corresponding node in another mesh.
	METHOD		The motion vectors are encoded into the bitstream in a
				certain order, defined by a breadth-first traversal of 
				the triangles of a mesh (see encoder). This order is
				recomputed by the decoder given the previous mesh, and
				each decoded motion vector is added to a node of the
				triangle being visited according to that order.
				Breadth-first traversal of triangles uses a fifo queue of 
				triangles. By ordering the nodes of stored triangles, the 
				traversal is speeded up. Visited triangles and nodes are
				marked, such that each node is processed exactly once.
	PRE			meshCurr		Mesh in the previous state
				nBoun			specifies number of boundary nodes
				bounIndex		specifies indices of boundary nodes
				bitstrmIn		pointer to input stream
				entrdecSet		pointer to entropy decoders
	POST		meshCurr		Mesh in the current state (+ motion vectors)
	RETURNS		none
	NOTES		Meshes are assumed to triangular and the two meshes
				are assumed to be in corresponcence, i.e. nodes/triangles 
				with same index correspond.
				Boundary nodes are assumed to be ordered counterclockwise.
				Triangle vertices are assumed to be ordered clockwise. 
==========================================================================-*/
void MeshDecoder::decMeshMotion ()
{
	TripletQueue queue;					// first-in-first-out triangle queue
	Poly2dRef *pref = NULL;					// polygon pointers
	Pt2dLink  *pb1, *pb2, *paa, *pcc;	// node pointers
	Point2d mvPred;						// prediction motion vector
	int i;								// loop index
	int code;							// read from bitstream
	Tri2d tri;							// a triangle


	// check input
	if (meshCurr.getSizePoints() == 0  ||  meshCurr.getSize() == 0)
		return_err("meshdecoder: number of nodes/triangles invalid\n");

	// decode motion range parameter (fcode)
	fCode = (UINT) bitstrmIn->getBits (NUMBITS_MOP_FCODE);
	// compute motion vector scale factor and motion range in half-pel units
	scaleFactor = 1 << (fCode - 1);
	motionRange = 32 << (fCode - 1);

	// initialize node and polygon markers
	meshCurr.markLinks(0);
	meshCurr.markRefs(0);

	// find the most top-left node on the boundary of the mesh
	int bmin = getTopLeftIndex();
	int b1 = bounIndex[bmin];				// top left node on boundary
	int b2 = bounIndex[Mod(bmin-1, nBoun)];	// next clockwise on boundary
	pb1 = (Pt2dLink *) meshCurr.addressC(b1);
	pb2 = (Pt2dLink *) meshCurr.addressC(b2);

	// find starting triangle (the one that contains b1 and b2)
	for (i=0; i<pb1->getSize(); i++) {
		pref = (Poly2dRef *) pb1->address(i);
		if (pref->address(0) == pb1 && pref->address(1) == pb2) {
			paa = pref->address(2);
			break;
		} 
		else if (pref->address(1) == pb1 && pref->address(2) == pb2) {
			paa = pref->address(0);
			break;
		}
		else if (pref->address(2) == pb1 && pref->address(0) == pb2) {
			paa = pref->address(1);
			break;
		}
	}
	// mark initial triangle
	pref->mark = 1;

	// decode motion vectors of nodes b1 and b2
	code = (int) bitstrmIn->getBits (1);
	if (code == 0) {	   				// nonzero motion vector
		mvPred = Point2d(0.,0.); 		// prediction vector x 2
		pb1->motion = decodeMV(mvPred);
	}
	else								// zero motion vector
		pb1->motion = Point2d();

	code = (int) bitstrmIn->getBits (1);
	if (code == 0) {	   				// nonzero motion vector
		mvPred = 2.0*(pb1->motion);		// prediction vector x 2
		pb2->motion = decodeMV(mvPred);
	}
	else								// zero motion vector
		pb2->motion = Point2d();

	pb1->mark = pb2->mark = 1; 			// mark initial nodes
	queue.append (pb1, pb2, paa);		// initialize the triangle queue 

	// now start the actual bread-first traversal of the mesh
	while (queue.empty() == 0)
	{
		// get and remove first triangle from the queue
		queue.remove(pb1, pb2, paa);

		// process node aa if it is not already marked processed
		if (paa->mark == 0) {
			code = (int) bitstrmIn->getBits (1);
			if (code == 0) {			// nonzero motion vector
				// compute mv prediction and round to half-pixel accuracy
				mvPred = pb1->motion + pb2->motion;
				mvPred.roundToInt();
				// decode motion vector and store it
				paa->motion = decodeMV(mvPred);
			}
			else						// zero motion vector
				paa->motion = Point2d();

			// mark this node processed
			paa->mark = 1;
		}

		// if there is a neighboring unmarked triangle having both 'aa' 
		// and 'b1' as nodes, queue it and mark it
		// note that current triangle is already marked
		int found = 0;
		for (i=0; i<paa->getSize(); i++) {
			// get another triangle connected to aa with b1
			pref = (Poly2dRef *) paa->address(i);
			if (pref->mark != 0)
				continue;				// skip if it is already marked
			if (pref->address(0) == pb1  || pref->address(1) == pb1  ||
				pref->address(2) == pb1) {
				found = 1;
				break;
			}
		}
		if (found != 0) {				// found a triangle!
			pref->mark = 1;				// mark it now

			// get the 3rd vertex of the triangle 
			for (i=0; i<3; i++) {
				pcc = (Pt2dLink *) pref->address(i);
				if (pcc != pb1  &&  pcc != paa)
					break;				// found 3rd vertex
			}
			// put the triangle (with vertices cw) at the end of the queue
			tri = Tri2d (pb1->getPoint(), paa->getPoint(), pcc->getPoint());
			if (tri.isCCW() == 1)
				queue.append (paa, pb1, pcc);
			else
				queue.append (pb1, paa, pcc);
		}

		// if there is a neighboring unmarked triangle having both 'aa' 
		// and 'b2' as nodes, queue it and mark it
		// note that current triangle is already marked
		found = 0;
		for (i=0; i<paa->getSize(); i++) {
			// get another triangle connected to aa with b2
			pref = (Poly2dRef *) paa->address(i);
			if (pref->mark != 0)
				continue;				// skip if it is already marked
			if (pref->address(0) == pb2  || pref->address(1) == pb2  ||
				pref->address(2) == pb2) {
				found = 1;
				break;
			}
		}
		if (found != 0) {				// found a triangle!
			pref->mark = 1;				// mark it now

			// get the 3rd vertex of the triangle 
			for (i=0; i<3; i++) {
				pcc = (Pt2dLink *) pref->address(i);
				if (pcc != pb2  &&  pcc != paa)
					break;				// found 3rd vertex
			}
			// put the triangle (with vertices cw) at the end of the queue
			tri = Tri2d (pb2->getPoint(), paa->getPoint(), pcc->getPoint());
			if (tri.isCCW() == 1)
				queue.append (paa, pb2, pcc);
			else
				queue.append (pb2, paa, pcc);
		}

	} // end-while

	// now apply motion vectors to node locations
	for (i=0; i<meshCurr.getSizePoints(); i++) {
		paa = (Pt2dLink *) meshCurr.addressC(i);
		paa->setPoint(paa->getPoint() + paa->motion);
	}

} // decMeshMotion


// descale a difference motion vector component for decoding
int deScaleMV (const int iVLC, const UINT uiResidual, const UINT uiScaleFactor)
{
	if (iVLC == 0 && uiResidual == 0)
		return 0;
	else if (uiScaleFactor == 1)
		return (iVLC);
	else {
		UINT uiAbsDiffMVcomponent = 
			abs(iVLC)*uiScaleFactor + uiResidual - uiScaleFactor + 1;
		return (sign (iVLC) * uiAbsDiffMVcomponent);
	}
}


// decode a motion vector (given a prediction vector)
// mvPred is the prediction vector x 2
Point2d MeshDecoder::decodeMV (Point2d &mvPred)
{
	double mvX, mvY;
	UINT uiResidual;


	// get difference code and residual from bitstream
	long lSymbol = entrdecSet->m_pentrdecMV->decodeSymbol();
	int iVLC = lSymbol - 32;
	if (fCode > 1 && iVLC != 0)	
		uiResidual = bitstrmIn->getBits(fCode - 1);
	else
		uiResidual = 0;

	// compute motion vector component in fractional pixel units
	mvX = mvPred.X() + (double) deScaleMV(iVLC, uiResidual, scaleFactor);
	// fit in the appropriate range [-range, range]
	if (mvX < -motionRange)
		mvX += 2 * motionRange;				
	else if (mvX > motionRange)
		mvX -= 2 * motionRange;

	// get difference code and residual from bitstream
	lSymbol = entrdecSet->m_pentrdecMV->decodeSymbol();
	iVLC = lSymbol - 32;
	if (fCode > 1 && iVLC != 0)
		uiResidual = bitstrmIn->getBits(fCode - 1);
	else
		uiResidual = 0;

	// compute motion vector component in fractional pixel units
	mvY = mvPred.Y() + (double) deScaleMV (iVLC, uiResidual, scaleFactor);
	// fit in the appropriate range [-range, range]
	if (mvY < -motionRange)
		mvY += 2 * motionRange;				
	else if (mvY > motionRange)
		mvY -= 2 * motionRange;

	// return motion vector at full pixel scale
	return ( Point2d(mvX / MOTVEC_FACTOR, mvY / MOTVEC_FACTOR) );
}


// Search for the next start-code in the bitstream, or return EOF
int MeshDecoder::findStartCode()
{
	// ensure byte alignment by flushing current byte
	bitstrmIn->flush ();

	if (bitstrmIn->eof()==EOF)
		return EOF;

	// read bytes until start-code found or EOF occurs
	while (bitstrmIn->peekBits(NUMBITS_SC) != SC)
	{
		bitstrmIn->getBits(8);
		if (bitstrmIn->eof()==EOF)
			return EOF;
	}
	return 0;
}


// Go to next byte-aligned position in the bitstream
// note: reads at least one 0; then some 1's if necessary (WD 4.0 rev 4)
int MeshDecoder::nextByteAlign()
{
	register int align = (bitstrmIn->getCounter())%8;

	if (align != 0)
		// ensure byte alignment by just flushing current byte
		bitstrmIn->flush ();
	else {
		// check if end of stream is reached
		if (bitstrmIn->eof() == EOF) {
			fprintf(stderr, "MeshDecoder: end of stream\n");
			return EOF;
		}
		// go to next byte aligned position by reading a whole byte
		register int code = (int) bitstrmIn->getBits(8);
		if (code != (int) 0x007F)
			return_errn("nextByteAlign: byte != 0x007F \n", 1);
//			return 1;
	}
	return 0;
}


// Compute mesh bounding box attibutes
void MeshDecoder::setBBox()
{
	if (meshCurr.getSizePoints() > 0) {
		// initialize
		Point2d p = meshCurr.point(0);
		xmin =  floor (p.X());
		xmax =  ceil (p.X());
		ymin =  floor (p.Y());
		ymax =  ceil (p.Y());

		// for all the other vertices in the mesh do ...
		for (int i=1; i<meshCurr.getSizePoints(); i++)
		{
			// get i'th vertex and check the coordinates
			p = meshCurr.point(i);
			double val =  floor (p.X());
			if (val < xmin)
				xmin = val;
			val =  ceil (p.X());
			if (val > xmax)
				xmax = val;
			val =  floor (p.Y());
			if (val < ymin)
				ymin = val;
			val =  ceil (p.Y());
			if (val > ymax)
				ymax = val;
		}
	}
	else
		xmin = xmax = ymin = ymax = 0.0;
}


// Compute index of most top left node on the boundary of the mesh
int MeshDecoder::getTopLeftIndex()
{
	if (nBoun > 0) {
		int index = -1;
		double xymin = DBL_MAX;
		double ymin  = DBL_MAX; 
		for (int n=0; n<nBoun; n++) {
			Point2d pt = meshCurr.point(bounIndex[n]);
			double xy = pt.X()+pt.Y();
			// find node with minimum x+y; y value breaks ties
			if ((xy < xymin) ||	(xy == xymin  &&  pt.Y() < ymin)) {	
				xymin = xy;
				index = n;
				ymin = pt.Y();
			}
		}
		return index;
	}
	else {
		fprintf(stderr, "meshdecoder: no nodes\n");
		return -1;
	}
}




////////// MAIN FUNCTION TO RUN STAND-ALONE //////////////////////////////////

#ifdef VISUAL_DECODER

int main (int argc, char *argv[])
{
	// Base of filename of mesh mappings without numbers of extension
	char meshDatFile[200];
	// Full filename of input bitstream file
	char bitStreamFile[200];
	// Full filename of output mesh file
	char tessFile[200];
	// Pointer to buffer holding file-input
	BYTE *inBuffer;
	// MOP-type (I or P)
	int type;
	// Frame number being decoded, number of first frame and frame increment
	int frame = 0;
	int firstFrame = 0;
	int frameInc = 1;
	// Mesh bounding box origin coordinates for composition
	double xOrig = 0.0;
	double yOrig = 0.0;


	// check number of arguments and read arguments from command line
	if (argc < 5) {
		fprintf(stderr, "2-D mesh geometry and motion decoder\n");
		fprintf(stderr, "Usage: %s <bitstream> <basename> <firstframe> <frameincr> [xo] [yo]\n", argv[0]);
		fprintf(stderr, "bitstream:  input mesh bitstream to be decoded\n");
		fprintf(stderr, "basename:   basename of output mesh-files\n");
		fprintf(stderr, "firstframe: number of first frame\n");
		fprintf(stderr, "frameinc:   frame increment value\n");
		fprintf(stderr, "xo,yo:      bounding-box origin coordinates\n");
		return 1;
	}
	sprintf (bitStreamFile, "%s", argv [1]);
	sprintf (meshDatFile, "%s", argv [2]);
	firstFrame = atoi(argv[3]);
	frameInc   = atoi(argv[4]);
	xOrig = yOrig = 0.0;
	if (argc == 7) {
		xOrig = atof(argv[5]);
		yOrig = atof(argv[6]);
	}

	DPRINT(("========================================================\n"));
	DPRINT(("2-D mesh geometry and motion decoder ------- version 4.0\n"));
	DPRINT(("========================================================\n\n"));
	DPRINT(("\tBitstreamfile      : %s\n", bitStreamFile));
	DPRINT(("\tBasename meshfiles : %s\n", meshDatFile));
	DPRINT(("\tFirst frame no.    : %d\n", firstFrame));
	DPRINT(("\tFrame increment    : %d\n", frameInc));
	DPRINT(("\tBounding-box offset: %.2lf,%.2lf\n\n", xOrig, yOrig));

	// open bitstream file and determine its size
	struct stat fstatus;
	if (stat (bitStreamFile, &fstatus) == -1) {
		fprintf(stderr, "Cannot open bitstream file\n");
		fprintf(stderr, "Usage: %s <bitstream> <basename> <firstframe> <frameincr> [xo] [yo]\n", argv[0]);
		fprintf(stderr, "bitstream:  input mesh bitstream to be decoded\n");
		fprintf(stderr, "basename:   basename of output mesh-files\n");
		fprintf(stderr, "firstframe: number of first frame\n");
		fprintf(stderr, "frameinc:   frame increment value\n");
		fprintf(stderr, "xo,yo:      bounding-box offset coordinates\n");
		return 1;
	}

	// read complete file into BYTE buffer
	// assume this file is not enormously big
	UINT finSize = (int) fstatus.st_size;
	inBuffer = new BYTE [finSize];
	int fd = open(bitStreamFile, O_RDONLY|O_BINARY);
	read(fd, inBuffer, finSize);
	close(fd);

	// declare mesh decoder and attach it to the input buffer
	MeshDecoder decoder(inBuffer, finSize);


	DPRINT(("========================================================\n\n"));

#ifdef TIMER
// SOLARIS PORTING 
/*	// Run time logging
	struct timeval tstart,tstop;
	// log the starting time
	gettimeofday(&tstart, (struct timezone *)NULL);*/
	// Run time logging
	unsigned long tstart,tstop;
	// log the starting time
	tstart = timeGetTime();
// SOLARIS PORTING - END
#endif

	frame = firstFrame;
	
	/////////////// MESH OBJECT PLANE LOOP ////////////////////////////
	
	// until end of bitstream file do
	while (decoder.getEndStrm() == 0)
	{
		// decode a mesh by decoding a I- or P-MOP
		int ret = decoder.decodeMesh(type);
		if (ret != 0)
			break;

		if (type == I_MOP)
			// move the mesh by specified offset
			decoder.meshCurr.move(xOrig, yOrig);

		// write mesh data of current frame
		sprintf(tessFile, "%s%d.tess", meshDatFile, frame);
		decoder.meshCurr.write(tessFile);
		frame+=frameInc;

	} // endwhile (frame loop)

	DPRINT(("\n========================================================\n\n"));

#ifdef TIMER
// SOLARIS PORTING
/*	// compute and print running time in hundredths of a second
	gettimeofday(&tstop, (struct timezone *)NULL);
	int runtime = 100*(tstop.tv_sec-tstart.tv_sec)
		+ (tstop.tv_usec-tstart.tv_usec)/10000;*/
	tstop = timeGetTime();
	int runtime = ( tstop - tstart ) / 100; 
	//tstop and tstart are in milliseconds
// SOLARIS PORTING - END
	frame = (frame-firstFrame)/frameInc + 1;
	printf("\tTiming info       : %d.%02d seconds, %d frames, %d.%02d fps\n",
		   runtime/100, runtime%100,
		   frame, ((10000*frame+runtime/2)/runtime)/100,
		   ((10000*frame+runtime/2)/runtime)%100);
#endif


	////////// cleanup my junk and exit ///////////////////////////////////
	delete [] inBuffer;
	return 0;

} // main


#endif // VISUAL_DECODER


// end of file

