/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* A. Tanju Erdem (Eastman Kodak Company)                             */
/* Candemir Toklu (University of Rochester)                           */
/*                                                                    */
/* and edited by                                                      */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* Eastman Kodak Company and the University of Rochester retain full  */
/* right to use the code for their own purposes, assign or donate the */
/* code to a third party and to inhibit third parties from using the  */
/* code for non MPEG-4 Visual (ISO/IEC 14496-2) conforming products.  */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1994-1997                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         pointClass.cpp
   AUTHOR       A. Tanju Erdem - Eastman Kodak Company and Candemir Toklu - 
                University of Rochester.
   DATE         September 1994
   UPDATE       14 August 1996 by Candemir Toklu
   DESCRIPTION  Member functions for Point2d class are defined in this
                file.

   SYNOPSIS     Include files : 
                                
                #include "pointClass.h"
                #include "general.h"
                               
*********************************************************************-*/
#include "pointClass.h"
#include "general.h"


//------------------------------------------------------------------------
// Constructors for class Point2d

Point2d::Point2d()  // Initialization
{
  x = y = 0.0;
  u = v = z = 0.0;
}

Point2d::Point2d( double a, double b)  // Initialization
{
  x = a;
  y = b;
  u = 0.0;
  v = 0.0;
  z = 0.0;
}

Point2d::Point2d( const Point2d& p)  // Initialization
{
  x = p.x;
  y = p.y;
  u = 0.0;
  v = 0.0;
  z = 0.0;
}

Point2d::~Point2d() {;}

Point2d& Point2d::operator=( const Point2d& p)
{
  x = p.x;
  y = p.y;
  u = p.u;
  v = p.v;
  z = p.z;
  return *this;
}

Point2d& Point2d::operator=( double d)
{
  x = d;
  y = d;
  u = 0.0;
  v = 0.0;
  z = 0.0;
  return *this;
}


//--------------------------------------------------------------------------
// Is the point inside a given convex polygon?
//--------------------------------------------------------------------------
int Point2d::isInside( Poly2d pg)
{
  int i;
  int size = pg.getSize();

  if( size == 0) return 0;

// Obtain the vectors from the point to the vertices of the polygon.

  for( i = 0; i < size; i++) {
    pg[i] -= *this;
  }

// If the point is inside the polygon, then angles between the consecutive 
// vectors will either all be >= 180 or <= 180. We determine whether an angle 
// is less or greater than or equal to 180 by using the cross-product operation.

  int newSign;
  int initialSign = 0;
  for( i = 0; i < size; i++)
  {
    if( ( newSign = SignE(pg[i] ^ pg[i+1])) !=0)
    {
      if( initialSign == 0) initialSign = newSign;
      if( newSign != initialSign) break;
    }     
  }
  return ( i == size);  // if i == size then the point is inside the polygon 
}


//--------------------------------------------------------------------------
// Is the point on a given polygon?
//--------------------------------------------------------------------------
int Point2d::isOn( Poly2d pg)
{
  int i;

  if( ! (*this).isInside(pg))
  {
    int size = pg.getSize();

    if( size == 0) return 0;

  // Obtain the vectors from the point to the vertices of the polygon.

    for( i = 0; i < size; i++) {
      pg[i] -= *this;
    }

    for( i = 0; i < size; i++)
    {
      if( pg[i] || pg[i+1])
      {
        double alpha;
        if( pg[i] != pg[i+1])
        {
          alpha = -pg[i].x / (pg[i+1].x - pg[i].x);
          return ((alpha >= 0) && (alpha <= 1));
        }
        else
        {
          return ( pg[i] == point(0,0));
        }         
      }     
    }
  }
  return 0;
}


//--------------------------------------------------------------------------
// Test whether the point is a vertex of a polygon
//--------------------------------------------------------------------------
int Point2d::isVertex(Poly2d &pgon)
{
	// for all vertices of polygon check ...
	for (int i=0; i<pgon.getSize(); i++)
	{
		if ((*this) == pgon[i])
			return(i+1);
	}
    return(0);
} // isVertex


//--------------------------------------------------------------------------
// Norm of a vector point
//--------------------------------------------------------------------------
Point2d Point2d::unit()
{
  return (*this) / (double) (*this).norm();
}

//--------------------------------------------------------------------------
// Distance-squared to a point
//--------------------------------------------------------------------------
double Point2d::distanceSqrTo( Point2d p)
{
  Point2d temp = *this - p;

  return temp.normSqr();
}

//--------------------------------------------------------------------------
// Distance to a point
//--------------------------------------------------------------------------
double Point2d::distanceTo( Point2d p)
{
  return sqrt( distanceSqrTo(p));
}

Point2d& Point2d::rotate( double a)
{
  double cosA = cos(a);
  double sinA = sin(a);
  Point2d tmp( x, y);
  x = tmp.x * cosA - tmp.y * sinA;
  y = tmp.x * sinA + tmp.y * cosA;
  return *this;
}

double Point2d::angle( const Point2d& p)
{
  return( atan2( (*this)^p, (*this)*p));
}

Point2d& Point2d::centerOfMass( Poly2d& p)
{
  x = y = 0.0;
  int psz = p.getSize();
  for( int i = 0; i < psz; i++)
  {
    x += p[i].X();
    y += p[i].Y();
  }
  if (psz > 0) {
	  x /= psz;
	  y /= psz;
  }
  return *this;
}

void Point2d::print()
{
  cout << "\n"
       << "Location (x, y) = (" << x << ", " << y << ")"
       << "    "
       << "Value (u, v, z) = (" << u << ", " << v << ", " << z << ")" << endl;
}

int Point2d::read(ifstream& fp)
{
  if(!fp) return 0;
  fp>>x;
  if(!fp) return 0;
  fp>>y;

  return 1;
}

int Point2d::write(ofstream& fp)
{
  if(!fp) return 0;
  fp<<x<<" ";
  if(!fp) return 0;
  fp<<y<<"\n";

  return 1;
}

int Point2d::read(string& filein)
{
  ifstream ifs;
  ifs.open(filein, ios::in);
  if( ! ifs) {
    cout << "\n  Could not open file " << filein 
         << " for input!" << endl;
    return 0;
  }
  if( ! read( ifs)) {
    cout << "\n  Error in reading file " << filein << endl;
    return 0;
  }

  ifs.close();

  return 1;
}

int Point2d::write(string& fileout)
{
  ofstream ofs;
  ofs.open(fileout, ios::out);
  if( ! ofs) {
    cout << "\n  Error: Could not open file " << fileout 
         << " for output!" << endl;
    return 0;
  }
  if( ! write( ofs)) {
    cout << "\n  Error in writing file " << fileout << endl;
    return 0;
  }

  ofs.close();

  return 1;
}

