/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* A. Tanju Erdem (Eastman Kodak Company)                             */
/* Candemir Toklu (University of Rochester)                           */
/*                                                                    */
/* and edited by                                                      */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* Eastman Kodak Company and the University of Rochester retain full  */
/* right to use the code for their own purposes, assign or donate the */
/* code to a third party and to inhibit third parties from using the  */
/* code for non MPEG-4 Visual (ISO/IEC 14496-2) conforming products.  */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1994-1997                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         polyTessClass.cpp
   AUTHOR       A. Tanju Erdem - Eastman Kodak Company and Candemir Toklu - 
                University of Rochester.
   DATE         September 1994
   UPDATE       14 August 1996 by Candemir Toklu
   				- various
   				Peter Van Beek, Fri Oct 18 13:09:46 1996
				- added member function neighbors
   				Peter Van Beek, Fri May  1 16:08:26 1998 
				- added quantize method
   DESCRIPTION  Member functions for Poly2dTess class are defined in this
                file.

   SYNOPSIS     Include files : 
                                
                #include "polyTessClass.h"
                #include "general.h"
                              
*********************************************************************-*/

#include "polyTessClass.h"
#include "general.h"


//------------------------------------------------------------------------

Poly2dTess::Poly2dTess()  // Empty tesselation
{
  sizeTess = 0;
  tess = NULL;
}

//------------------------------------------------------------------------
Poly2dTess::Poly2dTess( const Poly2d& p)  // A tesselation with a single polygon
{
  if( p.size == 0) {  // Empty tesselation
    sizeTess = 0;
    tess = NULL;
  }
  else {
    sizeTess = 1;
    tess = new Poly2dRef( p.size);  // corners = neighbors = NULL
    Pt2dLink *tmpLink;
    for( int i = 0; i < p.size; i++)
    {
      tmpLink = points.add( p.poly[i]);  // Add to the list of points
      tess->corner[i] = tmpLink;
      tmpLink->owner = (Poly2dRef **) new long;  // Only one owner polygon
      tmpLink->owner[0] = tess;  // The one owner is the newly defined polygon
    }
    tess->next = tess->prev = tess;  // Tesselation is closed

// A single polygon has no neighbors
  }
}

//------------------------------------------------------------------------
// Initialization with another tesselation
Poly2dTess::Poly2dTess(const Poly2dTess& from)
//: points( from.points)  // Copy the list of points
{
  sizeTess = 0;
  tess = NULL;
  *this = from;
}

//------------------------------------------------------------------------
Poly2dTess::~Poly2dTess()
{
  Poly2dRef *tmpRef;
  for( int i = 0; i < sizeTess; i++) {
    tmpRef = tess->prev;
    tmpRef->prev->next = tmpRef->next;
    tmpRef->next->prev = tmpRef->prev;
    delete tmpRef;  // Remove the current last reference from the list
  }
// The destructor for "points" is called automatically after this one.
}

//------------------------------------------------------------------------
Poly2dTess& Poly2dTess::destroy()
{
  Pt2dLink* tmpLink;
  Poly2dRef* tmpRef;
  int i;

  for( i = 0; i < sizeTess; i++) {
    tmpRef = tess->prev;
    tmpRef->prev->next = tmpRef->next;
    tmpRef->next->prev = tmpRef->prev;
    delete tmpRef;  // Remove the current last reference from the list
  }

  for( i = 0; i < points.size; i++) {
    tmpLink = points.list->prev;
    tmpLink->prev->next = tmpLink->next;
    tmpLink->next->prev = tmpLink->prev;
    delete tmpLink;  // Remove the current last point-link from the list
  }
  tess = NULL;
  sizeTess = 0;
  points.list = NULL;
  points.plast = NULL;
  points.ilast = 0;
  points.size = 0;
  return *this;
}

//------------------------------------------------------------------------
// Setting equal to another tesselation
Poly2dTess& Poly2dTess::operator=(const Poly2dTess& from) 
{
  if( this == &from) return *this;  // Cannot copy a tesselation onto itself

// Destruct the old tesselation
  int i;
  Poly2dRef *tmpRef;
  for( i = 0; i < sizeTess; i++)
  {
    tmpRef = tess->prev;
    tmpRef->prev->next = tmpRef->next;
    tmpRef->next->prev = tmpRef->prev;
    delete tmpRef;  // Remove the current last reference from the list
  }

// Copy the new list of points
  points = from.points;  // The owners, if any, are not copied at this stage

//  points.print();

// Then, construct a new tesselation that is identical to the given one
  if( from.sizeTess == 0)
  {
    tess = NULL;
    sizeTess = 0;
  }
  else
  {
// Create the list of references
    Poly2dRef *newRef;
    Poly2dRef *tmpRef = from.tess;
    tess = new Poly2dRef( tmpRef->sizeRef);  // Construct the 1st ref. in the list
    tess->sizeRef = tmpRef->sizeRef;
    tess->next = tess->prev = tess;  // Close the list
    for( i = 1; i < from.sizeTess; i++)
    {
      tmpRef = tmpRef->next;
      newRef = new Poly2dRef( tmpRef->sizeRef);
      newRef->prev = tess->prev;
      newRef->next = tess;
      tess->prev->next = newRef;
      tess->prev = newRef;
    }

    sizeTess = from.sizeTess;
    int j;

// Fill in the references
    tmpRef = from.tess;
    newRef = tess;
    for( i = 0; i < from.sizeTess; i++, tmpRef = tmpRef->next, newRef = newRef->next)
    {
      for( j = 0; j < tmpRef->sizeRef; j++)
      {
// Connect polygon to its corners
        newRef->corner[j] = addressC(from.where(tmpRef->corner[j]));
      }
    }

    Pt2dLink *tmpLink = from.points.list;
    Pt2dLink *newLink = points.list;
    for( i = 0; i < from.points.size; i++, tmpLink = tmpLink->next, newLink = newLink->next)
    {
      newLink->owner = (Poly2dRef **) new long[tmpLink->sizeLink];
      newLink->sizeLink = tmpLink->sizeLink;
      newLink->base = tmpLink->base; // C.T.
      newLink->mult = tmpLink->mult; // C.T.
      for( j = 0; j < tmpLink->sizeLink; j++)
      {
// Connect corner point to its owner polygon
        newLink->owner[j] = addressR(from.where(tmpLink->owner[j]));
      }
    }
  }
  return *this;
}

//--------------------------------------------------------------------------
Poly2dTess& Poly2dTess::append( Poly2d& p)
{
  Poly2dRef *newRef = new Poly2dRef( p.getSize());  // corners = neighbors = NULL
  Pt2dLink *tmpLink;
  Poly2dRef **newOwner;
  for( int i = 0; i < p.getSize(); i++)
  {
// Connect polygon to its corners
    tmpLink = points.add( p[i]);  // Add if not already in the list
    newRef->corner[i] = tmpLink;
// Connect corner point to its owner polygon
    newOwner = (Poly2dRef **) new long[tmpLink->sizeLink + 1];
    for( int j = 0; j < tmpLink->sizeLink; j++)
    {
      newOwner[j] = tmpLink->owner[j];  // Old owners
    }
    newOwner[tmpLink->sizeLink] = newRef;  // A new owner
    if( tmpLink->sizeLink) delete[] tmpLink->owner;  // Delete old pointers
    tmpLink->owner = newOwner;
    tmpLink->sizeLink += 1;
  }
  if( tess)
  {  // Add the new reference to the existing tesselation
    newRef->next = tess;
    newRef->prev = tess->prev;
    tess->prev->next = newRef;
    tess->prev = newRef;
  }
  else
  {  // Tesselation is empty
    tess = newRef;
    tess->next = tess->prev = tess;
  }
  sizeTess++;
  return *this;
}

//--------------------------------------------------------------------------
Poly2dTess& Poly2dTess::move( double dx, double dy)
{
   for( int i = 0; i < points.size; i++)
   {
       points[i].move(dx,dy);
   }
   return *this;
}

//--------------------------------------------------------------------------
Poly2dTess& Poly2dTess::move( const Point2d& p)
{
   for( int i = 0; i < points.size; i++)
   {
       points[i]+=p;
   }
   return *this;
}

//--------------------------------------------------------------------------
Poly2dTess& Poly2dTess::scale( double dx, double dy)
{
   for( int i = 0; i < points.size; i++)
   {
       points[i].scale(dx,dy);
   }
   return *this;
}

//--------------------------------------------------------------------------
// Quantize point coordinates with interval 'quant'
Poly2dTess& Poly2dTess::quantize( double quant)
{
   Point2d pt;

   for (int i = 0; i < points.size; i++)
   {
	   // get i'th point; divide, round and multiply ...
	   pt = points[i] / quant;
	   pt.roundToInt();
       points[i] = pt * quant;
   }
   return *this;
}

//--------------------------------------------------------------------------
Point2d& Poly2dTess::point( int indx)
{
  if( indx < 0 || indx >= points.size)
  {
    cout << "\nError: Illegan point index for the tesselation" << endl;
    exit(1);
  }

  return points[indx];
}

//--------------------------------------------------------------------------
Point2d& Poly2dTess::point( int indx, double& base, double& mult)
{

  if( points.size == 0) {
    cout << "\n  Error: Cannot index to an empty list!" << endl;
    exit( 1);
  }
  else if( indx < 0 || indx >= points.size) {
    cout << "\n  Error: Index value is out of range for the point list!"
         << endl;
    exit( 1);
  }
 
  {
    int i;
    Pt2dLink *tmpLink = points.plast;
    if( indx > points.ilast)
       for( i = 0; i < indx-points.ilast; i++) tmpLink = tmpLink->next;
    else
       for( i = 0; i < points.ilast-indx; i++) tmpLink = tmpLink->prev;
    points.plast = tmpLink;
    points.ilast = indx;
    base = tmpLink->base; // C.T.
    mult = tmpLink->mult; // C.T.
    return( tmpLink->here);
  }
}

//--------------------------------------------------------------------------
Poly2d Poly2dTess::polygon( int indx)
{
  if( indx < 0 || indx >= sizeTess)
  {
    cout << "\nError: Illegal polygon index for the tesselation" << endl;
    exit(1);
  }

  int i;

//  Move to the requested polygon
  Poly2dRef *tmp = tess;
  for( i = 0; i < indx; i++)
  {
    tmp = tmp->next;
  }

//  Build the polygon
  Poly2d p(tmp->sizeRef);
  for( i = 0; i < tmp->sizeRef; i++)
  {
    p[i] = tmp->corner[i]->here;
  }

  return p;
}


//--------------------------------------------------------------------------
Poly2dTess& Poly2dTess::reOrderOwners()
{
  Poly2dRef **newOwner;
  Point2d  *tmpList;
  Pt2dLink *tmpLink = points.list;
  Point2d center, tmpPt;
  double *numbers;
  int *newOrder;
  int  refSize, linkSize;
  int i,j;
  int Size = 1;

   do
   {
     if( tmpLink->sizeLink > Size)
        Size = tmpLink->sizeLink;
     tmpLink = tmpLink->next;
   } while(  tmpLink != points.list);
  
//  tmpLink = points.list;
  newOwner = (Poly2dRef **) new long[Size];
  tmpList = (Point2d *) new Point2d[Size];
  numbers = (double *) new double[Size];
  newOrder = (int *) new int[Size];

  do
  { 
     if( tmpLink->sizeLink > 1)
     {
       linkSize = tmpLink->sizeLink;
       center = tmpLink->here;
       for( i = 0; i< linkSize; i++)
       {
         tmpList[i] = ::point( 0,0);
         refSize = tmpLink->owner[i]->sizeRef;
         for( j = 0; j < refSize; j++)
           tmpList[i] += tmpLink->owner[i]->corner[j]->here;
         if( refSize) tmpList[i] /= refSize;
         else
         {
           cout << "\nWarning:Reference Size is zero and the reference is"
                << " a owner. Ordering continued." << endl;
         }
       }
 
       for( i = 0; i< linkSize; i++)
       {
         tmpPt = (tmpList[i] - center).unit();
         numbers[i] = Sign1at0( tmpPt.Y()) * (1 + tmpPt.X());
         newOrder[i] = i;
       }
       sortNumbers( numbers, newOrder, linkSize);
       for( i = 0; i < linkSize; i++)
         newOwner[linkSize - 1 - i] = tmpLink->owner[newOrder[i]];
       for( i = 0; i < linkSize; i++)
         tmpLink->owner[i] = newOwner[i];
     }
     tmpLink = tmpLink->next;
  } while( tmpLink != points.list);

  delete[] newOwner;
  delete[] tmpList;
  delete[] numbers;
  delete[] newOrder;
 
  return *this;
}

//--------------------------------------------------------------------------
int Poly2dTess::where( Pt2dLink* pLink) const
{
  return points.where( pLink);
}

//--------------------------------------------------------------------------
int Poly2dTess::where( Poly2dRef* pRef) const
{
  if( sizeTess == 0) {
    return -1;  // The tesselation is empty
  }
  else {
    int indx = 0;

    Poly2dRef* tmpRef = tess;
    while( tmpRef != pRef && indx < sizeTess) {
      indx++;
      tmpRef = tmpRef->next;
    }
    if( indx == sizeTess) {  // Polygon reference is not found in the list
      return -1;
    }
    else {
      return indx;   // Return the location of the polygon reference in the tess.
    }
  }
}

//--------------------------------------------------------------------------
int Poly2dTess::where( Point2d& p) const
{
  return points.where( p);
}

//--------------------------------------------------------------------------
Pt2dLink* Poly2dTess::addressC( int indx)
{
  return points.address( indx);
}

//--------------------------------------------------------------------------
Pt2dLink* Poly2dTess::addressC( const Point2d p)
{
  return points.address( p);
}

//--------------------------------------------------------------------------
Poly2dRef* Poly2dTess::addressR( int indx)
{
  if( sizeTess == 0) {
    cout << "\n  Error: Cannot index to an empty list!" << endl;
    exit( 1);
  }
  else if( indx < 0 || indx >= sizeTess) {
    cout << "\n  Error: Index value is out of range!" << endl;
    exit( 1);
  }
  else {
    int i;
    Poly2dRef *tmpRef = tess;
    for( i = 0; i < indx; i++) tmpRef = tmpRef->next;
    return tmpRef;
  }
  return(NULL);
}


//--------------------------------------------------------------------------
int Poly2dTess::getSizeLink( int indx)
{
  if( indx < 0 || indx >= points.size)
  {
    return -1;
  }
  else 
  {
    return addressC( indx)->sizeLink;
  }
}

//--------------------------------------------------------------------------
int Poly2dTess::getSizeRef( int indx)
{
  if( indx < 0 || indx >= sizeTess)
  {
    return -1;
  }
  else 
  {
    return addressR( indx)->sizeRef;
  }
}


//--------------------------------------------------------------------------
void Poly2dTess::markRefs(const char m)
{
	if (sizeTess != 0) {
		Poly2dRef* tmpRef = tess;
		do {
			tmpRef->mark = m;
			tmpRef = tmpRef->next;
		} while (tmpRef != tess);
	}
}


//--------------------------------------------------------------------------
void Poly2dTess::markLinks(const char m)
{
	points.markLinks(m);
}


//--------------------------------------------------------------------------
int Poly2dTess::read(ifstream& fp)
{
  Pt2dLink* tmpLink;
  Poly2dRef* tmpRef;
  double x,y;
  int i, j, index;
  
  // First, destruct the old tesselation. 
  // this->Poly2dTess();
  destroy();

  if(!fp) return 0;
  fp>>points.size;
  if(!fp) return 0;
  fp>>sizeTess;

  Poly2dRef *newRef;
  // Construct the 1st ref. in the list.
  tess = (Poly2dRef*) new Poly2dRef();
  tess->next = tess->prev = tess; // Close the list
  for( i = 1; i < sizeTess; i++)
  {
     newRef = (Poly2dRef*) new Poly2dRef();
     newRef->prev = tess->prev;
     newRef->next = tess;
     tess->prev->next = newRef;
     tess->prev = newRef;
  }
  
  Pt2dLink *newLink;
  points.list = (Pt2dLink*) new Pt2dLink();
  points.list->next = points.list->prev = points.list;
  for( i = 1; i < points.size; i++)
  {
     newLink = (Pt2dLink*) new Pt2dLink();
     newLink->prev = points.list->prev;
     newLink->next = points.list;
     points.list->prev->next = newLink;
     points.list->prev = newLink;
  }
  points.plast = points.list->prev;
  points.ilast = points.size - 1;
  
  tmpLink = points.list;
  for( i = 0; i < points.size; i++, tmpLink = tmpLink->next)
  {
    if(!fp) return 0;
    fp>>x;
    if(!fp) return 0;
    fp>>y;
    tmpLink->here = ::point( x,y);
    if(!fp) return 0;
    fp>>tmpLink->sizeLink;
    tmpLink->owner = (Poly2dRef**) new long[tmpLink->sizeLink];
    for( j = 0; j < tmpLink->sizeLink; j++)
    {
      if(!fp) return 0;
      fp>>index;
      tmpLink->owner[j] = addressR( index); 
    }
  }

  tmpRef = tess;
  for( i = 0; i < sizeTess; i++, tmpRef = tmpRef->next)
  {
    if(!fp) return 0;
    fp>>tmpRef->sizeRef;
    if(!fp) return 0;
    tmpRef->corner = (Pt2dLink**) new long[tmpRef->sizeRef];
    for( j = 0; j < tmpRef->sizeRef; j++)
    {
      if(!fp) return 0;
      fp>>index;
      tmpRef->corner[j] = addressC( index); 
    }
  }

  return 1;
}

//--------------------------------------------------------------------------
int Poly2dTess::write(ofstream& fp)
{
  fp.precision(18);
  if(!fp) return 0;
  fp<<points.size<<"\n";
  if(!fp) return 0;
  fp<<sizeTess<<"\n"; 
  
  Pt2dLink* tmpLink = points.list;
  do
  {
    if(!fp) return 0;
    fp<<tmpLink->here.X()<<" "<<tmpLink->here.Y()<<" ";
    if(!fp) return 0;
    fp<<tmpLink->sizeLink<<" ";
    if(!fp) return 0;
    for( int j = 0; j < tmpLink->sizeLink; j++)
    {
      if(!fp) return 0;
      fp<<where( tmpLink->owner[j])<<" ";
    }
    if(!fp) return 0;
    fp<<"\n";
    tmpLink = tmpLink->next;
  } while( tmpLink != points.list);

  if(!fp) return 0;
  fp<<"\n";

  Poly2dRef* tmpRef = tess;
  do
  {
    if(!fp) return 0;
    fp<<tmpRef->sizeRef<<" ";
    if(!fp) return 0;
    for( int j = 0; j < tmpRef->sizeRef; j++)
    {
      if(!fp) return 0;
      fp<<where( tmpRef->corner[j])<<" ";
    }
    if(!fp) return 0;
    fp<<"\n";
    tmpRef = tmpRef->next;
  } while( tmpRef != tess);

  return 1;
}


//--------------------------------------------------------------------------
// read in the tesselation from fin
int Poly2dTess::read(char fin[])
{
  ifstream ifs;

  ifs.open(fin, ios::in);
  if( ! ifs) {
    cout << "\n  Could not open file " << fin 
         << " for input!" << endl;
    return 0;
  }
  if( ! read( ifs)) {
    cout << "\n  Error in reading file" << endl;
    return 0;
  }
  ifs.close();
  return 1;
}

//--------------------------------------------------------------------------
// save the tesselation to fout
int Poly2dTess::write(char fout[])
{
  ofstream ofs;

  ofs.open(fout, ios::out);
  if( ! ofs) {
    cout << "\n  Error: Could not open file " << fout 
         << " for output!" << endl;
    return 0;
  }
  if( ! write( ofs)) {
    cout << "\n  Error in writing file" << endl;
    return 0;
  }
  ofs.close();
  return 1;
}


//--------------------------------------------------------------------------
void Poly2dTess::print()
{
  int i, j;
  Pt2dLink *tmpLink = points.list;
  i = 0;
  cout << "\nThe list of points and their owners is as follows." << endl;
  do
  {
    cout << "\n" << i++ << ": " << tmpLink->here << " -> ";
    for( j = 0; j < tmpLink->sizeLink; j++)
    {
      cout << where( tmpLink->owner[j]);
      if( j != tmpLink->sizeLink - 1) cout << ", ";
      else cout << endl;
    }
    tmpLink = tmpLink->next;
  } while( tmpLink != points.list);

  Poly2dRef *tmpRef = tess;
  i = 0;
  cout << "\nThe list of polygons and their corners is as follows." << endl;
  do
  {
    cout << "\n" << i++ << ": [" ;
    for( j = 0; j < tmpRef->sizeRef; j++)
    {
      cout << tmpRef->corner[j]->here;
      if( j != tmpRef->sizeRef - 1) cout << ", ";
      else cout << "] -> ";
    }
    for( j = 0; j < tmpRef->sizeRef; j++)
    {
      cout << where( tmpRef->corner[j]);
      if( j != tmpRef->sizeRef - 1) cout << ", ";
      else cout << endl;
    }
    tmpRef = tmpRef->next;
  } while( tmpRef != tess);
}


//--------------------------------------------------------------------------
// get neighbors of a node in a triangular or quadrangular tesselation 
// returns a list of points - with the node itself in the front
Pt2dList Poly2dTess::neighbors (Point2d& node, int nodeNum)
{
	Pt2dLink *tmpLink = addressC(nodeNum);
	int linkSize = getSizeLink(nodeNum);
	Poly2dRef *tmpRef;
	Pt2dLink *nodePtr;
	Pt2dList pList;

	pList.add(node);					// add current node first

	for (int j = 0; j < linkSize; j++)	// for all owner polygons
	{
		tmpRef = tmpLink->address(j);
										// add all points to the list
		for (int k = 0; k < tmpRef->getSize(); k++) {
			nodePtr = tmpRef->address(k);
			pList.add(nodePtr->getPoint());
		}
	}
	return pList;						// return pointlist
}


// Construct a uniform triangular mesh
Poly2dTess& Poly2dTess::makeUniform(int nhori, int nvert, double hsize, double vsize, int splitcode) 
{
	int i, j, n, nh = 0, nv, split = 0;
	double hpos, vpos;
	int ntotal = nhori*nvert;
	Pt2dLink  *pLink;
	Pt2dLink  **arrLink;
	Poly2dRef *tmpRef, *newRef, *nextRef;

	// Construct the point list from an array of Pt2dLinks
	arrLink = new Pt2dLink*[ntotal];
	for (int t = 0; t < ntotal; ++t)
	    arrLink[t] = new Pt2dLink;
	points.size  = ntotal;
	points.list  = points.plast = arrLink[0];
	points.ilast = 0;
	// Fill the values of each Pt2dLink
	vpos = 0.0;
	for (nv = 0, n = 0; nv < nvert; nv++, vpos+=vsize) {
		hpos = 0.0;
		for (nh = 0; nh < nhori; nh++, n++, hpos+=hsize) {
			(*arrLink[n]).here = Point2d(hpos, vpos);
			(*arrLink[n]).next = arrLink[(n+1)%ntotal];
			(*arrLink[n]).prev = arrLink[(n-1+ntotal)%ntotal];
		}
	}

	// Construct the polygon list from Poly2dRefs
	tmpRef = new Poly2dRef(3); 				// create first triangle
	sizeTess = 2*(nv-1)*(nh-1);
	tess = tmpRef;
	tess->next = tess->prev = tess;			// close the list
    for(i = 1; i < sizeTess; i++)
    {
		newRef = new Poly2dRef(3);
		newRef->prev = tess->prev;
		newRef->next = tess;
		tess->prev->next = newRef;
		tess->prev = newRef;
    }
	// Fill in the references to points and link these back
	tmpRef = tess;
	for (nv = 1; nv < nvert; nv++) {
		for (nh = 1; nh < nhori; nh++) {
			// decide local splitting direction in this rectangle
			switch (splitcode) {
			  case 0:
				split = 0; break;
			  case 1:
				split = 1; break;
			  case 2:
				split = ((nh & 1)? 0 : 1); break;
			  case 3:
				split = ((nh & 1)? 1 : 0); break;
			}
			// specify the two triangles lying inside one rectangle
			// put vertices in clockwise order!
			nextRef = tmpRef->next; 		// the next triangle
			switch (split) {
			  case 0:
				tmpRef->corner[0]  = arrLink[(nv-1)*nhori+nh-1];
				tmpRef->corner[1]  = arrLink[nv*nhori+nh];
				tmpRef->corner[2]  = arrLink[nv*nhori+nh-1];
				nextRef->corner[0] = arrLink[(nv-1)*nhori+nh-1];
				nextRef->corner[1] = arrLink[(nv-1)*nhori+nh];
				nextRef->corner[2] = arrLink[nv*nhori+nh];
				break;
			  case 1:
				tmpRef->corner[0]  = arrLink[(nv-1)*nhori+nh-1];
				tmpRef->corner[1]  = arrLink[(nv-1)*nhori+nh];
				tmpRef->corner[2]  = arrLink[nv*nhori+nh-1];
				nextRef->corner[0] = arrLink[nv*nhori+nh-1];
				nextRef->corner[1] = arrLink[(nv-1)*nhori+nh];
				nextRef->corner[2] = arrLink[nv*nhori+nh];
				break;
			}
			// now link the referenced points back to each triangle
			Poly2dRef **newOwner;
			for (i = 0; i < 3; i++) {
				pLink = tmpRef->corner[i];
				// copy the old owners of this point and add new one
				newOwner = (Poly2dRef **) new long[pLink->sizeLink + 1];
				for (j = 0; j < pLink->sizeLink; j++)
					newOwner[j] = pLink->owner[j];  	// old owners
				newOwner[pLink->sizeLink] = tmpRef;  	// the new owner
				// now delete the old owners 
				if (pLink->sizeLink) delete [] pLink->owner;  
				pLink->owner = newOwner;
				pLink->sizeLink += 1;
			}
			for (i = 0; i < 3; i++) {
				pLink = nextRef->corner[i];
				// copy the old owners of this point and add new one
				newOwner = (Poly2dRef **) new long[pLink->sizeLink + 1];
				for (j = 0; j < pLink->sizeLink; j++)
					newOwner[j] = pLink->owner[j];  	// old owners
				newOwner[pLink->sizeLink] = nextRef;  	// the new owner
				// now delete the old owners 
				if (pLink->sizeLink) delete [] pLink->owner;  
				pLink->owner = newOwner;
				pLink->sizeLink += 1;
			}

			tmpRef = nextRef->next;		// two triangles further
		}
	}

    delete [] arrLink;

	// I think that is all
	return *this;
}



// end of file
