/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Multiplexer
FILE:       JPEG.cpp
AUTHOR:     Matthew Leditschke (CSELT)
OVERVIEW:   Implementation of JPEGStream, the class that handles JPEG
input streams.
-------------------------------------------------------------------------*/
#include "defs.h"
#include "JPEG.h"
#include "odMux.h"

/*-----------------------------------------------------------------
JPEGStream constructor determines how the time between JPEG images
will be calculated, depending on the pDescriptor -> streamData.
  - if not given, defaults to 1 second.
  - if an floating point number is given, this gives interval
  - otherwise the pDescriptor -> streamData is the name of a file to read
    presentation times from.  Once the end of this file is reached,
	the last difference is used.
-----------------------------------------------------------------*/
JPEGStream :: JPEGStream (FILE *fStream, ESDesc *pDescriptor) :
	Stream (fStream, pDescriptor), m_bRAP (true)
{
	m_timeFile = NULL;
	m_prevChar[0] = 0;
	m_prevChar[1] = 0;
	m_readAhead = FALSE;

	// Handle the default when the pDescriptor -> streamData field is not given
	if ((pDescriptor -> streamData == NULL) || (pDescriptor -> streamData[0] == 0))
	{
		m_constant = TRUE;
		m_nextTime = 0;
		m_intervalTime = 1;
	}

	// Check if the stream data is numeric
	else if (isdigit(pDescriptor -> streamData[0]))
	{
		m_constant = TRUE;
		m_nextTime = 0;
		m_intervalTime = atof(pDescriptor -> streamData);
		if (m_intervalTime < 0.0001)
			throw ExceptionInvalidData (
				"visible/JPEG: The display interval time must be greater than 0.\n"
			);
	}

	// Otherwise assume that the pDescriptor -> streamData is the name of a file
	else
	{
		m_timeFile = fopen(pDescriptor -> streamData, "r");
		if (m_timeFile == NULL)
		{	static char szBuffer [150];
			sprintf(szBuffer, "visible/JPEG: Error opening '%s'.\n", pDescriptor -> streamData);
			throw ExceptionInvalidData (szBuffer);
		}
		m_constant = FALSE;
		m_prevTime[0] = 0;
		m_prevTime[1] = 0;
	}
}

/*-----------------------------------------------------------------
JPEGStream destructor closes the presentation time file if it is
still open.
-----------------------------------------------------------------*/
JPEGStream :: ~JPEGStream ()
{
	if (m_timeFile)
		fclose (m_timeFile);
}

/*-----------------------------------------------------------------
Time stamps are either calculated using a fixed increment, or they
are read from a file.
-----------------------------------------------------------------*/
long JPEGStream :: GetDTS ()
{
	float fTime;

	// Return -1 if the end of the file has been reached
	if (m_prevChar[0] == EOF)
		return -1;

	if (m_constant)
	{
		fTime = m_nextTime;
		m_nextTime += m_intervalTime;
	}

	else
	{
		// Read the next presentation time (in seconds) from the file
		int res = fscanf(m_timeFile, "%f", &fTime);

		// If the end is reached, switch to fixed increments
		if (res == EOF)
		{
			// The increment is the difference in the last two times
			m_constant = TRUE;
			if (m_prevTime[0] == m_prevTime[1])
				m_intervalTime = 1;
			else
				m_intervalTime = m_prevTime[0] - m_prevTime[1];
			fTime = m_prevTime[0] + m_intervalTime;
			m_nextTime = m_prevTime[0] + 2 * m_intervalTime;
		}
		else
		{
			// Keep track of the previous two times
			m_prevTime[1] = m_prevTime[0];
			m_prevTime[0] = fTime;

			//Abort if the times aren't always increasing
			if (m_prevTime[1] >= m_prevTime[0])
			{
				fprintf(stderr, "visible/JPEG: The presentation time aren't always increasing.\n");
				exit(1);
			}
		}
	}

	// Convert the time to milliseconds
	return (m_lDTS = (long)(fTime * 1000.0));
}

/*-------------------------------------------------------------
In this version only the first AU is RAP
-------------------------------------------------------------*/
bool JPEGStream :: IsRAP ()
{
	bool bResult = m_bRAP;
	m_bRAP = false;
	return bResult;
}

/*-----------------------------------------------------------------
GetChar returns a character and reads one next char. We are always
3 chars ahead, which are used to check whether this is end of AU.
-----------------------------------------------------------------*/
int JPEGStream :: GetChar ()
{
/*	This check is now disabled. A JPEGStream can only consist of a single picture

	// If the previous two characters read were the end code of a
	// picture, then return -1.
	if ((m_prevChar[1] == 0xFF) && (m_prevChar[0] == 0xD9))
	{
		m_prevChar[0] = 0; 
		m_prevChar[1] = 0;

		// Before returning -1, we need to check if reading the
		// next byte while reach the end of the file
		m_nextChar = getc();
		m_readAhead = TRUE;
		if (m_nextChar == EOF)
			m_prevChar[0] = m_nextChar;

		return -1;
	}
*/
	// Otherwise read in a new character (allowing for possible
	// read ahead), and update the list of the past two characters
	// read.
	int c = (m_readAhead) ? m_nextChar : getc();
	m_readAhead= FALSE;
	m_prevChar[1] = m_prevChar[0];
	m_prevChar[0] = c;

	return c;
}

static StreamFactory<JPEGStream> jpegFactory ("JPEG");