// -*- Mode: c++; tab-width: 4; c-basic-offset: 4; -*-

// This software module was adapted by Christopher Lee (MERL) 
// from DIBR/DepthImageProxy.cpp by Karl Anders ygard (Telenor R&D),
// and PointShop3D 2.0 by Richard Keiser, Oliver Knoll, Mark Pauly,
// and Matthias Zwicker (Computer Graphics Lab, ETH Zurich),
// in the course of development of the MPEG-4 Systems standard
// (ISO/IEC 14496-1).  This software module is an implementation
// of a part of one or more MPEG-4 Systems tools as specified by the MPEG-4
// Systems standard.  ISO/IEC gives users of the MPEG-4 Systems standard
// free license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4 Systems
// standard.  Those intending to use this software module in hardware or
// software products are advised that its use may infringe existing
// patents.  The original developer of this software module and his/her
// company, the subsequent editors and their companies, and ISO/IEC have no
// liability for use of this software module or modifications thereof in an
// implementation.  Copyright is not released for non MPEG-4 Systems
// conforming products.  The original developer retains full right to use
// the code for his/her own purpose, assign or donate the code to a third
// party and to inhibit third parties from using the code for non MPEG-4
// Systems conforming products.  This copyright notice must be included in
// all copies or derivative works.
//
// Copyright 2004.
//
// PointShop3D is available separately, under the GNU General Public License,
//  at http://graphics.ethz.ch/pointshop3d/index.html

#include "../../OpenGl.h"
#include "GL/gl.h"
#include "stream.h"
#include "DepthImageProxyV2.h"
#include "SimpleTextureProxyV2.h"
#include "../../appearance/ImageTextureProxy.h"
#include "PointTextureProxyV2.h"
#include "PS3D/Datatypes/MyDataTypes.h"
#include "PS3D/DIBR2-compat.h"
#include "PS3d/DataTypes/Matrix.h"
#include "../../base/OpenGLVisualRenderer.h"

#define NEAR_CLIPPING 0.5
//#define FAR_CLIPPING  400
#define FAR_CLIPPING  4000000
#define DEFAULT_FOV   45.0


// Render object description

/* virtual */ void
DepthImageProxyV2::RenderDescription (OpenGLVisualRenderer * pRenderer,
									  DWORD /* dwTime */ )
{
	RenderDescriptionBox (pRenderer, "DepthImageV2");
}

// Render object

/* virtual */ BOOL
DepthImageProxyV2::Render (const Matrix & /*rTransformation*/,
						   const CullFlag& /* rCullFace*/,
						   OpenGLVisualRenderer * pRenderer, DWORD /*dwTime*/)
{
	DepthImageV2 *pDepthImage = GetDepthImage ();

	if (pDepthImage->diTexture.GetNode () == NULL) {
		return FALSE;
	}
	// Get parameters in DepthImage node
	double nearPlane = pDepthImage->nearPlane;
	double farPlane = pDepthImage->farPlane;

	Vec2f FOV = pDepthImage->fieldOfView.GetValue ();
	double FOVx = FOV.x;
	double FOVy = FOV.y;
	Vec3f position = pDepthImage->position.GetValue ();
	Rotation orientation = pDepthImage->orientation.GetValue ();
	Bool ortho = pDepthImage->orthographic.GetValue ();
	BOOL orthographic = bool (ortho);


	// Determine whether this node has a previous and/or next node that is
	//  also a DepthImageProxyV2....
	checkPrevAndNextNodes();

	if ( m_surfelIterator == NULL ) {
		int NodeCode = pDepthImage->diTexture.GetNode ()->GetNodeCode ();
		if ( NodeCode == CODE_SimpleTextureV2 ) {
			SimpleTextureV2 *pST =
				(SimpleTextureV2 *) pDepthImage->diTexture.GetNode ();
			SimpleTextureProxyV2 *pSTP =
				(SimpleTextureProxyV2 *) pST->GetRenderingObject ();
			if ( pSTP->GetAnim() ) {
				if ( ! pSTP->GetImage (m_nCurrentListIdx) ) {
					return FALSE;
				}
				m_nCurrentListIdx = (1 + m_nCurrentListIdx) % m_nNumberOfFrame;
			} else {
				if ( ! pSTP->GetImage (0) ) {
					return FALSE;
				}
			}
			if (orthographic) {
				m_surfelIterator = pSTP->getSurfelIterator(-1,-1);
			} else {
				m_surfelIterator = pSTP->getSurfelIterator(pDepthImage->nearPlane,
														   pDepthImage->farPlane);
			}
		} else if (NodeCode == CODE_PointTextureV2) {
			PointTextureV2 *pPT =
				(PointTextureV2 *) pDepthImage->diTexture.GetNode ();
			PointTextureProxyV2 *pPTP =
				(PointTextureProxyV2 *) pPT->GetRenderingObject ();

			NodeField *pNF0 = (NodeField *) pPT->GetField (2);
			NodeField *pNF1 = (NodeField *) pPT->GetField (3);
			NodeField *pNF2 = (NodeField *) pPT->GetField (7);

			pPTP->SetDepthBitPerPixel (((SF < int >*) pNF0)->GetValue () + 1);
			pPTP->SetMaxDepth (pow (2, pPTP->GetDepthBitPerPixel ()) - 1);
			pPTP->SetHeight (((SF < int >*) pNF1)->GetValue ());
			pPTP->SetWidth (((SF < int >*) pNF2)->GetValue ());

			m_surfelIterator = pPTP->getSurfelIterator(pDepthImage->nearPlane,
													   pDepthImage->farPlane);
		}
	}

	// Initial transform according to the information in DepthImage
	glPushMatrix ();
	glTranslatef (position.x, position.y, position.z);
	glRotatef (orientation.theta * 180 / M_PI, orientation.x, orientation.y,
			   orientation.z);

	if (orthographic) {
		glTranslatef (0, 0, -nearPlane);
		glScalef (FOVx, FOVy, farPlane - nearPlane);
	} else {
		glTranslatef (0, 0, -nearPlane);
		double x = tan (FOVx / 2) * farPlane * 2;
		double y = tan (FOVy / 2) * farPlane * 2;
		double z = farPlane - nearPlane;
		glScalef (x, y, z);
	}

	QSize viewPortSize;
 	{
		GLint viewport[4];

		glGetIntegerv (GL_VIEWPORT, viewport);
		viewPortSize.width = viewport[2];
		viewPortSize.height  = viewport[3];
	}

	if ( prevDepthImageProxy ) {
		frameBuffer = prevDepthImageProxy->GetFrameBuffer ();
		surfelContext = prevDepthImageProxy->GetSurfelContext ();
	} else {
		if ( frameBuffer ) {
			QSize oldsz = frameBuffer->getSize();
			if ( oldsz.height != viewPortSize.height
				|| oldsz.width != viewPortSize.width ) {
				// free the old light
				SrfFreeLight (surfelContext, lightSourceHandle);
				SrfFree (surfelContext);
				frameBuffer->setSize(viewPortSize);
				surfelContext = NULL;
			}
		}
		if ( ! frameBuffer ) {
			frameBuffer = new FrameBuffer (viewPortSize);
		}
		if ( ! surfelContext ) {
			// instantiate the surfel rendering context
			surfelContext = SrfNew (frameBuffer, 1 /*superSampling*/);

			// light sources
			lightSourceHandle = SrfNewLight (surfelContext);
			SrfEnableLight (surfelContext, lightSourceHandle);

			{
				Vector3D lightDirection(0.0, 0.0, -1.0); // From Configuration.cpp
				SrfSetLightDirection (surfelContext,
									  lightSourceHandle, lightDirection);
			}

			// Defaults from SurfelGPRendererConfiguration via SurfelGPRenderer
			SrfDisableZBufferOption (surfelContext, SRF_BIDINORMALS);

			// SrfGPRenderer usese shading by default
			SrfEnable (surfelContext, SRF_SHD_LIGHTING);

			SrfSetBlendingThresholds (surfelContext, // SurfelGRenderer defaults:
									  0.4f,			 // constThreshold
									  2.0f,			 // distThreshold
									  8.0f);		 // angleThreshold
		}
	}
	{
		double aspectRatio = ( ((double)viewPortSize.width )
							   /((double)viewPortSize.height));
		double fieldOfView = DEFAULT_FOV;

		SrfSetFrustum (surfelContext,
					   fieldOfView,
					   aspectRatio,
					   NEAR_CLIPPING,
					   FAR_CLIPPING);

		GLdouble model_matrix[16];
		MyDataTypes::TransformationMatrix16f modelMatrix;
		MyDataTypes::TransformationMatrix16f modelMatrixTransp;

		glGetDoublev (GL_MODELVIEW_MATRIX, model_matrix);

		for ( int i = 0; i < 16; i++ ) {
			modelMatrix[i] = (float) model_matrix[i];
		}
		// make it a row major matrix
		MtrTranspose4x4f (modelMatrix,
						  modelMatrixTransp);

		// Lefthanded -> Righthanded
		modelMatrixTransp[8]  = -1.0f * modelMatrixTransp[8];
		modelMatrixTransp[9]  = -1.0f * modelMatrixTransp[9];
		modelMatrixTransp[10] = -1.0f * modelMatrixTransp[10];
		modelMatrixTransp[11] = -1.0f * modelMatrixTransp[11];
		SrfSetTrafo (surfelContext, modelMatrixTransp);
	}
	if ( ! prevNodeIsDepthImage ) {
		SrfBeginFrame (surfelContext);
		SrfSetSplatFunction (surfelContext, SRF_ELLIPTICALSURFEL);
	}
	int bbox[4];

	// bbox returns a bounding box containing all pixels that were hit during warping
	SrfWarp (surfelContext, m_surfelIterator, bbox);
	m_surfelIterator->reset();

	if ( ! nextNodeIsDepthImage ) {
		// always shade the whole framebuffer
		// NOTE: the bounding box includes the border
		bbox[0] = bbox[1] = 0;
		bbox[2] = viewPortSize.width - 1;
		bbox[3] = viewPortSize.height - 1;
		SrfShadeFrame (surfelContext, bbox, QRGBA(0,0,0,0)/*transparent*/);

	}

	if ( !nextNodeIsDepthImage ) {
		glMatrixMode (GL_MODELVIEW);
		glPushMatrix ();
		glLoadIdentity ();
		glMatrixMode (GL_PROJECTION);
		glPushMatrix ();
		glLoadIdentity ();
		gluOrtho2D (0.0, (GLfloat)viewPortSize.width, 0.0, (GLfloat)viewPortSize.height);
		glRasterPos2i (0,0);
		glDrawPixels (viewPortSize.width, viewPortSize.height, GL_RGBA, GL_UNSIGNED_BYTE,
					  frameBuffer->getImage());
		glPopMatrix();
		glMatrixMode (GL_MODELVIEW);
		glPopMatrix();
	}
	glPopMatrix ();
	glPopMatrix ();

	return 0;
}

void
DepthImageProxyV2::checkPrevAndNextNodes ()
{
	DepthImageV2 *pDepthImage = GetDepthImage ();
if ( ! checkedNodes ) {
		checkedNodes = TRUE;
		MediaObject* parent = pDepthImage->GetParent(0);
		if ( parent && (CODE_Group == parent->GetNodeCode()) ) {
			int numSiblings = parent->GetChildrenCount();
			int idxThis = 0;
			while ( idxThis < numSiblings && (pDepthImage != parent->GetChild(idxThis)) ) {
				idxThis += 1;
			}
			if ( idxThis < numSiblings ) {
				if ( idxThis > 0
					 && CODE_DepthImageV2 == parent->GetChild(idxThis-1)->GetNodeCode() ) {
					prevNodeIsDepthImage = TRUE;
					prevDepthImageProxy = (DepthImageProxyV2*)
						parent->GetChild(idxThis-1)->GetRenderingObject();
				}
				if ( (idxThis+1) < (numSiblings-1)
					 && CODE_DepthImageV2 == parent->GetChild(idxThis+1)->GetNodeCode() ) {
					nextNodeIsDepthImage = TRUE;
				}
			}
		}
	}
}


void
DepthImageProxyV2::DrawAxis (void)
{
	glDisable (GL_LIGHTING);
	glBegin (GL_LINES);
	glColor3f (1, 0, 0);
	glVertex3f (0, 0, 0);
	glVertex3f (10, 0, 0);

	glColor3f (0, 1, 0);
	glVertex3f (0, 0, 0);
	glVertex3f (0, 10, 0);

	glColor3f (0, 0, 1);
	glVertex3f (0, 0, 0);
	glVertex3f (0, 0, 10);

	glEnd ();
	glEnable (GL_LIGHTING);
}

// Traverse scene graph, checking for intersection

/* virtual */ BOOL
DepthImageProxyV2::Select (OpenGLVisualRenderer * pRenderer, DWORD /*dwTime*/,
						   unsigned int &depth)
{
	// I am not sure this would work really 2003.06.24 In Kyu Park

	GLuint selectBuffer[4];

	::glSelectBuffer (4, selectBuffer);
	::glRenderMode (GL_SELECT);

	if (::glRenderMode (GL_RENDER) > 0)
		if (depth > selectBuffer[1]) {
			depth = selectBuffer[1];

			return TRUE;
		}
#ifdef _DEBUG
	if (pRenderer->CheckGLError ("DepthImageProxyV2:"))
		TRACE ("DepthImageProxyV2::Render() [%08x]: ", this);
#endif

	return FALSE;
}

