// -*- Mode: c++; tab-width: 4; c-basic-offset: 4; -*-
// This software module was adapted, with permission, by Christopher Lee
// (Mitsubishi Electric Research Laboratories) from PointShop3D 2.0 by
//  Richard Keiser, Oliver Knoll, Mark Pauly, and Matthias Zwicker
// (Computer Graphics Lab, ETH Zurich).
// It was adapted in the course of development of the MPEG-4 Systems standard
// (ISO/IEC 14496-1).  This software module is an implementation
// of a part of one or more MPEG-4 Systems tools as specified by the MPEG-4
// Systems standard.  ISO/IEC gives users of the MPEG-4 Systems standard
// free license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4 Systems
// standard.  Those intending to use this software module in hardware or
// software products are advised that its use may infringe existing
// patents.  The original developer of this software module and his/her
// company, the subsequent editors and their companies, and ISO/IEC have no
// liability for use of this software module or modifications thereof in an
// implementation.  Copyright is not released for non MPEG-4 Systems
// conforming products.  The original developer retains full right to use
// the code for his/her own purpose, assign or donate the code to a third
// party and to inhibit third parties from using the code for non MPEG-4
// Systems conforming products.  This copyright notice must be included in
// all copies or derivative works.
//
// Copyright 2004.
//
// PointShop3D is available separately, under the GNU General Public License,
//  at http://graphics.ethz.ch/pointshop3d/index.html


//-------------------------------------------------------------------------
// Shader, shades samples in a z-buffer associated with a warper
// and writes result into an image buffer
//
// Note: 
// - this version does not support semi-transparent surfaces
// - performs shading in camera space, the phong illumination model is
// evaluated using cube reflectance maps 
//
// Shader_C_CubeReflMap.h 
//-------------------------------------------------------------------------

#ifndef SHADER_C_CUBEREFLMAP_H
#define SHADER_C_CUBEREFLMAP_H

#include "../../Surfel.h"
#include "../FrameBuffer.h"
#include "../SurfelPipeline_General/LSrList.h"
#include "../../DataTypes/MyDataTypes.h"
#include "Warper.h"

#define NOF_DBGCOLORS 8

// pack R,G,B components to format for display
#if 0
#define PACK_DIM_RGB(_c,_r,_g,_b) \
	_c = ((_r << 16) & 0x00ff0000) | \
		((_g << 8) & 0x0000ff00) | \
		(_b & 0x000000ff)
#else
#define PACK_DIM_RGB(_c,_r,_g,_b) _c = QRGB(_r,_g,_b)
#endif

// Global variables (for use by external shading functions)
extern float _shd_nx_c, _shd_ny_c, _shd_nz_c;
extern float _shd_x_c, _shd_y_c, _shd_z_c;
extern float _shd_vx, _shd_vy, _shd_vz;
extern float _shd_ndotv;
extern
	MyDataTypes::RGBTriple
	_shd_specularColor;
extern
	MyDataTypes::RGBTriple
	_shd_Id;
extern
	MyDataTypes::RGBTriple
	_shd_Ir;
extern float
	_shd_kA,
	_shd_kD,
	_shd_kS;
extern unsigned char
	_shd_shininess;
extern void *
	_shd_userdata;
extern int
	_shd_options;


// Variable and type declarations
typedef enum shdProperties
{
	SHD_BIDINORMALS = 1,
	SHD_DBGCOLORS = 2,
	SHD_ENVMAPPING = 4,
	SHD_LIGHTING = 8,
	SHD_PRESHADESPLAT = 16,
	SHD_PERPIXELSHADING = 32
} ShdProperties;

typedef enum shdShadingModel
{
	SHD_PHONG_H = 1,
	SHD_PHONG_R = 2,
	SHD_TORRANCESPARROW = 3,
	SHD_USERFNCT = 4
} ShdShadingModel;

typedef enum shdTextureFiltering
{
	SHD_TEXTUREFILTERING_NOINTERP = 1,
	SHD_TEXTUREFILTERING_LINEAR = 2
} ShdTextureFiltering;

typedef enum shdDiffColor
{
	SHD_DIFFCOL_DEFAULT = 0,
	SHD_DIFFCOL_MIPMAPLEVEL = 1,
	SHD_DIFFCOL_LODLEVEL = 2
} ShdDiffColor;

typedef void (*ShdLightSampleFnct) ();
typedef void (*ShdTextureFilteringFnct) ();

extern ShdLightSampleFnct
	_shd_lightsample;

typedef struct shader
{
	Warper *
		wrp;
	FrameBuffer *
		frameBuffer;

	LSrList *
		lsl;

	ShdLightSampleFnct
		shdLightSample;

	ShdLightSampleFnct
		shdUserShadingFunction;
	void *
		shdUserShadingData;

	int
		options;
	ShdDiffColor
		diffColor;
	int
		dbgcolors[NOF_DBGCOLORS];
} Shader;


// Function prototypes
Shader *
ShdNew (Warper * wrp, FrameBuffer * frameBuffer, LSrList * lsl);
void
ShdFree (Shader * shd);

void
ShdPrepareForShading (Shader * shd);
void
ShdShadeZBuffer (Shader * shader, int magfactor, int bbox[4],
				 QRgb backgroundColor);

void
ShdEnable (Shader * shd, ShdProperties p);
void
ShdDisable (Shader * shd, ShdProperties p);

void
ShdSetShadingModel (Shader * shd, ShdShadingModel s);
void
ShdSetUserShadingFunction (Shader * shd, ShdLightSampleFnct f, void *u);
void
ShdSetDiffColor (Shader * shd, ShdDiffColor c);


#endif // SHADER_C_CUBEREFLMAP_H

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
