/***********************************************************************
This software module was originally developed by Ioan Alexandru Salomie
(Vrije Universiteit Brussel VUB) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. VUB retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

  For more information or to receive an up to date version of this module,
  you may contact Ioan Alexandru Salomie at iasalomi@etro.vub.ac.be.
----------------------------------------------------------------------*/

#include "OpenGl/OpenGl.h"
#include <math.h>
#include "MeshGridProxy.h"
#include "opengl/appearance/Material2DProxy.h"
#include "opengl/AFX/BitWrapperProxy.h"
#include "IM1Decoders/AFX/AFX.h"
#include "IM1Decoders/AFX/MeshGridDecoder.h"
#include "IM1Decoders/AFX/MshGrd/DataDefs.h"
#include "IM1Decoders/AFX/MshGrd/MGMesh.h"
#include "IM1Decoders/AFX/MshGrd/MGGrid.h"

/***************************************************************************/
// Constructor - calls the inherited constructor from Node3DProxy
MeshGridProxy::MeshGridProxy(MeshGrid* pMG)
: AfxStreamProxy(pMG)
{
	m_bDecoded = false;
	m_failedDecode = false;
	m_bInit = false;
	m_lastTime = 0;
	m_lastUpdateMG = 0;
	m_lastUpdateGrid = 0;
	m_pInfo = NULL;
	m_level = 0;
	m_noTriangles = 0;
	m_indices = NULL;
	m_pMesh = NULL;
	m_pGrid = NULL;
}

MeshGridProxy::~MeshGridProxy()
{
	DeleteMGObjects();
}

// MeshGrid object destroy
void MeshGridProxy::DeleteMGObjects()
{
	delete m_indices; m_indices = NULL;
	delete m_pGrid;   m_pGrid = NULL;
	delete m_pMesh;   m_pMesh = NULL;
}

// traverse scene graph, checking for intersection
BOOL MeshGridProxy::Select(OpenGLVisualRenderer* pRenderer, DWORD dwTime, unsigned int& depth)
{
	GLuint selectBuffer[4];
	
	::glSelectBuffer(4, selectBuffer);
	::glRenderMode(GL_SELECT);
	
	Render(pRenderer, dwTime);
	
	if (::glRenderMode(GL_RENDER) > 0)
		if (depth >= selectBuffer[1]) {
			depth = selectBuffer[1];
			
			return TRUE;
		}
		
		return FALSE;
}

// update the object
bool MeshGridProxy::OnHierarchicalLevelChanged(MeshGrid *pMG)
{
	ASSERT(m_pMesh != NULL && m_pGrid != NULL);
	
	return m_pGrid->SetHierarchicalMode(pMG->hierarchicalLevel);
}

bool MeshGridProxy::OnGridCoordChanged(MeshGrid *pMG)
{
	ASSERT(m_pMesh != NULL && m_pGrid != NULL);
	
	// distribute
	m_pGrid->Distribute(pMG->hierarchicalLevel);	// update the mesh vertices
	m_pMesh->UpdateMGfromRG();
	
	return true;
}

bool MeshGridProxy::OnLevelChanged(MeshGrid *pMG)
{
	ASSERT(m_pMesh != NULL && m_pGrid != NULL);
	
	if (pMG->displayLevel < 0)
		pMG->displayLevel = 0;
	
	if (pMG->displayLevel >= m_pMesh->GetMaxNrLevels())
		pMG->displayLevel = m_pMesh->GetMaxNrLevels() - 1;
	
	if (pMG->displayLevel.GetValue() != m_level)
	{
		// the resolution level of the mesh changed, update the triangles
		m_level = pMG->displayLevel;
		GetTriangles();
	}
	
	return true;
}

bool MeshGridProxy::OnFilterChanged(MeshGrid *pMG)
{
	ASSERT(m_pMesh != NULL && m_pGrid != NULL);
	
	return m_pGrid->SetFilterType(pMG->filterType);
}

bool MeshGridProxy::OnVertexOffsetChanged(MeshGrid *pMG)
{
	ASSERT(m_pMesh != NULL && m_pGrid != NULL);
	
	int nrPts = pMG->vertexOffset.GetLength();
	if (nrPts <= 0)		return true;
	
	float *offsets = pMG->vertexOffset.GetArray().GetBuffer();
	
	// check for single layered and offsetAmplitude
	int count = 0;
	if (pMG->nSlices[2] == 1)
	{
		if (((int) offsets[count++]) == 1)
		{
			// offsetAmplitude is stored
			int sgn = (int) offsets[count++];
			
			// get the power 10 values
			float factor = 1.0;
			while (((int) offsets[count++]) == 1)
				factor *= 10.0f;
			
			// get the fractional offset
			float offset = factor * offsets[count++];
			int offsetAmplitude = (int) offset;
			ASSERT((float)offsetAmplitude == offset);
			
			if (sgn == 1) offsetAmplitude *= -1;
			
			m_pMesh->SetOffsetAmplitude(offsetAmplitude);
		}
	}
	ASSERT(nrPts > count);
	
	// update the mesh vertices
	CString error;
	m_pMesh->Update(&offsets[count], nrPts - count, error);
	
	return true;
}

// render object
BOOL MeshGridProxy::Render(OpenGLVisualRenderer* /* pRenderer */, DWORD dwTime)
{
	// called when the node is encoded with BIFS
	if (IsDecodeFailed()) return FALSE;
	
	MeshGrid *pMG = GetMeshGrid();
	ASSERT(pMG != NULL);
	
	if (!IsDecoded())
	{ 
		if (!Decode(pMG))
		{
			SetDecodeFailed(true);  return  FALSE;
		}
		
		SetFieldsInit(true);
		SetDecoded(true);
		
		OnFilterChanged(pMG);
		OnHierarchicalLevelChanged(pMG);
	}
	
	// update the mesh if necessary
	int lastUpdateTime = pMG->GetUpdateTime();
	if (lastUpdateTime > m_lastUpdateMG)
	{
		// some fields have changed
		m_lastUpdateMG = lastUpdateTime;
		
		if (pMG->displayLevel.WasFieldUpdated())
		{
			OnLevelChanged(pMG);
			pMG->displayLevel.ResetFieldUpdate();
		}
		
		if (pMG->filterType.WasFieldUpdated())
		{
			OnFilterChanged(pMG);
			pMG->filterType.ResetFieldUpdate();
		}
		
		if (pMG->hierarchicalLevel.WasFieldUpdated())
		{
			OnHierarchicalLevelChanged(pMG);
			pMG->hierarchicalLevel.ResetFieldUpdate();
		}
		
		if (pMG->vertexOffset.WasFieldUpdated())
		{
			OnVertexOffsetChanged(pMG);
			pMG->vertexOffset.ResetFieldUpdate();
		}
	}
	
	lastUpdateTime = pMG->gridCoord.GetNode()->GetUpdateTime();
	if (lastUpdateTime > m_lastUpdateGrid)
	{
		m_lastUpdateGrid = lastUpdateTime;
		OnGridCoordChanged(pMG);
	}
	
	// no action to be taken if the coord and texCoord fields are modified
	
	// display the result
	Render(dwTime);
	
	return TRUE;
}

BOOL MeshGridProxy::RenderFromStream(MediaStream* poStream, int /* id */, DWORD /* dwTime */)
{
	// it is called by BitWrapper
	// it should retrieve from the decoder the data and initialise the fields
	
	MeshGrid *pMG = GetMeshGrid();
	BitWrapper *pParent = (BitWrapper *)pMG->GetParent();
	ASSERT(pParent != NULL);
	
	// get the decoded data buffer
	int rnSize, rnStatus;
	DWORD rdvTime;
	BOOL rbEOU;
	BYTE * pOutput = (BYTE *)poStream->Fetch(rnSize, rdvTime, rbEOU, rnStatus);
	if (pOutput != NULL && rdvTime > m_lastTime)
	{
		// update the time
		m_lastTime = rdvTime;
		
		// retrieve tha data from the stream
		MGDataExchange dataExchange(pOutput, true);
		
		BYTE tag;
		DWORD len;
		void *pData = NULL;
		while ((pData = dataExchange.FetchNext(tag, len)) != NULL)
		{
			switch (tag)
			{
			case MGDataExchange::TAG_GRID_COORD:   GetGridCoord(pData, len);  break;
			case MGDataExchange::TAG_VERTEX_COORD: GetVertCoord(pData, len);  break;
			case MGDataExchange::TAG_TRIANGLES:    GetTriangles(pData, len);  break;
			case MGDataExchange::TAG_NORMALS:      GetNormals(pData, len);  break;
			case MGDataExchange::TAG_OFFSETS:      GetOffsets(pData, len);    break;
			case MGDataExchange::TAG_NO_VERTICES:  GetNoVertices(pData, len); break;
			default: ASSERT(0); return FALSE;
			}
		}
		
		CheckTextureIndices();
		
		// compute the bounding box and store the values in the MediaObjectProxy
		UpdateBbox();
		
		SetDecoded(true);
		pMG->isLoading = false;
		
		return TRUE;
	}
	
	return FALSE;
}

BOOL MeshGridProxy::RenderFromBuffer(SFString *buffer, int /* id */)
{
	// it is called by BitWrapper
	// it should decode the stream from the buffer and initialise the fields
	MeshGrid *pMG = GetMeshGrid();
	
	// get the node buffers
	Coordinate *grid = (Coordinate *) pMG->gridCoord.GetNode();
	ASSERT (grid != NULL);
	ZArray<Vec3f> &gridPoints = grid->point.GetArray();
	
	Coordinate *vertices = (Coordinate *) pMG->coord.GetNode();
	ASSERT (vertices != NULL);
	ZArray<Vec3f> &points = vertices->point.GetArray();
	
	// input data
	BYTE *data = (BYTE *)(LPCSTR) (*buffer);
	int length = buffer->GetLength();
	
	// decode the stream
	MGDecoder mgDecoder;
	if (!mgDecoder.Decode(data, length, gridPoints, points))
	{
		SetDecodeFailed(true);
		return FALSE;
	}
	
	// get the MeshGrid objects from the decoder
	mgDecoder.UnrefMGObjects(m_pMesh, m_pGrid);
	
	try
	{
		// fill in the fields of the node
		GetDisplayResolution();
		GetNoVertices();
		GetTriangles();
		GetOffsets();
		GetNormals();
		FillNodeFields();
		CheckTextureIndices();
		SetFieldsInit(true);
		
		// compute the bounding box and store the values in the MediaObjectProxy
		UpdateBbox();
	}
	catch (CMemoryException *)
	{
		TRACE0("Memory allocation error\n");
		return false;
	}
	
	// update the flags
	SetDecoded(true);
	pMG->isLoading = false;
	
	return TRUE;
}

BOOL MeshGridProxy::Render(DWORD /* dwTime */)
{
	if (!IsDecoded()) return TRUE;
	
	MeshGrid *pMG = GetMeshGrid();
	
	Coordinate *pCoordinate = (Coordinate *)pMG->coord.GetNode();
	TextureCoordinate* pTexCoord = (TextureCoordinate*)pMG->texCoord.GetNode(); //KM
	Normal *pNormal = (Normal *)pMG->normal.GetNode();
	MF<Vec3f> &points  = pCoordinate->point;
	MF<Vec2f> &texture = pTexCoord->point;
	MF<Vec3f> &vectors  = pNormal->vector;
	// MF<int>  &indices = pMG->coordIndex;
	MF<int>  &texIndices = pMG->texCoordIndex;
	
	int nrTextureCoords = texture.GetLength();
	int nrTextureIndices = texIndices.GetLength();
	
#ifdef _DEBUG
	int nrVertices = pMG->nVertices[m_level];
#endif
	
	// SF<Vec3f> sizeB, centerB;
	// MediaObjectProxy::GetBbox(sizeB, centerB);
	// Vec3f size = (Vec3f)sizeB, center = (Vec3f)centerB;
	
	// Enable vertex and normal arrays
	::glEnableClientState(GL_VERTEX_ARRAY);
	::glEnableClientState(GL_NORMAL_ARRAY);
	
	// Set pointers to buffers
	::glVertexPointer(3, GL_FLOAT, 0, points.GetArray().GetBuffer());
	::glNormalPointer(GL_FLOAT, 0, vectors.GetArray().GetBuffer());
	
	// Render
	::glDrawElements(GL_TRIANGLES, m_noTriangles * 3, GL_UNSIGNED_INT, m_indices);
	::glDisableClientState(GL_VERTEX_ARRAY);
	::glDisableClientState(GL_NORMAL_ARRAY);
	

	return TRUE;
}

bool MeshGridProxy::Decode(MeshGrid *pMG)
{
	// set the isLoading flag
	pMG->isLoading = true;
	
	// retrieve the info about the mesh
	MF<int> &nSlices = pMG->nSlices;
	MF<int> &nLevels = pMG->nLevels;
	
	CIPointUVW nrSlices(nSlices[0], nSlices[1], nSlices[2]);
	CIPointUVW nrLevels(nLevels[0], nLevels[1], nLevels[2]);
	
	// retrieve the grid points
	Coordinate *grid = (Coordinate *) pMG->gridCoord.GetNode();
	if (grid == NULL)
	{
		TRACE0("No grid points specified\n");   return false;
	}
	ZArray<Vec3f> &gridPoints = grid->point.GetArray();
	
	// retrieve the vertices
	Coordinate *vertices = (Coordinate *) pMG->coord.GetNode();
	ASSERT (vertices != NULL);
	ZArray<Vec3f> &points = vertices->point.GetArray();
	
	// retrieve the vertices' connectivity and offsets
	ZArray<int> &vertexLink = pMG->vertexLink.GetArray();
	ZArray<float> &vertexOffset = pMG->vertexOffset.GetArray();
	
	// decode the node
	MGDecoder mgDecoder;
	if (!mgDecoder.Decode(gridPoints, points, vertexLink, vertexOffset,
		nrSlices, nrLevels))
		return false;
	
	// get the MeshGrid objects from the decoder
	mgDecoder.UnrefMGObjects(m_pMesh, m_pGrid);
	
	try
	{
		// fill in the fields of the node
		GetDisplayResolution();
		GetNoVertices();
		GetTriangles();
		GetOffsets();
		GetNormals();
		CheckTextureIndices();
		SetFieldsInit(true);
		
		// compute the bounding box and store the values in the MediaObjectProxy
		UpdateBbox();
	}
	catch (CMemoryException *)
	{
		TRACE0("Memory allocation error\n");
		return false;
	}
	
	// update the flags
	SetDecoded(true);
	pMG->isLoading = false;
	
	return TRUE;
}

void MeshGridProxy :: CheckTextureIndices()
{
	// specify default texture indices if not present
	MeshGrid *pMG = GetMeshGrid();
	TextureCoordinate* pTexCoord = (TextureCoordinate*)pMG->texCoord.GetNode(); //KM
	if (pTexCoord != NULL)
	{
		MF<Vec2f> &texture = pTexCoord->point;
		if (texture.GetLength() > 0)
		{
			MF<int>  &texIndices = pMG->texCoordIndex;
			if (texIndices.GetLength() == 0)
			{
				Coordinate *pCoordinate = (Coordinate *)pMG->coord.GetNode();
				const MF<Vec3f> &points  = pCoordinate->point;
				
				texIndices.Expand(points.GetLength());
				for (int k = 0; k < points.GetLength(); k++)
				{
					texIndices.SetAt(k, k, FALSE);
				}
			}
		}
	}
}

void MeshGridProxy :: GetDisplayResolution()
{
	MeshGrid *pMG = GetMeshGrid();
	int maxNrLevels = m_pMesh->GetMaxNrLevels();
	/*
	if (pMG->displayLevel.GetValue() < 0 || pMG->displayLevel.GetValue() >= maxNrLevels)
	pMG->displayLevel = maxNrLevels - 1;
	
	  m_level = pMG->displayLevel.GetValue();
	*/
	m_level = pMG->displayLevel = maxNrLevels - 1;
}

void MeshGridProxy :: GetNoVertices()
{
	// retrieve the number of vertices
	MeshGrid *pMG = GetMeshGrid();
	int maxNrLevels = m_pMesh->GetMaxNrLevels();
	for (int level = 0; level < maxNrLevels; level++)
		pMG->nVertices.SetAt(level, m_pMesh->GetNrVertices(level));
}

void MeshGridProxy :: GetTriangles()
{
	// get the triangle list
	MeshGrid *pMG = GetMeshGrid();
	m_noTriangles = m_pMesh->GetNrPolygons(m_level);
	
	ZArray<int> &coordIndex = pMG->coordIndex.GetArray();
	coordIndex.Expand(m_noTriangles * 4);
	
	m_pMesh->GetTriangles(coordIndex.GetBuffer(), m_level);

	delete []m_indices;
	m_indices = new int[m_noTriangles * 3];
	for (int k = 0; k < m_noTriangles; k++)
	{
        m_indices[3 * k] = coordIndex[4 * k];
        m_indices[3 * k + 1] = coordIndex[4 * k + 1];
        m_indices[3 * k + 2] = coordIndex[4 * k + 2];
	}
}

void MeshGridProxy :: GetOffsets()
{
	// update the list of offsets of the VRML node
	// such that they are in the same order as the allocated vertices
	MeshGrid *pMG = GetMeshGrid();
	int noVertices = m_pMesh->GetTotalNrVertices();
	
	ZArray<float> &vertexOffset = pMG->vertexOffset.GetArray();
	vertexOffset.Expand(noVertices);
	
	m_pMesh->GetVertexOffset(vertexOffset.GetBuffer());
}

void MeshGridProxy :: GetNormals()
{
	// get the vertices normal vectors
	MeshGrid *pMG = GetMeshGrid();
	Normal *normals = (Normal *) pMG->normal.GetNode();
	ASSERT (normals != NULL);
	
	int noVertices = m_pMesh->GetTotalNrVertices();
	ZArray<Vec3f> &vector = normals->vector.GetArray();
	vector.Expand(noVertices);
	
	m_pMesh->GetVertexNormals(vector.GetBuffer(), m_level);
}

void MeshGridProxy :: FillNodeFields()
{
	MeshGrid *pMG = GetMeshGrid();
	const CIPointUVW &nLevels = m_pGrid->GetNrLevelsUVW();
	const CIPointUVW &nSlices = m_pGrid->GetMaxNrSlices();
	
	pMG->nLevels.InsertAt(0, nLevels.u, false);
	pMG->nLevels.InsertAt(1, nLevels.v, false);
	pMG->nLevels.InsertAt(2, nLevels.w, false);
	pMG->nSlices.InsertAt(0, nSlices.u, false);
	pMG->nSlices.InsertAt(1, nSlices.v, false);
	pMG->nSlices.InsertAt(2, nSlices.w, false);
}

bool MeshGridProxy :: GetGridCoord (void *pData, int size)
{
	ASSERT(size % (3 * sizeof(float)) == 0);
	int nElements = size / (3 * sizeof(float));
	
	Vec3f *pointArray = (Vec3f *)pData;
	MeshGrid *pMG = GetMeshGrid();
	Coordinate *coord = (Coordinate *) pMG->gridCoord.GetNode();
	MF<Vec3f> &points = coord->point;
	
	points.Expand(nElements);
	for (int k = 0; k < nElements; k++)
		points.SetAt(k, pointArray[k], FALSE);
	
	return true;
}

bool MeshGridProxy :: GetVertCoord (void *pData, int size)
{
	ASSERT(size % (3 * sizeof(float)) == 0);
	int nElements = size / (3 * sizeof(float));
	
	Vec3f *pointArray = (Vec3f *)pData;
	MeshGrid *pMG = GetMeshGrid();
	Coordinate *vertices = (Coordinate *) pMG->coord.GetNode();
	MF<Vec3f> &points = vertices->point;
	
	points.Expand(nElements);
	for (int k = 0; k < nElements; k++)
		points.SetAt(k, pointArray[k], FALSE);
	
	return true;
}

bool MeshGridProxy :: GetNormals (void *pData, int size)
{
	ASSERT(size % (3 * sizeof(float)) == 0);
	int nElements = size / (3 * sizeof(float));
	
	Vec3f *vectorArray = (Vec3f *)pData;
	MeshGrid *pMG = GetMeshGrid();
	Normal *normals = (Normal *) pMG->normal.GetNode();
	MF<Vec3f> &vector = normals->vector;
	
	vector.Expand(nElements);
	for (int k = 0; k < nElements; k++)
		vector.SetAt(k, vectorArray[k], FALSE);
	
	return true;
}

bool MeshGridProxy :: GetTriangles (void *pData, int size)
{
	ASSERT(size % (4 * sizeof(int)) == 0);
	int nrElements = size / sizeof(int);
	m_noTriangles = nrElements / 4;
	
	int *indices = (int *)pData;
	MeshGrid *pMG = GetMeshGrid();
	MF<int> &coordIndex = pMG->coordIndex;
	
	coordIndex.Expand(nrElements);

	//  FILE *fp = fopen("dumpTriangles.txt", "wt");
	int k;
	for (k = 0; k < nrElements; k++)
	{
		coordIndex.SetAt(k, indices[k], FALSE);
		//    fprintf(fp, "%d\n", indices[k]);
	}
	//  fclose(fp);

	// temporaty buffer used only for display
	delete []m_indices;
	m_indices = new int[m_noTriangles * 3];
	for (k = 0; k < m_noTriangles; k++)
	{
        m_indices[3 * k] = indices[4 * k];
        m_indices[3 * k + 1] = indices[4 * k + 1];
        m_indices[3 * k + 2] = indices[4 * k + 2];
	}
	
	return true;
}

bool MeshGridProxy :: GetOffsets (void *pData, int size)
{
	ASSERT(size % sizeof(int) == 0);
	int nrElements = size / sizeof(int);
	
	int *indices = (int *)pData;
	MeshGrid *pMG = GetMeshGrid();
	MF<float> &vertexOffset = pMG->vertexOffset;
	
	vertexOffset.Expand(nrElements);
	for (int k = 0; k < nrElements; k++)
		vertexOffset.SetAt(k, indices[k], FALSE);
	
	return true;
}

bool MeshGridProxy :: GetNoVertices (void *pData, int size)
{
	ASSERT(size % sizeof(int) == 0);
	int nrElements = size / sizeof(int);
	
	int *indices = (int *)pData;
	MeshGrid *pMG = GetMeshGrid();
	
	pMG->nVertices.Expand(nrElements);
	for (int k = 0; k < nrElements; k++)
		pMG->nVertices.SetAt(k, indices[k], FALSE);
	
	return true;
}

/// Get bounding box
void MeshGridProxy :: GetBbox(Vec3f& bboxSize, Vec3f& bboxCenter) const
{
	Coordinate* c = GetMeshGrid()->coord;
	const MF<Vec3f>& points=c->point;
	
	Vec3f mini(FLT_MAX,FLT_MAX,FLT_MAX),maxi(-FLT_MAX,-FLT_MAX,-FLT_MAX);
	
	for (int i = 0; i < points.GetLength(); i++)
	{
		if (points[i].x < mini.x) mini.x = points[i].x;
		if (points[i].y < mini.y) mini.y = points[i].y;
		if (points[i].z < mini.z) mini.z = points[i].z;
		
		if (points[i].x > maxi.x) maxi.x = points[i].x;
		if (points[i].y > maxi.y) maxi.y = points[i].y;
		if (points[i].z > maxi.z) maxi.z = points[i].z;
	}
	bboxSize = maxi - mini;
	bboxCenter = (maxi - mini) / 2;
}

void MeshGridProxy :: UpdateBbox()
{
	Vec3f size, center;
	GetBbox(size, center);
	SF<Vec3f> sizeB = (SF<Vec3f>)size, centerB = (SF<Vec3f>)center;
	SetBbox(sizeB, centerB);
}