/***********************************************************************
This software module was originally developed by Mikael Bourges-Sevenier
(Mindego Inc.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Mindego Inc. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Mikael Bourges-Sevenier at mikael@mindego.com.
----------------------------------------------------------------------*/

#include "OpenGl/OpenGl.h"
#include "NonLinearDeformerProxy.h"
#include "opengl/base/OpenGLVisualRenderer.h"
#include "opengl/geometry/IndexedFaceSetProxy.h"

/***************************************************************************/
// Constructor - calls the inherited constructor from Node3DProxy
NonLinearDeformerProxy::NonLinearDeformerProxy(NonLinearDeformer * pNLD)
  : GeometryProxy(pNLD), result(NULL), coords(NULL), normals(NULL),
	  _hasIFS(false)
{
}

NonLinearDeformerProxy::~NonLinearDeformerProxy()
{
	if(_hasIFS) { // not a good idea to do that but that works in C++ ;-)
		if(result) delete result;
		if(coords) delete coords;
		if(normals) delete normals;
	}
}

/* virtual */ BOOL
NonLinearDeformerProxy::Select(
		OpenGLVisualRenderer* pRenderer, // the renderer
    DWORD dwTime,                    // simulation time
    unsigned int& depth)             //
{
	// TODO [MBS] this object selection code crashes

  /*GLuint selectBuffer[4];

  ::glSelectBuffer(4, selectBuffer);
  ::glRenderMode(GL_SELECT);

  Render(pRenderer, dwTime);

  if (::glRenderMode(GL_RENDER) > 0)
    if (depth >= selectBuffer[1]) {
      depth = selectBuffer[1];

      return TRUE;
    }
	*/
  return FALSE;
}

// Render object
BOOL NonLinearDeformerProxy::Render(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
	/* MBS 7/5/03
	 * The problem with NonLinearDeformers (NLDs) is that we need access to the geometry of the children.
	 * IM1 code renders everything immediately, so we can't access geometry except for IndexedFaceSet (IFS) and other
	 * nodes with vertices specified explicitely (but not basic pre-defined shapes).
	 *
	 * NLD.geometry can contain any geometry node including an NLD. The downstream NLD always contains an IFS.
	 * We build an IFSProxy upon this IFS that shares meta information but copy vertices and normals.
	 * That way, upstream NLDs can reuse this IFSProxy and modify vertices and normals.
	 * Finally, only the root NLD renders the IFSProxy after applying all downstream NLDs to the IFS.
	 *
	 */


	// update coords and normals
	// create a new IFSProxy with these fields
	// render this IFSProxy
	if(result==NULL) {
		Update(dwTime);
		IndexedFaceSet *ifs=(IndexedFaceSet*) result->GetMediaObject();
		Coordinate *coord=(Coordinate*) ifs->coord.GetNode();
		coord->point=*coords;
	}

	glDisable(GL_CULL_FACE);
	result->Render(pRenderer,dwTime);

	return TRUE;
}

// Update IFSProxy object
BOOL NonLinearDeformerProxy::Update(DWORD dwTime)
{
	// TODO must test if we need to update IFSProxy

	m_nTimeStamp=(long) dwTime;

	NonLinearDeformer *pNLD=GetNonLinearDeformer();
	MediaObject *geometry=(MediaObject *) pNLD->geometry;
	if(!geometry)  return TRUE;
	if(geometry->GetNodeCode()==CODE_NonLinearDeformer)  {
		// access this downstream NLD's IFSProxy and apply transfo to it.
		//result=((NonLinearDeformerProxy*) ((NonLinearDeformer *) geometry)->GetRenderingObject())->GetIFSProxy();
		NonLinearDeformerProxy *def=(NonLinearDeformerProxy *) ((NonLinearDeformerProxy*) ((NonLinearDeformer *) geometry)->GetRenderingObject());
		def->Update(dwTime);
		result=def->result;
		coords=def->coords;
		normals=def->normals;
		_hasIFS=true;
	}

	else if(geometry->GetNodeCode()==CODE_IndexedFaceSet) {
		// we are the last downstream NLD
		// create an IFSProxy, copy vertices and normals, share all other fields. Then apply this NLD transformations
		IndexedFaceSet *child=(IndexedFaceSet*) geometry, *ifs=(IndexedFaceSet*) child->Clone(child->GetRoot());
		int i;
	
		result= (IndexedFaceSetProxy*) ifs->GetRenderingObject();

		const MF<Vec3f>& pts=((Coordinate*) child->coord.GetNode())->point;
		int num=pts.GetLength();

		if(coords==NULL)
			coords=new MF<Vec3f>();
		coords->Remove();

		for(i=0;i<num;i++)
			(*coords) += pts[i];

		if(child->normal.GetNode()) {
			const MF<Vec3f>& norms=((Normal*) child->normal.GetNode())->vector;
			num=norms.GetLength();
			if(normals==NULL)
				normals=new MF<Vec3f>();
			normals->Remove();

			for(i=0;i<num;i++)
				(*normals) += pts[i];
		}
	}	

	switch(pNLD->type) 
	{
		case 0: taper(); break;
		case 1: twist(); break;
		case 2: bend(); break;
	};

	return TRUE;
}

// Render object description (when viewing the scene graph)
/* virtual */ void
NonLinearDeformerProxy::RenderDescription(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
  GLfloat color[] = { 1.0f, 0.0f, 1.0f, 0.75f};

  ::glColor4fv(color);

  RenderDescriptionBox(pRenderer, "NonLinearDeformer");
}

/// Get bounding box
void NonLinearDeformerProxy::GetBbox(Vec3f& bboxSize, Vec3f& bboxCenter) const
{
	if(coords==NULL) {
		bboxSize=Vec3f(1,1,1);
		bboxCenter=Vec3f(0,0,0);
		return;
	}

	Vec3f mini(FLT_MAX,FLT_MAX,FLT_MAX),maxi(-FLT_MAX,-FLT_MAX,-FLT_MAX);
	int num=coords->GetLength();

	for(int i=0;i<num;i++) {
		const Vec3f& pt=(*coords)[i];
		if(pt.x<mini.x) mini.x=pt.x;
		if(pt.y<mini.y) mini.y=pt.y;
		if(pt.z<mini.z) mini.z=pt.z;

		if(pt.x>maxi.x) maxi.x=pt.x;
		if(pt.y<maxi.y) maxi.y=pt.y;
		if(pt.z<maxi.z) maxi.z=pt.z;
	}
	bboxSize=maxi-mini;
	bboxCenter=(maxi-mini)/2;
}

// taper along z-axis
void NonLinearDeformerProxy::taper()
{
	NonLinearDeformer *pNLD=GetNonLinearDeformer();
	float radius=pNLD->param;
	const MF<float>& exts=pNLD->extend;

	// transform vertices such that axis of object is z
	pNLD->axis.GetValue().Normalize();
	const Vec3f& n=pNLD->axis;

	// note: the matrix are transposed because Matrix::MultiplyPoint does v^T.M^T instead of M.v.
	Matrix  r(Vec3f(-n.x*n.y-n.z, -n.y*n.y, -n.y*n.z+n.x), Vec3f(n.x*n.x, n.x*n.y-n.z, n.x*n.z+n.y), Vec3f(n.x*n.z-n.y, n.y*n.z+n.x, n.z*n.z));
	Matrix r1(Vec3f(-n.x*n.y-n.z, n.x*n.x, n.x*n.z-n.y), Vec3f(-n.y*n.y, n.x*n.y-n.z, n.y*n.z+n.x), Vec3f(-n.y*n.z+n.x, n.x*n.z+n.y, n.z*n.z));
	int i, num=coords->GetLength();

	// Transform vertices by using the invert matrix because MultiplyPoint does v.M instead of M.v ;-(
	for(i=0;i<num;i++) 
		coords->SetAt(i,r.MultiplyPoint((*coords)[i]));
	
	// calc dz
	float zm=FLT_MAX, zM=-FLT_MAX;
	for(i=0;i<num;i++) {
		if(zm>(*coords)[i].z) zm=(*coords)[i].z;
		if(zM<(*coords)[i].z) zM=(*coords)[i].z;
	}
	float dz=zM-zm;
	if(dz==0) return;

	for(i=0;i<num;i++) {
		float z=(*coords)[i].z;
		int idx0=0, idx1=0;
		findKeys(z, zm, dz, exts, &idx0, &idx1);

		float pi0,pi1, ri0, ri1;
		if(idx0==idx1) {
			if(idx0==0) {
				pi0=0;       ri0=1;
				pi1=exts[0]; ri1=exts[1];
			}
			else if(idx0==exts.GetLength()-2) {
				pi0=exts[idx0]; ri0=exts[idx0+1];
				pi1=1;          ri1=1;
			}
		}
		else {
			pi0=exts[idx0]; ri0=exts[idx0+1];
			pi1=exts[idx1]; ri1=exts[idx1+1];
		}

		float ti,r=1;
		if(pi0 != pi1) {
			ti=(z-(zm+pi0*dz))/((pi1-pi0)*dz);
			r=ri1*ti+ri0*(1-ti);
		}

		// transform vertices from axis of object in z to the axis of object
		coords->SetAt(i,r1.MultiplyPoint(Vec3f((*coords)[i].x*r,(*coords)[i].y*r,z)));	
	}
}

// twist along z-axis
void NonLinearDeformerProxy::twist()
{
	NonLinearDeformer *pNLD=GetNonLinearDeformer();
	float radius=pNLD->param;
	const MF<float>& exts=pNLD->extend;

	// transform vertices such that axis of object is z
	pNLD->axis.GetValue().Normalize();
	const Vec3f& n=pNLD->axis;

	// note: the matrix are transposed because Matrix::MultiplyPoint does v^T.M^T instead of M.v.
	Matrix  r(Vec3f(-n.x*n.y-n.z, -n.y*n.y, -n.y*n.z+n.x), Vec3f(n.x*n.x, n.x*n.y-n.z, n.x*n.z+n.y), Vec3f(n.x*n.z-n.y, n.y*n.z+n.x, n.z*n.z));
	Matrix r1(Vec3f(-n.x*n.y-n.z, n.x*n.x, n.x*n.z-n.y), Vec3f(-n.y*n.y, n.x*n.y-n.z, n.y*n.z+n.x), Vec3f(-n.y*n.z+n.x, n.x*n.z+n.y, n.z*n.z));
	int i, num=coords->GetLength();

	for(i=0;i<num;i++) 
		coords->SetAt(i,r.MultiplyPoint((*coords)[i]));

	// calc dz
	float zm=FLT_MAX, zM=-FLT_MAX;
	for(i=0;i<num;i++) {
		if(zm>(*coords)[i].z) zm=(*coords)[i].z;
		if(zM<(*coords)[i].z) zM=(*coords)[i].z;
	}
	float dz=zM-zm;
	if(dz==0) return;
	
	for(i=0;i<num;i++) {
		float z=(*coords)[i].z;
		int idx0=0, idx1=0;
		findKeys(z, zm, dz, exts, &idx0, &idx1);

		float pi0,pi1, thi0, thi1;
		if(idx0==idx1) {
			if(idx0==0) {
				pi0=0;       thi0=0;
				pi1=exts[0]; thi1=exts[1];
			}
			else if(idx0==exts.GetLength()-2) {
				pi0=exts[idx0]; thi0=exts[idx0+1];
				pi1=1;          thi1=0;
			}
		}
		else {
			pi0=exts[idx0]; thi0=exts[idx0+1];
			pi1=exts[idx1]; thi1=exts[idx1+1];
		}

		float ti,th=0;
		if(pi0 != pi1) {
			ti=(z-(zm+pi0*dz))/((pi1-pi0)*dz);
			th=thi1*ti+thi0*(1-ti);
		}

		float ct=cos(th), st=sin(th);
		float xp=(*coords)[i].x*ct-(*coords)[i].y*st, yp=(*coords)[i].x*st+(*coords)[i].y*ct;

		// transform vertices from axis of object in z to the axis of object
		coords->SetAt(i,r1.MultiplyPoint(Vec3f(xp,yp,z)));	
	}
}

// isotropic bend a centerline parallel to y-axis
void NonLinearDeformerProxy::bend()
{
	NonLinearDeformer *pNLD=GetNonLinearDeformer();
	float k =pNLD->extend[0], k1=0;
	if(k!=0) 
		k1=1/k;

	float y0=pNLD->extend[1];
	float ym=pNLD->extend[2];
	float yM=pNLD->extend[3];

	// transform vertices 
	pNLD->axis.GetValue().Normalize();
	Vec3f n(0,pNLD->axis.GetValue().z,-pNLD->axis.GetValue().y);
	BOOL isX=(pNLD->axis.GetValue().x==1) && (pNLD->axis.GetValue().y==0) &&(pNLD->axis.GetValue().z==0);

	// note: the matrix are transposed because Matrix::MultiplyPoint does v^T.M^T instead of M.v.
	//Matrix  r(Vec3f(-n.x*n.y-n.z, -n.y*n.y, -n.y*n.z+n.x), Vec3f(n.x*n.x, n.x*n.y-n.z, n.x*n.z+n.y), Vec3f(n.x*n.z-n.y, n.y*n.z+n.x, n.z*n.z));
	//Matrix r1(Vec3f(-n.x*n.y-n.z, n.x*n.x, n.x*n.z-n.y), Vec3f(-n.y*n.y, n.x*n.y-n.z, n.y*n.z+n.x), Vec3f(-n.y*n.z+n.x, n.x*n.z+n.y, n.z*n.z));
	
	Matrix  r(Vec3f(n.x*n.x, n.x*n.y-n.z, n.x*n.z+n.y),Vec3f(n.x*n.y+n.z, n.y*n.y, n.y*n.z-n.x),Vec3f(n.x*n.z-n.y, n.y*n.z+n.x, n.z*n.z));
	Matrix r1(Vec3f(n.x*n.x, n.x*n.y+n.z, n.x*n.z-n.y),Vec3f(n.x*n.y-n.z, n.y*n.y, n.y*n.z+n.x),Vec3f(n.x*n.z+n.y, n.y*n.z-n.x, n.z*n.z));
	int num=coords->GetLength();

	// apply transformation
	for(int i=0;i<num;i++) {
		if(!isX)
			coords->SetAt(i,r.MultiplyPoint((*coords)[i]));
		float x=(*coords)[i].x, y=(*coords)[i].y, z=(*coords)[i].z;
		float th,ct,st;
		float yt,zt;
		if(y<ym) {
			th=k*(ym-y0);	ct=cos(th); st=sin(th);
			yt=-st*(z-k1)+y0+ct*(y-ym);
			zt=ct*(z-k1)+k1+st*(y-ym);
		}
		else if(y>yM) {
			th=k*(yM-y0);	ct=cos(th); st=sin(th);
			yt=-st*(z-k1)+y0+ct*(y-yM);
			zt=ct*(z-k1)+k1+st*(y-yM);
		}
		else {
			th=k*(y-y0);	ct=cos(th); st=sin(th);
			yt=-st*(z-k1)+y0;
			zt=ct*(z-k1)+k1;
		}

		// transform vertices from axis of object in z to the axis of object
		if(!isX) 
			coords->SetAt(i,r1.MultiplyPoint(Vec3f(x,yt,zt)));	
		else
			coords->SetAt(i,Vec3f(x,yt,zt));	
	}
}

void NonLinearDeformerProxy::findKeys(float z, float zm, float dz, const MF<float>& exts, int *p0, int *p1)
{
	if(exts.GetLength()<2) {
		*p0=*p1=0;
		return;
	}

	float z0=zm+exts[0]*dz,z1;
	if(z<z0) {
		*p0=*p1=0;
		return;
	}

	float zN=zm+exts[exts.GetLength()-2]*dz;
	if(z>zN) {
		*p0=*p1=exts.GetLength()-2;
		return;
	}

	for(int i=0;i<exts.GetLength()-2;i+=2) {
		z0=zm+exts[i]*dz;
		z1=zm+exts[i+2]*dz;
		if(z>=z0 && z<=z1) {
			*p0=i; 
			*p1=i+2;
			break;
		}
	}
	return;
}