/***********************************************************************
This software module was originally developed by Mikael Bourges-Sevenier
(Mindego Inc.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Mindego Inc. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Mikael Bourges-Sevenier at mikael@mindego.com.
----------------------------------------------------------------------*/

#include "OpenGl/OpenGl.h"
#include "NurbsCurve2DProxy.h"
#include "opengl/base/OpenGLVisualRenderer.h"

/***************************************************************************/
// Constructor - calls the inherited constructor from Node3DProxy
NurbsCurve2DProxy::NurbsCurve2DProxy(NurbsCurve2D * pNC)
  : GeometryProxy(pNC),m_nTimeStamp(-1)
{
	theNurb = gluNewNurbsRenderer();
	gluNurbsProperty(theNurb, GLU_SAMPLING_TOLERANCE, 25.0);
	gluNurbsProperty(theNurb, GLU_DISPLAY_MODE, GLU_FILL);
}

NurbsCurve2DProxy::~NurbsCurve2DProxy()
{
	gluDeleteNurbsRenderer(theNurb);
}

/* virtual */ BOOL
NurbsCurve2DProxy::Select(
		OpenGLVisualRenderer* pRenderer, // the renderer
    DWORD dwTime,                    // simulation time
    unsigned int& depth)             //
{
	// TODO [MBS] this object selection code crashes

  GLuint selectBuffer[4];

  ::glSelectBuffer(4, selectBuffer);
  ::glRenderMode(GL_SELECT);

  Render(pRenderer, dwTime);

  if (::glRenderMode(GL_RENDER) > 0)
    if (depth >= selectBuffer[1]) {
      depth = selectBuffer[1];

      return TRUE;
    }
  return FALSE;
}

// Render object
BOOL NurbsCurve2DProxy::Render(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
	NurbsCurve2D *pNC=GetNurbsCurve2D();
	Color *pColor=GetColor();
  const MF<float>& knots=pNC->knot;
	const MF<int>& colorIndex=pNC->colorIndex;
	BOOL isIndexed=false;
	
  const MF<Color3f>& colors = pColor->color;

	_needUpdate=long(pNC->GetUpdateTime()) > m_nTimeStamp ||
      (pColor && (long(pColor->GetUpdateTime()) > m_nTimeStamp));

	if (_needUpdate)
	{
		m_nTimeStamp=(long) dwTime;

		if(!m_nDisplayList) 
			m_nDisplayList = ::glGenLists(1);
		else
			::glDeleteLists(m_nDisplayList,1);

		m_nDisplayList = ::glGenLists(1);
		::glNewList(m_nDisplayList, GL_COMPILE_AND_EXECUTE);
		::glPushAttrib(GL_POLYGON_BIT | GL_LINE_BIT);

		// rendering Ze curve
		if(pColor && colorIndex.GetLength()) {
			isIndexed=TRUE;
			for(int i=0;i<colorIndex.GetLength();i++)
				glColor3fv((float*) &colors[colorIndex[i]].r);
		}

		int sz=pNC->controlPoint.GetLength(); // we use the 3D API so, in 2D, we use z=0.
		float *points=new float[sz*4];
		for(int i=0;i<sz;i++) {
			points[4*i+0]=pNC->controlPoint[i].x;
			points[4*i+1]=pNC->controlPoint[i].y;
			points[4*i+2]=0;
			points[4*i+3]=pNC->controlPoint[i].z;
		}

		// auto-generate knots if none
		float *k=(float*) &knots[0];
		int ksz=knots.GetLength();

		if(ksz==0) {
			ksz=2*pNC->order;
			k=new float[ksz];
			for(int i=0;i<pNC->order;i++) {
				k[i]=0;
				k[i+pNC->order]=1;
			}
		}

		glEnable(GL_AUTO_NORMAL);
		gluBeginCurve(theNurb);
			gluNurbsCurve(theNurb, ksz, k, 4, points, pNC->order, GL_MAP1_VERTEX_4);
		gluEndCurve(theNurb);

		::glPopAttrib();
		::glEndList();

		delete[] points;
	}
	else
		::glCallList(m_nDisplayList);

	return TRUE;
}

// Render object description (when viewing the scene graph)

/* virtual */ void
NurbsCurve2DProxy::RenderDescription(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
  int children = ((GetColor() ? 1 : 0));

  GLfloat color[] = { 1.0f, 0.0f, 1.0f, 0.75f};

  ::glColor4fv(color);

  RenderDescriptionBox(pRenderer, "NurbsCurve2D");

  ::glPushMatrix();

  ::glScalef(1.0f / children, 1.0f, 1.0f);
  ::glTranslatef(0.0f, -1.0f, 0.0f);

  if (/* pIndexedFaceSet-> */ GetColor()) {
    RenderDescriptionBox(pRenderer, "Color");
  }

  ::glPopMatrix();
}

/// Get bounding box
void NurbsCurve2DProxy::GetBbox(Vec3f& bboxSize, Vec3f& bboxCenter) const
{
  const MF<Vec3f>& points = GetControlPoint();
	
	Vec3f mini(FLT_MAX,FLT_MAX,FLT_MAX),maxi(-FLT_MAX,-FLT_MAX,-FLT_MAX);
	
	for(int i=0;i<points.GetLength();i++) {
		if(points[i].x<mini.x) mini.x=points[i].x;
		if(points[i].y<mini.y) mini.y=points[i].y;

		if(points[i].x>maxi.x) maxi.x=points[i].x;
		if(points[i].y<maxi.y) maxi.y=points[i].y;
	}
	bboxSize=maxi-mini;
	bboxCenter=(maxi-mini)/2;
}
