/***********************************************************************
This software module was originally developed by Mikael Bourges-Sevenier
(Mindego Inc.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Mindego Inc. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Mikael Bourges-Sevenier at mikael@mindego.com.
----------------------------------------------------------------------*/

#include "OpenGl/OpenGl.h"
#include "NurbsSurfaceProxy.h"
#include "opengl/base/OpenGLVisualRenderer.h"

/***************************************************************************/
// Constructor - calls the inherited constructor from Node3DProxy
NurbsSurfaceProxy::NurbsSurfaceProxy(NurbsSurface * pNS)
  : GeometryProxy(pNS),m_nTimeStamp(-1)
{
	theNurb = gluNewNurbsRenderer();
	gluNurbsProperty(theNurb, GLU_SAMPLING_METHOD, GLU_PATH_LENGTH);
	gluNurbsProperty(theNurb, GLU_SAMPLING_TOLERANCE, 90.0f);
	gluNurbsProperty(theNurb, GLU_DISPLAY_MODE, GLU_FILL);
}

NurbsSurfaceProxy::~NurbsSurfaceProxy()
{
	gluDeleteNurbsRenderer(theNurb);
}

/* virtual */ BOOL
NurbsSurfaceProxy::Select(
		OpenGLVisualRenderer* pRenderer, // the renderer
    DWORD dwTime,                    // simulation time
    unsigned int& depth)             //
{
	// TODO [MBS] this object selection code crashes

  GLuint selectBuffer[4];

  ::glSelectBuffer(4, selectBuffer);
  ::glRenderMode(GL_SELECT);

  Render(pRenderer, dwTime);

  if (::glRenderMode(GL_RENDER) > 0)
    if (depth >= selectBuffer[1]) {
      depth = selectBuffer[1];

      return TRUE;
    }
  return FALSE;
}

// Render object
BOOL NurbsSurfaceProxy::Render(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
	NurbsSurface *pNS=GetNurbsSurface();
	Color *pColor=GetColor();
  const MF<Vec4f>& points=pNS->controlPoint;
  const MF<float>& uknots=pNS->uKnot;
  const MF<float>& vknots=pNS->vKnot;
	const MF<int>& colorIndex=pNS->colorIndex;
	BOOL isIndexed=false;
	
  const MF<Color3f>& colors = pColor->color;

	_needUpdate=long(pNS->GetUpdateTime()) > m_nTimeStamp ||
      (pColor && (long(pColor->GetUpdateTime()) > m_nTimeStamp));

	if (_needUpdate)
	{
		m_nTimeStamp=(long) dwTime;

		if(m_nDisplayList) 
			::glDeleteLists(m_nDisplayList,1);

		m_nDisplayList = ::glGenLists(1);

		::glNewList(m_nDisplayList, GL_COMPILE_AND_EXECUTE);
		::glPushAttrib(GL_POLYGON_BIT);

		// rendering Ze Surface
		// auto-generate knots if none
		float *uk=(float*) &uknots[0];
		int uksz=uknots.GetLength();

		if(uksz==0) {
			uksz=2*pNS->uOrder;
			uk=new float[uksz];
			for(int i=0;i<pNS->uOrder;i++) {
				uk[i]=0;
				uk[i+pNS->uOrder]=1;
			}
		}

		float *vk=(float*) &vknots[0];
		int vksz=uknots.GetLength();

		if(vksz==0) {
			vksz=2*pNS->vOrder;
			vk=new float[vksz];
			for(int i=0;i<pNS->vOrder;i++) {
				vk[i]=0;
				vk[i+pNS->vOrder]=1;
			}
		}


		glEnable(GL_AUTO_NORMAL);
		glEnable(GL_NORMALIZE);

		gluBeginSurface(theNurb);
			gluNurbsSurface(theNurb, 
											uksz, uk,
											vksz, vk,
											4*4, 4, (float*) &(points[0].x), 
											pNS->uOrder,pNS->vOrder, 
											GL_MAP2_VERTEX_4);

		gluEndSurface(theNurb);

		::glPopAttrib();
		::glEndList();
	}
	else
		::glCallList(m_nDisplayList);

	return TRUE;
}

// Render object description (when viewing the scene graph)

/* virtual */ void
NurbsSurfaceProxy::RenderDescription(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
  int children = ((GetColor() ? 1 : 0));

  GLfloat color[] = { 1.0f, 0.0f, 1.0f, 0.75f};

  ::glColor4fv(color);

  RenderDescriptionBox(pRenderer, "NurbsSurface");

  ::glPushMatrix();

  ::glScalef(1.0f / children, 1.0f, 1.0f);
  ::glTranslatef(0.0f, -1.0f, 0.0f);

  if (/* pIndexedFaceSet-> */ GetColor()) {
    RenderDescriptionBox(pRenderer, "Color");
  }

  ::glPopMatrix();
}

/// Get bounding box
void NurbsSurfaceProxy::GetBbox(Vec3f& bboxSize, Vec3f& bboxCenter) const
{
  const MF<Vec4f> points = GetControlPoint();
	
	Vec3f mini(FLT_MAX,FLT_MAX,FLT_MAX),maxi(-FLT_MAX,-FLT_MAX,-FLT_MAX);
	
	for(int i=0;i<points.GetLength();i++) {
		if(points[i].x<mini.x) mini.x=points[i].x;
		if(points[i].y<mini.y) mini.y=points[i].y;
		if(points[i].z<mini.z) mini.z=points[i].z;

		if(points[i].x>maxi.x) maxi.x=points[i].x;
		if(points[i].y<maxi.y) maxi.y=points[i].y;
		if(points[i].z<maxi.z) maxi.z=points[i].z;
	}
	bboxSize=maxi-mini;
	bboxCenter=(maxi-mini)/2;
}
