/***********************************************************************
This software module was originally developed by Mikael Bourges-Sevenier
(Mindego Inc.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Mindego Inc. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Mikael Bourges-Sevenier at mikael@mindego.com.
----------------------------------------------------------------------*/

#include "OpenGl/OpenGl.h"
#include "PositionAnimatorProxy.h"

// Constructor, only calls the base constructor
PositionAnimatorProxy::PositionAnimatorProxy(MediaObject * poNode) :
InterpolatorNodeProxy(poNode)
{
}

BOOL PositionAnimatorProxy::Evaluate()
{
  PositionAnimator *pPA=(PositionAnimator*) GetMediaObject();

  // Check if we have any keys
  if (pPA->key.GetLength() == 0) return TRUE;

	float t=pPA->set_fraction;

	// don't do anything if out of fromTo range.
	const Vec2f& fromTo=pPA->fromTo.GetValue();
	if(fromTo.x == fromTo.y || fromTo.x>fromTo.y) return TRUE;
	if(t<fromTo.x && t>fromTo.y) return TRUE;

	t=(t-fromTo.x)/(fromTo.y-fromTo.x);

  if(t>=0 && t<=1) {
		if (pPA->keyValueType==0)
			interpolateLinear(t);
		else if (pPA->keyValueType==4)
			interpolateFit(t);
		else
			interpolateNURBS(t);
	}
  return TRUE;
}

// Bisection algorithm to find u=a*t^3+b*t^2+c*t+d
const float SEARCHTOL=0.001;

float PositionAnimatorProxy::bisec(float t, const Vec2f& a, const Vec2f& b, const Vec2f& c, const Vec2f& d)
{
  float ul=0,ur=1;
  float usearch, tsearch;

  do {
    usearch=(ul+ur)/2.;
    tsearch=usearch*(c.x+usearch*(b.x+usearch*a.x))+d.x;
    if (t<tsearch+SEARCHTOL) ur=usearch;
    else ul=usearch;
  } while (t>tsearch+SEARCHTOL || t<tsearch-SEARCHTOL);

  return usearch*(c.y+usearch*(b.y+usearch*a.y))+d.y;
}

// for piecewise linear paths
void PositionAnimatorProxy::interpolateLinear(float t)
{
  PositionAnimator *pPA=(PositionAnimator*) GetMediaObject();
  const MF<float>& ti=pPA->key;
  const MF<Vec3f>& vi=pPA->keyValue;
  int nKeys=ti.GetLength();
  int nValues=vi.GetLength();
  Vec3f pos;

  switch (pPA->keyType) {
  case 0: // interpolator
    {
      if (nKeys != nValues) return;
      if (t<ti[0] || t>ti[nKeys-1]) return;
      int i=0;
      for (i=0;i<nKeys-1;i++) {
        if (t>=ti[i] && t<ti[i+1])
          break;
      }
      float alpha=(t-ti[i])/(ti[i+1]-ti[i]);
      pos.x=alpha*vi[i+1].x+(1-alpha)*vi[i].x;
      pos.y=alpha*vi[i+1].y+(1-alpha)*vi[i].y;
      pos.z=alpha*vi[i+1].z+(1-alpha)*vi[i].z;
    }
    break;
  case 1: // discrete
    {
      pos=vi[t*nValues];
    }
    break;
  case 2: // linear
    {
      int i=t*(nValues-1);
      float alpha=(t-float(i)/float(nValues-1))*(nValues-1);
      pos.x=alpha*vi[i+1].x+(1-alpha)*vi[i].x;
      pos.y=alpha*vi[i+1].y+(1-alpha)*vi[i].y;
      pos.z=alpha*vi[i+1].z+(1-alpha)*vi[i].z;
    }
    break;
  case 3: // paced
    {
      int i=t*(nValues-1),j;
      Vec3f d;

      float di=0;
      for (j=0;j<i;j++) {
        d=Vec3f(vi[j+1].x-vi[j].x, vi[j+1].y-vi[j].y, vi[j+1].z-vi[j].z);
        di += sqrt(d.x*d.x+d.y*d.y+d.z*d.z);
      }
      float dn=di;
      for (j=i;j<nValues-1;j++) {
        d=Vec3f(vi[j+1].x-vi[j].x, vi[j+1].y-vi[j].y, vi[j+1].z-vi[j].z);
        dn += sqrt(d.x*d.x+d.y*d.y+d.z*d.z);
      }
      float tii=di/dn;

      d=Vec3f(vi[i+1].x-vi[i].x, vi[i+1].y-vi[i].y, vi[i+1].z-vi[i].z);
      float  di1  =  di  +  sqrt( d.x*d.x+d.y*d.y+d.z*d.z);
      float tii1=di1/dn;

      float alpha=(t-tii)/(tii1-tii);
      pos.x=alpha*vi[i+1].x+(1-alpha)*vi[i].x;
      pos.y=alpha*vi[i+1].y+(1-alpha)*vi[i].y;
      pos.z=alpha*vi[i+1].z+(1-alpha)*vi[i].z;
    }
    break;
  case 4: // velocity spline
    {
      const MF<Vec2f>& pts=pPA->keySpline;
      Vec2f a((pts[0].x-pts[1].x)*3+1,(pts[0].y-pts[1].y)*3+1);
      Vec2f b((pts[1].x-2*pts[0].x)*3,(pts[1].y-2*pts[0].y)*3);
      Vec2f c(pts[0].x*3,pts[0].y*3);
      Vec2f d(0,0);
      float u=bisec(t,a,b,c,d);
      int i=u*(nValues-1);
      float alpha=(u-float(i)/float(nValues-1))*(nValues-1);
      pos.x=alpha*vi[i+1].x+(1-alpha)*vi[i].x;
      pos.y=alpha*vi[i+1].y+(1-alpha)*vi[i].y;
      pos.z=alpha*vi[i+1].z+(1-alpha)*vi[i].z;
    }
    break;
  };

  pos.x += pPA->offset.GetValue().x;
  pos.y += pPA->offset.GetValue().y;
  pos.z += pPA->offset.GetValue().z;
	pPA->value_changed = pos;
}

// for cubic Bezier, quadratic Bezier and NURBS paths
void PositionAnimatorProxy::interpolateNURBS(float t)
{
  PositionAnimator *pPA=(PositionAnimator*) GetMediaObject();
  const MF<float>& ti=pPA->key;
  const MF<Vec3f>& vi=pPA->keyValue;
  int nKeys=ti.GetLength();
  int nValues=vi.GetLength();
  Vec3f pos;

  switch (pPA->keyType) {
  case 1: // discrete
    {
      int i=t*nValues;
      t=float(i)/float(nValues);
    }
    break;
  case 2: // linear
    {
      int i=t*(nValues-1);
      t=float(i)/float(nValues-1);
    }
    break;
  case 3: // paced
    {
      /*int i=t*(nValues-1),j;
      Vec3f d;

      float di=0;
      for(j=0;j<i;j++) {
        d=Vec3f(vi[j+1].x-vi[j].x, vi[j+1].y-vi[j].y, vi[j+1].z-vi[j].z);
        di += sqrt(d.x*d.x+d.y*d.y+d.z*d.z);
      }
      float dn=di;
      for(j=i;j<nValues-1;j++) {
        d=Vec3f(vi[j+1].x-vi[j].x, vi[j+1].y-vi[j].y, vi[j+1].z-vi[j].z);
        dn += sqrt(d.x*d.x+d.y*d.y+d.z*d.z);
      }
      float tii=di/dn;

      d=Vec3f(vi[i+1].x-vi[i].x, vi[i+1].y-vi[i].y, vi[i+1].z-vi[i].z);
      float di1 = di + sqrt(d.x*d.x+d.y*d.y+d.z*d.z);
      float tii1=di1/dn;

      float alpha=(t-tii)/(tii1-tii);
      pos.x=alpha*vi[i+1].x+(1-alpha)*vi[i].x;
      pos.y=alpha*vi[i+1].y+(1-alpha)*vi[i].y;
      pos.z=alpha*vi[i+1].z+(1-alpha)*vi[i].z;*/
    }
    break;
  case 4: // velocity spline
    {
      const MF<Vec2f>& pts=pPA->keySpline;
      Vec2f a((pts[0].x-pts[1].x)*3+1,(pts[0].y-pts[1].y)*3+1);
      Vec2f b((pts[1].x-2*pts[0].x)*3,(pts[1].y-2*pts[0].y)*3);
      Vec2f c(pts[0].x*3,pts[0].y*3);
      Vec2f d(0,0);
      t=bisec(t,a,b,c,d);
    }
    break;
  };

  // evaluate NURBS
  switch (pPA->keyValueType) {
  case 1: // quadratic bezier
    nurbs.init(Nurbs::QUADRATIC_BEZIER);
    break;
  case 2: // cubic bezier
    nurbs.init(Nurbs::CUBIC_BEZIER);
    break;
  case 3: // NURBS
    // key contains the knot vector
    // keyValue contains the control points
    const MF<float>& wi=pPA->weight;
    nurbs.init(nValues,nKeys,(float*) &ti[0],wi.GetLength(),(float*) &wi[0]);
    break;
  };

  int span=nurbs.findSpan(t);
  nurbs.basisFuns(span,t);
  nurbs.getPoint(span,0,vi,pos);

  pos.x += pPA->offset.GetValue().x;
  pos.y += pPA->offset.GetValue().y;
  pos.z += pPA->offset.GetValue().z;
	pPA->value_changed = pos;
}

// for fitting interpolating cubic spline to data
void PositionAnimatorProxy::interpolateFit(float)
{
}