/***********************************************************************
This software module was originally developed by Mikael Bourges-Sevenier
(Mindego Inc.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Mindego Inc. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Mikael Bourges-Sevenier at mikael@mindego.com.
----------------------------------------------------------------------*/

#include "OpenGl/OpenGl.h"
#include "SubdivisionSurfaceProxy.h"
#include "opengl/base/OpenGLVisualRenderer.h"
#include "sublib/include/subwrap.h"

static void getData(SubdivisionSurface* node, SurfaceData& data, TagData& tagData);

/***************************************************************************/
// Constructor - calls the inherited constructor from Node3DProxy
SubdivisionSurfaceProxy::SubdivisionSurfaceProxy(SubdivisionSurface * pSS)
  : GeometryProxy(pSS),
		m_pSSXL(NULL), _sw(0),f(0),v(0),vn(0),p(0),m_nTimeStamp(-1)
{
}

SubdivisionSurfaceProxy::~SubdivisionSurfaceProxy()
{
	delete m_pSSXL;
	
	if(f) delete[] f;
	if(v) delete[] v;
	if(vn) delete[] vn;
	if(p) delete[] p;
	f=NULL;
	v=vn=p=NULL;
	
	if(_sw)	delete _sw;
}

/* virtual */ BOOL
SubdivisionSurfaceProxy::Select(
		OpenGLVisualRenderer* pRenderer, // the renderer
    DWORD dwTime,                    // simulation time
    unsigned int& depth)             //
{
	// TODO [MBS] this object selection code crashes

  /*GLuint selectBuffer[4];

  ::glSelectBuffer(4, selectBuffer);
  ::glRenderMode(GL_SELECT);

  Render(pRenderer, dwTime);

  if (::glRenderMode(GL_RENDER) > 0)
    if (depth >= selectBuffer[1]) {
      depth = selectBuffer[1];

      return TRUE;
    }
	*/
  return FALSE;
}

// Render object

BOOL SubdivisionSurfaceProxy::Render(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
	SubdivisionSurface *pSS=GetSubdivisionSurface();
	Coordinate * pCoord = pSS->coord;
	Color* pColor = GetColor();
	TextureCoordinate* pTexCoord = (TextureCoordinate*)pSS->texCoord.GetNode(); //KM
	const MFNode<class SubdivSurfaceSector>& pSectors = pSS->sectors;
	
	_needUpdate=long(pSS->GetUpdateTime()) > m_nTimeStamp ||
		(pCoord && (long(pCoord->GetUpdateTime()) > m_nTimeStamp)) ||
		(pColor && (long(pColor->GetUpdateTime()) > m_nTimeStamp)) ||
		(pTexCoord && (long(pTexCoord->GetUpdateTime()) > m_nTimeStamp) ||
		m_nDisplayList==0);
	
	for(int i=0;i<pSectors.GetLength();i++)
		_needUpdate |= (long(pSS->sectors[i]->GetUpdateTime()) > m_nTimeStamp);
	
	if (_needUpdate)
	{
		m_nTimeStamp=(long) dwTime;
		
		int subdivisionLevel=pSS->subdivisionLevel;
		int loopflag = 1 - pSS->subdivisionType;

		const MF<int>& invisibleEdgeIndex = GetInvisibleEdgeIndex();

		// Do we need to render Loop/XLoop
		if(loopflag && pSectors.GetLength() == 0)
		{
			// For the moment always build a new SS
			delete m_pSSXL;
			m_pSSXL = BuildSSXL(pSS);

			assert(m_pSSXL != NULL);
			
			int iLevel = pSS->subdivisionLevel;

			assert(iLevel < m_pSSXL->cLevels());
			
			// Current level
			const Hull& theLevel = m_pSSXL->aLevels(iLevel);

			// Create a new display list
			if(m_nDisplayList) 
				::glDeleteLists(m_nDisplayList,1);
			
			m_nDisplayList = ::glGenLists(1);
			
			::glNewList(m_nDisplayList, GL_COMPILE_AND_EXECUTE);
			::glPushAttrib(GL_POLYGON_BIT);
			
			if((Bool)pSS->solid)
			{
				::glEnable(GL_CULL_FACE);
				::glCullFace(GL_FRONT);
			}
			else
			{
				::glDisable(GL_CULL_FACE);
			}

			for(int iFacet = 0; iFacet < theLevel.cFacets(); iFacet++)
			{
				const Facet& f = theLevel.aFacets(iFacet);

				assert(f.cVertices() == 3);

				::glBegin(GL_POLYGON);

				for(int iVertex = 0; iVertex < f.cVertices(); iVertex++)
				{
					const Vector3D& normal = theLevel.aNormals(f.aVertexNormals(iVertex));
					::glNormal3fv(reinterpret_cast<const float*>(&normal));

					if(theLevel.cColours() > 0)
					{
						const Vector3D& colour = theLevel.aColours(f.aVertexColours(iVertex));
						::glColor3fv(reinterpret_cast<const float*>(&colour));
					}

					if(theLevel.cUVWs() > 0)
					{
						const Vector3D& uvw = theLevel.aUVWs(f.aVertexUVWs(iVertex));
						::glTexCoord2fv(reinterpret_cast<const float*>(&uvw));
					}

					const Vector3D& vertex = theLevel.aVertices(f.aVertices(iVertex)).Vector();
					::glVertex3fv(reinterpret_cast<const float*>(&vertex));
				}

				::glEnd();
			}

			::glPopAttrib();
			::glEndList();
			
			_needUpdate = false;

			// We've handled everything; just return
			return TRUE;
		}
		
		SurfaceData data;
		TagData tagData;
		getData(pSS, data, tagData);
		
		// create a new mesh if necessary
		// MBS: if the mesh is recreated all the time, then deletion crashes
		//      if the mesh is updated, we have a out of memory error

		if(_sw) { delete _sw; _sw=NULL; }
		
		if(_sw == NULL)
			_sw = new SubWrap(subdivisionLevel, loopflag, data, tagData);
		//else
		//	_sw->update(subdivisionLevel, data, tagData);
		
		GetSubdivisionSurface()->solid=(_sw->isOpenMesh() ? 0 : 1);
		
		// resize if necessary
		int nvertices;
		int nfaces;
		_sw->getNumbers(subdivisionLevel, nvertices, nfaces);
		
		// get the mesh
		if(f==NULL) {
			f=new int[((loopflag==1) ? 4 : 5 )*nfaces]; // faces
			v=new float[nvertices*3]; // vertices
			vn=new float[nvertices*3]; // vertex normals
			p=new float[nvertices*2]; // texture params
		}
		else {
			if(sizeof(f) !=((loopflag==1) ? 4 : 5 )*nfaces) {
				delete[] f;
				f=new int[((loopflag==1) ? 4 : 5 )*nfaces]; // faces
			}
			if(sizeof(v) != nvertices*3) {
				delete[] v;
				delete[] vn;
				delete[] p;
				
				v=new float[nvertices*3]; // vertices
				vn=new float[nvertices*3]; // vertex normals
				p=new float[nvertices*2]; // texture params
			}
		}
		
		// copy values
		_sw->getFaces(subdivisionLevel, f); 	// faces are stored as <numVertices, vertex ids> for all faces
		_sw->getVertexData(subdivisionLevel, v, vn, p);
		
		// render mesh
		if(m_nDisplayList) 
			::glDeleteLists(m_nDisplayList,1);
		
		m_nDisplayList = ::glGenLists(1);
		
		::glNewList(m_nDisplayList, GL_COMPILE_AND_EXECUTE);
		::glPushAttrib(GL_POLYGON_BIT);
		
		//::glEnable(GL_CULL_FACE);
		
		for(int i=0,k=0;i<nfaces;i++) {
			int nvert=f[k];
			k++;
			::glBegin(GL_POLYGON);
			for(int j=0;j<nvert;j++) {
				int vert=f[k+j];
				::glNormal3fv   (vn+3*vert);
				::glTexCoord2fv (p+2*vert);
				::glVertex3fv   (v+3*vert);
			}
			::glEnd();
			k+=nvert;
		}
		::glPopAttrib();
		::glEndList();
		
		// end render mesh
		
		_needUpdate = false;
		
		// free tag info memory
		if(tagData.sectors) delete[] tagData.sectors;
	}
	else
		::glCallList(m_nDisplayList);
	
	return TRUE;
}

// Render object description (when viewing the scene graph)

/* virtual */ void
SubdivisionSurfaceProxy::RenderDescription(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
	int children = ((GetCoord() ? 1 : 0) +
		(GetColor() ? 1 : 0) +
		(GetTexCoord() ? 1 : 0));
	
	// TODO add sectors
	
	GLfloat color[] = { 1.0f, 0.0f, 1.0f, 0.75f};
	
	::glColor4fv(color);
	
	RenderDescriptionBox(pRenderer, "SubdivisionSurface");
	
	::glPushMatrix();
	
	::glScalef(1.0f / children, 1.0f, 1.0f);
	::glTranslatef(0.0f, -1.0f, 0.0f);
	
	if (/* pIndexedFaceSet-> */ GetCoord()) {
		RenderDescriptionBox(pRenderer, "Coordinate");
		
		::glTranslatef(1.0f, 0.0f, 0.0f);
	}
	
	if (/* pIndexedFaceSet-> */ GetColor()) {
		RenderDescriptionBox(pRenderer, "Color");
		
		::glTranslatef(1.0f, 0.0f, 0.0f);
	}
	
	if (/* pIndexedFaceSet-> */ GetTexCoord())
		RenderDescriptionBox(pRenderer, "TextureCoordinate");
	
	::glPopMatrix();
}

void getData(SubdivisionSurface* node, SurfaceData& data, TagData& tagData)
{
	// set surface data
	Coordinate* c = node->coord;
	TextureCoordinate* p = (TextureCoordinate*)node->texCoord.GetNode();//KM
	
	if(c)
	{
		data.noVtx = c->point.GetLength();
		data.pos=(float*) &c->point[0].x;
	}	
	
	if(p)
	{
		ASSERT(c && (c->point.GetLength() == p->point.GetLength()));
		data.param=(float*) &p->point[0].x;
	}
	//
	data.noFaces = node->coordIndex.GetLength();
	data.face = (int*) &(node->coordIndex[0]);  
	
	tagData.noCreaseLineIndex = node->creaseEdgeIndex.GetLength();
	tagData.creaseLineIndex= (int*) &node->creaseEdgeIndex[0];
	//
	tagData.noCornerIndex = node->cornerVertexIndex.GetLength();
	tagData.cornerIndex= (int*) &node->cornerVertexIndex[0];
	//
	tagData.noDartIndex = node->dartVertexIndex.GetLength();
	tagData.dartIndex= (int*) &node->dartVertexIndex[0];
	//
	tagData.noCreaseIndex = node->creaseVertexIndex.GetLength();
	tagData.creaseIndex= (int*) &node->creaseVertexIndex[0];
	
	//
	tagData.noSector = node->sectors.GetLength();
	tagData.sectors=new TagSectorInfo[tagData.noSector];
	for(int i=0;i<tagData.noSector;i++) {
		if (node->sectors.GetLength()>0) {
			SubdivSurfaceSector *sector=(SubdivSurfaceSector *) node->sectors[i];
			tagData.sectors[i].faceIndex =     (long) sector->faceIndex;
			tagData.sectors[i].vertexIndex =   (long) sector->vertexIndex;
			tagData.sectors[i].tag =           (long) sector->tag;
			tagData.sectors[i].flatness =      (float) sector->flatness;
			tagData.sectors[i].theta =         (float) sector->theta;
			tagData.sectors[i].normal =        (float*) &sector->normal;
			tagData.sectors[i].normalTension = (float) sector->normalTension;
		}
	}
}

/// Get bounding box
void SubdivisionSurfaceProxy::GetBbox(Vec3f& bboxSize, Vec3f& bboxCenter) const
{
	Coordinate* c = GetSubdivisionSurface()->coord;
	const MF<Vec3f>& points=c->point;
	
	Vec3f mini(FLT_MAX,FLT_MAX,FLT_MAX),maxi(-FLT_MAX,-FLT_MAX,-FLT_MAX);
	
	for(int i=0;i<points.GetLength();i++) {
		if(points[i].x<mini.x) mini.x=points[i].x;
		if(points[i].y<mini.y) mini.y=points[i].y;
		if(points[i].z<mini.z) mini.z=points[i].z;
		
		if(points[i].x>maxi.x) maxi.x=points[i].x;
		if(points[i].y<maxi.y) maxi.y=points[i].y;
		if(points[i].z<maxi.z) maxi.z=points[i].z;
	}
	bboxSize=maxi-mini;
	bboxCenter=(maxi-mini)/2;
}

SubdivisionSurfaceXLoop* SubdivisionSurfaceProxy::BuildSSXL(SubdivisionSurface* pSS)
{
	Coordinate* coord = pSS->coord;
	SFNode<class Color>* color = &pSS->color; // MBS 031804: changed SFColorNode to SFNode<class Color>
	TextureCoordinate* texCoord = (TextureCoordinate*)pSS->texCoord.GetNode(); //KM

	// max number of values, -1 no values 
	int coordMax = -1;
	int colorMax = -1;
	int texCoordMax = -1;
	
	// ignore empty nodes 
	if(coord && (coordMax = coord->point.GetLength()) == 0) coord = NULL;
	if(color && color->GetNode() != NULL && (colorMax = ((MF<Color3f>*)color->GetNode()->GetField(0))->GetLength()) == 0) color = NULL;
	if(texCoord && (texCoordMax = texCoord->point.GetLength()) == 0) texCoord = NULL;
	
	Hull control;                            // Build control hull from node data
	
	int iVertex;                                                  // Add vertices
	
	for(iVertex = 0; iVertex < coordMax; iVertex++)
	{
		const Vec3f& V = coord->point[iVertex];
		
		control.AddVertex(V.x, V.y, V.z);
	}
	
	if(texCoord != NULL)                                             // Add UVWs
	{
		int iUVW;
		
		for(iUVW = 0; iUVW < texCoordMax; iUVW++)
		{
			const Vec2f& UVW = texCoord->point[iUVW];
			
			control.AddUVW(UVW.x, UVW.y);
		}
	}
	else{                                             // Generate default mapping
		// TODO
	}
	
	if(color != NULL && color->GetNode() != NULL)                  // Add colours
	{
		MF<Color3f>& col = *(MF<Color3f>*)color->GetNode()->GetField(0);
		int iColour;
		
		for(iColour = 0; iColour < colorMax; iColour++)
		{
			const Color3f& RGB = col[iColour];
			
			control.AddColour(RGB[0], RGB[1], RGB[2]);
		}
	}
	
	MF<int>& coordIndex = pSS->coordIndex;
	MF<int>& texCoordIndex = pSS->texCoordIndex;
	MF<int>& colorIndex = pSS->colorIndex;
	int coordIndexMax = coordIndex.GetLength();
	int texCoordIndexMax = texCoordIndex.GetLength();
	int colorIndexMax = colorIndex.GetLength();
	
	int cVertices = 0;                                              // Add facets
	int iFirstVertex = 0;
	int iVertexIndex = 0;
	
	for(iVertexIndex = 0; iVertexIndex < coordIndexMax; iVertexIndex++)
	{
		if((cVertices > 0) &&                               // Find facet end:
			((coordIndex[iVertexIndex] == -1) ||            // Last vertex in facet
			(iVertexIndex + 1 == coordIndex.GetLength())))  // or last one in list
		{
			Facet F;
			
			assert(control.cFacets() == control.cNormals());
			
			int iF = control.cFacets();
			
			for(iVertex = 0; iVertex < cVertices; iVertex++)
			{
				int iV = coordIndex[iFirstVertex + iVertex];
				int iUVW = -1;
				int iColour = -1;
				
				if(control.cUVWs() > 0)
				{
					iUVW = (texCoordIndexMax > 0) ?
						texCoordIndex[iFirstVertex + iVertex] :      // Authored value
					iV;                                          // Default Value
				}
				
				if(control.cColours() > 0)
				{
					if((Bool)pSS->colorPerVertex)
					{
						iColour = (colorIndexMax > 0) ?
							colorIndex[iFirstVertex + iVertex] :    // Authored value
						iV;                                     // Default Value
					}
					else{
						iColour = (colorIndexMax > 0) ?
							colorIndex[iF] :                        // Authored value
						iF;                                     // Default Value
					}
				}
				
				assert((iUVW == -1) ||
					((iUVW >= 0) && (iUVW < control.cUVWs())));
				assert((iColour == -1) ||
					((iColour >= 0) && (iColour < control.cColours())));
				
				F.AddVertex(iV, -1, iUVW, iColour);    // TODO: ignoring author normals
			}
			
			if((Bool)pSS->ccw)                // Hull structure uses CW ordering; reverse
			{
				for(iVertex = 0; iVertex < cVertices / 2; iVertex++)
				{
					int iOtherVertex = cVertices - iVertex - 1;
					int tmp;
					
					tmp = F.aVertices(iVertex);
					F.aVertices(iVertex) = F.aVertices(iOtherVertex);
					F.aVertices(iOtherVertex) = tmp;
					
					tmp = F.aVertexUVWs(iVertex);
					F.aVertexUVWs(iVertex) = F.aVertexUVWs(iOtherVertex);
					F.aVertexUVWs(iOtherVertex) = tmp;
					
					tmp = F.aVertexColours(iVertex);
					F.aVertexColours(iVertex) = F.aVertexColours(iOtherVertex);
					F.aVertexColours(iOtherVertex) = tmp;
					
					// TODO: remember to reverse vertex normals when done
				}
			}
			
			int iWhere = control.AddFacet(F);
			
			assert(iWhere == iF);                        // Sanity chek; must be true
			
			iFirstVertex += cVertices + 1;                  // Prepare for next facet
			
			cVertices = 0;
		}
		else{                                         // Still no -1; keep counting
			cVertices++;
		}
	}
	
	control.Triangulate((Bool)pSS->convex);    // Loop subdivision requires triangles;

	MF<int>& invisibleEdgeIndex = pSS->invisibleEdgeIndex;
	int invisibleEdgeIndexMax = invisibleEdgeIndex.GetLength();
	
	// Tag invisible edges
	for(int iInvisibleEdge = 0; iInvisibleEdge < invisibleEdgeIndexMax; iInvisibleEdge += 2)
	{
		int iHead = invisibleEdgeIndex[iInvisibleEdge];
		int iTail = invisibleEdgeIndex[iInvisibleEdge + 1];
		int iEdge = control.iEdge(iHead, iTail);
		
		assert(iEdge != -1);
		
		control.aEdges(iEdge).SetInvisible();
	}

	MF<int>& creaseVertexIndex = pSS->creaseVertexIndex;
	int creaseVertexIndexMax = creaseVertexIndex.GetLength();

	// Add sharp vertices
	for(int iCreaseVertex = 0; iCreaseVertex < creaseVertexIndexMax; iCreaseVertex++)
	{
		// No variable sharpness
		control.aVertices(creaseVertexIndex[iCreaseVertex]).SetSharpness(100);
	}
	
	MF<int>& creaseEdgeIndex = pSS->creaseEdgeIndex;
	int creaseEdgeIndexMax = creaseEdgeIndex.GetLength();

	// Add sharp edges
	for(int iCreaseEdge = 0; iCreaseEdge < creaseEdgeIndexMax; iCreaseEdge += 2)
	{
		int iHead = creaseEdgeIndex[iCreaseEdge];
		int iTail = creaseEdgeIndex[iCreaseEdge + 1];
		int iEdge = control.iEdge(iHead, iTail);
		
		assert(iEdge != -1);
		
		// No variable sharpness
		control.aEdges(iEdge).SetSharpness(100);
	}
	
	return new SubdivisionSurfaceXLoop(control, pSS->subdivisionLevel);  // Create it
}
