/***********************************************************************
This software module was originally developed by Sebastian Amengual
(Universidad Politecnica de Madrid, UPM) in the course of development of 
the MPEG-4 Systems (ISO/IEC 14496-1) standard. This software module is
an implementation of a part of one or more MPEG-4 Systems 
(ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems
(ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 Systems 
(ISO/IEC 14496-1) free license to this software module or modifications
thereof for use in hardware or software products claiming conformance to
the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use this 
software module in hardware or software products are advised that its 
use may infringe existing patents. The original developer of this 
software module and his/her company, the subsequent editors and their
companies, and ISO/IEC have no liability for use of this software module
or modifications thereof in an implementation. Copyright is not 
released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
UPM retains full right to use the code for his/her own purpose, assign 
or donate the code to a third party and to inhibit third parties from 
using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming 
products. This copyright notice must be included in all copies or 
derivative works. Copyright  2005
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Francisco Moran at fmb@gti.ssr.upm.es.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Player3d
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  WaveSurfProxy
FILE:       WaveSurfProxy.cpp
AUTHOR:     Sebastin Amengual & Marcos Avils (UPM)
OVERVIEW:   Reads the bitstream written by WaveSurfDec.cpp and sends the 
            data inside it to WaveSurf in order to be decoded. Retrieves 
            the final mesh and renders.
-------------------------------------------------------------------------*/

#include "WaveSurfProxy.h"
#include "opengl/AFX/BitWrapperProxy.h"
#include "IM1Decoders/AFX/AFX.h"
#include "IM1Decoders/AFX/WaveSurf/WSSDataExchange.h"
#include "IM1Decoders/AFX/WaveSurf/WaveSurfDec.h"
#include "IM1Decoders/AFX/WaveSurf/WaveSurf.h"


//-----------------------------------------------------------------------------
// Name: WaveletSubdivisionSurfaceProxy()
// Desc: The Constructor creates a WaveSurf object, that will decode the 
//       details and add them to the subdivided base mesh.
//-----------------------------------------------------------------------------

WaveletSubdivisionSurfaceProxy::WaveletSubdivisionSurfaceProxy(WaveletSubdivisionSurface *pWSS)
    : AfxStreamProxy(pWSS), _normals(NULL), _indices(NULL)
{
    _pWaveSurf = new WaveSurf();
}


//-----------------------------------------------------------------------------
// Name: ~WaveletSubdivisionSurfaceProxy()
// Desc: Destructor.
//-----------------------------------------------------------------------------

WaveletSubdivisionSurfaceProxy::~WaveletSubdivisionSurfaceProxy()
{
    delete _pWaveSurf;
    delete[] _normals;
    delete[] _indices;
}


//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Renders the final mesh performing Vertex Array Rendering .
//-----------------------------------------------------------------------------

BOOL WaveletSubdivisionSurfaceProxy::Render(OpenGLVisualRenderer * /*pRenderer*/, DWORD /*dwTime*/)
{
    // Enable vertex and normal arrays
    ::glEnableClientState(GL_VERTEX_ARRAY);
    ::glEnableClientState(GL_NORMAL_ARRAY);

    // Set pointers to buffers
    ::glVertexPointer(3, GL_FLOAT, 0, _pWaveSurf ->GetCoord());
    ::glNormalPointer(GL_FLOAT, 0, _normals);

    // Render
    ::glDrawElements(GL_TRIANGLES, _pWaveSurf->GetCoordIndexLength() * 3 / 4, GL_UNSIGNED_INT, _indices);
    ::glDisableClientState(GL_VERTEX_ARRAY);
    ::glDisableClientState(GL_NORMAL_ARRAY);

    return TRUE;
}


//-----------------------------------------------------------------------------
// Name: RenderFromStream()
// Desc: Extracts al the info in the stream. Calls the proper methods of 
//       WaveSurf and calculates the normal per vertex of the final mesh in 
//       order to perform Gouraud shading.
//-----------------------------------------------------------------------------

BOOL WaveletSubdivisionSurfaceProxy::RenderFromStream(MediaStream *poStream, int /*id*/, DWORD /*dwTime*/)
{
    if (!poStream)
        return FALSE;

    int rnSize, rnStatus;
    DWORD rdvTime;
    BOOL rbEOU;

    BYTE *pOutput = static_cast<BYTE *>(poStream->Fetch(rnSize, rdvTime, rbEOU, rnStatus));
    if (!pOutput)
        return NULL;

    BYTE tag;
    DWORD len;
    void *pData = NULL;

    bool hasScaleCoeff(false);
    int nbBPSC(0), nbBPX(0), nbBPY(0), nbBPZ(0);
    int wtype(0);
    bool lift(false);
    int numberOfLevels(0);
    int lengthNbBits(0);
    float xMax(0.0f), yMax(0.0f), zMax(0.0f);
    bool isInLocalCoordinates(false);
    bool isPartial(false);

    char *pStream = NULL;
    unsigned int nStreamLength = 0;

    WSSDataExchange dataExchange(pOutput, true);
    pData = dataExchange.FetchNext(tag, len);
    while (pData)
    {
        switch (tag)
        {
            case WSSDataExchange::TAG_hasScaleCoeff:
                hasScaleCoeff = *static_cast<bool *>(pData);
                break;

            case WSSDataExchange::TAG_NbBpSC:
                nbBPSC = *static_cast<int *>(pData);
                break;

            case WSSDataExchange::TAG_NbBPX:
                nbBPX = *static_cast<int *>(pData);
                break;

            case WSSDataExchange::TAG_NbBPY:
                nbBPY = *static_cast<int *>(pData);
                break;

            case WSSDataExchange::TAG_NbBPZ:
                nbBPZ = *static_cast<int *>(pData);
                break;

            case WSSDataExchange::TAG_Wtype:
                wtype = *static_cast<int *>(pData);
                break;

            case WSSDataExchange::TAG_lift:
                lift = *static_cast<bool *>(pData);
                break;

            case WSSDataExchange::TAG_NumberOfLevels:
                numberOfLevels = *static_cast<int *>(pData);
                break;

            case WSSDataExchange::TAG_Xmax:
                xMax = *static_cast<float *>(pData);
                break;

            case WSSDataExchange::TAG_Ymax:
                yMax = *static_cast<float *>(pData);
                break;

            case WSSDataExchange::TAG_Zmax:
                zMax = *static_cast<float *>(pData);
                break;

            case WSSDataExchange::TAG_stream:
                pStream = static_cast<char *>(pData);
                nStreamLength = len;
                break;

            default:
                break;
        }

        pData = dataExchange.FetchNext(tag, len);
    }

    if (!pStream)
        return FALSE;


    WaveletSubdivisionSurface *pWSS = static_cast<WaveletSubdivisionSurface *>(GetMediaObject());
    IndexedFaceSet *pIndexedFaceSet = static_cast<IndexedFaceSet *>(pWSS->baseMesh.GetNode());

    const MF<int> &coordIndex = pIndexedFaceSet->coordIndex;
    MF<Vec3f> &coordArray = pIndexedFaceSet->coord->point;

    float *xArray = new float[coordArray.GetLength()];
    float *yArray = new float[coordArray.GetLength()];
    float *zArray = new float[coordArray.GetLength()];
    int *indexArray = new int[coordIndex.GetLength()];

    int i;
    for (i = 0; i < coordArray.GetLength(); i++)
    {
        xArray[i] = coordArray[i].x;
        yArray[i] = coordArray[i].y;
        zArray[i] = coordArray[i].z;
    }

    for (i = 0; i < coordIndex.GetLength(); i++)
        indexArray[i] = coordIndex[i];


    _pWaveSurf->Setup(nbBPX, nbBPY, nbBPZ, numberOfLevels, lengthNbBits, 
                      xMax, yMax, zMax, hasScaleCoeff, wtype, lift, 
                      isPartial, isInLocalCoordinates);
    _pWaveSurf->SetStream(pStream, nStreamLength);
    _pWaveSurf->SetMesh(xArray, yArray, zArray, indexArray, coordArray.GetLength(), coordIndex.GetLength());

    FillIndicesAndNormalsArrays();


    delete[] xArray;
    delete[] yArray;
    delete[] zArray;
    delete[] indexArray;

    return TRUE;
}


//-----------------------------------------------------------------------------
// Name: RenderFromBuffer()
// Desc: Not used.
//-----------------------------------------------------------------------------

BOOL WaveletSubdivisionSurfaceProxy::RenderFromBuffer(SFString *pBuffer, int /*id*/)
{
    if (!pBuffer)
        return FALSE;

    char *pStream = const_cast<char *>(static_cast<const char *>(*pBuffer));
    unsigned int nStreamLength = pBuffer->GetLength();

    WaveletSubdivisionSurface *pWSS = static_cast<WaveletSubdivisionSurface *>(GetMediaObject());
    IndexedFaceSet *pIndexedFaceSet = static_cast<IndexedFaceSet *>(pWSS->baseMesh.GetNode());

    const MF<int> &coordIndex = pIndexedFaceSet->coordIndex;
    MF<Vec3f> &coordArray = pIndexedFaceSet->coord->point;

    float *xArray = new float[coordArray.GetLength()];
    float *yArray = new float[coordArray.GetLength()];
    float *zArray = new float[coordArray.GetLength()];
    int *indexArray = new int[coordIndex.GetLength()];

    int i;
    for (i = 0; i < coordArray.GetLength(); i++)
    {
        xArray[i] = coordArray[i].x;
        yArray[i] = coordArray[i].y;
        zArray[i] = coordArray[i].z;
    }

    for (i = 0; i < coordIndex.GetLength(); i++)
        indexArray[i] = coordIndex[i];

    _pWaveSurf->SetStream(pStream, nStreamLength);
    _pWaveSurf->SetMesh(xArray, yArray, zArray, indexArray, coordArray.GetLength(), coordIndex.GetLength());

    FillIndicesAndNormalsArrays();

    delete[] xArray;
    delete[] yArray;
    delete[] zArray;
    delete[] indexArray;


    return TRUE;
}



//-----------------------------------------------------------------------------
// Name: Select()
// Desc: Only for debugging. Traverse scene graph, checking for intersection
//-----------------------------------------------------------------------------

BOOL WaveletSubdivisionSurfaceProxy::Select(OpenGLVisualRenderer* pRenderer, DWORD dwTime, unsigned int& depth)
{
    GLuint selectBuffer[4];

    ::glSelectBuffer(4, selectBuffer);
    ::glRenderMode(GL_SELECT);

    Render(pRenderer, dwTime);

    if (::glRenderMode(GL_RENDER) > 0)
    {
        if (depth >= selectBuffer[1])
        {
            depth = selectBuffer[1];
            return TRUE;
        }
    }

    return FALSE;
}

void WaveletSubdivisionSurfaceProxy::FillIndicesAndNormalsArrays()
{
    delete[] _indices;
    delete[] _normals;

    int numFaces = _pWaveSurf->GetCoordIndexLength() / 4;
    int numVertices = _pWaveSurf->GetCoordLength();

    _indices = new unsigned int[numFaces * 3];
    _normals = new Vec3f[numVertices];

    const int *wssIndices = _pWaveSurf->GetCoordIndex();

    // Vec3f implmentation sucks, doing the following with const Vec3f *wssCoords
    //     normal = (wssCoords[v1] - wssCoords[v0])
    // would be interpreted as
    //     float tmp = (const float *)(wssCoords[v1] - wssCoords[v0]);
    //     normal.operator = tmp;
    // which isn't what we wanted
    Vec3f *wssCoords = reinterpret_cast<Vec3f *>(const_cast<float *>(_pWaveSurf->GetCoord()));

    int v0, v1, v2;
    Vec3f normal;

    int i;
    for (i = 0; i < numFaces; i++)
    {
        _indices[3 * i] = v0 = wssIndices[4 * i];
        _indices[3 * i + 1] = v1 = wssIndices[4 * i + 1];
        _indices[3 * i + 2] = v2 = wssIndices[4 * i + 2];

        normal = wssCoords[v1] - wssCoords[v0]; 
        normal = normal.Cross(wssCoords[v2] - wssCoords[v0]);
        _normals[v0] += normal;
        _normals[v1] += normal;
        _normals[v2] += normal;
    }

    for (i = 0; i < numVertices; i++)
        _normals[i].Normalize();
}
