/********************************************************
*
*	$Author: mig $
*	$JustDate:  4/17/02 $
*	$Revision: 1.1 $
*	$Archive: afx/Solid/DensityTLS.cpp $
*
*********************************************************/
// AM #include <stdafx.h>
#include <assert.h>
#include "OpenGL/AFX/Solid/gdsgdl.h"
#include "OpenGL/AFX/Solid/VolumeTreeNode/SolidTools.h"
#include "OpenGL/AFX/Solid/Point.h"
#include "OpenGL/AFX/Solid/DensityTLS.h"


/********************************************************************************/
//
//  This file regroups most of the density tools including DensityMatrix and 
//  DensityArray Class.
//
//  Author : Eric Baillargeon, SGDL Systme Inc.
//  Date   : November 28th, 2001
//
//*******************************************************************************/

// [K_VolumeGridHighLimit , K_VolumeGridLowLimit] volume must resize grid to smaller than
// Global_Grid
const float K_VolumeGridHighLimit   = 8;
const float K_VolumeGridLowLimit    = 0.1;

//**************************************************************************
//
//	This function flag the current polygonising volume density points 
//  to already visited.
//
//	Author : Eric Baillargeon, SGDL Systems
//	Date : October 8th, 2001
//  Modified : April 4th, 2002
//
//  Input : Index i,j,k
//
//  Output :Min.Max x,y,z;
//
//  SL - Optimization for this function. A recursive algorithm
//  isn't possible because the density matrix is too large
//  and it causes stack overflow most of the time.  So I used an array
//  to simulate the recursive algorithm. 
// 
//*************************************************************************

void DensityMatrix::FlagNeighbours(int i, int j, int k)
{
    int       CurPt    = 0;
    int       SPLength = 0;
    int       l_i      = 0;
    int       l_j      = 0;
    int       l_k      = 0;
    int       Icpt     = 0;
    int       Jcpt     = 0;
    int       Kcpt     = 0;
    const int Adjust[3] = {-1, 0, 1}; 

    //Array to stock solid neighbour points
    unsigned char (*SP)[3] = 0;

    float x = C1.x + ( (float)i * ((float)dimension[0] / (float)division));
    float y = C1.y + ( (float)j * ((float)dimension[1] / (float)division));
    float z = C1.z - ( (float)k * ((float)dimension[2] / (float)division)); 

    //Init the MinMaxArray
    MinMaxArray[eBounds_MinX] = MinMaxArray[eBounds_MaxX] = x;
    MinMaxArray[eBounds_MinY] = MinMaxArray[eBounds_MaxY] = y;
    MinMaxArray[eBounds_MinZ] = MinMaxArray[eBounds_MaxZ] = z;    
    
    //I used unsigned char since the division of the density matrix 
    //isn't greater than 255.  SP = Solid Points.    
    SP = new unsigned char[Max_Division*Max_Division*Max_Division][3];
    assert(SP);   

    //Initialization
    MyPoints_HasBeenVisited[i][j][k]= true;
    SP[0][0] = i;
    SP[0][1] = j;
    SP[0][2] = k;
    SPLength = 1;

    //Check the neighbour points of each solid point founded.
    for (CurPt = 0; CurPt < SPLength; CurPt++)
    {
        i = SP[CurPt][0];
        j = SP[CurPt][1];
        k = SP[CurPt][2];        

        //Check all the 26 points around the point we found. The array SP 
        //grow each time we found a solid (dens > 0) neighbour point.
        for (Icpt = 0; Icpt < 3; Icpt++)
        {
            l_i = i + Adjust[Icpt];

            //Check to stay in the array limit...
            if ((l_i >= 0) && (l_i < division))
            {
                for (Jcpt = 0; Jcpt < 3; Jcpt++)
                {   
                    l_j = j + Adjust[Jcpt];

                    if ((l_j >= 0) && (l_j < division))
                    {
                        for (Kcpt = 0; Kcpt < 3; Kcpt++)
                        {
                            l_k = k + Adjust[Kcpt];
                            
                            if ((l_k >= 0) && (l_k < division))
                                SPLength += CheckForSolidPoint(l_i, l_j, l_k, SP[SPLength]);
                        }
                    }
                }
            }
        }
    }
        
    delete [] SP;
}

//**************************************************************************
//
//	This function first check the density of a point and if it isn't NULL,
//  it update the MinMaxArray and flag this point to already visited.
//
//	Author : Simon Lavallee, SGDL Systems
//	Date   : April 8th, 2002
//
//  Input  : Index i,j,k
//
//  Output : uch *SolidPoint -> Container to stock the point if it is solid
//           int -> Number of point added in the array (0 or 1)
// 
//*************************************************************************
inline int DensityMatrix::CheckForSolidPoint(int i, int j, int k, unsigned char *SolidPoint)
{    
    if (!MyPoints_HasBeenVisited[i][j][k] && 
        CheckDensity(MyPoints_density[i][j][k]))
    {       
        //We want to keep the max and the min x, y, z.
        CheckForMinMax(i, j, k);

        //Flag this point
        MyPoints_HasBeenVisited[i][j][k] = true;

        //Insert the point in the array
        SolidPoint[0] = i;
        SolidPoint[1] = j;
        SolidPoint[2] = k;

        //Return the number of added point
        return 1;
    }
    return 0;
}

/********************************************************************\
    Function:  CheckForMinMax
              	
    Input:     (i,j,k) index of a point
    Output:     
    
    Author:     Eric Baillargeon SGDL Systme Inc.
    Created:    2001/11/22	

		
*********************************************************************/

void DensityMatrix::CheckForMinMax(int i, int j, int k)
{
    float x = C1.x + ( (float)i * ((float)dimension[0] / (float)division));
    float y = C1.y + ( (float)j * ((float)dimension[1] / (float)division));
    float z = C1.z - ( (float)k * ((float)dimension[2] / (float)division));    
    
    MinMaxArray[eBounds_MaxX] = max(x, MinMaxArray[eBounds_MaxX], x);
    MinMaxArray[eBounds_MinX] = minFloat(x, MinMaxArray[eBounds_MinX]);

    MinMaxArray[eBounds_MaxY] = max(y, MinMaxArray[eBounds_MaxY], y);
    MinMaxArray[eBounds_MinY] = minFloat(y, MinMaxArray[eBounds_MinY]);

    MinMaxArray[eBounds_MaxZ] = max(z, MinMaxArray[eBounds_MaxZ], z);
    MinMaxArray[eBounds_MinZ] = minFloat(z, MinMaxArray[eBounds_MinZ]);
}
/********************************************************************\
    Function:  GetBboxAssociatedGrid
              	
    Input:     Current grid
    Output:    If grid has changed 
    
    Author:     Eric Baillargeon SGDL Systme Inc.
    Created:    2001/11/22	

		
*********************************************************************/

bool DensityMatrix::GetBboxAssociatedGrid(float& grid)
{
    float BBoxVolume(0.0f);
    float dx = MinMaxArray[eBounds_MaxX] - MinMaxArray[eBounds_MinX];
    float dy = MinMaxArray[eBounds_MaxY] - MinMaxArray[eBounds_MinY];
    float dz = MinMaxArray[eBounds_MaxZ] - MinMaxArray[eBounds_MinZ];
    BBoxVolume = dx * dy * dz;

    if (BBoxVolume < K_VolumeGridHighLimit)
    {
        grid =  max( ((BBoxVolume / K_VolumeGridHighLimit) + 0.35f) * Global_Grid * 0.65, 0.05f, 0.0f);
        return TRUE;
    }
    grid = Global_Grid;
    return FALSE;
}

//***************************************************		
//
//  Init the DensityMatrix class members
//
//	Author : Eric Baillargeon, SGDL Systems
//	Date : October 8th, 2001
//
//***************************************************	
void DensityMatrix::Init(Volume_Tree_Node &The_Node, DensityArray &UserDensArray)
{  
    Max_i               = Max_Division;
    Max_j               = Max_Division;
    Max_k               = Max_Division;
    Current_i           = 0x0000;
    Current_j           = 0x0000;
    Current_k           = 0x0000;
    MaxDensity          = 0x0000;
    division            = Max_Division;
    ShowAllDensities    = !UserDensArray.values;

    if (MyPoints_density == 0)
        MyPoints_density = new int[Max_Division][Max_Division][Max_Division];

    if (MyPoints_HasBeenVisited == 0)
        MyPoints_HasBeenVisited = new bool[Max_Division][Max_Division][Max_Division];

    if (!ShowAllDensities)
        VolumeDensityList.SetDefault(1000,-1);

    CalculateC1( dimension[0],  dimension[1],  dimension[2]);
    EvaluateDensity(The_Node,dimension[0],dimension[1],dimension[2]);

    // If the user did not define any DensityList ... get the default one.
    if (!ShowAllDensities)
    {   
        The_Node.ShowAllDensities = false;
        //Set the Density Matrix and Node currentDensity
        currentDensity = The_Node.currentDensity = UserDensArray;
        
    }
}

/***************************************************
//	
//  Evaluate the density points for a given Volume Tree;
//
//	Author : Eric Baillargeon, SGDL Systems
//	Date : October 9th, 2001
//
//  Input  : Volume_tree_Node
//           Bounding Box dx, dy , dz
//
//  Output : 
//
/***************************************************/
void DensityMatrix::EvaluateDensity(Volume_Tree_Node &The_Node, double dx ,double dy, double dz)
{
    double XInterval = dx / division;
    double YInterval = dy / division;
    double ZInterval = dz / division;

    Current_i = 0;
    Current_j = 0;
    Current_k = 0;

    for (int i = 0; i < Max_i ; i ++)
    {
        for (int j = 0; j < Max_j ; j ++)
        {       
            for (int k = 0; k < Max_k ; k ++)
            {
                MyPoints_density[i][j][k] = The_Node.EvaluateDensity(C1.x + (i * XInterval), C1.y + (j * YInterval), C1.z - (k * ZInterval));
                
                if (!ShowAllDensities)
                {
                    VolumeDensityList.values[ MyPoints_density[i][j][k] ] = MyPoints_density[i][j][k];
                    MaxDensity = maxInt(MaxDensity, MyPoints_density[i][j][k]);
                }
            }
        }
    }

    memset(MyPoints_HasBeenVisited, false, Max_i*Max_j*Max_k*sizeof(bool));

    // The array size is equivalent to the greatest density in the volume
    if (!ShowAllDensities)
    {
        VolumeDensityList.NumberOfDensity = MaxDensity + 1;
        VolumeDensityList.Sort();
    }
}

/***************************************************
//	
//  Compute the (-x, -y, z most) corner of the Bbox
//
//	Author : Eric Baillargeon, SGDL Systems
//	Date : October 9th, 2001
//
//  Input  :  Bounding Box dx, dy , dz
//
//  Output : 
//
/***************************************************/

void DensityMatrix::CalculateC1(double dx, double dy, double dz)
{
    C1.x = -(dx/2);
    C1.y = -(dy/2);
    C1.z =  (dz/2);
}


/***************************************************
//	
//  Return the index of a point inside the volume defined by the volumic tree
//
//	Author : Eric Baillargeon, SGDL Systems
//	Date : October 9th, 2001
//
//  Input  :  Index i,j,k
//
//  Output : 
//
/***************************************************/
void DensityMatrix::GetInsidePoint(int& i, int& j, int& k)
{   
    LookForInside(i,j,k);
    if (!(i == -1))
    {        
        FlagNeighbours(i,j,k);
    }
}


/***************************************************
//	
//  Look in the density matrix for a point inside the volume defined by the volumic tree
//
//	Author : Eric Baillargeon, SGDL Systems
//	Date : October 9th, 2001
//
//  Input  :  Index  i, j, k
//
//  Output : 
//
/***************************************************/
void DensityMatrix::LookForInside(int& id, int& jd, int& kd)
{ 
    for (int i = Current_i; i < Max_i; i++)
    {
        for (int j = 0 ; j < Max_j; j++)
        {
            for (int k = 0 ; k < Max_k; k++)
            {
                if (!(MyPoints_HasBeenVisited[i][j][k]) && this->CheckDensity(MyPoints_density[i][j][k]))
                {
                    Current_i = id = i;
                    Current_j = jd = j;
                    Current_k = kd = k;
                    
                    return;
                }
            }
        }
    }        
    id = -1;
}

/********************************************************************\
Function:  ResetToUnflag()

  Reset all the AsBeenVisited flags to false for further evaluations .. 
  
    Input:     
    Output:     
    
      Author:     Eric Baillargeon SGDL Systme Inc.
      Created:    2001/10/29	
      
        
*********************************************************************/

void DensityMatrix::ResetToUnflag()
{
    
    for (int k5 = 0; k5 < Max_i; k5++){
        for (int k6 = 0; k6 < Max_j; k6++){
            for (int k7 = 0; k7 < Max_k; k7++){
                
                MyPoints_HasBeenVisited[k5][k6][k7] = false;
            }
        }        
    }
}

/***************************************************
//	
//  DensityMatrix constructor
//
//	Author : Eric Baillargeon, SGDL Systems
//	Date : October 9th, 2001
//
//  Input  : 
//
//  Output : 
//
/***************************************************/
DensityMatrix::DensityMatrix()
{
   //Since we use unsigned char to stock index,
   //the division must be lower than 256.
   assert(Max_Division < 256);

   MyPoints_HasBeenVisited  = NULL;
   MyPoints_density         = NULL;

   ShowAllDensities         = true;   

   dimension[0] = 0;
   dimension[1] = 0;
   dimension[2] = 0;    
}

/***************************************************
//	
//  DensityMatrix destructor
//
//	Author : Eric Baillargeon, SGDL Systems
//	Date : October 9th, 2001
// Modified: 25-01-02
//
//  Input  : 
//
//  Output : 
//
/***************************************************/
DensityMatrix::~DensityMatrix()
{
   if (MyPoints_HasBeenVisited != NULL)
   {
      delete [] MyPoints_HasBeenVisited;
      MyPoints_HasBeenVisited = NULL;
   }

   if (MyPoints_density != NULL)
   {
      delete [] MyPoints_density;
      MyPoints_density = NULL;
   }
}

/********************************************************************\
Function:  maxInt

  Input:     int a, int b
  Output:    the maximum of a & b 
  
    Author:     Eric Baillargeon SGDL Systme Inc.
    Created:    2001/10/25	
     
      
*********************************************************************/
int maxInt(int a, int b)
{
    if (a > b)
        return a;
    else
        return b;
}
/********************************************************************\
Function:  minFloat

  Input:     float a, float b
  Output:    the minimum of a & b 
  
    Author:     Eric Baillargeon SGDL Systme Inc.
    Created:    2001/10/25	
     
      
*********************************************************************/
float minFloat(float a, float b)
{
    return ( (a < b) ? a : b);
}

//********************************************************************
//  Function:  DensityMatrix::CheckDensity()
//
//  Input:     int resulEval : F(x) bring to [0 1 2]
//             DensityArray densArray : The array of density
//  Output:    If resulEval define in densArray.
//  
//  Author:     Eric Baillargeon SGDL Systme Inc.
//  Created:    2001/10/25	
//  Modified: Simon Lavallee (17-04-2002)
//      
//*********************************************************************
bool DensityMatrix::CheckDensity(int resulEval)
{    
    if (ShowAllDensities)
        //If we want to show all density, just return true if
        //the density > 0 (optimisation)
        return (resulEval > 0);
    else
        //Else, call the global CheckDensity() (implemented in
        //SGDLTools.cpp)
        return ::CheckDensity(resulEval, currentDensity);    
}

/********************************************************************\
Function:  Sort()
            Sort a densityList for further utilisation.
  Input:    
  Output:    
  
    Author:     Eric Baillargeon SGDL Systme Inc.
    Created:    2001/10/25	
     
      
*********************************************************************/
void DensityArray::Sort()
{

    //Maximum to be decided ... 
  
    int tp[100];
    
    for (int k = 0; k < 100; k++)
    {
        tp[k] = -1;
    }
    int cpt = 0;       
    
    int max = 0;
    for (int i = 0; i < NumberOfDensity; i++)
    {        
        // Eliminate repetition : EX: 0 1 1 2 1 2 2 becomes 0 1 2
        if (values[i] > -1){
            int lolo = values[i];
            if (!(tp[lolo] == values[i])){          
                if (tp[lolo] == -1)
                {
                    cpt++;                    
                } //tp[values[i]]
                tp[values[i]] = values[i];
                
                // Get the max between the current max and the density beeing treated
                max = maxInt(max,values[i]);
            }          
        }
    }
    
    int WhereToPlace = 0;
    for (int j = 0; j <= max; j++){
        
        // regroup the density EX:  -1 1 2 -1 -1 5 becomes 1 2 5 -1
        if (tp[j] != -1){
            values[WhereToPlace] = tp[j];
            WhereToPlace++;                
        }
    }
    NumberOfDensity = cpt;  

}
/********************************************************************\
Function:  SetDefault

  Input:     numberOf       : How many elements in the array.
             defaultValue   : Value of the elements in the array.
  Output:    
  
    Author:     Eric Baillargeon SGDL Systme Inc.
    Created:    2001/10/25	
     
      
*********************************************************************/
void DensityArray::SetDefault(int numberOf,int defaultValue)
{
    if (values)
    {
        delete[] values;
    }
	if( numberOf != 0 ) {
    values = new int[numberOf];
    NumberOfDensity = numberOf;
    for (int i = 0; i < numberOf; i++)
    {
        values[i] = defaultValue;
    }
	}
	else
	{
		values = NULL;
	}
}
/********************************************************************\
Function:  DensityArray ( Constructor )

  Input:     MF<int>& UserList : The user density array;
  Output:    A copy of the user density array.
  
    Author:     Eric Baillargeon SGDL Systme Inc.
    Created:    2001/10/25	
     
      
*********************************************************************/
 /*
DensityArray::DensityArray(const MF<int>& UserList)
{
	int n;
    if ((n = UserList.GetLength()) > 0)
    {                
        values = new int[n];
        for (int i = 0; i < n; i++)
        {
            values[i] = UserList[i];
        }
    }
    else
    {
        values = NULL;
    }
    NumberOfDensity = n;
}
*/
/********************************************************************\
Function:  ValidateDensities
            This function validate the User Densities for the skin and 
            inside and put it in A_INT_Densities;

  Input:    int A_INT_Densities[],GvMFVec2f UserDensities
  change to Im1 - AM

  Output:    

    Author:     Eric Baillargeon SGDL Systme Inc.
    Created:    2001/10/25	


*********************************************************************/

void ValidateDensities(int A_INT_Densities[],int *UserDensities)
{
    if (UserDensities[K_INT_SKIN_INDEX] < 0)
    {
        UserDensities[K_INT_SKIN_INDEX] = K_INT_DEFAULTSKIN;
    }
    if (UserDensities[K_INT_INTERIOR_INDEX] < 0)
    {
        UserDensities[K_INT_INTERIOR_INDEX] = K_INT_DEFAULTINTERIOR;
    }  
    A_INT_Densities[K_INT_SKIN_INDEX]     = UserDensities[K_INT_SKIN_INDEX];
    A_INT_Densities[K_INT_INTERIOR_INDEX] = UserDensities[K_INT_INTERIOR_INDEX];

}


