//*******************************************************************
// SolidNode
//*******************************************************************
/***********************************************************************
This software module was originally developed by Alain Mignot (Mignot Informatique Graphique Inc),
Eric Paquet and Eric Baillargeon(SGDL Systemes Inc.)
 in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. (Mignot Informatique Graphique Inc. & SGDL Systemes Inc.)
 retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2004
************************************************************************/
#include "Opengl/AFX/Solid/stdafx.h"
#include <assert.h>
#include <ztl.h>
#include "Opengl/AFX/Solid/VolumeTreeNode/SolidNode.h"
#include "Opengl/AFX/Solid/VolumeTreeNode/SolidTools.h"
#include "Opengl/AFX/Solid/Polygonise.h"

//*******************************************************************
//  SolidNode constructor
//
//  By Eric Paquet
//  SGDL Systems inc 
//  Date : December 10th, 2001.
//*******************************************************************
SolidNode::SolidNode(const double coeff_ptr[],
			       int          coeff_number,
			       bool         is_solid,
				   const float  bbox_ptr[],
				   int          Node_Density[2]
                   ):
Volume_Tree_Node(bbox_ptr)
{
	assert(coeff_ptr	!= NULL &&
	       bbox_ptr	    != NULL && 
		   (coeff_number == K_QuadricNumberOfCoeff ||
		    coeff_number == K_QuarticNumberOfCoeff));

	init();

    coeff = new double[coeff_number];

	//Get a copy of the coefficients
	for ( int i = 0; i < coeff_number; i++)
        coeff[i] = coeff_ptr[i];

	nCoeff		  = coeff_number;
	solid		  = is_solid;
    density[0]	  = Node_Density[0];	
    density[1]	  = Node_Density[1];
	m_displaylist = 0;
    Manage_At_The_Top();
	
}

SolidNode::SolidNode(const float coeff_ptr[],
			       int         coeff_number,
			       bool        is_solid,
		           const float bbox_ptr[],
		           int         Node_Density[2]):
Volume_Tree_Node(bbox_ptr)
{
	assert(coeff_ptr     != NULL &&
	       bbox_ptr      != NULL && 
		   (coeff_number == K_QuadricNumberOfCoeff ||
		    coeff_number == K_QuarticNumberOfCoeff));

    init();
	coeff = new double[coeff_number];

	//Get a copy of the coefficients
	for ( int i = 0; i < coeff_number; i++)
		coeff[i] = static_cast<double>(coeff_ptr[i]);

	nCoeff		= coeff_number;
	solid		= is_solid;  
    density[0]	= Node_Density[0];
    density[1]	= Node_Density[1];
	m_displaylist = 0;
	Manage_At_The_Top();
	
}

SolidNode::SolidNode(void):
Volume_Tree_Node(false)
{
    init();
    solid       = false;
    coeff       = NULL;
    nCoeff      = 0;
    density[0]  = 1;
    density[1]  = 2;
	m_displaylist = 0;
	Manage_At_The_Top();
	
}

//******************************************************************
//	This function takes a point x, y, z and return true if the point
//	is inside the solid and false if the volume is outside.
//
//	Author : Eric Paquet
//	Date : October 9th, 2001
//	Modif : chloe ... for hybrid uses Solid_GetDensity
//******************************************************************
int SolidNode::EvaluateDensity(const double& x, const double& y, const double& z)const
{    
	assert(coeff != NULL &&
	      (nCoeff == K_QuadricNumberOfCoeff ||
	       nCoeff == K_QuarticNumberOfCoeff));

    int value = 2;

    if (solid)
    {
		//This call may modify is_inside
		Bounding_Box_Correction(bbox, &value, x, y, z);
    }

    if(value != 0)
    {
        if ( nCoeff == K_QuadricNumberOfCoeff )
		    value = DLz2_coe(coeff, x, y, z);
	    else if ( nCoeff == K_QuarticNumberOfCoeff )
		    value = DLz4_coe(coeff, x, y, z); 
    
        if (density[K_INT_INTERIOR_INDEX] != -1 && value > 0)
		    value = density[K_INT_INTERIOR_INDEX];
    }
    
    return value;
}

//*******************************************************************
//  SolidNode destructor 
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
SolidNode::~SolidNode(void)
{
    if ( ref == 0 && coeff != NULL)
        delete coeff;
}

//***************************************************
//	
//  SolidNode::EvaluateGradient()
//
//	Author : Simon Lavallee, SGDL Systems
//	Date   : March 27th, 2002
//
//  Input  : double x           -> X coordinate
//           double y           -> Y coordinate
//           double z           -> Z coordinate
//
//  Output : double MinValue    -> FieldValue of the implicit function
//           float  Color       -> Color of the point
//           double *Gradient   -> Gradient of the function
//
//  Desc   : This function evaluate the gradient
//           of the imlicit function at point x, y, z.
//           The gradient of the function return a vector
//           that is normal to the surface of the implicit
//           function.
//***************************************************
bool SolidNode::EvaluateGradient(const double& x,  const double& y, const double& z, 
                                double& Distance, double *Normal,   float *Color) const
{

    assert(coeff  != NULL &&
          (nCoeff == K_QuadricNumberOfCoeff || 
           nCoeff == K_QuarticNumberOfCoeff));
    
    double NormalLength = 0.0;
    double ToHighToBeTheGoodOne = 0.0;

    if (nCoeff == K_QuadricNumberOfCoeff)
    {
        //Get field value of the quadric (very approximative distance)
        Distance = DLz2_coeR(coeff, x, y, z);
        //Get the normal, using the gradient of the quadric
        Dlz2_Gradient(coeff, x, y, z, Normal);

        ToHighToBeTheGoodOne = 1.0;
    }
    
	else if (nCoeff == K_QuarticNumberOfCoeff)
    {
        //Get field value of the quartic (very approximative distance)
	    Distance = DLz4_coeR(coeff, x, y, z);
        //Get the normal, using the gradient of the quartic
        Dlz4_Gradient(coeff, x, y, z, Normal);

        ToHighToBeTheGoodOne = 10.0;
    }
    
    //For now, the distance we have is the field value of the 
    //implicit function that is actually not a very good approximation
    //of the distance between the point and the surface. The next step 
    //will project the point to the surface along the normal and return 
    //the distance between the initial and the final point.    
    
    if ((fabs(Distance) < ToHighToBeTheGoodOne) && !EpsilonEqual(Distance, 0.0))
    {
        //We compute the distance between the point and the surface only
        //if the field value isn't greater than one (that is certainly too 
        //far to be the good one) or isn't already on the surface (== 0.0).
        
        NormalLength = sqrt(Normal[0]*Normal[0] + 
                            Normal[1]*Normal[1] + 
                            Normal[2]*Normal[2]);

        //Normalize the vector
        Normal[0] /= NormalLength;
        Normal[1] /= NormalLength;
        Normal[2] /= NormalLength;

        //Get the distance between the point and the surface
        Distance = GetDistanceToSurface(coeff, nCoeff, Normal, 
                                        Distance, x, y, z);
    }

    if (solid && EpsilonGreaterOrEqual(Distance, 0.0))
    {
        //The gradient and the distance can be modified by the bbox.
        Bounding_Box_Gradient_Correction(bbox, Normal, &Distance, x, y, z);        
    }

    if (Color && !Undefined_Color(color))
    {
        //Get the color
        Color[0] = color[0];
        Color[1] = color[1];
        Color[2] = color[2];
    }

    return true;
}

//***************************************************
//	
//  SolidNode::GetFieldValue()
//
//	Author : Simon Lavallee, SGDL Systems
//	Date   : September 27th, 2002
//
//  Input  : double x           -> X coordinate
//           double y           -> Y coordinate
//           double z           -> Z coordinate
//
//  Output : bool               -> If we've been able to get the field Value.
//           double *fValue     -> The field value of the Implicit function
//
//  Desc   : Return the field value of the implicit function
//           at a given point.
//***************************************************
void SolidNode::GetPrimitives(ZArray<BoundingQuadric*> &outPrimitives,
                             const SL_Matrix &Transform, BBox &outGlobalBBox)
{
    if(nCoeff != K_QuadricNumberOfCoeff)
        return;

    BoundingQuadric *l_BQuadric = new BoundingQuadric;

    if(bbox != NULL)
    {
        l_BQuadric->BoundingBox.ma.Set((bbox[0] / 2.0) + 2*EPSILON, (bbox[1] / 2.0) + 2*EPSILON, (bbox[2] / 2.0) + 2*EPSILON);
        l_BQuadric->BoundingBox.mi.Set((-bbox[0] / 2.0) - 2*EPSILON, (-bbox[1] / 2.0) - 2*EPSILON, (-bbox[2] / 2.0) - 2*EPSILON);
    }

    else
        l_BQuadric->BoundingBox.SetEmpty();

    if(!Transform.IsIdentity())
    {
        SL_Matrix m = Transform;
        
        if(!l_BQuadric->BoundingBox.IsEmpty())
            l_BQuadric->BoundingBox *= m;

        SL_Matrix l_Quadric = QuadricMatrix(coeff[0], coeff[2], coeff[5],
                                         coeff[1]/2.0, coeff[4]/2.0, coeff[3]/2.0,
                                         coeff[6]/2.0, coeff[7]/2.0, coeff[8]/2.0,
                                         coeff[9]);

        m.Invert();
        l_Quadric.MultiplyLeft(m);
        m.Transpose();
        l_Quadric.MultiplyRight(m);

        l_BQuadric->Coefficients[0] = (double) l_Quadric.x[0][0];       //A
        l_BQuadric->Coefficients[2] = (double) l_Quadric.x[1][1];       //B
        l_BQuadric->Coefficients[5] = (double) l_Quadric.x[2][2];       //C
        l_BQuadric->Coefficients[1] = (double) l_Quadric.x[0][1] * 2.0; //D
        l_BQuadric->Coefficients[4] = (double) l_Quadric.x[1][2] * 2.0; //E
        l_BQuadric->Coefficients[3] = (double) l_Quadric.x[0][2] * 2.0; //F
        l_BQuadric->Coefficients[6] = (double) l_Quadric.x[0][3] * 2.0; //G
        l_BQuadric->Coefficients[7] = (double) l_Quadric.x[1][3] * 2.0; //H
        l_BQuadric->Coefficients[8] = (double) l_Quadric.x[2][3] * 2.0; //J
        l_BQuadric->Coefficients[9] = (double) l_Quadric.x[3][3];       //K
    }

    else    
        memcpy(l_BQuadric->Coefficients, coeff, K_QuadricNumberOfCoeff*sizeof(double));

//AM    outPrimitives.Append(l_BQuadric);
//AM remplace par:
		outPrimitives += l_BQuadric;

    if(!(l_BQuadric->BoundingBox).IsEmpty())
        outGlobalBBox.Union(l_BQuadric->BoundingBox);
}

//*******************************************************************
//  This function manages the variable At_The_Top
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void SolidNode::Manage_At_The_Top()
{
	At_The_Top = true;
}

//****************************************************
//	
//  Get the color of the point, based on the fact
//  that the point is in or outside the shape.
//
//	Author : Eric Baillargeon, SGDL Systems
//	Date : October 15th, 2001
//  Modified: Simon Lavallee, SGDL Systems
//            April 15th, 2002
//
//  Input  : x,y,z   
//
//  Output : float set_color[3] -> color to set
//           bool  is_inside    -> true if the point is inside
//                                 the shape.
//
//***************************************************
bool SolidNode::GetColor(const double &x, const double &y, const double &z, float set_color[])const
{
    bool is_inside = isInsideWithoutDensityCheck(x,y,z);
    
	//Set the color
	if ( is_inside && !Undefined_Color(color) )
	{
		set_color[0] = color[0];
		set_color[1] = color[1];
		set_color[2] = color[2];
	}

    return is_inside;
}

//*******************************************************************
//  This function sets the solidity of the SolidNode
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void SolidNode::Set_Solidity(bool to_set)
{
	solid = to_set;
}

//*******************************************************************
//  This function the SolidNode coeff(use for the gradient)
//
//  By Eric Baillargeon
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
double * SolidNode::GetCoeff(double x, double y , double z)
{
    return coeff;
}

//*******************************************************************
//  This function the SolidNode number of coeff (use for the gradient)
//
//  By Eric Baillargeon
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
int  SolidNode::GetNumOfCoeff(double x, double y , double z)
{
    return nCoeff;
}

//*******************************************************************
//  This function removes a reference from the SolidNode 
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void SolidNode::UnRef()
{
    ref--;

    if ( ref == 0 )
        delete this;
}

//*******************************************************************
//  This function set the graphic representation to the SolidNode 
//
//  By Alain Mignot
//  Mignot Informatique Graphique Inc.
//  Date : may 2004
//*******************************************************************
	void SolidNode::SetGraphic( SolidGraphic *Igraphic)
	{
		Graphic = Igraphic;
	}


//*******************************************************************
//  This function Get the graphic representation from the SolidNode 
//
//  By Alain Mignot
//  Mignot Informatique Graphique Inc.
//  Date : may 2004
//*******************************************************************
	SolidGraphic	*SolidNode::GetGraphic()
	{
		return( Graphic);
	}

//*******************************************************************
//  This function displays the graphic representation of the SolidNode 
//
//  By Alain Mignot
//  Mignot Informatique Graphique Inc.
//  Date : may 2004
//*******************************************************************
	bool SolidNode::Render(Matrix &rMatrix)
	{

	::glPushMatrix();
	if(!m_displaylist) {

		m_displaylist = ::glGenLists(1);
		
		::glNewList(m_displaylist, GL_COMPILE_AND_EXECUTE);
		::glPushAttrib(GL_POLYGON_BIT);


//		::glEnable(GL_CULL_FACE);
		::glShadeModel(GL_SMOOTH);
		::glEnable(GL_LIGHTING);
		if( Graphic != NULL)Graphic->Display();



		::glPopAttrib();
		::glEndList();


	}
	else
		::glCallList(m_displaylist);

	::glPopMatrix();
	return TRUE;


	}


//EOF***************************************************************