//*******************************************************************
//Volume_Tree_Binary_Op
//*******************************************************************

#include "..\\stdafx.h"
#include <assert.h>
#include "..\\VolumeTreeNode\\SolidTools.h"

//*******************************************************************
//  Volume_Tree_Binary_Op constructor 
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
Volume_Tree_Binary_Op::Volume_Tree_Binary_Op(SGDLBinaryOperator the_Op,
											 const Volume_Tree_Node *To_Be_V1,
											 const Volume_Tree_Node *To_Be_V2)
{
	assert(To_Be_V1 != NULL && To_Be_V2 != NULL);

	the_Operation = the_Op;
	v1 = const_cast<Volume_Tree_Node *>(To_Be_V1);
	v2 = const_cast<Volume_Tree_Node *>(To_Be_V2);    
	Manage_At_The_Top();
}

//*******************************************************************
//  Volume_Tree_Binary_Op constructor  
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
Volume_Tree_Binary_Op::Volume_Tree_Binary_Op(SGDLBinaryOperator the_Op,
											 const Volume_Tree_Node *To_Be_V1,
											 const Volume_Tree_Node *To_Be_V2,
                                             const float bbox_ptr[]):
Volume_Tree_Node(bbox_ptr)
{
	assert(To_Be_V1 != NULL && To_Be_V2 != NULL);

	the_Operation = the_Op;
	v1 = const_cast<Volume_Tree_Node *>(To_Be_V1);
	v2 = const_cast<Volume_Tree_Node *>(To_Be_V2);

	Manage_At_The_Top();
}

//**************************************************************************
//  This function return the density if the point x, y, z is inside and
//  false(0) if the point is outside
//
//  Author : Eric Paquet
//  Date : October 4th,2001
//
//
//***************************************************************************
int Volume_Tree_Binary_Op::EvaluateDensity(const double& x, const double& y, const double& z)const
{    
    int to_return = 0;
    int l_BoundingVolumeCount = 0;

    if(At_The_Top && !isInsideBoundingVolume(x, y, z))
        return 0;

	switch(the_Operation)
	{
	    //Beginning of Kleene operators
		case SGDLAnd :
		{
	        int v1_density = v1->EvaluateDensity(x, y, z) % WHICH_AIRE;
	        int v2_density = v2->EvaluateDensity(x, y, z) % WHICH_AIRE;
	        v1_density < v2_density ? to_return = v1_density : to_return = v2_density;
	        break;
		}
		case SGDLOr :
		{
			int v1_density = v1->EvaluateDensity(x, y, z) % WHICH_AIRE;
	        int v2_density = v2->EvaluateDensity(x, y, z) % WHICH_AIRE;
	        v1_density > v2_density ? to_return = v1_density : to_return = v2_density;
	        break;
		}
	    case SGDLImplication:
	    {
            //  >>| 0 1 2
            //  ---------
            //  0 | 2 2 2
            //  1 | 1 1 2
            //  2 | 0 1 2
            int v1_density = v1->EvaluateDensity(x, y, z) % WHICH_AIRE;
	        int v2_density = v2->EvaluateDensity(x, y, z) % WHICH_AIRE;

            if ( v1_density == 0 ||
                 v2_density == 2)
                to_return = 2;
            else if ( v1_density == 1 ||
                      v2_density == 1 )
                to_return = 1;
            else
                to_return = 0;
	        break;
            
	    }
	    case SGDLReciprocal:
	    {
            //  <<| 0 1 2
            //  ---------
            //  0 | 2 1 0
            //  1 | 1 1 1
            //  2 | 0 1 2
            int v1_density = v1->EvaluateDensity(x, y, z) % WHICH_AIRE;
	        int v2_density = v2->EvaluateDensity(x, y, z) % WHICH_AIRE;

            if ( v1_density == 1 ||
                 v2_density == 1 )
                to_return = 1;
            else if ( !(( v1_density == 2 ) ^ (v2_density == 2)) )
                to_return = 2;
            else
                to_return = 0;
            
	        break;
	    }
	    //End of Kleene operators
	    //Beginning of the general arithmetic operators
		case SGDLPlus :
		{
			to_return = v1->EvaluateDensity(x, y, z) + v2->EvaluateDensity(x, y, z);
	    break;
		}
		case SGDLMinus :
		{
			to_return = v1->EvaluateDensity(x, y, z) - v2->EvaluateDensity(x, y, z);
	    if ( to_return < 0 )
		to_return = 0;
	    break;
		}
		case SGDLMulti :
		{
			to_return = v1->EvaluateDensity(x, y, z) * v2->EvaluateDensity(x, y, z);
	    break;
		}
	    case SGDLExpo :
	    {
	        int v1_density = v1->EvaluateDensity(x, y, z);
	        int v2_density = v2->EvaluateDensity(x, y, z);
	        if ( v1_density == 0 && v2_density == 0 )
		    to_return = 0;
	        else
		    to_return = (int)pow(v1_density, v2_density);
	        break;
	    }
	    case SGDLPGCD :
	    {
	        to_return = GCD(v1->EvaluateDensity(x, y, z), v2->EvaluateDensity(x, y, z));

	        break;
	    }
	    case SGDLPPCM :
	    {
	        to_return = LCM(v1->EvaluateDensity(x, y, z), v2->EvaluateDensity(x, y, z));
	        break;
	    }
	    case SGDLMod :
	    {
	        int v2_density = v2->EvaluateDensity(x, y, z);
	        if ( v2_density != 0 )
		    to_return = v1->EvaluateDensity(x, y, z) % v2_density;
	        else
		    to_return = v1->EvaluateDensity(x, y, z);
	        break;
	    }
	    case SGDLAbsMinus :
	    {
	        to_return = v1->EvaluateDensity(x, y, z) - v2->EvaluateDensity(x, y, z);
	        if ( to_return < 0 )
		    to_return = -to_return;
	        break;
	    }
	    case SGDLMax:
	    {
	        int v1_density = v1->EvaluateDensity(x, y, z);
	        int v2_density = v2->EvaluateDensity(x, y, z);
	        v1_density > v2_density ? to_return = v1_density : to_return = v2_density;
	        break;
	    }
	    case SGDLMin:
	    {
	        int v1_density = v1->EvaluateDensity(x, y, z);
	        int v2_density = v2->EvaluateDensity(x, y, z);
	        v1_density < v2_density ? to_return = v1_density : to_return = v2_density;
	        break;
	    }
	    //End of the general arithmetic operators
	    //Beginning of test operators
	    case SGDLEquality :
	    {
	        //Return the value if it's true
	        int v1_density = v1->EvaluateDensity(x, y, z);
	        if ( v1_density == v2->EvaluateDensity(x, y, z) )
		    to_return = v1_density;
	        break;
	    }
	    case SGDLGreaterOrEqual :
	    {
	        //Return the second if this is true
	        int v2_density = v2->EvaluateDensity(x, y, z);
	        if ( v1->EvaluateDensity(x, y, z) >= v2_density )
		    to_return = v2_density;
	        break;
	    }
	    case SGDLGreater :
	    {
	        //Return the second if it's true
	        int v2_density = v2->EvaluateDensity(x, y, z);
	        if ( v1->EvaluateDensity(x, y, z) > v2_density )
		    to_return = v2_density;
	        break;
	    }
	    case SGDLLessOrEqual :
	    {
	        //Return the second if it's true
	        int v2_density = v2->EvaluateDensity(x, y, z);
	        if ( v1->EvaluateDensity(x, y, z) <= v2_density )
		    to_return = v2_density;
	        break;
	    }
	    case SGDLLess :
	    {
	        //Return the second if it's true
	        int v2_density = v2->EvaluateDensity(x, y, z);
	        if ( v1->EvaluateDensity(x, y, z) < v2_density )
		    to_return = v2_density;
	        break;
	    }
	    case SGDLDifference:
	    {
	        //Return the second if it's true
	        int v2_density = v2->EvaluateDensity(x, y, z);
	        if ( v1->EvaluateDensity(x, y, z) != v2_density )
		    to_return = v2_density;
	        break;
	    }
	    //End of test operators
		default:
		{
			assert(false); //Do not know this operation
		}
	}

	if ( At_The_Top )
		//This call may modify to_return
		Bounding_Box_Correction(bbox, &to_return, x, y, z);
	
	return to_return;

}


//***************************************************
//	
//  Volume_Tree_Binary_Op::EvaluateGradient()
//
//	Author : Simon Lavallee, SGDL Systems
//	Date   : March 27th, 2002
//
//  Input  : double x           -> X coordinate
//           double y           -> Y coordinate
//           double z           -> Z coordinate
//
//  Output : double &Distance   -> Distance between the surface and the point
//           float  Color       -> Color of the point
//           double *Normal     -> Normal of the point
//           bool               -> true if we've been able to get the normal
//
//  Desc   : This function computes the normal of each
//           of its volumes and keeps the one that the has the shortest
//           distance between the point and the surface. The normal can 
//           be modified depending on the operator.
//***************************************************
bool Volume_Tree_Binary_Op::EvaluateGradient(const double& x,  const double& y, const double& z, 
                                             double& Distance, double* Normal,  float *Color) const
{
    bool   IsInside        = isInside(x,y,z); //TBC    

    //The distance, normal and color of V1 and V2
    double DistanceV1      = INVALID_DISTANCE;
    double DistanceV2      = INVALID_DISTANCE;

    double NormalV1[3]     = {0.0, 0.0, 0.0};
    double NormalV2[3]     = {0.0, 0.0, 0.0};

    float  *ColorV1        = (Color) ? new float[3] : NULL;
    float  *ColorV2        = (Color) ? new float[3] : NULL;    

    //Get distance, normal and color of V1 and V2.
    bool   NormalV1IsValid = v1->EvaluateGradient(x, y, z, DistanceV1, NormalV1, ColorV1);
    bool   NormalV2IsValid = v2->EvaluateGradient(x, y, z, DistanceV2, NormalV2, ColorV2);    

    if (!NormalV1IsValid)
    {
        //V1 is a Constant or a GShell
        Distance  = DistanceV2;

        Normal[0] = NormalV2[0]; //x
        Normal[1] = NormalV2[1]; //y
        Normal[2] = NormalV2[2]; //z

        return NormalV2IsValid;
    }

    else if (!NormalV2IsValid)
    {
        //V2 is a constant or a GShell
        Distance  = DistanceV1;

        Normal[0] = NormalV1[0]; //x
        Normal[1] = NormalV1[1]; //y
        Normal[2] = NormalV1[2]; //z

        return NormalV1IsValid;
    }    

    switch(the_Operation)
    {        

        case SGDLPlus  :
        case SGDLOr    :
        case SGDLMulti :
        case SGDLAnd   :
        {
            //For those operators, just keep the normal of the volume that 
            //the distance to the surface is the nearest of 0.
            if (fabs(DistanceV1) < fabs(DistanceV2))
            {
                Normal[0] = NormalV1[0]; //x
                Normal[1] = NormalV1[1]; //y
                Normal[2] = NormalV1[2]; //z

                Distance  = DistanceV1;

                if (Color)
                {
                    Color[0] = ColorV1[0];
                    Color[1] = ColorV1[1];
                    Color[2] = ColorV1[2];
                }
            }

            else
            {
                Normal[0] = NormalV2[0]; //x
                Normal[1] = NormalV2[1]; //y
                Normal[2] = NormalV2[2]; //z

                Distance  = DistanceV2;

                if (Color)
                {
                    Color[0] = ColorV2[0];
                    Color[1] = ColorV2[1];
                    Color[2] = ColorV2[2];
                }
            }            
            break;
        }

        case SGDLImplication:
        {            
            if (fabs(DistanceV1) < fabs(DistanceV2))
            {
                Normal[0] = -NormalV1[0]; //x
                Normal[1] = -NormalV1[1]; //y
                Normal[2] = -NormalV1[2]; //z

                if ((IsInside  && (DistanceV1 < 0.0)) ||
                    (!IsInside && (DistanceV1 > 0.0)))
                    //Ensure we got a correct sign of the distance...
                    Distance = -DistanceV1;
                
                else
                    Distance = DistanceV1;

                if (Color)
                {
                    Color[0] = ColorV1[0];
                    Color[1] = ColorV1[1];
                    Color[2] = ColorV1[2];
                }
            }

            else
            {
                Normal[0] = NormalV2[0]; //x
                Normal[1] = NormalV2[1]; //y
                Normal[2] = NormalV2[2]; //z

                if ((IsInside  && (DistanceV2 < 0.0)) ||
                    (!IsInside && (DistanceV2 > 0.0)))
                    //Ensure we got a correct sign of the distance...
                    Distance = -DistanceV2;

                else
                    Distance = DistanceV2;

                if (Color)
                {
                    Color[0] = ColorV2[0];
                    Color[1] = ColorV2[1];
                    Color[2] = ColorV2[2];
                }
            }
            break;
        }        

        case SGDLReciprocal:
        {            
            if (fabs(DistanceV1) < fabs(DistanceV2))
            {
                if (EpsilonGreaterOrEqual(DistanceV2, 0.0))
                {
                    Normal[0] = NormalV1[0]; //x
                    Normal[1] = NormalV1[1]; //y
                    Normal[2] = NormalV1[2]; //z

                    Distance = DistanceV1;
                }
                
                else
                {
                    Normal[0] = -NormalV1[0]; //x
                    Normal[1] = -NormalV1[1]; //y
                    Normal[2] = -NormalV1[2]; //z

                    Distance  = -DistanceV1;
                }

                if (Color)
                {
                    Color[0] = ColorV1[0];
                    Color[1] = ColorV1[1];
                    Color[2] = ColorV1[2];
                }
            }

            else 
            {
                if (EpsilonGreaterOrEqual(DistanceV1, 0.0))
                {
                    Normal[0] = NormalV2[0]; //x
                    Normal[1] = NormalV2[1]; //y
                    Normal[2] = NormalV2[2]; //z

                    Distance = DistanceV2;
                }
                
                else
                {
                    Normal[0] = -NormalV2[0]; //x
                    Normal[1] = -NormalV2[1]; //y
                    Normal[2] = -NormalV2[2]; //z

                    Distance = -DistanceV2;
                }

                if (Color)
                {
                    Color[0] = ColorV2[0];
                    Color[1] = ColorV2[1];
                    Color[2] = ColorV2[2];
                }
            }
            break;            
        }

		case SGDLMod :            //To be implemented
        case SGDLPPCM :           //To be implemented
        case SGDLPGCD :           //To be implemented

		case SGDLAbsMinus :       //To be implemented
        case SGDLLess :           //To be implemented
        case SGDLLessOrEqual :    //To be implemented
        case SGDLGreater :        //To be implemented
        case SGDLGreaterOrEqual : //To be implemented
        case SGDLEquality :       //To be implemented
        case SGDLMin :            //To be implemented        
        case SGDLMax :            //To be implemented

		case SGDLDifference :  
        case SGDLMinus :
        {
            if (fabs(DistanceV1) < fabs(DistanceV2))
            {
                Normal[0] = NormalV1[0]; //x
                Normal[1] = NormalV1[1]; //y
                Normal[2] = NormalV1[2]; //z

                if ((IsInside && (DistanceV1 < 0.0)) ||
                    (!IsInside && (DistanceV1 > 0.0)))
                    Distance = -DistanceV1;
                else
                    Distance = DistanceV1;
            }
            else
            {
                Normal[0] = -NormalV2[0]; //x
                Normal[1] = -NormalV2[1]; //y
                Normal[2] = -NormalV2[2]; //z

                if ((IsInside && (DistanceV2 < 0.0)) ||
                    (!IsInside && (DistanceV2 > 0.0)))
                    Distance = -DistanceV2;
                else
                    Distance = DistanceV2;
            }

            if (Color)
            {
                Color[0] = ColorV1[0];
                Color[1] = ColorV1[1];
                Color[2] = ColorV1[2];
            }
            break;
        }

        case SGDLExpo :     //Not tested yet
        {            
            Normal[0] = NormalV1[0]; //x
            Normal[1] = NormalV1[1]; //y
            Normal[2] = NormalV1[2]; //z

            Distance  = DistanceV1;                
            break;
        }        
            
        default:
        {
            assert(false); //Do not know this operation or the operation
                           //isn't supported yet when calculating the normal.
            return false;
        }            
    }   

    if (At_The_Top)
    {   
        //The bbox can modify the distance and the normal
        Bounding_Box_Gradient_Correction(bbox, Normal, &Distance, x, y, z);     
    }

    if (Color)
    {
        delete ColorV1;
        delete ColorV2;
        
        if (!Undefined_Color(color))
        {
            //If this operation node has its own color, keep it.
            Color[0] = color[0];
            Color[1] = color[1];
            Color[2] = color[2];
        }
    }    

    return true;
}

//***************************************************
//	
//  Volume_Tree_Binary_Op::GetPrimitives()
//
//	Author : Simon Lavallee, SGDL Systems
//	Date   : March 27th, 2002
//
//  Input  : 
//
//  Output : Array<BoundingQuadric*> &outPrimitives -> array of bounding quadrics
//           const SL_Matrix &Transform                -> cumuled transformations
//           BBox &outGlobalBBox                    -> global bounding box
//
//  Desc   : This function retrieves all the transformed primitives (quadrics only)
//           and their respectives bounding box.
//***************************************************
void Volume_Tree_Binary_Op::GetPrimitives(ZArray<BoundingQuadric*> &outPrimitives,
                                          const SL_Matrix &Transform, BBox &outGlobalBBox)
{
    v1->GetPrimitives(outPrimitives, Transform, outGlobalBBox);
    v2->GetPrimitives(outPrimitives, Transform, outGlobalBBox);
    
    if(bbox != NULL)
    {
        BBox l_Bbox;
        l_Bbox.ma.Set(bbox[0] / 2.0, bbox[1] / 2.0, bbox[2] / 2.0);
        l_Bbox.ma.Set(-bbox[0] / 2.0, -bbox[1] / 2.0, -bbox[2] / 2.0);

        if(!Transform.IsIdentity())
            l_Bbox *= Transform;

        outGlobalBBox.Union(l_Bbox);
    }
}

/********************************************************************\
Function:  

  Input:     
  Output:     
  
    Author:	Eric Baillargeon SGDL Systme Inc.
    Created:	2001/10/26	
    
      
*********************************************************************/
double * Volume_Tree_Binary_Op::GetCoeff(double x, double y , double z)
{
    if (v1->isInside(x,y,z))
	return (v1->GetCoeff(x,y,z));
    else
	// Not from the left subtree, must be in the left one
	return (v2->GetCoeff(x,y,z));
}

/********************************************************************\
Function:  

  Input:     
  Output:     
  
    Author:	Eric Baillargeon SGDL Systme Inc.
    Created:	2001/10/26	
    
      
*********************************************************************/
int Volume_Tree_Binary_Op::GetNumOfCoeff(double x, double y , double z){
    if (v1->isInside(x,y,z))
	return (v1->GetNumOfCoeff(x,y,z));
    else
	// Not from the left subtree, must be in the left one
	return (v2->GetNumOfCoeff(x,y,z));
}

/***************************************************
//	
//  Volume_Tree_Binary_Op::GetColor()
//
//  Get the color of the point based on the fact that
//  is in or outside.
//
//	Author : Eric Baillargeon, SGDL Systems
//	Date : October 15th, 2001
//  Modified: Simon Lavallee (now just get the color)
//  Date: April 15th, 2002
//
//  Input  : x,y,z
//
//  Output : set_color -> color of the point
//           bool      -> if the point is inside the volume
//
/***************************************************/
bool Volume_Tree_Binary_Op::GetColor(const double &x, 
                                     const double &y, 
                                     const double &z, 
                                     float set_color[])const
{	
    //FIRST VOLUME
	float color_v1[] = {-1.0, -1.0, -1.0};
	bool inside_v1 = v1->GetColor(x, y, z, color_v1);
	
    //SECOND VOLUME
	float color_v2[] = {-1.0, -1.0, -1.0};
	bool inside_v2 = v2->GetColor(x, y, z, color_v2);

	if ( !Undefined_Color(color_v1) && !Undefined_Color(color_v2) )
	{
		//Do nothing here because the point is inside the two volume
        //so we can't determine wich color is the good one.
	}
	else if( !Undefined_Color(color_v1) )
	{
		set_color[0] = color_v1[0];
		set_color[1] = color_v1[1];
		set_color[2] = color_v1[2];
	}
	else if( !Undefined_Color(color_v2) )
	{
		set_color[0] = color_v2[0];
		set_color[1] = color_v2[1];
		set_color[2] = color_v2[2];
	}
	
    bool inside_volume = inside_v1 || inside_v2;

	if (!Undefined_Color(color) && (inside_volume || At_The_Top))
	{
        //If this node has its own color, keep it.
		set_color[0] = color[0];
		set_color[1] = color[1];
		set_color[2] = color[2];
	}

    return inside_volume;
}

//*******************************************************************
//  This function manages the operator = between two Volume_Tree_Binary_Op
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
const Volume_Tree_Binary_Op& Volume_Tree_Binary_Op::operator=(const Volume_Tree_Binary_Op& The_Other)
{
	the_Operation = The_Other.the_Operation;
	v1 = const_cast<Volume_Tree_Node *>(The_Other.v1);
	v2 = const_cast<Volume_Tree_Node *>(The_Other.v2);
	return *this;
}

//*******************************************************************
//  This function manages the variable At_The_Top
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void Volume_Tree_Binary_Op::Manage_At_The_Top()
{
	v1->At_The_Top = false;
	v2->At_The_Top = false;
	At_The_Top = true;
}

//*******************************************************************
//  This function adds a reference to a Volume_Tree_Binary_Op
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void Volume_Tree_Binary_Op::Ref()
{
    ref++;
    v1->Ref();
    v2->Ref();
}

//*******************************************************************
//  This function removes a reference to a Volume_Tree_Binary_Op
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void Volume_Tree_Binary_Op::UnRef()
{
    ref--;
    v1->UnRef();
    v2->UnRef();

    if ( ref == 0 )       
        delete this;       
}

//*******************************************************************
//  This function set the graphic representation to the SolidNode 
//
//  By Alain Mignot
//  Mignot Informatique Graphique Inc.
//  Date : may 2004
//*******************************************************************
	void Volume_Tree_Binary_Op::SetGraphic( SolidGraphic *Igraphic)
	{
		Graphic = Igraphic;
	}


//*******************************************************************
//  This function Get the graphic representation from the SolidNode 
//
//  By Alain Mignot
//  Mignot Informatique Graphique Inc.
//  Date : may 2004
//*******************************************************************
	SolidGraphic	*Volume_Tree_Binary_Op::GetGraphic()
	{
		return( Graphic);
	}

//*******************************************************************
//  This function displays the graphic representation of the SolidNode 
//
//  By Alain Mignot
//  Mignot Informatique Graphique Inc.
//  Date : may 2004
//*******************************************************************
	bool Volume_Tree_Binary_Op::Render(Matrix & rMatrix)
	{




	return FALSE;


	}
//EOF***************************************************************************