//*******************************************************************
//Volume_Tree_Matrix
//*******************************************************************

#include "../stdafx.h"
#include <assert.h>
#include "Opengl/AFX/Solid/VolumeTreeNode/SolidTools.h"

//*******************************************************************
//  Volume_Tree_Matrix constructor 
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
Volume_Tree_Matrix::Volume_Tree_Matrix(Volume_Tree_Node *The_Node,
						               const SL_Matrix &the_transfo)
{
	assert ( The_Node != NULL );
    v = The_Node;
	castOK = CAST_OK;

    The_Matrix = the_transfo;

    this->Minimize(The_Node);

    The_Matrix_Inverted = Invert(The_Matrix);
	Manage_At_The_Top();    
}

//*******************************************************************
//  Volume_Tree_Matrix constructor  
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
Volume_Tree_Matrix::Volume_Tree_Matrix(Volume_Tree_Node *The_Node,
				   const float bbox_ptr[],
				   const SL_Matrix &the_transfo):
Volume_Tree_Node(bbox_ptr)
{
//	assert ( The_Node != NULL );
    v = The_Node;
	castOK = CAST_OK;

    The_Matrix = the_transfo;

    this->Minimize(The_Node);

    The_Matrix_Inverted = Invert(The_Matrix);
	Manage_At_The_Top();
}

//*******************************************************************
//  This function compress two or more VolumeTreeMatrix in a row into
//  only one node with the product of all the matrix.
//
//  By Simon Lavallee
//  SGDL Systems inc
//  Date : October 7th, 2002.
//*******************************************************************
void Volume_Tree_Matrix::Minimize(Volume_Tree_Node *The_Node)
{
	if( The_Node == NULL)return;

#if 0
    Volume_Tree_Matrix *TriedVTMatrix = dynamic_cast<Volume_Tree_Matrix*>(The_Node);

    while((TriedVTMatrix != NULL) && (TriedVTMatrix != this))
    {
        The_Matrix *= TriedVTMatrix->Get_Matrix();
        v = TriedVTMatrix->v;
        delete TriedVTMatrix;
        TriedVTMatrix = dynamic_cast<Volume_Tree_Matrix*>(v);
    }
#else
	Volume_Tree_Matrix *TriedVTMatrix = (Volume_Tree_Matrix*)(The_Node);
		int cOK = TriedVTMatrix->castOK;
	if( cOK != CAST_OK) return; // remplace le dynamic cast

    while( (TriedVTMatrix != this))
    {
        The_Matrix *= TriedVTMatrix->Get_Matrix();
        v = TriedVTMatrix->v;
        delete TriedVTMatrix;
        TriedVTMatrix = (Volume_Tree_Matrix*)(v);
		cOK = TriedVTMatrix->castOK;
		if( cOK != CAST_OK) break; // remplace le dynamic cast

    }
#endif

}

//*******************************************************************
//  This function returns the density to the point x, y, z
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
int Volume_Tree_Matrix::EvaluateDensity(const double& x, const double& y, const double& z)const
{
    if(At_The_Top && !isInsideBoundingVolume(x, y, z))
        return 0;

	double new_Pt_x = x;
	double new_Pt_y = y;
	double new_Pt_z = z;

	int to_return = 0;

	if ( !(The_Matrix.IsIdentity()) )
	{
	    Multiply_Mat_Double(The_Matrix_Inverted, x, y, z, 
                            &new_Pt_x, &new_Pt_y, &new_Pt_z);
	}

	if( v!= NULL) {
		to_return = v->EvaluateDensity(new_Pt_x, new_Pt_y, new_Pt_z);

	if ( At_The_Top )
		Bounding_Box_Correction(bbox, &to_return, new_Pt_x, new_Pt_y, new_Pt_z);
	}
	return to_return;
}

//***************************************************
//	
//  Volume_Tree_Matrix::EvaluateGradient()
//
//	Author : Simon Lavallee, SGDL Systems
//	Date   : March 27th, 2002
//
//  Input  : const double &x   -> X coordinate
//           const double &y   -> Y coordinate
//           const double &z   -> Z coordinate
//
//  Output : double &Distance  -> Distance between the point 
//                                and the surface
//           float  *Color     -> Color of the point
//           double *Normal    -> Normal of the point
//           bool              -> If we've been able to get the normal
//
//  Desc   : This function evaluate the gradient of its volume and 
//           modify it, depending if the matrix has rotation.
//***************************************************
bool Volume_Tree_Matrix::EvaluateGradient(const double& x,  const double& y, const double& z, 
                                          double& Distance, double *Normal, float *Color) const
{
    //Get a copy of the points
    double new_Pt_x = x;
	double new_Pt_y = y;
	double new_Pt_z = z;

    //Flag to know if the normal calculation went well
    bool   NormalIsValid = false;

	if ( !(The_Matrix.IsIdentity()) )
	{
        //If the matrix isn't identity, multiply the point by
        //the inverted matrix since the function will be evaluated
        //as if it has no transformation matrix.
		Multiply_Mat_Double(The_Matrix_Inverted, x,	y, z,
							&new_Pt_x, &new_Pt_y, &new_Pt_z);
	}
    
    //Get the normal at this transformed point
    NormalIsValid = v->EvaluateGradient(new_Pt_x, new_Pt_y, new_Pt_z, 
                                        Distance, Normal, Color);

    if (The_Matrix.HasRotation() && NormalIsValid)
    {
        //If the matrix has rotation, we have to rotate the normal vector.

        //Get the rotation
        Point Rotate(0.0);
        The_Matrix.Decompose(Point(0.0), Point(0.0), Rotate, Point(0.0));
        
        //Rotate the normal
        Multiply_Mat_Double(RotationMatrix(Rotate),	Normal[0], Normal[1], Normal[2],
							&new_Pt_x,	&new_Pt_y,   &new_Pt_z);

        Normal[0] = new_Pt_x;
        Normal[1] = new_Pt_y;
        Normal[2] = new_Pt_z;
    }

    if (At_The_Top && NormalIsValid)
    {
        //The bounding box can modify the field value and the gradient
        Bounding_Box_Gradient_Correction(bbox, Normal, &Distance, x, y, z);        
    }

    if (Color && !Undefined_Color(color) && NormalIsValid)
    {
        Color[0] = color[0];
        Color[1] = color[1];
        Color[2] = color[2];
    }

	return NormalIsValid;
}

//***************************************************
//	
//  Volume_Tree_Matrix::GetPrimitives()
//
//	Author : Simon Lavallee, SGDL Systems
//	Date   : September 27th, 2002
//
//  Input  : Array<BoundingQuadric*> &outPrimitives
//           const SL_Matrix &Transform
//           BBox &outGlobalBBox
//
//  Output :
//
//  Desc   : Return the list of transformed primitives (quadrics only)
//           with they're associated primitives.
//***************************************************
void Volume_Tree_Matrix::GetPrimitives(ZArray<BoundingQuadric*> &outPrimitives,
                                       const SL_Matrix &Transform, BBox &outGlobalBBox)
{
    SL_Matrix m = Transform;

    if ( !(The_Matrix.IsIdentity()) )
    {
        m = m*The_Matrix;
    }

    v->GetPrimitives(outPrimitives, m, outGlobalBBox);

    if(bbox != NULL)
    {
        BBox l_Bbox;
        l_Bbox.ma.Set(bbox[0] / 2.0, bbox[1] / 2.0, bbox[2] / 2.0);
        l_Bbox.ma.Set(-bbox[0] / 2.0, -bbox[1] / 2.0, -bbox[2] / 2.0);

        if(!m.IsIdentity())
            l_Bbox *= m;

        outGlobalBBox.Union(l_Bbox);
    }
}

//*******************************************************************
//  This function returns the coeffs of the Volume_Tree_Matrix's VTN
//
//  By Eric Baillargeon
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
double * Volume_Tree_Matrix::GetCoeff(double x, double y , double z)
{
    return v->GetCoeff(x,y,z);
}

//*******************************************************************
//  This function returns the vumber of coeffs of the Volume_Tree_Matrix's VTN 
//
//  By Eric Baillargeon
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
int Volume_Tree_Matrix::GetNumOfCoeff(double x, double y , double z)
{
    return v->GetNumOfCoeff(x,y,z);
}

//*******************************************************************
//  This function set the color of a given point, based on the fact 
//  that is in or outside the volume.
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//  Modified : Simon Lavallee (April 15th 2002)
//*******************************************************************
bool Volume_Tree_Matrix::GetColor(const double &x, const double &y, const double &z, float set_color[])const
{  
    double new_Pt_x = x;
    double new_Pt_y = y;
    double new_Pt_z = z;
    
	if ( !(The_Matrix.IsIdentity()) )
    {		
		//Transform the point if the matrix isn't identity
        Multiply_Mat_Double(The_Matrix_Inverted, x, y, z,
                            &new_Pt_x, &new_Pt_y, &new_Pt_z);
    }    

	bool inside_volume = v->GetColor(new_Pt_x, new_Pt_y, new_Pt_z, set_color);

    if (!Undefined_Color(color) && (inside_volume || At_The_Top))
    {
        //If this node has his own color, keep it.
		set_color[0] = color[0];
		set_color[1] = color[1];
		set_color[2] = color[2];
    }

    return inside_volume;
}

//*******************************************************************
//  This function manages At_The_Top
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void Volume_Tree_Matrix::Manage_At_The_Top()
{
	if( v!= NULL)v->At_The_Top = false;
	At_The_Top = true;
}

//*******************************************************************
//  This function adds a reference to a Volume_Tree_Matrix
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void Volume_Tree_Matrix::Ref()
{
    ref++;

    if (v != 0)   //Ref only if v exist (SL SGDL-sys 30-01-02) TBM
       v->Ref();
}

//*******************************************************************
//  This function removes a reference to a Volume_Tree_Matrix
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void Volume_Tree_Matrix::UnRef()
{
    ref--;

    if (v != 0)   //UnRef only if v exist (SL SGDL-sys 30-01-02) TBM
       v->UnRef();

    if ( ref == 0 )
       delete this;
}


//*******************************************************************
//  This function set the graphic representation to the SolidNode 
//
//  By Alain Mignot
//  Mignot Informatique Graphique Inc.
//  Date : may 2004
//*******************************************************************
	void Volume_Tree_Matrix::SetGraphic( SolidGraphic *Igraphic)
	{
		Graphic = Igraphic;
	}


//*******************************************************************
//  This function Get the graphic representation from the SolidNode 
//
//  By Alain Mignot
//  Mignot Informatique Graphique Inc.
//  Date : may 2004
//*******************************************************************
	SolidGraphic	*Volume_Tree_Matrix::GetGraphic()
	{
		return( Graphic);
	}

//*******************************************************************
//  This function displays the graphic representation of the SolidNode 
//
//  By Alain Mignot
//  Mignot Informatique Graphique Inc.
//  Date : may 2004
//*******************************************************************
	bool Volume_Tree_Matrix::Render(Matrix & rMatrix)
	{
    ::glMatrixMode(GL_MODELVIEW);
	::glPushMatrix();

	GLfloat mat[16];

	::glGetFloatv(GL_MODELVIEW_MATRIX, &mat[0]);

//	::glLoadMatrixf( &rMatrix[0] );
//	::glGetFloatv(GL_MODELVIEW_MATRIX, &mat[0]);

//	::glScalef( 200., 100.,  100. );

//	::glGetFloatv(GL_MODELVIEW_MATRIX, &mat[0]);


	if(!m_displaylist) {

		m_displaylist = ::glGenLists(1);
		
		::glNewList(m_displaylist, GL_COMPILE_AND_EXECUTE);

		::glPushAttrib(GL_POLYGON_BIT);


//		::glEnable(GL_CULL_FACE);
		::glShadeModel(GL_SMOOTH);
		::glEnable(GL_LIGHTING);
		if( Graphic != NULL) Graphic->Display();



		::glPopAttrib();
		::glEndList();


	}
	else
		::glCallList(m_displaylist);

		::glPopMatrix();
	return FALSE;


	}





//EOF****************************************************************