//*******************************************************************
//  Volume_Tree_Unary_Op
//*******************************************************************

#include "../stdafx.h"
#include <assert.h>
#include "Opengl/AFX/Solid/VolumeTreeNode/SolidTools.h"

//*******************************************************************
//  Volume_Tree_Unary_Op constructor 
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
Volume_Tree_Unary_Op::Volume_Tree_Unary_Op(SGDLUnaryOperator the_Op, 
                                           const Volume_Tree_Node *To_Be_V)
{
    
	assert(To_Be_V != NULL);
	the_Operation = the_Op;
	v = const_cast<Volume_Tree_Node *>(To_Be_V);
	Manage_At_The_Top();
		m_displaylist = 0;
		castOK = CAST_OK_unary_op;
}

//*******************************************************************
//  Volume_Tree_Unary_Op constructor 
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
Volume_Tree_Unary_Op::Volume_Tree_Unary_Op(SGDLUnaryOperator the_Op,
										   const Volume_Tree_Node *To_Be_V,
                                           const float bbox_ptr[]):
Volume_Tree_Node(bbox_ptr)
{
	assert(To_Be_V != NULL);

	the_Operation = the_Op;
	v = const_cast<Volume_Tree_Node *>(To_Be_V);
	Manage_At_The_Top();
		m_displaylist = 0;
		castOK = CAST_OK_unary_op;
}

//*******************************************************************
//  This function return true(1) if the point x, y, z is inside and
//  false(0) if the point is outside
//
//  Author : Eric Paquet
//  Date : October 4th,2001
//
//  Modifs :
//*******************************************************************
int Volume_Tree_Unary_Op::EvaluateDensity(const double& x, const double& y, const double& z)const
{
    ASSERT(v != NULL);

    if(At_The_Top && !isInsideBoundingVolume(x, y, z))
        return 0;
    
    int to_return = 0;
    
    switch(the_Operation)
    {
        case SGDLNot :
	    {
            int dens = v->EvaluateDensity(x, y, z);
            if ( dens == 0 )
                to_return = 2;
            else 
                to_return = 0;            
            break;
	    }
        case SGDLSquareRoot :
        {
            to_return = static_cast<int>(sqrt(v->EvaluateDensity(x, y, z)));
            break;
        }
        case SGDLCubeRoot :
        {
            to_return = static_cast<int>(pow(v->EvaluateDensity(x, y, z), 1 / 3.0));
            break;
        }
        case SGDLEven :
        {
            int v_density = v->EvaluateDensity(x, y, z);
            if ( v_density % 2 == 0 )
                to_return = v_density;
            break;
        }
        case SGDLOdd :
        {
            int v_density = v->EvaluateDensity(x, y, z);
            if ( v_density % 2 != 0 )
                to_return = v_density;
            break;
        }

        default:
	    {
	    assert(false); //Do not know this operation
        }
    }
    
    if ( At_The_Top )
	    //This call may modify to_return
	    Bounding_Box_Correction(bbox, &to_return, x, y, z);	
    
    return to_return;
}

//***************************************************
//	
//  Volume_Tree_Unary_Op::EvaluateGradient()
//
//	Author : Simon Lavallee, SGDL Systems
//	Date   : March 27th, 2002
//
//  Input  : double x        -> X coordinate
//           double y        -> Y coordinate
//           double z        -> Z coordinate
//
//  Output : double Distance -> Distance between the point and the surface
//           double *Normal  -> Normal of the point
//           float  Color    -> Vertex color
//           bool            -> true if we've been able to compute the normal    
//
//  Desc   : This function evaluate the normal of its 
//           volume and modify it, depending on the operator.
//***************************************************
bool Volume_Tree_Unary_Op::EvaluateGradient(const double& x, const double& y, const double& z, 
                                            double& Distance, double *Normal, float *Color) const
{   
    bool NormalIsValid = v->EvaluateGradient(x, y, z, Distance, Normal, Color);

    if (NormalIsValid)
    {
        switch(the_Operation)
        {
            case SGDLNot :
            {
                //Reverse the gradient and the distance
                Normal[0] *= -1;
                Normal[1] *= -1;
                Normal[2] *= -1;

                Distance  *= -1;
                break;
            }

            case SGDLSquareRoot :
            case SGDLCubeRoot :            
            case SGDLEven :
            case SGDLOdd :
            {
                //These operators aren't supposed to modify the gradient
                break;
            }

            default:
            {
                assert(false); //Do not know this operation
            }
        }
     }
    
    if (At_The_Top && NormalIsValid)
    {
        //The gradient can be modified by the bounding box
        Bounding_Box_Gradient_Correction(bbox, Normal, &Distance, x, y, z);
    }

    if (Color && !Undefined_Color(color) && NormalIsValid)
    {
        //If this node has its own color, keep it.
        Color[0] = color[0];
        Color[1] = color[1];
        Color[2] = color[2];
    }

    return NormalIsValid;
}

//***************************************************
//	
//  Volume_Tree_Unary_Op::GetFieldValue()
//
//	Author : Simon Lavallee, SGDL Systems
//	Date   : September 27th, 2002
//
//  Input  : double x           -> X coordinate
//           double y           -> Y coordinate
//           double z           -> Z coordinate
//
//  Output : bool               -> If we've been able to get the field Value.
//           double *fValue     -> The field value of the Implicit function
//
//  Desc   : Return the field value of the implicit function
//           at a given point.
//***************************************************
void Volume_Tree_Unary_Op::GetPrimitives(ZArray<BoundingQuadric*> &outPrimitives,
                                         const SL_Matrix &Transform, BBox &outGlobalBBox)
{
    v->GetPrimitives(outPrimitives, Transform, outGlobalBBox);

    if(bbox != NULL)
    {
        BBox l_Bbox;
        l_Bbox.ma.Set(bbox[0] / 2.0, bbox[1] / 2.0, bbox[2] / 2.0);
        l_Bbox.ma.Set(-bbox[0] / 2.0, -bbox[1] / 2.0, -bbox[2] / 2.0);

        if(!Transform.IsIdentity())
            l_Bbox *= Transform;

        outGlobalBBox.Union(l_Bbox);
    }
}

/********************************************************************\
Function:  

  Input:     
  Output:     
  
    Author:	Eric Baillargeon SGDL Systme Inc.
    Created:	2001/10/26	
    
      
*********************************************************************/
double * Volume_Tree_Unary_Op::GetCoeff(double x, double y , double z)
{
    return (v->GetCoeff(x,y,z));    
}

/********************************************************************\
Function:  

  Input:     
  Output:     
  
    Author:	Eric Baillargeon SGDL Systme Inc.
    Created:	2001/10/26	
    
      
*********************************************************************/
int Volume_Tree_Unary_Op::GetNumOfCoeff(double x, double y , double z)
{
    return (v->GetNumOfCoeff(x,y,z));
}

//*******************************************************************
//  This function manages the operator = between two Volume_Tree_Unary_Op
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
const Volume_Tree_Unary_Op& Volume_Tree_Unary_Op::operator=(const Volume_Tree_Unary_Op& The_Other)
{
	the_Operation = The_Other.the_Operation;
	v = The_Other.v;
	return *this;
}

//*******************************************************************
//  This function return the color of a point based on the fact that is
//  in or ouside the volume.
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//  Modif: Simon Lavallee (04-16-2002) (Now just get the color)
//*******************************************************************
bool Volume_Tree_Unary_Op::GetColor(const double &x, const double &y, const double &z, float set_color[])const
{
	bool inside_volume = v->GetColor(x, y, z, set_color);
	
	if (!Undefined_Color(color) && (inside_volume || At_The_Top))
	{
		set_color[0] = color[0];
		set_color[1] = color[1];
		set_color[2] = color[2];
	}

    return inside_volume;
}

//*******************************************************************
//  This function manages At_The_Top
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void Volume_Tree_Unary_Op::Manage_At_The_Top()
{
	v->At_The_Top = false;
	At_The_Top = true;
}

//*******************************************************************
//  This adds a reference to a Volume_Tree_Unary_Op
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void Volume_Tree_Unary_Op::Ref()
{
    ref++;
    v->Ref();
}

//*******************************************************************
//  This removes a reference to a Volume_Tree_Unary_Op
//
//  By Eric Paquet
//  SGDL Systems inc
//  Date : December 10th, 2001.
//*******************************************************************
void Volume_Tree_Unary_Op::UnRef()
{
    ref--;
    v->UnRef();

    if ( ref == 0 )
        delete this;
}

//*******************************************************************
//  This function set the graphic representation to the SolidNode 
//
//  By Alain Mignot
//  Mignot Informatique Graphique Inc.
//  Date : may 2004
//*******************************************************************
	void Volume_Tree_Unary_Op::SetGraphic( SolidGraphic *Igraphic)
	{
		Graphic = Igraphic;
	}


//*******************************************************************
//  This function Get the graphic representation from the SolidNode 
//
//  By Alain Mignot
//  Mignot Informatique Graphique Inc.
//  Date : may 2004
//*******************************************************************
	SolidGraphic	*Volume_Tree_Unary_Op::GetGraphic()
	{
		return( Graphic);
	}

//*******************************************************************
//  This function displays the graphic representation of the SolidNode 
//
//  By Alain Mignot
//  Mignot Informatique Graphique Inc.
//  Date : may 2004
//*******************************************************************
	bool Volume_Tree_Unary_Op::Render( Matrix &rMatrix)
	{
		::glPushMatrix();
	if(!m_displaylist) {

		m_displaylist = ::glGenLists(1);
		
		::glNewList(m_displaylist, GL_COMPILE_AND_EXECUTE);
		::glPushAttrib(GL_POLYGON_BIT);


//		::glEnable(GL_CULL_FACE);
		::glShadeModel(GL_SMOOTH);
		::glEnable(GL_LIGHTING);
		if( Graphic != NULL)Graphic->Display();



		::glPopAttrib();
		::glEndList();


	}
	else
		::glCallList(m_displaylist);

	::glPopMatrix();
	return TRUE;

	}
//EOF***************************************************************************