/*=============================================================================


==============================================================================*/
#ifndef _SL_Matrix_h_
#define _SL_Matrix_h_

#include "Point.h"


//! a 4*4 float transformation matrix
//! many operations are designed to not use itermediate or stack return values for speed
class SL_Matrix {

public:
	enum Contents {
		HAS_TRANSLATION = 1,		//!< matrix has translation part

		HAS_SCALE		= 2,		//!< a scale part 
		
		HAS_NEGATIVE_SCALE = 4,		//!< negative scale 
		HAS_ROTATION	= 8,		//!< any rotation
		
		HAS_PERSPECTIVE	= 16,		//!< perspective part
		
		HAS_NON_UNIFORM_SCALE = 32,	//!< non uniform scale 
	

		HAS_ANY_SCALE = 2+4+32,
		HAS_ANY_TRANSFORM	= 1+2+4+8+32,
		HAS_ANYTHING	= 1+2+4+8+16+32,

	};	

		
	union {
		//! the 4*4 matrix
		float x[4][4];



	};
		unsigned char contents;

	bool HasRotation() const { return (contents & HAS_ROTATION) !=0; }

	// identity : contents = 0
	bool IsIdentity() const { return contents == 0; }

    // Fundamental operations
    void Invert();					//!< Invert the matrix
    void Transpose();				//!< Transpose the matrix
    
   // Assignment operators
    SL_Matrix& operator+= (const SL_Matrix& A);	//!< add-to
    SL_Matrix& operator-= (const SL_Matrix& A);	//!< subtract-from
    SL_Matrix& operator*= (const SL_Matrix& A);	//!< multiply by matrix

	SL_Matrix& MultiplyLeft(const SL_Matrix& A);	//!< multiply by SL_Matrix, A is left, this is right
    SL_Matrix& MultiplyRight(const SL_Matrix& A);	//!< multiply by SL_Matrix, A is right, this is left

	SL_Matrix& operator*= (float A);		//!< scale by scalar

	
	SL_Matrix& Scale(float X, float Y, float Z);
	SL_Matrix& Scale(const Point& scale);
	
	SL_Matrix& Translate(const Point& Location);
	SL_Matrix& Translate(float X, float Y, float Z);
	
	SL_Matrix& Rotate(const Point& Rotation);
	SL_Matrix& Rotate(float X, float Y, float Z);

	//! Acess row
	float*  operator[] (int row) { return(x[row]);}
	//! Acess row
	const float*  operator[] (int row) const { return(x[row]);}

   // Binary operators
    friend SL_Matrix operator+ (const SL_Matrix& A, const SL_Matrix& B);
    friend SL_Matrix operator- (const SL_Matrix& A, const SL_Matrix& B);
    friend SL_Matrix operator* (const SL_Matrix& A, float B);
    friend SL_Matrix operator* (const SL_Matrix& A, const SL_Matrix& B);
	//added by Chloe Mignot 2001/10/26 SGDL Systemes Inc
	friend bool operator== (const SL_Matrix& A, const SL_Matrix& B);


    friend SL_Matrix Invert(const SL_Matrix& M);		//!< Invert a given SL_Matrix
    friend SL_Matrix Transpose(const SL_Matrix& M);	//!< Transpose a given SL_Matrix

    // Create various types of SL_Matrix.
    friend SL_Matrix IdentityMatrix();

		//! set to identiy matrix 
	SL_Matrix& Identity();

	friend SL_Matrix TranslationMatrix(const Point& Location);
	friend SL_Matrix TranslationMatrix(float X, float Y, float Z);

    friend SL_Matrix ScaleMatrix(float X, float Y, float Z);
    friend SL_Matrix ScaleMatrix(const Point & Scale);

	friend SL_Matrix RotationMatrix(const Point& Rotation);
	friend SL_Matrix RotationMatrix (float X, float Y, float Z);

    friend SL_Matrix RotationXMatrix(float Angle);
    friend SL_Matrix RotationYMatrix(float Angle);
    friend SL_Matrix RotationZMatrix(float Angle);

	    friend SL_Matrix QuadricMatrix(float a, float b, float c, float d, float e,
				float f, float g, float h, float j, float k);

   int Decompose(Point &scale,Point &shear, Point &rotate, Point &translate) const;

};



#endif

