/***********************************************************************
This software module was originally developed by Dr. Michael Steliaros,
Superscape (UK) Ltd., in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Superscape (UK) Ltd. retains full
right to use the code for his/her own purpose, assign or donate the code
to a third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2001
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Dr. Michael Steliaros at msteliaros@superscape.com
----------------------------------------------------------------------*/

////---------------------------------------------------------------------
////    Basic dynamic storage class
////
////    Author:                 Sean T. Ellis (Superscape)
////    Creation date:          07-10-1996
////---------------------------------------------------------------------

#ifndef DYNARRAY_H
#define DYNARRAY_H

///----------------------------------------------------------------------
///
/// DynArray
/// Template class implementing a dynamically growable array, where all
/// the alements of the array must be contiguous with each other. This
/// allows a very fast lookup, but makes extension of large arrays more
/// time consuming. It should be used where lookups are considerably more
/// numerous than assignments. The "cActualElements" method is virtual and 
/// can be overridden to allow a different memory allocation strategy to be 
/// used.
///: prefix  : a
///: DynArray  : Default constructor. Array size is not specified.
///: ~DynArray  : Destructor. Destroys array and contents.
///: cElements    : Returns maximum index of accessed elements.
///: m_ppBlocks    : Pointer to array of pointers to storage blocks.
///: m_cBlocks    : Number of storage blocks allocated.
///: m_cElements  : Maximum index of accessed elements.
///
///  NOTE:  Reserve changes the amount of allocated storage. It reserves space
///      for exactly the given number of array elements. The reserved space
///      must be greater or equal to the number of elements currently in use.
///      The number of elements in use and their contents are unaffected.
///
///      SetcElements will set the number of active elements, but will
///      not change the number of reserved elements unless the reserved space
///      is insufficient to hold the new number of active elements.
///
///      Clear will change the number of elements in the array using
///      SetcElements, and clear any elements which are new or discarded. For
///      array reuse, using Clear() is to be preferred over using Truncate(0),
///      which will deallocate the array completely.

#include <stdlib.h>

template<class T> class DynArray
{
  public:
    DynArray() : m_pStorage(NULL),
                 m_cElements(0),
                 m_cReserved(0) { }
    DynArray(const DynArray<T>& aFrom) : m_pStorage(NULL),
                                         m_cElements(0),
                                         m_cReserved(0) { operator = (aFrom); }
    ~DynArray() { delete [] (m_pStorage); }

    DynArray<T> &operator=(const DynArray<T>& aFrom)
    {
      this->Truncate(aFrom.cElements());

      CopyArray(aFrom.m_pStorage, m_pStorage, aFrom.cElements());
      
      return(*this);
    }

    T& operator[](int iIndex) { return(m_pStorage[iIndex]); }

    const T&  operator[](int iIndex) const
    {
      return(const_cast<const T *>(m_pStorage)[iIndex]);
    }

    int cElements() const { return(m_cElements); }

    void Truncate(int cElements)
    {
      if(cElements != m_cElements)
      {
        int cOldStorage = cActualElements(m_cElements, m_cReserved);
        int cNewStorage = cActualElements(cElements, 0);

        if(cNewStorage != cOldStorage)
          SetStorageSize(cNewStorage,
                         (cElements < m_cElements) ? cElements : m_cElements);

        m_cElements = cElements;
      }
    }

    void Reserve(int cElements)
    {
      if(cElements != m_cElements)
        SetStorageSize(cElements, m_cElements);
    }

    void SetcElements(int cElements)
    {
      if(cElements != m_cElements)
      {
        int cOldStorage = cActualElements(m_cElements, m_cReserved);
        int cNewStorage = cActualElements(cElements, m_cReserved);

        if(cNewStorage != cOldStorage)
          SetStorageSize(cNewStorage,
                         (cElements < m_cElements) ? cElements : m_cElements);

        m_cElements = cElements;
      }
    }

    void Clear(int cElements = 0, const T& Value = T())
    {
      if(cElements != m_cElements)
      {
        int cOldElements = m_cElements;
        int cOldReserved = m_cReserved;
        
        SetcElements(cElements);

        if(cOldReserved == m_cReserved)
        {
          int i = (cElements < cOldElements) ? cElements : cOldElements;
          int iEnd = (cElements < cOldElements) ? cOldElements : cElements;

          while(i<iEnd)
            m_pStorage[i++] = Value;
        }              
      }
    }

    void Append(const T& elem)
    {
      if(m_cElements == m_cReserved)
        SetStorageSize(cActualElements(m_cElements + 1, m_cReserved),
                       m_cElements);

      (*this)[m_cElements++] = elem;
    }

    void Insert(int iIndex, const T& elem)
    {
      if(iIndex == m_cElements) { Append(elem); return; }
      
      if(m_cElements == m_cReserved)
        SetStorageSize(cActualElements(m_cElements + 1, m_cReserved),
                       m_cElements);

      CopyArray(&m_pStorage[iIndex], &m_pStorage[iIndex + 1],
                m_cElements - iIndex);

      m_cElements++;

      (*this)[iIndex] = elem;
    }

    void Remove(int iIndex)
    {
      m_pStorage[iIndex] = m_pStorage[m_cElements - 1];
      Truncate(m_cElements - 1);
    }

    void RemoveRetainOrder(int iIndex)
    {
      if(iIndex < m_cElements - 1)
        CopyArray(&m_pStorage[iIndex + 1], &m_pStorage[iIndex],
                  m_cElements - iIndex - 1);

      Truncate(m_cElements - 1);
    }

    const T* Pointer() const { return(m_pStorage); }

  protected:
    static int cActualElements(int cElements, int cReserved)  
    {
      if(cElements <= cReserved)
        return(cReserved);
      else{
        switch(cElements){
          case 0 :
          case 1 :
          case 2 : return(cElements);
          case 3 :
          case 4 : return(4);
          case 5 :
          case 6 :
          case 7 :
          case 8 : return(8);
          case 9 :
          case 10 :
          case 11 :
          case 12 :
          case 13 :
          case 14 :
          case 15 :
          case 16 : return(16);
          default : {
            int pwr2 = 32;

            while(cElements > pwr2)
              pwr2 += pwr2;
            
            return(pwr2);
          }
        }
      }
    }

    void SetStorageSize(int cNewStorage, int cElementsToCopy)
    {
      T *pStorage;

      if(cNewStorage != 0){
        pStorage = new T[cNewStorage];

        CopyArray(m_pStorage, pStorage, cElementsToCopy);
      }
      else
        pStorage=NULL;

      delete[] m_pStorage;

      m_pStorage = pStorage;
      m_cReserved = cNewStorage;
    }

    static void CopyArray(const T* paSource, T* paDestination, int cElements)
    {
      if (paSource < paDestination){
        paSource += cElements;
        paDestination += cElements;

        while(cElements-- > 0)
          *(--paDestination) = *(--paSource);
      }
      else if (paSource > paDestination){
        while(cElements-- > 0)
          *(paDestination++) = *(paSource++);
      }
    }

    T* m_pStorage;
    int m_cElements;
    int m_cReserved;
};

#endif // DYNARRAY_H
