/***********************************************************************
This software module was originally developed by Dr. Michael Steliaros,
Superscape (UK) Ltd., in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Superscape (UK) Ltd. retains full
right to use the code for his/her own purpose, assign or donate the code
to a third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2001
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Dr. Michael Steliaros at msteliaros@superscape.com
----------------------------------------------------------------------*/
////---------------------------------------------------------------------
////    Basic manifold hull storage classes
////
////    Author:                 Michael K. Steliaros (Superscape)
////    Last modification date: 28-11-2000
////    Creation date:          22-09-2000
////---------------------------------------------------------------------

#ifndef HULL_H
#define HULL_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "vector3d.h"
#include "dynarray.h"

//       iPred[0] \   /iSucc[1]
//                 \ /
//                  * iVertex[0] (tail)
//                  |
// iFacet[0] (left) |
//                  * iMidpoint (at next level)
//                  | iFacet[1] (right)
//                  |
//                  * iVertex[1] (head)
//                 / \
//       iSucc[0] /   \ iPred[1]

class Edge                                                // Winged edge class
{
  public:
    Edge() : m_iMidpoint(-1),
             m_Sharpness(0),
             m_bVisible(1)
    {
      m_iVertex[0] = -1;
      m_iVertex[1] = -1;
      m_iFacet[0] = -1;
      m_iFacet[1] = -1;
      m_iPred[0] = -1;
      m_iPred[1] = -1;
      m_iSucc[0] = -1;
      m_iSucc[1] = -1;
    }
    ~Edge() { }

    void SetiVertex(const int iVertex, const int bRight)
      { m_iVertex[bRight ? 1 : 0] = iVertex; }
    int iVertex(const int bRight)
      const { return(m_iVertex[bRight ? 1 : 0]); }
    int iOtherVertex(const int iVertex) const
    {
      assert((m_iVertex[0] == iVertex) || (m_iVertex[1] == iVertex));

      return(m_iVertex[(iVertex == m_iVertex[0]) ? 1 : 0]);
    }

    void SetiPred(const int iEdge, const int bRight)
      { m_iPred[bRight ? 1 : 0] = iEdge; }
    void SetiSucc(const int iEdge, const int bRight)
      { m_iSucc[bRight ? 1 : 0] = iEdge; }

    int iPred(const int bRight)
      const { return(m_iPred[bRight ? 1 : 0]); }
    int iSucc(const int bRight)
      const { return(m_iSucc[bRight ? 1 : 0]); }

    void MakeRing(const int iEdge)
      { m_iPred[0] = m_iPred[1] = m_iSucc[0] = m_iSucc[1] = iEdge; }

    void ReplaceReferences(const int iOldEdge, const int iNewEdge)
    {
      if(m_iPred[0] == iOldEdge) m_iPred[0] = iNewEdge;
      if(m_iPred[1] == iOldEdge) m_iPred[1] = iNewEdge;
      if(m_iSucc[0] == iOldEdge) m_iSucc[0] = iNewEdge;
      if(m_iSucc[1] == iOldEdge) m_iSucc[1] = iNewEdge;
    }

    void SetiFacet(const int iFacet, const int bRight)
      { m_iFacet[bRight ? 1 : 0] = iFacet; }
    int iFacet(const int bRight)
      const { return(m_iFacet[bRight ? 1 : 0]); }
    int iOtherFacet(const int iFacet) const
    {
      assert((m_iFacet[0] == iFacet) || (m_iFacet[1] == iFacet));

      return(m_iFacet[(iFacet == m_iFacet[0]) ? 1 : 0]);
    }

    void SetiMidpoint(const int iVertex)
      { m_iMidpoint = iVertex; }
    int iMidpoint()
      const { return(m_iMidpoint); }

    void IncSharpness()
      { m_Sharpness++; }
    void DecSharpness()
      { m_Sharpness--; }
    void SetSharpness(const int Sharpness)
      { m_Sharpness = Sharpness; }
    int Sharpness()
      const { return(m_Sharpness); }

    void SetbVisible(const int bVisible)
      { m_bVisible = bVisible; }
    void SetVisible()
      { SetbVisible(1); }
    void SetInvisible()
      { SetbVisible(0); }
    int bVisible()
      const { return(m_bVisible); }
    int bBoundary()
      const { return((m_iFacet[0] == -1) || (m_iFacet[1] == -1)); }
    int bSharp()
      const { return(bBoundary() || (m_Sharpness - 1 > 0)); }

    void Debug(FILE *fd) const
    {
      fprintf(fd, "%d %d\n"
                  "%7d \\   / %d\n"
                  "         \\ /\n"
                  "%9d *\n"
                  "%9d |\n"
                  "          * %d\n"
                  "          | %d\n"
                  "          * %d\n"
                  "         / \\\n"
                  "%7d /   \\ %d\n",
              m_Sharpness, m_bVisible,
              m_iPred[0], m_iSucc[1],
              m_iVertex[0],
              m_iFacet[0],
              m_iMidpoint,
              m_iFacet[1],
              m_iVertex[1],
              m_iSucc[0], m_iPred[1]);
    }

  private:
    int m_iVertex[2];                    // Edge vertices
    int m_iPred[2], m_iSucc[2];          // Wing links
    int m_iFacet[2];                     // Edge facets

    int m_iMidpoint;                     // Midpoint vertex (for next level)

    int m_Sharpness;                     // 0 - smooth; >0 -  sharpness
    int m_bVisible;                      // 1 - visible; 0 - invisible
};

class Vertex                                                    // Vertex class
{
  public:
    Vertex() : m_bDirty(0),
               m_Sharpness(0)
      { m_aEdges.Truncate(0); }
    Vertex(const Vector3D& Vector) : m_bDirty(0), m_Sharpness(0)
      { m_Vector = Vector; m_aEdges.Truncate(0); }
    ~Vertex()
      { m_aEdges.Truncate(0); }

    void SetVector(const Vector3D& Vector)
      { m_Vector = Vector; }
    Vector3D& Vector()
      { return(m_Vector); }
    const Vector3D& Vector()
      const { return(m_Vector); }

    void SetDirty()
      { m_bDirty = 1; }
    void SetClean()
      { m_bDirty = 0; }
    int bDirty()
      const { return(m_bDirty); }

    void IncSharpness()
      { m_Sharpness++; }
    void DecSharpness()
      { m_Sharpness--; }
    void SetSharpness(const int Sharpness)
      { m_Sharpness = Sharpness; }
    int Sharpness()
      const { return(m_Sharpness); }
    int bSharp()
      const { return(m_Sharpness - 1 > 0); }

    int cEdges()
      const { return(m_aEdges.cElements()); }
    int& aEdges(int iPos)
      { assert(iPos < m_aEdges.cElements()); return(m_aEdges[iPos]); }
    const int& aEdges(int iPos)
      const { assert(iPos < m_aEdges.cElements()); return(m_aEdges[iPos]); }

    int iEdgePos(const int iEdge) const
    {
      int Count;

      for(Count = 0; Count < m_aEdges.cElements(); Count++)
        if(m_aEdges[Count] == iEdge)
          return(Count);

      return(-1);
    }

    int AddEdge(const int iEdge)
    {
      int where = m_aEdges.cElements();

      m_aEdges.Append(iEdge);

      return(where);
    }
    void RemoveEdge(const int iPos)
    {
      m_aEdges.Remove(iPos);
    }
    void RemoveRetainOrderEdge(const int iPos)
    {
      m_aEdges.RemoveRetainOrder(iPos);
    }
    void TruncateaEdges(const int cElements)
    {
      m_aEdges.Truncate(cElements);
    }

    void Debug(FILE* fd) const
    {
      int Count;

      fprintf(fd, "[%.3f, %.3f, %.3f] %d [",
              m_Vector.X(),
              m_Vector.Y(),
              m_Vector.Z(),
              m_aEdges.cElements());

      for(Count = 0; Count < m_aEdges.cElements(); Count++)
        fprintf(fd, " %d", m_aEdges[Count]);

      fprintf(fd, "]\n");
    }

  private:
    Vector3D m_Vector;                // Vector

    DynArray<int> m_aEdges;           // Edges (valence)

    int m_bDirty;                     // 1 - dirty; 0 - clean

    int m_Sharpness;                  // 0 - smooth; >0 -  sharpness
};

class Facet                                                      // Facet class
{
  public:
    Facet() : m_iNormal(-1),
              m_bDirty(0)
    {
      m_aVertices.Truncate(0);
      m_aVertexNormals.Truncate(0);
      m_aVertexUVWs.Truncate(0);
      m_aVertexColours.Truncate(0);
    }
    Facet(const int cVertices,
          const int* aVertices = NULL,
          const int iNormal = -1) : m_iNormal(iNormal),
                                    m_bDirty(0)
    {
      int Count;

      assert(cVertices >= 3);

      m_aVertices.Truncate(cVertices);
      m_aVertexNormals.Truncate(cVertices);
      m_aVertexUVWs.Truncate(cVertices);
      m_aVertexColours.Truncate(cVertices);

      for(Count = 0; Count < cVertices; Count++){
        m_aVertices[Count] = (aVertices != NULL) ? aVertices[Count] : -1;
        m_aVertexNormals[Count] = iNormal;
        m_aVertexUVWs[Count] = -1;
        m_aVertexColours[Count] = -1;
      }
    }
    ~Facet()
    {
      m_aVertices.Truncate(0);
      m_aVertexNormals.Truncate(0);
      m_aVertexUVWs.Truncate(0);
      m_aVertexColours.Truncate(0);
    }

    int cVertices()
      const { return(m_aVertices.cElements()); }
    int& aVertices(const int iPos)
      { return(m_aVertices[iPos]); };
    int& aVertexNormals(const int iPos)
      { return(m_aVertexNormals[iPos]); };
    int& aVertexUVWs(const int iPos)
      { return(m_aVertexUVWs[iPos]); };
    int& aVertexColours(const int iPos)
      { return(m_aVertexColours[iPos]); };
    const int& aVertices(const int iPos)
      const { return(m_aVertices[iPos]); };
    const int& aVertexNormals(const int iPos)
      const { return(m_aVertexNormals[iPos]); };
    const int& aVertexUVWs(const int iPos)
      const { return(m_aVertexUVWs[iPos]); };
    const int& aVertexColours(const int iPos)
      const { return(m_aVertexColours[iPos]); };

    int iVertexPos(const int iVertex) const
    {
      int Count;

      for(Count = 0; Count < m_aVertices.cElements(); Count++)
        if(m_aVertices[Count] == iVertex)
          return(Count);

      return(-1);
    }

    int AddVertex(const int iVertex,
                  const int iNormal,
                  const int iUVW,
                  const int iColour)
    {
      int iPos = m_aVertices.cElements();

      m_aVertices.Append(iVertex);
      m_aVertexNormals.Append(iNormal);
      m_aVertexUVWs.Append(iUVW);
      m_aVertexColours.Append(iColour);

      return(iPos);
    }
    void RemoveVertex(const int iPos)
    {
      m_aVertices.Remove(iPos);
      m_aVertexNormals.Remove(iPos);
      m_aVertexUVWs.Remove(iPos);
      m_aVertexColours.Remove(iPos);
    }
    void RemoveRetainOrderVertex(const int iPos)
    {
      m_aVertices.RemoveRetainOrder(iPos);
      m_aVertexNormals.RemoveRetainOrder(iPos);
      m_aVertexUVWs.RemoveRetainOrder(iPos);
      m_aVertexColours.RemoveRetainOrder(iPos);
    }
    void TruncateaVertices(const int cElements)
    {
      m_aVertices.Truncate(cElements);
      m_aVertexNormals.Truncate(cElements);
      m_aVertexUVWs.Truncate(cElements);
      m_aVertexColours.Truncate(cElements);
    }

    void SetiNormal(const int iNormal) { m_iNormal = iNormal; };
    int iNormal() const { return(m_iNormal); };

    void SetDirty() { m_bDirty = 1; }
    void SetClean() { m_bDirty = 0; }
    int bDirty() const { return(m_bDirty); }

    void Debug(FILE* fd) const
    {
      int Count;

      fprintf(fd, "%d [", m_aVertices.cElements());

      for(Count = 0; Count < m_aVertices.cElements(); Count++)
        fprintf(fd, " %d", m_aVertices[Count]);

      fprintf(fd, "] [");

      for(Count = 0; Count < m_aVertexNormals.cElements(); Count++)
        fprintf(fd, " %d", m_aVertexNormals[Count]);

      fprintf(fd, "] [");

      for(Count = 0; Count < m_aVertexUVWs.cElements(); Count++)
        fprintf(fd, " %d", m_aVertexUVWs[Count]);

      fprintf(fd, "] [");

      for(Count = 0; Count < m_aVertexColours.cElements(); Count++)
        fprintf(fd, " %d", m_aVertexColours[Count]);

      fprintf(fd, "] %d\n", m_iNormal);
    }

  private:
    DynArray<int> m_aVertices;               // Vertices
    DynArray<int> m_aVertexNormals;          // Vertex normals
    DynArray<int> m_aVertexUVWs;             // Vertex texture coords
    DynArray<int> m_aVertexColours;          // Vertex colours

    int m_iNormal;                           // Facet normal

    int m_bDirty;                            // 1 - dirty; 0 - clean
};

class Hull                                                        // Hull class
{
  public:
    Hull() : m_strName(NULL),
             m_cEdgeHashes(0),
             m_cFreeEdgeHashes(0),
             m_aEdgeHashes(NULL)
    {
      m_aVertices.Truncate(0);
      m_aNormals.Truncate(0);
      m_aUVWs.Truncate(0);
      m_aColours.Truncate(0);
      m_aFacets.Truncate(0);
      m_aEdges.Truncate(0);
    }
    ~Hull()
    {
      if(m_strName != NULL) delete [] m_strName;

      m_aVertices.Truncate(0);
      m_aNormals.Truncate(0);
      m_aUVWs.Truncate(0);
      m_aColours.Truncate(0);
      m_aFacets.Truncate(0);
      m_aEdges.Truncate(0);

      if(m_aEdgeHashes != NULL) delete [] m_aEdgeHashes;
    }

    const Hull& operator=(const Hull& From);

    char* strName() const { return(m_strName); }

    void SetstrName(char strName[]);

    int AddVertex(const Vector3D& aVector);
    int AddVertex(const float X, const float Y, const float Z)
      { return(AddVertex(Vector3D(X, Y, Z))); }
    int AddVertex(const float aVector[3])
      { return(AddVertex(aVector[0], aVector[1], aVector[2])); }

    int VertexiPrimaryEdge(const int iVertex) const
    {
      const Vertex& V = m_aVertices[iVertex];
      int Count;

      for(Count = 0; Count < V.cEdges(); Count++){
        const int iPos = V.aEdges(Count);
        const Edge& E = m_aEdges[iPos];

        if((E.iFacet(0) == -1) && (E.iVertex(1) == iVertex))
          return(iPos);
      }

      return(V.aEdges(0));
    }

    Vector3D* VertexNormal(const int iVertex);
    Vector3D* VertexNormal(const int iVertex,
                           const int iStart,
                           const int cSpan);

    void SetVertexNormal(const int iVertex, const int iNormal);
    void SetVertexNormal(const int iVertex,
                         const int iStart,
                         const int cSpan,
                         const int iNormal);

    void MarkVertexDirty(const int iVertex);

    int AddNormal(const Vector3D& aNormal);
    int AddNormal(const float X, const float Y, const float Z)
      { return(AddNormal(Vector3D(X, Y, Z))); }
    int AddNormal(const float aNormal[3])
      { return(AddNormal(aNormal[0], aNormal[1], aNormal[2])); }

    int AddUVW(const Vector3D& aUVW);
    int AddUVW(const float U, const float V, const float W = 0.0)
      { return(AddUVW(Vector3D(U, V, W))); }
    int AddUVW(const float aUVW[3])
      { return(AddUVW(aUVW[0], aUVW[1], aUVW[2])); }

    int AddColour(const Vector3D& aColour);
    int AddColour(const float R, const float G, const float B)
      { return(AddColour(Vector3D(R, G, B))); }
    int AddColour(const float aColour[3])
      { return(AddColour(aColour[0], aColour[1], aColour[2])); }

    int AddEdge(const int iHead, const int iTail, const int iFacet);
    int AddEdge(const Edge& e)
    {
      int where = m_aEdges.cElements();

      m_aEdges.Append(e);

      return(where);
    }

    int iEdge(int iHead, int iTail) const;

    void SetiEdge(const int iHead, const int iTail, const int iEdge);

    Vertex& EdgeOtherVertex(const int iEdge, const int iVertex)
      { return(m_aVertices[m_aEdges[iEdge].iOtherVertex(iVertex)]); }
    const Vertex& EdgeOtherVertex(const int iEdge, const int iVertex)
      const { return(m_aVertices[m_aEdges[iEdge].iOtherVertex(iVertex)]); }

    Facet& EdgeOtherFacet(const int iEdge, const int iFacet)
      { return(m_aFacets[m_aEdges[iEdge].iOtherFacet(iFacet)]); }
    const Facet& EdgeOtherFacet(const int iEdge, const int iFacet)
      const { return(m_aFacets[m_aEdges[iEdge].iOtherFacet(iFacet)]); }

    void EdgeRingLink(const int iA, const int iB, const int iFacet)
    {
      int bARight = (m_aEdges[iA].iFacet(1) == iFacet) ? 1 : 0;
      int bBRight = (m_aEdges[iB].iFacet(1) == iFacet) ? 1 : 0;
      int iAPred = m_aEdges[iA].iPred(bARight);
      int iBPred = m_aEdges[iB].iPred(bBRight);
      int bAPredRight = (m_aEdges[iAPred].iFacet(1) == iFacet) ? 1 : 0;
      int bBPredRight = (m_aEdges[iBPred].iFacet(1) == iFacet) ? 1 : 0;

      m_aEdges[iAPred].SetiSucc(iB, bAPredRight);
      m_aEdges[iB].SetiPred(iAPred, bBRight);
      m_aEdges[iBPred].SetiSucc(iA, bBPredRight);
      m_aEdges[iA].SetiPred(iBPred, bARight);
    }

    int AddFacet(const Facet& F)
    {
      int iPos = m_aFacets.cElements();

      m_aFacets.Append(F);

      if(m_aEdgeHashes == NULL) InitHash(4 * cVertices());             // Guess

      FacetLinkEdges(iPos);

      FacetNormalRethink(iPos);

      return(iPos);
    }
    int AddFacet(const int cVertices,
                 const int aVertices[],
                 int iNormal = -1)
      { return(AddFacet(Facet(cVertices, aVertices, iNormal))); }

    int FacetiPrimaryEdge(const int iFacet) const
    {
      const Facet& F = m_aFacets[iFacet];
      const int iPrimaryEdge = iEdge(F.aVertices(0), F.aVertices(1));

      assert(iPrimaryEdge != -1);

      return(iPrimaryEdge);
    }

    int FacetSplit(const int iFacet, int iA, int iB);

    void FacetLinkEdges(const int iFacet, const int bInvisible = 0)
    {
      Facet& F = m_aFacets[iFacet];
      int iCurr = F.aVertices(0);
      int iNext = F.aVertices(1);

      const int iPrimaryEdge = AddEdge(iCurr, iNext, iFacet);

      assert(iPrimaryEdge != -1);

      if(bInvisible) m_aEdges[iPrimaryEdge].SetInvisible();

      for(int Count = 1; Count < F.cVertices(); Count++){

        iCurr = iNext;
        iNext = F.aVertices((Count + 1) % F.cVertices());

        const int iEdge = AddEdge(iCurr, iNext, iFacet);

        assert(iEdge != -1);

        EdgeRingLink(iPrimaryEdge, iEdge, iFacet);

        if(bInvisible) m_aEdges[iEdge].SetInvisible();
      }
    }

    void FacetUnlinkEdges(const int iFacet)
    {
      const int iPrimaryEdge = FacetiPrimaryEdge(iFacet);
      int iCurr = iPrimaryEdge;

      do{
        Edge& E = m_aEdges[iCurr];
        int bRight = (E.iFacet(1) == iFacet) ? 1 : 0;
        int iNext = E.iSucc(bRight);

        E.SetiFacet(-1, bRight);
        E.SetiPred(iCurr, bRight);
        E.SetiSucc(iCurr, bRight);

        iCurr = iNext;

      }while(iCurr != iPrimaryEdge);
    }

    void FacetNormalRethink(const int iFacet)
    {
      Facet& F = m_aFacets[iFacet];
      const Vector3D& V0 = m_aVertices[F.aVertices(F.cVertices() - 1)].Vector();
      const Vector3D& V1 = m_aVertices[F.aVertices(0)].Vector();
      const Vector3D& V2 = m_aVertices[F.aVertices(1)].Vector();
      Vector3D V10 = V1 - V0;
      Vector3D V20 = V2 - V0;
      Vector3D Normal = V20 % V10;

      Normal.Unit();

      if(F.iNormal() == -1){
        int Count;

        F.SetiNormal(AddNormal(Normal));

        for(Count = 0; Count < F.cVertices(); Count++)
          F.aVertexNormals(Count) = F.iNormal();
      }
      else{
        m_aNormals[F.iNormal()] = Normal;
      }
    }

    const float FacetArea(const int iFacet);

    int cVertices() const { return(m_aVertices.cElements()); }
    int cNormals() const { return(m_aNormals.cElements()); }
    int cUVWs() const { return(m_aUVWs.cElements()); }
    int cColours() const { return(m_aColours.cElements()); }
    int cEdges() const { return(m_aEdges.cElements()); }
    int cFacets() const { return(m_aFacets.cElements()); }

    Vertex& aVertices(const int iPos) { return(m_aVertices[iPos]); }
    Vector3D& aNormals(const int iPos) { return(m_aNormals[iPos]); }
    Vector3D& aUVWs(const int iPos) { return(m_aUVWs[iPos]); }
    Vector3D& aColours(const int iPos) { return(m_aColours[iPos]); }
    Edge& aEdges(const int iPos) { return(m_aEdges[iPos]); }
    Facet& aFacets(const int iPos) { return(m_aFacets[iPos]); }

    const Vertex& aVertices(const int iPos) const { return(m_aVertices[iPos]); }
    const Vector3D& aNormals(const int iPos) const { return(m_aNormals[iPos]); }
    const Vector3D& aUVWs(const int iPos) const { return(m_aUVWs[iPos]); }
    const Vector3D& aColours(const int iPos) const { return(m_aColours[iPos]); }
    const Edge& aEdges(const int iPos) const { return(m_aEdges[iPos]); }
    const Facet& aFacets(const int iPos) const { return(m_aFacets[iPos]); }

    void ReserveaVertices(const int cElements)
      { m_aVertices.Reserve(cElements); }
    void ReserveaNormals(const int cElements)
      { m_aNormals.Reserve(cElements); }
    void ReserveaUVWs(const int cElements)
      { m_aUVWs.Reserve(cElements); }
    void ReserveaColours(const int cElements)
      { m_aColours.Reserve(cElements); }
    void ReserveaEdges(const int cElements)
      { m_aEdges.Reserve(cElements); }
    void ReserveaFacets(const int cElements)
      { m_aFacets.Reserve(cElements); }

    void TruncateaVertices(const int cElements)
      { m_aVertices.Truncate(cElements); }
    void TruncateaNormals(const int cElements)
      { m_aNormals.Truncate(cElements); }
    void TruncateaUVWs(const int cElements)
      { m_aUVWs.Truncate(cElements); }
    void TruncateaColours(const int cElements)
      { m_aColours.Truncate(cElements); }
    void TruncateaEdges(const int cElements)
      { m_aEdges.Truncate(cElements); }
    void TruncateaFacets(const int cElements)
      { m_aFacets.Truncate(cElements); }

    void Triangulate(const int bConvex);
    void Smooth(const float CreaseAngle,
                const int CreaseSharpness = 0);

    void DebugVertex(FILE* fd, const int iVertex) const;
    void DebugEdge(FILE* fd, const int iEdge) const;
    void DebugFacet(FILE* fd, const int iFacet) const;
    void DebugFacetFull(FILE* fd, const int iFacet) const;
    void Debug(FILE* fd) const;

    void Stats(FILE* fd) const;

  private:
    char* m_strName;                // Hull name

    DynArray<Vertex> m_aVertices;   // Hull vertices
    DynArray<Vector3D> m_aNormals;  // Hull normals
    DynArray<Vector3D> m_aUVWs;     // Hull texture coords
    DynArray<Vector3D> m_aColours;  // Hull colours
    DynArray<Facet> m_aFacets;      // Hull facets
    DynArray<Edge> m_aEdges;        // Hull edges

    int m_cEdgeHashes;              // Hull edge hash table size
    int m_cFreeEdgeHashes;          // Hull edge hash table free size
    int* m_aEdgeHashes;             // Hull edge hash table

    long EdgeHash(const long a, const long b)
      const { return((a * 1664523L) ^ (1103515245L * b)); }

    void InitHash(int cHashes);
};

#endif // HULL_H
