/***********************************************************************
This software module was originally developed by Dr. Michael Steliaros,
Superscape (UK) Ltd., in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Superscape (UK) Ltd. retains full
right to use the code for his/her own purpose, assign or donate the code
to a third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2001
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Dr. Michael Steliaros at msteliaros@superscape.com
----------------------------------------------------------------------*/
////---------------------------------------------------------------------
////    Class definition file:  SSCOLOR.CPP
////    Associated header file: SSCOLOR.H
////
////    Author:					Sean T Ellis
////    Creation date:			19/06/1997
////---------------------------------------------------------------------

#include	<math.h>

#include	"sscolor.h"

///---------------------------------------------------------------------
///
/// CssColor::CssColor
/// Default constructor. Constructs a black (0,0,0).

CssColor::CssColor()
	: m_Red(0.0f), m_Green(0.0f), m_Blue(0.0f)
{
}

///---------------------------------------------------------------------
///
/// CssColor::CssColor
/// Copy constructor.
///: From	: The color from which to copy the information.

CssColor::CssColor(const CssColor &From)
	: m_Red(From.m_Red), m_Green(From.m_Green), m_Blue(From.m_Blue)
{
}

///---------------------------------------------------------------------
///
/// CssColor::CssColor
/// Element constructor. Constructs a color given three components.
///: Red		: Red component of color.
///: Green		: Green component of color.
///: Blue		: Blue component of color.

CssColor::CssColor(float Red, float Green, float Blue)
	: m_Red(Red), m_Green(Green), m_Blue(Blue)
{
	// Note: Ranges are not asserted here since it is possible for components to temporarily
	// exceed the range	0.0f...1.0f in intermediate results in calculations.
}

///---------------------------------------------------------------------
///
/// CssColor::CssColor
/// Element constructor. Constructs a color given three components 
/// packed into an unsigned int.
///: RGBA8	: Packed RGBA8 value representing this color.

CssColor::CssColor(unsigned int RGBA8)
	: m_Red  (((RGBA8>>ColorShift_Red  )&0xFF)/255.0f), 
	  m_Green(((RGBA8>>ColorShift_Green)&0xFF)/255.0f),
	  m_Blue (((RGBA8>>ColorShift_Blue )&0xFF)/255.0f)
{
}

///---------------------------------------------------------------------
///
/// CssColor::~CssColor
/// Destructor.

CssColor::~CssColor()
{
}

///---------------------------------------------------------------------
///
/// CssColor::GetPackedRGBA8
/// Return an unsigned int containing the 8-bit colour components.

unsigned int CssColor::GetPackedRGBA8() const
{
	unsigned int	R=static_cast<unsigned int>(m_Red  *255.0f);
	unsigned int	G=static_cast<unsigned int>(m_Green*255.0f);
	unsigned int	B=static_cast<unsigned int>(m_Blue *255.0f);

	return( (((R<256)?R:255)<<ColorShift_Red  )+
			(((G<256)?G:255)<<ColorShift_Green)+
			(((B<256)?B:255)<<ColorShift_Blue )+
			ColorMask_Alpha );
}

///---------------------------------------------------------------------
///
/// CssColor::SetPackedRGBA8
/// Sets the color from an unsigned int containing the 8-bit colour components.

void CssColor::SetPackedRGBA8(unsigned int RGBA8)
{
	m_Red  =(((RGBA8>>ColorShift_Red  )&0xFF)/255.0f);
	m_Green=(((RGBA8>>ColorShift_Green)&0xFF)/255.0f);
	m_Blue =(((RGBA8>>ColorShift_Blue )&0xFF)/255.0f);
}


///---------------------------------------------------------------------
///
/// CssColor::operator=
/// Copy from specified color.
///: From	: Color to copy information from.

CssColor &CssColor::operator=(const CssColor &From)
{
	m_Red  =From.m_Red;
	m_Green=From.m_Green;
	m_Blue =From.m_Blue;
	
	return(*this);
}


///---------------------------------------------------------------------
///
/// CssColor::operator+=
/// Add a color to this color. Transparency is taken into account on the
/// added color only.
///: From	: Color to add to this color.

CssColor &CssColor::operator+=(const CssColor &From)
{
	m_Red  +=From.m_Red;
	m_Green+=From.m_Green;
	m_Blue +=From.m_Blue;

	return(*this);
}

///---------------------------------------------------------------------
///
/// CssColor::operator-=
/// Subtract a color from this color.
///: From	: Color to subtract from this color.

CssColor &CssColor::operator-=(const CssColor &From)
{
	m_Red  -=From.m_Red;
	m_Green-=From.m_Green;
	m_Blue -=From.m_Blue;
	return(*this);
}

///---------------------------------------------------------------------
///
/// CssColor::operator*=
/// Sets this color to be the product of itself and another color.
///: From	: Color to multiply with this color.

CssColor &CssColor::operator*=(const CssColor &From)
{
	m_Red  *=From.m_Red;
	m_Green*=From.m_Green;
	m_Blue *=From.m_Blue;
	return(*this);
}

///---------------------------------------------------------------------
///
/// CssColor::operator*=
/// Sets this color to be the product of itself and a scalar value.
///: Scale	: Value to multiply with this color.

CssColor &CssColor::operator*=(const float Scale)
{
	m_Red  *=Scale;
	m_Green*=Scale;
	m_Blue *=Scale;

	return(*this);
}


///---------------------------------------------------------------------
///
/// CssColor::operator/=
/// Sets this color to be the quotient of itself and a scalar value.
///: Scale	: Value to multiply with this color.

CssColor &CssColor::operator/=(const float Scale)
{
	m_Red  /=Scale;
	m_Green/=Scale;
	m_Blue /=Scale;

	return(*this);
}

///---------------------------------------------------------------------
///
/// CssColor::operator~
/// Complements this color.

CssColor CssColor::operator~() const
{
	return(CssColor(1.0f-m_Red,1.0f-m_Green,1.0f-m_Blue));
}

///---------------------------------------------------------------------
///
/// CssColor::operator+
/// Add two colors.
///: From	: Color to add to this color.
///: return	: Result of adding From to this color.

CssColor CssColor::operator+(const CssColor &From) const
{
	CssColor Temp(*this);
	Temp+=From;
	return(Temp);
}

///---------------------------------------------------------------------
///
/// CssColor::operator-
/// Subtract two colors.
///: From	: Color to subtract from this color.
///: return	: Result of subtracting From from this color.

CssColor CssColor::operator-(const CssColor &From) const
{
	CssColor Temp(*this);
	Temp-=From;
	return(Temp);
}

///---------------------------------------------------------------------
///
/// CssColor::operator*
/// Product of two colors.
///: From	: Color to multiply with this color.
///: return	: The product of this color and From.

CssColor CssColor::operator*(const CssColor &From) const
{
	CssColor Temp(*this);
	Temp*=From;
	return(Temp);
}

///---------------------------------------------------------------------
///
/// CssColor::operator*
/// Product of a color and a scalar.
///: Scale	: Scale factor to multiply with this color.
///: return	: The product of this color and Scale.

CssColor CssColor::operator*(const float Scale) const
{
	CssColor Temp(*this);
	Temp*=Scale;
	return(Temp);
}

///---------------------------------------------------------------------
///
/// CssColor::operator/
/// Quotient of a color and a scalar.
///: Scale	: Scale factor to multiply with this color.
///: return	: The product of this color and Scale.

CssColor CssColor::operator/(const float Scale) const
{
	CssColor Temp(*this);
	Temp/=Scale;
	return(Temp);
}
