/***********************************************************************
This software module was originally developed by Dr. Michael Steliaros,
Superscape (UK) Ltd., in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Superscape (UK) Ltd. retains full
right to use the code for his/her own purpose, assign or donate the code
to a third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2001
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Dr. Michael Steliaros at msteliaros@superscape.com
----------------------------------------------------------------------*/

////---------------------------------------------------------------------
////    Superscape subdivision surface storage class
////
////    Author:                 Michael K. Steliaros (Superscape)
////    Last modification date: 28-11-2000
////    Creation date:          22-09-2000
////---------------------------------------------------------------------

#include <string.h>
#include <assert.h>
#include "subdivisionsurfacexloop.h"

#ifdef WIN32
#include <malloc.h>
#define alloca _alloca
#endif

void SubdivisionSurfaceXLoop::SetVertexVector(const int iVertex,
                                         const Vector3D& V)
{
  Hull& Control = m_aLevels[0];

  Control.aVertices(iVertex).SetVector(V);

  Control.MarkVertexDirty(iVertex);

  m_bDirty = 1;
}

void SubdivisionSurfaceXLoop::SetNormal(const int iNormal,
                                   const Vector3D& N)
{
  Hull& Control = m_aLevels[0];

  Control.aNormals(iNormal) = N;
}

void SubdivisionSurfaceXLoop::SetUVW(const int iUVW,
                                const Vector3D& UVW)
{
  Hull& Control = m_aLevels[0];

  Control.aUVWs(iUVW) = UVW;
}

void SubdivisionSurfaceXLoop::SetColour(const int iColour,
                                   const Vector3D& RGB)
{
  Hull& Control = m_aLevels[0];

  Control.aColours(iColour) = RGB;
}

void SubdivisionSurfaceXLoop::MarkDirty(void)
{
  Hull& Control = m_aLevels[0];

  int iVertex;

  for(iVertex = 0; iVertex < Control.cVertices(); iVertex++)
    Control.MarkVertexDirty(iVertex);

  m_bDirty = 1;
}

void SubdivisionSurfaceXLoop::SetEdgeSharpness(const int iHead,
                                          const int iTail,
                                          const int Sharpness)
{
  Hull& Control = m_aLevels[0];
  int iEdge = Control.iEdge(iHead, iTail);

  if(iEdge != -1){

    Control.aEdges(iEdge).SetSharpness(Sharpness);

    Control.MarkVertexDirty(iHead);
    Control.MarkVertexDirty(iTail);

    m_bDirty = 1;
  }
}

void SubdivisionSurfaceXLoop::SubdivideVertex(const int iPrevious,
                                         const int iCurrent,
                                         const int iVertex)
{
#define PERTURBATION
#ifdef PERTURBATION
  const Hull& Previous = m_aLevels[iPrevious];
  const Vertex& Old = Previous.aVertices(iVertex);

  if(Old.bSharp())
    return;

  Vector3D& V0 = m_aLevels[iCurrent].aVertices(iVertex).Vector();
  int aSharp[2], cSharp = 0;
  int Count;

  for(Count = 0; Count < Old.cEdges(); Count++){
    const Edge& E = Previous.aEdges(Old.aEdges(Count));

    if(E.bSharp())
      if(++cSharp <= 2)                          // Store first two sharp edges
        aSharp[cSharp - 1] = Old.aEdges(Count);
  }

  if(cSharp == 2){                   // Sharp vertex with exactly 2 sharp edges
    const Vector3D& V1 = Previous.EdgeOtherVertex(aSharp[0], iVertex).Vector();
    const Vector3D& V2 = Previous.EdgeOtherVertex(aSharp[1], iVertex).Vector();

    V0 *= 6.0f;
    V0 += V1;
    V0 += V2;
    V0 *= 1.0f / 8.0f;
  }
  else if(cSharp < 2){                                         // Smooth vertex
    Vector3D Vn;
    float n = 0.0f;

    for(Count = 0; Count < Old.cEdges(); Count++){             // V1 + ... + Vn
      int iEdge = Old.aEdges(Count);
      const Edge& E = Previous.aEdges(iEdge);

      if(E.bVisible()){

        Vn += Previous.EdgeOtherVertex(iEdge, iVertex).Vector();

        n++;
      }
    }

    float an = 3.0f + 2.0f * static_cast<float>(cos(2.0f * PI / n));
    an = an * an / 64.0f;
    an = 5.0f / 8.0f - an;                                              // a(n)

    an = (n - an * n) / an;                                         // alpha(n)

    V0 *= an;
    V0 += Vn;

    an = 1.0f / (n + an);

    V0 *= an;
  }
#endif
}

void SubdivisionSurfaceXLoop::SubdivideVertexNormalSmooth(const int iCurrent,
                                                     const int iVertex,
                                                     Vector3D& Normal)
{
  const Hull& Current = m_aLevels[iCurrent];
  const Vertex& V = Current.aVertices(iVertex);
  float i = 1.0f, n = static_cast<float>(V.cEdges());
  float C1 = static_cast<float>(cos(2.0f * PI / n)), C2;
  Vector3D T1;                        // T1 = C1 * V1 + C2 * V2 + ... + Cn * Vn
  Vector3D T2;                        // T2 = C2 * V1 + C3 * V2 + ... + C1 * Vn
  int Count;

  for(Count = 0; Count < V.cEdges(); Count++){
    const Vector3D& Vi =
      Current.EdgeOtherVertex(V.aEdges(Count), iVertex).Vector();

    C2 = static_cast<float>(cos(2.0f * PI * (++i) / n));

    T1 += Vi * C1;
    T2 += Vi * C2;

    C1 = C2;
  }

  Normal = T1 % T2;
  Normal.Unit();
}

void SubdivisionSurfaceXLoop::SubdivideVertexNormalSharp(const int iCurrent,
                                                    const int iVertex,
                                                    const int iStart,
                                                    const int iSpan,
                                                    Vector3D& Normal)
{
  const Hull& Current = m_aLevels[iCurrent];
  const Vertex& V = Current.aVertices(iVertex);
  const Vector3D& V0 = V.Vector();
  const Vector3D& V1 = Current.EdgeOtherVertex(
                         V.aEdges(iStart), iVertex).Vector();
  const Vector3D& Vm = Current.EdgeOtherVertex(
                         V.aEdges((iStart + iSpan - 1) % V.cEdges()),
                         iVertex).Vector();
  Vector3D T1, T2;

  T1 = V1 - Vm;

  switch(iSpan){
    case 2 :                                         // T2 = 2.0 * V0 - V1 - V2
      T2 = V0 + V0;
      T2 -= V1;
      T2 -= Vm;

      break;
    case 3 : {                                                  // T2 = V0 - V2
      const Vector3D& V2 = Current.EdgeOtherVertex(
                             V.aEdges((iStart + 1) % V.cEdges()),
                             iVertex).Vector();

      T2 = V0 - V2;

      break; }
    case 4 : {                 // T2 = 2.0 * V0 + V1 - 2.0 * V2 - 2.0 * V3 + Vm
      const Vector3D& V2 = Current.EdgeOtherVertex(
                             V.aEdges((iStart + 1) % V.cEdges()),
                             iVertex).Vector();
      const Vector3D& V3 = Current.EdgeOtherVertex(
                             V.aEdges((iStart + 2) % V.cEdges()),
                             iVertex).Vector();

      T2 = V0 + V0;
      T2 += V1;
      T2 -= V2;
      T2 -= V2;
      T2 -= V3;
      T2 -= V3;
      T2 += Vm;

      break; }
    default : {                       // T2 = w1 * V1 + w2 * V2 + ... + wm * Vm
      float Theta = PI / (float)(iSpan - 1);
      float W = static_cast<float>(sin(Theta));
      int Count, Count_1;

      T2 = V1 + Vm;
      T2 *= W;

      for(Count = 2, Count_1 = 1; Count < iSpan; Count++, Count_1++){
        const Vector3D& Vi = Current.EdgeOtherVertex(
                               V.aEdges((iStart + Count_1) % V.cEdges()),
                               iVertex).Vector();

        W = (2.0f * static_cast<float>(cos(Theta)) - 2.0f) *
            static_cast<float>(sin(static_cast<float>(Count_1) * Theta));

        T2 += Vi * W;
      }

      break; }
  }

  Normal = T1 % T2;
  Normal.Unit();
}

void SubdivisionSurfaceXLoop::SubdivideVertexNormal(const int iCurrent,
                                               const int iVertex)
{
  Hull& Current = m_aLevels[iCurrent];
  Vertex& V = Current.aVertices(iVertex);
  int* aSharp = (int *)alloca(V.cEdges() * sizeof(int));
  int* aSharpSpan = (int *)alloca(V.cEdges() * sizeof(int));
  int cSharp = 0;
  int Count;

  assert((aSharp != NULL) && (aSharpSpan != NULL));

  for(Count = 0; Count < V.cEdges(); Count++)
    aSharpSpan[Count] = 0;

  const int iPrimaryEdge = Current.VertexiPrimaryEdge(iVertex);
  int iCurr = iPrimaryEdge;
  int iPos = 0;

  do{
    const Edge& E = Current.aEdges(iCurr);
    int bRight = (E.iVertex(1) == iVertex);

    if(E.bSharp())
      aSharp[cSharp++] = iPos;

    if(cSharp > 0)                        // Keep track of the sharp span sizes
      aSharpSpan[cSharp - 1]++;

    V.aEdges(iPos++) = iCurr;             // Correct edge order in vertex array

    if(E.iPred(bRight) == iCurr) break; else iCurr = E.iPred(bRight);

  }while(iCurr != iPrimaryEdge);

  if(cSharp >= 2){                                              // Sharp vertex
    int iSpan;

    for(iSpan = 0; iSpan < cSharp; iSpan++){
      const Edge& E = Current.aEdges(V.aEdges(aSharp[iSpan]));
      int m = aSharpSpan[iSpan];

      if(E.iFacet(E.iVertex(1) == iVertex) != -1)
        m++;            // Include sharp edge from next span if not at boundary

      if(m >= 2){
        Vector3D Normal;

        SubdivideVertexNormalSharp(iCurrent, iVertex, aSharp[iSpan], m, Normal);

        iPos = Current.AddNormal(Normal);

        assert(iPos != -1);

        Current.SetVertexNormal(iVertex, aSharp[iSpan], m, iPos);
      }
    }
  }
  else{                                                        // Smooth vertex
    Vector3D Normal;

    SubdivideVertexNormalSmooth(iCurrent, iVertex, Normal);

    iPos = Current.AddNormal(Normal);

    assert(iPos != -1);

    Current.SetVertexNormal(iVertex, iPos);
  }
}

void SubdivisionSurfaceXLoop::SubdivideEdgeVector(const int iPrevious,
                                             const int iEdge,
                                             Vector3D& V)
{
  Hull& Previous = m_aLevels[iPrevious];
  const Edge& Old = Previous.aEdges(iEdge);
  int iHead = Old.iVertex(1);
  Vector3D& V0 = Previous.aVertices(iHead).Vector();
  int iTail = Old.iVertex(0);
  Vector3D& V1 = Previous.aVertices(iTail).Vector();

#ifdef PERTURBATION
  if(Old.bSharp()){
    V = V0 + V1;
    V *= 0.5f;
  }
  else{                                                          // Smooth edge
    Vector3D* V2 = NULL;
    Vector3D* V3 = NULL;
    Vector3D* V4 = NULL;
    Vector3D* V5 = NULL;
    int Count = 0, iCurr;
    Edge* E = &Previous.aEdges(iCurr = iEdge);

    do{                               // Clockwise at head from edge to find V4

      E = &Previous.aEdges(iCurr = E->iPred(E->iVertex(1) == iHead));

      if(E->bVisible()){

        V4 = &Previous.EdgeOtherVertex(iCurr, iHead).Vector();

        Count++;

        break;
      }

    }while(iCurr != iEdge);

    if(Count == 0){                                         // No visible edges
      V3 = V4 = &V0;
    }
    else{

      E = &Previous.aEdges(iCurr = iEdge);

      do{                         // Anticlockwise at head from edge to find V3

        E = &Previous.aEdges(iCurr = E->iSucc(E->iVertex(1) != iHead));

        if(E->bVisible()){

          V3 = &Previous.EdgeOtherVertex(iCurr, iHead).Vector();

          break;
        }

      }while(iCurr != iEdge);
    }

    Count = 0;

    E = &Previous.aEdges(iCurr = iEdge);

    do{                               // Clockwise at tail from edge to find V2

      E = &Previous.aEdges(iCurr = E->iPred(E->iVertex(1) == iTail));

      if(E->bVisible()){

        V2 = &Previous.EdgeOtherVertex(iCurr, iTail).Vector();

        Count++;

        break;
      }

    }while(iCurr != iEdge);

    if(Count == 0){                                         // No visible edges
      V2 = V5 = &V1;
    }
    else{

      E = &Previous.aEdges(iCurr = iEdge);

      do{                         // Anticlockwise at tail from edge to find V5

        E = &Previous.aEdges(iCurr = E->iSucc(E->iVertex(1) != iTail));

        if(E->bVisible()){

          V5 = &Previous.EdgeOtherVertex(iCurr, iTail).Vector();

          break;
        }

      }while(iCurr != iEdge);
    }

    assert((V2 != NULL) && (V3 != NULL) && (V4 != NULL) && (V5 != NULL));

    if(Old.bVisible()){
      V = V0 + V1;
      V *= 6.0f;
      V += *V2;
      V += *V3;
      V += *V4;
      V += *V5;
      V *= 1.0f / 16.0f;
    }
    else{
      V = V0 + V1;
      V *= 2.0f;
      V += *V2;
      V += *V3;
      V += *V4;
      V += *V5;
      V *= 1.0f / 8.0f;
    }
  }
#else
  V = V0 + V1;
  V *= 0.5f;
#endif
}

void SubdivisionSurfaceXLoop::SubdivideEdgeUVW(const int iCurrent,
                                          const int iHead,
                                          const int iTail,
                                          const int iMidpoint,
                                          const int iLeft,
                                          const int iRight)
{
  Hull& Current = m_aLevels[iCurrent];
  int iHLeftUVW = -1, iTLeftUVW = -1;
  int iHRightUVW = -1, iTRightUVW = -1;
  int iMLeftPos = -1;
  int iMRightPos = -1;
  int iSort = 0;

  if(iLeft != -1){

    iMLeftPos = Current.aFacets(iLeft).iVertexPos(iMidpoint);

    assert(iMLeftPos != -1);

    const Facet& Left = Current.aFacets(iLeft);

    int iHLeftPos = Left.iVertexPos(iHead);
    int iTLeftPos = Left.iVertexPos(iTail);

    assert((iHLeftPos != -1) && (iTLeftPos != -1));

    iHLeftUVW = Left.aVertexUVWs(iHLeftPos);
    iTLeftUVW = Left.aVertexUVWs(iTLeftPos);

    if((iHLeftUVW != -1) && (iTLeftUVW != -1))
      iSort += 1;
  }

  if(iRight != -1){

    iMRightPos = Current.aFacets(iRight).iVertexPos(iMidpoint);

    assert(iMRightPos != -1);

    const Facet& Right = Current.aFacets(iRight);

    int iHRightPos = Right.iVertexPos(iHead);
    int iTRightPos = Right.iVertexPos(iTail);

    assert((iHRightPos != -1) && (iTRightPos != -1));

    iHRightUVW = Right.aVertexUVWs(iHRightPos);
    iTRightUVW = Right.aVertexUVWs(iTRightPos);

    if((iHRightUVW != -1) && (iTRightUVW != -1))
      iSort += 2;
  }

  if(iSort == 0) return;             // No facets on either side; nothing to do

  float Al, Bl;

  {                                               // Midpoint ratios along edge
    const Vector3D& Head = Current.aVertices(iHead).Vector();
    const Vector3D& Tail = Current.aVertices(iTail).Vector();
    const Vector3D& Midpoint = Current.aVertices(iMidpoint).Vector();
    Vector3D A, B;
    float ABl;

    A = Head - Midpoint; Al = A.Magnitude();
    B = Tail - Midpoint; Bl = B.Magnitude();

    if((ABl = Al + Bl) > 0.0f){ Al /= ABl; Bl /= ABl; }
  }

  switch(iSort){
    case 1 : {                                               // Left facet only
      Facet& Left = Current.aFacets(iLeft);

      if(iHLeftUVW != iTLeftUVW){                     // Different; interpolate
        Vector3D& A = Current.aUVWs(iHLeftUVW);
        Vector3D& B = Current.aUVWs(iTLeftUVW);

        Left.aVertexUVWs(iMLeftPos) =
          Current.AddUVW(Bl * A[0] + Al * B[0],
                         Bl * A[1] + Al * B[1],
                         Bl * A[2] + Al * B[2]);
      }
      else{                                     // Identical; don't interpolate
        Left.aVertexUVWs(iMLeftPos) = iHLeftUVW;
      }

      break; }
    case 2 : {                                              // Right facet only
      Facet& Right = Current.aFacets(iRight);

      if(iHRightUVW != iTRightUVW){                   // Different; interpolate
        Vector3D& A = Current.aUVWs(iHRightUVW);
        Vector3D& B = Current.aUVWs(iTRightUVW);

        Right.aVertexUVWs(iMRightPos) =
          Current.AddUVW(Bl * A[0] + Al * B[0],
                         Bl * A[1] + Al * B[1],
                         Bl * A[2] + Al * B[2]);
      }
      else{                                     // Identical; don't interpolate
        Right.aVertexUVWs(iMRightPos) = iHRightUVW;
      }

      break; }
    default : {                                     // = 3; left + right facets
      Facet& Left = Current.aFacets(iLeft);
      Facet& Right = Current.aFacets(iRight);

      if((iHLeftUVW == iHRightUVW) && (iTLeftUVW == iTRightUVW)){
        if(iHLeftUVW != iTLeftUVW){                   // Different; interpolate
          Vector3D& A = Current.aUVWs(iHLeftUVW);
          Vector3D& B = Current.aUVWs(iTLeftUVW);

          Left.aVertexUVWs(iMLeftPos) =
          Right.aVertexUVWs(iMRightPos) =
            Current.AddUVW(Bl * A[0] + Al * B[0],
                           Bl * A[1] + Al * B[1],
                           Bl * A[2] + Al * B[2]);
        }
        else{                                   // Identical; don't interpolate
          Left.aVertexUVWs(iMLeftPos) =
          Right.aVertexUVWs(iMRightPos) = iHLeftUVW;
        }
      }
      else{                                        // Different left from right
        if(iHLeftUVW != iTLeftUVW){                   // Different; interpolate
          Vector3D& A = Current.aUVWs(iHLeftUVW);
          Vector3D& B = Current.aUVWs(iTLeftUVW);

          Left.aVertexUVWs(iMLeftPos) =
            Current.AddUVW(Bl * A[0] + Al * B[0],
                           Bl * A[1] + Al * B[1],
                           Bl * A[2] + Al * B[2]);
        }
        else{                                   // Identical; don't interpolate
          Left.aVertexUVWs(iMLeftPos) = iHLeftUVW;
        }

        if(iHRightUVW != iTRightUVW){                 // Different; interpolate
          Vector3D& A = Current.aUVWs(iHRightUVW);
          Vector3D& B = Current.aUVWs(iTRightUVW);

          Right.aVertexUVWs(iMRightPos) =
            Current.AddUVW(Bl * A[0] + Al * B[0],
                           Bl * A[1] + Al * B[1],
                           Bl * A[2] + Al * B[2]);
        }
        else{                                   // Identical; don't interpolate
          Right.aVertexUVWs(iMRightPos) = iHRightUVW;
        }
      }

      break; }
  }
}

void SubdivisionSurfaceXLoop::SubdivideEdgeColour(const int iCurrent,
                                             const int iHead,
                                             const int iTail,
                                             const int iMidpoint,
                                             const int iLeft,
                                             const int iRight)
{
  Hull& Current = m_aLevels[iCurrent];
  int iHLeftColour = -1, iTLeftColour = -1;
  int iHRightColour = -1, iTRightColour = -1;
  int iMLeftPos = -1;
  int iMRightPos = -1;
  int iSort = 0;

  if(iLeft != -1){

    iMLeftPos = Current.aFacets(iLeft).iVertexPos(iMidpoint);

    assert(iMLeftPos != -1);

    const Facet& Left = Current.aFacets(iLeft);

    int iHLeftPos = Left.iVertexPos(iHead);
    int iTLeftPos = Left.iVertexPos(iTail);

    assert((iHLeftPos != -1) && (iTLeftPos != -1));

    iHLeftColour = Left.aVertexColours(iHLeftPos);
    iTLeftColour = Left.aVertexColours(iTLeftPos);

    if((iHLeftColour != -1) && (iTLeftColour != -1))
      iSort += 1;
  }

  if(iRight != -1){

    iMRightPos = Current.aFacets(iRight).iVertexPos(iMidpoint);

    assert(iMRightPos != -1);

    const Facet& Right = Current.aFacets(iRight);

    int iHRightPos = Right.iVertexPos(iHead);
    int iTRightPos = Right.iVertexPos(iTail);

    assert((iHRightPos != -1) && (iTRightPos != -1));

    iHRightColour = Right.aVertexColours(iHRightPos);
    iTRightColour = Right.aVertexColours(iTRightPos);

    if((iHRightColour != -1) && (iTRightColour != -1))
      iSort += 2;
  }

  if(iSort == 0) return;             // No facets on either side; nothing to do

  float Al, Bl;

  {                                               // Midpoint ratios along edge
    const Vector3D& Head = Current.aVertices(iHead).Vector();
    const Vector3D& Tail = Current.aVertices(iTail).Vector();
    const Vector3D& Midpoint = Current.aVertices(iMidpoint).Vector();
    Vector3D A, B;
    float ABl;

    A = Head - Midpoint; Al = A.Magnitude();
    B = Tail - Midpoint; Bl = B.Magnitude();

    if((ABl = Al + Bl) > 0.0f){ Al /= ABl; Bl /= ABl; }
  }

  switch(iSort){
    case 1 : {                                               // Left facet only
      Facet& Left = Current.aFacets(iLeft);

      if(iHLeftColour != iTLeftColour){               // Different; interpolate
        Vector3D& A = Current.aColours(iHLeftColour);
        Vector3D& B = Current.aColours(iTLeftColour);

        Left.aVertexColours(iMLeftPos) =
          Current.AddColour(Bl * A[0] + Al * B[0],
                            Bl * A[1] + Al * B[1],
                            Bl * A[2] + Al * B[2]);
      }
      else{                                     // Identical; don't interpolate
        Left.aVertexColours(iMLeftPos) = iHLeftColour;
      }

      break; }
    case 2 : {                                              // Right facet only
      Facet& Right = Current.aFacets(iRight);

      if(iHRightColour != iTRightColour){             // Different; interpolate
        Vector3D& A = Current.aColours(iHRightColour);
        Vector3D& B = Current.aColours(iTRightColour);

        Right.aVertexColours(iMRightPos) =
          Current.AddColour(Bl * A[0] + Al * B[0],
                            Bl * A[1] + Al * B[1],
                            Bl * A[2] + Al * B[2]);
      }
      else{                                     // Identical; don't interpolate
        Right.aVertexColours(iMRightPos) = iHRightColour;
      }

      break; }
    default : {                                     // = 3; left + right facets
      Facet& Left = Current.aFacets(iLeft);
      Facet& Right = Current.aFacets(iRight);

      if((iHLeftColour == iHRightColour) && (iTLeftColour == iTRightColour)){
        if(iHLeftColour != iTLeftColour){             // Different; interpolate
          Vector3D& A = Current.aColours(iHLeftColour);
          Vector3D& B = Current.aColours(iTLeftColour);

          Left.aVertexColours(iMLeftPos) =
          Right.aVertexColours(iMRightPos) =
            Current.AddColour(Bl * A[0] + Al * B[0],
                              Bl * A[1] + Al * B[1],
                              Bl * A[2] + Al * B[2]);
        }
        else{                                   // Identical; don't interpolate
          Left.aVertexColours(iMLeftPos) =
          Right.aVertexColours(iMRightPos) = iHLeftColour;
        }
      }
      else{                                        // Different left from right
        if(iHLeftColour != iTLeftColour){             // Different; interpolate
          Vector3D& A = Current.aColours(iHLeftColour);
          Vector3D& B = Current.aColours(iTLeftColour);

          Left.aVertexColours(iMLeftPos) =
            Current.AddColour(Bl * A[0] + Al * B[0],
                              Bl * A[1] + Al * B[1],
                              Bl * A[2] + Al * B[2]);
        }
        else{                                   // Identical; don't interpolate
          Left.aVertexColours(iMLeftPos) = iHLeftColour;
        }

        if(iHRightColour != iTRightColour){           // Different; interpolate
          Vector3D& A = Current.aColours(iHRightColour);
          Vector3D& B = Current.aColours(iTRightColour);

          Right.aVertexColours(iMRightPos) =
            Current.AddColour(Bl * A[0] + Al * B[0],
                              Bl * A[1] + Al * B[1],
                              Bl * A[2] + Al * B[2]);
        }
        else{                                   // Identical; don't interpolate
          Right.aVertexColours(iMRightPos) = iHRightColour;
        }
      }

      break; }
  }
}

int SubdivisionSurfaceXLoop::SubdivideEdge(const int iPrevious,
                                      const int iCurrent,
                                      const int iEdge)
{
  Hull& Previous = m_aLevels[iPrevious];
  Hull& Current = m_aLevels[iCurrent];
  Vector3D Midpoint;
  int iMidpoint = -1;

  SubdivideEdgeVector(iPrevious, iEdge, Midpoint);

  if((iMidpoint = Current.AddVertex(Midpoint)) != -1){
    Edge& Old = Current.aEdges(iEdge);
    int iHead = Old.iVertex(1), iTail = Old.iVertex(0);
    int iLeft = Old.iFacet(0), iRight = Old.iFacet(1);
    Vertex& Tail = Current.aVertices(iTail);
    int iNewEdge = Current.cEdges();
    Edge E;

    Previous.aEdges(iEdge).SetiMidpoint(iMidpoint);   // Previous level pointer

    E.SetiVertex(iTail, 0);
    E.SetiVertex(iMidpoint, 1);

    E.SetiFacet(iLeft, 0);
    E.SetiFacet(iRight, 1);

    E.SetSharpness(Old.Sharpness());

    E.SetbVisible(Old.bVisible());

    Old.SetiVertex(iMidpoint, 0);

    if(iRight == -1){
      E.SetiPred(iNewEdge, 1);
      E.SetiSucc(iNewEdge, 1);
    }
    else{
      E.SetiPred(iEdge, 1);
      E.SetiSucc(Old.iSucc(1), 1);
      Old.SetiSucc(iNewEdge, 1);
    }

    if(iLeft == -1){
      E.SetiPred(iNewEdge, 0);
      E.SetiSucc(iNewEdge, 0);
    }
    else{
      E.SetiPred(Old.iPred(0), 0);
      E.SetiSucc(iEdge, 0);
      Old.SetiPred(iNewEdge, 0);
    }

    Current.AddEdge(E);                                         // Add new edge

    Current.SetiEdge(iHead, iMidpoint, iEdge);               // Fix edge hashes
    Current.SetiEdge(iMidpoint, iTail, iNewEdge);

    int Count = Tail.iEdgePos(iEdge);

    assert(Count != -1);

    Tail.aEdges(Count) = iNewEdge;

    if(E.iPred(0) != iNewEdge)            // Fix edge neighbours' back-pointers
      Current.aEdges(E.iPred(0)).ReplaceReferences(iEdge, iNewEdge);
    if(E.iSucc(1) != iNewEdge)
      Current.aEdges(E.iSucc(1)).ReplaceReferences(iEdge, iNewEdge);

    Vertex& Midpoint = Current.aVertices(iMidpoint);

    Midpoint.AddEdge(iEdge);
    Midpoint.AddEdge(iNewEdge);

    for(Count = 0; Count < 2; Count++){     // Insert new vertex in facet lists
      int iFacet = E.iFacet(Count);

      if(iFacet != -1){
        Facet& F = Current.aFacets(iFacet);

        F.AddVertex(iMidpoint, F.iNormal(), -1, -1);

        int iA = F.aVertices(0), iB;
        int Index;

        for(Index = F.cVertices() - 1; Index > 0 ; Index--){

          iB = F.aVertices(Index - 1);

          if(((iA == iHead) && (iB == iTail)) ||
             ((iA == iTail) && (iB == iHead))){

            F.aVertices(Index) = iMidpoint;
            F.aVertexNormals(Index) = F.iNormal();
            F.aVertexUVWs(Index) = -1;
            F.aVertexColours(Index) = -1;

            break;
          }

          F.aVertices(Index) = iA = iB;
          F.aVertexNormals(Index) = F.aVertexNormals(Index - 1);
          F.aVertexUVWs(Index) = F.aVertexUVWs(Index - 1);
          F.aVertexColours(Index) = F.aVertexColours(Index - 1);
        }
      }
    }

    SubdivideEdgeUVW(iCurrent, iHead, iTail, iMidpoint, iLeft, iRight);
    SubdivideEdgeColour(iCurrent, iHead, iTail, iMidpoint, iLeft, iRight);
  }

  return(iMidpoint);
}

void SubdivisionSurfaceXLoop::SubdivideEdgeVisibility(const int iPrevious,
                                                 const int iCurrent,
                                                 const int iEdge)
{
                                                  //  Previous a-bcde - AB-CDEF
  const int VisTable[] = { 0x00,                  //           0-0000 - 00-0000
                           0x0A,                  //  e     b  0-0001 - 00-1010
                           0x05,                  //   \   /   0-0010 - 00-0101
                           0x0C,                  //    \ /    0-0011 - 00-1100
                           0x0A,                  //     |     0-0100 - 00-1010
                           0x0A,                  //     a     0-0101 - 00-1010
                           0x19,                  //     |     0-0110 - 01-1001
                           0x19,                  //    / \    0-0111 - 01-1001
                           0x05,                  //   /   \   0-1000 - 00-0101
                           0x26,                  //  d     c  0-1001 - 10-0110
                           0x05,                  //           0-1010 - 00-0101
                           0x26,                  //     |     0-1011 - 10-0110
                           0x03,                  //     V     0-1100 - 00-0011
                           0x26,                  //           0-1101 - 10-0110
                           0x19,                  //    Next   0-1110 - 01-1001
                           0x0F,                  //           0-1111 - 00-1111
                           0x00,                  // F       C 1-0000 - 00-0000
                           0x00,                  //  \  A  /  1-0001 - 00-0000
                           0x00,                  //   \ | /   1-0010 - 00-0000
                           0x03,                  //    \|/    1-0011 - 00-0011
                           0x00,                  //     M     1-0100 - 00-0000
                           0x00,                  //    /|\    1-0101 - 00-0000
                           0x00,                  //   / | \   1-0110 - 00-0000
                           0x03,                  //  /  B  \  1-0111 - 00-0011
                           0x00,                  // E       D 1-1000 - 00-0000
                           0x00,                  //           1-1001 - 00-0000
                           0x00,                  //           1-1010 - 00-0000
                           0x03,                  //           1-1011 - 00-0011
                           0x0C,                  //           1-1100 - 00-1100
                           0x0C,                  //           1-1101 - 00-1100
                           0x0C,                  //           1-1110 - 00-1100
                           0x0F };                //           1-1111 - 00-1111
  const Hull& Previous = m_aLevels[iPrevious];
  Hull& Current = m_aLevels[iCurrent];
  const Edge& Old = Previous.aEdges(iEdge);

  if(!Old.bBoundary()){
    int b = Old.iSucc(1);
    int c = Old.iPred(1);
    int d = Old.iSucc(0);
    int e = Old.iPred(0);
    int M = Old.iMidpoint();
    int Visibility = Old.bVisible();

    Visibility <<= 1; Visibility |= Previous.aEdges(b).bVisible();
    Visibility <<= 1; Visibility |= Previous.aEdges(c).bVisible();
    Visibility <<= 1; Visibility |= Previous.aEdges(d).bVisible();
    Visibility <<= 1; Visibility |= Previous.aEdges(e).bVisible();

    Visibility = VisTable[Visibility];

    int iPos;

    if(Visibility & 0x20)                                                  // A
      if((iPos = Current.iEdge(M, Old.iVertex(1))) != -1)
        Current.aEdges(iPos).SetVisible();
    if(Visibility & 0x10)                                                  // B
      if((iPos = Current.iEdge(M, Old.iVertex(0))) != -1)
        Current.aEdges(iPos).SetVisible();
    if(Visibility & 0x08)                                                  // C
      if((iPos = Current.iEdge(M, Previous.aEdges(b).iMidpoint())) != -1)
        Current.aEdges(iPos).SetVisible();
    if(Visibility & 0x04)                                                  // D
      if((iPos = Current.iEdge(M, Previous.aEdges(c).iMidpoint())) != -1)
        Current.aEdges(iPos).SetVisible();
    if(Visibility & 0x02)                                                  // E
      if((iPos = Current.iEdge(M, Previous.aEdges(d).iMidpoint())) != -1)
        Current.aEdges(iPos).SetVisible();
    if(Visibility & 0x01)                                                  // F
      if((iPos = Current.iEdge(M, Previous.aEdges(e).iMidpoint())) != -1)
        Current.aEdges(iPos).SetVisible();
  }
}

void SubdivisionSurfaceXLoop::SubdivideFacet(const int iCurrent,
                                        const int iFacet)
{
  Hull& Current = m_aLevels[iCurrent];
  Facet& Old = Current.aFacets(iFacet);
  int OldcVertices = Old.cVertices();
  int cVertices = OldcVertices / 2;
  int* aVertices = (int *)alloca(cVertices * sizeof(int));
  int* aVertexUVWs = (int *)alloca(cVertices * sizeof(int));
  int* aVertexColours = (int *)alloca(cVertices * sizeof(int));

  assert((aVertices != NULL) &&
        (aVertexUVWs != NULL) &&
        (aVertexColours != NULL));

  int Count, Index;

  Current.FacetUnlinkEdges(iFacet);

  for(Count = 0, Index = 0; Count < OldcVertices; Count++){
    int iVertex = Old.aVertices(Count);

    if(Count % 2 == 1){

      aVertices[Index] = iVertex;
      aVertexUVWs[Index] = Old.aVertexUVWs(Count);
      aVertexColours[Index] = Old.aVertexColours(Count);

      Index++;
    }
  }

  for(Count = 0, Index = 0; Count < OldcVertices; Count++){
    if(Count % 2 == 0){
      Facet F(3);

      F.aVertices(0) = Current.aFacets(iFacet).aVertices(Count);
      F.aVertices(1) = aVertices[Index];
      F.aVertices(2) = aVertices[(Index + cVertices - 1) % cVertices];

      F.aVertexUVWs(0) = Current.aFacets(iFacet).aVertexUVWs(Count);
      F.aVertexUVWs(1) = aVertexUVWs[Index];
      F.aVertexUVWs(2) = aVertexUVWs[(Index + cVertices - 1) % cVertices];

      F.aVertexColours(0) = Current.aFacets(iFacet).aVertexColours(Count);
      F.aVertexColours(1) = aVertexColours[Index];
      F.aVertexColours(2) = aVertexColours[(Index + cVertices - 1) % cVertices];

      Current.AddFacet(F);

      Index++;
    }
  }

  Facet& F = Current.aFacets(iFacet);    // Same as Old but could be rellocated

  for(Count = 0; Count < cVertices; Count++){          // Fix facet information
    F.aVertices(Count) = aVertices[Count];
    F.aVertexNormals(Count) = F.iNormal();
    F.aVertexUVWs(Count) = F.aVertexUVWs(Count * 2 + 1);
    F.aVertexColours(Count) = F.aVertexColours(Count * 2 + 1);
  }

  F.TruncateaVertices(cVertices);               // Halve the number of vertices

  Current.FacetNormalRethink(iFacet);               // Rethink the facet normal

  Current.FacetLinkEdges(iFacet, 1);
}

void SubdivisionSurfaceXLoop::Subdivide()
{
  int iPrevious, iCurrent;

  for(iPrevious = 0, iCurrent = 1;
      iPrevious < m_aLevels.cElements() - 1;
      iPrevious++, iCurrent++){

    Hull& Previous = m_aLevels[iPrevious];

    m_aLevels[iCurrent] = Previous;                 // Duplicate previous level

    Hull& Current = m_aLevels[iCurrent];

    if(Current.strName() != NULL){                  // Add '+' to start of name
      char* name = (char *)alloca(strlen(Current.strName()) + 2);

      sprintf(name, "+%s", Current.strName());

      Current.SetstrName(name);
    }

    Current.ReserveaVertices(2 * Current.cVertices());         // Reserve space
    Current.ReserveaEdges(3 * Current.cEdges());
    Current.ReserveaFacets(4 * Current.cFacets());

    Current.TruncateaNormals(Current.cFacets());     // Remove unnecessary ones

    int Count;

    for(Count = 0; Count < Previous.cVertices(); Count++)   // Perturb vertices
      if(Previous.aVertices(Count).cEdges() > 0)
        SubdivideVertex(iPrevious, iCurrent, Count);

    for(Count = 0; Count < Previous.cEdges(); Count++)   // Split+perturb edges
      SubdivideEdge(iPrevious, iCurrent, Count);

    for(Count = 0; Count < Previous.cFacets(); Count++)         // Split facets
      SubdivideFacet(iCurrent, Count);

    for(Count = 0; Count < Previous.cEdges(); Count++)   // Fix edge visibility
      SubdivideEdgeVisibility(iPrevious, iCurrent, Count);

    for(Count = 0; Count < Current.cVertices(); Count++)  // Fix vertex normals
      if(Current.aVertices(Count).cEdges() > 0)
        SubdivideVertexNormal(iCurrent, Count);

    for(Count = 0; Count < Current.cVertices(); Count++)// Dec vertex sharpness
      Current.aVertices(Count).DecSharpness();

    for(Count = 0; Count < Current.cEdges(); Count++)     // Dec edge sharpness
      Current.aEdges(Count).DecSharpness();
  }
}

void SubdivisionSurfaceXLoop::RethinkVertex(const int iPrevious,
                                       const int iCurrent,
                                       const int iVertex)
{
  const Hull& Previous = m_aLevels[iPrevious];
  Hull& Current = m_aLevels[iCurrent];
  const Vertex& VPrevious = Previous.aVertices(iVertex);
  Vertex& VCurrent = Current.aVertices(iVertex);
  int Count;

  VCurrent.SetVector(VPrevious.Vector());

  SubdivideVertex(iPrevious, iCurrent, iVertex);              // Rethink vertex

  Current.MarkVertexDirty(iVertex);                // Mark vertex at next level

  for(Count = 0; Count < VPrevious.cEdges(); Count++){    // Rethink neighbours
    int iEdge = VPrevious.aEdges(Count), iMidpoint;
    const Edge& E = Previous.aEdges(iEdge);

    if((iMidpoint = E.iMidpoint()) != -1){
      Vertex& Midpoint = Current.aVertices(iMidpoint);

      SubdivideEdgeVector(iPrevious, iEdge, Midpoint.Vector());

      Current.MarkVertexDirty(iMidpoint);     // Mark neighbourds at next level
    }
  }
}

void SubdivisionSurfaceXLoop::RethinkVertexNormal(const int iCurrent,
                                             const int iVertex)
{
  Hull& Current = m_aLevels[iCurrent];
  Vertex& V = Current.aVertices(iVertex);
  int* aSharp = (int *)alloca(V.cEdges() * sizeof(int));
  int* aSharpSpan = (int *)alloca(V.cEdges() * sizeof(int));
  int cSharp = 0;
  int Count;

  assert((aSharp != NULL) && (aSharpSpan != NULL));

  for(Count = 0; Count < V.cEdges(); Count++)
    aSharpSpan[Count] = 0;

  const int iPrimaryEdge = Current.VertexiPrimaryEdge(iVertex);
  int iCurr = iPrimaryEdge;
  int iPos = 0;

  do{
    const Edge& E = Current.aEdges(iCurr);
    int bRight = (E.iVertex(1) == iVertex);

    if(E.bSharp())
      aSharp[cSharp++] = iPos;

    if(cSharp > 0)                        // Keep track of the sharp span sizes
      aSharpSpan[cSharp - 1]++;

    iPos++;

    if(E.iPred(bRight) == iCurr) break; else iCurr = E.iPred(bRight);

  }while(iCurr != iPrimaryEdge);

  if(cSharp >= 2){                                              // Sharp vertex
    for(int iSpan = 0; iSpan < cSharp; iSpan++){
      const Edge& E = Current.aEdges(V.aEdges(aSharp[iSpan]));
      int m = aSharpSpan[iSpan];

      if(E.iFacet(E.iVertex(1) == iVertex) != -1)
        m++;            // Include sharp edge from next span if not at boundary

      if(m >= 2){
        Vector3D* N = Current.VertexNormal(iVertex, aSharp[iSpan], m);

        assert(N != NULL);

        SubdivideVertexNormalSharp(iCurrent, iVertex, aSharp[iSpan], m, *N);
      }
    }
  }
  else{                                                        // Smooth vertex
    Vector3D* N = Current.VertexNormal(iVertex);

    assert(N != NULL);

    SubdivideVertexNormalSmooth(iCurrent, iVertex, *N);
  }
}

void SubdivisionSurfaceXLoop::Rethink()
{
  if(m_bDirty){
    Hull& Control = m_aLevels[0];
    int iLevel, Count;

    m_bDirty = 0;

    for(Count = 0; Count < Control.cFacets(); Count++){
      Facet& F = Control.aFacets(Count);

      if(F.bDirty()){                      // Dirty facet; rethink facet normal

        Control.FacetNormalRethink(Count);

        F.SetClean();
      }
    }

    for(iLevel = 1; iLevel < m_aLevels.cElements(); iLevel++){
      Hull& Previous = m_aLevels[iLevel - 1];

      for(Count = 0; Count < Previous.cVertices(); Count++){
        Vertex& V = Previous.aVertices(Count);

        if(V.bDirty()){                                // Dirty vertex; rethink

          if(V.cEdges() > 0)
            RethinkVertex(iLevel - 1, iLevel, Count);

          V.SetClean();
        }
      }

      Hull& Current = m_aLevels[iLevel];

      for(Count = 0; Count < Current.cVertices(); Count++){
        Vertex& V = Current.aVertices(Count);

        if(V.bDirty()){                  // Dirty vertex; rethink vertex normal
          if(V.cEdges() > 0)
            RethinkVertexNormal(iLevel, Count);
        }
      }

      for(Count = 0; Count < Current.cFacets(); Count++){
        Facet& F = Current.aFacets(Count);

        if(F.bDirty()){                    // Dirty facet; rethink facet normal

          Current.FacetNormalRethink(Count);

          F.SetClean();
        }
      }
    }
  }
}

void SubdivisionSurfaceXLoop::Debug(FILE* fd) const
{
  int Count;

  fprintf(fd, "levels: %d\n", m_aLevels.cElements());

  for(Count = 0; Count < m_aLevels.cElements(); Count++){

    fprintf(fd, "%d: ", Count);

    m_aLevels[Count].Debug(fd);
  }
}
