/***********************************************************************
This software module was originally developed by Dr. Michael Steliaros,
Superscape (UK) Ltd., in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Superscape (UK) Ltd. retains full
right to use the code for his/her own purpose, assign or donate the code
to a third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2001
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Dr. Michael Steliaros at msteliaros@superscape.com
----------------------------------------------------------------------*/

////---------------------------------------------------------------------
////    Basic 3D vector class
////
////    Author:                 Michael K. Steliaros (Superscape)
////    Last modification date: 24-11-2000
////
////    Original author:        Paul C. Gregory (Superscape)
////    Creation date:          04-01-1996
////---------------------------------------------------------------------

#ifndef VECTOR3D_H
#define VECTOR3D_H

#include <stdio.h>
#include <assert.h>
#include <math.h>

#ifndef PI
#define PI 3.14159265358979323846f
#endif

class Vector3D
{
  public:
    Vector3D() : m_X(0.0f), m_Y(0.0f), m_Z(0.0f) { }
    Vector3D(const Vector3D& V) : m_X(V.m_X), m_Y(V.m_Y), m_Z(V.m_Z) { }
    Vector3D(float X, float Y, float Z) : m_X(X), m_Y(Y), m_Z(Z) { }
    ~Vector3D() { }
            
    float X() const { return(m_X); } 
    void X(const float X) { m_X = X; }
    float Y() const { return(m_Y); } 
    void Y(const float Y) { m_Y = Y; }
    float Z() const { return(m_Z); }
    void Z(const float Z) { m_Z = Z; }

    float& operator[](unsigned int i)               // Array access to elements
    { 
      switch(i){
        case 0 : return(m_X);
        case 1 : return(m_Y);
        case 2 : return(m_Z);
        default : return(m_Z);
      }
    }

    Vector3D& operator=(const Vector3D& V)                        // Assignment
      { m_X = V.m_X; m_Y = V.m_Y; m_Z = V.m_Z; return(*this); }

    Vector3D operator-() const                                      // Negation
      { return(Vector3D(-m_X, -m_Y, -m_Z)); }

    Vector3D operator+(const Vector3D& V) const                          // Sum
      { return(Vector3D(m_X + V.m_X, m_Y + V.m_Y, m_Z + V.m_Z)); }
    Vector3D& operator+=(const Vector3D& V)
      { m_X += V.m_X; m_Y += V.m_Y; m_Z += V.m_Z; return(*this); }

    Vector3D operator-(const Vector3D& V) const                   // Difference
      { return(Vector3D(m_X - V.m_X, m_Y - V.m_Y, m_Z - V.m_Z)); }
    Vector3D& operator-=(const Vector3D& V)
      { m_X -= V.m_X; m_Y -= V.m_Y; m_Z -= V.m_Z; return(*this); }

    float operator*(const Vector3D& V) const                     // Dot product
      { return(m_X * V.m_X + m_Y * V.m_Y + m_Z * V.m_Z); }

    Vector3D operator*(const float Scale) const                        // Scale
      { return(Vector3D(m_X * Scale, m_Y * Scale, m_Z * Scale)); }
    Vector3D operator/(const float Scale) const
      { return(operator*(1.0f / Scale)); }

    Vector3D& operator*=(const float Scale)                       // Scale self
      { m_X *= Scale; m_Y *= Scale; m_Z *= Scale; return(*this); }
    Vector3D& operator/=(const float Scale)
      { return(operator*=(1.0f / Scale)); } 

    Vector3D& operator*=(const Vector3D& Scale)              // Component scale
      { m_X *= Scale.m_X; m_Y *= Scale.m_Y; m_Z *= Scale.m_Z; return(*this); }

    float Magnitude2() const                                 // Vector length^2
      { return((m_X * m_X) + (m_Y * m_Y) + (m_Z * m_Z)); }
    float Magnitude() const                                    // Vector length
      { return(static_cast<float>(sqrt(Magnitude2()))); }

    void fabs()                                            // Make elements +ve
    {
      if(m_X < 0.0f) m_X = -m_X;
      if(m_Y < 0.0f) m_Y = -m_Y;
      if(m_Z < 0.0f) m_Z = -m_Z;
    }

    void Unit()                                             // Make unit length
    {
      float len = Magnitude();

      if(len == 0.0f){
        m_X = len;
        m_Y = len;
        m_Z = len;
      }
      else{
        len = 1.0f / len;
        m_X *= len;
        m_Y *= len;
        m_Z *= len;
      }
    }

    Vector3D operator%(const Vector3D& V) const                // Cross product
    {
      return(Vector3D((m_Y * V.m_Z) - (m_Z * V.m_Y),
                      (m_Z * V.m_X) - (m_X * V.m_Z),
                      (m_X * V.m_Y) - (m_Y * V.m_X)));
    }
    Vector3D& operator%=(const Vector3D& V)
    {
      Vector3D T(*this);

      m_X = (T.m_Y * V.m_Z) - (T.m_Z * V.m_Y);
      m_Y = (T.m_Z * V.m_X) - (T.m_X * V.m_Z);
      m_Z = (T.m_X * V.m_Y) - (T.m_Y * V.m_X);

      return(*this);
    }

    float operator^(const Vector3D& V) const                    // Vector angle
    {
      Vector3D T = *this % V;
      float x = *this * V;
      float y = T.Magnitude();
      int cx = 0, cy = 0;

      if(x > 0.0f) cx += 1; else if(x < 0.0f) cx += 2;
      if(y > 0.0f) cy += 1; else if(y < 0.0f) cy += 2;

      switch((cx << 2) | cy){
        case 1 :                                             // x == 0 && y > 0
          return(PI / 2.0f);
        case 2 :                                             // x == 0 && y < 0
          return(-PI / 2.0f);
        case 5 :                                             // x > 0  && y > 0
          return((y > x) ? PI / 2.0f - static_cast<float>(atan(x / y)) :
                           static_cast<float>(atan(y / x)));
        case 6 :                                             // x > 0  && y < 0
          return(-((-y > x) ? PI / 2.0f - static_cast<float>(atan(x / -y)) :
                              static_cast<float>(atan(-y / x))));
        case 8 :                                            // x < 0  && y == 0
          return(PI);
        case 9 :                                             // x < 0  && y > 0
          return(PI - ((y > -x) ? PI / 2.0f - static_cast<float>(atan(-x / y)) :
                                  static_cast<float>(atan(y / -x))));
        case 10 :                                            // x < 0  && y < 0
          return(((y < x) ? PI / 2.0f - static_cast<float>(atan(x / y)) :
                            static_cast<float>(atan(y / x)) - PI));
        default :
          return(0.0f);
      }
    }

    int operator==(const Vector3D& Cmp) const                // Vector equality
      { return((m_X == Cmp.m_X) && (m_Y == Cmp.m_Y) && (m_Z == Cmp.m_Z)); }

    int operator!=(const Vector3D& Cmp) const              // Vector inequality
      { return((m_X != Cmp.m_X) || (m_Y != Cmp.m_Y) || (m_Z != Cmp.m_Z)); }

    void Debug(FILE* fd)
      const { fprintf(fd, "[%.3f, %.3f, %.3f]\n", m_X, m_Y, m_Z); }
      
  protected:
    float m_X;                                               // Vector elements
    float m_Y;
    float m_Z;
};

#endif // VECTOR3D_H
