#include "subwrap.h"

#include "compat.h"
#include <set>

#include "tagflatmesh.h"
#include "vertex.h"
#include "subtri.h"
#include "subquad.h"
#include "tagmesh.h"

// TODO:
// - split faces
// - texture coordinates
// - tags, normals, etc.

static void dataToFlatMesh(const SurfaceData& data, bool split, TagFlatMesh& fm);

static void addTags(const TagData& data, TagFlatMesh& fm);

static bool
toFlatMesh(FlatMesh* im,
           int numCoords, float* coords, float* params,
           int numIndices, int* indices,
           bool split);



template<class Mesh>
void count(Mesh* m, int level, int& noVertices, int& noFaces);

template<class Mesh>
void getFacesFunc(Mesh* m, int level, int* f);

template<class Mesh>
void getVertexDataFunc(Mesh* m, int level, float* pos, float* normal, float* param);

//-------------------------------------------------------
//    SubWrap(int l, const SurfaceData& SurfaceData);
SubWrap::SubWrap(int l, 
                 int loopFlag,
                 const SurfaceData& data) : _tm(0), _qm(0), _fm(0)
{
    // steps:
    // 1. create a flat mesh
    // 2. create a mesh (from the flat mesh)
    // 3. subdivide the mesh
    
    // create a flat mesh
    _fm = new TagFlatMesh();
    dataToFlatMesh(data, (loopFlag==1) ? true : false, *_fm);
    
    // build a mesh
    if(loopFlag==1) 
    {
        _tm = new TagMeshTp<Tri>(*_fm);
        _tm->subdivide(l);
    } 
    else
    {
        _qm = new TagMeshTp<Quad>(*_fm);
        _qm->subdivide(l);
    }    
}

//-------------------------------------------------------
//    SubWrap(int l, const SurfaceData& SurfaceData);
SubWrap::SubWrap(int l, 
                 int loopFlag,
                 const SurfaceData& data,
                 const TagData& tagData) : _tm(0), _qm(0), _fm(0)
{
    // steps:
    // 1. create a flat mesh
    // 2. create a mesh (from the flat mesh)
    // 3. subdivide the mesh
    
    // create a flat mesh
    _fm = new TagFlatMesh();
    dataToFlatMesh(data, (loopFlag==1) ? true : false, *_fm);
    addTags(tagData, *_fm);

    // build a mesh
    if(loopFlag==1) 
    {
        _tm = new TagMeshTp<Tri>(*_fm);
        _tm->subdivide(l);
    } 
    else
    {
        _qm = new TagMeshTp<Quad>(*_fm);
        _qm->subdivide(l);
    }    
}


//-------------------------------------------------------
//    ~SubWrap();
SubWrap::~SubWrap()
{
    // release memory
    if(_tm)
        delete _tm;
    if(_qm)
        delete _qm;
    if(_fm)
        delete _fm;
}

//-------------------------------------------------------
void SubWrap::update(int l,
                     const SurfaceData& surfaceData,
            const TagData& tagData)
{
    // make sure we have the correct numbers
    assert(_fm);
    assert(surfaceData.noVtx == _fm->vert_v.size());
    assert(tagData.noSector == _fm->sectorInfoVec.size());

    // update vertices
    float* params = surfaceData.param;
    float* coords = surfaceData.pos;
    int i;
    for(i = 0; i < _fm->vert_v.size(); ++i)
    {
        Vertex* vp = _fm->vert_v[i];
        if(params)
            vp->setPos(cvec5f(coords[3*i+0], coords[3*i+1], coords[3*i+2], params[2*i+0], params[2*i+1]));
        else
            vp->setPos(cvec5f(coords[3*i+0], coords[3*i+1], coords[3*i+2], 0, 0));
    }

    // update tags
    //::ofstream os("updatelog.txt");
    for(i = 0; i < _fm->sectorInfoVec.size(); ++i)
    {
        const TagData& d = tagData;
        SectorInfo* si = _fm->sectorInfoVec[i].second;
        si->setSectorTag((SectorInfo::SectorTagType)d.sectors[i].tag);
        si->setTheta(d.sectors[i].theta);
        si->setModifiedFlatness(d.sectors[i].flatness);
        si->setModifiedNormal(cvec3f(d.sectors[i].normal[0], d.sectors[i].normal[1], d.sectors[i].normal[2]));
        si->setModifiedNormalT(d.sectors[i].normalTension);
    }

    // update surface
    if(_tm)
    {
        _tm->subdivide(l);
    } 
    else if(_qm)
    {
        _qm->subdivide(l);
    }    
}

// number of faces and vertices
//-------------------------------------------------------
//    void getNumbers(int level, int& noVertices, int& noFaces) const;
void SubWrap::getNumbers(int level, int& noVertices, int& noFaces) const
{
    // count faces and vertices on level l
    //   number vertices on level l
    if(_tm != 0)
        count(_tm, level, noVertices, noFaces);
    if(_qm != 0)
        count(_qm, level, noVertices, noFaces);
}


//-------------------------------------------------------
//    void getFaces(int level, 
void SubWrap::getFaces(int level, int* f) const 
{
    // get faces on level l
    if(_tm != 0)
        getFacesFunc(_tm, level, f);
    if(_qm != 0)
        getFacesFunc(_qm, level, f);
}


//-------------------------------------------------------
//    void getVertexData(int level, float* pos, float* normal, float* param) const;
void SubWrap::getVertexData(int level, float* pos, float* normal, float* param) const
{
    // get vertices on level l
    //    positions
    //    normals
    //    parameter values
    if(_tm != 0)
        getVertexDataFunc(_tm, level, pos, normal, param);
    if(_qm != 0)
        getVertexDataFunc(_qm, level, pos, normal, param);
}

//-------------------------------------------------------
void SubInfo::showInfo(::ostream& os)
{
    os<<"BaseVertex::cnt(): "<<BaseVertex::cnt()<<
        ", Tri::cnt(): "<<Tri::cnt()<<
        ", Quad::cnt(): "<<Quad::cnt()<<
        ", SectorInfo::cnt(): "<<SectorInfo::cnt()
        <<endl;
}

//-------------------------------------------------------
void dataToFlatMesh(const SurfaceData& data, bool split, TagFlatMesh& im)
{
    toFlatMesh(&im, data.noVtx, data.pos, data.param, data.noFaces, data.face, split);
}

//-------------------------------------------------------
void addTags(const TagData& d, TagFlatMesh& im)
{
    int i;
    // creaseLineIndex
    for(i = 0; i < d.noCreaseLineIndex; ++i)
        im.creaseEdgeVec.push_back(d.creaseLineIndex[i]);
    // cornerIndex
    for(i = 0; i < d.noCornerIndex; ++i)
        im.cornerVertVec.push_back(d.cornerIndex[i]);
    // creaseIndex
    for(i = 0; i < d.noCreaseIndex; ++i)
        im.creaseVertVec.push_back(d.creaseIndex[i]);
    // dartIndex
    for(i = 0; i < d.noDartIndex; ++i)
        im.dartVertVec.push_back(d.dartIndex[i]);
    // sectors
    for(i = 0; i < d.noSector; ++i)
    {        
        pair<int, int> ip(d.sectors[i].faceIndex, d.sectors[i].vertexIndex);
        SectorInfo* si = new SectorInfo();
        si->setSectorTag((SectorInfo::SectorTagType)d.sectors[i].tag);
        si->setTheta(d.sectors[i].theta);
        si->setModifiedFlatness(d.sectors[i].flatness);
        si->setModifiedNormal(cvec3f(d.sectors[i].normal[0], d.sectors[i].normal[1], d.sectors[i].normal[2]));
        si->setModifiedNormalT(d.sectors[i].normalTension);
        SectorInfo::ref(si);

        im.sectorInfoVec.push_back(pair<pair<int, int>, SectorInfo*>(ip, si));

        typedef pair<pair<int, int>, SectorInfo*> FlatSectorType;
        vector<FlatSectorType> sectorInfoVec;

    }
}


//-------------------------------------------------------
template<class Mesh>
void count(Mesh* m, int level, int& noVertices, int& noFaces)
{
    noFaces = 0;
    noVertices = 0;
    
    // clear indices
    typename Mesh::FaceIterType it;
    int v;
    for(it = m->faceBegin(level); it != m->faceEnd(level); ++it)
    {
        if(it.depth() == level) 
        {
            for(v = 0; v < (*it)->noVtx(); ++v)
            {
                (*it)->index(v) = -1;
            }
        }
    }
    
    // enumerate all vertices
    for(it = m->faceBegin(level); it != m->faceEnd(level); ++it)
    {
        if(it.depth() == level) 
        {
            ++noFaces;
            for(v = 0; v < (*it)->noVtx(); ++v)
            {
                if((*it)->index(v) == -1)
                    (*it)->index(v) = noVertices++;
            }
        }
    }
}


//-------------------------------------------------------
template<class Mesh>
void getFacesFunc(Mesh* m, int level, int* f)
{
    // enumerate all vertices
    int t = 0;
    
    // clear flags
    typename Mesh::FaceIterType it;
    for(it = m->faceBegin(level); it != m->faceEnd(level); ++it)
    {
        if(it.depth() == level) 
        {
            // number of vertices
            f[t++] = (*it)->noVtx();
            // vertex indices
            int v;
            if(!(*it)->orientation())
            {
                for(v = 0; v < (*it)->noVtx(); ++v)
                    f[t++] = (*it)->index((*it)->noVtx()-1-v);
            }
            else
            {
                for(v = 0; v < (*it)->noVtx(); ++v)
                    f[t++] = (*it)->index(v);
            }
        }
    }
}


//-------------------------------------------------------
template<class Mesh>
void getVertexDataFunc(Mesh* m, int level, float* pos, float* normal, float* param)
{
    // enumerate all vertices
    int t = 0;
    int cnt = 0;
    
    // enumerate all vertices
    typename Mesh::FaceIterType it;
    for(it = m->faceBegin(level); it != m->faceEnd(level); ++it)
    {
        if(it.depth() == level) 
        {
            // vertex indices
            int v;
            for(v = 0; v < (*it)->noVtx(); ++v)
                if((*it)->index(v) == cnt)
                {
                    if(pos!=0)
                    {
                        pos[3*t+0] = (*it)->vert(v)->getPos(level).x();
                        pos[3*t+1] = (*it)->vert(v)->getPos(level).y();
                        pos[3*t+2] = (*it)->vert(v)->getPos(level).z();
                    }
                    //
                    if(normal!=0)
                    {
                        normal[3*t+0] = (*it)->normal(v).x();
                        normal[3*t+1] = (*it)->normal(v).y();
                        normal[3*t+2] = (*it)->normal(v).z();
                    }
                    //
                    if(param!=0)
                    {
                        param[3*t+0] = (*it)->vert(v)->getPos(level).u();
                        param[3*t+1] = (*it)->vert(v)->getPos(level).v();
                        param[3*t+2] = 0;
                    }
                    ++cnt;
                    ++t;
                }
        }
    }
}

//-------------------------------------------------------
// 
bool
toFlatMesh(FlatMesh* im,
           int numCoords, float* coords, float* params,
           int numIndices, int* indices,
           bool split)
{
    
    
    set<uint> vset;
    pair< set<uint>::iterator,bool> res;
    
    int i; 
    for( i = 0; i < numCoords; i++ ) {
        Vertex* vp =  new Vertex();
        if(params)
            vp->setPos(cvec5f(coords[3*i+0], coords[3*i+1], coords[3*i+2], params[2*i+0], params[2*i+1]));
        else
            vp->setPos(cvec5f(coords[3*i+0], coords[3*i+1], coords[3*i+2], 0, 0));
        Vertex::ref(vp);
        im->vert_v.push_back( vp );
    }
    
    i = 0;
    if( numIndices == 0 ) { im->Cleanup(); return false; }
    
    if(!split) { 
        while( i < numIndices ) {
            vset.erase(vset.begin(), vset.end());
            if( im->poly_v.size() > 0 && im->poly_v[im->poly_v.size()-1].novtx() < 3) {
                // skip polygons with < 3 vertices
                cerr << "skipping polygon with " << 
                    im->poly_v[im->poly_v.size()-1].novtx() << " vertices "; 
                for(uint l = 0; l < im->poly_v[im->poly_v.size()-1].novtx(); l++ )
                    cerr << im->index_v[im->poly_v[im->poly_v.size()-1].start()+l] << " ";
                cerr << endl;
                im->poly_v[im->poly_v.size()-1] = IPoly(im->index_v.size(),0 );
                im->triindex_v.push_back(-1);
            } else { 
                im->poly_v.push_back( IPoly(im->index_v.size(),0 ) );
                im->triindex_v.push_back(im->poly_v.size()-1);
            }
            while( indices[i] != -1 ) { 	
                res = vset.insert(indices[i]);
                // skip duplicate vertices
                if(res.second) { 
                    im->poly_v[im->poly_v.size()-1].incnovtx();
                    im->index_v.push_back(indices[i++]);       
                } else { 
                    cerr << "\t\t skipping duplicate vertex " << endl;
                    i++;
                }
            } 
            i++; // skip -1
        }
        // remove the last polygon if its size is zero 
        assert(im->poly_v.size() > 0 );
        if( im->poly_v.back().novtx() == 0 ) { 
            if( im->poly_v.size() == 1) { im->Cleanup(); return false;}
            im->poly_v.erase(im->poly_v.end());
        }
    } else { // split all polygons into triangles
        int first_vertex, second_vertex, third_vertex;
        while( i < numIndices ) {
            while( i < numIndices && indices[i] == -1 ) 
                i++;
            first_vertex = indices[i];
            i++;      
            while(  i < numIndices && indices[i] != -1 && indices[i] == 
                first_vertex) i++;
            if(i < numIndices && indices[i] != -1)
                second_vertex = indices[i];
            else continue;
            i++;
            while( i < numIndices && indices[i] != -1  && 
                (indices[i] == second_vertex || indices[i] == first_vertex) )
                i++;
            if(i < numIndices && indices[i] != -1)
                third_vertex = indices[i];
            else continue;
            assert( i < numIndices && indices[i] != -1 && 
                first_vertex < numIndices && 
                second_vertex < numIndices );
            i++;
            im->poly_v.push_back(IPoly(im->index_v.size(),3));      
            im->triindex_v.push_back(im->poly_v.size()-1);
            im->index_v.push_back(first_vertex); 
            im->index_v.push_back(second_vertex);
            im->index_v.push_back(third_vertex);
            while( i < numIndices && indices[i] != -1) {
                second_vertex = third_vertex; 
                while( i < numIndices && indices[i] != -1 && 
                    (indices[i] == first_vertex || indices[i] == second_vertex))
                    i++;
                if(i < numIndices && indices[i] != -1)
                    third_vertex = indices[i];
                else break;
                i++;
                im->poly_v.push_back(IPoly(im->index_v.size(),3));      
                im->index_v.push_back(first_vertex); 
                im->index_v.push_back(second_vertex);
                im->index_v.push_back(third_vertex);
                
            }
            assert(indices[i] == -1);
        }
    }
    
    return true;
}

//-------------------------------------------------------
//
template<class Mesh> bool isOpenFunc(Mesh* m);

bool SubWrap::isOpenMesh() const
{
    if(_tm)
        return isOpenFunc(_tm);
    else if(_qm)
        return isOpenFunc(_qm);
    else
        return 0;
}

template<class Mesh>
bool isOpenFunc(Mesh* m)
{
    bool io = false;
    typename Mesh::FaceIterType it;
    for(it = m->faceBegin(0); it != m->faceEnd(0); ++it)
    {
        VnoType v;
        for(v = 0; v < (*it)->noVtx(); ++v)
        {
            if((*it)->vertexTag(v) != Mesh::FaceType::NOTAG_VERTEX)
                io = true;
        }
    }
    return io;
}
    
