// -*- c++ -*-

/***********************************************************************
This software module was originally developed by Riitta Vaananen
(Helsinki University of Technology) in the course of development of
the MPEG-4 Systems (ISO/IEC 14496-1) standard.  This software module
is an implementation of a part of one or more MPEG-4 Systems (ISO/IEC
14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1)
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1)
free license to this software module or modifications thereof for use
in hardware or software products claiming conformance to the MPEG-4
Systems (ISO/IEC 14496-1). Those intending to use this software module
in hardware or software products are advised that its use may infringe
existing patents.  The original developer of this software module and
his/her company, the subsequent editors and their companies, and
ISO/IEC have no liability for use of this software module or
modifications thereof in an implementation.  Copyright is not released
for non MPEG-4 Systems (ISO/IEC 14496- 1) conforming products.
Helsinki University of Technology retains full right to use the code
for their own purpose, assign or donate the code to a third party and
to inhibit third parties from using the code for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. This copyright notice must be
included in all copies or derivative works. Copyright  1999.
************************************************************************/


#ifndef __ACOUSTICSCENEPROXY_H__
#define __ACOUSTICSCENEPROXY_H__

#include "OpenGl/group/GroupProxy.h"

// SB %%%% 99/05/10
#include "opengl/AdvancedAudio/AudioProcess/ProcessSound.h"

class AcousticScene;
class AcousticSurface;
class Auralize;

/** AcousticScene renderer node.

    Declared in opengl/AdvancedAudio/AcousticSceneProxy.h*/
class AcousticSceneProxy
  : public Node3DProxy
{
	friend class ProcessDirectiveSound;

public:

  /// Constructor

                    AcousticSceneProxy(AcousticScene* pAcousticScene);

					~AcousticSceneProxy();

  // Render object and its children

  virtual BOOL		Render(const Matrix& rTransformation, const CullFlag& rCullFlag,
			       OpenGLVisualRenderer* pRenderer, DWORD dwTime);

 /// Render object description

  virtual void		RenderDescription(OpenGLVisualRenderer* pRenderer, DWORD dwTime);

  /** Fecthes all the AcousticScene nodes in the scene graph

	Adds a pointer to corresponding media object to an array 
	of AcousticScenes, adds a transformation matrix of concatenated 
	transforms above this AcousticScene to an array of transformation
	matrices.
  */
	virtual int		FindAcousticScenes(
						ZArray<AcousticScene*>& rAcousticScenes, 
						ZArray<Matrix>& rASceneTransformations,
						const Matrix& rTransform);

	/** Looks for Acoustic shapes
	
	Recursively looks for IndexedFaceSets in the sibling 
    sub trees of the corresponding media object (node). */
	void FindAcousticShapes(Group* pGroup);
	
	/** Recursively looks for IndexedFaceSets in the sibling 
      sub trees of the corresponding media object (node). */
	inline AcousticScene * GetAcousticScene()
		{ return (AcousticScene *) GetMediaObject(); }

	 /** Builds up an array of acoustic polygons 
	 
	   Builds up an array of acoustic polygons (by calling 
      MakeAcousticSurfaceArray()) first time a sound stream 
      is attached to a scene. After that the array of 
      polygons is updated  at each scene update.*/
	void MakeAcousticPolygons();


	void AddToAcousticSurfaceArray(Shape* acoustic_shape);

	 /** Builds up the array of acoustic surfaces.
	 
	   Actually builds up the array of surfaces (IndexedFaceSets with
      AcousticMaterial), or updates it if this is an update.*/	 
	void MakeAcousticSurfaceArray(BOOL is_update);

	/** Looks for transformations in parents of AcousticScene
	 
	   Recursively looks for the transformations in the 
      parent nodes of this node and stores the Transform nodes in
      AS_transf. */
	void FindTransformsInParents(MediaObject* AS);

	  /** Updates the array of acoustic polygons
	  
		Looks for the new transformations in the parents, finds the acoustic shapes 
      and calls MakeAcousticSurfaceArray again. */
	void UpdateAcousticPolygons();


	Matrix GetASTransform(void);

	 /** polygons_done
	 
	   Used in MakeAcousticPolygons to indicate whether this 
      is an update. */
	BOOL polygons_done;

	  /** acousticshapes
	  
		Array of Shape nodes (in the sibling sub trees of 
      the MediaObject corresponding to this object) that 
      have IndexedFaceSet node in its geometry field and 
      AcousticMaterial in its appearance field. */
	ZArray<Shape*> acousticshapes;

	  /** trAShape
	  
		Array of transform Matrices that define the concatenated 
      transformations above the shape nodes in the acousticshapes
      array. */
	ZArray<Matrix> trAShape;
//	ZArray<Transform*> transforms;

	 /** transforms_AS
	 
	   Array of transformations above the corresponding media object
      (AcousticScene node).  */
	ZArray<Transform*> transforms_AS;

 /** AS_transf
 
   Concatenated transformation matrix above the corresponding 
      media object (AcousticScene node). */
	Matrix AS_transf;

	 /** a_surfaces
	 
	   Array of AcousticSurface objects that include all the 
      relevant information about the IndexedFaceSets that have 
      AcousticMaterial attached to them. */
	ZArray<AcousticSurface*> a_surfaces;
//	ZArray<int*> ashape_IDs;
//	ZArray<int*> transform_IDs;


	// riitta 03082000
//	ZArray<Transform*> transforms_AS_old;
//	ZArray<Shape*> acousticshapes_old;
	ZArray<Matrix> trAShape_old;
	MF<Bool> transform_changes_AShape;
//	Bool transform_changes_AScene;

};

/** class for handling one AcousticSurface.

  Handles all the information in an IndexedFaceSet 
    object that has some acoustical relevance (AcousticMaterial 
    attached to it). 

    Declared in opengl/AdvancedAudio/AcousticSceneProxy.h*/
class 
AcousticSurface
{
	friend AcousticSceneProxy;
	friend ProcessDirectiveSound;
	friend Auralize;
	friend ImageSource;
	friend SoundSource;

public:

	/// Constructor
	AcousticSurface::AcousticSurface(AcousticMaterial* a_mat, Vec3f* poly_coords, int nof_v);

	/** Destructor.
	Deletes filter coefficient arrays.*/
	~AcousticSurface();

private:

	/** Puts filter coefficients to arrays.

	And sets the number of coefficients variable*/
	void SetFilters(AcousticMaterial* a_mat);

	/** Computes normals for a polygon*/
	void SetNormals(Vec3f* coor);

	/** Updates surface if it has changed. */
	void UpdateSurface(/*AcousticMaterial* a_mat,*/ Vec3f* poly_coords, int nof_v);	
	
	/** vertices of the polygon */
	Vec3f* vertices;
	int nof_vert;

	/** Number of reflection filter coefficients */
	int nof_refcoeffs;

	/** Number of transmission filter coefficients. */
	int nof_transcoeffs;


	/** a coefficients of reflectivity filter*/
	float* ref_a;

	/** b coefficients of reflectivity filter*/
	float* ref_b;

	/** a coefficients of transmission filter*/
	float* trans_a;

	/** b coefficients of transmission filter*/
	float* trans_b;

	/** Normal of the polygon*/
	Vec3f normal;

	/** A point on the surface*/
	Vec3f surface_point;

	// riitta 03072000:

	int num_ref_freq;
	int num_trans_freq;
	int num_ref_param;
	int num_trans_param;
	float* ref_freq;
	float* trans_freq;
	float* ref_param;
	float* trans_param;
	Bool filter_form_ref;
	Bool filter_form_trans;

};



#include "AcousticSceneProxy.i"

#endif // __ACOUSTICSCENEPROXY_H__

