#ifndef __AURALIZE_H__
#define __AURALIZE_H__
/***********************************************************************
This software module was originally developed by Riitta Vaananen
(Helsinki University of Technology) in the course of development of
the MPEG-4 Systems (ISO/IEC 14496-1) standard.  This software module
is an implementation of a part of one or more MPEG-4 Systems (ISO/IEC
14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1)
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1)
free license to this software module or modifications thereof for use
in hardware or software products claiming conformance to the MPEG-4
Systems (ISO/IEC 14496-1). Those intending to use this software module
in hardware or software products are advised that its use may infringe
existing patents.  The original developer of this software module and
his/her company, the subsequent editors and their companies, and
ISO/IEC have no liability for use of this software module or
modifications thereof in an implementation.  Copyright is not released
for non MPEG-4 Systems (ISO/IEC 14496- 1) conforming products.
Helsinki University of Technology retains full right to use the code
for their own purpose, assign or donate the code to a third party and
to inhibit third parties from using the code for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. This copyright notice must be
included in all copies or derivative works. Copyright  1999.
************************************************************************/


//#define MAXFILENAMELEN 256
#define NOF_AIRCOEFFS ((int)180)
#define MAX_NOF_REF ((int)100)

#include <stdlib.h>
#include <stdio.h>
#include <fstream.h>
#include <math.h>
#include <string.h>
#include "Include/defs.h"
#include "core/ztypes.h"
#include "opengl/AdvancedAudio/AudioProcess/Reverb.h"

#include "opengl/AdvancedAudio/AudioProcess/Spatialize.h"
class ImageSource;
class SoundSource;
class DirectiveSound;
class AcousticScene;

// riitta 03032000
class IIRFilt;


/** class for equalization filters*/ // riitta 03032000

class EQFilters
{

public:
	/// Constructor

	EQFilters(float* g, float* f, int fs, int n);

	/// Destructor
	~EQFilters();

	void EmptyFilters();
	
	void ComputeCoefficients(float* f, float* g, int n);

	float operator()(float input);

	ZArray<IIRFilt*> filters;

	
private:
	int nof_filters;

};

/** class for sound directivity filtering */

class SoundDirectivity
{
public:

	/** Class Constructor. */	
	SoundDirectivity(int sf, 
		float viewp_from_presenter[4], 
		Vec3f sound_loc_from_presenter, 
		Vec3f sound_dir_from_presenter, 
		DirectiveSound* DS,
		int num_param, 
		int nof_angles);

	/// Destructor

	~SoundDirectivity();

	/** Operator; Processes each sound sample.
	*/

	float operator()(float input);

	/** Updates the filter coefficients.*/

	void UpdateCoeffs(float viewp_from_presenter[4], 
		Vec3f sound_loc_from_presenter, 
		Vec3f sound_dir_from_presenter);

	/** Interpolates the filter coefficients.
	*/

	void InterpolateCoeffs();

	// riitta 03062000; added
	void InterpolateGain();

	/// Set the (linear) interpolation gain used in the coefficient interpolation

    void SetInterpAndOffset();

private:
	float* coeffs_b;
	float* coeffs_a; 
	float* DL_filter;
	float* angles;
	float* curr_coeffs_b;
	float* curr_coeffs_a;
	float GetAngle(float viewp[4], Vec3f sound_loc, Vec3f sound_dir);
	int pDL;
	int nof_coeff;
	int no_angles;
	float angle;
	int offset_index;
	float interp_gain;

	// riitta 03032000
	IIRFilt* dir_filt;

	IIRFilt* filt_1;
	IIRFilt* filt_2;
	ZArray<IIRFilt*> filt_arr;

	Bool filter_form;

	ZArray<EQFilters*> eq_array; 

	//riitta 03072000:
	int no_param;
};

/** class for generic IIR filtering.
*/

class IIRFilt
{
public:
	/** Constructor.
	
	Initializes the filter, sets up filter coefficients and the order of the filter
	*/
	
	IIRFilt(float* coeff_b, float* coeff_a, int order);

	/// Destructor
	~IIRFilt();

	/** Operator.

	Processes the given input sample by the IIR filter with the current coefficients.	
	*/

	float operator()(float input);
	
	/** Gives new values to the filter coefficients.
	
	This is done at the update rate of the auralization process.
	*/
	void UpdateCoeffs(float* coeff_b, float* coeff_a);

	void EmptyFilter();

private:
	float* coeffs_a;
	float* coeffs_b;
	int order;
	int no_coeffs;
	float* DL_IIR;
	int pDL;



};

/** class for auralization processing. */

class Auralize
{
	friend class SoundSource;
public:

	/** Constructor. */

	Auralize(int sf, int update,
		float viewp[4],
		DirectiveSound* DS,
		Vec3f sound_dir,
		AcousticScene* pAScene,
		ZArray<ImageSource*>& imagesources,
		SoundSource* source);
	
	/** class Destructor.
		
	delay lines and other arrays with arbitrary lengh are deleted
	*/
	~Auralize();

	/** Operator for sound processing of each input sample.	*/

	float* operator()(float input);

	/** Updates all the parameters involved in the auralization process.
	*/
	
	void UpdateAll(float[4],
		Vec3f sound_dir,
		AcousticScene* pAScene, 
		ZArray<ImageSource*>& reflections, 
		SoundSource* source);

			    
	/**	Sets up the air absorption coefficient table when needed.
	*/
	void SetAirabsCoeffs(float air_coeffs[NOF_AIRCOEFFS]);
		  
private:
	ofstream* os;
	int update_period;

	/** Reverebrator.
	
	Pointer to a LateReverb object that is used for reverberating the direct sound.
	*/
	LateReverb* reverb;

	/** Function for setting up and updating parameters related to reflections. 
	*/
  
	void ReflectionParameters(ZArray<ImageSource*>& /* ImageSource** */ ref, 
		float viewp[4]);

	/** Distance computation.
	
	  Used for both direct sound and reflections
	*/
	void GetDistance(float viewp[4], 
		Vec3f s_loc, 
		float* d_samp_old, 
		float* d_samp, 
		float* curr_dist);
	/** Get current air absorption coefficients.
	
	  For both direct sound and early reflections */

	void GetAirabsCoeffs(float curr_dist, 
		float* air_b, 
		float* air_a); // Sets air absorption current coefficients
	
	/** Distance attenuation computation.

	Used for computing distances of direct sound and image sources. */
	void GetDistAtt(float curr_dist, 
		float* dist_gain, 
		float *dist_gain_old);      //computes distance dependent attenuation

	/** Air absorption object.

	For direct sound; implemented with IIRFilt class
	*/
	IIRFilt* airabs;

	/** Sound directivity object.

	For direct sound; implemented with SoundDirectivity class.
	*/
	SoundDirectivity* dir;


	AmPan* pan; // direct sound

	/** pointer to an object for direction of arrival filtering.
	
	For 3D spatialization of direct sound; implemented by amplitude panning, 
	so actually ralizes only 2D spatialization.
	*/
	PanITD* panitd;

	int ref_len;

	/// Array of pointers to spatialization objects of reflections.

	ZArray<PanITD*> ref_panitd;

	/// Array of pointers to objects for air absorption filtering of reflections.
	
	ZArray<IIRFilt*> ref_airabs;

	/// Array of pointers to objects for air reflectivity filtering of each reflection.

	ZArray<IIRFilt*> ref_filt;

	// riitta 03072000
	ZArray<EQFilters*> ref_eq;

	SoundSource* soundsource;

	ZArray<IIRFilt*> transfilts;

	float out[2];
	float* DL_dist;
	float airabscoeffs[180];
	float* air_b;
	float* air_a;
	float att_dist; 
	float curr_dist;
	float speedOfSound; 
	int samp_freq;
	Bool spatialize;
	Bool roomEffect;
	int is_directive; 
	int max_dist_delay;
	int count;
	float dist_samp;
	float dist_samp_old;
	float dist_gain;
	float dist_gain_old;
	float* dist_samp_ref;
	float* dist_samp_ref_old;
	float* dist_gain_ref;
	float* dist_gain_ref_old;
	BOOL* vis_ref;

	BOOL reflections_initialized;
	int nof_ref;
	int pDL;
	int nof_images; 
	int nof_angles;
	Bool useairabs;

	// riitta 03112000
	Bool dist_delay_rendering;
	Bool dist_gain_rendering;
};

#endif
