#ifndef _PROCESSSOUND_H_
#define _PROCESSSOUND_H_
/***********************************************************************
This software module was originally developed by Riitta Vaananen
(Helsinki University of Technology) in the course of development of
the MPEG-4 Systems (ISO/IEC 14496-1) standard.  This software module
is an implementation of a part of one or more MPEG-4 Systems (ISO/IEC
14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1)
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1)
free license to this software module or modifications thereof for use
in hardware or software products claiming conformance to the MPEG-4
Systems (ISO/IEC 14496-1). Those intending to use this software module
in hardware or software products are advised that its use may infringe
existing patents.  The original developer of this software module and
his/her company, the subsequent editors and their companies, and
ISO/IEC have no liability for use of this software module or
modifications thereof in an implementation.  Copyright is not released
for non MPEG-4 Systems (ISO/IEC 14496- 1) conforming products.
Helsinki University of Technology retains full right to use the code
for their own purpose, assign or donate the code to a third party and
to inhibit third parties from using the code for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. This copyright notice must be
included in all copies or derivative works. Copyright  1999.
************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <fstream.h>
#include <math.h>
#include <string.h>

class Presenter;

#include "opengl/audio/SoundProxy.h"
#include "opengl/AdvancedAudio/AudioProcess/Spatialize.h" 

#ifdef ADV_AUDIO

// SOLARIS PORTING - Case typos
#include "opengl/AdvancedAudio/DirectiveSoundProxy.h"
#include "opengl/AdvancedAudio/AudioProcess/Auralize.h" 
#include "opengl/AdvancedAudio/AudioProcess/Reverb.h"
#include "opengl/AdvancedAudio/spatc7/Spatializer.h"
#include "opengl/AdvancedAudio/spatc7/Spatparameters.h"

class Auralize;

class AcousticScene;
class ImageSource;
class AcousticSurface;
//class IIRFiltParam;
class IIRFilt;

#endif // ADV_AUDIO

#ifdef ADV_AUDIO

/** class for holding visibility information of a sound reflection
(image source) and pointer to an acoustic surface with respect to
which this image source is computed */
class ImageSource
{

	friend class Auralize;
	friend class ProcessDirectiveSound;

public:
	
	 ImageSource(){};

	 /** Constructor.

	sets up image source location information, pointer to the acoustic surface,
	off which this image source generates a reflection, and visibility*/
	ImageSource(Vec3f loc, int ord, AcousticSurface* surf);

	/** Destructor. */
	~ImageSource();

	/** Sets the value of the visibility flag of this image source. */
	void SetVisibility(float viewp[4], ZArray<AcousticSurface*>&
		a_s, Vec3f s_l, int nofsurf);

	/** Computes the visibility.*/
	BOOL Visibility(float viewp[4], Vec3f loc, AcousticSurface* surf, Vec3f* y);
	


private:
	Vec3f location;
	int order;
	BOOL visibility;
	AcousticSurface* a_surf;

};




/** class for direct sound information. */
/* Stores the information about the location and visibility of the sound source,
and holds information the acoustic surfaces that are currently 
between the source and the listener, and stores the corresponding 
transmission filters to and array.*/
class SoundSource : public ImageSource
{
	friend class Auralize;
	friend class ProcessDirectiveSound;
	friend class IIRFilt;

public:
	/** Constructor.

	Sets up the transmission filter coefficient, visibility, and source location information.*/
	SoundSource(Vec3f loc, float sfreq);

	/** Destructor.

	Deletes the filter coefficient arrays, if they are != NULL.*/
	~SoundSource();

	/** Checks visibility of sources.
	
	  Checks whether this source is currently visible, or if there are
	acoustically relevant surfaces between the source and the listener */
	virtual void SetVisibility(float viewp[4],  
		ZArray<AcousticSurface*>&
		a_s, 
		Vec3f l, 
		int nofsurf);

private:
	Vec3f location;
	BOOL visibility;
	int order;
	float sf;

	/** array of IIRFilters 
	
	  Transmission of IIRFilters for the surfaces that are 
	currently between the source and the listener
	*/
	ZArray<IIRFilt*> iirfilts;

	// riitta 03072000
	ZArray<EQFilters*> eqfilts;

	// array of acoustic surfaces this source is obstructed by
	ZArray<AcousticSurface*> asurf;
};

#endif // ADV_AUDIO

/** class for processing of distance dependent gain.*/
class EllipseSound
{
public:

	/** Constructor. */
	EllipseSound(float viewp_from_presenter[4],
		Vec3f sound_loc_from_presenter, 
		Vec3f sound_dir_from_presenter, 
		float ellipse_definitions[4]);
	
	/// Destructor

	~EllipseSound();

	/** Operator. */
	float operator()(float input);

	/** Updates the gain. */
	void ChangeGain(float viewp_from_presenter[4], 
		Vec3f sound_loc_from_presenter, 
		Vec3f sound_dir_from_presenter, 
		float ellipse_definitions[4]);

private:
	Vec3f rel_pos;
	float gain;
	void GetRelativePosition(float viewp_from_presenter[4], 
		Vec3f sound_loc_from_presenter, 
		Vec3f sound_dir_from_presenter, 
		Vec3f* pos); // returns vector with: x,y,z, assumed sound direction (0 0 1)

	float ComputeGain(float ellip_def[4], Vec3f pos); // retuns current gain 
};

/** class for Sound node processing.

Takes care of the sound processing according to the ellipse and spatialize 
information in the fields of the Sound node.
*/
 
class ProcessSound3D 
{
public:

	/** Constructor. */
	ProcessSound3D(float viewp[4], 
		Sound* S3D, 
		Matrix tr);
	
	/** Destructor.

	Delets the gain computation and 3D spatialization objects
	*/
	~ProcessSound3D(void);

	/** Operator.
	
	  Processes the sound at each input audio sample. */
	float* operator()(float input);

	/** Updates the relative locations of Sound and Viewpoint, and calls the updates
	of gain computation and amplitude panning functions. */
	void Update(float viewp[4], 
		Sound* S3D, 
		Matrix tr);


private:
	float out[2];

	/** pointer to amplitude panning object.

	This is used only when the spatialize flag == TRUE.
	*/
	AmPan* pan;

	/** pointer to the ellipse related gain computation.
	*/
	EllipseSound* ellip;

	Vec3f sound_loc;
	Vec3f sound_dir;
	float ellipse_def[4];
	float intensity;
	float priority;
	Bool spatialize;

	/** Computes the relative positions of the viewpoint and the sound source. */
	void GetSoundInfo(Sound* S3D, 
		Matrix tr);
};

#ifdef ADV_AUDIO

/**	class for DirectiveSound node processing. */
class ProcessDirectiveSound 
{
	friend class ProcessSound;
	friend class AudioRenderer;

public:

	/** Constructor. */
	
	ProcessDirectiveSound(int s_freq, 
		float viewp[4], 
		DirectiveSound* DS, 
		Matrix tr, int update, 
		ZArray<AcousticScene*>& pAScene, 
		ZArray<Matrix>& trAScene);

	/** Destructor.
	
	  Deletes the Auralization object used by this class.
	*/
	~ProcessDirectiveSound(void);

	// Change to provate (riitta 161199)
	/** Computes the sound location.*/
//	void GetSoundInfo(Matrix tr, DirectiveSound* DS);

	/** Updates parameters at the update rate of the auralization process. */
	void Update(float viewp[4], 
		DirectiveSound* DS, 
		Matrix tr, 
		ZArray<AcousticScene*>& pAScene, 
		ZArray<Matrix>& trAScene);

	/** Operator. Does the actual processing*/
		float* operator()(float input);


private:

	// riitta 161199:
	void EmptyImagesources();

	/** Gets image sources according to acoustic surface information under 
	the AcousticScene above this sound node. */
	void MakeImageSourceArray(float viewp[4], 
		AcousticScene* pAS);

	/** Computes the sound location.*/
	void GetSoundInfo(Matrix tr, DirectiveSound* DS);

	/** Updates the locations of the image sources.

	Also checks their visibility. */
	void UpdateImageSourceArray(float viewp[4], 
		AcousticScene* pAS);

	static void ISLocation(Vec3f* loc, 
		AcousticSurface* surf, 
		Vec3f s_l);


	int update_period;

	/** SoundSource object for holding the direct sound information. */
	SoundSource* directsound;

	/** Array of pointers to ImageSource objects. */
	ZArray<ImageSource*> imagesources;

	BOOL InsideBbox(AcousticScene* pAS, 
		Matrix& trAS);
	AcousticScene* WhichAcousticScene(ZArray<AcousticScene*>& pAScene, 
		ZArray<Matrix>& trAScene);
	AcousticScene* pCurrentAcousticScene;
	AcousticScene* pNewAcousticScene;
	
	/** Array of pointers to the AcousticScene node above this DirectiveSound. */
	Vec3f view_coord;
	float out_gain;
	
	float size_x;
	float out[2];

	/** pointer to the Auralize object which controls the actual sound 
	processing at each sample. */
	Auralize* aural;

	int samp_freq;
	Vec3f old_sound_loc;
	Vec3f sound_loc;
	Vec3f sound_dir;
	float intensity;
	Bool roomEffect;
};

#endif // ADV_AUDIO

/** class that selects the correct sound processing class according
to the type of the sound source. 

This object is initialized by the AudioRenderer for each sound stream.
*/
class ProcessSound 
{
	// Riitta 021199
	friend class AudioRenderer;
public:

	/** Constructor. */
	ProcessSound(int stream_number, 
		int s_freq, 
		Presenter* pres, 
		int update, 
		MediaObject* parent_sound);

	/** Destructor.

	deletes the correct sound processing function. */
	~ProcessSound(void);

	/** Updates sound information (direction, location), and 
	calls the Update function of the correct sound procesing object. */
	void Update(int stream_id, 
		Presenter* pres);

	/** Transfroms a 3D point with given transformation matrix mat*/
	static Vec3f TransformPoints(Matrix mat, 
		Vec3f vec);


	/** Calls the correct sound processing function according to the source type*/
	 virtual float* operator()(float insample);

#if ADV_AUDIO
	 void process(int stream_id,void* input,short* output,int spat_buf_length);
	 int getSourceType(void)
	 {	return source_type;
	 }
#endif

	
private:
	Time old_time;
	int source_type; // 0 = Sound3D, 1 = Sound2D, 2 = DirectiveSound

	/// holds current viewpoint information (location, direction)
	float viewp[4];

	/// sound transformation matrix
	Matrix tr;
	float out[2];
	MediaObject* parentSound;

#ifdef ADV_AUDIO

	/** pointer to an object for processing of DirectiveSound.*/
	ProcessDirectiveSound* dirsound;

	DirectiveSound* DS;

	float* perceptualsound;
	CSpatializer* m_spat;
	CSpatParameters* m_spatparameters;
	Presenter* m_pres;
    void GetSoundPositionAndAngles(float* azimuth,float* dirAngle,float* x,float* y,float*z);

#endif // ADV_AUDIO

	/** pointer to an object for processing of Sound.
	
	This is instanciated if the source type is Sound, otehrwise remains NULL. */
	ProcessSound3D* sound_3D;	
	Sound* S3D;

	/** Sets up the correct sound processing object. */
	void SetupSoundProcessing(int stream_id, 
		int s_freq, 
		float viewp[4],
		int update,
		Presenter* pres);

	/** Looks for sound source associated to this audio stream. */
	void GetSoundInfo(int stream_id, 
		Presenter* pres);
};

#endif
