// This software module was originally developed by Matthew Leditschke
// (CSELT) in the course of development of the MPEG-4 Systems standard
// (ISO/IEC 14496-1).  This software module is an implementation of a part
// of one or more MPEG-4 Systems tools as specified by the MPEG-4 Systems
// standard.  ISO/IEC gives users of the MPEG-4 Systems standard free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4 Systems
// standard.  Those intending to use this software module in hardware or
// software products are advised that its use may infringe existing
// patents.  The original developer of this software module and his/her
// company, the subsequent editors and their companies, and ISO/IEC have no
// liability for use of this software module or modifications thereof in an
// implementation.  Copyright is not released for non MPEG-4 Systems
// conforming products.  The original developer retains full right to use
// the code for his/her own purpose, assign or donate the code to a third
// party and to inhibit third parties from using the code for non MPEG-4
// Systems conforming products.  This copyright notice must be included in
// all copies or derivative works.
//
// For more information or to receive an up to date version of this module,
// mail karlo@nta.no or check <http://televr.fou.telenor.no/compositor/>.
//
// Copyright 1997.

#include "OpenGl/OpenGl.h"
#include <assert.h>
#include "AudioRenderer.h"
#include "Presenter.h"
#include "Stream.h"
#include "OpenGl/AdvancedAudio/AudioProcess/ProcessSound.h"

#ifdef WIN32

//#include <mmsystem.h>
#include "Clock.h"
#include "opengl/base/dx-errors.h"

// This is a symbol defined in DirectX 3
#ifndef DSBCAPS_GETCURRENTPOSITION2
#define DSBCAPS_GETCURRENTPOSITION2 0x00010000
#endif

// Stefano Battista 98/11/12
#include "opengl/audio/AudioObjectProxy.h"
#include "opengl/audio/SoundProxy.h"
#include "opengl/audio/Sound2DProxy.h"

// The interface assumes that audio samples are presented in
// fixed size frames.  A number of frames are grouped together
// into a block before presenting to the DirectSound routines.
// The DirectSound streaming buffer contains a number of blocks.
const unsigned int FRAMES_PER_BLOCK = 3;
const unsigned int BLOCKS_PER_BUFFER = 9;

// This is the number of blocks to wait before starting the
// buffer playing.
const unsigned int BLOCKS_BEFORE_START_PLAYING = 3;

// riitta 200999; added the following two definitions 
#define NOF_FRAMES 4		// update interval for sound processing
#define OUTPUTRATE 8000		// samplingrate of the output audio
							// (needed in the initialization of the 
							// primary DirectSound buffer)

//
// A class containing all of the details pertaining to a stream.
//
class StreamInfo
{
public:
  PCMWAVEFORMAT pcmwf;		// The format of the audio being played
  DSBUFFERDESC dsbdesc;		// The DirectSound buffer descriptor
  LPDIRECTSOUNDBUFFER pSBuf;	// The DirectSound streaming buffer object
  char* preBuffer;			// Collect frames in this buffer
  unsigned int muiPrebufferSize;	// The size of the preBuffer
  unsigned int muiPrebufferFrameWrite;	// Index of the next free frame in the prebuffer
  // unsigned int ;		// Where to write the next byte of data
  unsigned int frameSize;		// The size of a frame (in bytes)
  BOOL startedPlaying;		// TRUE once the buffer has been started playing
  unsigned int muiCountAllBlocks;	// The number of blocks that have been played
  
  // SB 99/03/08
  unsigned int muiPointRead;
  unsigned int muiPointWrite;
  
  // Constructor, set default values
  StreamInfo() :
		pSBuf(NULL),
      preBuffer(NULL),
      muiPrebufferFrameWrite(0),
      startedPlaying(FALSE),
      muiCountAllBlocks(0),
      muiPointRead(0),
      muiPointWrite(0)
    {}
};


// ========================================================================

//
// AudioRenderer::AudioRenderer
//
// Class constructor - given a pointer to the presenter.
//
// Initialise the DirectSound routines.
//

AudioRenderer::AudioRenderer(Presenter* pPresenter, HWND hWnd)
{
  char errorString[200];
  errorString[0] = 0;
  m_pPresenter = pPresenter;
  
  // Initialise the array of sound streams
  m_streams = NULL;
  m_streamsSize = 0;
  m_numStreams = 0;
  
  //RIITTA:
//  pan = NULL;
//  ellip = NULL;
  // buffer length for spatializer
  buf_spat = NULL;
  spat_buf_length = 0;
//  viewp = new float[4];
//  ell_def = new float[4];
  
//  x = NULL;
//  tr = NULL;
//  count = 0;
//  spat = 0;
  
  // Create the DirectSound object, the interface to the sound card
  HRESULT hResult = DirectSoundCreate(NULL, &m_pDirectSound, NULL);
  if (hResult != DS_OK) {
    sprintf (errorString, "DirectSoundCreate() failed.\n  %s",
      HRESULTtoText(hResult));
    return;
  }
  
  // Set the cooperative level
  if (errorString[0] == 0)
  {
    hResult = m_pDirectSound->SetCooperativeLevel(hWnd, DSSCL_PRIORITY);
    if (hResult != DS_OK)
      sprintf (errorString, "SetCooperativeLevel() failed.\n  %s",
					 HRESULTtoText(hResult));
  }
  
  // Create a primary buffer (in order to set the output format)
  DSBUFFERDESC primaryDesc;
  memset (&primaryDesc, 0, sizeof (DSBUFFERDESC));
  primaryDesc.dwSize = sizeof (DSBUFFERDESC);
  primaryDesc.dwFlags = DSBCAPS_PRIMARYBUFFER;
  LPDIRECTSOUNDBUFFER primarySB;
  hResult = m_pDirectSound->CreateSoundBuffer(&primaryDesc, &primarySB, NULL);
  if (hResult != DS_OK)
    sprintf (errorString, "CreateSoundBuffer() failed.\n  %s",
    HRESULTtoText(hResult));
  else
  {
    // Specify the new format
    WAVEFORMATEX wFmt;
    wFmt.wFormatTag		 = WAVE_FORMAT_PCM;
    //		wFmt.nChannels		 = 1;
    wFmt.nChannels		 = 2;
	wFmt.nSamplesPerSec	 = OUTPUTRATE;
	//    wFmt.nSamplesPerSec	 = 8000;
    //		wFmt.nSamplesPerSec	 = 22050;
    wFmt.wBitsPerSample	 = 16;
    wFmt.nBlockAlign	 = wFmt.nChannels * wFmt.wBitsPerSample / 8;
    wFmt.nAvgBytesPerSec = wFmt.nSamplesPerSec * wFmt.nBlockAlign;
    wFmt.cbSize			 = 0;
    hResult = primarySB->SetFormat(&wFmt);
    if (hResult != DS_OK)
      sprintf (errorString, "CreateSoundBuffer() failed.\n  %s",
					 HRESULTtoText(hResult));
  }
  
  // If an error has occurred, display a message and then exit
  if (errorString[0] != 0)
  {
    //		CFatalError fatalErrorDialog;
    //		fatalErrorDialog.m_message = errorString;
    //		fatalErrorDialog.DoModal();
    
    m_initOK = FALSE;
  }
  else
    m_initOK = TRUE;
  
  // Tell audio renderer to adjust clock first time around
  m_bFirst = TRUE;
} /* AudioRenderer::AudioRenderer */


// ========================================================================

//
// AudioRenderer::~AudioRenderer
//
// Class destructor - shuts down the DirectSound routines.
//

AudioRenderer::~AudioRenderer()
{
  //	TRACE_PRINT("Entering AudioRenderer::~AudioRenderer():\n");
  
  // Terminate the session (closing down all streams)

	// riitta 200999
	// Terminate is already called from the OpenGLPresenter
//	Terminate();
  

	// riitta 200999
	// delete all the existing sound processing classes
	// and related material

	int i=0;			
	for(i=0;i<sound_process.GetLength();i++){
		if(sound_process[i] != NULL){
				delete sound_process[i];
				sound_process[i] = NULL;
			}
		}
		
//	delete [] sound_process;
		sound_process.Clear();
		count.Remove();

  // Delete the array of streams
	if (m_streams)
		free(m_streams);
  
  // Release the main direct sound object
	if (m_pDirectSound != NULL)
		m_pDirectSound->Release();

	if(buf_spat){
		delete [] buf_spat;
		buf_spat = NULL;
	}
	
  //	TRACE_PRINT("Leaving AudioRenderer::~AudioRenderer():\n");
} /* AudioRenderer::~AudioRenderer */


// ========================================================================

//
// AudioRenderer::Init
//
// Initialise the renderer.
//

void AudioRenderer::Init()
{
  // Empty the list ofactive streams
  m_numStreams = 0;
} /* AudioRenderer::Init */


// ========================================================================

//
// AudioRenderer::Terminate
//
// Stops rendering and performs cleanups.
//

void AudioRenderer::Terminate()
{
  //	TRACE_PRINT("Entering AudioRenderer::Terminate():\n");
  
  // Close down all of the streams which haven't already been closed.
  for (int i = 0; i < m_numStreams; i++)
    if (m_streams[i] != NULL)
      CloseOutputStream(i);



//		COMPLAIN("in AR  terminate");



    //	TRACE_PRINT("Leaving AudioRenderer::Terminate():\n");
} /* AudioRenderer::Terminate */


// ========================================================================

//
// AudioRenderer::GetCapabilities
//
// Get the capabilities of this audio renderer.
//

void AudioRenderer::GetCapabilities (Capabilities* pCaps)
{
  pCaps->AddFormatCapability(/* FORMAT_PCM */ MediaFormat::PCM, 0);
} /* AudioRenderer::GetCapabilities */


// ========================================================================

//
// AudioRenderer::OpenOutputStream
//
// Open a stream within the AudioRenderer which audio data can
// be sent to, returning the number of this stream.  When opening
// this stream, the format of the audio sample that are to be sent
// to this stream must be given, comprising the sample rate,
// number of channels, bits per sample and sample format.  The size
// of audio frames (in terms of the number of samples) that will be
// given to QueueFrameForPlaying() is also required.
//
// If the stream couldn't be opened, -1 is returned, otherwise the
// stream number is returned.
//

int AudioRenderer::OpenOutputStream(MediaStream* pStream,
                                    unsigned int sampleRate,
                                    unsigned int numChannels,
                                    unsigned int bitsPerSample,
                                    unsigned int /*sampleFormat*/,
                                    unsigned int frameSize,
									MediaObject* parent_sound) // riitta 2000999
{
  StreamInfo* pNewStream = new StreamInfo;
  
  // Set up wave format structure describing the audio format
  memset(&(pNewStream->pcmwf), 0, sizeof(PCMWAVEFORMAT));
  pNewStream->pcmwf.wf.wFormatTag = WAVE_FORMAT_PCM;
  pNewStream->pcmwf.wf.nChannels = (unsigned short)numChannels;
  pNewStream->pcmwf.wf.nSamplesPerSec = sampleRate;
  pNewStream->pcmwf.wf.nBlockAlign = (unsigned short)((int)((bitsPerSample + 7) / 8) * numChannels);
  //RIITTA
  pNewStream->pcmwf.wf.nAvgBytesPerSec = pNewStream->pcmwf.wf.nSamplesPerSec * pNewStream->pcmwf.wf.nBlockAlign;
  pNewStream->pcmwf.wBitsPerSample = (unsigned short)bitsPerSample;
  //COMPLAIN("NUM_BITS %d", sizeof(short));
  
  // Set up DSBUFFERDESC structure.
  memset(&(pNewStream->dsbdesc), 0, sizeof(DSBUFFERDESC));
  pNewStream->dsbdesc.dwSize = sizeof(DSBUFFERDESC);
  pNewStream->dsbdesc.dwFlags = DSBCAPS_CTRLVOLUME | DSBCAPS_GETCURRENTPOSITION2; 
  
  // Create the streaming buffer (output buffer)
  pNewStream->dsbdesc.dwBufferBytes = frameSize * FRAMES_PER_BLOCK
    * BLOCKS_PER_BUFFER
    * pNewStream->pcmwf.wf.nBlockAlign;
  
  pNewStream->dsbdesc.lpwfxFormat = (LPWAVEFORMATEX)&(pNewStream->pcmwf);
  
  TRACE("[AR::OpenOutputStream] Initial Buffer Size %8d bytes \n",
    pNewStream->dsbdesc.dwBufferBytes);
  //	TRACE_PRINT1("OpenOutputStream(): Initial buffer size = %d milliseconds.\n",
  //				1000 * pNewStream->dsbdesc.dwBufferBytes / pNewStream->pcmwf.wf.nAvgBytesPerSec);
  
  // Create the DirectSound buffer
  HRESULT hResult = m_pDirectSound->CreateSoundBuffer(&(pNewStream->dsbdesc),
    &(pNewStream->pSBuf),
    NULL);
  if (hResult != DS_OK)
  {
    // Creation failed, return -1.
    delete pNewStream;
    return -1;
  }
  
  // Allocate the pre-buffer space for this stream 
  pNewStream->frameSize = frameSize * pNewStream->pcmwf.wf.nBlockAlign;
  pNewStream->muiPrebufferSize = FRAMES_PER_BLOCK * pNewStream->frameSize;
  pNewStream->preBuffer = new char [pNewStream->muiPrebufferSize]; //in bytes
  
  // The audio stream has been successfully created.  Add it to the
  // internal array of streams.
  
  // Make the array larger, if required
  if (m_numStreams == m_streamsSize)
  {
    m_streamsSize += 5;
    m_streams = (StreamInfoPtr*)realloc(m_streams,
      sizeof(StreamInfoPtr) * m_streamsSize);
  }
  
  // Add the new stream to this array
  m_streams[m_numStreams] = pNewStream;
  m_numStreams++;
  
  // Adjust clock according to first audio stream
  if (m_bFirst)
  {
    ((MediaStreamImp *) pStream)->AdjustClock();
    m_bFirst = FALSE;
  }
  
  // riitta 200999
	//COMPLAIN("In AUdioRenderer, m_numStreams-1: %d", m_numStreams-1);
	SetupBuffer(sampleRate, m_numStreams-1);
	int frame_length = (int)(spat_buf_length/2);
	int update_period = frame_length * NOF_FRAMES;
		
	// riitta210999
	// If you want to test the audio output WITHOUT any audio 
	// processing, comment out the following instantiation of 
	// ProcessSound objects, and take comments away from the
	// "sound_process += NULL" statement
	sound_process += new ProcessSound(
		m_numStreams-1,
		sampleRate, 
		m_pPresenter, 
		update_period, 
		parent_sound);
	
//	sound_process += NULL;
	count += 0;

//Setup spatialize and sound ellipses
//  SetupSpatialize(8000, m_numStreams - 1);
  
  // Return the index of this new stream
  return (m_numStreams - 1);
} /* AudioRenderer::OpenOutputStream */




// ========================================================================

//
// AudioRenderer::CloseOutputStream
//
// Close down the given internal channel in the audio renderer.
//

void AudioRenderer::CloseOutputStream(int strNum)
{
  // Only proceed if this stream hasn't already been closed down
  if (m_streams[strNum] == NULL)
    return;
  
  // Ensure that the given stream number actually exists
  assert(strNum < m_numStreams);
  
  // Stop all of the sound buffers playing
  m_streams[strNum]->pSBuf->Stop();
  
  // Release this sound buffer associated with this stream
  if (m_streams[strNum]->pSBuf != NULL)
    m_streams[strNum]->pSBuf->Release();
  
  delete m_streams[strNum]->preBuffer;
  delete m_streams[strNum];
  m_streams[strNum] = NULL;
  
} /* AudioRenderer::CloseOutputStream */


// ========================================================================

//
// AudioRenderer::GetBufferDelayTime
//
// Get the amount of delay time (in milliseconds) that the buffering
// on this stream imposes.
//

unsigned int AudioRenderer::GetBufferDelayTime(int strNum)
{
  // Ensure that the given stream number actually exists
  
  
  assert(strNum < m_numStreams);
  assert(m_streams[strNum] != NULL);
  
  return (1000 * m_streams[strNum]->frameSize * FRAMES_PER_BLOCK * BLOCKS_BEFORE_START_PLAYING
    / m_streams[strNum]->pcmwf.wf.nSamplesPerSec
    / m_streams[strNum]->pcmwf.wf.nBlockAlign);
} /* AudioRenderer::GetBufferDelayTime */


// ========================================================================
//
// AudioRenderer::QueueFrameForPlaying
//
// Put the given frame of audio samples in the queue of samples
// to be played.  Samples from different channels will be combined
// together by the audio renderer.  The number of samples in the
// frame is that given to OpenOutputStream().
//
// If TRUE is returned, the data has successfully been queued, and the
// memory pointed to can be reused.  If FALSE is returned, the data
// couldn't be queued because of a lack of buffer space.  The frame
// of data should be queued at a later stage.
//
// If bitsPerSample is no more than 8, the samples are to be stored
// in an array of unsigned chars.  If bitsPerSample is no more than
// 16, the samples are to be stored in an array of signed shorts.
// Currently bitsPerSample can't exceed 16.
//
// The data is to be stored channel interleaved.
//

BOOL AudioRenderer::QueueFrameForPlaying(
                                         int strNum,
                                         void* ptrToSamples)
{
  HRESULT hResult;
  BOOL frameUsed = FALSE;
  
  int i=0;
  
  // Ensure that the given stream number actually exists
  // COMPLAIN("in queueframe...: s");
  assert(strNum < m_numStreams);

  // riitta 210999
  /* The AudioSourceThread() may still be running and send frames to QueueFrameForPlaying
   even if the CloseOutputStream has been called (from the Terminate() method), i.e., when 
	exiting the player. Therefore we don't want to assert the following, but just to test
	whether this output stream still exists, and if it doesn't, then we'll just return 
	"frameUsed" without any output or processing.
	*/
  // assert(m_streams[strNum] != NULL);
  if(!m_streams[strNum])
	  return frameUsed;

  // Ensure that some samples are given
  assert(ptrToSamples != NULL);
  
  TRACE("[AR::Queue] Time %8d \n", m_pPresenter->GetClockTime());
  
  // ***************************************************************
  // STEP 1:
  // Add this frame to the pre-buffer if there is room
  // ***************************************************************
  
  if (m_streams[strNum]->muiPrebufferFrameWrite < FRAMES_PER_BLOCK)
  {

	  // riitta 210999
	// the following processing loop is changed to the one below that 
	  //update coefficients, now appr. at rate 20 Hz (every 480th sample)
/*
      float y = 0.0;

      GetSoundAndViewPositions(strNum);
      CleanSpatBuffer();

    if(pan)
      pan->ChangeGains(viewp, sound_loc);
    if(ellip)
      ellip->ChangeGain(viewp, sound_loc, sound_dir, ell_def);
    
    // ========================================================
    // Spatialization loop
    short* temp = NULL;
    temp = (short*) (ptrToSamples);
    // begin loop >>>>
    for (i = 0; i < spat_buf_length; i += 2) 
    {
      if (ellip)
        y = (*ellip)((float) (*temp));
      else
        y = (float) (*temp);
      
      if (pan) 
      {
        x = (*pan)(y);
        buf_spat[i] = x[0];
        buf_spat[i+1] = x[1];
      } 
      else 
      {
        //if no spatialization
        buf_spat[i] = y;
        buf_spat[i+1] = y;
      }
      temp++;
    }
    // end loop <<<<
    // ========================================================
*/


// riitta 200999:
///////////////////////////////////////////////////////////////////////////

	  // check whether the update interval has been reached	
	  if(count[strNum]==NOF_FRAMES){
		  if(sound_process[strNum])
			  sound_process[strNum]->Update(strNum, m_pPresenter);	
		  count.SetAt(strNum,0);
	  }
	  count.SetAt(strNum,count[strNum]+1);


//  test if perceptual processing applies 

	  int itest=sound_process[strNum]->getSourceType();
	  if(itest==3)
	  { short* temp = NULL;
		temp = (short*) (ptrToSamples);
		if(sound_process[strNum])
			sound_process[strNum]->process(strNum,ptrToSamples,buf_spat,spat_buf_length);
	  }
	  else 
	  {	short* temp = NULL;
		float* x = NULL;
		temp = (short*) (ptrToSamples);

		for(i=0;i<spat_buf_length;i=i+2){	
			if(sound_process[strNum])
				x = (*sound_process[strNum])((float) (*temp));

			else{				
				float x2[2];
				x2[0] = 0.0;
				x2[1] = 0.0;
				x2[0] = *temp;
				x2[1] = *temp;
				x = (float*)x2;				
			}
			buf_spat[i] = x[0];
			buf_spat[i+1] = x[1];	
			temp ++;
		} // for
	  } // end if

///////////////////////////////////////////////////////////////////////////
    memcpy(
      m_streams[strNum]->preBuffer + 
      m_streams[strNum]->muiPrebufferFrameWrite * 
      m_streams[strNum]->frameSize, 
      buf_spat,
      m_streams[strNum]->frameSize);
    
    m_streams[strNum]->muiPrebufferFrameWrite++;
    frameUsed = TRUE;
    
  } // end if (m_streams[strNum]->nextFreeFrame < FRAMES_PER_BLOCK)
  
  // ***************************************************************
  // step 2:
  // If the pre-buffer is full, try to send it to the DirectSound
  // streaming buffer.
  // ***************************************************************
  
  if (m_streams[strNum]->muiPrebufferFrameWrite == FRAMES_PER_BLOCK)
  {
    // Determine the current play cursor
    unsigned long playCursor, writeCursor;
    hResult = m_streams[strNum]->pSBuf->GetCurrentPosition(
      &playCursor, &writeCursor);
    char* errText = HRESULTtoText(hResult);
    assert(hResult == DS_OK);
    
    /*** THIS IS BROKEN SB 99/03/08
    // Determine the amount of free space in the buffer
    unsigned long writeLength;
    if (m_streams[strNum]->nextWrite > playCursor)
    {
    // Writing at the end of the buffer, wrapping around to the start
    writeLength = 
    m_streams[strNum]->dsbdesc.dwBufferBytes - 
    m_streams[strNum]->nextWrite + playCursor;
    }
    else if ((m_streams[strNum]->nextWrite == playCursor) &&
    (m_streams[strNum]->startedPlaying == FALSE))
    {
    // Same write and read point - buffer is empty
    writeLength = m_streams[strNum]->dsbdesc.dwBufferBytes;
    }
    else
    {
    // Writing behind the play pointer
    writeLength = playCursor - m_streams[strNum]->nextWrite;
    }
    SB 99/03/08 ***/
    
    // Determine the amount of free space in the buffer
    unsigned int uiLengthWrite;
    // unsigned int uiLengthRead;
    m_streams[strNum]->muiPointRead = playCursor;
    if (m_streams[strNum]->muiPointWrite > 
      m_streams[strNum]->muiPointRead)
    {
      // Writing at the end of the buffer, wrapping around to the start
      uiLengthWrite = 
        m_streams[strNum]->dsbdesc.dwBufferBytes - 
        m_streams[strNum]->muiPointWrite + 
        m_streams[strNum]->muiPointRead;
    }
    else if (m_streams[strNum]->muiPointWrite < 
      m_streams[strNum]->muiPointRead)
    {
      // Writing behind the play pointer
      uiLengthWrite = m_streams[strNum]->muiPointRead - 
        m_streams[strNum]->muiPointWrite;
    }
    else if (m_streams[strNum]->muiPointWrite == 
      m_streams[strNum]->muiPointRead)
    {
      if (m_streams[strNum]->startedPlaying == FALSE)
      {
        // Same write and read point - buffer is empty
        uiLengthWrite = m_streams[strNum]->dsbdesc.dwBufferBytes;
      }
      else 
      {
        // uiLengthWrite = 0;
        uiLengthWrite = m_streams[strNum]->dsbdesc.dwBufferBytes;
      }
    }
    // SB $$$$ REALLY DIRTY !!!!
    uiLengthWrite = m_streams[strNum]->dsbdesc.dwBufferBytes;
    
    // ========================================================
    // if the Pre Buffer is full:
    // LOCK/COPY/UNLOCK (BEGIN)
    // ========================================================
    
    // If there is space for writing in the Buffer,
    // copy data from the Prebuffer
    if (uiLengthWrite >= m_streams[strNum]->muiPrebufferSize)
    {
      TRACE("[AR::Queue] before Lock/Copy/Unlock frame # %4d \n",
        m_streams[strNum]->muiCountAllBlocks);
      
      // Lock the buffer before writing to it
      LPBYTE lpWrite1, lpWrite2;
      DWORD dwLen1, dwLen2;
      DWORD bytesToWrite = m_streams[strNum]->muiPrebufferSize;
      hResult = m_streams[strNum]->pSBuf->Lock(
        m_streams[strNum]->muiPointWrite,
        bytesToWrite,
        (void **) &lpWrite1, 
        &dwLen1,
        (void **) &lpWrite2, 
        &dwLen2, 
        0);
      
      if (hResult == DSERR_BUFFERLOST)
      {
        m_streams[strNum]->pSBuf->Restore();
        hResult = m_streams[strNum]->pSBuf->Lock(
          m_streams[strNum]->muiPointWrite,
          bytesToWrite,
          (void **) &lpWrite1, 
          &dwLen1,
          (void **) &lpWrite2, 
          &dwLen2, 
          0);
      }
      
      errText = HRESULTtoText(hResult);
      assert(hResult == DS_OK);
      
      // Copy over the data into the buffer.  It may have to be done
      // in two copies, as the free space may be wrapped over the end of
      // the circular buffer.
      if (bytesToWrite <= dwLen1)
      {
        memcpy(lpWrite1, m_streams[strNum]->preBuffer, bytesToWrite);
        TRACE("  -- at %8d length %4d \n", lpWrite1, bytesToWrite);
      }
      else
      {
        memcpy(lpWrite1, m_streams[strNum]->preBuffer, dwLen1);
        memcpy(lpWrite2, m_streams[strNum]->preBuffer + dwLen1,
          bytesToWrite - dwLen1);
        TRACE("  -- at %8d length %4d \n", lpWrite1, dwLen1);
        TRACE("  -- at %8d length %4d \n", lpWrite2, dwLen2);
      }
      
      // Unlock the buffer
      hResult = m_streams[strNum]->pSBuf->Unlock(
        lpWrite1, dwLen1,
        lpWrite2, dwLen2);
      errText = HRESULTtoText(hResult);
      assert(hResult == DS_OK);
      
      TRACE("[AR::Queue] after  Lock/Copy/Unlock frame # %4d \n",
        m_streams[strNum]->muiCountAllBlocks);
      
      // Empty the pre-buffer
      m_streams[strNum]->muiPrebufferFrameWrite = 0;
      TRACE("[AR::Queue] Reset the Prebuffer counter \n");
      
      // Update the pointer to the next byte to write
      m_streams[strNum]->muiPointWrite += bytesToWrite;
      if (m_streams[strNum]->muiPointWrite >= m_streams[strNum]->dsbdesc.dwBufferBytes)
      {
        m_streams[strNum]->muiPointWrite -= m_streams[strNum]->dsbdesc.dwBufferBytes;
        TRACE("[AR::Queue] Reset the BUFFER counter \n");
      }
      
      // If the DirectSound streaming buffer is now full, start playing
      // Start the buffer playing (in loop mode)
      m_streams[strNum]->muiCountAllBlocks++;
      if (m_streams[strNum]->muiCountAllBlocks == BLOCKS_BEFORE_START_PLAYING)
      {
        TRACE("[AR::Queue] Play the DSound buffer (LOOP) \n");
        
        m_streams[strNum]->startedPlaying = TRUE;
        hResult = m_streams[strNum]->pSBuf->Play(0, 0, DSBPLAY_LOOPING);
        if (hResult != DS_OK)
          return 0;
      }
    }
    // ========================================================
    // if the Pre Buffer is full:
    // LOCK/COPY/UNLOCK (END)
    // ========================================================
    
  }
  
  // Return TRUE if frame sent, FALSE otherwise.
  return frameUsed;
} /* AudioRenderer::QueueSamplesForPlaying */

#else

// ========================================================================

//
// AudioRenderer::AudioRenderer
//
// Class constructor - given a pointer to the presenter.
//
// Initialise the DirectSound routines.
//
AudioRenderer::AudioRenderer(Presenter* pPresenter, HWND /* hWnd */)
: m_pPresenter(pPresenter)
{
  
} /* AudioRenderer::AudioRenderer */


// ========================================================================

//
// AudioRenderer::~AudioRenderer
//
// Class destructor - shuts down the DirectSound routines.
//

AudioRenderer::~AudioRenderer()
{
} /* AudioRenderer::~AudioRenderer */


// ========================================================================

//
// AudioRenderer::Init
//
// Initialise the renderer.
//

void AudioRenderer::Init()
{
  //	m_pPresenter->GetExecutive()->GetClock()->Start();
} /* AudioRenderer::Init */


// ========================================================================

//
// AudioRenderer::Terminate
//
// Stops rendering and performs cleanups.
//

void AudioRenderer::Terminate()
{
} /* AudioRenderer::Terminate */


// ========================================================================

//
// AudioRenderer::GetCapabilities
//
// Get the capabilities of this audio renderer.
//

void AudioRenderer::GetCapabilities (Capabilities* /* pCaps */)
{
} /* AudioRenderer::GetCapabilities */


// ========================================================================

//
// AudioRenderer::OpenOutputStream
//
// Open a stream within the AudioRenderer which audio data can
// be sent to, returning the number of this stream.  When opening
// this stream, the format of the audio sample that are to be sent
// to this stream must be given, comprising the sample rate,
// number of channels, bits per sample and sample format.  The size
// of audio frames (in terms of the number of samples) that will be
// given to QueueFrameForPlaying() is also required.
//
// If the stream couldn't be opened, -1 is returned, otherwise the
// stream number is returned.
//

int AudioRenderer::OpenOutputStream(MediaStream * /* pStream */,unsigned int /* sampleRate */,unsigned int /* numChannels */,unsigned int /* bitsPerSample */,unsigned int /* sampleFormat */,unsigned int /* frameSize */, MediaObject *)

{
	return 0; // SOLARIS PORTING
} /* AudioRenderer::OpenOutputStream */


// ========================================================================

//
// AudioRenderer::CloseOutputStream
//
// Close down the given internal channel in the audio renderer.
//

void AudioRenderer::CloseOutputStream(int strNum)
{
} /* AudioRenderer::CloseOutputStream */


// ========================================================================

//
// AudioRenderer::GetBufferDelayTime
//
// Get the amount of delay time (in milliseconds) that the buffering
// on this stream imposes.
//

unsigned int AudioRenderer::GetBufferDelayTime(int strNum)
{
	return 0; // SOLARIS PORTING
} /* AudioRenderer::GetBufferDelayTime */


// ========================================================================

//
// AudioRenderer::QueueFrameForPlaying
//
// Put the given frame of audio samples in the queue of samples
// to be played.  Samples from different channels will be combined
// together by the audio renderer.  The number of samples in the
// frame is that given to OpenOutputStream().
//
// If TRUE is returned, the data has successfully been queued, and the
// memory pointed to can be reused.  If FALSE is returned, the data
// couldn't be queued because of a lack of buffer space.  The frame
// of data should be queued at a later stage.
//
// If bitsPerSample is no more than 8, the samples are to be stored
// in an array of unsigned chars.  If bitsPerSample is no more than
// 16, the samples are to be stored in an array of signed shorts.
// Currently bitsPerSample can't exceed 16.
//
// The data is to be stored channel interleaved.
//

BOOL AudioRenderer::QueueFrameForPlaying(int strNum,
                                         void* ptrToSamples)
{
	return 0; // SOLARIS PORTING
} /* AudioRenderer::QueueSamplesForPlaying */

#endif

// ========================================================================

//
// AudioRenderer::GetPresenter()
//
// Get Presenter of renderer.
//

Presenter* AudioRenderer::GetPresenter() const
{
  return m_pPresenter;
} /* AudioRenderer::GetPresenter */

// ========================================================================


#ifdef WIN32

void AudioRenderer::SetupBuffer(int /*s_freq*/, int stream_id)
{
	if(!buf_spat){
		int t = 0;
		spat_buf_length = (m_streams[stream_id]->frameSize)/2;
		buf_spat = new short[spat_buf_length];
		for(t=0;t<spat_buf_length;t++){
			buf_spat[t] = 0;
		}
	}
}


void AudioRenderer::DeleteSoundProcessing(int stream_id)
{
	if(stream_id != -1)
		if(sound_process[stream_id] != NULL){
				delete sound_process[stream_id];
				sound_process[stream_id] = NULL;
				COMPLAIN("sound processing deleted");
		}
}

#endif
