// This software module was originally developed by Lars Nilsson (Telenor
// R&D) in the course of development of the MPEG-4 Systems standard
// (ISO/IEC 14496-1). This current version includes approach developed by
// Vctor Fernndez-Carbajales Caete (GTI-EPS-UAM/OLGA). This software 
// module is an implementation of a part of one or more MPEG-4 Systems 
// tools as specified by the MPEG-4 Systems standard. ISO/IEC gives users
// of the MPEG-4 Systems standard free license to this software module or
// modifications thereof for use in hardware or software products claiming
// conformance to the MPEG-4 Systems standard. Those intending to use 
// this software module in hardware or software products are advised that
// its use may infringe existing patents.  The original developer of this 
// software module and his/her company, the subsequent editors and their 
// companies, and ISO/IEC have no liability for use of this software module
// or modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 Systems conforming products.  The original developer 
// retains full right to use the code for his/her own purpose, assign or 
// donate the code to a third party and to inhibit third parties from using
// the code for non MPEG-4 Systems conforming products.  
// This copyright notice must be included in all copies or derivative works.
//
// For more information or to receive an up to date version of this module,
// mail karlo@nta.no or check <http://televr.fou.telenor.no/compositor/>.
//
// Copyright 1997, 1998.

#include "OpenGl/OpenGl.h"
#include "Matrix.h"


// Constructor

Matrix::Matrix()
{
  m_elements[15] = m_elements[10] = m_elements[5] = m_elements[0] = 1.0f;
  m_elements[1] = m_elements[4] = m_elements[8] = m_elements[12] = 0.0f;
  m_elements[2] = m_elements[6] = m_elements[9] = m_elements[13] = 0.0f;
  m_elements[3] = m_elements[7] = m_elements[11] = m_elements[14] = 0.0f;
}

// Constructor

Matrix::Matrix(float matrix[16])
{
  for (int i = 0; i < 16; ++i)
    m_elements[i] = matrix[i];
}

// Constructor

Matrix::Matrix(const Vec3f& rXAxis,
	       const Vec3f& rYAxis,
	       const Vec3f& rZAxis)
{
  m_elements[ 0] = rXAxis.x;
  m_elements[ 1] = rXAxis.y;
  m_elements[ 2] = rXAxis.z;
  m_elements[ 3] = 0.0f;

  m_elements[ 4] = rYAxis.x;
  m_elements[ 5] = rYAxis.y;
  m_elements[ 6] = rYAxis.z;
  m_elements[ 7] = 0.0f;

  m_elements[ 8] = rZAxis.x;
  m_elements[ 9] = rZAxis.y;
  m_elements[10] = rZAxis.z;
  m_elements[11] = 0.0f;

  m_elements[12] = 0.0f;
  m_elements[13] = 0.0f;
  m_elements[14] = 0.0f;
  m_elements[15] = 1.0f;
}

// Translate

void
Matrix::Translate(float fX, float fY, float fZ)
{
  float tmp[3];

  tmp[0] = m_elements[12];
  tmp[1] = m_elements[13];
  tmp[2] = m_elements[14];

  for (int i = 0; i < 3; i++) 
    tmp[i] += (fX * m_elements[i] + 
	       fY * m_elements[i + 4] + 
	       fZ * m_elements[i + 8]);
	       
  m_elements[12] = tmp[0];
  m_elements[13] = tmp[1];
  m_elements[14] = tmp[2];
}

// Scale

void
Matrix::Scale(float fX, float fY, float fZ)
{
  float tmp[3];

  tmp[0] = fX; 
  tmp[1] = fY; 
  tmp[2] = fZ; 

  for (int i = 0; i < 3; i++) 
    for (int j = 0; j < 3; j++) 
      m_elements[i * 4 + j] = m_elements[j + 4 * i] * tmp[i];
}

// Rotate

void
Matrix::Rotate(float fTheta, float fX, float fY, float fZ)
{
  fTheta *= 0.5f;

  float mag;

  mag = ::sqrt(fX * fX + fY * fY + fZ * fZ);
  
  if (mag != 1.0f && mag != 0.0f) {
    fX /= mag;
    fY /= mag;
    fZ /= mag;
  }

  float theta = ::cos(fTheta);
  float tmp   = ::sin(fTheta);

  float fXsin = fX * tmp;
  float fYsin = fY * tmp;
  float fZsin = fZ * tmp;

  float aa = fXsin * fXsin, ab = fXsin * fYsin, ac = fXsin * fZsin, ad = fXsin * theta;
  float                     bb = fYsin * fYsin, bc = fYsin * fZsin, bd = fYsin * theta;
  float                                         cc = fZsin * fZsin, cd = fZsin * theta;
  float                                                             dd = theta * theta;
  
  float T[3][3] = {{aa - bb - cc + dd,     2.0 * (ab + cd),      2.0 * (ac - bd)},
		   {  2.0 * (ab - cd), - aa + bb - cc + dd,      2.0 * (bc + ad)},
		   {  2.0 * (ac + bd),     2.0 * (bc - ad), - aa - bb + cc + dd}};
  
  for (int i = 0; i < 3; i++)  {
    float shites[3] = {0.0f, 0.0f, 0.0f};
    
    for (int j = 0; j < 3; j++)
      for (int k = 0; k < 3; k++)
        shites[j] += m_elements[k * 4 + i] * T[j][k];

    m_elements[i]     = shites[0];          
    m_elements[i + 4] = shites[1];          
    m_elements[i + 8] = shites[2];          
  }
}

// Multiply point with matrix

Vec3f
Matrix::MultiplyPoint(const Vec3f& rPoint) const
{
  Vec3f tmp;

  tmp.x = (rPoint[0] * m_elements[ 0] +
	   rPoint[1] * m_elements[ 4] +
	   rPoint[2] * m_elements[ 8]);
  
  tmp.y = (rPoint[0] * m_elements[ 1] +
	   rPoint[1] * m_elements[ 5] +
	   rPoint[2] * m_elements[ 9]);
  
  tmp.z = (rPoint[0] * m_elements[ 2] +
	   rPoint[1] * m_elements[ 6] +
	   rPoint[2] * m_elements[10]);
  
  tmp.x += m_elements[12];
  tmp.y += m_elements[13];
  tmp.z += m_elements[14];
  
  return tmp;
}


// Multiply vector with matrix

Vec3f
Matrix::MultiplyVector(const Vec3f& rPoint) const
{
  Vec3f tmp;
  
  tmp.x = (rPoint[0] * m_elements[ 0] +
	   rPoint[1] * m_elements[ 4] +
	   rPoint[2] * m_elements[ 8]);
  
  tmp.y = (rPoint[0] * m_elements[ 1] +
	   rPoint[1] * m_elements[ 5] +
	   rPoint[2] * m_elements[ 9]);
  
  tmp.z = (rPoint[0] * m_elements[ 2] +
	   rPoint[1] * m_elements[ 6] +
	   rPoint[2] * m_elements[10]);
  
  return tmp;
}

void Matrix::operator =(const Rotation& q)
{
	LoadIdentityTransform();

  float xx2=q.x*q.x*2;
  float yy2=q.y*q.y*2;
  float zz2=q.z*q.z*2;
  float xy2=q.x*q.y*2;
  float zw2=q.z*q.theta*2;
  float xz2=q.x*q.z*2;
  float yw2=q.y*q.theta*2;
  float yz2=q.y*q.z*2;
  float xw2=q.x*q.theta*2;
  
  m_elements[0]=1-yy2-zz2;   m_elements[4]=  xy2+zw2;  m_elements[8]=  xz2-yw2;
  m_elements[1]=  xy2-zw2;   m_elements[5]=1-xx2-zz2;  m_elements[9]=  yz2+xw2;
  m_elements[2]=  xz2+yw2;   m_elements[6]=  yz2-xw2;  m_elements[10]=1-xx2-yy2;
}

// MP
void Matrix::LoadIdentityTransform(){

	m_elements[0]= 1;   m_elements[4]=  0;  m_elements[8]=  0;  m_elements[12]=  0;
	m_elements[1]= 0;   m_elements[5]=  1;  m_elements[9]=  0;  m_elements[13]=  0;
	m_elements[2]= 0;   m_elements[6]=  0;  m_elements[10]= 1;  m_elements[14]= 0;
	m_elements[3]=0;   m_elements[7]= 0;  m_elements[11]= 0;  m_elements[15]= 1;

}
// Construct a rotation of a given angle about a given axis.
// Derived from Eric Haines's SPD (Standard Procedural
// Database).
Matrix::Matrix(const Rotation& rot)
{
    LoadIdentityTransform();
	if (rot.theta == 0.0) {
		return;
	}

    double cosine = cos(rot.theta);
    double sine = sin(rot.theta);
    double one_minus_cosine = 1 - cosine;

/*    m_elements[0] = rot.x * rot.x + (1.0 - rot.x * rot.x) * cosine;
    m_elements[1] = rot.x * rot.y * one_minus_cosine + rot.z * sine;
    m_elements[2] = rot.x * rot.z * one_minus_cosine - rot.y * sine;
    m_elements[3] = 0;

    m_elements[4] = rot.x * rot.y * one_minus_cosine - rot.z * sine;
    m_elements[5] = rot.y * rot.y + (1.0 - rot.y * rot.y) * cosine;
    m_elements[6] = rot.y * rot.z * one_minus_cosine + rot.x * sine;
    m_elements[7] = 0;

    m_elements[8] = rot.x * rot.z * one_minus_cosine + rot.y * sine;
    m_elements[9] = rot.y * rot.z * one_minus_cosine - rot.x * sine;
    m_elements[10] = rot.z * rot.z + (1.0 - rot.z * rot.z) * cosine;
    m_elements[11] = 0;
*/
	m_elements[0] = rot.x * rot.x + (1.0 - rot.x * rot.x) * cosine;
    m_elements[4] = rot.x * rot.y * one_minus_cosine + rot.z * sine;
    m_elements[8] = rot.x * rot.z * one_minus_cosine - rot.y * sine;
    m_elements[12] = 0;

    m_elements[1] = rot.x * rot.y * one_minus_cosine - rot.z * sine;
    m_elements[5] = rot.y * rot.y + (1.0 - rot.y * rot.y) * cosine;
    m_elements[9] = rot.y * rot.z * one_minus_cosine + rot.x * sine;
    m_elements[13] = 0;

    m_elements[2] = rot.x * rot.z * one_minus_cosine + rot.y * sine;
    m_elements[6] = rot.y * rot.z * one_minus_cosine - rot.x * sine;
    m_elements[10] = rot.z * rot.z + (1.0 - rot.z * rot.z) * cosine;
    m_elements[14] = 0;

    m_elements[3] = 0;
    m_elements[7] = 0;
    m_elements[11] = 0;
    m_elements[15] = 1;

}

// Construct a traslation of a given vector.
Matrix::Matrix(const Vec3f& tra)
{
    LoadIdentityTransform();
    m_elements[12] =  tra.x;
	m_elements[13] =  tra.y;
	m_elements[14] = tra.z;
}

// Construct a scale of a given vector, with orientation before scale.
// Victor Begin {
Matrix::Matrix(const Vec3f& sca, Rotation& scaOri)
{
    Matrix m2;
	m2.LoadIdentityTransform();
    m2.m_elements[0] =  sca.x;
	m2.m_elements[5] =  sca.y;
	m2.m_elements[10] = sca.z;
	Matrix m1=Matrix(scaOri);
	Matrix m12= m1 * m2;
	scaOri.theta = scaOri.theta * -1;
	Matrix m3=Matrix(scaOri);
	Matrix m13=m12 * m3;
	LoadIdentityTransform();
	for(int i = 0;i < 15;i++)
		m_elements[i] =  m13.m_elements[i];
}
// } Victor End

Matrix Matrix::operator*=(const Matrix&m)
{
	Matrix ret = *this;
	int a,b;

	for (int i = 0; i < 16; i++){
	  float subt = 0;
	  int ii=(int)i/4;
	  for (int k = 0; k < 4; k++){
		  a=k*4+i-4*ii;
		  b=k+4*ii;
		  subt += ret.m_elements[a] * m.m_elements[b];
	  }
	   m_elements[i] = subt;
	}
/*
	for (int i = 4; i < 8; i++)
	  float subt = 0;
	  for (int k = 0; k < 4; k++)
		subt += ret.m_elements[k+4] * m.m_elements[k*4+i-4];
	   m_elements[i] = subt;
	}

	for (int i = 8; i < 12; i++)
	  float subt = 0;
	  for (int k = 0; k < 4; k++)
		subt += ret.m_elements[k+8] * m.m_elements[k*4+i-8];
	   m_elements[i] = subt;
	}
	for (int i = 12; i < 16; i++)
	  float subt = 0;
	  for (int k = 0; k < 4; k++)
		subt += ret.m_elements[k+12] * m.m_elements[k*4+i-12];
	   m_elements[i] = subt;
	}
	*/
	return *this;
}

Matrix Matrix::operator*(const Matrix&m)
{
	Matrix rez ;
	int a,b;
	
	for (int i = 0; i < 16; i++){
	  float subt = 0;
	  int ii=(int)(i/4);
	  for (int k = 0; k < 4; k++){
  		  a=k*4+i-4*ii;
		  b=k+4*ii;
		  subt += m_elements[a] * m.m_elements[b];
	  }
	   rez.m_elements[i] = subt;
	}
	return rez;
}

BOOL Matrix::operator==(const Matrix& m)
{
  if(m_elements[0] == m.m_elements[0] && m_elements[1] == m.m_elements[1] &&
     m_elements[2] == m.m_elements[2] && m_elements[3] == m.m_elements[3] &&
     m_elements[4] == m.m_elements[4] && m_elements[5] == m.m_elements[5] &&
     m_elements[6] == m.m_elements[6] && m_elements[7] == m.m_elements[7] &&
     m_elements[8] == m.m_elements[8] && m_elements[9] == m.m_elements[9] &&
     m_elements[10] == m.m_elements[10] && m_elements[11] == m.m_elements[11] &&
     m_elements[12] == m.m_elements[12] && m_elements[13] == m.m_elements[13] &&
     m_elements[14] == m.m_elements[14] && m_elements[15] == m.m_elements[15])
     return TRUE;
  return FALSE;
}
#ifdef _DEBUG
    
// Debugging output

void
Matrix::Print(FILE* pStream)
{
  fprintf(pStream, "%f, %f, %f, %f\n%f, %f, %f, %f\n%f, %f, %f, %f\n%f, %f, %f, %f\n",
	  m_elements[ 0], m_elements[ 4], m_elements[ 8], m_elements[12], 
	  m_elements[ 1], m_elements[ 5], m_elements[ 9], m_elements[13], 
	  m_elements[ 2], m_elements[ 6], m_elements[10], m_elements[14], 
	  m_elements[ 3], m_elements[ 7], m_elements[11], m_elements[15]);
}

#endif

Matrix Matrix::Inverse()
{
	Matrix ret;
	ret.LoadIdentityTransform();
	float det=this->GetDeterminant();
	ret.m_elements[0]=(m_elements[5]*m_elements[10]-m_elements[9]*m_elements[6])/det;
	ret.m_elements[1]=(m_elements[2]*m_elements[9]-m_elements[1]*m_elements[10])/det;
	ret.m_elements[2]=(m_elements[1]*m_elements[6]-m_elements[2]*m_elements[5])/det;
	ret.m_elements[4]=(m_elements[6]*m_elements[8]-m_elements[4]*m_elements[10])/det;
	ret.m_elements[5]=(m_elements[0]*m_elements[10]-m_elements[2]*m_elements[8])/det;
	ret.m_elements[6]=(m_elements[4]*m_elements[2]-m_elements[0]*m_elements[6])/det;
	ret.m_elements[8]=(m_elements[4]*m_elements[9]-m_elements[5]*m_elements[8])/det;
	ret.m_elements[9]=(m_elements[1]*m_elements[8]-m_elements[0]*m_elements[9])/det;
	ret.m_elements[10]=(m_elements[0]*m_elements[5]-m_elements[1]*m_elements[4])/det;
	ret.m_elements[12]=-(m_elements[4]*m_elements[9]*m_elements[14]+m_elements[5]*m_elements[10]*m_elements[12]+m_elements[8]*m_elements[13]*m_elements[6]-m_elements[6]*m_elements[9]*m_elements[12]-m_elements[13]*m_elements[10]*m_elements[4]-m_elements[5]*m_elements[8]*m_elements[14]);
	ret.m_elements[13]=(m_elements[0]*m_elements[9]*m_elements[14]+m_elements[1]*m_elements[10]*m_elements[12]+m_elements[8]*m_elements[13]*m_elements[2]-m_elements[2]*m_elements[9]*m_elements[12]-m_elements[13]*m_elements[10]*m_elements[0]-m_elements[1]*m_elements[8]*m_elements[14]);
	ret.m_elements[14]=-(m_elements[0]*m_elements[5]*m_elements[14]+m_elements[1]*m_elements[6]*m_elements[12]+m_elements[4]*m_elements[13]*m_elements[2]-m_elements[2]*m_elements[5]*m_elements[12]-m_elements[13]*m_elements[6]*m_elements[0]-m_elements[1]*m_elements[4]*m_elements[14]);
	return ret;
}
float Matrix::GetDeterminant()
{
	return m_elements[0]*m_elements[5]*m_elements[10]+m_elements[4]*m_elements[9]*m_elements[2]+m_elements[1]*m_elements[6]*m_elements[8]-m_elements[2]*m_elements[5]*m_elements[8]-m_elements[6]*m_elements[9]*m_elements[0]-m_elements[1]*m_elements[4]*m_elements[10];
}