// -*- c++ -*-

// This software module was originally developed by Karl Anders ygard
// (Telenor R&D) in the course of development of the MPEG-4 Systems
// standard (ISO/IEC 14496-1).  This software module is an implementation
// of a part of one or more MPEG-4 Systems tools as specified by the MPEG-4
// Systems standard.  ISO/IEC gives users of the MPEG-4 Systems standard
// free license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4 Systems
// standard.  Those intending to use this software module in hardware or
// software products are advised that its use may infringe existing
// patents.  The original developer of this software module and his/her
// company, the subsequent editors and their companies, and ISO/IEC have no
// liability for use of this software module or modifications thereof in an
// implementation.  Copyright is not released for non MPEG-4 Systems
// conforming products.  The original developer retains full right to use
// the code for his/her own purpose, assign or donate the code to a third
// party and to inhibit third parties from using the code for non MPEG-4
// Systems conforming products.  This copyright notice must be included in
// all copies or derivative works.
//
// Copyright 1997, 1998.

#include "OpenGl/OpenGl.h"
#include "MediaObjectProxy.h"
#include "opengl/base/OpenGLPresenter.h"
#include "opengl/base/OpenGLVisualRenderer.h"

// SOLARIS PORTING - Removed "::" in the gl calls

// Get object to world matrix

/* virtual */ void
MediaObjectProxy::ObjectToWorldMatrix(DWORD dwTime, Matrix& rMatrix) const
{
  // use the Node ptr to climb up the graph to the Root
  // to do this, just use the first parent, and keep going

  MediaObject * pNode = (MediaObject *) GetMediaObject();

  // are we at the root of the scene graph?
  if (pNode->GetParentsCount() > 0)
    // no, keep going
    ((MediaObjectProxy *) pNode->GetParent(0)->GetRenderingObject())->
        ObjectToWorldMatrix(dwTime, rMatrix);
}

// Get viewpoint to world matrix

/* virtual */ void
MediaObjectProxy::ViewpointToWorldMatrix(DWORD dwTime, Matrix& rMatrix) const
{
  // use the Node ptr to climb up the graph to the Root
  // to do this, just use the first parent, and keep going

  MediaObject * pNode = (MediaObject *) GetMediaObject();
  /* We only bother with the first parent; LoadProjectionMatrix() will
     only be invoked in relation to Viewpoints, and Viewpoints may not be
     instanced multiply. */
  // are we at the root of the scene graph?
  
  if (! pNode->IsRoot())
  {
    // no, keep going
    ((MediaObjectProxy *) pNode->GetParent(0)->GetRenderingObject())->
        ViewpointToWorldMatrix(dwTime, rMatrix);
  }
}

// Render object description box

void MediaObjectProxy::RenderDescriptionBox(
  OpenGLVisualRenderer* pRenderer, 
  const char* pString) const
{
  GLfloat color[] = { 0.0f, 0.0f, 0.0f, 0.75f};
  char buffer[128];
  float height;
  float width;
    
  if (pRenderer->DrawSceneGraph() == -1) {
    ::sprintf(buffer, "%08x", GetMediaObject());
    pString = buffer;
  }

  height = pRenderer->GetFontHeight();
  width = pRenderer->GetStringWidth(pString);

  ::glRectf(0.1f, 0.1f, 0.9f, 0.9f);

  ::glPushMatrix();
  ::glPushAttrib(GL_CURRENT_BIT);

  ::glTranslatef(0.5f, 0.25f, 0.0f);

  float scale = 1.0f / (7.0f * height);

  ::glScalef(scale / ((width * scale <= 0.7f) ? 1.0f : (width * scale) / 0.7f),
             1.0f / (2.0f * height),
             1.0f);

  ::glTranslatef(-width / 2.0f,
		 pRenderer->GetFontBaseline(),
		 0.0f);

  glColor4fv(color);

  pRenderer->RenderString(pString);

  ::glPopAttrib();
  ::glPopMatrix();
}


// Render object description

/* virtual */ void MediaObjectProxy::RenderDescription(
  OpenGLVisualRenderer* pRenderer, 
  DWORD /* dwTime */)
{
  GLfloat color[] = { 0.0f, 1.0f, 1.0f, 0.75f};

  glColor4fv(color);

  RenderDescriptionBox(pRenderer, "MediaObject");
}


// Collision detection

MediaObject * MediaObjectProxy::DetectCollision(
  const SF<Vec3f> & rTransformedSize, 
  const SF<Vec3f> & rTransformedCenter,
  const SF<Vec3f> & rSize, 
  const SF<Vec3f> & rCenter,
  const Matrix& rTransform,
  const MediaObject * /* pCallingNode */)
{
  MediaObject * poMediaObject = GetMediaObject();

  for (int i = /* m_parents.GetLength() */ poMediaObject->GetParentsCount() - 1; i >= 0; --i)
  {
    MediaObject * pIntersectingNode;
    
    pIntersectingNode = /* m_parents.ElementAt */ 
      ((MediaObjectProxy *) poMediaObject->GetParent(i)->GetRenderingObject())->
        DetectCollision(
          rTransformedSize, 
          rTransformedCenter,
          rSize, 
          rCenter,
          rTransform,
          GetMediaObject() /* this */ );
    
    if (pIntersectingNode)
      return pIntersectingNode;
  }
  
  return NULL;
}


// Render object

/* virtual */ BOOL MediaObjectProxy::Render(
  const Matrix& /* rTransformation */, 
  const CullFlag& /* rCullFace */,
	OpenGLVisualRenderer* pRenderer, 
  DWORD /* dwTime */)
{
  // SB #### do the rendering without using MediaObject vars/methods

  if (pRenderer->DrawBBoxes())
  {
    /* MediaObject * pNode = (MediaObject *) GetMediaObject(); */

    Vec3f size;
    Vec3f center;

// SOLARIS PORTING
// Created new objects to avoid memory dereferencing
  SF<Vec3f> sf_size, sf_center;
  sf_size = (SF<Vec3f>) size;
  sf_center = (SF<Vec3f>) center;

    /* pNode-> */ GetBbox( sf_size, sf_center);

  size = (Vec3f)sf_size;
  center = (Vec3f)sf_center;

// SOLARIS PORTING - END

    ::glPushAttrib(GL_FOG_BIT);
    ::glDisable(GL_FOG);

    ::glPushMatrix();
    ::glTranslatef(center.x, center.y, center.z);

    if (size.x > -1)
    {
      size *= 0.5f;

      ::glBegin(GL_LINE_STRIP);
      glVertex3f( size.x,  size.y,  size.z);
      glVertex3f(-size.x,  size.y,  size.z);
      glVertex3f(-size.x, -size.y,  size.z);
      glVertex3f( size.x, -size.y,  size.z);
	       			  
      glVertex3f( size.x,  size.y,  size.z);
      glVertex3f( size.x,  size.y, -size.z);
      glVertex3f( size.x, -size.y, -size.z);
      glVertex3f( size.x, -size.y,  size.z);
      ::glEnd();			  
				  
      ::glBegin(GL_LINE_STRIP);	  
      glVertex3f(-size.x, -size.y,  size.z);
      glVertex3f(-size.x, -size.y, -size.z);
      glVertex3f( size.x, -size.y, -size.z);
      ::glEnd();			  
				  
      ::glBegin(GL_LINE_STRIP);	  
      glVertex3f(-size.x,  size.y,  size.z);
      glVertex3f(-size.x,  size.y, -size.z);
      glVertex3f( size.x,  size.y, -size.z);
      ::glEnd();			  
				  
      ::glBegin(GL_LINES);		  
      glVertex3f(-size.x,  size.y, -size.z);
      glVertex3f(-size.x, -size.y, -size.z);
      ::glEnd();
    }

    ::glBegin(GL_POINTS);
    glVertex3f(0.0f, 0.0f, 0.0f);
    ::glEnd();

    ::glPopMatrix();
    ::glPopAttrib();
  }
  
  return TRUE;
}

// Test if object is inside frustum

CullFlag MediaObjectProxy::InsideFrustum(
  const Matrix& transform,
	OpenGLVisualRenderer* pRenderer)
{
  /* MediaObject * pNode = (MediaObject *) GetMediaObject(); */

  const Vec3f& left = pRenderer->GetLeftFrustumPlane();
  const Vec3f& right = pRenderer->GetRightFrustumPlane();
  const Vec3f& top = pRenderer->GetTopFrustumPlane();
  const Vec3f& bottom = pRenderer->GetBottomFrustumPlane();

  SF<Vec3f> sizeSF, centerSF;
  GetTransformedBbox(sizeSF, centerSF);
  Vec3f size = sizeSF;
  Vec3f center = centerSF;

  Vec3f bll; // back, lower, left
  Vec3f fur; // front, upper, right
  fur = bll = transform.MultiplyPoint(center);

  Vec3f x (transform[0], transform[1], transform[ 2]);
  Vec3f y (transform[4], transform[5], transform[ 6]);
  Vec3f z (transform[8], transform[9], transform[10]);

  x *= size.x;
  y *= size.y;
  z *= size.z;

  size.x = ::fabs(x.x) + ::fabs(y.x) + ::fabs(z.x);
  size.y = ::fabs(x.y) + ::fabs(y.y) + ::fabs(z.y);
  size.z = ::fabs(x.z) + ::fabs(y.z) + ::fabs(z.z);

  size *= 0.5f;

  bll -= size;
  fur += size;

  /***
  {
     // Draw frustum culling bounding boxes

    ::glPushMatrix();
    ::glLoadIdentity();

    ::glBegin(GL_LINE_STRIP);
    glVertex3f(fur.x, fur.y, fur.z);
    glVertex3f(bll.x, fur.y, fur.z);
    glVertex3f(bll.x, bll.y, fur.z);
    glVertex3f(fur.x, bll.y, fur.z);
	       			  
    glVertex3f(fur.x, fur.y, fur.z);
    glVertex3f(fur.x, fur.y, bll.z);
    glVertex3f(fur.x, bll.y, bll.z);
    glVertex3f(fur.x, bll.y, fur.z);
    ::glEnd();			  
				  
    ::glBegin(GL_LINE_STRIP);	  
    glVertex3f(bll.x, bll.y, fur.z);
    glVertex3f(bll.x, bll.y, bll.z);
    glVertex3f(fur.x, bll.y, bll.z);
    ::glEnd();			  
				  
    ::glBegin(GL_LINE_STRIP);	  
    glVertex3f(bll.x, fur.y, fur.z);
    glVertex3f(bll.x, fur.y, bll.z);
    glVertex3f(fur.x, fur.y, bll.z);
    ::glEnd();			  
				  
    ::glBegin(GL_LINES);		  
    glVertex3f(bll.x, fur.y, bll.z);
    glVertex3f(bll.x, bll.y, bll.z);
    ::glEnd();
    ::glPopMatrix();
  }
  ***/
  
  // First, check if behind frustum
  if (bll.z < 0.0f)
  {
    CullFlag intersection = COMPLETELY_INSIDE;
    Vec3f fll = bll;
    Vec3f bur = fur;
    
    bur.z = bll.z;
    fll.z = fur.z;
    
    // In front of frustum

    if (bll.y > 0.0f)
    {
      if (bll * top > 0.0f)
        return COMPLETELY_OUTSIDE;
    }
    else
    {
      if (fll * bottom > 0.0f)
        intersection = PARTIALLY_INSIDE;
    }
    if (bur.y < 0.0f)
    {
      if (bur * bottom > 0.0f)
        return COMPLETELY_OUTSIDE;
    }
    else
    {
      if (intersection == COMPLETELY_INSIDE)
        if (fur * top > 0.0f)
          intersection = PARTIALLY_INSIDE;
    }
    
    if (bll.x > 0.0f)
    {
      if (bll * right > 0.0f)
        return COMPLETELY_OUTSIDE;
    }
    else 
    {
      if (intersection == COMPLETELY_INSIDE)
        if (fll * left > 0.0f) 
          intersection = PARTIALLY_INSIDE;
    }
    if (bur.x < 0.0f)
    {
      if (bur * left > 0.0f)
        return COMPLETELY_OUTSIDE;
    }
    else 
    {
      if (intersection == COMPLETELY_INSIDE)
        if (fur * right > 0.0f)
          intersection = PARTIALLY_INSIDE;
    }
    return intersection;
  }
  else
    return COMPLETELY_OUTSIDE;
}

// Update bounding box

void MediaObjectProxy::UpdateBbox()
{
  // for this NodeProxy:
  // get the number of Children;
  // initialize Size and Center with the BBox of the first Child;
  // check all other Children and get the minimum common BBox including all

  // SB #### $$$$
  int i;
  /***
  i = GetChildren().GetLength() - 1;
  ***/

  if (i > 0)
  {
    Vec3f center;
    Vec3f size;
  // SB ####
  /***
    GetChildren()[--i]->
      GetTransformedBbox((SF<Vec3f> &) size, (SF<Vec3f> &) center);
  ***/
    
    Vec3f fur(center);
    Vec3f bll(center);
    if (size.x > -1.0f)
    {
      size *= 0.5f;
      fur += size;
      bll -= size;
    }
    
    while (i > 0)
    {
  // SB ####
  /***
      GetChildren()[--i]->
        GetTransformedBbox((SF<Vec3f> &) size, (SF<Vec3f> &) center);
  ***/
      if (size.x > 0.0f)
      {
        float tmp;
        size *= 0.5f;
        if (fur.x < (tmp = center.x + size.x)) 
          fur.x = tmp;
        if (bll.x > (tmp = center.x - size.x)) 
          bll.x = tmp;
        if (fur.y < (tmp = center.y + size.y))
          fur.y = tmp;
        if (bll.y > (tmp = center.y - size.y)) 
          bll.y = tmp;
        if (fur.z < (tmp = center.z + size.z)) 
          fur.z = tmp;
        if (bll.z > (tmp = center.z - size.z)) 
          bll.z = tmp;
      }
      else 
      {
        if (fur.x < center.x) 
          fur.x = center.x;
        if (bll.x > center.x)
          bll.x = center.x;
        if (fur.y < center.y)
          fur.y = center.y;
        if (bll.y > center.y)
          bll.y = center.y;
        if (fur.z < center.z)
          fur.z = center.z;
        if (bll.z > center.z)
          bll.z = center.z;
      }
    }      
    
    Vec3f new_size(fur);
    Vec3f new_center(fur);
    new_size -= bll;
    new_center += bll;
    new_center *= 0.5;
    
// SOLARIS PORTING
// Created new objects to avoid memory dereferencing
  SF<Vec3f> sf_size, sf_center;
  sf_size = (SF<Vec3f>) size;
  sf_center = (SF<Vec3f>) center;

    /* pNode-> */ GetBbox( sf_size, sf_center);

  size = (Vec3f)sf_size;
  center = (Vec3f)sf_center;

// SOLARIS PORTING - END

    if (new_center != center || new_size != size) 
    {
      // SOLARIS PORTING
      // Created new objects to avoid memory dereferencing
      SF<Vec3f> sf_new_size, sf_new_center;
      sf_new_size = (SF<Vec3f>) new_size;
      sf_new_center = (SF<Vec3f>) new_center;

      SetBbox( sf_new_size, sf_new_center);	

      new_size = (Vec3f)sf_size;
      new_center = (Vec3f)sf_center;

      // SOLARIS PORTING - END

    }
  }
  else
  {
	//ML - changed this line to remove a warning message
    //SetBbox((SF<Vec3f> &) Vec3f(), (SF<Vec3f> &) Vec3f());
    // SOLARIS PORTING
	SF<Vec3f> tmp;
    SetBbox( tmp, tmp);
    // SOLARIS PORTING - END
  }
}
