/***************************************************************************/
/*                                                                         */
/* BodyProxy Class Implementation										   */
/*                                                                         */
/***************************************************************************/
/*                                                                         */
/* Purpose : The BodyProxy is an the interpreter of the Body Node. It is   */
/* in charge of managing the main body functions concerning Body Animation */
/*                                                                         */
/* Authors : Fabrice Vergnenegre, Tolga Capin, LIG - EPFL                  */
/*                                                                         */
/***************************************************************************/

// includes for that class implementation
#include "OpenGl/OpenGl.h"
#include "BodyProxy.h"
//#include "defs.h"
//#include "opengl/base/OpenGLPresenter.h"
#include "opengl/fba/BAPProxy.h"
#include "opengl/fba/BDPProxy.h"
#include "opengl/group/TransformProxy.h"
#include "opengl/geometry/SphereProxy.h"
#include "opengl/geometry/ShapeProxy.h"
//#include "Nodes.h"

#include "HAnimParser.h"

// defines for that class implementation
#define DEFAULT_VRML_BODY "SphereManSegments.wrl"

/***************************************************************************/
// Constructor - calls the inherited constructor from Node3DProxy
BodyProxy::BodyProxy(Body * pBody)
 : Node3DProxy(pBody),
   firstUpdatePassed(false),
   bodyIsDefinedInBDP(false)
{
}

/***************************************************************************/
// ConstructBody - constructs the hierarchy defining the body under the
// bodySceneGraph node (Node contained in Body node)
void BodyProxy::ConstructBody()
{
	TRACE("[BodyProxy::ConstructBody] Launches parser\n");

	// Launches a VRML-like parser and hierarchy constructor
	HAnimParser defaultBodyParser(DEFAULT_VRML_BODY, GetBody()->GetExecutive());

	TRACE("[BodyProxy::ConstructBody] Parser finished \n");

	poBapProxy = (BAPProxy *) GetBAP()->GetRenderingObject();
	// is the BAPProxy here ?
	if (poBapProxy) {
		TRACE("BAPPROXY IS HERE\n");
		poBapProxy->SetPointerTable(defaultBodyParser.getPointerTable());
	}

	// attaches the created body to the renderedBody Node
    // Tolga: 00/03/24: The following was modified to have second parameter
	//        GetBDP(). I am not sure if this is true.
	// GetBody()->renderedBody.Append(defaultBodyParser.getResultBody());
	GetBody()->renderedBody.Append(defaultBodyParser.getResultBody(), GetBody());
	this->UpdateBbox();

	TRACE("[BodyProxy::ConstructBody] Attached finished\n");
}

/***************************************************************************/
BOOL BodyProxy::Render(	const Matrix& rTransformation,
						const CullFlag& rCullFlag,
						OpenGLVisualRenderer* pRenderer, 
						DWORD dwTime)
{
	TRACE("[BodyProxy::Render]\n");

	// First do an update
	Update(pRenderer, dwTime);

	/*
	MediaObject * pObject = GetFirstChild();
	NodeProxy* pProxy = (NodeProxy*) pObject->GetRenderingObject();
	pProxy->CullOrRender(rTransformation, rCullFlag, pRenderer, dwTime);

	// default function for return value
	return (MediaObjectProxy::Render(rTransformation,rCullFlag,pRenderer, dwTime));*/

	ZArray<NodeProxy*> nodes;
  
	for (MediaObject * pObject = GetFirstChild(); pObject; pObject = GetNextSibling()) {
		NodeProxy* pProxy = (NodeProxy*) pObject->GetRenderingObject();

		int i = 0;
		float drawingOrder = pProxy->GetDrawingOrder();

		while (nodes.GetLength() > i && nodes[i]->GetDrawingOrder() > drawingOrder)
			++i;

		nodes.InsertAt(i, pProxy);
	}

	for (int j = 0; j < nodes.GetLength(); ++j) {
    TRACE("[BodyProxy::Render] Cull/Render node %4d \n", j);
    nodes[j]->CullOrRender(rTransformation, rCullFlag,
			   pRenderer, dwTime);
	}

	return Node3DProxy::Render(rTransformation, rCullFlag, pRenderer, dwTime);
}

/***************************************************************************/
void BodyProxy::Select(	OpenGLVisualRenderer* pRenderer, 
						DWORD dwTime,
						MediaObjectProxy*& intersectingObject, 
						unsigned int& depth)
{
	TRACE("[BodyProxy::Select - 1] \n");
	for (MediaObject * pObject = GetFirstChild(); pObject; pObject = GetNextSibling())
		((MediaObjectProxy*) pObject->GetRenderingObject())->Select(pRenderer, dwTime, intersectingObject, depth);
}

/***************************************************************************/
BOOL BodyProxy::RenderLights(	OpenGLVisualRenderer* pRenderer, 
								DWORD dwTime)
{
	TRACE("[BodyProxy::RenderLights] \n");
	for (MediaObject * pObject = GetFirstChild(); pObject; pObject = GetNextSibling()) {
		TRACE("[BodyProxy::RenderLights] body child renderlights invoked\n");
		((NodeProxy*) pObject->GetRenderingObject())->RenderLights(pRenderer, dwTime);
	}

	return TRUE;
}

/***************************************************************************/
// Select - checks activation of sensors ?
void BodyProxy::Select(	const Matrix& rTransformation, 
						const CullFlag& rCullFlag,
						OpenGLVisualRenderer* pRenderer, 
						DWORD dwTime,
						MediaObjectProxy*& intersectingObject,
						ZArray<MediaObjectProxy*>& activatedSensors, 
						unsigned int& depth,
						const ZArray<MediaObjectProxy*>& localSensors)
{
	TRACE("[BodyProxy::Select - 2] \n");

	 // Face* pFace = (Face*) GetMediaObject();
	ZArray<MediaObjectProxy*> sensors;
	ZArray<MediaObjectProxy*> nodes;

	// sort objects by drawing order and find pointing device sensors

	for (MediaObject * pObject = GetFirstChild(); pObject; pObject = GetNextSibling()) {
		TRACE("[BodyProxy::Select - 2] Child pass\n");
		MediaObjectProxy* pProxy = (MediaObjectProxy*) pObject->GetRenderingObject();
		int i = 0;
		float drawingOrder = pProxy->GetDrawingOrder();
    
		while (nodes.GetLength() > i && nodes[i]->GetDrawingOrder() > drawingOrder)
			++i;
    
		nodes.InsertAt(i, pProxy);

		if (pProxy->IsPointingDeviceSensor())
			sensors += pProxy;
	}

	// then traverse scene graph further down

	for (int j = 0; j < nodes.GetLength(); ++j) {
		TRACE("[BodyProxy::Select - 2] Node Child Pass\n");
		nodes[j]->CullOrSelect(	rTransformation, rCullFlag,
								pRenderer, dwTime,intersectingObject,
								activatedSensors, depth,
								sensors.GetLength() ? sensors : localSensors);
	}
}

/***************************************************************************/
// Update - called by the presenter. Updates the Body node, must transmit
// the update message to the children (BapProxy and BDPProxy)
BOOL BodyProxy::Update(	VisualRenderer* pRenderer, 
						DWORD dwTime)
{
	TRACE("[BodyProxy::Update] \n");

	// gets a pointer to the BAPPRoxy
	poBapProxy = (BAPProxy *) GetBAP()->GetRenderingObject();

	// checks the BDPPRoxy first
	// if there is a defined body, it is better that it is constructed
	// first
	if ((BDPProxy *) GetBDP()) {
		poBdpProxy = (BDPProxy *) GetBDP()->GetRenderingObject();
		poBdpProxy->Update(GetBody(), dwTime);
	}

	// is the BAPProxy here ?
	if (poBapProxy) {
		poBapProxy->Update(pRenderer, dwTime);
	}

	// first pass and no defined Body => construct the body
	if (!firstUpdatePassed) {

		// IF BDPBody, BDP:Update has already created a body
		// and set the BOOL to TRUE
		if (!bodyIsDefinedInBDP)
			ConstructBody();
		else
			UpdateBbox();

		firstUpdatePassed=true;
	}

	poBapProxy->ApplyBAPValues();

	return TRUE;
}

/***************************************************************************/
// AddChild - Add a child to a MediaObject and returns a pointer to the new
// child appened.
MediaObject * BodyProxy::AddChild(MediaObject * pObjectToAdd, MediaObject * where)
{
	// set the parents of added object
	pObjectToAdd->SetParent(where);

	// return the pointer to the new object added
	return pObjectToAdd;
}

/***************************************************************************/
// RenderDescription - Draws the description of the object when called by
// the main application (View scene graph).
void BodyProxy::RenderDescription(	OpenGLVisualRenderer* pRenderer, 
									DWORD dwTime)
{
	TRACE("[BodyProxy::RenderDescription] \n");

	// draws the body blue box
	GLfloat color[] = { 0.0f, 0.0f, 1.0f, 0.75f};
	// SOLARIS PORTING
	//::glColor4fv(color);	
	glColor4fv(color);	
	RenderDescriptionBox(pRenderer, "Body");

	// saves the OGL transform matrix stack
	::glPushMatrix();

	// scales the further children boxes
	int divisor = ( ( GetBAP() ? 1:0 ) +
		            ( GetBDP() ? 1:0 ) +
					( GetChildCount())
				  );

	::glScalef(1.0f / divisor, 1.0f, 1.0f);
	::glTranslatef(0.0f, -1.0f, 0.0f);

	// first draw the bap child
	poBapProxy = (BAPProxy *) GetBAP()->GetRenderingObject();
	if (poBapProxy) {
		poBapProxy->RenderDescription(pRenderer, dwTime);
		::glTranslatef(1.0f, 0.0f, 0.0f);
	}

	// then the eventual BDP child
	if ((BDPProxy *) GetBDP()) {
		poBdpProxy = (BDPProxy *) GetBDP()->GetRenderingObject();
		poBdpProxy->RenderDescription(pRenderer, dwTime);
		::glTranslatef(1.0f, 0.0f, 0.0f);
	}

	// then the rendered Body node
	for (MediaObject * pObject = GetFirstChild(); pObject ; pObject = GetNextSibling()) {
		TRACE("PASS INTO SIBLING\n");
		NodeProxy* np = (NodeProxy*) pObject->GetRenderingObject();
		np->RenderDescription(pRenderer, dwTime);
		::glTranslatef(1.0f, 0.0f, 0.0f);
	}


	// restores the OGL transform matrix stack
	::glPopMatrix();
}

void BodyProxy::UpdateBbox()
{
  MediaObject * child = GetFirstChild();

  if (child) {
    Vec3f center;
    Vec3f size;
    
    ((MediaObjectProxy *) (child->GetRenderingObject())) ->
      GetTransformedBbox((SF<Vec3f>) size, (SF<Vec3f>) center);
    
    Vec3f fur(center);
    Vec3f bll(center);
    
    if (size.x > -1.0f) {
      size *= 0.5f;
      
      fur += size;
      bll -= size;
    }
    
    while ((child = GetNextSibling())) {
      ((MediaObjectProxy *) (child->GetRenderingObject())) ->
        GetTransformedBbox((SF<Vec3f>) size, (SF<Vec3f>) center);
      
      if (size.x > 0.0f) {
        float tmp;
        
        size *= 0.5f;
        
        if (fur.x < (tmp = center.x + size.x)) 
          fur.x = tmp;
        
        if (bll.x > (tmp = center.x - size.x)) 
          bll.x = tmp;
        
        if (fur.y < (tmp = center.y + size.y))
          fur.y = tmp;
        
        if (bll.y > (tmp = center.y - size.y)) 
          bll.y = tmp;
        
        if (fur.z < (tmp = center.z + size.z)) 
          fur.z = tmp;
        
        if (bll.z > (tmp = center.z - size.z)) 
          bll.z = tmp;
      } else {
        if (fur.x < center.x) 
          fur.x = center.x;
        
        if (bll.x > center.x)
          bll.x = center.x;
        
        if (fur.y < center.y)
          fur.y = center.y;
        
        if (bll.y > center.y)
          bll.y = center.y;
        
        if (fur.z < center.z)
          fur.z = center.z;
        
        if (bll.z > center.z)
          bll.z = center.z;
      }
    }      
    
    Vec3f new_size(fur);
    Vec3f new_center(fur);
    
    new_size -= bll;
    
    new_center += bll;
    new_center *= 0.5;
    
    GetBbox((SF<Vec3f>) size, (SF<Vec3f>) center);

    if (new_center != center || new_size != size) 
      SetBbox((SF<Vec3f>) new_size, (SF<Vec3f>) new_center);	
  }
  else
  {
	SetBbox((SF<Vec3f>) Vec3f(), (SF<Vec3f>) Vec3f());
  }
}
