#include "OpenGl/OpenGl.h"
#include <stdio.h>
#include <stdlib.h>

#include "SegmentParser.h"

//#include "defs.h"
#include "opengl/appearance/AppearanceProxy.h"
#include "opengl/appearance/MaterialProxy.h"
#include "opengl/group/TransformProxy.h"
#include "opengl/geometry/IndexedFaceSetProxy.h"
#include "opengl/geometry/ShapeProxy.h"
//#include "Nodes.h"

#include "Vector3d.h"
#include "MyFace.h"

/**************************************************************************/
Vector3d** SegmentParser::CalculateFaceNormals()
{
	// construct an array of Vector3d* storing face normals
	Vector3d ** faceNrmls = new Vector3d*[nOfFaces];

	// computes normals and fills array
	for (int i=0 ; i<nOfFaces ; i++) {
		Vector3d a(*points[(*faces[i])[1]]-*points[(*faces[i])[0]]);
		Vector3d b(*points[(*faces[i])[2]]-*points[(*faces[i])[1]]);
		Vector3d normal(a^b);
		if (ccw) 
			faceNrmls[i] = new Vector3d(normal);
		else {
			Vector3d inversedNormal = Vector3d(0.f,0.f,0.f) - normal;
			faceNrmls[i] = new Vector3d( inversedNormal);
		}
	}

	// return the computed array
	return faceNrmls;
}

/**************************************************************************/
Vector3d** SegmentParser::CalculateVerticesNormals()
{
	// creates a new array of normals for the vertices
	Vector3d ** verticesNormals = new Vector3d*[nOfPoints];

	// computes face normals
	Vector3d ** faceNormals = CalculateFaceNormals();

	// for each vertex, calculate its average normal
    // PORTING to CC 5.3
    int i = 0;
//	for (int i=0 ; i<nOfPoints ; i++) {
	for ( i=0 ; i<nOfPoints ; i++) {
		Vector3d normal;
		int nbShared = 0;

		// sum the normals that share this vertex
		for (int j=0 ; j<nOfFaces ; j++)
			for (int k=0 ; k<faces[j]->GetNbPoints() ; k++)
				if ( (*faces[j])[k]==i ) {
					normal = normal + (*faceNormals[j]);
					nbShared++;
					break;
				}

		// average the obtained normal
		if (nbShared!=0) {
			normal=normal / (float)nbShared;
		}

		verticesNormals[i] = new Vector3d(normal);
	}

	// wipes out the local face normals - no more use
	for (i=0 ; i<nOfFaces ; i++)
		delete faceNormals[i];
	delete faceNormals;

	// returns the computed array
	return verticesNormals;
}

/**************************************************************************/
//
// IM1 Segment Topology
//
//                      +-------------------+
//                      |    Transform      |
//                      +--------+----------+
//                               |
//                               |     +-----------------+
//                               +---->|      Shape      |
//                                     +-----+------+----+
//                                           |      |
//                        +--------------+   |      |    +----------------+   * normal
//                        |  Appearance  |<--+      +--->| IndexedFaceSet |   * coord
//                        +------+-------+               +----------------+   * coordIndex
//                               |
//                               |     +------------+
//                               +---->|  Material  
//                                     +------------+
//
SegmentParser::SegmentParser(char * filename, Executive * pExecutive, Transform * p)
  : nbConsecutiveIndices(0),
	localSegmentTransform(NULL),
	points(NULL),
	faces(NULL),
	normals(NULL),
	normalIndex(NULL),
	ccw(TRUE),
	creaseAngle(0)
{
	TRACE("[SegmentParser] - Constructor\n");

	// opens the file and saves its filename (for reopen)
	fileName = new char[strlen(filename)+1];
	strcpy(fileName,filename);
	parsedFile = fopen(fileName,"rt");

	// parse the eventual material, if no, then default IM1 Material
	// is constructed (with default values, specified in Core)
	TRACE("[SegmentParser] - Parse Material\n");
	Material * pMaterial = ParseMaterial();

	// parse geometry (points, faces, normals, normalIndex, eventually
	// additional transformations + segment information
	TRACE("[SegmentParser] - ParseGeometry\n");
	ParseGeometry();

	// calculate the vertices normals in case they are not defined for this segment
	if (normals==NULL) {
		TRACE("[SegmentParser] - Generating normals\n");
		normals = CalculateVerticesNormals();
		nOfNormals=nOfPoints;
	}

	// Create the needed Nodes and their Proxies
	// and create a new MaterialProxy for the new Material (parsed or default)
	TRACE("[SegmentParser] - Creating segment local Nodes and their Proxies\n");
	Transform * pTransform = CreateTransformNodeAndItsProxy(pExecutive);
	Shape * pShape = CreateShapeNodeAndItsProxy(pExecutive);
	Appearance * pAppearance = CreateAppearanceNodeAndItsProxy(pExecutive);
	IndexedFaceSet * pIndexedFaceSet = CreateIndexedFaceSetNodeAndItsProxy(pExecutive);
	CreateMaterialProxy(pMaterial, pExecutive);

	// Create a new Coordinate node
	Coordinate * pCoordinate = new Coordinate;
	pCoordinate->SetExecutive(pExecutive);

	// Create a new Normal node
	Normal * pNormal = new Normal;
	pNormal->SetExecutive(pExecutive);

	// append the coordinates points
	MF<Vec3f> coordinates;
	TRACE("[SegmentParser] - Appending %d points\n",nOfPoints);
    // PORTING to CC 5.3
    int i = 0;
//	for (int i=0 ; i<nOfPoints ; i++)
	for ( i=0 ; i<nOfPoints ; i++)
		coordinates.Append(Vec3f(points[i]->GetX(),points[i]->GetY(),points[i]->GetZ()));

	// append the normals
	MF<Vec3f> thenormals;
	TRACE("[SegmentParser] - Appending %d normals\n",nOfNormals);
	for (i=0 ; i<nOfNormals ; i++) {
		Vec3f localNormal(normals[i]->GetX(),normals[i]->GetY(),normals[i]->GetZ());
		localNormal.Normalize();
		thenormals.Append(localNormal);
	}

	// append the indexed faces
	MF<int> indices;
	TRACE("[SegmentParser] - Appending %d faces\n",nOfFaces);
	for (i=0 ; i<nOfFaces ; i++) {
		for (int j=0 ; j<faces[i]->GetNbPoints() ; j++)
			indices.Append((*faces[i])[j]);
		indices.Append(-1);
	}

	// eventually append the indexedNormals
	if (normalIndex!=NULL) {
		TRACE("[SegmentParser] - Appening %d indexedNormals\n",nOfNormalIndex);
		MF<int> normalIndices;

		for (i=0 ; i<nOfNormalIndex; i++) {
			for (int j=0 ; j<normalIndex[i]->GetNbPoints() ; j++)
				normalIndices.Append((*normalIndex[i])[j]);
			normalIndices.Append(-1);
		}
		// Tolga: 00/03/28 changed the following line
		// pIndexedFaceSet->normalIndex = *((NodeField *) &(normalIndices));
		pIndexedFaceSet->normalIndex = normalIndices;
		pIndexedFaceSet->normalPerVertex = FALSE;
	}
	else
		pIndexedFaceSet->normalPerVertex = TRUE;
	
	// link IndexedFaceSet normals
    // Tolga: 00/03/28 changed the following line
	// pNormal->vector = *((NodeField *) &(thenormals));
	pNormal->vector = thenormals;
    //ML - replaced this line with the next
	//pIndexedFaceSet->normal = pNormal;
    pIndexedFaceSet->normal.SetObject(pNormal, pIndexedFaceSet);

	// link IndexedFaceSet coordinates
	pCoordinate->point=coordinates;
    //ML - replaced this line with the next
	//pIndexedFaceSet->coord=pCoordinate;
    pIndexedFaceSet->coord.SetObject(pCoordinate, pIndexedFaceSet);

	// link IndexedFaceSet coordIndex
    // Tolga: 00/03/28 changed the following line
	// pIndexedFaceSet->coordIndex = *((NodeField *) &(indices));
	pIndexedFaceSet->coordIndex = indices;

	// Set IndexedFaceSet values and makes it updated
	pIndexedFaceSet->creaseAngle=creaseAngle;
	pIndexedFaceSet->ccw=ccw;
	pIndexedFaceSet->SetUpdateTime();

	// Create the nested Node hierarchy (See scheme above)
    //ML - replaced these three lines with the next three lines
    //pAppearance->material = pMaterial;
    //pShape->appearance = pAppearance;
	//pShape->geometry=pIndexedFaceSet;
    pAppearance->material.SetObject(pMaterial, pAppearance);
    pShape->appearance.SetObject(pAppearance, pShape);
	pShape->geometry.SetObject(pIndexedFaceSet, pShape);

	// Has this segment local transformations ?
	// normally, the ParseEventualTransformations has been processed before
	// by a call to ParseGeometry
	if (localSegmentTransform!=NULL)
		ApplySegmentTransformations(pTransform);

	// Complete this segment local hierarchy by linking shape to the transf. node
	pTransform->AddChild(pShape);

	// link local Transform to the global Joint Transform
	p->AddChild(pTransform);
	pTransform->SetParent(p);
}

/**************************************************************************/
Transform * SegmentParser::CreateTransformNodeAndItsProxy(Executive * pExecutive)
{
	Transform * pTransform = new Transform();
	TransformProxy * pTransformProxy = new TransformProxy(pTransform);
	pTransform->SetRenderingObject(pTransformProxy);
	pTransform->SetExecutive(pExecutive);
	// Tolga: 00/03/28. Commented out the following. Core is updated so it may work
	// pTransform->CreateChildren();			// to remove when core debugged
	return pTransform;
}

/**************************************************************************/
Shape * SegmentParser::CreateShapeNodeAndItsProxy(Executive * pExecutive)
{
	Shape * pShape = new Shape();
	ShapeProxy * pShapeProxy = new ShapeProxy(pShape);
	pShape->SetRenderingObject(pShapeProxy);
	pShape->SetExecutive(pExecutive);
	return pShape;
}

/**************************************************************************/
Appearance * SegmentParser::CreateAppearanceNodeAndItsProxy(Executive * pExecutive)
{
	Appearance * pAppearance = new Appearance();
	AppearanceProxy * pAppearanceProxy = new AppearanceProxy(pAppearance);
	pAppearance->SetRenderingObject(pAppearanceProxy);
	pAppearance->SetExecutive(pExecutive);
	return pAppearance;
}

/**************************************************************************/
IndexedFaceSet * SegmentParser::CreateIndexedFaceSetNodeAndItsProxy(Executive * pExecutive)
{
	IndexedFaceSet * pIndexedFaceSet = new IndexedFaceSet();
	IndexedFaceSetProxy * pIndexedFaceSetProxy = new IndexedFaceSetProxy(pIndexedFaceSet);
	pIndexedFaceSet->SetRenderingObject(pIndexedFaceSetProxy);
	pIndexedFaceSet->SetExecutive(pExecutive);
	return pIndexedFaceSet;
}

/**************************************************************************/
void SegmentParser::CreateMaterialProxy(Material * pMaterial, Executive * pExecutive)
{
	MaterialProxy * pMaterialProxy = new MaterialProxy(pMaterial);
	pMaterial->SetRenderingObject(pMaterialProxy);
	pMaterial->SetExecutive(pExecutive);
}

/**************************************************************************/
SegmentParser::~SegmentParser()
{
	TRACE("[SegmentParser] - Destructor\n");

	// close the file
	fclose(parsedFile);

	// cleans the points
    // PORTING to CC 5.3
    int i = 0;
//	for (int i=0 ; i<nOfPoints ; i++)
	for ( i=0 ; i<nOfPoints ; i++)
		delete points[i];
	delete points;

	// cleans the faces
	for (i=0 ; i<nOfFaces ; i++)
		delete faces[i];
	delete faces;

	// cleans the normals
	for (i=0 ; i<nOfNormals ; i++)
		delete normals[i];
	delete normals;

	// cleans the normalIndex
	for (i=0 ; i<nOfNormalIndex ; i++)
		delete normalIndex[i];
	delete normalIndex;
}

/**************************************************************************/
void SegmentParser::ApplySegmentTransformations(Transform* pTransform)
{
	TRACE("[SegmentParser] - ApplySegmentTransformations\n");

	// Is there a translation ?
	if (localSegmentTransform->translation!=NULL) {
		pTransform->translation=(Vec3f( localSegmentTransform->translation[0],
										localSegmentTransform->translation[1],
										localSegmentTransform->translation[2]));
	}

	// And a rotation ?
	if (localSegmentTransform->rotation!=NULL) {
		pTransform->rotation=(Rotation( localSegmentTransform->rotation[0],
										localSegmentTransform->rotation[1],
										localSegmentTransform->rotation[2],
										localSegmentTransform->rotation[3]));
	}

	// A scaling ?
	if (localSegmentTransform->scale!=NULL) {
		pTransform->scale=(Vec3f(	localSegmentTransform->scale[0],
									localSegmentTransform->scale[1],
									localSegmentTransform->scale[2]));
		TRACE("TRANSFORMATION SCALE %f %f %f\n",localSegmentTransform->scale[0],
									localSegmentTransform->scale[1],
									localSegmentTransform->scale[2]);
	}

	// A scaleorientation ?
	if (localSegmentTransform->scaleorientation!=NULL) {
		pTransform->scaleOrientation=(Rotation( localSegmentTransform->scaleorientation[0],
												localSegmentTransform->scaleorientation[1],
												localSegmentTransform->scaleorientation[2],
												localSegmentTransform->scaleorientation[3]));
	}
}

/**************************************************************************/
// Returns a string without crap (ie spaces and tabs) in the beginning
// of the given char*
char * SegmentParser::RemoveCrap(char * s)
{
	char * pointer = s;

	// moves pointer until a nonspace or nontab car.
	while ((*pointer==' ') || (*pointer=='\t'))
		pointer++;

	// returns the pointer
	return pointer;
}

/**************************************************************************/
// Place the file handle to the line beginning with the given char *
void SegmentParser::GotoLine(char * s)
{
	char * cleanBuffer;

	// parse until find the string
	do {
		fgets(buffer,SIZE_OF_PARSING_BUFFER,parsedFile);
		cleanBuffer=RemoveCrap(buffer);
	} while ( (!feof(parsedFile)) && (strncmp(s,cleanBuffer,strlen(s))) );
}

/**************************************************************************/
Material * SegmentParser::ParseMaterial()
{
	char * cleanBuffer;
	float a,s,t,dif[3],spe[3],emi[3];

	// constructs a new material (default values)
	Material * m = new Material;

	// first, move to the material definition in the file
	GotoLine("material");

	// parse material until finished
	int stop=-1;
	while (stop==-1) {
		fgets(buffer,SIZE_OF_PARSING_BUFFER,parsedFile);
		cleanBuffer=RemoveCrap(buffer);

		if (sscanf(cleanBuffer,"ambientIntensity %f\n",&a)!=1)
			if (sscanf(cleanBuffer,"diffuseColor %f %f %f\n",&(dif[0]),&(dif[1]),&(dif[2])) !=3)
				if (sscanf(cleanBuffer,"specularColor %f %f %f\n",&(spe[0]),&(spe[1]),&(spe[2]))!=3)
					if (sscanf(cleanBuffer,"emissiveColor %f %f %f\n",&(emi[0]),&(emi[1]),&(emi[2]))!=3)
						if (sscanf(cleanBuffer,"shininess %f\n",&s)!=1)
							if (sscanf(cleanBuffer,"transparency %f\n",&t)!=1)
								stop=0;
							else m->transparency=0.f;
						else m->shininess=s;
					else m->emissiveColor=Color3f(emi[0],emi[1],emi[2]);
				else m->specularColor=Color3f(spe[0],spe[1],spe[2]);
			else m->diffuseColor=Color3f(dif[0],dif[1],dif[2]);
		else m->ambientIntensity=a;
	}

	// return constructed material
	return m;
}

/**************************************************************************/
void SegmentParser::ParseGeometry()
{
	// initializes the points and faces numbers
	nOfPoints=0;
	nOfFaces=0;
	nOfNormals=0;
	nOfNormalIndex=0;

	// first parse the vertices
	ParseCoordinates();

	// then the faces
	ParseFaces();

	// then additional eventual transformations
	ParseEventualTransformations();

	// then eventually the normals
	//ParseNormals();
}

/**************************************************************************/
void SegmentParser::ParseEventualTransformations()
{
	// locals
	char * cleanBuffer;
	float tempo[4];
	BOOL noTransformationFound = TRUE;
	char CCW[8];
	
	// constructs Object ans stes all of his fields
	// to NULL pointers (security). Doing that, the 
	// segment will know what transformations have to
	// be set to the local Transform.
	localSegmentTransform = new MyTransformations;
	localSegmentTransform->translation=NULL;
	localSegmentTransform->rotation=NULL;
	localSegmentTransform->scale=NULL;
	localSegmentTransform->scaleorientation=NULL;

	// browse until end of file
	while (!feof(parsedFile)) {
		fgets(buffer,SIZE_OF_PARSING_BUFFER,parsedFile);
		cleanBuffer=RemoveCrap(buffer);

		// try to match one of the transformations
		if (sscanf(cleanBuffer,"translation %f %f %f\n",
			&tempo[0],&tempo[1],&tempo[2])==3) {
			localSegmentTransform->translation=new float[3];
			localSegmentTransform->translation[0]=tempo[0];
			localSegmentTransform->translation[1]=tempo[1];
			localSegmentTransform->translation[2]=tempo[2];
			noTransformationFound=FALSE;
		}
		else
			if (sscanf(cleanBuffer,"rotation %f %f %f %f\n",
				&tempo[0],&tempo[1],&tempo[2],&tempo[3])==4) {
				localSegmentTransform->rotation=new float[4];
				localSegmentTransform->rotation[0]=tempo[0];
				localSegmentTransform->rotation[1]=tempo[1];
				localSegmentTransform->rotation[2]=tempo[2];
				localSegmentTransform->rotation[3]=tempo[3];
				noTransformationFound=FALSE;
			}
			else
				if (sscanf(cleanBuffer,"scale %f %f %f\n",
					&tempo[0],&tempo[1],&tempo[2])==3) {
					localSegmentTransform->scale=new float[3];
					localSegmentTransform->scale[0]=tempo[0];
					localSegmentTransform->scale[1]=tempo[1];
					localSegmentTransform->scale[2]=tempo[2];
					noTransformationFound=FALSE;
				}
				else
					if (sscanf(cleanBuffer,"scaleOrientation %f %f %f %f\n",
						&tempo[0],&tempo[1],&tempo[2],&tempo[3])==4) {
						localSegmentTransform->scaleorientation=new float[4];
						localSegmentTransform->scaleorientation[0]=tempo[0];
						localSegmentTransform->scaleorientation[1]=tempo[1];
						localSegmentTransform->scaleorientation[2]=tempo[2];
						localSegmentTransform->scaleorientation[3]=tempo[3];
						noTransformationFound=FALSE;
					}
					else
						if (sscanf(cleanBuffer,"creaseAngle %f\n",&creaseAngle)==1)
						{
						}
						else
							if (sscanf(cleanBuffer,"ccw %s\n",CCW)==1)
								if (!strncmp("FALSE",CCW,4)) ccw=FALSE;
	}

	// if none of these transformations were found delete the object
	// and set it to NULL (security)
	
	TRACE("SEGMENT CCW=%d creaseAngle=%4.4f\n",ccw,creaseAngle);

	if (noTransformationFound) {
		delete localSegmentTransform;
		localSegmentTransform=NULL;
	}
}

/**************************************************************************/
void SegmentParser::ProcessNewBufferOfFaces(int * b, int nb)
{
	// parse the buffer
	for (int i=0 ; i<nb ; i++) {

		if (b[i]!=-1)
			bufferOfIndices[nbConsecutiveIndices++]=b[i];
		else {
			if (nbConsecutiveIndices!=0) {
				AddFace(bufferOfIndices,nbConsecutiveIndices);
				nbConsecutiveIndices=0;
			}
		}
	}
}

/**************************************************************************/
void SegmentParser::ProcessNewBufferOfNormalIndex(int * b, int nb)
{
	// parse the buffer
	for (int i=0 ; i<nb ; i++) {

		if (b[i]!=-1)
			bufferOfIndices[nbConsecutiveIndices++]=b[i];
		else {
			if (nbConsecutiveIndices!=0) {
				AddNormalIndex(bufferOfIndices,nbConsecutiveIndices);
				nbConsecutiveIndices=0;
			}
		}
	}
}

/**************************************************************************/
void SegmentParser::ParseNormals()
{
	char * cleanBuffer;
	float x,y,z;
	BOOL normalsSpecified=TRUE;

	// closes and reopens file to be at beginning
	fclose(parsedFile);
	parsedFile=fopen(fileName,"rt");

	// search for normal
	GotoLine("normal");

	// did Gotoline find a line with "normal"
	if (!feof(parsedFile)) {
		// yes, see if field is NULL
		cleanBuffer=RemoveCrap(buffer);
		char * nulled = RemoveCrap(cleanBuffer+6);
		normalsSpecified=(strncmp("NULL",nulled,4));
	}
	else
		normalsSpecified=FALSE;
		
	if (normalsSpecified) {

		TRACE("Normals for that segment are specified\n");
		// yes, so parse them
		normals = new Vector3d*[MAX_NUMBER_OF_NORMALS_PER_SEGMENT];

		// gets first line
		fgets(buffer,SIZE_OF_PARSING_BUFFER,parsedFile);
		cleanBuffer=RemoveCrap(buffer);

		// makes it match for the first line
		sscanf(cleanBuffer,"vector [ %f %f %f,\n",&x,&y,&z);
		AddNormal((float)x,(float)y,(float)z);
		
		int stop = -1;
		// loops for the rest of lines
		do {
			fgets(buffer,SIZE_OF_PARSING_BUFFER,parsedFile);
			cleanBuffer=RemoveCrap(buffer);

			if (sscanf(cleanBuffer,"%f %f %f,\n",&x,&y,&z)!=3)
				stop=0;

			// adds the point
			if (stop!=0)
				AddNormal(x,y,z);

		} while (stop == -1);

		// then parse normal Index
		ParseNormalIndex();
	}
}

/**************************************************************************/
void SegmentParser::ParseNormalIndex()
{
	// local indexed vertices for parsing
	int v[8];

	char * cleanBuffer;

	normalIndex = new MyFace*[MAX_NUMBER_OF_NORMAL_INDEX_PER_SEGMENT];
	
	// Positions the file to the coordIndex field
	GotoLine("normalIndex");

	// makes it match for the first line
	cleanBuffer=RemoveCrap(buffer);
	sscanf(cleanBuffer,"normalIndex [ %d, %d, %d, %d, %d, %d, %d, %d\n",
		&v[0],&v[1],&v[2],&v[3],&v[4],&v[5],&v[6],&v[7]);

	ProcessNewBufferOfNormalIndex(v,8);

	int stop = -1;
	int nb = 0;

	// loops for the rest of lines
	do {
		fgets(buffer,SIZE_OF_PARSING_BUFFER,parsedFile);
		cleanBuffer=RemoveCrap(buffer);

		if (sscanf(cleanBuffer,"%d, %d, %d, %d, %d, %d, %d, %d ]\n",
			&v[0],&v[1],&v[2],&v[3],&v[4],&v[5],&v[6],&v[7]
			)!=8)
		if (sscanf(cleanBuffer,"%d, %d, %d, %d, %d, %d, %d ]\n",
			&v[0],&v[1],&v[2],&v[3],&v[4],&v[5],&v[6]
			)!=7)
		if (sscanf(cleanBuffer,"%d, %d, %d, %d, %d, %d ]\n",
			&v[0],&v[1],&v[2],&v[3],&v[4],&v[5]
			)!=6)
		if (sscanf(cleanBuffer,"%d, %d, %d, %d, %d ]\n",
			&v[0],&v[1],&v[2],&v[3],&v[4]
			)!=5)
		if (sscanf(cleanBuffer,"%d, %d, %d, %d ]\n",
			&v[0],&v[1],&v[2],&v[3]
		    )!=4)
		if (sscanf(cleanBuffer,"%d, %d, %d ]\n",
			&v[0],&v[1],&v[2]
			)!=3)
		if (sscanf(cleanBuffer,"%d, %d ]\n",
			&v[0],&v[1]
			)!=2)
		if (sscanf(cleanBuffer,"%d ]\n",
			&v[0]
			)!=1)
		    stop=0;
		else nb=1;
		else nb=2;
	    else nb=3;
	    else nb=4;
		else nb=5;
		else nb=6;
		else nb=7;
		else nb=8;

		// adds the point
		if (stop!=0)
			ProcessNewBufferOfNormalIndex(v,nb);
	} while (stop == -1);
}

/**************************************************************************/
void SegmentParser::ParseCoordinates()
{
	// local buffers for reading adn parsing
	char * cleanBuffer;

	// local floats for temporarily storing the points coordinates
	float x,y,z;

	// Assigns a sufficient space for coordinates
	points = new Vector3d*[MAX_NUMBER_OF_POINTS_PER_SEGMENT];

	// Positions the file to the Coordinate field
	GotoLine("coord");

	// gets a line and removes spaces
	fgets(buffer,SIZE_OF_PARSING_BUFFER,parsedFile);
	cleanBuffer=RemoveCrap(buffer);

	// makes it match for the first line
	sscanf(cleanBuffer,"point [ %f %f %f,\n",&x,&y,&z);
	AddPoint((float)x,(float)y,(float)z);

	int stop = -1;
	// loops for the rest of lines
	do {
		fgets(buffer,SIZE_OF_PARSING_BUFFER,parsedFile);
		cleanBuffer=RemoveCrap(buffer);

		if (sscanf(cleanBuffer,"%f %f %f,\n",&x,&y,&z)!=3)
		stop=0;

		// adds the point
		if (stop!=0)
			AddPoint(x,y,z);

	} while (stop == -1);
}

/**************************************************************************/
void SegmentParser::ParseFaces()
{
	// local buffers for reading and parsing
	char * cleanBuffer;

	// local indexed vertices for parsing
	int v[8];

	// Assigns a sufficient space for faces
	faces = new MyFace*[MAX_NUMBER_OF_TRIANGLES_PER_SEGMENT];

	// Positions the file to the coordIndex field
	GotoLine("coordIndex");

	// makes it match for the first line
	cleanBuffer=RemoveCrap(buffer);
	sscanf(cleanBuffer,"coordIndex [ %d, %d, %d, %d, %d, %d, %d, %d\n",
		&v[0],&v[1],&v[2],&v[3],&v[4],&v[5],&v[6],&v[7]);
	ProcessNewBufferOfFaces(v,8);

	int stop = -1;
	int nb = 0;

	// loops for the rest of lines
	do {
		fgets(buffer,SIZE_OF_PARSING_BUFFER,parsedFile);
		cleanBuffer=RemoveCrap(buffer);

		if (sscanf(cleanBuffer,"%d, %d, %d, %d, %d, %d, %d, %d ]\n",
			&v[0],&v[1],&v[2],&v[3],&v[4],&v[5],&v[6],&v[7]
			)!=8)
		if (sscanf(cleanBuffer,"%d, %d, %d, %d, %d, %d, %d ]\n",
			&v[0],&v[1],&v[2],&v[3],&v[4],&v[5],&v[6]
			)!=7)
		if (sscanf(cleanBuffer,"%d, %d, %d, %d, %d, %d ]\n",
			&v[0],&v[1],&v[2],&v[3],&v[4],&v[5]
			)!=6)
		if (sscanf(cleanBuffer,"%d, %d, %d, %d, %d ]\n",
			&v[0],&v[1],&v[2],&v[3],&v[4]
			)!=5)
		if (sscanf(cleanBuffer,"%d, %d, %d, %d ]\n",
			&v[0],&v[1],&v[2],&v[3]
		    )!=4)
		if (sscanf(cleanBuffer,"%d, %d, %d ]\n",
			&v[0],&v[1],&v[2]
			)!=3)
		if (sscanf(cleanBuffer,"%d, %d ]\n",
			&v[0],&v[1]
			)!=2)
		if (sscanf(cleanBuffer,"%d ]\n",
			&v[0]
			)!=1)
		    stop=0;
		else nb=1;
		else nb=2;
	    else nb=3;
	    else nb=4;
		else nb=5;
		else nb=6;
		else nb=7;
		else nb=8;

		// adds the point
		if (stop!=0)
			ProcessNewBufferOfFaces(v,nb);
	} while (stop == -1);
}
