#include "OpenGl/OpenGl.h"
#include "ColorInterpolatorProxy.h"

ColorInterpolatorProxy::ColorInterpolatorProxy(MediaObject * poNode) :
InterpolatorNodeProxy(poNode)
{}

inline BOOL ColorInterpolatorProxy::SetValue(
    ColorInterpolator * poCI, int iIndex)
{
  poCI->value_changed = poCI->keyValue[iIndex];
  return TRUE;
}

BOOL ColorInterpolatorProxy::Evaluate()
{
  ColorInterpolator * poCI = 
    (ColorInterpolator *) GetMediaObject();

  // Check if we have any keys
  if (poCI->key.GetLength() > 0)
  {
    // If we only have one key
    if (poCI->key.GetLength() == 1)
    {
      return SetValue(poCI, 0);
    }
    else // more keys
    {
      // If set_fraction is smaller than smallest key
      if (poCI->set_fraction <= poCI->key[0])
      {
        return SetValue(poCI, 0);
      }
      else
      {
        // If set_fraction is larger than largest key
        if (poCI->set_fraction >= poCI->key[poCI->key.GetLength() - 1])
        {
          return SetValue(poCI, poCI->key.GetLength() - 1);
        }
        else
        {
          // Search the key space for fraction
          for (int i = poCI->key.GetLength() - 2; i >= 0 ; --i)
          {
            if (poCI->key[i] == poCI->set_fraction)
            {
              // Found exact match
              return SetValue(poCI, i);
            }
            else
            {
              if (poCI->key[i] < poCI->set_fraction)
              {
                // Didn't find exact match; must interpolate
                return InterpolateValue( 
                  i, i + 1, 
                  ((poCI->key[i + 1] - poCI->set_fraction) / 
                  (poCI->key[i + 1] - poCI->key[i])) );
              }
            }
          } // for
        }
      }
    } // one or more keys
  } // some key

  return FALSE;
}

BOOL ColorInterpolatorProxy::InterpolateValue(
    int nMinIndex, int nMaxIndex, float fFraction)
{  
  ColorInterpolator * poCI = 
    (ColorInterpolator *) GetMediaObject();

  Color3f tempColor;
  float minH, minS, minV;
  float maxH, maxS, maxV;

  // interpolation is to be done in the HSV color space; convert
    
  RGBToHSV(
    poCI->keyValue[nMinIndex].r, 
    poCI->keyValue[nMinIndex].g, 
    poCI->keyValue[nMinIndex].b,
	  minH, minS, minV);
    
  RGBToHSV(
    poCI->keyValue[nMaxIndex].r, 
    poCI->keyValue[nMaxIndex].g, 
    poCI->keyValue[nMaxIndex].b,
	  maxH, maxS, maxV);

  // interpolate in HSV and convert back to RGB
    
  HSVToRGB(maxH + (minH - maxH) * fFraction,
           maxS + (minS - maxS) * fFraction,
           maxV + (minV - maxV) * fFraction,
           tempColor.r, tempColor.g, tempColor.b);
    
  poCI->value_changed = tempColor;

  return TRUE;
}

/* Convert RGB to HSV

   From [Foley et al, Computer Graphics, Principles and Practice,
   pp. 592].

   fR, fG, fB == [0, 1],
   rH == [0, 360] or -1 (undefined),
   rS, rV == [0, 1]. */

/* static */ void
ColorInterpolatorProxy::RGBToHSV(
    float fR, float fG, float fB,
    float& rH, float& rS, float& rV)
{
  float max = max3(fR, fG, fB);
  float min = min3(fR, fG, fB);

  // max is the value V

  rV = max;

  // find saturation; saturation is 0 if rgb are all 0

  if (max != 0.0f)
    rS = (max - min) / max;
  else
    rS = 0.0f;

  // find hue; hue is undefined if saturation is 0

  if (rS == 0.0f)
    rH = -1.0f;
  else {
    float delta = max - min;

    if (fR == max)
      rH = (fG - fB) / delta;
    else
      if (fG == max)
	rH = 4.0f + (fR - fG) / delta;
    
    rH = rH * 60.0f;

    if (rH < 0.0f) 
      rH = rH + 360.0f;
  }
}

/* Convert HSV to RGB

   From [Foley et al, Computer Graphics, Principles and Practice,
   pp. 593].

   rH == [0, 360] or -1 (undefined),
   rS, rV == [0, 1],
   fR, fG, fB == [0, 1]. */

/* static */ void
ColorInterpolatorProxy::HSVToRGB(
    float fH, float fS, float fV,
    float& rR, float& rG, float& rB)
{
  if (fS == 0.0f) {
    assert(fH == -1.0f);

    rR = fV;
    rG = fV;
    rB = fV;
  } else {
    float p, q, t;
    float frac;
    int i;

    if (fH == 360.0f)
      fH = 0.0f;

    fH = fH / 60.0f;
    i = int(::floor(fH));

    frac = fH - i;
    p = fV * (1.0f - fS);
    q = fV * (1.0f - (fS * frac));
    t = fV * (1.0f - (fS * (1 - frac)));

    switch (i) {
    case 0:
      rR = fV;
      rG = t;
      rB = p;
      break;

    case 1:
      rR = q;
      rG = fV;
      rB = p;
      break;

    case 2:
      rR = p;
      rG = fV;
      rB = t;
      break;

    case 3:
      rR = p;
      rG = q;
      rB = fV;
      break;

    case 4:
      rR = t;
      rG = p;
      rB = fV;
      break;

    case 5:
      rR = fV;
      rG = p;
      rB = q;
      break;
    }
  }
}
