/***********************************************************************
This software module was originally developed by James D. K. Kim
(Samsung) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Samsung retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact James D. K. Kim at jacobkim@sait.samsung.co.kr.
----------------------------------------------------------------------*/

#include "OpenGl/OpenGl.h"
#include "CoordinateInterpolatorProxy.h"
#include "../../IM1Decoders/AFX/jwafxstreamdecoder.h"
#include "Stream.h"

CoordinateInterpolatorProxy::CoordinateInterpolatorProxy(MediaObject * poNode) :
CompressedInterpolatorProxy(poNode)
{}

inline BOOL CoordinateInterpolatorProxy::SetValue(
    CoordinateInterpolator * poCI, int iIndex)
{
  int arraySize = poCI->keyValue.GetLength() / poCI->key.GetLength();
  int offset = arraySize * iIndex;

  // sanity test
  assert(poCI->keyValue.GetLength() % poCI->key.GetLength() == 0);

  // Get rid of all old values
  poCI->value_changed.Remove();

  // Set coordinates directly
  for (int j = 0; j < arraySize; ++j)
    poCI->value_changed.Append(poCI->keyValue[offset + j]);

  return TRUE;
}

BOOL CoordinateInterpolatorProxy::Evaluate()
{
  CoordinateInterpolator * poCI = 
    (CoordinateInterpolator *) GetMediaObject();

  // Check if we have any keys
  if (poCI->key.GetLength() > 0)
  {
    // If we only have one key
    if (poCI->key.GetLength() == 1)
    {
      return SetValue(poCI, 0);
    }
    else // more keys
    {
      // If set_fraction is smaller than smallest key
      if (poCI->set_fraction <= poCI->key[0])
      {
        return SetValue(poCI, 0);
      }
      else
      {
        // If set_fraction is larger than largest key
        if (poCI->set_fraction >= poCI->key[poCI->key.GetLength() - 1])
        {
          return SetValue(poCI, poCI->key.GetLength() - 1);
        }
        else
        {
          // Search the key space for fraction
          for (int i = poCI->key.GetLength() - 2; i >= 0 ; --i)
          {
            if (poCI->key[i] == poCI->set_fraction)
            {
              // Found exact match
              return SetValue(poCI, i);
            }
            else
            {
              if (poCI->key[i] < poCI->set_fraction)
              {
                // Didn't find exact match; must interpolate
                return InterpolateValue( 
                  i, i + 1, 
                  ((poCI->key[i + 1] - poCI->set_fraction) / 
                  (poCI->key[i + 1] - poCI->key[i])) );
              }
            }
          } // for
        }
      }
    } // one or more keys
  } // some key

  return FALSE;
}

BOOL CoordinateInterpolatorProxy::InterpolateValue(
    int nMinIndex, int nMaxIndex, float fFraction)
{
  CoordinateInterpolator * poCI = 
    (CoordinateInterpolator *) GetMediaObject();

  int arraySize = poCI->keyValue.GetLength() / poCI->key.GetLength();
  int min = arraySize * nMinIndex;
  int max = arraySize * nMaxIndex;

  // Sanity test

  assert(poCI->keyValue.GetLength() % poCI->key.GetLength() == 0);

  // Get rid of all old values

  poCI->value_changed.Remove();

  // Interpolate coordinates

  for (int i = 0; i < arraySize; ++i) {
    Vec3f tempCoordinate;

    tempCoordinate.x = (poCI->keyValue[max + i].x + 
      (poCI->keyValue[min + i].x - poCI->keyValue[max + i].x) * fFraction);
    tempCoordinate.y = (poCI->keyValue[max + i].y + 
      (poCI->keyValue[min + i].y - poCI->keyValue[max + i].y) * fFraction);
    tempCoordinate.z = (poCI->keyValue[max + i].z + 
      (poCI->keyValue[min + i].z - poCI->keyValue[max + i].z) * fFraction);

    poCI->value_changed.Append(tempCoordinate);
  }

  return TRUE;
}

// SOWOO 20030714 begin
#if defined SAIT_CONFORMANCE_2003

BOOL CoordinateInterpolatorProxy::RenderFromStream(MediaStream* poStream, int id, DWORD dwTime)
{
  BYTE * pData = NULL;
  int nSize = 0;
  BOOL bEOU;
  int nStatus;
  BOOL bSkip = TRUE;
  BOOL bWait = FALSE;
  long lDelta = 0;
  BOOL bDontClose = TRUE;
  DWORD rdvTime;
  
  pData = (unsigned char *) poStream->Fetch(
    nSize, rdvTime, bEOU, nStatus, bSkip, bWait, lDelta, bDontClose);
  
  TRACE("[BitWrapperProxy::Enable] Fetch returned Size %4d \n", nSize);
  
  if (pData)
  {
    JWAFXCoordinateInterpolatorDecoder* pDecoder = new JWAFXCoordinateInterpolatorDecoder;
    
    // Set key and keyValues
    pDecoder->SetValues(pData);
    
    // Test
    pDecoder->Write("AFX-CoordinateInterpolator.debug.vrml", id);

    // Enable
    Enable(pData);

    return TRUE;
  }

  return FALSE;
}

void CoordinateInterpolatorProxy::RenderFromBuffer(SFString *pBuffer, int id)
{
  BYTE * pData = NULL;
  int nSize = 0;

  JWAFXCoordinateInterpolatorDecoder* pDecoder = new JWAFXCoordinateInterpolatorDecoder;

  // Decode bitstream
  pDecoder->Decode((LPBYTE) pBuffer->GetValue(), pBuffer->GetLength(), (LPBYTE) pData, nSize);

  // Test
  pDecoder->Write("AFX-CoordinateInterpolator.debug.vrml", id);

  // Enable
  Enable(pData);
}

void CoordinateInterpolatorProxy::Enable (unsigned char *pData)
{
	// variables: indices
	int i = 0;
	int j = 0;

	CoordinateInterpolator* poCoordInterp = (CoordinateInterpolator*) GetMediaObject();

	if(poCoordInterp)
	{
		int* pData_NumberOfKeys = (int*) pData;
		int *pData_NumberOfCoords = (int*) (pData + sizeof(int));
		float* pData_Keys = (float*) (pData + 2*sizeof(int));
		int nNumberOfKeys = *pData_NumberOfKeys;
		float* pData_KeyValues = (float*) (pData + 2*sizeof(int) + nNumberOfKeys*sizeof(float));
		int nNumberOfCoords = *pData_NumberOfCoords;

		MF<float>* pKeys = (MF<float>*) poCoordInterp->GetField(1);
		MF<Vec3f>* pKeyValues = (MF<Vec3f>*) poCoordInterp->GetField(2);

		Vec3f dummy;

		// key data
		for(i = 0; i < nNumberOfKeys; i++)
			pKeys->SetAt(i, pData_Keys[i], FALSE);

		// key value data
		for(i = 0; i < nNumberOfKeys; i++)
		{
			for(j = 0; j < nNumberOfCoords; j++)
			{
				dummy.x = pData_KeyValues[i*3*nNumberOfCoords+j*3];
				dummy.y = pData_KeyValues[i*3*nNumberOfCoords+j*3+1];
				dummy.z = pData_KeyValues[i*3*nNumberOfCoords+j*3+2];
				pKeyValues->SetAt(i*nNumberOfCoords+j, dummy, FALSE);
			}
		}
	}
}
#endif SAIT_CONFORMANCE_2003
// SOWOO 20030714 end