/***********************************************************************
This software module was originally developed by James D. K. Kim
(Samsung) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Samsung retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2003
************************************************************************

For more information or to receive an up to date version of this module,
you may contact James D. K. Kim at jacobkim@sait.samsung.co.kr.
----------------------------------------------------------------------*/

#include "OpenGl/OpenGl.h"
#include "PositionInterpolatorProxy.h"
#include "../../IM1Decoders/AFX/jwafxstreamdecoder.h"
#include "Stream.h"

PositionInterpolatorProxy::PositionInterpolatorProxy(MediaObject * poNode) :
CompressedInterpolatorProxy(poNode)
{}

inline BOOL PositionInterpolatorProxy::SetValue(
    PositionInterpolator * poPI, int iIndex)
{
  poPI->value_changed = poPI->keyValue[iIndex];
  return TRUE;
}

BOOL PositionInterpolatorProxy::Evaluate()
{
  PositionInterpolator * poPI = 
    (PositionInterpolator *) GetMediaObject();

  // Check if we have any keys
  if (poPI->key.GetLength() > 0)
  {
    // If we only have one key
    if (poPI->key.GetLength() == 1)
    {
      return SetValue(poPI, 0);
    }
    else // more keys
    {
      // If set_fraction is smaller than smallest key
      if (poPI->set_fraction <= poPI->key[0])
      {
        return SetValue(poPI, 0);
      }
      else
      {
        // If set_fraction is larger than largest key
        if (poPI->set_fraction >= poPI->key[poPI->key.GetLength() - 1])
        {
          return SetValue(poPI, poPI->key.GetLength() - 1);
        }
        else
        {
          // Search the key space for fraction
          for (int i = poPI->key.GetLength() - 2; i >= 0 ; --i)
          {
            if (poPI->key[i] == poPI->set_fraction)
            {
              // Found exact match
              return SetValue(poPI, i);
            }
            else
            {
              if (poPI->key[i] < poPI->set_fraction)
              {
                // Didn't find exact match; must interpolate
                return InterpolateValue( 
                  i, i + 1, 
                  ((poPI->key[i + 1] - poPI->set_fraction) / 
                  (poPI->key[i + 1] - poPI->key[i])) );
              }
            }
          } // for
        }
      }
    } // one or more keys
  } // some key

  return FALSE;
}

BOOL PositionInterpolatorProxy::InterpolateValue(
    int nMinIndex, int nMaxIndex, float fFraction)
{
  PositionInterpolator * poPI = 
    (PositionInterpolator *) GetMediaObject();

  Vec3f tempPosition;

  assert(poPI->keyValue.GetLength() == poPI->key.GetLength());

  // Interpolate positions

  tempPosition.x = (poPI->keyValue[nMaxIndex].x + 
    (poPI->keyValue[nMinIndex].x - poPI->keyValue[nMaxIndex].x) * fFraction);
  tempPosition.y = (poPI->keyValue[nMaxIndex].y + 
    (poPI->keyValue[nMinIndex].y - poPI->keyValue[nMaxIndex].y) * fFraction);
  tempPosition.z = (poPI->keyValue[nMaxIndex].z + 
    (poPI->keyValue[nMinIndex].z - poPI->keyValue[nMaxIndex].z) * fFraction);
  
  poPI->value_changed = tempPosition;

  return TRUE;
}


#if defined SAIT_CONFORMANCE_2003

BOOL PositionInterpolatorProxy::RenderFromStream(MediaStream* poStream, int id, DWORD dwTime)
{
  BYTE * pData = NULL;
  int nSize = 0;
  BOOL bEOU;
  int nStatus;
  BOOL bSkip = TRUE;
  BOOL bWait = FALSE;
  long lDelta = 0;
  BOOL bDontClose = TRUE;
  DWORD rdvTime;
  
  pData = (unsigned char *) poStream->Fetch(
    nSize, rdvTime, bEOU, nStatus, bSkip, bWait, lDelta, bDontClose);
  
  TRACE("[BitWrapperProxy::Enable] Fetch returned Size %4d \n", nSize);
  
  if (pData)
  {
    JWAFXPositionInterpolatorDecoder* pDecoder = new JWAFXPositionInterpolatorDecoder;
    
    // Set key and keyValues
    pDecoder->SetValues(pData);
    
    // Test
    pDecoder->Write("AFX-PositionInterpolator.debug.vrml",id);

    // Enable
    Enable(pData);

    return TRUE;
  }

  return FALSE;
}

void PositionInterpolatorProxy::RenderFromBuffer(SFString *pBuffer, int id)
{
  BYTE * pData = NULL;
  int nSize = 0;

  JWAFXPositionInterpolatorDecoder* pDecoder = new JWAFXPositionInterpolatorDecoder;
  
  // Decode the bitstream
  pDecoder->Decode((LPBYTE) pBuffer->GetValue(), pBuffer->GetLength(), (LPBYTE) pData, nSize);
  
  // Test
  pDecoder->Write("AFX-PositionInterpolator.debug.vrml", id);

  // Enable
  Enable(pData);
}

void PositionInterpolatorProxy::Enable (unsigned char *pData)
{
	PositionInterpolator* poPosInterp = (PositionInterpolator*) GetMediaObject();

	if(poPosInterp)
	{
		int* pData_NumberOfKeys = (int*) pData;
		float* pData_Keys = (float*) (pData + sizeof(int));
		int nNumberOfKeys = pData_NumberOfKeys[0];
		float* pData_KeyValues = (float*) (pData + sizeof(float)*nNumberOfKeys + sizeof(int));


		MF<float>* pKeys = (MF<float>*) poPosInterp->GetField(1);
		MF<Vec3f>* pKeyValues = (MF<Vec3f>*) poPosInterp->GetField(2);

		Vec3f dummy;
		for(int i=0,j=0; i<nNumberOfKeys; i++)
		{
			pKeys->SetAt(i, pData_Keys[i], FALSE);

			dummy.x = pData_KeyValues[j++];
			dummy.y = pData_KeyValues[j++];
			dummy.z = pData_KeyValues[j++];
			pKeyValues->SetAt(i, dummy, FALSE);
		}
	}
}
#endif SAIT_CONFORMANCE_2003
