//This software module was originally developed by ENST in the course of   //
//development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.             //
//This software module is an implementation of a part of one or more MPEG-4 //
//Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems        //
//(ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 Systems     //
//(ISO/IEC 14496-1) free license to this software module or modifications   //
//thereof for use in hardware or software products claiming conformance to  //
//the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use this software//
//module in hardware or software products are advised that its use may      //
//infringe existing patents.                                                //
//The original developer of this software module and his/her company, the   //
//subsequent editors and their companies, and ISO/IEC have no liability     //
//for use of this software module or modifications thereof in an            //
//implementation. Copyright is not released for non MPEG-4 Systems          //
//(ISO/IEC 14496-1) conforming products. ENST retains full right     //
//to use the code for his/her own purpose, assign or donate the code to a   //
//third party and to inhibit third parties from using the code for non      //
//MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright      //
//notice must be included in all copies or derivative works.                //
//Copyright  2002.                                                         //
//////////////////////////////////////////////////////////////////////////////
//
// File:  AnimationStreamProxy.cpp
//
// Author: Jean Le Feuvre, ENST 2002
//
//
// Description:
//		Proxy object for AnimationStreamImp
//
// Created: September, 2002.
//
//
///////////////////////////////////////////////////////////////////////////

#include "defs.h"
#include "Presenter.h"
#include "AnimationStreamProxy.h"
#include "Nodes-INIT.h"
#include "Stream.h"
#include "ODImp.h"


AnimationStreamProxy *Presenter::InitAnimationStream(AnimationStreamImp* pObj)
{
	return new AnimationStreamProxy(pObj);
}


AnimationStreamProxy::AnimationStreamProxy(MediaObject* pMO) : Traversable(pMO) 
{
	m_startTime = 0;
	m_cycle = 0;
	GetPresenter()->RegisterTimedNode(this);
	m_bIsRegistered = TRUE;
	m_bNeedsUnregister = FALSE;
	m_RestartStatus = ST_None;
	m_bRegWithMC = FALSE;
	m_bSetSpeed = FALSE;
}

AnimationStreamProxy::~AnimationStreamProxy()
{
	if (m_bRegWithMC) {
		MediaControlImp *pCtrl = GetAnimationStream()->GetObjectMediaControl();
		if (pCtrl) pCtrl->RemoveUser();
	}
	if (m_bIsRegistered) GetPresenter()->UnregisterTimedNode(this);
}

void AnimationStreamProxy::Render(Effects *effects) 
{
	AnimationStreamImp *pAS = GetAnimationStream();

	if (!pAS->GetOD()) return;

	//scene invalidation is done by the anim decoder
	if (pAS->HasStreamFinished()) {
		if (IsLooping()) {
			/* cf note in TimedNode.h */
			switch (m_RestartStatus) {
			case ST_None:
				m_RestartStatus = ST_Stoped;
				pAS->Stop();
				/*force redraw to called again*/
				GetPresenter()->InvalidateScene(NULL);
				break;
			case ST_Stoped:
				m_RestartStatus = ST_Opened;
				pAS->Start();
				m_bSetSpeed = TRUE;
				GetPresenter()->InvalidateScene(NULL);
				break;
			}
		} else if ((BOOL) (Bool) pAS->isActive) {
			//filter in case this is due to a mediaControl restart, in which case the OD has been closed already
			//if not closed, this is a regular end of stream so deactivate
			if (pAS->GetOD()->IsPlaying()) {
				//Note: it may happen that the clock is ruled by a media control whose object has a longer duration. 
				//In this case we don't want to deactivate the node before the other object is finished
				MediaControlImp *pCtrl = pAS->GetMediaControl();
				if (!pCtrl || !pCtrl->GetOD()->IsPlaying() )
					Deactivate();
			}
		}
	} else {
		m_RestartStatus = ST_None;
	}

	//not running
	if (! (BOOL) (Bool) pAS->isActive) {
		//has been stoped, unregister
		if (m_bNeedsUnregister && m_bIsRegistered) {
			GetPresenter()->UnregisterTimedNode(this);
			m_bNeedsUnregister = FALSE;
			m_bIsRegistered = FALSE;
		}
		return;
	}

	/*if OD is attached, get duration*/
	if (m_cycle==0) {
		m_cycle = pAS->GetOD()->GetDuration();
		pAS->duration_changed = m_cycle;
	}

	if (m_bSetSpeed && (m_RestartStatus != ST_Opened) ) {
		pAS->SetObjectSpeed(pAS->speed);
		m_bSetSpeed = FALSE;
	}
}

BOOL AnimationStreamProxy::IsLooping()
{
	AnimationStreamImp *pAS = GetAnimationStream();
	MediaControlImp *pCtrl = pAS->GetObjectMediaControl();
	if (pCtrl) return (BOOL) (Bool) pCtrl->loop;
	return (BOOL) (Bool) pAS->loop;
}

void AnimationStreamProxy::Activate()
{
	AnimationStreamImp *pAS = GetAnimationStream();
	MediaControlImp *pCtrl = pAS->GetObjectMediaControl();

	pAS->isActive = TRUE;
	if (pCtrl) {
		m_bRegWithMC = TRUE;
		pCtrl->AddUser();
	}
	//and open
	pAS->OpenStream();
	/*we cannot set the speed if the object is not running (that would mean a lot of modifs in the Core)
	so set it at next Render() */
	m_bSetSpeed = TRUE;
	m_iNumLoop = 0;
}

void AnimationStreamProxy::Deactivate()
{
	AnimationStreamImp *pAS = GetAnimationStream();
	MediaControlImp *pCtrl = pAS->GetObjectMediaControl();

	pAS->isActive = FALSE;
	if (m_bRegWithMC && pCtrl) {
		m_bRegWithMC = TRUE;
		pCtrl->RemoveUser();
	}
	//and close
	pAS->CloseStream();
}


double AnimationStreamProxy::GetSpeed()
{
	double speed;
	AnimationStreamImp *pAS = GetAnimationStream();
	speed = pAS->GetMediaControl() ? pAS->GetMediaControl()->mediaSpeed : m_Speed;
	if (speed<0) speed *= -1;
	return speed;
}

void AnimationStreamProxy::UpdateTimeInfo()
{
	AnimationStream *pAS = (AnimationStream *) GetObject();

	/*not active, store startTime*/
	if (! (Bool) pAS->isActive) {
		m_startTime = (double) pAS->startTime;
		m_Speed = pAS->speed;
	}

	double currentTime = GetFlexTime() / 1000.;
	// Only proceed if the start time has been reached
	if (currentTime < m_startTime || (pAS->startTime < 0)) return;

	double speed = GetSpeed();

	//if active and not paused, check for deactivation
	if (speed && (Bool) pAS->isActive) {
		//if stoptime is reached (>startTime) deactivate	
		if ((pAS->stopTime > m_startTime) && (currentTime >= pAS->stopTime) ) {
			Deactivate();
			return;
		}
		//end of cycle detection is done through end of stream detection (see Render() )
	}

	/*we're (about to be) active: VRML:
	"A time-dependent node is inactive until its startTime is reached. When time now becomes greater than or 
	equal to startTime, an isActive TRUE event is generated and the time-dependent node becomes active 	*/
	assert(m_bIsRegistered);

	// activate
	if (! (BOOL) (Bool) pAS->isActive) Activate();
}


void AnimationStreamINIT::OnUpdate()
{

	((AnimationStreamProxy*)GetRenderingObject())->OnUpdate();
}

void AnimationStreamProxy::OnUpdate()
{
	AnimationStream *pAS = (AnimationStream *) GetObject();


	//cf note in TimedNodes.h
	PresenterContext ft_ctx;
	ft_ctx.SetTime(GetFlexTime());
	pAS->startTime.ApplyOffset(&ft_ctx);
	pAS->stopTime.ApplyOffset(&ft_ctx);


	if ((Bool) pAS->isActive) UpdateTimeInfo();

	if (!m_bIsRegistered) {
		GetPresenter()->RegisterTimedNode(this);
		m_bIsRegistered = TRUE;
	}
}
