/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  Audible.cpp
//
// Author: Jean Le Feuvre - ENST 2002
//
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Audible.h"
#include "Presenter.h"
#include "AudioRenderer.h"
#include "ODImp.h"


Traversable* AudioGroup::GetChild(int n)
{
	MediaObject *pChild = (*GetObject()->GetChildrenField())[n];
	if (pChild)
		return (Traversable*) pChild->GetRenderingObject(); 
	return NULL;
}

AudibleNode::AudibleNode(MediaObject* pMO) : AudioProd(pMO)
{
	m_registeredWithRenderer = FALSE;
	m_registeredWithParent = FALSE;
	m_store = NULL;
	m_storeLen = 0;
	m_isFinished = FALSE;
	m_bIsConfigured = FALSE;
	m_bStreamOpened = FALSE;
	m_nBytesPerSec = 0;
	m_nODID = -1;
}

AudibleNode::~AudibleNode()
{
	//stream close is done by StreamConsumer
	if (m_registeredWithRenderer) {
		GetPresenter()->GetAudioRenderer()->RemoveSourceNode(this);
		m_registeredWithRenderer = FALSE;
	}
	if (m_store) delete m_store;
	if (m_registeredWithParent) GetPresenter()->InvalidateScene(NULL);
}

void AudibleNode::DataUsed(DWORD nBytesGiven)
{
	StreamConsumer *pObj = (StreamConsumer *) GetObject();

	if (m_storeLen > 0) {
		m_storeOffset += nBytesGiven;
		if (m_storeOffset < m_storeLen) return;
	}
	if (m_isFinished) return;

	//OK release
	pObj->GetStream()->Release();
	m_storeOffset = m_storeLen = 0;
}

void AudibleNode::GetData(DWORD nBytesRequired, LPBYTE& rpData, DWORD& rnBytesGiven,DWORD& timeStamp)
{
	StreamConsumer *pObj = (StreamConsumer *) GetObject();

	rpData = NULL;
	rnBytesGiven = 0;
	timeStamp=0;

	if ((pObj->GetStream() == NULL) || m_isFinished) return;
	if (!pObj->GetOD()->IsOpen()) return;

	int dataSize;
	DWORD rdwTime;
	int  Status;
	LPVOID dataPtr;
    BOOL bEOU;

	//use stored CU
	if (m_storeLen > 0) {
		rpData = m_store + m_storeOffset;
		rnBytesGiven = m_storeLen - m_storeOffset;
		if (rnBytesGiven>nBytesRequired) rnBytesGiven = nBytesRequired;
		//update relative time
		timeStamp = m_storeTime + m_storeOffset * 1000 / m_nBytesPerSec;
		return;
	} 

	//first time fetch
	if (m_bIsFirstFrame) {
		dataPtr = pObj->Fetch(dataSize, rdwTime, bEOU, Status, FALSE, FALSE, 0, FALSE);

		if (dataPtr != NULL && rdwTime != 0) {
			m_bIsFirstFrame	= false;
//			GetStream()->AdjustClock(rdwTime);
		}
	} 
	//otherwise fetch next and discards whatever is late
	else {
		dataPtr = pObj->GetStream()->FetchNext(dataSize, rdwTime, bEOU, FALSE);
	}

	// If there is no data, or the data returned has already been played, return zero bytes
	if (dataPtr == NULL) return;

//	TRACE("audio fetched %d at %d\n", rdwTime, GetTime(0));

	//If end of stream has been reached, close the stream and return zero bytes
	if (dataPtr == END_OF_STREAM) {
		if (!m_isFinished) {
			m_isFinished = TRUE;
			GetPresenter()->InvalidateScene(NULL);
		}
		return;
	}

	// If data block returned has 0 size, return zero bytes
	if (dataSize == 0) {
		pObj->GetStream()->Release();
		return;
	}

	timeStamp = rdwTime;
	rpData = (LPBYTE)dataPtr;
	rnBytesGiven = dataSize;

	m_storeTime = rdwTime;
	m_storeOffset = 0;
	if (!m_store || (m_storeLen < rnBytesGiven)) {
		m_storeLen = rnBytesGiven;
		// Make the partial store large enough
		if (m_store) delete m_store;
		m_store = new unsigned char [m_storeLen];
	}

	//Copy over the remaining samples for next call
	memcpy(m_store, rpData, m_storeLen);
	if (nBytesRequired<rnBytesGiven) rnBytesGiven = nBytesRequired;

}

StreamConsumer *AudibleNode::GetStreamConsumer()
{
	//WARNING: all nodes using this class shall be stream consumers
	return (StreamConsumer *) GetObject();
}

MediaStreamImp *AudibleNode::GetStream()
{
	//WARNING: all nodes using this class shall be stream consumers
	StreamConsumer *pObj = (StreamConsumer *) GetObject();
	if (pObj) return pObj->GetStream();
	return NULL;
}


BOOL AudibleNode::OpenObject()
{
	//WARNING: all nodes using this class shall be stream consumers
	StreamConsumer *pObj = (StreamConsumer *) GetObject();
	
	//not ready
	if (!pObj->GetStream()) return FALSE;

	pObj->OpenStream();
	m_bStreamOpened = TRUE;

	/*if ODID has changed force audio reconfig*/
	if (m_nODID != pObj->GetObjectDescriptorID()) {
		m_nBytesPerSec = 0;
		m_nODID = pObj->GetObjectDescriptorID();
	}
	m_bIsConfigured = FALSE;
	return TRUE;
}

BOOL AudibleNode::RegisterObject(AudioGroup *parent)
{
	/*already setup*/
	if (m_bIsConfigured) return TRUE;

	/*try to init config*/
	if (!m_nBytesPerSec) {
		StreamConsumer *pObj = (StreamConsumer *) GetObject();
		//Read the output of the decoder to get the audio format
		Decoder *pIstDec = pObj->GetDecoder(); 

		//some audio streams don't use decoder config (mp3 ...), thus we shall wait for a frame to figure out the config...
		m_nSamplesPerSec = pIstDec->GetAttributes(AUDIO_FREQUENCY,NULL);
		m_nBitsPerSample = pIstDec->GetAttributes(BITS_PER_SAMPLE,NULL);
		m_nChannels = pIstDec->GetAttributes(NUM_CHANNELS,NULL);

		m_nBytesPerSec = m_nSamplesPerSec * m_nChannels * m_nBitsPerSample / 8;
		if (!m_nBytesPerSec) return FALSE;
		m_bIsFirstFrame = TRUE;
		m_isFinished = FALSE;
	}

	//register with parent
	if (parent && !m_registeredWithParent) {
		parent->AddInput(this);
		m_registeredWithParent = TRUE;
	} 
	//register with renderer
	else if (!m_registeredWithRenderer) {
		GetPresenter()->GetAudioRenderer()->AddSourceNode(this);
		m_registeredWithRenderer = TRUE;
	}
	m_bIsConfigured = TRUE;
	return TRUE;
}


void AudibleNode::Reset()
{
	if (m_storeLen>0 && GetStream() ) GetStream()->Release();
	m_storeLen = 0;
	m_bIsFirstFrame = true;
	m_isFinished = false;
}

void AudibleNode::CloseObject()
{
	Reset();
	m_isFinished = TRUE;

	if (!m_bStreamOpened) return;
	((StreamConsumer *) GetObject())->CloseStream();
	m_bStreamOpened = FALSE;

	//if register with main renderer, unregister
	if (m_registeredWithRenderer) {
		GetPresenter()->GetAudioRenderer()->RemoveSourceNode(this);
		m_registeredWithRenderer = FALSE;
	}
}
