/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  Audible.h
//
// Author: Jean Le Feuvre - ENST 2002
//
//
///////////////////////////////////////////////////////////////////////////


#ifndef AUDIBLE_H
#define AUDIBLE_H

#include "Traversable.h"
#include "StreamCo.h"
#include "Stream.h"

//virtual class used by all audio nodes producing data
class AudioProd : public Traversable
{
public:
	AudioProd(MediaObject* pMO) : Traversable(pMO) 
	{
		lastR = lastL = 0;
		fracSample = fracSampRatio = 0;
		m_pSound2D = NULL;
	}
	~AudioProd() {}
	
	//retuns audio data in ptr, never more than bytesWanted
	virtual void GetData(DWORD bytesWanted, LPBYTE &ptr, DWORD &bytes_available, DWORD &time_stamp) = 0;
	//drops bytesUsed from current audio data
	virtual void DataUsed(DWORD bytesUsed) = 0;
	
	//retuns audio volume, between 0.0 and 1.0
	virtual float GetIntensity() = 0;
	virtual float GetSpeed() = 0;

	//retuns audio left pan, between 0.0 and 1.0. Mono objects shall return 1.0
	virtual float GetPanLeft() = 0;
	//retuns audio right pan, between 0.0 and 1.0.  Mono objects shall return 0.0
	virtual float GetPanRight() = 0;

	//retunrs whether stream is muted (not active) or not
	virtual BOOL IsMuted() = 0;

	//returns clock time - this is only used to discard audio units based on their TS, thus nodes that can't get 
	//a unique object clock (such as AudioMix or audioBuffer) can emulate their clock
	//audioDelay specifies the delay between the object and the mixer - used by the audio renderer to specify 
	//the delay of the sound card. 
	virtual DWORD GetTime(unsigned int audioDelay) = 0;
	
	//the object using this audio source (only Sound2D in 2D) - some objects may not have a sound2D associated
	class Sound2D *m_pSound2D;

	//audio configuration
	DWORD m_nSamplesPerSec, m_nBitsPerSample, m_nChannels;

	//SR conversion stuff
	signed short lastR, lastL;
	unsigned int fracSample;
	unsigned int fracSampRatio;
};

/*base class for all audio grouping nodes */
class AudioGroup : public AudioProd
{
public:
	AudioGroup(MediaObject* pMO) : AudioProd(pMO) {}
	~AudioGroup() {}

	Traversable* GetChild(int n);

	//inputs are added in declarative order. if @src is NULL this means the input isn't
	//ready and another traversal is needed
	virtual void AddInput(AudioProd *src) = 0;
};


/*base class for AudioSource and AudioClip*/
class AudibleNode : public AudioProd
{

public:
	AudibleNode(MediaObject* pMO);
	~AudibleNode();
	
	//opens object - returns FALSE if object not ready
	BOOL OpenObject();
	//register with audio parent/renderer - returns FALSE if audio config isn't known
	BOOL RegisterObject(AudioGroup *parent);
	//stops object - this is used at destruction time and by AudioClip in case oif Loop=TRUE only
	void CloseObject();

	virtual void GetData(DWORD bytesWanted, LPBYTE &ptr, DWORD &bytes_available, DWORD &time_stamp);
	virtual void DataUsed(DWORD bytesUsed);


	StreamConsumer *GetStreamConsumer();
	MediaStreamImp *GetStream();

	//default methods to override by sub-classes
	virtual float GetIntensity() { return 1.0; }
	virtual float GetSpeed() { return 1.0; }
	//if not active data is fetched and discarded
	virtual BOOL IsMuted() = 0;

	//retuns audio left pan, between 0.0 and 1.0.  Mono objects shall return 1.0
	virtual float GetPanLeft() {return 1.0; }
	//retuns audio right pan, between 0.0 and 1.0.  Mono objects shall return 0.0
	virtual float GetPanRight() {return 1.0; }

	virtual DWORD GetTime(unsigned int audioDelay) { return audioDelay + GetStream()->GetClock(); }

	
	void Reset();

protected:
	BOOL m_isFinished;
	BOOL m_bIsMuted;

private:
	DWORD m_nBytesPerSec;
	BOOL m_registeredWithRenderer;
	//this is to filter calls to add_source - Note this DOESN'T work if the node is a USE
	BOOL m_registeredWithParent;
	BOOL m_bIsFirstFrame;
	BOOL m_bIsConfigured;
	BOOL m_bStreamOpened;


	//storage of current audio CU
	DWORD m_storeTime;
	DWORD m_storeLen, m_storeOffset;
	unsigned char* m_store;

	int m_nODID;
};

#endif

