/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2003.
*/

///////////////////////////////////////////////////////////////////////////
// File:  AudioBufferProxy.cpp
//
// Author: Jean Le Feuvre - ENST 2003
//
//
///////////////////////////////////////////////////////////////////////////

#include "defs.h"
#include "Nodes-INIT.h"
#include "presenter.h"
#include "AudioRenderer.h"
#include "AudioBufferProxy.h"


AudioBufferProxy * Presenter::InitAudioBuffer(AudioBuffer* pObj)
{
	return (new AudioBufferProxy(pObj));
}

AudioBufferProxy::AudioBufferProxy(MediaObject* pMO) : AudioGroup(pMO)
{
	GetPresenter()->RegisterTimedNode(this);
	m_bIsRegistered = TRUE;
	m_bNeedsUnregister = FALSE;

	m_bDone = FALSE;
	m_lastCount = 0;
	m_startTime = 0;
	m_pOutBuffer = NULL;
	m_lastLength = 0.0;
	m_bIsInit = FALSE;
	m_registeredWithRenderer = FALSE;
	m_CurTS = 0;
	m_Mixer.SetPresenter(GetPresenter());
}

AudioBufferProxy::~AudioBufferProxy()
{
	if (m_bIsRegistered) GetPresenter()->UnregisterTimedNode(this);
	if (m_pOutBuffer) delete [] m_pOutBuffer;
	if (m_registeredWithRenderer) GetPresenter()->GetAudioRenderer()->RemoveSourceNode(this);
}

void AudioBufferProxy::CheckMixer()
{
	int i, j;
	m_Mixer.m_Lock.P();

	//check if all object in mixer are valid
	for (i=0; i<m_Mixer.m_Streams.GetLength(); i++) {
		BOOL bFound = FALSE;
		for (j = 0; j<GetObject()->GetChildCount(); j++) {
			AudioProd *src = (AudioProd *) GetObject()->GetChild(j);
			if (src==m_Mixer.m_Streams[i]) bFound = TRUE;
		}
		if (!bFound) {
			m_Mixer.m_Streams.DeleteFrom(i);
			i--;
		}
	}
	//the rest is done in subgraph traversal
	m_Mixer.m_Lock.V();
}

void AudioBufferProxy::AddInput(AudioProd *src)
{
	if (!src) {
		m_LastUpdateTime = 0;
		GetPresenter()->InvalidateScene(NULL);
		return;
	}

	if (m_Mixer.AddSource(src)) {
		m_nSamplesPerSec = m_Mixer.m_nSamplesPerSec;
		m_nBitsPerSample = m_Mixer.m_nBitsPerSample;
		m_nChannels = m_Mixer.m_nChannels;
		m_BytesPerSec = m_nSamplesPerSec * m_nChannels * m_nBitsPerSample / 8;
		//destroy the content (a nice opt would be to rewrite the current buffer with the new cfg)
		if (m_pOutBuffer) delete [] m_pOutBuffer;
		m_pOutBuffer = NULL;
	}
	m_bIsInit = TRUE;
}


void AudioBufferProxy::Render(Effects *effects)
{
	AudioBuffer *pAB = (AudioBuffer *) GetObject();

	DWORD now = 1 + GetObject()->GetUpdateTime();
	//rebuild or not the graph...
	if (!m_bNeedsUnregister && now != m_LastUpdateTime) {
		m_Mixer.SetOutputChannels(pAB->numChan);
		CheckMixer();

		Effects *newEffects=new Effects(effects);
		newEffects->m_pAudioParent = this;

		//render all children
		int count = GetObject()->GetChildCount();
		for (int i=0; i<count; i++) {
			Traversable *pTrav = GetChild(i);
			if (pTrav) pTrav->Render(newEffects);
		}
		delete newEffects;

		//no input and children, this means one of the child is not ready
		if (count && ! m_Mixer.m_Streams.GetLength()) {
			GetPresenter()->InvalidateScene(NULL);
			return;
		}
		m_LastUpdateTime = now;
	}

	//register audio object with parent
	if (effects->m_pAudioParent) {
		effects->m_pAudioParent->AddInput(this);		
	} else {
		GetPresenter()->GetAudioRenderer()->AddSourceNode(this);
		m_registeredWithRenderer = TRUE;
	}

	//not running
	if (! (BOOL) (Bool) pAB->isActive) {
		//has been stoped, unregister
		if (m_bNeedsUnregister && m_bIsRegistered) {
			GetPresenter()->UnregisterTimedNode(this);
			m_bNeedsUnregister = FALSE;
			m_bIsRegistered = FALSE;
		}
	}

}


void AudioBufferProxy::Activate()
{
	AudioBuffer *pAB = (AudioBuffer *) GetObject();
	pAB->isActive = TRUE;
	pAB->duration_changed = pAB->length;
	m_bDone = FALSE;
	m_currentReadPos = 0;

	//reset buffer if length changed
	if (m_lastLength != pAB->length) {
		m_lastLength = pAB->length;
		if (m_pOutBuffer) delete [] m_pOutBuffer;
		m_pOutBuffer = NULL;
	}
}

void AudioBufferProxy::Deactivate()
{
	AudioBuffer *pAB = (AudioBuffer *) GetObject();
	pAB->isActive = FALSE;
	m_bNeedsUnregister = TRUE;
	GetPresenter()->InvalidateScene(NULL);
}



void AudioBufferProxy::UpdateTimeInfo()
{
	AudioBuffer *pAB = (AudioBuffer *) GetObject();
	if (! (Bool) pAB->isActive && (pAB->startTime != m_startTime)) {
		m_startTime = (double) pAB->startTime;
	}

	double currentTime = GetFlexTime() / 1000.0;
	// Only proceed if the start time has been reached
	if (currentTime < m_startTime || (m_startTime < 0)) return;

	//if active check de-activation
	if ( (pAB->pitch != 0.0) && (Bool) pAB->isActive ) {
		if ((pAB->stopTime > m_startTime) && (currentTime >= pAB->stopTime) ) {
			Deactivate();
			return;
		}
		//end of buffer and no loop
		if (m_bDone) {
			Deactivate();
			return;
		}
	}		
	/*we're (about to be) active: VRML:
	"A time-dependent node is inactive until its startTime is reached. When time now becomes greater than or 
	equal to startTime, an isActive TRUE event is generated and the time-dependent node becomes active 	*/

	// If this is the first time that the start time has been reached
	if (! (Bool) pAB->isActive) {
		Activate();
	}
}

void AudioBufferINIT::OnUpdate()
{
	//not clear in the specs: is it allowed to restart buffering while the node is active ? doubt so
	if ((Bool) isActive) return;

	AudioBufferProxy *pABP = (AudioBufferProxy *) GetRenderingObject();
	pABP->OnUpdate();
}

void AudioBufferProxy::OnUpdate()
{
	AudioBuffer *pAB = (AudioBuffer *) GetObject();

	//cf note in TimedNodes.h
	PresenterContext ft_ctx;
	ft_ctx.SetTime(GetFlexTime());
	pAB->startTime.ApplyOffset(&ft_ctx);
	pAB->stopTime.ApplyOffset(&ft_ctx);

	if ((Bool) pAB->isActive) UpdateTimeInfo();

	if (!m_bIsRegistered) GetPresenter()->RegisterTimedNode(this);
	m_bIsRegistered = TRUE;
	m_bNeedsUnregister = FALSE;
}


void AudioBufferProxy::GetData(DWORD bytesWanted, LPBYTE &ptr, DWORD &bytes_available, DWORD &time_stamp)
{
	bytes_available = time_stamp = 0;
	ptr = NULL;

	if (!m_pOutBuffer) {
		AudioBuffer *pAB = (AudioBuffer *) GetObject();
		m_bDone = FALSE;
		m_pOutSize = pAB->length * m_BytesPerSec;
		m_pOutBuffer = new unsigned char [m_pOutSize];
		memset(m_pOutBuffer, 0, m_pOutSize);
		m_currentWritePos = 0;
		m_currentReadPos = 0;
		m_lastLength = pAB->length;
	}
	if (m_bDone) return;

	if (m_currentWritePos<m_pOutSize) {
		unsigned int written;
		if (m_currentWritePos + bytesWanted < m_pOutSize) {
			written = m_Mixer.GetMix(m_pOutBuffer + m_currentWritePos, bytesWanted, FALSE);
		} else {
			written = m_Mixer.GetMix(m_pOutBuffer + m_currentWritePos, m_pOutSize - m_currentWritePos, FALSE);
		}
		m_currentWritePos += written;
		//consider this is end of stream
		if (!written && m_currentWritePos) m_currentWritePos = m_pOutSize;
	}

	if ( ! (Bool) ((AudioBuffer *) GetObject())->isActive) return;

	//no support for negative pitch 
	//NB: we assume the buffer is always full enough for read
	time_stamp = m_CurTS = m_currentReadPos * 1000 / m_BytesPerSec;
	if (m_currentReadPos + bytesWanted < m_currentWritePos) {
		bytes_available = bytesWanted;
		ptr = m_pOutBuffer + m_currentReadPos;
	} else {
		//can be 0
		bytes_available = m_currentWritePos - m_currentReadPos;
		ptr = m_pOutBuffer + m_currentReadPos;
	}
}

void AudioBufferProxy::DataUsed(DWORD bytesUsed)
{	
	m_currentReadPos += bytesUsed;
	if (m_currentReadPos>=m_pOutSize) {
		m_currentReadPos = m_pOutSize;
		m_bDone = TRUE;
		AudioBuffer *pAB = (AudioBuffer *) GetObject();
		if ((Bool) pAB->loop) {
			m_bDone = FALSE;
			m_currentReadPos = 0;
		}
	}
}

float AudioBufferProxy::GetSpeed()
{
	AudioBuffer *pAB = (AudioBuffer *) GetObject();
	return (float) pAB->pitch;
}

BOOL AudioBufferProxy::IsMuted()
{
	AudioBuffer *pAB = (AudioBuffer *) GetObject();
	return ! (BOOL) (Bool) pAB->isActive;
}
DWORD AudioBufferProxy::GetTime(unsigned int audioDelay)
{
	return m_CurTS;
}

