/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2003
*/

///////////////////////////////////////////////////////////////////////////
// File:  AudioMixer.cpp
//
// Author: Jean Le Feuvre - ENST 2003
//
///////////////////////////////////////////////////////////////////////////

#include "defs.h"
#include "AudioMixer.h"
#include "Audible.h"
#include "Presenter.h"


AudioMixer::AudioMixer()
{
	//this is the only supported bps
	m_nBitsPerSample = 16;
	//default init
	m_nSamplesPerSec = 44100;
	m_nChannels = 2;
	m_bChannelsForced = FALSE;
	m_pPresenter = NULL;
}

AudioMixer::~AudioMixer()
{
}

float AudioMixer::GetOutputIntensity()
{
	m_Lock.P();
	float intens;
	switch (m_Streams.GetLength()) {
	case 1:
		intens = m_Streams[0]->GetIntensity();
		break;
	case 0:
		intens = 0.0f;
		break;
	default:
		intens = 1.0f;
		break;
	}
	m_Lock.V();
	return intens;
}

BOOL AudioMixer::IsSourceSupported(AudioProd *src)
{
	if (src->m_nBitsPerSample!=16) return FALSE;
	if (src->m_nChannels>2) return FALSE;
	if (src->m_nSamplesPerSec>48000) return FALSE;
	return TRUE;
}

void AudioMixer::RemoveSource(AudioProd *src)
{
	m_Lock.P();
	int i= m_Streams.Find(src);
	if (i>=0) m_Streams.DeleteFrom(i);
	m_Lock.V();
}

void AudioMixer::RemoveAll()
{
	m_Lock.P();
	m_Streams.Clear();
	m_bChannelsForced = FALSE;
	m_Lock.V();
}

void AudioMixer::SetOutputChannels(unsigned int numChannels)
{
	if (!numChannels || numChannels>2) return;
	m_Lock.P();
	m_bChannelsForced = TRUE;
	m_nChannels = numChannels;
	m_Lock.V();
}

BOOL AudioMixer::AddSource(AudioProd *src)
{
	BOOL bChanged = FALSE;

	m_Lock.P();

	assert(src->m_nSamplesPerSec);
	assert(src->m_nChannels);
	assert(src->m_nBitsPerSample);

	//if we already have some streams update to max SR/numChan
	if (m_Streams.GetLength()) {
		//Update mixer settings
		if (m_nSamplesPerSec < src->m_nSamplesPerSec) {
			m_nSamplesPerSec = src->m_nSamplesPerSec;
			bChanged = TRUE;
		}
		if (!m_bChannelsForced && (m_nChannels < src->m_nChannels)) {
			m_nChannels = src->m_nChannels;
			bChanged = TRUE;
		}
	} else {
		m_nSamplesPerSec = src->m_nSamplesPerSec;
		if (!m_bChannelsForced) m_nChannels = src->m_nChannels;
		bChanged = TRUE;
	}
	if (bChanged) {
		for (int i=0; i<m_Streams.GetLength(); i++) {
			m_Streams[i]->fracSampRatio = m_Streams[i]->m_nSamplesPerSec * 0xffff / m_nSamplesPerSec;
			m_Streams[i]->fracSample = 0;
		}
	}
	src->fracSampRatio = src->m_nSamplesPerSec * 0xffff / m_nSamplesPerSec;

	//add if not present
	if (m_Streams.Find(src) < 0) m_Streams += src;

	m_Lock.V();
	
	return bChanged;
}

//min delay (objectclock - timestamp) - if exceeded, input samples are dropped (audio renderer too late)
#define AUDIO_TOLERANCE_SKIP_MS		40
//min delay (timestamp - objectclock) - if exceeded blank is inserted to output (audio renderer too early)
#define AUDIO_TOLERANCE_WAIT_MS		200

#define ENABLE_AUDIO_WAIT	0

unsigned int AudioMixer::GetMix(void *output, unsigned int output_len, BOOL checkTiming, unsigned int audioDelayMS)
{
	int Bps;
	LPBYTE inptr;
	DWORD insize, ts, object_ck, skipBytes, waitBytes;

	m_Lock.P();

	m_audio_delay_ms = audioDelayMS;

	int count = m_Streams.GetLength();

	//direct copy
	if ((count==1) 
		&& (m_Streams[0]->GetIntensity()==1.0) 
		&& (m_Streams[0]->GetSpeed()==1.0) 
		&& (m_Streams[0]->m_nChannels==m_nChannels) ) {
		
		memset(output, 0, output_len);
		LPBYTE cur = (LPBYTE) output;
		DWORD remain = output_len;
		BOOL checkClock = checkTiming;

		while (remain) {
			m_Streams[0]->GetData(remain, inptr, insize, ts);
			
			if (!insize) {
//				TRACE("Not enough data in input (%d missing)\n", remain);
				break;
			}
			//resync buffer on first fill
			skipBytes = 0;
			waitBytes = 0;
			if (checkClock) {
				object_ck = m_Streams[0]->GetTime(m_audio_delay_ms);
				if (object_ck > ts + AUDIO_TOLERANCE_SKIP_MS) {

					Bps = m_Streams[0]->m_nSamplesPerSec *  m_Streams[0]->m_nChannels *  m_Streams[0]->m_nBitsPerSample / 8;
					skipBytes = (object_ck - ts) * Bps / 1000;

					if (skipBytes > insize) {
						m_Streams[0]->DataUsed(insize);
						continue;
					}
				}
#if ENABLE_AUDIO_WAIT
				/*this should help resync after fast-forward or for silent-detection enabled codecs*/
				else if (object_ck + AUDIO_TOLERANCE_WAIT_MS < ts ) {
					Bps = m_nSamplesPerSec *  m_nChannels *  m_nBitsPerSample / 8;
					waitBytes = (ts - object_ck) * Bps / 1000;
					if (waitBytes>=remain) break;
					remain -= waitBytes;
					if (remain<insize) insize = remain;
					cur += waitBytes;
				}
#endif
			
			}

			if (! m_Streams[0]->IsMuted()) {
				memcpy(cur, inptr + skipBytes, insize - skipBytes);
			}

			//move forward
			cur += insize;
			remain -= insize;
			m_Streams[0]->DataUsed(insize);
			checkClock = FALSE;
		}	
		m_Lock.V();
		return output_len - remain;
	}
	
	//reset mix buffer
	int out_size = 	8 * output_len / m_nBitsPerSample;
	memset(output, 0, sizeof(signed short) * out_size);

	//mix all streams
	unsigned int done = 0;
	for (int i=0; i<count; i++) {
		unsigned int res = MixStream(output, output_len, i, m_Streams[i]->GetPanLeft(), m_Streams[i]->GetPanRight(), checkTiming);
		if (res>done) done = res;
	}
	m_Lock.V();
	return done;
}

unsigned int AudioMixer::MixStream(void *output, unsigned int output_len, unsigned int streamIndex, float panLeft, float panRight, BOOL checkTiming)
{
	LPBYTE inptr;
	int Bps;
	DWORD insize, ts, object_ck, skipBytes, tolerance;


	if (! m_Streams.GetLength()) return 0;
	if (streamIndex >= (unsigned int) m_Streams.GetLength()) return 0;

	m_Lock.P();

	LPBYTE cur = (LPBYTE ) output;
	unsigned int read;
	DWORD remain = output_len;
	BOOL checkClock = checkTiming;

	int i = (int) streamIndex;
	float speed = m_Streams[i]->GetSpeed();
	if (speed<0) speed = 1.0;
	
	if (speed==0) remain = 0;
	else tolerance = AUDIO_TOLERANCE_SKIP_MS / speed;

	if (speed != 1.0) checkClock = FALSE;

	while (remain) {
		m_Streams[i]->GetData(remain, inptr, insize, ts);

		if (!insize) {
//			TRACE("Not enough data in input\n");
			break;
		}
		//resync buffer on first fill
		skipBytes = 0;
		if (checkClock) {
			object_ck = m_Streams[i]->GetTime(m_audio_delay_ms);
			if (object_ck > ts - tolerance) {
				Bps = m_Streams[i]->m_nSamplesPerSec * m_Streams[i]->m_nChannels * m_Streams[i]->m_nBitsPerSample / 8;
				skipBytes = (object_ck - ts) * Bps / 1000;

				if (skipBytes > insize) {
					m_Streams[i]->DataUsed(insize);
					continue;
				}
			} 
#if ENABLE_AUDIO_WAIT
			/*this should help resync after fast-forward or for silent-detection enabled codecs*/
			else if ((speed==1.0) && object_ck + AUDIO_TOLERANCE_WAIT_MS < ts) {
				Bps = m_nSamplesPerSec *  m_nChannels *  m_nBitsPerSample / 8;
				unsigned int waitBytes = (ts - object_ck) * Bps / 1000;
				if (waitBytes>=remain) break;
				remain -= waitBytes;
				cur += waitBytes;
				if (remain<insize) insize = remain;
			}
#endif
		}
		int written = AudioMix((signed short *) cur, inptr + skipBytes, insize - skipBytes, m_Streams[i], panLeft, panRight, read, speed);
		
		cur += written;
		remain -= written;
		if (!read) {
			m_Streams[i]->DataUsed(read + skipBytes);
		} else {
			m_Streams[i]->DataUsed(read + skipBytes);
		}
		checkClock = FALSE;
	}	

	m_Lock.V();
	return output_len - remain;
}

#define SHORTCLIP(t) t = (t) > SHRT_MAX ? SHRT_MAX : ( ( (t) <SHRT_MIN) ? SHRT_MIN : (t) ) 

int AudioMixer::AudioMix(signed short *dst, void *src, unsigned int src_len, AudioProd *stream, float volL, float volR, unsigned int &src_read, float speed)
{
	unsigned int curPos = 0;
	short S1, S2, S3, S4, *input;
	int tL, tR;

	input = (short *) src;
	src_read = src_len;
	int written = 0;

	if (stream->IsMuted()) {
		volL = volR = 0.0;
	} else {
		volL *= stream->GetIntensity();
		volR *= stream->GetIntensity();
	}
	S3 = S4 = 0;

	if (m_nChannels==2) {
		if (stream->m_nChannels==2) {
			for (int i=0; i<(int) src_len/4; i++) {
				if (curPos != 0) {
					S1 = *(input + (curPos>>1) - 2);
					S2 = *(input + (curPos>>1) - 1);
				} else {
					S1 = stream->lastL;
					S2 = stream->lastR;
				}
				S3 = *(input + (curPos>>1));
				S4 = *(input + (curPos>>1) + 1);
				tL = S1 * (0xffff - stream->fracSample);
				tR = S2 * (0xffff - stream->fracSample);
				tL = tL + S3 * stream->fracSample;
				tR = tR + S4 * stream->fracSample;
				tL>>=16;
				tR>>=16;
				if (volL != 1.0) tL *= volL;
				if (volR != 1.0) tR *= volR;

				//mix and clip
				tL += * dst;
				tR += * (dst + 1);
				SHORTCLIP(tL);
				SHORTCLIP(tR);
				*dst++ = (short) tL;
				*dst++ = (short) tR;
				stream->fracSample += stream->fracSampRatio;
				written += 4;
				if (stream->fracSample >=0xffff) {
					stream->fracSample -= 0xffff;
					curPos +=4*speed;
					if (curPos > src_len) break;
				}
			}
			stream->lastL = S3;
			stream->lastR = S4;
			src_read = i ? curPos : src_len;
			return written;
		}

		for (int i=0; i<(int) src_len/4; i++) {
			if (curPos != 0) {
				S1 = *(input + (curPos>>1) - 1);
			} else {
				S1 = stream->lastL;
			}
			S3 = *(input + (curPos>>1));
			tL = S1 * (0xffff - stream->fracSample);
			tL += S3 * stream->fracSample;
			tL>>=16;
			if (volL != 1.0) tL *= volL;
			tR = tL;
			//mix and clip
			tL += * dst;
			tR += * (dst + 1);
			SHORTCLIP(tL);
			SHORTCLIP(tR);
			*dst++ = (short) tL;
			*dst++ = (short) tR;
			written += 4;
			stream->fracSample += stream->fracSampRatio;
			if (stream->fracSample >=0xffff) {
				stream->fracSample -= 0xffff;
				curPos+=2*speed;
				if (curPos > src_len) break;
			}
		}

		stream->lastL = S3;
		//in case we switch to stereo input (audioBuffer ...)
		stream->lastR = S3;
		src_read = i ? curPos : src_len;
		return written;
	}
	//mono out
	
	//stereo in - skip right channel
	if (stream->m_nChannels==2) {
		for (int i=0; i<(int) src_len/2; i++) {
			if (curPos != 0) {
				S1 = *(input + (curPos>>1) - 2);
			} else {
				S1 = stream->lastL;
			}
			S3 = *(input + (curPos>>1));
			//in case we switch to stereo output (audioBuffer ...)
			S4 = *(input + (curPos>>1) + 1);
			tL = S1 * (0xffff - stream->fracSample);
			tL += S3 * stream->fracSample;
			tL>>=16;
			if (volL != 1.0) tL *= volL;

			//mix and clip
			tL += * dst;
			SHORTCLIP(tL);
			*dst++ = (short) tL;
			written += 2;
			stream->fracSample += stream->fracSampRatio;
			if (stream->fracSample >=0xffff) {
				stream->fracSample -= 0xffff;
				curPos+=4*speed;
				if (curPos > src_len) break;
			}
		}
		stream->lastL = S3;
		//in case we switch to stereo output (audioBuffer ...)
		stream->lastR = S3;
		src_read = i ? curPos : src_len;
		return written;
	}
	//mono in
	for (int i=0; i<(int) src_len/2; i++) {
		if (curPos != 0) {
			S1 = *(input + (curPos>>1) - 1);
		} else {
			S1 = stream->lastL;
		}
		S3 = *(input + (curPos>>1));
		tL = S1 * (0xffff - stream->fracSample);
		tL += S3 * stream->fracSample;
		tL>>=16;
		if (volL != 1.0) tL *= volL;
		//mix and clip
		tL += * dst;
		SHORTCLIP(tL);
		*dst++ = (short) tL;
		written += 2;
		stream->fracSample += stream->fracSampRatio;
		if (stream->fracSample >=0xffff) {
			stream->fracSample -= 0xffff;
			curPos+=2*speed;
			if (curPos > src_len) break;
		}
	}
	stream->lastL = S3;
	//in case we switch to stereo input (audioBuffer ...)
	stream->lastR = S3;
	src_read = i ? curPos : src_len;
	return written;
}
