/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  AudioSourceProxy.cpp
//
// Author: Jean Le Feuvre - ENST 2002
//
//
///////////////////////////////////////////////////////////////////////////



#include "defs.h"
#include <assert.h>
#include "AudioSourceProxy.h"
#include "Presenter.h"
#include "Nodes-INIT.h"
#include "ODImp.h"

AudioSourceProxy * Presenter::InitAudioSource(AudioSource* pObj)
{
	return (new AudioSourceProxy(pObj));
}

AudioSourceProxy::AudioSourceProxy(MediaObject* pMO) : AudibleNode(pMO)
{
	m_bIsActive = FALSE;
	m_bIsMuted = FALSE;
	m_RestartStatus = ST_None;

	GetPresenter()->RegisterTimedNode(this);
	m_bIsRegistered = TRUE;
	m_bNeedsUnregister = FALSE;
	m_bRegWithMC = FALSE;
	m_cycle = 0;
	m_Speed = 1;
}

AudioSourceProxy::~AudioSourceProxy()
{
	if (m_bIsRegistered) GetPresenter()->UnregisterTimedNode(this);
	if (m_bRegWithMC) {
		AudioSource *pAS = (AudioSource *) GetObject();
		MediaControlImp *pCtrl = pAS->GetObjectMediaControl();
		if (pCtrl) pCtrl->RemoveUser();
	}
}

void AudioSourceProxy::Render(Effects *effects)
{
	AudioSource *pAS = (AudioSource *) GetObject();

	if (!pAS->GetOD()) return;

	if (m_isFinished==TRUE) {
		if (IsLooping()) {
			/* cf note in TimedNode.h */
			switch (m_RestartStatus) {
			case ST_None:
				pAS->Stop();
				/*force a render to fall back into next case*/
				GetPresenter()->InvalidateScene(NULL);
				m_RestartStatus = ST_Stoped;
				break;
			case ST_Stoped:
				pAS->Start();
				m_iNumLoop++;
				m_RestartStatus = ST_Opened;
				/*force a render to fall back into next case*/
				GetPresenter()->InvalidateScene(NULL);
				break;
			case ST_Opened:
				Reset();
				break;
			}
			return;
		} else if ( m_bIsActive) {
			//filter in case this is due to a mediaControl restart, in which case the OD has been closed already
			//if not closed, this is a regular end of stream so deactivate
			if (pAS->GetOD()->IsPlaying()) {
				//Note: it may happen that the clock is ruled by a media control whose object has a longer duration. 
				//In this case we don't want to deactivate the node before the other object is finished
				MediaControlImp *pCtrl = pAS->GetMediaControl();
				if (!pCtrl || !pCtrl->GetOD()->IsPlaying() )
					Deactivate();
				else {
					//this is to force fetch of audio data in case of a restart of the media
					m_isFinished=FALSE;
				}
			}
		}
	}
	m_RestartStatus = ST_None;

	/*if active and config isn't known, invalidate parent or scene and return*/

	if (m_bIsActive) {
		if (!RegisterObject(effects->m_pAudioParent)) {
				if (effects->m_pAudioParent) 
					effects->m_pAudioParent->AddInput(NULL);
				else 
					GetPresenter()->InvalidateScene(NULL);
			return;
		}

		if (m_cycle==0)
			m_cycle = pAS->GetOD()->GetDuration();

		if (m_bSetSpeed && (m_RestartStatus != ST_Opened)) {
			pAS->SetObjectSpeed(pAS->speed);
			m_bSetSpeed = FALSE;
		}
	}
	m_bIsMuted = effects->m_isSwitchedOff;

	//has been stoped, unregister
	if (m_bNeedsUnregister && m_bIsRegistered) {
		GetPresenter()->UnregisterTimedNode(this);
		m_bNeedsUnregister = FALSE;
		m_bIsRegistered = FALSE;
	}

}


BOOL AudioSourceProxy::IsLooping()
{
	AudioSource *pAS = (AudioSource *) GetObject();
	MediaControlImp *pCtrl = pAS->GetObjectMediaControl();
	if (pCtrl) return (BOOL) (Bool) pCtrl->loop;
	//AudioSource doesn't have a loop field
	return FALSE;
}

void AudioSourceProxy::Activate()
{
	AudioSource *pAS = (AudioSource *) GetObject();
	MediaControlImp *pCtrl = pAS->GetObjectMediaControl();

	m_bIsActive = OpenObject();
	if (!m_bIsActive) return;

	m_iNumLoop = 0;
	m_bSetSpeed = TRUE;

	if (pCtrl) {
		m_bRegWithMC = TRUE;
		pCtrl->AddUser();
	}
	//invalidate scene graph to force a call to Render()
	GetPresenter()->InvalidateScene(NULL);
}

void AudioSourceProxy::Deactivate()
{
	AudioSource *pAS = (AudioSource *) GetObject();
	MediaControlImp *pCtrl = pAS->GetObjectMediaControl();

	if (!m_bIsActive) return;
	m_bIsActive = FALSE;
	m_bNeedsUnregister = TRUE;

	if (pCtrl && m_bRegWithMC) {
		m_bRegWithMC = FALSE;
		pCtrl->RemoveUser();
	}
	CloseObject();
	Reset();
	//invalidate scene graph to force a call to Render()
	GetPresenter()->InvalidateScene(NULL);
}

void AudioSourceProxy::UpdateTimeInfo()
{
	AudioSource *pAS = (AudioSource *) GetObject();

	if (!pAS->GetOD()) return;

	if (!m_bIsActive) {
		m_startTime = (double) pAS->startTime;
		m_Speed = pAS->speed;
	}


	double currentTime = GetFlexTime() / 1000.0;
	// Only proceed if the start time has been reached
	if ( (currentTime < m_startTime) || (m_startTime<0) ) return;
	
	double speed = (double) GetSpeed();

	//running, check for deactivation
	if (speed && m_bIsActive) {
		if ((pAS->stopTime > m_startTime) && (currentTime >= pAS->stopTime) ) {
			Deactivate();
			return;
		}
		//end of cycle detection is done through end of stream detection (see Render() )
	}


	/*we're (about to be) active: VRML:
	"A time-dependent node is inactive until its startTime is reached. When time now becomes greater than or 
	equal to startTime, an isActive TRUE event is generated and the time-dependent node becomes active 	*/

	// play
	if (!m_bIsActive) Activate();
}

void AudioSourceINIT::OnUpdate()
{
	((AudioSourceProxy *)GetRenderingObject())->OnUpdate();
}

void AudioSourceProxy::OnUpdate()
{
	AudioSource *pAS = (AudioSource *) GetObject();
	//cf note in TimedNodes.h
	PresenterContext ft_ctx;
	ft_ctx.SetTime(GetFlexTime());
	pAS->startTime.ApplyOffset(&ft_ctx);
	pAS->stopTime.ApplyOffset(&ft_ctx);

	if (m_bIsActive) {
		UpdateTimeInfo();
	} else {
		if (m_startTime != pAS->startTime) {
			if (!m_bIsRegistered) GetPresenter()->RegisterTimedNode(this);
			m_bIsRegistered = TRUE;
			m_bNeedsUnregister = FALSE;
		}
	}
}

BOOL AudioSourceProxy::IsMuted()
{
	if (m_bIsMuted) return TRUE;
	if (! m_bIsActive) return TRUE;
	AudioSource *pAS = (AudioSource *) GetObject();
	MediaControl *pCtrl = pAS->GetMediaControl();
	if (pCtrl && (Bool) pCtrl->mute) return TRUE;
	return FALSE;
}

float AudioSourceProxy::GetIntensity() 
{
	float ret = 1;
	if (m_pSound2D) {
		ret = m_pSound2D->intensity;
		if (ret>1.0) ret = 1.0;
	}
	return ret;
}

float AudioSourceProxy::GetSpeed()
{
	AudioSource *pAS = (AudioSource *) GetObject();
	MediaControl *pCtrl = pAS->GetMediaControl();
	if (pCtrl) return pCtrl->mediaSpeed;
	return pAS->speed;
}
