/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  Background2DProxy.cpp
//
// Author: Jean Le Feuvre - ENST 2002
//
// Description:
//   proxy for Background2D node.
//
// Created: August 2002.
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-INIT.h"
#include "Background2DProxy.h"
#include "DrawableImage.h"
#include "Bindable.h"
#include "GroupNodeProxy.h"
#include "Presenter.h"

Background2DProxy* Presenter::InitBackground2D(Background2D* pObj)
{
	return (new Background2DProxy(pObj)); 
}



Background2DProxy::Background2DProxy(MediaObject* pMO) : DrawableImage(pMO)
{
	m_firstTime = TRUE;
}

Background2DProxy::~Background2DProxy() 
{
}

void Background2DProxy::DeleteObject() 
{
	BackgroundStatus *status;
	while (m_Surfaces.GetLength() ) {
		status = m_Surfaces[0];
		status->m_pStack->RemoveBindable(this);
		m_Surfaces.DeleteFrom(0);
		delete status;
	}
}

void Background2DProxy::Render(Effects *effects)
{
	BackgroundStatus *status;

	//this will register the background
	status = GetStatus(*effects);
	if (!status) return;

	if (effects->m_bShouldRebuild) status->m_Context.m_bNeedsDrawing = TRUE;

	// The first time that this is called, the stream may need to be started.
	if (m_firstTime) {
		if ((GetBackground2D()->GetUrl(0)) != NULL)
			GetBackground2D()->OpenStream();

		//if this is the top of the stack, bound it
		if (status->m_pStack->IsOnTop(this))
			GetBackground2D()->isBound = TRUE;
		
		m_firstTime = FALSE;
		
		//here we have a pb: it is not possible to register the background with its parent stack
		//until the first draw() since sevral node use background stacks (in VRML only one is defined
		//thus the pb doesn't occur). If we get here this means we need to invalidate the scene at next rendering
		//to draw the background first (direct mode only)
		if (!effects->m_bDrawBackground && !effects->m_bUseBoundTracking) {
			GetPresenter()->InvalidateScene(NULL);
		}
	}

	//do nothing if the Background is not bound
	if ( !(Bool) GetBackground2D()->isBound ) 
		return;

	//not drawing and no bounds tracking: nothing to do
	if (!effects->m_bDrawBackground && !effects->m_bUseBoundTracking) 
		return;

	//only update if no parent (clipping) group, otherwise the bounds of the background are setup by the parent 
	if (! effects->m_ParentGroup) {
		VisualPath _screen;
		BRect screenBRect;
		_screen.AddRectangle(status->m_Context.m_pVisualSurface->GetWidth(), status->m_Context.m_pVisualSurface->GetHeight());
		_screen.GetBoundingRect(&screenBRect, NULL, 0);
		
		//the background always fills the parent layer
		status->m_Context.m_boundingRect.SetDimensions(screenBRect);
		status->m_Context.m_clippedBoundingRect = status->m_Context.m_boundingRect.GetClipper();
	}


	//update texture 
	if ((GetBackground2D()->GetUrl(0)) != NULL) {
		Update(&status->m_Context);
		//texture is not ready, don't draw backGroound but invalidate next render cycle
		if (!m_pTexture) {
			GetPresenter()->InvalidateScene(NULL);
			return;
		}
	} else {
		Color3f col = GetBackground2D()->backColor;
		if (! (col == status->m_previousBackColor)) {
			status->m_previousBackColor = col;
			status->m_Context.m_bNeedsDrawing = TRUE;
		}
	}

	if (effects->m_bDrawBackground)
		FinalizeRender(effects, &status->m_Context);
}

void Background2DProxy::Draw(DrawableContext *pContext)
{
	VisualPath path;
	assert(pContext->m_pVisualSurface);

	//this means there's no background activated and we're on top of the stack
	if (! (Bool) GetBackground2D()->isBound) return;

	if (pContext->m_clippedBoundingRect.IsEmpty()) return;

	pContext->m_pVisualSurface->ResetStyles();
	pContext->m_pVisualSurface->SetClip(& pContext->m_clippedBoundingRect);

	if ((GetBackground2D()->GetUrl(0))!=NULL) {
		if (!m_pTexture) return;

		path.AddRectangle(pContext->m_boundingRect.left, pContext->m_boundingRect.top, pContext->m_boundingRect.width, pContext->m_boundingRect.height);

		pContext->m_pVisualSurface->SetBrushStyle(0, GetBackground2D()->backColor, 0);
		m_pTexture->SetTransparency(0);
		m_pTexture->SetTextureRepeat(FALSE, FALSE);
		pContext->m_pVisualSurface->TextureVisualPath(&path, m_pTexture);
		Release();
	} else {
		//use only the given rect
		path.AddRectangle(pContext->m_clippedBoundingRect.left, pContext->m_clippedBoundingRect.top, 
						pContext->m_clippedBoundingRect.width, pContext->m_clippedBoundingRect.height);
		
		//draw with linewidth 2 (this is after some testing with Gdiplus)
		pContext->m_pVisualSurface->SetLineStyle(GetBackground2D()->backColor, 0.0, 2.0);
		pContext->m_pVisualSurface->SetBrushStyle(1, GetBackground2D()->backColor, 0);
		pContext->m_pVisualSurface->DrawVisualPath(&path);
	}
	pContext->m_bNeedsDrawing = FALSE;
}


void Background2D::Onset_bind() 
{
	Background2DProxy *pBackOnTop, *pBackground2DProxy;
	pBackground2DProxy = ((Background2DProxy*) GetRenderingObject());

	//notify all stacks using this node
	for (int i=0; i<pBackground2DProxy->m_Surfaces.GetLength();i++) {
		BackgroundStatus *status = pBackground2DProxy->m_Surfaces[i];
		BindableStack *pBackStack = status->m_pStack;
		BOOL isOnTop = pBackStack->IsOnTop(pBackground2DProxy);

		GetPresenter()->InvalidateScene(NULL);
	
		if (! (Bool) set_bind) {
			//generate an isBound=FALSE (only in the first iteration)
			if ((Bool) isBound) 
				isBound = FALSE;

			//if on top remove from top and pop the stack
			if (isOnTop) {
				pBackStack->RemoveFromTop(pBackground2DProxy);
				pBackOnTop = (Background2DProxy *) pBackStack->GetTop();
				if ((pBackOnTop != NULL) && (pBackOnTop != pBackground2DProxy)) {
					pBackOnTop->GetBackground2D()->set_bind = TRUE;
					//the core assumes events are only written through the scene graph, so
					//force manual update
					pBackOnTop->GetBackground2D()->Onset_bind();
				}
			}
		} else {
			//we're getting active
			if (! (Bool) isBound) {
				status->m_Context.m_bNeedsDrawing = TRUE;
				isBound = TRUE;
			}

			//if we're already on top do nothing
			if (!isOnTop) {
				//get previous top if any
				pBackOnTop = (Background2DProxy *) pBackStack->GetTop();
				//move to top
				pBackStack->MoveToTop(pBackground2DProxy);
				//unbind prev top if not this one
				if ( (pBackOnTop != NULL) && (pBackOnTop != pBackground2DProxy)) {
					pBackOnTop->GetBackground2D()->set_bind = FALSE;
					//the core assumes events are only written through the scene graph, so
					//force manual update
					pBackOnTop->GetBackground2D()->Onset_bind();
				}

			}
		}
	}
}


BackgroundStatus *Background2DProxy::GetStatus(Effects & effects)
{
	BackgroundStatus *status;
	BindableStack *stack;

	stack = NULL;
	if (effects.m_ParentGroup) {
		stack = effects.m_ParentGroup->GetBackgroundStack();
	}
	if (!stack) {
		stack = effects.m_pSurface->GetBackgroundStack();
	}
	if (!stack) return NULL;

	for (int i=0; i<m_Surfaces.GetLength(); i++) {
		status = m_Surfaces[i];
		if (status->m_pStack == stack) return status;
	}
	//not found, register it
	status = new BackgroundStatus;
	status->m_Context.m_pDrawable = this;
	status->m_Context.ResetHandlers();
	status->m_Context.m_pAppearance = NULL;
	status->m_Context.m_pTextureProxy = NULL;
	status->m_Context.m_hasStreamFinished = FALSE;
	status->m_Context.m_transform.SetIdentity();
	status->m_Context.m_bIsTransparent = FALSE;
	status->m_Context.m_bNeedsDrawing = TRUE;
	status->m_Context.m_isTextNode = FALSE;

	status->m_Context.m_pVisualSurface = effects.m_pSurface;
	status->m_pStack = stack;

	m_Surfaces += status;
	stack->AddBindable(this);
	status->m_previousBackColor = Color3f(0.0, 0.0, 0.0);
	return status;
}

DrawableContext *Background2DProxy::GetContext(BindableStack *stack)
{
	BackgroundStatus *status;
	for (int i=0; i<m_Surfaces.GetLength(); i++) {
		status = m_Surfaces[i];
		if (status->m_pStack == stack) return &status->m_Context;
	}
	return NULL;
}
