/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  BitmapProxy.cpp
//
// Author: Jean Le Feuvre, ENST 2002
//
// Description:
//   proxy object for Bitmap node.
//
// Created: august 2002.
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-INIT.h"
#include "BitmapProxy.h"
#include "DrawableImage.h"
#include "Stream.h"
#include "Presenter.h"

BitmapProxy * Presenter::InitBitmap(Bitmap* pObj)
{
	return new BitmapProxy(pObj);
}

BitmapProxy::BitmapProxy(MediaObject* pMO) : GeometryNode(pMO)
{
}

void BitmapProxy::Render(Effects *effects)
{
	//no rotation
	if (effects->transform.HasRotations()) return;

	DrawableContext *pContext=InitRenderContext(effects);
	if (!pContext) return;
	Matrix2D mat;

	VisualTexture *pTexture = GetTexture(pContext);
	//texture not init...
	if (!pTexture || !pTexture->m_Width || !pTexture->m_Height) 
		return;

	BuildGraph(pContext);

	//compute overall scaling
	Vec2f scale = GetBitmap()->scale;
	mat.AddScale((scale.x>0) ? scale.x : 1.0, (scale.y>0) ? scale.y : 1.0, 0.0);
	mat.AddTransform(&effects->transform);

	m_Path.GetBoundingRect(& pContext->m_boundingRect, &mat, 0);
	//default for clipping rect is bounding rect
	pContext->m_clippedBoundingRect =  pContext->m_boundingRect;

	//bitmap is transparent if the image is
	if (pContext->m_pClipper) {
		pContext->m_bIsTransparent = TRUE;
	} else {
		pContext->m_bIsTransparent = pTexture->m_bIsTransparent;
		//or has material keying
		MaterialKey *mk = GetMaterialKey(pContext);
		if (mk  && (BOOL) (Bool) mk->isKeyed) {
			pContext->m_bIsTransparent = TRUE;
		}
	}

	FinalizeRender(effects, pContext);
}

// BitmapProxy::Draw
void BitmapProxy::Draw(DrawableContext * pContext)
{
	if (!pContext->m_pTextureProxy) return;

	if (pContext->m_clippedBoundingRect.IsEmpty()) {
		pContext->m_pTextureProxy->Release();
		return;
	}

	VisualTexture *pTexture = GetTexture(pContext);	
	if (!pTexture) return;

	MaterialKey *mk = GetMaterialKey(pContext);
	if (mk  && ! (BOOL) (Bool) mk->isKeyed) mk = NULL;

	//set texture transparency (init the brush)
	pTexture->SetTransparency(pContext->gProps.transparency);
	
	//consider the bitmap as a textured rectangle if the associated texture is valid for texturing
	//(aka supported color space) and no materialKey is used (not supported through textures yet)
	if (!mk && pTexture->HasValidBrush()) {

		if (!SetupDraw(pContext)) return;
		
		//override linestyle to empty
		pContext->m_pVisualSurface->SetLineStyle(Color3f(0.0), 0.0, 0.0);

		Matrix2D mat;
		Vec2f scale = GetBitmap()->scale;
		mat.AddScale((scale.x>0) ? scale.x : 1.0, (scale.y>0) ? scale.y : 1.0, 0.0);
		mat.AddTransform(&pContext->m_transform);
		pContext->m_pVisualSurface->SetTransform(&mat);
		//also note we need to reset our path since a single bitmap object can be used with
		//several textures, thus will have different sizes
		BuildGraph(pContext);
		
		DrawTexture(pContext, NULL, TRUE);
		return;
	}

	pContext->m_clippedBoundingRect.Intersection(pContext->m_pVisualSurface->GetClipper());

	pContext->m_pVisualSurface->DrawBitmap(&pContext->m_clippedBoundingRect, 
											&pContext->m_boundingRect, 
											pTexture, mk);	

	pContext->m_pTextureProxy->Release();

}


void BitmapProxy::BuildGraph(DrawableContext *pContext)
{
	Matrix2D mat;

	VisualTexture *pTexture = GetTexture(pContext);

	m_Path.Reset();

	float w = pTexture->m_Width;
	float h = pTexture->m_Height;
	pContext->m_pVisualSurface->InitObjectMatrix(mat);
	mat.ReverseApply(w, h);
	m_Path.AddRectangle(w, h);
}
