/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File: ColorGradient.cpp
//
// Author: Jean Le Feuvre
//
// Description:
//		Object used to store color gradient (pen and brush)
//
// Created: August 2002
//
///////////////////////////////////////////////////////////////////////////

#include "defs.h"
#include "ColorGradient.h"

ColorGradient::ColorGradient()
{
	m_pBrush = NULL;
	m_pLinearBrush = NULL;
	m_pPath = NULL;
}

ColorGradient::~ColorGradient()
{
	Reset();
}

void ColorGradient::Reset()
{
	if (m_pBrush) delete m_pBrush;
	m_pBrush = NULL;
	if (m_pLinearBrush) delete m_pLinearBrush;
	m_pLinearBrush = NULL;
	if (m_pPath) delete m_pPath;
	m_pPath = NULL;

	m_OriginalLinearMatrix.SetIdentity();
	m_Matrix.SetIdentity();
}

BOOL ColorGradient::SetVisualPath(VisualPath *path)
{
	if (!path) return FALSE;
	Reset();
	m_pBrush = new Gdiplus::PathGradientBrush(&path->m_Path);
	m_pPath = path->m_Path.Clone();
	return TRUE;
}

BOOL ColorGradient::SetGradientPoints(PointF* pts, unsigned int numPoints)
{
	if (!pts) return FALSE;
	Reset();
	Gdiplus::PointF *g_pts = new Gdiplus::PointF[numPoints];
	for (unsigned int i=0;i<numPoints;i++) {
		g_pts[i].X = pts[i].x;
		g_pts[i].Y = pts[i].y;
	}
	m_pBrush = new Gdiplus::PathGradientBrush(g_pts, numPoints);
	delete [] g_pts;
	return TRUE;
}

BOOL ColorGradient::SetGradientColors(Color3f *col, unsigned int numColors, float transparency)
{
	unsigned int alpha = 255 * (1-transparency);
	Gdiplus::Color *colors;
	colors = new Gdiplus::Color[numColors];
	for (unsigned int i=0;i<numColors;i++) {
		Gdiplus::ARGB c = Gdiplus::Color::MakeARGB(alpha, 255*col[i].r, 255*col[i].g, 255*col[i].b);
		colors[i].SetValue(c);
	}
	m_pBrush->SetSurroundColors(colors, (int *) &numColors);
	delete [] colors;
	return TRUE;
}


BOOL ColorGradient::SetGradientColors(Color4 *col, unsigned int numColors)
{
	Gdiplus::Color *colors;
	colors = new Gdiplus::Color[numColors];
	for (unsigned int i=0;i<numColors;i++) {
		Gdiplus::ARGB c = Gdiplus::Color::MakeARGB(col[i].a, col[i].r, col[i].g, col[i].b);
		colors[i].SetValue(c);
	}
	m_pBrush->SetSurroundColors(colors, (int *) &numColors);
	delete [] colors;
	return TRUE;
}

BOOL ColorGradient::SetBlendColors(Color4 *col, float *blendFactors, unsigned int numColors)
{
	Gdiplus::Color *colors;
	Gdiplus::REAL *reals;
	colors = new Gdiplus::Color[numColors];
	reals = new Gdiplus::REAL[numColors];

	//gdiplus take blends only from bounds to center
	int k = 0;
	for (unsigned int i=numColors; i>0; i--) {
		Gdiplus::ARGB c = Gdiplus::Color::MakeARGB(col[i-1].a, col[i-1].r, col[i-1].g, col[i-1].b);
		colors[k].SetValue(c);
		reals[k] = 1 - blendFactors[i-1];
		k++;
	}
	m_pBrush->SetInterpolationColors(colors, reals, numColors);
	m_BackColor = col[numColors-1];

	delete [] colors;
	delete [] reals;
	return TRUE;

}

BOOL ColorGradient::SetCenterColor(float x, float y, Color3f col, float transparency)
{
	if (!m_pBrush) return FALSE;
	Gdiplus::PointF pt(x, y);
	m_pBrush->SetCenterPoint(pt);

	Gdiplus::Color g_col(255*(1-transparency), 255*col.r, 255*col.g, 255*col.b);
	m_pBrush->SetCenterColor(g_col);
	return TRUE;
}

BOOL ColorGradient::SetLinearGradient(float startx, float starty, float endx, float endy, Color3f start_col, Color3f end_col, float transparency)
{
	Reset();
	Gdiplus::PointF start(startx, starty);
	Gdiplus::PointF end(endx, endy);
	Gdiplus::Color startCol(255*(1-transparency), 255*start_col.r, 255*start_col.g, 255*start_col.b);
	Gdiplus::Color endCol(255*(1-transparency), 255*end_col.r, 255*end_col.g, 255*end_col.b);
	m_pLinearBrush = new Gdiplus::LinearGradientBrush(start, end, startCol, endCol); 

	m_pLinearBrush->GetTransform(&m_OriginalLinearMatrix.m_Matrix);
	return TRUE;
}

BOOL ColorGradient::SetLinearGradient(float startx, float starty, float endx, float endy, Color4 *col, float *blendFactors, unsigned int numColors)
{
	Reset();

	if (numColors<2) return FALSE;
	Gdiplus::PointF start(startx, starty);
	Gdiplus::PointF end(endx, endy);
	Gdiplus::Color startCol(col[0].a, col[0].r, col[0].g, col[0].b);
	Gdiplus::Color endCol(col[numColors-1].a, col[numColors-1].r, col[numColors-1].g, col[numColors-1].b);
	m_pLinearBrush = new Gdiplus::LinearGradientBrush(start, end, startCol, endCol); 

	m_pLinearBrush->GetTransform(&m_OriginalLinearMatrix.m_Matrix);

	if (numColors==2) return TRUE;

	Gdiplus::Color *colors;
	colors = new Gdiplus::Color[numColors];

	for (unsigned int i=0; i<numColors; i++) {
		Gdiplus::ARGB c = Gdiplus::Color::MakeARGB(col[i].a, col[i].r, col[i].g, col[i].b);
		colors[i].SetValue(c);
	}
	m_pLinearBrush->SetInterpolationColors(colors, blendFactors, numColors);
	delete [] colors;
	return TRUE;
}

void ColorGradient::SetTransform(Matrix2D *pMat)
{
	m_Matrix.SetIdentity();
	if (!pMat) return;
	if (m_pLinearBrush) m_Matrix.m_Matrix.Multiply(&m_OriginalLinearMatrix.m_Matrix);
	m_Matrix.AddTransform(pMat);
}

BOOL ColorGradient::SetRepeatMode(enum GRADIENT_TILE mode)
{
	if (!m_pBrush && !m_pLinearBrush) return FALSE;
	switch (mode) {
	case GRADIENT_TILE_FLIP:
		if (m_pBrush) m_pBrush->SetWrapMode(Gdiplus::WrapModeTileFlipXY);
		if (m_pLinearBrush) m_pLinearBrush->SetWrapMode(Gdiplus::WrapModeTileFlipXY);
		break;
	case GRADIENT_TILE_BASE:
		if (m_pBrush) m_pBrush->SetWrapMode(Gdiplus::WrapModeTile);
		if (m_pLinearBrush) m_pLinearBrush->SetWrapMode(Gdiplus::WrapModeTile);
		break;
	case GRADIENT_TILE_PAD:
	default:
		if (m_pBrush) m_pBrush->SetWrapMode(Gdiplus::WrapModeClamp);
		//a linear gradient brush is always tiled in GDIplus :(
//		if (m_pLinearBrush) m_pLinearBrush->SetWrapMode(Gdiplus::WrapModeClamp);
		break;
	}
	return TRUE;
}
