/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File: ColorTools.cpp
//
// Author: Jean Le Feuvre
//
// Description:
//		color tools (yuv->rgb, blending, ...)
//			original YUV2RGB routines from xvid
// Created: February 2003
//
///////////////////////////////////////////////////////////////////////////

#include "defs.h"
#include "ColorTools.h"



#define MIN(X, Y) ((X)<(Y)?(X):(Y))
#define MAX(X, Y) ((X)>(Y)?(X):(Y))

#define MIN3(x,y,z) (((x) < (y)) ? min((x),(z)) : min((y),(z)))

BOOL SetUpCopyParameters(CopyParameters& cpParams, SrcImage const& srcImage, SurfaceDesc const& surfaceDesc)
{
	// Determine the region of the source frame which needs to be copied.
	//  - take into account offset and dimensions of destination surface.
	cpParams.srcStartX = (srcImage.left < 0) ? (-1 * srcImage.left) : 0;
	cpParams.srcStartY = (srcImage.top < 0)  ? (-1 * srcImage.top)  : 0;
	cpParams.copyWidth  = MIN3(srcImage.width - cpParams.srcStartX,
							   surfaceDesc.dwWidth - srcImage.left,
							   surfaceDesc.dwWidth);
	cpParams.copyHeight = MIN3(srcImage.height - cpParams.srcStartY,
							   surfaceDesc.dwHeight - srcImage.top,
							   surfaceDesc.dwHeight);

	// Determine the top-left point in the destination frame
	cpParams.destStartX = (srcImage.left < 0) ? 0 : srcImage.left;
	cpParams.destStartY = (srcImage.top < 0)  ? 0 : srcImage.top;

	// Only proceed if the image will actually appear
	if ((cpParams.srcStartX >= srcImage.width) ||
		(cpParams.srcStartY >= srcImage.height) ||
		(srcImage.left >= (int)surfaceDesc.dwWidth) ||
		(srcImage.top >= (int)surfaceDesc.dwHeight))
		return TRUE;

	
	return FALSE;
} 

int RGB_Y[256];
int B_U[256];
int G_U[256];
int G_V[256];
int R_V[256];

#define SCALEBITS_OUT	13
#define FIX_OUT(x)		((unsigned short) ((x) * (1L<<SCALEBITS_OUT) + 0.5))

static int is_init = 0;

/**/
void yuv2rgb_init(void) 
{
	if (!is_init) {
		is_init = 1;
		for(int i = 0; i < 256; i++) {
			RGB_Y[i] = FIX_OUT(1.164) * (i - 16);
			B_U[i] = FIX_OUT(2.018) * (i - 128);
			G_U[i] = FIX_OUT(0.391) * (i - 128);
			G_V[i] = FIX_OUT(0.813) * (i - 128);
			R_V[i] = FIX_OUT(1.596) * (i - 128);
		}
	}
}


void yuv2rgb_555(unsigned char *dst, int dst_stride,
				 unsigned char *y_src, unsigned char * u_src, unsigned char * v_src,
				 int y_stride, int uv_stride, int width, int height, unsigned char const_a, MaterialKey *mk)
{
	yuv2rgb_init();

	const unsigned int dst_dif = 2 * dst_stride - 2 * width;
	int y_dif = 2 * y_stride - width;

	unsigned char *dst2 = dst + dst_stride;
	unsigned char *y_src2 = y_src + y_stride;
	unsigned int x, y;
	unsigned char a;

	if (height < 0) return;

	for (y = height / 2; y; y--) {
		int r, g, b, r2, g2, b2;
		r = g = b = r2 = g2 = b2 = 0;

		for (x = 0; x < (unsigned int) width / 2; x++) {
			int u, v, rgb;
			int b_u, g_uv, r_v, rgb_y;

			u = u_src[x];
			v = v_src[x];

			b_u = B_U[u];
			g_uv = G_U[u] + G_V[v];
			r_v = R_V[v];

			rgb_y = RGB_Y[*y_src];
			b = (b & 0x7) + ((rgb_y + b_u) >> SCALEBITS_OUT);
			g = (g & 0x7) + ((rgb_y - g_uv) >> SCALEBITS_OUT);
			r = (r & 0x7) + ((rgb_y + r_v) >> SCALEBITS_OUT);
			rgb = (int) *(unsigned short *) dst;

			a = mk ? GetAlphaKey(mk, r, g, b) : const_a;
			b = ApplyAlphaMask(b, a, ( (rgb << 3) & 0xf8) );
			g = ApplyAlphaMask(g, a, ( (rgb >> 2) & 0xf8) );
			r = ApplyAlphaMask(r, a, ( (rgb >> 7) & 0xf8) );
			*(unsigned short *) dst = MAKE_RGB555(r, g, b);

			y_src++;
			rgb_y = RGB_Y[*y_src];
			b = (b & 0x7) + ((rgb_y + b_u) >> SCALEBITS_OUT);
			g = (g & 0x7) + ((rgb_y - g_uv) >> SCALEBITS_OUT);
			r = (r & 0x7) + ((rgb_y + r_v) >> SCALEBITS_OUT);
			rgb = (int) *(unsigned short *) (dst + 2);
			a = mk ? GetAlphaKey(mk, r, g, b) : const_a;
			b = ApplyAlphaMask(b, a, ( (rgb << 3) & 0xf8) );
			g = ApplyAlphaMask(g, a, ( (rgb >> 2) & 0xf8) );
			r = ApplyAlphaMask(r, a, ( (rgb >> 7) & 0xf8) );
			*(unsigned short *) (dst + 2) = MAKE_RGB555(r, g, b);
			y_src++;

			rgb_y = RGB_Y[*y_src2];
			b2 = (b2 & 0x7) + ((rgb_y + b_u) >> SCALEBITS_OUT);
			g2 = (g2 & 0x7) + ((rgb_y - g_uv) >> SCALEBITS_OUT);
			r2 = (r2 & 0x7) + ((rgb_y + r_v) >> SCALEBITS_OUT);
			rgb = (int) *(unsigned short *) dst2;
			a = mk ? GetAlphaKey(mk, r2, g2, b2) : const_a;
			b2 = ApplyAlphaMask(b2, a, ( (rgb << 3) & 0xf8) );
			g2 = ApplyAlphaMask(g2, a, ( (rgb >> 2) & 0xf8) );
			r2 = ApplyAlphaMask(r2, a, ( (rgb >> 7) & 0xf8) );
			*(unsigned short *) (dst2) = MAKE_RGB555(r2, g2, b2);
			y_src2++;

			rgb_y = RGB_Y[*y_src2];
			b2 = (b2 & 0x7) + ((rgb_y + b_u) >> SCALEBITS_OUT);
			g2 = (g2 & 0x7) + ((rgb_y - g_uv) >> SCALEBITS_OUT);
			r2 = (r2 & 0x7) + ((rgb_y + r_v) >> SCALEBITS_OUT);
			rgb = (int) *(unsigned short *) (dst2 + 2);
			a = mk ? GetAlphaKey(mk, r2, g2, b2) : const_a;
			b2 = ApplyAlphaMask(b2, a, ( (rgb << 3) & 0xf8) );
			g2 = ApplyAlphaMask(g2, a, ( (rgb >> 2) & 0xf8) );
			r2 = ApplyAlphaMask(r2, a, ( (rgb >> 7) & 0xf8) );
			*(unsigned short *) (dst2 + 2) = MAKE_RGB555(r2, g2, b2);
			y_src2++;

			dst += 4;
			dst2 += 4;
		}

		dst += dst_dif;
		dst2 += dst_dif;

		y_src += y_dif;
		y_src2 += y_dif;

		u_src += uv_stride;
		v_src += uv_stride;
	}
}


void yuv2rgb_565(unsigned char * dst, int dst_stride,
				 unsigned char* y_src, unsigned char* u_src, unsigned char* v_src,
				 int y_stride, int uv_stride, int width, int height, unsigned char const_a, MaterialKey *mk)
{
	yuv2rgb_init();

	const unsigned int dst_dif = 2 * dst_stride - 2 * width;
	int y_dif = 2 * y_stride - width;

	unsigned char *dst2 = dst + dst_stride;
	unsigned char *y_src2 = y_src + y_stride;
	unsigned int x, y, rgb;
	unsigned char a;

	if (height < 0) return;

	for (y = height / 2; y; y--) {
		int r, g, b, r2, g2, b2;
		r = g = b = r2 = g2 = b2 = 0;

		for (x = 0; x < (unsigned int) width / 2; x++) {
			int u, v;
			int b_u, g_uv, r_v, rgb_y;

			u = u_src[x];
			v = v_src[x];

			b_u = B_U[u];
			g_uv = G_U[u] + G_V[v];
			r_v = R_V[v];

			rgb_y = RGB_Y[*y_src];
			b = clip_col( (rgb_y + b_u) >> SCALEBITS_OUT);
			g = clip_col( (rgb_y - g_uv) >> SCALEBITS_OUT);
			r = clip_col( (rgb_y + r_v) >> SCALEBITS_OUT);
			rgb = (int) *(unsigned short *) dst;
			a = mk ? GetAlphaKey(mk, r, g, b) : const_a;
			b = ApplyAlphaMask(b, a, ( (rgb << 3) & 0xf8) );
			g = ApplyAlphaMask(g, a, ( (rgb >> 3) & 0xfc) );
			r = ApplyAlphaMask(r, a, ( (rgb >> 8) & 0xf8) );
			*(unsigned short *) dst = MAKE_RGB565(r, g, b);

			y_src++;
			rgb_y = RGB_Y[*y_src];
			b = clip_col( (rgb_y + b_u) >> SCALEBITS_OUT);
			g = clip_col( (rgb_y - g_uv) >> SCALEBITS_OUT);
			r = clip_col( (rgb_y + r_v) >> SCALEBITS_OUT);
			rgb = (int) *(unsigned short *) (dst+2);
			a = mk ? GetAlphaKey(mk, r, g, b) : const_a;
			b = ApplyAlphaMask(b, a, ( (rgb << 3) & 0xf8) );
			g = ApplyAlphaMask(g, a, ( (rgb >> 3) & 0xfc) );
			r = ApplyAlphaMask(r, a, ( (rgb >> 8) & 0xf8) );
			*(unsigned short *) (dst + 2) = MAKE_RGB565(r, g, b);
			y_src++;

			rgb_y = RGB_Y[*y_src2];
			b2 = clip_col( (rgb_y + b_u) >> SCALEBITS_OUT);
			g2 = clip_col( (rgb_y - g_uv) >> SCALEBITS_OUT);
			r2 = clip_col( (rgb_y + r_v) >> SCALEBITS_OUT);
			rgb = (int) *(unsigned short *) dst2;
			a = mk ? GetAlphaKey(mk, r2, g2, b2) : const_a;
			b2 = ApplyAlphaMask(b2, a, ( (rgb << 3) & 0xf8) );
			g2 = ApplyAlphaMask(g2, a, ( (rgb >> 3) & 0xfc) );
			r2 = ApplyAlphaMask(r2, a, ( (rgb >> 8) & 0xf8) );
			*(unsigned short *) (dst2) = MAKE_RGB565(r2, g2, b2);
			y_src2++;

			rgb_y = RGB_Y[*y_src2];
			b2 = clip_col( (rgb_y + b_u) >> SCALEBITS_OUT);
			g2 = clip_col( (rgb_y - g_uv) >> SCALEBITS_OUT);
			r2 = clip_col( (rgb_y + r_v) >> SCALEBITS_OUT);
			rgb = (int) *(unsigned short *) (dst2+2);
			a = mk ? GetAlphaKey(mk, r2, g2, b2) : const_a;
			b2 = ApplyAlphaMask(b2, a, ( (rgb << 3) & 0xf8) );
			g2 = ApplyAlphaMask(g2, a, ( (rgb >> 3) & 0xfc) );
			r2 = ApplyAlphaMask(r2, a, ( (rgb >> 8) & 0xf8) );
			*(unsigned short *) (dst2 + 2) = MAKE_RGB565(r2, g2, b2);
			y_src2++;

			dst += 4;
			dst2 += 4;
		}

		dst += dst_dif;
		dst2 += dst_dif;
		y_src += y_dif;
		y_src2 += y_dif;
		u_src += uv_stride;
		v_src += uv_stride;
	}
}

void yuv2rgb_24(unsigned char *dst, int dststride, unsigned char *y_src, unsigned char *u_src, unsigned char * v_src, 
				 int y_stride, int uv_stride, int width, int height, unsigned char const_a, MaterialKey *mk)
{
	yuv2rgb_init();

	const unsigned int dst_dif = 2 * dststride - 3 * width;
	int y_dif = 2 * y_stride - width;
	
	unsigned char *dst2 = dst + dststride;
	unsigned char *y_src2 = y_src + y_stride;
	unsigned int x, y;
	unsigned char a;
	if (height < 0) return;

	for (y = height / 2; y; y--) {
		for (x = 0; x < (unsigned int)width / 2; x++) {
			int u, v;
			int b_u, g_uv, r_v, rgb_y;
			int r, g, b;

			u = u_src[x];
			v = v_src[x];

			b_u = B_U[u];
			g_uv = G_U[u] + G_V[v];
			r_v = R_V[v];

			rgb_y = RGB_Y[*y_src];
			b = clip_col( (rgb_y + b_u) >> SCALEBITS_OUT);
			g = clip_col( (rgb_y - g_uv) >> SCALEBITS_OUT);
			r = clip_col( (rgb_y + r_v) >> SCALEBITS_OUT);
			a = mk ? GetAlphaKey(mk, r, g, b) : const_a;
			dst[0] = ApplyAlphaMask(b, a, dst[0]);
			dst[1] = ApplyAlphaMask(g, a, dst[1]);
			dst[2] = ApplyAlphaMask(r, a, dst[2]);
			y_src++;

			rgb_y = RGB_Y[*y_src];
			b = clip_col( (rgb_y + b_u) >> SCALEBITS_OUT);
			g = clip_col( (rgb_y - g_uv) >> SCALEBITS_OUT);
			r = clip_col( (rgb_y + r_v) >> SCALEBITS_OUT);
			a = mk ? GetAlphaKey(mk, r, g, b) : const_a;
			dst[3] = ApplyAlphaMask(b, a, dst[3]);
			dst[4] = ApplyAlphaMask(g, a, dst[4]);
			dst[5] = ApplyAlphaMask(r, a, dst[5]);
			y_src++;

			rgb_y = RGB_Y[*y_src2];
			b = clip_col( (rgb_y + b_u) >> SCALEBITS_OUT);
			g = clip_col( (rgb_y - g_uv) >> SCALEBITS_OUT);
			r = clip_col( (rgb_y + r_v) >> SCALEBITS_OUT);
			a = mk ? GetAlphaKey(mk, r, g, b) : const_a;
			dst2[0] = ApplyAlphaMask(b, a, dst2[0]);
			dst2[1] = ApplyAlphaMask(g, a, dst2[1]);
			dst2[2] = ApplyAlphaMask(r, a, dst2[2]);
			y_src2++;

			rgb_y = RGB_Y[*y_src2];
			b = clip_col( (rgb_y + b_u) >> SCALEBITS_OUT);
			g = clip_col( (rgb_y - g_uv) >> SCALEBITS_OUT);
			r = clip_col( (rgb_y + r_v) >> SCALEBITS_OUT);
			a = mk ? GetAlphaKey(mk, r, g, b) : const_a;
			dst2[3] = ApplyAlphaMask(b, a, dst2[3]);
			dst2[4] = ApplyAlphaMask(g, a, dst2[4]);
			dst2[5] = ApplyAlphaMask(r, a, dst2[5]);
			y_src2++;

			dst += 6;
			dst2 += 6;
		}

		dst += dst_dif;
		dst2 += dst_dif;

		y_src += y_dif;
		y_src2 += y_dif;

		u_src += uv_stride;
		v_src += uv_stride;
	}
}


void yuv2rgb_32(unsigned char *dst, int dststride, unsigned char *y_src, unsigned char *v_src, unsigned char * u_src,
				 int y_stride, int uv_stride, int width, int height, unsigned char const_a, MaterialKey *mk)
{
	yuv2rgb_init();

	const unsigned int dst_dif = 2 * dststride - 4 * width;
	int y_dif = 2 * y_stride - width;
	
	unsigned int x, y;
	unsigned char *dst2 = dst + dststride;
	unsigned char *y_src2 = y_src + y_stride;
	unsigned char a;
	for (y = height / 2; y; y--) {
		for (x = 0; x < (unsigned int)width / 2; x++) {
			int u, v;
			int b_u, g_uv, r_v, rgb_y;
			int r, g, b;

			u = u_src[x];
			v = v_src[x];

			b_u = B_U[u];
			g_uv = G_U[u] + G_V[v];
			r_v = R_V[v];

			rgb_y = RGB_Y[*y_src];
			b = clip_col( (rgb_y + b_u) >> SCALEBITS_OUT );
			g = clip_col( (rgb_y - g_uv) >> SCALEBITS_OUT );
			r = clip_col( (rgb_y + r_v) >> SCALEBITS_OUT );
			a = mk ? GetAlphaKey(mk, r, g, b) : const_a;
			dst[0] = ApplyAlphaMask(r, a, dst[0]);
			dst[1] = ApplyAlphaMask(g, a, dst[1]);
			dst[2] = ApplyAlphaMask(b, a, dst[2]);
			dst[3] = 255;
			y_src++;

			rgb_y = RGB_Y[*y_src];
			b = clip_col( (rgb_y + b_u) >> SCALEBITS_OUT);
			g = clip_col( (rgb_y - g_uv) >> SCALEBITS_OUT);
			r = clip_col( (rgb_y + r_v) >> SCALEBITS_OUT);
			a = mk ? GetAlphaKey(mk, r, g, b) : const_a;
			dst[4] = ApplyAlphaMask(r, a, dst[4]);
			dst[5] = ApplyAlphaMask(g, a, dst[5]);
			dst[6] = ApplyAlphaMask(b, a, dst[6]);
			dst[7] = 255;
			y_src++;

			rgb_y = RGB_Y[*y_src2];
			b = clip_col( (rgb_y + b_u) >> SCALEBITS_OUT);
			g = clip_col( (rgb_y - g_uv) >> SCALEBITS_OUT);
			r = clip_col( (rgb_y + r_v) >> SCALEBITS_OUT);
			a = mk ? GetAlphaKey(mk, r, g, b) : const_a;
			dst2[0] = ApplyAlphaMask(r, a, dst2[0]);
			dst2[1] = ApplyAlphaMask(g, a, dst2[1]);
			dst2[2] = ApplyAlphaMask(b, a, dst2[2]);
			dst2[3] = 255;
			y_src2++;

			rgb_y = RGB_Y[*y_src2];
			b = clip_col( (rgb_y + b_u) >> SCALEBITS_OUT);
			g = clip_col( (rgb_y - g_uv) >> SCALEBITS_OUT);
			r = clip_col( (rgb_y + r_v) >> SCALEBITS_OUT);
			a = mk ? GetAlphaKey(mk, r, g, b) : const_a;
			dst2[4] = ApplyAlphaMask(r, a, dst2[4]);
			dst2[5] = ApplyAlphaMask(g, a, dst2[5]);
			dst2[6] = ApplyAlphaMask(b, a, dst2[6]);
			dst2[7] = 255;
			y_src2++;

			dst += 8;
			dst2 += 8;
		}

		dst += dst_dif;
		dst2 += dst_dif;

		y_src += y_dif;
		y_src2 += y_dif;
		u_src += uv_stride;
		v_src += uv_stride;
	}
}

void yuva2rgb_32(unsigned char *dst, int dststride, unsigned char *y_src, unsigned char *v_src, unsigned char * u_src, unsigned char *a_src,
				 int y_stride, int uv_stride, int width, int height)
{
	yuv2rgb_init();

	const unsigned int dst_dif = 2 * dststride - 4 * width;
	int y_dif = 2 * y_stride - width;
	
	unsigned int x, y;
	unsigned char *dst2 = dst + dststride;
	unsigned char *y_src2 = y_src + y_stride;
	unsigned char *a_src2 = a_src + y_stride;
	
	for (y = height / 2; y; y--) {
		for (x = 0; x < (unsigned int)width / 2; x++) {
			int u, v;
			int b_u, g_uv, r_v, rgb_y;
			int r, g, b;

			u = u_src[x];
			v = v_src[x];

			b_u = B_U[u];
			g_uv = G_U[u] + G_V[v];
			r_v = R_V[v];

			rgb_y = RGB_Y[*y_src];
			b = (rgb_y + b_u) >> SCALEBITS_OUT;
			g = (rgb_y - g_uv) >> SCALEBITS_OUT;
			r = (rgb_y + r_v) >> SCALEBITS_OUT;
			dst[0] = MAX(0,MIN(255, r));
			dst[1] = MAX(0,MIN(255, g));
			dst[2] = MAX(0,MIN(255, b));
			dst[3] = *a_src;
			y_src++;
			a_src++;

			rgb_y = RGB_Y[*y_src];
			b = (rgb_y + b_u) >> SCALEBITS_OUT;
			g = (rgb_y - g_uv) >> SCALEBITS_OUT;
			r = (rgb_y + r_v) >> SCALEBITS_OUT;
			dst[4] = MAX(0,MIN(255, r));
			dst[5] = MAX(0,MIN(255, g));
			dst[6] = MAX(0,MIN(255, b));
			dst[7] = *a_src;
			y_src++;
			a_src++;

			rgb_y = RGB_Y[*y_src2];
			b = (rgb_y + b_u) >> SCALEBITS_OUT;
			g = (rgb_y - g_uv) >> SCALEBITS_OUT;
			r = (rgb_y + r_v) >> SCALEBITS_OUT;
			dst2[0] = MAX(0,MIN(255, r));
			dst2[1] = MAX(0,MIN(255, g));
			dst2[2] = MAX(0,MIN(255, b));
			dst2[3] = *a_src2;
			y_src2++;
			a_src2++;

			rgb_y = RGB_Y[*y_src2];
			b = (rgb_y + b_u) >> SCALEBITS_OUT;
			g = (rgb_y - g_uv) >> SCALEBITS_OUT;
			r = (rgb_y + r_v) >> SCALEBITS_OUT;
			dst2[4] = MAX(0,MIN(255, r));
			dst2[5] = MAX(0,MIN(255, g));
			dst2[6] = MAX(0,MIN(255, b));
			dst2[7] = *a_src2;
			y_src2++;
			a_src2++;

			dst += 8;
			dst2 += 8;
		}

		dst += dst_dif;
		dst2 += dst_dif;

		y_src += y_dif;
		y_src2 += y_dif;
		a_src += y_dif;
		a_src2 += y_dif;

		u_src += uv_stride;
		v_src += uv_stride;
	}
}
