/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File: CompositeSurface.cpp
//
// Authors: Jean Le Feuvre ENST 2002
//
// Modified: 
//
// Description:
//   Composition surface used by Compositetexture2D (Layer2D doesn't use
// composition surface for speed purposes)
//
// Created: September 2002
//
///////////////////////////////////////////////////////////////////////////

#include "defs.h"
#include "CompositeSurface.h"
#include "drawable.h"
#include "Presenter.h"


CompositeSurface::CompositeSurface(Presenter *pPresenter) : VisualSurface()
{
	m_pTexture = NULL;
	m_Width = m_Height = 0;
	m_bIsClearing = FALSE;
	m_pPresenter = pPresenter;
	m_pPresenter->GetVisualRenderer()->GetRenderingOption(VIDEO_FLAG_PIXELMETRICS, m_bUsePixelMetrics);
}

CompositeSurface::~CompositeSurface()
{
	if (m_pTexture) delete m_pTexture;
}

BOOL CompositeSurface::InitGraphRender(BOOL bForTextDraw)
{
	if (!m_pTexture || !m_pTexture->m_pBitmap) return FALSE;
	m_pGraph = new Gdiplus::Graphics(m_pTexture->m_pBitmap);

	int highSpeed, antiAlias, textAntiAlias;

	m_pPresenter->GetVisualRenderer()->GetRenderingOption(VIDEO_FLAG_HIGHSPEED, highSpeed);
	m_pPresenter->GetVisualRenderer()->GetRenderingOption(VIDEO_FLAG_ANTIALIAS, antiAlias);
	m_pPresenter->GetVisualRenderer()->GetRenderingOption(VIDEO_FLAG_TEXTANTIALIAS, textAntiAlias);

	//set rendering options
	if (m_bIsClearing) {
		//don't waste time when clearing
		m_pGraph->SetCompositingQuality(Gdiplus::CompositingQualityHighSpeed);
		m_pGraph->SetSmoothingMode(Gdiplus::SmoothingModeHighSpeed);
	} else {
		m_pGraph->SetCompositingQuality(highSpeed ? Gdiplus::CompositingQualityHighSpeed : Gdiplus::CompositingQualityHighQuality);

		if (bForTextDraw && textAntiAlias) {
			m_pGraph->SetSmoothingMode(Gdiplus::SmoothingModeHighQuality);
			m_pGraph->SetTextRenderingHint(Gdiplus::TextRenderingHintAntiAlias);
		} else {
			m_pGraph->SetSmoothingMode(antiAlias ? Gdiplus::SmoothingModeHighQuality : Gdiplus::SmoothingModeHighSpeed);
			if (antiAlias) {
				m_pGraph->SetTextRenderingHint(Gdiplus::TextRenderingHintAntiAlias);
			} else if (highSpeed) {
				m_pGraph->SetTextRenderingHint(Gdiplus::TextRenderingHintSingleBitPerPixel);
			} else {
				m_pGraph->SetTextRenderingHint(Gdiplus::TextRenderingHintSingleBitPerPixelGridFit);
			}
		}

		m_pGraph->SetInterpolationMode(highSpeed ? Gdiplus::InterpolationModeLowQuality : Gdiplus::InterpolationModeHighQualityBicubic);
	}

	Gdiplus::Matrix final;
	final.Scale(1, -1, Gdiplus::MatrixOrderAppend);
	final.Translate(m_Width/2, m_Height/2, Gdiplus::MatrixOrderAppend);
	m_pGraph->SetTransform(&final);

	//force the background to be erased when clearing since we use alpha channel
	if (m_bIsClearing) {
		m_pGraph->SetCompositingMode(Gdiplus::CompositingModeSourceCopy);
	}
	return TRUE;
}

void CompositeSurface::ResetAntialiasing()
{
	if (!m_pGraph) return;
	m_pGraph->SetSmoothingMode(Gdiplus::SmoothingModeHighSpeed);
}

void CompositeSurface::ReleaseGraphRender()
{
	delete m_pGraph;
	m_pGraph = NULL;
}

BOOL CompositeSurface::InitObjectMatrix(Matrix2D &mat)
{
	Gdiplus::Matrix final;
	mat.SetIdentity();
	if (!m_bUsePixelMetrics)
		mat.AddScale(m_Width/2, m_Height/2, 0.0);

	return m_bUsePixelMetrics;
}

void CompositeSurface::Clear(BRect const* pBRect)
{
	VisualPath path;
	if (pBRect == NULL) {
		path.AddRectangle(m_Width, m_Height);
	} else {
		path.AddRectangle(pBRect->left, pBRect->top, pBRect->width, pBRect->height);
	}
	ResetStyles();
	SetBrushStyle(1, Color3f(0.0, 0.0, 0.0), 1.0);
	m_bIsClearing = TRUE;
	DrawVisualPath(&path, TRUE);
	m_bIsClearing = FALSE;
}


float CompositeSurface::GetWidth()
{
	if (!m_bUsePixelMetrics) return 2.0;
	return m_Width;
}

float CompositeSurface::GetHeight()
{
	if (m_bUsePixelMetrics) return m_Height;
	return 2.0 * m_Width / m_Height;
}

//not used by composite surface since texture transform are needed to perform this task, and
//texture transform are context-based
BOOL CompositeSurface::DeviceToSurfaceCoords(int input_X, int input_Y, float *output_X, float *output_Y)
{
	*output_X = input_X;
	*output_Y = input_Y;
	return TRUE;
}

void CompositeSurface::DrawBitmap(BRect *, BRect *, VisualTexture *, MaterialKey *)
{
	//not supported in composite texture
	return;
}

BOOL CompositeSurface::Setup(unsigned int pixel_width, unsigned int pixel_height)
{
	m_Width = pixel_width;
	m_Height = pixel_height;

	if (!m_pTexture) m_pTexture = new VisualTexture();;
	if (!m_pTexture->InitOffscreen(m_Width, m_Height)) return FALSE;

	return TRUE;
}

