/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  CompositeTexture2DProxy.cpp
//
// Author: Jean Le Feuvre ENST 
//
// Modified:
//
// Description:
//   The definition of the proxy object used by the presenter
//   to render the CompositeTexture2D node.
//
// Created: September, 2002.
//
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "CompositeTexture2DProxy.h"
#include "ViewportProxy.h"
#include "Presenter.h"
#include "GeometryNode.h"


CompositeTexture2DProxy *Presenter::InitCompositeTexture2D(CompositeTexture2D* pObj)
{
	return new CompositeTexture2DProxy(pObj);
}

CompositeTexture2DProxy::CompositeTexture2DProxy(MediaObject* pMO) : DrawableImage(pMO)
{
	m_pSurface = NULL;
	m_LastDrawTime = 0;
	m_pTexture = NULL;
}
CompositeTexture2DProxy::~CompositeTexture2DProxy()
{
	//set texture to NULL since it is destroyed with the surface
	m_pTexture = NULL;
	if (m_pSurface) delete m_pSurface;
}

void CompositeTexture2DProxy::Update(DrawableContext *pContext)
{
	Effects *effects = new Effects();

	float width, height;
	
	width = GetCompositeTexture()->pixelWidth;
	height = GetCompositeTexture()->pixelHeight;

	if (!m_pSurface) {
		m_pSurface = new CompositeSurface(GetPresenter());
		GetPresenter()->RegisterSurface(m_pSurface);

		if (width<=0) {
			width = 200;
			GetCompositeTexture()->pixelWidth = 200;
		}
		if (height<=0) {
			height = 200;
			GetCompositeTexture()->pixelHeight = 200;
		}
		m_pSurface->Setup(width, height);
	}
	//we should forbid dynamic changes of texture resolution
	if ( (width != m_pSurface->GetWidth()) || (height != m_pSurface->GetHeight()) ) {
		m_pSurface->Setup(width, height);
	}

	//texture is already rendered
	DWORD now = GetObject()->GetClock() + 1;
	if (m_LastDrawTime == now) return;

	//texture has changed, update the brush object
	m_pTexture = m_pSurface->GetTexture();
		
	effects->m_pSurface = m_pSurface;
	effects->m_bUseBoundTracking = GetPresenter()->m_defEffects.m_bUseBoundTracking;

	m_pSurface->ResetDrawList();

	m_pSurface->InitDraw(effects);

	//since there may be any kind of updates inside the texture we must perform prerender no matter what
	RenderChildren(effects);

	//draw has changed the surface, invalidate
	if (m_pSurface->TerminateDraw(effects)) {
		m_pTexture->Invalidate();
		TRACE("CTexture changed\n");
	}

	//then draw the texture
	m_LastDrawTime = now;

	pContext->m_bUpdateTexture = TRUE;

	BRect visibleRect;
	visibleRect.MakeEmpty();

	//set active viewport in image coordinates top-left=(0, 0), not in BIFS
	//Note that position and orientation fields are already applied when rendering
	ViewportProxy *pVP = (ViewportProxy *) m_pSurface->GetViewportStack()->GetTop();
	if (pVP && (Bool) pVP->GetViewport()->isBound) {
		Vec2f size = pVP->GetViewport()->size;
		if (size.x >=0 && size.y>=0) {
			width = size.x;
			height = size.y;
		}
	} 
	visibleRect.width = width;
	visibleRect.height = height;
	visibleRect.left = -1 * width/2;
	visibleRect.top = height/2;

	//translate to (0, 0) coords
	visibleRect.left += m_pSurface->GetWidth() / 2;
	visibleRect.top = m_pSurface->GetHeight() / 2 - visibleRect.top;
	//and to pixel metrics
	VisualPath path;
	path.AddRectangle(visibleRect.left, visibleRect.top, visibleRect.width, visibleRect.height);
	Matrix2D mat;
	m_pSurface->InitObjectMatrix(mat);
	path.GetBoundingRect(&visibleRect, &mat, 0.0);


	m_pTexture->SetActiveRect(&visibleRect);

	BOOL rS, rT;
	GetTextureRepeat(rS, rT);
	m_pTexture->SetTextureRepeat(rS, rT);

	delete effects;
}



void CompositeTexture2DProxy::RenderChildren(Effects *effects)
{
	//perform composite render and drawing 
	effects->m_pSurface = m_pSurface;
	effects->m_ParentGroup = NULL;

	int i;
	
	//check for modifs of the children
	ResortSensors();
	int length= GetObject()->GetChildCount();

	//fill effects sensor list
	for (i = 0; i <m_sensors.GetLength(); i++) {
		Traversable * pTrav=m_sensors[i];
		effects->AddUserEventHandler((UserEventHandlerProxy *)pTrav, &effects->transform);
	}

	//render children
	for (i=0; i<length; i++) {
		Traversable * pTrav=GetChild(i);
		if(pTrav) pTrav->Render(effects);
	}
}

//classify all our sensors in a dedicated array
void CompositeTexture2DProxy::ResortSensors()
{
	DWORD now = 1 + GetObject()->GetUpdateTime();
	if (now==m_LastUpdateTime) return;
	m_LastUpdateTime = now;
	m_sensors.Clear();

	for (int i = 0; i<GetObject()->GetChildCount(); i++) {
		Traversable * pTrav=GetChild(i);
		if(pTrav && pTrav->IsSensor() ) 
			m_sensors += pTrav;
	}
}

Traversable *CompositeTexture2DProxy::GetChild(int n) 
{
	MediaObject *pChild = (*GetObject()->GetChildrenField())[n];
	if (pChild)
		return (Traversable*) pChild->GetRenderingObject(); 
	return NULL;
}


DrawableContext *CompositeTexture2DProxy::IsPointInTexture(DrawableContext *pContext, float x, float y)
{
	if (!m_pSurface) return NULL;

	//get active width and height 
	float width, height;
	width = m_pSurface->GetTexture()->GetActiveWidth();
	height = m_pSurface->GetTexture()->GetActiveHeight();

	//reverse coords to our local coordinate system. Everything must be computed in the original bounds
	BRect bounds;
	((GeometryNode *)pContext->m_pDrawable)->GetOriginalBounds(bounds);

	//reverse our coords
	pContext->m_transform.ReverseApply(x, y);

	//center the coords
	x -= bounds.left + bounds.width/2;
	y += bounds.height/2 - bounds.top;

	//get texture transforms if any
	if (!pContext->m_pAppearance) return NULL;
	TextureTransform *tx_trans = (TextureTransform *) pContext->m_pAppearance->textureTransform.GetNode();
	if (tx_trans) {
		Vec2f transl = (Vec2f)tx_trans->translation;
		Vec2f center = (Vec2f)tx_trans->center;
		Vec2f scale = (Vec2f)tx_trans->scale;
		if (!scale.x) scale.x=0.01;
		if (!scale.y) scale.y=0.01;

		Matrix2D mat;
		//order as specified in VRML. Note: the spec isn't really clear regarding the 
		//scale factor: since it is applied before the rotation there is no angle used 
		//hence we always scale along (S, T)
		//build everything as reverse (cf spec)
		mat.AddRotation(-tx_trans->rotation, center.x, center.y);
		mat.AddScale(1/scale.x, 1/scale.y, 0.0);
		mat.AddTranslation( - transl.x, - transl.y);

		//undo texture transform
		mat.ReverseApply(x, y);
	}

	//undo repeatS and repeatT
	BOOL repeatS = FALSE;
	BOOL repeatT = FALSE;
	GetTextureRepeat(repeatS, repeatT);

	if (repeatS) {
		//switch to x origin in 0
		x += bounds.width / 2;
		//reverse repeatS
		while (x>width) x -= width;
		//switch to x origin in middle of texture
		x -= width/2;
	} else {
		x *= width;
		x /= bounds.width;
	}
	if (repeatT) {
		//switch to y origin in 0
		y += bounds.height / 2;
		//reverse repeatT
		while (y>height) y -= height;
		//switch to y origin in middle of texture
		y -= height/2;
	} else {
		y *= height;
		y /= bounds.height;
	}

	return m_pSurface->FindNode(x, y);
}

BOOL CompositeTexture2DProxy::HasSensors()
{
	if (!m_pSurface) return FALSE;
	return m_pSurface->HasSensors();
}
