//////////////////////////////////////////////////////////////////////////////
//This software module was originally developed by CSELT in the course of   //
//development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.             //
//This software module is an implementation of a part of one or more MPEG-4 //
//Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems        //
//(ISO/IEC 14496-1) standard. ISO/IEC gives users of the MPEG-4 Systems     //
//(ISO/IEC 14496-1) free license to this software module or modifications   //
//thereof for use in hardware or software products claiming conformance to  //
//the MPEG-4 Systems (ISO/IEC 14496-1). Those intending to use this software//
//module in hardware or software products are advised that its use may      //
//infringe existing patents.                                                //
//The original developer of this software module and his/her company, the   //
//subsequent editors and their companies, and ISO/IEC have no liability     //
//for use of this software module or modifications thereof in an            //
//implementation. Copyright is not released for non MPEG-4 Systems          //
//(ISO/IEC 14496-1) conforming products. CSELT S.pA. retains full right     //
//to use the code for his/her own purpose, assign or donate the code to a   //
//third party and to inhibit third parties from using the code for non      //
//MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright      //
//notice must be included in all copies or derivative works.                //
//Copyright  1999.                                                         //
//This work has been partly developed within the framework of the ESPRIT    //
//European Research Project  "MPEG-4 PC".                                   //
//////////////////////////////////////////////////////////////////////////////
//
// File: CopyRGBA.cpp
//
// Author: Matthew Leditschke
//
// Modified: Gianluca Di Cagno  gianluca.dicagno@cselt.it
// 				Jean Le Feuvre - ENST 2002
//
// Description:
//   The implementation of the functions which copy an RGBA image
//   to the offscreen RGB surface in the visual renderer.
//
// Created: April, 1997
//
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "VisualRender-GDIP.h"
#include "ColorTools.h"


unsigned char GetAlphaKey(MaterialKey *mk, unsigned char r, unsigned char g, unsigned char b)
{
	Color3f col = mk->keyColor;
	unsigned char c_r, c_b, c_g, y, u, v;
	unsigned char alpha = (unsigned char) (255 * (float) (1 - mk->transparency));

	if (! (BOOL) (Bool) mk->isRGB) {
		y = 255*col.r;
		u = 255*col.g;
		v = 255*col.b;
		ConvertYUVtripleToRGB(y, u, v, c_r, c_g, c_b);
	} else {
		c_r = 255*col.r;
		c_g = 255*col.g;
		c_b = 255*col.b;
	}
	int res = (abs(c_r-r) + abs(c_g-g) + abs(c_b-b)) / 3;

	if (res <= 255*mk->lowThreshold) return 0;
	if (res >= 255*mk->highThreshold) return alpha;
	if (mk->highThreshold == mk->lowThreshold) return alpha;
	res -= 255*mk->lowThreshold;
	res = res * alpha / (mk->highThreshold - mk->lowThreshold);
	return (unsigned char) res;
}

// ===============================================================================

//
// CopyRGBAToRGB555
//
// Copy the given RGBA image to the given surface, in RGB555 format, using
// the alpha mask to determine which pixels are copied.
//
// The top-left corner of the given image is copied to the point given by
// (top,left), where (0,0) is the top-left corner of the given surface.
//

void CopyRGBAToRGB555(SrcRGBAImage const& srcImage,
					  SurfaceDesc const& surfaceDesc)
{
	CopyParameters cpParams;
	if (SetUpCopyParameters(cpParams, srcImage, surfaceDesc))
		return;

	// For every row of pixels...
	for (unsigned int srcY = cpParams.srcStartY; srcY < cpParams.srcStartY + cpParams.copyHeight; srcY++)
	{
		// Get a pointer to the start of this line in the destination frame
		unsigned short* pDest = (unsigned short*)((unsigned char*)((unsigned char*)surfaceDesc.lpSurface + (cpParams.destStartY + srcY - cpParams.srcStartY) * surfaceDesc.lPitch
								+ 2 * cpParams.destStartX));

		// Copy across the line of data, performing alpha masking if required
		for (unsigned int srcX = cpParams.srcStartX; srcX < cpParams.srcStartX + cpParams.copyWidth; srcX++)
		{
			// The offset to get to the desired pixel in the image
			unsigned int pixelOffset = srcY * srcImage.pitch + srcX * srcImage.ptrInc;

			// Get the red, green and blue values for the new pixel
			int red   = *(srcImage.pR + pixelOffset);
			int green = *(srcImage.pG + pixelOffset);
			int blue  = *(srcImage.pB + pixelOffset);

			// Apply the alpha mask - if required
			int alpha;
			if (srcImage.mk) {
				alpha = GetAlphaKey(srcImage.mk, red, green, blue);
			} else {
				if (srcImage.pA != NULL) 
					alpha = *(srcImage.pA + pixelOffset);
				else
					alpha  = srcImage.const_alpha;
			}

			if (alpha < 255)
			{
				// Get the RGB values underneath this one
				int oldRed   = (*pDest & 0x7c00) >> 7;
				int oldGreen = (*pDest & 0x03e0) >> 2;
				int oldBlue  = (*pDest & 0x001f) << 3;

				// Apply the alpha mask factor
				ApplyAlphaMaskToRGB(red, green, blue, alpha, oldRed, oldGreen, oldBlue);
			}

			// Combine the values, 5 bits red, 5 bits green and 5 bits blue
			*pDest++ = MAKE_RGB555(red, green, blue);
		}
	}
} /* CopyRGBAToRGB555 */


// ===============================================================================

//
// CopyRGBAToRGB565
//
// Copy the given RGBA image to the given surface, in RGB565 format, using
// the alpha mask to determine which pixels are copied.
//
// The top-left corner of the given image is copied to the point given by
// (top,left), where (0,0) is the top-left corner of the given surface.
//

void CopyRGBAToRGB565(SrcRGBAImage const& srcImage,
					  SurfaceDesc const& surfaceDesc)
{
	CopyParameters cpParams;
	if (SetUpCopyParameters(cpParams, srcImage, surfaceDesc))
		return;

	// For every row of pixels...
	for (unsigned int srcY = cpParams.srcStartY; srcY < cpParams.srcStartY + cpParams.copyHeight; srcY++)
	{
		// Get a pointer to the start of this line in the destination frame
		unsigned short* pDest = (unsigned short*)((unsigned char*)((unsigned char*)surfaceDesc.lpSurface + (cpParams.destStartY + srcY - cpParams.srcStartY) * surfaceDesc.lPitch
								+ 2 * cpParams.destStartX));

		// Copy across the line of data, performing alpha masking if required
		for (unsigned int srcX = cpParams.srcStartX; srcX < cpParams.srcStartX + cpParams.copyWidth; srcX++)
		{
			// The offset to get to the desired pixel in the image
			unsigned int pixelOffset = srcY * srcImage.pitch + srcX * srcImage.ptrInc;

			// Get the red, green and blue values for the new pixel
			int red   = *(srcImage.pR + pixelOffset);
			int green = *(srcImage.pG + pixelOffset);
			int blue  = *(srcImage.pB + pixelOffset);

			// Apply the alpha mask - if required
			int alpha;
			if (srcImage.mk) {
				alpha = GetAlphaKey(srcImage.mk, red, green, blue);
			} else {
				if (srcImage.pA != NULL) 
					alpha = *(srcImage.pA + pixelOffset);
				else
					alpha  = srcImage.const_alpha;
			}

			if (alpha < 255)
			{
				// Get the RGB values underneath this one
				int oldRed   = (*pDest & 0xf800) >> 8;
				int oldGreen = (*pDest & 0x07e0) >> 3;
				int oldBlue  = (*pDest & 0x001f) << 3;

				// Apply the alpha mask factor
				ApplyAlphaMaskToRGB(red, green, blue, alpha, oldRed, oldGreen, oldBlue);
			}

			// Combine the values, 5 bits red, 6 bits green and 5 bits blue
			*pDest++ = MAKE_RGB565(red, green, blue);
		}
	}
} /* CopyRGBAToRGB565 */


// ===============================================================================

//
// CopyRGBAToRGB24bpp
//
// Copy the given RGBA image to the given surface, in RGB24bpp format, using
// the alpha mask to determine which pixels are copied.
//
// The top-left corner of the given image is copied to the point given by
// (top,left), where (0,0) is the top-left corner of the given surface.
//

void CopyRGBAToRGB24bpp(SrcRGBAImage const& srcImage,
						SurfaceDesc const& surfaceDesc)
{
	CopyParameters cpParams;
	if (SetUpCopyParameters(cpParams, srcImage, surfaceDesc))
		return;

	// For every row of pixels...
	for (unsigned int srcY = cpParams.srcStartY; srcY < cpParams.srcStartY + cpParams.copyHeight; srcY++)
	{
		// Get a pointer to the start of this line in the destination frame
		unsigned char* pDest = (unsigned char*)surfaceDesc.lpSurface
								+ (cpParams.destStartY + srcY - cpParams.srcStartY) * surfaceDesc.lPitch
								+ 3 * cpParams.destStartX;

		// Copy across the line of data, performing alpha masking if required
		for (unsigned int srcX = cpParams.srcStartX; srcX < cpParams.srcStartX + cpParams.copyWidth; srcX++)
		{
			// The offset to get to the desired pixel in the image
			unsigned int pixelOffset = srcY * srcImage.pitch + srcX * srcImage.ptrInc;

			// Get the red, green and blue values for the new pixel
			int red   = *(srcImage.pR + pixelOffset);
			int green = *(srcImage.pG + pixelOffset);
			int blue  = *(srcImage.pB + pixelOffset);

			// Apply the alpha mask - if required
			int alpha;
			if (srcImage.mk) {
				alpha = GetAlphaKey(srcImage.mk, red, green, blue);
			} else {
				if (srcImage.pA != NULL) 
					alpha = *(srcImage.pA + pixelOffset);
				else
					alpha  = srcImage.const_alpha;
			}

			if (alpha < 255)
			{
				// Get the RGB values underneath this one
				int oldRed   = pDest[2];
				int oldGreen = pDest[1];
				int oldBlue  = pDest[0];

				// Apply the alpha mask factor
				ApplyAlphaMaskToRGB(red, green, blue, alpha, oldRed, oldGreen, oldBlue);
			}

			// Store the 8 bit values for blue, green and red
			*pDest++ = (unsigned char) blue;
			*pDest++ = (unsigned char) green;
			*pDest++ = (unsigned char) red;
		}
	}
} /* CopyRGBAToRGB24bpp */


// ===============================================================================

//
// CopyRGBAToRGB32bpp
//
// Copy the given RGBA image to the given surface, in RGB32bpp format, using
// the alpha mask to determine which pixels are copied.
//
// The top-left corner of the given image is copied to the point given by
// (top,left), where (0,0) is the top-left corner of the given surface.
//

void CopyRGBAToRGB32bpp(SrcRGBAImage const& srcImage,
						SurfaceDesc const& surfaceDesc)
{
	CopyParameters cpParams;
	if (SetUpCopyParameters(cpParams, srcImage, surfaceDesc))
		return;

	// For every row of pixels...
	for (unsigned int srcY = cpParams.srcStartY; srcY < cpParams.srcStartY + cpParams.copyHeight; srcY++)
	{
		// Get a pointer to the start of this line in the destination frame
		unsigned long* pDest = (unsigned long*)((unsigned char*)((unsigned char*)surfaceDesc.lpSurface + (cpParams.destStartY + srcY - cpParams.srcStartY) * surfaceDesc.lPitch
								+ 4 * cpParams.destStartX));

		// Copy across the line of data, performing alpha masking if required
		for (unsigned int srcX = cpParams.srcStartX; srcX < cpParams.srcStartX + cpParams.copyWidth; srcX++)
		{
			// The offset to get to the desired pixel in the image
			unsigned int pixelOffset = srcY * srcImage.pitch + srcX * srcImage.ptrInc;

			// Get the red, green and blue values for the new pixel
			int red   = *(srcImage.pR + pixelOffset);
			int green = *(srcImage.pG + pixelOffset);
			int blue  = *(srcImage.pB + pixelOffset);

			// Apply the alpha mask - if required
			int alpha;
			if (srcImage.mk) {
				alpha = GetAlphaKey(srcImage.mk, red, green, blue);
			} else {
				if (srcImage.pA != NULL) 
					alpha = *(srcImage.pA + pixelOffset);
				else
					alpha  = srcImage.const_alpha;
			}

			if (alpha < 255)
			{
				// Get the RGB values underneath this one
				int oldRed   = (*pDest & 0xff0000) >> 16;
				int oldGreen = (*pDest & 0xff00) >> 8;
				int oldBlue  = (*pDest & 0xff);

				// Apply the alpha mask factor
				ApplyAlphaMaskToRGB(red, green, blue, alpha, oldRed, oldGreen, oldBlue);
			}

			// Combine the values, 8 bits red, 8 bits green and 8 bits blue
			*pDest++ = MAKE_RGB32(red, green, blue);
		}
	}
} /* CopyRGBAToRGB32bpp */

