/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  Curve2DProxy.cpp
//
// Author: Jean Le Feuvre
//
// Description:
//   The implementation of the proxy object used by the presenter
//   to render the Curve node.
//
// Created: August, 2002.
//
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-INIT.h"
#include "Curve2DProxy.h"
#include "EmptyProxies.h"

#include "Presenter.h"
#include <math.h>
#include <assert.h>


Curve2DProxy* Presenter::InitCurve2D(Curve2D* pObj)
{
	return (new Curve2DProxy(pObj));
}

Curve2DProxy* Presenter::InitXCurve2D(XCurve2D* pObj)
{
	return (new Curve2DProxy(pObj));
}

Curve2DProxy::Curve2DProxy(MediaObject* pMO) : GeometryNode(pMO)
{
}

Curve2DProxy::~Curve2DProxy()
{
}

Coordinate2D *Curve2DProxy::GetCoordinate2D() const
{
	if (GetObject()->GetNodeCode() == CODE_Curve2D) {
		return (Coordinate2D*) ((Curve2D*)GetObject())->point; 
	}
	return (Coordinate2D*) ((XCurve2D*)GetObject())->point; 
}

int Curve2DProxy::GetType(int i) const
{
	if (GetObject()->GetNodeCode() == CODE_Curve2D) {
		return ((Curve2D*)GetObject())->type[i]; 
	}
	return ((XCurve2D*)GetObject())->type[i]; 
}

int Curve2DProxy::GetNumTypes() const
{
	if (GetObject()->GetNodeCode() == CODE_Curve2D) {
		return ((Curve2D*)GetObject())->type.GetLength(); 
	}
	return ((XCurve2D*)GetObject())->type.GetLength(); 
}

float Curve2DProxy::GetFineness() const
{
	if (GetObject()->GetNodeCode() == CODE_Curve2D) {
		return ((Curve2D*)GetObject())->fineness; 
	}
	return ((XCurve2D*)GetObject())->fineness; 
}


void Curve2DProxy::Render(Effects *effects)
{
	if (!GetCoordinate2D()) return;

	BOOL filled = TRUE;
	if (effects->m_pAppearanceProxy) {
		Material2D *pMat = (Material2D *) (MediaObject *) ((Appearance *) effects->m_pAppearanceProxy->GetObject())->material;
		if (!pMat || (! (Bool) pMat->filled) ) filled = FALSE;
	}

	//check for geometry change
	BOOL buildIt = effects->m_bShouldRebuild ? TRUE : FALSE;
	DWORD now = 1 + GetObject()->GetUpdateTime();
	if (now > m_LastUpdateTime) {
		m_LastUpdateTime = now;
		buildIt = TRUE;
	}

	now = 1 + GetCoordinate2D()->GetUpdateTime();
	if (now > m_LastUpdateTime) {
		m_LastUpdateTime = now;
		buildIt = TRUE;
	}
	if (buildIt) {
		BuildGraph(filled);
		m_bHasChanged = TRUE;
	}

	DrawableContext *pContext = InitRenderContext(effects);
	if (!pContext) return;

	m_Path.GetBoundingRect(&pContext->m_boundingRect, &pContext->m_transform, pContext->gProps.lineWidth, pContext->gProps.lineScaled);	
	pContext->m_clippedBoundingRect = pContext->m_boundingRect.GetClipper();
	
	FinalizeRender(effects, pContext);
}

void Curve2DProxy::GetOriginalBounds(BRect &bounds)
{
	m_Path.GetBoundingRect(&bounds, NULL, 0);
}

void Curve2DProxy::AddToClipper(VisualClipper *pClipper, DrawableContext *pContext, BOOL bInside, BOOL bUseXOR)
{
	pClipper->AddPath(&m_Path, &pContext->m_transform, bInside, bUseXOR);
}

void Curve2DProxy::Draw(DrawableContext * pContext)
{
	if (!SetupDraw(pContext)) return;
	DrawTexture(pContext);
	pContext->m_pVisualSurface->DrawVisualPath(&m_Path);
}


//max number of points defining a segment of curve (aka max N-Bezier suported: N=20)
#define MAX_POINTS	20

static void FillArray(PointF *pts, PointF orig, PointF ctl_orig, PointF ctl_end, PointF end, BOOL IsQuadratic)
{
	pts[0].x = orig.x;
	pts[0].y = orig.y;
	if (IsQuadratic) {
		pts[1].x = ctl_end.x;
		pts[1].y = ctl_end.y;
		pts[2].x = end.x;
		pts[2].y = end.y;
	} else {
		pts[1].x = ctl_orig.x;
		pts[1].y = ctl_orig.y;
		pts[2].x = ctl_end.x;
		pts[2].y = ctl_end.y;
		pts[3].x = end.x;
		pts[3].y = end.y;
	}
}

void Curve2DProxy::BuildGraph(BOOL filled)
{
	float fineness;
	int offX, offY, bezRes;
	PointF pts[MAX_POINTS];
	PointF orig, ctl_orig, ctl_end, end, begin;

	//setup the bezier resolution
	VisualRenderer *pRender = GetPresenter()->GetVisualRenderer();
	pRender->GetRenderingOption(VIDEO_FLAG_BEZIERRESOLUTION, bezRes);
	m_Path.Reset(bezRes);

	//browse all types
	int NumTypes = GetNumTypes();
	int NumPoints = GetCoordinate2D()->point.GetLength();
	fineness = GetFineness();

	//setup origin
	orig.x = GetCoordinate2D()->point[0].x;
	orig.y = GetCoordinate2D()->point[0].y;
	//avoid compiler warning
	ctl_orig = orig;

	begin = orig;
	//if no types we shall ignore the init
	int curIndex = NumTypes ? 1 : 0;

	bool HasBeenClosed = FALSE;

	for (int i=0; i<NumTypes; i++) {
		offX = offY = 0;
		int type = GetType(i);
		switch (type) {
		//relative moveTo, 1 point
		case 8:
			offX = orig.x;
			offX = orig.y;
		//moveTo, 1 point
		case 0:
			//if first point is moveTo, ignore it since we already inited the curve2D
			if (i) {
				//close shape if filled, except if already done (closePath in curve2D v2)
				if (!HasBeenClosed) {
					if (filled) m_Path.AddLine(orig.x, orig.y, begin.x, begin.y);
					if (i) m_Path.StartNewFigure();
				}
				HasBeenClosed = FALSE;

				orig.x = offX + GetCoordinate2D()->point[curIndex].x;
				orig.y = offY + GetCoordinate2D()->point[curIndex].y;
				//store start of curve
				begin = orig;
				curIndex += 1;
			}
			break;
		//relative lineTo, 1 point
		case 9:
			offX = orig.x;
			offX = orig.y;
		//lineTo, 1 point
		case 1:
			end.x = offX + GetCoordinate2D()->point[curIndex].x;
			end.y = offY + GetCoordinate2D()->point[curIndex].y;
			m_Path.AddLine(orig.x, orig.y, end.x, end.y);
			orig = end;
			curIndex += 1;
			break;
		
		//relative curveTo, 3 points
		case 10:
			offX = orig.x;
			offX = orig.y;
		//curveTo, 3 points
		case 2:
			ctl_orig.x = offX + GetCoordinate2D()->point[curIndex].x;
			ctl_orig.y = offY + GetCoordinate2D()->point[curIndex].y;
			ctl_end.x = offX + GetCoordinate2D()->point[curIndex+1].x;
			ctl_end.y = offY + GetCoordinate2D()->point[curIndex+1].y;
			end.x = offX + GetCoordinate2D()->point[curIndex+2].x;
			end.y = offY + GetCoordinate2D()->point[curIndex+2].y;

			FillArray(pts, orig, ctl_orig, ctl_end, end, 0);
			m_Path.AddBezier(pts, 4, fineness);

			curIndex += 3;
			ctl_orig = ctl_end;
			orig = end;
			break;
		//relative nextCurveTo, 2 points
		case 11:
			offX = orig.x;
			offX = orig.y;
		//nextCurveTo, 2 points
		case 3:
			ctl_orig.x = 2*orig.x - ctl_orig.x;
			ctl_orig.y = 2*orig.y - ctl_orig.y;
			ctl_end.x = offX + GetCoordinate2D()->point[curIndex].x;
			ctl_end.y = offY + GetCoordinate2D()->point[curIndex].y;
			end.x = offX + GetCoordinate2D()->point[curIndex+1].x;
			end.y = offY + GetCoordinate2D()->point[curIndex+1].y;

			FillArray(pts, orig, ctl_orig, ctl_end, end, 0);
			m_Path.AddBezier(pts, 4, fineness);
			curIndex += 2;
			ctl_orig = ctl_end;
			orig = end;
			break;

		//relative ArcTo
		case 12:
		case 13:
			offX = orig.x;
			offX = orig.y;
		//ArcTo
		case 4:
		case 5:
			ctl_orig.x = offX + GetCoordinate2D()->point[curIndex].x;
			ctl_orig.y = offY + GetCoordinate2D()->point[curIndex].y;
			ctl_end.x = offX + GetCoordinate2D()->point[curIndex+1].x;
			ctl_end.y = offY + GetCoordinate2D()->point[curIndex+1].y;
			end.x = offX + GetCoordinate2D()->point[curIndex+2].x;
			end.y = offY + GetCoordinate2D()->point[curIndex+2].y;
			//clockwise direction
			if ((type==5) || (type==13))
				m_Path.AddArc(orig, end, ctl_orig, ctl_end, TRUE);
			else
				m_Path.AddArc(orig, end, ctl_orig, ctl_end);
			curIndex += 3;
			orig = end;
			break;

		//closePath
		case 6:
			m_Path.CloseFigure();
			m_Path.StartNewFigure();
			HasBeenClosed = TRUE;
			continue;
		//quadratic arc, 2 points
		case 14:
			offX = orig.x;
			offX = orig.y;
		//quadratic arc, 2 points
		case 7:
			ctl_end.x = offX + GetCoordinate2D()->point[curIndex].x;
			ctl_end.y = offY + GetCoordinate2D()->point[curIndex].y;
			end.x = offX + GetCoordinate2D()->point[curIndex+1].x;
			end.y = offY + GetCoordinate2D()->point[curIndex+1].y;
			FillArray(pts, orig, ctl_orig, ctl_end, end, 1);
			m_Path.AddBezier(pts, 3, fineness);
			curIndex += 2;
			ctl_orig = ctl_end;
			orig = end;
			break;

		}
		//reset the closePath state
		HasBeenClosed = FALSE;
	}


	//if we still have points draw a final bezier here
	int remain = NumPoints-curIndex;
	if ((remain > 0) && (remain < MAX_POINTS-1)) {
		for (i=0;i<remain;i++) {
			pts[i].x = GetCoordinate2D()->point[i+curIndex].x;
			pts[i].y = GetCoordinate2D()->point[i+curIndex].y;
		}
		m_Path.AddBezier(pts, remain, fineness);
	}
}
