/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File: ColorTools.h
//
// Author: Jean Le Feuvre
//
// Description:
//		direct draw tools
//
// Created: February 2003
//
///////////////////////////////////////////////////////////////////////////



#include "defs.h"

#include <windows.h>
#include <vfw.h>
#include <assert.h>
#include <ddrawex.h>
#include <mmsystem.h>

#include "DirectDraw.h"
#include "dx-errors.h"
#include "ColorTools.h"
#include "trace.h"

enum
{
	/*supported planar formats*/
	FCC_YV12 = mmioFOURCC('Y','V','1','2'),
	FCC_I420 = mmioFOURCC('I','4','2','0'),
	FCC_IYUV = mmioFOURCC('I','Y','U','V'),
	/*supported packed formats*/
	FCC_UYVY = mmioFOURCC('U','Y','V','Y'),
	FCC_Y422 = mmioFOURCC('Y','4','2','2'),
	FCC_UYNV = mmioFOURCC('U','Y','N','V'),
	FCC_YUY2 = mmioFOURCC('Y','U','Y','2'),
	FCC_YUNV = mmioFOURCC('Y','U','N','V'),
	FCC_V422 = mmioFOURCC('V','4','2','2'),
	FCC_YVYU = mmioFOURCC('Y','V','Y','U')
};

static BOOL IsYUVFormatSupported(DWORD fmt)
{
	switch (fmt) {
	case FCC_YV12:
	case FCC_I420:
	case FCC_IYUV:
	case FCC_UYVY:
	case FCC_Y422:
	case FCC_UYNV:
	case FCC_YUY2:
	case FCC_YUNV:
	case FCC_V422:
	case FCC_YVYU:
		return TRUE;
	}
	return FALSE;
}

static BOOL IsPlanarYUV(DWORD fmt)
{
	switch (fmt) {
	case FCC_YV12:
	case FCC_I420:
	case FCC_IYUV:
		return TRUE;
	}
	return FALSE;
}

static DWORD GetYUVEq(DWORD fmt) 
{
	switch (fmt) {
	case FCC_I420:
	case FCC_IYUV:
		return FCC_YV12;
	case FCC_UYVY:
	case FCC_Y422:
	case FCC_UYNV:
		return FCC_UYVY;
	case FCC_YUY2:
	case FCC_YUNV:
	case FCC_V422:
		return FCC_YUY2;
	case FCC_YVYU:
		return FCC_YVYU;
	default:
		return fmt;
	}
}

// Constructor - set all pointers to NULL.
DirectDraw::DirectDraw()
{
	lpDD = NULL;
	m_pPrimary = NULL;
	m_pOffscreen = NULL;
	m_pPool = NULL;
	m_PoolWidth = m_PoolHeight = 0;
	m_bPoolIsYUV = FALSE;
	lpClipper = NULL;
	m_bFullScreen = FALSE;
	m_Width = m_Height = 0;
	m_hWnd = NULL;
	m_fsWidth = m_fsHeight = 0;
	m_bCheckFormats = TRUE;
	m_prefer4CC = 0;
	m_fourcc = 0;
}

DirectDraw::~DirectDraw()
{
	Shutdown();
}

void DirectDraw::Shutdown()
{
	if (m_bFullScreen) {
		lpDD->RestoreDisplayMode();
		lpDD->SetCooperativeLevel(m_hWnd, DDSCL_NORMAL);
	}

	if (m_pOffscreen != NULL) m_pOffscreen->Release();
	m_pOffscreen = NULL;
	
	if (m_pPool != NULL) m_pPool->Release();
	m_pPool = NULL;
	m_PoolWidth = m_PoolHeight = 0;
	m_bPoolIsYUV = FALSE;
	
	if (lpClipper != NULL) lpClipper->Release();
	lpClipper = NULL;
	
	if (m_pPrimary != NULL) m_pPrimary->Release();
	m_pPrimary = NULL;

	if (lpDD != NULL) lpDD->Release();
	lpDD = NULL;
}


const char *DirectDraw::SetFullScreen(BOOL bFS)
{
	if (bFS == m_bFullScreen) return NULL;
	/*shutdown all surfaces*/
	Shutdown();	
	m_bFullScreen = bFS;
	return Init(m_hWnd, m_Width, m_Height);
}

const char* DirectDraw::Init (HWND hwnd, unsigned int compositeWidth, unsigned int compositeHeight)
{
    HRESULT ddrval;
	static char errorMessage[200];

#ifdef USE_DX_3
    ddrval = DirectDrawCreate (NULL, &lpDD, NULL);
#else
	ddrval = DirectDrawCreateEx(NULL, (VOID**)&lpDD, IID_IDirectDraw7, NULL);
#endif
    if (ddrval != DD_OK) goto exit;

	m_hWnd = hwnd;

	if (m_bCheckFormats) {
		DWORD numCodes;
		DWORD *codes;
		ddrval = lpDD->GetFourCCCodes(&numCodes, NULL);
		codes = (DWORD *)malloc(numCodes*sizeof(DWORD));
		ddrval = lpDD->GetFourCCCodes(&numCodes, codes);
		
		m_numFormats = 0;
		for (unsigned int i=0; i<numCodes; i++) {
			if (IsYUVFormatSupported(codes[i])) {
				m_Formats[m_numFormats] = codes[i];
				m_numFormats++;
			}
		}
		free(codes);
		m_bCheckFormats = FALSE;
	}

	// Set cooperative level
	if (!m_bFullScreen) {
		ddrval = lpDD->SetCooperativeLevel (m_hWnd, DDSCL_NORMAL);
	    if (ddrval != DD_OK) goto exit;
	} else {
		ddrval = lpDD->SetCooperativeLevel(m_hWnd, DDSCL_ALLOWMODEX | DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE);
	    if (ddrval != DD_OK) goto exit;

		// Enumerate surfaces and get the closest match
		GetDisplayMode();


		/*in FS try 16bpp first (faster, and better support for overlay by most graphics cards)
		then 32 (24bpp is sometimes extremely slow on some cards) */
#ifdef USE_DX_3
		ddrval = lpDD->SetDisplayMode(m_fsWidth, m_fsHeight, 16);
		if (ddrval != DD_OK) {
			ddrval = lpDD->SetDisplayMode(m_fsWidth, m_fsHeight, 32);
			if (ddrval != DD_OK) {
				ddrval = lpDD->SetDisplayMode(m_fsWidth, m_fsHeight, 24);
				if (ddrval != DD_OK) goto exit;
			}
		}
#else
		ddrval = lpDD->SetDisplayMode(m_fsWidth, m_fsHeight, 16, 0, 0);
		if (ddrval != DD_OK) {
			ddrval = lpDD->SetDisplayMode(m_fsWidth, m_fsHeight, 32, 0, 0);
			if (ddrval != DD_OK) {
				ddrval = lpDD->SetDisplayMode(m_fsWidth, m_fsHeight, 24, 0, 0);
				if (ddrval != DD_OK) goto exit;
			}
		}
#endif
	}

	/*Create a primary surface  */
#ifdef USE_DX_3
	DDSURFACEDESC surfaceDesc;
#else
	DDSURFACEDESC2 surfaceDesc;
#endif

	memset (&surfaceDesc, 0, sizeof(surfaceDesc));
	surfaceDesc.dwSize = sizeof(surfaceDesc);
	surfaceDesc.dwFlags = DDSD_CAPS;
	surfaceDesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

	ddrval = lpDD->CreateSurface (&surfaceDesc, &m_pPrimary, NULL);
    if (ddrval != DD_OK) goto exit;

	/*
	 * Determine what pixel formats this display can handle.
	 *  - in particular, what is the RGB format.
	 */
	DDPIXELFORMAT pixelFmt;
	memset (&pixelFmt, 0, sizeof(pixelFmt));
	pixelFmt.dwSize = sizeof(pixelFmt);
	ddrval = m_pPrimary->GetPixelFormat (&pixelFmt);
    if (ddrval != DD_OK) goto exit;

	errorMessage[0] = 0;
	if (pixelFmt.dwFlags & DDPF_RGB) {
		// Handle 16 bits per pixel displays
		if (pixelFmt.dwRGBBitCount == 16) {
			if ((pixelFmt.dwRBitMask == 0xf800) &&
				(pixelFmt.dwGBitMask == 0x07e0) &&
				(pixelFmt.dwBBitMask == 0x001f))
				rgbPixelFormat = RGB565;

			else if ((pixelFmt.dwRBitMask == 0x7c00) &&
				(pixelFmt.dwGBitMask == 0x03e0) &&
				(pixelFmt.dwBBitMask == 0x001f))
				rgbPixelFormat = RGB555;

			else
				sprintf (errorMessage,
						 "Unsupported 16 bpp RGB masks: %04x %04x %04x",
						 pixelFmt.dwRBitMask,
						 pixelFmt.dwGBitMask,
						 pixelFmt.dwBBitMask);
		}

		// Handle 24 bits per pixel displays
		else if (pixelFmt.dwRGBBitCount == 24)
		{
			if ((pixelFmt.dwRBitMask == 0xff0000) &&
				(pixelFmt.dwGBitMask == 0x00ff00) &&
				(pixelFmt.dwBBitMask == 0x0000ff))
				rgbPixelFormat = RGB24;
			else
				sprintf (errorMessage,
						 "Unsupported 24 bpp RGB masks: %06x %06x %06x",
						 pixelFmt.dwRBitMask,
						 pixelFmt.dwGBitMask,
						 pixelFmt.dwBBitMask);
		}
		// Handle 32 bits per pixel displays
		else if (pixelFmt.dwRGBBitCount == 32)
		{
			if ((pixelFmt.dwRBitMask == 0xff0000) &&
				(pixelFmt.dwGBitMask == 0x00ff00) &&
				(pixelFmt.dwBBitMask == 0x0000ff))
				rgbPixelFormat = RGB32;
			else
				sprintf (errorMessage,
						 "Unsupported 32 bpp RGB masks: %08x %08x %08x",
						 pixelFmt.dwRBitMask,
						 pixelFmt.dwGBitMask,
						 pixelFmt.dwBBitMask);
		}
		else
			strcpy (errorMessage, "Can only handle 16, 24 and 32 bit per pixel displays");
	}
	else
		strcpy (errorMessage, "This doesn't appear to be an RGB display!!\n");


	if (errorMessage[0] != 0) return (errorMessage);

    /* Create a clipper  */
    ddrval = lpDD->CreateClipper (0, &lpClipper, NULL);
    if (ddrval != DD_OK) goto exit;
    ddrval = lpClipper->SetHWnd (0, hwnd);
    if (ddrval != DD_OK) goto exit;
    ddrval = m_pPrimary->SetClipper (lpClipper);
    if (ddrval != DD_OK) goto exit;
    lpClipper->Release();
	
	// create the offscreen RGB surface
	ddrval = Resize(compositeWidth, compositeHeight);

	if (ddrval != DD_OK) {
		strcpy (errorMessage, "Error in Resize():\n");
		strcat (errorMessage, HRESULTtoText(ddrval));
		return (errorMessage);
	}

exit:
    if (ddrval != DD_OK)
    {
		strcpy (errorMessage, "Error in DirectDrawCreate():\n");
		strcat (errorMessage, HRESULTtoText(ddrval));
        return (errorMessage);
    }

	return NULL;
}

HRESULT DirectDraw::Resize (unsigned int newWidth, unsigned int newHeight)
{
#ifdef USE_DX_3
	DDSURFACEDESC surfaceDesc;
#else
	DDSURFACEDESC2 surfaceDesc;
#endif
	HRESULT returnValue;

	// Delete the current and currently existing surfaces
	if (m_pOffscreen != NULL) m_pOffscreen->Release();
	m_pOffscreen = NULL;

    memset (&surfaceDesc, 0, sizeof(surfaceDesc));
    surfaceDesc.dwSize = sizeof(surfaceDesc);
    surfaceDesc.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
    surfaceDesc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
    surfaceDesc.dwWidth = newWidth;
    surfaceDesc.dwHeight = newHeight;

    returnValue = lpDD->CreateSurface (&surfaceDesc, &m_pOffscreen, NULL);

    if (returnValue != DD_OK) return FALSE;
    
	if (m_bFullScreen) {
		newWidth = m_fsWidth;
		newHeight = m_fsHeight;
	} else {
		m_Width = newWidth;
		m_Height = newHeight;
	}

	if(m_pOffscreen) returnValue = EraseSurface(NULL);
	return returnValue;
}



HRESULT DirectDraw::RenderSurface(RECT* pDestRect)
{
	HRESULT returnValue;
	
	returnValue = m_pPrimary->Blt(pDestRect, m_pOffscreen, NULL, DDBLT_WAIT, NULL);

	if (returnValue == DDERR_SURFACELOST) {
		returnValue = m_pPrimary->Restore();
	}
	return returnValue;
}


HRESULT DirectDraw::LockSurface(DDSURFACEDESC &surfaceDesc)
{
	HRESULT returnValue;

#ifdef USE_DX_3
	DDSURFACEDESC ddsd;
#else
	DDSURFACEDESC2 ddsd;
#endif

	if (!m_pOffscreen) return DDERR_NOTINITIALIZED;
	returnValue = DD_OK;
	ddsd.dwSize = sizeof(ddsd);
	
	while (1) {
		returnValue = m_pOffscreen->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
		if (returnValue == DD_OK) break;
	}
	memset(&surfaceDesc, 0, sizeof(surfaceDesc));
	surfaceDesc.dwSize = sizeof(surfaceDesc);
	surfaceDesc.dwWidth = ddsd.dwWidth;
	surfaceDesc.dwHeight = ddsd.dwHeight;
	surfaceDesc.lPitch = ddsd.lPitch;
	surfaceDesc.lpSurface = ddsd.lpSurface;
	return returnValue;
}

HRESULT DirectDraw::UnlockSurface()
{
	if (!m_pOffscreen) return DDERR_NOTINITIALIZED;
	return m_pOffscreen->Unlock(NULL);
}


HRESULT DirectDraw::EraseSurface(RECT const* pRect, BOOL erasePrimary)
{
	HRESULT res;
	DDSURFACEDESC surfaceDesc;

	res = LockSurface(surfaceDesc);
	if(res != DD_OK) return res;
	
	if (pRect == NULL) {
		memset(surfaceDesc.lpSurface, 0, surfaceDesc.lPitch * surfaceDesc.dwHeight);
	} else {
		int bytesPerPixel = (int)((surfaceDesc.ddpfPixelFormat.dwRGBBitCount - 1) / 8) + 1;
		for (int y = pRect->top; y < pRect->bottom; y++) {
			memset((char*)surfaceDesc.lpSurface
					+ y * surfaceDesc.lPitch
					+ pRect->left * bytesPerPixel,
				   0, (pRect->right - pRect->left) * bytesPerPixel);
		}
	}
	res = UnlockSurface();
	if (erasePrimary) RenderSurface(NULL);
	return res;
}

HRESULT DirectDraw::GetHDC(HDC &hDC)
{
	if (!m_pOffscreen) return DDERR_NOTINITIALIZED;
	HRESULT returnValue = m_pOffscreen->IsLost();
	if (!returnValue) {
		returnValue = m_pOffscreen->GetDC(&hDC);
	}
	return returnValue;
}

// Release the handle to the device context for the offscreen RGB surface.
HRESULT DirectDraw::ReleaseHDC(HDC& hDC)
{
	if (!m_pOffscreen) return DDERR_NOTINITIALIZED;
	return m_pOffscreen->ReleaseDC(hDC);
}


//enum is done in increasing resolution. The first resolution that is larger than our scene
//is kept
#ifdef USE_DX_3
HRESULT WINAPI EnumScreenModes( LPDDSURFACEDESC pDesc, LPVOID lpContext)
#else
HRESULT WINAPI EnumScreenModes( LPDDSURFACEDESC2 pDesc, LPVOID lpContext)
#endif
{
	DirectDraw *pObj = (DirectDraw *) lpContext;
	
	TRACE("%d %d vs %d %d\n", pObj->m_Width, pObj->m_Height, pDesc->dwWidth, pDesc->dwHeight);
	
	//check W and H
	if (pObj->m_Width <= pDesc->dwWidth  && pObj->m_Height <= pDesc->dwHeight
		//check FSW and FSH
		&& pObj->m_fsWidth > pDesc->dwWidth && pObj->m_fsHeight > pDesc->dwHeight) {

		pObj->m_fsWidth = pDesc->dwWidth;
		pObj->m_fsHeight = pDesc->dwHeight;
		return DDENUMRET_CANCEL;
	}
	return DDENUMRET_OK;
}

BOOL DirectDraw::GetDisplayMode()
{
	HRESULT hr;
	if (! lpDD) return FALSE;

	m_fsWidth = m_fsHeight = 50000;

#ifdef USE_DX_3
	hr = lpDD->EnumDisplayModes(NULL, NULL, this,  (LPDDENUMMODESCALLBACK) EnumScreenModes);
#else
	hr = lpDD->EnumDisplayModes(NULL, NULL, this,  (LPDDENUMMODESCALLBACK2) EnumScreenModes);
#endif
	if (FAILED(hr)) return FALSE;

	return TRUE;
}


BOOL DirectDraw::HasYUVHardware()
{
	if (m_bPoolIsYUV) return TRUE;
	return FALSE;
}

HRESULT DirectDraw::ScalePool(RECT src, RECT dstScaled)
{
	/*this happens due to float rounding. since we always draw the offscreen at 0, 0 checking is easy...*/
	if (src.right > (int) m_PoolWidth) src.right = m_PoolWidth;
	if (src.bottom > (int) m_PoolHeight) src.bottom = m_PoolHeight;
	return m_pOffscreen->Blt(&dstScaled, m_pPool, &src, 0, NULL);
}

HRESULT DirectDraw::CreatePoolSurface(unsigned int width, unsigned int height, BOOL isYUV)
{
#ifdef USE_DX_3
	DDSURFACEDESC ddsd;
	LPDIRECTDRAWSURFACE	pPool;    
#else
	DDSURFACEDESC2 ddsd;
	LPDIRECTDRAWSURFACE7 pPool;
#endif
	HRESULT res;

	/*no YUV HW found*/
	if (!m_prefer4CC) isYUV = FALSE;

	/*same surface as what we have, keep it*/
	if (m_pPool && (isYUV==m_bPoolIsYUV) && (width<=m_PoolWidth) && (height<=m_PoolHeight)) return DD_OK; 

	if (isYUV && m_numFormats) {
		pPool = NULL;
		memset (&ddsd, 0, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		ddsd.ddpfPixelFormat.dwSize = sizeof(ddsd.ddpfPixelFormat);

		ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_PIXELFORMAT;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OVERLAY | DDSCAPS_VIDEOMEMORY;
		ddsd.ddpfPixelFormat.dwFlags = DDPF_FOURCC;
		ddsd.dwWidth = width;
		ddsd.dwHeight = height;

		if (m_prefer4CC) {
			ddsd.ddpfPixelFormat.dwFourCC = m_prefer4CC;
			res = lpDD->CreateSurface(&ddsd, &pPool, NULL);
			if (res != DD_OK) pPool = NULL;
		}
		if (!pPool) {
			for (int i=0; i<m_numFormats; i++) {
				ddsd.ddpfPixelFormat.dwFourCC = m_Formats[i];
				res = lpDD->CreateSurface(&ddsd, &pPool, NULL);
				if (res == DD_OK) break;
			}
		}
		if (res != DD_OK) pPool = NULL;

		if (pPool) {
			if (m_pPool) m_pPool->Release();
			m_pPool = pPool;
			m_PoolWidth = width;
			m_PoolHeight = height;
			m_fourcc = GetYUVEq(ddsd.ddpfPixelFormat.dwFourCC);
			m_bPoolIsYUV = TRUE;
			return res;
		}
		/*failed to create YUV, fall back to RGB*/
	}

	/*don't create an RGB surface if the current one fits our needs*/
	if (m_pPool && !m_bPoolIsYUV && (width<=m_PoolWidth) && (height<=m_PoolHeight)) return DD_OK; 

	if (m_pPool) m_pPool->Release();
	m_pPool = NULL;
	m_bPoolIsYUV = FALSE;

	/*create RGB pool*/

    memset(&ddsd, 0, sizeof(ddsd));
    ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
    ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
    ddsd.dwWidth = width;
    ddsd.dwHeight = height;

	res = lpDD->CreateSurface(&ddsd, &m_pPool, NULL);

	if (res != DD_OK) {
		m_pPool = NULL;
	} else {
		m_PoolWidth = width;
		m_PoolHeight = height;
	}
	return res;
}


HRESULT DirectDraw::LockPool(DDSURFACEDESC &surfaceDesc)
{
	HRESULT res;
#ifdef USE_DX_3
	DDSURFACEDESC ddsd;
#else
	DDSURFACEDESC2 ddsd;
#endif
	
	if (!m_pPool) return DDERR_NOTINITIALIZED;
	res = DD_OK;
	ddsd.dwSize = sizeof(surfaceDesc);
	
	while (1) {
		res = m_pPool->Lock(NULL, &ddsd, DDLOCK_WAIT, NULL);
		if (res == DD_OK) break;
	}
	memset(&surfaceDesc, 0, sizeof(surfaceDesc));
	surfaceDesc.dwSize = sizeof(surfaceDesc);
	surfaceDesc.dwWidth = ddsd.dwWidth;
	surfaceDesc.dwHeight = ddsd.dwHeight;
	surfaceDesc.lPitch = ddsd.lPitch;
	surfaceDesc.lpSurface = ddsd.lpSurface;
	return res;
}

HRESULT DirectDraw::UnlockPool()
{
	if (!m_pPool) return DDERR_NOTINITIALIZED;
	return m_pPool->Unlock(NULL);
}


HRESULT DirectDraw::WriteYUV(unsigned char *pY, unsigned char *pU, unsigned char *pV, unsigned int width, unsigned int height, unsigned int pitch)
{
	DDSURFACEDESC ddsd;
	HRESULT res;
	res = LockPool(ddsd);
	if (res!=DD_OK) return res;

	if (IsPlanarYUV(m_fourcc)) {

		if ( (ddsd.lPitch==(int)pitch) && (width==pitch)) {
			memcpy(ddsd.lpSurface, pY, width*height);
			memcpy((unsigned char *)ddsd.lpSurface + width*height, (m_fourcc != FCC_YV12) ? pU : pV, width*height/4);
			memcpy((unsigned char *)ddsd.lpSurface + 5*width*height/4, (m_fourcc != FCC_YV12) ? pV : pU, width*height/4);
		} else {

			unsigned int i;
			unsigned char *dst, *src, *dst2, *src2, *dst3, *src3;

			src = pY;
			dst = (unsigned char *) ddsd.lpSurface;
			
			src2 = (m_fourcc != FCC_YV12) ? pU : pV;
			dst2 = (unsigned char *) ddsd.lpSurface + ddsd.lPitch*ddsd.dwHeight;
			src3 = (m_fourcc != FCC_YV12) ? pV : pU;
			dst3 = (unsigned char *) ddsd.lpSurface + 5*ddsd.lPitch*ddsd.dwHeight/4;
			for (i=0; i<height; i++) {
				memcpy(dst, src, width);
				src += pitch;
				dst += ddsd.lPitch;
				if (i<height/2) {
					memcpy(dst2, src2, width/2);
					src2 += pitch/2;
					dst2 += ddsd.lPitch/2;
					memcpy(dst3, src3, width/2);
					src3 += pitch/2;
					dst3 += ddsd.lPitch/2;
				}
			}
		}
	} else if (m_fourcc==FCC_UYVY) {
		unsigned int i, j;
		unsigned char *dst, *y, *u, *v;
		for (i=0; i<height; i++) {
			y = pY + i*pitch;
			u = pU + (i/2) * pitch/2;
			v = pV + (i/2) * pitch/2;
			dst = (unsigned char *) ddsd.lpSurface + i*ddsd.lPitch;

			for (j=0; j<width/2;j++) {
				*dst = *u;
				dst++;
				u++;
				*dst = *y;
				dst++;
				y++;
				*dst = *v;
				dst++;
				v++;
				*dst = *y;
				dst++;
				y++;
			}
		}
	} else if (m_fourcc==FCC_YUY2) {
		unsigned int i, j;
		unsigned char *dst, *y, *u, *v;
		for (i=0; i<height; i++) {
			y = pY + i*pitch;
			u = pU + (i/2) * pitch/2;
			v = pV + (i/2) * pitch/2;
			dst = (unsigned char *) ddsd.lpSurface + i*ddsd.lPitch;

			for (j=0; j<width/2;j++) {
				*dst = *y;
				dst++;
				y++;
				*dst = *u;
				dst++;
				u++;
				*dst = *y;
				dst++;
				y++;
				*dst = *v;
				dst++;
				v++;
			}
		}
	} else if (m_fourcc==FCC_YVYU) {
		unsigned int i, j;
		unsigned char *dst, *y, *u, *v;
		for (i=0; i<height; i++) {
			y = pY + i*pitch;
			u = pU + (i/2) * pitch/2;
			v = pV + (i/2) * pitch/2;
			dst = (unsigned char *) ddsd.lpSurface + i*ddsd.lPitch;

			for (j=0; j<width/2;j++) {
				*dst = *y;
				dst++;
				y++;
				*dst = *v;
				dst++;
				v++;
				*dst = *y;
				dst++;
				y++;
				*dst = *u;
				dst++;
				u++;
			}
		}
	}

	return UnlockPool();
}

void DirectDraw::SetPreferedYUVFormat(int format)
{
	if (format==m_prefer4CC) return;
	if (m_pPool) m_pPool->Release();
	m_pPool = NULL;
	m_prefer4CC = format;
	m_fourcc = m_prefer4CC;
}

#define NUM_TEST	10
/*tolerance for planar vs packed format*/
#define PLANAR_TOLERANCE	2

void DirectDraw::CheckYUVHardwareSupport()
{
	if (!m_pPrimary || !m_pOffscreen) return;

	HRESULT res;
	unsigned char *yuvbuf, *y, *u, *v;
	unsigned int w, h;
	int k;
	/*don't try with too high resolutions since there may not be enough memory on the card (cf ReadMe)*/
	w = 320;
	h = 240;
	yuvbuf = (unsigned char *) malloc(sizeof(char)*3*w*h/2);
	memset(yuvbuf, 0, sizeof(char)*3*w*h/2);

	y = yuvbuf;
	u = yuvbuf + w*h;
	v = yuvbuf + 5*w*h/4;

	DWORD now, minTimePacked = 0xFFFFFFFF, minTimePlanar = 0xFFFFFFFF;
	DWORD bestPacked = 0, bestPlanar = 0;

	BOOL checkPacked = TRUE;

	timeBeginPeriod(1);

	for (int i=0; i<m_numFormats; i++) {
		/*check planar first*/
		if (!checkPacked && !IsPlanarYUV(m_Formats[i])) goto go_on;
		/*then check packed */
		if (checkPacked && IsPlanarYUV(m_Formats[i])) goto go_on;

		m_prefer4CC = m_Formats[i];
		res = CreatePoolSurface(w, h, TRUE);
		if (res != DD_OK) goto rem_fmt;
		/*YUV format is not supported at all*/
		if (!m_bPoolIsYUV) break;
		/*current format is not supported*/
		if ((DWORD) m_fourcc != m_Formats[i]) goto rem_fmt;

		now = timeGetTime();

		/*copy YUV*/
		res = WriteYUV(y, u, v, w, h, w);
		if (res != DD_OK) goto rem_fmt;

		for (k=0; k<NUM_TEST; k++) {
			/*blit*/
			res = m_pOffscreen->Blt(NULL, m_pPool, NULL, 0, NULL);
			if (res != DD_OK) goto rem_fmt;
		}
		now = timeGetTime() - now;

		if (!checkPacked) {
			if (now<minTimePlanar) {
				minTimePlanar = now;
				bestPlanar = m_prefer4CC;
			}
		} else {
			if (now<minTimePacked) {
				minTimePacked = now;
				bestPacked = m_prefer4CC;
			}
		}

go_on:
		/* reset pool settings*/
		m_bPoolIsYUV = 0;
		if (checkPacked && (i+1==m_numFormats)) {
			i = -1;
			checkPacked = FALSE;
		}
		continue;

rem_fmt:
		for (int j=i; j<m_numFormats-1; j++) {
			m_Formats[j] = m_Formats[j+1];
		}
		i--;
		m_numFormats--;
	}
	timeEndPeriod(1);

	free(yuvbuf);

	if (bestPlanar && (minTimePlanar < minTimePacked + PLANAR_TOLERANCE)) {
		m_prefer4CC = bestPlanar;
	} else {
		m_prefer4CC = bestPacked;
	}
	m_fourcc = m_prefer4CC;

	if (m_pPool) m_pPool->Release();
	m_pPool = NULL;
	m_bPoolIsYUV = 0;
}