/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  DrawableImage.cpp
//
// Author: Jean Le Feuvre - ENST 2002
//
// Description:
//   Proxy fetching data from visual decoders
//   all natural texture proxies derive from this one
//
// Created: August 2002.
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "Stream.h"
#include "Clock.h"
#include "DrawableImage.h"
#include "trace.h"
#include "resource.h"

#include "Presenter.h"
#include "Caps.h"

#include "ShapeProxy.h"


enum {VIDEO_DELAY = 0};

DrawableImage::DrawableImage(MediaObject* pMO) : DrawableNode(pMO)
{
	m_frameWasFetched=FALSE;
	m_LastFrameTime = (unsigned long) -1;
	m_bShouldRedraw = FALSE;
	m_lastFetchTime = 0;
	m_bLastWasEndOfStream = FALSE;

	m_bUpdateNext = FALSE;
	m_pTexture = NULL;	
}

// DrawableImage::~DrawableImage
DrawableImage::~DrawableImage()
{
	if (m_pTexture) delete m_pTexture;	
}

// This method is called when the end of a stream is detected in either
// the Render() method or the Draw() method.
void DrawableImage::StopStream()
{
	if(m_frameWasFetched && GetStream())
		GetStream()->Release(FALSE);

	GetStreamConsumer()->CloseStream();
	m_frameWasFetched=FALSE;
}


float DrawableImage::GetTransparency(DrawableContext *pContext) const
{
	if ((pContext->m_pAppearance == NULL) || (pContext->m_pAppearance->material == NULL)) return 0.0;
	return ((Material2D*)(MediaObject*) pContext->m_pAppearance->material)->transparency;
}


// fetches the next texture frame if any, check for texture transforms
void DrawableImage::Update(DrawableContext *pContext)
{
	MediaStreamImp *pStream;
	LPVOID dataPtr = NULL;
	int dataSize = 0;

	pContext->m_bUpdateTexture = FALSE;
	pContext->m_hasStreamFinished = FALSE;
	pStream = GetStream();
	
	// Only proceed once a stream has been attached
	if (pStream == NULL) return;

	// Decide if the stream contains a new frame to display
	
	DWORD rdwTime;
	BOOL bEOU;
	MediaFormat frameFormat = GetStreamConsumer()->GetStreamFormat().m_code;
	//decoder isn't up-to-date
	if (frameFormat == MediaFormat :: UNKNOWN) {
		CapabilitiesImp caps;
		DecoderParams par;
		GetStreamConsumer()->GetCapabilities (&caps);
		BOOL bFormatOK = GetStreamConsumer()->GetDecoder()->SetFormat(caps, par);
		if (!bFormatOK || par.m_format==MediaFormat :: UNKNOWN) return;
		GetStreamConsumer()->SetStreamFormat (par.m_format);
		//ok
		frameFormat = GetStreamConsumer()->GetStreamFormat().m_code;
	}
	
	//to avoid compiler warning
	rdwTime = m_LastFrameTime - 1;
	BOOL bIsUSE = FALSE;

	DWORD now = GetPresenter()->GetTime();
	int status = 1;

	if (m_lastFetchTime != now + 1) {
		//in case the same frame was fetched at last call, release the current one
		if(m_frameWasFetched) {
			pStream->Release(FALSE);
			m_frameWasFetched=FALSE;
		}	
		//fetch new frame 
		dataPtr = GetStreamConsumer()->Fetch(dataSize, rdwTime, bEOU, status, TRUE, FALSE, 0, 1);


		m_bUpdateNext = FALSE;
		if (dataPtr) m_lastFetchTime = now + 1;
	}
	//we're in the same render cycle as last visit -> this is a USE
	else {
		bIsUSE = TRUE;
	}

	if (bIsUSE && m_bLastWasEndOfStream) {
		pContext->m_hasStreamFinished = TRUE;
		pContext->m_bUpdateTexture = m_bUpdateNext;
		return;
	}

	if ( (status == -1) || (dataPtr == END_OF_STREAM) ) {
		pContext->m_hasStreamFinished = TRUE;
		m_bLastWasEndOfStream = TRUE;
	} else {
		m_bLastWasEndOfStream = FALSE;
	}

	// If there is no data, no updating is required
	if (((dataPtr == NULL) || (dataPtr == END_OF_STREAM) ) ) {
		if (bIsUSE) pContext->m_bUpdateTexture = m_bUpdateNext;
		return;
	}

	// If data block returned has 0 size, no updating is required
	if (dataSize == 0) {
		assert(!dataPtr);
		pStream->Release(FALSE);
		return;
	}

	//if the timing hasn't changed we are fetching the same frame, don't update 
	//if the node doesn't need update (inactive movie texture) don't update
	if (!ShouldBuildBrush() || (m_LastFrameTime==1+rdwTime)) {
		/*release since we're not redrawing the texture*/
		pStream->Release(FALSE);
		if (m_bShouldRedraw) pContext->m_bUpdateTexture = TRUE;
		return;
	}
//	TRACE("fetched frame %d\n", rdwTime);

	m_frameWasFetched = TRUE;
	m_LastFrameTime = 1+rdwTime;
	pContext->m_bUpdateTexture = m_bUpdateNext = TRUE;

	// get a pointer to the current instantiated decoder
	// so that we could know how to retrieve image size,
	// pitch,offset.
	
	Decoder * pIstDec =GetStreamConsumer()->GetDecoder();
	
	float left,top,height,width;
	DecoderImage const* pImgDetails;
	unsigned int pitch, offset;

	offset = pIstDec->GetDataOffset(NULL);


	//JLF - there is absolutely NO documentation on OFFSETX/Y attributes, i don't know why the previous versions 
	//assumed top = -OFFSETY . The new version uses top=OFFSETY, and assumes OFFSETX and OFFSETY are the top-left 
	//coordinates of the first valid pixel in the decoded buffer. These are then passed to the texture object
	if(offset==0) {
		width = pIstDec->GetAttributes(FRAME_WIDTH,(LPBYTE) dataPtr);
		height = pIstDec->GetAttributes(FRAME_HEIGHT,(LPBYTE) dataPtr);
		left = (int)pIstDec->GetAttributes(OFFSETX,(LPBYTE) dataPtr);
		top = (int) pIstDec->GetAttributes(OFFSETY,(LPBYTE) dataPtr);
		pitch = pIstDec->GetAttributes(PITCH,(LPBYTE) dataPtr);
	} else {
		pImgDetails = (DecoderImage*) dataPtr; 
		left = pImgDetails->offsetX; 
		top = -pImgDetails->offsetY;
		width = pImgDetails->width; 
		height = pImgDetails->height;
		pitch = pImgDetails->pitch;
	}
	
	//init texture object since this is a new frame at least for this texture object
	if (!m_pTexture) m_pTexture = new VisualTexture;

	m_pTexture->Init(width, height, (unsigned char *) dataPtr + offset, pitch, frameFormat);
	
	//convert texture to supported output format - in Gdiplus only RGB texturing is possible
	//also we don't support alpha + stretching on bitmaps ...
	if ( (pContext->m_pDrawable->GetObject()->GetNodeCode() != CODE_Bitmap) || pContext->gProps.transparency)
		m_pTexture->SetTextureActive();
}

void DrawableImage::Release()
{
	//once the draw is called the texture MUST be released, do not wait for next render since it may lock
	//the stream (2 objects using the same OD)
	if(m_frameWasFetched) {
		GetStream()->Release(FALSE);
		m_frameWasFetched=FALSE;
	}
	m_bShouldRedraw = FALSE;
}

void DrawableImage::GetTextureRepeat(BOOL &repeatS, BOOL &repeatT)
{
	ImageTexture *pImage;
	MovieTexture *pMovie;
	CompositeTexture2D *pComp;
	DrawableImage *surface_b;

	//default values are TRUE in VRML
	repeatS = repeatT = TRUE;

	MediaObject *pNode = GetObject();
	switch (pNode->GetNodeCode()) {
	case CODE_ImageTexture:
		pImage = (ImageTexture *) pNode;
		repeatS = (BOOL) (Bool) pImage->repeatS;
		repeatT = (BOOL) (Bool) pImage->repeatT;
		break;
	case CODE_MovieTexture:
		pMovie = (MovieTexture *) pNode;
		repeatS = (BOOL) (Bool) pMovie->repeatS;
		repeatT = (BOOL) (Bool) pMovie->repeatT;
		break;
	case CODE_CompositeTexture2D:
		pComp = (CompositeTexture2D *) pNode;
		//not yet reflected in v1 templates - TO FIX
#if 0
		switch (pComp->repeatSandT) {
		case 3:
			repeatS = repeatT = TRUE;
			break;
		case 2:
			repeatS = FALSE;
			repeatT = TRUE;
			break;
		case 1:
			repeatS = TRUE;
			repeatT = FALSE;
			break;
		default:
			repeatS = repeatT = FALSE;
			break;
		}
#endif		
		break;

	//this is not specified currentl, just use the settings from surfaceB since this surface
	//defines the geometric aspect of MatteTexture
	case CODE_MatteTexture:
		pNode = (MediaObject *) ((MatteTexture *) pNode)->surfaceB;
		if (!pNode) return;
		surface_b = (DrawableImage *) (RenderingObject *) pNode->GetRenderingObject();
		if (!surface_b) return;
		surface_b->GetTextureRepeat(repeatS, repeatT);
		break;
	default:
		return;
	}
}

