/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  GeometryNode.cpp
//
// Author: Jean Le Feuvre - ENST 2002
//
// Description:
//   base class for all geometry nodes.
//
// Created: August 2002.
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "GeometryNode.h"
#include "Presenter.h"
#include "DrawableImage.h"
#include "EmptyProxies.h"
#include "DrawableImage.h"
#include "VisualTexture.h"
#include "VisualSurface.h"
#include "GenericGradient.h"


#include "ColorMatrix.h"

BOOL GeometryNode::HasChanged(DrawableContext *pContext)
{
	/*timing update in the core is based on presenter time, not scene time
	(cf MediaObject :: SetUpdateTime) */
	DWORD  presTime = GetPresenter()->GetTime();

	if (pContext->m_bUpdateTexture) return TRUE;

	//if the object is signaled as changed redraw it
	if (m_bHasChanged) return TRUE;

	//if appearance has changed
	if (pContext->m_pAppearance == NULL) return FALSE;
	if (pContext->m_pAppearance->GetUpdateTime() >= presTime) return TRUE;

	//if texture ask the texture object if it should redraw
	if (pContext->m_pTextureProxy != NULL) {
		if ( ((DrawableNode *) pContext->m_pTextureProxy)->HasChanged(pContext) ) return TRUE;
	}

	TextureTransform *tx_trans = GetTextureTransform(pContext);
	if (tx_trans && tx_trans->GetUpdateTime() >= presTime) return TRUE;

	//check material key
	MaterialKey *pKey = GetMaterialKey(pContext);
	if (pKey) {
		if (pKey->GetUpdateTime() >= presTime) return TRUE;
		return FALSE;
	}
	// Check if the material node has been updated
	Material2D* pMaterial = (Material2D*) (MediaObject*) pContext->m_pAppearance->material; 
	if (pMaterial == NULL) return FALSE;
	if (pMaterial->GetUpdateTime() >= presTime) return TRUE;

	//Check if the LineProperties node has been updated
	Material2DProxy *pMat = (Material2DProxy *) pMaterial->GetRenderingObject();
	
	if (pMat->GetLineProperties() && (pMat->GetLineProperties()->GetUpdateTime() >= presTime)) 
		return TRUE;
	if (pMat->GetXLineProperties() && (pMat->GetXLineProperties()->GetUpdateTime() >= presTime)) 
		return TRUE;

	return FALSE;
}

MaterialKey *GeometryNode::GetMaterialKey(DrawableContext *pContext) const
{
	if (!pContext->m_pAppearance) return NULL;
	// Check if the material node has been updated
	if (pContext->m_pAppearance->material==NULL) return NULL;
	if (pContext->m_pAppearance->material->GetNodeCode() != CODE_MaterialKey) return NULL;
	return (MaterialKey *) (MediaObject*) pContext->m_pAppearance->material;
}

void GeometryNode::SetupDrawableContext(DrawableContext *pContext)
{
	Material2D *pMaterial;
	if (pContext->m_pAppearance == NULL) return;
	if (pContext->m_pAppearance->material == NULL) return;
	if (pContext->m_pAppearance->material->GetNodeCode() != CODE_Material2D) return;
	pMaterial = (Material2D *) (MediaObject *) pContext->m_pAppearance->material;

	pContext->gProps.emissiveColor = pMaterial->emissiveColor;
	pContext->gProps.transparency = pMaterial->transparency;
	pContext->gProps.filled = (Bool) pMaterial->filled;
	//this is for pointSet2D and indexedLineSet2D
	pContext->gProps.lineTransparency = pContext->gProps.transparency;
	pContext->gProps.lineColor = pMaterial->emissiveColor;

	if (pMaterial->lineProps == NULL_OBJECT) {
		//this is a bug in the spoec: by default line width is 1.0, but in meterMetrics this means half of the screen :)
		Matrix2D mat;
		float lW = pContext->m_pVisualSurface->InitObjectMatrix(mat) ? 1 : 0.01;

		pContext->gProps.lineWidth = lW;
		//if filled and no line props, draw no line
		if (pContext->gProps.filled) {
			switch (pContext->m_pDrawable->GetObject()->GetNodeCode()) {
			case CODE_IndexedLineSet2D:
			case CODE_PointSet2D:
				break;
			default:
				pContext->gProps.lineWidth = 0;
				break;
			}
		}
		return;
	}
	
	pContext->gProps.hasLineProperties = TRUE;
	LineProperties *pLP = ((Material2DProxy*)pMaterial->GetRenderingObject())->GetLineProperties();
	XLineProperties *pXLP = ((Material2DProxy*)pMaterial->GetRenderingObject())->GetXLineProperties();

	pContext->gProps.lineStyle = pLP ? pLP->lineStyle : pXLP->lineStyle;
	pContext->gProps.lineColor = pLP ? pLP->lineColor : pXLP->lineColor;
	pContext->gProps.lineWidth = pLP ? pLP->width : pXLP->width;
	
	if (pXLP) {
		pContext->gProps.lineTransparency = pXLP->transparency;
		pContext->gProps.lineScaled = (Bool) pXLP->isScalable;
		pContext->gProps.lineInside = ! (Bool) pXLP->isCenterAligned;
		pContext->gProps.lineScaled = (Bool) pXLP->isScalable;
		pContext->gProps.lineCap = pXLP->lineCap;
		pContext->gProps.lineJoin = pXLP->lineJoin;
		pContext->gProps.miterLimit = pXLP->miterLimit;
	}
}

Appearance *GeometryNode::GetAppearanceWithMaterial(DrawableContext *pContext) const
{
	if (pContext->m_pAppearance == NULL) return NULL;
	if (pContext->m_pAppearance->material == NULL) return NULL;
	if (pContext->m_pAppearance->material->GetNodeCode() != CODE_Material2D) return NULL;
	return pContext->m_pAppearance;
}


TextureTransform *GeometryNode::GetTextureTransform(DrawableContext *pContext) const
{
	if (!pContext->m_pAppearance) return NULL;
	return (TextureTransform *) pContext->m_pAppearance->textureTransform.GetNode();
}

// Return the Appearance.texture proxy object
VisualTexture *GeometryNode::GetTexture(DrawableContext * pContext) const
{
	if (!pContext->m_pTextureProxy) return NULL;
	return pContext->m_pTextureProxy->m_pTexture;
}

//check if transparent, if so rendering should be skipped
BOOL GeometryNode::CheckTransparentSkip(DrawableContext * pContext, BOOL skipFill)
{
	//this is in case a sensor is being inserted on a transparent geometry: we shall build the
	//bounds no matter what
	if (pContext->m_Handlers.GetLength()) return FALSE;
	
	if (!pContext->m_pTextureProxy && !pContext->m_pGradient) {
		if (pContext->gProps.transparency > 1 - ALMOST_ZERO) {
			//to refine
			if (pContext->gProps.lineTransparency > 1 - ALMOST_ZERO) return TRUE;
		}
		if (pContext->gProps.lineWidth == 0) {
			if (skipFill) return TRUE;
			if (pContext->gProps.filled == FALSE) return TRUE;
		}
	}
	//should draw
	return FALSE;
}



// The method will store a copy of all of the effects parameters
// in a DrawableContext taken from the Visual Surface
DrawableContext * GeometryNode::InitRenderContext(Effects *effects)
{
	assert(effects->m_pSurface);
	/*switched-off geometry nodes are not rendered*/
	if (effects->m_isSwitchedOff) return NULL;

	//Get a empty context from the current surface
	DrawableContext *pContext = effects->m_pSurface->GetDrawableContext();
	
	//Save effects
	pContext->m_transform.Set(effects->transform);
	pContext->m_pDrawable=this;

	// save the pointer of the appearance and shape objects
	if (effects->m_pAppearanceProxy) {
		pContext->m_pAppearance = (Appearance *) effects->m_pAppearanceProxy->GetObject();
	}
	//if clipper, just add the graph and don't draw
	else if (effects->m_ClipperMode) {
		assert(effects->m_pClipper);
		BOOL bInside, bXOR;
		bXOR = FALSE;
		if (effects->m_ClipperMode >= 10) {
			bXOR = TRUE;
			effects->m_ClipperMode -= 10;
		}
		bInside = (BOOL) (effects->m_ClipperMode - 1);
		AddToClipper(effects->m_pClipper, pContext, bInside, bXOR);
		effects->m_pSurface->RemoveLastContext();
		return NULL;
	}

	MediaObject *pText = NULL;
	if (pContext->m_pAppearance) pText = pContext->m_pAppearance->texture;

	if (effects->m_pColorTransform) {
		pContext->m_pColorMatrix = new ColorMatrix();
		pContext->m_pColorMatrix->Set(* effects->m_pColorTransform);
	}

	//WARNING: we do NOT handle gradient as a texture but as a gradient
	//(eg no possible use of MatteTexture with gradient nodes)
	pContext->m_pTextureProxy = NULL;
	pContext->m_pGradient = NULL;

	if (pText) {
		switch (pText->GetNodeCode()) {
		case CODE_RadialGradient:
		case CODE_LinearGradient:
			pContext->m_pGradient = (GenericGradient *) pText->GetRenderingObject();
			break;
		case CODE_PixelTexture:
		case CODE_ImageTexture:
		case CODE_MovieTexture:
		case CODE_MatteTexture:
		case CODE_CompositeTexture2D:
			pContext->m_pTextureProxy = (DrawableImage *) pText->GetRenderingObject();
			break;
		//unknown texture field
		default:
			break;
		}
	}

	// save the pointers of the sensor which affects  the drawable
	for (int i=0; i<effects->m_Handlers.GetLength(); i++) 
		pContext->AddUserEventHandler(effects->m_Handlers[i]);

	SetupDrawableContext(pContext);

	//Update texture info
	if (pContext->m_pTextureProxy) {
		pContext->m_pTextureProxy->Update(pContext);
		//texture not ready, we skip the node drawing (we could draw without texture also)
		if (pContext->m_pTextureProxy->m_pTexture == NULL) {
			effects->m_pSurface->RemoveLastContext();
			return NULL;
		}
	}
	//update gradient info
	else if (pContext->m_pGradient) {
		pContext->m_pGradient->Update(pContext);
	}

	//IndexedLineSet2D and PointSet2D ignores fill flag
	BOOL skipFill = FALSE;
	if (GetObject()->GetNodeCode()==CODE_IndexedLineSet2D) skipFill = TRUE;
	else if (GetObject()->GetNodeCode()==CODE_PointSet2D) skipFill = FALSE;

	if (CheckTransparentSkip(pContext, skipFill)) {
		effects->m_pSurface->RemoveLastContext();
		return NULL;
	}

	//setup clipper if needed
	if (effects->m_pClipper) {
		pContext->m_pClipper = effects->m_pClipper->Clone();
	}
	//and force redraw if needed
	if (effects->m_bInvalidateAll) 
		m_bHasChanged = effects->m_bInvalidateAll;
	return pContext;
}

//returns TRUE if the point is over the shape. Function only called if the point
//is already over the bounds of the shape
BOOL GeometryNode::IsPointOver(DrawableContext *pContext, float x, float y)
{

	//VRML 4.6.7.3: The appearance properties of the geometry do not affect activation of the sensor.
	//In particular, transparent materials or textures shall be treated as opaque with 
	//respect to activation of pointing-device sensors
	return (BOOL) m_Path.IsPointInPath(x, y, &pContext->m_transform);
}


BOOL GeometryNode::SetupDraw(DrawableContext *pContext)
{
	assert(pContext->m_pVisualSurface);

	if (pContext->m_clippedBoundingRect.IsEmpty()) {
		if (pContext->m_pTextureProxy) pContext->m_pTextureProxy->Release();
		return FALSE;
	}

	pContext->m_pVisualSurface->ResetStyles();
	if (pContext->m_pColorMatrix) 
		pContext->m_pVisualSurface->SetColorMatrix(pContext->m_pColorMatrix);

	if (pContext->m_pClipper)
		pContext->m_pVisualSurface->SetShapeClipper(pContext->m_pClipper);


	pContext->m_pVisualSurface->SetLineStyle(pContext->gProps.lineColor,
												pContext->gProps.lineTransparency, 
												pContext->gProps.lineWidth, 
												pContext->gProps.lineCap, 
												pContext->gProps.lineJoin, 
												pContext->gProps.lineStyle, 
												pContext->gProps.miterLimit,
												pContext->gProps.lineScaled,
												pContext->gProps.lineInside);

	pContext->m_pVisualSurface->SetBrushStyle(pContext->gProps.filled, pContext->gProps.emissiveColor, pContext->gProps.transparency);
	pContext->m_pVisualSurface->SetTransform(&pContext->m_transform);
	pContext->m_pVisualSurface->SetClip(& pContext->m_clippedBoundingRect);
	return TRUE;
}

BOOL GeometryNode::DrawTexture(DrawableContext *pContext, VisualPath *a_path, BOOL noAntiAlias)
{
	assert(pContext->m_pVisualSurface);
	VisualTexture *pTexture = GetTexture(pContext);
	Matrix2D mat;

	if (!pTexture) {
		//check whether this is a gradient or not
		return DrawGradient(pContext, a_path);
	}
	//set the transparency, this will create the associated TextureBrush if possible
	pTexture->SetTransparency(pContext->gProps.transparency);

	if (!pTexture->m_pTexture) {
		//texture is not ready!! invalidates next render
		m_LastUpdateTime = 0;
		return FALSE;
	}
	//setup texture transform
	TextureTransform *tx_trans = GetTextureTransform(pContext);
	if (tx_trans) {
		Vec2f transl = (Vec2f)tx_trans->translation;
		Vec2f center = (Vec2f)tx_trans->center;
		Vec2f scale = (Vec2f)tx_trans->scale;
		if (!scale.x) scale.x=0.01;
		if (!scale.y) scale.y=0.01;

		//order as specified in VRML. Note: the spec isn't really clear regarding the 
		//scale factor: since it is applied before the rotation there is no angle used 
		//hence we always scale along (S, T)
		//build everything as reverse (cf spec)
		mat.AddScale(1/scale.x, 1/scale.y, 0.0);
		mat.AddRotation(- tx_trans->rotation, center.x, center.y);
		mat.AddTranslation(- transl.x * pTexture->GetActiveWidth(), - transl.y * pTexture->GetActiveHeight());
		pTexture->SetTextureTransform(&mat);
	}

	pContext->m_pVisualSurface->SetBrushStyle(pContext->gProps.filled, pContext->gProps.emissiveColor, pContext->gProps.transparency);

	//if filled draw the material first
	if (pContext->gProps.filled && pTexture->m_bIsTransparent)
		pContext->m_pVisualSurface->DrawVisualPath(a_path? a_path : &m_Path, TRUE);

	pContext->m_pVisualSurface->TextureVisualPath(a_path? a_path : &m_Path, pTexture, noAntiAlias);
	//reset fill flag
	pContext->m_pVisualSurface->SetBrushStyle(0, pContext->gProps.emissiveColor, pContext->gProps.transparency);
	
	//release texture object
	pContext->m_pTextureProxy->Release();

	mat.SetIdentity();
	pTexture->SetTextureTransform(&mat);
	
	return TRUE;
}


BOOL GeometryNode::DrawGradient(DrawableContext *pContext, VisualPath *a_path)
{
	if (!pContext->m_pGradient) return FALSE;
	Color3f col(0);
	BRect bounds;
	Matrix2D mat, mat2;

	//get original bounds of the object
	GetOriginalBounds(bounds);
	if (bounds.IsEmpty()) return FALSE;
	//init gradient matrix
	pContext->m_pGradient->InitMatrix(&bounds, mat);


	//setup texture transform
	TextureTransform *tx_trans = GetTextureTransform(pContext);
	if (tx_trans) {
		Vec2f transl = (Vec2f)tx_trans->translation;
		Vec2f center = (Vec2f)tx_trans->center;
		Vec2f scale = (Vec2f)tx_trans->scale;
		if (!scale.x) scale.x=0.01;
		if (!scale.y) scale.y=0.01;

		//order as specified in VRML. Note: the spec isn't really clear regarding the 
		//scale factor: since it is applied before the rotation there is no angle used 
		//hence we always scale along (S, T)
		//build everything as reverse (cf spec)
		mat2.AddRotation(-tx_trans->rotation, center.x, center.y);
		mat2.AddScale(1/scale.x, 1/scale.y, 0.0);
		mat2.AddTranslation( - transl.x, - transl.y);

		mat.AddTransform(&mat2);
	}
	//add full transform
	mat.AddTransform(&pContext->m_transform);
	
	pContext->m_pGradient->m_Gradient.SetTransform(&mat);

	//if filled draw the material first
	pContext->m_pVisualSurface->SetBrushStyle(1, pContext->gProps.emissiveColor, pContext->gProps.transparency);
	if (pContext->gProps.filled)
		pContext->m_pVisualSurface->DrawVisualPath(a_path? a_path : &m_Path, TRUE);

	pContext->m_pVisualSurface->DrawVisualPath(a_path? a_path : &m_Path, TRUE, &pContext->m_pGradient->m_Gradient);

	//reset fill flag
	pContext->m_pVisualSurface->SetBrushStyle(0, col, pContext->gProps.transparency);

	return TRUE;
}