/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  GroupNodeProxy.cpp
//
// Author: Jean Le Feuvre - ENST 2002
//
// Description:
//   base class for all grouping nodes (nodes with children field)
//
// Created: August 2002.
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-INIT.h"
#include "GroupNodeProxy.h"
#include "TextProxy.h"


Traversable* GroupNodeProxy::GetChild(int n)
{
	MediaObject *pChild = (*GetObject()->GetChildrenField())[n];
	if (pChild)
		return (Traversable*) pChild->GetRenderingObject(); 
	return NULL;
}


// Apply the Render method to all of the children of this node.
void GroupNodeProxy::RenderChildren(Effects *effects)
{
	Effects *newEffects=new Effects(effects);

	int i;
	//check for modifs of the children
	ResortSensors();
	int length= GetObject()->GetChildCount();

	//fill effects sensor list
	for (i = 0; i <m_sensors.GetLength(); i++) {
		Traversable * pTrav=m_sensors[i];
		newEffects->AddUserEventHandler((UserEventHandlerProxy *)pTrav,&effects->transform);
	}

	//note: sensors may be prerendered too (anchor always is)
	for (i=0; i<length; i++) {
		ChildGroup *cg = NULL;
		//if the parent group is this object, this is a node with postprocessing (form, layer and layout) 
		//so start new child
		if (effects->m_ParentGroup == this) cg = StartChild();
		
		Traversable * pTrav=GetChild(i);
		if(pTrav) pTrav->Render(newEffects);

		if (cg) {
			//should be done in  a cleaner way: grouping nodes may 
			//compute bounds
			cg->ComputeBounds();
			//store orignal bounds
			cg->m_FinalBounds = cg->m_OriginalBounds;
		}
	}
	delete newEffects;
}

//classify all our sensors in a dedicated array
void GroupNodeProxy::ResortSensors()
{
	DWORD now = 1 + GetObject()->GetUpdateTime();
	if (now==m_LastUpdateTime) return;
	m_LastUpdateTime = now;
	m_sensors.Clear();

	int length = GetObject()->GetChildCount();
	for (int i = 0; i<length; i++) {
		Traversable * pTrav=GetChild(i);
		//skip time sensors
		if(pTrav && pTrav->IsSensor()) 
			m_sensors += pTrav;
	}
}

ChildGroup *GroupNodeProxy::StartChild()
{
	ChildGroup *cg = new ChildGroup();
	cg->m_ContextList.Clear();
	m_Groups += cg;
	return cg;
}

void GroupNodeProxy::ResetChildren()
{
	while (m_Groups.GetLength()) {
		ChildGroup *cg = m_Groups[0];
		m_Groups.DeleteFrom(0);
		delete cg;
	}
}


void ChildGroup::ComputeBounds()
{
	m_bIsTextGroup = TRUE;
	m_ascent = m_descent = 0;

	m_OriginalBounds.MakeEmpty();
	for (int i=0; i<m_ContextList.GetLength(); i++) {
		DrawableContext *ct = m_ContextList[i];

		m_OriginalBounds.Union(&ct->m_boundingRect);

		if (!m_bIsTextGroup) continue;
		if (!ct->m_isTextNode) {
			m_bIsTextGroup = FALSE;
			continue;
		}
		float a, d;
		((TextProxy *)ct->m_pDrawable)->GetAscentAndDescent(a, d);
		if (a>m_ascent) m_ascent = a;
		if (d>m_descent) m_descent = d;
	}
}

void ChildGroup::UpdateAndDraw(Effects *parentEffects, BRect *parentClipper)
{
	Matrix2D mat;
	VisualPath path;
	BRect clip;

	//get the final transform of this child group
	mat.AddTranslation(m_FinalBounds.left - m_OriginalBounds.left, m_FinalBounds.top - m_OriginalBounds.top);

	for (int i=0; i<m_ContextList.GetLength(); i++) {
		DrawableContext *ct = m_ContextList[i];
		//update bounds
		mat.Apply(ct->m_boundingRect.left, ct->m_boundingRect.top);
		mat.Apply(ct->m_clippedBoundingRect.left, ct->m_clippedBoundingRect.top);

		//update object matrix
		ct->m_transform.AddTransform(&mat);

		//update transform matrix with parent matrix
		ct->m_transform.AddTransform(&parentEffects->transform);
		//and handlers too
		for (int j=0; j<ct->m_Handlers.GetLength(); j++) {
			ct->m_Handlers[j]->m_Matrix.AddTransform(&parentEffects->transform);
		}

		//get bounds in parent context
		path.Reset();
		path.AddRectangle(ct->m_boundingRect.left, ct->m_boundingRect.top, ct->m_boundingRect.width, ct->m_boundingRect.height);
		path.GetBoundingRect(&ct->m_boundingRect, &parentEffects->transform, 0.0);

		//get clipped bounds in parent context
		path.Reset();
		path.AddRectangle(ct->m_clippedBoundingRect.left, ct->m_clippedBoundingRect.top, ct->m_clippedBoundingRect.width, ct->m_clippedBoundingRect.height);
		path.GetBoundingRect(&ct->m_clippedBoundingRect, &parentEffects->transform, 0.0);

		//get parent context clipper
		path.Reset();
		path.AddRectangle(parentClipper->left, parentClipper->top, parentClipper->width, parentClipper->height);
		path.GetBoundingRect(&clip, &parentEffects->transform, 0.0);
		
		ct->m_clippedBoundingRect.Intersection(&clip);

		//register node in the parent group (this is needed for nested clipping objects)
		if (parentEffects->m_ParentGroup) {
			parentEffects->m_ParentGroup->AddToDrawableContextList(ct);
		}
		//otherwise draw the node
		//NOTE: since the Render is always called in draw order (from last to first)
		//the context in the list are also in the right order. 
		else if (!parentEffects->m_bUseBoundTracking) {
			ct->m_pDrawable->Draw(ct);
		}
	}
}

//extensions for Form and Layout nodes. These nodes require post processing to the
//children drawableContext, hence the group node should have a list of 
//children drawablecontexts.
void GroupNodeProxy::AddToDrawableContextList(DrawableContext* pDrawableContext)
{
	int last = m_Groups.GetLength();
	if (!last) return;
	ChildGroup *cg = m_Groups[last-1];
	cg->m_ContextList += pDrawableContext;
}

