/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  Layer2DProxy.cpp
//
// Author:  Jean Le Feuvre ENST
//
// Description:
//   proxy object for Layer2D node.
//
// Created: August 2002.
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "MediaObj.h"
#include "Presenter.h"
#include "nodes.h"
#include "Layer2DProxy.h"
#include "Background2DProxy.h"
#include "ViewportProxy.h"
#include "drawable.h"
#include "bindable.h"

Layer2DProxy* Presenter::InitLayer2D(Layer2D* pObj)
{
	return new  Layer2DProxy(pObj);
}

Layer2DProxy::Layer2DProxy(MediaObject* pMO) : GroupNodeProxy(pMO) 
{
	m_pBackgroundStack = new BindableStack;
	m_pViewportStack = new BindableStack;
	m_bFirst = TRUE;
	m_drawableContext.m_pColorMatrix = NULL;

}
Layer2DProxy::~Layer2DProxy()
{
	delete m_pBackgroundStack; 
	delete m_pViewportStack; 
}

void Layer2DProxy::Render(Effects *effects)
{
	DWORD now;
	int i;
	Effects *newEffects = new Effects(effects);
	Matrix2D mat;

	newEffects->m_ParentGroup = this;

	//children are rendered with no transform then updated
	newEffects->transform.SetIdentity();


	//first time register bindables
	if (m_bFirst) {
		if (GetLayer2D()->background != NULL_OBJECT) {
			Background2DProxy *pBck = (Background2DProxy *) ((MediaObject *)GetLayer2D()->background)->GetRenderingObject();
			newEffects->m_bDrawBackground = FALSE;
			pBck->Render(newEffects);
			ResetChildren();
		}
		Viewport *pV = (Viewport *) (MediaObject *) GetLayer2D()->viewport;
		if (pV != NULL_OBJECT) {
			m_pViewportStack->AddBindable(pV->GetRenderingObject());
			if (! (Bool) pV->isBound) pV->isBound = TRUE;
		}
	}
	BOOL updateRequired = FALSE;
	now = 1 + GetObject()->GetUpdateTime();

	if (now != m_LastUpdateTime) updateRequired = TRUE;
	//DON'T modify m_LastUpdateTime it is used in the GroupNodeProxy::PreRenderChildren


	now = timeGetTime();

	ViewportProxy *pVP = (ViewportProxy *) m_pViewportStack->GetTop();
	Background2DProxy *pLayerBack = (Background2DProxy *) m_pBackgroundStack->GetTop();

	
	float iWidth = effects->m_pSurface->GetWidth();
	float iHeight = effects->m_pSurface->GetHeight();
	Vec2f size = (Vec2f)GetLayer2D()->size;

	m_drawableContext.m_boundingRect.height = (size.y>=0) ? size.y : iHeight;
	m_drawableContext.m_boundingRect.width = (size.x>=0) ? size.x : iWidth;
	m_drawableContext.m_boundingRect.left = - m_drawableContext.m_boundingRect.width / 2;
	m_drawableContext.m_boundingRect.top = m_drawableContext.m_boundingRect.height / 2;
	m_drawableContext.m_clippedBoundingRect = m_drawableContext.m_boundingRect;
	
	//setup effects
	BRect ClippingRect;
	if (pVP && (Bool) pVP->GetViewport()->isBound) {
		pVP->SetupEffects(newEffects, size.x, size.y, ClippingRect);
		m_drawableContext.m_boundingRect.Intersection(&ClippingRect);
		m_drawableContext.m_clippedBoundingRect = m_drawableContext.m_boundingRect;
	}

	//background should be first rendered
	DrawableContext *pBackContext = NULL;
	if (pLayerBack) {
		ChildGroup *cg = StartChild();

		newEffects->m_bDrawBackground = TRUE;
		pLayerBack->Render(newEffects);
		newEffects->m_bDrawBackground = FALSE;

		//force bounds to be only the layer size
		if (cg->m_ContextList.GetLength()) {
			DrawableContext *tmp = cg->m_ContextList[0];
			tmp->m_boundingRect.SetDimensions(m_drawableContext.m_boundingRect);
			tmp->m_clippedBoundingRect = tmp->m_boundingRect;
		}

		cg->ComputeBounds();
		//store orignal bounds
		cg->m_FinalBounds = cg->m_OriginalBounds;

		//special case in bounds tracking mode: we shall get a context from the visual surface
		//otherwise the background won't be used
		if (newEffects->m_bUseBoundTracking) 
			pBackContext = newEffects->m_pSurface->GetDrawableContext();

		if (cg->m_ContextList.GetLength()){
			DrawableContext *tmp;
			tmp = cg->m_ContextList[0];
		}
	}

	//prerender all children
	GroupNodeProxy::RenderChildren(newEffects);

	//check if we haven't missed background/viewport. If so don't render layout
	//and invalidate scene
	//NB: this is needed because bindable may not be registered in the stack before we call
	//Render
	if (m_bFirst) {
		m_bFirst = FALSE;
		if (!pLayerBack) {
			pLayerBack = (Background2DProxy *) m_pBackgroundStack->GetTop();
			if (pLayerBack) {
				ResetChildren();
				GetPresenter()->InvalidateScene(NULL);
				goto exit;
			}
		}
		if (!pVP) {
			pVP = (ViewportProxy *) m_pViewportStack->GetTop();
			if (pVP) {
				ResetChildren();
				GetPresenter()->InvalidateScene(NULL);
				goto exit;
			}
		}
	}

	//setup the cliping rect on all registered children and update transform 
	for (i=0; i<m_Groups.GetLength(); i++) {
		m_Groups[i]->UpdateAndDraw(effects, &m_drawableContext.m_clippedBoundingRect);
	}
	ResetChildren();


	//update the background context 
	if (pBackContext) {
		DrawableContext *pStaticContext = pLayerBack->GetContext(m_pBackgroundStack);
		pBackContext->m_bIsTransparent = FALSE;
		pBackContext->m_bNeedsDrawing = pStaticContext->m_bNeedsDrawing;
		pBackContext->m_boundingRect = pStaticContext->m_boundingRect;
		pBackContext->m_clippedBoundingRect = pStaticContext->m_clippedBoundingRect;
		pBackContext->m_transform.Set(pStaticContext->m_transform);
		pBackContext->m_pDrawable = pStaticContext->m_pDrawable;
	}

exit:
	delete newEffects;
}