/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  LayoutProxy.cpp
//
// Author:  Jean Le Feuvre ENST
// Description:
//   proxy object for Layout node.
//
// Created: August 2002.
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "MediaObj.h"
#include "Presenter.h"
#include "nodes.h"
#include "LayoutProxy.h"
#include "drawable.h"
#include "TextProxy.h"

LayoutProxy* Presenter::InitLayout(Layout* pObj)
{
	return new  LayoutProxy(pObj);
}

LayoutProxy::LayoutProxy(MediaObject* pMO): GroupNodeProxy(pMO)
{
	m_nAllocatedLength = 0;
	m_nLength=0;
	m_LastUpdateTime=0xffffffff;
	m_previousPos=NULL;
	m_nScrolledOut=0;
	m_AllScrolledOut=FALSE;
	m_maxAscent=0;
	m_maxHeight=0;
	m_startTime=0;
	m_drawableContext.m_pColorMatrix = NULL;
}

// LayoutProxy::~LayoutProxy
LayoutProxy::~LayoutProxy()
{
	ResetChildren();
	if (m_previousPos) delete [] m_previousPos;
}

const char * LayoutProxy::GetJustify(int i) const
{
	if (i>GetLayout()->justify.GetLength())
		return "";
	else
	return (LPCSTR)(const ZString&)GetLayout()->justify[i];
}

//Render the children in the order specified by the Layout order field.
void LayoutProxy::Render(Effects *effects)
{
	Effects *newEffects = new Effects(effects);

	m_nLength= GetObject()->GetChildCount();
	if (m_nLength > m_nAllocatedLength-1) {
		// Delete the previous arrays - if required
		if (m_previousPos)
			delete [] m_previousPos;
		m_previousPos=new int [m_nLength+1];
		m_nAllocatedLength = m_nLength+1;
	}

	newEffects->m_ParentGroup=this;
	//render with a null transform so that we can compute the align/distr
	newEffects->transform.SetIdentity();

	//create the Layout bounding rect which is referred in the GroupIndexList as 0 
	InitLayoutContext();

	GroupNodeProxy::RenderChildren(newEffects);

	if (m_startTime==0)
		m_startTime = GetFlexTime();

	ApplyFormats();
	ScrollObjects(GetFlexTime());

	//setup the cliping rect on all registered
	BRect *rc = &m_Groups[0]->m_OriginalBounds;
	for (int i=0; i<m_Groups.GetLength(); i++) {
		m_Groups[i]->UpdateAndDraw(effects, rc);
	}
	ResetChildren();
	delete newEffects;
}


void LayoutProxy::InitLayoutContext()
{
	ChildGroup *cg = StartChild();
	cg->m_OriginalBounds.height = ((Vec2f)GetLayout()->size).y;
	cg->m_OriginalBounds.width = ((Vec2f)GetLayout()->size).x;
	cg->m_OriginalBounds.left = -cg->m_OriginalBounds.width/2;
	cg->m_OriginalBounds.top = cg->m_OriginalBounds.height/2;
	//store final bounds
	cg->m_FinalBounds = cg->m_OriginalBounds;

	AddToDrawableContextList(&m_drawableContext);
}


// LayoutProxy::ApplyFormats
// returns a pointer to the indexes of DrawableContext taken from 
// the Contextset of nodes from the groupIndex field of a Form
// each group is separated by a -1
void LayoutProxy::ApplyFormats()
{
	// now in m_drawableContextList there is the list of objects to 
	GetMaxChildrenHeight(m_maxHeight,m_maxAscent,m_maxDescent,m_maxWidth);	
	BRect *pLayoutBRect = &m_Groups[0]->m_FinalBounds;
	
	m_step = m_maxHeight * GetLayout()->spacing;
	
	float start_y, lineHeight;
	CString major, minor;
	major = GetJustify(0);
	minor = GetJustify(1);

	BOOL ttb = (Bool)(GetLayout()->topToBottom);

	// horizontal aligment
	if ((Bool)(GetLayout()->horizontal)) {
		if (minor=="FIRST") {
			lineHeight = max(m_maxAscent, m_maxHeight ) + m_maxDescent;
		} else {
			lineHeight = max(m_maxAscent + m_maxDescent, m_maxHeight);
		}

		//minor-justification is for the whole layout
		if (minor=="MIDDLE") {
			if (ttb)
				start_y = pLayoutBRect->height/2;
			else
				start_y = m_maxDescent + m_maxAscent - pLayoutBRect->height/2;
		}
		else if (minor=="BEGIN") {
			start_y = ttb ? m_maxDescent : m_maxDescent + m_maxAscent;
		}
		else if (minor=="END") {
			start_y = ttb ? pLayoutBRect->height : -pLayoutBRect->height + 2 * m_maxDescent + m_maxAscent;
		}
		else {
			start_y = m_maxAscent;
		}
		
		//set left coordinate of the first object looking to the major axis attribute
		CString csJustify = GetJustify(0);
		if(csJustify=="FIRST"||csJustify=="BEGIN") {
			m_Groups[1]->m_FinalBounds.left = pLayoutBRect->left;
		}
		else if(csJustify=="MIDDLE") {
			m_Groups[1]->m_FinalBounds.left = pLayoutBRect->left + (pLayoutBRect->width - m_ChildrenWidth)/2;			

		}
		else if(csJustify=="END") {
			m_Groups[1]->m_FinalBounds.left = pLayoutBRect->left + pLayoutBRect->width - m_ChildrenWidth;
		}          
		//set top coordinate of the first object looking to the minor axis attribute
		//set also left and top of the other objects taking into account the space
		//between them
		csJustify = GetJustify(1);

		if(csJustify=="FIRST") {
			for (int i=1;i<m_Groups.GetLength();i++) {
				if (m_Groups[i]->m_bIsTextGroup) {
					m_Groups[i]->m_FinalBounds.top = pLayoutBRect->top - m_maxHeight + m_Groups[i]->m_ascent;
				} else {
					m_Groups[i]->m_FinalBounds.top = pLayoutBRect->top - m_maxHeight + m_Groups[i]->m_FinalBounds.height;
				}
				if (i==1) continue;
				
				m_Groups[i]->m_FinalBounds.left = m_Groups[i-1]->m_FinalBounds.left +
													m_Groups[i-1]->m_FinalBounds.width +
													GetLayout()->spacing;
			}
			// to add for wrap: set size here to max(maxAscent,maxHeight)+maxDescent
		}
		//top aligment
		else if(csJustify=="BEGIN") {
			for (int i=1; i<m_Groups.GetLength(); i++) {
				m_Groups[i]->m_FinalBounds.top = pLayoutBRect->top;
				if(i==1) continue;
				m_Groups[i]->m_FinalBounds.left = m_Groups[i-1]->m_FinalBounds.left +
													m_Groups[i-1]->m_FinalBounds.width +
													GetLayout()->spacing;
			}
			// to add for wrap: set size here to max(maxAscent+maxDescent,maxHeight)			
		}
		//(top-bottom)/2 aligment
		else if(csJustify=="MIDDLE") {
			for (int i=1; i<m_Groups.GetLength(); i++) {
				m_Groups[i]->m_FinalBounds.top = pLayoutBRect->top - (m_maxHeight - m_Groups[i]->m_FinalBounds.height)/2;
				if(i==1) continue;
				m_Groups[i]->m_FinalBounds.left = m_Groups[i-1]->m_FinalBounds.left +
													m_Groups[i-1]->m_FinalBounds.width +
													GetLayout()->spacing;
			}
			// to add for wrap: set size here to max(maxAscent+maxDescent,maxHeight)
		}
		//bottom aligment
		else if(csJustify=="END") {
			for (int i=1; i<m_Groups.GetLength();i++) {
				m_Groups[i]->m_FinalBounds.top = pLayoutBRect->top - m_maxHeight + m_Groups[i]->m_FinalBounds.height;
				if (i==1) continue;
				
				m_Groups[i]->m_FinalBounds.left = m_Groups[i-1]->m_FinalBounds.left + 
													m_Groups[i-1]->m_FinalBounds.width +
													GetLayout()->spacing;
			}
		}
	}
	//Vertical aligment
	else {
		CString csJustify = GetJustify(0);
		//NEEDS SPEC CHECKING !!!
		if(csJustify=="FIRST")
			m_Groups[1]->m_FinalBounds.top = pLayoutBRect->top;
		else if(csJustify=="BEGIN")
			m_Groups[1]->m_FinalBounds.top = pLayoutBRect->top;
		else if(csJustify=="MIDDLE")
			m_Groups[1]->m_FinalBounds.top = pLayoutBRect->top - (pLayoutBRect->height - m_ChildrenHeight)/2;
		else if(csJustify=="END")
			m_Groups[1]->m_FinalBounds.top = pLayoutBRect->top - pLayoutBRect->height + m_ChildrenHeight;
		//NEEDS SPEC CHECKING !!!!

		csJustify = GetJustify(1);

		if(csJustify=="FIRST"||csJustify=="BEGIN") {
			for (int i=1; i<m_Groups.GetLength(); i++) {
				m_Groups[i]->m_FinalBounds.left = m_Groups[0]->m_FinalBounds.left;
				if(i==1) continue;
				m_Groups[i]->m_FinalBounds.top = m_Groups[i-1]->m_FinalBounds.top - 
													m_Groups[i-1]->m_FinalBounds.height - m_step;
			}
		}	
		else if(csJustify=="MIDDLE") {
			for (int i=1; i<m_Groups.GetLength(); i++) {
				m_Groups[i]->m_FinalBounds.left = - (m_Groups[i]->m_FinalBounds.width)/2;
				if(i==1) continue;
				m_Groups[i]->m_FinalBounds.top = m_Groups[i-1]->m_FinalBounds.top - 
													m_Groups[i-1]->m_FinalBounds.height - 
													m_step;
			}
		}
		else if(csJustify=="END") {
			for (int i=1; i<m_Groups.GetLength();i++) {
				m_Groups[i]->m_FinalBounds.left = m_Groups[0]->m_FinalBounds.left +
													m_Groups[0]->m_FinalBounds.width -
													m_Groups[i]->m_FinalBounds.width;
				if(i==1) continue;
				m_Groups[i]->m_FinalBounds.top = m_Groups[i-1]->m_FinalBounds.top -
													m_Groups[i-1]->m_FinalBounds.height -			
													m_step;
			}
		}
	}			
}

void LayoutProxy::ScrollObjects(DWORD currentTime)
{
	BOOL loop =(Bool)GetLayout()->loop;
	BOOL scrollVertical =(Bool)GetLayout()->scrollVertical;
	float scrollRate = GetLayout()->scrollRate;
	BRect *pLayoutBRect = &m_Groups[0]->m_FinalBounds;

	DWORD dwUpdateTime = GetLayout()->GetUpdateTime();	
	if (dwUpdateTime != m_LastUpdateTime) {
		m_LastUpdateTime = dwUpdateTime;
		for (int i=1; i<m_Groups.GetLength();i++) {
			if (scrollVertical)
				m_previousPos[i] = m_Groups[i]->m_FinalBounds.top;
			else
				m_previousPos[i] = m_Groups[i]->m_FinalBounds.left;
		}
	}
    BOOL bScroll = FALSE;
	
	if ((currentTime - m_startTime)>=1000) {
		m_startTime=currentTime;
		bScroll=TRUE;
	
	}
	for (int i=1; i<m_Groups.GetLength(); i++) {
		BRect *pBRect = &m_Groups[i]->m_FinalBounds;
		if (bScroll)
			m_previousPos[i] += scrollRate;
		if (scrollVertical)
			pBRect->top = m_previousPos[i];
		else
			pBRect->left = m_previousPos[i];

		if (loop) {
			if (!pBRect->Overlaps(pLayoutBRect) && m_AllScrolledOut) {
				if (scrollVertical) {
					if (i==1)
						m_previousPos[i]=(pLayoutBRect->top - pLayoutBRect->height);
					else
						m_previousPos[i] = m_previousPos[i-1] - m_Groups[i-1]->m_FinalBounds.height - m_step;
				} else {
					m_previousPos[i] = m_previousPos[i] - pLayoutBRect->width - m_maxWidth;//add space here
				}
			} else if (!pBRect->Overlaps(pLayoutBRect) && !m_AllScrolledOut) {
				pBRect->MakeEmpty();
				++m_nScrolledOut;
			}
		}
	}		
	
	if(	m_AllScrolledOut) {
		m_AllScrolledOut=FALSE;
		m_nScrolledOut=0;
	}
	if (m_nScrolledOut==m_Groups.GetLength()-1)
		m_AllScrolledOut=TRUE;
	else 
		m_nScrolledOut=0;
}

void LayoutProxy::GetMaxChildrenHeight(float &maxHeight,float &maxAscent,float &maxDescent,float &maxWidth)
{
	m_ChildrenHeight = maxHeight = m_Groups[1]->m_FinalBounds.height;
	m_ChildrenWidth = maxWidth = m_Groups[1]->m_FinalBounds.width;

	maxAscent = maxDescent = 0;
	if (m_Groups[1]->m_bIsTextGroup) {
		maxAscent = m_Groups[1]->m_ascent;
		maxDescent = m_Groups[1]->m_descent;
	}
	
	for (int i=2; i<m_Groups.GetLength(); i++) {
		if (m_Groups[i]->m_FinalBounds.height > maxHeight)
			maxHeight = m_Groups[i]->m_FinalBounds.height;
		
		if (m_Groups[i]->m_FinalBounds.width > maxWidth)
			maxWidth = m_Groups[i]->m_FinalBounds.width;
		
		if (m_Groups[i]->m_bIsTextGroup) {
			(m_Groups[i]->m_ascent > maxAscent) ? maxAscent = m_Groups[i]->m_ascent : 1;
			(m_Groups[i]->m_descent > maxDescent) ? maxDescent = m_Groups[i]->m_descent : 1;
		}
		m_ChildrenWidth += m_Groups[i]->m_FinalBounds.width;
		m_ChildrenHeight += m_Groups[i]->m_FinalBounds.height;
	}
	//add interobjects spacing
	m_ChildrenWidth += (m_Groups.GetLength() - 2) * GetLayout()->spacing;
	m_ChildrenHeight += (m_Groups.GetLength() - 2) * GetLayout()->spacing;
}
