/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  Matrix2D.cpp
//
// Author: Jean Le Feuvre
//
// Description:
//   2D transform matrix wrapper for GDIplus.
//
// Created: August, 2002.
//
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <stdio.h>
#include <math.h>
#include "Matrix2D.h"

Matrix2D::Matrix2D()
{
	m_Matrix.Reset();
}

Matrix2D::Matrix2D(const Matrix2D& mat)
{
	Set(mat);
}
void Matrix2D::SetIdentity()
{
	m_Matrix.Reset();
}

void Matrix2D::Set(float m1, float m2, float m3, float m4, float m5, float m6)
{
	/*Gdiplus inverting Y, the coef ARE NOT THE REGULAR ONES (matrix used is the symetric)
	regular (BIFS) matrix is:
		m1 m2 m3
		m4 m5 m6
		0  0  1
	GDIplus matrix is
		m1 m4 0
		m2 m5 0
		m3 m6 1	
	*/
	m_Matrix.SetElements(m1, m4, m2, m5, m3, m6);
}


void Matrix2D::Set(float *coefs)
{
	Set(coefs[0], coefs[1], coefs[2], coefs[3], coefs[4], coefs[5]);
}

void Matrix2D::Set(const Matrix2D &mat) 
{   
	float cf[6];
	mat.m_Matrix.GetElements(cf);
	m_Matrix.SetElements(cf[0], cf[1], cf[2], cf[3], cf[4], cf[5]);
}

void Matrix2D::AddTranslation(float dx, float dy)
{
	m_Matrix.Translate(dx, dy, Gdiplus::MatrixOrderAppend);
}

void Matrix2D::AddScale(float sx, float sy, float orientation, float center_x, float center_y)
{
	if ((sx== 1) && (sy==1)) return;

	Gdiplus::PointF pt(center_x, center_y);
	//rotate coords
	if (orientation != 0.0) m_Matrix.RotateAt(-1*orientation*180/PI, pt, Gdiplus::MatrixOrderAppend);
	//apply scaling
	m_Matrix.Scale(sx, sy, Gdiplus::MatrixOrderAppend);
	//restore coords
	if (orientation != 0.0) m_Matrix.RotateAt(orientation*180/PI, pt, Gdiplus::MatrixOrderAppend);
}

void Matrix2D::AddRotation(float angle, float center_x, float center_y)
{
	if (angle == 0.0) return;
	Gdiplus::PointF pt(center_x, center_y);
	m_Matrix.RotateAt(angle*180/PI, pt, Gdiplus::MatrixOrderAppend);
}

void Matrix2D::AddTransform(Matrix2D *trans)
{
	if (!trans) return;
	m_Matrix.Multiply(&trans->m_Matrix, Gdiplus::MatrixOrderAppend);
}


void Matrix2D::Apply(float& x, float& y) const
{
	Gdiplus::PointF pt(x, y);
	m_Matrix.TransformPoints(&pt, 1);
	x = pt.X;
	y = pt.Y;
}

void Matrix2D::Apply(long& x, long& y) const
{
	Gdiplus::Point pt(x, y);
	m_Matrix.TransformPoints(&pt, 1);
	x = pt.X;
	y = pt.Y;
}
void Matrix2D::Apply(int& x, int& y) const
{
	Gdiplus::Point pt(x, y);
	m_Matrix.TransformPoints(&pt, 1);
	x = pt.X;
	y = pt.Y;
}

BOOL Matrix2D::ReverseApply(float& x, float& y) 
{
	if (!m_Matrix.IsInvertible()) return FALSE;

	Gdiplus::PointF pt(x, y);
	Gdiplus::Matrix *mat = m_Matrix.Clone();
	mat->Invert();
	mat->TransformPoints(&pt, 1);
	x = pt.X;
	y = pt.Y;
	delete mat;
	
	return TRUE;
}

BOOL Matrix2D::Inverse()
{
	if (!m_Matrix.IsInvertible()) return FALSE;
	m_Matrix.Invert();
	return TRUE;
}


BOOL Matrix2D::HasRotations()
{
	float coefs[6];
	m_Matrix.GetElements(coefs);
	if (coefs[1] != 0.0) return 1;
	if (coefs[2] != 0.0) return 1;
	return 0;
}

void Matrix2D::GetElements(float *m)
{
	float coefs[6];
	m_Matrix.GetElements(coefs);
	m[0] = coefs[0];
	m[1] = coefs[2];
	m[2] = coefs[4];
	m[3] = coefs[1];
	m[4] = coefs[3];
	m[5] = coefs[5];
}

int operator== (const Matrix2D& mat1,
				const Matrix2D& mat2)
{
	
	return mat1.m_Matrix.Equals(&mat2.m_Matrix);
}

int operator!= (const Matrix2D& mat1,
				const Matrix2D& mat2)
{
	return !operator==(mat1, mat2);
}

