/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// Author: marc Patoureaux marc@envivio.com
//				Jean Le Feuvre ENST 2002
//
// Description:
//   The implementation of the proxy object used by the presenter
//   to render the MatteTexture node.
//   all functions are not implemented, but the most difficult ones are
//   adding new functions wouldn't had any complexity.
//
// Created: 18 november 2001.
// Modified: september 2002 - full alpha support (given you do have an input alpha texture :)
//   note that alpha surface are replaces by the R value of regular 3 colour surfaces
//   in case the alpha_surface is an RGB surface, otherwiase the alpha channel is used
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-INIT.h"
#include "MatteTextureProxy.h"
#include "Stream.h"
#include "Presenter.h"

enum {VIDEO_DELAY = 0};

static LPCSTR s_token[] = {
    "INVALID",
    "OFFSET",
    "SCALE",
    "BLUR",
    "BIAS",
    "COLOR_MATRIX",
    "REPLACE_ALPHA",
    "MULTIPLY_ALPHA",
    "CROSS_FADE",
    "BLEND",
    "ADD",
    "A",
    "B",
    "REVEAL"
};


MatteTextureProxy * Presenter::InitMatteTexture(MatteTexture* pObj)
{
	return new MatteTextureProxy(pObj);
}

MatteTextureProxy::MatteTextureProxy(MediaObject* pMO) : DrawableImage(pMO)
{
	m_bOverwrite = FALSE;
	m_LastUpdateTime = 0;
}

MatteTextureProxy::~MatteTextureProxy() 
{
	//if we are in overwrite mode disable the texture object otherwise it would be deleted twice
	if (m_bOverwrite) m_pTexture = NULL;
}	

void MatteTextureProxy::DeleteObject()
{
	MediaObject * a_mediaObject;
    RenderingObject * a_renderingObject;
    
	//do NOT delete other textures, this shall be signaled by the core (because of DEF/USE)
	return;

    a_mediaObject = (MediaObject*) GetMatteTexture()->surfaceA;
    if (a_mediaObject) {
        a_renderingObject = a_mediaObject->GetRenderingObject();
        if (a_renderingObject) {
            ((DrawableImage* ) a_renderingObject)->DeleteObject();
        }
    }
    a_mediaObject = (MediaObject*) GetMatteTexture()->surfaceB;
    if (a_mediaObject) {
        a_renderingObject = a_mediaObject->GetRenderingObject();
        if (a_renderingObject) {
            ((DrawableImage* ) a_renderingObject)->DeleteObject();
        }
    }
    a_mediaObject = (MediaObject*) GetMatteTexture()->alphaSurface;
    if (a_mediaObject) {
        a_renderingObject = a_mediaObject->GetRenderingObject();
        if (a_renderingObject) {
            ((DrawableImage* ) a_renderingObject)->DeleteObject();
        }
    }
}

void MatteTextureProxy::Update(DrawableContext * pContext)
{
	//1 - call update on each surfaces
    MediaObject * a_mediaObject;
	RenderingObject *a_renderingObject;

	surface_a = surface_b = surface_alpha = NULL;
	BOOL NeedsCompute = FALSE;

    a_mediaObject = (MediaObject*) GetMatteTexture()->surfaceA;
    if (a_mediaObject) {
        a_renderingObject = a_mediaObject->GetRenderingObject();
        if (a_renderingObject) {
			surface_a = (DrawableImage* ) a_renderingObject;
            surface_a->Update(pContext);
			if (pContext->m_bUpdateTexture) NeedsCompute = TRUE;
        }
    }
    a_mediaObject = (MediaObject*) GetMatteTexture()->surfaceB;
	//we shall have at least B
    if (!a_mediaObject) return;
    a_renderingObject = a_mediaObject->GetRenderingObject();
    if (a_renderingObject) {
		surface_b = (DrawableImage* ) a_renderingObject;
        surface_b->Update(pContext);
		if (pContext->m_bUpdateTexture) NeedsCompute = TRUE;
    }

    a_mediaObject = (MediaObject*) GetMatteTexture()->alphaSurface;
    if (a_mediaObject) {
        a_renderingObject = a_mediaObject->GetRenderingObject();
        if (a_renderingObject) {
			surface_alpha = (DrawableImage* ) a_renderingObject;
            surface_alpha->Update(pContext);
			if (pContext->m_bUpdateTexture) NeedsCompute = TRUE;
        }
    }
	pContext->m_bUpdateTexture = FALSE;
	//surface B isn't ready, return
	if (!surface_b->m_pTexture) return;

	DWORD now = 1 + GetObject()->GetUpdateTime();
	if (!NeedsCompute && m_LastUpdateTime == now) return;
	m_LastUpdateTime = now;
	pContext->m_bUpdateTexture = TRUE;
	if (m_bOverwrite) surface_b->Invalidate();

	//2 - setup output image from surface B
	if (!m_bOverwrite && m_pTexture) delete m_pTexture;
    m_bOverwrite = GetMatteTexture()->overwrite.GetValue();
	if (m_bOverwrite) {
		m_pTexture = surface_b->m_pTexture;	
	} else {
		m_pTexture = surface_b->m_pTexture->Clone();
		if (!m_pTexture) return;
	}
	
	//3 - parse effects
    parseOperation((LPCSTR) GetMatteTexture()->operation.GetValue());

    switch(m_token) {
    case BIAS :
	    ApplyBIAS();
	    break;
	case SCALE :
	    ApplySCALE();
	    break;
    case BLUR :
        ApplyBLUR();
        break;
    case REPLACE_ALPHA:
		ApplyREPLACE_ALPHA();
		break;
	case CROSS_FADE : 
   		ApplyCROSS_FADE();
		break;
	case REVEAL:
        ApplyREVEAL();
        break;
    default :
        break;
	}

	//finally apply texture repeat
	BOOL rS, rT;
	GetTextureRepeat(rS, rT);
	m_pTexture->SetTextureRepeat(rS, rT);
}


void MatteTextureProxy::Release() 
{
	//release each texture
	if (surface_a) surface_a->Release();
	if (surface_b) surface_b->Release();
	if (surface_alpha) surface_alpha->Release();
}



void MatteTextureProxy::parseOperation(LPCSTR operation) {
    for (int i = INVALID ; i < NBOTOKEN; i++) {
        if (!stricmp(operation,s_token[i])) {
            m_token = (oToken) i;
            return;
        }
    }
    m_token = INVALID;
}

void MatteTextureProxy::ApplyBIAS()
{
	float a_parameter[4];
    int i, length;
	length = GetMatteTexture()->parameter.GetLength();
    for ( i = 0; i < 4; i ++) {
		a_parameter[i] = (i < length) ? GetMatteTexture()->parameter[i] : 0;
    }
	m_pTexture->TranslateColors(a_parameter);
}

void MatteTextureProxy::ApplySCALE()
{
    float a_parameter[4];
    int i, length;
	length = GetMatteTexture()->parameter.GetLength();
    for (i=0; i<4; i++) {
		a_parameter[i] = (i<length) ? GetMatteTexture()->parameter[i] : 0;
    }
	m_pTexture->ScaleColors(a_parameter);
}

float b_kernel(float x, float s) {
    if (s != 0) {
        return exp(-x*x/(2*s*s));
    } else if (x==0) {
        return 10;
    } else {
        return 0;
    }
}

void MatteTextureProxy::ApplyBLUR() 
{
	unsigned int width, height;
	BOOL hasAlpha = surface_b->m_pTexture->m_bIsTransparent;
	//we need an extra backup since one blur step modifies several pixel in the source
	VisualTexture *text_temp = surface_b->m_pTexture->Clone();
	if (!text_temp) return;

	width = surface_b->m_pTexture->m_Width;
	height = surface_b->m_pTexture->m_Height;
	unsigned int i, j;
	int l, m;

    // This is a basic integration of the blur kernel function !!!!!
    // this can be better done...
    float a_parameter[2];

    switch (GetMatteTexture()->parameter.GetLength()) {
    case 0:
		delete text_temp;
        return;
    case 1:
        a_parameter[0] = GetMatteTexture()->parameter[0];
        a_parameter[1] = a_parameter[0];
        break;
    default :
        a_parameter[0] = GetMatteTexture()->parameter[0];
        a_parameter[1] = GetMatteTexture()->parameter[1];
		break;
    }

    if (a_parameter[0] == 0 && a_parameter[1] == 0) {
		if (!m_bOverwrite) {
			delete m_pTexture;
			m_pTexture = text_temp;
		} else {
			delete text_temp;
		}
        return;
	}


    unsigned int Xcount = 0;
    while (b_kernel(Xcount+0.5, a_parameter[0]) > 0.05) {
        Xcount++;
    }
    unsigned int Ycount = 0;
    while (b_kernel(Ycount+0.5, a_parameter[1]) > 0.05) {
        Ycount++;
    }

    if (Xcount == 0 && Ycount == 0) {
		if (!m_bOverwrite) {
			delete m_pTexture;
			m_pTexture = text_temp;
		} else {
			delete text_temp;
		}
        return;
	}
    
	float * a_XRatio = new float[2*Xcount +1];
    float * a_YRatio = new float[2*Ycount +1];
    float Xsum = 0;
    float Ysum = 0;

    for ( i = 0; i <= Xcount; i++) {
        a_XRatio[Xcount+i] = 0;
        for ( int f = -5; f < 5; f++) {
            if ((10 * i + f) < 0) continue;
            a_XRatio[Xcount+i] += b_kernel((float)(10*i + f) / 10.0, a_parameter[0]) * 0.1;
        }
        Xsum += 2*a_XRatio[Xcount+i];
        if (i == 0) {
            a_XRatio[Xcount-i] *= 2;
        } else {
            a_XRatio[Xcount-i] = a_XRatio[Xcount+i];
        }
    }
    for ( i = 0; i <= Ycount; i++) {
        a_YRatio[Ycount+i] = 0;
        for ( int f = -5; f < 5; f++) {
            if ((10 * i + f) < 0) continue;
            a_YRatio[Ycount+i] += b_kernel((float)(10*i + f) / 10.0, a_parameter[1]) * 0.1;
        }
        Ysum += 2*a_YRatio[Ycount+i];
        if (i == 0) {
            a_YRatio[Ycount-i] *= 2;
        } else {
            a_YRatio[Ycount-i] = a_YRatio[Ycount+i];
        }
    }

    for (int z = -1 * Xcount; z <= (int) Xcount; z++) {
        a_XRatio[Xcount+z] /= Xsum;
    }
    for (z = -1 * Ycount; z <= (int) Ycount; z++) {
        a_YRatio[Ycount+z] /= Ysum;
    }

    for (i=0; i<width; i++) {
        for (j=0; j<height; j++) {
            Color4 color = 0;
            for (l = -1 * Xcount; l <= (int) Xcount; l++) {
                for (m = -1 * Ycount; m <= (int) Ycount; m++) {
                    if ((j+m < 0)  || (j+m >= height) || (i+l < 0) || (i+l >= width))
						continue;

					Color4 a_col = text_temp->GetPixel(i+l, j+m);
                    color.r += a_XRatio[Xcount+l] * a_YRatio[Ycount+m] * a_col.r;
                    color.g += a_XRatio[Xcount+l] * a_YRatio[Ycount+m] * a_col.g;
                    color.b += a_XRatio[Xcount+l] * a_YRatio[Ycount+m] * a_col.b;
					if (hasAlpha) {
	                    color.a += a_XRatio[Xcount+l] * a_YRatio[Ycount+m] * a_col.a;
					} else {
						color.a = 255;
					}
                }
            }
			m_pTexture->SetPixel(i, j, color);
        }
    }
    delete[] a_XRatio;
    delete[] a_YRatio;
	delete text_temp;
}

void MatteTextureProxy::ApplyCROSS_FADE() 
{
	if (!surface_a->m_pTexture) return;

    float a_fraction = (float) GetMatteTexture()->fraction;
    if (a_fraction > 1) { 
        a_fraction = 1;
    } else if (a_fraction < 0) {
        a_fraction = 0;
    }
	
    float scaleX = (float) surface_a->m_pTexture->m_Width / (float) surface_b->m_pTexture->m_Width;
    float scaleY = (float) surface_a->m_pTexture->m_Height / (float) surface_b->m_pTexture->m_Height;

	unsigned int width, height;
	width = surface_b->m_pTexture->m_Width;
	height = surface_b->m_pTexture->m_Height;

    int X_A, Y_A;
    for (unsigned int Y = 0; Y < height; Y++) {
        Y_A = (int) (scaleY * Y);
        for (unsigned int X = 0; X < width; X++) {
			X_A = (int) (scaleX*X);
			Color4 col_a, col_b;
			col_a = surface_a->m_pTexture->GetPixel(X_A, Y_A);
			col_b = surface_b->m_pTexture->GetPixel(X, Y);

			col_b.r = (unsigned char) ( a_fraction * col_b.r + (1 - a_fraction) * col_a.r);
			col_b.g = (unsigned char) ( a_fraction * col_b.g + (1 - a_fraction) * col_a.g);
			col_b.b = (unsigned char) ( a_fraction * col_b.b + (1 - a_fraction) * col_a.b);
			col_b.a = (unsigned char) ( a_fraction * col_b.a + (1 - a_fraction) * col_a.a);
			m_pTexture->SetPixel(X, Y, col_b);
        }
    }
}

void MatteTextureProxy::ApplyREPLACE_ALPHA()
{
	if (!surface_alpha->m_pTexture) return;

    float scaleX = (float) surface_alpha->m_pTexture->m_Width / (float) surface_b->m_pTexture->m_Width;
    float scaleY = (float) surface_alpha->m_pTexture->m_Height / (float) surface_b->m_pTexture->m_Height;

	unsigned int width, height;
	width = surface_b->m_pTexture->m_Width;
	height = surface_b->m_pTexture->m_Height;
	BOOL useR = !surface_alpha->m_pTexture->m_bIsTransparent;

    int X_A, Y_A;
    for (unsigned int Y = 0; Y < height; Y++) {
        Y_A = (int) (scaleY * Y);
        for (unsigned int X = 0; X < width; X++) {        
            X_A = (int) (scaleX*X);
			Color4 col_a, col_b;
			col_a = surface_alpha->m_pTexture->GetPixel(X_A, Y_A);
			col_b = surface_b->m_pTexture->GetPixel(X, Y);

			if (useR) col_a.a = col_a.r;

            col_b.r = col_b.r * col_a.a / 255;
            col_b.g = col_b.g * col_a.a / 255;
            col_b.b = col_b.b * col_a.a / 255;
//			col_b.a = col_b.a * col_a.a / 255;

			m_pTexture->SetPixel(X, Y, col_b);
        }
    }
}

void MatteTextureProxy::ApplyREVEAL() 
{
	if (!surface_a->m_pTexture) return;
	if (!surface_alpha->m_pTexture) return;

    float a_fraction  = (float) GetMatteTexture()->fraction;
    float a_parameter = (float) GetMatteTexture()->parameter[0];

	unsigned int width, height;
	width = surface_b->m_pTexture->m_Width;
	height = surface_b->m_pTexture->m_Height;
	BOOL useR = !surface_alpha->m_pTexture->m_bIsTransparent;

    float scaleXA = (float) surface_a->m_pTexture->m_Width /  (float) width;
    float scaleYA = (float) surface_a->m_pTexture->m_Height / (float) height;
    float scaleXAlpha = (float) surface_alpha->m_pTexture->m_Width /  (float) width;
    float scaleYAlpha = (float) surface_alpha->m_pTexture->m_Height / (float) height;

    int A_X, A_Y, Alpha_X, Alpha_Y;

    for (unsigned int Y=0; Y<height; Y++) {
		Alpha_Y = (int) (scaleYAlpha * Y);
        A_Y = (int) (scaleYA * Y);
        for (unsigned int X = 0; X < width; X++) {
			Alpha_X = (int) (scaleXAlpha * X);
            A_X = (int) (scaleXA * X);
            
			Color4 col_a, col_b, col_alpha;
			col_a = surface_a->m_pTexture->GetPixel(A_X, A_Y);
			col_b = surface_b->m_pTexture->GetPixel(X, Y);
			col_alpha = surface_alpha->m_pTexture->GetPixel(Alpha_X, Alpha_Y);

			if (useR) col_alpha.a = col_alpha.r;
			
			if (col_alpha.a < 255 * (a_fraction - a_parameter)) {
                col_b = col_a;
            } else if (col_alpha.a < 255 * (a_fraction + a_parameter)) {
                float a_ratio = 0.5 + ((((float)col_alpha.a)/255.0) - a_fraction) / a_parameter / 2;
                if (a_ratio > 1) { 
                    a_ratio = 1;
                } else if (a_ratio < 0) {
                    a_ratio = 0;
                }
                col_b.r = a_ratio * col_b.r + (1-a_ratio)*col_a.r;
                col_b.g = a_ratio * col_b.g + (1-a_ratio)*col_a.g;
                col_b.b = a_ratio * col_b.b + (1-a_ratio)*col_a.b;
                col_b.a = a_ratio * col_b.a + (1-a_ratio)*col_a.a;
            }
 			m_pTexture->SetPixel(X, Y, col_b);
       }
    }
}