/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  MovieTextureProxy.cpp
//
// Author: Jean Le Feuvre ENST 2002
//
// Description:
//   proxy object for MovieTexture node.
//
// Created: august 2002.
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-INIT.h"
#include "MovieTextureProxy.h"
#include "Stream.h"
#include "Presenter.h"
#include "ShapeProxy.h"
#include "ODImp.h"

//if 1 the movieTexture will show the first frame of the movie before being activated (VRML)
//if 0 the movietexture will not open the stream until activated (less resources)
#define ENABLE_FIRST_FRAME_FETCH	1

MovieTextureProxy * Presenter::InitMovieTexture(MovieTexture* pObj)
{
	return new MovieTextureProxy(pObj);
}

MovieTextureProxy::MovieTextureProxy(MediaObject* pMO) : DrawableImage(pMO)
{
	/*init default*/
	m_startTime = 0;
	m_RestartStatus = ST_None;
	m_bNeedsUnregister = FALSE;
	m_bFirstTime = TRUE;
	m_bFirstFrameFetched = FALSE;
	m_bRegWithMC = FALSE;
	m_cycle = -1;
	m_bOpened = FALSE;
	m_iNumLoop = 0;
	/*register as timed node*/
	GetPresenter()->RegisterTimedNode(this);
	m_bIsRegistered = TRUE;
	m_bSetSpeed = FALSE;
}

MovieTextureProxy::~MovieTextureProxy()
{
	//unregister time node
	if (m_bIsRegistered) GetPresenter()->UnregisterTimedNode(this);
	//unregister from media control if needed
	if (m_bRegWithMC) {
		MediaControlImp *pCtrl = GetMovieTexture()->GetObjectMediaControl();
		if (pCtrl) pCtrl->RemoveUser();
	}
}	
	
BOOL MovieTextureProxy::ShouldBuildBrush()
{
	MovieTexture *pMT=GetMovieTexture();
	if ((Bool) pMT->isActive) return TRUE;

	//if first time use the first frame as a brush and don't update
	if (!m_bFirstFrameFetched) return TRUE;
	return FALSE; 
}

void MovieTextureProxy::Update(DrawableContext * pContext)
{
	MovieTexture *pMT=GetMovieTexture();
	if (!pMT->GetOD()) return;

	DrawableImage::Update(pContext);

	/*end of stream*/
	if(pContext->m_hasStreamFinished==TRUE) {
		if (IsLooping()) {
			/* cf note in TimedNode.h */
			switch (m_RestartStatus) {
			case ST_None:
				DrawableImage::Release();
				pMT->Stop();
				m_RestartStatus = ST_Stoped;
				break;
			case ST_Stoped:
				TRACE("restarting movie texture\n");
				pMT->Start();
				m_iNumLoop++;
				m_RestartStatus = ST_Opened;
				break;
			}
			return;
		} else if ((Bool)pMT->isActive) {
			//filter in case this is due to a mediaControl restart, in which case the OD has been closed already
			//if not closed, this is a regular end of stream so deactivate
			if (pMT->GetOD()->IsPlaying() ) {
				//Note: it may happen that the clock is ruled by a media control whose object has a longer duration. 
				//In this case we don't want to deactivate the node before the other object is finished
				MediaControlImp *pCtrl = pMT->GetMediaControl();
				if (!pCtrl || !pCtrl->GetOD()->IsPlaying() )
					Deactivate();
			}
		}
	} else {
		m_RestartStatus = ST_None;
	}


	//not running
	if (m_bFirstFrameFetched && ! (BOOL) (Bool) pMT->isActive) {
		//has been stoped, unregister
		if (m_bNeedsUnregister && m_bIsRegistered) {
			GetPresenter()->UnregisterTimedNode(this);
			m_bNeedsUnregister = FALSE;
			m_bIsRegistered = FALSE;
		}
		return;
	}

	if (!m_pTexture) return;

	BOOL rS, rT;
	GetTextureRepeat(rS, rT);
	m_pTexture->SetTextureRepeat(rS, rT);

	//first frame fetched, set duration
	if (!m_bFirstFrameFetched) {
		/*VRML: "A value of "-1" implies the movie has not yet loaded 
		or the value is unavailable for some reason." by default duration is -1 in IM1*/
		m_cycle = pMT->GetOD()->GetDuration();
		pMT->duration_changed = m_cycle;
		m_bFirstFrameFetched = TRUE;
		
		//if not active close object
		if (!(BOOL) (Bool) pMT->isActive && m_bOpened) {
			m_bOpened = FALSE;
			DrawableImage::StopStream();
			pMT->isActive = FALSE;
		}
	}
	if (m_bSetSpeed && (m_RestartStatus != ST_Opened)) {
		pMT->SetObjectSpeed(pMT->speed);
		m_bSetSpeed = FALSE;
	}
}

BOOL MovieTextureProxy::IsLooping()
{
	MovieTexture *pMT = GetMovieTexture();
	MediaControlImp *pCtrl = pMT->GetObjectMediaControl();
	if (pCtrl) return (BOOL) (Bool) pCtrl->loop;
	return (BOOL) (Bool) pMT->loop;
}

void MovieTextureProxy::Activate()
{
	MovieTexture *pMT = GetMovieTexture();
	MediaControlImp *pCtrl = pMT->GetObjectMediaControl();

	pMT->isActive = TRUE;
	if (pCtrl) {
		m_bRegWithMC = TRUE;
		pCtrl->AddUser();
	}

	//and open
	if (!m_bOpened) {
		pMT->OpenStream();
		m_bOpened = TRUE;
	}
	m_bSetSpeed = TRUE;
	m_iNumLoop = 0;
	GetPresenter()->InvalidateScene(NULL);
}

void MovieTextureProxy::Deactivate()
{
	MovieTexture *pMT = GetMovieTexture();
	MediaControlImp *pCtrl = pMT->GetObjectMediaControl();

	pMT->isActive = FALSE;

	m_bNeedsUnregister = TRUE;
	if (m_bRegWithMC && pCtrl) {
		pCtrl->RemoveUser();
		m_bRegWithMC = FALSE;
	}

	//and stop 
	if (m_bOpened) {
		m_bOpened = FALSE;
		DrawableImage::StopStream();
	}
	GetPresenter()->InvalidateScene(NULL);
}

double MovieTextureProxy::GetSpeed()
{
	double speed;
	MovieTexture *pMT = GetMovieTexture();
	speed = pMT->GetMediaControl() ? pMT->GetMediaControl()->mediaSpeed : m_Speed;
	if (speed<0) speed *= -1;
	return speed;
}


void MovieTextureProxy::UpdateTimeInfo()
{
	MovieTexture *pMT = GetMovieTexture();

	/*not active, store start time and speed*/
	if ( ! (Bool) pMT->isActive) {
		m_startTime = (double) pMT->startTime;
		m_Speed = pMT->speed;
	}

	double currentTime = GetFlexTime() / 1000.;
	// Only proceed if the start time has been reached
	if (currentTime < m_startTime || (m_startTime < 0)) {
		/*opens stream only at first access to fetch first frame, but don't register with control since we're not
		active*/

#if ENABLE_FIRST_FRAME_FETCH
		if (m_bFirstTime) {
			m_bFirstTime = FALSE;
			m_bOpened = TRUE;
			pMT->OpenStream();
			//invalidate scene since we just started a new stream
			GetPresenter()->InvalidateScene(NULL);
		}
#endif
		return;
	}

	// if active and not paused
	if (GetSpeed() && (BOOL) (Bool) pMT->isActive) {
		//if stoptime is reached (>startTime) deactivate	
		if ((pMT->stopTime > m_startTime) && (currentTime >= pMT->stopTime) ) {
			Deactivate();
			return;
		}

		//end of cycle detection is done through end of stream detection (see Render() )
	}

	/*we're (about to be) active: VRML:
	"A time-dependent node is inactive until its startTime is reached. When time now becomes greater than or 
	equal to startTime, an isActive TRUE event is generated and the time-dependent node becomes active 	*/
	assert(m_bIsRegistered);

	// activate
	if (! (BOOL) (Bool) pMT->isActive) {
		Activate();
	}

	//we're playing, invalidate next frame
	GetPresenter()->InvalidateScene(NULL);
}

void MovieTextureINIT::OnUpdate()
{
	((MovieTextureProxy *)GetRenderingObject())->OnUpdate();
}

void MovieTextureProxy::OnUpdate()
{
	MovieTexture *pMT = GetMovieTexture();
	//cf note in TimedNodes.h
	PresenterContext ft_ctx;
	ft_ctx.SetTime(GetFlexTime());
	pMT->startTime.ApplyOffset(&ft_ctx);
	pMT->stopTime.ApplyOffset(&ft_ctx);

	//update state if we're active
	if ((Bool) GetMovieTexture()->isActive) UpdateTimeInfo();

	//re-register node 
	if (!m_bIsRegistered) GetPresenter()->RegisterTimedNode(this);
	m_bIsRegistered = TRUE;
	m_bNeedsUnregister = FALSE;
}
