/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File: Presenter.cpp
//
// Author: Jean Le Feuvre - ENST 2002
//
// Description:
//   The implementation of a class which represents a Presenter,
//   composed of a visual and audio rendering engine.
//
// Created: August 2002
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>
#include "Presenter.h"
#include "Executiv.h"
#include "MediaObj.h"
#include "Caps.h"
#include "Clock.h"
#include "trace.h"
#include "Nodes-INIT.h"
#include "UserEventHandlerProxy.h"
#include "AudioRenderer.h"

#include "FontEngine.h"

#include "App.h"

#ifdef WRITE_TRACE
FILE* traceFile = NULL;
#endif


// Presenter::Presenter
//
// Class constructor
//  - instantiates the visual and audio renderers
Presenter::Presenter(Executive* pExecutive, HWND hWnd)
  : PresenterFlex (pExecutive, hWnd)
{

#ifdef WRITE_TRACE
	traceFile = fopen("\\osmose_trace.txt", "wt");
#endif
	
	//overload node code for rendering proxy creation
	OVERLOAD_NODE(AnchorINIT);
	OVERLOAD_NODE(AnimationStreamINIT);
	OVERLOAD_NODE(AppearanceINIT);
	OVERLOAD_NODE(AudioBufferINIT);
	OVERLOAD_NODE(AudioClipINIT);
	OVERLOAD_NODE(AudioSourceINIT);
	OVERLOAD_NODE(Background2DINIT);
	OVERLOAD_NODE(BitmapINIT);
	OVERLOAD_NODE(CircleINIT);
	OVERLOAD_NODE(ConditionalINIT);
	OVERLOAD_NODE(ColorINIT);
	OVERLOAD_NODE(ColorInterpolatorINIT);
	OVERLOAD_NODE(Coordinate2DINIT);
	OVERLOAD_NODE(Curve2DINIT);
	OVERLOAD_NODE(DiscSensorINIT);
	OVERLOAD_NODE(EllipseINIT);
	OVERLOAD_NODE(FontStyleINIT);
	OVERLOAD_NODE(FormINIT);
	OVERLOAD_NODE(GroupINIT);
	OVERLOAD_NODE(IndexedLineSet2DINIT);
	OVERLOAD_NODE(IndexedFaceSet2DINIT);
	OVERLOAD_NODE(InlineINIT);
	OVERLOAD_NODE(ImageTextureINIT);
	OVERLOAD_NODE(Layer2DINIT);
	OVERLOAD_NODE(LayoutINIT);
	OVERLOAD_NODE(LinePropertiesINIT);
	OVERLOAD_NODE(Material2DINIT);
	OVERLOAD_NODE(MaterialKeyINIT);
	OVERLOAD_NODE(MatteTextureINIT);
	OVERLOAD_NODE(MediaControlINIT);
	OVERLOAD_NODE(MediaSensorINIT);
	OVERLOAD_NODE(MovieTextureINIT);

	OVERLOAD_NODE(SynthesizedTextureINIT);

	OVERLOAD_NODE(OrderedGroupINIT);
	OVERLOAD_NODE(PixelTextureINIT);
	OVERLOAD_NODE(PointSet2DINIT);
	OVERLOAD_NODE(PositionInterpolator2DINIT);
	OVERLOAD_NODE(PlaneSensor2DINIT);
    OVERLOAD_NODE(ProximitySensor2DINIT);
	OVERLOAD_NODE(RectangleINIT);
	OVERLOAD_NODE(ScalarInterpolatorINIT);
	OVERLOAD_NODE(ScriptINIT);
	OVERLOAD_NODE(ShapeINIT);
	OVERLOAD_NODE(Sound2DINIT);
	OVERLOAD_NODE(SwitchINIT);
	OVERLOAD_NODE(TemporalTransformINIT);
	OVERLOAD_NODE(TemporalGroupINIT);
	OVERLOAD_NODE(TextINIT);
	OVERLOAD_NODE(TermCapINIT);
	OVERLOAD_NODE(TimeSensorINIT);
	OVERLOAD_NODE(TouchSensorINIT);
	OVERLOAD_NODE(Transform2DINIT);
	OVERLOAD_NODE(ValuatorINIT);
	OVERLOAD_NODE(WorldInfoINIT);
	OVERLOAD_NODE(CompositeTexture2DINIT);
	OVERLOAD_NODE(TextureTransformINIT);
	
	OVERLOAD_NODE(GradientRadialINIT);
	OVERLOAD_NODE(GradientLinearINIT);

	OVERLOAD_NODE(AdvancedFontStyleINIT);
	OVERLOAD_NODE(TransformMatrix2DINIT);
	OVERLOAD_NODE(XLinePropertiesINIT);
	OVERLOAD_NODE(ViewportINIT);
	OVERLOAD_NODE(Clipper2DINIT);
	OVERLOAD_NODE(XCurve2DINIT);
	OVERLOAD_NODE(ColorTransformINIT);


	m_pScene = NULL;
	m_actualFrameRate = 0;
	m_currentTime = 0;
	m_scenesDrawnSinceStart = 0;
	// The default frame rate is 15 frames per second
	m_frameRate = 15;
	m_newFrameRate = 15;
	m_bUseSplitTracking = FALSE;

	//default value for executive init
	m_sceneUpdatePeriod = 30.0;

	m_iNumMediaSensors = 0;	

	// Remember the pointer to the Executive
	m_pExecutive = pExecutive;

	timeBeginPeriod(1);
 
	// Instantiate the renderers
	m_pVisualRenderer = VisualRenderer::Create(this, (void*)hWnd);
	m_pAudioRenderer = new AudioRenderer(this, (void*)hWnd);

	m_pFontEngine = new FontEngine();
	m_pFontEngine->Init();

	RegisterSurface(m_pVisualRenderer);
}

Presenter::~Presenter()
{
	timeEndPeriod(1);
	 
	UnregisterSurface(m_pVisualRenderer);

	// Delete the visual and audio renderers
	delete m_pVisualRenderer;
	delete m_pAudioRenderer;
	delete m_pFontEngine;
    

#ifdef WRITE_TRACE
	fclose(traceFile);
#endif
}

FontEngine *Presenter::GetFontEngine()
{
	if (m_pFontEngine->IsInit()) return m_pFontEngine;
	return NULL;
}

void Presenter::GetCapabilities (Capabilities* pCaps)
{
	int i, j;
	CapabilitiesImp caps;

	// Get the capabilities of the two renderers
	m_pVisualRenderer->GetCapabilities(pCaps); 
	m_pAudioRenderer->GetCapabilities(&caps);
	
	// Combine the capabilities into a single object
	j = pCaps->GetFormatCapabilityLength();
	for (i = 0; i < caps.GetFormatCapabilityLength(); i++)
		pCaps->AddFormatCapability(caps.GetFormatCapability(i), j++);
}

void Presenter::Init()
{
	m_defEffects.transform.SetIdentity();
	m_defEffects.ResetHandlers();
	m_defEffects.m_ParentGroup=NULL;
	m_defEffects.m_bShouldRebuild=FALSE;
	m_defEffects.m_pColorTransform = NULL;

	// init audio renderer
	m_pAudioRenderer->Init();

	// Call the method in the base class
	PresenterBase::Init();
}

void Presenter :: Start ()
{
	// Remember the pointer to the root scene object
	m_pScene = GetScene();
	int height,width;

	m_LastCursorType = -1;
	m_fZoom = 1.0;
	m_trans_x = m_trans_y = 0;

	m_defEffects.transform.SetIdentity();

	m_bShouldRebuild = TRUE;
	m_bDrawNextFrame = TRUE;
	m_pGrabbedContext = NULL;
	m_pGrabbedNode = NULL;
	m_bTrackingOn = FALSE;

	//get BIFS Author scene size 
	height=((Root *)m_pScene)->GetPixelHeight();
	width=((Root *)m_pScene)->GetPixelWidth();
	//setup our resolution
	m_pVisualRenderer->SetBIFSResolution(width, height, ((Root *)m_pScene)->IsPixelMetric());
	//if we have a size tell the app
	if(height!=0 && width!=0) {
		GetExecutive()->GetApplication()->SetSize(width, height);
	}

	GetExecutive()->GetApplication()->OnConnect(RESPONSE_OK);

	m_scenesDrawnSinceRateChage = 0;

	m_pVisualRenderer->Init();
    
	m_sceneUpdatePeriod = 1000.0 / m_frameRate;
	ResetFrameRateCalculation();
	m_scenesDrawnSinceStart = 0;
	AskScreenAccess();
	m_pVisualRenderer->Clear();
	m_pVisualRenderer->Paint();
	ReleaseScreenAccess();

	/*audio is started after first frame*/
	m_pAudioRenderer->Pause();

	//first run shall indicate full compute
	m_defEffects.m_bShouldRebuild = TRUE;

	m_startingTime = timeGetTime();
	m_currentTime = 0;
	m_bIsPaused = FALSE;
	m_bStepMode = FALSE;

	//force update of frameRate so that the clock granularity of the executive is updated
	m_frameRate = 0;

	PresenterBase::Start();
}

void Presenter::Terminate()
{
	//stop rendering step
	PresenterBase::Terminate();


	// Stop the visual renderer
	m_pVisualRenderer->Terminate();

	//pause audio loop
	m_pAudioRenderer->Pause();
    
	//Added for the correct restart
	m_eventHandlersArray.Clear();

	m_mtxEventArray.P();
	while (m_eventArray.GetLength()) {
		delete m_eventArray[0];
		m_eventArray.DeleteFrom(0);
	}
	m_mtxEventArray.V();

	//clear all time sensor aray
	m_TimedNodeArray.Clear();
	//this is to ensure the next scene will use a correct time offset
	m_currentTime = 0;

	//To be reworked
	GetExecutive()->GetApplication()->OnFinish();
}


void Presenter::Reset()
{
	//this is to ensure the next scene will use a correct time offset
	m_startingTime = timeGetTime();
	m_currentTime = 0;
}

// Add the given TimeSensorProxy node to the list
void Presenter::RegisterTimedNode(TimedNode *pNode)
{
	if (m_TimedNodeArray.Find(pNode) <0) {
		m_TimedNodeArray += pNode;
	}
}


// Remove the given UserEventHandlerProxy node from the
// list of nodes which need to be informed of user events.
void Presenter::UnregisterTimedNode(TimedNode *pNode)
{
	m_TimedNodeArray -= pNode;
}

// Add the given UserEventHandlerProxy node to the list
// of nodes which need to be informed of user events.
void Presenter::RegisterEventHandler(UserEventHandlerProxy* pNode)
{
	// Only add this node if it is not already in the list
	BOOL inList = FALSE;
	for (int i = 0; i < m_eventHandlersArray.GetLength(); i++)
		if (m_eventHandlersArray[i] == pNode)
			inList = TRUE;

	if (!inList)
		m_eventHandlersArray += pNode;

}


// Remove the given UserEventHandlerProxy node from the
// list of nodes which need to be informed of user events.
void Presenter::UnregisterEventHandler(UserEventHandlerProxy* pNode)
{
	m_eventHandlersArray -= pNode;
}

// Called by the executive when a user event occurs.
void Presenter::OnUserEvent(USER_EVENT event, int x, int y, UINT controlKeysStatus)
{
	// Only deal with events when the presenter is active
	if (!IsActive()) return;

/*
	EventInfo *pEvent = new EventInfo(event, x, y, controlKeysStatus);
	GetExecutive () -> LockScene ();
	ExecuteEvent(pEvent);
	delete pEvent;
	GetExecutive () -> UnlockScene ();
	GetExecutive () ->ActivateRoutes();
	return;
*/

	m_mtxEventArray.P();
	m_eventArray += new EventInfo(event, x, y, controlKeysStatus);
	m_mtxEventArray.V();
}


void Presenter::ExecuteEvent(EventInfo* pEvent)
{
	static float prevX, prevY;
	
	if ( pEvent->m_event==EVENT_MOUSE_DOWN) {
		pEvent->m_event=EVENT_MOUSE_DOWN;
	}

	//1 - translate to BIFS coordinates
	float X, Y;
	GetVisualRenderer()->DeviceToSurfaceCoords(pEvent->m_x, pEvent->m_y, &X, &Y);
	pEvent->m_x = X;
	pEvent->m_y = Y;

	// Look through the drawable nodes for a node under this point
	DrawableContext *pContext=NULL;

	int i;
	//2 - locate node

	//SAFETY CHECK: WE CANNOT HANDLE OBJECT INSERTION AT SAME DEPTH LEVEL WHILE DRAGGING
	//are undefined since this pointer may indicate another node
	if (m_bTrackingOn) {
		if (m_pGrabbedContext && m_pGrabbedContext->m_pDrawable!=m_pGrabbedNode) {
			m_bTrackingOn = 0;
			m_pGrabbedContext = NULL;
		}
	}
	
	//tracking is off
	if (!m_bTrackingOn) {
		//store context
		pContext = m_pVisualRenderer->FindNode(pEvent->m_x, pEvent->m_y);
		m_pGrabbedContext = pContext;
		if (pContext) m_pGrabbedNode = pContext->m_pDrawable;
	} else {
		//restore context 
		pContext = m_pGrabbedContext;
	}

	//3- mark all sensors of the context to skip deactivation
	pEvent->m_pContext = pContext;
	if (pContext) {	
		// make sure all sensors of active node are marked
		for (int i=0; i<pContext->m_Handlers.GetLength(); i++) 
			pContext->m_Handlers[i]->m_pHandler->SkipSecondPass = TRUE;

		//also notify the app we're above a sensor
		int type = SENSOR_NONE;
		int code = pContext->m_Handlers[pContext->m_Handlers.GetLength()-1]->m_pHandler->GetObject()->GetNodeCode();

		switch (code) {
		case CODE_Anchor:
			type = SENSOR_ANCHOR;
			break;
		case CODE_PlaneSensor2D:
			type = SENSOR_PLANE;
			break;
		case CODE_DiscSensor:
			type = SENSOR_DISC;
			break;
		case CODE_ProximitySensor2D:
			type = SENSOR_PROXIMITY;
			break;
		case CODE_TouchSensor:
			type = SENSOR_TOUCH;
			break;
		}
		if (type != SENSOR_NONE) {
			if (m_LastCursorType != type) {
				GetExecutive()->GetApplication()->OverSensor(type);
				m_LastCursorType = type;
			}
		}
	}

	if (!pContext && (m_LastCursorType != SENSOR_NONE)) {
		GetExecutive()->GetApplication()->OverSensor(SENSOR_NONE);
		m_LastCursorType = SENSOR_NONE;
	}

	//4 - deactivate all registered nodes
	pEvent->m_pContext = NULL;
	BOOL doIt;
	int count;
	for (i = 0; i < m_eventHandlersArray.GetLength(); i++) {
		doIt = !m_eventHandlersArray[i]->SkipSecondPass;
		//restore BEFORE since the node may unregister itself
		m_eventHandlersArray[i]->SkipSecondPass = FALSE;
		//quick hack to avoid rewriting all user events handling
		count = m_eventHandlersArray.GetLength();
		if (doIt) 
			m_eventHandlersArray[i]->OnUserEvent(pEvent);
		//check if the node has not removed itself. if so don't forget to rewind i
		if (count != m_eventHandlersArray.GetLength()) i-= 1;
	}	
	
	if (!pContext) return;
	pEvent->m_pContext = pContext;
	//5 - send event to all sensors attached to the drawable node
	for (i=pContext->m_Handlers.GetLength(); i>0; i--) {
		pContext->m_Handlers[i-1]->m_pHandler->SkipSecondPass = FALSE;
		pContext->m_Handlers[i-1]->m_pHandler->OnUserEvent(pEvent, &pContext->m_Handlers[i-1]->m_Matrix);
	}
}


void Presenter::ProcessEvents()
{
	m_mtxEventArray.P();

	while (m_eventArray.GetLength() > 0) {
					
		// Get the first event
		EventInfo* pEvent = m_eventArray[0];
		// We have finished with this event, remove it from the array
		m_eventArray.DeleteFrom(0);
	
		ExecuteEvent(pEvent);
		
		delete pEvent;

		//don't process if we're to exit...
		if (!IsActive()) break;
	}
	m_mtxEventArray.V();
}

void Presenter::SetVideoRenderMode(int mode)
{
	m_mtxScreenAccess.P(); 	

	if (mode<=0)	{
		m_VideoMode = 0;
	} else if (mode>=2) {
		m_VideoMode = 2;
	} else {
		m_VideoMode = 1;
	}
	m_mtxScreenAccess.V(); 	

}

// This is a one shot run (prerender, events processing and draw)
//called by the base class
void Presenter::Render()
{
	if (m_bIsPaused) {
		Sleep(m_sceneUpdatePeriod);
		return;
	}
	/*not (re)loaded yet*/
	else if (!GetScene()) {
		if (!m_currentTime) m_startingTime = timeGetTime();
		Sleep(m_sceneUpdatePeriod);
		return;
	}

	DWORD startTimeStep = timeGetTime();

	if (m_bShouldRebuild) m_bDrawNextFrame = TRUE;

	// If the frame rate has changed, reset the FPS related variables (don't reset startTime)
 	if (m_frameRate != m_newFrameRate) {
		m_frameRate = m_newFrameRate;
		m_sceneUpdatePeriod = 1000.0 / m_frameRate;
		m_scenesDrawnSinceRateChage = 0;
		m_bDrawNextFrame = TRUE;
		GetExecutive()->SetClocksGranularity(GetIterationInterval());
	}
	
	
	//should be called in the base class, this is scene-specific flextime
//	PresenterDrivenEvents();

	GetExecutive () -> LockScene ();

	//Process user events on last drawn nodes,and before activating routes of course
	ProcessEvents();

	//activate all previous routes. This must be done BEFORE any processng that may create
	//route otherwise we will mix up route timing and break the cascade model.
	GetExecutive()->ActivateRoutes();

	//force fixed frame rate
	if (m_VideoMode==1) m_bDrawNextFrame = TRUE;
	//cf presenter.h
	else if (m_iNumMediaSensors>0) m_bDrawNextFrame = TRUE;

	//draw frame
	if (m_bDrawNextFrame) {
		//don't redraw next frame unless notified
		m_bDrawNextFrame = FALSE;

		m_mtxScreenAccess.P(); 	

		// Render the scene
		m_defEffects.m_bShouldRebuild = m_bShouldRebuild;
		m_defEffects.m_bUseBoundTracking = m_VideoMode ? TRUE : FALSE;
		m_pVisualRenderer->DrawScene(m_defEffects);	

		//unless notified the graphics settings are unchanged
		m_bShouldRebuild = FALSE;
		
		m_mtxScreenAccess.V();

	}

	//Update the TimeSensor nodes in the scene.
	for (int i=0; i<m_TimedNodeArray.GetLength(); i++) {
		m_TimedNodeArray[i]->UpdateTimeInfo();
	}

	DWORD endTimeStep = timeGetTime();

	//presenter time must be change last before unlocking the scene, otherwise the nodes updateTime
	//will be set to the previous cycle
    m_currentTime = endTimeStep - m_startingTime;
	GetExecutive () -> UnlockScene ();

	m_scenesDrawnSinceStart++;
	if (m_scenesDrawnSinceStart) m_pAudioRenderer->Resume();

	m_scenesDrawnSinceRateChage += 1;

	if (m_bStepMode) {
		Pause();
		m_bStepMode = FALSE;
	}

	UpdateFrameRateCalculation(endTimeStep - m_startingTime);

	/* figure out how long to sleep till next simulation tick so we don't kill the CPU
	sleep till next step if too short */
	endTimeStep -= startTimeStep;
	i=1;
	while (m_sceneUpdatePeriod*i < endTimeStep) i++;
	startTimeStep = m_sceneUpdatePeriod*i - endTimeStep;
	Sleep(startTimeStep);
}

void Presenter::ShouldRebuildGraph()
{
	m_bShouldRebuild = TRUE;
}

void Presenter::Paint(void * /*info*/)
{ 
	m_mtxScreenAccess.P();
	m_pVisualRenderer->Paint();
	m_mtxScreenAccess.V();
}

void Presenter::ResetFrameRateCalculation()
{
	m_nextElapsedTime = NUM_ELAPSED_TIMES - 1;
	for (int i = 0; i < (int)NUM_ELAPSED_TIMES; i++)
		m_elapsedTimes[i] = 0;
	m_actualFrameRate = 0;
}

void Presenter::UpdateFrameRateCalculation(unsigned int displayTime)
{
	// Put the given time into the ring buffer of the most recent
	// NUM_ELAPSED_TIMES times.
	m_nextElapsedTime = (m_nextElapsedTime + 1) % NUM_ELAPSED_TIMES;
	m_elapsedTimes[m_nextElapsedTime] = displayTime;

	// Look back through the recent display times, for a frame which
	// was displayed at least a second (1000 ms) ago.  This is used to
	// determine the average frame rate for the past second.
	int i = m_nextElapsedTime;
	int frames = 0;
	while ((displayTime - m_elapsedTimes[i] < 1000) &&
		   (frames < NUM_ELAPSED_TIMES))
	{
		i = (i == 0) ? (NUM_ELAPSED_TIMES - 1) : (i - 1);
		frames++;
	}

	// And now calculate the frame rate
	if (displayTime == m_elapsedTimes[i])
		m_actualFrameRate = m_frameRate;
	else
		m_actualFrameRate = 1000.0 * (float)frames / (float)(displayTime - m_elapsedTimes[i]);
}

void Presenter::DeleteObject(MediaObject* pObject)
{
	Traversable* pProxy = (Traversable*)(pObject->GetRenderingObject());
	if (pProxy != NULL) delete pProxy;

}


void Presenter::RegisterSurface(VisualSurface *pSurface)
{
	if (!IsSurfaceRegistered(pSurface))
		m_Surfaces += pSurface;
}

void Presenter::UnregisterSurface(VisualSurface *pSurface)
{
	m_Surfaces -= pSurface;
}

BOOL Presenter::IsSurfaceRegistered(VisualSurface *pSurface)
{
	int i = m_Surfaces.Find(pSurface);
	return (i<0) ? FALSE : TRUE;
}

void Presenter::RemoveDrawable(DrawableNode *pNode)
{
	for (int i=0; i<m_Surfaces.GetLength(); i++) {
		m_Surfaces[i]->NodeDeleted(pNode);
	}
	if (m_pGrabbedNode == pNode) {
		SetTracking(FALSE);
	}
	m_bDrawNextFrame = TRUE;
}



void Presenter::InvalidateScene(MediaObject *pObject)
{
	//no need to check ...
	if (m_bDrawNextFrame) return;

	//no object so we must invalidate
	if (!pObject) {
		m_bDrawNextFrame=TRUE;
		return;
	}
	//otherwise filter objects
	Traversable *pTrav = (Traversable *) pObject->GetRenderingObject();
	if (!pTrav) return;

	if (pTrav->ChangesInvalidateScene()) {
		m_bDrawNextFrame = TRUE;
//		TRACE("%s needs redraw (%d)\n", pObject->GetName(), pObject->GetUpdateTime());
		return;
	}
}

void Presenter::SetScalingFactor(float x_factor, float y_factor)
{
	if (!GetExecutive()->GetScene()) return;
	m_mtxScreenAccess.P(); 

	m_ARMatrix.SetIdentity();
	m_ARMatrix.AddScale(x_factor, y_factor, 0.0);
	SetZoom(m_fZoom);
	
	m_mtxScreenAccess.V();
}
void Presenter::SetZoom(float fZoom) 
{
	float ratio;
	if (fZoom<=0) fZoom = 0.001;
	if (fZoom != m_fZoom) {
		ratio = fZoom/m_fZoom;
		m_trans_x *= ratio;
		m_trans_y *= ratio;
		m_fZoom = fZoom;
	}
	m_defEffects.transform.SetIdentity();
	m_defEffects.transform.AddTransform(&m_ARMatrix);
	m_defEffects.transform.AddScale(m_fZoom, m_fZoom, 0);
	m_defEffects.transform.AddTranslation(m_trans_x, -m_trans_y);
	m_bShouldRebuild = TRUE;
}

void Presenter::AddTranslation(int tx, int ty)
{
	float _x, _y;
	if (!m_pScene) return;
	_x = tx;
	_y = ty;
	if (! ((Root *)m_pScene)->IsPixelMetric()) {
		_x /= m_pVisualRenderer->GetScreenWidth();
		_y /= m_pVisualRenderer->GetScreenHeight();
	}
	m_trans_x += _x;
	m_trans_y += _y;
	SetZoom(m_fZoom);
}

void Presenter::ResetTranslation()
{
	m_trans_x = 0;
	m_trans_y = 0;
	SetZoom(m_fZoom);
}

UserEventHandlerProxy::~UserEventHandlerProxy()
{
	GetPresenter()->UnregisterEventHandler(this);
}

static int pauseTime;
void Presenter::Pause()
{
	if (m_bIsPaused) return;
	m_pAudioRenderer->Pause();
	GetExecutive()->Pause();
	pauseTime = timeGetTime();
	m_bIsPaused = TRUE;
}
void Presenter::Resume()
{
	if (!m_bIsPaused) return;
	m_bIsPaused = FALSE;
	/*adjust start time*/
	m_startingTime += timeGetTime() - pauseTime;
	m_pAudioRenderer->Resume();
	GetExecutive()->Resume();
}
void Presenter::Step()
{
	m_bStepMode = TRUE;
	Resume();
}

int Presenter::GetIterationInterval()
{
	//use half the sleep for clock granularity
	return (int) (m_sceneUpdatePeriod/2); 
}
