/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File: Presenter.h
//
// Authors: Jean Le Feuvre ENST 2002
//
///////////////////////////////////////////////////////////////////////////


#ifndef PRESENTER_H
#define PRESENTER_H

#include "../FlexTime/PresenterFlex.h"//PAB+:
#include "ztl.h"
#include "Im1defs.h"


#include "VisualRender.h"

#include "TimedNode.h"
#include "drawable.h"

class AudioRenderer;
class Executive;
class Capabilities;
class MediaObject;
class UserEventHandlerProxy;

class FontEngine;

#define NUM_ELAPSED_TIMES	30

class EventInfo
{
public:
	EventInfo(USER_EVENT event, float x, float y, UINT controlKeysStatus)
	{
		m_event = event;
		m_x = x;
		m_y = y;
		m_controlKeysStatus = controlKeysStatus;
		m_pContext = NULL;
	}
	USER_EVENT m_event;
	float m_x, m_y;
	UINT m_controlKeysStatus;
	//we shall pass a pointer to the current context of the sensor (for local coord).
	//If this pointer is NULL, this means the sensor is NOT under the mouse (deactivation)
	DrawableContext *m_pContext;
};


// The Presenter class
class Presenter : public PresenterFlex
{
public:
	
	Presenter(Executive* pExecutive, HWND hWnd);
	~Presenter();

	//user events
	ZArray<EventInfo*> m_eventArray;

	// Get the capabilities of the video and audio renderers
	void GetCapabilities (Capabilities* pPresentationCaps);

	// Initialise the rendering parameters (top transform and effects)
	void Init();

	//starts scene rendering
	void Start();

	//stops scene rendering
	void Terminate();

	// Return the current presenter time, which is the time since the
	// presenter was started, and this is only updated before a scene
	// is displayed.  The time returned is in milliseconds.
	DWORD GetTime() { return m_currentTime; }


	//Called by the executive when a user event occurs.
	void OnUserEvent(USER_EVENT event, int x, int y, UINT uControlKeysStatus);

	//frame rendering, called by PresBase in loop
	void Render();

	// Pass the paint message from the Executive onto the visual renderer
	void Paint(void *info);

	void InvalidateScene(MediaObject *pObject);
	void Reset();

	// Add the given UserEventHandlerProxy node to the list
	// of nodes which need to be informed of user events.
	void RegisterEventHandler(UserEventHandlerProxy* pNode);

	// Remove the given UserEventHandlerProxy node from the
	// list of nodes which need to be informed of user events.
	void UnregisterEventHandler(UserEventHandlerProxy* pNode);

	//timesensors node registration (this avoids a complete tree browsing for time updates)
	void RegisterTimedNode(TimedNode *pNode);
	void UnregisterTimedNode(TimedNode *pNode);

	//Sets BIFS rendering frame rate 
	void SetTargetFrameRate(float newFrameRate) { if (newFrameRate > 0) m_newFrameRate = newFrameRate; }

	//Set video rendering mode
	void SetVideoRenderMode(int mode);

	// Get the visual renderer's current target frame rate.
	float GetTargetFrameRate() const { return m_newFrameRate; }

	virtual int GetIterationInterval();

	void SetZoom(float fZoom);
	void AddTranslation(int tx, int ty);
	void ResetTranslation();

	// Return the actual frame rate that the visual renderer is producing.
	// This will differ from the target frame rate if there is not enough
	// computing power.  The rate returned will have been obtained by
	// averaging over the last second.
	float GetActualFrameRate() const { return m_actualFrameRate; }

	VisualRenderer *GetVisualRenderer () {return m_pVisualRenderer;}
	AudioRenderer *GetAudioRenderer () {return m_pAudioRenderer;}
	Executive* GetExecutive() {return m_pExecutive;}

	void AskScreenAccess(){m_mtxScreenAccess.P();}
	void ReleaseScreenAccess(){m_mtxScreenAccess.V();}
	DWORD GetScenesDrawn(){return  m_scenesDrawnSinceStart;}

	//called by the renderer when scalable zooming is on. DO NOT call from the app
	void SetScalingFactor(float x_factor, float y_factor);

	//indicates a change in renderer settings that may affect some nodes (currently used for curve2D bezier resolution)
	void ShouldRebuildGraph();
	
	//This is the generic object destroyer, to be called by the destructors of the nodes.
	void DeleteObject(class MediaObject* pObject);

	void RemoveDrawable(DrawableNode *pNode);

	void RegisterSurface(VisualSurface *pSurface);
	void UnregisterSurface(VisualSurface *pSurface);
	BOOL IsSurfaceRegistered(VisualSurface *pSurface);

	// The default rendering effects
	Effects m_defEffects;

	void SetBoundsTrackingMode(BOOL bUseSplitTracking) { m_bUseSplitTracking = bUseSplitTracking; }
	BOOL GetBoundsTrackingMode() { return m_bUseSplitTracking; }

	void SetTracking(BOOL bOn) {m_bTrackingOn = bOn;}
	FontEngine *GetFontEngine();

	/*there is a problem in the case the presentation has a mediaSensor on audio and no other animation:
	the scene is never invalidated thus the MediaSensor::Render never called. We thus force the rendering mode 
	to fixed frame-rate in this case */
	void MediaSensorAdded() { m_iNumMediaSensors++; }
	void MediaSensorRemoved() {if (m_iNumMediaSensors) m_iNumMediaSensors--; }

	void Pause();
	void Resume();
	void Step();

	// The following functions are called by the constructors of the
	// nodes, and are used to create the proxy object associated with
	// every node which is used to render it.
	// All of these functions take a single argument, being a pointer
	// to the node.
	
	class AnchorProxy * InitAnchor(class Anchor*);
	class AnimationStreamProxy * InitAnimationStream(class AnimationStreamImp*);
	class AppearanceProxy * InitAppearance(class Appearance*);
	class AudioBufferProxy * InitAudioBuffer(class AudioBuffer*);
	class AudioClipProxy * InitAudioClip(class AudioClip*);
	class AudioSourceProxy * InitAudioSource(class AudioSource*);
	class Background2DProxy * InitBackground2D(class Background2D*);
	class BitmapProxy *  	InitBitmap(class Bitmap *);
	class CircleProxy		 * InitCircle(class Circle*);
	class CompositeTexture2DProxy *InitCompositeTexture2D(class CompositeTexture2D *);
	class ConditionalProxy * InitConditional(class Conditional *);
	class ColorProxy * InitColor(class Color*);
	class ColorInterpolatorProxy * InitColorInterpolator(class ColorInterpolator*);
	class Coordinate2DProxy * InitCoordinate2D(class Coordinate2D*);
	class CoordinateInterpolator2DProxy * InitCoordinateInterpolator2D(class CoordinateInterpolator2D*);
	class Curve2DProxy * InitCurve2D(class Curve2D*);
	class DiscSensorProxy * InitDiscSensor(class DiscSensor*);
	class FontStyleProxy * InitFontStyle(class FontStyle*);
	class FormProxy * InitForm(class Form*);
	class GroupProxy * InitGroup(class Group*);
	class IndexedLineSet2DProxy * InitIndexedLineSet2D(class IndexedLineSet2D*);
	class IndexedFaceSet2DProxy * InitIndexedFaceSet2D(class IndexedFaceSet2D*);
	class ImageTextureProxy *  	InitImageTexture(class ImageTexture *);
	class LayoutProxy * InitLayout(class Layout*);
	class Layer2DProxy * InitLayer2D(class Layer2D*);
	class InlineProxy * InitInline(class InlineImp*);
	class LinePropertiesProxy * InitLineProperties(class LineProperties*);
	class Material2DProxy * InitMaterial2D(class Material2D*);
	class MaterialKeyProxy * InitMaterialKey(class MaterialKey*);
	class MatteTextureProxy *InitMatteTexture(class MatteTexture *);
	class MediaControlProxy *InitMediaControl(class MediaControlImp *);
	class MediaSensorProxy *InitMediaSensor(class MediaSensorImp *);
	class MovieTextureProxy *  	InitMovieTexture(class MovieTexture *);

	class SynthesizedTextureProxy *InitSynthesizedTexture( class SynthesizedTexture* pObj);


	class OrderedGroupProxy * InitOrderedGroup(class OrderedGroup*);
	class PointSet2DProxy * InitPointSet2D(class PointSet2D*);
	class PixelTextureProxy * InitPixelTexture(class PixelTexture*);
	class PlaneSensor2DProxy * InitPlaneSensor2D(class PlaneSensor2D*);
	class ProximitySensor2DProxy * InitProximitySensor2D(class ProximitySensor2D*);
	class PositionInterpolator2DProxy * InitPositionInterpolator2D(class PositionInterpolator2D*);
	class RectangleProxy * InitRectangle(class Rectangle*);
	class ScriptProxy * InitScript(class ScriptImp*);
	class ScalarInterpolatorProxy * InitScalarInterpolator(class ScalarInterpolator*);
	class ShapeProxy * InitShape(class Shape*);
	class Sound2DProxy * InitSound2D(class Sound2D*);
	class SwitchProxy * InitSwitch(class Switch*);
	class TextProxy * InitText(class Text*);
	class TermCapProxy * InitTermCap(class TermCap*);
	class TimeSensorProxy * InitTimeSensor(class TimeSensor*);
	class TouchSensorProxy * InitTouchSensor(class TouchSensor*);
	class Transform2DProxy * InitTransform2D(class Transform2D*);
	class ValuatorProxy * InitValuator(class Valuator*);
	class WorldInfoProxy * InitWorldInfo(class WorldInfo*);
	class TemporalGroupProxy * InitTemporalGroup(class TemporalGroup*);
	class TemporalTransformProxy * InitTemporalTransform(class TemporalTransform*);
	class TextureTransformProxy *InitTextureTransform(TextureTransform *pObj);

	//v6 nodes (ATG)
	class AdvancedFontStyleProxy * InitAdvancedFontStyle(class XFontStyle *);
	class Clipper2DProxy * InitClipper2D(class Clipper2D*);
	class ColorTransformProxy * InitColorTransform(class ColorTransform *);
	class EllipseProxy *InitEllipse(class Ellipse*); // MBS 031223: changed EllipseNode to Ellipse
	class GradientLinearProxy *InitGradientLinear(class LinearGradient*);
	class GradientRadialProxy *InitGradientRadial(class RadialGradient*);
	class TransformMatrix2DProxy * InitTransformMatrix2D(class TransformMatrix2D*);
	class ViewportProxy * InitViewport(class Viewport *);
	class XLinePropertiesProxy * InitXLineProperties(class XLineProperties *);
	//XCurve2D uses the same proxy as curve2D
	class Curve2DProxy * InitXCurve2D(class XCurve2D*);
    

	// Private implementation details
	//
private:
	void ProcessEvents();

	void ExecuteEvent(EventInfo* pEvent);

	// The output frame rate (frames per second)
	float m_frameRate;

	// This holds the new value of the frame rate until it is made
	// the current frame rate.
	float m_newFrameRate;

	// This stores the actual frame rate as calculated by the main loop
	// of the presenter.
	float m_actualFrameRate;

	// The time that has elapsed since the presenter was started.
	DWORD m_currentTime;

	// Reset the frame rate calculation variables
	void ResetFrameRateCalculation();

	// Update the frame rate calculation, by giving the next time at
	// which a frame has been presented.
	void UpdateFrameRateCalculation(unsigned int nextTime);

	// Data members pertaining to the calculation of the actual frame
	// rate.
	unsigned int m_nextElapsedTime;
	unsigned int m_elapsedTimes[NUM_ELAPSED_TIMES];
    
	
	// The array of events that need to be processed
	ZMutex m_mtxEventArray;

	// The array of nodes to be notified of user events
	ZArray<UserEventHandlerProxy*> m_eventHandlersArray;
	ZArray<TimedNode *> m_TimedNodeArray;

	ZArray<VisualSurface *> m_Surfaces;

	// Other variables relating to the timing of the presenter
	unsigned long m_scenesDrawnSinceRateChage;
	unsigned long m_startingTime;
	float m_sceneUpdatePeriod;
	DWORD m_scenesDrawnSinceStart;

	VisualRenderer *m_pVisualRenderer;
	AudioRenderer *m_pAudioRenderer;
	Executive *m_pExecutive;
	MediaObject *m_pScene;
	ZMutex m_mtxScreenAccess;

	DrawableContext *m_pGrabbedContext;
	DrawableNode *m_pGrabbedNode;

	int m_LastCursorType;
	BOOL m_bShouldRebuild;

	//if set to TRUE next frame is redrawn
	BOOL m_bDrawNextFrame;
	int m_VideoMode;
	
	//zoom factor - 0 means no zoom (eg 1)
	float m_fZoom;
	float m_trans_x, m_trans_y;
	Matrix2D m_ARMatrix;

	BOOL m_bTrackingOn;
	BOOL m_bUseSplitTracking;

	FontEngine *m_pFontEngine;

	BOOL m_bStepMode, m_bIsPaused;
	int m_iNumMediaSensors;
};


#endif // PRESENTER_H
