/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  TimeSensorProxy.cpp
//
// Author: Jean Le Feuvre - ENST 2002
//
// Description:
//   proxy object for TimeSensor node.
//
// Created: August 2002.
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-INIT.h"
#include "TimeSensorProxy.h"
#include <math.h>
#include "stream.h"
#include "Presenter.h"

TimeSensorProxy* Presenter::InitTimeSensor(TimeSensor* pObj)
{
	return new  TimeSensorProxy(pObj); 
}

TimeSensorProxy::TimeSensorProxy(MediaObject* pMO) : Traversable(pMO)
{
	m_NumCycles = 0;
	m_StartTime = -1;
	m_CycleInterval = 0.0;
	m_bNeedsUnregister = FALSE;

	if (pMO->GetID() >= 0) {
		GetPresenter()->RegisterTimedNode(this);
		m_bIsRegistered = TRUE;
	} else {
		m_bIsRegistered = FALSE;
	}
}

TimeSensorProxy::~TimeSensorProxy()
{
	if(m_bIsRegistered) GetPresenter()->UnregisterTimedNode(this);
}

void TimeSensorProxy::Render(Effects *effects)
{

	if (m_bIsRegistered && m_bNeedsUnregister) {
		((Presenter*)GetPresenter())->UnregisterTimedNode(this);
		m_bIsRegistered = FALSE;
		m_bNeedsUnregister = FALSE;
		assert(! (Bool)GetTimeSensor()->isActive);
	}
}


void TimeSensorProxy::Deactivate()
{
	TimeSensor *pTS = GetTimeSensor();
	pTS->isActive = FALSE;
	m_bNeedsUnregister = TRUE;
	m_NumCycles = 0;
	TRACE("deactivating %d\n", pTS->GetID());
}


void TimeSensorProxy::UpdateTimeInfo()
{

	TimeSensor *pTS = GetTimeSensor();

	double currentTime = GetFlexTime() / 1000.0;
	if (! (Bool) pTS->enabled) {
		//to be turned off
		if ((Bool) pTS->isActive) {
			pTS->time = currentTime;
			Deactivate();
		}
		return;
	}

	//not activated, store timing
	if (! (Bool) pTS->isActive) {
		m_StartTime = pTS->startTime;
		m_CycleInterval = pTS->cycleInterval;
		if (!m_CycleInterval) m_CycleInterval = FLT_MAX;
	}
	
	// Only proceed if the start time has been reached
	if ((currentTime < m_StartTime) || (m_StartTime<0)) return;

	//time since start of current cycle
	double cycleTime = currentTime - m_StartTime - (m_NumCycles) * m_CycleInterval;
	// Update the fraction_changed eventOut field,
	double newFraction = fmod(cycleTime, m_CycleInterval);
	newFraction /= m_CycleInterval;

	//sensor is active
	if ((Bool) pTS->isActive) {
		// Update the time eventOut field
		pTS->time = currentTime;

		/*VRML:
			"f = fmod( (now - startTime) , cycleInterval) / cycleInterval
			if (f == 0.0 && now > startTime) fraction_changed = 1.0
			else fraction_changed = f"
		*/
		if (!newFraction && (currentTime > m_StartTime) ) newFraction = 1.0;

		assert(newFraction >= 0);
		/*check for deactivation*/
		
		//if stoptime is reached (>startTime) deactivate	
		if ((pTS->stopTime > m_StartTime) && (currentTime >= pTS->stopTime) ) {
			//set fraction to current
			assert(newFraction <= 1.0);
			pTS->fraction_changed = newFraction;
			Deactivate();
			return;
		}
		//if current cycle is over (no loop), deactivate
		if (!(BOOL) (Bool) pTS->loop) {
			if (cycleTime >= m_CycleInterval) {
				/*set fraction to end of cycle*/
				pTS->fraction_changed = 1.0;
				Deactivate();
				return;
			}
		}
		//set fraction to current
		assert(newFraction <= 1.0);
		pTS->fraction_changed = newFraction;
	}
	/*we're (about to be) active: VRML:
	"A time-dependent node is inactive until its startTime is reached. When time now becomes greater than or 
	equal to startTime, an isActive TRUE event is generated and the time-dependent node becomes active 	*/

	if (!(Bool) pTS->isActive) {
		m_bNeedsUnregister = FALSE;
		pTS->isActive = TRUE;
		pTS->time = currentTime;
		pTS->cycleTime = currentTime;
		assert(newFraction <= 1.0);
		pTS->fraction_changed = newFraction;
		TRACE("activating %d\n", pTS->GetID());
	}

	//compute cycle time
	if ((BOOL) (Bool) pTS->loop && (cycleTime >= m_CycleInterval) ) {
		int inc = 1 + (cycleTime - m_CycleInterval) / m_CycleInterval;
		m_NumCycles += inc;
		cycleTime -= inc*m_CycleInterval;
		//set cycleTime to theoretical value at begining of cycle
		pTS->cycleTime = currentTime - cycleTime;
	}
}



//overload of OnUpdate for timed nodes in order to reset or not the sensor
void TimeSensorINIT::OnUpdate () 
{
	((TimeSensorProxy *) GetRenderingObject())->OnUpdate();
}

void TimeSensorProxy::OnUpdate()
{
	TimeSensor *pTS = GetTimeSensor();
	//cf note in TimedNodes.h
	PresenterContext ft_ctx;
	ft_ctx.SetTime(GetFlexTime());
	pTS->startTime.ApplyOffset(&ft_ctx);
	pTS->stopTime.ApplyOffset(&ft_ctx);


	//if we're active, update time info in case enabled/stopTime has been changed
	if ((Bool) pTS->isActive) UpdateTimeInfo();

	//if sensor is inactive, modified values will be updated at next UpdateTimeInfo

	//if enabled, check if we need to register. the rest will be done in UpdateTimeInfo
	if ((Bool) pTS->enabled) {
		m_bNeedsUnregister = FALSE;
		if (!m_bIsRegistered) {
			((Presenter*)GetPresenter())->RegisterTimedNode(this);
			m_bIsRegistered = TRUE;
		}
	}
}