/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  Traversable.h
//
// Author: Jean Le Feuvre - ENST 2002
//
//
///////////////////////////////////////////////////////////////////////////


#ifndef TRAVERSABLE_H
#define TRAVERSABLE_H


#include "MediaObj.h"
#include "Executiv.h"
#include "VisualClipper.h"
#include "ztypes.h"
#include "RendObj.h"
#include "PresBase.h"
#include "ColorMatrix.h"

class Presenter;
class AudioRenderer;
class EventInfo;
class UserEventHandlerProxy;


//this class is used to store sensor context
class SensorContext
{
public:
	SensorContext(UserEventHandlerProxy *pHandler, Matrix2D *matrix)
	{
		m_pHandler = pHandler;
		if (matrix) m_Matrix.Set(*matrix);
	}
	UserEventHandlerProxy *m_pHandler;
	Matrix2D m_Matrix;
};

// This class is used to store all of the attributes which effect how
// a node in the scene graph can be displayed.
class Effects
{
public:
	Effects();
	Effects(Effects *pEffects);
	~Effects(); 

	Matrix2D transform;
	
	//set to TRUE to force background drawing otherwise background
	//is skipped (normal use). This is set at the VisualSurface level only
	BOOL m_bDrawBackground;

	//indicates a change in renderer settings (speed, curve resolution)
	//set at the top level and untouched during all prerender
	//current nodes affected are Curve2D (bezier reolution)
	BOOL m_bShouldRebuild;
	
	//if set to true, the renderer only attempts to redraw the minimal
	//amount of the scene. Otherwise the full scene is redrawn. This shall be set at the
	//top-level effects since it affects the whole scene graph rendering
	BOOL m_bUseBoundTracking;
	
	//set to indicate all nodes under these effects shall be redrawn
	BOOL m_bInvalidateAll;

	//current shape info is propagated here
	class AppearanceProxy *m_pAppearanceProxy;
	class GeometryNode *m_pGeometryProxy;

	//this is used for geometry prerender in order to distinguish between real objects
	//and clipping objects
	//0: not a clipper, 1: clipper.inside = FALSE, 2: clipper.inside=TRUE
	unsigned int m_ClipperMode;

	//current clipper
	VisualClipper *m_pClipper;

	//current drawing surface (changed at CompositeTexture Prerender steps)
	class VisualSurface *m_pSurface;
	
	//the eventual special node (layout,form) who affects the 
	//rendering of his children
	class GroupNodeProxy *m_ParentGroup;
	ColorMatrix *m_pColorTransform;


	void AddUserEventHandler(UserEventHandlerProxy *handler, Matrix2D *matrix);
	void RemoveHandler(UserEventHandlerProxy *handler);
	void ResetHandlers();

	ZArray<SensorContext *> m_Handlers;

	BOOL m_isSwitchedOff;

	//Audio parent
	class AudioGroup *m_pAudioParent;
};


class Traversable : public RenderingObject
{
	//
	// Public functions
	//
public:

	Traversable(MediaObject* pMO);
	virtual ~Traversable() {}

	//the rendering fonction, called once and traverse all scene graph
	//depending on the rendering mode it will draw or just perform pre-render (bounds,
	//effects & appearance, texture updates)
	virtual void Render(Effects *effects) = 0;
	
	//returns associated node
	MediaObject* GetObject() const { return m_pMediaObject; }
	
	// Return TRUE if this is a TouchSensor node (which will handle
	// the given event), otherwise return FALSE.  The default
	// implementation returns FALSE.This is called in the prerendering
	// step, and the pointer of the sensor is then stored in the drawable
	// context if TRUE is returned.
	virtual BOOL IsSensor() { return FALSE; }

	//returns wether modification to this node triggers scene redraw or not
	//default is true; Nodes which acts as routers (valuator, interpolators)
	//shouldn't return TRUE
	virtual BOOL ChangesInvalidateScene() { return TRUE; }

	//placeholder for flextime - ignored for now
	DWORD GetFlexTime(PresenterContext *flexTimeContext = NULL);

	//
	// Protected implementation details
	//
protected:

	// Return a pointer to the presenter
	Presenter* GetPresenter() const { return m_pPresenter; }

	//
	// Private implementation details
	//
private:

	// A pointer to the MediaObject derived object that this
	// object is attached to.
	MediaObject* m_pMediaObject;

	// Remember the presenter that this object can use.
	Presenter* m_pPresenter;
};


#endif // TRAVERSABLE_H
