/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File: VisualPath.cpp
//
// Author: Jean Le Feuvre - ENST 2002
//
// Description:
//		Object used to store visual path (curves, points, lines, Text)
//
// Created: August 2002
//
///////////////////////////////////////////////////////////////////////////

#include "defs.h"
#include "VisualPath.h"
#include <math.h>
//for line styles
#include "VisualRender.h"



VisualPath::VisualPath()
{
	m_pFont = NULL;
	Reset(2);
}

void VisualPath::Reset(unsigned int quality_level)
{
	m_Path.Reset();
	if (m_pFont) delete m_pFont;
	m_pFont = NULL;
	m_FontStyle = 0;
	m_BezierResolution = quality_level;
	m_bHasText = FALSE;
	m_pFE = NULL;
}

void VisualPath::AddCircle(float radius)
{
	m_Path.AddEllipse(-radius, -radius, 2*radius, 2*radius);
}

void VisualPath::AddEllipse(float a_axis, float b_axis)
{
	m_Path.AddEllipse(-a_axis/2, -b_axis / 2, a_axis, b_axis);
}

void VisualPath::AddRectangle(float width, float height)
{
	Gdiplus::RectF rec(-width/2, -height/2, width, height);
	m_Path.AddRectangle(rec);
}

void VisualPath::AddRectangle(float x, float y, float width, float height)
{

	Gdiplus::RectF rec(0, 0, width, height);
	Gdiplus::GraphicsPath path;
	path.AddRectangle(rec);
	Gdiplus::Matrix mat;
	mat.Translate(x, y-height);
	path.Transform(&mat);
	m_Path.AddPath(&path, TRUE);
}

void VisualPath::AddLine(float startX, float startY, float endX, float endY)
{
	m_Path.AddLine(startX, startY, endX, endY);
}

//a "decent" N bezier algorithm since GDI only support 3rd order curves
//NB: doesn't compute the last point (mu=1.0)
static Gdiplus::PointF NBezier(PointF *points, int n, double mu)
{
	int k,kn,nn,nkn;
	double blend,muk,munk;
	Gdiplus::PointF b(0.0, 0.0);
	muk = 1;
	munk = pow(1-mu,(double)n);
	for (k=0;k<=n;k++) {
		nn = n;
		kn = k;
		nkn = n - k;
		blend = muk * munk;
		muk *= mu;
		munk /= (1-mu);
		while (nn >= 1) {
			blend *= nn;
			nn--;
			if (kn > 1) {
				blend /= (double)kn;
				kn--;
			}
			if (nkn > 1) {
				blend /= (double)nkn;
				nkn--;
			}
		}
		b.X += points[k].x * blend;
		b.Y += points[k].y * blend;
	}
	return (b);
}

void VisualPath::AddBezier(PointF *points, int n, float fineness)
{
	Gdiplus::PointF start, end;
	double mu;
	int numPoints;

	if (n<2) return;
	if (fineness < 0.0) fineness = 0.0;
	if (fineness > 1.0) fineness = 1.0;

	//compute number of points based on fineness
	numPoints = m_BezierResolution * fineness;
	//compute sampling interval
	mu = 0.0;
	if (numPoints) mu = 1/(double)numPoints;

	start.X = points[0].x;
	start.Y = points[0].y;
	//don't do first and last points
	for (int i=1; i<numPoints; i++) {
		end = NBezier(points, n-1, i*mu);
		m_Path.AddLine(start, end);
		start = end;
	}
	//draw last one
	end.X = points[n-1].x;
	end.Y = points[n-1].y;
	m_Path.AddLine(start, end);
}

//Note: we may get bad authored arcs were the 2 points don't belong to the same ellipse. 
//this is undefined behaviour. We therefore don't check the final point, only compute the ellipse
//based on start and focals, and use the final point to compute the sweep angle, which means that the 
//final ending point may be different from the specified one
void VisualPath::AddArc(PointF start, PointF end, PointF focal_a, PointF focal_b, BOOL ClockWise)
{
	Gdiplus::PointF C, A, E, Fa, Fb;
	Gdiplus::Matrix mat, inv;
	double angle, start_angle, end_angle, sweep;
	float axis_w, axis_h, tmp;
	//get the center
	C.X = (focal_b.x + focal_a.x)/2;
	C.Y = (focal_b.y + focal_a.y)/2;
	//get the matrix
	angle = atan2(focal_b.y-focal_a.y, focal_b.x-focal_a.x);
	mat.Rotate(angle*180/PI);
	mat.Translate(C.X, C.Y, Gdiplus::MatrixOrderAppend);

	A.X = start.x;
	A.Y = start.y;
	E.X = end.x;
	E.Y = end.y;
	Fa.X = focal_a.x;
	Fa.Y = focal_a.y;
	Fb.X = focal_b.x;
	Fb.Y = focal_b.y;
	//make our ellipse centered and // to X axis
	mat.Invert();
	mat.TransformPoints(&A, 1);
	mat.TransformPoints(&E, 1);
	mat.TransformPoints(&Fa, 1);
	mat.TransformPoints(&Fb, 1);

	//start angle and end angle
	start_angle = atan2(A.Y, A.X);
	end_angle = atan2(E.Y, E.X);
	//get axis size
	//r1+r2 = 2a;
	tmp = (A.X - Fa.X) * (A.X - Fa.X) + (A.Y - Fa.Y) * (A.Y - Fa.Y); 
	axis_w = sqrt(tmp);
	tmp = (A.X - Fb.X) * (A.X - Fb.X) + (A.Y - Fb.Y) * (A.Y - Fb.Y); 
	axis_w += sqrt(tmp);
	axis_w /= 2;
	//b2 = a2 - c2;
	axis_h = sqrt(axis_w*axis_w - Fa.X*Fa.X);

	sweep = end_angle-start_angle;

	//gdiplus draws arc anti-clockwise
	if (ClockWise) {
		if (sweep>0) sweep -= 2*PI;
	} else {
		if (sweep<0) sweep += 2*PI;
	}

	//start a sub-graph for local transform
	Gdiplus::GraphicsPath subPath;
	subPath.AddArc(-axis_w, -axis_h, 2*axis_w, 2*axis_h, start_angle*180/PI, sweep*180/PI);
	//reinvers our transform
	mat.Invert();
	subPath.Transform(&mat);
	//add the new path and specify it is part of the current path
	m_Path.AddPath(&subPath, TRUE);
}

void VisualPath::AddPoint(float xPos, float yPos)
{
	//NB: no "point" in GDI plus so we draw a quite short line ...
	m_Path.AddLine(xPos, yPos, xPos+0.005, yPos+0.005);
}

void VisualPath::StartNewFigure()
{
	m_Path.StartFigure();
}
	//closes current figure in the path
void VisualPath::CloseFigure()
{
	m_Path.CloseFigure();
}

BOOL VisualPath::IsPointInPath(float x, float y, Matrix2D *matrix)
{
	Gdiplus::PointF pt(x, y);
	if (matrix) {
		Gdiplus::Matrix *mat = matrix->m_Matrix.Clone();
		mat->Invert();
		mat->TransformPoints(&pt, 1);
		delete mat;
	}
	return m_Path.IsVisible(pt, NULL);
}

#define SECURITY_BOUNDS		1

//NOTE: using Gdiplus pen for BRect compute gives a much larger cliper than needed, so we do it by hand
void VisualPath::GetBoundingRect(BRect *pClipBRect, Matrix2D *matrix, float pen_width, BOOL pen_is_scaled)
{
	Gdiplus::RectF rect;
	Gdiplus::Matrix *mat;
	Gdiplus::Pen *pen;

	if (!pClipBRect) return;
	pClipBRect->MakeEmpty();

	mat = NULL;
	if (matrix) {
		mat = matrix->m_Matrix.Clone();
	} else {
		mat = new Gdiplus::Matrix;
	}
	//revert our coord system
	mat->Scale(1, -1, Gdiplus::MatrixOrderAppend);

	pen = NULL;
	if (pen_width && pen_is_scaled) {
		pen = new Gdiplus::Pen(Gdiplus::Color(255, 255, 255), pen_width);
		if (pen_is_scaled) pen->SetTransform(&matrix->m_Matrix);
	}
	m_Path.GetBounds(&rect, mat, pen);

	float x, y, w, h;
	//revert our coord system back
	x = rect.X;
	y = -rect.Y;
	w = rect.Width;
	h = rect.Height;

	//get max pen size once scaled (this is needed because of scalable zooming)
	if (pen_width && !pen_is_scaled) {
		if (matrix) {
			float m[6];
			matrix->GetElements(m);
			if (m[0]<0) m[0]*=-1;
			if (m[4]<0) m[4]*=-1;
			pen_width *= max(m[0], m[4]);
		}
		x -= SECURITY_BOUNDS*pen_width;
		y += SECURITY_BOUNDS*pen_width;
		w += 2*SECURITY_BOUNDS*pen_width;
		h += 2*SECURITY_BOUNDS*pen_width;
	}

	pClipBRect->SetDimensions(x, y, w, h);
	if (mat) delete mat;
	if (pen) delete pen;
}


void VisualPath::Transform(Matrix2D *mat)
{
	if (!mat) return;
	m_Path.Transform(& mat->m_Matrix);
}

VisualPath *VisualPath::Clone()
{
	VisualPath *clone = new VisualPath();
	clone->Reset(m_BezierResolution);
	clone->m_bHasText = m_bHasText;
	clone->m_Path.AddPath(&m_Path, TRUE);
	return clone;
}

