/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File: VisRend-ENST-GDIP.cpp
//
// Author: Jean Le Feuvre
//
// Description:
//   The implementation of a class which performs visual rendering using
//   DirectDraw and GDIplus
//
// Created: August, 2002
//
// Note: GDI+ handles coord in the following fashion:
//			origin (0,0) is the top-left corner of the client area
//			Y axis increments from top to bottom
//
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include <assert.h>


#include "VisualRender-GDIP.h"
#include "Caps.h"
#include "Executiv.h"
#include "Presenter.h"
#include "Clock.h"
#include "dx-errors.h"
#include "ColorTools.h"

#include <stdio.h>

#include "VisualTexture.h"

//this is for the MPEG logo resource, otherwise DON'T include any files from the GUI itself
//in the renderer
#include "resource.h"

#include "App.h"


#include "trace.h"

#define BEZIER_DEFAUT_POINTS	90


VisualRendererGDI::VisualRendererGDI(Presenter* pPresenter, HWND hWnd)
{
	m_initOK = TRUE;

	m_hWnd = hWnd;
	m_pPresenter = pPresenter;

	m_BezierResolution = BEZIER_DEFAUT_POINTS;
	m_bNoFill = FALSE;

    //default size  just to init DirectDraw for Windowed Application the a
	//resize will be called.
	RECT WindowSize;
	GetClientRect(hWnd,&WindowSize);
	
	m_bIsClearing = FALSE;
	m_bKeepRatio = TRUE;
	m_bScalableZoom = FALSE;

	m_bIsActive = FALSE;
	m_bUseYUVHW = FALSE;
	m_bUsePixelMetrics = TRUE;
	m_bHasSize = TRUE;
	m_BWidth = m_origBWidth  = WindowSize.right;
	m_BHeight = m_origBHeight = WindowSize.bottom;

	// Initialise the DirectDraw routines
	m_pDirectDraw = new DirectDraw;
	const char* errorString = m_pDirectDraw->Init(hWnd, m_BWidth, m_BHeight);
	
	// If an error has occurred, display a message and then exit the
	// application
	if (errorString != NULL) {

		GetPresenter()->GetExecutive()->GetApplication()->OnFatalError(VIDEO_INIT_FAILED, errorString);		
		m_initOK = FALSE;
	}
	//setup rect
	UpdateScreenSizeInfo();
	//check YUV hardware
	m_pDirectDraw->CheckYUVHardwareSupport();
	m_bUseYUVHW = m_pDirectDraw->m_numFormats ? TRUE : FALSE;

	// Erase the output window
	if (m_initOK) Clear();

	//default is speed priority
	m_bHighSpeed = FALSE;
	m_bAntiAlias = FALSE;

	m_CaptureMode = VIDEO_CAPTURE_STOP;
}

VisualRendererGDI::~VisualRendererGDI()
{
	delete m_pDirectDraw;
}

void VisualRendererGDI::Init()
{
	VisualRenderer::Init();
	m_bIsActive = TRUE;
}


void VisualRendererGDI::Terminate()
{
	VisualRenderer::Terminate();

	m_bIsActive = FALSE;
	if (m_clearOnEnd) {
		Paint();
	}
}

// Get the capabilities of this visual renderer.
void VisualRendererGDI::GetCapabilities (Capabilities* pCaps)
{	
	//prefered formats in priority order (RGB is first because of texture mapping)
	pCaps->AddFormatCapability(MediaFormat (MediaFormat :: RGB), 0);
	pCaps->AddFormatCapability(MediaFormat (MediaFormat :: YUV), 1);
	pCaps->AddFormatCapability(MediaFormat (MediaFormat :: YUVA), 2);
	pCaps->AddFormatCapability(MediaFormat (MediaFormat :: RGBA), 3);
	pCaps->AddFormatCapability(MediaFormat (MediaFormat :: BGR), 4);
}

//copy the offscreen surface to the screen. we redraw the full screen because:
//1 - float to pixel conversion  will introduce some overhead, if not some inaccuracy
//2 - we actually don't know what Gdiplus is doing in terms of float->pixel
//conversion, thus it is better to blit the full DC at a time
//3 - since this is just a simple blit it shouldn't take too long anyway
//4 - it allows to visually make sure the offscreen surface is always consistent
void VisualRendererGDI::Paint()
{
	RECT destRect;
	HRESULT returnValue;

	if (!IsWindow(m_hWnd)) return;
	if (!m_origBWidth || !m_origBHeight) return;
	if (!m_outputWidth || !m_outputHeight) return;
	

	//update output location
	UpdateDisplayRect();
	memcpy(&destRect, &m_OutputRect, sizeof(RECT));

	if (! m_bIsActive && m_clearOnEnd)
		DisplayLogo();

	returnValue = m_pDirectDraw->RenderSurface(&destRect);  
	if (returnValue!=DD_OK) PostError(returnValue);	

	CaptureFrame();
}

// Completely erase the specifed region of the offscreen drawing
// area, but do not display the surface.  If the given rectangle
// is NULL, erase the entire offscreen area.
void VisualRendererGDI::Clear(BRect const* pBRect)
{
	HRESULT res;

	if (pBRect == NULL) {
		res = m_pDirectDraw->EraseSurface(NULL);
		if (res!=DD_OK) PostError(res);
	} else {
		VisualPath path;
		path.AddRectangle(pBRect->left, pBRect->top, pBRect->width, pBRect->height);
		ResetStyles();
		SetBrushStyle(1, Color3f(0.0, 0.0, 0.0), 0.0);
		m_bIsClearing = TRUE;
		DrawVisualPath(&path, TRUE);
		m_bIsClearing = FALSE;
	}
}

//inits the Gdiplus::Graphics object or return false otherwise
BOOL VisualRendererGDI::InitGraphRender(BOOL bForTextDraw) 
{
	HRESULT res;

	res = m_pDirectDraw->GetHDC(m_hDC);
	if (res != DD_OK) return FALSE;

	m_pGraph = new Gdiplus::Graphics(m_hDC);

	//set rendering options
	if (m_bIsClearing) {
		//don't waste time when clearing
		m_pGraph->SetCompositingQuality(Gdiplus::CompositingQualityHighSpeed);
		m_pGraph->SetSmoothingMode(Gdiplus::SmoothingModeHighSpeed);
	} else {
		m_pGraph->SetCompositingQuality(m_bHighSpeed ? Gdiplus::CompositingQualityHighSpeed : Gdiplus::CompositingQualityHighQuality);
		
		if (bForTextDraw && m_bTextAntiAlias) {
			m_pGraph->SetSmoothingMode(Gdiplus::SmoothingModeHighQuality);
			m_pGraph->SetTextRenderingHint(Gdiplus::TextRenderingHintAntiAlias);
		} else {
			m_pGraph->SetSmoothingMode(m_bAntiAlias ? Gdiplus::SmoothingModeHighQuality : Gdiplus::SmoothingModeHighSpeed);

			if (m_bAntiAlias) {
				m_pGraph->SetTextRenderingHint(Gdiplus::TextRenderingHintAntiAlias);
			} else if (m_bHighSpeed) {
				m_pGraph->SetTextRenderingHint(Gdiplus::TextRenderingHintSingleBitPerPixel);
			} else {
				m_pGraph->SetTextRenderingHint(Gdiplus::TextRenderingHintSingleBitPerPixelGridFit);
			}
		}
		m_pGraph->SetInterpolationMode(m_bHighSpeed ? Gdiplus::InterpolationModeLowQuality : Gdiplus::InterpolationModeHighQualityBicubic);
	}

	Gdiplus::Matrix final;
	final.Scale(1, -1, Gdiplus::MatrixOrderAppend);
	final.Translate(m_BWidth/2, m_BHeight/2, Gdiplus::MatrixOrderAppend);
	m_pGraph->SetTransform(&final);

	return TRUE;
}

void VisualRendererGDI::ResetAntialiasing()
{
	if (!m_pGraph) return;
	m_pGraph->SetSmoothingMode(Gdiplus::SmoothingModeHighSpeed);
}


//releases the Gdiplus::Graphics object
void VisualRendererGDI::ReleaseGraphRender()
{
	if (!m_pGraph) return;
	delete m_pGraph;
	m_pGraph = NULL;
	m_pDirectDraw->ReleaseHDC(m_hDC);
}

//computes default scaling of all objects - this is needed in case we work with MeterMetrics
BOOL VisualRendererGDI::InitObjectMatrix(Matrix2D &mat)
{
	Gdiplus::Matrix final;
	mat.SetIdentity();
	if (!m_bUsePixelMetrics)
		mat.AddScale(m_origBWidth/2, m_origBWidth/2, 0.0);

	return m_bUsePixelMetrics;
}

/*
	TODO: fix bug when resizing screen (crashes)
*/
void VisualRendererGDI::DrawBitmap(BRect *clippedDest, BRect *dest, VisualTexture *texture, MaterialKey *mk)
{
	MediaFormat format = texture->m_OriginalFormat;
	unsigned int stride = texture->m_hstride;
	int BPP = 3;
	BRect clipped_final = *clippedDest;
	BRect final = *dest;

	Matrix2D mat;
	InitObjectMatrix(mat);
	clipped_final.width += clipped_final.left;
	clipped_final.height += clipped_final.top;
	mat.Apply(clipped_final.left, clipped_final.top);
	mat.Apply(clipped_final.width, clipped_final.height);
	clipped_final.width -= clipped_final.left;
	clipped_final.height -= clipped_final.top;

	final.width += final.left;
	final.height += final.top;
	mat.Apply(final.left, final.top);
	mat.Apply(final.width, final.height);
	final.width -= final.left;
	final.height -= final.top;
	
	float w_scale = final.width / texture->m_Width;
	float h_scale = final.height / texture->m_Height;

	//switch to device coord
	clipped_final.top = m_BHeight / 2 - clipped_final.top;
	clipped_final.left += m_BWidth / 2;
	final.top = m_BHeight / 2 - final.top;
	final.left += m_BWidth / 2;
	//make sure we lie in the final rect
	if (clipped_final.left<0) {
		clipped_final.width += clipped_final.left;
		clipped_final.left = 0;
	}
	if (clipped_final.top<0) {
		clipped_final.height += clipped_final.top;
		clipped_final.top = 0;
	}
	if (clipped_final.left + clipped_final.width > m_BWidth) {
		clipped_final.width = m_BWidth - clipped_final.left;
		clipped_final.left = m_BWidth - clipped_final.width;
	}
	if (clipped_final.top + clipped_final.height > m_BHeight) {
		clipped_final.height = m_BHeight - clipped_final.top;
		clipped_final.top = m_BHeight - clipped_final.height;
	}

	int start_x = (clipped_final.left - final.left) / w_scale;
	int start_y = (clipped_final.top - final.top) / h_scale;
	final = clipped_final;

	RECT rc;

	//now make sure we're performing the right rounding
	rc.left = final.left;
	if ((double)rc.left != (double)final.left) rc.left += 1;
	rc.top = final.top;
	if ((double)rc.top != (double)final.top) rc.top += 1;
	rc.right = final.left + final.width;
	if ((double) rc.right != (double) (final.left + final.width)) rc.right +=1;
	rc.bottom = final.top + final.height;
	if ((double) rc.bottom != (double) (final.top + final.height)) rc.bottom +=1;

	int width = rc.right - rc.left;
	if (width < 0) return;
	width /= w_scale;
	int height = rc.bottom - rc.top;
	if (height < 0) return;
	height /= h_scale;

	if (!width || !height) return;

	char *data_ptr = texture->LockBits();

	if ((format == MediaFormat :: YUV) || (format == MediaFormat :: YUVA)) {
		//switch to RGB if available (hstride is then 3*hstride)
		if (texture->m_bIsConverted) {
			if (format==MediaFormat :: YUV) {
				format = MediaFormat :: RGB;
				format = MediaFormat :: BGR;
				stride = texture->m_hstride*3;
			} else {
				stride = texture->m_hstride*3;
				format = MediaFormat :: RGBA;
			}
		} else {
			SrcYUVAImage yuv_image;

			yuv_image.width = width;
			yuv_image.height = height;
			yuv_image.orig_width = texture->m_Width;
			yuv_image.orig_height = texture->m_Height;
			yuv_image.scaleX = w_scale;
			yuv_image.scaleY = h_scale;
			yuv_image.left = rc.left;
			yuv_image.top = rc.top;
			yuv_image.pitch = stride;
			
			//using meter metrics with bitmaps is indeed hasardous :)
			yuv_image.pY = (unsigned char *) data_ptr;
			yuv_image.pU = (unsigned char *) data_ptr + texture->m_Width*texture->m_Height;
			yuv_image.pV = (unsigned char *) data_ptr + 5 * texture->m_Width*texture->m_Height / 4;

			/* update pointers given startx and starty */
			yuv_image.pY += ((int) start_y) * texture->m_hstride + (int) start_x;
			yuv_image.pU += ((int) start_y/2) * texture->m_hstride/2 + (int) start_x / 2;
			yuv_image.pV += ((int) start_y/2) * texture->m_hstride/2 + (int) start_x / 2;
			
			if (format == MediaFormat :: YUVA) {
				yuv_image.pA = yuv_image.pY + 3 * texture->m_Width*texture->m_Height / 2;
				yuv_image.pA += ((int) start_y) * texture->m_hstride + (int) start_x;
			} else {
				yuv_image.pA = NULL;
				yuv_image.const_alpha = texture->GetAlphaBlend();
			}
			yuv_image.mk = mk;
			DisplayYUV420AImage(yuv_image);
			texture->UnlockBits();
			return;
		}
	} 
	switch (format) {
	case MediaFormat :: RGBA:
		BPP = 4;
	case MediaFormat :: RGB:
	case MediaFormat :: BGR:
		break;
	default:
		texture->UnlockBits();
		return;
	}

	SrcRGBAImage rgb_image;

	rgb_image.orig_width = texture->m_Width;
	rgb_image.orig_height = texture->m_Height;
	rgb_image.width	= width;
	rgb_image.height = height;
	rgb_image.scaleX = w_scale;
	rgb_image.scaleY = h_scale;
	rgb_image.left = rc.left;
	rgb_image.top = rc.top;
	rgb_image.pitch = stride;
	rgb_image.const_alpha = texture->GetAlphaBlend();
	rgb_image.mk = mk;

	//using meter metrics with bitmaps is indeed hasardous :)
	data_ptr += ((int) start_y) * stride + ((int) start_x)*BPP;
	
	rgb_image.pA = NULL;
	switch (format) {
	case MediaFormat :: RGBA:
		rgb_image.pR = (unsigned char *) data_ptr + 2;
		rgb_image.pG = (unsigned char *) data_ptr + 1;
		rgb_image.pB = (unsigned char *) data_ptr ;
		rgb_image.pA = (unsigned char *) data_ptr + 3;

		break;
	case MediaFormat :: RGB:
		rgb_image.pR = (unsigned char *) data_ptr;
		rgb_image.pG = (unsigned char *) data_ptr + 1;
		rgb_image.pB = (unsigned char *) data_ptr + 2;
		break;
	case MediaFormat :: BGR:
		rgb_image.pB = (unsigned char *) data_ptr;
		rgb_image.pG = (unsigned char *) data_ptr + 1;
		rgb_image.pR = (unsigned char *) data_ptr + 2;
		break;
	} 
	rgb_image.ptrInc = BPP;
	DisplayRGBAImage(rgb_image);
	texture->UnlockBits();
}



// Display RGBA Image, possibly with transparency & material key if not stretched
//coordinates are in screen system (top-left = 0 0)
void VisualRendererGDI::DisplayRGBAImage(SrcRGBAImage const& srcImage)
{
	//copy image pointer
	SrcRGBAImage flippedSrcImage = srcImage;
	BOOL bUsePool = TRUE;
	// Lock the surface for writing
	DDSURFACEDESC ddSurfaceDesc;

	//Test if scaling is required from source to dest, 
	//if so copy pixel to third surface and then 
	//blit to secondary..
	HRESULT res;

	if ((srcImage.scaleX==1)&&(srcImage.scaleY==1)) bUsePool = FALSE;
	
	if (bUsePool) {
		/*transparency not supported with stretch*/
		flippedSrcImage.const_alpha = 255;
		flippedSrcImage.mk = NULL;

		/*check for a pool surface with the current used size*/
		res = m_pDirectDraw->CreatePoolSurface(srcImage.width, srcImage.height, FALSE);
		/*and lock it if existing*/
		if (res == DD_OK) res = m_pDirectDraw->LockPool(ddSurfaceDesc);
	} else {
		res = m_pDirectDraw->LockSurface(ddSurfaceDesc);
	} 

	if(res != DD_OK) {
		PostError(res);
		return;
	}

	if (bUsePool) flippedSrcImage.top = flippedSrcImage.left = 0;

	SurfaceDesc surfaceDesc;
	surfaceDesc.dwWidth   = ddSurfaceDesc.dwWidth;
	surfaceDesc.dwHeight  = ddSurfaceDesc.dwHeight;
	surfaceDesc.lPitch    = ddSurfaceDesc.lPitch;
	surfaceDesc.lpSurface = ddSurfaceDesc.lpSurface;

	// Copy the image (or a portion of it)
	switch (m_pDirectDraw->GetRGBpixelFormat())
	{
	case RGB555:
		CopyRGBAToRGB555(flippedSrcImage, surfaceDesc);
		break;
	case RGB565:
		CopyRGBAToRGB565(flippedSrcImage, surfaceDesc);
		break;
	case RGB24:
		CopyRGBAToRGB24bpp(flippedSrcImage, surfaceDesc);
		break;
	case RGB32:
		CopyRGBAToRGB32bpp(flippedSrcImage, surfaceDesc);
		break;
	default:
		assert(0);		// Shouldn't get here.
		break;
	}

	// Finished writing to the surface, so unlock it
	if (!bUsePool) {
		res = m_pDirectDraw->UnlockSurface();
	} else {
		res = m_pDirectDraw->UnlockPool();
	}
	if(res != DD_OK) {
		PostError(res);
		return;
	}
	if (!bUsePool) return;
	if(((srcImage.height) <=1) || ((srcImage.width)<=1)) return;

	RECT rectSrc,rectDst;
	rectSrc.left = flippedSrcImage.left;
	rectSrc.top = flippedSrcImage.top;
	rectSrc.right = rectSrc.left+flippedSrcImage.width;
	rectSrc.bottom = rectSrc.top+flippedSrcImage.height;
	rectDst.left = srcImage.left;
	rectDst.top = srcImage.top;
	rectDst.right = rectDst.left + flippedSrcImage.width * srcImage.scaleX;
	rectDst.bottom = rectDst.top + flippedSrcImage.height * srcImage.scaleY;
	res = m_pDirectDraw->ScalePool(rectSrc, rectDst);
	if(res != DD_OK) PostError(res);
}

// Display YUVA Image, possibly with transparency & material key if not stretched 
//coordinates are in screen system (top-left = 0 0)
void VisualRendererGDI::DisplayYUV420AImage(SrcYUVAImage const& srcImage)
{
	DDSURFACEDESC ddSurfaceDesc;
	HRESULT res;
	BOOL usePool;

	//copy image pointer
	SrcYUVAImage flippedSrcImage = srcImage;

	/*if no stretch and materialKey, disable hardware (can't get DirectDraw to work with alpha & color keys...)*/
	usePool = ((srcImage.scaleX==1)&&(srcImage.scaleY==1)) ? FALSE : TRUE;

	BOOL tryHardware = TRUE;
	if (!usePool && (srcImage.mk!=NULL)) tryHardware = FALSE;

	/*check for YUV pool*/
	if (tryHardware) {
		res = m_pDirectDraw->CreatePoolSurface(srcImage.width, srcImage.height, srcImage.pA ? FALSE : m_bUseYUVHW);
		if (res!= DD_OK) {
			PostError(res);
			return;
		}
	}

	/*do we have hardware YUV support ? */
	if (tryHardware && m_pDirectDraw->HasYUVHardware()) {
		usePool = TRUE;
		res = m_pDirectDraw->WriteYUV(srcImage.pY, srcImage.pU, srcImage.pV, srcImage.width, srcImage.height, srcImage.pitch);
		assert(res==DD_OK);
	} else {
		if (usePool) {
			/*transparency not supported with stretch*/
			flippedSrcImage.const_alpha = 255;
			flippedSrcImage.mk = NULL;

			res = m_pDirectDraw->LockPool(ddSurfaceDesc);
		} else {
			res = m_pDirectDraw->LockSurface(ddSurfaceDesc);
		}
		if(res != DD_OK) {
			PostError(res);
			return;
		}
		
		//if the image is scaled copy it starting from the top of the POOL surface
		if (usePool) flippedSrcImage.top = flippedSrcImage.left = 0;
		
		SurfaceDesc surfaceDesc;
		surfaceDesc.dwWidth   = ddSurfaceDesc.dwWidth;
		surfaceDesc.dwHeight  = ddSurfaceDesc.dwHeight;
		surfaceDesc.lPitch    = ddSurfaceDesc.lPitch;
		surfaceDesc.lpSurface = ddSurfaceDesc.lpSurface;
		
		// Copy the image (or a portion of it)
		switch (m_pDirectDraw->GetRGBpixelFormat()) {
		case RGB555:
			CopyYUVToRGB555(flippedSrcImage, surfaceDesc);
			break;
		case RGB565:
			CopyYUVToRGB565(flippedSrcImage, surfaceDesc);
			break;
		case RGB24:
			CopyYUVToRGB24bpp(flippedSrcImage, surfaceDesc);
			break;
		case RGB32:
			CopyYUVToRGB32bpp(flippedSrcImage, surfaceDesc);
			break;
		default:
			assert(0);		// Shouldn't get here.
			break;
		}
		if (usePool) {
			res = m_pDirectDraw->UnlockPool();
		} else {
			res = m_pDirectDraw->UnlockSurface();
		}
	}

	if (res != DD_OK) {
		PostError(res);
		return;
	}

	//scale it if required (hardware scale if supported)
	if (!usePool) return;
	if (((srcImage.height) <=1) || ((srcImage.width)<=1)) return;

	RECT rectSrc,rectDst;
	rectSrc.left = rectSrc.top = 0;
	rectSrc.right = flippedSrcImage.width;
	rectSrc.bottom = flippedSrcImage.height;
	rectDst.left = srcImage.left;
	rectDst.top = srcImage.top;
	rectDst.right = rectDst.left + flippedSrcImage.width * srcImage.scaleX;
	rectDst.bottom = rectDst.top + flippedSrcImage.height * srcImage.scaleY;

	res = m_pDirectDraw->ScalePool(rectSrc, rectDst);
	if(res != DD_OK) PostError(res);
}

void VisualRendererGDI::PostError(HRESULT returnValue) 
{
	GetPresenter()->GetExecutive()->Error(VIDEO_RENDER_ERROR, HRESULTtoText(returnValue));
}

void VisualRendererGDI::UpdateScreenSizeInfo()
{
	RECT rec;
	double ratio;
	HRESULT res;
	
	GetClientRect(m_hWnd, &rec);
	int width = rec.right - rec.left;
 	int height = rec.bottom - rec.top;

	m_OriginX = 0;
	m_OriginY = 0;

	//store window rect
	GetWindowRect(m_hWnd, &m_WindowRect);

	//store new size even if not init
	m_outputWidth = width;
	m_outputHeight = height;

	//if no size info, just resize
	if (!m_bHasSize) {
		m_BWidth = m_origBWidth = m_outputWidth;
		m_BHeight = m_origBHeight = m_outputHeight;
		res = m_pDirectDraw->Resize(m_BWidth, m_BHeight);
		return;
	}
	//no ratio
	if (!m_bKeepRatio) {
		m_OriginX = 0;
		m_OriginY = 0;
		goto compute_scale;
	} 

	//no init yet
	if (!m_origBWidth || !m_origBHeight) return;
	
	//compute ratio
	ratio = (double) m_origBHeight / (double) m_origBWidth;

	rec.bottom = m_outputWidth;
	rec.right = m_outputHeight;

	//scale matches height
	if (width * ratio > height) {
		m_outputHeight = height;
		m_outputWidth = height * m_origBWidth;
		m_outputWidth /= m_origBHeight;
	}
	//scale matches width
	else {
		m_outputWidth = width;
		m_outputHeight = width * m_origBHeight;
		m_outputHeight /= m_origBWidth;
	}
	m_OriginX = (width - m_outputWidth)/2;
	m_OriginY = (height - m_outputHeight)/2;

	/*erase if needed*/
	if ((rec.right > (int) m_outputWidth)  || (rec.bottom > (int) m_outputHeight))
		m_pDirectDraw->EraseSurface(NULL, TRUE);


	m_BWidth = m_origBWidth;
	m_BHeight = m_origBHeight;

	//scalable zooming part
compute_scale:
	if (!m_outputWidth || !m_outputHeight) return;

	if (!m_bScalableZoom) {
		res = m_pDirectDraw->Resize(m_BWidth, m_BHeight);
		m_pPresenter->SetScalingFactor(1.0, 1.0);
		return;
	}
	//if we use scalable zooming resize DirectDraw and ask the presenter to rescale the scene 

	m_BWidth = m_outputWidth;
	m_BHeight = m_outputHeight;
	res = m_pDirectDraw->Resize(m_BWidth, m_BHeight);

	float x_factor = m_outputWidth;
	x_factor /= m_origBWidth;
	float y_factor = m_outputHeight;
	y_factor /= m_origBHeight;
	m_pPresenter->SetScalingFactor(x_factor, y_factor);
}

void VisualRendererGDI::UpdateDisplayRect()
{
	POINT pt;
	pt.x = m_OriginX;
	pt.y = m_OriginY;
	ClientToScreen(m_hWnd, &pt);

	m_OutputRect.left = pt.x;
	m_OutputRect.top = pt.y;
	m_OutputRect.right = pt.x + m_outputWidth;
	m_OutputRect.bottom= pt.y + m_outputHeight;

}

void VisualRendererGDI::SetBIFSResolution(unsigned int width, unsigned int height, BOOL usePixelMetrics)
{
	m_bUsePixelMetrics = usePixelMetrics;

	//if no size specified use the win rect
	if (!width || !height) {
		m_bHasSize = FALSE;
		UpdateScreenSizeInfo();
		unsigned int max = m_outputWidth > m_outputHeight ? m_outputWidth : m_outputHeight;
		m_origBWidth = m_origBHeight = max;
		GetPresenter()->GetExecutive()->GetApplication()->SetSize(max, max);
		return;
	}
	m_bHasSize = TRUE;

	m_BWidth = m_origBWidth = width;
	m_BHeight = m_origBHeight = height;
	HRESULT res = m_pDirectDraw->Resize(width, height);
	if (res != DD_OK) PostError(res);
	UpdateScreenSizeInfo();
}


BOOL VisualRendererGDI::DeviceToSurfaceCoords(int input_X, int input_Y, float *output_X, float *output_Y)
{
	BOOL isInScreen = TRUE;
	int width = m_WindowRect.right - m_WindowRect.left;
	int height = m_WindowRect.bottom - m_WindowRect.top;

	//from win32 to BIFS coords
	*output_X = input_X - width / 2;
	*output_Y = height/2 - input_Y;

	if (!m_bHasSize) {
		if (m_bUsePixelMetrics) return TRUE;
		*output_X /= width/2;
		*output_Y /= width/2;
	}

	if ((input_X < m_OriginX) || ( input_X > m_OriginX + (int) m_outputWidth) 
		|| (input_Y < m_OriginY) || (input_Y > m_OriginY + (int) m_outputHeight) ) {
		isInScreen = FALSE;
	}

	//now apply scaling back to BIFS
	if (m_bScalableZoom) {
		float ft = (float) m_BWidth / (float) m_outputWidth;
		*output_X *= ft;
		ft = (float) m_BHeight / (float) m_outputHeight;
		*output_Y *= ft;
	} else {
		float ft = (float) m_origBWidth / (float) m_outputWidth;
		*output_X *= ft;
		ft = (float) m_origBHeight / (float) m_outputHeight;
		*output_Y *= ft;
	}
	//scale back to original in meters
	if (!m_bUsePixelMetrics) {
		*output_X /= m_origBWidth / 2;
		*output_Y /= m_origBWidth / 2;
	}
	return isInScreen;
}


void VisualRendererGDI::DisplayLogo()
{
	HDC hdc;
	CBitmap MPEGLogo;
	BITMAP bm;
	HRESULT res;
	
	HDC hdcImage;
	MPEGLogo.LoadBitmap(IDB_LOGO);
    hdcImage = CreateCompatibleDC(NULL);
	res = m_pDirectDraw->GetHDC(hdc);
	if (res != DD_OK) return;                

	SelectObject(hdcImage,(MPEGLogo));
	
    // get size of the bitmap
    GetObject(MPEGLogo, sizeof(bm), &bm);
    
	StretchBlt(hdc, 0, 0, m_BWidth, m_BHeight, hdcImage, 0, 0, bm.bmWidth, bm.bmHeight, SRCCOPY);
    DeleteDC(hdcImage);
	m_pDirectDraw->ReleaseHDC(hdc);
	MPEGLogo.DeleteObject();
}

BOOL VisualRendererGDI::SetRenderingOption(enum renderingOptions option, int value)
{
	const char *ret;
	Presenter *pres = GetPresenter();
	
	switch (option) {
	case VIDEO_FLAG_HIGHSPEED:
		if (m_bHighSpeed == (BOOL) value) return TRUE;
		m_bHighSpeed = (BOOL) value;
		//notify the presenter our settings have changed
		GetPresenter()->ShouldRebuildGraph();
		break;
	case VIDEO_FLAG_ANTIALIAS:
		if (m_bAntiAlias == (BOOL) value) return FALSE;
		m_bAntiAlias = (BOOL) value;
		break;
	case VIDEO_FLAG_TEXTANTIALIAS:
		if (m_bTextAntiAlias == (BOOL) value) return FALSE;
		m_bTextAntiAlias = (BOOL) value;
		break;
	case VIDEO_FLAG_SCALABLEZOOM:
		if (m_bScalableZoom == (BOOL) value) return TRUE;
		m_bScalableZoom = (BOOL) value;
		
		//rescale rendering area - NEEDS SCREEN ACCESS
		pres->AskScreenAccess();
		//forces recompute of scaling ratio
		UpdateScreenSizeInfo();
		pres->ReleaseScreenAccess();
		break;
	case VIDEO_FLAG_KEEPRATIO:
		if (!m_bHasSize) return FALSE;
		if (m_bKeepRatio == (BOOL) value) return TRUE;
		m_bKeepRatio = (BOOL) value;
		//forces recompute of scaling ratio - NEEDS SCREEN ACCESS
		pres->AskScreenAccess();
		UpdateScreenSizeInfo();
		UpdateDisplayRect();
		pres->ReleaseScreenAccess();
		break;
	case VIDEO_FLAG_RESTOREASPECT:
		if (!m_bHasSize) return TRUE;
		GetPresenter()->GetExecutive()->GetApplication()->SetSize(m_origBWidth, m_origBHeight);
		break;
	case VIDEO_FLAG_BEZIERRESOLUTION:
		m_BezierResolution = value;
		break;
	case VIDEO_FLAG_NOFILL:
		m_bNoFill = (BOOL) value;
		//notify the presenter our settings have changed
		GetPresenter()->ShouldRebuildGraph();
		break;
	case VIDEO_FLAG_FULLSCREEN:
		pres->AskScreenAccess();
		ret = m_pDirectDraw->SetFullScreen((BOOL) value);
		UpdateScreenSizeInfo();
		pres->ReleaseScreenAccess();
		if (ret) {
			pres->AskScreenAccess();
			GetPresenter()->GetExecutive()->Error(VIDEO_RENDER_ERROR, ret);
			m_pDirectDraw->SetFullScreen(FALSE);
			UpdateScreenSizeInfo();
			pres->ReleaseScreenAccess();
			return FALSE;
		}
		break;
	case VIDEO_FLAG_YUV:
		if (!m_pDirectDraw->m_numFormats) {
			m_bUseYUVHW = FALSE;
			return FALSE;
		}
		m_bUseYUVHW = (BOOL) value;
		break;
	case VIDEO_FLAG_YUV_FORMAT:
		if (value) {
			pres->AskScreenAccess();
			m_pDirectDraw->SetPreferedYUVFormat(value);
			pres->ReleaseScreenAccess();
		}
		break;
	default:
		return VisualRenderer::SetRenderingOption(option, value);
	}
	return TRUE;
}

BOOL VisualRendererGDI::GetRenderingOption(enum renderingOptions option, int &value)
{
	switch (option) {
	case VIDEO_FLAG_HIGHSPEED:
		value = m_bHighSpeed;
		break;
	case VIDEO_FLAG_ANTIALIAS:
		value = m_bAntiAlias;
		break;
	case VIDEO_FLAG_TEXTANTIALIAS:
		value = m_bTextAntiAlias;
		break;
	case VIDEO_FLAG_SCALABLEZOOM:
		value = m_bScalableZoom;
		break;
	case VIDEO_FLAG_BEZIERRESOLUTION:
		value = m_BezierResolution;
		if (m_bHighSpeed) value /= 3;
		if (value<10) value=10;
		break;
	case VIDEO_FLAG_PIXELMETRICS:
		value = m_bUsePixelMetrics;
		break;
	case VIDEO_FLAG_YUV:
		value = m_pDirectDraw->HasYUVHardware();
		break;
	case VIDEO_FLAG_YUV_FORMAT:
		value = m_pDirectDraw->m_fourcc;
		break;
	default:
		return VisualRenderer::SetRenderingOption(option, value);
	}
	return TRUE;
}

float VisualRendererGDI::GetWidth()
{
	return m_bUsePixelMetrics ? m_BWidth : 2.0;
}

float VisualRendererGDI::GetHeight()
{
	if (m_bUsePixelMetrics) return m_BHeight;
	if (m_bHasSize) return 2.0 * m_origBHeight / m_origBWidth;
	return 2.0;
}



static int GetEncoderClsid(const WCHAR* format, CLSID* pClsid)
{
   UINT  num = 0;          // number of image encoders
   UINT  size = 0;         // size of the image encoder array in bytes

   Gdiplus::ImageCodecInfo* pImageCodecInfo = NULL;

   Gdiplus::GetImageEncodersSize(&num, &size);
   if(size == 0)
      return -1;  // Failure

   pImageCodecInfo = (Gdiplus::ImageCodecInfo *)(malloc(size));
   if(pImageCodecInfo == NULL)
      return -1;  // Failure

   Gdiplus::GetImageEncoders(num, size, pImageCodecInfo);

   for(UINT j = 0; j < num; ++j)
   {
      if( wcscmp(pImageCodecInfo[j].MimeType, format) == 0 )
      {
         *pClsid = pImageCodecInfo[j].Clsid;
         free(pImageCodecInfo);
         return j;  // Success
      }    
   }

   free(pImageCodecInfo);
   return -1;  // Failure
}

void VisualRendererGDI::Screenshot(const char *pathandname, int mode)
{
	m_CaptureMode = VIDEO_CAPTURE_STOP;
	if ((mode == VIDEO_CAPTURE_STOP) || !pathandname || !strlen(pathandname)) return;

	
	strcpy(m_savePath, pathandname);

	if (strstr(m_savePath, "png") || strstr(m_savePath, "PNG")) {
		GetEncoderClsid(L"image/png", &m_EncoderClsid);
		strcpy(m_Ext, ".png");
	} else if (strstr(m_savePath, "jpg") || strstr(m_savePath, "JPG")) {
		GetEncoderClsid(L"image/jpg", &m_EncoderClsid);
		strcpy(m_Ext, ".jpg");
	} else if (strstr(m_savePath, "bmp") || strstr(m_savePath, "BMP")) {
		GetEncoderClsid(L"image/bmp", &m_EncoderClsid);
		strcpy(m_Ext, ".bmp");
	} else {
		return;
	}

	while (m_savePath[strlen(m_savePath)-1] != '.')
		m_savePath[strlen(m_savePath)-1] = 0;
	m_savePath[strlen(m_savePath)-1] = 0;
	
	m_CaptureMode = mode;
	m_NumFrame = 0;
	if (m_CaptureMode == VIDEO_CAPTURE_SHOT) {
		GetPresenter()->InvalidateScene(NULL);
	}
}

void VisualRendererGDI::CaptureFrame()
{
	// Lock the surface for writing
#ifndef USE_DX_3

	if ((m_CaptureMode == VIDEO_CAPTURE_STOP) || !m_bIsActive) return;


	Gdiplus::Bitmap *bmp = Gdiplus::Bitmap::FromDirectDrawSurface7(m_pDirectDraw->m_pOffscreen);

	char name[MAX_PATH];
	unsigned short _name[MAX_PATH];
	m_NumFrame++;
	if (m_CaptureMode == 2) {
		sprintf(name, "%s%s", m_savePath, m_Ext);
	} else {
		if (m_NumFrame<10) {
			sprintf(name, "%s_0%d%s", m_savePath, m_NumFrame, m_Ext);
		} else {
			sprintf(name, "%s_%d%s", m_savePath, m_NumFrame, m_Ext);
		}
	}
	MultiByteToWideChar(CP_UTF8, 0, name, strlen(name)+1, 
				_name, sizeof(_name)/sizeof(_name[0]) );

	bmp->Save(_name, &m_EncoderClsid, NULL);
	delete bmp;

	if (m_CaptureMode == VIDEO_CAPTURE_SHOT) m_CaptureMode = VIDEO_CAPTURE_STOP;

#endif

}


