/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File: VisualRender-GDIP.h
//
// Author: Jean Le Feuvre
//
//
// Description:
//   The definition of a class which performs visual rendering using
//   DirectDraw ans GDIplus
//
// Created: August, 2002
//
//
///////////////////////////////////////////////////////////////////////////


#ifndef VISUAL_RENDER_GDIP_H
#define VISUAL_RENDER_GDIP_H


#include "VisualRender.h"
#include "DirectDraw.h"

#include "OsmoseGdiplus.h"


// This class contains the parameters which specify a source image which
// is to be displayed.  Other classes build on this to make actual images
// which can be displayed.
class SrcImage
{
public:
	SrcImage() 
	{ left = top = width = height = pitch = 0; scaleX=scaleY=1; mk = NULL;}

	/*original image size*/
	unsigned int orig_width;
	unsigned int orig_height;

	// The line pitch for this image
	unsigned int pitch;

	// Where to display the top-left corner of this image.  The origin
	// of the scene is in the middle of the screen, with increasing X
	// to the right, and increasing Y upwards.
	int left, top;

	// The width and height portion to display
	unsigned int width, height;

	float scaleX, scaleY;

	//if 255 no alpha blending
	unsigned char const_alpha;

	//material key for bitmap drawing
	MaterialKey *mk;
};


// This class specifies the parameters for an RGBA image which is
// to be displayed.  Pointers to the four bands which make up the
// image are to be specified (although the alpha band is optional).
class SrcRGBAImage : public SrcImage
{
public:
	SrcRGBAImage()
		{ pR = pG = pB = pA = NULL; ptrInc = 0; }

	// Pointers to the start of each band in the image
	unsigned char* pR;
	unsigned char* pG;
	unsigned char* pB;
	// If this is NULL, there is no alpha layer
	unsigned char* pA;

	// How much to increment one of the above pointers to get
	// to the next pixel in the same band
	int ptrInc;
};


// This class specifies the parameters for a YUVA image which is to be
// displayed.  The Y, U, V and A bands are stored separately.
//
// The width and height parameters from SrcImage refer to the dimensions
// of the Y band.  The U and V bands are half this (rounding away from
// zero).
class SrcYUVAImage : public SrcImage
{
public:
	SrcYUVAImage()
	{ pY = pU = pV = pA = NULL; }

	// Pointers to the bands of the image,
	unsigned char* pY;
	unsigned char* pU;
	unsigned char* pV;
	unsigned char* pA;	// If this is NULL, there is no alpha
};


// The definition of a class which describes the parameters of the
// destination surface.
class SurfaceDesc
{
public:
	DWORD	dwHeight;
    DWORD	dwWidth;
    LONG	lPitch;
    LPVOID	lpSurface;
};

class VisualRendererGDI : public VisualRenderer
{
	//
	// Public functions
	//
public:


	// Class constructor - given a pointer to the presenter and a handle
	// to the window to display in.
	VisualRendererGDI(Presenter* pPresenter, HWND hWnd);

	// Class destructor.
	virtual ~VisualRendererGDI();

	// Methods inherited from the VisualRenderer base class.
	virtual void Init();
	virtual void Terminate();
	virtual void Paint();
	virtual void Clear(BRect const* pUpdateRect = NULL);
	virtual void GetCapabilities (Capabilities* pPresentationCaps);

	void PostError(HRESULT);
	
	void DisplayLogo();

	virtual void DrawBitmap(BRect *clippedDest, BRect *dest, VisualTexture *texture, MaterialKey *mk=NULL);

	//set a given option (cf enum renderingOptions )
	virtual BOOL SetRenderingOption(enum renderingOptions option, int value); 
	virtual BOOL GetRenderingOption(enum renderingOptions option, int &value);

	virtual BOOL DeviceToSurfaceCoords(int input_X, int input_Y, float *output_X, float *output_Y);
	virtual void SetBIFSResolution(unsigned int width, unsigned int height, BOOL usePixelMetrics);
	virtual void UpdateScreenSizeInfo(); 

	//returns physical offscreen surface size
	virtual unsigned int GetScreenWidth() {return m_BWidth; }
	virtual unsigned int GetScreenHeight() {return m_BHeight; }

	virtual float GetWidth();
	virtual float GetHeight();
	

	//inits the Gdiplus::Graphics object or return false otherwise
	virtual BOOL InitGraphRender(BOOL bForTextDraw = FALSE);
	//releases the Gdiplus::Graphics object
	virtual void ReleaseGraphRender();
	//computes default scaling of all objects (this is needed in case we work with MeterMetrics
	virtual BOOL InitObjectMatrix(Matrix2D &mat);

	virtual void ResetAntialiasing();

	virtual void Screenshot(const char *pathandname, int mode);

private:

	void UpdateDisplayRect();

	void CaptureFrame();

	// Display the given RGBA image, which includes where to display
	// the top-left pixel in the image.  This uses the BIFS coordinate
	// system, with the origin of the output display is in the middle
	// of the window, with increasing X and Y to the right and up.
	void DisplayRGBAImage(SrcRGBAImage const& srcImage);

	// Display the given YUV4:2:0 image, which includes where to display
	// the top-left pixel in the image.  This uses the BIFS coordinate
	// system, with the origin of the output display is in the middle
	// of the window, with increasing X and Y to the right and up.
	void DisplayYUV420AImage(SrcYUVAImage const& srcImage);

	BOOL m_initOK;

	// The dimensions of the output window
	unsigned int m_outputWidth, m_outputHeight;
	//the dimensions of the output BIFS
	unsigned int m_BWidth, m_BHeight;
	//the original BIFS resolution
	unsigned int m_origBWidth, m_origBHeight;

	unsigned int m_BezierResolution;

	BOOL m_bKeepRatio;
	RECT m_OutputRect, m_WindowRect;
	int m_OriginX, m_OriginY;

	BOOL m_bScalableZoom;	

	BOOL m_bUsePixelMetrics, m_bHasSize, m_bNoFill;
	
	//handle to display window
	HWND m_hWnd;
	BOOL m_bIsClearing;


	HDC m_hDC;

	unsigned int m_CaptureMode;
	char m_savePath[MAX_PATH];
	char m_Ext[10];
	CLSID m_EncoderClsid;
	unsigned int m_NumFrame;

	BOOL m_bUseYUVHW;

	//these are publics for access by Compositetexture
public:
	//DirectDraw object (public for YUV format selection...)
	DirectDraw *m_pDirectDraw;

	BOOL m_bHighSpeed, m_bAntiAlias;	
	BOOL m_bIsActive;
	BOOL m_bTextAntiAlias;
};


#endif // VISUAL_RENDER_GDIP_H
