/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File: VisualSurface.cpp
//
// Authors: Jean Le Feuvre ENST 2002
//
// Modified: 
//
// Description:
//   The definition of an abstract base class which performs visual
//   graph sorting and Gdiplus rendering on any surface
//		This is used for the main graph and CompositeTexture2D
//
// Created: September 2002
//
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "VisualSurface.h"
#include "bindable.h"
#include "CompositeTexture2DProxy.h"
#include "Background2DProxy.h"
#include "ViewportProxy.h"
#include "UserEventHandlerProxy.h"
#include "Presenter.h"

#include "trace.h"

#define CONTEXT_ALLOC_STEP	500

VisualSurface::VisualSurface()
{
	m_numberOfDrawables = 0;
	m_contextArray=NULL;
	m_drawableContextSize=0;
	BackgroundStack = new BindableStack;
	ViewportStack = new BindableStack;

	m_pShapeClipper = NULL;
	m_pPen = NULL;
	m_pBrush = NULL;
	m_pMatrix = NULL;
	m_pCliper = NULL;
	m_bScaleLine = FALSE;
	m_contextArray = NULL;
	m_LastUsedBoundsTracking = FALSE;
	m_bUseSplitTracking = FALSE;
	m_bHadBck = FALSE;
}

VisualSurface::~VisualSurface()
{
	for (unsigned int i=0;i<m_drawableContextSize;i++) {
		m_contextArray[i]->Reset();
		delete m_contextArray[i];	
	}
	
	free(m_contextArray);
	delete BackgroundStack; 
	delete ViewportStack; 

	while (m_SensorNodes.GetLength()) {
		SensorClipped *ptr = m_SensorNodes[0];
		delete ptr;
		m_SensorNodes.DeleteFrom(0);
	}

	ResetStyles();
}

void VisualSurface::Reset()
{
	while (m_SensorNodes.GetLength()) {
		SensorClipped *ptr = m_SensorNodes[0];
		delete ptr;
		m_SensorNodes.DeleteFrom(0);
	}
	m_PrevNodes.Clear(); 
	m_numberOfDrawables = 0;
}

//returns a new context
DrawableContext *VisualSurface::GetDrawableContext()
{
	// If the array of drawable nodes is not large enough, make it larger
	if (m_numberOfDrawables >= m_drawableContextSize) {
		m_drawableContextSize += CONTEXT_ALLOC_STEP;
		m_contextArray = (DrawableContext**) realloc(m_contextArray, m_drawableContextSize * sizeof(DrawableContext*));
		for (unsigned int i=m_numberOfDrawables;i<m_drawableContextSize;i++)
			m_contextArray[i]=new DrawableContext;	
	}
	//reset the context
	m_contextArray[m_numberOfDrawables]->Reset();
	m_contextArray[m_numberOfDrawables]->m_pVisualSurface = this;
	return m_contextArray[m_numberOfDrawables++];
}

void VisualSurface::RemoveLastContext()
{
	if (!m_numberOfDrawables) return;
	m_numberOfDrawables --;
}


void VisualSurface::InitDraw(Effects *effects, BOOL bIsCompositeTexture)
{
	effects->m_pSurface = this;

	bIsCompositeTexture = FALSE;

	//store top level matrix and screen bounds
	m_TopLevelMatrix.Set(effects->transform);

	float width, height;
	width = GetWidth();
	height = GetHeight();
	VisualPath screen;
	screen.AddRectangle(width, height);
	screen.GetBoundingRect(&m_ClippingRect, NULL, 0);

	
	m_pViewport = NULL;

	//1 - setup viewport
	ViewportProxy *pVP = (ViewportProxy *) GetViewportStack()->GetTop();
	if (pVP && (Bool) pVP->GetViewport()->isBound) {
		pVP->SetupEffects(effects, GetWidth(), GetHeight(), m_ClippingRect);
		m_pViewport = pVP;
	} 

	//2 - reset all nodes previously registered
	for (int j=0; j<m_PrevNodes.GetLength(); j++) {
		//reset bounds when changing mode
		if (m_LastUsedBoundsTracking != effects->m_bUseBoundTracking)
			m_PrevNodes[j]->ResetPreviousBounds();
		
		//move last drawn bounds to previous
		m_PrevNodes[j]->FlushBoundsList(effects->m_bShouldRebuild, this);
	}

	m_bUseSplitTracking = GetPresenter()->GetBoundsTrackingMode();
	m_AreaToClear.SetRectEmulation(!m_bUseSplitTracking);

	m_LastUsedBoundsTracking = effects->m_bUseBoundTracking;
	//if tracking draw background later
	if (effects->m_bUseBoundTracking) return;

	//3 - draw background
	Background2DProxy *pBack = (Background2DProxy *) GetBackgroundStack()->GetTop();
	if (pBack && (Bool) pBack->GetBackground2D()->isBound) {
		DrawableContext *ct = pBack->GetContext(BackgroundStack);
		//force rect to be recomputed each time (cf Background2DProxy::Render)
		ct->m_boundingRect.MakeEmpty();
		ct->m_clippedBoundingRect = ct->m_boundingRect;
		effects->m_bDrawBackground = TRUE;
		pBack->Render(effects);
		effects->m_bDrawBackground = FALSE;
	} else {
		Clear();
	}

}

BOOL VisualSurface::TerminateDraw(Effects *effects)
{
	int j, k;
	RectCollect toDraw, toErase;

	toDraw.SetRectEmulation(!m_bUseSplitTracking);
	toErase.SetRectEmulation(!m_bUseSplitTracking);

	while (m_SensorNodes.GetLength()) {
		SensorClipped *ptr = m_SensorNodes[0];
		delete ptr;
		m_SensorNodes.DeleteFrom(0);
	}
	m_NodesToDraw.Clear();

	ViewportProxy *pVP = (ViewportProxy *) GetViewportStack()->GetTop();
	if (pVP && (Bool) pVP->GetViewport()->isBound) {
		//if we missed the viewport when setting up the top matrix, redraw
		if (m_pViewport != pVP) {
			GetPresenter()->InvalidateScene(NULL);
			return FALSE;
		}
	}

	BOOL isEmpty;
	for (unsigned int i=0; i<m_numberOfDrawables; i++) {		
		//clip to surface
		m_contextArray[i]->m_clippedBoundingRect.Intersection(&m_ClippingRect);
		isEmpty = m_contextArray[i]->m_clippedBoundingRect.IsEmpty();

		//store bounds in any cases, whether the bounds tracking is off or not
		//otherwise we may not turn it on dynamically
		m_contextArray[i]->m_pDrawable->StoreBounds(m_contextArray[i]);

		//if empty skip but don't reset node
		if (isEmpty && effects->m_bUseBoundTracking) continue;

		//register sensor
		if (!isEmpty) RegisterUserSensor(m_contextArray[i]);
		
		//register node to draw
		m_NodesToDraw += m_contextArray[i];

		//if bound tracking is off the nodes are already drawn so nothing to do
		if (!effects->m_bUseBoundTracking) {
			m_contextArray[i]->m_pDrawable->ResetPreviousBounds();
			continue;
		}

		m_contextArray[i]->UpdateDrawInfo();
		
		
		if (m_contextArray[i]->m_bNeedsDrawing)
			toDraw.Union(m_contextArray[i]->m_clippedBoundingRect);
	
		//if transparent add transparent area
		if (m_contextArray[i]->m_bIsTransparent) {
			toErase.Union(m_contextArray[i]->m_clippedBoundingRect);
		}
		//otherwise try to remove any sensor hidden below
		else {
			RemoveHiddenSensors(m_contextArray[i]);
		}
	}

	//if bound tracking is off the nodes are already drawn so nothing to do
	if (!effects->m_bUseBoundTracking) return TRUE;

	//garbage collection phase 1: clear all nodes that were drawn previously and are not now
	//or have moved 
	BRect refreshRect;
	for (j=0; j<m_PrevNodes.GetLength(); j++) {
		while (m_PrevNodes[j]->GetPreviousBound(&refreshRect, this)) {
			toDraw.Union(refreshRect);
			toErase.Union(refreshRect);
		}
	}
	//don't clear the previous nodes list, it will be needed in case nothing has changed in this frame

	//garbage collection phase 2: clean all deleted nodes
	m_AreaToClear.UpdateCollection();
	if (!m_AreaToClear.IsEmpty()) {
		toDraw.Merge(m_AreaToClear);
		toErase.Merge(m_AreaToClear);
		m_AreaToClear.MakeEmpty();
	}


	BOOL drawBack = FALSE;
	Background2DProxy *pBack = (Background2DProxy *) GetBackgroundStack()->GetTop();
	DrawableContext *pBackContext = NULL;
	
	//bck needs drawing
	if (pBack) {
		m_bHadBck = TRUE;
		pBackContext = pBack->GetContext(BackgroundStack);
		if (pBackContext->m_bNeedsDrawing) {
			drawBack = TRUE;
			toDraw.MakeEmpty();
			toErase.MakeEmpty();
			toDraw.Union(pBackContext->m_clippedBoundingRect);
		}
	} 
	//previous bck has been deleted
	else if (m_bHadBck) {
		m_bHadBck = FALSE;
		drawBack = TRUE;
		toDraw.MakeEmpty();
		toErase.MakeEmpty();
		toDraw.Union(m_ClippingRect);
	}

	if (!drawBack) {
		//update areas
		toErase.UpdateCollection();
		toDraw.UpdateCollection();

		//don't erase more than what we need
		toErase.Intersection(toDraw);
	}

	//nothing to do
	if (!drawBack && toDraw.IsEmpty()) {
		return FALSE;
	}

	//mark opaque areas to speed up
	MarkOpaqueAreas(&toDraw);

	//we have the full background to draw
	if (drawBack) {
		if (pBack && (Bool) pBack->GetBackground2D()->isBound) {
			pBack->Draw(pBackContext);
		} else {
			Clear(&m_ClippingRect);
		}
	}
	//we're erasing things
	else if (!toErase.IsEmpty()) {
		//mark opaque areas to speed up
		MarkOpaqueAreas(&toErase);
		for (k=0;k<toErase.m_List.GetLength(); k++) {
			//opaque area, skip
			if (toErase.m_List[k].opaqueNodeIndex>=0) continue;

			if (pBackContext) {
				pBackContext->m_clippedBoundingRect = toErase.m_List[k];
				pBack->Draw(pBackContext);
			} else {
				Clear(&toErase.m_List[k]);
			}
		}
	}

	//then draw each node
	int drawn = 0;

	for (j = 0; j < m_NodesToDraw.GetLength(); j++) {
		BRect bounds = m_NodesToDraw[j]->m_clippedBoundingRect;

		for (k=0;k<toDraw.m_List.GetLength();k++) {
			//if we are below the opaque node for this rect, skip
			if (j < toDraw.m_List[k].opaqueNodeIndex) continue;

			//clip the node
			m_NodesToDraw[j]->m_clippedBoundingRect = bounds;
			m_NodesToDraw[j]->m_clippedBoundingRect.Intersection(&toDraw.m_List[k]);

			if (!m_NodesToDraw[j]->m_clippedBoundingRect.IsEmpty()) {
				m_NodesToDraw[j]->m_pDrawable->Draw(m_NodesToDraw[j]);
				drawn ++;
			}
		}

		//restore original clipped rect (for sensors)
		m_NodesToDraw[j]->m_clippedBoundingRect = bounds;

		//wether we drawned or not the node keep track of it
		if (m_PrevNodes.Find(m_NodesToDraw[j]->m_pDrawable) < 0) 
			m_PrevNodes += m_NodesToDraw[j]->m_pDrawable;
	}
//	TRACE("drawn %d nodes\n", drawn);
	return j ? TRUE : FALSE;
}

//removes all nodes covered by given top
//NOTE: we cannot remove the context if hidden because we need to flush bounds list
//once all bounds are known, and a given node may have an instance hidden while another is visible
//we thus rely on the opaque areas' thing to skip hidden nodes instances drawning
void VisualSurface::RemoveHiddenSensors(DrawableContext *pTop)
{

	for (int i=0; i<m_NodesToDraw.GetLength()-1; i++) {
		if (! (m_NodesToDraw[i]->m_clippedBoundingRect <= pTop->m_clippedBoundingRect) ) continue;

		//remove associated sensor if any
		if (! m_NodesToDraw[i]->m_Handlers.GetLength()) continue;

		for (int k=0; k<m_SensorNodes.GetLength(); k++) {
			SensorClipped *pTmp = m_SensorNodes[k];
			if (pTmp->pContext == m_NodesToDraw[i]) {
				m_SensorNodes.DeleteFrom(k);
				delete pTmp;
				break;
			}
		}
	}
}

//figure out if they are opaque areas among those to redraw. if so mark them
void VisualSurface::MarkOpaqueAreas(RectCollect *toDraw)
{
	int i, k;

	//for each area to redraw
	for (k=0;k<toDraw->m_List.GetLength();k++) {
		toDraw->m_List[k].opaqueNodeIndex = -1;

		//locate the top-most node
		for (i=m_NodesToDraw.GetLength()-1; i>=0; i--) {
			if (!toDraw->m_List[k].Overlaps(m_NodesToDraw[i]->m_clippedBoundingRect)) continue;

			//which opaquely covers the given area and store its index
			if (!m_NodesToDraw[i]->m_bIsTransparent && (toDraw->m_List[k] <= m_NodesToDraw[i]->m_clippedBoundingRect)) {
				DrawableContext *pct = m_NodesToDraw[i];
				toDraw->m_List[k].opaqueNodeIndex = i;
				break;
			}
		}
	}
}


void VisualSurface::RegisterUserSensor(DrawableContext *pContext)
{

	//notify all previous sensor a new node is on top
	for (int i=0; i<m_SensorNodes.GetLength(); i++) {
		if (m_SensorNodes[i]->pContext->m_boundingRect.Overlaps(&pContext->m_boundingRect)) {
			m_SensorNodes[i]->m_NodesOnTop += pContext;
		}
	}
	
	//then check if this is a sensor. If so add it
	int length=pContext->m_Handlers.GetLength();
	if (length) {
		//check the state of the sensor
		for (int i=0; i<length; i++) {
			if (pContext->m_Handlers[i]->m_pHandler->IsEnabled()) {
				goto register_top;
			} 
		}
		//reset sensor list
		pContext->ResetHandlers();
	}
	//no texture
	if (!pContext->m_pTextureProxy) return;
	//NB: this means we don't track user interaction on matteTexture using CompositeTexture
	//as one of its surface :)
	if (pContext->m_pTextureProxy->GetObject()->GetNodeCode() != CODE_CompositeTexture2D) return;

	if ( !((CompositeTexture2DProxy *)pContext->m_pTextureProxy)->HasSensors()) return;

	
register_top:
	SensorClipped *pSensor;
	pSensor = new SensorClipped;
	pSensor->pContext = pContext;
	m_SensorNodes += pSensor;
}


//call when the node is deleted to remove it from the list of nodes
void VisualSurface::NodeDeleted(DrawableNode *pNode)
{
	//NB: we only keep one instance of the node in the prev list
	int i = m_PrevNodes.Find(pNode);
	if (i>=0) {
		m_PrevNodes.DeleteFrom(i);
		GetPresenter()->InvalidateScene(NULL);
	}
	//also remove associated sensors
	for (i=0; i<m_SensorNodes.GetLength(); i++) {
		SensorClipped *pSensor = m_SensorNodes[i];
		if (pSensor->pContext->m_pDrawable == pNode) {
			delete pSensor;
			m_SensorNodes.DeleteFrom(i);
			i--;
		} else {
			for (int j=0; j<pSensor->m_NodesOnTop.GetLength(); j++) {
				if (pSensor->m_NodesOnTop[j]->m_pDrawable == pNode) {
					pSensor->m_NodesOnTop.DeleteFrom(j);
					j--;
				}
			}
		}
	}
}

// VisualRenderer::FindNode
// Go through the list of drawable nodes, in front to back depth
// order (opposite to the drawing order), looking for the first which
// is under the given point (which is in BIFS coordinates).
DrawableContext * VisualSurface::FindNode(float x, float y)
{
	if (!HasSensors()) return NULL;

	DrawableContext *pTemp;

	//Look through the list, starting with the front most
	int i = m_SensorNodes.GetLength() - 1;
restart:
	for (; i >= 0; i--) {
		//the point shall be over the bounding rect of the node
		if (!m_SensorNodes[i]->pContext->IsPointInBounds(x, y)) continue;

		for (int k=m_SensorNodes[i]->m_NodesOnTop.GetLength(); k>0; k--) {
			pTemp = m_SensorNodes[i]->m_NodesOnTop[k-1];
			//if not in bounds an dnot in shape skip it
			if (!pTemp->IsPointInBounds(x, y) ) continue;
			if (!pTemp->m_pDrawable->IsPointOver(pTemp, x, y) ) continue;

			//we're over another node: 
			//1 - it has no sensor, send back no sensor
			if (!pTemp->m_Handlers.GetLength()) return NULL;
			//2 - it has sensors: skip current sensor (we could locate next one too but that would get messy :)
			if (i) i--;
			goto restart;
		}

		//and over the shape itself
		pTemp = m_SensorNodes[i]->pContext;
		if (! pTemp->m_pDrawable->IsPointOver(pTemp, x, y) ) continue;

		//check if has sensors
		if (m_SensorNodes[i]->pContext->m_Handlers.GetLength()) {
			return m_SensorNodes[i]->pContext;
		}
		//check for composite texture
		if (!m_SensorNodes[i]->pContext->m_pTextureProxy) return NULL;
		if (m_SensorNodes[i]->pContext->m_pTextureProxy->GetObject()->GetNodeCode() != CODE_CompositeTexture2D) return NULL;
		
		return ((CompositeTexture2DProxy *) m_SensorNodes[i]->pContext->m_pTextureProxy)->IsPointInTexture(m_SensorNodes[i]->pContext, x, y);
    }
	return NULL;
}

void VisualSurface::DrawVisualPath(VisualPath *path, BOOL no_line, ColorGradient *gradient)
{
	Matrix2D mat;
	Gdiplus::Matrix final;

	if (!path) return;

	if (!InitGraphRender(path->m_bHasText)) return;

	InitObjectMatrix(mat);

	//apply the same matrix if needed (only the scaling part is used by GDIplus)
	if (m_pPen && !no_line) {
		final.Reset();
		float height, width;
		height = width = m_pPen->GetWidth();
		float coefs[6];

		mat.Apply(width, height);
		width += height;
		width /= 2;
		m_pPen->SetWidth(width);

		if (m_bScaleLine && m_pMatrix) {
			m_pMatrix->GetElements(coefs);
			coefs[4] = coefs[3];
		}
		else
			m_TopLevelMatrix.GetElements(coefs);

		if (coefs[0]<0) coefs[0] *= -1;
		if (coefs[4]<0) coefs[4] *= -1;
		width *= (coefs[0] > coefs[4]) ? coefs[0] : coefs[4];
		m_pPen->SetWidth(width);


		float test;
		test = m_pPen->GetWidth();
	
	}


	if (m_pMatrix) mat.m_Matrix.Multiply(m_pMatrix);

	//Note: Gdiplus applies the transform matrix to the graphics cliper: we should check what is the fastest:
	//* clone path and apply transform to path only (what we are doing)
	//* don't clone path and build a graph clipper from the inverse transfom (pb: is it always possible?)
	//* any other suggestion
	Gdiplus::GraphicsPath *_path = path->m_Path.Clone();
	_path->Transform(& mat.m_Matrix);

	//debug, to show cliper
#if 0
	if (m_pCliper) {
		Gdiplus::Pen pen(Gdiplus::Color(0, 0, 0), 1.0);
		m_pGraph->DrawRectangle(&pen, *m_pCliper);
	}
#endif

	//cliper
	if (m_pCliper) m_pGraph->SetClip(*m_pCliper, Gdiplus::CombineModeReplace);
	
	//shaped clipper
	if (m_pShapeClipper)
		m_pGraph->SetClip(m_pShapeClipper->m_pRegion, m_pShapeClipper->m_bIsClipping ? Gdiplus::CombineModeReplace : Gdiplus::CombineModeExclude);

	//filled brush
	if (m_pBrush) {
		if (gradient) {
			if (gradient->m_pBrush) {
				gradient->m_pBrush->SetTransform(&gradient->m_Matrix.m_Matrix);
				
				//Gdiplus doesn't implement GetGraphicsPath for brushes, so do it ourselves
				Gdiplus::GraphicsPath *pPath = gradient->m_pPath->Clone();
				pPath->Transform(&gradient->m_Matrix.m_Matrix);
				m_pGraph->SetClip(pPath, Gdiplus::CombineModeExclude);
				
				if (m_pShapeClipper)
					m_pGraph->SetClip(m_pShapeClipper->m_pRegion, m_pShapeClipper->m_bIsClipping ? Gdiplus::CombineModeReplace : Gdiplus::CombineModeExclude);


				//Gdiplus path brush don't always fill the full exceed their bounds, 
				//so force a fill with the back color
				Gdiplus::Color col(gradient->m_BackColor.a, gradient->m_BackColor.r, gradient->m_BackColor.g, gradient->m_BackColor.b);
				Gdiplus::SolidBrush *pBrush = new Gdiplus::SolidBrush(col);
				m_pGraph->FillPath(pBrush, _path);
				delete pBrush;
				delete pPath;
				m_pGraph->ResetClip();
				if (m_pCliper) m_pGraph->SetClip(*m_pCliper, Gdiplus::CombineModeReplace);

				if (m_pShapeClipper)
					m_pGraph->SetClip(m_pShapeClipper->m_pRegion, m_pShapeClipper->m_bIsClipping ? Gdiplus::CombineModeReplace : Gdiplus::CombineModeExclude);

				m_pGraph->FillPath(gradient->m_pBrush, _path);

			} else if (gradient->m_pLinearBrush) {
				gradient->m_pLinearBrush->SetTransform(&gradient->m_Matrix.m_Matrix);
				m_pGraph->FillPath(gradient->m_pLinearBrush, _path);
			}
		} else {
			m_pGraph->FillPath(m_pBrush, _path);
		}
	}

	//outline
	if (m_pPen && !no_line) {
		//can use both linear or complex
		if (gradient) {
			if (gradient->m_pLinearBrush) {
				if (m_pMatrix) gradient->m_pLinearBrush->SetTransform(m_pMatrix);
				m_pPen->SetBrush(gradient->m_pLinearBrush);
			} else {
				if (m_pMatrix) gradient->m_pBrush->SetTransform(m_pMatrix);
				m_pPen->SetBrush(gradient->m_pBrush);
			}
		}
		m_pGraph->DrawPath(m_pPen, _path);
	}
	//flush
	m_pGraph->Flush(Gdiplus::FlushIntentionSync);

	delete _path;
	ReleaseGraphRender();
}



void VisualSurface::TextureVisualPath(VisualPath *path, VisualTexture *texture, BOOL noAntiAlias)
{
	BOOL bPixelMetrics;
	Matrix2D mat;
	Gdiplus::Matrix final;

	if (!path || !texture) return;

	if (!InitGraphRender(path->m_bHasText)) return;

	if (noAntiAlias) ResetAntialiasing();

	bPixelMetrics = InitObjectMatrix(mat);
	BRect rect;
	path->GetBoundingRect(&rect, &mat, 0);

	if (m_pMatrix) mat.m_Matrix.Multiply(m_pMatrix);
		
	//Note: Gdiplus applies the transform matrix to the graphics cliper: we should check what is the fastest:
	//* clone path and apply transform to path only (what we are doing)
	//* don't clone path and build a graph clipper from the inverse transfom (pb: is it always possible?)
	//* any other suggestion
	Gdiplus::GraphicsPath *_path = path->m_Path.Clone();
	_path->Transform(&mat.m_Matrix);

	//cliper
	if (m_pCliper) m_pGraph->SetClip(*m_pCliper, Gdiplus::CombineModeReplace);

	//shaped clipper
	if (m_pShapeClipper) m_pGraph->SetClip(m_pShapeClipper->m_pRegion, m_pShapeClipper->m_bIsClipping ? Gdiplus::CombineModeReplace : Gdiplus::CombineModeExclude);

	final.Reset();

	//setup texture transform
	//get width/height of the texture in new coord system
	Gdiplus::GraphicsPath txpath;
	InitObjectMatrix(mat);
	
	float width, height;
	width = texture->GetActiveWidth();
	height = texture->GetActiveHeight();
	mat.ReverseApply(width, height);
	Gdiplus::RectF rc(0, 0, width, height);
	txpath.AddRectangle(rc);
	txpath.GetBounds(&rc, &texture->m_Matrix.m_Matrix, NULL);

	width = rc.Width;
	height = rc.Height;

	//set to top-left
	float coef[6];
	texture->m_Matrix.GetElements(coef);
	final.Translate((coef[0]>=0) ? rect.left : -rect.left, (coef[4]>=0) ? rect.top : -rect.top, Gdiplus::MatrixOrderAppend);
//	final.Translate(rect.left, rect.top, Gdiplus::MatrixOrderAppend);

	final.Multiply(&texture->m_Matrix.m_Matrix, Gdiplus::MatrixOrderAppend);

	mat.ReverseApply(rect.width, rect.height);
	//compute texture coords
	//set tiling
	float ratio_s, ratio_t;
	ratio_s = rect.width;
	ratio_s /= width;
	ratio_t = rect.height;
	ratio_t /= height;

	//both directions
	if (texture->m_RepeatS && texture->m_RepeatT) {
		final.Scale(1, -1);
	}
	//S tiling
	else if (texture->m_RepeatS ) {
		final.Scale(1, -ratio_t);
	}
	//T tiling
	else if (texture->m_RepeatT) {
		final.Scale(ratio_s, -1);
	}
	//no tiling
	else {
		final.Scale(ratio_s, -ratio_t);
	}

	if (m_pMatrix) {
		//in meter metrics we must use an hybrid matrix for bitmap brush, where the translation
		//is converted to pixels but scale and rotation are untouched. 
		if (!bPixelMetrics) {
			float cf1[6], cf2[6];
			Matrix2D temp;
			mat.m_Matrix.GetElements(cf1);
			m_pMatrix->GetElements(cf2);
						
			temp.m_Matrix.SetElements(cf2[0], cf2[1], cf2[2], cf2[3], cf1[0] * cf2[4], cf1[3] * cf2[5]); 			
			final.Multiply(&temp.m_Matrix, Gdiplus::MatrixOrderAppend);
		} else {
			final.Multiply(m_pMatrix, Gdiplus::MatrixOrderAppend);
		}
	}

	//compute texture coords from BRect
	texture->m_pTexture->SetTransform(&final);

	//and fill path
	m_pGraph->FillPath(texture->m_pTexture, _path);

	texture->m_pTexture->ResetTransform();

	m_pGraph->Flush(Gdiplus::FlushIntentionSync);
	delete _path;

	ReleaseGraphRender();
}

void VisualSurface::SetLineStyle(Color3f pixColor, float transparency, float width, int EndCap, int JoinCap, int DashStyle, float miterLimit, BOOL IsScaled, BOOL LineInsideShape)
{
	BYTE alpha;

	if (!width) return;
	if (!m_pPen) delete m_pPen;
	//color
	ApplyColor(pixColor, transparency);
	alpha = 255 - (BYTE) (255*transparency);
	Gdiplus::Color col(alpha, 255*pixColor.r, 255*pixColor.g, 255*pixColor.b);
	m_pPen = new Gdiplus::Pen(col, width);

	m_bScaleLine = IsScaled;

	m_pPen->SetAlignment(LineInsideShape ? Gdiplus::PenAlignmentInset : Gdiplus::PenAlignmentCenter);
	//end cap
	switch (EndCap) {
	case LINE_CAP_SQUARE:
		m_pPen->SetEndCap(Gdiplus::LineCapSquare);
		m_pPen->SetStartCap(Gdiplus::LineCapSquare);
		m_pPen->SetDashCap(Gdiplus::DashCapFlat);
		break;
	case LINE_CAP_ROUND:
		m_pPen->SetEndCap(Gdiplus::LineCapRound);
		m_pPen->SetStartCap(Gdiplus::LineCapRound);
		m_pPen->SetDashCap(Gdiplus::DashCapRound);
		break;
	case LINE_CAP_FLAT:
	default:
		m_pPen->SetEndCap(Gdiplus::LineCapFlat);
		m_pPen->SetStartCap(Gdiplus::LineCapFlat);
		m_pPen->SetDashCap(Gdiplus::DashCapFlat);
		break;
	}
	//join cap
	switch (JoinCap) {
	case LINE_JOIN_ROUND:
		m_pPen->SetLineJoin(Gdiplus::LineJoinRound);
		break;
	case LINE_JOIN_MITER:
		m_pPen->SetLineJoin(Gdiplus::LineJoinMiter);
		m_pPen->SetMiterLimit(miterLimit);
		break;
	case LINE_JOIN_BEVEL:
	default:
		m_pPen->SetLineJoin(Gdiplus::LineJoinBevel);
		break;
	}

	//dash styles
	float dashValues[3] = {5, 5, 2};
	switch (DashStyle) {
	case DASH_DASH_DASH_DOT:
		//not specified at all in specs
		m_pPen->SetDashStyle(Gdiplus::DashStyleCustom);
		
		m_pPen->SetDashPattern(dashValues, 3);
		break;
	default:
		if (DashStyle > DASH_DASH_DOT_DOT) DashStyle = DASH_SOLID;
		m_pPen->SetDashStyle( (Gdiplus::DashStyle) DashStyle);
		break;
	}
}

void VisualSurface::SetBrushStyle(BOOL Filled, Color3f pixColor, float transparency)
{
	if (m_pBrush) delete m_pBrush;
	m_pBrush = NULL;
	if (!Filled) return;

	ApplyColor(pixColor, transparency);

	BYTE alpha = 255 - (BYTE) (255*transparency);
	Gdiplus::Color col(alpha, 255*pixColor.r, 255*pixColor.g, 255*pixColor.b);
	m_pBrush = new Gdiplus::SolidBrush(col);
}

void VisualSurface::SetTransform(Matrix2D *matrix)
{
	//delete matrix if any
	if (m_pMatrix) {
		delete m_pMatrix;
		m_pMatrix = NULL;
	}
	if (matrix) m_pMatrix = matrix->m_Matrix.Clone();	
}

void VisualSurface::SetClip(BRect *pClipBRect)
{
	if (m_pCliper) {
		delete m_pCliper;
		m_pCliper = NULL;
	}
	if (!pClipBRect) return;

	BRect rc = *pClipBRect;
	Matrix2D mat;
	InitObjectMatrix(mat);
	rc.width += rc.left;
	rc.height += rc.top;
	mat.Apply(rc.left, rc.top);
	mat.Apply(rc.width, rc.height);
	rc.width -= rc.left;
	rc.height -= rc.top;

	m_pCliper = new Gdiplus::RectF(rc.left, rc.top - rc.height, rc.width, rc.height ); 

}

void VisualSurface::ResetStyles()
{
	if (m_pPen) delete m_pPen;
	m_pPen = NULL;
	if (m_pMatrix) delete m_pMatrix;
	m_pMatrix = NULL;
	if (m_pBrush) delete m_pBrush;
	m_pBrush = NULL;
	if (m_pCliper) delete m_pCliper;
	m_pCliper = NULL;
	m_pColorMatrix = NULL;
	m_pShapeClipper = NULL;
}

void VisualSurface::SetColorMatrix(ColorMatrix *pMatrix)
{
	m_pColorMatrix = pMatrix;
}
void VisualSurface::SetShapeClipper(VisualClipper *pClip)
{
	m_pShapeClipper = pClip;
}

void VisualSurface::ApplyColor(Color3f &col, float &transparency)
{
	if (m_pColorMatrix) 
		m_pColorMatrix->Apply(col, transparency);
}
