/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File: VisualSurface.h
//
// Authors: Jean Le Feuvre ENST 2002
//
// Modified: 
//
// Description:
//   The definition of an abstract base class which performs visual
//   rendering on any surface
//		This is used for the main graph, layer2D and CompositeTexture2D
//
//	NOTE: this file is Gdiplus specific to avoid code duplication
// Created: September 2002
//
//
///////////////////////////////////////////////////////////////////////////


#ifndef VISUALSURFACE_H
#define VISUALSURFACE_H


#include "ztypes.h"
#include "RectCollect.h"
#include "drawable.h"
#include "ztl.h"
#include "Nodes.h"
#include "ColorGradient.h"

#include "OsmoseGdiplus.h"

class Presenter;
class DrawableNode;
class BRect;
class BindableStack;
class VisualTexture;
class ViewportProxy;

//Join styles as specified in XLineProperties
enum
{
	LINE_JOIN_MITER = 0,
	LINE_JOIN_ROUND,
	LINE_JOIN_BEVEL,
};

//End styles  as specified in XLineProperties
enum
{
	LINE_CAP_FLAT = 0,
	LINE_CAP_ROUND,
	LINE_CAP_SQUARE,
};

//Line styles  as specified in LineProperties/XLineProperties
enum
{
	DASH_SOLID = 0,
	DASH_DASH,
	DASH_DOT,
	DASH_DASH_DOT,
	DASH_DASH_DASH_DOT,
	DASH_DASH_DOT_DOT,
};

class SensorClipped
{
public:
	DrawableContext *pContext;
	ZArray<DrawableContext *> m_NodesOnTop;
};

class VisualSurface
{

	// Public functions
public:
	VisualSurface();
	virtual ~VisualSurface();


	void Reset();


	// Completely erase the specifed region of the offscreen drawing
	// area, but do not display the surface.  If the given rectangle
	// is NULL, erase the entire offscreen area.
	virtual void Clear(BRect const* pUpdateRect = NULL) = 0;


	//return width , height in pixels or meters
	virtual float GetWidth() = 0;
	virtual float GetHeight() = 0;

	//translation from device coordinates to BIFS coordinates. 
	//returns FALSE if device is not in BIFS screen
	virtual BOOL DeviceToSurfaceCoords(int input_X, int input_Y, float *output_X, float *output_Y) = 0;

	// Call this function before calling Render() of the root of
	// the scene, to ensure that the list of drawable nodes is empty.
	void ResetDrawList() { m_numberOfDrawables = 0; }

	// This function is called by a renderable node which needs to
	// be rendered.	
	void AddToDrawList(DrawableNode *pDrawableNode);

	// Drawable Context list management
	DrawableContext *GetDrawableContext();
	void RemoveLastContext();

	//returns the background stack
	BindableStack *	GetBackgroundStack() { return BackgroundStack;}
	BindableStack *	GetViewportStack() { return ViewportStack;}
		
	//draw background if needed and setup world matrix according to top-level viewport
	//if compositeTexture is TRUE, viewport fit/align params are ignored
	void InitDraw(Effects *effects, BOOL bIsCompositeTexture = FALSE);
	
	//if object track changes is enabled, compute the min rect(s) to redraw
	//and draw the scene
	//returns TRUE if some drawing has ocured (note that if bounds tracking isn't used
	//it always returns true)
	BOOL TerminateDraw(Effects *effects);

	//call when the node is deleted to remove it from the list of nodes
	void NodeDeleted(DrawableNode *pNode);

	// Go through the list of drawable nodes, in front to back depth
	// order, looking for the first which is under the given point
	// (which is in BIFS coordinates).
	DrawableContext * FindNode(float x, float y);

	// Return a pointer to the Presenter
	Presenter* GetPresenter() const { return m_pPresenter; }
	void SetPresenter(Presenter *pPresenter) { m_pPresenter = pPresenter; }

	//reset line, brush, cliper and transform info
	virtual void ResetStyles();

	const BRect *GetSurfaceClipper() {return &m_ClippingRect; }

	//draw the specified path and reset styles afterwards
	void DrawVisualPath(VisualPath *path, BOOL no_line=FALSE, ColorGradient *gradient=NULL);
	
	//fill the specified path with the given Visualtexture object - outline is not drawn and styles not reset
	void TextureVisualPath(VisualPath *path, VisualTexture *texture, BOOL noAntiAlias = FALSE);
	
	//draw the specified texture as a bitmap (no fliping, no tiling, no rotation) in the given rectangle (in 
	//BIFS coordinates)
	//clippedDest: clipped destination rect. 
	//dest: destination rect. 
	//texture: texture to draw
	//MaterialKey: material key node used for bitmap, or NULL
	//All styles and cliper are ignored in this function
	//Note: bitmap is not supported in composite texture
	virtual void DrawBitmap(BRect *clippedDest, BRect *dest, VisualTexture *texture, MaterialKey *mk=NULL) = 0;

	//set current line style
	void SetLineStyle(Color3f pixColor, float transparency, float width, int EndCap=LINE_CAP_FLAT, int JoinCap=LINE_JOIN_BEVEL, int DashStyle=DASH_SOLID, float miterLimit=0.0, BOOL IsScaled=FALSE, BOOL LineInsideShape = FALSE);

	//set current brush style (except texture brush)
	void SetBrushStyle(BOOL Filled, Color3f pixelColor, float transparency);
	//set current world transform matrix
	void SetTransform(Matrix2D *);
	//set clipping rectangle
	void SetClip(BRect *pClipBRect);

	//setup a pointer to the active color matrix (not duplicated)
	void SetColorMatrix(ColorMatrix *pMatrix);

	void SetShapeClipper(VisualClipper *pClipper);

	//computes scaling applied to all objects (this is needed in case we work with MeterMetrics)
	//return TRUE if using pixelMetrics
	virtual BOOL InitObjectMatrix(Matrix2D &mat) = 0;

	//area(s) to be cleared at each step. Filled by nodes when destroyed. 
	RectCollect m_AreaToClear;

	BOOL HasSensors() 
	{
		return m_SensorNodes.GetLength() ? TRUE : FALSE;
	}

	BRect GetClipper() { return m_ClippingRect; }

	//top-level transform matrix (this is used to compute pen scaling and gradients)
	Matrix2D m_TopLevelMatrix;

	// Protected implementation details
	//
protected:

	// Remember a pointer to the presenter that this renderer is part of
	Presenter* m_pPresenter;

	//list of contexts
	DrawableContext **m_contextArray;
	// The allocated size of this array
	unsigned int m_drawableContextSize;
	// The number of items in this array
	unsigned int m_numberOfDrawables;

	// This is TRUE if the screen should be cleared when the end
	// of a file is reached.
	BOOL m_clearOnEnd;
    
	//background list. 
	BindableStack * BackgroundStack;
	//viewport list
	BindableStack * ViewportStack;

	//inits the Gdiplus::Graphics object or return false otherwise
	virtual BOOL InitGraphRender(BOOL bForTextDraw = FALSE) = 0;
	//releases the Gdiplus::Graphics object
	virtual void ReleaseGraphRender() = 0;

	//for bitmap (Gdiplus seems to have pbs with antialiased bitmaps)
	virtual void ResetAntialiasing() = 0;

	//this shall be init/released by derived classes
	Gdiplus::Graphics *m_pGraph;

private:
	void ApplyColor(Color3f &col, float &transparency);
	
	//GDIPlus variables setup for each object
	Gdiplus::Pen *m_pPen;
	Gdiplus::Brush *m_pBrush;
	Gdiplus::Matrix *m_pMatrix;
	Gdiplus::RectF *m_pCliper;
	BOOL m_bScaleLine;
	
	ColorMatrix *m_pColorMatrix;
	VisualClipper *m_pShapeClipper;


	//clipping rect as defined by the viewport if any or full surface rect
	BRect m_ClippingRect;

	//node garbage collector: this cannot be done through context since nodes
	//may "disapear" from the graph
	ZArray<DrawableNode *> m_PrevNodes;

	void RegisterUserSensor(DrawableContext *pContext);
	ZArray<SensorClipped *> m_SensorNodes;
	/*remove all hidden sensors (but keep context)*/
	void RemoveHiddenSensors(DrawableContext *pTop);
	/*find Z-order of first opaque node instance covering the areas to redraw*/
	void MarkOpaqueAreas(RectCollect *toDraw);
	ZArray<DrawableContext *> m_NodesToDraw;

	ViewportProxy *m_pViewport;

	BOOL m_LastUsedBoundsTracking;
	BOOL m_bUseSplitTracking;

	BOOL m_bHadBck;
};


#endif // VISUALSURFACE_H

