/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File: VisualTexture.cpp
//
// Author: Jean Le Feuvre
//
// Description:
//		Wrapper for texture brush
//
// Created: August 2002
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "ColorTools.h"
#include "VisualTexture.h"
#include "ColorMatrix.h"

VisualTexture::VisualTexture()
{
	m_pTexture = NULL;
	m_pBitmap = NULL;
	m_pOriginalPixels = NULL;
	m_pRGBPixels = NULL;
	m_Width = m_Height = 0;
	m_hstride = 0;
	m_pVisbleRect = NULL;
	//init color matrix to identity
	memset(&m_colorMatrix, 0, sizeof(Gdiplus::ColorMatrix));
	m_colorMatrix.m[0][0] = 1.0;
	m_colorMatrix.m[1][1] = 1.0;
	m_colorMatrix.m[2][2] = 1.0;
	m_colorMatrix.m[3][3] = 1.0;
	m_colorMatrix.m[4][4] = 1.0;

	m_RepeatS = m_RepeatT = TRUE;
	m_bIsConverted = FALSE;
	m_bIsTransparent = FALSE;
	m_bIsBGR = FALSE;
	m_Matrix.SetIdentity();
	m_pBitmapData = NULL;

	m_PrevMatrix.Reset();
	m_bHadMatrix = FALSE;
}

VisualTexture::~VisualTexture()
{
	//just in case
	UnlockBits();
	if (m_pBitmap) delete m_pBitmap;
	//free conversion memory
	if (m_pRGBPixels) free(m_pRGBPixels);
	if (m_pVisbleRect) delete m_pVisbleRect;
}

void VisualTexture::Reset()
{
	m_colorMatrix.m[4][0] = m_colorMatrix.m[4][1] = m_colorMatrix.m[4][2] = m_colorMatrix.m[4][3] = 0.0;
	m_colorMatrix.m[1][1] = 1.0;
	m_colorMatrix.m[3][3] = 1.0;
	if (m_bIsBGR) {
		m_colorMatrix.m[0][2] = 1.0;
		m_colorMatrix.m[2][0] = 1.0;
	} else {
		m_colorMatrix.m[0][0] = 1.0;
		m_colorMatrix.m[2][2] = 1.0;
	}
}

void VisualTexture::Invalidate()
{
	if (m_pTexture) delete m_pTexture;
	m_pTexture = NULL;
}

BOOL VisualTexture::CreateBrush()
{
	if (m_pTexture) delete m_pTexture;
	if (!m_pVisbleRect) {
		Gdiplus::Rect src_rect(m_OffsetX, m_OffsetY, m_Width, m_Height);
		m_pTexture = new Gdiplus::TextureBrush(m_pBitmap, src_rect, &m_Properties);
	} else {
		m_pTexture = new Gdiplus::TextureBrush(m_pBitmap, *m_pVisbleRect, &m_Properties);
	}
	m_pTexture->SetWrapMode(Gdiplus::WrapModeTile);
	if (!m_pTexture) return FALSE;

//	TRACE("Creating brush\n");
	return TRUE;
}

BOOL VisualTexture::SetTransparency(float transparency, ColorMatrix *pMatrix)
{
	if (!m_pBitmap) return FALSE;

	//creating the brush is extremely time consuming, so filter calls
	if (pMatrix) {
		ColorMatrix mat;
		mat.SetGDIPLUS(m_colorMatrix);
		mat.Multiply(*pMatrix);
		//same color transform and texture already present : same as last time
		if (m_pTexture && (mat==m_PrevMatrix)) return TRUE;
		m_PrevMatrix.Set(mat);
		m_bHadMatrix = TRUE;
		m_Properties.SetColorMatrix(&mat.m_Matrix);
	} else {
		//same color transform and texture already present : same as last time
		if (m_pTexture && (m_colorMatrix.m[3][3] == 1.0 - transparency) && !m_bHadMatrix) return TRUE;
		m_colorMatrix.m[3][3] = 1.0 - transparency;
		m_bHadMatrix = FALSE;
		m_Properties.SetColorMatrix(&m_colorMatrix);
	}
	return CreateBrush();
}



BOOL VisualTexture::Init(unsigned int width, unsigned int height, unsigned char *pixel_data, unsigned int horiz_stride, MediaFormat colorFormat, unsigned int offset_x, unsigned int offset_y)
{
	if (!width || !height || !pixel_data || !horiz_stride) return FALSE;
	
	Gdiplus::PixelFormat pformat;

	m_bIsTransparent = FALSE;
	BOOL isYUV = FALSE;
	BOOL deleteSurface = TRUE;
	m_bIsBGR = FALSE;

	int BPP = 0;
	//make sure we have identity matrix
	m_colorMatrix.m[0][0] = 1.0;
	m_colorMatrix.m[0][2] = 0.0;
	m_colorMatrix.m[2][2] = 1.0;
	m_colorMatrix.m[2][0] = 0.0;

	pformat = PixelFormat24bppRGB;

	switch (colorFormat) {
	case MediaFormat :: GRAYSCALE:
		BPP = 1;
		break;
	case MediaFormat :: ALPHAGRAYSCALE:
		pformat = PixelFormat32bppARGB;
		BPP = 2;
		m_bIsTransparent = TRUE;
		break;
	case MediaFormat :: RGB:
		BPP = 3;
		break;
	case MediaFormat :: RGBA:
		m_bIsTransparent = TRUE;
		pformat = PixelFormat32bppARGB;
		BPP = 4;
		break;
	case MediaFormat :: BGR:
		BPP = 3;
		m_bIsBGR = TRUE;
		break;
	//YUV
	case MediaFormat :: YUVA:
		m_bIsTransparent = TRUE;
	case MediaFormat :: YUV:
		if (m_OriginalFormat == colorFormat) 
			deleteSurface = FALSE;
		isYUV = TRUE;
		break;
	default:
		return FALSE;
	}
	m_OffsetX = offset_x;
	m_OffsetY = offset_y;

	m_pOriginalPixels = (char *) pixel_data;

	//delete offscreen surface
	if (deleteSurface || (m_Width != width) || (m_Height != height)) {
		if (m_pRGBPixels) free(m_pRGBPixels);
		m_pRGBPixels = NULL;
	}

	m_bIsConverted = FALSE;
	m_Width = width;
	m_Height = height;
	m_OriginalFormat = colorFormat;
	m_hstride = horiz_stride;

	if (m_pTexture) delete m_pTexture;
	m_pTexture = NULL;


	if (isYUV) return TRUE;

//setup bitmap
	if (m_pBitmap) delete m_pBitmap;
	m_pBitmap = NULL;

	//GDIplus limitation: 
	//1 - horiz_stride shall be multiple of 4 
	//2 - grayscale and alpha grayscale use RGB format so do the translation
	if ((BPP>2) && !(horiz_stride%4) ) {
		//Gdiplus BUG: this Bitmap constructor reverses RGB , but uses colorMatrix correctly
		if (!m_bIsBGR) {
			m_colorMatrix.m[0][0] = 0.0;
			m_colorMatrix.m[0][2] = 1.0;
			m_colorMatrix.m[2][2] = 0.0;
			m_colorMatrix.m[2][0] = 1.0;
		}
		m_bIsBGR = 0;
		m_pBitmap = new Gdiplus::Bitmap(m_Width, m_Height, horiz_stride, pformat, pixel_data);
	}
	else {
		if (!BPP) return FALSE;
		m_pBitmap = new Gdiplus::Bitmap(m_Width, m_Height, pformat);
		Gdiplus::Color col;
		Gdiplus::ARGB val;
		unsigned char *ptr = pixel_data;
		unsigned int i, j;
		unsigned char a, r, g, b;
		for (j=0; j<m_Height; j++) {
			for (i=0; i<m_Width; i++) {
				switch (BPP) {
				case 1:
					val = Gdiplus::Color::MakeARGB(255, *ptr, *ptr, *ptr);
					ptr ++;
					break;
				case 2:
					r = *ptr;
					ptr ++;
					a = *ptr;
					ptr ++;
					val = Gdiplus::Color::MakeARGB(a, r, r, r);
					break;
				case 3:
					r = *ptr;
					ptr ++;
					g = *ptr;
					ptr ++;
					b = *ptr;
					ptr ++;
					if (m_bIsBGR) {
						val = Gdiplus::Color::MakeARGB(255, b, g, r);
					} else {
						val = Gdiplus::Color::MakeARGB(255, r, g, b);
					}
					break;
				case 4:
					r = *ptr;
					ptr ++;
					g = *ptr;
					ptr ++;
					b = *ptr;
					ptr ++;
					a = *ptr;
					ptr ++;
					val = Gdiplus::Color::MakeARGB(a, r, g, b);
					break;
				default:
					val = Gdiplus::Color::MakeARGB(255, 255, 255, 255);
					break;
				}
				col.SetValue(val);
				m_pBitmap->SetPixel(i, j, col);
			}
		}
		//since we performed the conversion we're now in RGB
		m_bIsBGR = FALSE;
		//our original format is either 24 or 32 bpp RGB
		m_OriginalFormat = (BPP==4) ? MediaFormat :: RGBA : MediaFormat :: RGB;
	}

	return TRUE;
}


//creates an empty offscreen bitmap
BOOL VisualTexture::InitOffscreen(unsigned int width, unsigned int height)
{
	if (!width || !height) return FALSE;
	
	m_bIsTransparent = TRUE;

	//make sure we have identity matrix
	m_colorMatrix.m[0][0] = 1.0;
	m_colorMatrix.m[0][2] = 0.0;
	m_colorMatrix.m[2][2] = 1.0;
	m_colorMatrix.m[2][0] = 0.0;

	m_OffsetX = 0;
	m_OffsetY = 0;
	m_pOriginalPixels = NULL;

	if (m_pRGBPixels) free(m_pRGBPixels);
	m_pRGBPixels = NULL;
	m_bIsConverted = FALSE;
	m_Width = width;
	m_Height = height;
	m_OriginalFormat = MediaFormat :: RGBA;
	m_hstride = 4*width;
	if (m_pTexture) delete m_pTexture;
	m_pTexture = NULL;
	if (m_pBitmap) delete m_pBitmap;
	m_pBitmap = NULL;

	m_bIsBGR = 0;
	m_pBitmap = new Gdiplus::Bitmap(m_Width, m_Height, PixelFormat32bppARGB);

	return TRUE;
}


BOOL VisualTexture::SetTextureCoordinate(float start_x, float start_y)
{
	start_x = start_y = 0;
	return FALSE;
}

void VisualTexture::SetTextureRepeat(BOOL repeat_s, BOOL repeat_t)
{
	m_RepeatS = repeat_s;
	m_RepeatT = repeat_t;
}


VisualTexture *VisualTexture::Clone()
{
	VisualTexture *clone;

	//activate if needed
	if (!m_bIsConverted)
		if (!SetTextureActive()) return NULL;

	if (!m_pBitmap) return NULL;

	clone = new VisualTexture;
	clone->m_Width = m_Width;
	clone->m_Height = m_Height;
	clone->m_pOriginalPixels = NULL;
	clone->m_hstride = m_hstride;
	clone->m_OffsetX = m_OffsetX;
	clone->m_OffsetY = m_OffsetX;
	clone->m_OriginalFormat = m_bIsTransparent ? MediaFormat :: RGBA : MediaFormat :: RGB;
	Gdiplus::PixelFormat pformat = m_bIsTransparent ? PixelFormat32bppARGB : PixelFormat24bppRGB;

	clone->m_pRGBPixels = NULL;
	clone->m_RepeatS = m_RepeatS;
	clone->m_RepeatT = m_RepeatT;
	clone->m_colorMatrix = m_colorMatrix;
	clone->m_bIsBGR = m_bIsBGR;

	//clone returns a CONVERTED image
	clone->m_colorMatrix.m[0][0] = 1.0;
	clone->m_colorMatrix.m[0][2] = 0.0;
	clone->m_colorMatrix.m[2][2] = 1.0;
	clone->m_colorMatrix.m[2][0] = 0.0;
	clone->m_bIsBGR = FALSE;

	clone->m_Matrix.Set(m_Matrix);
	Gdiplus::Rect rc(0, 0, m_Width, m_Height);
	//finally clone the bitmap
	clone->m_pBitmap = m_pBitmap->Clone(rc, pformat);
	return clone;
}

BOOL VisualTexture::SetTextureActive()
{
	if (!m_pOriginalPixels) return FALSE;
	if (m_bIsConverted) return TRUE;

	int BPP = 3;
	Gdiplus::PixelFormat pformat;

	switch (m_OriginalFormat) {
	case MediaFormat :: YUV:
		pformat = PixelFormat24bppRGB;
		break;
	case MediaFormat :: YUVA:
		pformat = PixelFormat32bppARGB;
		BPP = 4;
		break;
	default:
		return TRUE;
	}
	//init mem array
	if (!m_pRGBPixels) {
		m_pRGBPixels = (char *) malloc(sizeof(char) * m_Width * m_Height * BPP);
	}

	unsigned char *pY = (unsigned char *) m_pOriginalPixels;
	unsigned char *pU = pY + m_Width*m_Height;
	unsigned char *pV = pY + 5 * m_Width*m_Height / 4;
	if (BPP==4) {
		unsigned char *pA = pY+ 3 * m_Width*m_Height / 2;
		yuva2rgb_32((unsigned char *) m_pRGBPixels, 4*m_Width, pY, pU, pV, pA, m_Width, m_Width/2, m_Width, m_Height);
	} else {
		yuv2rgb_24((unsigned char *) m_pRGBPixels, 3*m_Width, pY, pU, pV, m_Width, m_Width/2, m_Width, m_Height, 255, NULL);
	}

	//setup bitmap
	if (m_pBitmap) delete m_pBitmap;

	if ((BPP>2) && !(m_hstride%4) ) {
		m_pBitmap = new Gdiplus::Bitmap(m_Width, m_Height, BPP*m_hstride, pformat, (unsigned char *)m_pRGBPixels);
	} else {
		m_pBitmap = new Gdiplus::Bitmap(m_Width, m_Height, pformat);
		Gdiplus::Color col;
		Gdiplus::ARGB val;
		unsigned char *ptr = (unsigned char *)m_pRGBPixels;
		unsigned int i, j;
		unsigned char a, r, g, b;
		for (j=0; j<m_Height; j++) {
			for (i=0; i<m_Width; i++) {
				switch (BPP) {
				case 3:
					r = *ptr;
					ptr ++;
					g = *ptr;
					ptr ++;
					b = *ptr;
					ptr ++;
					val = Gdiplus::Color::MakeARGB(255, r, g, b);
					break;
				case 4:
					r = *ptr;
					ptr ++;
					g = *ptr;
					ptr ++;
					b = *ptr;
					ptr ++;
					a = *ptr;
					ptr ++;
					val = Gdiplus::Color::MakeARGB(a, r, g, b);
					break;
				default:
					val = Gdiplus::Color::MakeARGB(255, 255, 255, 255);
				}
				col.SetValue(val);
				m_pBitmap->SetPixel(i, j, col);
			}
		}
	}

	m_bIsConverted = TRUE;

	//reset color matrix RGB
	m_colorMatrix.m[0][0] = 1.0;
	m_colorMatrix.m[0][2] = 0.0;
	m_colorMatrix.m[2][2] = 1.0;
	m_colorMatrix.m[2][0] = 0.0;
	m_bIsBGR = FALSE;
	return TRUE;
}


unsigned int VisualTexture::GetAlphaBlend()
{
	return (255*m_colorMatrix.m[3][3]);
}

void VisualTexture::SetTextureTransform(Matrix2D *matrix)
{
	if (!matrix) return;
	m_Matrix.SetIdentity();
	m_Matrix.Set(*matrix);
}


Color4 VisualTexture::GetPixel(unsigned int x, unsigned int y)
{
	Color4 ret(0, 0, 0, 0);
	if (!m_pBitmap) return ret;

	Gdiplus::Color col;
	m_pBitmap->GetPixel(x, y, &col);
	ret.g = col.GetG();
	if (m_bIsBGR) {
		ret.b = col.GetR();
		ret.r = col.GetB();
	} else {
		ret.r = col.GetR();
		ret.b = col.GetB();
	}
	ret.a = col.GetA();
	return ret;
}

void VisualTexture::SetPixel(unsigned int x, unsigned int y, Color4 col)
{
	if (!m_pBitmap) return;
	Gdiplus::Color _col;
	Gdiplus::ARGB val;
	if (m_bIsBGR) {
		val = Gdiplus::Color::MakeARGB(col.a, col.r, col.g, col.b);
	} else {
		val = Gdiplus::Color::MakeARGB(col.a, col.b, col.g, col.r);
	}
	_col.SetValue(val);
	m_pBitmap->SetPixel(x, y, _col);
}


BOOL VisualTexture::TranslateColors(float *color_offsets)
{
	if (!m_pBitmap) return FALSE;
	Reset();

	//Note: we don't revert R/B colors since the brush already has the color matrix setup
	m_colorMatrix.m[4][0] = color_offsets[0];
	m_colorMatrix.m[4][1] = color_offsets[1];
	m_colorMatrix.m[4][2] = color_offsets[2];
	m_colorMatrix.m[4][3] = color_offsets[3];
	return TRUE;
}

BOOL VisualTexture::ScaleColors(float *color_scales)
{
	if (!m_pBitmap) return FALSE;
	Reset();

	//Note: we don't revert R/B colors since the brush already has the color matrix setup
	m_colorMatrix.m[0][0] = color_scales[0];
	m_colorMatrix.m[1][1] = color_scales[1];
	m_colorMatrix.m[2][2] = color_scales[2];
	m_colorMatrix.m[3][3] = color_scales[3];
	return TRUE;
}

void VisualTexture::ResetColorsOp()
{
	Reset();
	m_Properties.ClearColorMatrix();
}

char *VisualTexture::LockBits()
{
	//already locked
	if (m_pBitmapData) return (char *)m_pBitmapData->Scan0;

	switch (m_OriginalFormat) {
	//shouldn't be possible
	case MediaFormat :: YUV:
	case MediaFormat :: YUVA:
		//if YUVtoRGB has been performed return new pointer
		if (m_pRGBPixels) return m_pRGBPixels;
		//otherwise return original (shouldn't be NULL)
		return m_pOriginalPixels;
	default:
		break;
	}
	if (m_pOriginalPixels) return m_pOriginalPixels;

	//we should have a bitmap pointer since we're in RGB now
	if (!m_pBitmap) return NULL;

	Gdiplus::Rect rc(m_OffsetX, m_OffsetY, m_Width, m_Height);
	m_pBitmapData = new Gdiplus::BitmapData;
	//lock in read mode in current format
	m_pBitmap->LockBits(&rc, Gdiplus::ImageLockModeRead, m_pBitmap->GetPixelFormat(), m_pBitmapData);

	return (char *)m_pBitmapData->Scan0;
}

void VisualTexture::UnlockBits()
{
	if (m_pBitmapData) {
		m_pBitmap->UnlockBits(m_pBitmapData);
		delete m_pBitmapData;
		m_pBitmapData = NULL;
	}
}

BOOL VisualTexture::HasColorTransform()
{
	//no check on alpha scaling
	if (m_colorMatrix.m[1][1] != 1.0) return TRUE;
	if (m_colorMatrix.m[4][1] != 0.0) return TRUE;
	if (m_colorMatrix.m[4][3] != 0.0) return TRUE;
	if (m_bIsBGR) {
		if (m_colorMatrix.m[0][2] != 1.0) return TRUE;
		if (m_colorMatrix.m[2][0] != 1.0) return TRUE;
		if (m_colorMatrix.m[4][0] != 0.0) return TRUE;
		if (m_colorMatrix.m[4][2] != 0.0) return TRUE;
	} else {
		if (m_colorMatrix.m[0][0] != 1.0) return TRUE;
		if (m_colorMatrix.m[2][2] != 1.0) return TRUE;
		if (m_colorMatrix.m[4][0] != 0.0) return TRUE;
		if (m_colorMatrix.m[4][2] != 0.0) return TRUE;
	}
	if (!m_pOriginalPixels) return TRUE;
	return FALSE;
}

void VisualTexture::SetActiveRect(BRect *rect)
{
	if (m_pVisbleRect) {
		//if same active rectangle, don't destroy brush object
		if (rect) {
			if ( (rect->left==m_pVisbleRect->X) && (rect->top==m_pVisbleRect->Y)
				&& (rect->width==m_pVisbleRect->Width) && (rect->height==m_pVisbleRect->Height) )
				return;
		}
		delete m_pVisbleRect;
	}
	m_pVisbleRect = NULL;
	Invalidate();
	if (!rect) return;

	//this MUST be in our image
	if (rect->left<0) rect->left = 0;
	if (rect->top<0) rect->top = 0;
	if (rect->left + rect->width > m_Width) rect->width = m_Width - rect->left;
	if (rect->top + rect->height > m_Height) rect->height = m_Height - rect->top;

	m_pVisbleRect = new Gdiplus::RectF(rect->left, rect->top, rect->width, rect->height);
}

unsigned int VisualTexture::GetActiveWidth()
{
	if (!m_pVisbleRect) return m_Width;
	return m_pVisbleRect->Width;
}
unsigned int VisualTexture::GetActiveHeight()
{
	if (!m_pVisbleRect) return m_Height;
	return m_pVisbleRect->Height;
}
