/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File: VisualTexture.h
//
// Author: Jean Le Feuvre
//
// Description:
//		Wrapper for texture brush
//
// Created: August 2002
//
///////////////////////////////////////////////////////////////////////////

#include "BRect.h"

#ifndef VISUALTEXTURE_H_
#define VISUALTEXTURE_H_

#include "OsmoseGdiplus.h"
#include "Matrix2D.h"
#include "ColorMatrix.h"
#include "Capabilities.h"


class VisualTexture
{
public:
	VisualTexture();
	~VisualTexture();

	//this shall be called before each draw and will init the phisical brush used for
	//texture mapping. Calling it too often will waste time. It is currently called once 
	//either in GeometryNode::DrawTexture or in Background2D::Draw
	//if colorMatrix is set the brush will change the image color space accordingly
	BOOL SetTransparency(float transparency, ColorMatrix *pMatrix = NULL);
	
	//width, height: size of the texture 
	//pixel_data: pointer to original data
	//horiz_stride: size in bytes of between 2 consecutive line starts
	//colorFormat: color format of the data
	//offset_x, offset_y: texture top-left corner in the image (this is for shape coding)
	BOOL Init(unsigned int width, unsigned int height, unsigned char *pixel_data, unsigned int horiz_stride, MediaFormat colorFormat, unsigned int offset_x=0, unsigned int offset_y=0);
	//creates an empty offscreen bitmap
	BOOL InitOffscreen(unsigned int width, unsigned int height);

	//turns the pixel data into supported texture. This is used because Gdiplus doesn't support YUV textures
	//The function is called only if the texture is to be applied to something else than bitmap node
	BOOL SetTextureActive();

	void SetTextureTransform(Matrix2D *matrix);
	BOOL SetTextureCoordinate(float start_x, float start_y);
	void SetTextureRepeat(BOOL repeat_s, BOOL repeat_t);
	//set the rect that will be used for the brush. NULL means all texture is used 
	//this is given in pixel coordinates, left-top = (0, 0)
	void SetActiveRect(BRect *rect);

	//reset the brush object (this is used by compositeTexture)
	void Invalidate();

	Gdiplus::TextureBrush *m_pTexture;

	unsigned int GetAlphaBlend();

	unsigned int m_Width, m_Height;
	BOOL m_RepeatS, m_RepeatT;

	//original pixel format
	MediaFormat m_OriginalFormat;
	//original pixel array except for clone textures where it is set to NULL
	char *m_pOriginalPixels;
	//original data horizontal stride
	unsigned int m_hstride;
	//transformed RGB24 or RGB32 array if needed (YUV -> RGB texture)
	char *m_pRGBPixels;


	unsigned int m_OffsetX, m_OffsetY;

	//set to true for textures with alpha layer
	BOOL m_bIsTransparent;
	
	//texture transform matrix
	Matrix2D m_Matrix;

	//returns a full copy of the texture. The cloned texture is always activated
	VisualTexture *Clone();
	Color4 GetPixel(unsigned int x, unsigned int y);
	void SetPixel(unsigned int x, unsigned int y, Color4 col);

	//color manipulation routines - floats are 4 float (R, G, B, A)
	BOOL TranslateColors(float *color_offsets);
	BOOL ScaleColors(float *color_scales);
	void ResetColorsOp();

	//Locks/unlocks data ptr in texture.
	//NOTE: it would be nice to optimize rendering of bitmap through GDIplus by 
	//specifying in which format the pixel should be lock (GDIplus::Bitmap feature)
	//for now we only lock in RGB24 or RGBA32 for RGB - TO DO
	char *LockBits();
	void UnlockBits();

	BOOL HasColorTransform();
	BOOL HasValidBrush() {return m_pTexture ? TRUE : FALSE; }

	unsigned int GetActiveWidth();
	unsigned int GetActiveHeight();

//private:
	void Reset();

	Gdiplus::ColorMatrix m_colorMatrix;
	Gdiplus::Bitmap *m_pBitmap;

	BOOL m_bIsConverted, m_bIsBGR;
	Gdiplus::ImageAttributes m_Properties;

	BOOL CreateBrush();
	Gdiplus::BitmapData *m_pBitmapData;

	Gdiplus::RectF *m_pVisbleRect;

	ColorMatrix m_PrevMatrix;
	BOOL m_bHadMatrix;
};


#endif	//VISUALTEXTURE_H_
