/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File: drawable.cpp
//
// Authors:  Jean Le Feuvre - ENST 2002
//
// Description: context methods and drawable nodes.
//
// Created: August 2002.
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes.h"
#include "GroupNodeProxy.h"
#include "drawable.h"
#include "Presenter.h"
#include "Trace.h"

BoundsInfo::BoundsInfo(DrawableContext *pContext)
{
	m_clippedBounds = pContext->m_clippedBoundingRect;
	m_bounds = pContext->m_boundingRect;
	m_pSurface = pContext->m_pVisualSurface;
	m_pAppearance = pContext->m_pAppearance;
}


DrawableNode::DrawableNode(MediaObject* pMO) : Traversable(pMO)
{
	m_LastFlushTime = 0;
	m_bHasChanged = FALSE;
}

DrawableNode::~DrawableNode() 
{
	//we shall never delete a node while rendering
//	assert(m_PreviousBounds.GetLength()==0);
	ResetPreviousBounds();

	//clean all previous instances
	while (m_CurrentBounds.GetLength()) {
		BoundsInfo *bi = m_CurrentBounds[0];
		if (GetPresenter()->IsSurfaceRegistered(bi->m_pSurface)) {
			bi->m_pSurface->m_AreaToClear.Union(bi->m_clippedBounds);
		}
		m_CurrentBounds.DeleteFrom(0);
		delete bi;
	}
	GetPresenter()->RemoveDrawable(this);
}

BOOL DrawableNode::IsPointOver(DrawableContext * /*pDrawableContext*/, float /*x*/, float /*y*/) 
{
	return TRUE; 
}

void DrawableNode::FlushBoundsList(BOOL bReset, VisualSurface *pSurface)
{
	/*reset previous bounds on this surface*/
	for (int i=0; i<m_PreviousBounds.GetLength(); i++) {
		BoundsInfo *bi = m_PreviousBounds[i];
		if (pSurface==bi->m_pSurface) {
			m_PreviousBounds.DeleteFrom(i);
			delete bi;
			i--;
		}
	}

	while (m_CurrentBounds.GetLength()) {
		BoundsInfo *bi = m_CurrentBounds[0];
		m_CurrentBounds.DeleteFrom(0);
		if (bReset) {
			delete bi;
		} else {
			m_PreviousBounds += bi;
		}
	}
	m_bHasChanged = FALSE;
}

void DrawableNode::ResetPreviousBounds()
{
	while (m_PreviousBounds.GetLength()) {
		BoundsInfo *bi = m_PreviousBounds[0];
		m_PreviousBounds.DeleteFrom(0);
		delete bi;
	}
}

void DrawableNode::StoreBounds(DrawableContext *pContext)
{
	if (pContext->m_clippedBoundingRect.IsEmpty()) return;
	BoundsInfo *bi = new BoundsInfo(pContext);
	m_CurrentBounds += bi;
}

BOOL DrawableNode::HasSameBounds(DrawableContext *pContext)
{
	assert(m_CurrentBounds.GetLength());

	for(int i=0; i<m_PreviousBounds.GetLength(); i++) {
		BoundsInfo *bi = m_PreviousBounds[i];
		//must be on the same surface
		if (bi->m_pSurface != pContext->m_pVisualSurface) continue;
		//must have the same appearance (otherwise the same geometry used with
		//different appearances won't be redrawn)
		if (bi->m_pAppearance != pContext->m_pAppearance) continue;
		//and the same not clipped bounding rect
		if (bi->m_bounds != pContext->m_boundingRect) continue;
		if (bi->m_clippedBounds == pContext->m_clippedBoundingRect) {
			m_PreviousBounds.DeleteFrom(i);
			delete bi;
			return TRUE;
		}
	}
	return FALSE;
}

BOOL DrawableNode::GetPreviousBound(BRect *rc, VisualSurface *pSurface)
{
	for(int i=0; i<m_PreviousBounds.GetLength(); i++) {
		BoundsInfo *bi = m_PreviousBounds[i];
		if (bi->m_pSurface != pSurface) continue;
		*rc = bi->m_clippedBounds;
		m_PreviousBounds.DeleteFrom(i);
		delete bi;
		return TRUE;
	}
	return FALSE;
}


void DrawableNode::FinalizeRender(Effects *effects, DrawableContext *pContext)
{
	if(effects->m_ParentGroup) {
		effects->m_ParentGroup->AddToDrawableContextList(pContext);
	} else if (!effects->m_bUseBoundTracking) {
		Draw(pContext);
	}
}


DrawableContext::DrawableContext()
{
	m_pColorMatrix = NULL;
	m_pClipper = NULL;
}


DrawableContext::~DrawableContext()
{
	if (m_pColorMatrix) delete m_pColorMatrix;
}


void DrawableContext::AddUserEventHandler(SensorContext *handler)
{
	if (!handler) return;

	if (m_Handlers.Find(handler)<0) {
		m_Handlers += new SensorContext(handler->m_pHandler, &handler->m_Matrix);
	}
}

void DrawableContext::ResetHandlers()
{
	while (m_Handlers.GetLength()) {
		SensorContext *ptr = m_Handlers[0];
		m_Handlers.DeleteFrom(0);
		delete ptr;
	}
}

void DrawableContext::UpdateDrawInfo()
{
	BOOL moved;
	//will draw if needed
	m_bNeedsDrawing = m_pDrawable->HasChanged(this);
	//if not, will draw if we can't find the same bounds
	moved = !m_pDrawable->HasSameBounds(this);
	if (!m_bNeedsDrawing) m_bNeedsDrawing = moved;
}

BOOL DrawableContext::IsPointInBounds(float x, float y)
{
	return m_clippedBoundingRect.IsPointIn(x, y);
}

void DrawableContext::Reset()
{
	m_bUpdateTexture = FALSE;
	m_hasStreamFinished = FALSE;
	m_bIsTransparent = TRUE;
	m_isTextNode = FALSE;
	m_boundingRect.MakeEmpty();
	m_clippedBoundingRect.MakeEmpty();
	m_bNeedsDrawing = FALSE;
	m_bCanSkipRedraw = FALSE;

	ResetHandlers();

	m_pDrawable = NULL;
	m_pAppearance = NULL;
	m_pGradient = NULL;
	m_pTextureProxy = NULL;
	m_pVisualSurface = NULL;
	if (m_pColorMatrix) {
		delete m_pColorMatrix;
		m_pColorMatrix = NULL;
	}

	if (m_pClipper) {
		delete m_pClipper;
		m_pClipper = NULL;
	}

	//setup default aspect
	memset(&gProps, 0, sizeof(gProps));
	gProps.emissiveColor = Color3f(0.8, 0.8, 0.8);
	gProps.lineWidth = 1.0;
	gProps.lineColor = Color3f(0.8, 0.8, 0.8);

	gProps.lineCap = LINE_CAP_FLAT;
	gProps.lineJoin = LINE_JOIN_BEVEL;
	//default is 4.0
	gProps.miterLimit = 4.0;

}