/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  drawable.h
//
// Author:  Jean Le Feuvre ENST 2002
//
// Description:
//   drawable node and drawable context definition
//
// Created: September 2002
//
///////////////////////////////////////////////////////////////////////////


#ifndef DRAWABLE_H
#define DRAWABLE_H

#include "BRect.h"
#include "Matrix2D.h"
#include "Traversable.h"
#include "ztl.h"

class ShapeProxy;
class Appearance;
class GenericGradient;
class UserEventHandlerProxy;
class DrawableImage;
class DrawableContext;

class BoundsInfo
{
public:
	BoundsInfo(DrawableContext *pContext);

	BRect m_clippedBounds;
	BRect m_bounds;
	VisualSurface *m_pSurface;
	Appearance *m_pAppearance;
};

class DrawableNode : public Traversable
{
public:

	DrawableNode(MediaObject* pMO);
	~DrawableNode();
	
	//draw the node in the given context
	virtual void Draw(DrawableContext *pContext) = 0;

	// Return TRUE if the given point, in BIFS coordinates, is above the
	// current shape.  Return FALSE otherwise. This functions is only called when the point is
	//already over the shape's bounding rect
	virtual BOOL IsPointOver(DrawableContext *pContext, float x, float y);	

	//store the final bound in the given context
	void StoreBounds(DrawableContext *pContext);

	//moves all bounds in the current list to the previous list
	//if bReset is TRUE the current bounds are all reseted (forces the node to redraw)
	void FlushBoundsList(BOOL bReset, VisualSurface *pSurface);

	//returns TRUE if the same clipped rect was present in the last rendering step
	BOOL HasSameBounds(DrawableContext *pContext);

	//returns TRUE while there are some bound info left from the previous step
	//pSurface is used to discriminate between drawing surfaces
	BOOL GetPreviousBound(BRect *rc, VisualSurface *pSurface);

	//returns TRUE if the node shape/appearance has changed. 
	//geometry nodes override this function
	virtual BOOL HasChanged(DrawableContext * /*pContext*/) {return FALSE;}

	void ResetPreviousBounds();

protected:
	void FinalizeRender(Effects *effects, DrawableContext *pContext);
	BOOL m_bHasChanged;

private:

	//bound tracking structures.
	ZArray<BoundsInfo *> m_CurrentBounds;
	ZArray<BoundsInfo *> m_PreviousBounds;

	//used to reset the tracking list
	DWORD m_LastFlushTime;
};

//this is used to store colors, line width and co
typedef struct
{
	//appearance info
	Color3f emissiveColor;
	float transparency;
	BOOL filled;

	//line info
	int lineStyle;
	Color3f lineColor;
	float lineWidth;
	float lineTransparency;
	BOOL lineScaled;
	BOOL lineInside;
	int lineCap;
	int lineJoin;
	float miterLimit;
	BOOL hasLineProperties;

} GraphicsProperties;


class DrawableContext
{
public:
	DrawableContext();
	~DrawableContext();

	void Reset();
	  	
	//set to TRUE if texture has been updated
	BOOL m_bUpdateTexture;
	BOOL m_hasStreamFinished;
	
	//only nodes covering entirely their bounds are not transparent. This is always set to
	//TRUE at init
	BOOL m_bIsTransparent;
	
	// NOTE The bounding rectangles uses the BIFS coordinate system, where the origin of the output display is 
	//in the middle of the window, with increasing X and Y to the right and up.

	//clipped bounding rect
	BRect m_clippedBoundingRect;
	//bounding rect
	BRect m_boundingRect;

	BOOL IsPointInBounds(float x, float y);
	
	GraphicsProperties gProps;

	//absolute transform matrix in BIFS coordinates
	Matrix2D m_transform;

	void AddUserEventHandler(SensorContext *handler);
	void ResetHandlers();
	ZArray<SensorContext *> m_Handlers;

	VisualSurface *m_pVisualSurface;

	VisualClipper *m_pClipper;

	//used by text and layout
	BOOL m_isTextNode;
	
	//update draw info will decide if the node instance should be drawn or node
	//the result is stored in m_bNeedsDrawing
	void UpdateDrawInfo();
	BOOL m_bNeedsDrawing;
	BOOL m_bCanSkipRedraw;


	DrawableNode *m_pDrawable;
	Appearance *m_pAppearance;
	DrawableImage *m_pTextureProxy;
	GenericGradient *m_pGradient;

	ColorMatrix *m_pColorMatrix;
};	

#endif	//DRAWABLE_H