// -*- c++ -*-

// This software module was originally developed by Matthew Leditschke
// (CSELT), Zvi Lifshitz (VDOnet Corp.) and Karl Anders ygard (Telenor
// R&D) in the course of development of the MPEG-4 Systems standard
// (ISO/IEC 14496-1).  This software module is an implementation of a part
// of one or more MPEG-4 Systems tools as specified by the MPEG-4 Systems
// standard.  ISO/IEC gives users of the MPEG-4 Systems standard free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4 Systems
// standard.  Those intending to use this software module in hardware or
// software products are advised that its use may infringe existing
// patents.  The original developer of this software module and his/her
// company, the subsequent editors and their companies, and ISO/IEC have no
// liability for use of this software module or modifications thereof in an
// implementation.  Copyright is not released for non MPEG-4 Systems
// conforming products.  The original developer retains full right to use
// the code for his/her own purpose, assign or donate the code to a third
// party and to inhibit third parties from using the code for non MPEG-4
// Systems conforming products.  This copyright notice must be included in
// all copies or derivative works.
//
// For more information or to receive an up to date version of this module,
// mail karlo@nta.no or check <http://televr.fou.telenor.no/compositor/>.
//
// Copyright 1997, 1998.

#ifndef __PRESENTER_H__
#define __PRESENTER_H__

class Executive;
class VisualRenderer;
class AudioRenderer;
class Capabilities;

// class Root;
#include "Root.h"

#include "NodeFld.h"
#include "Im1defs.h"

// SOLARIS PORTING Core 5.7 rc1
// Stefano Battista 98/11/17
#include "PresBase.h"
//#include "../FlexTime/PresenterFlex.h"

// SB @@@@ 99/01/22
#include "Nodes.h"

//added by Helge Drumm TU Ilmenau, 6.12.2006, Shadow Rendering
#include "../SGManager/SGManager.h"


#ifdef MPGJ
//MPEGJ - MERGE
#include "../mpegj-player/MPEGJTerminalInterface.h"
	// for pointer table
// END - MPEGJ - PA
#endif // MPEGJ

/** Presenter.

    This class runs in its own thread and controls the scene presentation.
    The object itself only provides the thread and the timing, while the
    presentation hard work is done by the \Ref{MediaObject}s and the
    \Ref{Renderer}s.  This works as following:

    \begin{enumerate}

    \item The \Ref{Executive} instantiates the \Ref{Presenter}

    \item The \Ref{Presenter} instantiates the visual and audio renderers

    \item When \Ref{SceneConstructor} has finished constructing the scene
      out of the BIFS, the \Ref{Presenter}'s Init() function is called.
      This function calls the initialisation functions of the
      \Ref{Renderer}s, and starts the \Ref{Presenter}'s thread.

    \item The \Ref{Presenter}'s thread runs in a loop, which every x
      milliseconds calls the Render() function of the scene's root object.

    \item Each \Ref{MediaObject} renders itself and its child nodes.

    \item At the end, the \Ref{Presenter}'s Terminate() function is called.
      This performs cleanup stuff, like erasing the window, and terminates
      the \Ref{Presenter}'s thread.

    \end{enumerate} */

// SB @@@@ cannot use pure virtual methods

// SOLARIS PORTING - to Core 5.7rc1
class Presenter : public PresenterBase
{
  // SB @@@@ variables from Zvi's code 98/11/22 >>>>
private:
	DWORD m_dwIterations, m_dwOffset;
	BOOL m_bClockActive;
	DWORD m_dwStartTime;

	//ML - added the next four lines
	WorldInfo* m_pWorldInfo;
// WIN-SOL release
#ifdef WIN32
	CString* m_pWorldInfoTitle;
	CString* m_pWorldInfoInfo;
#else
	char* m_pWorldInfoTitle;
	char* m_pWorldInfoInfo;
#endif

    void ExtractWorldInfo();

protected:
	FILE *m_fLogFile;
  // SB @@@@ variables from Zvi's code 98/11/22 <<<<

public:

	#ifdef MPGJ
	// MPEGJ - MERGE
	/*** handle to the function pointer  ***/
	MPJTerminalDispatchTablePtr di;
	/**** - end MPEGJ - PA */
	#endif

  /// Constructor

   //added by Helge Drumm TU Ilmenau, 6.12.2006, Shadow Rendering
   // SGManager handle pointer; 
   SGManager *sgm; 



// SOLARIS PORTING - changed PresenterBase to PresenterFlex
  Presenter(Executive* pExecutive, HWND hWnd)
    : PresenterBase(pExecutive, hWnd),
      m_pExecutive(pExecutive),
	  m_pWorldInfo(NULL),
	  m_pWorldInfoTitle(NULL),
	  m_pWorldInfoInfo(NULL)
  {
	#ifdef MPGJ
	// MPEGJ - MERGE
	/***** load function table */
	di = MPJGetTerminalDispatchTable();
	// END - MPEGJ - PA
	#endif // MPEGJ
	
	//added by Helge Drumm TU Ilmenau, 6.12.2006, Shadow Rendering
	// SGManger create object, 
	sgm = new SGManager();

  }

  /// Destructor

  virtual ~Presenter()
  {
  }

  /// Get pointer to visual renderer
  // pure virtual in Presenter
  virtual VisualRenderer * GetVisualRenderer() const /* = 0 */
  { return NULL; };

  /// Get pointer to audio renderer
  // pure virtual in Presenter
  virtual AudioRenderer * GetAudioRenderer() const /* = 0 */
  { return NULL; };

  /// Get pointer to scene
  inline Root* GetScene() const
  {
    return GetExecutive()->GetScene();
  }

  /// Get pointer to executive
  inline Executive*	GetExecutive() const
  {
    return m_pExecutive;
  }

  /***************************************************************************************/
  // SB @@@@ pure virtual methods inherited from PresenterBase >>>>

  /// Get capabilities of presenter
  // pure virtual in PresenterBase
  virtual void GetCapabilities(Capabilities *pPresentationCapabilities);

  /// Handle User Events
  // fake: pure virtual in PresenterBase
  virtual void OnUserEvent(
    USER_EVENT /* oEvent */,
    int /* x */, int /* y */,
    UINT /* uiControlKeyStatus */) {};

  /// Get current time of presenter
  // pure virtual in PresenterBase
  virtual DWORD GetTime() /* const = 0 */
  {
#ifndef BIFSENC
    Root * poRoot = m_pExecutive->GetScene();
    if (poRoot != NULL)
      return poRoot->GetTime();
    else
      return 0;
#else
    return 0;
#endif
  };

  /// Get iteration interval
  // pure virtual in PresenterBase
  virtual int GetIterationInterval() { return 1; }

  /// Render
  virtual void Render() {};

  // SB @@@@ pure virtual methods inherited from PresenterBase <<<<
  /***************************************************************************************/

  // SB @@@@ methods from Zvi's code 98/11/22 >>>>
	DWORD GetClockTime ();
	long GetClockOffset () {return (long) m_dwOffset;}
	void SetClock (BOOL bOn);
	BOOL ClockActive () {return m_bClockActive;}
  // SB @@@@ methods from Zvi's code 98/11/22 <<<<

  //ML - moved the bodies of the next four function to Presenter.cpp
  virtual void Init();
  virtual void Start();
  virtual BOOL Stop();
  virtual void Terminate();

  // TraverseSG, traverse the scenegraph and create all NodeReference objects 
  //added by Helge Drumm TU Ilmenau, 6.12.2006, Shadow Rendering

  MediaObject* TraverseSG(int iNodeCode,MediaObject* pStartNode,RBaseList *refList); 
  
  
  /// Paint
  // pure virtual in Presenter
  virtual void		Paint(void * /* rectangle */) /* = 0 */ {};

  /// Resize the display
  virtual void		ReSize(int /* nWidth */, int /* nHeight */) /* = 0 */ {};

  /// Output message
  virtual void		Logf(LPCSTR szFormat, ...) /* = 0 */ ;

  /// Handle mouse events
  virtual void		OnMouseEvent(
    USER_EVENT /* event */,
    int /* nX */,
    int /* nY */,
    UINT /* uModifierState */) /* = 0 */ {};

  //ML - added the next three functions
// WIN-SOL release
#ifdef WIN32
  const CString* GetWorldInfoTitle() const {return m_pWorldInfoTitle;}
  const CString* GetWorldInfoInfo() const  {return m_pWorldInfoInfo;}
#else
  const char* GetWorldInfoTitle() const {return m_pWorldInfoTitle;}
  const char* GetWorldInfoInfo() const  {return m_pWorldInfoInfo;}
#endif

  void SetWorldInfo(WorldInfo* pWorldInfo);


  /***************************************************************************************/
  /// methods to initialize rendering Nodes within the Presenter

  //ML - change InitCLASSNAME to CreateCLASSNAMEProxy

  // init MediaObject
  virtual RenderingObject* CreateMediaObjectProxy(MediaObject * pObject);

  // Create GroupNode (base for grouping Nodes)
  //ML - removed this function
  // virtual void		InitGroupNode(GroupNode * pObject);

  //  StreamConsumer (base for streaming Nodes)
  virtual RenderingObject* CreateStreamConsumerProxy(StreamConsumer * pObject);

  // init Node
  // virtual  void		InitNode(Node * pObject);

  // init Node3D
  // virtual  void		InitNode3D(Node3D * pObject);

  // Create Inline

  virtual RenderingObject* CreateInlineProxy(Inline* pObject);

  // Create Switch

  virtual RenderingObject* CreateSwitchProxy(Switch* pObject);

  // init Group

  virtual RenderingObject* CreateGroupProxy(Group * pObject);

  // init Transform

  virtual RenderingObject* CreateTransformProxy(Transform* pObject);

  // init Billboard

  virtual RenderingObject* CreateBillboardProxy(Billboard* pObject);

  // init Anchor

  virtual RenderingObject* CreateAnchorProxy(Anchor* pObject);

  // init Collision

  virtual RenderingObject* CreateCollisionProxy(Collision* pObject);

  // init Layout

  virtual RenderingObject* CreateLayoutProxy(Layout* pObject);

  // init Form

  virtual RenderingObject* CreateFormProxy(Form* pObject);

  // init LOD

  virtual RenderingObject* CreateLODProxy(LOD* pObject);

  // init Node2D
  // virtual  void		InitNode2D(Node2D* pObject);

  // init Conditional

  virtual RenderingObject* CreateConditionalProxy(Conditional* pObject);

  // init Inline2D

#ifdef PRE_FDIS
  virtual RenderingObject* CreateInline2DProxy(Inline2D* pObject);
#endif // PRE_FDIS

  // init Switch2D

#ifdef PRE_FDIS
  virtual RenderingObject* CreateSwitch2DProxy(Switch2D* pObject);
#endif // PRE_FDIS

  // init Group2D

#ifdef PRE_FDIS
  virtual RenderingObject* CreateGroup2DProxy(Group2D* pObject);
#endif // PRE_FDIS

  // init Transform2D

  virtual RenderingObject* CreateTransform2DProxy(Transform2D* pObject);

  // init TextureNode2D
  // virtual void		InitTextureNode2D(TextureNode2D* pObject);

  // init Anchor2D

#ifdef PRE_FDIS
  virtual RenderingObject* CreateAnchor2DProxy(Anchor2D* pObject);
#endif // PRE_FDIS

  // init VideoObject2D

#ifdef PRE_FDIS
  virtual RenderingObject* CreateVideoObject2DProxy(VideoObject2D* pObject);
#endif // PRE_FDIS

  // init Image2D

#ifdef PRE_FDIS
  virtual RenderingObject* CreateImage2DProxy(Image2D* pObject);
#endif // PRE_FDIS

  // init Bindable
  // virtual void		InitBindable(Bindable* pObject);

  // init Viewpoint

  virtual RenderingObject* CreateViewpointProxy(Viewpoint* pObject);

  // init WorldInfo

  virtual RenderingObject* CreateWorldInfoProxy(WorldInfo* pObject);

  // init Fog

  virtual RenderingObject* CreateFogProxy(Fog* pObject);

  // init Background

  virtual RenderingObject* CreateBackgroundProxy(Background* pObject);

  // init NavigationInfo

  virtual RenderingObject* CreateNavigationInfoProxy(NavigationInfo* pObject);

  // init InterpolatorNode
  // virtual void		InitInterpolatorNode(InterpolatorNode* pObject);

  // init AnimationStream

  virtual RenderingObject* CreateAnimationStreamProxy(AnimationStream* pObject);

  // init TermCap

  virtual RenderingObject* CreateTermCapProxy(TermCap* pObject);

  // init DragSensor
  // virtual void		InitDragSensor(DragSensor* pObject);

  // init PlaneSensor

  virtual RenderingObject* CreatePlaneSensorProxy(PlaneSensor* pObject);

  // init PlaneSensor2D

  virtual RenderingObject* CreatePlaneSensor2DProxy(PlaneSensor2D* pObject);

  // init SphereSensor

  virtual RenderingObject* CreateSphereSensorProxy(SphereSensor* pObject);

  // init CylinderSensor

  virtual RenderingObject* CreateCylinderSensorProxy(CylinderSensor* pObject);

  // init DiscSensor

  virtual RenderingObject* CreateDiscSensorProxy(DiscSensor* pObject);

  // init TouchSensor

  virtual RenderingObject* CreateTouchSensorProxy(TouchSensor* pObject);

  // init TimeSensor

  virtual RenderingObject* CreateTimeSensorProxy(TimeSensor* pObject);

  // init ProximitySensor

  virtual RenderingObject* CreateProximitySensorProxy(ProximitySensor* pObject);

  // init VisibilitySensor

  virtual RenderingObject* CreateVisibilitySensorProxy(VisibilitySensor* pObject);

  // init Appearance

  virtual RenderingObject* CreateAppearanceProxy(Appearance* pObject);

  // init MaterialNode
  // virtual void		InitMaterialNode(MaterialNode* pObject);

  // init Material

  virtual RenderingObject* CreateMaterialProxy(Material* pObject);

  // init TextureNode
  // virtual void		InitTextureNode(TextureNode* pObject);

  // init ImageTexture

  virtual RenderingObject* CreateImageTextureProxy(ImageTexture* pObject);

  // init MovieTexture

  virtual RenderingObject* CreateMovieTextureProxy(MovieTexture* pObject);

  // init PixelTexture

  virtual RenderingObject* CreatePixelTextureProxy(PixelTexture* pObject);

  // init TextureTransform

  virtual RenderingObject* CreateTextureTransformProxy(TextureTransform* pObject);

  // init FontStyle

  virtual RenderingObject* CreateFontStyleProxy(FontStyle* pObject);

  // init Material2D

  virtual RenderingObject* CreateMaterial2DProxy(Material2D* pObject);

  // init LineProperties

  virtual RenderingObject* CreateLinePropertiesProxy(LineProperties* pObject);

  // init Coordinate

  virtual RenderingObject* CreateCoordinateProxy(Coordinate* pObject);

  // init Coordinate2D

  virtual RenderingObject* CreateCoordinate2DProxy(Coordinate2D* pObject);

  // init Normal

  virtual RenderingObject* CreateNormalProxy(Normal* pObject);

  // init TextureCoordinate

  virtual RenderingObject* CreateTextureCoordinateProxy(TextureCoordinate* pObject);

  // init Color

  virtual RenderingObject* CreateColorProxy(Color* pObject);

  // init Shape

  virtual RenderingObject* CreateShapeProxy(Shape* pObject);

  // init Geometry
  // virtual void		InitGeometry(Geometry* pObject);

  // init Box

  virtual RenderingObject* CreateBoxProxy(Box* pObject);

  // init Sphere

  virtual RenderingObject* CreateSphereProxy(Sphere* pObject);

  // init Cylinder

  virtual RenderingObject* CreateCylinderProxy(Cylinder* pObject);

  // init Cone

  virtual RenderingObject* CreateConeProxy(Cone* pObject);

  // init IndexedFaceSet

  virtual RenderingObject* CreateIndexedFaceSetProxy(IndexedFaceSet* pObject);

  // init IndexedLineSet

  virtual RenderingObject* CreateIndexedLineSetProxy(IndexedLineSet* pObject);

  // init PointSet

  virtual RenderingObject* CreatePointSetProxy(PointSet* pObject);

  // init StreamingText

#ifdef PRE_FDIS
  virtual RenderingObject* CreateStreamingTextProxy(StreamingText* pObject);
#endif // PRE_FDIS

  // init Text

  virtual RenderingObject* CreateTextProxy(Text* pObject);

  // init Circle

  virtual RenderingObject* CreateCircleProxy(Circle* pObject);

  // init Rectangle

  virtual RenderingObject* CreateRectangleProxy(class Rectangle* pObject);

  // init IndexedFaceSet2D

  virtual RenderingObject* CreateIndexedFaceSet2DProxy(IndexedFaceSet2D* pObject);

  // init IndexedLineSet2D

  virtual RenderingObject* CreateIndexedLineSet2DProxy(IndexedLineSet2D* pObject);

  // init PointSet2D

  virtual RenderingObject* CreatePointSet2DProxy(PointSet2D* pObject);

  // init ElevationGrid

  virtual RenderingObject* CreateElevationGridProxy(ElevationGrid* pObject);

  // init Extrusion

  virtual RenderingObject* CreateExtrusionProxy(Extrusion* pObject);

  // init Light
  // virtual void		InitLight(Light* pObject);

  // init DirectionalLight

  virtual RenderingObject* CreateDirectionalLightProxy(DirectionalLight* pObject);

  // init PointLight

  virtual RenderingObject* CreatePointLightProxy(PointLight* pObject);

  // init SpotLight

  virtual RenderingObject* CreateSpotLightProxy(SpotLight* pObject);

  // init Face

  virtual RenderingObject* CreateFaceProxy(Face* pObject);

  // init FIT

  virtual RenderingObject* CreateFITProxy(FIT* pObject);

  // init FDP

  virtual RenderingObject* CreateFDPProxy(FDP* pObject);

  // init FAP

  virtual RenderingObject* CreateFAPProxy(FAP* pObject);

  // init Viseme

  virtual RenderingObject* CreateVisemeProxy(Viseme* pObject);

  // init Expression

  virtual RenderingObject* CreateExpressionProxy(Expression* pObject);

  // init FaceDefTables

  virtual RenderingObject* CreateFaceDefTablesProxy(FaceDefTables* pObject);

  // init FaceDefMesh

  virtual RenderingObject* CreateFaceDefMeshProxy(FaceDefMesh* pObject);

  // init FaceDefTransform

  virtual RenderingObject* CreateFaceDefTransformProxy(FaceDefTransform* pObject);

   // init Sound

  virtual RenderingObject* CreateSoundProxy(Sound* pObject);

  // init Sound2D

  virtual RenderingObject* CreateSound2DProxy(Sound2D* pObject);

  // init AudioObject

  // virtual void		InitAudioObject(AudioObject* pObject);

  // init AudioSource

  virtual RenderingObject* CreateAudioSourceProxy(AudioSource* pObject);

  // init AudioSwitch

  virtual RenderingObject* CreateAudioSwitchProxy(AudioSwitch* pObject);

  // init AudioFX

  virtual RenderingObject* CreateAudioFXProxy(AudioFX* pObject);

  // init AudioMix

  virtual RenderingObject* CreateAudioMixProxy(AudioMix* pObject);

  // init AudioDelay

  virtual RenderingObject* CreateAudioDelayProxy(AudioDelay* pObject);

  // init Script

  virtual RenderingObject* CreateScriptProxy(Script* pObject);

  // Delete object

  virtual void		DeleteObject(MediaObject* pObject);

	// **************************************************************************
public:

  // SB @@@@ added on 99/01/20 for new Nodes >>>>

  virtual RenderingObject* CreateAudioBufferProxy(AudioBuffer * poNode);
  virtual RenderingObject* CreateLayer2DProxy(Layer2D * poNode);
  virtual RenderingObject* CreateLayer3DProxy(Layer3D * poNode);
  virtual RenderingObject* CreateOrderedGroupProxy(OrderedGroup * poNode);

  virtual RenderingObject* CreateAudioClipProxy(AudioClip * poNode);
  virtual RenderingObject* CreateCompositeTexture2DProxy(CompositeTexture2D * poNode);
  virtual RenderingObject* CreateCompositeTexture3DProxy(CompositeTexture3D * poNode);

  virtual RenderingObject* CreateBackground2DProxy(Background2D * poNode);
  virtual RenderingObject* CreateBitmapProxy(Bitmap * poNode);
  virtual RenderingObject* CreateColorInterpolatorProxy(ColorInterpolator * poNode);
  virtual RenderingObject* CreateCoordinateInterpolatorProxy(CoordinateInterpolator * poNode);
  virtual RenderingObject* CreateCoordinateInterpolator2DProxy(CoordinateInterpolator2D * poNode);
  virtual RenderingObject* CreateCurve2DProxy(Curve2D * poNode);
  virtual RenderingObject* CreateListeningPointProxy(ListeningPoint * poNode);
  virtual RenderingObject* CreateNormalInterpolatorProxy(NormalInterpolator * poNode);
  virtual RenderingObject* CreateOrientationInterpolatorProxy(OrientationInterpolator * poNode);
  virtual RenderingObject* CreatePositionInterpolatorProxy(PositionInterpolator * poNode);
  virtual RenderingObject* CreatePositionInterpolator2DProxy(PositionInterpolator2D * poNode);
  virtual RenderingObject* CreateProximitySensor2DProxy(ProximitySensor2D * poNode);
  virtual RenderingObject* CreateQuantizationParameterProxy(QuantizationParameter * poNode);
  virtual RenderingObject* CreateScalarInterpolatorProxy(ScalarInterpolator * poNode);
  virtual RenderingObject* CreateValuatorProxy(Valuator * poNode);

  // SB @@@@ added on 99/01/20 for new Nodes <<<<


  //ML - removed the following descriptor nodes
/*
  // SB @@@@ added on 99/01/20 for descriptors Nodes >>>>
  virtual void InitObjectDescriptor(MediaObject * poNode);
  virtual void InitES_Descriptor(MediaObject * poNode);
  virtual void InitDecoderConfigDescriptor(MediaObject * poNode);
  virtual void InitBIFSDecoderConfig(MediaObject * poNode);
  virtual void InitIPI_Descriptor(MediaObject * poNode);
  virtual void InitIP_InformationDataSet(MediaObject * poNode);
  virtual void InitSupplementaryContentIdentifier(MediaObject * poNode);
  virtual void InitQoS_Descriptor(MediaObject * poNode);
  virtual void InitQoS_Qualifier(MediaObject * poNode);
  virtual void InitExtensionDescriptor(MediaObject * poNode);
  virtual void InitShortExtensionDescriptor(MediaObject * poNode);
  virtual void InitIPMP_DescriptorPointer(MediaObject * poNode);
  virtual void InitALConfigDescriptor(MediaObject * poNode);
  virtual void InitIPMPDescriptor(MediaObject * poNode);
  // SB @@@@ added on 99/01/20 for descriptors Nodes <<<<
*/

#ifdef ADV_AUDIO
  virtual RenderingObject* CreateAcousticMaterialProxy(AcousticMaterial * poNode);
  virtual RenderingObject* CreateAcousticSceneProxy(AcousticScene * poNode);
  virtual RenderingObject* CreateDirectiveSoundProxy(DirectiveSound * poNode);
  virtual RenderingObject* CreatePerceptualParametersProxy(PerceptualParameters * poNode);
#endif // ADV_AUDIO

  //ML - added the following nodes
  virtual RenderingObject* CreateBAPProxy(BAP*);
  virtual RenderingObject* CreateBDPProxy(BDP*);
  virtual RenderingObject* CreateBodyProxy(Body*);
  virtual RenderingObject* CreateBodyDefTableProxy(BodyDefTable*);
  virtual RenderingObject* CreateBodySegmentConnectionHintProxy(BodySegmentConnectionHint*);
  virtual RenderingObject* CreateHierarchical3DMeshProxy(Hierarchical3DMesh*);
  virtual RenderingObject* CreateMaterialKeyProxy(MaterialKey*);
  virtual RenderingObject* CreateServerCommandProxy(ServerCommand*);

  virtual RenderingObject* CreateTemporalTransformProxy(TemporalTransform*);
  virtual RenderingObject* CreateTemporalGroupProxy(TemporalGroup*);
  virtual RenderingObject* CreateMediaSensorProxy(MediaSensor*);
  virtual RenderingObject* CreateMediaControlProxy(MediaControl*);

  virtual RenderingObject* CreateApplicationWindowProxy(ApplicationWindow*);

  // SOLARIS PORTING - Core 5.7rc1 - Added these three methods
  virtual RenderingObject* CreateInputSensorProxy(InputSensor*);
  virtual RenderingObject* CreateMatteTextureProxy(MatteTexture*);
  virtual RenderingObject* CreateMediaBufferProxy(MediaBuffer*);

  // SOLARIS PORTING - seems to have been omitted
  //virtual RenderingObject* CreateParticleInitConeProxy(ParticleInitCone *);
  // SOLARIS PORTING - Core 5.9rc9 - END of new nodes

  // SOLARIS PORTING - Merging code with 2D player
  virtual RenderingObject* CreateClipper2DProxy(Clipper2D *);
  virtual RenderingObject* CreateColorTransformProxy(ColorTransform *);
  virtual RenderingObject* CreateTransformMatrix2DProxy(TransformMatrix2D *);
  virtual RenderingObject* CreateViewportProxy(Viewport *);
  virtual RenderingObject* CreateXCurve2DProxy(XCurve2D *);
  virtual RenderingObject* CreateXLinePropertiesProxy(XLineProperties *);

// **************************************************************************

	// AFX nodes
  virtual RenderingObject* CreateCoordinateInterpolator4DProxy(CoordinateInterpolator4D *);
  virtual RenderingObject* CreatePositionInterpolator4DProxy(PositionInterpolator4D *);
	virtual RenderingObject* CreateBitWrapperProxy(BitWrapper *);
  virtual RenderingObject* CreateDepthImageProxy(DepthImage *);
  virtual RenderingObject* CreateFFDProxy(FFD *);
  virtual RenderingObject* CreateMeshGridProxy(MeshGrid *);
  virtual RenderingObject* CreateNonLinearDeformerProxy(NonLinearDeformer *);
  virtual RenderingObject* CreateNurbsCurveProxy(NurbsCurve *);
  virtual RenderingObject* CreateNurbsCurve2DProxy(NurbsCurve2D *);
  virtual RenderingObject* CreateNurbsSurfaceProxy(NurbsSurface *);
  virtual RenderingObject* CreateOctreeImageProxy(OctreeImage *);
  virtual RenderingObject* CreatePointTextureProxy(PointTexture *);
  virtual RenderingObject* CreatePositionAnimatorProxy(PositionAnimator *);
  virtual RenderingObject* CreatePositionAnimator2DProxy(PositionAnimator2D *);
  virtual RenderingObject* CreateProceduralTextureProxy(ProceduralTexture *);
  virtual RenderingObject* CreateSBBoneProxy(SBBone *);
  virtual RenderingObject* CreateSBMuscleProxy(SBMuscle *);
  virtual RenderingObject* CreateSBSegmentProxy(SBSegment *);
  virtual RenderingObject* CreateSBSiteProxy(SBSite *);
  virtual RenderingObject* CreateSBSkinnedModelProxy(SBSkinnedModel *);
  virtual RenderingObject* CreateSBVCAnimationProxy(SBVCAnimation *);
  virtual RenderingObject* CreateScalarAnimatorProxy(ScalarAnimator *);
  virtual RenderingObject* CreateSimpleTextureProxy(SimpleTexture *);
  virtual RenderingObject* CreateSubdivisionSurfaceProxy(SubdivisionSurface *);
  virtual RenderingObject* CreateSubdivSurfaceSectorProxy(SubdivSurfaceSector *);
  virtual RenderingObject* CreateWaveletSubdivisionSurfaceProxy(WaveletSubdivisionSurface *);

	// removed AFX nodes per 68th meeting resolutions
  virtual RenderingObject* CreateImplicitProxy(Implicit *) ;
  RenderingObject* CreateXXLFM_AppearanceProxy(XXLFM_Appearance *) { return NULL; }
  RenderingObject* CreateXXLFM_BlendListProxy(XXLFM_BlendList *) { return NULL; }
  RenderingObject* CreateXXLFM_FrameListProxy(XXLFM_FrameList *) { return NULL; }
  RenderingObject* CreateXXLFM_LightMapProxy(XXLFM_LightMap *) { return NULL; }
  RenderingObject* CreateXXLFM_SurfaceMapListProxy(XXLFM_SurfaceMapList *) { return NULL; }
  RenderingObject* CreateXXLFM_ViewMapListProxy(XXLFM_ViewMapList *) { return NULL; }
  RenderingObject* CreateXXParticlesProxy(XXParticles *) { return NULL; }
  RenderingObject* CreateXXParticleInitBoxProxy(XXParticleInitBox *) { return NULL; }
  RenderingObject* CreateXXPlanarObstacleProxy(XXPlanarObstacle *) { return NULL; }
  RenderingObject* CreateXXPointAttractorProxy(XXPointAttractor *) { return NULL; }
  virtual RenderingObject* CreateQuadricProxy(Quadric *);
  virtual RenderingObject* CreateSolidRepProxy(SolidRep *);

  virtual RenderingObject* CreateFootPrintSetProxy(FootPrintSet *);
  virtual RenderingObject* CreateFootPrintProxy(FootPrint *);
  virtual RenderingObject* CreateBuildingPartProxy(BuildingPart *);
  virtual RenderingObject* CreateRoofProxy(Roof *);
  virtual RenderingObject* CreateFacadeProxy(Facade *);

  // end removed AFX nodes


// end of AFX nodes

	// version 6 nodes
  virtual RenderingObject* CreateSynthesizedTextureProxy(SynthesizedTexture *);
	virtual RenderingObject* CreateXEllipseProxy(XEllipse *);
  virtual RenderingObject* CreateLinearGradientProxy(LinearGradient *);
  virtual RenderingObject* CreateRadialGradientProxy(RadialGradient *);
  virtual RenderingObject* CreatePathLayoutProxy(PathLayout *);
  virtual RenderingObject* CreateXFontStyleProxy(XFontStyle *);

	// end version 6 nodes

	// version 7 nodes
  virtual RenderingObject* CreateDepthImageV2Proxy(DepthImageV2 *);
  virtual RenderingObject* CreateSBVCAnimationV2Proxy( SBVCAnimationV2* );
  virtual RenderingObject* CreateMorphShapeProxy( MorphShape* );
  virtual RenderingObject* CreateMultiTextureProxy( MultiTexture* );
  virtual RenderingObject* CreateMultiTextureCoordinateProxy(MultiTextureCoordinate* );
  virtual RenderingObject* CreatePointTextureV2Proxy(PointTextureV2 *);
  virtual RenderingObject* CreateSimpleTextureV2Proxy(SimpleTextureV2 *);

	// advanced audio bifs
  virtual RenderingObject* CreateAdvancedAudioBufferProxy(AdvancedAudioBuffer *);
  virtual RenderingObject* CreateAudioChannelConfigProxy(AudioChannelConfig *);
  virtual RenderingObject* CreateSurroundingSoundProxy(SurroundingSound *);
  virtual RenderingObject* CreateTransform3DAudioProxy(Transform3DAudio *);
  virtual RenderingObject* CreateWideSoundProxy(WideSound *);
  //added by Helge Drumm TU Ilmenau, 6.12.2006, Shadow Rendering
  virtual RenderingObject *CreateShadowProxy(Shadow* pObject);
	// end version 7 nodes
private:

  /// Executive

  Executive*		m_pExecutive;
};


#endif // __PRESENTER_H__
