// This software module was originally developed by 
// Gabriel Abrantes (AT&T) in the course
// of development of the MPEG-4 Systems standard (ISO/IEC 14496-1).  This
// software module is an implementation of a part of one or more MPEG-4
// Systems tools as specified by the MPEG-4 Systems standard.  ISO/IEC
// gives users of the MPEG-4 Systems standard free license to this software
// module or modifications thereof for use in hardware or software products
// claiming conformance to the MPEG-4 Systems standard.  Those intending to
// use this software module in hardware or software products are advised
// that its use may infringe existing patents.  The original developer of
// this software module and his/her company, the subsequent editors and
// their companies, and ISO/IEC have no liability for use of this software
// module or modifications thereof in an implementation.  Copyright is not
// released for non MPEG-4 Systems conforming products.  The original
// developer retains full right to use the code for his/her own purpose,
// assign or donate the code to a third party and to inhibit third parties
// from using the code for non MPEG-4 Systems conforming products.  This
// copyright notice must be included in all copies or derivative works.
//
// Copyright 1998.


#include <assert.h>
#include <time.h> 

#include "PhonemeFAP_Converter.h"
#include "Tts_interface.h"

#include "Core/Nodes.h"

#include "OpenGl/Fba/FapProxy.h"

//Constructor

PhonemeFAP_Converter::PhonemeFAP_Converter(): 
	TTS_Face(NULL),
	frames_lock(0),
	last_frame(0)
{
}

// Destructor

PhonemeFAP_Converter::~PhonemeFAP_Converter()
{
	if(this) {
		ClearFrames();
		TTS_BookMarks.Clear();
		TTSFAPInterface.Clear();
	}
}





void PhonemeFAP_Converter::SetTTSFace(Face* ttsface)
{
	if(this) TTS_Face = ttsface;
}



void PhonemeFAP_Converter::SetTTSFAPInterface(ZArray<TtsFAPInterface*> *TttsFAPI)
{
	if(this) {
		TTSFAPInterface <<= (*TttsFAPI);
		Phonemes_to_FAP();
	}
}




void PhonemeFAP_Converter::ClearFrames()
{
	while(frames_lock);
	frames_lock = 1;
	TTS_FAframes.Clear();
	last_frame = 0;
	frames_lock = 0;
}



// convert Phonemes to FAP

void PhonemeFAP_Converter::Phonemes_to_FAP()
{
	ClearFrames();

	Phoneme_2_FA_Frames();

	ExtraFrames4Bookmarks();

	TTSFAPInterface.Clear();
}




void PhonemeFAP_Converter::Phoneme_2_FA_Frames()
{
	int no_phonemes = (TTSFAPInterface).GetLength();
	if(!no_phonemes) return;

	for(int i=0;i<no_phonemes;i++) {
		
		for(int j=0;j<NO_MPEG_Phonemes;j++) {
			
			if(!strcmp(MPEG_Phonemes_Alphabet[j],(TTSFAPInterface)[i]->phonemeSymbol)) {
				
				TTS_FA_frame* frame_out;
				frame_out = (TTS_FA_frame*) calloc(1,sizeof(TTS_FA_frame));
				
				
				// check bookmarks, bookmark starts at the same time as phoneme
				
				for(int kk=0;kk<NO_BOOKMARKS;kk++) {
					if(strlen((TTSFAPInterface)[i]->bookmark[kk])) 
						ParseBookmark((TTSFAPInterface)[i]->bookmark[kk],(int)((TTSFAPInterface)[i]->start_time));
					else kk = NO_BOOKMARKS+1;
				}
				
				
				// 1st option: Phonemes to visemes FAP
				
				frame_out->expression_set = 0;
				frame_out->viseme_set = 1;
				
				frame_out->viseme.select1 = Phoneme_to_Visemes[j];
				frame_out->viseme.select2 = 0;
				frame_out->viseme.blend   = 0;
				
				
				frame_out->time_stamp = (int)((TTSFAPInterface)[i]->start_time);
				
				// add bookmark info, if available
				
				BookmarksToFrame(frame_out);
				
				// following only used because face animation implementation is 
				// uncomplete
				
				// 2nd option: Phonemes to low level FAP
				AddLowLevelFAPs4VisExp(frame_out);

				AddFAPFrame(frame_out);	
				
			} 
			else {
				
				// if no match, then interpolate in the fill function, between gaps !!!
				
			}
		}
		
	}

	// insert one frame at the end of the sentence after duration of last phoneme
				
	TTS_FA_frame* frame_out;
	frame_out = (TTS_FA_frame*) calloc(1,sizeof(TTS_FA_frame));

	frame_out->time_stamp = (int)(TTSFAPInterface[TTSFAPInterface.GetLength()-1]->start_time +
								  TTSFAPInterface[TTSFAPInterface.GetLength()-1]->phonemeDuration );
	// add bookmark info, if available
	BookmarksToFrame(frame_out);			
	AddLowLevelFAPs4VisExp(frame_out);
	AddFAPFrame(frame_out);

}



void PhonemeFAP_Converter::AddLowLevelFAPs4VisExp(TTS_FA_frame* FAframe)
{
	if(FAframe->viseme_set) {
	
		int l=0;
		for(int k=2;k<68;k++) {
			if(ttsvisemes_mask[k])
				FAframe->faps[k] = 0.9*ttsvisemes_fap[FAframe->viseme.select1][l++];
		}
	}
	
	if(FAframe->expression_set) {
		int mex1 =0;
		int mex2 =0;
		int ex1,ex2;
		for(int k=2;k<68;k++) {
			if(ttsexpressions_mask[FAframe->expression.select1][k])
				ex1 = ttsexpressions_fap[FAframe->expression.select1][mex1++];
			else ex1= 0;
			if(ttsexpressions_mask[FAframe->expression.select2][k])
				ex2 = ttsexpressions_fap[FAframe->expression.select2][mex2++];
			else ex2= 0;
			
			if( (ex1 || ex2) && !FAframe->faps[k])
				FAframe->faps[k] = (int)(
				(((float)FAframe->expression.intensity1)/63.0f)
				*(float)ex1
				+ 
				(((float)FAframe->expression.intensity2)/63.0f)
				*(float)ex2
				);
		}
	}

}




void PhonemeFAP_Converter::ExtraFrames4Bookmarks()
{
	int no_frames = TTS_FAframes.GetLength();
	if(!no_frames) return;
	int lt = TTS_FAframes[no_frames-1]->time_stamp;
	int tim = 1,stop = 0;

	while(!stop) {
		
		TTS_FA_frame* frame_out;
		frame_out = (TTS_FA_frame*) calloc(1,sizeof(TTS_FA_frame));
		
		frame_out->time_stamp = lt + (int)(CLOCKS_PER_SEC*(0.066f)*tim);
		
		stop = BookmarksToFrame(frame_out);
		AddLowLevelFAPs4VisExp(frame_out);		

		AddFAPFrame(frame_out);
		tim++;
	}

}


int PhonemeFAP_Converter::GetInt(char *bm,int *bc)
{
	int l = *bc;
	int b = strlen(bm);
	if(l>=b) return -1; // error

	while(isspace(bm[*bc])) (*bc)++;

	char word[8]; int i=0;
	while(!isspace(bm[*bc]) && (*bc)<b) word[i++] = bm[(*bc)++];
	word[i]='\0';

	return atoi(word);
}




void PhonemeFAP_Converter::ParseBookmark(char *bm,int start_time)
{
	int bc = 0,bml = TTS_BookMarks.GetLength();

	// eliminate triangle bookmarks that have elapsed && ended reset bookmarks
	int i = 0;
	while(i<bml) {
		if( (start_time >= TTS_BookMarks[i]->start_time + TTS_BookMarks[i]->T) 
			&& (TTS_BookMarks[i]->C==2) ) {
			TTS_BookMarks.DeleteFrom(i);
			bml--;
		}
		else if( (start_time > TTS_BookMarks[i]->start_time + TTS_BookMarks[i]->T) 
			&& (TTS_BookMarks[i]->FAP>2) && !TTS_BookMarks[i]->FAP_value ) {
			TTS_BookMarks.DeleteFrom(i);
			bml--;
		}
		else if( (start_time > TTS_BookMarks[i]->start_time + TTS_BookMarks[i]->T) 
			&& (TTS_BookMarks[i]->FAP==2) && !TTS_BookMarks[i]->expression.intensity1
			&& !TTS_BookMarks[i]->expression.intensity2) {
			TTS_BookMarks.DeleteFrom(i);
			bml--;
		}
		else i++;
	}

	// bookmark to be added
	TTS_BookMark* tts_bm = (TTS_BookMark*) calloc(1,sizeof(TTS_BookMark));
	
	GetInt(bm,&bc); // get FAP name of string

	tts_bm->FAP = GetInt(bm,&bc);


	if(tts_bm->FAP>2) //low level FAPs
		tts_bm->FAP_value = GetInt(bm,&bc);
	else if(tts_bm->FAP==1) { // ignore viseme
		free(tts_bm);
		return;	
	}
	else if(tts_bm->FAP==2) { // expression
		tts_bm->expression.select1    = GetInt(bm,&bc);
		tts_bm->expression.intensity1 = GetInt(bm,&bc);
		tts_bm->expression.select2    = GetInt(bm,&bc);
		tts_bm->expression.intensity2 = GetInt(bm,&bc);
	}

	tts_bm->T = GetInt(bm,&bc);
	tts_bm->C = GetInt(bm,&bc);

	tts_bm->start_time = start_time;

	
	// Check if there is a bookmark with this FAP, if so delete it !
	// but first compute as & gs_1
	
	i = 0;
	while(i<bml) {
		
		if(TTS_BookMarks[i]->FAP==tts_bm->FAP) {
			
			if(TTS_BookMarks[i]->FAP==2) { // expression
				
				if( (TTS_BookMarks[i]->expression.select1==tts_bm->expression.select1)
					&& (TTS_BookMarks[i]->expression.select2==tts_bm->expression.select2)) {
					
					if(start_time >= (TTS_BookMarks[i]->start_time + TTS_BookMarks[i]->T) ) {
						tts_bm->gs_1[0] = 0;
						tts_bm->gs_1[1] = 0;
						tts_bm->as[0]   = TTS_BookMarks[i]->expression.intensity1;
						tts_bm->as[1]   = TTS_BookMarks[i]->expression.intensity2;
					}
					else {
						float t1,t2,t3;
						t1 = start_time;
						t2 = TTS_BookMarks[i]->start_time;
						t3 = TTS_BookMarks[i]->T;
						
						tts_bm->gs_1[0] = Tangent( 
							TTS_BookMarks[i]->as[0],
							TTS_BookMarks[i]->expression.intensity1,
							TTS_BookMarks[i]->gs_1[0],
							(t1 - t2)/t3);
						
						tts_bm->gs_1[1] = Tangent( 
							TTS_BookMarks[i]->as[1],
							TTS_BookMarks[i]->expression.intensity2,
							TTS_BookMarks[i]->gs_1[1],
							(t1 - t2)/t3);
						
						tts_bm->as[0]   = Interpolation(
							TTS_BookMarks[i]->C,
							TTS_BookMarks[i]->as[0],
							TTS_BookMarks[i]->expression.intensity1,
							TTS_BookMarks[i]->gs_1[0],
							(t1 - t2)/t3);
						
						tts_bm->as[1]   = Interpolation(
							TTS_BookMarks[i]->C,
							TTS_BookMarks[i]->as[1],
							TTS_BookMarks[i]->expression.intensity2,
							TTS_BookMarks[i]->gs_1[1],
							(t1 - t2)/t3);
						
					}	
					
				}
				
			}
			else if(TTS_BookMarks[i]->FAP>2) { //low level FAP
				
				if(start_time >= (TTS_BookMarks[i]->start_time + TTS_BookMarks[i]->T) ) {
					tts_bm->gs_1[0] = 0;
					tts_bm->as[0]   = TTS_BookMarks[i]->FAP_value;
				}
				else {
					float t1,t2,t3;
					t1 = start_time;
					t2 = TTS_BookMarks[i]->start_time;
					t3 = TTS_BookMarks[i]->T;
					
					tts_bm->gs_1[0] = Tangent( 
						TTS_BookMarks[i]->as[0],
						TTS_BookMarks[i]->FAP_value,
						TTS_BookMarks[i]->gs_1[0],
						(t1 - t2)/t3);
					
					tts_bm->as[0]   = Interpolation(
						TTS_BookMarks[i]->C,
						TTS_BookMarks[i]->as[0],
						TTS_BookMarks[i]->FAP_value,
						TTS_BookMarks[i]->gs_1[0],
						(t1 - t2)/t3);
				}	
			}

			TTS_BookMarks.DeleteFrom(i);
			break;
		}
		i++;
	}

	TTS_BookMarks+=tts_bm;
}



int PhonemeFAP_Converter::BookmarksToFrame(TTS_FA_frame* frame)
{
	int bml = TTS_BookMarks.GetLength();
	if(!bml) return 1;
	
	for(int i=0;i<bml;i++) {

		if( (frame->time_stamp >= (TTS_BookMarks[i]->start_time + TTS_BookMarks[i]->T)) && (TTS_BookMarks[i]->C!=2) ) {

			if(TTS_BookMarks[i]->C!=2) { 
				
				if(TTS_BookMarks[i]->FAP==2) { // expression
					frame->expression_set = 1;
					frame->expression.select1    = TTS_BookMarks[i]->expression.select1;
					frame->expression.select2    = TTS_BookMarks[i]->expression.select2;
					frame->expression.intensity1 = TTS_BookMarks[i]->expression.intensity1;
					frame->expression.intensity2 = TTS_BookMarks[i]->expression.intensity2;
				}
				else if(TTS_BookMarks[i]->FAP>2) //low level FAP
					frame->faps[TTS_BookMarks[i]->FAP-1] = TTS_BookMarks[i]->FAP_value;
			}
		}
		
		if( (frame->time_stamp == TTS_BookMarks[i]->start_time) && !TTS_BookMarks[i]->T) {

			if(TTS_BookMarks[i]->C!=2) { 
				
				if(TTS_BookMarks[i]->FAP==2) { // expression
					frame->expression_set = 1;
					frame->expression.select1    = TTS_BookMarks[i]->expression.select1;
					frame->expression.select2    = TTS_BookMarks[i]->expression.select2;
					frame->expression.intensity1 = TTS_BookMarks[i]->expression.intensity1;
					frame->expression.intensity2 = TTS_BookMarks[i]->expression.intensity2;
				}
				else if(TTS_BookMarks[i]->FAP>2) //low level FAP
					frame->faps[TTS_BookMarks[i]->FAP-1] = TTS_BookMarks[i]->FAP_value;
			}
		}

		else if( TTS_BookMarks[i]->start_time == frame->time_stamp) {
			
			if(TTS_BookMarks[i]->FAP==2) { // expression
				frame->expression_set = 1;
				frame->expression.select1    = TTS_BookMarks[i]->expression.select1;
				frame->expression.select2    = TTS_BookMarks[i]->expression.select2;
				frame->expression.intensity1 = TTS_BookMarks[i]->as[0];
				frame->expression.intensity2 = TTS_BookMarks[i]->as[1];
			}
			else if(TTS_BookMarks[i]->FAP>2) //low level FAP
				frame->faps[TTS_BookMarks[i]->FAP-1] = TTS_BookMarks[i]->as[0];
		}
		
		
		else if( (TTS_BookMarks[i]->start_time + TTS_BookMarks[i]->T) >= frame->time_stamp) {
			
			float t1,t2,t3;
			t1 = frame->time_stamp;
			t2 = TTS_BookMarks[i]->start_time;
			t3 = TTS_BookMarks[i]->T;
			
			if(TTS_BookMarks[i]->FAP==2) { // expression
				frame->expression_set = 1;
				frame->expression.select1    = TTS_BookMarks[i]->expression.select1;
				frame->expression.select2    = TTS_BookMarks[i]->expression.select2;
				frame->expression.intensity1 = Interpolation(	
					TTS_BookMarks[i]->C,
					TTS_BookMarks[i]->as[0],
					TTS_BookMarks[i]->expression.intensity1,
					TTS_BookMarks[i]->gs_1[0],
					(t1-t2)/t3 );
				frame->expression.intensity2 = Interpolation(	
					TTS_BookMarks[i]->C,
					TTS_BookMarks[i]->as[1],
					TTS_BookMarks[i]->expression.intensity2,
					TTS_BookMarks[i]->gs_1[1],
					(t1-t2)/t3 );
			}
			else if(TTS_BookMarks[i]->FAP>2) //low level FAP
				frame->faps[TTS_BookMarks[i]->FAP-1] = Interpolation(	
					TTS_BookMarks[i]->C,
					TTS_BookMarks[i]->as[0],
					TTS_BookMarks[i]->FAP_value,
					TTS_BookMarks[i]->gs_1[0],
					(t1-t2)/t3 );
		}
	}


	int stop = 1;
	for(i=0;i<bml;i++)
		if( (frame->time_stamp < (TTS_BookMarks[i]->start_time + TTS_BookMarks[i]->T))) 
			stop = 0;

	return stop;

}



int PhonemeFAP_Converter::Tangent(float as,float a,float gs_1,float t)
{
	return (6*pow(t,2)-6*t)*(as-a) + (3*pow(t,2)-4*t+1)*gs_1;
}



int PhonemeFAP_Converter::Interpolation(int function,float as,float a,float gs_1,
										float t)
{
	function =1;
	// linear
	if(function==1) return (as+1*t*(a-as));

	// triangle
	else if (function==2) {
		if(t<0.5f) return (as + 2*t*(a-as));
		else return (as + 2*(1-t)*(a-as));
	}

	// Hermite
	else if (function==3) {
		float gs = Tangent(as,a,gs_1,t);
		return ( ( 2*pow(t,3) - 3*pow(t,2) + 1 )*as + 
				 ( 3*pow(t,2) - 2*pow(t,2) )*a + 
				 ( 1*pow(t,3) - 2*pow(t,2) + t)*gs ) ;
	}

	else return 0;
}




void PhonemeFAP_Converter::AddFAPFrame(TTS_FA_frame* frame_out)
{
	// wait for release of FA-frames-structure
	while(frames_lock);

	// lock FA-frames-structure
	frames_lock = 1;

	// insert 1 more frame
	TTS_FAframes+=frame_out;

	// delete all frames with TS<clock(), thus not needed anymore

	// unlock FA-frames-structure
	frames_lock = 0;
}



 
// FA gets 1 frame from the FA-frames-structure
void PhonemeFAP_Converter::Get1FA_Frame()
{
	while(frames_lock);
	frames_lock = 1;

	if( (!TTS_FAframes.GetLength()) || (TTS_FAframes[0]->time_stamp>clock()) ){
		frames_lock = 0;		
		return;
	}

	FAP* pFAP = TTS_Face->fap;


	// search a frame  with the highest TS equal or lower than actual time
	// if there is none, do nothing
	

	int cont = 1;
	int frame = 1;

	int found,frame_no,frame_cnt = TTS_FAframes.GetLength();
	time_t current_time;

	found = 0;
	if(frame_cnt>1) {

		int k = last_frame;if(!k) k=1;

		while(k<frame_cnt) {

			current_time = clock();
			if( (TTS_FAframes[k]->time_stamp >= current_time) && 
				(TTS_FAframes[k-1]->time_stamp < current_time)) {
				
				frame_no = k;

				if( ((TTS_FAframes[k]->time_stamp-current_time) > 
					(TTS_FAframes[k-1]->time_stamp-current_time) )
					&& ((k-1)>last_frame) ) 
				frame_no = k-1;
				
				found = 1;
				break;
			}
			k++;
		}
		if(!found) {
			frames_lock = 0;
			return;
		}
	}

	last_frame = frame_no;
	// set the FAPs using selected frame

	if(TTS_FAframes[frame_no]->expression_set) {
		
		Expression* pExpress = pFAP->expression;
		
		if(pExpress) {
			// pExpress->SetExpression_select1(TTS_FAframes[frame_no]->expression.select1);
      pExpress->expression_select1 = 
        TTS_FAframes[frame_no]->expression.select1;
			// pExpress->SetExpression_intensity1(TTS_FAframes[frame_no]->expression.intensity1);
      pExpress->expression_intensity1 = 
        TTS_FAframes[frame_no]->expression.intensity1;
			// pExpress->SetExpression_select2(TTS_FAframes[frame_no]->expression.select2);
      pExpress->expression_select2 = 
        TTS_FAframes[frame_no]->expression.select2;
			// pExpress->SetExpression_intensity2(TTS_FAframes[frame_no]->expression.intensity2);
      pExpress->expression_intensity2 = 
        TTS_FAframes[frame_no]->expression.intensity2;
			pExpress->expression_def = FALSE;
			pExpress->init_face = FALSE;
		}
	}
	if(TTS_FAframes[frame_no]->viseme_set) {
		
		Viseme* pVis = pFAP->viseme;
		
		if(pVis) {
			pVis->viseme_select1 = 
        TTS_FAframes[frame_no]->viseme.select1;
			pVis->viseme_select2 = 
        TTS_FAframes[frame_no]->viseme.select2;
			pVis->viseme_blend = 
        TTS_FAframes[frame_no]->viseme.blend;
			pVis->viseme_def = FALSE;
		}
	}

  FAPProxy * poFapProxy = 
    (FAPProxy *) pFAP->GetRenderingObject();
	// low level FAPs
	for (int i=2;i<68;i++) 
    poFapProxy->SetFAPValue(i, TTS_FAframes[frame_no]->faps[i]);

	frames_lock = 0;
}

