// This software module was originally developed by Gabriel Abrantes
// (AT&T) in the course of development of the MPEG-4 Systems
// standard (ISO/IEC 14496-1).  This software module is an implementation
// of a part of one or more MPEG-4 Systems tools as specified by the MPEG-4
// Systems standard.  ISO/IEC gives users of the MPEG-4 Systems standard
// free license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4 Systems
// standard.  Those intending to use this software module in hardware or
// software products are advised that its use may infringe existing
// patents.  The original developer of this software module and his/her
// company, the subsequent editors and their companies, and ISO/IEC have no
// liability for use of this software module or modifications thereof in an
// implementation.  Copyright is not released for non MPEG-4 Systems
// conforming products.  The original developer retains full right to use
// the code for his/her own purpose, assign or donate the code to a third
// party and to inhibit third parties from using the code for non MPEG-4
// Systems conforming products.  This copyright notice must be included in
// all copies or derivative works.


// Festival client for IM1-3DPlayer support
// Copyright 1998 AT&T
//
// Program "th_client" opens a connection to a Festival TTS server.
// Text sentences are sent to the server. 
//
// The server returns zero or more data packets followed by a status
// code, which is either "OK" if successful or "ER" if an error occurred.
// Data packets begin either with "LP" or "WV".  LP indicates an ASCII
// message or printed Lisp expression, while "WV" indicates that binary
// speech data follows.  In both cases, packets terminate with the
// string "ft_StUfF_key".
//
// This program depends on code in the Scheme library file
// "talking_head.scm" to set up the Festival server.  This code
// modifies the synthesis procedure, including a function which
// sends word and phoneme timing information to the client.  This
// info comes in event messages like these for the word "three":
//	"EVENT WORD three 0.5255 cd"
//	"EVENT PHONEME th 0.5255 0.1101 1 109.125"
//	"EVENT PHONEME r 0.6356 0.0705 1 0"
//	"EVENT PHONEME ii 0.7061 0.1720 1 95.692"
//
// Word events list name and start time (sec).
// Phoneme events list name, start time (sec) and duration (sec).
//
// EVENT SPEECH indicates that Festival will start speaking
//

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <ctype.h>
#include <winsock2.h>
#include <time.h>
#include <errno.h>

#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <process.h>

#include "../base/ztl.h"
#include "tts_festival_client.h"

#define F_SILENCE  101
#define F_PHONEME  201
#define F_WORD	   301
#define F_SPEECH   401
#define F_UTTER    501


// Local functions.
void th_collect_server_response(int server_fd,char*);
int  th_command_and_reply(int server_fd, const char* buf);
int  th_sentence_command_and_reply(int server_fd, const char* buf,
								   int face_on, ZArray<TtsFAPInterface*> *TTSFAPInterface,
								   int *start_tts);
int  th_connect_to_server(char* server_name, long server_port);
int  th_enable_server_support(int server_fd, char* libdir);



//These 3 following variables are set by the file FestivalClientDefs.txt

// Host where the TTS server runs for web and other demos.
char server_name_default[256] = "/0";

// The default is 1314
long server_port_default = 1314; 

// The default library directory containing "talking_head.scm"
char th_libdir_default[256] = "/0";


// The default library directory containing festival executable
char festival_location[256] = "/0";


int GetWord(FILE *fp,char *s)
{
	int c;
	while(c = getc(fp),isspace(c));
	ungetc(c,fp);
	
	while(c = getc(fp),!isspace(c)) 
		*s++ = c;
	*s='\0';
	
	return ungetc(c,fp);
}



int StartServer()
{
	int proc;
	char *args[3];
	args[0] = "festival";
	args[1] = "--server";
	args[2] = NULL;

	// read default parameters

	FILE *fp_defs = fopen("FestivalClientDefs.txt","rb");
	if(!fp_defs) {
		MessageBox(NULL,"Can't find file - FestivalClientDefs.txt - !","TTS in IM1 3D-Player",MB_OK);
		return 0;
	}
	
	char  trash[512];
	
	while(ungetc(getc(fp_defs),fp_defs) == '#') 
		fgets(trash,512,fp_defs);

	GetWord(fp_defs,server_name_default);
	fscanf(fp_defs,"%d", &server_port_default);
	GetWord(fp_defs,th_libdir_default);
	GetWord(fp_defs,festival_location);
	
	fclose(fp_defs);

	// make it read the directory from a file, so there will be no need to change 
	// the code, use c++ functions to read from file, it's easier

	_flushall();
	// put Festival working in server mode
	proc = _spawnv(_P_NOWAIT,festival_location,args);

	if(proc==-1)  return 0;	

	return proc;
}


int ConnectoServer()
{
    int status = 0;
	
	// read this from init file

    char* server_name = server_name_default;
    long  server_port = server_port_default;
    char* th_libdir = th_libdir_default;
	
	char word[256];

    int server_fd = th_connect_to_server(server_name, server_port);

    if (server_fd < 0) {
		sprintf(word,"Connection to server failed!");
		MessageBox(NULL,word,"TTS in IM1 3D-Player",MB_OK);
		return -1;
    }

    status = th_enable_server_support(server_fd, th_libdir);
    if (status < 0) {
		sprintf(word,"Failed to enable Festival support in server!");
		MessageBox(NULL,word,"TTS in IM1 3D-Player",MB_OK);
		return 0;
    }

	return server_fd;
}


void SaySentence(int server_fd,char *sentence,int face_on, 
				 ZArray<TtsFAPInterface*> *TTSFAPInterface,int *start_tts)
{
	char cmd[1024];

    if (server_fd < 0) return;
	
	sprintf(cmd,"(SayText \"%s\")\n",sentence);
	
	int status = th_sentence_command_and_reply(server_fd, cmd,face_on,TTSFAPInterface,start_tts);
}



void CloseConnection(int server_fd)
{
    if (server_fd >= 0)
		closesocket(server_fd);
}



int CloseServer(int pid)
{
	return pid;	
}



int th_connect_to_server(char* server_name, long server_port)
{
    struct sockaddr_in server_addr;
    struct hostent* server_host;
    int    server_fd;
	

	WORD wVersionRequested;
	WSADATA wsaData;
	int err;

	wVersionRequested = MAKEWORD( 2, 0 ); 
	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		return -1;
	}

    
	if (!(server_host = gethostbyname(server_name))) {
		int e = WSAGetLastError ();
		return -1;
    }



    if ((server_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		int e = WSAGetLastError ();
		return -1;
    }

	memset(&server_addr, 0, sizeof(server_addr));
	
	// sin_family
    server_addr.sin_family = AF_INET;
	
	memmove(&server_addr.sin_addr,server_host->h_addr, server_host->h_length);

	// sin_port
    server_addr.sin_port = htons((unsigned short)server_port);
	
	// try get connected during 10 seconds
	int connection_fail = 1;
	int try_cnt = 0;

	while(connection_fail && (try_cnt<10) ) {
		connection_fail = connect(server_fd, (struct sockaddr *)&server_addr, sizeof(server_addr));
		try_cnt++;
		Sleep(1000);
	}

	if(connection_fail) {
		closesocket(server_fd);
		return -1;
	}

    return server_fd;		// returns -1 on error
}



int th_enable_server_support(int server_fd, char* libdir)
{
    const char* err = "Error in th_enable_server_support: ";
		
    int n = strlen(libdir);
    if (n < 1) {
		return -1;
    }
    while (libdir[--n] == '/')	// remove trailing slash(es)
		libdir[n] = 0;

	
	char libpath[1024];
	sprintf(libpath,"%s/th_prolog.scm",libdir);

	char cmd[1024];
	sprintf(cmd,"(load \"%s\")\n",libpath);

    int status = th_command_and_reply(server_fd, cmd);
    if (status != 0) {
		printf(" %s Failed to load %s \n",err,libpath);
		return -1;
    }
	
    return 0;
}






int th_command_and_reply(int server_fd, const char* buf)
{
    const char* err = "Error in th_command_and_reply: ";
	
    // Check args
    if (server_fd < 0) return -1;
    if (buf == 0) return -1;

    int nbuf = strlen(buf);
    if (nbuf > 0) {
		
		// Send buffer to server
		int rc = send(server_fd, buf, nbuf,0);

		if (rc < nbuf) return -1;

		if (buf[nbuf-1] != '\n') {	// be sure to send a newline
			rc = send(server_fd, "\n", 1,0);
			if (rc != 1) return -1;
		}
		
		// Read replies from server until "OK" or "ER" (error)
		while (1) {				// Break from within the loop
			
			// First 3 bytes may be the final status (OK or ER)...
			char reply[3+1];
			char c;
			for(int l=0;l<4;l++) reply[l]=0;
			
			int i = 0;
			while (i < 3) {
				int rc = recv(server_fd, &c, 1,0);
				if (rc < 1) return -1;

				reply[i++] = c;
				reply[i] = 0;
				if (c == '\n') break;
			}
			
			if (strncmp(reply, "OK\n", 3) == 0) break;

			if (strncmp(reply, "ER\n", 3) == 0) return +1;	// Scheme error; NOT fatal
			
			// Or reply may identify the type of response to follow
			char response[1024];
			
			if (strncmp(reply, "WV\n", 3) == 0) {
				th_collect_server_response(server_fd,response);
				if (!strlen(response)) return -1;
			}
			else if (strncmp(reply, "LP\n", 3) == 0) {
				th_collect_server_response(server_fd,response);
				if (!strlen(response)) return -1;
			}
		}
    }
	
    return 0;
}




void GetWord(char *buffer,int *wrd_cnt,char *word_out)
{
	int maxcnt = strlen(buffer);

	while(isspace(buffer[*wrd_cnt]))
		(*wrd_cnt)++;

	int i=0;
	while(!isspace(buffer[*wrd_cnt]))
		word_out[i++] = buffer[(*wrd_cnt)++];

	word_out[i]='\0';
}




int GetResponseInfo(char *response,ZArray<TtsFAPInterface*> *TTSFAPInterface)
{
	int  wrd_cnt = 1;
	char word[32];
	char phoneme[8];

	static int wBegin = 0;

	if(!strncmp(response,"#<UTT",5)) return F_UTTER;

	GetWord(response,&wrd_cnt,word);

	if(!strcmp(word,"EVENT")) {

		GetWord(response,&wrd_cnt,word);
		
		if(!strcmp(word,"WORD")) {
			wBegin = 1;
			return F_WORD;
		}
		else if(!strcmp(word,"SPEECH"))
			return F_SPEECH;

		else if(!strcmp(word,"PHONEME")) {

			GetWord(response,&wrd_cnt,phoneme);

			GetWord(response,&wrd_cnt,word);
			double start_time = atof(word);

			GetWord(response,&wrd_cnt,word);
			double duration = atof(word);

			TtsFAPInterface* phoneme_out = (TtsFAPInterface*) calloc(1,sizeof(TtsFAPInterface));
			
			strcpy(phoneme_out->phonemeSymbol,phoneme);
			phoneme_out->start_time = (float)start_time;		
			phoneme_out->phonemeDuration = (float)duration;
					
			if(wBegin) {
				phoneme_out->wordBegin = 1;
				wBegin = 0;
			}

			*TTSFAPInterface+=phoneme_out;

			if(!strcmp(phoneme,"pau")) return F_SILENCE;
			
		}
		return F_PHONEME;		
	}
	else return -1;

}



int th_sentence_command_and_reply(int server_fd, const char* buf,
								  int face_on, ZArray<TtsFAPInterface*> *TTSFAPInterface,
								  int *start_tts)
{	
	if (buf == 0) return -1;

	int nbuf = strlen(buf);
    if (nbuf > 0) {
		
		// Send buffer to server
		int rc = send(server_fd, buf, nbuf,0);

		if (rc < nbuf) return -1;

		if (buf[nbuf-1] != '\n') {	// be sure to send a newline
			rc = send(server_fd, "\n", 1,0);
			if (rc != 1) return -1;
		}

		// Read replies from server until "OK" or "ER" (error)
		while (1) {				// Break from within the loop
			
			// First 3 bytes may be the final status (OK or ER)...
			char reply[3+1],c;

			for(int l=0;l<4;l++) reply[l] = 0;
			
			int i = 0;
			while (i < 3) {
				int rc = recv(server_fd, &c, 1,0);
				if (rc < 1) return -1;

				reply[i++] = c;
				reply[i] = 0;
				if (c == '\n') break;
			}
			
			if (!strncmp(reply, "OK\n", 3)) break;

			if (!strncmp(reply, "ER\n", 3)) return +1;
			
			// Or reply may identify the type of response to follow
			char response[1024];
			
			if (!strncmp(reply, "WV\n", 3)) {
				th_collect_server_response(server_fd,response);
				if (!strlen(response)) return -1;
			}
			else if (!strncmp(reply, "LP\n", 3)) {
				th_collect_server_response(server_fd,response);
				if (!strlen(response)) return -1;
				else if(face_on) {
					int resp = GetResponseInfo(response,TTSFAPInterface);
					if(resp==F_SPEECH) {
						*start_tts = clock();
						return 0;					
					}
					else if(resp==F_UTTER) {
						i = 0;
						
						// collect garbage (OK) from previous sentence
						while (i++ < 3) recv(server_fd, &c, 1,0);
					}
					
				}
			}
		}
    }
	
    return 0;
}


void th_collect_server_response(int fd,char *s)
{
    const char *key = "ft_StUfF_key";	// This string marks end of response.
    const int nkey = strlen(key);

    char ch;
    int k = 0;
	int i;
	
	int l=0;
    // Read until key is found
    while (k < nkey) {
		int n = recv(fd, &ch, 1,0);
		if (n < 0) {
			s[0]='\0';
			return;
		}
		if (ch == key[k]) k++;// Have seen this much of the key
		else {
			// Save partial key (if any)
			for (i=0; i<k; i++)
				s[l++] = key[i];
			// Save current char,
			// except drop 'X' in place of final key char
			if (!((ch=='X') && (k==(nkey-1))))  s[l++] = ch;
			k = 0;
		}
    }
	s[l]='\0';
}
