/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  UIStream Decoder
FILE:       UIStreamDec.cpp
AUTHOR:     Cyril Concolato (ENST/COMELEC)
OVERVIEW:   Implementation of UIStreamDecoder.
-------------------------------------------------------------------------*/
#include "defs.h"
#include <assert.h>
#include <Executiv.h>
#include "Presenter.h"
#include "UIStreamDec.h"
#include "Root.h"
#include "Nodes.h"
#include "Except.h"
#include "Inline.h"
#include "ODP4.h"
#include "MemBifs.h"

#define pCoder this


#include "InputSensorImp.h"


//-----------------------------------------------------------------
// UIStreamDecoder constructor
//-----------------------------------------------------------------
UIStreamDecoder :: UIStreamDecoder( 
		Executive *pExecutive, 
		InlineImp *pSource, 
		FILE *fList,
		LPCSTR szListTitle,
		int flags,
		int objectTypeIndication
) : SysCoder (fList, szListTitle, flags, objectTypeIndication),
	m_pExecutive(pExecutive), 
	m_pSource(pSource),
	m_pScene (pSource ? pSource -> GetScene () : NULL),
	m_pBufferDecoder(NULL),
	m_pNode(NULL),
	m_pNodeBuffer(NULL),
	m_nNodeBufferLength(0),
	m_devSpecInfo(NULL)
{
}

UIStreamDecoder :: ~UIStreamDecoder() {
	m_pDeviceFilter->DeviceListener_Detach_Req(m_nDeviceId);
	if (m_pBufferDecoder)
		delete m_pBufferDecoder;
}

/*-----------------------------------------------------------------
Parses updates from the stream, and update the scene accordingly.
-----------------------------------------------------------------*/
void UIStreamDecoder :: ParseFrame (LPCBYTE pData, int nLength, DWORD dwTS)
{
	Reset (pData, nLength);
	dwTS;
	TRACEF ("AT %d", dwTS);
	TRACENEWLINE;
	try
	{			
		if (m_pNode->IsEnabled()) {
			currentField = 0;
			semBuffer.P();
			UpdateDecoder();
			if (m_pBufferDecoder) {
				m_pBufferDecoder->ParseUpdateFrame ();
				ExecuteUpdate();
			}
			semBuffer.V();
		}
	}
	catch (BifsException err)
	{
		err;
		Error (SCENE_PARSING_ERROR);
	}
}

/*----------------------------------------------------------------
Parse the data in the DecSpecificInfo field of the device descriptor,
and configure the decoder appropriately.
-----------------------------------------------------------------*/
bool UIStreamDecoder :: Configure (const BYTE *decspecificInfo, int nInfoLength)
{
	UIConfig *pUiConfig = NULL;
	bool bResult = true;
	try
	{	if (nInfoLength > 0)
		{	ODDecoder odDec (
				m_pExecutive,
				decspecificInfo,
				nInfoLength,
				m_pExecutive -> GetTraceFile (),
				"UIConfig",
				m_pExecutive -> GetTraceFlags (),
				GetVersion ()
			);

			pUiConfig = new UIConfig;
			pUiConfig -> SetLength (nInfoLength * 8);
			pUiConfig -> EncDec (&odDec, DECODE, NULL);

			m_szDeviceName = pUiConfig -> deviceName;
			if (pUiConfig ->devSpecInfo.GetLength() != 0)			
				m_devSpecInfo = pUiConfig -> devSpecInfo;
			
			m_nbOfFields = -1;
			if (!strcmp(m_szDeviceName,"KeySensor")) {
				m_nbOfFields  = 7;
			}			
			if (!strcmp(m_szDeviceName,"StringSensor")) {
				m_nbOfFields  = 2;
			}
			if (!strcmp(m_szDeviceName,"Mouse")) {
				m_nbOfFields  = 5;
			}
			if (m_nbOfFields > 0) {
				pFields = new NodeFieldPtr[m_nbOfFields];
				pMediaObjects = new MediaObjectPtr[m_nbOfFields];
				pEntries = new EntryPtr[m_nbOfFields];
				pNPos = new int[m_nbOfFields];
				for (int i=0;i<m_nbOfFields;i++) {
				  pFields[i] = NULL;
				}
				bResult = true;
			} else {
				m_pExecutive -> Error (INVALID_UI_CONFIG);
				bResult = false;				
			}
		}		
	}
	catch (BifsException)
	{
		m_pExecutive -> Error (INVALID_UI_CONFIG);
		bResult = false;
	}	

	if (bResult) {
		m_pDeviceFilter = GetExecutive()->GetDeviceFilter();
		m_pDeviceFilter -> DeviceListener_Attach_Req (this,m_szDeviceName,m_devSpecInfo);
		delete pUiConfig;
	}
	return bResult;
}

void UIStreamDecoder::UpdateBuffer(const BYTE *nodeBuffer, int nodeBufferLength) {
	
	semBuffer.P();
	m_nNodeBufferLength = nodeBufferLength;
	m_pNodeBuffer = new BYTE[m_nNodeBufferLength];
	memcpy(m_pNodeBuffer,nodeBuffer,m_nNodeBufferLength);
	semBuffer.V();
}

void UIStreamDecoder::UpdateDecoder() {
	try
	{	
		if (m_nNodeBufferLength) {
			if (m_pBufferDecoder)  {
				delete m_pBufferDecoder;
				m_pBufferDecoder = NULL;
			}
			
			m_pBufferDecoder =
				new MemBIFSDecoder (
					GetExecutive (),
					m_pNodeBuffer,
					m_nNodeBufferLength,
					GetExecutive () -> GetTraceFile (),
					"InputSensor",
					GetExecutive () -> GetTraceFlags (),
					0,
					m_pNode->GetRoot ()
			);			
			m_pBufferDecoder -> SetUICoder (this);
		} else {
			if (m_pBufferDecoder) delete m_pBufferDecoder;
			m_pBufferDecoder = NULL;
		}

	} catch (BifsException)	{
		#ifdef PLAYER
		GetExecutive () -> Error (INVALID_UI_BUFFER);
		#endif
	}
}

void UIStreamDecoder::DataReceive_Ind(const BYTE* pStreamData, const int nStreamDataLength, const int nErrorFlag) {
	nErrorFlag;
	ParseFrame(pStreamData, nStreamDataLength, GetExecutive()->GetPresenter() ->GetTime()/1000);
}

void UIStreamDecoder::SetFields(NodeField *pField, MediaObject *pMediaObject, const BifsFieldTable* pEntry, int nPos) {
	pFields[currentField] = pField;
	pMediaObjects[currentField] = pMediaObject;
	pEntries[currentField] = pEntry;
	pNPos[currentField] = nPos;
	currentField++;
}

void UIStreamDecoder::ExecuteUpdate()
{
	int index;
	int nValues = ParseInt(8,"numberOfValues");
	for (int i = 0; i<nValues; i++) 
	{
		if ((index = ParseInt(8,"DeviceFieldIndex"))>=0) 
		{
			int type = ParseInt(6,"fieldType");

			if (pFields[index])
			{
				FIELD_TYPE nodeType;
				if ((nodeType= pFields[index]->GetType()) >= 32  ) 
				{						
					switch (nodeType) 
					{
					case MFBOOL:
						((MF<Bool> *)pFields[index]) -> UIReplaceElement ((BIFSDecoder *)this,
																		pMediaObjects[index],
																		pNPos[index]
																		);
						break;
					case MFFLOAT:
						((MF<float> *)pFields[index]) -> UIReplaceElement ((BIFSDecoder *)this,
																		pMediaObjects[index],
																		pNPos[index]
																		);
						break;
					case MFTIME:
						((MF<Time> *)pFields[index]) -> UIReplaceElement ((BIFSDecoder *)this,
																		pMediaObjects[index],
																		pNPos[index]
																		);
						break;
					case MFINT32:
						((MF<int> *)pFields[index]) -> UIReplaceElement ((BIFSDecoder *)this,
																		pMediaObjects[index],
																		pNPos[index]
																		);
						break;
					case MFSTRING:
						((MF<ZString> *)pFields[index]) -> UIReplaceElement ((BIFSDecoder *)this,
																			pMediaObjects[index],
																			pNPos[index]
																			);
						break;
					case MFVEC3F:
						((MF<Vec3f> *)pFields[index]) -> UIReplaceElement ((BIFSDecoder *)this,
																		pMediaObjects[index],
																		pNPos[index]
																		);
						break;
					case MFCOLOR:
						((MF<Color3f> *)pFields[index]) -> UIReplaceElement ((BIFSDecoder *)this,
																			pMediaObjects[index],
																			pNPos[index]
																			);
						break;
					case MFROTATION:
						((MF<Rotation> *)pFields[index]) -> UIReplaceElement ((BIFSDecoder *)this,
																			pMediaObjects[index],
																			pNPos[index]
																			);
						break;
					case MFIMAGE:
					case MFNODE:
						break;
					}
				} 
				else {
					pFields[index]->Parse((BIFSDecoder *)this,pMediaObjects[index]);
				} 
				
				pMediaObjects[index]->UpdateField(pEntries[index]);
				pFields[index]->OnFieldUpdate();

			}
			else 
			{
				switch (type) {
				case 0:
					ParseInt(1,"bool");
					break;
				case 1:
					ParseInt(32,"float");
					break;
				case 3:
					ParseInt(32,"int");
					break;
				case 4:
					ZString string;
					ParseSFString(string);
					break;
				}

			}
		}
	}
}
