/*
This software module was originally developed by Five Bats Research
in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Five Bats Research retains full right
to use the code for its purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1999.
*/
#ifndef INCLUDED_MPJSCENE_H
#define INCLUDED_MPJSCENE_H

class Executive;
struct MPJSceneDispatchTable;
struct FieldValueStruct;
class MPJNode;

// TEMP to prevent multiple definition
#ifndef MPJException_DEFINED
#define MPJException_DEFINED
class MPJException
{
public:
	MPJException( int error ) : m_err(error){}
	int getErrorCode() const { return m_err; }
private:
	int m_err;
};
#endif


/*
	class MPJScene: The primary interface to the MPEG-J scene API
	implementation.

	There is usually only one instance of this class, and it should
	be initialized from MPEGJTerminal by calling initialize().

	The static methods may be called
	without a MPJScene instance, so it is OK to use them in code
	that must run even when MPEG-J (or the scene API) is not present.
*/
class MPJScene
{
	public:

		/*
			initialize scene API, load DLL. This should normally be called
			when MPEGJTerminal is created.

			@param e The Executive running in IM1
			@return A boolean which is true if the scene API
			        services are available (the DLL loaded and properly
			        initialized)
		*/
		static bool initialize( Executive *e );

		/*
			notify mpeg-j that the scene was replaced

			@param e The Executive running in IM1
			@param initialScene Set to true if this is the initial scene,
								otherwise false
		*/
		static void sceneReplaced( Executive *e, bool initialScene );
		
		/*
			obtain an integer handle that is associated with the current
			scene

			@return the handle that is associated with the current
			        scene. It will be zero if there is no current
					scene.
		*/
		static int getSceneHandle();

		/*
			obtain an MPJScene instance given a scene handle

			@param sceneHandle the handle that is associated with the current
			        scene.
			@return an MPJScene instance given a scene handle. This will be
					zero if there is no current scene.
		*/
		static MPJScene *getScene( int sceneHandle );

		/*
			find a node instance given its node ID within the scene

			@param nodeID The defID of the desired node.
			@param outNodeHandle A handle that is associated with the desired node.

			@return An error code that will be nonzero if an error occurred.
		*/
		int findNode( int nodeID, int *outNodeHandle );

		/*
			get a node ID from the node name

			@param defName The defName of the desired node.
			@param outDefID The defID of the node

			@return An error code that will be nonzero if an error occurred.
		*/
		int getNodeIDFromName( char *defName, int *outDefID );

		/*
			notify the java terminal about an event out that it was interested in

			@param nodeHandle The handle for the node that contains the event.
			@param eventOutID The ID of the event field
			@val The new value of the field.

			@return An error code that will be nonzero if an error occurred.
		*/
		int doEventNotification( int nodeHandle, int eventOutID, struct FieldValueStruct *val );

		/*
			call back into java terminal to request the value of a new field, the value is placed in newFieldValue

			@param param1 an opaque parameter for use by the java terminal to identify the new node
			@param param2 an opaque parameter for use by the java terminal to identify the new node
			@param fieldID the ID of the event field
			@newFieldValue the field value is given in this object

			@return An error code that will be nonzero if an error occurred.
		*/
		int doGetNewFieldValue( int param1, int param2, int fieldID, struct FieldValueStruct *newFieldValue );

	private:
		MPJScene( Executive *e, MPJSceneDispatchTable *t );
		MPJSceneDispatchTable *getDispatchTable() { return s_dispatch; }
		virtual ~MPJScene();

		static MPJSceneDispatchTable *s_dispatch;
		Executive             *m_executive;
		static MPJScene       *s_instance;
};

#endif
