/********************************************************************
 This software module was originally developed by Luca Deri (Finsiel
- I - SOMMIT Consortium) and edited by Roberto Becchini (SIA - I -
SOMMIT Consortium), Barbara Silano (SIA - I - SOMMIT Consortium),
Gianluca De Petris (CSELT - I - SOMMIT Consortium) in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1). This software
module is an implementation of part of one or more MPEG-4 Systems
(ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC
14496-1). ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1)
free license to this software module or modifications thereof for use
in hardware or software products claiming conformance to the MPEG-4
Systems (ISO/IEC 14496-1). Those intending to use this software
module in hardware or software products are advised that its use may
infringe existing patents. THE ORIGINAL DEVELOPER OF THIS SOFTWARE
MODULE AND HIS/HER COMPANY, THE SUBSEQUENT EDITORS AND THEIR
COMPANIES, AND ISO/IEC HAVE NO LIABILITY FOR USE OF THIS SOFTWARE
MODULE OR MODIFICATIONS THEREOF. No license to this software module
is granted for non MPEG-4 Systems (ISO/IEC 14496-1) conforming
products. The SOMMIT Consortium retains full right to use the
software module for his/her own purpose, assign or donate the
software module to a third party and to inhibit third parties from
using the code for non MPEG-4 Systems (ISO/IEC 14496-1) conforming
products. This copyright notice must be included in all copies or
derivative works of the software module. Copyright 1999
********************************************************************/


package org.iso.mpeg.mpegj.resource;

import java.awt.*;
import java.util.*;
import com.ice.jni.registry.Registry;
import com.ice.jni.registry.RegistryValue;
import com.ice.jni.registry.RegistryKey;
import javax.comm.*;


/**
 *
 *
 *   <p>[Adaptive Audio-Visual Session Format ISO/IEC JTC1/SC29/WG11 - MPEG97/2566]<p>
 *    A CapabilityManager monitors latency and bandwidth of network connections and manages
 *    peripheral devices.<p>
 *    Functionality:<br><ol>
 *    <li> Interacts with DMIF.
 *    <li> Interacts with presentation devices.
 *    <li> Throws profile exceptions.
 *    <li> Interacts with input devices.
 *    <li> Throws MPEG4Events.
 *    </ol><p>
 *    The CapabilityManager class handles capability through the CapabilityManager. The
 *    CapabilityManager offers a simple and generic way to handle capabilities.
 *    The CapabilityManager implements some interfaces in order to simplify developers work
 *    when basic capabilities need to be managed. [An interface incapsulates
 *    a coherent set of services and attributes, i.e. a Role, without exlicitly binding
 *    this function to that of any particular object or code].<br>
 *    The interfaces are:
 *    <ol>
 *
 *    <li> DynamicCapability<br>
 *    This interface is used to provide developers a simple way to access basic dynamic
 *    terminal capabilities. A capability is dynamic if it can change value/status at
 *    runtime. For instance memory usage is a dynamic capability.<p>
 *
 *    <li> StaticCapability<br>
 *    This interface is used to provide developers a simple way to access basic dynamic
 *    terminal capabilities. A capability is considered static if its value/static cannot
 *    change at runtime. For instance the audio card type is a static capability.<p>
 *
 *    <li> CapabilityObserver<br>
 *    This interface is used to provide developers a few methods that allow them to
 *    simply monitor the value of common capabilities such as memory and resource usage.<p>
 *    </ol>
 *
 * @author  Luca Deri (l.deri@tecsiel.it)
 * @version 1.0, 11/1998
 */

public class CapabilityManager implements DynamicCapability,
  StaticCapability, DynamicCapabilityObserver, TerminalProfileManager {

    static final short capabilityCode_getFreeTerminalMemory    = 0;
    static final short capabilityCode_getFreeApplicationMemory = 1;
    static final short capabilityCode_getTerminalLoad          = 2;
    static final short capabilityCode_getNetworkLoad           = 3;

    /**
     *
     * The variables below are used to store the observer objects that need
     * to be notified.
     *
     *
     */
  private CapabilityObserver freeAppMemObservers, freeTermMemObservers,
                             termLoadObservers, nwLoadObservers;

    /////////////////////////////////////////////

    /**
    *
    * notifyApplicationFreeMemory
    *
    * @param     freeApplicationMemory     application memory threshold
    * @param     objToNotify               object to notify when the specified threshold is exceeded
    * @exception IllegalArgumentException  the freeApplicationMemory is either negative or out of range.
    *
    */
  public void notifyApplicationFreeMemory(long freeApplicationMemory, Observer objToNotify)
       throws IllegalArgumentException  {

	 if(freeApplicationMemory < 0) throw(new IllegalArgumentException());
	 freeAppMemObservers.addObserver(objToNotify, freeApplicationMemory,
					 false /* below */);
  }

    /* *************************************** */

    /**
    *
    * notifyTerminalFreeMemory
    *
    * @param     freeTerminalMemory        terminal memory threshold
    * @param     objToNotify               object to notify when the specified threshold is exceeded
    * @exception IllegalArgumentException  the freeTerminalMemory is either negative or out of range.
    *
    */
  public void notifyTerminalFreeMemory(long freeTerminalMemory, Observer objToNotify)
       throws IllegalArgumentException  {

	 if(freeTerminalMemory < 0) throw(new IllegalArgumentException());
	 freeTermMemObservers.addObserver(objToNotify, freeTerminalMemory,
					  false /* below */);
  }

    /* *************************************** */

    /**
    *
    * notifyTerminalLoad
    *
    * @param     cpuLoadPercentage         terminal CPU load percentage threshold
    * @param     objToNotify               object to notify when the specified threshold is exceeded
    * @exception IllegalArgumentException  the cpuLoadPercentage is either negative or out of range.
    *
    */
  public void notifyTerminalLoad(long cpuLoadPercentage, Observer objToNotify)
       throws IllegalArgumentException  {

	 if(cpuLoadPercentage < 0) throw(new IllegalArgumentException());
	 termLoadObservers.addObserver(objToNotify, cpuLoadPercentage,
				       true /* above */);
  }

    /* *************************************** */

    /**
    *
    * notifyTerminalNetworkLoad
    *
    * @param     nwLoadPercentage          network load percentage threshold
    * @param     objToNotify               object to notify when the specified threshold is exceeded
    * @exception IllegalArgumentException  the nwLoadPercentage is either negative or out of range.
    *
    */
  public void notifyTerminalNetworkLoad(long nwLoadPercentage, Observer objToNotify)
       throws IllegalArgumentException  {
	 if(nwLoadPercentage < 0) throw(new IllegalArgumentException());
	 nwLoadObservers.addObserver(objToNotify, nwLoadPercentage,
				     true /* above */);
  }


  /* *************************************** */


    /**
    *
    * deleteObserver
    *
    * @param  objToDeregister  observer object previously registered using registerXXX methods
    *
    */
  public void deleteObserver(Observer objToDeregister) {
    // Since we don't know where's regitsered the observer, we do a full deregistr.
    freeAppMemObservers.deleteObserver(objToDeregister);
    freeTermMemObservers.deleteObserver(objToDeregister);
    termLoadObservers.deleteObserver(objToDeregister);
    nwLoadObservers.deleteObserver(objToDeregister);
  }

  /* *************************************** */


    /**
    *
    * Number of Terminal CPUs
    *
    * @return  the number of CPUs that equip the terminal or -1 is an error occurs
    *
    */
  public int getNumCPUs() {
// SOALRIS PORTING
/*    try {
      RegistryKey rKey = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("HARDWARE").openSubKey("DESCRIPTION").openSubKey("System").openSubKey("CentralProcessor");
      return(rKey.getNumberSubkeys());
    } catch(Exception ex) {
      return(1);
    }*/
    return 1;
  }

    /* *************************************** */

    /**
    *
    * Terminal CPU Speed
    *
    * @return  the speed of the 'idx' CPU or -1 is an error occurs
    *
    */
  public int getCPUSpeed(int cpuIdx) {
// SOLARIS PORTING
/*    try {
      RegistryValue value = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("HARDWARE").openSubKey("DESCRIPTION").openSubKey("System").openSubKey("CentralProcessor").openSubKey(String.valueOf(cpuIdx)).getValue("~MHZ");
      byte speed[] = value.getByteData();

      return(256+(int)speed[3]);
    } catch(Exception ex) {
      return(-1);
    }*/
    return 333;
  }

    /* *************************************** */

   /**
    *
    * Terminal CPU type (Pentium for instance)
    *
    * @param   idx   the index of the CPU (idx >= 0)
    * @return  the type of CPU
    * @exception IllegalArgumentException   the index is either negative or out of range.
    *
    */
    public String getCPUType(int cpuIdx) throws IllegalArgumentException {
// SOALRIS POT<S-Del>RTING
/*    RegistryValue value;

    try {
        if(areWeRunningWindows95())
            value = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("Enum").openSubKey("Root").openSubKey("*PNP0C01").openSubKey("0000").getValue("CPU");
        else
            value = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("HARDWARE").openSubKey("DESCRIPTION").openSubKey("System").openSubKey("CentralProcessor").openSubKey(String.valueOf(cpuIdx)).getValue("Identifier");

      return(new String(value.getByteData()));
    } catch(Exception ex) {
        throw(new IllegalArgumentException());
    }*/
    return new String("SPARC");
  }

    /* *************************************** */

    /**
    *
    * Terminal Mouse Type
    *
    * @return  the mouse type or null if an error occurred
    *
    */
  public String getMouseType() {
// SOLARIS PORTING
/*    RegistryValue value;

    try {
        if(areWeRunningWindows95())
            value = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("System").openSubKey("CurrentControlSet").openSubKey("Services").openSubKey("Class").openSubKey("Mouse").openSubKey("0000").getValue("DriverDesc");
        else
            value = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("SOFTWARE").openSubKey("Microsoft").openSubKey("Windows NT").openSubKey("CurrentVersion").openSubKey("WOW").openSubKey("boot.description").getValue("mouse.drv");

      return(new String(value.getByteData()));
    } catch(Exception ex) {
      return(null);
    }*/
    return new String("Standard PS/2 Port Mouse");
  }

    /* *************************************** */

    /**
    *
    * Terminal Display Color Depth
    *
    * @return  the display color depth (bits per pixel)
    *
    */
  public int getDisplayColorDepth() {
    return(Toolkit.getDefaultToolkit().getColorModel().getPixelSize());
  }


    /* *************************************** */

    /**
    *
    * Terminal Display Type (e.g. VGA)
    *
    * @return  the display type
    *
    */
  public String getDisplayType() {
// SOLARIS PORTING
/*    try {
      RegistryValue value;

      if(areWeRunningWindows95())
          value = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("Enum").openSubKey("Monitor").openSubKey("Default_Monitor").openSubKey("0001").getValue("DeviceDesc");
      else
          value = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("SOFTWARE").openSubKey("Microsoft").openSubKey("Windows NT").openSubKey("CurrentVersion").openSubKey("WOW").openSubKey("boot.description").getValue("display.drv");
      return(new String(value.getByteData()));
    } catch(Exception ex) {
      return(null);
    }*/
    return new String("Super VGA 1024x768");
  }

    /* *************************************** */

    /**
    *
    * Terminal Keyboard Type
    *
    * @return  the keyboard type, NULL if the keyboard is not present.
    *
    */public String getKeyboardType() {
// SOLARIS PORTING
/*    try {
      RegistryValue value = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("SOFTWARE").openSubKey("Microsoft").openSubKey("Windows NT").openSubKey("CurrentVersion").openSubKey("WOW").openSubKey("boot.description").getValue("keyboard.typ");
      return(new String(value.getByteData()));
    } catch(Exception ex) {
      return(null);
    }*/
    return new String("Sun Sparc");
  }

    /* *************************************** */

    /**
    *
    * Terminal Network Type
    *
    * @return  the network type or null if an error occurred.
    *
    */
    public String getNetworkType() {
// SOLARIS PORTING
/*    RegistryValue value;

    try {
        if(areWeRunningWindows95())
            value = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("Enum").openSubKey("Network").openSubKey("MSTCP").openSubKey("0001").getValue("DeviceDesc");
        else
            value = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("SOFTWARE").openSubKey("Microsoft").openSubKey("Windows NT").openSubKey("CurrentVersion").openSubKey("WOW").openSubKey("boot.description").getValue("network.drv");

      return(new String(value.getByteData()));
    } catch(Exception ex) {
      return(null);
    }*/
    return new String("TCP/IP");
  }

    /* *************************************** */

    /**
    *
    * Terminal parallel IEEE-1284 port(s)
    *
    * @return  the number of parallel ports present on the terminal.
    *
    */
  public int getNumParallelPorts() {
    Enumeration e = CommPortIdentifier.getPortIdentifiers();
    int num = 0;

    if(e != null)
          while(e.hasMoreElements()) {
    	CommPortIdentifier identifier = (CommPortIdentifier)e.nextElement();
    	if(identifier.getPortType() == CommPortIdentifier.PORT_PARALLEL)
    	  num++;
      }

      return(num);
    }

    /* *************************************** */

    /**
    *
    * Terminal serial RS-232 port(s)
    *
    * @return  the number of serial ports present on the terminal.
    *
    */
    public int getNumSerialPorts() {
      Enumeration e = CommPortIdentifier.getPortIdentifiers();
      int num = 0;

      if(e != null)
    	while(e.hasMoreElements()) {
    	  CommPortIdentifier identifier = (CommPortIdentifier)e.nextElement();
    	  //System.out.println(identifier.getPortType()+" - "+identifier.isCurrentlyOwned());
    	  if(identifier.getPortType() == CommPortIdentifier.PORT_SERIAL)
    	    num++;
	  }

      return(num);
    }

    /* *************************************** */

   /**
    *
    * Terminal screen size
    *
    * @return  the screen size (pixel x pixel)
    *
    */
      public Dimension getScreenSize() {
	    return(Toolkit.getDefaultToolkit().getScreenSize());
      }

    /* *************************************** */

    /**
    *
    * Terminal screen depth
    *
    * @return  the screen depth (bits/inch)
    *
    */
      public short getScreenDepth() {
	    return((short)Toolkit.getDefaultToolkit().getColorModel().getPixelSize());
      }

    /* *************************************** */

    /**
    *
    * Terminal screen resolution
    *
    * @return  the screen resolution (dots/inch)
    *
    */
      public short getScreenResolution() {
	    return((short)Toolkit.getDefaultToolkit().getScreenResolution());
      }

    /* *************************************** */

    /**
    *
    * Terminal OS language
    *
    * @return  the language of the OS version installed on the terminal.
    *
    */   public String getOSLanguage() {
	    return(System.getProperty("user.region"));
      }

      /* *************************************** */

    /**
    *
    * Terminal OS type (for instance JavaPC, Windows, MacOS)
    *
    * @return  the OS type installed on the terminal.
    *
    */
    public String getOSType() {
    	return(System.getProperty("os.name")+
    	       " v."+System.getProperty("os.version"));
    }

    /* *************************************** */

    /**
    *
    * Terminal architecture (for instance Alpha, x86)
    *
    * @return  the terminal architecture type.
    *
    */
    public String getTerminalArchitecture() {
      return(System.getProperty("os.arch"));
    }

    /* *************************************** */

  private String[] getAvailableDrivers(String[] matchingStrings, String win95Prefix) {
// SOLARIS PORTING
/*    if(areWeRunningWindows95())
        return(getAvailableWin95Drivers(matchingStrings, win95Prefix));

    try {
	RegistryKey rKey;
	String rsp[] = new String[1];
	int idx;
	Enumeration key, val;
	Hashtable hashTable = new Hashtable();
	Stack s = new Stack();

    rsp[0] = "";

	if(matchingStrings == null)
	  return(rsp);

	rKey = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("SOFTWARE").openSubKey("Microsoft").openSubKey("Windows NT").openSubKey("CurrentVersion").openSubKey("Drivers32");

	for(val = rKey.valueElements(); val.hasMoreElements();) {
	  String obj = (String)val.nextElement();
	  for(int i=0; i<matchingStrings.length; i++)
	    if(obj.startsWith(matchingStrings[i])) {
	      hashTable.put(new String(rKey.getStringValue(obj)), new String(obj));
	      break;
	    }
	}

    rKey = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("SOFTWARE").openSubKey("Microsoft").openSubKey("Windows NT").openSubKey("CurrentVersion").openSubKey("drivers.desc");

	for(val = rKey.valueElements(); val.hasMoreElements();) {
	  String obj = (String)val.nextElement();
	  String value = rKey.getStringValue(obj);

	  if((obj != null) && (hashTable.get(obj) != null)) {
	    s.push(new String(value));
	  }
	}

	idx = 0;
	rsp = new String[s.size()];
	while(!s.isEmpty())
	  rsp[idx++] = (String)s.pop();

	return(rsp);
      } catch(Exception ex) {
    	String errrsp[] = new String[1];
        errrsp[0] = "";
    	return(errrsp);
      }*/
    	String errrsp[] = new String[1];
        errrsp[0] = "";
    	return(errrsp);
    }

  /* *************************************** */

  private String[] getAvailableWin95Drivers(String[] matchingStrings, String win95Prefix) {
// SOLARIS PORTING
/*      try {
    	RegistryKey rKey;
    	String rsp[] = new String[1];
    	int idx=0;
    	Enumeration key, val;
    	Hashtable hashTable = new Hashtable();
    	Stack s = new Stack();

        rsp[0] = "";

    	if(matchingStrings == null)
    	  return(rsp);

        rKey = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("System").openSubKey("CurrentControlSet").openSubKey("control").openSubKey("MediaResources").openSubKey(win95Prefix);

        if(rKey.getNumberSubkeys() == 0)
            return(rsp);
        else
            rsp = new String[rKey.getNumberSubkeys()];

    	for(val = rKey.keyElements(); val.hasMoreElements();) {
    	  String obj = (String)val.nextElement();
    	  String value = new String(rKey.openSubKey(obj).getValue("Description").getByteData());
    	  rsp[idx++] = new String(value);
          //System.out.println(win95Prefix+"-"+value+"-");
    	}

    	return(rsp);
      } catch(Exception ex) {
    	String errrsp[] = new String[1];
        errrsp[0] = "";
    	return(errrsp);
      }*/
    	String errrsp[] = new String[1];
        errrsp[0] = "";
    	return(errrsp);
    }


    /* *************************************** */

    /**
    *
    * Terminal Audio Enc/Dec
    *
    * @return  the available audio drivers.
    *
    */
    public String[] getAudioDrivers() {
      String[] matchingStrings = { "msacm", "wave" };

      return(getAvailableDrivers(matchingStrings, "wave"));
    }

    /* *************************************** */

    /**
    *
    * Terminal MIDI Enc/Dec
    *
    * @return  the available MIDI drivers.
    *
    */
    public String[] getMIDIDrivers() {
      String[] matchingStrings = { "midi" };

      return(getAvailableDrivers(matchingStrings, "midi"));
    }

    /* *************************************** */

    /**
    *
    * Terminal Video Enc/Dec
    *
    * @return  the available video drivers.
    *
    */
    public String[] getVideoDrivers() {
      String[] matchingStrings = { "vidc." };

      return(getAvailableDrivers(matchingStrings, "icm"));
    }

    /* *************************************** */


    /**
    *
    * Terminal modem type
    *
    * @return  the modem speed ("28K", "56K" etc.) or "<no modem available>" is the modem is not installed.
    *
    */
    public String getModemType() {
// SOLARIS PORTING
/*      try {
    	RegistryValue value;

        if(areWeRunningWindows95()) {
        	value = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("System").openSubKey("CurrentControlSet").openSubKey("Services").openSubKey("Class").openSubKey("Modem").openSubKey("0000").getValue("Model");
    	} else {
        	value = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("System").openSubKey("ControlSet001").openSubKey("Services").openSubKey("Modem").openSubKey("Enum").getValue("0");
        	StringTokenizer modemId = new StringTokenizer(new String(value.getByteData()), "\\");
        	value = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("System").openSubKey("ControlSet001").openSubKey("Enum").openSubKey(modemId.nextToken()).openSubKey(modemId.nextToken()).openSubKey(modemId.nextToken()).getValue("FriendlyName");
    	}

       	return(new String(value.getByteData()));
      } catch(Exception ex) {
    	return("<no modem available>");
      }*/
      return new String("LAN Connection");
    }

    /////////////////////////////////////////////

    /**
    *
    * Terminal free memory (RAM)
    *
    * @return  the amount of free memory (bytes) installed on the terminal.
    *
    */
    public long getFreeTerminalMemory() {
      return((long)CPUmonitor.getFreeSystemMemory());
    }

    /* *************************************** */

    /**
    *
    * Terminal total memory (RAM)
    *
    * @return  the total amount of memory (bytes) installed on the terminal.
    *
    */
    public long getTotalTerminalMemory() {
      return((long)CPUmonitor.getAvailSystemMemory());
    }

    /* *************************************** */

    /**
    *
    * Total application memory (RAM)
    *
    * @return  the total amount of memory (bytes).
    *
    */
  public long getTotalApplicationMemory() {
    return(Runtime.getRuntime().totalMemory());
  }

    /* *************************************** */

    /**
    *
    * Free application memory (RAM)
    *
    * @return  the amount of available (at application level e.g. JavaVM)
    * free memory (bytes).
    *
    */
  public long getFreeApplicationMemory() {
    return(Runtime.getRuntime().freeMemory());
  }

    /* *************************************** */

    /**
    *
    * Terminal CPU Load
    *
    * @return  the load (percentage) of the terminal CPUs
    *
    */
  public long getTerminalLoad() {
    return((long)CPUmonitor.getCPUload());
  }

    /* *************************************** */

    /**
    *
    * Terminal Network Load
    *
    * @return  the network load (percentage).
    *
    */
  public long getNetworkLoad() {
    System.out.println("WARNING: missing method implementation !!! (getNetworkLoad)");
    return 0;
  }

    /////////////////////////////////////////////

  private boolean areWeRunningWindows95() {
// SOLARIS PORTING
/*        try {
    	RegistryValue value = Registry.getTopLevelKey("HKEY_LOCAL_MACHINE").openSubKey("SOFTWARE").openSubKey("Microsoft").openSubKey("Windows").openSubKey("CurrentVersion").getValue("ProductName");
    	String osName = new String(value.getByteData());
    	if(osName.equals("Microsoft Windows 95"))
    	    return(true);
      } catch(Exception ex) { ;}
      return(false);*/
    return(false);
  }

  /////////////////////////////////////////////

    /**
    *
    * Scene Description Profile
    *
    * @return  the actual scene description profile.
    *
    */
  public short getSceneDescriptionProfile() {
    return(TerminalProfileManager.Profile_Simple2D);
  }

    /* *************************************** */
   
  public short getGraphicsCombinationProfile() {
    return(TerminalProfileManager.Profile_2D);
  }    

    /* *************************************** */

    public short getVisualProfile() {
      return(TerminalProfileManager.VisualProfile_Main);
    }

    /* *************************************** */

    public short getAudioProfile() {
      return(TerminalProfileManager.AudioProfile_Main);
    }

    /* *************************************** */

    public CapabilityManager ()
      {
    	freeAppMemObservers = new CapabilityObserver(this,
    				   capabilityCode_getFreeApplicationMemory);
    	freeTermMemObservers = new CapabilityObserver(this,
    				   capabilityCode_getFreeTerminalMemory);
    	termLoadObservers = new CapabilityObserver(this,
    				   capabilityCode_getTerminalLoad);
    	nwLoadObservers = new CapabilityObserver(this,
    				   capabilityCode_getNetworkLoad);

    //	CPUmonitor.getCPUload(); // Dummy call, used just to initialise the counter
      }
  }
