/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1999.
************************************************************************/
package org.iso.mpeg.mpegj.scene;

/**
 * An interface that acts as a proxy for a BIFS node in the scene graph. Only nodes in the scene that have been instanced with DEF may have a Node proxy.
 */
public interface Node {
    /**
     * Updates the value of an eventIn or exposedField in this node of the scene graph. This is a synchronous call that will not return until the field is updated in the scene.
     *
     * @param inID The inID of the field to be updated.
     * @param newValue The desired new value for the field.
     * @exception BadParameterException Thrown if the selected field ID is not valid for this node, or if the value is not an appropriate type for the field.
     * @exception InvalidNodeException Thrown if the node is no longer valid (due to removal or replacement).
     * @see org.iso.mpeg.mpegj.scene.EventIn
     */
    public void sendEventIn(int inID, FieldValue newValue) throws BadParameterException, InvalidNodeException;
    
    /**
     * Reads the current value of an eventOut or exposedField in this node of the scene graph.
     *
     * @param outID The outID of the field to be read.
     * @return a FieldValue containing the value read from the desired field.
     * @exception BadParameterException Thrown if the selected field ID is not valid for this node.
     * @exception InvalidNodeException Thrown if the node is no longer valid (due to removal or replacement).
     * @see org.iso.mpeg.mpegj.scene.EventOut
     */
    public FieldValue getEventOut(int outID) throws BadParameterException, InvalidNodeException;
    
    /**
     * Install a listener for an eventOut or exposedField in this node of the scene graph.
     *
     * @param outID The outID of the field to be monitored.
     * @param listener The listener to notify of changes.
     * @exception BadParameterException Thrown if the selected field ID is not valid for this node.
     * @exception InvalidNodeException Thrown if the node is no longer valid (due to removal or replacement).
     * @see org.iso.mpeg.mpegj.scene.EventOut
     * @see org.iso.mpeg.mpegj.scene.EventOutListener
     */
    public void addEventOutListener(int outID, EventOutListener l) throws BadParameterException, InvalidNodeException;
    
    /**
     * Remove a listener previously added for an eventOut or exposedField in this node of the scene graph.
     *
     * @param outID The outID of the field no longer to be monitored.
     * @param listener The listener that is to be removed.
     * @exception BadParameterException Thrown if the selected field ID is not valid for this node.
     * @exception InvalidNodeException Thrown if the node is no longer valid (due to removal or replacement).
     * @see org.iso.mpeg.mpegj.scene.EventOut
     * @see org.iso.mpeg.mpegj.scene.EventOutListener
     */
    public void removeEventOutListener(int outID, EventOutListener l) throws BadParameterException, InvalidNodeException;

    /**
     * Obtain the node type of the node.
     *
     * @return an int indicating the node type.
     * @exception InvalidNodeException Thrown if the node is no longer valid.
     * @see org.iso.mpeg.mpegj.scene.NodeType
     */
    public int getNodeType() throws InvalidNodeException;
}