/*
	This software module was originally developed by ENST in the course of
development of the MPEG-4 Systems (ISO/IEC 14496-1) standard.
	This software module is an implementation of a part of one or more MPEG-4
Systems (ISO/IEC 14496-1) tools as specified by the MPEG-4 Systems (ISO/IEC 14496-1) 
standard. ISO/IEC gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license 
to this software module or modifications thereof for use in hardware or software 
products claiming conformance to the MPEG-4 Systems (ISO/IEC 14496-1). 
Those intending to use this software module in hardware or software products are 
advised that its use may infringe existing patents.
	The original developer of this software module and his/her company, the 
subsequent editors and their companies, and ISO/IEC have no liability for use of 
this software module or modifications thereof in an implementation. Copyright is 
not released for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. 
	ENST retains full right to use the code for his/her own purpose, assign or 
donate the code to a third party and to inhibit third parties from using the code 
for non MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright    
notice must be included in all copies or derivative works.              
						Copyright  2002.
*/

///////////////////////////////////////////////////////////////////////////
// File:  JSWrapper.cpp
//
// Author: Jean Le Feuvre, ENST
// Description:
//   JS init and bidings, SF/MF type operations
//
// Created: December, 2002.
//	
//	NOTE1: since the core doesn't handle text parsing, SFNode creation is not possible
//	from within the script
//
//	NOTE2: since the core doesn't handle node or route locating without ID, route add/delete
//	is not supported from within the script
//
//	NOTE3: possible operations on SFVec4f from within the script are not defined in the spec, thus
//	not implemented
//
//	NOTE4: Since JavaScript 1.3, the toString method on any objects returns [object ClassName] - this is the case
//	in this implementation too for all MF<T>... Usage of toSource() should be prefered
//
///////////////////////////////////////////////////////////////////////////



#include "defs.h"
#include "Executiv.h"
#include "App.h"
#include "ScriptImp.h" 
#include "JSFieldWrapper.h"

static ScriptImp *GetScriptNode(JSContext *c, JSObject *ptr)
{
  jsval p;
  if (! JS_GetProperty(c, ptr, "_this", &p )) return NULL;

  ScriptImp *pObj  = (ScriptImp *) JSVAL_TO_PRIVATE(p);
  return pObj;
}

static JSClass globalClass = {
    "global", 0,
    JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,   JS_FinalizeStub
};

//to do
static void OnJSError(JSContext *c, const char *msg, JSErrorReport *)
{
	ScriptImp *pObj = (ScriptImp *) JS_GetContextPrivate(c);
	pObj->GetExecutive()->Error(SCRIPT_ERROR, msg);
}


static JSBool Print(JSContext *c, JSObject *p, uintN argc, jsval *argv, jsval *)
{
	uintN i;
	char buf[5000];
	ScriptImp *pObj = GetScriptNode(c, p);
	if (!pObj) return JS_FALSE;

	strcpy(buf, "");
	for (i = 0; i < argc; i++) {
		JSString *str = JS_ValueToString(c, argv[i]);
		if (!str) return JS_FALSE;
		strcat(buf, " ");
		strcat(buf, JS_GetStringBytes(str));
	}
	pObj->GetExecutive()->Error(SCRIPT_MESSAGE, buf);
	return JS_TRUE;
}

static JSFunctionSpec globalFunctions[] = {
    {"print",           Print,          0},
    { 0 }
};

/*
				BIFS browser API definitions
*/
static JSClass browserClass = {
    "Browser", 0,
    JS_PropertyStub,  JS_PropertyStub,  JS_PropertyStub, JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  JS_FinalizeStub
};


static JSBool getName(JSContext *c, JSObject *obj, uintN, jsval *, jsval *rval)
{
	ScriptImp *script = GetScriptNode(c, obj);
	const char *name = script->GetExecutive()->GetApplication()->GetApplicationName();
	*rval = STRING_TO_JSVAL(JS_InternString(c, name ));
	return JS_TRUE;
}
static JSBool getVersion(JSContext*c, JSObject*obj, uintN, jsval *, jsval *rval)
{
	ScriptImp *script = GetScriptNode(c, obj);
	if (!script) return JS_FALSE;
	const char *name = script->GetExecutive()->GetApplication()->GetApplicationVersion();
	*rval = STRING_TO_JSVAL(JS_InternString( c, name ));
	return JS_TRUE;
}
static JSBool getCurrentSpeed(JSContext *c, JSObject*, uintN, jsval *, jsval *rval)
{
	//no speed support for MPEG4
	*rval = DOUBLE_TO_JSVAL(JS_NewDouble( c, 0.0 ));
	return JS_TRUE;
}
static JSBool getCurrentFrameRate(JSContext *c, JSObject*, uintN, jsval*, jsval*rval)
{
	//no "framerate" support for MPEG4 BIFS because of UPDATES
	*rval = DOUBLE_TO_JSVAL(JS_NewDouble(c, 0.0 ));
	return JS_TRUE;
}
static JSBool getWorldURL(JSContext*c, JSObject*obj, uintN, jsval *, jsval *rval)
{
	ScriptImp *script = GetScriptNode(c, obj);
	if (!script) return JS_FALSE;
	const char *name = script->GetExecutive()->GetApplication()->GetWorldURL();
	*rval = STRING_TO_JSVAL(JS_InternString( c, name ));
	return JS_TRUE;
}
static JSBool replaceWorld(JSContext*, JSObject*, uintN, jsval *, jsval *)
{
	//NOT SUPPORTED: we need SFNode creation from text string in the core, not done yet
	return JS_TRUE;
}
static JSBool addRoute(JSContext*, JSObject*, uintN, jsval *, jsval *)
{
	//NOT SUPPORTED: the core Route handling mechanism should be changed to be more generic
	return JS_TRUE;
}
static JSBool deleteRoute(JSContext*, JSObject*, uintN, jsval *, jsval *)
{
	//NOT SUPPORTED: the core Route handling mechanism should be changed to be more generic
	return JS_TRUE;
}

static JSBool loadURL(JSContext*c, JSObject*obj, uintN argc, jsval *argv, jsval *)
{
	if ((int)argc < 1) return JS_FALSE;

	ScriptImp *script = GetScriptNode(c, obj);
	if (!script) return JS_FALSE;
	if (!JSVAL_IS_OBJECT(argv[0])) return JS_FALSE;
	
	JSObject *p;
	JS_ValueToObject(c, argv[0], &p);

	MFWrapper *f = (MFWrapper *) JS_GetPrivate(c, p);
	if (!f) return JS_FALSE;
	jsuint len;
	JS_GetArrayLength(c, f->m_pList, &len);
	jsval item;

	for (unsigned int i=0; i<len; i++) {
		JS_GetElement(c, f->m_pList, (jsint) i, &item);

		if (JSVAL_IS_STRING(item)) {
			JSString *str = JSVAL_TO_STRING(item);
			if (script->GetExecutive()->GetApplication()->Navigate(JS_GetStringBytes(str)) ) return JS_TRUE;
		}
	}
	return JS_TRUE;
}

static JSBool setDescription(JSContext*, JSObject*, uintN, jsval *, jsval *)
{
	//to do...
	return JS_TRUE;
}


static JSFunctionSpec browserFunctions[] = {
  {"getName", getName, 0},
  {"getVersion", getVersion, 0},
  {"getCurrentSpeed", getCurrentSpeed, 0},
  {"getCurrentFrameRate", getCurrentFrameRate, 0},
  {"getWorldURL", getWorldURL, 0},
  {"replaceWorld", replaceWorld, 0},
  {"addRoute", addRoute, 0},
  {"deleteRoute", deleteRoute, 0},
  {"loadURL", loadURL, 0},
  {"setDescription", setDescription, 0},
  {0}
};



/*
				script eventOut API 
*/
JSBool eventOut_setProperty(JSContext *c, JSObject *obj, jsval id, jsval *val)
{
	JSString *str = JS_ValueToString(c, id);
	if (!str) return JS_FALSE;
	const char *eventName = JS_GetStringBytes(str);

	ScriptImp *script = GetScriptNode(c, obj);
	if (!script) return JS_FALSE;

	ScriptNodeField *pField = NULL;
	for (int i=0; i<script->EventOutList.GetLength(); i++) {
		pField = script->EventOutList[i];
		if (!strcmp(pField->m_Name, eventName)) goto found;
	}
	return JS_FALSE;

found:
	script->JSValToNodeField(c, *val, pField->m_pNodeField, script);
	return JS_TRUE;
}


/*
				generic SF field (all but nodes) JS functions
*/

// Generic field destructor
static void field_finalize(JSContext *c, JSObject *obj)
{
	JSField *ptr = (JSField *) JS_GetPrivate(c, obj);
	if (ptr) delete ptr;
	JS_SetPrivate(c, obj, 0 );
}

// Generic SF field -> string
static JSBool field_toString(JSContext *c, JSObject *obj, uintN, jsval *, jsval *rval)
{
	JSField *f = (JSField *) JS_GetPrivate(c, obj);
	if (!f || !f->m_pField) return JS_FALSE;
	ZString str;
	char temp[1000];

	switch (f->m_pField->GetType()) {
	case SFVEC2F:
	case MFVEC2F:
	{
		Vec2f val = ((JS<Vec2f> *)f)->m_value;
		sprintf(temp, "%f %f", val.x, val.y);
		str+=temp;
		break;
	}
	case SFVEC3F:
	case MFVEC3F:
	{
		Vec3f val = ((JS<Vec3f> *)f)->m_value;
		sprintf(temp, "%f %f %f", val.x, val.y, val.z);
		str+=temp;
		break;
	}
	case SFVEC4F:
	case MFVEC4F:
	{
		Vec4f val = ((JS<Vec4f> *)f)->m_value;
		sprintf(temp, "%f %f %f %f", val.x, val.y, val.z, val.q);
		str+=temp;
		break;
	}
	case SFROTATION:
	case MFROTATION:
	{
		Rotation val = ((JS<Rotation> *)f)->m_value;
		sprintf(temp, "%f %f %f %f", val.x, val.y, val.z, val.theta);
		str+=temp;
		break;
	}
	case SFCOLOR:
	case MFCOLOR:
	{
		Color3f val = ((JS<Color3f> *)f)->m_value;
		sprintf(temp, "%f %f %f", val.r, val.g, val.b);
		str+=temp;
		break;
	}
	case SFIMAGE:
	case MFIMAGE:
	{
		Image val = ((JS<Image> *)f)->m_value;
		sprintf(temp, "%dx%dx%d", val.width, val.height, val.numComponents+1);
		str+=temp;
		break;
	}
	}
	JSString *s = JS_NewStringCopyZ(c, (LPCSTR) str);
	if (!s) return JS_FALSE;
	*rval = STRING_TO_JSVAL(s); 
	return JS_TRUE; 
}

/*
			BIFS Node JS functions
*/

static JSBool SFNodeConstructor(JSContext *, JSObject *, uintN , jsval *, jsval *)
{
	//NOT SUPPORTED: we need SFNode creation from text string in the core, not done yet
	return JS_FALSE;
}
static void node_finalize(JSContext *c, JSObject *obj)
{
	JSField *ptr = (JSField *) JS_GetPrivate(c, obj);
	if (ptr) delete ptr;
	JS_SetPrivate(c, obj, 0 );
}
static JSBool node_toString(JSContext *c, JSObject *obj, uintN, jsval *, jsval *rval)
{
	JSNode *f = (JSNode *) JS_GetPrivate(c, obj);
	if (!f || !f->m_pNode) return JS_FALSE;
	ZString str;

	//should make GetName always defined ...
	const char *n = f->m_pNode->GetNameID();
	if (n) {
		str += "DEF ";
		str += n;
		str += " ";
	}

	str += f->m_pNode->GetName();
	str += " {";
	//here we should dump fields i guess
	str += "}";

	JSString *s = JS_NewStringCopyZ(c, (LPCSTR) str);
	if (!s) return JS_FALSE;
	*rval = STRING_TO_JSVAL(s); 
	return JS_TRUE; 
}
static JSBool node_getProperty(JSContext *c, JSObject *obj, jsval id, jsval *vp)
{
	JSNode *ptr = (JSNode *) JS_GetPrivate(c, obj);
	MediaObject *n = ptr->m_pNode;
	JSString *str = 0;
	ScriptImp *script = GetScriptNode(c, JS_GetParent(c, obj) );

	if (n && JSVAL_IS_STRING(id) && ((str = JSVAL_TO_STRING(id)) != 0) && script) {
		char *fieldName = JS_GetStringBytes(str);
		int iFieldIndex = n->GetFieldID(fieldName);
		//field not found...
		if (iFieldIndex<0) return JS_FALSE;
		NodeField *pField = n->GetField(iFieldIndex);
		
		if (pField) {
			*vp = script->NodeFieldToJSVal(pField, false, n);
		}
		return (*vp == JSVAL_NULL) ? JS_FALSE : JS_TRUE;
    }
	return JS_FALSE;
}

static JSBool node_setProperty(JSContext *c, JSObject *obj, jsval id, jsval *vp)
{
	JSNode *ptr = (JSNode *) JS_GetPrivate(c, obj);
	MediaObject *n = ptr->m_pNode;
	ScriptImp *script = GetScriptNode(c, JS_GetParent(c, obj) );

	if (n && JSVAL_IS_STRING(id)) {
		JSString *str = JSVAL_TO_STRING(id);
		char *fieldName = JS_GetStringBytes(str);
		int FieldIndex = n->GetFieldID(fieldName);
		if (FieldIndex<0) return JS_FALSE;
		NodeField *pField = n->GetField(FieldIndex);

		//translate to JS obj
		script->JSValToNodeField(c, *vp, pField, n);
	}
	return JS_TRUE;
}
static JSClass SFNodeClass = {
	"SFNode", JSCLASS_HAS_PRIVATE,
	JS_PropertyStub,  JS_PropertyStub,  node_getProperty, node_setProperty,
	JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  node_finalize
};
static JSFunctionSpec SFNodeMethods[] = {
	{"toString",        node_toString,       0},
	{0}
};



/*
			SFVec2f class functions
*/
static JSBool SFVec2fConstructor(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *)
{
	jsdouble x = 0.0, y = 0.0;
	if (argc > 0) JS_ValueToNumber(c, argv[0], &x);
	if (argc > 1) JS_ValueToNumber(c, argv[1], &y);
	JS<Vec2f> *v = new JS<Vec2f>(Vec2f(x, y));
	JS_SetPrivate(c, obj, v);
	return JS_TRUE;
}
/*
		NOTE on getProperty: it appears that the JS engine calls getProperties for other ids
	than what declared in the properties struct. If FALSE is returned in this case, the toString() method
	(and maybe others) will fail...
*/
static JSBool vec2f_getProperty(JSContext *c, JSObject *obj, jsval id, jsval *vp)
{
	JS<Vec2f> *val = (JS<Vec2f> *) JS_GetPrivate(c, obj);
	if (JSVAL_IS_INT(id)) {
		switch (JSVAL_TO_INT(id)) {
		case 0: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.x)); break;
		case 1: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.y)); break;
		default: return JS_FALSE;
		}
	}
	return JS_TRUE;
}

static JSBool vec2f_setProperty(JSContext *c, JSObject *obj, jsval id, jsval *vp)
{
	JS<Vec2f> *ptr = (JS<Vec2f> *) JS_GetPrivate(c, obj);
	Vec2f val = ptr->m_value;
	jsdouble d;
	if (JSVAL_IS_INT(id) && JSVAL_TO_INT(id) >= 0 && JSVAL_TO_INT(id) < 2 && JS_ValueToNumber(c, *vp, &d)) {
		switch (JSVAL_TO_INT(id)) {
		case 0: val.x = d; break;
		case 1: val.y = d; break;
		default: return JS_FALSE;
		}
		*ptr = val;
		return JS_TRUE;
    }
	return JS_FALSE;
}
static JSClass SFVec2fClass = {
  "SFVec2f", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  vec2f_getProperty, vec2f_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  field_finalize
};
static JSPropertySpec SFVec2fProps[] = {
	{"x",       0,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{"y",       1,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{0}
};
/*
			Vec2f methods
*/
static JSBool vec2f_add(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if (argc<=0 || !JSVAL_IS_OBJECT(argv[0]) || (&SFVec2fClass != JS_GetClass(JSVAL_TO_OBJECT(argv[0]))))
		return JS_FALSE;

	Vec2f v1 = ((JS<Vec2f> *) JS_GetPrivate(c, obj))->m_value;
    Vec2f v2 = ((JS<Vec2f> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(argv[0])))->m_value;
	JSObject *pNew = JS_NewObject(c, &SFVec2fClass, 0, JS_GetParent(c, obj));  
	v1 += v2;
	JS<Vec2f> *v3 = new JS<Vec2f>(v1);
	JS_SetPrivate(c, pNew, v3);
	*rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}
static JSBool vec2f_subtract(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if (argc<=0 || !JSVAL_IS_OBJECT(argv[0]) || (&SFVec2fClass != JS_GetClass(JSVAL_TO_OBJECT(argv[0]))))
		return JS_FALSE;

	Vec2f v1 = ((JS<Vec2f> *) JS_GetPrivate(c, obj))->m_value;
    Vec2f v2 = ((JS<Vec2f> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(argv[0])))->m_value;
	JSObject *pNew = JS_NewObject(c, &SFVec2fClass, 0, JS_GetParent(c, obj));  
	v2 *= -1;
	v1 += v2;
	JS<Vec2f> *v3 = new JS<Vec2f>(v1);
	JS_SetPrivate(c, pNew, v3);
	*rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}
static JSBool vec2f_negate(JSContext *c, JSObject *obj, uintN , jsval *, jsval *rval)
{
	Vec2f v1 = ((JS<Vec2f> *) JS_GetPrivate(c, obj))->m_value;
	JSObject *pNew = JS_NewObject(c, &SFVec2fClass, 0, JS_GetParent(c, obj));  
	v1 *= -1;
	JS<Vec2f> *v3 = new JS<Vec2f>(v1);
	JS_SetPrivate(c, pNew, v3);
	*rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}
static JSBool vec2f_multiply(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if (argc<=0) return JS_FALSE;

	Vec2f v1 = ((JS<Vec2f> *) JS_GetPrivate(c, obj))->m_value;
	JSObject *pNew = JS_NewObject(c, &SFVec2fClass, 0, JS_GetParent(c, obj));  
    jsdouble d = 0;
	JS_ValueToNumber(c, argv[0], &d );
	v1 *= (float) d;
	JS<Vec2f> *v3 = new JS<Vec2f>(v1);
	JS_SetPrivate(c, pNew, v3);
	*rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}
static JSBool vec2f_divide(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if (argc<=0) return JS_FALSE;

	Vec2f v1 = ((JS<Vec2f> *) JS_GetPrivate(c, obj))->m_value;
	JSObject *pNew = JS_NewObject(c, &SFVec2fClass, 0, JS_GetParent(c, obj));  
    jsdouble d = 0;
	JS_ValueToNumber(c, argv[0], &d );
	d = 1 / d;
	v1 *= (float) d;
	JS<Vec2f> *v3 = new JS<Vec2f>(v1);
	JS_SetPrivate(c, pNew, v3);
	*rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}
static JSBool vec2f_length(JSContext *c, JSObject *obj, uintN, jsval *, jsval *rval)
{
	Vec2f v1 = ((JS<Vec2f> *) JS_GetPrivate(c, obj))->m_value;
	*rval = DOUBLE_TO_JSVAL(JS_NewDouble(c, v1.GetLength()) );
	return JS_TRUE;
}
static JSBool vec2f_normalize(JSContext *c, JSObject *obj, uintN, jsval *, jsval *rval)
{
	Vec2f v1 = ((JS<Vec2f> *) JS_GetPrivate(c, obj))->m_value;
	JSObject *pNew = JS_NewObject(c, &SFVec2fClass, 0, JS_GetParent(c, obj));  
	v1.Normalize();
	JS<Vec2f> *v3 = new JS<Vec2f>(v1);
	JS_SetPrivate(c, pNew, v3);
	*rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}
static JSBool vec2f_dot(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if (argc<=0 || !JSVAL_IS_OBJECT(argv[0]) || (&SFVec2fClass != JS_GetClass(JSVAL_TO_OBJECT(argv[0]))))
		return JS_FALSE;

	Vec2f v1 = ((JS<Vec2f> *) JS_GetPrivate(c, obj))->m_value;
    Vec2f v2 = ((JS<Vec2f> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(argv[0])))->m_value;
	*rval = DOUBLE_TO_JSVAL(JS_NewDouble(c, v1 * v2 ) );
	return JS_TRUE;
}

//as defined in VRML
static JSFunctionSpec SFVec2fMethods[] = {
    {"add",             vec2f_add,      1},
    {"divide",          vec2f_divide,   1},
    {"dot",             vec2f_dot,      1},
    {"length",          vec2f_length,   0},
    {"multiply",        vec2f_multiply, 1},
    {"normalize",       vec2f_normalize,0},
    {"subtract",        vec2f_subtract, 1},
	//defined for Vec3f but not Vec2f in VRML, let's add it
    {"negate",          vec2f_negate,   0},
	{"toString",        field_toString,       0},
	{0}
};



/*
			SFVec3f class functions
*/
static JSBool SFVec3fConstructor(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *)
{
	jsdouble x = 0.0, y = 0.0, z = 0.0;
	if (argc > 0) JS_ValueToNumber(c, argv[0], &x);
	if (argc > 1) JS_ValueToNumber(c, argv[1], &y);
	if (argc > 2) JS_ValueToNumber(c, argv[2], &z);
	JS<Vec3f> *v = new JS<Vec3f>(Vec3f(x, y, z));
	JS_SetPrivate(c, obj, v);
	return JS_TRUE;
}
static JSBool vec3f_getProperty(JSContext *c, JSObject *obj, jsval id, jsval *vp)
{
	JS<Vec3f> *val = (JS<Vec3f> *) JS_GetPrivate(c, obj);
	if (JSVAL_IS_INT(id)) {
		switch (JSVAL_TO_INT(id)) {
		case 0: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.x)); break;
		case 1: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.y)); break;
		case 2: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.z)); break;
		default: return JS_FALSE;
		}
	}
	return JS_TRUE;
}

static JSBool vec3f_setProperty(JSContext *c, JSObject *obj, jsval id, jsval *vp)
{
	JS<Vec3f> *ptr = (JS<Vec3f> *) JS_GetPrivate(c, obj);
	Vec3f val = ptr->m_value;
	jsdouble d;
	if (JSVAL_IS_INT(id) && JSVAL_TO_INT(id) >= 0 && JSVAL_TO_INT(id) < 3 && JS_ValueToNumber(c, *vp, &d)) {
		switch (JSVAL_TO_INT(id)) {
		case 0: val.x = d; break;
		case 1: val.y = d; break;
		case 2: val.z = d; break;
		default: return JS_FALSE;
		}
		*ptr = val;
		return JS_TRUE;
    }
	return JS_FALSE;
}
static JSClass SFVec3fClass = {
  "SFVec3f", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  vec3f_getProperty, vec3f_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  field_finalize
};
static JSPropertySpec SFVec3fProps[] = {
	{"x",       0,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{"y",       1,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{"z",       2,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{0}
};
/*
			Vec3f methods
*/
static JSBool vec3f_add(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if (argc<=0 || !JSVAL_IS_OBJECT(argv[0]) || (&SFVec3fClass != JS_GetClass(JSVAL_TO_OBJECT(argv[0]))))
		return JS_FALSE;

	Vec3f v1 = ((JS<Vec3f> *) JS_GetPrivate(c, obj))->m_value;
    Vec3f v2 = ((JS<Vec3f> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(argv[0])))->m_value;
	JSObject *pNew = JS_NewObject(c, &SFVec3fClass, 0, JS_GetParent(c, obj));  
	v1 += v2;
	JS<Vec3f> *v3 = new JS<Vec3f>(v1);
	JS_SetPrivate(c, pNew, v3);
	*rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}
static JSBool vec3f_subtract(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if (argc<=0 || !JSVAL_IS_OBJECT(argv[0]) || (&SFVec3fClass != JS_GetClass(JSVAL_TO_OBJECT(argv[0]))))
		return JS_FALSE;

	Vec3f v1 = ((JS<Vec3f> *) JS_GetPrivate(c, obj))->m_value;
    Vec3f v2 = ((JS<Vec3f> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(argv[0])))->m_value;
	JSObject *pNew = JS_NewObject(c, &SFVec3fClass, 0, JS_GetParent(c, obj));  
	v2 *= -1;
	v1 += v2;
	JS<Vec3f> *v3 = new JS<Vec3f>(v1);
	JS_SetPrivate(c, pNew, v3);
	*rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}
static JSBool vec3f_negate(JSContext *c, JSObject *obj, uintN , jsval *, jsval *rval)
{
	Vec3f v1 = ((JS<Vec3f> *) JS_GetPrivate(c, obj))->m_value;
	JSObject *pNew = JS_NewObject(c, &SFVec3fClass, 0, JS_GetParent(c, obj));  
	v1 *= -1;
	JS<Vec3f> *v3 = new JS<Vec3f>(v1);
	JS_SetPrivate(c, pNew, v3);
	*rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}
static JSBool vec3f_multiply(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if (argc<=0) return JS_FALSE;

	Vec3f v1 = ((JS<Vec3f> *) JS_GetPrivate(c, obj))->m_value;
	JSObject *pNew = JS_NewObject(c, &SFVec3fClass, 0, JS_GetParent(c, obj));  
    jsdouble d = 0;
	JS_ValueToNumber(c, argv[0], &d );
	v1 *= (float) d;
	JS<Vec3f> *v3 = new JS<Vec3f>(v1);
	JS_SetPrivate(c, pNew, v3);
	*rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}
static JSBool vec3f_divide(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if (argc<=0) return JS_FALSE;

	Vec3f v1 = ((JS<Vec3f> *) JS_GetPrivate(c, obj))->m_value;
	JSObject *pNew = JS_NewObject(c, &SFVec3fClass, 0, JS_GetParent(c, obj));  
    jsdouble d = 0;
	JS_ValueToNumber(c, argv[0], &d );
	d = 1 / d;
	v1 *= (float) d;
	JS<Vec3f> *v3 = new JS<Vec3f>(v1);
	JS_SetPrivate(c, pNew, v3);
	*rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}
static JSBool vec3f_length(JSContext *c, JSObject *obj, uintN, jsval *, jsval *rval)
{
	Vec3f v1 = ((JS<Vec3f> *) JS_GetPrivate(c, obj))->m_value;
	*rval = DOUBLE_TO_JSVAL(JS_NewDouble(c, v1.GetLength()) );
	return JS_TRUE;
}
static JSBool vec3f_normalize(JSContext *c, JSObject *obj, uintN, jsval *, jsval *rval)
{
	Vec3f v1 = ((JS<Vec3f> *) JS_GetPrivate(c, obj))->m_value;
	JSObject *pNew = JS_NewObject(c, &SFVec3fClass, 0, JS_GetParent(c, obj));  
	v1.Normalize();
	JS<Vec3f> *v3 = new JS<Vec3f>(v1);
	JS_SetPrivate(c, pNew, v3);
	*rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}
static JSBool vec3f_dot(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if (argc<=0 || !JSVAL_IS_OBJECT(argv[0]) || (&SFVec3fClass != JS_GetClass(JSVAL_TO_OBJECT(argv[0]))))
		return JS_FALSE;

	Vec3f v1 = ((JS<Vec3f> *) JS_GetPrivate(c, obj))->m_value;
    Vec3f v2 = ((JS<Vec3f> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(argv[0])))->m_value;
	*rval = DOUBLE_TO_JSVAL(JS_NewDouble(c, v1 * v2 ) );
	return JS_TRUE;
}
static JSBool vec3f_cross(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if (argc<=0 || !JSVAL_IS_OBJECT(argv[0]) || (&SFVec3fClass != JS_GetClass(JSVAL_TO_OBJECT(argv[0]))))
		return JS_FALSE;

	Vec3f v1 = ((JS<Vec3f> *) JS_GetPrivate(c, obj))->m_value;
    Vec3f v2 = ((JS<Vec3f> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(argv[0])))->m_value;
	JSObject *pNew = JS_NewObject(c, &SFVec3fClass, 0, JS_GetParent(c, obj));  
	JS<Vec3f> *v3 = new JS<Vec3f>(v1.Cross(v2));
	JS_SetPrivate(c, pNew, v3);
	*rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}

//as defined in VRML
static JSFunctionSpec SFVec3fMethods[] = {
    {"add",             vec3f_add,      1},
    {"cross",           vec3f_cross,    1},
    {"divide",          vec3f_divide,   1},
    {"dot",             vec3f_dot,      1},
    {"length",          vec3f_length,   0},
    {"multiply",        vec3f_multiply, 1},
    {"normalize",       vec3f_normalize,0},
    {"subtract",        vec3f_subtract, 1},
    {"negate",          vec3f_negate,   0},
	{"toString",        field_toString,       0},
	{0}
};


/*
			SFVec4f class functions
*/
static JSBool SFVec4fConstructor(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *)
{
	jsdouble x = 0.0, y = 0.0, z = 0.0, q = 0.0;
	if (argc > 0) JS_ValueToNumber(c, argv[0], &x);
	if (argc > 1) JS_ValueToNumber(c, argv[1], &y);
	if (argc > 2) JS_ValueToNumber(c, argv[2], &z);
	if (argc > 3) JS_ValueToNumber(c, argv[3], &q);
	JS<Vec4f> *v = new JS<Vec4f>(Vec4f(x, y, z, q));
	JS_SetPrivate(c, obj, v);
	return JS_TRUE;
}
static JSBool vec4f_getProperty(JSContext *c, JSObject *obj, jsval id, jsval *vp)
{
	JS<Vec4f> *val = (JS<Vec4f> *) JS_GetPrivate(c, obj);
	if (JSVAL_IS_INT(id)) {
		switch (JSVAL_TO_INT(id)) {
		case 0: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.x)); break;
		case 1: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.y)); break;
		case 2: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.z)); break;
		case 3: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.q)); break;
		default: return JS_FALSE;
		}
	}
	return JS_TRUE;
}

static JSBool vec4f_setProperty(JSContext *c, JSObject *obj, jsval id, jsval *vp)
{
	JS<Vec4f> *ptr = (JS<Vec4f> *) JS_GetPrivate(c, obj);
	Vec4f val = ptr->m_value;
	jsdouble d;
	if (JSVAL_IS_INT(id) && JSVAL_TO_INT(id) >= 0 && JSVAL_TO_INT(id) < 3 && JS_ValueToNumber(c, *vp, &d)) {
		switch (JSVAL_TO_INT(id)) {
		case 0: val.x = d; break;
		case 1: val.y = d; break;
		case 2: val.z = d; break;
		case 3: val.q = d; break;
		default: return JS_FALSE;
		}
		*ptr = val;
		return JS_TRUE;
    }
	return JS_FALSE;
}
static JSClass SFVec4fClass = {
  "SFVec4f", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  vec4f_getProperty, vec4f_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  field_finalize
};
static JSPropertySpec SFVec4fProps[] = {
	{"x",       0,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{"y",       1,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{"z",       2,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{"q",       3,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{0}
};
/*
			Vec4f methods: none for now, i can't get my hands on anything in the AFX spec specifying
		the supported methods...
*/
static JSFunctionSpec SFVec4fMethods[] = {
	{"toString",        field_toString,       0},
	{0}
};


/*
			SFRotation class functions
*/
static JSBool SFRotationConstructor(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *)
{
	jsdouble x = 0.0, y = 0.0, z = 0.0, a = 0.0;
	JS<Rotation> *r;

	//first constructor: SFRotation(float, float, float, float)
	if (JSVAL_IS_NUMBER(argv[0]) ) {
		if (argc > 0) JS_ValueToNumber(c, argv[0], &x);
		if (argc > 1) JS_ValueToNumber(c, argv[1], &y);
		if (argc > 2) JS_ValueToNumber(c, argv[2], &z);
		if (argc > 3) JS_ValueToNumber(c, argv[3], &a);
		r = new JS<Rotation>(Rotation(x, y, z, a));
		JS_SetPrivate(c, obj, r);
		return JS_TRUE;
	}
	if (argc != 2) return JS_FALSE;
	if (! JSVAL_IS_OBJECT(argv[0]) 
		|| ( &SFVec3fClass != JS_GetClass( JSVAL_TO_OBJECT(argv[0]) )) )
		return JS_FALSE;
	
	JS<Vec3f> *v = (JS<Vec3f> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(argv[0]));

	//second constructor: SFRotation(Vec3f, float)
	if (JSVAL_IS_NUMBER(argv[1]) ) {
		JS_ValueToNumber( c, argv[1], &a);
		r = new JS<Rotation>(Rotation(v->m_value.x, v->m_value.y, v->m_value.z, a));
		JS_SetPrivate(c, obj, r);
		return JS_TRUE;
	}
	//last constructor: new SFRotation(Vec3f, Vec3f): rotation from v1 to v2
	if (! JSVAL_IS_OBJECT(argv[1]) 
		|| ( &SFVec3fClass != JS_GetClass( JSVAL_TO_OBJECT(argv[1]) )) )
		return JS_FALSE;

	JS<Vec3f> *v2 = (JS<Vec3f> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(argv[1]));
	Vec3f i = v->m_value;
	i.Cross(v2->m_value);
	a = acos( v->m_value * v2->m_value / (v->m_value.GetLength() * v2->m_value.GetLength()) );
	r = new JS<Rotation>(Rotation(i.x, i.y, i.z, a));
	JS_SetPrivate(c, obj, r);
	return JS_TRUE;
	
}
static JSBool rotation_getProperty(JSContext *c, JSObject *obj, jsval id, jsval *vp)
{
	JS<Rotation> *val = (JS<Rotation> *) JS_GetPrivate(c, obj);
	if (JSVAL_IS_INT(id)) {
		switch (JSVAL_TO_INT(id)) {
		case 0: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.x)); break;
		case 1: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.y)); break;
		case 2: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.z)); break;
		case 3: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.theta)); break;
		default: return JS_FALSE;
		}
	}
	return JS_TRUE;
}

static JSBool rotation_setProperty(JSContext *c, JSObject *obj, jsval id, jsval *vp)
{
	JS<Rotation> *ptr = (JS<Rotation> *) JS_GetPrivate(c, obj);
	Rotation val = ptr->m_value;
	jsdouble d;
	if (JSVAL_IS_INT(id) && JSVAL_TO_INT(id) >= 0 && JSVAL_TO_INT(id) < 3 && JS_ValueToNumber(c, *vp, &d)) {
		switch (JSVAL_TO_INT(id)) {
		case 0: val.x = d; break;
		case 1: val.y = d; break;
		case 2: val.z = d; break;
		case 3: val.theta = d; break;
		default: return JS_FALSE;
		}
		*ptr = val;
		return JS_TRUE;
    }
	return JS_FALSE;
}
static JSClass SFRotationClass = {
  "SFRotation", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  rotation_getProperty, rotation_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  field_finalize
};
static JSPropertySpec SFRotationProps[] = {
	{"x",       0,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{"y",       1,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{"z",       2,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{"angle",   3,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{0}
};

/*
			Rotation methods
*/

static JSBool rotation_getAxis(JSContext *c, JSObject *obj, uintN , jsval *, jsval *rval)
{
    JS<Rotation> *r = (JS<Rotation> *) JS_GetPrivate(c, obj);
	//retuns axis in an SFVec3f object
	JSObject *pNew = JS_NewObject(c, &SFVec3fClass, 0, JS_GetParent(c, obj));  
	Vec3f v;
	v.x = r->m_value.x;
	v.y = r->m_value.y;
	v.z = r->m_value.z;
	JS<Vec3f> *val = new JS<Vec3f>(v);
	JS_SetPrivate(c, pNew, val);
	*rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}

static JSBool rotation_inverse(JSContext *c, JSObject *obj, uintN , jsval *, jsval *rval)
{
    JS<Rotation> *r = (JS<Rotation> *) JS_GetPrivate(c, obj);
	JSObject *pNew = JS_NewObject(c, &SFRotationClass, 0, JS_GetParent(c, obj));  
	Rotation _r = r->m_value;
	_r.theta *= -1;
	JS<Rotation> *val = new JS<Rotation>(_r);
	JS_SetPrivate(c, pNew, val);
    *rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}

static JSBool rotation_multiply(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if (argc!=1) return JS_FALSE;
	if (! JSVAL_IS_OBJECT(argv[0]) || (&SFRotationClass == JS_GetClass( JSVAL_TO_OBJECT(argv[0]) ) ) )return JS_FALSE;

    JS<Rotation> *r1 = (JS<Rotation> *) JS_GetPrivate(c, obj);
    JS<Rotation> *r2 = (JS<Rotation> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(argv[0]) );
	JSObject *pNew = JS_NewObject(c, &SFRotationClass, 0, JS_GetParent(c, obj));  
	Rotation r3 = r1->m_value;
	r3 *= r2->m_value;
	JS<Rotation> *val = new JS<Rotation>(r3);
	JS_SetPrivate(c, pNew, val);
    *rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}

static JSBool rotation_multVec(JSContext *, JSObject *, uintN , jsval *, jsval *)
{
	//not implemented yet
	return JS_FALSE;
}

static JSBool rotation_setAxis(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *)
{
	if (argc!=1) return JS_FALSE;
	if (! JSVAL_IS_OBJECT(argv[0]) || (&SFVec3fClass == JS_GetClass( JSVAL_TO_OBJECT(argv[0]) ) ) )return JS_FALSE;
    JS<Rotation> *r = (JS<Rotation> *) JS_GetPrivate(c, obj);
    JS<Vec3f> *vec = (JS<Vec3f> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(argv[0]) );

	Rotation rot = r->m_value;
	rot.x = vec->m_value.x;
	rot.y = vec->m_value.y;
	rot.z = vec->m_value.z;
	//according to VRML this modifies directly the rotation
	*r = rot;
	return JS_TRUE;
}

static JSBool rotation_slerp(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	if (argc!=2) return JS_FALSE;
	if (! JSVAL_IS_OBJECT(argv[0]) || (&SFRotationClass == JS_GetClass( JSVAL_TO_OBJECT(argv[0]) ) ) )return JS_FALSE;

    JS<Rotation> *r1 = (JS<Rotation> *) JS_GetPrivate(c, obj);
    JS<Rotation> *r2 = (JS<Rotation> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(argv[0]) );
	jsdouble interp;
	JS_ValueToNumber(c, argv[1], &interp);
	Rotation newR = r1->m_value.Slerp(r2->m_value, (float) interp);
	JS<Rotation> *val = new JS<Rotation>(newR);
	JSObject *pNew = JS_NewObject(c, &SFRotationClass, 0, JS_GetParent(c, obj));  
	JS_SetPrivate(c, pNew, val);
    *rval = OBJECT_TO_JSVAL(pNew);
	return JS_TRUE;
}
static JSFunctionSpec SFRotationMethods[] = {
    {"getAxis",         rotation_getAxis,    0},
    {"inverse",         rotation_inverse,    0},
    {"multiply",        rotation_multiply,   1},
    {"multVec",         rotation_multVec,    1},
    {"setAxis",         rotation_setAxis,    1},
    {"slerp",           rotation_slerp,      2},
	{"toString",        field_toString,       0},
	{0}
};

/*
			SFColor class functions
*/
static JSBool SFColorConstructor(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *)
{
	jsdouble r = 0.0, g = 0.0, b = 0.0;
	if (argc > 0) JS_ValueToNumber(c, argv[0], &r);
	if (argc > 1) JS_ValueToNumber(c, argv[1], &g);
	if (argc > 2) JS_ValueToNumber(c, argv[2], &b);
	JS<Color3f> *v = new JS<Color3f>(Color3f(r, g, b));
	JS_SetPrivate(c, obj, v);
	return JS_TRUE;
}
static JSBool color_getProperty(JSContext *c, JSObject *obj, jsval id, jsval *vp)
{
	JS<Color3f> *val = (JS<Color3f> *) JS_GetPrivate(c, obj);

	if (JSVAL_IS_INT(id)) {
		switch (JSVAL_TO_INT(id)) {
		case 0: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.r)); break;
		case 1: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.g)); break;
		case 2: *vp = DOUBLE_TO_JSVAL(JS_NewDouble(c, val->m_value.b)); break;
		default: return JS_FALSE;
		}
	}
	//not ours, return TRUE!!
	return JS_TRUE;
}

static JSBool color_setProperty(JSContext *c, JSObject *obj, jsval id, jsval *vp)
{
	JS<Color3f> *ptr = (JS<Color3f> *) JS_GetPrivate(c, obj);
	Color3f val = ptr->m_value;
	jsdouble d;
	if (JSVAL_IS_INT(id) && JSVAL_TO_INT(id) >= 0 && JSVAL_TO_INT(id) < 3 && JS_ValueToNumber(c, *vp, &d)) {
		switch (JSVAL_TO_INT(id)) {
		case 0: val.r = d; break;
		case 1: val.g = d; break;
		case 2: val.b = d; break;
		default: return JS_FALSE;
		}
		*ptr = val;
		return JS_TRUE;
    }
	return JS_FALSE;
}
static JSClass SFColorClass = {
  "SFColor", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub,  color_getProperty, color_setProperty,
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  field_finalize
};
static JSPropertySpec SFColorProps[] = {
	{"r",       0,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{"g",       1,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{"b",       2,       JSPROP_ENUMERATE | JSPROP_PERMANENT},
	{0}
};
/*
			Color methods
*/

static JSBool color_setHSV(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *)
{
	if (argc != 3) return JS_FALSE;
	jsdouble h, s, v;
	JS<Color3f> *col = (JS<Color3f> *) JS_GetPrivate(c, obj );
	JS_ValueToNumber( c, argv[0], &h );
	JS_ValueToNumber( c, argv[1], &s );
	JS_ValueToNumber( c, argv[2], &v );
	Color3f newcol;
	newcol.r = h;
	newcol.g = s;
	newcol.b = v;
	newcol.fromHSV();
	*col = newcol;
	return JS_TRUE;
}

static JSBool color_getHSV(JSContext *c, JSObject *obj, uintN , jsval *, jsval *rval)
{
	Color3f col = * (Color3f *) (JS<Color3f> *) JS_GetPrivate(c, obj );
	float hsv[3];
	col.toHSV();
	hsv[0] = col.r;
	hsv[1] = col.g;
	hsv[2] = col.b;
	jsval vec[3];
	jsdouble *d;
	d = JS_NewDouble(c, col.r);
	vec[0] = DOUBLE_TO_JSVAL(d);
	d = JS_NewDouble(c, col.g);
	vec[1] = DOUBLE_TO_JSVAL(d);
	d = JS_NewDouble(c, col.b);
	vec[2] = DOUBLE_TO_JSVAL(d);
	JSObject *arr = JS_NewArrayObject(c, 3, vec);
	*rval = OBJECT_TO_JSVAL(arr);
	return JS_TRUE;
}
static JSFunctionSpec SFColorMethods[] = {
    {"setHSV",          color_setHSV,   3, 0, 0},
    {"getHSV",          color_getHSV,   0, 0, 0},
	{"toString",        field_toString,       0, 0, 0},
	{0, 0, 0, 0, 0}
};



/*
			MF<*> class functions
	
	"*" can be of any type, including nodes (not tested yet)
	This class is just an object container where all SF items are stored as object or native JS types
*/

static JSBool MFArrayCons(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	MFWrapper *ptr = new MFWrapper();
	ptr->m_pList = JS_NewArrayObject(c, (jsint) argc, argv);
	JS_SetPrivate(c, obj, ptr);
	*rval = OBJECT_TO_JSVAL(obj);
	return obj == 0 ? JS_FALSE : JS_TRUE;
}

static void array_finalize(JSContext *c, JSObject *obj)
{
	MFWrapper *ptr = (MFWrapper *) JS_GetPrivate(c, obj);
	if (ptr) delete ptr;
	JS_SetPrivate(c, obj, 0 );
}

JSBool array_getElement(JSContext *c, JSObject *obj, jsval id, jsval *rval)
{
	MFWrapper *ptr = (MFWrapper *) JS_GetPrivate(c, obj);
	if (JSVAL_IS_INT(id)) {
		int k = JSVAL_TO_INT(id);
		JS_GetElement(c, ptr->m_pList, k, rval);
		k = JSVAL_TO_INT(*rval);
	}
	return JS_TRUE;
}

//this could be overloaded for each MF type...
JSBool array_setElement(JSContext *c, JSObject *obj, jsval id, jsval *rval)
{
	MFWrapper *ptr = (MFWrapper *) JS_GetPrivate(c, obj);
	int ind = JSVAL_TO_INT(id);
	int k, iFieldIndex;
	JSBool ret;
	k = JSVAL_TO_INT(*rval);
	
	ret = JS_SetElement(c, ptr->m_pList, ind, rval);

	if (!ptr->m_pField) {
		ret = JSVAL_IS_STRING(*rval);
		JSString *str = JSVAL_TO_STRING(*rval);
		LPCSTR tmp;
		tmp = JS_GetStringBytes(str);
		return ret;
	}
	BOOL changed = TRUE;
	jsdouble d;

	switch (ptr->m_pField->GetType()) {
	case MFBOOL:
	{
		if (! JSVAL_IS_BOOLEAN(*rval)) return JS_FALSE;
		Bool val = JSVAL_TO_BOOLEAN(*rval);
		MF<Bool> *f = (MF<Bool> *) ptr->m_pField;
		if ((f->GetLength()>ind) && (val == (*f)[ind])) changed = FALSE;
		if (changed) f->SetAt(ind, val);
		break;
	}
	case MFINT32:
	{
		if (! JSVAL_IS_INT(*rval)) return JS_FALSE;
		int val = JSVAL_TO_INT(*rval);
		MF<int> *f = (MF<int> *) ptr->m_pField;
		if ((f->GetLength()>ind) && (val == (*f)[ind])) changed = FALSE;
		if (changed) f->SetAt(ind, val);
		break;
	}
	case MFFLOAT:
	{
		if (! JSVAL_IS_NUMBER(*rval)) return JS_FALSE;
		JS_ValueToNumber(c, *rval, &d );
		MF<float> *f = (MF<float> *) ptr->m_pField;
		if ((f->GetLength()>ind) && ((float) d == (*f)[ind])) changed = FALSE;
		if (changed) f->SetAt(ind, (float) d);
		break;
	}
	case MFTIME:
	{
		if (! JSVAL_IS_NUMBER(*rval)) return JS_FALSE;
		JS_ValueToNumber(c, *rval, &d );
		MF<float> *f = (MF<float> *) ptr->m_pField;
		if ((f->GetLength()>ind) && ((float) d == (*f)[ind])) changed = FALSE;
		if (changed) f->SetAt(ind, (float) d);
		break;
	}
	//MF string is assumed to have changed ...
	case MFSTRING:
	{
		//don't check type, just convert to string
		JSString *str = JS_ValueToString(c, *rval);
		if (!str) return JS_FALSE;
		MF<ZString> *f = (MF<ZString> *) ptr->m_pField;
		f->SetAt(ind, JS_GetStringBytes(str));
		break;
	}
	//MF url is assumed to have changed ...
	case MFURL:
	{
		//don't check type, just convert to string
		JSString *str = JS_ValueToString(c, *rval);
		MF<Url> *f = (MF<Url> *) ptr->m_pField;
		//blindly set the urlValue field and rely on user to understand "od:" scheme
		Url newVal;
		newVal.urlValue = (ZString) JS_GetStringBytes(str);
		f->SetAt(ind, newVal);
		f->OnFieldUpdate();
		//this is dirty but since "url" type doesn't exist we have to do it that way to reload the url field
		iFieldIndex = ptr->m_pOwner->GetFieldID("url");
		if (iFieldIndex>=0) {
			ptr->m_pOwner->UpdateField(ptr->m_pOwner->GetFieldTable() + iFieldIndex, TRUE);
		} else {
			changed = FALSE;
		}
		break;
	}

	//non native types
	case MFVEC2F:
	{
		if (! JSVAL_IS_OBJECT(*rval)) return JS_FALSE;
		JS<Vec2f> * item = (JS<Vec2f> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(*rval) );
		MF<Vec2f> *f = (MF<Vec2f> *) ptr->m_pField;
		if ((f->GetLength()>ind) && (item->m_value == (*f)[ind])) changed = FALSE;
		if (changed) f->SetAt(ind, item->m_value);
		break;
	}
	case MFVEC3F:
	{
		if (! JSVAL_IS_OBJECT(*rval)) return JS_FALSE;
		JS<Vec3f> * item = (JS<Vec3f> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(*rval) );
		MF<Vec3f> *f = (MF<Vec3f> *) ptr->m_pField;
		if ((f->GetLength()>ind) && (item->m_value == (*f)[ind])) changed = FALSE;
		if (changed) f->SetAt(ind, item->m_value);
		break;
	}
	case MFVEC4F:
	{
		if(!JSVAL_IS_OBJECT(*rval))return JS_FALSE;
		JS<Vec4f> * item = (JS<Vec4f> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(*rval) );
		MF<Vec4f> *f = (MF<Vec4f> *) ptr->m_pField;
		if ((f->GetLength()>ind) && (item->m_value == (*f)[ind])) changed = FALSE;
		if (changed) f->SetAt(ind, item->m_value);
		break;
	}
	case MFROTATION:
	{
		if (! JSVAL_IS_OBJECT(*rval)) return JS_FALSE;
		JS<Rotation> * item = (JS<Rotation> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(*rval) );
		MF<Rotation> *f = (MF<Rotation> *) ptr->m_pField;
		if ((f->GetLength()>ind) && (item->m_value == (*f)[ind])) changed = FALSE;
		if (changed) f->SetAt(ind, item->m_value);
		break;
	}
	case MFCOLOR:
	{
		if (!JSVAL_IS_OBJECT(*rval)) return JS_FALSE;
		JS<Color3f> * item = (JS<Color3f> *) JS_GetPrivate(c, JSVAL_TO_OBJECT(*rval) );
		MF<Color3f> *f = (MF<Color3f> *) ptr->m_pField;
		if ((f->GetLength()>ind) && (item->m_value == (*f)[ind])) changed = FALSE;
		if (changed) f->SetAt(ind, item->m_value);
		break;
	}
	case MFNODE:
	{
		if (!JSVAL_IS_OBJECT(*rval)) return JS_FALSE;
		JSNode *item = (JSNode *) JS_GetPrivate(c, JSVAL_TO_OBJECT(*rval) );
		MFGenericNode *f = (MFGenericNode *) ptr->m_pField;
		if ((f->GetLength()>ind) && (item->m_pNode == (*f)[ind])) changed = FALSE;
		if (changed) f->SetAt(ind, item->m_pNode, ptr->m_pOwner);
		break;
	}

	default:
		return JS_FALSE;
	}
	if (changed) {
		ptr->m_pOwner->SetUpdateTime();
		ptr->m_pOwner->GetExecutive()->InvalidateScene(ptr->m_pOwner);
	}
	return ret;
}

static JSBool array_toString(JSContext *c, JSObject *obj, uintN, jsval *, jsval *rval)
{
	MFWrapper *arr = (MFWrapper *) JS_GetPrivate(c, obj);
	if (!arr) return JS_FALSE;
	jsuint len;
	JS_GetArrayLength(c, arr->m_pList, &len);

	ZString str;

	jsval item;
	for (unsigned int i=0; i<len;i++) {
		JS_GetElement(c, arr->m_pList, i, &item);
		JSString * const s = JS_ValueToString(c, item);
		str += JS_GetStringBytes(s);
		str += ", ";
	}
	JSString *s = JS_NewStringCopyZ(c, (LPCSTR) str);
	if (!s) return JS_FALSE;
	*rval = STRING_TO_JSVAL(s); 
	return JS_TRUE;
}

JSBool array_setLength(JSContext *c, JSObject *obj, jsval , jsval *val)
{
	MFWrapper *ptr = (MFWrapper *) JS_GetPrivate(c, obj);
	if (!JSVAL_IS_INT(*val) || JSVAL_TO_INT(*val) < 0) return JS_FALSE;
	int len = JSVAL_TO_INT(*val);
	return JS_SetArrayLength(c, ptr->m_pList, len);
}
JSBool array_getLength(JSContext *c, JSObject *obj, jsval , jsval *val)
{
	MFWrapper *ptr = (MFWrapper *) JS_GetPrivate(c, obj);
	jsuint len;
	JSBool ret = JS_GetArrayLength(c, ptr->m_pList, &len);
	*val = INT_TO_JSVAL(len);
	return ret;
}
static JSPropertySpec MFArrayProp[] = {
	{ "length", 0, JSPROP_PERMANENT, array_getLength, array_setLength },
	{ 0, 0, 0, 0, 0 } 
};
static JSFunctionSpec MFArrayMethods[] = {
	{"toString",        array_toString,       0},
	{0}
};

static JSClass MFInt32Class = {
  "MFInt32", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub, array_getElement,  array_setElement,  
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  array_finalize,
};
static JSClass MFBoolClass = {
  "MFBool", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub, array_getElement,  array_setElement,  
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  array_finalize,
};
static JSClass MFTimeClass = {
  "MFTime", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub, array_getElement,  array_setElement,  
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  array_finalize,
};
static JSClass MFFloatClass = {
  "MFFloat", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub, array_getElement,  array_setElement,  
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  array_finalize,
};
static JSClass MFUrlClass = {
  "MFUrl", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub, array_getElement,  array_setElement,  
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  array_finalize,
};
static JSClass MFStringClass = {
  "MFString", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub, array_getElement,  array_setElement,  
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  array_finalize,
};


/*
		NOTE on constructors: i follow the annex C of VRML on ECMA script regarding MF constructions:
			new MF<T>(T1, T2, T3...)
*/
/*
			MFVec2f class functions
*/
static JSBool MFVec2fCons(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	MFWrapper *ptr = new MFWrapper();
	ptr->m_pList = JS_NewArrayObject(c, 0, 0);
	JS_SetArrayLength(c, ptr->m_pList, argc);
	JS_SetPrivate(c, obj, ptr);

	jsval val;
	JSObject *item;
	for (unsigned int i=0; i<argc; i++) {
		if (!JSVAL_IS_OBJECT(argv[i]) || ( &SFVec2fClass != JS_GetClass( JSVAL_TO_OBJECT(argv[i])) ) ) {
			item = JS_ConstructObject(c, &SFVec2fClass, 0, obj);
			val = OBJECT_TO_JSVAL(item);
			JS_SetElement(c, ptr->m_pList, i, &val);
		} else {
			JS_SetElement(c, ptr->m_pList, i, &argv[i]);
		}
	}
	*rval = OBJECT_TO_JSVAL(obj);
	return obj == 0 ? JS_FALSE : JS_TRUE;
}

static JSClass MFVec2fClass = {
  "MFVec2f", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub, array_getElement,  array_setElement,  
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  array_finalize,
};

/*
			MFVec3f class functions
*/
static JSBool MFVec3fCons(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	MFWrapper *ptr = new MFWrapper();
	ptr->m_pList = JS_NewArrayObject(c, 0, 0);
	JS_SetArrayLength(c, ptr->m_pList, argc);
	JS_SetPrivate(c, obj, ptr);

	jsval val;
	JSObject *item;
	for (unsigned int i=0; i<argc; i++) {
		if (!JSVAL_IS_OBJECT(argv[i]) || ( &SFVec3fClass != JS_GetClass( JSVAL_TO_OBJECT(argv[i])) ) ) {
			item = JS_ConstructObject(c, &SFVec3fClass, 0, obj);
			val = OBJECT_TO_JSVAL(item);
			JS_SetElement(c, ptr->m_pList, i, &val);
		} else {
			JS_SetElement(c, ptr->m_pList, i, &argv[i]);
		}
	}
	*rval = OBJECT_TO_JSVAL(obj);
	return obj == 0 ? JS_FALSE : JS_TRUE;
}
static JSClass MFVec3fClass = {
  "MFVec3f", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub, array_getElement,  array_setElement,  
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  array_finalize,
};

/*
			MFVec4f class functions
*/
static JSBool MFVec4fCons(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	MFWrapper *ptr = new MFWrapper();
	ptr->m_pList = JS_NewArrayObject(c, 0, 0);
	JS_SetArrayLength(c, ptr->m_pList, argc);
	JS_SetPrivate(c, obj, ptr);

	jsval val;
	JSObject *item;
	for (unsigned int i=0; i<argc; i++) {
		if (!JSVAL_IS_OBJECT(argv[i]) || ( &SFVec4fClass != JS_GetClass( JSVAL_TO_OBJECT(argv[i])) ) ) {
			item = JS_ConstructObject(c, &SFVec4fClass, 0, obj);
			val = OBJECT_TO_JSVAL(item);
			JS_SetElement(c, ptr->m_pList, i, &val);
		} else {
			JS_SetElement(c, ptr->m_pList, i, &argv[i]);
		}
	}
	*rval = OBJECT_TO_JSVAL(obj);
	return obj == 0 ? JS_FALSE : JS_TRUE;
}
static JSClass MFVec4fClass = {
  "MFVec4f", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub, array_getElement,  array_setElement,  
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  array_finalize,
};



/*
			MFRotation class functions
*/
static JSBool MFRotationCons(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	MFWrapper *ptr = new MFWrapper();
	ptr->m_pList = JS_NewArrayObject(c, 0, 0);
	JS_SetArrayLength(c, ptr->m_pList, argc);
	JS_SetPrivate(c, obj, ptr);

	jsval val;
	JSObject *item;
	for (unsigned int i=0; i<argc; i++) {
		if (!JSVAL_IS_OBJECT(argv[i]) || ( &SFRotationClass != JS_GetClass( JSVAL_TO_OBJECT(argv[i])) ) ) {
			item = JS_ConstructObject(c, &SFRotationClass, 0, obj);
			val = OBJECT_TO_JSVAL(item);
			JS_SetElement(c, ptr->m_pList, i, &val);
		} else {
			JS_SetElement(c, ptr->m_pList, i, &argv[i]);
		}
	}
	*rval = OBJECT_TO_JSVAL(obj);
	return obj == 0 ? JS_FALSE : JS_TRUE;
}
static JSClass MFRotationClass = {
  "MFRotation", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub, array_getElement,  array_setElement,  
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  array_finalize,
};


/*
			MFColor class functions
*/
static JSBool MFColorCons(JSContext *c, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
	MFWrapper *ptr = new MFWrapper();
	ptr->m_pList = JS_NewArrayObject(c, 0, 0);
	JS_SetArrayLength(c, ptr->m_pList, argc);
	JS_SetPrivate(c, obj, ptr);

	jsval val;
	JSObject *item;
	for (unsigned int i=0; i<argc; i++) {
		if (!JSVAL_IS_OBJECT(argv[i]) || ( &SFColorClass != JS_GetClass( JSVAL_TO_OBJECT(argv[i])) ) ) {
			item = JS_ConstructObject(c, &SFColorClass, 0, obj);
			val = OBJECT_TO_JSVAL(item);
			JS_SetElement(c, ptr->m_pList, i, &val);
		} else {
			JS_SetElement(c, ptr->m_pList, i, &argv[i]);
		}
	}
	*rval = OBJECT_TO_JSVAL(obj);
	return obj == 0 ? JS_FALSE : JS_TRUE;
}
static JSClass MFColorClass = {
  "MFColor", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub, array_getElement,  array_setElement,  
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  array_finalize,
};

static JSClass MFNodeClass = {
  "MFNode", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub,  JS_PropertyStub, array_getElement,  array_setElement,  
  JS_EnumerateStub, JS_ResolveStub,   JS_ConvertStub,  array_finalize,
};


//inits all JS stuff
BOOL ScriptImp::InitJSAPI()
{
	JS_SetErrorReporter(m_context, OnJSError);

	m_pObj = JS_NewObject(m_context, &globalClass, 0, 0 );
	JS_InitStandardClasses(m_context, m_pObj);
	JS_DefineFunctions(m_context, m_pObj, globalFunctions );

	JS_DefineProperty(m_context, m_pObj, "FALSE", BOOLEAN_TO_JSVAL(false), 0, 0, JSPROP_READONLY | JSPROP_PERMANENT );
	JS_DefineProperty(m_context, m_pObj, "TRUE", BOOLEAN_TO_JSVAL(true), 0, 0, JSPROP_READONLY | JSPROP_PERMANENT );
	JS_DefineProperty(m_context, m_pObj, "_this", PRIVATE_TO_JSVAL(this), 0, 0, JSPROP_READONLY | JSPROP_PERMANENT );

	m_pBrowser = JS_DefineObject(m_context, m_pObj, "Browser", &browserClass, 0, 0 );
	JS_DefineProperty(m_context, m_pBrowser, "_this", PRIVATE_TO_JSVAL(this), 0, 0, JSPROP_READONLY | JSPROP_PERMANENT );
	JS_DefineFunctions(m_context, m_pBrowser, browserFunctions);

	//define types
	//SF Types

	JS_InitClass(m_context, m_pObj, 0, &SFVec2fClass, SFVec2fConstructor, 0, SFVec2fProps, SFVec2fMethods, 0, 0);
	JS_InitClass(m_context, m_pObj, 0, &SFVec3fClass, SFVec3fConstructor, 0, SFVec3fProps, SFVec3fMethods, 0, 0);
	JS_InitClass(m_context, m_pObj, 0, &SFVec4fClass, SFVec4fConstructor, 0, SFVec4fProps, SFVec4fMethods, 0, 0);
	JS_InitClass(m_context, m_pObj, 0, &SFRotationClass, SFRotationConstructor, 0, SFRotationProps, SFRotationMethods, 0, 0);
	JS_InitClass(m_context, m_pObj, 0, &SFColorClass, SFColorConstructor, 0, SFColorProps, SFColorMethods, 0, 0);

	// SFNode
	JS_InitClass( m_context, m_pObj, 0, &SFNodeClass, SFNodeConstructor, 1, 0, SFNodeMethods, 0, 0);

	//all MF types are ArrayClass object filled with SF objects
	JS_InitClass(m_context, m_pObj, 0, &MFInt32Class, MFArrayCons, 0, MFArrayProp, 0, 0, 0);
	JS_InitClass(m_context, m_pObj, 0, &MFBoolClass, MFArrayCons, 0, MFArrayProp, 0, 0, 0);
	JS_InitClass(m_context, m_pObj, 0, &MFFloatClass, MFArrayCons, 0, MFArrayProp, 0, 0, 0);
	JS_InitClass(m_context, m_pObj, 0, &MFTimeClass, MFArrayCons, 0, MFArrayProp, 0, 0, 0);
	JS_InitClass(m_context, m_pObj, 0, &MFStringClass, MFArrayCons, 0, MFArrayProp, 0, 0, 0);
	JS_InitClass(m_context, m_pObj, 0, &MFUrlClass, MFArrayCons, 0, MFArrayProp, 0, 0, 0);

	JS_InitClass(m_context, m_pObj, 0, &MFVec2fClass, MFVec2fCons, 0, MFArrayProp, MFArrayMethods, 0, 0);
	JS_InitClass(m_context, m_pObj, 0, &MFVec3fClass, MFVec3fCons, 0, MFArrayProp, 0, 0, 0);
	JS_InitClass(m_context, m_pObj, 0, &MFVec4fClass, MFVec4fCons, 0, MFArrayProp, 0, 0, 0);
	JS_InitClass(m_context, m_pObj, 0, &MFRotationClass, MFRotationCons, 0, MFArrayProp, 0, 0, 0);
	JS_InitClass(m_context, m_pObj, 0, &MFColorClass, MFColorCons, 0, MFArrayProp, 0, 0, 0);

	JS_InitClass(m_context, m_pObj, 0, &MFNodeClass, MFArrayCons, 0, MFArrayProp, 0, 0, 0);

	return TRUE;
}




//used in the following...
#define SETUP_JS_MFFIELD	ptr = (MFWrapper *) JS_GetPrivate(m_context, obj); \
							ptr->m_pField = pField;		\
							ptr->m_pOwner = pNode;

jsval ScriptImp::NodeFieldToJSVal(NodeField *pField, BOOL protect, MediaObject *pNode)
{
	JSObject *obj = NULL;
	JSField *pJSField = NULL;
	jsdouble *d;
	MFWrapper *ptr;

	double test;

	switch (pField->GetType()) {
//simple types mapped to JS native types
    case SFBOOL:
		return BOOLEAN_TO_JSVAL( (Bool) * ((SF<Bool> *) pField));
	case SFINT32:
		return INT_TO_JSVAL( *((SF<int> *) pField));
    case SFFLOAT:
		d = JS_NewDouble(m_context, (double) *((SF<float> *) pField));
		if (protect) JS_AddRoot(m_context, d);
		return DOUBLE_TO_JSVAL(d);
	case SFTIME:
		test = *((SF<Time> *) pField);
		d = JS_NewDouble(m_context, (double) *((SF<Time> *) pField));
		if (protect) JS_AddRoot(m_context, d);
		return DOUBLE_TO_JSVAL(d);
	case SFSTRING:
    {
		JSString *s = JS_NewStringCopyZ(m_context, (LPCSTR) (ZString) (* (SF<ZString> *)pField) );
		if (protect) JS_AddRoot(m_context, s);
		return STRING_TO_JSVAL( s );
    }
	//we handle URL as a string in the script engine (thus using "od:")
	case SFURL:
    {
		Url url = * (SF<Url> *)pField;
		ZString str;
		if (url.ODid >= 0) {
			char msg[30];
			sprintf(msg, "od:%d", url.ODid);
			str += msg;
		} else {
			str += url.urlValue;
		}
		JSString *s = JS_NewStringCopyZ(m_context, (LPCSTR) str );
		if (protect) JS_AddRoot(m_context, s);
		return STRING_TO_JSVAL( s );
    }
//other SF types 
    case SFVEC2F:
		pJSField = new JS<Vec2f>(* ((SF<Vec2f> *) pField) );
		pJSField->SetParentField(pField, pNode);
		obj = JS_NewObject(m_context, &SFVec2fClass, 0, m_pObj);
		break;
    case SFVEC3F:
		pJSField = new JS<Vec3f>(* ((SF<Vec3f> *)pField) );
		pJSField->SetParentField(pField, pNode);
		obj = JS_NewObject(m_context, &SFVec3fClass, 0, m_pObj);
		break;
    case SFVEC4F:
		pJSField = new JS<Vec4f>(* ((SF<Vec4f> *)pField) );
		pJSField->SetParentField(pField, pNode);
		obj = JS_NewObject(m_context, &SFVec4fClass, 0, m_pObj);
		break;
    case SFROTATION:
		pJSField = new JS<Rotation>(* ((SF<Rotation> *)pField) );
		pJSField->SetParentField(pField, pNode);
		obj = JS_NewObject(m_context, &SFRotationClass, 0, m_pObj);
		break;
    case SFCOLOR:
		pJSField = new JS<Color3f>(* ((SF<Color3f> *)pField) );
		pJSField->SetParentField(pField, pNode);
		obj = JS_NewObject(m_context, &SFColorClass, 0, m_pObj);
		break;
		
    case SFNODE:
		pJSField = new JSNode;
		pJSField->SetParentField(pField, pNode);
		obj = JS_NewObject(m_context, &SFNodeClass, 0, m_pObj);
		break;

//MF mapped to native types
	case MFBOOL:
	{
		obj = JS_ConstructObject(m_context, &MFBoolClass, 0, m_pObj);
		SETUP_JS_MFFIELD;
		MF<Bool> *f = (MF<Bool> *) pField;
		for (int i = 0; i<f->GetLength(); i++) {
			int val = (*f)[i];
			jsval newVal = BOOLEAN_TO_JSVAL(val);
			JS_SetElement(m_context, obj, (jsint) i, &newVal);
		}
		break;
	}
	case MFINT32:
	{
		obj = JS_ConstructObject(m_context, &MFInt32Class, 0, m_pObj);
		SETUP_JS_MFFIELD;
		MF<int> *f = (MF<int> *) pField;
		for (int i = 0; i<f->GetLength(); i++) {
			int val = (*f)[i];
			jsval newVal = INT_TO_JSVAL(val);
			JS_SetElement(m_context, obj, (jsint) i, &newVal);
		}
		JSString *str = JS_ValueToString(m_context, OBJECT_TO_JSVAL(obj));
		LPCSTR tmp;
		tmp = JS_GetStringBytes(str);
		break;
	}
	case MFFLOAT:
	{
		obj = JS_ConstructObject(m_context, &MFFloatClass, 0, m_pObj);
		SETUP_JS_MFFIELD;
		MF<float> *f = (MF<float> *) pField;
		for (int i = 0; i<f->GetLength(); i++) {
			d = JS_NewDouble(m_context, (*f)[i]);
			jsval newVal = DOUBLE_TO_JSVAL(d);
			JS_SetElement(m_context, obj, (jsint) i, &newVal);
		}
		break;
	}
	case MFTIME:
	{
		obj = JS_ConstructObject(m_context, &MFTimeClass, 0, m_pObj);
		SETUP_JS_MFFIELD;
		MF<Time> *f = (MF<Time> *) pField;
		for (int i = 0; i<f->GetLength(); i++) {
			d = JS_NewDouble(m_context, (*f)[i]);
			jsval newVal = DOUBLE_TO_JSVAL(d);
			JS_SetElement(m_context, obj, (jsint) i, &newVal);
		}
		break;
	}
	case MFSTRING:
	{
		obj = JS_ConstructObject(m_context, &MFStringClass, 0, m_pObj);
		SETUP_JS_MFFIELD;
		MF<ZString> *f = (MF<ZString> *) pField;
		for (int i = 0; i<f->GetLength(); i++) {
			JSString *s = JS_NewStringCopyZ(m_context, (LPCSTR) (*f)[i]);
			jsval newVal = STRING_TO_JSVAL( s );
			JS_SetElement(m_context, obj, (jsint) i, &newVal);
		}
		break;
	}
	case MFURL:
	{
		obj = JS_ConstructObject(m_context, &MFUrlClass, 0, m_pObj);
		SETUP_JS_MFFIELD;
		MF<Url> *f = (MF<Url> *) pField;
		for (int i = 0; i<f->GetLength(); i++) {
			ZString str;
			if ((*f)[i].ODid >= 0) {
				char msg[30];
				sprintf(msg, "od:%d", (*f)[i].ODid);
				str += msg;
			} else {
				str += (*f)[i].urlValue;
			}
			JSString *s = JS_NewStringCopyZ(m_context, (LPCSTR) str);
			jsval newVal = STRING_TO_JSVAL( s );
			JS_SetElement(m_context, obj, (jsint) i, &newVal);
		}
		break;
	}

//other MF types	
	case MFVEC2F:
	{
		obj = JS_ConstructObject(m_context, &MFVec2fClass, 0, m_pObj);
		SETUP_JS_MFFIELD;
		MF<Vec2f> *f = (MF<Vec2f> *) pField;
		for (int i = 0; i<f->GetLength(); i++) {
			JS<Vec2f> *val = new JS<Vec2f>((*f)[i]);
			val->SetParentField(pField, pNode, i);
			JSObject *pf = JS_NewObject(m_context, &SFVec2fClass, 0, obj);
			JS_SetPrivate(m_context, pf, val);
			jsval newVal = OBJECT_TO_JSVAL(pf);
			JS_SetElement(m_context, obj, (jsint) i, &newVal);
		}
		break;
	}
	case MFVEC3F:
	{
		obj = JS_ConstructObject(m_context, &MFVec3fClass, 0, m_pObj);
		SETUP_JS_MFFIELD;
		MF<Vec3f> *f = (MF<Vec3f> *) pField;
		for (int i = 0; i<f->GetLength(); i++) {
			JS<Vec3f> *val = new JS<Vec3f>((*f)[i]);
			val->SetParentField(pField, pNode, i);
			JSObject *pf = JS_NewObject(m_context, &SFVec3fClass, 0, obj);
			JS_SetPrivate(m_context, pf, val);
			jsval newVal = OBJECT_TO_JSVAL(pf);
			JS_SetElement(m_context, obj, (jsint) i, &newVal);
		}
		break;
	}
	case MFVEC4F:
	{
		obj = JS_ConstructObject(m_context, &MFVec4fClass, 0, m_pObj);
		SETUP_JS_MFFIELD;
		MF<Vec4f> *f = (MF<Vec4f> *) pField;
		for (int i = 0; i<f->GetLength(); i++) {
			JS<Vec4f> *val = new JS<Vec4f>((*f)[i]);
			val->SetParentField(pField, pNode, i);
			JSObject *pf = JS_NewObject(m_context, &SFVec4fClass, 0, obj);
			JS_SetPrivate(m_context, pf, val);
			jsval newVal = OBJECT_TO_JSVAL(pf);
			JS_SetElement(m_context, obj, (jsint) i, &newVal);
		}
		break;
	}
	case MFROTATION:
	{
		obj = JS_ConstructObject(m_context, &MFRotationClass, 0, m_pObj);
		SETUP_JS_MFFIELD;
		MF<Rotation> *f = (MF<Rotation> *) pField;
		for (int i = 0; i<f->GetLength(); i++) {
			JS<Rotation> *val = new JS<Rotation>((*f)[i]);
			val->SetParentField(pField, pNode, i);
			JSObject *pf = JS_NewObject(m_context, &SFRotationClass, 0, obj);
			JS_SetPrivate(m_context, pf, val);
			jsval newVal = OBJECT_TO_JSVAL(pf);
			JS_SetElement(m_context, obj, (jsint) i, &newVal);
		}
		break;
	}

	case MFCOLOR:
	{
		obj = JS_ConstructObject(m_context, &MFColorClass, 0, m_pObj);
		SETUP_JS_MFFIELD;
		MF<Color3f> *f = (MF<Color3f> *) pField;
		for (int i = 0; i<f->GetLength(); i++) {
			JS<Color3f> *val = new JS<Color3f>((*f)[i]);
			val->SetParentField(pField, pNode, i);
			JSObject *pf = JS_NewObject(m_context, &SFColorClass, 0, obj);
			JS_SetPrivate(m_context, pf, val);
			jsval newVal = OBJECT_TO_JSVAL(pf);
			JS_SetElement(m_context, obj, (jsint) i, &newVal);
		}
		break;
	}

	case MFNODE:
	{
		obj = JS_ConstructObject(m_context, &MFNodeClass, 0, m_pObj);
		SETUP_JS_MFFIELD;
		MFGenericNode *f = (MFGenericNode *) pField;
		for (int i = 0; i<f->GetLength(); i++) {
			JSNode *val = new JSNode;
			val->SetParentField(pField, pNode, i);
			JSObject *pf = JS_NewObject(m_context, &SFNodeClass, 0, obj);
			JS_SetPrivate(m_context, pf, val);
			jsval newVal = OBJECT_TO_JSVAL(pf);
			JS_SetElement(m_context, obj, (jsint) i, &newVal);
		}
		break;
	}

	//not supported
    default:
		return JSVAL_NULL;
    }

	if (!obj) return JSVAL_NULL;
	//store field associated with object if needed
	if (pJSField) JS_SetPrivate(m_context, obj, pJSField);
	
	//protection from garbage collector
	if (protect) JS_AddRoot(m_context, obj);
	return OBJECT_TO_JSVAL(obj);
}


void ScriptImp::CheckFieldChange(BOOL changed, MediaObject *pNode, NodeField *pField)
{
	if (!changed) return;
	if (pField) pField->OnFieldUpdate();
	if (pNode) {
		pNode->SetUpdateTime();
		pNode->OnUpdate();
		pNode->GetExecutive()->InvalidateScene(pNode);
	}
}

void ScriptImp::JSValToNodeField(JSContext *cx, jsval _val, NodeField *pField, MediaObject * pNode)
{
	jsdouble d;

	if (JSVAL_IS_NULL(_val) || JSVAL_IS_VOID(_val)) return;

	/*
		VRML: If directOutput is FALSE (the default), the script may only affect the rest of the world via events 
		sent through its eventOuts. The results are undefined if directOutput is FALSE and the script sends 
		events directly to a node to which it has access.

		** This implementation ignores directOutput **
	*/
	BOOL changed = FALSE;

	switch (pField->GetType()) {
	case SFBOOL:
	{
		if (!JSVAL_IS_BOOLEAN(_val)) return;
		Bool v = JSVAL_TO_BOOLEAN(_val);
		changed = ! ( (Bool) *((SF<Bool> *) pField) == v);
		*((SF<Bool> *) pField) = v;
		CheckFieldChange(changed, pNode, pField);
		return;
	}
	case SFINT32:
	{
		if (! JSVAL_IS_INT(_val) ) return;
		SF<int> p = JSVAL_TO_INT(_val);
		changed = ! (* ((SF<int> *) pField) == p);
		*((SF<int> *) pField) = p;
		CheckFieldChange(changed, pNode, pField);
		return;
	}
	case SFFLOAT:
	{
		if (! JSVAL_IS_NUMBER(_val) ) return;
		JS_ValueToNumber(cx, _val, &d );
		changed = ! (*((SF<float> *) pField) == (float) (double) d);
		*((SF<float> *) pField) = (float) (double) d;
		CheckFieldChange(changed, pNode, pField);
		return;
    }

	case SFTIME:
	{
		if (! JSVAL_IS_NUMBER(_val) )return;
		JS_ValueToNumber( cx, _val, &d );
		changed = ! (*((SF<Time> *) pField) == (double) d);
		*((SF<Time> *) pField) = (double) d;
		CheckFieldChange(changed, pNode, pField);
		return;
    }
	case SFSTRING:
	{
		if (! JSVAL_IS_STRING(_val)) return;
		JSString *str = JSVAL_TO_STRING(_val);
		*((SF<ZString> *) pField) = JS_GetStringBytes(str);
		CheckFieldChange(changed, pNode, pField);
		return;
	}
	case SFURL:
	{
		if (! JSVAL_IS_STRING(_val)) return;
		JSString *str = JSVAL_TO_STRING(_val);
		((Url *) (SF<Url> *) pField)->urlValue = (ZString) JS_GetStringBytes(str);
		( (Url *) (SF<Url> *) pField)->ODid = -1;
		CheckFieldChange(changed, pNode, pField);
		if (pNode) {
			int iFieldIndex = pNode->GetFieldID("url");
			if (iFieldIndex>=0) {
				pNode->UpdateField(pNode->GetFieldTable() + iFieldIndex, TRUE);
			}
		}
		return;
	}

	default:
		break;
	}

	//from here we must have an object
	if (! JSVAL_IS_OBJECT(_val)) return;
	JSObject *obj = JSVAL_TO_OBJECT(_val) ;

	switch (pField->GetType()) {
	case SFVEC2F:
	{
		if ( &SFVec2fClass != JS_GetClass( obj ) ) return;
		JS<Vec2f> *p = (JS<Vec2f> *) JS_GetPrivate(cx, obj);
		changed = ! ( (Vec2f) ( *((SF<Vec2f> *)pField) ) == p->m_value);
		*((SF<Vec2f> *)pField) = p->m_value;
		CheckFieldChange(changed, pNode, pField);
		return;
	}
	case SFVEC3F:
	{
		if ( &SFVec3fClass != JS_GetClass( obj ) ) return;
		JS<Vec3f> *p = (JS<Vec3f> *) JS_GetPrivate(cx, obj);
		changed = ! ( (Vec3f) ( *((SF<Vec3f> *)pField) ) == p->m_value);
		*((SF<Vec3f> *)pField) = p->m_value;
		CheckFieldChange(changed, pNode, pField);
		return;
	}
	case SFVEC4F:
	{
		if ( &SFVec4fClass != JS_GetClass( obj ) ) return;
		JS<Vec4f> *p = (JS<Vec4f> *) JS_GetPrivate(cx, obj);
		changed = ! ( (Vec4f) ( *((SF<Vec4f> *)pField) ) == p->m_value);
		*((SF<Vec4f> *)pField) = p->m_value;
		CheckFieldChange(changed, pNode, pField);
		return;
	}
	case SFCOLOR:
	{
		if ( &SFColorClass != JS_GetClass( obj ) ) return;
		JS<Color3f> *p = (JS<Color3f> *) JS_GetPrivate(cx, obj);
		changed = ! ( (Color3f) ( *((SF<Color3f> *)pField) ) == p->m_value);
		*((SF<Color3f> *)pField) = p->m_value;
		CheckFieldChange(changed, pNode, pField);
		return;
	}
	case SFNODE:
	{
		if ( &SFNodeClass != JS_GetClass( obj ) ) return;
		JSNode *pNode;
		pNode = (JSNode *) JS_GetPrivate(cx, obj);
		//TODO
		return;
	}
	default:
		break;
	}

	//from here wa handle only MF fields
	if ( ( &MFBoolClass != JS_GetClass( obj ) )
		&& ( &MFInt32Class != JS_GetClass( obj ) )
		&& ( &MFFloatClass != JS_GetClass( obj ) )
		&& ( &MFTimeClass != JS_GetClass( obj ) )
		&& ( &MFStringClass != JS_GetClass( obj ) )
		&& ( &MFUrlClass != JS_GetClass( obj ) )
		&& ( &MFVec2fClass != JS_GetClass( obj ) )
		&& ( &MFVec3fClass != JS_GetClass( obj ) )
		&& ( &MFVec4fClass != JS_GetClass( obj ) )
		&& ( &MFRotationClass != JS_GetClass( obj ) )
		&& ( &MFColorClass != JS_GetClass( obj ) )
		) return;

	MFWrapper *ptr = (MFWrapper *) JS_GetPrivate(cx, obj);
	jsuint len;
	JS_GetArrayLength(m_context, ptr->m_pList, &len);
	jsval item;

	//Note: we could check if the field has changed (check each slot in the MF) or not but that's a bit of a pain
	//so empty and copy... to optimize
	((MFGenericNode *)pField)->Remove();


	BOOL bIsURL = FALSE;
	for (unsigned int i=0; i<len; i++) {
		JS_GetElement(m_context, ptr->m_pList, (jsint) i, &item);

		switch (pField->GetType()) {
//native to MF
		case MFBOOL:
		{
			if (JSVAL_IS_BOOLEAN(item)) {
				((MF<Bool> *) pField)->Append((Bool) JSVAL_TO_BOOLEAN(item), FALSE);
			}
			break;
		}
		case MFINT32:
		{
			if (JSVAL_IS_INT(item)) {
				((MF<int> *) pField)->Append((int) JSVAL_TO_INT(item), FALSE);
			}
			break;
		}
		case MFFLOAT:
		{
			if (JSVAL_IS_NUMBER(item)) {
				JS_ValueToNumber(m_context, item, &d);
				((MF<float> *) pField)->Append((float) d, FALSE);
			}
			break;
		}
		case MFTIME:
		{
			if (JSVAL_IS_NUMBER(item)) {
				JS_ValueToNumber(m_context, item, &d);
				((MF<Time> *) pField)->Append((Time) d, FALSE);
			}
			break;
		}
		case MFSTRING:
		{
			if (JSVAL_IS_STRING(item)) {
				JSString *str = JSVAL_TO_STRING(item);
				ZString s = JS_GetStringBytes(str);
				((MF<ZString> *) pField)->Append((ZString) s, FALSE);
			}
			break;
		}
		case MFURL:
		{
			if (JSVAL_IS_STRING(item)) {
				JSString *str = JSVAL_TO_STRING(item);
				Url s;
				s.urlValue = (ZString) JS_GetStringBytes(str);
				((MF<Url> *) pField)->Append(s, FALSE);
			}
			bIsURL = TRUE;
			break;
		}

//other MF types
		case MFVEC2F:
		{
			if (JSVAL_IS_OBJECT(item)) {
				JS<Vec2f> *val = (JS<Vec2f> *) JS_GetPrivate(cx, JSVAL_TO_OBJECT(item));
				((MF<Vec2f> *) pField)->Append( val->m_value, FALSE);
			}
			break;
		}
		case MFVEC3F:
		{
			if (JSVAL_IS_OBJECT(item)) {
				JS<Vec3f> *val = (JS<Vec3f> *) JS_GetPrivate(cx, JSVAL_TO_OBJECT(item));
				((MF<Vec3f> *) pField)->Append( val->m_value, FALSE);
			}
			break;
		}
		case MFVEC4F:
		{
			if (JSVAL_IS_OBJECT(item)) {
				JS<Vec4f> *val = (JS<Vec4f> *) JS_GetPrivate(cx, JSVAL_TO_OBJECT(item));
				((MF<Vec4f> *) pField)->Append( val->m_value, FALSE);
			}
			break;
		}
		case MFROTATION:
		{
			if (JSVAL_IS_OBJECT(item)) {
				JS<Rotation> *val = (JS<Rotation> *) JS_GetPrivate(cx, JSVAL_TO_OBJECT(item));
				((MF<Rotation> *) pField)->Append( val->m_value, FALSE);
			}
			break;
		}
		case MFCOLOR:
		{
			if (JSVAL_IS_OBJECT(item)) {
				JS<Color3f> *val = (JS<Color3f> *) JS_GetPrivate(cx, JSVAL_TO_OBJECT(item));
				((MF<Color3f> *) pField)->Append( val->m_value, FALSE);
			}
			break;
		}
		default:
			return;
		}
	}

	//check URL changes
	if (bIsURL && pNode) {
		int iFieldIndex = pNode->GetFieldID("url");
		if (iFieldIndex>=0) {
			pNode->UpdateField(pNode->GetFieldTable() + iFieldIndex, TRUE);
		}
	}

	//and update field
	CheckFieldChange(TRUE, pNode, pField);
}

void ScriptImp::InitFields()
{
	int i;

	ScriptNodeField *pField;

	//init eventIns
    for (i = 0; i < EventInList.GetLength(); i++) {
    }

	//init fields
    for ( i = 0; i < FieldList.GetLength(); i++) { 
		pField = FieldList[i];
		jsval val = NodeFieldToJSVal(pField->m_pNodeField, false, this);
		JS_DefineProperty(m_context, m_pObj, (LPCSTR) pField->m_Name, val, 0, 0, JSPROP_PERMANENT);
    }

	//init eventOuts
    for ( i = 0; i < EventOutList.GetLength(); i++) { 
		pField = EventOutList[i];
		jsval val = NodeFieldToJSVal( pField->m_pNodeField, false, this);
		if (JSVAL_IS_OBJECT(val)) {
			JS_DefineProperty(m_context, JSVAL_TO_OBJECT(val), "_eventOut", PRIVATE_TO_JSVAL((long int) (pField->m_Name) ), 0, 0, JSPROP_READONLY | JSPROP_PERMANENT );
		}
		JS_DefineProperty(m_context, m_pObj, (LPCSTR) pField->m_Name, val, 0, eventOut_setProperty, JSPROP_PERMANENT );
    }
}
