/*-------------------------------------------------------------------------- 
// 
//  Copyright (C) IBM Corporation, 1996, 1997, 1998
//  Licensed Materials, Program Property of IBM 
// 
//  This software module is an implementation of a part of one or more
//  MPEG-4 tools as specified by the MPEG-4.  ISO/IEC gives users of the
//  MPEG-4 free license to this software module or modifications thereof
//  for use in hardware or software products claiming conformance to the
//  MPEG-4.  Those intending to use this software module in hardware or
//  software products are advised that its use may infringe existing
//  patents.  The original developer of this software module and his/her
//  company, the subsequent editors and their companies, and ISO/IEC have
//  no liability for use of this software module or modifications thereof
//  in an implementation.  Copyright is not released for non MPEG-4
//  conforming products.  IBM Corp. retains full right to use the code for
//  his/her own purpose, assign or donate the code to a third party and to
//  inhibit third parties from using the code for non MPEG-4 conforming
//  products.  This copyright notice must be included in all copies or
//  derivative works.
// 
//  AUTHORS:
// 
//  Andre Gueziec
// 
//  IBM T. J. Watson Research Center 
//  P.O.Box 704 
//  Yorktown Heights, New York 10598 
// 
//--------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include "MPG_Simplify.h"
extern FILE* logfp;

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpTopoSimplifyTriangleMesh                                           |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpTopoSimplifyTriangleMesh(int         nV,
                                 float*      v,
                                 int*        new_nV,
				 int         nT,
                                 int*        t,
                                 int*        new_nT,
				 float*      colors,
                                 float*      normals,
                                 float*      texCoords,
				 MPG_Binding colorBinding,
                                 MPG_Binding normalBinding,
                                 MPG_Binding texCoordBinding,
				 float       tolerance,
                                 int*        vertex_parents,
				 float*      error_volume,
                                 int         move_vertices,
                                 int         simplify_boundary)
{    
  int success = 0;

  int nE;

  EdgeS *e = NULL;

  Htable edge_table[1];

  /* create a simplification data structure where all the relevant arrays
     are copied */

  SimpData simpData[1];
    
  bzero(simpData, sizeof (SimpData));
  
  bzero(edge_table, sizeof (Htable));

  /* the first step is to build the table of edges, and to detect
     a few singular cases. The rest of the singular (non-manifold) cases 
     would be detected when simplifying and pivoting about vertices */

  if (_smpBuildEdgeTable(nV, nT, t, &nE, edge_table, &e))
 
    {
      simpData->nT       = nT; /* number of vertices, edges and triangles  */
      simpData->nE       = nE; /* before simplification */
      simpData->nV       = nV;
      
      simpData->move_vertices          = move_vertices;
      simpData->simplify_boundary      = simplify_boundary;
      simpData->tolerance              = tolerance;

      // simplification data that references existing arrays 

      simpData->edge_array      = e;
      simpData->e_table         = edge_table;
      simpData->vert            = (Vertex *) v;
      simpData->triangles       = t;
      simpData->vertex_father   = vertex_parents;
      simpData->err_volume      = error_volume;

      // properties 

      simpData->colors    = colors; 
      simpData->normals   = normals;
      simpData->texCoords = texCoords;

      simpData->colorBinding    = colorBinding;
      simpData->normalBinding   = normalBinding;
      simpData->texCoordBinding = texCoordBinding;
      
      if ((simpData->_smpCreateSimplificationDataStructure()) &&
	  (simpData->_smpSimplifyManifoldSurface()))
	{
	  /* free only the simplification data structures that are not
             exported by the driver routine,
	     to avoid running out of memory in _smpBuildSimplifiedSurface(): */

	  simpData->_smpPartialFreeSimplificationDataStructure();

	  /* build a simplified surface using the vertex_father and
             triangle_father arrays */

	  if (simpData->_smpBuildSimplifiedSurface(*new_nV, *new_nT))
            success = 1;

	  simpData->_smpFreeSimplificationDataStructure();
	}
        
      _smpFreeHashTable(edge_table);
      if (e) free((void*)e);
    }
  
  return success;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpBuildEdgeTable                                                     |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpBuildEdgeTable(int nV,   int nT,   int *t,
		       int *nE, Htable *e_table, EdgeS **edges)

{
  int success = 0;

  /* detect degenerate triangles */

  if (_smpNoDegenerateTriangles(nV, nT, t))
    
      /* detect vertices that are not referenced by any triangle */

    if (_smpNoStandaloneVertices(nV,nT, t))

      /* put the edges in a table, 
	 return an error if any edge is non-manifold */

      if (_smpIndexManifoldEdges(nE, edges, e_table, nT, t))
	
	  success = 1;
	

  return success;

}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpIndexManifoldEdges                                                 |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpIndexManifoldEdges(int *n_edges, EdgeS **new_edges, Htable *e_table, int nT, int *t)
{

  /* build the edges at most twice */

  /* 1) allocate an array of edges that is big enough,
     and build the edges 

     if error: only array is allocated
     if success: array and hash table are allocate
     */

  int 
    nE = nT * 3;

  *new_edges =  (EdgeS *) malloc(nE * sizeof (EdgeS));

  if (!(*new_edges)) 
    {
      fprintf(logfp,"\"_smpIndexManifoldEdges\": allocation of \"new_edges\" failed\n");
      goto error;
    }

  bzero((char *)e_table, sizeof(Htable)); /* reset the hash table */

  if (!_smpFillEdgeTable(&nE, *new_edges, e_table, nT, t))
    goto error;

  /* 2) verify that all the edges are manifold, otherwise print a message
   and return an error */

  {
    int i = 0;
    EdgeS *e = *new_edges;

    while (i < nE && (SurfEdgeLabel(e) <=2))
      {
	i++;
	e++;
      }
    if (i < nE)
      {
	/* non-manifold edges were detected */
	fprintf(logfp,"\"_smpIndexManifoldEdges\": non manifold edges were detected\n");
	goto error;
      }
  }

  /* 3) rebuild the edges with the proper number allocated */
  *new_edges =  (EdgeS *) realloc(*new_edges, nE * sizeof (EdgeS));

  if (!(*new_edges)) 
    {
      fprintf(logfp,"\"_smpIndexManifoldEdges\": re-allocation of \"new_edges\" failed\n");
      goto error;
    }

  if (!_smpFillEdgeTable(&nE, *new_edges, e_table, nT, t))
    goto error;

  *n_edges = nE;

  return 1;

error:
  
  if (e_table ) _smpFreeHashTable(e_table);
  bzero ((char *)e_table, sizeof (Htable));

  if (*new_edges) free((void*)(*new_edges)); *new_edges = NULL;
	
  return 0;

}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpFillEdgeTable                                                      |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

/* this routine fills a pre-allocated array of edges 
   and builds/ rebuilds a hash-table */

int _smpFillEdgeTable(int *n_edges, EdgeS *edges, Htable *e_table, int nT, int *t)
{

  /* this routine works for both a manifold and a non manifold surface
     but the number of edges must be known in advance */


  EdgeS 
    the_edge[1], 
    *edge_in_table;
	     
  int
    nE                   = *n_edges,
    *the_triangle        = t,
    *incident_tri        = SurfEdgeGetTriangles(the_edge),
    *endpoints           = SurfEdgeGetVertices (the_edge),
    second_tri,
    v0,                    /* surface edge end-points */
    v1,
    *triangles_in_table;

   
  register int i;

  register U_char k;

  /* frees the hash-table if we start with a pre-existing one */

  if (e_table) _smpFreeHashTable(e_table);

  bzero((char *)e_table, sizeof(Htable));


  /* initialize the number of faces per edge to 1 */
       
  SurfEdgeLabel(the_edge) = 1; 

  /* initialize the hash table with roughly as much entries as
     the number of edges */

   if (!_smpInitHashTable(e_table, 
			  _smpFlog2(nE), 
			  (int (*)(KEYTYPE))_smpEdgeSKey, 
			  (int (*)(RECTYPE, RECTYPE))_smpEdgeSFilter, 
			  NULL)) goto error;
  
   for (the_triangle = t, nE = i = 0; i < nT; i++, the_triangle += 3)
      {
	  
	for (k=0;k<3;k++)
		    
	  {
	    v0 = the_triangle[k];        /* surface edge end-points */
	    v1 = the_triangle[(k+1)%3];

   
	    if (v0>=v1) 
	      { 
		endpoints[0]     = v0;           endpoints[1] = v1; 
		incident_tri[0]  =  i;        incident_tri[1] = -1;
		second_tri       =  0;
	      }
	    else
	      { 
		endpoints[0]     = v1;           endpoints[1] = v0; 
		incident_tri[0]  = -1;        incident_tri[1] = i;
		second_tri       =  1;
	      }  

	    edge_in_table = (EdgeS *)_smpFindRecord(e_table,(KEYTYPE)the_edge);

	    if (edge_in_table)
	      {
		  
		triangles_in_table = SurfEdgeGetTriangles(edge_in_table);
			
		SurfEdgeLabel(edge_in_table) += 1;

		triangles_in_table[second_tri] = i;

	      }
	    else
	      {
		memcpy(edges+nE, the_edge, sizeof(EdgeS));

		/* add the edge to the hash table */

		if (!_smpAddRecord(e_table, (RECTYPE)(edges+nE), (KEYTYPE)(edges+nE)))
		  goto error;

		/* increment the number of edges */
		nE++;

	      }
	  }
      }

    /* now the final number of edges is known. 
       I return this number to the caller routine */

    *n_edges     = nE;

    return 1;

error:
  
    fprintf(logfp,"\"_smpIndexManifoldEdges\": not enough memory\n");

    if (e_table) _smpFreeHashTable(e_table);
  
    return 0;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpNoDegenerateTriangles                                              |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpNoDegenerateTriangles(int nV, int nT, int *t)
{
  int 
    *t1            = t;
  
  register int i;

  /* take each triangle in turn */

  for (i=0 ; i<nT; i++, t1+=3)

    {
      /* if all three triangle corners are different then continue
	 otherwise, exit and report an error */
	
      /* of course, all the indices should be  0 <=   < nV

	 (we never know what sort of strange input people could create) */

      if (t1[0] < 0 || t1[0] >= nV || t1[1] < 0 || t1[1] >= nV || t1[2] < 0 || t1[2] >= nV)
	goto error;
	
      if (t1[0] == t1[1] || t1[1] == t1[2] || t1[0] == t1[2])
	goto error;
    }
  
  return 1;

error:

  return 0;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpNoStandaloneVertices                                               |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpNoStandaloneVertices(int nV, int nT, int *t)
{
 
  register int i;

  /* allocate the new arrays using the old counts since we potentially
     eliminate standalone vertices, the arrays cannot grow larger
     during this procedure (the vertex array may grow larger later on */

  int *valence = (int *) _smpAllocateZero(nV * sizeof (int));

  if (!valence) 
    {
      fprintf(logfp,"\"_smpNoStandaloneVertices\": allocation of \"valence\" failed\n");
      goto error;
    }

  _smpComputeVertexValence(nT, t, valence);
  
  for (i=0; i<nV; i++)

    if (valence[i]==0)
   
      goto error;

  if (valence)                 free((void*) valence);
  
  return 1;
  
error:
  if (valence)                 free((void*) valence);
  
  return 0;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpInitHashTable                                                      |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpInitHashTable(Htable *table, int numbits, int (*fcode)(KEYTYPE), 
		      int (*filter)( RECTYPE, RECTYPE), KEYTYPE (*compute_key)(RECTYPE))
{
  register int i;

  /* relation between size and number of bits */
  table->size = 1 << numbits;

  /* the codes will be used as hash entries modulo table->mask */
  table->mask = table->size-1;

  /* alloc the table of entries */

  table->bucket = (Nodetype **) _smpAllocateZero (sizeof(Nodetype *) * table->size);

  if (!table->bucket) goto error;

  for(i=0;i<table->size;i++)
    table->bucket[i] = (Nodetype *) NULL;

  table->fcode  = fcode;  /* function that computes the code from the key */
  table->filter = filter;/* filtering function:
	returns 1 if the filter accepts the record
	returns 0 if the record is rejected, in which case the next
	      record must be found and processed */

  
  if (compute_key == NULL) /* function that computes the key from the record */
    table->compute_key = _smpIdentifyRecord2Key;
  else 
    table->compute_key = compute_key;

  /* initialize the first block of entries */

  table->nodes_per_block = 1 << 10 ;

  table->block    = (Block *) NULL;

  /* initialize the free node */

  table->freenode = (Nodetype *) NULL;

  return 1;

error:

  bzero(table, sizeof (Htable));
  return 0;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpNewBlock                                                           |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpNewBlock(Htable *table)
{
  Block    *newblock = NULL;

  newblock = (Block *) _smpAllocateZero (sizeof (Block));

  if (!newblock) goto error;

  newblock->next = table->block;
  newblock->rec  = NULL;

  table->block = newblock;

  return 1;

error:

  return 0;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   *_smpgetnode                                                           |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

Nodetype *_smpgetnode(Htable *table)
{
  Nodetype *retval;

  register int i;

 if (table->freenode)
   {
     retval          = table->freenode;
     table->freenode = table->freenode->next;
   }
  else
    {
      /* create a new block */

      /* this can potentially fail if there is not enough memory */

      if (!_smpNewBlock(table)) goto error;

      /* allocate a new array of nodes */

      table->block->rec= (RECTYPE) _smpAllocateZero (sizeof(Nodetype)
						   * table->nodes_per_block);

      if (!table->block->rec) goto error;

      /* link the nodes newly created  */

      for(retval= (Nodetype *) table->block->rec,
	  i = 0; i < table->nodes_per_block-1; i++, retval++)
	{
	  retval->next = retval+1 ;
	}

      /* the last node has to point to NULL */

      retval->next    = (Nodetype *) NULL;

      retval = (Nodetype *) table->block->rec ;
      table->freenode = retval+1;
    }

  return(retval);

error:

  return (NULL);
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   *_smpnewnode                                                           |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

Nodetype *_smpnewnode(Htable *table, RECTYPE rec, int code)
{
  
  Nodetype *retval = _smpgetnode(table);

  /* if _smpgetnode(table) returns a NULL pointer, we should not try to access it */

  if (retval)
    {
      retval->code     = code;
      retval->rec      = rec;
    }
  return(retval);
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpHashTableReset                                                     |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void _smpHashTableReset(Htable *table)
{
  /* resets the hash table by cleaning all buckets and
     deleting all records */

  Block 
    *block=table->block,
    *next;

  while(block)
    {
      next=block->next;
      if (block->rec) free((void*) block->rec);
      free((void*) block);
      block=next;
    }
  table->block = NULL;

  /* clear  buckets */

  if (table->size > 0 && table->bucket) 
    bzero((char *)table->bucket, table->size * sizeof(Nodetype *)) ;

  return;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpAddRecord                                                          |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpAddRecord(Htable *table, RECTYPE rec, KEYTYPE key)
{
  Nodetype  *node = NULL;

  register int i;

  int code= table->fcode(key);

  node=_smpnewnode(table,rec,code);

  if (!node) goto error;

  /* add the node containing the record to the appropriate bucket */

  i=node->code & table->mask;

  node->next = table->bucket[i];
  table->bucket[i]=node;

  return 1;

error:

  return 0;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   *_smpFindRecord                                                        |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

char *_smpFindRecord(Htable *table, KEYTYPE key)
{
  register int i;
  
  table->search_code = table->fcode (key);

  i = table->search_code & table->mask;

  table->search_key = key;

  table->search_node= table->bucket[i];

  return (_smpNextRecord(table));
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   *_smpNextRecord                                                        |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

char *_smpNextRecord(Htable *table)
{
  U_char found=0;

  Nodetype  *node=table->search_node;

  /* if a filter function was provided, the first node with the same code AND such
     that the filter function comparing the node's record and the search key "agrees"
     is selected */

  if (table->filter)
    {
      while (node && !found)
      {
	if (node->code == table->search_code &&
	    table->filter(node->rec,table->compute_key(table->search_key)))
	  found=1;
	else node = node->next;
      }
    }

  /* if no filtering function was provided, the next node with the same code is
     retrieved */


  else
    while (node && !found)
      {
	if (node->code == table->search_code)
	  found=1;
	else node = node->next;
      }

  if (found)
    {
      table->search_node = node->next;
      return ((char *)node->rec);
    }

  else return ((char *) NULL);

}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpIdentifyRecord2Key                                                 |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

KEYTYPE _smpIdentifyRecord2Key(RECTYPE rec)
{
  return (KEYTYPE) rec;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpFlog2                                                              |
  |                                                                          |
  +--------------------------------------------------------------------------+*/
/* 
   Log base 2 (integer form)
   
   The following relation holds exactly

   2 ^ Log2(N) <= N < 2 * 2 ^ Log2(N)

   if N==0 then the result is by definition -1, but the true result
   is of course -infty
*/

int _smpFlog2(int n)
{
   int d=0;
   if (n==0) return (-1);
   while(n>1)
        {d++;
         n/=2;
        }
   return(d);
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpEdgeSKey                                                           |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

/* compute the Hash-Table code from the key in the case of a surface edge */

int _smpEdgeSKey(EdgeS *e)
{
  return(e->v[1]^(e->v[0] << 8));
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpEdgeSFilter                                                        |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpEdgeSFilter(EdgeS *e1, EdgeS *e2)
{
  if ((e1->v[0] == e2->v[0]) &&
      (e1->v[1] == e2->v[1])) return(1);
  else
    return(0);
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpEdgeSFilter                                                        |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

EdgeS *_smpFindEdge(Htable *e, int vA, int vB)
{
  EdgeS edge;
  
  if (vA > vB)
    { 
      edge.v[0]=vA;edge.v[1]=vB;
    }
  else
    { 
      edge.v[0]=vB;edge.v[1]=vA;
    }

  return (EdgeS *) _smpFindRecord(e,(KEYTYPE) &edge);
 
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpInitArray                                                          |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void _smpInitArray(char *array, char *data, int n, int size)
{
  
  register int i;

  for(i=0;i<n;i++, array += size) 
    memcpy(array, data, size);

  return;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpInitFather                                                         |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void _smpInitFather(int n, int *father)
{
  register int i;

  for(i=0;i<n;i++) father[i] = i; 
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpFather                                                             |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpFather(int i, int *father)
{
  static int Ci, Fi;

  /*--- path traversal ---*/
  for(Ci=i;(Fi=father[Ci])!=Ci;Ci=Fi);

  /*--- path compression ---*/
  for(;(Fi=father[i])!=Ci;i=Fi) father[i] = Ci;

  return(Ci);
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpCreateSimplificationDataStructure                                  |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpCreateSimplificationDataStructure()
{
 
  edge2index               = NULL;
  index2edge               = NULL;
  edge_father              = NULL;
  triangle_father          = NULL;
  face_normals             = NULL;
  face_area                = NULL;
  compactness              = NULL;
  valence                  = NULL;
  boundary_vert            = NULL;
  tol_volume               = NULL;
  vx_color                 = NULL;
  vx_normal                = NULL;
  vx_texCoord              = NULL;
  vx_color_error           = NULL;
  vx_normal_error          = NULL;
  vx_texCoord_error        = NULL;
  edge_heap                = NULL;
 

  edge2index = (int *) _smpAllocateZero(nE * sizeof (int));

  if (!edge2index) goto error;

  index2edge = (int *) _smpAllocateZero(nE * sizeof (int));
  
  if (!index2edge) goto error;

  edge_father = (int *) _smpAllocateZero(nE * sizeof (int));
  
  if (!edge_father) goto error;

  triangle_father = (int *) _smpAllocateZero(nT * sizeof (int));
  
  if (!triangle_father) goto error;

  /* initialize the father arrays */

  _smpInitFather(nV, vertex_father);

  _smpInitFather(nE, edge_father);

  _smpInitFather(nT, triangle_father);

  removed_component_father = NULL;

  removed_component_father = (int *) _smpAllocateZero(nT * sizeof (int));
  
  if (!removed_component_father) goto error;

  _smpInitFather(nT, removed_component_father);


  face_normals = (Vertex *) _smpAllocateZero(nT * sizeof (Vertex));

  if (!face_normals) goto error;

  face_area   = (float  *) _smpAllocateZero(2*nT * sizeof (float));

  if (!face_area) goto error;

  compactness   = (float  *) _smpAllocateZero(nT * sizeof (float));

  if (!compactness) goto error;

  /* compute the triangle normals, areas and compactness */

  _smpTrianglesNormalAreaCompactness(nT, (Face *)triangles, vert,
				     face_normals, face_area, compactness);

  /* save the initial areas for later use */

  memcpy(face_area+nT, face_area, nT * sizeof (float));
  
 
  /* use default values for the maximum angle between face normals 
     and maximum ratio of compactness values */

  min_scalprod_nor  = (float) cos ((double)SIMPLIFY_MAX_ANGLE_NOR);

  compactness_ratio = SIMPLIFY_COMPACTNESS_RATIO;



  /* allocate and compute the vertex valence */

  valence = (U_short *) _smpAllocateZero(nV * sizeof (U_short));

  if (!valence) goto error;

  _smpComputeVertexValence(nT, (int *)triangles, valence);


  red_valence = (U_short *) _smpAllocateZero(nV * sizeof (U_short));

  if (!red_valence) goto error;

  memcpy(red_valence, valence,  nV * sizeof (U_short));


  ori_valence = (U_short *) _smpAllocateZero(nV * sizeof (U_short));

  if (!ori_valence) goto error;

  memcpy(ori_valence, valence,  nV * sizeof (U_short));


  /* allocate and compute the "boundary_vert" array that flags boundary vertices with a flag value of 1 */

  boundary_vert = (U_char *)  _smpAllocateZero(nV * sizeof (U_char));

  if (!boundary_vert) goto error;

  _smpFlagBoundaryVertices(nE, edge_array, boundary_vert);

  tol_volume = (float *) _smpAllocateZero(nV * sizeof (float));
  
  if (!tol_volume) goto error;

  /* otherwise initialize the tolerance array */

   _smpInitArray((char *)tol_volume,(char *) &tolerance, nV, sizeof (float));


  /* allocate memory for per-vertex properties */


   if (colorBinding == PB_PER_VERTEX && colors)
     {
       vx_color = (Vertex *) colors;

       vx_color_error = (float *) malloc (nV * sizeof (float));

       if (!vx_color_error) goto error;

       bzero ((void *) vx_color_error, nV * sizeof (float));

       vx_color_tol = (float)SIMPLIFY_VERTEX_COLOR_TOL;
     }
   
   
   if (normalBinding == PB_PER_VERTEX && normals)
     {
       vx_normal = (Vertex *) normals;

       vx_normal_error = (float *) malloc (nV * sizeof (float));

       if (!vx_normal_error) goto error;

       bzero ((void *) vx_normal_error, nV * sizeof (float));

       vx_normal_tol = SIMPLIFY_VERTEX_NORMAL_TOL;
     }
   
   
   if (texCoordBinding == PB_PER_VERTEX && texCoords)
     {
       vx_texCoord = (float *) texCoords;

       vx_texCoord_error = (float *) malloc (nV * sizeof (float));

       if (!vx_texCoord_error) goto error;

       bzero ((void *) vx_texCoord_error, nV * sizeof (float));

       vx_texCoord_tol = (float)SIMPLIFY_VERTEX_TEXCOORD_TOL;
     }
   
   

  /* allocate and initialize the edge heap */

   edge_heap = new MPG_Heap(MPG_Lex::LT,MPG_Lex::EQ,nE);

  if (!edge_heap) goto error;

  heap_outside_index     = nE+1;
      
  /* initialize the heap indices */
  
  bzero ((void *) edge2index, nE *sizeof (int));

  /* initialize various counters: */
  num_edg_collapsed           = 0;
  num_edg_weights             = 0;
  edg_rejected_4_topology     = 0;
  edg_rejected_4_geometry     = 0;
  edg_rejected_4_tolerance    = 0;

  /* initialize the number of edges left to be tested: */
  num_edg_remaining_4_testing = nE;

  valence_max = SIMPLIFY_VALENCE_MAX;

  /* initialize texture area (when texture coordinates per corner are provided) */

  if (texCoordBinding == PB_PER_CORNER)
    {
      register float *uv = NULL;
      register int i;

      _collapsed_texture_area = 0.;
      _texture_area           = 0.;
      
      for (uv = texCoords,  i = 0; i < nT; i++, uv +=6)
	{
	  register float area;
	  TEXMAP_AREA(uv, area);
	  _texture_area += area;
	}
    }
  
  
  return 1;

error:

  if (edge2index)               
    {
      free((void*) edge2index);
      edge2index = NULL;
    }
 
  if (index2edge)
    {
      free((void*) index2edge);
      index2edge = NULL;
    }

  if (edge_father) 
    {
      free((void*) edge_father);
      edge_father = NULL;
    }

  if (triangle_father) 
    {
      free((void*) triangle_father);
      triangle_father = NULL;
    }

  if (face_normals) 
    {
      free((void*) face_normals);
      face_normals = NULL;
    }

  if (face_area)
    {
      free((void*) face_area);
      face_area = NULL;
    }

  if (compactness) 
    {
      free((void*) compactness);
      compactness = NULL;
    }
 
  if (valence)   
    {
      free((void*) valence);
      valence = NULL;
    }

  if (red_valence)   
    {
      free((void*) red_valence);
      red_valence = NULL;
    }

  if (ori_valence)   
    {
      free((void*) ori_valence);
      ori_valence = NULL;
    }

  if (removed_component_father)   
    {
      free((void*) removed_component_father);
      removed_component_father = NULL;
    }

  if (boundary_vert)  
    {
      free((void*) boundary_vert);
      boundary_vert = NULL;
    }

  if (tol_volume) 
    {
      free((void*) tol_volume);
      tol_volume = NULL;
    }

  if (vx_color_error) 
    {
      free((void*) vx_color_error);
      vx_color_error = NULL;
    }
  
  if (vx_normal_error) 
    {
      free((void*) vx_normal_error);
      vx_normal_error = NULL;
    }
  
  if (vx_texCoord_error) 
    {
      free((void*) vx_texCoord_error);
      vx_texCoord_error = NULL;
    }
  
  if (edge_heap)   
    {
      delete (edge_heap);
      edge_heap = NULL;
    }

  return 0;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpFreeSimplificationDataStructure                                    |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpFreeSimplificationDataStructure()
{

  if (edge2index)               
    {
      free((void*) edge2index);
      edge2index = NULL;
    }
 
  if (index2edge)
    {
      free((void*) index2edge);
      index2edge = NULL;
    }

  if (edge_father) 
    {
      free((void*) edge_father);
      edge_father = NULL;
    }

  if (triangle_father) 
    {
      free((void*) triangle_father);
      triangle_father = NULL;
    }

  if (removed_component_father) 
    {
      free((void*) removed_component_father);
      removed_component_father = NULL;
    }

  if (face_normals) 
    {
      free((void*) face_normals);
      face_normals = NULL;
    }

  if (face_area)
    {
      free((void*) face_area);
      face_area = NULL;
    }

  if (compactness) 
    {
      free((void*) compactness);
      compactness = NULL;
    }
 
  if (valence)   
    {
      free((void*) valence);
      valence = NULL;
    }

  if (red_valence)   
    {
      free((void*) red_valence);
      red_valence = NULL;
    }

  if (ori_valence)   
    {
      free((void*) ori_valence);
      ori_valence = NULL;
    }

  if (boundary_vert)  
    {
      free((void*) boundary_vert);
      boundary_vert = NULL;
    }

  if (tol_volume) 
    {
      free((void*) tol_volume);
      tol_volume = NULL;
    }

  if (vx_color_error) 
    {
      free((void*) vx_color_error);
      vx_color_error = NULL;
    }
  
  if (vx_normal_error) 
    {
      free((void*) vx_normal_error);
      vx_normal_error = NULL;
    }
  
  if (vx_texCoord_error) 
    {
      free((void*) vx_texCoord_error);
      vx_texCoord_error = NULL;
    }
  
  if (edge_heap)   
    {
      delete (edge_heap);
      edge_heap = NULL;
    }

  return 1;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpPartialFreeSimplificationDataStructure                             |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpPartialFreeSimplificationDataStructure()
{

  if (edge2index)               
    {
      free((void*) edge2index);
      edge2index = NULL;
    }
 
  if (index2edge)
    {
      free((void*) index2edge);
      index2edge = NULL;
    }

  if (edge_father) 
    {
      free((void*) edge_father);
      edge_father = NULL;
    }

  if (compactness) 
    {
      free((void*) compactness);
      compactness = NULL;
    }
 
  if (valence)   
    {
      free((void*) valence);
      valence = NULL;
    }

  if (boundary_vert)  
    {
      free((void*) boundary_vert);
      boundary_vert = NULL;
    }

  if (tol_volume) 
    {
      free((void*) tol_volume);
      tol_volume = NULL;
    }

  if (edge_heap)   
    {
      delete (edge_heap);
      edge_heap = NULL;
    }

  return 1;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpTrianglesNormalAreaCompactness                                     |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpTrianglesNormalAreaCompactness( int nT, Face *t, Vertex *v, Vertex *t_normal,
					float *t_area, float *t_compactness)
{
  int success = 1;

  static Vertex triVert[3];
  
  register int i;
  register U_char j,k;

  /* compute the face normals */
   
  for(i=0;i<nT;i++,t++)
    {
      float
	sum_lengths_sq = 0.;

      for (j=0;j<3;j++)
	  for (k=0;k<3;k++)
	    triVert[j][k] = v[t[0][j]][k];
      
       t_area[i] =_smpTriangleNormal(triVert, t_normal[i]) / 2.;

       /* compute the squared lenghts of the three edges: */
  
       for(j=0;j<3;j++)
	 {
	   float len_edg_sq = 0, tmp;

	   for(k=0;k<3;k++)
	     {
	       tmp         = triVert[(j+1)%3][k] - triVert[j][k];
	       len_edg_sq += tmp * tmp;
	     }

	   sum_lengths_sq += len_edg_sq;

	 }

       t_compactness[i] = FOUR_SQRT_3 * t_area[i] / sum_lengths_sq;

    }

  return success;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpTriangleNormal                                                     |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

/* computation of the triangle normal in single precision
   return twice the area */

float _smpTriangleNormal(Vertex *tri, Vertex normal)

{
  Vertex x1, x2, x3;
  float norm;
  int origin;
  U_char k;
	
  _smpTriangleBasisVectors(tri, x1, x2, origin);

  VECPROD3(x1, x2, x3);

  norm = NORM3(x3);
	
  for (k=0;k<3;k++) normal   [k] = x3[k] / norm;

  return norm;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpNormalize3                                                         |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

float _smpNormalize3(float *vec)
{
  register U_char k;
  float norm =  NORM3(vec);
  for (k=0;k<3;k++) vec[k] /= norm;
  return norm;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpFlagBoundaryVertices                                               |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

/* as the name indicates, mark the boundary vertices with a 
value of 1 in the array "boundary_vert"
the interior vertices are marked with 0 */

int _smpFlagBoundaryVertices(int nE, EdgeS *edges, U_char *boundary_vert)
{
  /* loop through edges. 

     if an edge is a boundary edge, then mark both vertices as boundary vertices */

  register int i;

  for (i=0; i<nE; i++, edges++)
    
    
    if (SurfEdgeLabel(edges) == 1)
      {
	int *endpoints = SurfEdgeGetVertices(edges);

	boundary_vert[endpoints[0]] = boundary_vert[endpoints[1]] = 1;

      }
    
  return 1;
}


/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpAddEdge2Heap                                                       |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

/* add a surface edge to the heap and record the index associated
   with the edge */

int SimpData::_smpAddEdge2Heap(int the_edge)
{
  int 
    added      = 0;

  /* verify that the Heap is not already full and that the edge does
     not have an index that forbids adding to the Heap */

  if (edge2index[the_edge] >=0)
    {
      /* compute the representatives of the edge vertices */
      int 
	*v = SurfEdgeGetVertices(edge_array+the_edge),
	v0 = FATHER(v[0], vertex_father),
	v1 = FATHER(v[1], vertex_father);

      /* 1) to compute the key to the heap, we choose the edge length to start with */
      float 
	weight = DIST3(vert[v0], vert[v1]);

      /* 1.1) we add the sum of the errors at the two vertices */

      weight += err_volume[v0] + err_volume[v1];

      /* 2) add the edge to the heap */

      union { float f; void* p; } w;
      w.f = weight;
      index2edge[ edge2index[the_edge]= edge_heap->add(w.p)] = the_edge;
      
      added = 1;
	
    }

  return added;

}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpRemoveLowestWeightEdgeFromHeap                                     |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpRemoveLowestWeightEdgeFromHeap()
{
  
  int 
    idx  = edge_heap->delMin();
	
  if (idx != -1)
  
    return index2edge[idx];

  else
    return -1;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpBuildSimplifiedSurface                                             |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpBuildSimplifiedSurface(int& new_nV, int& new_nT)

/*...........................................................................*/
/*
   
    Use the information contained in the vertex and triangle fathers
    to construct a simplified surface:
    
    only the vertices and triangles that are their own "father" stay in 
    the simplified surface. the vertex_father and triangle_father arrays
    are used for interpolating properties.
    

*/
/*...........................................................................*/

{
 
  register int i,j;
  register U_char k;

  vector <int> new_v_number((size_t)nV,0);
  vector <int> new_t_number((size_t)nT,0);
  
  /* I) build the new vertices and copy the vertex errors and data */

  new_nV = 0;

  for (i=0; i< nV; i++) if (i == FATHER(i,vertex_father)) 
    
    {
      new_v_number[i] = new_nV;
      new_nV += 1;
    }

  

  /* copy the vertices and other vertex attributes and update the vertex_father array */

  
  for(j=i=0;i<nV;i++) if (i == vertex_father[i]) 
    {
      if (j<i)
	{
	  memcpy(vert[j], vert[i], sizeof(Vertex));

	  err_volume[j] = err_volume[i];
 
	  // treat the PER_VERTEX properties right here
      
	  if (vx_color_error)
	    memcpy(colors+3*j, colors+3*i, sizeof(Vertex));
	  if (vx_normal_error)
	    memcpy(normals+3*j, normals+3*i, sizeof(Vertex));
	  if (vx_texCoord_error)
	    memcpy(texCoords+2*j, texCoords+2*i, 2 * sizeof (float));

	}

     
      j++;
    }

  // shuffle the vertex fathers 

  for(i=0;i<nV;i++) vertex_father[i] = new_v_number[vertex_father[i]];
 
  /* II) build the new triangles  */



  new_nT = 0;

  for (i=0; i< nT; i++) if (i == FATHER(i,triangle_father)) 
    
    {
      new_t_number[i] = new_nT;
      new_nT += 1;
    }


  vector <int> new_t((size_t)(3*new_nT),0); 

  for(j=i=0;i<nT;i++) if (i == triangle_father[i]) 
    {
      // if (j < i) bug here! 
      for (k=0;k<3;k++) new_t[j*3+k] = vertex_father[ triangles[i * 3 + k]];
      j++;
    }
    
  // treat the per-corner properties
  // for texture coordinates, just discard the corner properties of the removed faces 
  
  if (texCoordBinding == PB_PER_CORNER)
    {
      for(j=i=0;i<nT;i++) if (i == triangle_father[i]) 
	{
	  for (k=0;k<3;k++) texCoords[j * 3 + k] = texCoords[i * 3 + k];
	  j++;
	}
    }
        
  /* re-shuffle the triangle_father array */
  
  for(i=0;i<nT;i++) triangle_father[i] = new_t_number[triangle_father[i]];
 
  // for colors and normals per corner, redistribute corner properties

  if (colorBinding == PB_PER_CORNER)
    _smpRedistributePerCornerProperty(3, colors, triangles, nT, new_t.begin(), new_nT,
				      vertex_father, triangle_father, face_area+nT);

  
  if (normalBinding == PB_PER_CORNER)
    {
      _smpRedistributePerCornerProperty(3, normals, triangles, nT, new_t.begin(), new_nT, 
					vertex_father, triangle_father, face_area+nT);

      // then normalize the resulting normals
      
      for (i=0;i<3*nT;i++) _smpNormalize3(normals+3*i);
    }


  memcpy(triangles, new_t.begin(), new_nT * 3 * sizeof (int));


  // redistribute the per-face attributes
  
  if (colorBinding == PB_PER_FACE) 
    _smpRedistributePerFaceProperty(3, colors, nT, new_nT, triangle_father, face_area+nT);

  if (normalBinding == PB_PER_FACE) 
    {
      _smpRedistributePerFaceProperty(3, normals, nT, new_nT, triangle_father, face_area+nT);
      
      // then normalize the resulting normals
      
      for (i=0;i<nT;i++) _smpNormalize3(normals+3*i);
	
    }

  return 1;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpMarkEdgesAdjacent2Boundary                                         |
  |                                                                          |
  +--------------------------------------------------------------------------+*/


/*
  I am  using two different flags for edges adjacent to
  the surface boundary. 

  an edge incident to two boundary vertices gets a flag of -2

  an edge incident to a single boundary vertex gets a flag  of -1.

 */

int SimpData::_smpMarkEdgesAdjacent2Boundary()
{

  int 
    e  = 0,
    edges_adjacent_2_boundary;

  U_char 
    boundary1,
    boundary2;

  EdgeS *current_edge    = edge_array;

  /* for each edge of the surface, determine whether it is a boundary
     edge of the first or second type. If yes, label it accordingly
     using the edge2index array */

  for (edges_adjacent_2_boundary = nE,
       e=0; e<nE; e++, current_edge++)
    {
      
      /* if the edge label is 1, then it is a boundary edge of the second kind */

      if (SurfEdgeLabel(current_edge) == 1) edge2index[e] = -2;

      else
	{
	  int 
	    *v = SurfEdgeGetVertices(current_edge);

	  boundary1        = boundary_vert[v[0]];
	  boundary2        = boundary_vert[v[1]];

	  if (boundary1 || boundary2) edge2index[e] = -1;
	  
	  else  edges_adjacent_2_boundary--;
	}
      
    }


  return edges_adjacent_2_boundary;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpBuildEdgeHeap                                                      |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpBuildEdgeHeap()
{
  
  int 
    i = 0,
    n_edges_added = 0;

  while (i < nE)
    { 
      n_edges_added += _smpAddEdge2Heap(i);

      i++;

    } /* end while */

  return (n_edges_added);
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpRotateParentTriangle                                               |
  |                                                                          |
  +--------------------------------------------------------------------------+*/


/* performs a circular permutation (rotation) of the 
   vertex indices of a triangle such that the parent of the 
   first index is the same as "parent_vertex" */

int _smpRotateParentTriangle(int  parent_vertex, int *vertex_fathers, 
			     int *tri_v, int *tri_v_parents, int *tri_v_rotated)
{

  /* the triangle must be a parent triangle itself (otherwise
	 all three triangle vertices are not necessarily different */

  tri_v_parents[0] = FATHER(tri_v[0], vertex_fathers);
  tri_v_parents[1] = FATHER(tri_v[1], vertex_fathers);
  tri_v_parents[2] = FATHER(tri_v[2], vertex_fathers);

  if (tri_v_parents[0] == parent_vertex)
    {
      tri_v_rotated[0] = tri_v[0];
      tri_v_rotated[1] = tri_v[1];
      tri_v_rotated[2] = tri_v[2];
    }
  else if (tri_v_parents[1] == parent_vertex)
    {
      /* rotate the vertex parents */
     
	  tri_v_rotated[0] = tri_v_parents[0]; /* use tri_v_rotated[0] as a temporary buffer */
      
      tri_v_parents[0] = tri_v_parents[1];
      tri_v_parents[1] = tri_v_parents[2];
      tri_v_parents[2] = tri_v_rotated[0];  

      /* rotate the vertices */
      tri_v_rotated[0] = tri_v[1];
      tri_v_rotated[1] = tri_v[2];
      tri_v_rotated[2] = tri_v[0];
    }
  else 
    {
      /* rotate the vertex parents */
      tri_v_rotated[0] = tri_v_parents[0]; /* use tri_v_rotated[0] as a temporary buffer */

      tri_v_parents[0] = tri_v_parents[2];
      tri_v_parents[2] = tri_v_parents[1];  
      tri_v_parents[1] = tri_v_rotated[0];
	      
      /* rotate the vertices */
      tri_v_rotated[0] = tri_v[2];
      tri_v_rotated[1] = tri_v[0];
      tri_v_rotated[2] = tri_v[1];
    }

  return 1;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpRotateParentEdge                                                   |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpRotateParentEdge(int parent_vertex, int *vertex_fathers,
			 int *edge_v, int *edge_t, int *edge_v_parents, 
			 int *edge_v_rotated, int *edge_t_rotated)
{
  

  /* rotate a parent edge with respect to a parent vertex, so that
     the child of the parent vertex will be first (index [0])
     among the two edge vertices 
     ( currently, by default the first vertex is the vertex with the
       highest number )
  */

  edge_v_parents[0] = FATHER(edge_v[0], vertex_fathers);
  edge_v_parents[1] = FATHER(edge_v[1], vertex_fathers);

  /* if the vertices are in the right order, just copy the vertices */

  if (edge_v_parents[0] == parent_vertex)
    {
      edge_v_rotated[0] = edge_v[0];
      edge_v_rotated[1] = edge_v[1];
      edge_t_rotated[0] = edge_t[0];
      edge_t_rotated[1] = edge_t[1];
    }
  else
    {
      edge_v_rotated[0] = edge_v_parents[0];  /* use edge_v_rotated[0] as a temporary buffer */

      edge_v_parents[0] = edge_v_parents[1];
      edge_v_parents[1] = edge_v_rotated[0];
      
      edge_v_rotated[0] = edge_v[1];
      edge_v_rotated[1] = edge_v[0];
      edge_t_rotated[0] = edge_t[1];
      edge_t_rotated[1] = edge_t[0];

    }

  return 1;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpDirectedParentVertexStar                                           |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

/* 
   compute a vertex star and store the triangles in either
   CLOCKWISE or COUNTER-CLOCKWISE direction
 */

int SimpData::_smpDirectedParentVertexStar(int parent_vertex, int first_triangle,
					   int valence, int *star, int direction)
{   


  /* labeling system for directed triangle stars 

	          v*1\      /v*0  
	           e*1\  *1/e*0   
	       *val0-1 \  / *0    
	     ___________\/___________ we omit on purpose the first vertex, because in the case
	     v*val0-1   v0            of a boundary edge, we know it already


	     */

  int 
    *vstar = star   + valence,
    *estar = vstar  + valence,
    /* rotated triangle vertices: */
    tri_v_rotated[3],  tri_v_parents[3], 
    /* rotated edge vertices and adjacent triangles: */
    edge_v_rotated[2], edge_v_parents[2], edge_t_rotated[2],
    star_triangle = first_triangle,
    i;

     
  for (i = 0; i < valence; i++)
    {
      star[i] = star_triangle;

      _smpRotateParentTriangle(parent_vertex, vertex_father,
			       triangles + 3 * star_triangle,
			       tri_v_parents, tri_v_rotated);

      /* find previous or next edges and get their parents */

      estar[i] = _smpNextParentEdge(tri_v_rotated, direction);

      _smpRotateParentEdge(parent_vertex, vertex_father,
			   (int *) SurfEdgeGetVertices (edge_array + estar[i]),
			   (int *) SurfEdgeGetTriangles(edge_array + estar[i]),
			   edge_v_parents, edge_v_rotated, edge_t_rotated);

      vstar[i] = edge_v_parents[1];
      /* this vertex number should be the same as tri_v_parents[direction] */

      if (i < valence-1) 
	/* otherwise, the next triangle would be -1 and finding its father 
	   would produce a segmentation fault or another bad thing */
	star_triangle = _smpNextParentTriangle(edge_t_rotated, direction);
    } 


  return 1;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpManifoldLinkTest                                                   |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

/*
   detect duplicate vertices in the link of an edge.
   In case duplicate vertices are detected, collapsing an edge
   would generate a singular vertex.
  */

int _smpManifoldLinkTest(int *link0, int *link1, int val0, int val1)
{
  /* among the vertices of the link of v0, link0 and of the link of v1, link1,
     is there any duplicate vertex ? */

  int 
    link_is_manifold = 1,
    val              = val0 + val1;

  /* since 
	val0 + val1 <= SIMPLIFY_VALENCE_MAX4,
	I can use a static array here. */

  static int
    link[SIMPLIFY_VALENCE_MAX4];

  {
    register int i;
      
    /* gather vertices of link0 and link1 in a single array, sort
       the array, and determine whether there are any duplicate
       vertices */

    memcpy(link,        link0, val0 * sizeof (int));
    memcpy(link + val0, link1, val1 * sizeof (int));

    qsort(link, val, sizeof(int), (int (*)(const void*, const void*)) _smpCmpIntSmallFirst);

    i = 0;

    while ((i < val -1 ) && (link_is_manifold))
      {
	if (link[i+1] == link[i]) /* if there are duplicate vertices,
				     they must be contiguous in the
				     sorted array */
				       
	  link_is_manifold = 0;
	i++;
      }
     
  }

  return link_is_manifold; 

}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpCmpIntSmallFirst                                                   |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

/*
    compare integers. Using this function in qsort will
	result in listing the smaller integers first
  */

int _smpCmpIntSmallFirst(char *s1, char *s2)
{
  int 
    *f1 = (int *)s1,
    *f2 = (int *)s2;

 if (*f2 > *f1)
    return(-1);

  else if (*f2 < *f1)
    return(1);

  else
    return(0);
  
}

#undef  COMPOSEVEC3
#define COMPOSEVEC3(u,v,lambda,w) {(w)[0] = (u)[0] + (lambda) * (v)[0]; (w)[1] = (u)[1] + (lambda) * (v)[1]; \
                                   (w)[2] = (u)[2] + (lambda) * (v)[2];}


/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpSimplifiedVertexLocation                                           |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

/* 
   determine the location of a potential simplified vertex
 */

int SimpData::_smpSimplifiedVertexLocation(int v0, int v1, 
					   U_short   val0, U_short   val1, 
					   U_short   star_val,
					   Vertex   *star_vert, Face     *star_face,
					   Plane    *star_plane, float    *star_areas,
					   float    *work, int method)
{

  if (method == 0)
    {
      /* simplified location is v0 */
      memcpy(simplified_vertex,     vert[v0], sizeof (Vertex));
      /* alternate location is v1 */
      memcpy(simplified_vertex_alt, vert[v1], sizeof (Vertex));
    }

  else if (method == 1)
    {
      /* take the vertex that minimizes the sum of distances to the
	 vertices of the edge star plus the planes defined by the
	 triangles of the edge star and that preserves the volume of
	 the edge star */

      float  
	star_volume,   /* volume of the edge star (x 6) */
	sign;
	
      int    
	star_val_1      = star_val + 1,
	size_planes     = (star_val_1) * sizeof (Plane);

      float  *star_plane_sav = (float *) (star_plane + star_val_1);

      static Vertex 
	v, z     = {0.,0.,1.};

      star_volume = (float) _smpStarPlaneEquationsAndVolume(star_vert, star_face, 
							    star_plane, star_areas, 
							    star_val);

      /* compute the equation of the plane p such that the star of any
	 vertex on that plane would have the same volume as star_volume */

      _smpPlaneForIdenticalVolume(val0, val1, star_plane[star_val], star_vert, 
				  star_face,  star_volume);

      /* now change again the coordinate system such that either
	 (0,0,1) or (0,0,-1) corresponds to the normal of P, 

	 change the equations of the planes. THE VERTICES ARE LEFT UNCHANGED */

      /* compute the householder vector */

      sign = (star_plane[star_val][2] > 0.) ? 1. : -1.;

      COMPOSEVEC3(star_plane[star_val], z, sign, v);

      /* Apply the Householder transform to the planes
	 while saving their previous positions */
      
      memcpy(star_plane_sav, star_plane, size_planes);
		     
      _smpHouseholderPreMultiplication((float *)star_plane, 4, v, 3, star_val_1, work);
	
      /* weight the distances to planes supporting the triangles by triangle area*/
     
      _smpPositionVertexLSConstraintsWArea(star_plane, star_areas, star_val, simplified_vertex);
      
      /* change back the position of the simplified vertex to
	 the original coordinate system */

      _smpHouseholderPreMultiplication(simplified_vertex, 3, v, 3, 1, work);

      /* idem with the planes */
     
      memcpy(star_plane, star_plane_sav, size_planes);
		
    } 

  else if (method == 2)
    {
      /* take the mid-point of the edge endpoints */

      register U_char j;

      for (j=0;j<3;j++)

	simplified_vertex[j] = (vert[v0][j]  + vert[v1][j] ) /2.;
    }
  
  return 1;
}


#undef COMPOSEVEC3
/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpPositionVertexLSConstraintsWArea                                   |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int  _smpPositionVertexLSConstraintsWArea(Plane *star_plane, float *star_areas, U_short star_val, 
					  Vertex simplified_vertex)
{

  /* same as before, but the area of the triangles is used to weight the least squares */

  int retval = 1;

  static double 
    A1[SIMPLIFY_VALENCE_MAX4],
    A2[SIMPLIFY_VALENCE_MAX4],
    B [SIMPLIFY_VALENCE_MAX4];

  register int i;
  
  /*1) each of the planes provides a LS equation, plus the minimum
    distance to the origin which provides 2 equations */

  double 
    
    /* height of the plane (+/-) z + delta = 0 */

    delta = 
    star_plane[star_val][3] *
    star_plane[star_val][2]; 

  /* 2) write the quadratic constraints associated with the planes */

  /* 2.1 
     multiply the coefficients of the planar equations with the weights:
     areas of triangles */
  
  for (i = 0; i < star_val; i++)
    {
      A1[i] = star_plane[i][0] * star_areas[i];
      A2[i] = star_plane[i][1] * star_areas[i];
      B [i] = (delta * star_plane[i][2] - star_plane[i][3]) * star_areas[i] ;
    }

  {
    /* write the normal equations */
    static double a,b,c,d,e,x,y;

    for (a=b=c=d=e=0.,i=0;i<star_val ;i++)
      {
	a += A1[i] * A1[i];
	b += A1[i] * A2[i];
	c += A2[i] * A2[i];
	d += A1[i] * B [i];
	e += A2[i] * B [i];
      }

    /* solve the normal equations */

    _smpSolveSymmetric2x2Eqn(a,b,c,d,e, &x, &y);
     
    simplified_vertex[0] = (float)  x;
    
    simplified_vertex[1] = (float)  y;
    
    simplified_vertex[2] = (float) -delta;
    
  }

  return retval;
}



/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpSolveQuadraticEquation                                             |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpSolveQuadraticEquation(double A, double B, double C, 
			       double *sol1, double *sol2, double eps)
{
  /*
  
    returns the number of real solutions of a second degree equation as well as the solutions

   everything below eps is supposed to be 0
   
   */

  int num_sol;


  double 
    A_abs = (A > 0.0) ? A : -A,
    B_abs = (B > 0.0) ? B : -B,
    C_abs = (C > 0.0) ? C : -C;

  A *= 2.;

  if ( A_abs < eps * B_abs) /* compare with eps * abs(B)  because I compute -B/A */
    {
      /* B x + C = 0 */
      
      if ( B_abs < eps * C_abs)
	num_sol = 0;

      else 
	{
	  num_sol = 1;
	  *sol1   = - C / B;
	}
    }

  else
    {
      double delta  =  B * B - 2. * A * C;

      if ( delta < - eps )
	num_sol = 0;

     else
       {
	 /* the discriminant is zero or positive */
	
	 if (delta < eps * B_abs)
	   {
	     num_sol = 1;
	      
	     *sol1 = -B / A;
	   }
	 else
	   { 
	     delta = sqrt (delta);

	     num_sol = 2;

	      *sol1 = (-B - delta ) /A;

	      *sol2 = (-B + delta ) /A;
	    }
       }
    }

  return num_sol;

}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpSolveSymmetric2x2Eqn                                               |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpSolveSymmetric2x2Eqn(double a,  double b, double d, double e, double f,
			     double *x, double *y)
{
  int retval = 0;

  static double eps = 1e-13;
  
   if (b < eps && b > -eps)
    {
      *x = (a < eps && a > -eps) ? 0. : e / a;
      *y = (d < eps && d > -eps) ? 0. : f / d;

      retval = 2;
    }
  else
    {
      double 
	theta = (d - a) / 2. / b,
	
	t     = (theta > 0.) ? 
	1. / (theta + sqrt (theta * theta + 1.)) : 
	-1./ (sqrt (theta * theta + 1.) - theta),

	c     = 1. / (sqrt (1. + t*t)),
        s     = t * c,

	tb   =     t * b,
	A    = a - tb,
	C    = d + tb,
	Rx   = (A < eps && A > -eps) ? 0. : (c * e - s * f) / A,
	Ry   = (C < eps && C > -eps) ? 0. : (s * e + c * f) / C;
      
      *x =  c * Rx + s * Ry ; /* apply the inverse rotation */
      *y = -s * Rx + c * Ry ; /*          "    "            */
	
      retval = 1;
    }

  return retval;
}


/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpHouseholderPreMultiplication                                       |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpHouseholderPreMultiplication(
			     float *A,  /* input matrix, given column
				           by column */
			     int mA,    /* leading dimension of matrix a */
			     float *v,  /* householder vector    */
			     int m,     /* dimension of v        */
			     int n,     /* second dimension of A */
			     float *w   /* workspace, dimension n */
				 ) 
{
  int retval = 0;
  
  register int i,j  ;
  
  float *w1 = w;

  /* 1) compute beta = -2 vT v: */

  float beta = 0; 

  for (i=0;i<m;i++) beta += v[i]*v[i];

  beta = -2./beta;

  /* 2) compute w = beta AT v */
  
  if (w1 == NULL)
    {
      w1 = (float *) malloc (n * sizeof (float));
    }
  
  if (w1)
    { 
      float *A1 = A;

      for (i=0;i< n; i++, A1 += mA) 
      
	{
	  for (w1[i]=0., j=0; j<m;j++)
	
	    w1[i] += A1[j]*v[j];

	  w1[i] *= beta;
	}

      /* 3) compute A = A + w vT */

      A1 = A;

      for (i=0;i< n; i++, A1 += mA) 
	
	for (j=0; j<m;j++)
	  
	  A1[j] += w1[i] * v[j];
	
      retval = 1;

      if (w != w1)
	free((char *) w1);
    
    }
  else fprintf(logfp,
		 " _smpHouseholderPreMultiplication: allocation of w1 failed.\n");

  return retval;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpPlaneForIdenticalVolume                                            |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void     _smpPlaneForIdenticalVolume(U_short val0, U_short val1, 
				     Plane p, Vertex *star_vert, 
				     Face *star_face, float star_volume)
{
  static Vertex  n;
      
  U_short i, val02 = val0 - 2, val013 = val0 + val1 - 3;
  
  bzero((char *)p, sizeof(Vertex));

  for(i=1; i<=val02; i++)
    {
      U_short v1  = star_face[i][1],
              v2  = star_face[i][2];
      VECPROD3(star_vert[v1], star_vert[v2], n);
      ADDVEC3(p, n);
    }

  for(i=val0; i<=val013; i++)
    {
      U_short v1  = star_face[i][1],
              v2  = star_face[i][2];
      VECPROD3(star_vert[v1], star_vert[v2], n);
      ADDVEC3(p, n);
    }
 
  p[3] = -star_volume;

  // NormalizePlaneEquation(p):

  float norm = _smpNormalize3((float *) p);

  if (norm > 0.) p[3] /= norm;

  return;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpStarPlaneEquationsAndVolume                                        |
  |                                                                          |
  +--------------------------------------------------------------------------+*/


double _smpStarPlaneEquationsAndVolume(Vertex *star_vert,  Face *star_face, 
				       Plane  *star_plane, float *star_areas, 
				       int   star_val)
{
  double star_volume_x_6 = 0.;

  register U_short j;

  for(j=0;j<star_val;j++)
    {
      /* determine the equation of the plane when the normal is known.

	 (n,v)  + d  = 0  is the equation, n known vector, d unknown value.
	 (n,v1) + d1 = 0;
	 (n,v2) + d2 = 0; Three data points to determine the optimum d.
	 (n,v3) + d3 = 0;
	 d           = -[ (n,v1) + (n,v2) + (n,v3) ]/3.;
	 */

      register U_char k;

      star_plane[j][3] = 0.;

      for(k=0;k<3;k++)
	
	star_plane[j][3] -= SCALPROD3(star_plane[j], star_vert[star_face[j][k]]);
	
      star_plane[j][3] /= 3.;

      
      /* for volume computation, retrieve the face area from the global array
	 the volume is equal to one third of the triangle area times the
	 height of the origin with respect to the triangle */

      star_volume_x_6  -= star_areas[j] * star_plane[j][3] / 3.;
    }

  return (star_volume_x_6 * 6.);
}


/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpClosestPointOnEdge                                                 |
  |                                                                          |
  +--------------------------------------------------------------------------+*/



#undef  EDGE_BARYCENTRIC_COORDINATES
#define EDGE_BARYCENTRIC_COORDINATES(u,v,a,b) { a = SCALPROD3(u,v); b = SCALPROD3(u,u); }

float _smpClosestPointOnEdge(Vertex X, Vertex XP, Vertex A, Vertex B, 
			     float *t)/* barycentric coordinates: t, 1-t */
{
  float d;
  Vertex BA,BX;
  float BA_BX,BA_2;

  MAKEVEC3(B,A,BA);
  MAKEVEC3(B,X,BX);
  EDGE_BARYCENTRIC_COORDINATES(BA,BX,BA_BX,BA_2);

  if (BA_BX <= 0.)
    {
      *t = 0.;
      d = NORM3(BX);
      memcpy((void *)XP,(void *)B, sizeof (Vertex));
    }
  else if (BA_2 <= BA_BX)
    {
      *t = 1.;
      d = DIST3(A,X);
      memcpy((void *)XP,(void *)A, sizeof (Vertex));
    }
  else
    {
      register U_char j;
      float t0, t1;

      *t = BA_BX/BA_2;

      t0 = *t;
      t1 = 1. -t0;
      
      for (j=0;j<3;j++)
	XP[j] = t0 * A[j] + t1 * B[j];
     
      d = DIST3(X,XP);

    }


  return(d);
}
#undef  EDGE_BARYCENTRIC_COORDINATES

#define EPS_VERTEX_OUTSIDE_TRIANGLE -1e5 

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpClosestPointOnTriangle                                             |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpClosestPointOnTriangle(Vertex *tri, Vertex w, Vertex wp, Vertex bary, float *dist)
{
  int inside = _smpTriangle3DBarycentricCoordinates(tri, w, wp, bary, dist);
	
  if (!inside)
    {
      static float eps= EPS_VERTEX_OUTSIDE_TRIANGLE;
  
      /* either 2 barycentric coordinates are negative */

      if (bary[0] < eps && bary[1] < eps)
	{
	  bary[0] = bary[1] = 0.;
	  bary[2] = 1.;
	  memcpy(wp, tri[2], sizeof(Vertex));
	  *dist = DIST3(w,wp);
	}
	
      else if (bary[0] < eps && bary[2] < eps)
	{
	  bary[0] = bary[2] = 0.;
	  bary[1] = 1.;
	  memcpy(wp, tri[1], sizeof(Vertex));
	  *dist = DIST3(w,wp);
	}

      else if (bary[1] < eps && bary[2] < eps)
	{
	  bary[1] = bary[2] = 0.;
	  bary[0] = 1.;
	  memcpy(wp, tri[0], sizeof(Vertex));
	  *dist = DIST3(w,wp);
	}

      /* or 1 barycentric coordinate is negative */

      else if (bary[0] < eps)
	{
	  *dist =  _smpClosestPointOnEdge(w, wp, tri[1], tri[2], &bary[1]);
	  bary[2] = 1. - bary[1];
	  bary[0] = 0.;
	} 

      else if (bary[1] < eps)
	{
	  *dist =  _smpClosestPointOnEdge(w, wp, tri[0], tri[2], &bary[0]);
	  bary[2] = 1. - bary[0];
	  bary[1] = 0.;
	}
      
      else if (bary[2] < eps)
	{
	  *dist =  _smpClosestPointOnEdge(w, wp, tri[0], tri[1], &bary[0]);
	  bary[1] = 1. - bary[0];
	  bary[2] = 0.;
	}
    }

  return (inside);
}


/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpTriangle3DBarycentricCoordinates                                   |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpTriangle3DBarycentricCoordinates(Vertex *tri, Vertex w, Vertex wp, Vertex bary, float *residual)

/* compute the barycentric coordinates of a vector with respect to a point
   the projection wp and the residual are only valid if the function
   returns 1 (vertex projects inside)
*/
{
  int inside;

  static Vertex x1, x2;

  double a = 0., b = 0., c = 0., d = 0., e = 0., x3, x, y;

  register U_char j;

  int origin;

  /*
     x1 is the vector representing the shortest edge of the triangle,
     which is best numerically

   Min 
            
   ||  |  x1[0]  x2[0] |               | w[0] - tri[origin][0] |  ||^2
   ||  |  x1[1]  x2[1] |  *  | b1 | -  | w[1] - tri[origin][1] |  ||  
   ||  |  x1[2]  x2[2] |     | b2 |    | w[2] - tri[origin][2] |  ||  


   using the normal equations 
   */

  _smpTriangleBasisVectors(tri, x1, x2, origin);

  for (j=0; j<3;j++)
    {
      x3 = w[j] - tri[origin][j];

      a += x1[j] * x1[j];
      b += x1[j] * x2[j];
      c += x2[j] * x2[j];
      d += x1[j] * x3;
      e += x2[j] * x3;
    }
  
  _smpSolveSymmetric2x2Eqn(a,b,c,d,e, &x, &y);

  bary [origin] = 1. - x - y; 
    
  bary [(origin + 1)%3] = x; 
     
  bary [(origin + 2)%3] = y; 
     
  _smpMakeBarycenter(3, tri, wp, bary);
      
  /* determine whether w projects inside or outside tri */

  inside = !(bary[0] < EPS_VERTEX_OUTSIDE_TRIANGLE || 
	     bary[1] < EPS_VERTEX_OUTSIDE_TRIANGLE ||  
	     bary[2] < EPS_VERTEX_OUTSIDE_TRIANGLE);

  *residual   = DIST3(w, wp);

  return inside;

}

#undef EPS_VERTEX_OUTSIDE_TRIANGLE
/*+--------------------------------------------------------------------------+
  |                                                                          |
  |  _smpSimplifiedBoundaryVertexLocation                                    |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData:: _smpSimplifiedBoundaryVertexLocation( int v0, int v1,
						     int *vstar0, int *vstar1, 
						     int val0, int val1, 
						     int method, int parent)
{
  int success = 1;

  if (method == 0)

    {
      memcpy(simplified_vertex, vert[parent], sizeof (Vertex));
    } 

  else if (method == 1)
    {

      /* place the simplified vertex so that the length of the boundary
	 will be preserved.  The vertex must be situated on an ellipse 

      ___________\/___________\/__________
      v*val0-1   v0           v1          v*val1-1
      == vl                                  == vr

      Firstly, we retrieve the vertices vl, v0, v1, vr from the stars.

      we compute the length of the polygonal arc vl--v0--v1--vr, and
      call 2a this length
     
      The focal points of the ellipsoid are vr and vl, so we note 2c the distance
      vr-vl.

      */

      float 
	a, a2, c, 
	*ve = (float *) simplified_vertex;

      Vertex 
	v_zero, v_one, vr, vl, vm;
  
      memcpy(v_zero, vert[v0], sizeof (Vertex));
      memcpy(v_one,  vert[v1], sizeof (Vertex));
  
      memcpy(vl, vert[vstar0[val0-1]], sizeof (Vertex));
      memcpy(vr, vert[vstar1[val1-1]], sizeof (Vertex));
      
      /* compute the parameters of the ellipsoid */

      a  = 
	(
	 DIST3(vert[v0],             vert[v1])  +
	 DIST3(vert[vstar0[val0-1]], vert[v0])  +
	 DIST3(vert[vstar1[val1-1]], vert[v1])
	 ) 
	/ 2.;

      a2 = a * a;
  
      c  = DIST3(vl,vr)/2.;
  
      /* Compute the middle of v_zero and v_one */

      vm[0] = (v_zero[0] + v_one[0]) / 2.;
      vm[1] = (v_zero[1] + v_one[1]) / 2.;
      vm[2] = (v_zero[2] + v_one[2]) / 2.;
 
      /* Then, our first next test is to determine whether a and c are
	 sufficiently different, otherwise, taking the middle of v_zero and
	 v_one will certainly work since we will have the same cord length
	 along the boundary */

      if (1. - c/a < 1e-6) 

	/* limit of the floating point precision */

	{
	  memcpy(ve, vm, sizeof(Vertex));
	}

      else /* if (a > c ) */
	{
	  float 
	    t,  x,  y, 
	    ye;

	  Vertex 
	    vp;
  
	  /* then we consider the middle of v0 and v1, vm,
	     we project vm on the segment vr vl and obtain vp

	     finally, we project vp on the ellipsoid defined by vr, vl, a
	     and c along the direction vp--->vm


                             ve


                             vm
                             |y
      vl________________0____|_____________vr
      -c                     x              c
                             vp


			     by projecting vm onto the segment (vl,vc), we obtain
			     vp, and the coordinates of vm in the coordinate system formed by
			     0(the middle of vr and vl) vr and vm.

      x and y are the coordinates of vm in this coordinate system
      We keep x unchanged, and compute ye such that

      x^2/a^2 + ye^2/b^2 = 1, with b^2 = a^2-c^2

                                     x^2
      hence        ye^2    =  (  1 -  ---  ) ( a^2 - c^2 )
                                     a^2

      We take for ye the positive square root of ye^2 

      and finally ve is obtained by

      ve-vp = ye/y (vm-vp) 

      ve    = vp + ye/y (vm-vp) */

 
	  /* compute x and y of vm with respect to vr and vl */

	  y  = _smpClosestPointOnEdge(vm, vp, vl, vr, &t);


	  /* what do we do if y == 0, in which case vm is accidentally
	     exactly on the line segment vr-vl?

	     In that case, we still have a non-degenerate ellipse since we
	     ruled out the fact that a and c are the same. Then either
	     v_zero and v_one cannot be on the line segment vr-vl.

	     So we project instead v0 on the segment vr vl 
	 
	     v0
	 ___|________________
       vl   vp0      vp      vr

                                ----->
       and we use the direction vp0 v0 for projecting the point vp on the 
       ellipse

       */


	  if (y/c < 1e-6 )
	    {
	      /* limit of floating print precision */
	      /* In that case, either v_zero or v_one cannot be on the line vl, vr,
		 so we replace vm with v_zero */
	  
	      memcpy(vm, v_zero, sizeof(Vertex));
	  
	      y  = _smpClosestPointOnEdge(vm, vp, vl, vr, &t);

	      if (y/c < 1e-6 )
		{
		  memcpy(vm, v_one, sizeof(Vertex));
	  
		  y  = _smpClosestPointOnEdge(vm, vp, vl, vr, &t);
		}
	    }

	  /* in case vm does not project orthogonally on the segment (vl,vr)
	     the routine ClosestPointOnEdge projects vm either on vl and assigns
	     1 to t or on vr and assigns 0 to t 
	     we need to determine the angle that vm,vp makes with (vl,vr)
	     */

	  if (c > 0.0 /* c == 0 is handled as the general case */
	      && (t == 0. || t == 1.))
	    {
	      Vertex u,v;
	      double 
		cos_uv, cos_uv_2, sin_uv_2,
		A, B, C, sol1, sol2, eps = 1e-12;

	      MAKEVEC3(vl,vr,u);
	      _smpNormalize3(u);
	  
	      MAKEVEC3(vp,vm,v);
	      _smpNormalize3(v);

	      cos_uv   = u[0] * v[0] + u[1]*v[1] + u[2]*v[2]; /* FScalProd(u,v); */

	      if (cos_uv < 0.)
		cos_uv = -cos_uv;

	      cos_uv_2 = cos_uv * cos_uv;

	      sin_uv_2 = 1. - cos_uv_2;

	      /* ye is the solution of the following equation of the second
		 degree:

		 x^2/a^2 + y^2/(a^2-c^2) = 1 <=>
		 
		 (ye cos_uv + c)^2/a^2 + (ye sin_uv)^2/(a^2-c^2) = 1 <=>

		 ye^2( cos_uv_2 + a^2/(a^2-c^2) sin_uv_2) +
		 ye  (2 c cos_uv) +
		 c^2-a^2                                  = 0


		 A ye^2 + cb ye + C = 0 */

	      A = cos_uv_2 + sin_uv_2 * a2 / (a+c)/(a-c);
	      B = 2. * cos_uv * c;
	      C = (c+a)*(c-a); /* c^2-a^2 */

	      _smpSolveQuadraticEquation(A,B,C,&sol1, &sol2, eps);

	      /* there is only one positive solution */

	      if (sol1 > 0.)
		ye = sol1;
	      else if (sol2 > 0.)
		ye = sol2;
	      else 
		{
		  ye = 0;
		  fprintf(logfp,"\"SimplifiedBoundaryVertexLocation\":");
		  fprintf(logfp," Quadratic equation has no positive root\n");
		}
	    } /*endif vm project on an extremity */

	  else /* vm projects normally inside the segment */
	    {
	      double tmp;
	      /* determine x from t: */

	      /* vp = t * vl + (1-t) * vr 
		 t == 1 <==> x = -c
		 t == 0 <==> x =  c
	 
		 x = c - 2 c t = c ( 1-2t) */
	      
	      x = c * (1. - 2. * t);
  

	      /* ye = sqrt ((double) (a2 - c * c)*( 1. - x*x/a2)); */

	      /* since a and c are likely to be very close, a more robust
		 computation is to avoid computations such as a2 - c^2
		 and compute sqrt(a2) if a is readily available */

	        tmp = (double) (a -c) * (a+c) * (a-x) * (a+x);

	      /* this is to avoid floating point exceptions when 
		 computing the square root */

		if (tmp >0.0) ye = sqrt (tmp) / a;

		else ye = 0.0;

	    }
	
	  if (y != 0.0)
	   
	    ye /= y;

	  /* do the following vector composition without
	     a function call ve    = vp + ye/y (vm-vp) */
    
	  ve[0] = vp[0] + ye * (vm[0] - vp[0]);
	  ve[1] = vp[1] + ye * (vm[1] - vp[1]);
	  ve[2] = vp[2] + ye * (vm[2] - vp[2]);

	  /* test whether the boundary length is preserved */
	  {

	    float 
	      l_before = 2. * a,
	      l_after  = DIST3(vl,ve) + DIST3(ve,vr);

	    if (l_before -l_after > 1e-6 || l_before - l_after < -1e-6)
	      success = 0;
	   

	  }

	} /* if (a>c) otherwise, vm, the middle is good for ve */

    } /* if (method = 1, length preservation) otherwise */
  
  return success;
}


/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpFastCompactness3DTriangle2                                         |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

/* compute the triangle compactness and (positive) area
   this is a fast method, in comparison with the method that 
   orthogonalizes the triangle and computes the area afterwards 
   
   this fast method uses a cross product 
 */

float _smpFastCompactness3DTriangle2(Vertex *tri, float *triangle_area)
{
  
  /* compute the squared lenghts of the three edges: */
  
  double sum_lengths_sq = 0., tmp;
   
  {

    register U_char i,j;

    for(i=0;i<3;i++)
      {
	double len_edg_sq = 0;

	for(j=0;j<3;j++)
	  {
	    tmp         = tri[(i+1)%3][j] - tri[i][j];
	    len_edg_sq += tmp * tmp;
	  }

	sum_lengths_sq += len_edg_sq;

      }
  }
      

  {
    double 
      m0 =    
	       tri[1][1] - tri[0][1],
      m1 =
	       tri[2][2] - tri[1][2],
      m2 =
	       tri[1][2] - tri[0][2],
      m3 =
	       tri[2][1] - tri[1][1],
      m4 =
	       tri[2][0] - tri[1][0],
      m5 =
	       tri[1][0] - tri[0][0],
      n0  =
	       m0 * m1 -
	       m2 * m3,
		 
      n1  =
	       m2 * m4 -
	       m5 * m1,
			 
      n2  =
	       m5 * m3 -
	       m0 * m4;
    
    tmp            = n0 * n0 + n1 * n1 + n2 * n2;

    if (tmp > 0.0) *triangle_area = sqrt (tmp) / 2.;

    else *triangle_area = 0;
     
	       
  }       

  
  return (float) (FOUR_SQRT_3 * *triangle_area / sum_lengths_sq);

}


/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpBoundaryCollapse2KndGeometricallyOk                                |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpBoundaryCollapse2KndGeometricallyOk(int v0, int v1,
					
					    /* this procedure must be run twice, once for the
					       star of the vertex v0 and once for the star of the vertex v1
					       Here, v1 rather stands for the first vertex of the star
					       */

					int *star0, int *vstar0, int val0, Vertex *s_normal0, float *s_area0,
					int direction, float min_scalprod_nor, float compactness_ratio) 

{ 
  int collapsible = 1, i = 0;

  Vertex faceVert[3];
  
  float 
    *s_comp0 = s_area0 + val0,
    min_compactness_before,
    min_compactness_after;

  /* labeling system for directed triangle stars 

	          v*1\      /v*0  
	           e*1\  *1/e*0   
	       *val0-1 \  / *0    
	     ___________\/___________   we omit on purpose the first vertex, because in the case
	     v*val0-1   v0          v1  of a boundary edge, we know it already


	     */

  /* retrieve the compactness of the first triangle */

  min_compactness_before = compactness[star0[0]];

  min_compactness_after  = 1; /* initialization of the compactness
				 after simplification */
  
  while (collapsible && i < val0-1)
    {
   
      /* retrieve the face compactness before simplification */
   
      min_compactness_before = MIN ( compactness[star0[i+1]], min_compactness_before);

      /* recompute the compactness after changing the first vertex */

      memcpy(faceVert[0], simplified_vertex, sizeof (Vertex));
    
      if (direction == DIRECTION_CLOCKWISE)
	{
	  /* reverse the order of triangle vertices, so that the normal
	     that we will compute will respect the surface orientation */
	  memcpy(faceVert[2], vert[vstar0[i]], sizeof (Vertex));
	  memcpy(faceVert[1], vert[vstar0[i+1]], sizeof (Vertex));
	}

      else
	{
	  memcpy(faceVert[1], vert[vstar0[i]], sizeof (Vertex));
	  memcpy(faceVert[2], vert[vstar0[i+1]], sizeof (Vertex));
	}

      s_comp0[i+1] = _smpFastCompactness3DTriangle2(faceVert, s_area0+i+1);

      min_compactness_after = MIN ( s_comp0[i+1], min_compactness_after);

      /* recompute the face normal assuming that the vertex v0 has
         moved and save this normal for future use*/

      _smpTriangleNormal(faceVert, s_normal0[i+1]);


      /* check whether the triangle normal orientation is consistent */

      collapsible = ((SCALPROD3(s_normal0[i+1], face_normals[star0[i+1]])) > min_scalprod_nor);

      i++;
    }

  if (collapsible)

    /* check whether the minimum triangle compactness is too much degraded */

    collapsible = min_compactness_after > (compactness_ratio * min_compactness_before);
  
  return collapsible;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpRemoveEdgesFromHeap                                                |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpRemoveEdgesFromHeap(int *estar, U_short val)
{
  int 
    num_removed = 0;

  register U_short i;
 
  /* remove all edges of the star */
 
  for(i=0;i<val;i++)
    {
      int the_edge = estar[i];
      int idx      = edge2index[the_edge];
      if (idx >=0 && idx < edge_heap->length())
	{
	  num_removed++;
	  edge_heap->del(idx);
	}
    }

  return num_removed;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpReinstateEdgesInHeap                                               |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int  SimpData::_smpReinstateEdgesInHeap(int *estar, U_short val, int *num_edg_added)

     /* return the number of edges that effectively have been reinstated
	(were in the queue before, were deleted and are now reintered */
{
  
  int 
    num_edg_reinstated = 0;

  register int i;

  *num_edg_added = 0;

  /* add all edges of the star to the heap  */
 
  for(i=0;i<val;i++)
    {
      int
	outside = (edge2index[estar[i]] == heap_outside_index),
	added    = _smpAddEdge2Heap(estar[i]);
		 
      if (added) 
	{
	  *num_edg_added += 1;
	  if (outside) num_edg_reinstated++;
	  
	}
    }

  return num_edg_reinstated;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpCollapseBoundaryEdge2ndKnd                                         |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpCollapseBoundaryEdge2ndKnd(int edg_num, int v0, int v1, int val0, int val1, 
					     int *star0, int *star1, int *vstar0, int *vstar1, int *estar0, int *estar1,
					     Vertex *s_normal0, Vertex *s_normal1, float *s_area0, float *s_area1,
					     int parent, int child)
{
  int 
    n_edges_added[2],
    n_edges_reinstated = 0;
 

    /*   e*:
	          v*1\      /v*0  \      /
	           e*1\  *1/e*0    \ *1 /
	       *val0-1 \  / *0   *0 \  / *val1-1
	     ___________\/___________\/__________
	     v*val0-1   v0    vs      v1
             vl                                  vr
	     */


  /* father relationships:

                    /\
              e*0[0]  \e*1[0]
                  /    \
                 /<-----\		  
                /     ---\--> *1[1]
		   *1[0]
            v0  <-------  v1


	    This is only true if val1 > 1
	    otherwise if val1 ==1 we must have val0 > 1
     */

  /* 1) change the red_valence counting the number of remaining faces incident to a
     given vertex */

  _smpNoInteriorVerticesInRemovedFaceSetUpdateRedValence(edg_num);

  /* update the component information for the removed face set */
  _smpRemovedFaceForest(edg_num, 0, 1);

  /* 2) update the vertex, triangle and edge fathers */

  vertex_father[child]         = parent;

  if (val1 > 1)
    {
      edge_father   [estar1[0]] = edge_father   [estar0[0]];
      
      triangle_father[star1[0]] = triangle_father[star1[1]];
    }
  else
    {
      edge_father   [estar0[0]] = edge_father   [estar1[0]];
      
      triangle_father[star0[0]] = triangle_father[star0[1]];
    }

  /* 3) change the valences of the vertices */

  valence[parent]               = val0 + val1 - 2;
  valence[vstar0[0]] --;

  /* 4) remove from the edge heap all edges that have changed   */
  
  _smpRemoveEdgesFromHeap(estar0, val0);

  _smpRemoveEdgesFromHeap(estar1, val1);

  /* mark the 2 discarded edges. 
     
     verify that estar1[0] was not already counted as outside the heap */

  if (edge2index[estar1[0]] != heap_outside_index)
    num_edg_remaining_4_testing --;
  
  edge2index[estar1[0]] = -4;
  edge2index[edg_num]   = -7;
  

  /* 5) reinstate edges in the heap */

  n_edges_reinstated = 
    _smpReinstateEdgesInHeap(estar0, val0, n_edges_added);
  
  /* reinstate all but the first edge (that collapses) in the star of v1 */

  n_edges_reinstated += 
    _smpReinstateEdgesInHeap(estar1+1, val1-1, n_edges_added+1);
 
  /* 6) update the number of edges left to be tested */

  num_edg_remaining_4_testing += n_edges_reinstated;

  /* 7) update the coordinates of the simplified vertex */

  memcpy(vert[parent], simplified_vertex, sizeof(Vertex));

  /* 8) replace the triangle normals, areas, and compactness in the simplified stars */

  {
    register int i;

    float 
      *s_comp0 = s_area0 + val0,
      *s_comp1 = s_area1 + val1;

    /* replace for v0 */
    for(i=1;i<val0;i++)
      {
	face_area[star0[i]]    = s_area0[i];
	compactness[star0[i]]  = s_comp0[i];
	memcpy(face_normals[star0[i]], s_normal0[i], sizeof(Vertex));
      }
    /* replace for v1 */

    for(i=1;i<val1;i++)
      {
	face_area[star1[i]]    = s_area1[i];
	compactness[star1[i]]  = s_comp1[i];
	memcpy(face_normals[star1[i]], s_normal1[i], sizeof(Vertex));
      }
  }
  
  /* 9. update removed texture area (when texture coordinates per corner are used) */

  if (texCoordBinding == PB_PER_CORNER) _collapsed_texture_area += _inc_collapsed_texture_area;
 
  return n_edges_added[0] + n_edges_added[1];
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpCollapsibilityTestsBoundaryEdge2ndKind                             |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpCollapsibilityTestsBoundaryEdge2ndKind(int edge_num, int v0, int v1, int val0, int val1, 
							 int move_simplified_vertex)
{
  int 
    collapsible = 0,
    parent      = v0,
    child       = v1,
    *edge_t     = (int *) SurfEdgeGetTriangles (edge_array+edge_num),
    *star0      = NULL;

  Vertex *s_normal0 = NULL;

  float *s_area0    = NULL;

  /* perform collapsibility tests particular to a boundary edge of the
     second kind, such that both edge vertices are boundary vertices

     In addition of the two vertices touching a boundary, the edge must
     effectively belong to the boundary, which means that exactly one
     triangle must be missing, with an index of -1. If both triangles had
     a -1 index, though, we would not have a valid edge */

  if (boundary_vert[v0] && boundary_vert[v1] && (edge_t[0] == -1 || edge_t[1] == -1))
    {
 
      /* we only accept the edges such that both end vertices are located on the
	 boundary */
      
      int 
	first_triangle = edge_t[0];
      
      /* perform the valence test. For this particular case, the valence
	 test is 1 <= v(v0) + v(v1) -2 <= valence_max */

      if ((val0 + val1 -2 >= 1) && (val0 + val1 -2 < valence_max))
	{
	 
	  /* build the edge star:
	     contrary to the procedure applied to the surface interior vertices,
	     we compute the vertex star in a directed fashion.
	     Here is the labeling system:

	          v*1\      /v*0  \      /
	           e*1\  *1/e*0    \ *1 /
	       *val0-1 \  / *0   *0 \  / *val1-1
	     ___________\/___________\/__________
	     v*val0-1   v0           v1

	     */

	  /* allocate storage space for the vertex stars */

	  star0 = (int *) _vertex_star_buffer;

	  bzero (star0, (val0 + val1) * 3 * sizeof (int));

	  /*
	    
	    we next rotate the edge so that edge_t[0] is a valid triangle,     t[0]
	    i.e., so that the representation above holds true              v[0]-----v[1]
                                                                                t[1]
										*/  
	  if (first_triangle == -1)
	    {
	      /* rotate the edge, i.e. echange v0, v1, val0 and val1 
		 using first_triangle as a temporary storage location
		 
		 assign edge_t[1] to first_triangle
		 */

	      first_triangle = v0;
	      v0             = v1;
	      v1             = first_triangle;
	      
	      first_triangle = val0;
	      val0           = val1;
	      val1           = first_triangle;

	      first_triangle = edge_t[1];
	    }

	  
	  {int 
	      *vstar0        = star0  + val0,
	      *estar0        = vstar0 + val0,
	      *star1         = estar0 + val0,
	      *vstar1        = star1 + val1,
	      *estar1        = vstar1 + val1;
	      
	    first_triangle = FATHER(first_triangle, triangle_father);

	    _smpDirectedParentVertexStar(v0, first_triangle, val0, star0, DIRECTION_COUNTER_CLOCKWISE);

	    /* compute the star of v1 in reverse order */

	    _smpDirectedParentVertexStar(v1, first_triangle, val1, star1, DIRECTION_CLOCKWISE);

	   
	    /* 
	       up to now, we found a parent triangle 
	       and we located the next edge, "n_edgeCCW" or "n_edgeCW" 
	       depending on whether we are rotating counterclockwise
	       or clockwise around "parent vertex".
	       
	       The parent of this edge must be stored in the e_star. 

	                              _/                                                          
	                    n_edgeCCW/\
		   parent vertex    /  \  
	                        \_ /____\                                                         
	                           n_edge\_                                                      
	                             CW       

	       Similarly to before, we need to orient the edge so that 
	       the parent of the first edge vertex is precisely "parent vertex"
	       */  
	
	    /* the manifold test is very similar to the test we
	       perform for an interior edge case we compare the
	       vertices of the links and determine whether duplicate
	       vertices exist */
	      
	    if (
		_smpManifoldLinkTest(vstar0+1, vstar1+1, val0-1, val1-1) &&
		_smpManifoldRemovedFaceSet(edge_num) &&
		_smpRemovedFaceForest(edge_num, 1, 0) && 
		// colors per faces: are the disappearing colors represented in the neighborhood?
		(colorBinding   != PB_PER_FACE   ||_smpColorPerFaceOk(star0[0], -1 , val0-1, star0+1, val1-1, star1+1)) && 
		// texCoords per corner: how large is the texture area that will disappear?
		(texCoordBinding!= PB_PER_CORNER ||_smpTexCoordPerCornerOk(star0[0], -1 , val0-1, star0+1, val1-1, star1+1))
		)
	      {
		  Vertex 
		    *s_normal1;
		  float 
		    *s_area1 ;

		  s_normal0 = (Vertex *) _edge_star_buffer_vx;

		  bzero (s_normal0, ( val0 + val1 ) * sizeof (Vertex));
		  
		  s_normal1 = s_normal0 + val0;
		       
		  s_area0 = (float *) _edge_star_buffer_fl;

		  bzero (s_area0, 2 * ( val0 + val1 ) * sizeof (float));

		  s_area1 = s_area0 + 2 * val0;

		  /* position the simplified vertex so that the length of the boundary
		     will be preserved.  */

		  if (_smpSimplifiedBoundaryVertexLocation(v0, v1, vstar0, vstar1, val0, val1, 
							   move_simplified_vertex, parent))
		    {  
		      int the_position = 0, num_positions = 1 + !(move_simplified_vertex);

			      
		      while (the_position < num_positions && (!collapsible))
			{
		     
			    
			  /* compute the difference in orientation for the
			     triangle normals before and after simplification */
		      
			  /* compute the minimum compactnesses before and after simplification */

			  if ((_smpBoundaryCollapse2KndGeometricallyOk(v0, v1, star0, vstar0, val0, 
								       s_normal0, s_area0,
								       DIRECTION_COUNTER_CLOCKWISE,
								       min_scalprod_nor, 
								       compactness_ratio) )
			      &&
			      (_smpBoundaryCollapse2KndGeometricallyOk(v1, v0, star1, vstar1, val1, 
								       s_normal1, s_area1, DIRECTION_CLOCKWISE,
								       min_scalprod_nor, 
								       compactness_ratio) )
			      ) /* geometry test */


			    {
			      /* compute the tolerance after simplification */
			    
			      if (_smpErrorWithinToleranceBoundary2ndKnd(v0, v1, vstar0, vstar1,
									 val0, val1, parent, child))
				{
				
				  num_edg_collapsed +=2;

				  num_edg_weights += 

				    _smpCollapseBoundaryEdge2ndKnd(edge_num, v0, v1, val0, val1,
								   star0,  star1,  vstar0, vstar1, 
								   estar0, estar1, s_normal0, s_normal1,
								   s_area0, s_area1, parent, child); 

				  collapsible = 1;

				}
			      else /* reject for tolerance off limits */
				edg_rejected_4_tolerance ++;
			   
			    }
			  else /* reject for geometry */
			    edg_rejected_4_geometry ++;

			  the_position++;
			  /* now use the second (alternate) position */
			  memcpy(simplified_vertex, simplified_vertex_alt, sizeof (Vertex));
			} /*end while: if ! move vertices, try two positions */
		    }
	      }

	    else /* reject for topology */
	    
	      {
		edge2index[edge_num] = -3; 
			
		/* this condition will not be modified even if the
		   surface is simplified around the edge */

		edg_rejected_4_topology ++;
	      }
	  }
	}
      else   
	edg_rejected_4_topology ++;

    } /* end if (edge_t[0] == -1 || edge_t[1] == -1) */

  return collapsible;
      
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpManifoldLinkTest1stKnd                                             |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpManifoldLinkTest1stKnd(int v0, int val0, int *star0, int *star1, int *link1, int val1)
{

  int link_is_manifold = 1;

  /* 
     v0 is a  boundary vertex
     v1 is an interior vertex


     | /\vt
     |/  \<-----this triangle is *1[0]
     v0___\_v1
     |\   /
     | \ /<-----this triangle is *1[1]
     |  vb


     this test works by taking 
     all vertices in the link of v0, except v1, and verify that they are
     all different from the vertices of the link of v1, except vt, v0 and vb
     
     we take from vstar0 as many vertices as val0 (remember that vstar0 has one element
     more than star0, because v0 is a boundary vertex)

     we skip the vertices 0, 1, and val1-1 from the link of v1, i.e.,
     we take from vstar1 as many vertices as val1-3
     
   */
  
  int 
    val = val0 + val1 -3;

  /* since 
	val0 + val1 <= SIMPLIFY_VALENCE_MAX4,
	I can use a static array here. */

  static int
    link[SIMPLIFY_VALENCE_MAX4];

    {
    
      int 
	i  = 0,
	
	tri_v_rotated[3],  tri_v_parents[3], 
    
	/* rotated edge vertices and adjacent triangles: */
	edge_v_rotated[2], edge_v_parents[2], edge_t_rotated[2],
    
	triangle = star1[0],
	edge;

      do
	{
	  triangle = FATHER (triangle,  triangle_father);

	  _smpRotateParentTriangle(v0, vertex_father,
				   triangles + 3 * triangle,
				   tri_v_parents, tri_v_rotated);

	  edge = _smpNextParentEdge(tri_v_rotated, DIRECTION_COUNTER_CLOCKWISE);

	  _smpRotateParentEdge(v0, vertex_father,
			       (int *) SurfEdgeGetVertices (edge_array + edge),
			       (int *) SurfEdgeGetTriangles(edge_array + edge),
			       edge_v_parents, edge_v_rotated, edge_t_rotated);
	  star0[i]  = triangle;
	  link[i++] = edge_v_parents[1];

	  triangle = edge_t_rotated[0];

	}
      while (triangle != -1);


      /* now repeat the same operation in the clockwise direction
	 starting at vb (for Vertex_at_Bottom) */

      triangle = star1[1];

      do
	{
	  triangle = FATHER (triangle, triangle_father);

	  _smpRotateParentTriangle(v0, vertex_father,
				   triangles + 3 * triangle,
				   tri_v_parents, tri_v_rotated);

	  edge = _smpNextParentEdge(tri_v_rotated, DIRECTION_CLOCKWISE);

	  _smpRotateParentEdge(v0, vertex_father,
			       (int *) SurfEdgeGetVertices (edge_array + edge),
			       (int *) SurfEdgeGetTriangles(edge_array + edge),
			       edge_v_parents, edge_v_rotated, edge_t_rotated);

	  star0[i]  = triangle;
	  link[i++] = edge_v_parents[1];

	  triangle = edge_t_rotated[1];
	}
      while (triangle != -1);


      /* gather vertices of link0 and link1 in a single array, sort
	 the array, and determine whether there are any duplicate
	 vertices */

      memcpy(&link[i], link1+2, (val1 -3) * sizeof (int));

      qsort(link, val, sizeof(int), (int (*)(const void*, const void*)) _smpCmpIntSmallFirst);

      i = 0;

      while ((i < val -1 ) && (link_is_manifold))
	{
	
  
	  if (link[i+1] == link[i]) /* if there are duplicate vertices,
				       they must be contiguous in the
				       sorted array */
				       
	    link_is_manifold = 0;
	  i++;
	}

    }

  return (link_is_manifold);
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpBoundaryCollapse1stKndGeometricallyOk                              |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpBoundaryCollapse1stKndGeometricallyOk(int v0, int *star0, int *vstar0, 
							int val0, Vertex *s_normal0, float *s_area0,
							float min_scalprod_nor, float compactness_ratio) 

{ 
  int collapsible = 1, i = 1;

  Vertex faceVert[3];
  
  float 
    *s_comp0 = s_area0 + val0,
    min_compactness_before,
    min_compactness_after;

  /*
    labeling system          *[0]\  /
                         v*[0]____\/____
			      *[1]/\
                                 /  \
				 *[2]

    both triangles *[0] and *[1] disappear after edge collapse

   */


  /* compute the compactness of the collapsed triangles */
  
  min_compactness_before = MIN ( compactness[star0[1]], compactness[star0[0]]);

  min_compactness_after  = 1; /* initialization of the compactness
				 after simplification */
  
  while (collapsible && i < val0-1)
    {
 
      /* retrieve the face compactness before simplification */
      min_compactness_before = MIN ( compactness[star0[i+1]], min_compactness_before);

      /* recompute the compactness, and area, after changing the first vertex */
  
      memcpy(faceVert[0], simplified_vertex, sizeof (Vertex));
      memcpy(faceVert[1], vert[vstar0[i]], sizeof (Vertex));
      memcpy(faceVert[2], vert[vstar0[i+1]], sizeof (Vertex));

      s_comp0[i+1] = _smpFastCompactness3DTriangle2(faceVert, s_area0+i+1);

      min_compactness_after = MIN ( s_comp0[i+1], min_compactness_after);

      /* recompute the face normal assuming that the vertex v0 has
         moved and save this normal for future use*/

      _smpTriangleNormal(faceVert, s_normal0[i+1]);

      /* check whether the triangle normal orientation is consistent */

      collapsible = (SCALPROD3(s_normal0[i+1], face_normals[star0[i+1]]) > min_scalprod_nor);

      i++;
    }

  if (collapsible)

    /* check whether the minimum triangle compactness is too much degraded */

    collapsible = min_compactness_after > (compactness_ratio * min_compactness_before);
  
  return collapsible;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpCollapseBoundaryEdge1stKnd                                         |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpCollapseBoundaryEdge1stKnd(int edg_num, int v0, int v1, int val0, int val1, 
					     int *star1, int *vstar1, int *estar1,
					     Vertex *s_normal1, float *s_area1)
{
  int 
    n_edges_added      = 0,
    n_edges_reinstated = 0,
    et, eb,
    tri_v_rotated[3],  tri_v_parents[3];
 
   /*                     v*[val1-1]
       labeling system          *[0]\  /
                            v*[0]____\/____
	   		         *[1]/\
                               v*[1]/  \
				    *[2]

				 both triangles *[0] and *[1] disappear after edge collapse

   */


  /* father relationships:

                    /\
                et /  \e*1[val1-1]
                  /    \
                 /<-----\		  
                /     ---\--> *1[val1-1]
		   *1[0]
            v0  <-------  v1
               \   *1[1]/
            eb  \    --/------> *1[2]
                 \<---/e*1[1]
                  \  /
                   \/
     */
	 
  /* 1) change the red_valence counting the number of remaining faces incident to a
     given vertex */
  _smpNoInteriorVerticesInRemovedFaceSetUpdateRedValence(edg_num);

  /* update the component information for the removed face set */
  _smpRemovedFaceForest(edg_num, 0, 1);



  /* 2) update the vertex, triangle and edge fathers */

  /* 2.1) First retrieve the edges et and eb on "top" and "bottom" of
     v0, since these edges were not pre-stored */

  /* take the triangle star1[0], orient it with respect to v0
     and use the oriented triangle vertices to determine the next edge counter clockwise starting
     from v0: et */

  _smpRotateParentTriangle(v0, vertex_father, triangles + 3 * star1[0],
			   tri_v_parents, tri_v_rotated);

  et =  (int) (((EdgeS *) _smpFindEdge(e_table, tri_v_rotated[0], tri_v_rotated[2])) - edge_array);

  et =  FATHER (et, edge_father);

  /* do the same with the triangle star1[1] to determine the next edge clockwise starting
     from v0: eb */

  _smpRotateParentTriangle(v0, vertex_father, triangles + 3 * star1[1], tri_v_parents, tri_v_rotated);

  eb =  (int) (((EdgeS *) _smpFindEdge(e_table, tri_v_rotated[0], tri_v_rotated[1])) - edge_array);

  eb =  FATHER (eb, edge_father);

  /* 2.2) */

  vertex_father[v1]   = v0;

  edge_father   [estar1[0]]      = edge_father   [estar1[1]] = eb;
  edge_father   [estar1[val1-1]] = et;

  triangle_father[star1[0]] = star1[val1-1];
  triangle_father[star1[1]] = star1[2];
   
  /* 3) change the valences of the vertices */

  valence[v0]               = val0 + val1 - 4;
  valence[vstar1[1]]      --;
  valence[vstar1[val1-1]] --;

  /* 4) remove from the edge heap all edges that have changed   */
  
  _smpRemoveEdgesFromHeap(estar1, val1);

  /* mark the 2 discarded edges. 
     
     verify that they were  not already counted as outside the heap */

  if (edge2index[estar1[1]] != heap_outside_index)
    num_edg_remaining_4_testing --;
   
  if (edge2index[estar1[val1-1]] != heap_outside_index)
    num_edg_remaining_4_testing --;
  
  edge2index[estar1[1]] = edge2index[estar1[val1-1]] = -4;
  edge2index[edg_num]   = -7;
  

  /* 5) reinstate all but the two edges and last edge  in the star of v1 */

  n_edges_reinstated += 
    _smpReinstateEdgesInHeap(estar1+2, val1-3, &n_edges_added);
 
  /* 6) update the number of edges left to be tested */

  num_edg_remaining_4_testing += n_edges_reinstated;

  /* 7) update the coordinates of the simplified vertex" no need since
   its coordinates do not change*/

  /* 8) replace the triangle normals, areas and compactness in the simplified star of v1 */

  {
    register int i;

    float *s_comp1 = s_area1 + val1;
    
    /* replace for v1 */

    for(i=2;i<val1;i++)
      {
	face_area[star1[i]]    = s_area1[i];
	compactness[star1[i]]  = s_comp1[i];
	memcpy(face_normals[star1[i]], s_normal1[i], sizeof(Vertex));
      }
  }

  /* 9. update removed texture area (when texture coordinates per corner are used) */

  if (texCoordBinding == PB_PER_CORNER) _collapsed_texture_area += _inc_collapsed_texture_area;
 
  return n_edges_added;
}


/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpCollapsibilityTestsBoundaryEdge1stKind                             |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpCollapsibilityTestsBoundaryEdge1stKind(int edge_num, int v0, int v1, int val0, int val1)
{
   /* perform collapsibility tests particular to a boundary edge of
     the first kind, such that exactly one of the edge vertices is a
     boundary vertex */

  int 
    collapsible = 0;

  int 
    *star0 = NULL,
    *star1 = NULL;

  Vertex 
    *s_normal1 = NULL;

  float 
    *s_area1 = NULL;

  /* the valence of v0 becomes val0 + val1 - 4 */
  
  if ((val0 + val1 -4 >= 1) && (val0 + val1 -4 < valence_max))
    {
      int
	*edge_t     = (int *) SurfEdgeGetTriangles (edge_array+edge_num),
	first_tri   = edge_t[0],
	second_tri  = edge_t[1],
	/* this is the counter-clockwise order with respect to v1 */
	*vstar1, 
	*estar1;

      /* permute v0 and v1  so that v0 is the boundary vertex */

      if (boundary_vert[v1])
	{
	  int 
	    tmp = v0;
	  v0 = v1;
	  v1 = tmp;

	  tmp = val0; val0 = val1; val1 = tmp;
	  first_tri = edge_t[1];      second_tri = edge_t[0];
	}
 
      /* initialize the storage pointer for the buffer star1 */

      star1 = (int *) _vertex_star_buffer;
      star0 = star1 + 3 * val1;

      bzero (star1, 3 * val1 * sizeof (int));
      bzero (star0, val0 * sizeof (int));

      /*
                   | / first_triangle
                   |/
                   v0----------------v1
                    \
                     \
                   boundary
                   */

    
      estar1 = star1  + val1;
      vstar1 = estar1 + val1;

      first_tri = FATHER(first_tri, triangle_father);

      /* first star elements (index 0)*/
      star1[0]  = first_tri;
      estar1[0] = edge_num;
      vstar1[0] = v0;
	  
      /* complete the star from indices 1 to val1-1 */
      _smpParentVertexStar(v1, second_tri, (U_short) val1, star1); 

      if ( // take advantage of this test to fill star0 with the appropriate faces
	  _smpManifoldLinkTest1stKnd(v0, val0, star0, star1, vstar1, val1) &&
	  _smpManifoldRemovedFaceSet(edge_num) && 
	  _smpRemovedFaceForest(edge_num, 1, 0) && 
	  // colors per faces: are the disappearing colors represented in the neighborhood?
	  (colorBinding    != PB_PER_FACE   || _smpColorPerFaceOk(star1[0], star1[1] , val0-2, star0, val1-2, star1+2)) && 
	  // texCoords per corner: how large is the texture area that will disappear ?
	  (texCoordBinding != PB_PER_CORNER || _smpTexCoordPerCornerOk(star1[0], star1[1] , val0-2, star0, val1-2, star1+2)) 
	  )
	{

	  /* the simplified vertex is already positioned, since it must be v0 */
	  
	  memcpy(simplified_vertex, vert[v0], sizeof (Vertex));

	  /* initialize the storage pointer for the buffer s_normal1 */

	  s_normal1 = (Vertex *) _edge_star_buffer_vx;

	  bzero(s_normal1, val1 * sizeof (Vertex));


	  /* initialize the storage pointer for the buffer s_area1 */

	  s_area1 = (float *) _edge_star_buffer_fl;

	  bzero (s_area1, 2 * val1 * sizeof (float));

		  
	  if (_smpBoundaryCollapse1stKndGeometricallyOk(v1, star1, vstar1, val1, 
							s_normal1, s_area1, min_scalprod_nor, 
							compactness_ratio))

	    {
	      /* perform the error tolerance test */
		      
	      if (_smpErrorWithinToleranceBoundary1stKnd(v0, v1, vstar1, val1))
		{
		  num_edg_collapsed +=3;
		  
		  num_edg_weights += _smpCollapseBoundaryEdge1stKnd(edge_num, v0, v1, val0, val1, 
								    star1, vstar1, estar1, s_normal1, s_area1);
			 
		  collapsible = 1;

		}
	      else /* reject for tolerance off limits */
		edg_rejected_4_tolerance ++;
	    }

	  else /* reject for geometry */
	    edg_rejected_4_geometry ++;

	}
      else
	{ 
	  /* this condition will not change if the surface is
	     further simplified */

	  edge2index[edge_num] = -3; 
			
	  edg_rejected_4_topology ++;
	}
    }
  else   
    edg_rejected_4_topology ++;

  return collapsible;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpCollapseTopologicallyFeasible                                      |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int  _smpCollapseTopologicallyFeasible(int *vstar0, int *vstar1, 
				       U_short val0, U_short val1)
{

  register U_short i, j;
  int feasible = 1;

  /* count each vertex only once 
                        *1[val1-1]  /\ *0[1]  counted twice
                           *1[0]   /  \ *0[0]
                                   \  /
			      *1[1] \/ *0[val0-1] counted twice
   */

  {
    /* verify that no two vertices of the link are the same */

    val0 -= 2;
    val1 -= 2;

    i = 2;
    while(( i <= val0) && (feasible))
      {

	j = 2;
	while ((j <= val1) && (feasible))
	  {

	    if (vstar0[i] == vstar1[j])
	      feasible = 0;
	    j++;
	  }
	i++;
      }
  }

  return(feasible);

}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpBuildParentEdgeStars                                               |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void SimpData::_smpBuildParentEdgeStars(EdgeS *edg, int v0, int v1, U_short val0,
					U_short val1, int *star0, int *star1)


/*...........................................................................*/
/*
   The following procedure computes the star of v0 and v1 belonging to
   an "active" surface data structure, i.e. the surface is undergoing a
   simplification process.

   the ordering of the stars is done such that:

         e2    e1                     e5    e4
          \ t2 /  t1                  \ t5 /
      t3   \  /                   t0   \  /  t4
  e3 _______\/_______e0       e0 _______\/_______e3
            /\v0     v1        v0       /\v1    
      t4   /  \  t0               t1   /  \  t3
          / t5 \                      / t2 \
         e4     e5                   e1     e2
*/
/*...........................................................................*/

{
  int *estar0 = star0  + val0, *vstar0 = estar0 + val0;
  int *estar1 = star1  + val1, *vstar1 = estar1 + val1,
      e_num   = (int) (edg - edge_array);

  int *edg_triangles = SurfEdgeGetTriangles(edg);
  /* complete first the straighforward portion of the stars
         /\
        /  \
       /    \    father(edg_triangles[0]);
      /      \           
   v0/edg->num\ v1         
     \        /          
      \      /           
       \    / father(edg_triangles[1]);
        \  / 
	 \/ 
	 */

  star0 [0] = FATHER(edg_triangles[1], triangle_father);
  vstar0[0] = v1;
  estar0[0] = e_num; /* the simplifiable edge is always a root edge */

  _smpParentVertexStar(v0, edg_triangles[0], val0, star0);
 
  star1 [0] = FATHER(edg_triangles[0], triangle_father);
  vstar1[0] = v0;
  estar1[0] = e_num;
  
  _smpParentVertexStar(v1, edg_triangles[1], val1, star1);
 
  return;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpParentVertexStar                                                   |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void SimpData::_smpParentVertexStar(int vf, int t0, U_short val, int *star)

/*...........................................................................*/
/*
   an "active" vertex is a vertex belonging to the inside of a surface
   ( not a boundary vertex ) undergoing a simplification process
*/
/*...........................................................................*/
{
  int *estar = star  + val;
  int *vstar = estar + val;
  register int i;
  int t = t0;
  
  /* complete the star from 1 to the valence,
     the 0-elements have been already completed */

  for(i=1;i<val;i++)
    {
      register U_char j=0;
      int vA,vAf,vB,vBf,ef, e_num, *endpoints, *facing_tris;
      EdgeS *edg, *edgf;

      t = FATHER(t, triangle_father);

      /* complete the triangle part of the vertex star */
      star[i] = t;
      do 
	{
	  vA = triangles[3 * t + j];
	  vB = triangles[3 * t + (j+2)%3];
	  vAf  = FATHER(vA, vertex_father);
	}
      while((j++ < 3) && (vAf != vf));
       
      /* now complete the edge portion of the vertex star */

      edg   = _smpFindEdge(e_table,vA,vB);

      e_num = (int) (edg - edge_array); /* differential pointer is
					   also the edge number */

      ef    = FATHER (e_num, edge_father);

      estar[i] = ef;

      /* find out the parents of the vertices of ef */

      edgf = edge_array+ef;

      endpoints   = SurfEdgeGetVertices(edgf);
      facing_tris = SurfEdgeGetTriangles(edgf);

      vAf  = FATHER(endpoints[0], vertex_father);
      vBf  = FATHER(endpoints[1], vertex_father);
      
      /* and the vertex portion of the vertex star */

      if (vAf == vf) 
	{
	  vstar[i] = vBf;
	  /*                          /\f0  ^ccw direction
             the next triangle is  v0/__\v1 |          v0 = vf, v0 > v1 
                                     \  /   |
                                      \/f1   */
 
	  t        = facing_tris[0];
	}
      else
	{
	  vstar[i] = vAf;
	  t        = facing_tris[1];
	 }
    
    }

  return;
}

static void  BuildEdgeStarVal1Is3(Vertex *, Face *, int, int, int, int, int, int *, Vertex *);

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   BuildEdgeStarVal1Is3                                                   |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

static void  BuildEdgeStarVal1Is3(Vertex *vert, Face *face, 
				  int v0, int v1, int val0, int val01, 
				  int val02, int *vstar0, Vertex *surfvert)
{
  register int i,i1;
  
  /* example of numbering for the case when val0 = 6, val1 = 3 

          v1         _v0
          \        _//|
           \ t1  _/ / |  The interior vertices come last
        t2  \  _/t0/  |  The indices of triangles are shifted by one
     v2______\/___/ t6|  with respect to the vertex star
           v5/\_  \v6 |
        t3  /   \t5\  |
           /  t4  \_\ |
          /         \\|
         v3          v4      */

        
  /* 1) copy the locations of the vertices */

  /* 1.1) star of v0 */
  for (i=0, i1=1; i< val01; i++, i1++)
    memcpy(vert[i], surfvert[vstar0[i1]], sizeof(Vertex));
  memcpy(vert[val01], surfvert[v0], sizeof(Vertex));

  /* 1.2) star of v1 */
  memcpy(vert[val0], surfvert[v1], sizeof(Vertex));

  /* 2) create the faces or connections between vertices */

 face[0][0] = val01; face[0][1] = val0; face[0][2] = 0;
  
  for (i = 1, i1 = 0; i <= val02; i++, i1++)
    {
      face[i][0] =  val01; face[i][1] =  i1; face[i][2] =  i;
    }
  face[val01][0] = val01; 
  face[val01][1] = val02;
  face[val01][2] = val0;
  
  face[val0][0] = val0;
  face[val0][1] = val02;
  face[val0][2] = 0; 

  return;
}


/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpBuildEdgeStar                                                      |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void SimpData::_smpBuildEdgeStar(Vertex *star_vert, Face *star_face, Plane *star_plane, float *star_areas, 
				 float *star_comp, int v0, int v1, U_short val0, U_short val1, 
				 int *star0, int *star1)

/*...........................................................................*/
/*
   
   represent the edge star with origin at the centroid of the edge link

   copy the vertices and faces of both vertex stars in a single array:
   here is the new numbering of the vertices:

       1      0                    val0+val1-4 
        \    /               \    /   ..
         \  /                 \  /
   .. ____\/____           ____\/____val0+1    
    val0-1/\v0      val0+val1-3/\
         /  \                 /  \
        /    \               /    \val0
       ..    val0 -2             

    The number of triangles in star_val is val0 + val1 -2
    The number of vertices  is             val0 + val1 -4 + 2,
    since v0 and v1 belong to the edge star
    In a vertex star, there is one more vertex (the center) than 
    there are triangles.
*/
/*...........................................................................*/

{
  /* there are two cases: either val1 is exactly 3 or it is > 3 */
 
  int  star_val    = val0 + val1 - 2,
           val013  = star_val    -1,
           val014  = val013      -1,
           val01   = val0        -1,
           val02   = val01       -1,
          *vstar0  = star0       + 2 * val0, 
          *vstar1  = star1       + 2 * val1;


  if (val1 == 3)
    {
      BuildEdgeStarVal1Is3(star_vert, star_face, v0, v1, val0, val01, val02, vstar0, vert);
    }
  else
    {
  
      register int i,i1;
 
      /* 0) copy the locations of the vertices */

      /* 0.1) star of v0 */
      
      for (i=0, i1=1; i< val01; i++, i1++)
	memcpy(star_vert[i], vert[vstar0[i1]], sizeof(Vertex));

      memcpy(star_vert[val01], vert[v0], sizeof(Vertex));

      /* 0.2) star of v1 */

      for (i=val0, i1=2; i<= val014; i++, i1++)
	memcpy(star_vert[i], vert[vstar1[i1]], sizeof(Vertex)); 
    
      memcpy(star_vert[val013], vert[v1], sizeof(Vertex));

      /* 1) copy the faces */
  
      /* numbering of the faces :
                         val0 +
    \ 1  /               \val1/
   2 \  / 0               \-3/
  ____\/____           ____\/____
      /\                   /\
  .. /  \ val0-1          /  \val0+1
    /    \               /val0\    */

      star_face[0][0] = val01; star_face[0][1] = val013; star_face[0][2] = 0;
  
      for (i = 1, i1 = 0; i <= val02; i++, i1++)
	{
	  star_face[i][0] =  val01; star_face[i][1] =  i1; star_face[i][2] =  i;
	}
      star_face[val01][0] = val01; 
      star_face[val01][1] = val02;
      star_face[val01][2] = val013;
  
      star_face[val0][0] = val013;
      star_face[val0][1] = val02;
      star_face[val0][2] = val0; 
 
  
      for (i = val0+1, i1 = val0; i <= val014; i++, i1++)
	{
	  star_face[i][0] = val013; star_face[i][1] = i1; star_face[i][2] = i;
	}
      
      star_face[val013][0] = val013;
      star_face[val013][1] = val014;
      star_face[val013][2] = 0;
    
    }
  /*  2) copy the normals of the faces and plug them into the 
      equations of the planes  */

  _smpCopyFaceNormalsAreasCompactness(star_plane, star_areas, star_comp, 
				      star0, star1, val0, val013, face_normals, face_area, compactness);

 
 
  return;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpCopyFaceNormalsAreasCompactness                                    |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void _smpCopyFaceNormalsAreasCompactness(Plane *plane, float *s_area, float *s_comp, 
					 int *star0, int *star1, int val0, int val013, 
					 Vertex *t_normal, float *t_area, float *t_comp)
{
  register int i, i1;
  int val01 = val0-1;

  /*  copy the normals of the faces and plug them into the 
     equations of the planes  */
   
  for (i=0, i1=1; i< val01; i++, i1++)
    {
      memcpy(plane[i], t_normal[star0[i1]], sizeof(Vertex));
      s_area[i] = t_area[star0[i1]];
      s_comp[i] = t_comp[star0[i1]];
    }

  memcpy(plane[val01], t_normal[star0[0]], sizeof(Vertex));
  s_area[val01] = t_area[star0[0]];
  s_comp[val01] = t_comp[star0[0]];

  for (i=val0, i1=2; i<= val013; i++, i1++)
    {
      memcpy(plane[i], t_normal[star1[i1]], sizeof(Vertex));
      s_area[i] = t_area[star1[i1]];
      s_comp[i] = t_comp[star1[i1]];
    }

  return;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpMinFloatArray                                                      |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

float _smpMinFloatArray(float *array, int n)
{
  float min = array[0];
 
  register int i;
 
  for(i=1;i<n;i++) 
    if (array[i] < min)
      min = array[i];
  
  return min;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpCollapseGeometricallyFeasible                                      |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpCollapseGeometricallyFeasible(Vertex *svert, Face *sface, 
				      Plane *splane, float *old_comp, float new_comp_min,
				      Vertex *snor0, Vertex *snor1, U_short val0, U_short sval, 
				      float min_scalprod, float compactness_ratio)
{
  int feasible = 1;

  // check whether any triangle normal might have switched its orientation 

  if (_smpNoFlipOverCheck(1,   val0-1, snor0, splane, min_scalprod) == 1 &&
      _smpNoFlipOverCheck(val0,  sval, snor1, splane, min_scalprod) == 1)
    {
      float old_comp_min = _smpMinFloatArray(old_comp,sval);
     
      feasible = (new_comp_min  > (compactness_ratio * old_comp_min));
    }
  else   feasible = 0;
  

  return feasible;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpNoFlipOverCheck                                                    |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpNoFlipOverCheck(U_short first_face, U_short last_face, 
			Vertex *nor, Plane *plane, float min_scalprod_nor)
{
  
  register U_short i1=0, i=first_face;

  int orientation_consistent;

  do
    {
      orientation_consistent = 
	(SCALPROD3(nor[i1],plane[i]) > min_scalprod_nor);

      i++;
      i1++;
    }
  while ((orientation_consistent ) && (i < last_face));

  return orientation_consistent;

}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpSimplifiedStarNormals                                              |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void _smpSimplifiedStarNormals(Face *sface, Vertex *svert, Vertex *snor, 
			       float *sarea, float *scomp, Vertex simpvert, 
			       U_short first_face, U_short last_face)

/*...........................................................................*/
/*
   compute the face normals in the simplified star as well as the compactness
   and area of the faces.
*/
/*...........................................................................*/
{
  register U_short i,i1;
  register U_char j,k;

  static Vertex triVert[3];

  /* introduce the simplified vertex */
  
  memcpy(triVert[0], simpvert, sizeof (Vertex));
      
  for(i=first_face, i1=0; i < last_face; i++, i1++)
    {
      double sum_lengths_sq = 0.0;

      /* get the second and third vertices of the face.. */
      U_short v1  = sface[i][1],
              v2  = sface[i][2];

      /* ..and copy them at the 2nd and 3rd positions of faceVert */

      memcpy(triVert[1], svert[v1], sizeof (Vertex));
   
      memcpy(triVert[2], svert[v2], sizeof (Vertex));
   
      /* triangle normal and area */

      sarea[i1] = _smpTriangleNormal(triVert, snor[i1]) / 2.;


      /* compute the squared lenghts of the three edges: */
  
       for(j=0;j<3;j++)
	 {
	   double len_edg_sq = 0, tmp;

	   for(k=0;k<3;k++)
	     {
	       tmp         = triVert[(j+1)%3][k] - triVert[j][k];
	       len_edg_sq += tmp * tmp;
	     }

	   sum_lengths_sq += len_edg_sq;

	 }

       scomp[i1] = FOUR_SQRT_3 * sarea[i1] / sum_lengths_sq;

    }

  return;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpReplaceFaceNormals                                                 |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void _smpReplaceFaceNormals(int *star, Vertex *new_nor, float *new_area, float *new_comp, Vertex *nor, 
			    float *area, float *comp, U_short val)
{

/*...........................................................................*/
/*
  
   NOTE the count of the star goes from 0 to val-1
        the corresponding new normals go from 2 to val-1 
	(triangle 0 and 1 are omitted since they will be removed)
*/
/*...........................................................................*/
  
  register U_short i,i1;
 
  for(i=2, i1=0; i<val; i++, i1++)
    {
      memcpy(nor[star[i]], new_nor[i1], sizeof(Vertex));
      area[star[i]] = new_area[i1];
      comp[star[i]] = new_comp[i1];
    }

  return;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpCollapseEdge                                                       |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpCollapseEdge(int v0, int v1, int *star0, int *star1, 
			       U_short val0, U_short val1, Vertex *nor0, Vertex *nor1, 
			       float *area0, float *area1, float *comp0, float *comp1)

/*...........................................................................*/
/*
    
    Perform the simplification of an interior edge, and update the edge heap, by removing
    all edges of the star and adding back the new ones, keyed with the
    new distances.
 
 father(*0[1])  /\
             /\/  \
              /\   \
             / *0[1]\
            /________\       
         v0 \        /v1
             \ *O[0]/
              \   \/
               \  /\/
                \/  father(*0[1])
  

                           /                             \
       The father of both /                      is       \estar1[val1-1]
                         /estar0[1]                        \
                        /                                   \
                       /________ estar0[0]                   \



       The father  of     /     is  \
                         /           \
                        /estar1[1]    \estar0[val0-1]
                       /               \

    NOTE:
    In the *0 or star0 all triangles are parents of themselves 
    (root of the disjoint-set forest)

    in the surface, all vertices, triangles and edges are their original.
    We have to extract the parent each time.
*/
/*...........................................................................*/
{
  int 
    *estar0 = star0  + val0, *vstar0 = estar0 + val0,
    *estar1 = star1  + val1, *vstar1 = estar1 + val1,
    n_edges_added[2],
    n_edges_reinstated = 0;

  /* change the red_valence counting the number of remaining faces incident to a
     given vertex */
  
  _smpNoInteriorVerticesInRemovedFaceSetUpdateRedValence(estar0[0]);
  /* update the component information for the removed face set */
  _smpRemovedFaceForest(estar0[0], 0, 1);

/* 1) perform the topological transformations */

  triangle_father[star1[0]] = star1[val1-1];
  triangle_father[star1[1]] = star1[2];

  vertex_father[v1]         = v0; 

  /* the remaining edges must be valid, and the way we choose to pick
     the triangle fathers dictates the way we choose edge father */
    
  edge_father[estar0[0]] = 
    edge_father[ estar1[val1-1] ] = estar0[1];

  edge_father[estar1[1]] = estar0[val0-1];

  /* 2) change the valences of the vertices */

  valence[v0]            = val0 + val1 - 4;

  /* v0 is not the only vertex whose valence changes:
     the valence of vstar0[1] and vstar1[1] decreases by one 

       /\vstar0[1]            |
      /  \                    |
     /____\           --->    |
     \    /                   |
      \  /                    |
       \/vstar1[1]            | */

  valence[vstar0[1]] --;
  valence[vstar1[1]] --;
  

  /* 3 remove the old edge lengths from the heap and put back the new ones */

  /* 3.1 remove all edges in the star of v0 but v0 v1, because it was
   already removed from the heap (current edge under simplification) */

  _smpRemoveEdgesFromHeap(estar0+1, val0-1);

  /* 3.2 remove all edges but v0-v1 in the star of v1 */

  _smpRemoveEdgesFromHeap(estar1+1, val1-1);

  /* 4. mark the three edges that are to be discarded from the 
     structure as collapsed with a -4 index */

  /* update the number of edges left to be tested:
     the three collapsed edges wont be tested any more 
     one among them was removed already : estar0[1]
     for the last two, we want to verify that they were not
     already counted as outside the heap */

  if (edge2index[estar1[val1-1]] != heap_outside_index)
    num_edg_remaining_4_testing --;
  if (edge2index[estar1[1]] != heap_outside_index)
    num_edg_remaining_4_testing --;

  edge2index[estar1[val1-1]] = edge2index[estar1[1]] = -4;

  edge2index[estar0[0]] = -7;

  /* 5.3 reinstate in the heap all the edges but 0  in estar0
     that are their own parent, that is, all the edges that are a root
     of the disjoint set forest) */

  n_edges_reinstated  = _smpReinstateEdgesInHeap(estar0+1, val0-1, n_edges_added);
  
  /* 2.3 reinstate in the heap all edges but 0 and 1 and val1-1 in estar1 */

  n_edges_reinstated += _smpReinstateEdgesInHeap(estar1+2, val1-3, n_edges_added+1);
 
  /* update the number of edges left to be tested */

  num_edg_remaining_4_testing += n_edges_reinstated;

  /* 3. copy the simplified vertex in the right place */

  memcpy(vert[v0], simplified_vertex, sizeof(Vertex));

  /* 4. replace the face normals, areas and compactness for the simplified stars with the new ones */

  _smpReplaceFaceNormals(star0, nor0, area0, comp0, face_normals, face_area, compactness, val0);
  _smpReplaceFaceNormals(star1, nor1,  area1, comp1, face_normals, face_area, compactness, val1);

  /* 5. update removed texture area (when texture coordinates per corner are used) */

  if (texCoordBinding == PB_PER_CORNER) _collapsed_texture_area += _inc_collapsed_texture_area;
 
  return n_edges_added[0] + n_edges_added[1];
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpSimplifyManifoldSurface                                            |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpSimplifyManifoldSurface()
{
  int 
    e,
    num_edg_tested                 = 0,                   /* number of edges extracted from the heap */
    num_edg_added_to_heap          = 0,
    num_edg_adjacent_2_boundary    = 0;

  /* allocate buffers for storing various information on the vertex and edge stars */

  /* since the valence of the simplified vertex can be as low as  v0 + v1 - 4 (for an interior vertex)
     and is less than SIMPLIFY_VALENCE_MAX, assuming that v0 and v1 >= 1 we
     can be sure that the valence of v0 and v1 that are accepted for processing is <= SIMPLIFY_VALENCE_MAX+3.
     conservatively, we use SIMPLIFY_VALENCE_MAX+4 as a bound */

  int 
    vertex_star_buffer_size = 6 * SIMPLIFY_VALENCE_MAX4 * sizeof (int),
    edge_star_buffer_flsize = 9 * SIMPLIFY_VALENCE_MAX4 * sizeof (float),
    edge_star_buffer_vxsize = 4 * SIMPLIFY_VALENCE_MAX4 * sizeof (Vertex),
    edge_star_buffer_plsize = 4 * SIMPLIFY_VALENCE_MAX4 * sizeof (Plane),
    edge_star_buffer_fasize = 2 * SIMPLIFY_VALENCE_MAX4 * sizeof (Face);

  void
    *edge_star_buffer_pl = NULL,
    *edge_star_buffer_fa = NULL;


  _vertex_star_buffer = NULL;
  _edge_star_buffer_fl = NULL,
  _edge_star_buffer_vx = NULL,
    
  _vertex_star_buffer = (int *)    _smpAllocateZero(vertex_star_buffer_size);

  if (!_vertex_star_buffer) goto error;

  _edge_star_buffer_fl = (void*) _smpAllocateZero(edge_star_buffer_flsize);

  if (!_edge_star_buffer_fl) goto error;

  _edge_star_buffer_vx = (void*) _smpAllocateZero(edge_star_buffer_vxsize);

  if (!_edge_star_buffer_vx) goto error;

  edge_star_buffer_pl = (void*) _smpAllocateZero(edge_star_buffer_plsize);

  if (!edge_star_buffer_pl) goto error;

  edge_star_buffer_fa = (void*) _smpAllocateZero(edge_star_buffer_fasize);

  if (!edge_star_buffer_fa) goto error;

  /* 1) mark the edges touching the surface boundary as not simplifiable: */

  if (!simplify_boundary)
   
    num_edg_adjacent_2_boundary = _smpMarkEdgesAdjacent2Boundary();

  /* 2) fill a Heap with edges, keyed with the weight of the edge */

  num_edg_added_to_heap =  _smpBuildEdgeHeap();

  num_edg_weights += num_edg_added_to_heap;

  do
    {
      e = _smpRemoveLowestWeightEdgeFromHeap();

      if (e != -1) 
	{	
		  
	  /* variable set to 1 if the edge is effectively collapsed */
		  
	  int 
	    collapsible = 0;
	  EdgeS 
	    *edg        = edge_array+e;
	  int 
	    *endpoints  = SurfEdgeGetVertices(edg),
	    v0          = FATHER(endpoints[0], vertex_father),
	    v1          = FATHER(endpoints[1], vertex_father);
	  U_short 
	    val0        = valence[v0],
	    val1        = valence[v1];

	  num_edg_tested ++;
	  /* mark the edge as removed from the heap: give an index such that HeapDelete would fail + 1*/
		      
	  edge2index[e] = heap_outside_index;
		      
	  /* decrement the number of edges remaining for testing */
			
	  num_edg_remaining_4_testing --;

	  /* treat a boundary vertex of the second kind */

	  if (simplify_boundary && boundary_vert[v0] && boundary_vert[v1])
		    
	    collapsible =
	      _smpCollapsibilityTestsBoundaryEdge2ndKind(e, v0, v1, val0, val1, move_vertices);

	  /* treat a boundary vertex of the fisrt kind */

	  else if (simplify_boundary && (boundary_vert[v0] || boundary_vert[v1]))
		    
	    collapsible =
	      _smpCollapsibilityTestsBoundaryEdge1stKind(e, v0, v1, val0, val1);

	  /* treat an interior edge */
	  else
	    {
	      /* perform the edge valence test for an interior edge */
		      
	      if ( (val0 + val1 - 4 >= 3) && (val0 + val1 - 4 < valence_max))
		{
		    
		  /* BUILD THE 2 VERTEX STARS */
		      
		  int
		    *star0  = (int *) _vertex_star_buffer,
		    *estar0 = star0  + val0, *vstar0 = estar0 + val0,
		    *star1  = vstar0 + val0, *estar1 = star1  + val1,
		    *vstar1 = estar1 + val1;
		  
		  bzero(star0, 3 * (val0 + val1) * sizeof (int));
			 
		  /* 3) construct the edge star */
		      
		  _smpBuildParentEdgeStars(edg,v0,v1,val0,val1,star0,star1);
		      
		  /* 3.1) test whether the simplification is topologically feasible */
		      
		  if (_smpCollapseTopologicallyFeasible(vstar0, vstar1, val0, val1) &&
		      _smpManifoldRemovedFaceSet(e) &&
		      _smpRemovedFaceForest(e, 1, 0))
		    {
		      /* 3.2) compute a simplified vertex, and test whether the simplification is
			 geometrically possible, if not, go to END */
			  
		      /* the number of vertices is given by val0 + val1 - 2 and the number of
			 triangles is also val0 + val1 - 2 */
			  
		      U_short  
			sval  = val0 + val1 -2, 
			sval1 = sval + 1,
			val02 = val0-2, val12 = val1-2;
			
		      /* BUILD THE EDGE STAR */
			  
		      Vertex 
			*svert       = (Vertex *) _edge_star_buffer_vx,
			*snor0       = svert + sval1,
			*snor1       = snor0 + val02;
		      Face 
			*sface       = (Face  *) edge_star_buffer_fa;
		      Plane 
			*splane      = (Plane *) edge_star_buffer_pl;
		      float 
			*scomp0      = (float *) _edge_star_buffer_fl,
			*sarea0      = scomp0 + val02,
			*scomp1      = sarea0 + val02,
			*sarea1      = scomp1 + val12,
			*scomp       = sarea1 + val12,
			*sarea       = scomp  + sval,
			*wspace      = sarea  + sval,
			min_comp_after;

		      bzero((char *)svert,  edge_star_buffer_vxsize);
		      bzero((char *)sface,  edge_star_buffer_fasize);
		      bzero((char *)splane, edge_star_buffer_plsize);
		      bzero((char *)scomp0, edge_star_buffer_flsize);
			      
		      _smpBuildEdgeStar(svert, sface, splane, sarea, scomp,
					v0, v1, val0, val1, star0, star1);
			  
		      _smpSimplifiedVertexLocation(v0, v1, val0, val1,
						   sval, svert, sface, splane, sarea, wspace,
						   move_vertices);
			    
		     
		      if (// colors per faces: are the disappearing colors represented in the neighborhood?
			  (colorBinding    != PB_PER_FACE   || _smpColorPerFaceOk(star0[0], star0[1], val0-2, star0+2,
										  val1-2, star1+2)) && 
			  // texCoords per corner: how large is the texture area that will disappear ?
			  // (globally & incrementally)
			  (texCoordBinding != PB_PER_CORNER || _smpTexCoordPerCornerOk(star0[0], star0[1], val0-2, star0+2,
										       val1-2, star1+2)))
			{
			/* if we cannot move the simplified vertex, we try successively 
			   the two positions of v0 and v1 */
			int the_position = 0, num_positions = 1 + !(move_vertices);
			      
			while (the_position < num_positions && (!collapsible))
			  {
			    
			    /* to determine whether the collapse is geometrically
			       possible, the new vertex normals are needed */
			  
			    _smpSimplifiedStarNormals(sface, svert, snor0, sarea0, scomp0, 
						      simplified_vertex, 1, val0-1);
			    _smpSimplifiedStarNormals(sface, svert, snor1, sarea1, scomp1, 
						      simplified_vertex, val0, sval);
			  
			    /* minimum triangle compactness after simplification :*/
			    min_comp_after = MIN( _smpMinFloatArray( scomp0, val02 ), 
						  _smpMinFloatArray( scomp1, val12 ) );
			  
			    if (_smpCollapseGeometricallyFeasible(svert, sface, splane, scomp, 
								  min_comp_after, snor0, snor1, val0, sval,
								  min_scalprod_nor, compactness_ratio))

			      {
				/* 3.3) if collapsible, test whether the new errors are within the
				   tolerance limits  */
			
				collapsible = _smpErrorWithinTolerance(star0, star1, val0, val1);
			      
				if (collapsible)
				
				  {
				  
				    /* 3.4) simplify (collapse the edge ) and replace old edges in
				       the heap with new ones */
				  
				    num_edg_collapsed += 3;
				  
				    num_edg_weights += 
				      _smpCollapseEdge( v0, v1, star0, star1,
							val0, val1, snor0, snor1, sarea0, sarea1,
							scomp0, scomp1);
				  }
				else
				  {
				    /* rejected for tolerance off limits the edge can be
				       reinserted afterwards if its length changes*/
				      
				    edg_rejected_4_tolerance ++;
				  }
				
			      }
			    else
			      {/* geometrically infeasible */

				edg_rejected_4_geometry ++;
			      }
			
			    the_position++;
			    /* now use the second (alternate) position */
			    memcpy(simplified_vertex, simplified_vertex_alt, sizeof (Vertex));
			  } /*end while: if ! move vertices, try two positions */

		      } else edg_rejected_4_geometry ++;
		    }
		  else 
		    { /* topologically infeasible */

		      edge2index[e] = -3; 
			
		      /* this condition will not be modified even if the surface is simplified around 
			 the edge */

		      edg_rejected_4_topology ++;  
		    } 
		      
			
		}
	      else 
		{

		  /* If the valence is too high or too low, reject also for topology, However, this condition
		     might change */

		  edg_rejected_4_topology ++; 
		}
	    } /* end treat interior vertex */

	} /* end if (e != -1) */
	
    }    while(e != -1);


  free((void*)_vertex_star_buffer);

  free(_edge_star_buffer_fl);

  free(_edge_star_buffer_vx);

  free(edge_star_buffer_pl);

  free(edge_star_buffer_fa);


  return 1;

error:

  if (_vertex_star_buffer)    free((void*)_vertex_star_buffer);

  if (_edge_star_buffer_fl)   free(_edge_star_buffer_fl);

  if (_edge_star_buffer_vx)   free(_edge_star_buffer_vx);

  if (edge_star_buffer_pl)    free(edge_star_buffer_pl);

  if (edge_star_buffer_fa)    free(edge_star_buffer_fa);

  return 0;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpAllocateZero                                                       |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void* _smpAllocateZero(int size)
{
  void* buf = (void*) malloc (size);

  if (buf)
    {
      bzero (buf, size);
    }
  return buf;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpFloatDataBoundingBoxDiagonal                                       |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

float _smpFloatDataBoundingBoxDiagonal(float *data, int n_data, int 
				       data_dim)
{
  float diagonal = 0.0;
  
  float
    *cur_data = data,
    *bbox     = NULL;
  
  register int i,j;
  
  bbox = (float *) _smpAllocateZero (2 * data_dim * sizeof (float));
  
  if (!bbox) goto error;
  
  /* initialize the bounding box to the first array values */
  
  /* the bbox is organized as follows
     
     bbox[0]               MININUM
     bbox[1]                DATA
     VALUES
     FOR
     EACH
     bbox[data_dim-1]       COMPONENT
     
     
     bbox[data_dim]        MAXIMUM
     bbox[data_dim+1]       DATA
     VALUES
     FOR
     EACH
     bbox[2*data_dim-1]     COMPONENT
     */
  
  for(j=0;j<data_dim;j++)
    {
      bbox[j] = bbox[j+data_dim] = data[j];
    }
  
  /* update the bounding box if data values are smaller than the minimum or 
     larger than the maximum */
  
  cur_data = data + data_dim;
  
  for (i=1;i<n_data;i++,cur_data+=data_dim)
    
    for(j=0;j<data_dim;j++)
      {
	if (cur_data[j] < bbox[j]) bbox[j] = cur_data[j];
	
	else if (cur_data[j] > bbox[j+data_dim]) bbox[j+data_dim] = cur_data[j];
      }
  
  /* once the bounding box is complete, compute its diagonal. If the 
     dimension of the data is one, we just compute max-min and thus avoid the 
     computation of squares and of a square root */
  
  if (data_dim == 1) diagonal = bbox[1] - bbox[0];
  
  else
    {
      double
	one_coordinate,
	double_diagonal = 0.0;
      
      for (j=0;j<data_dim;j++)
	{
	  one_coordinate   = bbox[j+data_dim] - bbox[j];
	  double_diagonal += one_coordinate * one_coordinate;
	}
      
      if (double_diagonal>0.0) double_diagonal = sqrt(double_diagonal);
      
      diagonal = (float) double_diagonal;
    }
  
  free((void*)bbox);
  return diagonal;
  
error:
  
  free((void*)bbox);
  
  return 0.0;
}


/* tests for the purpose of determining whether the set of removed faces
   has no interior vertices and is a manifold */


/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpNoInteriorVerticesInRemovedFaceSetUpdateRedValence                 |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpNoInteriorVerticesInRemovedFaceSetUpdateRedValence(int the_collapsed_edge)
{
  /* the red valence changes: this is the valence of the original
     vertices where the collapsed faces are counted as missing faces:
     We are interested in the original elements here, not in their
     fathers so we take the triangles that actually get removed: */

    int 
      the_face_num,
      *the_face,
      *adjacent_faces = (int *)SurfEdgeGetTriangles(edge_array+the_collapsed_edge);

    register U_char j,k;

    /* for each face, for each corner */

    for (k=0;k<2;k++) 
      {
      	
	if (adjacent_faces[k] != -1) /*applies to valid faces only */
	  
	  {
	    the_face_num = FATHER(adjacent_faces[k],triangle_father);
	    
	    the_face     = triangles + 3 * the_face_num;
	    
	    for(j=0;j<3;j++)
	      
	      {		
		red_valence[the_face[j]] -= 1;
				
	      }	  
	  }
      
      }
    
    return 1;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpManifoldRemovedFaceSet                                             |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpManifoldRemovedFaceSet(int the_collapsed_edge)
{
  /* this test verifies that the set of removed faces is a manifold */

  /* if the removed faces were adjacent, this is what we would do:

         v2
	     /\adjface[0]
	  v0/__\v1
	    \  /
	     \/adjface[1]
         v3

     for each vertex v0, v1, v2, v3, rotate about the vertex
	 and count the number of "color" changes, from red 
	 faces (remaining) to blue faces (removed)


	the trouble is that the representatives of such faces,
	which are the faces that eventually become blue,
	are not necessarily adjacent
  */

  register U_char
    nvd = 0, /* number of different vertices among the six vertices */
    j,k;

  int 
    v_counts[6],
    v[6],
    val[6],
    old_father[2],
    *adjfaces   = SurfEdgeGetTriangles(edge_array + the_collapsed_edge),

    valid_adjfaces[2],
    num_valid_adjfaces = 0,
    not_an_interior_vertex = 1,
    is_manifold = 1;

  /* we temporarily mark the two faces adjacent to the edge as beeing
     updated by the simplification: in fact, what will happen is that
     the fathers of the adjacent faces will be removed these are the
     faces that we should check for "manifoldness". They can be
     disconnected. hence there can be at most six different vertices
     to check

           /\v2
          /__\    old_father[0]
	 v0  v1

           /\v5
          /__\    old_father[1]
	 v3  v4

	*/

  /* permute the adjfaces such that the first adjface is
     always valid

     if adjfaces[0] is not valid adjfaces[1] must be valid
     */

  if (adjfaces[0] == -1) {
	valid_adjfaces[0]  = adjfaces[1];
	num_valid_adjfaces = 1;
  }
 
  else if (adjfaces[1] == -1) {
	valid_adjfaces[0]  = adjfaces[0];
	num_valid_adjfaces = 1;
  }

  else {
	valid_adjfaces[0]  = adjfaces[0];
	valid_adjfaces[1]  = adjfaces[1];
	num_valid_adjfaces = 2;
  }


  for(j=0;j<num_valid_adjfaces;j++)
    
    /* note that adjface can be -1 for either face, if the
       collapsed edge is a boundary edge */
    old_father[j] = FATHER(valid_adjfaces[j], triangle_father);	 
	  
     
  /* update the triangle_father in a second pass, to avoid corrupting
     the entire triangle_father relationship when calling FATHER() subsequently */

 for(j=0;j<num_valid_adjfaces;j++)   
      
   triangle_father[old_father[j]] = -1;
	 
    
  /* collect v0,v1,v2,v3,v4,v5 */

 for(j=0;j<3;j++)
   
   {
     v[j]   = triangles[3*old_father[0]+j]; v_counts[j] = 1; 
     val[j] = red_valence[v[j]];
   }
 nvd = 3;
 
 /* in v_counts, we count the number of times that a particular vertex
    is shared among the removed faces (can be either one or two) */


 if (num_valid_adjfaces == 2)
   {
  
     for(j=0;j<3;j++)
       {	  
	 int not_a_duplicate_vertex = 1;

	 v[nvd] = triangles[3*old_father[1]+j];
  
	 k = 0;

	 while (not_a_duplicate_vertex && k < nvd)
	   {
	     if (v[nvd] == v[k]) 
	       {					
		 v_counts[k] ++;
		 not_a_duplicate_vertex = 0;
	       }
	     k++;
	  }
	 if (not_a_duplicate_vertex) 
	   {
	     v_counts[nvd] = 1; 
	     val[nvd] = red_valence[v[nvd]];   
	     nvd++;
	   }
       }
   }

 j = 0;

 while (j<3 &&  (is_manifold = 
	  
		 _smpManifoldVertexRemovedSet(v[j], triangles, old_father[0]))) j++;
  
 /* adjfaces[1] must be used as a first triangle for v3,v4 and v5 */

 if (num_valid_adjfaces == 2 && (is_manifold))
   {
     while (j<nvd &&  (is_manifold = 	  
		       _smpManifoldVertexRemovedSet(v[j], triangles, old_father[1]))) j++;
   }

 /* we restore the original triangle fathers */
 
 for(j=0;j<num_valid_adjfaces ;j++)
   triangle_father[old_father[j]] = old_father[j];

  


  /* verify that the set of removed faces has no interior vertices */
  {
	  
    /* for each of the nvd vertices, verify that the blue valence (the
       number of removed faces sharing that vertex) is strictly larger
       than the v_counts we do not explicitely compute the blue
       valence, but instead take the difference between the original
       valence stored in "ori_valence" and red_valence */
	
  
    k = 0;  
    while (not_an_interior_vertex && k < nvd)
      {
	not_an_interior_vertex = (v_counts[k] < val[k]);
	k ++;
      }
  }


  return (is_manifold && not_an_interior_vertex);
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   RedTriangle                                                            |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

#define RedTriangle(t) (triangle_father[(t)] == (t))

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   TrianglesSameColor                                                     |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

#define TrianglesSameColor(t1,t2) \
	((RedTriangle(t1)) == (RedTriangle(t2))) /* red == red || blue == blue */

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpManifoldVertexRemovedSet                                           |
  |                                                                          |
  +--------------------------------------------------------------------------+*/


int SimpData::_smpManifoldVertexRemovedSet(int v, int *triangles, int first_tri)
{
  int 
	  the_tri           = first_tri,
	  next_tri          =        -1,
	  num_color_changes =         0,
	  is_manifold       =         1;

	  /* val               = valence[v];

         the valence cannot be used because it is updated
		 during simplification */
	  

  /* case of an interior vertex */

  if (!(boundary_vert[v]))
    {
      /* rotate about the vertex v and count the face color changes
	 accross edges. Red triangles are remaining triangles Blue
	 triangles are removed triangles A maximum number of two color
	 changes can be observed otherwise the set of blue faces is
	 not manifold
	  
	 when next_tri == first_tri, we know that the loop was completed
	 (we cannot use the valence to count the number of triangles) */

      while (num_color_changes <=2 && next_tri != first_tri &&
	     ((next_tri = _smpNextTriangle(triangles + 3 * the_tri, e_table, v, 1)) != -1)
	     /* Surface NextCCWTriangle */
	     )
	{
	  num_color_changes += !(TrianglesSameColor(the_tri, next_tri));

	  the_tri  = next_tri;
	}

      is_manifold = (num_color_changes <=2);
    }

  /* case of a boundary vertex */

  else 
    {
      /* rotate about v in clockwise direction until the
	 boundary is found or the valence is reached */

      /* then rotate in counterclockwise direction until
	 either the boundary is found or the valence is reached */
    
      while (num_color_changes <=1 &&
	     ((next_tri = _smpNextTriangle(triangles + 3 * the_tri, e_table, v, 0)) != -1)    
	     )
	{
	  num_color_changes += !(TrianglesSameColor(the_tri, next_tri));

	  the_tri = next_tri;
	}
      /* when precisely next_tri -1, if the_tri is a blue triangle,
	 there is a color change there (remember that we want the set
	 of blue triangles to be a manifold) */

      /* however, for a boundary vertex, we allow a single change of
	 color only in the interior, not counting the changes of color
	 on the boundary */

      the_tri           = first_tri;
      next_tri          = -1;

      /* 
	 example:
	         4
	        _/_
               1/\3
                 2 
		 first_tri =2: test one edge to the left and 2 edges to
		 the right 

	 first_tri = 1: test 0 edges to the left(exept the boundary)
         + 3 edges to the right
         however the two boundary tests (5 tests in total)
		 are important, and in order to stop when next_tri == -1
		 we need to have a test i<val instead of i < val-1
		 */

      while (num_color_changes <=1 &&
	     ((next_tri = _smpNextTriangle(triangles + 3 * the_tri, e_table, v, 1)) != -1)
	     )
	{
	  num_color_changes += !(TrianglesSameColor(the_tri, next_tri));

	  the_tri = next_tri;
	}

      /* if the last triangle is blue, we count a color change at the 
	 boundary */

      /*if (num_color_changes <=2 && next_tri == -1)
	num_color_changes += !(RedTriangle(the_tri));*/

      is_manifold = (num_color_changes <=1);
    }

  return is_manifold;
}

#undef RedTriangle

#undef TrianglesSameColor


/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpRemovedFaceForest                                                  |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpRemovedFaceForest(int the_collapsed_edge, int verify_forest, int update_representative)
{
  /* this test verifies that the set of removed faces is a forest
     (meaning, that there we do not connect the same face component
     again when adding the faces corresponding to "the_collapsed_edge"
     to the removed face set */

  register U_char j,k;

  int 
    old_father[2],
    *adjfaces                   = (int *) SurfEdgeGetTriangles(edge_array + the_collapsed_edge),

    valid_adjfaces[2],
    face_neighbors[6],
    neighbor_representatives[6],
    num_valid_representatives   = 0,
    num_valid_adjfaces          = 0,
    is_forest                   = 1,
    removed_faces_are_adjacent  = 0;



  /* permute the adjfaces such that the first adjface is
	always valid

	if adjfaces[0] is not valid adjfaces[1] must be valid
	*/

  if (adjfaces[0] == -1) {
	valid_adjfaces[0]  = adjfaces[1];
	num_valid_adjfaces = 1;

  }
 
  else if (adjfaces[1] == -1) {
	valid_adjfaces[0]  = adjfaces[0];
	num_valid_adjfaces = 1;
  }

  else {
	valid_adjfaces[0]  = adjfaces[0];
	valid_adjfaces[1]  = adjfaces[1];
	num_valid_adjfaces = 2;
  }


  for(j=0;j<num_valid_adjfaces;j++)
    
    old_father[j] = FATHER(valid_adjfaces[j], triangle_father);	 
	  
   if ( num_valid_adjfaces == 1) old_father[1] = -1;

  /* now we now which two faces will be added to the removed face set.
     we need to collect the three neighbors of such faces */

   for(j=0;j<3;j++)
    
     {
       /*     v2
       n[0]   /\ n[2]
             /__\    old_father[0]
	    v0  v1
              n[1]
           /\
          /__\    old_father[1]
	   
	
	*/

       face_neighbors[j] = _smpNextTriangle(triangles + 3 * old_father[0],
					    e_table, triangles[3*old_father[0]+j], 1);
       
     }
   

  if (num_valid_adjfaces == 2)
   
      for(j=0;j<3;j++)
	{
	  face_neighbors[j+3] = _smpNextTriangle(triangles + 3 * old_father[1],
						 e_table, triangles[3*old_father[1]+j], 1);
	  
	  if (face_neighbors[j+3] == old_father[0]) removed_faces_are_adjacent = 1;
	}

  /* for each neighbor of old_father[0], 
     we retrieve the removed component number of the neighbor.
     
     we do the same for the old_father[1] 


     even if  the "old_father" faces are not neighbors,
     the neighbors and the old_father form a connected set of
     removed triangles
     
  
         __  __o_f[1]
       /\x /\  /
      /__\/_x\/
     o_f[0]

    the neighbors that correspond to the other
     old_father should not be counted. But they aren't, since such faces
     are not marked as being removed yet.
     */

  num_valid_representatives = 0;
    
  {
    int *neighbors = face_neighbors;

    for(k=0; k< num_valid_adjfaces;k++, neighbors +=3)
      {
	
	/* replace the neighbors with their component representative */

	for (j=0;j<3;j++)

	  if (neighbors[j] != -1) /* if this neighbor is defined */
	  
	    if (neighbors[j] != FATHER(neighbors[j], triangle_father)) /* if it was removed */
	      
	    /* we don't want to count the same neighbor twice so,
	       if k = 1, this neighbor should be different from
	       the three neighbors of the first valid_adjface */
		
	    if (k == 0 || ((neighbors[j] != face_neighbors[0]) && (neighbors[j] != face_neighbors[1]) &&
		(neighbors[j] != face_neighbors[2])))
	      {
		neighbor_representatives[num_valid_representatives] = 
		  FATHER(neighbors[j],removed_component_father);

		num_valid_representatives ++;
	      }
	      
		
      }
   
    /* among all the representatives found, they all must be different */

    if (verify_forest)

      /* if removed faces are not adjacent this test is not necessary */
      if (removed_faces_are_adjacent)
      
	is_forest = _smpAreValuesAllDifferent(num_valid_representatives, (char *)neighbor_representatives,
					      sizeof (int));
	   
    if (update_representative)

      /* set the representatives of the num_valid_representatives - 1
	 last neighbors as the first representative so that they all
	 point to the same thing */
      {
	for (j=0;j<num_valid_representatives;j++)
	  removed_component_father[neighbor_representatives[j]] = old_father[0];
	
	/* also that the representatives of the faces we are removing */
	for (j=1;j< num_valid_adjfaces;j++)
	  removed_component_father[ old_father[j]] = old_father[0];
      }

  } /* num_valid_representatives = 0 */

  return is_forest;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpAreValuesAllDifferent                                              |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpAreValuesAllDifferent(int n_elements, char *array_of_elements, int element_size)
{
  int are_different = 1;

  register int i, j;

  char *el_i, *el_j;

  i = 0;

  el_i = array_of_elements;

  while (are_different && i<n_elements)
    {
      j = i+1;

      el_j = el_i + element_size;

      while (are_different && j<n_elements)
	{
	  are_different = (memcmp(el_i, el_j, element_size) != 0);
	   
	  j++;
	  el_j += element_size;
	}
      i++;
      el_i += element_size;
    }

  return are_different;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpNextTriangle                                                       |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int _smpNextTriangle(int *triangle, Htable *e_table, int theVertex, U_char ccw)
{
  int nextTri = -1;
  U_char theCorner = 0;


  while (theCorner < 3 && triangle[theCorner] != theVertex) theCorner++;

  if (theCorner < 3)
    {
      int nextVert = (ccw) ? triangle[(theCorner+2)%3] : triangle[(theCorner+1)%3];

      EdgeS the_edge[1], *edge_in_table;

      if (theVertex > nextVert)
	{
	  the_edge->v[0] = theVertex;  the_edge->v[1] = nextVert;
	}
      else
	{
	  the_edge->v[0] = nextVert;   the_edge->v[1] = theVertex;
	}
      edge_in_table = (EdgeS *)_smpFindRecord(e_table,(KEYTYPE)the_edge);
      
      if (edge_in_table)
	{
	  int 
	    *incident_tri  = SurfEdgeGetTriangles(edge_in_table);

	  if (ccw)
	    {
	      if (theVertex > nextVert) nextTri = incident_tri[0];
	      else                      nextTri = incident_tri[1];
	    }
	  else
	    {
	      if (theVertex > nextVert) nextTri = incident_tri[1];
	      else                      nextTri = incident_tri[0];
	    }
	} 
      else
	fprintf(logfp,"\"_smpNextTriangle\"(): edge %d %d not found in edge table\n", theVertex, nextVert);

    }
  else
    fprintf(logfp,"\"_smpNextTriangle\"(): vertex %d not in triangle\n");

  return nextTri;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpRedistributePerFaceProperty                                        |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void _smpRedistributePerFaceProperty(int nD, float *property, int nF, int new_nF, int *f_map, 
				     float *face_areas)
{
  register int i,j;
  
  float 
    *old_property = property,
    *face_prop_averages_current;

  int total_new_num_elements = new_nF * nD;
  
  vector <float> face_prop_weights ((size_t)new_nF,0.);
  vector <float> face_prop_averages((size_t)total_new_num_elements,0.);

  for (i=0; i < nF; i++, old_property += nD)
    {
      int new_face = f_map[i];
      face_prop_weights[new_face] += face_areas[i];

      face_prop_averages_current = face_prop_averages.begin() + new_face * nD;

      for (j=0; j<nD;j++)
	face_prop_averages_current[j] += face_areas[i] * old_property[j];
    }

  // then loop on all the averages, and divide them with the weigths

  for (face_prop_averages_current = face_prop_averages.begin(),
       i=0; i< new_nF; i++, face_prop_averages_current+=nD)
    {
      for (j=0;j<nD;j++)
	face_prop_averages_current[j] /= face_prop_weights[i];
    }

  // finally, copy the new properties in the beginning of the existing array

  memcpy(property, face_prop_averages.begin(), total_new_num_elements * sizeof (float));

  return;
}



/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpRedistributePerCornerProperty                                      |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void _smpRedistributePerCornerProperty(int nD, float *property, int *t, int nT, int *new_t, int new_nT, 
				       int *v_map, int *t_map, float *face_areas)
{
  register int i,j,k;
  
  float 
    *old_property = property,
    *corner_prop_averages_current;

  int new_nC = new_nT*3;

  vector <float> corner_prop_weights ((size_t)new_nC,0.);
  vector <float> corner_prop_averages((size_t)(new_nC*nD),0.);

  for (i=0; i < nT; i++,  t+=3)
    {
      int new_face   = t_map[i];
      
      // for each old corner, get the new vertex
      // and the new face, which gives the new corner
      
      for (j=0;j<3;j++, old_property += nD)
	{
	  int new_vertex = v_map[t[j]];
	 
	  // fetch the new corner inside the new face 

	  int new_corner = new_face*3;

	  for (k=0;k<3;k++, new_corner++)
	    if (new_t[new_corner] == new_vertex) break;

	  
	  corner_prop_weights[new_corner] += face_areas[i];

	  corner_prop_averages_current = corner_prop_averages.begin() + new_corner * nD;
 
	  for (j=0; j<nD;j++)
	    corner_prop_averages_current[j] += face_areas[i] * old_property[j];
	}
    }

  // then loop on all the averages, and divide them with the weigths

  for (corner_prop_averages_current = corner_prop_averages.begin(),
       i=0; i< new_nC; i++, corner_prop_averages_current+=nD)
    {
      for (j=0;j<nD;j++)
	corner_prop_averages_current[j] /= corner_prop_weights[i];
    }

  // finally, copy the new properties in the beginning of the existing array

  memcpy(property, corner_prop_averages.begin(), new_nC*nD * sizeof (float));

  return;
}

#undef  EXTRACT_TRIANGLE_FROM_VERTEX_STAR
#define EXTRACT_TRIANGLE_FROM_VERTEX_STAR( vstar, vertex_array, i) a_triangle; {\
  memcpy(a_triangle[1], (vertex_array)[(vstar)[(i)]],   sizeof(Vertex));\
  memcpy(a_triangle[2], (vertex_array)[(vstar)[(i)+1]], sizeof(Vertex));}

#undef  UPDATE_MAX_CONSTRAINT
#define UPDATE_MAX_CONSTRAINT(potential_err_pivot)                           \
{	                                                                     \
              if ((potential_err_pivot) > largest_err_pivot)                 \
		largest_err_pivot = (potential_err_pivot);                   \
	      edge_collapsible = (largest_err_pivot <= tol_pivot);}

#undef  SUBVEC3
#define SUBVEC3(u,v) {(u)[0] -= (v)[0]; (u)[1] -= (v)[1];  (u)[2] -= (v)[2];}

#undef  SCALEVEC3
#define SCALEVEC3(u,s) {(u)[0] *= (s); (u)[1] *= (s);  (u)[2] *= (s);}



static int _smpImproveSimpVertPosAndErrorBBox(Vertex simpvert,
					      float *largest_error,
					      int  num_tiling_vert,
					      Vertex  *tiling_vert,
					      float  *tiling_error);

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpErrorWithinTolerance                                               |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpErrorWithinTolerance(int *star0, int *star1, int val0, int val1)

{
  /* The purpose of this method is to construct simultaneously a tiling of
     e* (edge undergoing collapsibility tests) denoted til(e*) and a
     tiling of vs* (potential simplified vertex) denoted til(vs*), such
     that there is a one to one mapping between
     til(e*) and til(vs*) 
     

     numbering of the faces : t0,t1,...t_val0+val1-3 triangles of e*
     ~~~~~~~~~~~~~~~~~~~~~~
                             val0 +
        \ 1  /               \val1/
       2 \  / 0               \-3/
      ____\/____           ____\/____
          /\                   /\
      .. /  \ val0-1          /  \val0+1
        /    \               /val0\      

	q1,...,qval0-2,....,qval0,...,q_val0+val1-3 triangles of vs*
	*/

  /* 1) determine the closest point to the vertex vs on the triangles t0
     and t_val0-1 */

  /* 2) determine the closest point to the vertex v0 on the
     triangles q1,...,qval0-2. */

  /* 3) determine the closest point to the vertex v1 on the triangles
     qval0,...,q_val0+val1-3 */

  /* 4) Take in turn each edge (vs,v0),(vs,v1),..., (vs,v_val0-2)
     call the ith edge ei
     determine the closest point on each edge incident to v_val0-1:

     (v_val0-1, v0),..., (v_val0-1, v_val0-2), and (v_val0-1, v_val0+val1-3)
     If the closest point is located in the inside of either edge, 
     determine a crossing point.
     */


  /* 5) Take in turn each edge (vs,v_val0-2),...,(vs,v_val0 + val1-4) and
     (vs,v0),

     call the ith edge ei
     determine the closest point on each edge incident to v_val0+val1-3:

     */


  /* Firstly determine numbers of elements in old and new configurations */
   
  int 
    edge_collapsible = 1,

    /* vertex stars */
    *vstar0 = star0 + 2 * val0, 
    *vstar1 = star1 + 2 * val1,
    i,j, i_min = -1;

  static  Vertex 
    vs[1],             /* local copy of the simplified vertex */
    a_triangle[3],
    wp[2],
    alpha[SIMPLIFY_VALENCE_MAX1],
  /* val0 + val1 -4 <= valence max <=> val0, val1 <= valence_max +1 
     because val0 >=3 and val1 >=3 */

    /* save all vertices of the tiling, properly scaled and represented at the simplified vertex */

    tiling_vert[SIMPLIFY_TILING_VERTICES_MAX],
    tiling_vert_max;

  static float
    dp[2],  /* distance to the projected point + error */
    ep[2],  /* error at the projected point */
    tiling_vert_errors[SIMPLIFY_TILING_VERTICES_MAX],
    tiling_work_space [SIMPLIFY_TILING_VERTICES_MAX],
    tiling_vert_max_error;

  float 
    tol_pivot, largest_err_pivot = 0., d_min,
    lambda1, onelambda1, lambda2, onelambda2, one_over_alpha;

  int 
    num_tiling_vert = 0;

  /* set the maximum tolerance for the star */

  tol_pivot = tol_volume[vstar0[0]];

  for (i=1;i<val0-1;i++) /* "val0-1" instead of "val0" serves to avoid using the last vertex of the star twice 
			    when taking into account both loops */
    if (tol_volume[vstar0[i]] < tol_pivot)
      tol_pivot = tol_volume[vstar0[i]];

  for (i=0;i<val1-1;i++)
    if (tol_volume[vstar1[i]] < tol_pivot)
      tol_pivot = tol_volume[vstar1[i]];

  
  /* verify that the maximum tolerance is not already exceeded at
     the vertices */

  i = 0;
  while (edge_collapsible && i < val0-1) /* otherwise the last vertex is counted twice */
  /* the equality (star_err = tol) is allowed */  
    if (err_volume[vstar0[i++]] > tol_pivot)
      edge_collapsible = 0;

  i = 0; 
  while (edge_collapsible && i < val1-1) /* the last vertex is the second vertex of the previous loop */
    if (err_volume[vstar1[i++]] > tol_pivot)
      edge_collapsible = 0;


  if (edge_collapsible)
    {
      int 
	which_triangle;
      Vertex 
	*qi, *t0, *t1;
 
      /* local copy of the simplified vertex, translated by the due amount */
      memcpy(vs[0], simplified_vertex, sizeof (Vertex));
     
      /* 1) */  

      /* then fabricate the triangle t0 = v1 v0 vval0-1 */
      memcpy(a_triangle[0], vert[vstar0[0]], sizeof (Vertex));
      t0 =  EXTRACT_TRIANGLE_FROM_VERTEX_STAR(vstar1, vert, 0);

      
      /* find the point closest to vs on t0 and record the projection as
	 well as the distance */
       
      _smpClosestPointOnTriangle(t0, vs[0], wp[0], alpha[0], &dp[0]);
      
      ep[0] = 
  
	  alpha[0][0] * err_volume[vstar0[0]] +
	  alpha[0][1] * err_volume[vstar1[0]] +
	  alpha[0][2] * err_volume[vstar1[1]];

      dp[0] = dp[0] + ep[0];

      /*  find the point closest to vs on t1 and record the same info 
	  triangle t1 = (v0, v1, v2) */
      
      memcpy(a_triangle[0], vert[vstar1[0]], sizeof (Vertex));
      t1 =  EXTRACT_TRIANGLE_FROM_VERTEX_STAR(vstar0, vert,0);

      _smpClosestPointOnTriangle(t1, vs[0], wp[1], alpha[1], &dp[1]);
   
      ep[1]  = 
	
	  alpha[1][0] * err_volume[vstar1[0]] +
	  alpha[1][1] * err_volume[vstar0[0]] +
	  alpha[1][2] * err_volume[vstar0[1]];

      dp[1] = dp[1] + ep[1];
	  
      /* the triangle with the smallest potential error wins: */

      if (dp[0] < dp[1])

	{
	  which_triangle = 0;

	  /* 1) set values for the normal, color and data at the potential simplified vertex location */
	     
	  if (move_vertices)
	    _smpErrorTestVertexPropertiesInit(
		      vstar0[0], vstar1[0], vstar1[1], alpha[0][0], alpha[0][1], alpha[0][2]);
	  else
	    /* since the simplified vertex does not move, just use the data values defined at v0 */
	    _smpErrorTestVertexPropertiesInit( vstar1[0], 0, 0, 1., 0., 0.);

	  /* 2) compute the minimum distance from any triangle on the Gaussian Sphere to the origin */

	  _smpPenaltyFactor4ErrorOnNormals( val0, val1, vstar0, vstar1);

	  edge_collapsible = /*  simpData,
	                  simplified_vertex,     new_v2,     new_v3, alpha_simp_vertex, alpha_n_v2, alpha_n_v3,
			  old_v1,                old_v2,     old_v3, alpha_o_v1,        alpha_o_v2, alpha_o_v3
			  */
	    _smpErrorTestVertexPropertiesUpdate( 
                      vstar1[0], 0, 0, 1., 0., 0., 
		      vstar0[0], vstar1[0], vstar1[1], alpha[0][0], alpha[0][1], alpha[0][2]);

	}
      else
	{
	  which_triangle = 1;

	   if (move_vertices)
	     _smpErrorTestVertexPropertiesInit( 
		      vstar1[0], vstar0[0], vstar0[1], alpha[1][0], alpha[1][1], alpha[1][2]);
	   else
	     /* since the simplified vertex does not move, just use the data values defined at v0 */
	     _smpErrorTestVertexPropertiesInit( vstar1[0], 0, 0, 1., 0., 0.);

	  _smpPenaltyFactor4ErrorOnNormals( val0, val1, vstar0, vstar1);

	  edge_collapsible = 
	    _smpErrorTestVertexPropertiesUpdate( 
		       vstar1[0], 0, 0, 1., 0., 0., 
		       vstar1[0], vstar0[0], vstar0[1], alpha[1][0], alpha[1][1], alpha[1][2]);
	}

#ifdef _IMPROVE_POSITIONING

      if (move_vertices)
	{
	  /* this is the first constraint */

	  /* copy      wp[which_triangle] - vs[0]                to tiling_vert[0] 
	     copy   the associated error  ep[which_triangle]     to tiling_vert_errors[0] */
      
	  memcpy(tiling_vert[0], wp[which_triangle], sizeof(Vertex));
	  SUBVEC3(tiling_vert[0], vs[0]);

	  tiling_vert_errors[0] = ep[which_triangle];
	}
#endif

       UPDATE_MAX_CONSTRAINT(dp[which_triangle]);

      /* 2)  determine the closest point to the vertex v0 on the
	 triangles q1,...,qval0-2. */

      if (edge_collapsible)
	{
	 

	  /* the triangles q1,...,val0-2 are constructed explicitely
	     with the triples of vertices:
	     (vs,v0,v1), vs(v_val03, v_val02) */

	  memcpy(a_triangle[0], vs[0], sizeof(Vertex));

	  i = 1;
	  d_min = 0.0;

	  while (i< val0-1)
	    {
	      qi = EXTRACT_TRIANGLE_FROM_VERTEX_STAR(vstar0, vert, i);

	      _smpClosestPointOnTriangle(qi, vert[vstar1[0]], wp[0], alpha[i], &dp[0]);

	      if (alpha[i][0]>0.)

		/* If v0 does not project on the link, then the projection is useful and can produce
		   a valid constraint */

		{
		  one_over_alpha = 1. / alpha[i][0];

		  ep[0]    = (err_volume[vstar1[0]]
			      - alpha[i][1] * err_volume[vstar0[i]]
			      - alpha[i][2] * err_volume[vstar0[i+1]]) * one_over_alpha;
		  
		  dp[0]    = dp[0] * one_over_alpha + ep[0];
		    
		  if (i_min  == -1 || dp[0] < d_min) { d_min = dp[0]; i_min = i;

#ifdef _IMPROVE_POSITIONING

		  if (move_vertices)
		    {
		      /* in a temp storage put
			 (vert[vstar1[0]] - wp[0]) / alpha[i][0] 
			 and ep[0] 
			 and save them as the constraint for which the minimum is found 
			 alternatively, we can use the tiling_vert array as storage space
			 and only increment the num_tiling_vert counter once the constraint
			 is implemented 
			 also, we need to store the magnitude of the constraint and
			 keep track of the constraint of max magnitude as we store the constraints.
			 */

		      memcpy(  tiling_vert[num_tiling_vert], vert[vstar1[0]], sizeof(Vertex));
		      SUBVEC3(  tiling_vert[num_tiling_vert], wp[0]);
		      SCALEVEC3(tiling_vert[num_tiling_vert], one_over_alpha);

		      tiling_vert_errors[num_tiling_vert] = ep[0];

		    } /* end if (move_vertices) */
#endif 

		  }
		}
	      else
		{
		  /* otherwise v0 projects onto the link edge of "triangle"
		     the epsilon value at the simplified vertex cannot compensate
		     but it is possible that the error values at the link are high enough
		     to treat that constraint */

		  if (alpha[i][1] * err_volume[vstar0[i]]   + 
		      alpha[i][2] * err_volume[vstar0[i+1]] >= dp[0])
		    {
		      /* look no further for the closest triangle */
		      i_min = i;
		      i = val0-1; /* next i will be = val0 and the loop will immediately stop */
		      d_min = 0.;
		    }
		}
	      i++;
	    }
	  
	  /* the triangle corresponding to the smallest error wins */

	  if (i_min >=0)
	    {
	      /* if vertex0 does not project on the rim,

		 i.e., 

		 if (alpha[i_min][0] > 0.)


		 a potential new error value is generated at the pivot
		 as follows: */
	     
	      UPDATE_MAX_CONSTRAINT(d_min);

	      if (edge_collapsible)

		    edge_collapsible = 
		      _smpErrorTestVertexPropertiesUpdate( 
							vstar1[0], vstar0[i_min], vstar0[i_min+1], 
							alpha[i_min][0], alpha[i_min][1], alpha[i_min][2],
							vstar1[0], 0, 0, 1., 0., 0.);

		  /* even though v0 == vstar1[0] is the same vertex, and the
		     data is not allowed to be modified at at vertex, 
		     we still have to perform this test in case the simplified
		     vertex would have been allowed to move: we would then verify
		     that the old data is still compatible with the new directions */

	    }

	  else edge_collapsible = 0; /* no projection was useful */
	 
	  /* 3) */

	  if (edge_collapsible)
	    {
	      /* take each edge (vs,v1) .. (vs,v_val0-1)
		 whereby v1, v_val0-1 refer to indices to
		 the star the first vertex vstar0 */
	
	      i = 1;
	      while (edge_collapsible && i<val0)
		{
		 
		  j = 0;
		  while (edge_collapsible && j<val0)
		    {
		
		      d_min = _smpSegmentIntersection(vs[0], 
						      vert[vstar0[i]], 
						      vert[vstar1[0]],
						      vert[vstar0[j]], 
						      &lambda1, &lambda2, !j, 0, 1);

		      
		      /* If the closest point is located in the
			 inside of either edge, determine a crossing point. */

		      if (lambda2 >= 0. && lambda2 <= 1.) /* inside 2 */
			{
			  onelambda2 = 1. -lambda2;

			  ep[0]  = lambda2  * err_volume[vstar1[0]] + 
			      (onelambda2)  * err_volume[vstar0[j]];

			  if (lambda1 > 0. && lambda1 <= 1.) /* inside 1 */
		      
			    {
			      onelambda1 = 1.-lambda1;
			      /* treat separately the error and the constraint vector */
			        
			      one_over_alpha   = 1./ lambda1;

			      ep[0] = (ep[0] - onelambda1 * err_volume[vstar0[i]]) 
				* one_over_alpha;

			      d_min = d_min * one_over_alpha + ep[0];
		     
#ifdef _IMPROVE_POSITIONING
			      /*  (lambda2 *  vstar1[0] + (1-lambda2) vstar0[j]) -
				  (lambda1 *  vs[0]     + (1-lambda1) vstar0[i]) */

			      if (move_vertices)
			       { 
				register U_char k;

				for (k=0;k<3;k++)
				  tiling_vert[num_tiling_vert][k] = lambda2 * vert[vstar1[0]][k] +
				                               (onelambda2) * vert[vstar0[j]][k] -
				                                    lambda1 * vs[0][k] -
				                               (onelambda1) * vert[vstar0[i]][k];
				SCALEVEC3(tiling_vert[num_tiling_vert], one_over_alpha);
			      
				tiling_vert_errors[num_tiling_vert] = ep[0];  

			       }
#endif	  
			      UPDATE_MAX_CONSTRAINT(d_min);

			      if (edge_collapsible)
				edge_collapsible = 

				  _smpErrorTestVertexPropertiesUpdate( 
						       vstar1[0], vstar0[i], 0, lambda1, onelambda1, 0.,
						       vstar1[0], vstar0[j], 0, lambda2, onelambda2, 0.);

			    } 
			  
			  /* LAMBDA1 = 0 IS AGAIN A SPECIAL CASE 
			     if lambda1 = 0
			     we can't treat that constraint and the edge is
			     not collapsible */

			  else if (lambda1 == 0.) 
			    {
			      /* the error at vs cannot accomodate that constraint,
				 but maybe it is already satisfied */

			      if (err_volume[vstar0[i]] < d_min)
				edge_collapsible = 0;
			    }
			}

		    

		      j++; if (i==j) j++;
		    }
		  i++;
		}

	      /* 4) determine the closest point to the vertex v1 on
		   the triangles qval0,...,q_val0+val1-3 */


	      if (edge_collapsible)
		{
	
		  d_min = 0.0;

		  i_min = -1;

		  /* the triangles qval0,...,q_val0+val1-3 are
		     constructed explicitely using the vertices of
		     vstar1 and the triples of vertices with the
		     triples of vertices: (vs,v1,v2), (vs,v_val1-2,
		     v_val1-1) */

		  i = 1;

		  while (i<val1-1)
		    {
		      
		      qi = EXTRACT_TRIANGLE_FROM_VERTEX_STAR(vstar1, vert, i);

		      _smpClosestPointOnTriangle(qi, vert[vstar0[0]], wp[0], alpha[i], &dp[0]);

		      if  (alpha[i][0]>0.) 

			{  
			  one_over_alpha = 1. / alpha[i][0];

			  ep[0] = (err_volume[vstar0[0]]
				   - alpha[i][1] * err_volume[vstar1[i]]
				   - alpha[i][2] * err_volume[vstar1[i+1]]) * one_over_alpha;

			  dp[0]  = dp[0] * one_over_alpha + ep[0];
		    
			  if (i_min == -1 || dp[0] < d_min) {d_min = dp[0]; i_min = i;

#ifdef _IMPROVE_POSITIONING
			  if (move_vertices)
			   {
			    memcpy(  tiling_vert[num_tiling_vert], vert[vstar0[0]], sizeof(Vertex));
			    SUBVEC3(  tiling_vert[num_tiling_vert], wp[0]);
			    SCALEVEC3(tiling_vert[num_tiling_vert], one_over_alpha);

			    tiling_vert_errors[num_tiling_vert] = ep[0];
			   }
#endif 

			  }
			}
		      else
			{
			  /* otherwise v1 projects onto the link edge of "triangle"
			     the epsilon value at the simplified vertex cannot compensate
			     but it is possible that the error values at the link are high enough
			     to treat that constraint */
			  if (alpha[i][1] * err_volume[vstar1[i]]   + 
			      alpha[i][2] * err_volume[vstar1[i+1]] >= dp[0])
			    {
			      /* look no further for the closest triangle */
			      i_min = i;
			      i = val1-1; /* and the loop will immediately stop  */
			      d_min = 0.;
			    }

			}
		      i++;
		    }
	  
		  /* the closest triangle wins */
		  
		  if (i_min >=0)
		    
		    {
		      UPDATE_MAX_CONSTRAINT(d_min);
		
		      if (edge_collapsible)

			edge_collapsible = 
			  _smpErrorTestVertexPropertiesUpdate( 
		            vstar1[0], vstar1[i_min], vstar1[i_min+1], 
		            alpha[i_min][0], alpha[i_min][1], alpha[i_min][2],
		            vstar0[0], 0, 0, 1., 0., 0.);
	     
		    }
		  else edge_collapsible = 0; /* no projection was useful */
		  	 
		  /* 5) */

		  if (edge_collapsible)
		    {
		      /* Take in turn each edge (vs,v1),(vs,v1),...,
			 (vs,v_val1-1) wherein v1, v2... refer to star 1

			 original vertices of the star:
			 0,1,2,...,val1-1
			 we take
			     1,...,val1-1 */

		      i = 1;
		    
		      while (edge_collapsible && i<val1)
			{
			  int first_j = 1;

			  /* i == 1      and j == 0 was treated before in 3)
			     i == val1-1 and j == 0 as well */

			  j = 
			    (i == 1)      ? 2 : 
			    (i == val1-1) ? 1 : 
			    0;

			  while (edge_collapsible && j<val1)
			    {
		  
			      /* Determine the closest point on each
				 edge incident to "v1" */

			      d_min = _smpSegmentIntersection(vs[0], 
							      vert[vstar1[i]],
							      vert[vstar0[0]],
							      vert[vstar1[j]],
							      &lambda1, &lambda2, first_j, 0, 1);
		      
		      
			      if (lambda2 >= 0. && lambda2 <= 1.) /* inside 2 */
				{
				  onelambda2 = 1.-lambda2;

				  ep[0] =  lambda2 * err_volume[vstar0[0]] +
				      (onelambda2) * err_volume[vstar1[j]];
				
				  if (lambda1 >= 0. && lambda1 <= 1.) /* inside 1 */
				    { 
				      onelambda1 = 1.-lambda1;
				      /* treat separately the error and the constraint vector */
			        
				      one_over_alpha   = 1./ lambda1;

				      ep[0] = (ep[0] - onelambda1 * err_volume[vstar1[i]]) 
					* one_over_alpha;

				      d_min = d_min * one_over_alpha + ep[0];

#ifdef _IMPROVE_POSITIONING	

				      /*  (lambda2 *  vstar0[0] + (1-lambda2) vstar1[j]) -
					  (lambda1 *  vs[0]     + (1-lambda1) vstar1[i]) */
				      
				      if (move_vertices)
				       { 
					register U_char k;

					for (k=0;k<3;k++)
					  tiling_vert[num_tiling_vert][k] = 
                                                                    lambda2 * vert[vstar0[0]][k] +
				                               (onelambda2) * vert[vstar1[j]][k] -
				                                    lambda1 * vs[0][k] -
				                               (onelambda1) * vert[vstar1[i]][k];
					SCALEVEC3(tiling_vert[num_tiling_vert], one_over_alpha);
			   
					/* report error */
					tiling_vert_errors[num_tiling_vert] = ep[0];  

				       }
#endif 		
				      UPDATE_MAX_CONSTRAINT(d_min);

				      if (edge_collapsible)
					edge_collapsible = 
					  _smpErrorTestVertexPropertiesUpdate( 
				            vstar1[0], vstar1[i], 0, lambda1, onelambda1, 0.,
				            vstar0[0], vstar1[j], 0, lambda2, onelambda2, 0.);

				    }
				  else if (lambda1 == 0.)
				    {
				      /* the error at vs cannot accomodate that constraint,
					 but maybe it is already satisfied */

				      if (err_volume[vstar1[i]] < d_min)
					edge_collapsible = 0;
				    }
				}
			      first_j = 0;
			      j++;
			      if (i==j) j++;
			    }
			  i++;
			}
		    }
		}
	    }
	}
    }
    


  if (edge_collapsible)
    {

#ifdef  _IMPROVE_POSITIONING

      if (move_vertices)
	{

	  /* use the list of constraints that were collected and improve
	     upon the position of the simplified vertex if possible and
	     upon the error */


	  _smpImproveSimpVertPosAndErrorBBox(
					  simplified_vertex,
					  &largest_err_pivot, 
					  /* magnitude of the largest (first) constraint + first error,
					     so that we don't have to recompute it */
					  num_tiling_vert,
					  /* the constraint of maximum magnitude (mag vector + error)
					     is the first constraint */
					  tiling_vert,
					  tiling_vert_errors); 

	} /* end if (move_vertices) */

	
#endif 

      /* update the tolerance and error values 
	 only the error at the pivot will change */

      
      err_volume[vstar1[0]] = largest_err_pivot;

      /* update the tolerance volume */
 	  
      tol_volume[vstar1[0]] =  tol_pivot;

      _smpErrorTestVertexPropertiesEnd( vstar1[0]);

    }

  return edge_collapsible;
}

#undef SUBVEC3
#undef SCALEVEC3

#ifdef _IMPROVE_POSITIONING 

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpImproveSimpVertPosAndErrorBBox                                     |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

static int _smpImproveSimpVertPosAndErrorBBox(Vertex simplified_vertex,
					      float *largest_error,
					      int  num_tiling_vert,
					      Vertex  *tiling_vert,
					      float  *tiling_error)
{
  int improved = 0;

  Vertex new_position, bbox_low, bbox_hi;

  double new_magnitude = 0.0, magnitude, coord_dist;

  /* 1) computation of the bounding box of the set of spheres */

  register int i;

  register U_char j;

  for (j=0;j<3;j++)
    {
      bbox_low[j] = tiling_vert[0][j] - tiling_error[0];
      bbox_hi [j] = tiling_vert[0][j] + tiling_error[0];
    }

  for (i=1;i< num_tiling_vert;i++)

    for (j=0;j<3;j++)
      {
	if ((tiling_vert[i][j] - tiling_error[i]) < bbox_low[j]) 
	  bbox_low[j] = tiling_vert[i][j] - tiling_error[i];

	if ((tiling_vert[i][j] + tiling_error[i]) > bbox_hi[j]) 
	  bbox_hi[j] = tiling_vert[i][j] + tiling_error[i];
      }

  /* 2) compute the new position as the center of the bounding box */

  for (j=0;j<3;j++) new_position[j] = 0.5 * (bbox_low[j] + bbox_hi[j]);

  /* 3) compute the new largest error */

  
  for (i=0;i< num_tiling_vert;i++)
    {
      magnitude = 0.0;

      for (j=0;j<3;j++)
	{
	  coord_dist = tiling_vert[i][j] - new_position[j];
	  magnitude += coord_dist * coord_dist;
	}

      magnitude = sqrt (magnitude) + tiling_error[i];

      if (magnitude >  new_magnitude)
	new_magnitude = magnitude;
    }

  /* 4) verify whether the distance was reduced */

  if (new_magnitude < *largest_error) improved = 1;
  
  /* 5) move the simplified vertex and update the magnitude */

  if (improved)
    {
      /* actually add the new position to the simplified vertex location */

       ADDVEC3(simplified_vertex, new_position);
		
       *largest_error = new_magnitude;
		  
    }

  return improved;
}

#endif 
	

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpErrorWithinToleranceBoundary2ndKnd                                 |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpErrorWithinToleranceBoundary2ndKnd(int v0, int v1, 
						     int *vstar0, int *vstar1, int val0, int val1,
						     int parent, int child)
{
  int edge_collapsible = 1;

  /* we construct a tiling of (e*) and a tiling of (vs*)
     such that there is a one to one mapping between the two tilings */
  
 
  /*   e*:
	          v*1\      /v*0  \      /
	           e*1\  *1/e*0    \ *1 /
	       *val0-1 \  / *0   *0 \  / *val1-1
	     ___________\/___________\/__________
	     v*val0-1   v0    vs      v1
             vl                                  vr
	     */

  /* The difference between the interior vertex and boundary vertex
    approaches is that vs v0 and v1 are projected on edges exclusively.
    That limits the number of edge intersections that must be computed
   */

  /* 1) determine the closest point to the vertex vs on the edge v0,v1 */

  /* 2) determine the closest point to the vertex v0 on the edge 
     (vl,vs) */

  /* 3) determine the closest point to the vertex v1 on the edge (vs,vr) */

  /* 4) take in turn each edge (vs,v*1)    (vs,v*val0-1) 
     and intersect them with each edge (v0,v*0),  (v0, v*val0-1) */


  /* 5) do the same as 4) for the star of the vertex v1 
   */

  register int i;


  float 
    potential_err_pivot,
    largest_err_pivot = 0.,

    /* set the maximum tolerance for the star */
  
    tol_pivot = MIN(tol_volume[v0], tol_volume[v1]);

  for (i=0;i<val0;i++)
    if (tol_volume[vstar0[i]] < tol_pivot)
      tol_pivot = tol_volume[vstar0[i]];

  for (i=1;i<val1;i++) 

    /* the first vertex of the star of v1 is the same as the first
       vertex of the star of v0: vstar1[0] = vstar0[0] */

    if (tol_volume[vstar1[i]] < tol_pivot)
      tol_pivot = tol_volume[vstar1[i]];

  /* verify that the maximum tolerance is not already exceeded at
     the vertices */

  i = 0;

  while (edge_collapsible && i < val0)
  /* the equality (star_err == tol) is allowed */  
    if (err_volume[vstar0[i++]] > tol_pivot)
      edge_collapsible = 0;

  i = 1;
  while (edge_collapsible && i < val1)
    if (err_volume[vstar1[i++]] > tol_pivot)
      edge_collapsible = 0;
  
  if (err_volume[v0] > tol_pivot || err_volume[v1] > tol_pivot)
    edge_collapsible = 0;

  if (edge_collapsible)
    {
      float 
	t, dist;

      Vertex wp[2];

      bzero ((char *)wp, 2 *sizeof (Vertex));

      
      /* 1 determine the closest point to the vertex vs on the edge v0,v1 */
     
      dist = _smpClosestPointOnEdge(simplified_vertex, wp[0],
				    vert[v0], vert[v1],&t);

      potential_err_pivot = dist + t * err_volume[v0] + (1.-t) * err_volume[v1];
     
      UPDATE_MAX_CONSTRAINT (potential_err_pivot);

      if (edge_collapsible)
	{
	  if (move_vertices)
	    _smpErrorTestVertexPropertiesInit( v0, v1, 0, t, 1.-t, 0.);
	  else
	    _smpErrorTestVertexPropertiesInit( parent, 0, 0, 1., 0., 0.);
	  
	  _smpPenaltyFactor4ErrorOnNormalsBoundary2ndKind( v0, v1, val0, val1, vstar0, vstar1);

	  _smpErrorTestVertexPropertiesUpdate( 
					      v0, 0, 0, 1., 0., 0., 
					      v0, v1, 0, t, 1.-t, 0.); 
	  
	  /* 2) determine the closest point to the vertex v0 on the edge 
	     (vl,vs) */

	 
	  dist = _smpClosestPointOnEdge(vert[v0], wp[0],
					vert[vstar0[val0-1]], simplified_vertex, 
					&t);

	  /* t * err[vl] + 1-t * err[vs] >= dist + err[v0] */

	  if (t < 1.)
	    {
	      potential_err_pivot = 
		
		(dist + err_volume[v0] -t * err_volume[vstar0[val0-1]] )/ (1.-t);

	      UPDATE_MAX_CONSTRAINT (potential_err_pivot);
	    }

	  /*
	    v0
	    if (1-t == 0 <==> t = 1)                             |
	    that means that v0 projects orthogonally on vl      vl____________vs

	    we can't treat this constraint and the edge is not collapsible 
	    unless err[vl] >= dist + err[v0]
	    */
	  
	  else if (err_volume[vstar0[val0-1]] < dist + err_volume[v0])
	    
	    edge_collapsible = 0;

	  if (edge_collapsible)
	    {
	      edge_collapsible =
		_smpErrorTestVertexPropertiesUpdate( 
						    v0, vstar0[val0-1], 0, 1.-t, t, 0.,  /* vs, vl */
						    v0, 0, 0, 1., 0., 0.);    

	      /* 3) determine the closest point to the vertex v1 on the edge (vs,vr) */

	      dist = _smpClosestPointOnEdge(vert[v1], wp[0],
					    vert[vstar1[val1-1]], simplified_vertex, &t);

	      /* t * err[vr] + 1-t * err[vs] >= dist + err[v1] */
                                
	      if (t < 1.)
		{
		  potential_err_pivot = 
		    (dist + err_volume[v1] -t * err_volume[vstar1[val1-1]])
		    / (1.-t);

		  UPDATE_MAX_CONSTRAINT (potential_err_pivot);
		}
	  
	      else if (err_volume[vstar1[val1-1]] < dist + err_volume[v1])
		
		edge_collapsible = 0;

	
	      if (edge_collapsible)

		edge_collapsible = _smpErrorTestVertexPropertiesUpdate( 
							   v0, vstar1[val1-1], 0, 1.-t, t, 0.,  /* vs, vl */
							   v1, 0, 0, 1., 0., 0.);    
      

	      if (edge_collapsible)
		{
		  float 
		    lambda1, lambda2;

		  int j;

		  /* 4) take in turn each edge (vs,v*1)    (vs,v*val0-1) 
		     and intersect them with each edge (v0,v*0),  (v0, v*val0-1) */

		  i = 1;
		  while (edge_collapsible && i<val0)
		    {
		 
		      j = 0;
		      while (edge_collapsible && j<val0)
			{
		
			  dist = _smpSegmentIntersection(simplified_vertex,
							 vert[vstar0[i]], 
							 vert[v0], 
							 vert[vstar0[j]],
							 &lambda1, &lambda2, !j, 0, 1);
					       
			  /* If the closest point is located in the
			     inside of either edge, determine a crossing point. */

			  if (lambda2 >= 0. && lambda2 <= 1.) /* inside 2 */
			    {
			      dist += lambda2   * err_volume[v0] + 
				(1. -lambda2)  * err_volume[vstar0[j]];


			      if (lambda1 >= 0. && lambda1 <= 1.)
				
				edge_collapsible = 
				  _smpErrorTestVertexPropertiesUpdate( 
                                        v0, vstar0[i], 0, lambda1, 1. -lambda1, 0.,
				        v0, vstar0[j], 0, lambda2, 1. -lambda2, 0.);


			      if (lambda1 > 0. && lambda1 <= 1.) /* inside 1 */
		      
				{
				  potential_err_pivot = 
				    (dist - (1. -lambda1)  * err_volume[vstar0[i]])
				    / lambda1;

	    
				  UPDATE_MAX_CONSTRAINT(potential_err_pivot);
			      
				} 
			  
			      /* if lambda1 = 0 we can't accomodate
				 that constraint with the error value at the simplified vertex
				 and the edge might be declared "non-collapsible"

				 vs_______v*1
                                          |
				       v0_|_v*0

				       dist == 0 can occur when the two vertices
				       vstar0[j] and vstar0[i] are the same
				       but we prohibit that */

			      else if (lambda1 == 0.) 
				{
				  /* the error at vs cannot accomodate
				     that constraint, but maybe it is
				     already satisfied */

				  if (err_volume[vstar0[i]] < dist)
			    
				    edge_collapsible = 0;
				}
			    }

		    

			  j++;
			  if (i == j) j++; /* we prohibit (i == j) */
			}
		      i++;
		    }

		  if (edge_collapsible)
		    {
		      /* 5) do the same as 4) for the star of the vertex v1:

			 take in turn each edge (vs,v*1)    (vs,v*val1-1) 
			 and intersect them with each edge (v1,v*0),  (v1, v*val1-1) */

		      i = 1;
		      while (edge_collapsible && i<val1)
			{
		 
			  j = 0;
			  while (edge_collapsible && j<val1)
			    {
		
			      dist = _smpSegmentIntersection(simplified_vertex,
							     vert[vstar1[i]], 
							     vert[v1], 
							     vert[vstar1[j]],
							     &lambda1, &lambda2, !j, 0, 1);
					       
			      			      
			      if (lambda2 >= 0. && lambda2 <= 1.) /* inside 2 */
				{

				  dist += 
				    lambda2        * err_volume[v1] +
				    (1. -lambda2)  * err_volume[vstar1[j]];


				  if (lambda1 >= 0. && lambda1 <= 1.)

				    edge_collapsible = 
				      _smpErrorTestVertexPropertiesUpdate( 
				        v0, vstar1[i], 0, lambda1, 1. -lambda1, 0.,
				        v1, vstar1[j], 0, lambda2, 1. -lambda2, 0.);

				  if (lambda1 > 0. && lambda1 <= 1.) /* inside 1 */
		      
				    {
				      potential_err_pivot = 
					(dist - (1. -lambda1)  * err_volume[vstar1[i]])
					/ lambda1;

				      
				      UPDATE_MAX_CONSTRAINT(potential_err_pivot);
			      
				    } 
			  
				  else if (lambda1 == 0.) 
				    {
				      /* the error value at vs cannot accomodate that constraint,
					 but maybe it is already satisfied */
				
				      if (err_volume[vstar1[i]] < dist)
					edge_collapsible = 0;
				    }
				}

			      j++;
			      if (i == j) j++; /* we prohibit (i == j) */
			    }
			  i++;
			}
		    }
		}
	    }
	}
    }

  /* update the tolerance and error volumes in case the edge will be simplified */

  if (edge_collapsible) 
    {

      err_volume[parent] = largest_err_pivot;

      /* update the tolerance volume */
 	  
      tol_volume[parent] =  tol_pivot;
	
      _smpErrorTestVertexPropertiesEnd( parent);

    }
  
  return edge_collapsible;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpErrorWithinToleranceBoundary1stKnd                                 |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpErrorWithinToleranceBoundary1stKnd(int v0, int v1, int *vstar1, int val1)
{
  int edge_collapsible = 1;

  register int i;
    
  float
    tol_pivot, potential_err_pivot, largest_err_pivot = 0., 
    dist,
    lambda1, lambda2;

  /* set the maximum tolerance for the star */
  
  tol_pivot = tol_volume[v1];
  
  for (i=0;i<val1;i++)
    if (tol_volume[vstar1[i]] < tol_pivot)
      tol_pivot = tol_volume[vstar1[i]];

  /* verify that the maximum tolerance is not already exceeded at
     the vertices */

  i = 0; 
  while (edge_collapsible && i < val1)
    if (err_volume[vstar1[i++]] > tol_pivot)
      edge_collapsible = 0;
  
  if (err_volume[v1] > tol_pivot)
    edge_collapsible = 0;
  
  if (edge_collapsible)
    {
      Vertex
	triangle[3],
	vs[1],/* local copy of the coordinates of the simplified vertex */
	wp[2],
	alpha[SIMPLIFY_VALENCE_MAX1];  /* val0 + val1 -4 <= valence max <=> val0, val1 <= valence_max +1 
					  because val0 >=3 and val1 >=3 */
      float
	dp;

      int 
	i_min         = -1,
	link_error_ok = 0;

      bzero((char *)triangle, 3 * sizeof (Vertex));

      bzero((char *)vs, sizeof (Vertex));

      bzero((char *)wp, 2 * sizeof (Vertex));

      bzero((char *)alpha, SIMPLIFY_VALENCE_MAX1 * sizeof (Vertex));

      _smpErrorTestVertexPropertiesInit( vstar1[0], 0, 0, 1., 0., 0.);
      _smpPenaltyFactor4ErrorOnNormalsBoundary1stKind( v1, vstar1, val1);

      /* the simplified is already positioned, since it must be v0 */
      memcpy(vs[0], simplified_vertex, sizeof (Vertex));

      /* 1) determine the closest point to the vertex v1 on the triangles of star1 after simplification
       */
      
      memcpy(triangle[0], vs[0], sizeof (Vertex));

      /*                  v*[val1-1]
    labeling system          *[0]\  /
                         v*[0]____\/____
			      *[1]/\
                            v*[1]/  \
				 *[2]

				 both triangles *[0] and *[1] disappear after edge collapse

				 */
      i    = 1;

      dist = 0.;

      while (i<val1-1)
	{
	  memcpy(triangle[1], vert[vstar1[i]],   sizeof(Vertex));
	  memcpy(triangle[2], vert[vstar1[i+1]], sizeof(Vertex));
	
	  _smpClosestPointOnTriangle(triangle, vert[v1], wp[0], alpha[i], &dp);
      
	  dp += err_volume[v1];

	  if (alpha[i][0]>0.)
	    {
	      dp = (dp - 
		alpha[i][1] * err_volume[vstar1[i]] -
		alpha[i][2] * err_volume[vstar1[i+1]])
		/ alpha[i][0];

	      if (i_min == -1 || dp < dist)  {dist = dp; i_min = i;}
		
	    }
	  else
	    {
	      /* otherwise v1 projects onto the link edge of "triangle"
		 the epsilon value at the simplified vertex cannot compensate
		 but it is possible that the error values at the link are already sufficient */

	      if (alpha[i][1] * err_volume[vstar1[i]]   + 
		  alpha[i][2] * err_volume[vstar1[i+1]] >= dp)
		{
		  /* look no further for the closest triangle */
		  i_min = i;
		  i = val1-1; /* so that the loop will immediately stop*/
		  dist = 0;
		}
	    }
	 
	  i++;
	}
	  
      /* the closest triangle wins */
      
      if (i_min >=0)
	       
	{
	  potential_err_pivot = dist;

	  UPDATE_MAX_CONSTRAINT(potential_err_pivot);	
   
	  if (edge_collapsible)

	    edge_collapsible = 
	      _smpErrorTestVertexPropertiesUpdate( 
						  v0, vstar1[i_min], vstar1[i_min+1], 
						  alpha[i_min][0], alpha[i_min][1], alpha[i_min][2],
						  v1, 0, 0, 1., 0., 0.);

	}

      else 
	edge_collapsible = 0; /* v1 did not project inside of the new star */


      if (edge_collapsible)
	{
	  /* 2)
	     
                /                 v*[val1-1]     v*[val1-2]
               /    __/                    \    /
	      /  __/                        \  /
	   v0/__/________                ____\/____
           ==\  \__                     v*0  /\
           vs \    \__                      /  \
               \      \                    /    \
                \                         v*1    v*2

		take in turn each edge (vs,v*2) ..., (vs, v*[val1-2])
		
		and "intersect" them with each edge (v1, v*[1]) ... (v1, v*[val1-1])
		*/
	  int j;

	  i = 2;

	  while (edge_collapsible && i < val1-1)
	    {
	      j = 1;
	      
	      while (edge_collapsible && j < val1)
		{         
		  dist =  _smpSegmentIntersection(simplified_vertex,
						  vert[vstar1[i]], 
						  vert[v1], 
						  vert[vstar1[j]],
						  &lambda1, &lambda2, (j==1), 0, 1);
		     
		  /* If the closest point is located in the
		     inside of either edge, determine a crossing point. */

		  if (lambda2 >= 0. && lambda2 <= 1.) /* inside 2 */
		    {
		      dist += 
			lambda2        * err_volume[v1] +
		        (1. -lambda2)  * err_volume[vstar1[j]];


		      if (lambda1 >= 0. && lambda1 <= 1.)
			
			edge_collapsible = 
			  _smpErrorTestVertexPropertiesUpdate(
				        v0, vstar1[i], 0, lambda1, 1. -lambda1, 0.,
				        v1, vstar1[j], 0, lambda2, 1. -lambda2, 0.);


		      if (lambda1 > 0. && lambda1 <= 1.) /* inside 1 */
		      
			{
			  potential_err_pivot = 
			    (dist + (lambda1 -1.) * err_volume[vstar1[i]])
			    / lambda1;
	    
			  UPDATE_MAX_CONSTRAINT(potential_err_pivot);
			}
		      else if (lambda1 == 0.)
			{
			  /* the error at vs cannot accomodate that constraint,
			     but maybe it is already satisfied */

			  if (err_volume[vstar1[i]] < dist)
			    edge_collapsible = 0;
			}
		    }
		  j++;

		  if (i == j) j++; /* we prohibit (i == j) */
		
		}
	      i++;
	    }
	}
	
    }

  if (edge_collapsible)
    {
      /* update the tolerance and error values */

      /* only the error at the pivot will change */
      
      err_volume[v0] = largest_err_pivot;

      /* update the tolerance volume */
 	  
      tol_volume[v0] =  tol_pivot;	

      _smpErrorTestVertexPropertiesEnd( v0);

    }

  return edge_collapsible;
}

#undef UPDATE_MAX_CONSTRAINT

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpErrorTestVertexPropertiesInit                                        |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void SimpData::_smpErrorTestVertexPropertiesInit(int old_v1, int old_v2, int old_v3, 
						 float alpha_o_v1, float alpha_o_v2, float alpha_o_v3) 
{
  vx_normal_psve = vx_color_psve = vx_texCoord_psve = 0.;

  if (vx_normal_error)
    {
      register U_char j;

      /* compute the normal at the potential simplified vertex */

      for (j=0;j<3;j++)
	
	psv_normal[j] = 
	  alpha_o_v1 * vx_normal[old_v1][j] +
	  alpha_o_v2 * vx_normal[old_v2][j] +
	  alpha_o_v3 * vx_normal[old_v3][j];
	  
      /* normalize the normal */
      _smpNormalize3((float *)psv_normal);

 
      
    }

  /* vertex colors */
  if (vx_color_error)
    {
      register U_char j;

      /* compute the color at the potential simplified vertex */  

      for (j=0;j<3;j++)
	
	psv_color[j] = 
	  alpha_o_v1 * vx_color[old_v1][j] +
	  alpha_o_v2 * vx_color[old_v2][j] +
	  alpha_o_v3 * vx_color[old_v3][j];
	  
    }

  /* vertex uv coordinates */

  if (vx_texCoord_error)
    {
      register U_char j;

      /* compute the uv coordinates at the potential simplified vertex */  

      for (j=0;j<2;j++)
	
	psv_texCoord[j] = 
	  alpha_o_v1 * vx_texCoord[old_v1*2+j] +
	  alpha_o_v2 * vx_texCoord[old_v2*2+j] +
	  alpha_o_v3 * vx_texCoord[old_v3*2+j];
    }

   
  return;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpPenaltyFactor4ErrorOnNormals                                       |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpPenaltyFactor4ErrorOnNormals(int val0, int val1, 
					       int *vstar0, int *vstar1)
{
  int success = 0,i;

  /* compute the normal penalty factor: minimum distance from the
     origin to a triangle whose vertices are the vertex normals we do
     this for all triangles before collapse and for all triangles
     after collapse and take the minimum */
 
  if (vx_normal_error)
    {
      static Vertex a_triangle[3], g;
      Vertex *ti, *qi;
      float sq_norm_centroid, sq_norm_centroid_min = -1;

    
      /* we have an interior edge */

      /* 1) visit all triangles before collapse */

      /* star of v0 */
      memcpy(a_triangle[0], vx_normal[vstar1[0]],        sizeof (Vertex)); /* v0*/
      
      /* treat the first triangle separately for easier initialization of sq_norm_centroid_min */

      ti = EXTRACT_TRIANGLE_FROM_VERTEX_STAR(vstar0, vx_normal, 0);
      _smpMakeCentroid(3, ti, g);
      sq_norm_centroid_min = g[0]*g[0] + g[1]*g[1] + g[2]*g[2];

      for (i=1;i<val0-1;i++)
	{
	  ti = EXTRACT_TRIANGLE_FROM_VERTEX_STAR(vstar0, vx_normal, i);
	  /* take the triangle centroid */
	  _smpMakeCentroid(3, ti, g);
	  /* take the squared norm (we will compute the squared root on the smaller value) */
	  
	  sq_norm_centroid = g[0]*g[0] + g[1]*g[1] + g[2]*g[2];

	  if (sq_norm_centroid < sq_norm_centroid_min) sq_norm_centroid_min = sq_norm_centroid;
	}

      /* star of v1 */
      memcpy(a_triangle[0], vx_normal[vstar0[0]],        sizeof (Vertex)); /* v1 */
      
      for (i=0;i<val1-1;i++)
	{
	  ti = EXTRACT_TRIANGLE_FROM_VERTEX_STAR(vstar1, vx_normal, i);
	  _smpMakeCentroid(3, ti, g);
	  sq_norm_centroid = g[0]*g[0] + g[1]*g[1] + g[2]*g[2]; 
	  if (sq_norm_centroid < sq_norm_centroid_min) sq_norm_centroid_min = sq_norm_centroid;
	}
      
      
      /* visit all the triangles after collapse */   
      
      /* use the normal at the potential simplified vertex */
      memcpy(a_triangle[0], psv_normal, sizeof (Vertex));
      
      for (i=1;i<val0-1;i++)
	{
	  qi = EXTRACT_TRIANGLE_FROM_VERTEX_STAR(vstar0, vx_normal, i);
	  
	  _smpMakeCentroid(3, qi, g);
	  
	  sq_norm_centroid = g[0]*g[0] + g[1]*g[1] + g[2]*g[2];
	  
	  if (sq_norm_centroid < sq_norm_centroid_min) sq_norm_centroid_min = sq_norm_centroid;
	}
      
      for (i=1;i<val1-1;i++)
	{
	  qi = EXTRACT_TRIANGLE_FROM_VERTEX_STAR(vstar1, vx_normal, i);
	  
	  _smpMakeCentroid(3, qi, g);
	  
	  sq_norm_centroid = g[0]*g[0] + g[1]*g[1] + g[2]*g[2];
	  
	  if (sq_norm_centroid < sq_norm_centroid_min) sq_norm_centroid_min = sq_norm_centroid;
	}

      /* compute the penalty value */
      normal_error_penalty = (sq_norm_centroid_min > 0.0) ? sqrt(sq_norm_centroid_min) : 0.0;

      success = 1;
    }

  return success;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpPenaltyFactor4ErrorOnNormalsBoundary2ndKind                        |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpPenaltyFactor4ErrorOnNormalsBoundary2ndKind(
						    int v0, int v1,
						    int val0, int val1, 
						    int *vstar0, int *vstar1)
{
  int success = 0,i;

  if (vx_normal_error)
    {
      static Vertex a_triangle[3], g;
      Vertex *ti, *qi;
      float sq_norm_centroid, sq_norm_centroid_min = -1;

    
      /* we have a boundary edge of the second kind */

      /* 1) visit all triangles of the edge star before collapse */

      /* first (central) triangle: */
      memcpy(a_triangle[0], vx_normal[v0],        sizeof (Vertex));
      memcpy(a_triangle[1], vx_normal[v1],        sizeof (Vertex));
      memcpy(a_triangle[2], vx_normal[vstar0[0]], sizeof (Vertex));
    
      _smpMakeCentroid(3, a_triangle, g);
      sq_norm_centroid_min = g[0]*g[0] + g[1]*g[1] + g[2]*g[2];

      /* star of v0 */

      /* memcpy(a_triangle[0], vx_normal[v0], sizeof (Vertex));
	 this was not necessary, vx_normal[v0] was already copied at that
	 memory location a few lines above */
      
      for (i=0;i<val0-1;i++)
	{
	  ti = EXTRACT_TRIANGLE_FROM_VERTEX_STAR(vstar0, vx_normal, i);
	  /* take the triangle centroid */
	  _smpMakeCentroid(3, ti, g);
	  /* take the squared norm (we will compute the squared root on the smaller value) */
	  
	  sq_norm_centroid = g[0]*g[0] + g[1]*g[1] + g[2]*g[2];

	  if (sq_norm_centroid < sq_norm_centroid_min) sq_norm_centroid_min = sq_norm_centroid;
	}

      /* star of v1 */    
      memcpy(a_triangle[0], vx_normal[v1], sizeof (Vertex));
      
 
      for (i=0;i<val1-1;i++)
	{
	  ti = EXTRACT_TRIANGLE_FROM_VERTEX_STAR(vstar1, vx_normal, i);
	  _smpMakeCentroid(3, ti, g);
	  sq_norm_centroid = g[0]*g[0] + g[1]*g[1] + g[2]*g[2]; 
	  if (sq_norm_centroid < sq_norm_centroid_min) sq_norm_centroid_min = sq_norm_centroid;
	}
      
      
      /* visit all the triangles after collapse */   
      
      /* use the normal at the potential simplified vertex */
      memcpy(a_triangle[0], psv_normal, sizeof (Vertex));
      
      for (i=0;i<val0-1;i++)
	{
	  qi = EXTRACT_TRIANGLE_FROM_VERTEX_STAR(vstar0, vx_normal, i);
	  
	  _smpMakeCentroid(3, qi, g);
	  
	  sq_norm_centroid = g[0]*g[0] + g[1]*g[1] + g[2]*g[2];
	  
	  if (sq_norm_centroid < sq_norm_centroid_min) sq_norm_centroid_min = sq_norm_centroid;
	}
      
      for (i=0;i<val1-1;i++)
	{
	  qi = EXTRACT_TRIANGLE_FROM_VERTEX_STAR(vstar1, vx_normal, i);
	  
	  _smpMakeCentroid(3, qi, g);
	  
	  sq_norm_centroid = g[0]*g[0] + g[1]*g[1] + g[2]*g[2];
	  
	  if (sq_norm_centroid < sq_norm_centroid_min) sq_norm_centroid_min = sq_norm_centroid;
	}

      /* compute the penalty value */
      normal_error_penalty = (sq_norm_centroid_min > 0.0) ? sqrt(sq_norm_centroid_min) : 0.0;

      success = 1;
    }

  return success;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpPenaltyFactor4ErrorOnNormalsBoundary1stKind                        |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpPenaltyFactor4ErrorOnNormalsBoundary1stKind(
						    int v1, int *vstar1, int val1)
{
  int success = 0,i;

  if (vx_normal_error)
    {
      static Vertex a_triangle[3], g;
      Vertex *ti, *qi;
      float sq_norm_centroid, sq_norm_centroid_min = -1;

    
      /* we have a boundary edge of the first kind */

      /* 1) visit all triangles before collapse */

      /* first  triangle: */
      memcpy(a_triangle[0], vx_normal[v1],             sizeof (Vertex));
      memcpy(a_triangle[1], vx_normal[vstar1[val1-1]], sizeof (Vertex));
      memcpy(a_triangle[2], vx_normal[vstar1[0]],      sizeof (Vertex)); /* v0 */
      
      _smpMakeCentroid(3, a_triangle, g);
      sq_norm_centroid_min = g[0]*g[0] + g[1]*g[1] + g[2]*g[2];

      /* star of v1 */

      /*                  v*[val1-1]
                             *[0]\  /
                         v*[0]____\/____
			      *[1]/\
                            v*[1]/  \
				 *[2]

				 both triangles *[0] and *[1] disappear after edge collapse */

      /* memcpy(a_triangle[0], vx_normal[v1], sizeof (Vertex));
	 this was not necessary */
      
      for (i=0;i<val1-1;i++)
	{  
	  /*                               
                                 \* / *
                              ____\/____
			        * /\*
                                 / *\
				     
           */
	  ti = EXTRACT_TRIANGLE_FROM_VERTEX_STAR(vstar1, vx_normal, i);
	  /* take the triangle centroid */
	  _smpMakeCentroid(3, ti, g);
	  /* take the squared norm (we will compute the squared root on the smaller value) */
	  
	  sq_norm_centroid = g[0]*g[0] + g[1]*g[1] + g[2]*g[2];

	  if (sq_norm_centroid < sq_norm_centroid_min) sq_norm_centroid_min = sq_norm_centroid;
	}

      /* visit all the triangles after collapse */   
      
      /* use the normal at the potential simplified vertex */
      memcpy(a_triangle[0], psv_normal, sizeof (Vertex));
      
      for (i=1;i<val1-1;i++)
	{ 
	  /*                   val1-1      
                                 \* / *
                              ____\/____
			          /\*
                                 / *\
				1    2
           */
	  qi = EXTRACT_TRIANGLE_FROM_VERTEX_STAR(vstar1, vx_normal, i);
	  
	  _smpMakeCentroid(3, qi, g);
	  
	  sq_norm_centroid = g[0]*g[0] + g[1]*g[1] + g[2]*g[2];
	  
	  if (sq_norm_centroid < sq_norm_centroid_min) sq_norm_centroid_min = sq_norm_centroid;
	}
      
      /* compute the penalty value */
      normal_error_penalty = (sq_norm_centroid_min > 0.0) ? sqrt(sq_norm_centroid_min) : 0.0;

      success = 1;
    }

  return success;
}


#undef  EXTRACT_TRIANGLE_FROM_VERTEX_STAR

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpErrorTestVertexPropertiesUpdate                                    |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpErrorTestVertexPropertiesUpdate(int simp_vertex, 
						  int new_v2, int new_v3,
						  float alpha_sv, float alpha_n_v2, float alpha_n_v3,
						  int old_v1, int old_v2, int old_v3, 
						  float alpha_o_v1, float alpha_o_v2, float alpha_o_v3)
{
  int do_collapse = 1;

  float 
    new_error_link,
    new_error_sv,
    old_error;

  /* interpolate vertex normals and decide if they are sufficiently similar */

  if (vx_normal_error)
    {
      /* we are going to divide with the normal_error_penalty if it is zero
	 we know that the vertex is not collapsible */

      if (normal_error_penalty == 0.0) do_collapse = 0;
      
      else {
	

      VertexD 
	old_normal, 
	new_normal;

      double 
	normal_difference = 0.;
	
      register U_char j;

      for (normal_difference = 0.,
	     j=0;j<3;j++)
	{
	  old_normal[j] = 
	    alpha_o_v1 * vx_normal[old_v1][j] +
	    alpha_o_v2 * vx_normal[old_v2][j] +
	    alpha_o_v3 * vx_normal[old_v3][j];
	  
	  new_normal[j] =
	    alpha_sv   * psv_normal[j] +
	    alpha_n_v2 * vx_normal[new_v2][j] +
	    alpha_n_v3 * vx_normal[new_v3][j];
	}

      /* the two interpolated normals must be normalized-> 97/9/3 NO, not any more:
	 now we divide with the normal_error_penalty */
      {
	double 
	  tmp_diff;
	  
	for (normal_difference=0., j=0;j<3;j++)
	  {
	    tmp_diff = new_normal[j]- old_normal[j];
	  
	    normal_difference += tmp_diff * tmp_diff;
	  }

      }
      normal_difference = sqrt (normal_difference) / normal_error_penalty;
	  
      old_error = 
	alpha_o_v1 * vx_normal_error[old_v1] +
	alpha_o_v2 * vx_normal_error[old_v2] +
	alpha_o_v3 * vx_normal_error[old_v3];
	
      /* can the sum of the normal difference and the old error at the
	 current mesh point be explained by a new error at the simplified
	 vertex, and if it is the case, what is this new error ? */

      new_error_link = 
	alpha_n_v2 * vx_normal_error[new_v2] +
	alpha_n_v3 * vx_normal_error[new_v3];


      /* there is a special case if the simplified vertex is associated with an
	 alpha that's zero: in this case, the error must be explained with the error at
	 the link */

      if (alpha_sv == 0. )
	{
	  if (new_error_link < old_error + normal_difference)
	    do_collapse = 0;
	}
      else /* alpha_sv * error_sv + new_error_link > old_error + normal_difference */
	{
	  new_error_sv = (old_error + normal_difference - new_error_link) / alpha_sv;

	  if (new_error_sv > vx_normal_psve) vx_normal_psve = new_error_sv;

	  do_collapse = (vx_normal_psve <= vx_normal_tol);
	}
      } /* end else if (normal_penalty != 0) */

    }

  /* interpolate vertex colors and decide if they are sufficiently similar */

  if (do_collapse && vx_color_error)
    {
      Vertex 
	old_color, 
	new_color;
      float 
	color_difference = 0.,
	tmp_col_diff;
	
      register U_char j;

      for (color_difference = 0.,
	     j=0;j<3;j++)
	{
	  old_color[j] = 
	    alpha_o_v1 * vx_color[old_v1][j] +
	    alpha_o_v2 * vx_color[old_v2][j] +
	    alpha_o_v3 * vx_color[old_v3][j];
	  
	  new_color[j] =
	    alpha_sv   * psv_color[j] +
	    alpha_n_v2 * vx_color[new_v2][j] +
	    alpha_n_v3 * vx_color[new_v3][j];

	  tmp_col_diff =  new_color[j] - old_color[j];
	  
	  color_difference += tmp_col_diff * tmp_col_diff;
	    

	}

      color_difference = sqrt (color_difference);
	  
      old_error = 
	alpha_o_v1 * vx_color_error[old_v1] +
	alpha_o_v2 * vx_color_error[old_v2] +
	alpha_o_v3 * vx_color_error[old_v3];
	
      /* can the sum of the color difference and the old error at the
	 current mesh point be explained by a new error at the simplified
	 vertex, and if it is the case, what is this new error ? */

      new_error_link = 
	alpha_n_v2 * vx_color_error[new_v2] +
	alpha_n_v3 * vx_color_error[new_v3];


      /* there is a special case if the simplified vertex is associated with an
	 alpha that's zero: in this case, the error must be explained with the error at
	 the link */

      if (alpha_sv == 0. )
	{
	  if (new_error_link < old_error + color_difference)
	    do_collapse = 0;
	}
      else /* alpha_sv * error_sv + new_error_link > old_error + color_difference */
	{
	  new_error_sv = (old_error + color_difference - new_error_link) / alpha_sv;

	  if (new_error_sv > vx_color_psve) vx_color_psve = new_error_sv;

	  do_collapse = (vx_color_psve <= vx_color_tol);
	}
    }

  /* perform same operation on vertex (u,v) coordinates */

  if (do_collapse && vx_texCoord_error)
    {
      float
	old_uv[2],
	new_uv[2];

      float
	uv_difference = 0.,
	tmp_uv_diff;
	
      register U_char j;

      for (uv_difference = 0., j=0;j<2;j++)
	{
	  old_uv[j] = 
	    alpha_o_v1 * vx_texCoord[old_v1*2+j] +
	    alpha_o_v2 * vx_texCoord[old_v2*2+j] +
	    alpha_o_v3 * vx_texCoord[old_v3*2+j];
	  
	  new_uv[j] =
	    alpha_sv   * psv_texCoord[j] +
	    alpha_n_v2 * vx_texCoord[new_v2*2+j] +
	    alpha_n_v3 * vx_texCoord[new_v3*2+j];

	  tmp_uv_diff =  new_uv[j] - old_uv[j];
	  
	  uv_difference += tmp_uv_diff * tmp_uv_diff;
	}

      uv_difference = sqrt (uv_difference);
	  
      old_error = 
	alpha_o_v1 * vx_texCoord_error[old_v1] +
	alpha_o_v2 * vx_texCoord_error[old_v2] +
	alpha_o_v3 * vx_texCoord_error[old_v3];
	
      new_error_link = 
	alpha_n_v2 * vx_texCoord_error[new_v2] +
	alpha_n_v3 * vx_texCoord_error[new_v3];

      if (alpha_sv == 0. )
	{
	  if (new_error_link < old_error + uv_difference)
	    do_collapse = 0;
	}

      else
	{
	  new_error_sv = (old_error + uv_difference - new_error_link) / alpha_sv;

	  if (new_error_sv > vx_texCoord_psve) vx_texCoord_psve = new_error_sv;

	  do_collapse = (vx_texCoord_psve <= vx_texCoord_tol);
	}
    }


  return do_collapse;
}


/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpErrorTestVertexPropertiesEnd                                       |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void SimpData::_smpErrorTestVertexPropertiesEnd(int v0)
{

  /* if we worry about vertex normals */

  if (vx_normal_error)
    {
      vx_normal_error[v0] = vx_normal_psve;
      
      /* use the new vertex normal: */
      memcpy(vx_normal[v0], psv_normal,sizeof(Vertex));

    }

  /* vertex colors: */

  if (vx_color_error)
    {
      vx_color_error[v0]  = vx_color_psve;

      /* use the new vertex color: */
      memcpy(vx_color[v0], psv_color,sizeof(Vertex));
    }

  /* vertex uv texture coordinates: */

  if (vx_texCoord_error)
    {
      vx_texCoord_error[v0]  = vx_texCoord_psve;

      memcpy((void *) (vx_texCoord+v0*2), (void *) psv_texCoord, 2 * sizeof (float));
    }

  return;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpMakeCentroid                                                       |
  |                                                                          |
  +--------------------------------------------------------------------------+*/


void _smpMakeCentroid(int nV, Vertex *vv, Vertex centroid)
{
  register int i;
  register U_char j;

  /* initialize centroid */

  for(j=0;j<3;j++)
    centroid[j]=0.0;

  /* recursive mean estimation */

  for(i=0;i<nV;i++)
    {
      for(j=0;j<3;j++)
	
	centroid[j] =
	  (i * centroid[j] +vv[i][j] ) / (float) (i+1);
    }

  return;
}


/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpMakeBarycenter                                                     |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

void _smpMakeBarycenter(int nV, Vertex *vv, Vertex bary, float *bary_coord)
{
  register int i;
  register U_char j;
 
  /* initialize barycenter */

  for(j=0;j<3;j++)
    bary[j]=0.0;

  for(i=0;i<nV;i++)
    {
      for(j=0;j<3;j++)
	bary[j]+= bary_coord[i]*vv[i][j];
    }

  return;
}



/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpSegmentIntersection                                                |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

float _smpSegmentIntersection(Vertex A, Vertex B, Vertex C, Vertex D,
			      float *lambda, float *mu, int changed_AorB, int changed_C, int changed_D)
{
  /* check whether for two segments in R3, the intersection is in the
    interior of each of them.

    lamda values of 1 and 0 qualify as being in the interior, except
    when one of two edges is of length zero, in which case we set both
    lambda and mu to be -1, to reflect the fact that there is no
    intersection */

  float dist = 0;

  static Vertex w[3];
  static float work[2], max_z, min_z, sign = 0., scale;

  register U_char i,j;


 
  /* if both Vertices A and B did not change since the previous call
     of the routine, the scaling factor, householder vector and sign
     are still correct and do not need to be re-evaluated (this spares
     us the computation of one square root */

  /* if C or D changed, we only have to recompute the portion
     corresponding to C or D 
     in surface simplification, often the only variable that changes
     is D */

  if (changed_AorB)
    {

      for (j=0;j<3;j++)
	{
	  /* substract A from the rest of the vertices */

	  w[0][j] = B[j]-A[j];
	  w[1][j] = C[j]-A[j];
	  w[2][j] = D[j]-A[j];
	}

      /* record the norm of the first vector and scale the 
	 vertices accordingly */

      scale = w[0][0] * w[0][0] + w[0][1] * w[0][1] + w[0][2] * w[0][2];
      if (scale > 0.0) scale = sqrt (scale);

      /* particular case if the scale is zero */

      if (scale > 0.)
	{
	  /* apply a scaling factor so that w[0][2] will be negative */

	  sign = (w[0][2] > 0.) ? -1. : 1.;

	  sign /= scale;

	  for (i=0;i<3;i++) 
	    for (j=0;j<3;j++)
	      w[i][j] *= sign; /* apply the scaling */
    
      /* determine the Householder vector for transforming
	 w[0] to z */
	  
	  w[0][2] -= 1.;
     
	  _smpHouseholderPreMultiplication((float *) w[1], 3,(float *) w[0], 3, 2, work);
   

	  /* now the first segment is the same as (0,0,1) */
	}
    }

  else 
    {
      if (changed_C)
	{
	  for (j=0;j<3;j++)
	    w[1][j] = (C[j]-A[j]) * sign; /* apply scaling */

	  _smpHouseholderPreMultiplication((float *) w[1], 3, /* apply Householder */
					  (float *) w[0], 3, 1, work);
	}

      if (changed_D)
	{
	  for (j=0;j<3;j++)
	    w[2][j] = (D[j]-A[j]) * sign; /* apply scaling */

	  _smpHouseholderPreMultiplication((float *) w[2], 3, (float *) w[0], 3, 1, work); 
	}
    }
	

#define xC_ w[1][0]
#define yC_ w[1][1]
#define zC_ w[1][2]
#define xD_ w[2][0]
#define yD_ w[2][1]
#define zD_ w[2][2]
     
  if (scale > 0.)
    {
     /* there are two cases in which there is no intersection 
	if zC and zD >1 or if zC and zD < 0 */

     if (zC_> zD_) {max_z = zC_; min_z = zD_;}
     else          {max_z = zD_; min_z = zC_;}
    
     if ((min_z > 1.) || (max_z  < 0.))
	
       {
	  *lambda = *mu = -1.;
       }
     
     else
       {
	 static double xDC, yDC, zDC, dlambda, dmu, a, half_b, c, tmp;
	 /* x = xD + mu xDC 
	    y = yD + mu yDC */

	 xDC = xC_ - xD_ ;
	 yDC = yC_ - yD_ ;
	 zDC = zC_ - zD_ ;
	
	 c      = xD_ * xD_ + yD_ * yD_;
	 a      = xDC * xDC + yDC * yDC;
	 
	 if (a > 0.)
	   {
	     half_b  = xD_  * xDC + yD_ * yDC;

	     dmu     = -half_b / a;
	     *mu     = dmu;                     /* mu = -b/(2a) */

             tmp     =  c + half_b * dmu;
             if (tmp >0.0) tmp = sqrt (tmp);
             else tmp= 0.0;

	     dist    =  scale * tmp; /* c - b^2/4a   */

	     /* lambda is given by the 1-z_value at that point */

	     *lambda = 1. - zD_ - dmu * zDC;
	   }

	 /* otherwise, either the two segment are parallel or the
	    second segment is of zero length */
	 
	 else if (min_z < max_z)
	   {

	     /* the second segment is parallel to the first *and*
		there is an intersection */

	     min_z   = MAX(min_z,0.);
	     max_z   = MIN(max_z,1.);

	     dlambda = (max_z + min_z)/2.; 
	   
	     /* this is definitely between [0,1] and corresponds to
		a point inside CD to */

	     *mu = (dlambda - zD_) / zDC;
	     
	     *lambda = 1. - dlambda;
	     dist    = scale * sqrt (c);
	     
	   }
	 else
	   {
	     /* the second segment is degenerate */
	     *lambda = *mu = -1.;
	   }

       }
    }/* endif (scale > 0) */

#undef xC_ 
#undef yC_ 
#undef zC_ 
#undef xD_ 
#undef yD_ 
#undef zD_ 

    
  else
    {
      /* the first edge is degenerate. There is no mutual intersection */

      *lambda = *mu = -1.;

    }

  return dist;
}

/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpColorPerFaceOk                                                     |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpColorPerFaceOk(int fa, int fb, int n_adj0, int *neighboring_faces0, int n_adj1, int *neighboring_faces1)

  /* for each face of the two that disappear, compute the closest
     color (Euclidean distance in RGB space) among the neighboring
     faces, and make sure that the distance is less than a threshold
     value? */

{
  int OK                  = 1;
  int num_collapsed_faces = 2;
  int two_faces[2];

  U_char f = 0;

  two_faces[0] = fa;

  // when the parameter fb is set to -1, this indicates that only one face is collapsed
  if (fb < 0) num_collapsed_faces = 1;
  else two_faces[1] = fb;

  while (OK && f < num_collapsed_faces) 
    {
      if (two_faces[f] >= 0) // verify that we have a valid face index
	{
	  float          *the_color = colors + 3 * two_faces[f];          // color of the collapsed face
	  register float *adj_color = colors + 3 * neighboring_faces0[0]; // color of first neighboring face
	  

	  register float min_distance = DIST3 (the_color, adj_color);
	  register float distance     = 0.;

	  register int   i = 0;
	  
	  // compute the minimum distance to the colors of the neigboring faces

	  for (i = 1; i < n_adj0; i++)
	    {
	      adj_color    = colors + 3 * neighboring_faces0[i];
	      distance     = DIST3 (the_color, adj_color);
	      if (distance < min_distance) min_distance = distance;
	    }

	  for (i = 0; i < n_adj1; i++)
	    {
	      adj_color    = colors + 3 * neighboring_faces1[i];
	      distance     = DIST3 (the_color, adj_color);
	      if (distance < min_distance) min_distance = distance;
	    }
	  
	  if (min_distance > SIMPLIFY_FACE_COLOR_TOL) OK = 0;
     
	}
      f++;
    }

  return OK;
}


/*+--------------------------------------------------------------------------+
  |                                                                          |
  |   _smpTexCoordPerCornerOk                                                |
  |                                                                          |
  +--------------------------------------------------------------------------+*/

int SimpData::_smpTexCoordPerCornerOk(int fa, int fb, int n_adj0, int *neighboring_faces0, 
				      int n_adj1, int *neighboring_faces1)


  /* compute the texture area that disappears, and its proportion of the entire texture area in
     the edge star
     verify that the ratio of disappearing texture is small
     verify also that the total texture area that disappeared is less than a threshold */

{
  int num_collapsed_faces = 2;
  int two_faces[2];

  register int i = 0;

  U_char     f = 0;

  two_faces[0] = fa;

  // when the parameter fb is set to -1, this indicates that only one face is collapsed
  if (fb < 0) num_collapsed_faces = 1;
  else two_faces[1] = fb;

  // compute the texture area of the collapsed faces 

  float 
    star_texture_area          = 0.,
    area                       = 0.,
    *uv                        = NULL;

  _inc_collapsed_texture_area  = 0.;

  for (f = 0; f < num_collapsed_faces; f++)
    {
      uv  = texCoords + 6 * two_faces[f];
      TEXMAP_AREA(uv, area);
      _inc_collapsed_texture_area += area;
    }

  star_texture_area = _inc_collapsed_texture_area;
	 
  for (i = 0; i < n_adj0; i++)
    {
      uv  = texCoords + 6 * neighboring_faces0[i];
      TEXMAP_AREA(uv, area);
      star_texture_area += area;
    }

  for (i = 0; i < n_adj1; i++)
    {
      uv  = texCoords + 6 * neighboring_faces1[i];
      TEXMAP_AREA(uv, area);
      star_texture_area += area;
    }

  return 
    (
     ( (_collapsed_texture_area + _inc_collapsed_texture_area) < SIMPLIFY_TOT_COLLAPSED_TEXTURE_AREA_RATIO * _texture_area )
     &&
     (_inc_collapsed_texture_area < SIMPLIFY_INC_COLLAPSED_TEXTURE_AREA_RATIO * star_texture_area)
     );
}

