// This software module was originally developed by Karl Anders ygard
// (Telenor R&D) in the course of development of the MPEG-4 Systems
// standard (ISO/IEC 14496-1).  This software module is an implementation
// of a part of one or more MPEG-4 Systems tools as specified by the MPEG-4
// Systems standard.  ISO/IEC gives users of the MPEG-4 Systems standard
// free license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4 Systems
// standard.  Those intending to use this software module in hardware or
// software products are advised that its use may infringe existing
// patents.  The original developer of this software module and his/her
// company, the subsequent editors and their companies, and ISO/IEC have no
// liability for use of this software module or modifications thereof in an
// implementation.  Copyright is not released for non MPEG-4 Systems
// conforming products.  The original developer retains full right to use
// the code for his/her own purpose, assign or donate the code to a third
// party and to inhibit third parties from using the code for non MPEG-4
// Systems conforming products.  This copyright notice must be included in
// all copies or derivative works.
//
// Copyright 1997, 1998.

#include "OpenGl/OpenGl.h"
#include "TextureNodeProxy.h"
#include "Stream.h"


// static members

/* static */ BOOL TextureNodeProxy::initialised = FALSE;
/* static */ struct TextureNodeProxy::yuvconvtable TextureNodeProxy::convtab[256];
/* static */ int TextureNodeProxy::cliptab[768];

// destructor

/* virtual */
TextureNodeProxy::~TextureNodeProxy()
{
  if (m_nTextureObject)
#if !defined(GL_VERSION_1_1)
    ::glDeleteTexturesEXT(1, &m_nTextureObject);
#else
    ::glDeleteTextures(1, &m_nTextureObject);
#endif
}

// Render object description

/* virtual */ void
TextureNodeProxy::RenderDescription(OpenGLVisualRenderer* pRenderer, DWORD /* dwTime */)
{
  RenderDescriptionBox(pRenderer, "TextureNode");
}

// Make an RGB Bitmap out of the YUV  

void
TextureNodeProxy::Yuv2Rgb(unsigned char* rgbdata,
			       unsigned char* yptr,
			       unsigned char* uptr,
			       unsigned char* vptr,
			       int lumcols, int lumrows,
			       int chrcols, int chrrows)
{
  // subsampling factors for chrominance register 

  int hsub = lumcols / chrcols;  
  int vsub = lumrows / chrrows; 

  // row loop

  for (int j = 1; j <= lumrows; ++j) {
    // loop chrominance columns 

    for (int i = chrcols; i > 0; --i) {
      // calculate intermediate values that can be recycled

      int u = *uptr++ - 128;
      int v = *vptr++ - 128;

      int u1 = (137 * u) /  100;
      int u2 = (698 * u) / 1000;
      int v2 = (336 * v) / 1000;
      int v3 = (173 * v) /  100;

      // exploit horizontal downsampling by recycling values   

      for (int k = hsub; k > 0; --k) { 
	int y = *yptr++;
	int tmp;

	tmp = y      + v3; *rgbdata++ = tmp < 255 ? (tmp > 0 ? tmp : 0) : 255; // r
	tmp = y - u2 - v2; *rgbdata++ = tmp < 255 ? (tmp > 0 ? tmp : 0) : 255; // g
	tmp = y + u1     ; *rgbdata++ = tmp < 255 ? (tmp > 0 ? tmp : 0) : 255; // b
      }
    }

    if (j % vsub) {
      vptr -= chrcols;
      uptr -= chrcols;
    }
  }
}

void
TextureNodeProxy::Yuv4222Rgb(unsigned char* rgbdata,
				  unsigned char* yptr,
				  unsigned char* uptr,
				  unsigned char* vptr,
				  int lumcols, int lumrows)
{
  if (!initialised) {
    int i;

    for (i = 0; i < 256; ++i) {
      convtab[i].u1 = int(float(i - 128) * 1.37f);
      convtab[i].u2 = int(float(i - 128) * 0.698f);
      convtab[i].v1 = int(float(i - 128) * 0.336f);
      convtab[i].v2 = int(float(i - 128) * 1.73f);
    }
   
    for (i = 0; i < 256; ++i)
      cliptab[i] = 0;
    for (; i < 512; ++i)
      cliptab[i] = i - 256;
    for (; i < 768; ++i)
      cliptab[i] = 255;

    initialised = TRUE;
  }

  // row loop

  for (int j = lumrows >> 1; j > 0; --j) {
    unsigned char* tuptr = uptr;
    unsigned char* tvptr = vptr;
    int i;

    // loop chrominance columns 

    for (i = lumcols >> 1; i > 0; --i) {
      int u = *tuptr++;
      int v = *tvptr++;

      // horizontal downsampling

      for (int k = 2; k > 0; --k) { 
	int y = *yptr++;

	*rgbdata++ = (cliptab + 256)[y                 + convtab[v].v2]; // r
	*rgbdata++ = (cliptab + 256)[y - convtab[u].u2 - convtab[v].v1]; // g
	*rgbdata++ = (cliptab + 256)[y + convtab[u].u1                ]; // b
      }
    }

    for (i = lumcols >> 1; i > 0; --i) {
      // calculate intermediate values that can be recycled

      int u = *uptr++;
      int v = *vptr++;

      // horizontal downsampling   

      for (int k = 2; k > 0; --k) { 
	int y = *yptr++;

	*rgbdata++ = (cliptab + 256)[y                 + convtab[v].v2]; // r
	*rgbdata++ = (cliptab + 256)[y - convtab[u].u2 - convtab[v].v1]; // g
	*rgbdata++ = (cliptab + 256)[y + convtab[u].u1                ]; // b
      }
    }
  }
}

// Enable old texture

BOOL
TextureNodeProxy::Enable(OpenGLVisualRenderer* pRenderer)
{
  // bind texture object

#if !defined(GL_VERSION_1_1)
  if (m_nTextureObject == 0)
    ::glGenTexturesEXT(1, &m_nTextureObject);
    
  ::glBindTextureEXT(GL_TEXTURE_2D, m_nTextureObject);
#else
  if (m_nTextureObject == 0)
    ::glGenTextures(1, &m_nTextureObject);
    
  ::glBindTexture(GL_TEXTURE_2D, m_nTextureObject);
#endif


/*--------------------------------------------------------------------------
set render texture yes/no decision for multipass shadowrendering
- added by H.Drumm 11-2005 
--------------------------------------------------------------------------*/

unsigned int state = pRenderer->GetTextureEnabled();
if (state != 0)	
{

  pRenderer->SetCurrentTexture(this);

  ::glEnable(GL_TEXTURE_2D);

  // We set these parameters every frame so the renderer can change min and
  // mag functions on the fly.

  ::glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, pRenderer->GetTextureMinFunc());
  ::glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, pRenderer->GetTextureMagFunc());

  // Scale texture coordinates if texture size is not 2^n

  if (m_nTextureHeight != m_nHeight ||
      m_nTextureWidth != m_nWidth) {
    ::glMatrixMode(GL_TEXTURE);
    ::glPushMatrix();
    ::glTranslatef(1.0f - float(m_nWidth) / float(m_nTextureWidth), 0.0f, 0.0f);
    ::glScalef(float(m_nWidth) / float(m_nTextureWidth),
	       float(m_nHeight) / float(m_nTextureHeight), 1.0f);
    ::glMatrixMode(GL_MODELVIEW);
  }

#ifdef _DEBUG
  if (pRenderer->CheckGLError("TextureNodeProxy:")) {
    TRACE("Texture width: %d, %d\n", m_nTextureWidth, m_nTextureHeight);
    TRACE(" width: %d, %d\n", m_nWidth, m_nHeight);
    TRACE("TextureNodeProxy::Enable() [%08x]: ", this);
      
    return FALSE;
  }
#endif

  return TRUE;
} // -------------------------------------------------------------------------

}

// Enable new texture

BOOL
TextureNodeProxy::Enable(OpenGLVisualRenderer* pRenderer, TextureType eType,
			 int nWidth, int nHeight, unsigned char* pTextureData)
{
  // StreamConsumer * pTexture = (StreamConsumer *) GetMediaObject();

  // bind texture object

#if !defined(GL_VERSION_1_1)
  if (m_nTextureObject == 0)
    ::glGenTexturesEXT(1, &m_nTextureObject);
    
  ::glBindTextureEXT(GL_TEXTURE_2D, m_nTextureObject);
#else
  if (m_nTextureObject == 0)
    ::glGenTextures(1, &m_nTextureObject);
    
  ::glBindTexture(GL_TEXTURE_2D, m_nTextureObject);
#endif

  ::glEnable(GL_TEXTURE_2D);

  pRenderer->SetCurrentTexture(this);

  // mar 20061121: Flip the texture (OpenGL reads the texture from top to bottom)
  int nPitch = nWidth * (eType == RGBA ? 4 : 3);
  unsigned char *pLine = new unsigned char [nPitch];

  int nHalfHeight = nHeight >> 1;
  for (int i = 0; i < nHalfHeight; i++)
  {
    ::memcpy(pLine, pTextureData + nPitch * i, nPitch);
    ::memcpy(pTextureData + nPitch * i, pTextureData + nPitch * (nHeight - i - 1), nPitch);
    ::memcpy(pTextureData + nPitch * (nHeight - i - 1), pLine, nPitch);
  }

  delete[] pLine;

  // We set these parameters every frame so the renderer can change min and
  // mag functions on the fly.

  m_nWidth = nWidth;
  m_nHeight = nHeight;
  m_bRescale = FALSE;

  // textures must have height and width == 2^n
  
  // align width
  
  {
    int numshifts = -1;
    
    for (; nWidth; nWidth >>= 1)
      ++numshifts;
    
    nWidth = 1 << numshifts;
    if (nWidth < m_nWidth)
      nWidth <<= 1;
  }	  
  
  // align height
  
  {
    int numshifts = -1;
    
    for (; nHeight; nHeight >>= 1)
      ++numshifts;
    
    nHeight = 1 << numshifts;
    if (nHeight < m_nHeight)
      nHeight <<= 1;
  }	    

  // has the texture environment been set up properly?
  
  if (nWidth != m_nTextureWidth ||
      nHeight != m_nTextureHeight) {
    ::glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    
    ::glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,
		      /* pTexture-> */ GetRepeatT() == TRUE ? GL_REPEAT : GL_CLAMP);
    ::glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,
		      /* pTexture-> */ GetRepeatS() == TRUE ? GL_REPEAT : GL_CLAMP);
    
    // Check whether we have space for the texture

    for (;;) {
      GLint width;
    
      if (eType == RGBA)
#if !defined(GL_VERSION_1_1)
	::glTexImage2D(GL_PROXY_TEXTURE_2D_EXT, 0,
		       4, nWidth, nHeight,
		       0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
#else
	::glTexImage2D(GL_PROXY_TEXTURE_2D, 0,
		       4, nWidth, nHeight,
		       0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
#endif
      else
#if !defined(GL_VERSION_1_1)
	::glTexImage2D(GL_PROXY_TEXTURE_2D_EXT, 0,
		       3, nWidth, nHeight,
		       0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
#else
	::glTexImage2D(GL_PROXY_TEXTURE_2D, 0,
		       3, nWidth, nHeight,
		       0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
#endif

#if !defined(GL_VERSION_1_1)
      ::glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D_EXT, 0,
				 GL_TEXTURE_WIDTH, &width);
#else
      ::glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0,
				 GL_TEXTURE_WIDTH, &width);
#endif
    
      if (width)
        break;
    
      // We don't

      nWidth >>= 1;
      nHeight >>= 1;
    
      m_bRescale = TRUE;
    }

    if (m_bRescale) {
      unsigned char* pImage = NULL;
      
      // Okay, rescale texture
      
      if (eType == RGBA) {
	pImage = (unsigned char*) ::malloc(4 * nWidth * nHeight);
	
	if (pImage)
	  ::gluScaleImage(GL_RGBA,
			  m_nWidth, m_nHeight,
			  GL_UNSIGNED_BYTE, pTextureData,
			  nWidth, nHeight,
			  GL_UNSIGNED_BYTE, pImage);
      } else {
	pImage = (unsigned char*) ::malloc(3 * nWidth * nHeight);
	
	if (pImage)
	  ::gluScaleImage(GL_RGB,
			  m_nWidth, m_nHeight,
			  GL_UNSIGNED_BYTE, pTextureData,
			  nWidth, nHeight,
			  GL_UNSIGNED_BYTE, pImage);
      }
      
      pTextureData = pImage;
      m_nWidth = nWidth;
      m_nHeight = nHeight;
    }

    // did texture have height and width == 2^n?
    
    if (nHeight != m_nHeight ||
	nWidth != m_nWidth)
      if (eType == RGBA) {
#if defined(_IRIX) || defined(MESA)
	// SGI and Mesa conveniently allows us to send a NULL pointer,
	// which will set up the texture, but won't transmit any texture
	// data

	::glTexImage2D(GL_TEXTURE_2D, 0,
		       4, nWidth, nHeight,
		       0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
#else
	{
	  // this is terribly stupid, but currently has to be this way to
	  // set up texturing correctly
	  
	  unsigned char* buffer = new unsigned char[nWidth * nHeight * 4];
	  
	  ::glTexImage2D(GL_TEXTURE_2D, 0,
			 4, nWidth, nHeight,
			 0, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
	  delete buffer;
	}
#endif
      } else {
#if defined(_IRIX) || defined(MESA)
	// SGI and Mesa conveniently allows us to send a NULL pointer,
	// which will set up the texture, but won't transmit any texture
	// data

	::glTexImage2D(GL_TEXTURE_2D, 0,
		       3, nWidth, nHeight,
		       0, GL_RGB, GL_UNSIGNED_BYTE, NULL);
#else
	{
	  // this is terribly stupid, but currently has to be this way to
	  // set up texturing correctly
	  
	  unsigned char* buffer = new unsigned char[nWidth * nHeight * 3];
	  
	  ::glTexImage2D(GL_TEXTURE_2D, 0,
			 3, nWidth, nHeight,
			 0, GL_RGB, GL_UNSIGNED_BYTE, buffer);
	  delete buffer;
	}
#endif
      }

    m_nTextureWidth = nWidth;
    m_nTextureHeight = nHeight;
  }
  
  // set texture
  
  if (eType == RGBA)
    if (m_nTextureHeight != m_nHeight ||
	m_nTextureWidth != m_nWidth) {
#if !defined(GL_VERSION_1_1)
      ::glTexSubImage2DEXT(GL_TEXTURE_2D, 0,
			   m_nTextureWidth - m_nWidth, 0, m_nWidth, m_nHeight,
			   GL_RGBA, GL_UNSIGNED_BYTE, pTextureData);
#else
      ::glTexSubImage2D(GL_TEXTURE_2D, 0,
			m_nTextureWidth - m_nWidth, 0, m_nWidth, m_nHeight,
			GL_RGBA, GL_UNSIGNED_BYTE, pTextureData);
#endif
    } else
      ::glTexImage2D(GL_TEXTURE_2D, 0,
		     4, m_nTextureWidth, m_nTextureHeight,
		     0, GL_RGBA, GL_UNSIGNED_BYTE, pTextureData);
  else
    if (m_nTextureHeight != m_nHeight ||
	m_nTextureWidth != m_nWidth) {
#if !defined(GL_VERSION_1_1)
      ::glTexSubImage2DEXT(GL_TEXTURE_2D, 0,
			   m_nTextureWidth - m_nWidth, 0, m_nWidth, m_nHeight,
			   GL_RGB, GL_UNSIGNED_BYTE, pTextureData);
#else
      ::glTexSubImage2D(GL_TEXTURE_2D, 0,
			m_nTextureWidth - m_nWidth, 0, m_nWidth, m_nHeight,
			GL_RGB, GL_UNSIGNED_BYTE, pTextureData);
#endif
    } else
      ::glTexImage2D(GL_TEXTURE_2D, 0,
		     3, m_nTextureWidth, m_nTextureHeight,
		     0, GL_RGB, GL_UNSIGNED_BYTE, pTextureData);

  if (m_bRescale && pTextureData)
    ::free(pTextureData);

  // We set these parameters every frame so the renderer can change min and
  // mag functions on the fly.

  ::glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, pRenderer->GetTextureMinFunc());
  ::glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, pRenderer->GetTextureMagFunc());

  // Scale texture coordinates if texture size is not 2^n

  if (m_nTextureHeight != m_nHeight ||
      m_nTextureWidth != m_nWidth) {
    ::glMatrixMode(GL_TEXTURE);
    ::glPushMatrix();
    ::glTranslatef(1.0f - float(m_nWidth) / float(m_nTextureWidth), 0.0f, 0.0f);
    ::glScalef(float(m_nWidth) / float(m_nTextureWidth),
	       float(m_nHeight) / float(m_nTextureHeight), 1.0f);
    ::glMatrixMode(GL_MODELVIEW);
  }

#ifdef _DEBUG
  if (pRenderer->CheckGLError("TextureNodeProxy:")) {
    TRACE("Texture width: %d, %d\n", m_nTextureWidth, m_nTextureHeight);
    TRACE(" width: %d, %d\n", m_nWidth, m_nHeight);
    TRACE("TextureNodeProxy::Enable() [%08x]: ", this);
      
    return FALSE;
  }
#endif

  return TRUE;
}

